/*
 * Copyright (C) 2004 State of California, Department of
 * Water Resources.
 *
 * See the file FilterGUI.java in the package dwr.des.digitalfilter
 * for licensing information.
 */

package gov.ca.dwr.des.digitalfilter;

/**
 * This class utilizes FFT.java to transform data in the time domain
 * into the frequency domain, remove specified frequencies, and then
 * transform the data back to the time domain.
 *
 * @author Aaron Miller <br>
 * last modified by $Author: miller $
 * @version $Revision: 1.2 $ <br>
 * $Date: 2004/09/03 21:20:06 $
 */

public class DigitalFilter {

	double [] data, psd, pbar, xmask, time, period;

	/**
	 * Constructs and executes the DigitalFilter based on parameters passed in.
	 * This would probably be better suited as a static method.
	 * @param params parameters for this DigitalFilter.
	 */
	public DigitalFilter(DigitalFilterParams params){
		this.data = params.getFilterIO().getData(false);
		this.time = params.getFilterIO().getTime(false);
		psd = new double[data.length];
		pbar = new double[data.length];
		xmask = new double[data.length];
		period = new double[data.length];


		// zero tapper of raw series
		double ah;
		int ntap = params.getTaper();
		int maxd = params.getAdjNumData();
		int ntotal = params.getTotalNum();

		for(int i = 0; i<ntap; i++){
			ah=0.5-0.5*Math.cos(Math.PI*(double)(i)/(double)(ntap));
			data[i]=ah*data[i];
			data[maxd-i-1] = ah*data[maxd-i-1];
		}

		// zero padding to series
		int nout = ntotal;
		int nt = 2;
		while(nt < ntotal){
			nt=nt*2;
		}
		if(nt != ntotal){
			for(int m = ntotal+1; m<= nt; m++){
				data[m] = 0;
			}
		}
		ntotal = nt;
		params.setAdjustedTotalNum(ntotal);
		// set max and min frequencies
		int nn=2*ntotal;
		int n2=ntotal/2;

		// forward fft
		FFT.realft(data,n2,+1);

		// save power spectral density
		for(int i = 0;i < n2+1;i++){
			psd[i]=data[2*i] * data[2*i];
		}

		// initialized mask
		for(int i = 0; i<xmask.length; i++){
			xmask[i] = 1.0;
		}

		// apply frequency filters
		// freqPt[] = {nlps, nlstp, nhps, nhstp, dnl, dnh, nttl, ntth}
		//              0      1     2      3     4    5    6     7
		int freqPt [] = params.getFreqFiltCoef();
		double arg, a;
		if(freqPt[0] != 0 || freqPt[1] != 0){ // low pass portion
			for(int j = freqPt[0]; j<=freqPt[6]; j++){
				arg = (double)(freqPt[1]-j)/(double)freqPt[4];
				if(j < freqPt[1] - 1)
					a = 0.5-(0.5*Math.cos(Math.PI * arg));
				else
					a = 0;
				data[j-1] = a*data[j-1];
				xmask[j-1] = a;
			}
		}

		if(freqPt[2] != 0 || freqPt[3] != 0){ // high pass portion
			for(int j = freqPt[7]+1; j<=freqPt[2]; j++){
				arg = (double)(freqPt[3]-j)/(double)freqPt[5];
				if(j >= freqPt[3])
					a = 0.5-0.5*Math.cos(Math.PI * arg);
				else
					a = 0;
				data[j-1] = a*data[j-1];
				xmask[j-1] = a;
			}
		}


		// inverse FFT
		FFT.realft(data,n2,-1);

		// apply "1/n" factor
		a = 2/(double)ntotal;
		for(int k = 0;k<ntotal;k++){
			data[k] = a*data[k];
		}

		// set zeros for zero padding
		if(nout < ntotal){
			for(int i=nout+1; i<ntotal; i++){
				data[i]=0;
			}
		}

		// power spectral density preparation and output
			// apply moving average
			double dt6 = params.getTimestep();
			int imov = params.getPSDMoveLen();
			double df = params.getFrequencyStep();

			double f, tp;
			for(int j = imov; j<n2-imov; j++){
				pbar[j] = 0;
				for(int k = j-imov; k<=j+imov; k++){
					pbar[j] = pbar[j] + psd[k];
				}
				pbar[j] = pbar[j]/(double)((2*imov+1)*ntotal);
			}

			// set other data arrays for output

		for (int j = imov; j < n2 - imov + 1; j++) {
			f = df * (double) (j);
			tp = 1 / f;
			period[j] = tp;
		}

		params.getFilterIO().setPBARData(pbar);
		params.getFilterIO().setXMaskData(xmask);
		params.getFilterIO().setPeriodData(period);
	}
}
