/*
 * Copyright (C) 2004 State of California, Department of
 * Water Resources.
 *
 * See the file FilterGUI.java in the package dwr.des.digitalfilter
 * for licensing information.
 */

package gov.ca.dwr.des.digitalfilter;

import java.util.*;
import java.io.*;
import java.text.*;
/**
 * This class handles all of the data input and output to the digital filter.
 *
 * @author Aaron Miller <br>
 * last modified by $Author: miller $
 * @version $Revision: 1.2 $ <br>
 * $Date: 2004/09/03 21:20:06 $
 */

public class DigitalFilterIO {

	private double data[];
	private double time[];
	private double period[];
	private double xmask[];
	private double psd[];
	private double pbar[];
	private int maxd;
	private double time_hr;
	// interleave array
	private double iArray[];
	// interleaved number of dimensions
	private int iNumDim;
	// interleaved x dimension
	private int iXDim;
	// interleaved y dimension
	private int iYDim;

	private DigitalFilterParams params;
	private BufferedWriter psdOut;

	/**
	 * Constructs a DigitalFilterIO with a data array, starting and ending time, and a timestep.
	 * @param data array of data.
	 * @param stTime data start time.
	 * @param endTime data end time.
	 * @param timestep data timestep.
	 */
	public DigitalFilterIO(double data[], double stTime, double endTime, double timestep){
		this.data = data;
		time = new double[(int)Math.round((endTime-stTime)/timestep)+1];
		for (int i = 0; i<time.length; i++){
			time[i] = stTime + timestep * (i+1);
		}
		maxd = data.length;
		time_hr = timestep;
	}
	/**
	 * Constructs a DigitalFilterIO with a data and time array.
	 * @param data array of data.
	 * @param time array of corresponding time in hours.
	 * @param timestep timestep of data in hours.
	 */
	public DigitalFilterIO(double data[], double time[], double timestep){
		this.data = data;
		this.time = time;
		maxd = time.length;
		time_hr = timestep;
	}

	/**
	 * Constructs a DigitalFilterIO for GR.
	 *
	 * @param iData    Interleaved data array.
	 * @param numDim   Number of dimesions in the data array.
	 * @param xDim     Location of X data in array.
	 * @param yDim     Location of Y data in array.
	 * @param timestep Data timestep in hours.
	 */
	public DigitalFilterIO(double iData[], int numDim, int xDim, int yDim, double timestep) {
	    iArray = iData;
	    maxd = iData.length/numDim;
	    time = new double[maxd];
	    data = new double[maxd];
	    for(int i = 0; i<maxd; i++){
	        time[i] = iData[i * numDim + xDim];
	        data[i] = iData[i * numDim + yDim];
	    }
	    time_hr = timestep;
	    iNumDim = numDim;
	    iXDim = xDim;
	    iYDim = yDim;
	}

	/**
	 * Constructs a DigitalFilterIO with a simple file tab or space delimited.
	 * The format is:  [time in hours or minutes][value]<br>
	 *
	 * @param filename the path and name of the file containing data.
	 * @param timestep the timestep in hours.
	 * @throws IOException if something happens during the read.
	 */
	public DigitalFilterIO(String filename, double timestep) throws IOException {
		maxd = input(filename);
		time_hr = timestep;
	}

	/**
	 * Checks and adjusts the size of the data array.
	 * @param size the size the array should be.
	 */
	void adjustArraySize(int size) {
		if (size > data.length) {
			double newArray[] = new double[size];
			System.arraycopy(data, 0, newArray, 0, data.length);
			data = newArray;
		}
	}

	/**
	 * Checks and adjusts the size of the data array.
	 * @param size the size of the usable data array.
	 * @param ntotal the calculated 2^n that will ecompass the data.
	 */
	void adjustArraySizeForMirror(int size, int ntotal){
	    double reverseArray[] = new double[size];
	    double newArray[] = new double[ntotal*2];

	    for(int i = 0; i<size; i++){
	        reverseArray[i] = data[size-1-i];
	    }

	    System.arraycopy(reverseArray, 0, newArray, 0, size);
	    System.arraycopy(data, 0, newArray, size, size);
	    System.arraycopy(reverseArray, 0, newArray, size*2, size);
	    data = newArray;
	    return;
	}

	/**
	 * Returns the length of the data array.
	 * @param trim true to return length of usable array,
	 * false to return length of working array.
	 * @return the length of array.
	 */
	public int getDataLength(boolean trim){
	    if(trim){
	        int ntotal = params.getAdjustedTotalNum();
	        int ntap = params.getTaper();
	        int num;
	        if(params.getPadType() == DigitalFilterParams.PAD_ZERO){
	            if(ntotal < maxd){
	                num = ntotal - 2*ntap;
	            }
	            else{
	                num = maxd - 2*ntap;
	            }
	            return num;
	        }
	    }
		return maxd;
	}

	/**
	 * Returns data array.
	 * @param trim true to return copy or adjusted data array so that only usable
	 * data is in array, false to return working array.
	 * @return pointer of copy of data array.
	 */
	public double[] getData(boolean trim) {
		if(trim){
			int ntotal = params.getAdjustedTotalNum();
			int ntap = params.getTaper();
			int num;

			if(params.getPadType() == DigitalFilterParams.PAD_MIRROR){
				for(int i = 0; i<maxd; i++){
					data[i] = data[i+maxd];
				}
				return data;
			}
			else if(params.getPadType() == DigitalFilterParams.PAD_ZERO){
				for(int i = 0; i<(maxd-ntap); i++){
					data[i] = data[i+ntap];
				}
			}

			if(ntotal < maxd){
				num = ntotal - 2*ntap;
			}
			else{
				num = maxd - 2*ntap;
			}

			double newdata[] = new double[num];
			System.arraycopy(data,0,newdata,0,num);
			return newdata;
		}
		return data;
	}

	/**
	 * Returns time array.
	 * @param trim true to return copy or adjusted time array so that only usable
	 * data is in array, false to return working array.
	 * @return pointer of copy of time array.
	 */
	public double[] getTime(boolean trim) {

		if(trim){
			int ntotal = params.getAdjustedTotalNum();
			int ntap = params.getTaper();
			int num;

			if(params.getPadType() == DigitalFilterParams.PAD_MIRROR){
			    return time;
			}
			else if(params.getPadType() == DigitalFilterParams.PAD_ZERO){
				for(int i = 0; i<(maxd-ntap); i++){
					time[i] = time[i+ntap];
				}
			}

			if(ntotal < maxd){
				num = ntotal - 2*ntap;
			}
			else{
				num = maxd - 2*ntap;
			}

			double newtime[] = new double[num];
			System.arraycopy(time,0,newtime,0,num);
			return newtime;
		}
		return time;
	}

	/**
	 * Returns a new array containing interleaved data specific to GR.
	 * @return interleaved data array.
	 */
	public double[] getInterleavedData(){
		int ntap = params.getTaper();
	    int len = getDataLength(true);
        double[] newArray = new double[len * iNumDim];
	    if(params.getPadType() == DigitalFilterParams.PAD_ZERO){
	        for(int i = 0; i<len; i++){
	            newArray[i * iNumDim + iXDim] = time[i+ntap];
	            newArray[i * iNumDim + iYDim] = data[i+ntap];
	        }
	    }
	    else{
	        for(int i = 0; i<len; i++){
	            newArray[i * iNumDim + iXDim] = time[i];
	            newArray[i * iNumDim + iYDim] = data[i + len];
	        }
	    }
	    return newArray;
	}

	/**
	 * Returns the array of PSD values.
	 * @return the PSD array.
	 */
	public double[] getPSD() {
		return pbar;
	}

	/**
	 * Returns the array of period values.
	 * @return the period array.
	 */
	public double[] getPeriod(){
		return period;
	}

	/**
	 * Returns the data array length that was initally calculated.
	 * @return the data array length.
	 */
	public int getMaxd() {
		return maxd;
	}

	/**
	 * Returns the timestep associated with the data array.
	 * @return the timestep in hours.
	 */
	public double getTimestep() {
		return time_hr;
	}

	/**
	 * Extracts data from a file.  Uses a simple file tab or space delimited.
	 * The format is:  [time in hours or minutes][value]
	 * @param filename the path and filename to the file containing the data.
	 * @return the number of data points read.
	 * @throws IOException if an I/O error occurs.
	 */
	private int input(String filename) throws IOException {
		int arrSiz = 70000;
		data = new double[arrSiz];
		time = new double[arrSiz];

		String tmpstr;
		StringTokenizer tokenizer;
		int i = 0;
		BufferedReader in = new BufferedReader(new FileReader(filename));
		while ((tmpstr = in.readLine()) != null) {
			tokenizer = new StringTokenizer(tmpstr);
			time[i] = Double.parseDouble(tokenizer.nextToken());
			data[i] = Double.parseDouble(tokenizer.nextToken());
			i++;
		}
		return i;
	}

	/**
	 * Sets the xmask array.
	 * @param xmask xmask array.
	 */
	void setXMaskData(double xmask[]) {
		this.xmask = xmask;
	}

	/**
	 * Sets the PSD array.
	 * @param psd psd array.
	 */
	void setPSDData(double psd[]) {
		this.psd = psd;
	}

	/**
	 * Sets the PBAR array.
	 * @param pbar pbar array.
	 */
	void setPBARData(double pbar[]) {
		this.pbar = pbar;
	}

	/**
	 * Sets the period data array.
	 * @param period period array.
	 */
	public void setPeriodData(double period[]){
		this.period = period;
	}

	/**
	 * Writes out the psd array to the specified file.
	 * @param filename filename to write data to.
	 * @throws IOException if an I/O error occurs.
	 */
	public void writePSDOut(String filename) throws IOException {
		double dt6 = time_hr;
		int imov = params.getPSDMoveLen();
		double df = params.getFrequencyStep();
		int n2 = params.getAdjustedTotalNum()/2;
		psdOut = new BufferedWriter(new FileWriter(filename));
		double f, tp;
		DecimalFormat fmt = new DecimalFormat("#####0.###E0");
		for (int j = imov; j < n2 - imov + 1; j++) {
			f = df * (double) (j);
			tp = 1.0 / f;
			psdOut.write(fmt.format(f) + "\t" + fmt.format(tp) +
				"\t" + fmt.format(pbar[j]) + "\t" + fmt.format(xmask[2 * j - 1]) + "\n");
		}
		psdOut.close();
	}

	/**
	 * Writes out the data array to the specified file.
	 * @param filename filename to write data to.
	 * @throws IOException if an I/O error occurs.
	 */
	public void writeDataOut(String filename) throws IOException {
		BufferedWriter out = new BufferedWriter(new FileWriter(filename));
		int ntotal = params.getAdjustedTotalNum();
		int ntap = params.getTaper();

		double time_tmp[] = getTime(true);
		double data_tmp[] = getData(true);

		for (int j = 0; j < getDataLength(true); j++){
			out.write(time_tmp[j]+"\t"+data_tmp[j]+"\n");
		}
		out.close();
	}

	/**
	 * Registers DigitalFilterParams object with this object.
	 * @param params DigitalFilterParams object.
	 */
	public void setParams(DigitalFilterParams params) {
		this.params = params;
	}
}
