/*
 * Copyright (C) 2004 State of California, Department of
 * Water Resources.
 *
 * See the file FilterGUI.java in the package dwr.des.digitalfilter
 * for licensing information.
 */

package gov.ca.dwr.des.digitalfilter;

/**
 * This class defines the parameters of the digital filter.
 *
 * @author Aaron Miller <br>
 * last modified by $Author: miller $
 * @version $Revision: 1.2 $ <br>
 * $Date: 2004/09/03 21:20:06 $
 */

public class DigitalFilterParams {

	// moving average length
	int moveLen;

	double time_hr;

	//taper parameters
	int ntaper, taperType;

	//padding parameters
	int padType;

	//band parameters
	int nlps, nlstp, nhps, nhstp; // point positions

	int dnl, dnh; // transition length

	int nttl, ntth;

	//filter type
	String filterType;

	int ntotal, maxd, adjNTotal, adjMaxd;

	double df;

	DigitalFilterIO filterIO;

/**
 * Constructs an object containing parameters specific to the digital filter.
 * Padding types appropriate to this method are:<br>
 * dwr.des.digitalFilter.DigitalFilterParams.PAD_ZERO<br>
 * dwr.des.digitalFilter.DigitalFilterParams.PAD_MIRROR
 * @param filterIO the location of the data to be filtered.
 * @param padTypeIn the type of padding to apply to data.
 */
	public DigitalFilterParams(DigitalFilterIO filterIO, int padTypeIn) {
		filterIO.setParams(this);
		this.filterIO = filterIO;
		this.maxd = filterIO.getMaxd();
		this.time_hr = filterIO.getTimestep();
		this.padType = padTypeIn;

		// convert timestep into hourly value
		int jb = 1;

		adjMaxd = maxd;
		if(padType == PAD_MIRROR){
		    adjMaxd = maxd * 3;
		}

		do {
			ntotal = (int) Math.round(Math.pow(2.0, (double) jb));
			jb++;
		} while (ntotal < adjMaxd);

		if(padType == PAD_MIRROR){
		    filterIO.adjustArraySizeForMirror(maxd,ntotal);
		}
		// check to make sure array is big enough, if not then adjust it
		else if (filterIO.getData(false).length < ntotal * 2) {
			filterIO.adjustArraySize(ntotal * 2);
		}

		df = 1.0 / ((double) ntotal * time_hr);

	}


	/**
	 * Sets the filter to remove high frequencies.
	 * <pre>
	 *      |    stop   |transition|   pass
	 *      |    band   |   band   |   band
	 *      |___________|__________|__________
	 *                 low        low
	 *                 stop       pass
	 *                 (pt1)      (pt2)
	 *      ---- decreasing frequency ---->
	 * </pre>
	 * @param lowStop The frequency at which to start the transition from the stop band (pt1)
	 * @param lowPass The frequency at which to start the pass band (pt2)
	 */
	public void setLowPass(double lowStop, double lowPass) {
		nlps = 2 * (int) ((1. / lowPass) / df) + 1;
		nlstp = 2 * (int) ((1. / lowStop) / df) + 1;
		nhps = 0;
		nhstp = 0;
		dnl = nlstp - nlps;
		dnh = 0;
		nttl = ntotal;
		ntth = 0;
		filterType = "Type: LOW PASS FILTER:\n"
				+ "----------------------\n"
				+ "Stop: time = " + lowStop + " hr, freq. = " + 1. / lowStop
				+ " cph, n = " + nlstp + "\n" + "Pass: time = " + lowPass + " hr, freq. = " + 1. / lowPass
				+ " cph, n = " + nlps;
	}

	/**
	 * Sets the filter to remove low frequencies.
	 * <pre>
	 *      |    pass   |transition|   stop
	 *      |    band   |   band   |   band
	 *      |___________|__________|__________
	 *                 high       high
	 *                 pass       stop
	 *                 (pt1)      (pt2)
	 *      ---- decreasing frequency ---->
	 * </pre>
	 * @param highPass The frequency at which to start the transition to the stop band (pt1)
	 * @param highStop The frequency at which to start the stop band (pt2)
	 */
	public void setHighPass(double highPass, double highStop) {
		nlps = 0;
		nlstp = 0;
		nhps = 2 * (int) ((1. / highPass) / df) + 1;
		nhstp = 2 * (int) ((1. / highStop) / df) + 1;
		dnl = 0;
		dnh = nhps - nhstp;
		nttl = 0;
		ntth = 0;
		filterType = "Type: HIGH PASS FILTER:\n"
				+ "-----------------------\n"
				+ "Pass: time = " + highPass + " hr, freq. = " + 1. / highPass
				+ " cph, n = " + nhps + "\n" + "Stop: time = " + highStop + " hr, freq. = " + 1. / highStop
				+ " cph, n = " + nhstp;
	}

	/**
	 * Sets the filter to isolate a specifide range of frequencies.
	 * <pre>
	 *      |    stop   |transition|   pass   |transition|   stop
	 *      |    band   |   band   |   band   |   band   |   band
	 *      |___________|__________|__________|__________|__________
	 *                 low        low        high       high
	 *                 stop       pass       pass       stop
	 *                 (pt1)      (pt2)      (pt3)      (pt4)
	 *      ---- decreasing frequency ---->
	 * </pre>
	 * @param lowStop The frequency at which to start the transition from the stop band (pt1)
	 * @param lowPass The frequency at which to start the pass band (pt2)
	 * @param highPass The frequency at which to start the transition to the stop band (pt3)
	 * @param highStop The frequency at which to start the stop band (pt4)
	 */
	public void setBandPass(double lowStop, double lowPass, double highPass, double highStop) {
		nlps = 2 * (int) ((1. / lowPass) / df) + 1;
		nlstp = 2 * (int) ((1. / lowStop) / df) + 1;
		nhps = 2 * (int) ((1. / highPass) / df) + 1;
		nhstp = 2 * (int) ((1. / highStop) / df) + 1;
		dnl = nlstp - nlps;
		dnh = nhps - nhstp;
		//		nttl = ntotal-1;
		nttl = ntotal;
		ntth = 0;
		filterType = "Type: BAND PASS FILTER:\n"
				+ "-----------------------\n"
				+ "Low stop: time = " + lowStop + " hr, freq. = " + 1. / lowStop
				+ " cph, n = " + nlstp + "\n" + "Low pass: time = " + lowPass + " hr, freq. = " + 1. / lowPass
				+ " cph, n = " + nlps + "\n" + "High pass: time = " + highPass + " hr, freq. = " + 1. / highPass
				+ " cph, n = " + nhps + "\n" + "High stop: time = " + highStop + " hr, freq. = " + 1. / highStop
				+ " cph, n = " + nhstp;
	}

	/**
	 * Sets the filter to remove a specified range of frequencies.
	 * <pre>
	 *      |    pass   |transition|   stop   |transition|   pass
	 *      |    band   |   band   |   band   |   band   |   band
	 *      |___________|__________|__________|__________|__________
	 *                 high       high       low        low
	 *                 pass       stop       stop       pass
	 *                 (pt1)      (pt2)      (pt3)      (pt4)
	 *      ---- decreasing frequency ---->
	 * </pre>
	 * @param highPass The frequency at which to start the transition to the stop band (pt1)
	 * @param highStop The frequency at which to start the stop band (pt2)
	 * @param lowStop The frequency at which to start the transition to the pass band (pt3)
	 * @param lowPass The frequency at which to start pass band (pt4)
	 */
	public void setBandStop(double highPass, double highStop, double lowStop, double lowPass) {
		nlps = 2 * (int) ((1. / lowPass) / df) + 1;
		nlstp = 2 * (int) ((1. / lowStop) / df) + 1;
		nhps = 2 * (int) ((1. / highPass) / df) + 1;
		nhstp = 2 * (int) ((1. / highStop) / df) + 1;
		dnl = nlstp - nlps;
		dnh = nhps - nhstp;
		//		nttl = nlstp;
		nttl = nhstp;
		ntth = nhstp;
		filterType = "Type: BAND STOP FILTER:\n"
				+ "-----------------------\n"
				+ "High pass: time = " + highPass + " hr, freq. = " + 1. / highPass
				+ " cph, n = " + nhps + "\n" + "High stop: time = " + highStop + " hr, freq. = " + 1. / highStop
				+ " cph, n = " + nhstp + "\n" + "Low stop: time = " + lowStop + " hr, freq. = " + 1. / lowStop
				+ " cph, n = " + nlstp + "\n" + "Low pass: time = " + lowPass + " hr, freq. = " + 1. / lowPass
				+ " cph, n = " + nlps;
	}

	/**
	 * Sets the type and number of points for tapering data.
	 * Acceptable taper types are:<br>
	 * dwr.des.DigitalFilterParams.TAPER_NONE<br>
	 * dwr.des.DigitalFilterParams.TAPER_COS
	 *
	 * @param taper taper type
	 * @param npoints number of taper points
	 */
	public void setTaper(int taper, int npoints) {
		taperType = taper;
		if(taper == NUM_POINT_AUTO || npoints < 0){
		    ntaper = (int)Math.round(maxd * 0.03);
		}else{
			ntaper = npoints;
		}
	}

	/**
	 * Sets the type of data padding.
	 * Acceptable padding types are:<br>
	 * dwr.des.DigitalFilterParams.PAD_ZERO<br>
	 * dwr.des.DigitalFilterParams.PAD_MIRROR
	 *
	 * @param padding type of data padding
	 */
	public void setPadding(int padding) {
		padType = padding;
	}

	/**
	 * Sets half the number of points to use in the moving average applied
	 * to the PSD.  The moving avereage applied to the PSD is a centered
	 * moving average and the number passed into the method is the length of
	 * a single leg.  If 2 is passed into the method the moving average will
	 * an averaging length of 5.<br>
	 * <pre>
	 * (val[-2] + val[-1] + val[0] + val[1] + val[2]) / (2*leg+1)
	 * </pre>
	 * @param leg length of a single leg
	 */
	public void setPSDMovingAveLen(int leg) {
		moveLen = leg;
	}

	/**
	 * Sets filter defaults to typical tidal filter.  setLowPass(30, 40) -
	 * setTaper(TAPER_COS, 100) - setPadding(PAD_ZERO) - setPSDMovingAveLen(2)
	 */
	public void setDefaults() {
		setLowPass(30, 40);
		//		setHighPass(30, 40);
		//		setBandPass(9,11,14,18);
        //		setBandStop(9, 11, 14, 18);
		//		setTaper(DigitalFilterParams.TAPER_COS,DigitalFilterParams.NUM_POINT_AUTO);
		setTaper(DigitalFilterParams.TAPER_COS, 100);
		setPadding(DigitalFilterParams.PAD_ZERO);
		setPSDMovingAveLen(2);
	}

	/**
	 * Returns the integer associated with the taper flag set.
	 * @return taper for this parameter set
	 */
	public int getTaper() {
		return ntaper;
	}

	/**
	 * Returns the integer associated with the padding type set.
	 * @return padding type for this parameter set
	 */
	public int getPadType(){
	    return padType;
	}

	/**
	 * Returns the time step in hours.
	 * @return the time step for this paramenter set
	 */
	double getTimestep() {
		return time_hr;
	}

	/**
	 * Returns the Power Spectrum Density moving average length.
	 * @return PSD moving average leg length
	 */
	public int getPSDMoveLen() {
		return moveLen;
	}

	/**
	 * Returns the <strong>total</strong> number of data points.
	 * @return total number of data points
	 */
	int getTotalNum() {
		return ntotal;
	}

	/**
	 * Returns adjusted number of <strong>usable</strong> data points.
	 * @return adjusted number of usable data points
	 */
	int getAdjNumData() {
		return adjMaxd;
	}

	/**
	 * Returns the fundamental frequency of the data set.
	 * @return fundamental frequency
	 */
	public double getFrequencyStep() {
		return df;
	}

	/**
	 * Returns an array of array indexing values specific to use in DigitalFilter.java.
	 * @return array of indexing parameters
	 */
	int[] getFreqFiltCoef() {
		return new int[] { nlps, nlstp, nhps, nhstp, dnl, dnh, nttl, ntth };
	}

	/**
	 * Returns the DigitalFilterIO object.
	 * @return the DigitalFilterIO object
	 */
	public DigitalFilterIO getFilterIO() {
		return filterIO;
	}

	/**
	 * Sets the adjusted number of <strong>total</strong> data points.
	 * @param adjTotal adjusted total number of data points
	 */
	void setAdjustedTotalNum(int adjTotal) {
		adjNTotal = adjTotal;
	}

	/**
	 * Returns the adjusted number of <strong>total</strong> data points.
	 * @return the adjusted total number data points
	 */
	public int getAdjustedTotalNum() {
		return adjNTotal;
	}

	/**
	 * Returns the string value of this DigitalFilterParams object.
	 * @return a string representation of this object
	 */
	public String toString() {
		StringBuffer stringBuf = new StringBuffer("DATA Characteristics:\n"
				+ "--------------------\n"
				+ "Sampling interval = " + time_hr + " hours\n"
				+ "Record Length = " + (time_hr * maxd / 24.)+ " days\n"
				+ "Number of data points = " + maxd + "\n"
				+ "Record length analyzed = " + ntotal + "\n"
				+ "Padded points = " + (ntotal - maxd) + "\n"
				+ "Nyquist frequency = " + (1. / (2. * time_hr))
				+ " cph, period = " + (2. * time_hr) + " hr\n"
				+ "Fundamental frequency = " + df + " cph, period = "
				+ (1. / df / 24.) + " day\n\n"
				+ "FILTER Characteristics:\n"
				+ "-----------------------\n");
		if (taperType == DigitalFilterParams.TAPER_COS)
			stringBuf.append("Data taper type: Cosign\n");
		else
			stringBuf.append("Data taper type: None\n");
		stringBuf.append("Points tapered:" + getTaper() + "\n");
		if (padType == DigitalFilterParams.PAD_ZERO)
			stringBuf.append("Data padding: Zero\n\n");
		else
			stringBuf.append("Data padding: Mirror\n\n");
		stringBuf.append(filterType);
		return stringBuf.toString();
	}

	/**
	 * Automatically select number of taper points.
	 */
	public static int NUM_POINT_AUTO = -1;
	/**
	 * Taper flag for no taper.
	 */
	public static int TAPER_NONE = 1;
	/**
	 * Taper flag for a cosign taper.
	 */
	public static int TAPER_COS = 2;
	/**
	 * Padding flag to use zeros.
	 */
	public static int PAD_ZERO = 3;
	/**
	 * Padding flag to use mirroring.
	 */
	public static int PAD_MIRROR = 4;
	/**
	 * Low pass filter has been set.
	 */
	static int LOW_PASS_FILTER = 10;
	/**
	 * High pass filter has been set.
	 */
	static int HIGH_PASS_FILTER = 11;
	/**
	 * Band pass filter has been set.
	 */
	static int BAND_PASS_FILTER = 12;
	/**
	 * Band stop filter has been set.
	 */
	static int BAND_STOP_FILTER = 13;

}
