/*
 * Copyright (C) 2004 State of California, Department of
 * Water Resources.
 *
 * See the file FilterGUI.java in the package dwr.des.digitalfilter
 * for licensing information.
 */

package gov.ca.dwr.des.digitalfilter;

/**
 * This class provides Fast Fourier Transform methods.  The algorithms
 * for these methods are very similar to those found in "Numerical Recipes,
 * The Art of Scientific Computing" by William H. Press et al.  Changes were
 * made to these methods to allow for array indexing to start with 0.
 *
 * @author Aaron Miller <br>
 * last modified by $Author: miller $
 * @version $Revision: 1.2 $ <br>
 * $Date: 2004/09/03 21:20:06 $
 */

public class FFT {

/**
 * Calculates discrete Fourier Transform, or the inverse.
 * Replaces data[] by its discrete Fourier transform if isign is 1, or
 * replaces data[] by nn times it's inverse discrete Fourier transform
 * if isign is -1.
 *
 * @param data  complex array of len nn, or a real array of len 2*nn
 * @param nn    length of data - must be an integer power of 2
 * @param isign 1 for FFT, or -1 for inverse FFT
 */
    public static void four1(double data[],int nn,int isign){

		double tempR, tempI;
		int n = 2 * nn;
		int m;
		int j = 0;
		int mmax = 2;

		for(int i = 0; i<n-1; i+=2){
			if(j > i){
				tempR=data[j];
				tempI=data[j+1];
				data[j]=data[i];
				data[j+1]=data[i+1];
				data[i]=tempR;
				data[i+1]=tempI;
				int junk = 0;
			}
			m = n/2;
			while((m >= 1) && (j >= m)){
				j = j - m;
				m = m / 2;
			}
			j = j + m;
		}
		long count = 0;
		while (n > mmax){
			int istep = 2 * mmax;
			double theta = 6.28318530717959/(double)(isign*mmax);
			double wtemp = Math.sin(0.5*theta);
			double wpr = -2.0 * wtemp * wtemp;
			double wpi = Math.sin(theta);
			double wr = 1.0;
			double wi = 0.0;
			for(m = 0; m < mmax-1; m+=2){
				for(int i = m; i < n-1; i+=istep){
					j = i + mmax;
					tempR = wr * data[j] - wi * data[j+1];
					tempI = wr * data[j+1] + wi * data[j];
					data[j] = data[i] - tempR;
					data[j+1] = data[i+1] - tempI;
					data[i] = data[i] + tempR;
					data[i+1] = data[i+1] + tempI;
					count++;
				}
				wtemp = wr;
				wr = wr * wpr - wi * wpi + wr;
				wi = wi * wpr + wtemp * wpi + wi;
			}
			mmax = istep;
		}
		int junk = 0;
	}

    /**
     * Calculates the Fourier Transform of a set of 2n real-valued data points.
     * Replaces this data[] by the positive frequency half of it's complex Fourier
     * Transform.  The real-valued first and last components of the complex
     * transform are returned as elements data[0] and data[1] respectively.  n must
     * be a power of 2.  This routine also calculates the inverse transform of a
     * complex data array if it is the transform of real data.  (Result in this case
     * must be multiplied by 1/n).
     *
     * @param data  array of real-valued data points
     * @param n     length of data - must be an integer power of 2
     * @param isign 1 for FFT, or -1 for inverse FFT
     */
	public static void realft(double data[], int n, int isign){

		double wtemp;
		int i1,i2,i3,i4;
		double h1r,h1i,h2r,h2i,c2;

		double theta = 6.28318530717959/2.0/(double)n;
		double c1 = 0.5;

		if(isign == 1){
			c2 = -0.5;
			four1(data,n,+1);
		}
		else{
			c2 = 0.5;
			theta = -theta;
		}

		double wpr = -2.0 * Math.pow(Math.sin(0.5*theta), 2.0);
		double wpi = Math.sin(theta);
		double wr = 1.0 + wpr;
		double wi = wpi;
		int n2p3 = 2 * n + 1;
		for(int i = 1; i < n/2; i++){
			i1 = 2 * i;
			i2 = i1 + 1;
			i3 = n2p3 - i2;
			i4 = i3 + 1;

			h1r = c1 * (data[i1] + data[i3]);
			h1i = c1 * (data[i2] - data[i4]);
			h2r = -c2 * (data[i2] + data[i4]);
			h2i = c2 * (data[i1] - data[i3]);

			data[i1] = h1r + wr * h2r - wi * h2i;
			data[i2] = h1i + wr * h2i + wi * h2r;
			data[i3] = h1r - wr * h2r + wi * h2i;
			data[i4] = -h1i + wr * h2i + wi * h2r;

			wtemp = wr;
			wr = wr * wpr - wi * wpi + wr;
			wi = wi * wpr + wtemp * wpi + wi;
		}
		if(isign == 1){
			h1r = data[0];
			data[0] = h1r + data[1];
			data[1] = h1r - data[1];
		}
		else{
			h1r = data[0];
			data[0] = c1 * (h1r + data[1]);
			data[1] = c1 * (h1r - data[1]);
			four1(data,n,-1);
		}
	}
}
