/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-07-19
 * Date started:    2007-05-29
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.crrl.data;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.DimensionOrganizer;
import gov.usgs.gr.tools.TimeMarker;
import gov.usgs.gr.visual.Graph;
import gov.usgs.gr.visual.InterpolatingColorScheme;
import gov.usgs.gr.visual.PageOfGraphs;
import gov.usgs.gr.visual.TextMarkup;
import gov.usgs.gr.visual.Zoom;
import gov.usgs.gr.visual.gl.GlColorBar;
import gov.usgs.gr.visual.gl.GlParticleAnimation;
import java.awt.Component;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.*;
import java.io.*;
import java.nio.IntBuffer;
import java.util.*;
import javax.swing.*;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;

//==============================================================================
/**
 * A data file object for reading data in CRRL Particle format.
 */
public class ParticleDataFile extends gov.usgs.gr.data.TimeDataFile {

    /** An instance of the controller associated with this object. */
    GrController m_grCon = null;
    /** An instance of the page associated with this object. */
    PageOfGraphs m_page = null;
    /** Whether to add the objects to the current page after reading the file.*/
    protected boolean m_bModifyPage = true;
    /** The time step of the animation. */
    protected double m_animationTS = 1d / 24d / 60d / 4d;
    /**
     * The longest gap allowed between detections before considering
     * a particle to have left the system.
     */
    protected double m_gapLimit = 1d / 16d;
    /** A list of times when a particle entered the system. */
    protected List<Double> m_entryTimes = new ArrayList<Double>();
    /** A list of times when a particle exited the system. */
    protected List<Double> m_exitTimes = new ArrayList<Double>();
    /** The color scheme to use for the particles. */
    protected InterpolatingColorScheme m_ics = new InterpolatingColorScheme();
    /**
     * Whether to show a dialog when giving notification messages to
     * the user.
     */
    protected boolean m_bUseDialogNotification = true;
    /** The size for particles. */
    protected double m_particleScale = 10;
    /** The symbol type for particles. */
    protected String m_particleSymbolType = "Bull's Eye";
    /** The number of trails for particles. */
    protected int m_particleNumTrails = 10;
    /** The trail line width for particles. */
    protected int m_particleTrailLineWidth = 1;
    /** The alpha fade for particles. */
    protected float m_particleAlphaFade = 0.8f;
    /** The sleep duration for the time marker. */
    protected int m_tmSleep = 45;
    /** Whether to assign color by tag number instead of depth. */
    protected boolean m_bColorByTag = false;
    /** Whether to show the tag number next to the particle. */
    protected boolean m_bShowParticleTags = false;
    /** The color of the tag number next to the particle. */
    protected String m_particleTagColor = "black";
    /** The color of the tag number next to the particle. */
    protected String m_particleTagFont = "glutBitmapHelvetica12";
    /** The index for generating new particle colors. */
    protected int m_colorGenIndex = 0;
    /** The RGB cube corners to use in spacing out colors. */
    protected double[][] m_colorGenCorners = {{1, 0, 0}, {0, 1, 1}, {1, 1, 0},
        {0, 0, 1}, {0, 1, 0}, {1, 0, 1}};
    /**
     * The list of particle colors that have been generated and stored by tag.
     */
    protected HashMap<Integer, IntBuffer> m_tagColors =
        new HashMap<Integer, IntBuffer>();
    /** Whether to show the options panel in the GUI. */
    protected boolean m_bShowOptionsPanel = false;
    /** The panel created by this class for displaying on a toolbar. */
    protected JPanel m_panel = new JPanel();
    
    
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/ParticleDataFile.gif";

//------------------------------------------------------------------------------
    /**
     * Create an empty data file object.
     */
    public ParticleDataFile() {

        m_numHeaderLines = 1;

        m_inputDelimiters = "\n\r\f,";

        m_ics.setColorAtValue("(255, 255, 255) 70");
        m_ics.setColorAtValue("(255, 0, 255) 80");
        m_ics.setColorAtValue("(255, 0, 0) 90");
        m_ics.setColorAtValue("(255, 255, 0) 105");


        // Check if any default Particle options were set at start up from a
        // config file in the Gr/config directory.
        setDefaults("gov.usgs.crrl.data.ParticleDataFile");
    }

//------------------------------------------------------------------------------
    /**
     * Shows a dialog for configuring this file.
     * @param frame The frame to attach the dialog box to.  Usually the frame of
     *              the window that is controlling opening the file.
     */
    public void showOptionsDialog(Frame frame) {

    }

//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     * @return  The ordered list of GrProperty objects.
     */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        return props;
    }

//------------------------------------------------------------------------------
    /**
     * Set properties from a list of values.
     * @param props The list of GrProperty objects.
     */
    public void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        // Iterate through the list of properties.
        Iterator iter = props.keySet().iterator();
        if (props instanceof GrHashList) {
            iter = ((GrHashList) props).listIterator();
        }
        while (iter.hasNext()) {
            String key = (String) iter.next();
            Object obj = props.get(key);

            // Find the String value let the setProperty method handle it.
            if (obj instanceof GrProperty) {
                p = (GrProperty) obj;

                if (p.value instanceof String) {
                    setProperty(key, (String) p.value);
                }
                if (key.equals("animationTS")) {
                    m_animationTS = TimeReference.convertToDecimalDay((String)
                        p.value);
                }
                if (key.equals("gapLimit")) {
                    m_gapLimit = TimeReference.convertToDecimalDay((String)
                        p.value);
                }
                if (key.equals("particleScale")) {
                    m_particleScale = (Double) p.value;
                }
                if (key.equals("particleSymbolType")) {
                    m_particleSymbolType = (String) p.value;
                }
                if (key.equals("particleNumTrails")) {
                    m_particleNumTrails = (Integer) p.value;
                }
                if (key.equals("particleTrailLineWidth")) {
                    m_particleTrailLineWidth = (Integer) p.value;
                }
                if (key.equals("particleAlphaFade")) {
                    m_particleAlphaFade = (Float) p.value;
                }
                if (key.equals("tmSleep")) {
                    m_tmSleep = (Integer) p.value;
                }
                if (key.equals("colorByTag")) {
                    m_bColorByTag = (Boolean) p.value;
                }
                if (key.equals("showParticleTags")) {
                    m_bShowParticleTags = (Boolean) p.value;
                }
                if (key.equals("particleTagColor")) {
                    m_particleTagColor = (String) p.value;
                }
                if (key.equals("particleTagFont")) {
                    m_particleTagFont = (String) p.value;
                }
                if (key.equals("showOptionsPanel")) {
                    m_bShowOptionsPanel = (Boolean) p.value;
                }
            }
            if (obj instanceof String) {
                String value = (String) obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
    /**
     * Set a single property.
     * @param key   The name of the property being set.
     * @param value The value of the property being set.
     */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("modifyPage")) {
            m_bModifyPage = Boolean.parseBoolean(value);
        }
        if (key.equals("animationTS")) {
            m_animationTS = TimeReference.convertToDecimalDay(value);
        }
        if (key.equals("gapLimit")) {
            m_gapLimit = TimeReference.convertToDecimalDay(value);
        }
        if (key.equals("particleScale")) {
            m_particleScale = Double.parseDouble(value);
        }
        if (key.equals("particleSymbolType")) {
            m_particleSymbolType = value;
        }
        if (key.equals("particleNumTrails")) {
            m_particleNumTrails = Integer.parseInt(value);
        }
        if (key.equals("particleTrailLineWidth")) {
            m_particleTrailLineWidth = Integer.parseInt(value);
        }
        if (key.equals("particleAlphaFade")) {
            m_particleAlphaFade = Float.parseFloat(value);
        }
        if (key.equals("tmSleep")) {
            m_tmSleep = Integer.parseInt(value);
        }
        if (key.equals("colorByTag")) {
            m_bColorByTag = Boolean.parseBoolean(value);
        }
        if (key.equals("showParticleTags")) {
            m_bShowParticleTags = Boolean.parseBoolean(value);
        }
        if (key.equals("particleTagColor")) {
            m_particleTagColor = value;
        }
        if (key.equals("particleTagFont")) {
            m_particleTagFont = value;
        }
        if (key.equals("showOptionsPanel")) {
            m_bShowOptionsPanel = Boolean.parseBoolean(value);
        }
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0) {
            setTreeIcon();
        }
        return (javax.swing.ImageIcon) TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null) {
            icon = super.setTreeIcon();
        }
        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
    /**
     * Read the data through a reader attached to an open file.
     * <p><b> Pre Conditions </b><br>
     *      The header has been read with readHeader()
     * <p><b> Post Conditions </b><br>
     *      The file needs to be closed.
     * @param br    The buffered reader object to use for reading.
     * @throws IOException When file I/O errors occur.
     */
    protected void readData() throws IOException {

        // Remove any data that were children of this object.
        clear();

        // Create lists to hold values
        List<GregorianCalendar> dateList = new ArrayList<GregorianCalendar>();
        List<Double> xList = new ArrayList<Double>();
        List<Double> yList = new ArrayList<Double>();
        List<Double> zList = new ArrayList<Double>();

        List<GlParticleAnimation> glpList =
            new ArrayList<GlParticleAnimation>();

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        String dateDescriptor = "ddMMMyyyy:HH:mm:ss";

        // Read each line of the data file.
        int lineNumber = m_numHeaderLines;
        int firstOutOfOrderLine = -1;
        String[] tokens;
        int partNum = Integer.MIN_VALUE;
        long baseTick = Long.MAX_VALUE;
        long baseMillis = 0;
        long lastMillis = 0;
        long gapLimitMillis = (long)(m_gapLimit * 24d * 60d * 60d * 1000d);
        long animationTsMillis =
            (long)(m_animationTS * 24d * 60d * 60d * 1000d); 
        
        try {
            while ((tokens = AsciiIO.parseLine(m_br.readLine(),
                m_inputDelimiters)) != null) {
                
                lineNumber++;

                if (tokens.length < 6) {
                    continue;
                }
                String dateToken = tokens[2];

                // Create the date object.
                GregorianCalendar date = new GregorianCalendar(utcZone, locale);

                // Parse the current time
                SimpleDateFormat formatter =
                    new SimpleDateFormat(dateDescriptor);
                formatter.setTimeZone(utcZone);
                formatter.setLenient(true);
                date.clear();
                ParsePosition pos = new ParsePosition(0);
                date.setTime(formatter.parse(dateToken, pos));
                long dateMillis = date.getTime().getTime();
                long elapsedDateMillis = dateMillis - baseMillis;

                int parsedPartNum =
                    (int) Math.round(1000 * Double.parseDouble(tokens[1]));
                if (parsedPartNum != partNum ||
                    elapsedDateMillis > gapLimitMillis ||
                    elapsedDateMillis < 0) {

                    GlParticleAnimation glp = createParticle(partNum, dateList,
                        xList, yList, zList);
                    if (glp != null) {
                        glpList.add(glp);
                    }
                    partNum = parsedPartNum;
                    baseTick = Long.MAX_VALUE;
                }

                long tick = Long.parseLong(tokens[0]);
                if (tick < baseTick || elapsedDateMillis > animationTsMillis) {
                    
                    // Set the base tick and millis
                    baseTick = tick;
                    baseMillis = dateMillis;
                }

                // Add the difference from the base tick/time to the base.
                long tickDiff = tick - baseTick;
                long diffMillis = (int) (tickDiff / 12);

                date.clear();
                date.setTime(new Date(baseMillis + diffMillis));

                dateList.add(date);

                xList.add(Double.parseDouble(tokens[3]));
                yList.add(Double.parseDouble(tokens[4]));
                zList.add(Double.parseDouble(tokens[5]));
            }
        } catch (Exception e) {
            System.err.println("Could not read line " + lineNumber);
            e.printStackTrace();
            return;
        }

        if (m_bModifyPage) {
            populatePage(glpList);
        }
    }

//------------------------------------------------------------------------------
    /**
     */
    protected GlParticleAnimation createParticle(int partNum,
        List<GregorianCalendar> dateList, List<Double> xList,
        List<Double> yList, List<Double> zList) {

        if (dateList.size() == 0) {
            return null;
        }
        // Define the start and end dates.
        GregorianCalendar startDate =
            (GregorianCalendar) dateList.get(0).clone();
        GregorianCalendar lastDate =
            (GregorianCalendar) dateList.get(dateList.size() - 1).clone();
        int referenceYear = startDate.get(GregorianCalendar.YEAR);

        // Add NaN to the beginning and end so it will disappear when inactive.
        dateList.add(0, new GregorianCalendar());
        dateList.add(dateList.size(), new GregorianCalendar());
        xList.add(0, Double.NaN);
        xList.add(xList.size(), Double.NaN);
        yList.add(0, Double.NaN);
        yList.add(yList.size(), Double.NaN);
        zList.add(0, Double.NaN);
        zList.add(zList.size(), Double.NaN);

        int numDates = dateList.size();

        // Create an array to hold the interleaved x-y points.
        double[] array = new double[4 * numDates];

        // Interleave the date and data in each array.
        for (int i = 0; i < numDates; i++) {
            GregorianCalendar date =
                (GregorianCalendar) dateList.get(i).clone();

            array[(4 * i)] = julianDecimal(date, referenceYear);
            array[(4 * i) + 1] = xList.get(i);
            array[(4 * i) + 2] = yList.get(i);
            array[(4 * i) + 3] = zList.get(i);
        }

        // Space the times out for the first and last NaN points so they
        // have the spacing as the first two and last two real points.
        double firstStep = array[8] - array[4];
        if (firstStep <= 0) {
            firstStep = m_animationTS;
        }
        if (array.length > 12) {
            array[0] = array[4] - firstStep;
        } else {
            array[0] = array[4] - m_animationTS;
        }
        double lastStep = array[array.length - 8] - array[array.length - 12];
        if (lastStep <= 0) {
            lastStep = m_animationTS;
        }
        array[array.length - 4] = array[array.length - 8] + lastStep;

        m_entryTimes.add(array[4]);
        m_exitTimes.add(array[array.length - 8]);

        // Create a time series from the interleaved array and add it to the
        // series list.
        TimeSeries series = new TimeSeries(array, 4);
        series.setStartDate(startDate);
        series.setProperty("title", "" + partNum);

        // Shift the data to be referenced from the most appropriate year.
        series.syncToReferenceYear(findAppropriateReferenceYear(startDate,
            lastDate));

        GlParticleAnimation glp = new GlParticleAnimation();
        glp.add(series);
        glp.add(m_ics);
        glp.setProperty("title", "" + partNum);
        glp.setProperty("scale", "" + m_particleScale);
        if (m_bColorByTag) {
            glp.setProperty("seriesNums", "1, 1, 1, -1, -1");
            glp.setProperty("dimNums", "0, 1, 2, -1, -1");
            
            int[] color = getTagColor(partNum);
            
            String colorString = "(" + color[0] + "," + color[1] + ","
                + color[2] + ")";
            glp.setProperty("color", colorString);
        }
        else {
            glp.setProperty("seriesNums", "1, 1, 1, -1, 1");
            glp.setProperty("dimNums", "0, 1, 2, -1, 3");
            glp.setProperty("color", "yellow");
        }
        glp.setProperty("symbolType", m_particleSymbolType);
        glp.setProperty("trailLineWidth", "" + m_particleTrailLineWidth);
        glp.setProperty("numTrails", "" + m_particleNumTrails);
        glp.setProperty("alphaFade", "" + m_particleAlphaFade);
        if (m_bShowParticleTags) {
            TextMarkup textMarkup = new TextMarkup();
            String tagString = " ";
            if ((partNum % 1000) == 0)
                tagString += (partNum / 1000);
            else
                tagString += partNum;
            textMarkup.setText("" + tagString);
            textMarkup.setProperty("color", m_particleTagColor);
            textMarkup.setProperty("screenFont", m_particleTagFont);
            glp.add(textMarkup);
        }
        add(glp);

        dateList.clear();
        xList.clear();
        yList.clear();
        zList.clear();

        return glp;
    }

//------------------------------------------------------------------------------
    /**
     */
    protected int[] getTagColor(int partNum) {
            
        int[] color = null;
        
        // Check if the partNum already has a color.
        Iterator<Integer> iter = m_tagColors.keySet().iterator();
        while (iter.hasNext()) {
            Integer key = iter.next();
            if (key == partNum) {
                color = m_tagColors.get(key).array();
                break;
            }
        }

        // Generate a new color for the partNum if one wasn't found.
        if (color == null) {
            color = generateNextColor();
            m_tagColors.put(partNum, IntBuffer.wrap(color));
        }
        
        return color;
    }

//------------------------------------------------------------------------------
    /**
     */
    protected int[] generateNextColor() {
        
        int[] color = {0, 0, 0};
        double[] dColor = {0, 0, 0};
        double[] origin = {0, 0, 0};
        double dist = 0;
        double allowableDist = 0.5;
        double v0 = m_colorGenCorners[m_colorGenIndex][0];
        double v1 = m_colorGenCorners[m_colorGenIndex][1];
        double v2 = m_colorGenCorners[m_colorGenIndex][2];

        while (dist < allowableDist) {
            dColor[0] = Math.random();
            dColor[1] = Math.random();
            dColor[2] = Math.random();

            dColor[0] = dColor[0] * (dColor[0] + 1) * 0.5;
            dColor[1] = dColor[1] * (dColor[1] + 1) * 0.5;
            dColor[2] = dColor[2] * (dColor[2] + 1) * 0.5;

            if (v0 < 0)
                dColor[0] *= -1;
            dColor[0] = v0 + dColor[0];
            if (v1 < 0)
                dColor[1] *= -1;
            dColor[1] = v1 + dColor[1];
            if (v2 < 0)
                dColor[2] *= -1;
            dColor[2] = v2 + dColor[2];

            double bCoord = (dColor[0] + dColor[1] + dColor[2]) / 3;
            double[] bPoint = new double[3];
            bPoint[0] = bCoord;
            bPoint[1] = bCoord;
            bPoint[2] = bCoord;
            dist = distPointPoint(dColor, bPoint);
            allowableDist = 0.6 * (1 - distPointPoint(origin, bPoint));
            allowableDist = Math.min(allowableDist, 0.15);
        }

        color[0] = Math.min((int)Math.round(dColor[0] * 255), 255);
        color[1] = Math.min((int)Math.round(dColor[1] * 255), 255);
        color[2] = Math.min((int)Math.round(dColor[2] * 255), 255); 

        // Advance the index for the corner of the RGB cube.
        m_colorGenIndex++;
        if (m_colorGenIndex > 5)
            m_colorGenIndex = 0;        
        
        return color;
    }

//------------------------------------------------------------------------------
    /**
     */
    protected double distPointLine(double[] l0, double[] l1, double[] p) {
        
        double[] p1 = intersectionPointLine(l0, l1, p);
        return distPointPoint(p, p1);
}

//------------------------------------------------------------------------------
    /**
     */
    protected double[] intersectionPointLine(double[] l0, double[] l1,
        double[] p) {
        
        // Dist is ABS(A * x0 + B * y0 + C) / SQRT(A^2 + B^2)
        // for line Ax + By + C = 0  and point (x0, y0)
        
        return new double[2];
}
    
    
//------------------------------------------------------------------------------
    /**
     */
    protected double distPointPoint(double[] p0, double[] p1) {
        
        if (p0.length == 3) {
            double lx = p1[0] - p0[0];
            double ly = p1[1] - p0[1];
            double lz = p1[2] - p0[2];
            return Math.sqrt((lx * lx) + (ly * ly) + (lz * lz));
        }
        if (p0.length == 2) {
            double lx = p1[0] - p0[0];
            double ly = p1[1] - p0[1];
            return Math.sqrt((lx * lx) + (ly * ly));
        }
        if (p0.length == 1) {
            return Math.abs(p1[0] - p0[0]);
        }
        
        return 0;
}

//------------------------------------------------------------------------------
    /**
     *
     */
    public void populatePage(List<GlParticleAnimation> glpList) {

        GrController grCon = Gr.getController();
        TimeSeries series = null;
        TimeSeries entrySeries = null;
        TimeSeries sharedSeries = null;
        TimeMarker tm = null;

        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject) iter.next();
            if (gro instanceof TimeSeries) {
                series = (TimeSeries) gro;
            }
        }

        if (grCon == null) {
            return;
        }
        try {
            // Make sure the animation template is in place.
            grCon.createAnimation();
            PageOfGraphs page = grCon.getPage();

            synchronized (page) {

                List<Graph> list = page.getGraphList();
                Zoom tZoom = null;

                for (Graph graph : list) {
                    GrHashList props = graph.getProperties();
                    GrProperty p = (GrProperty) props.get("aspect1to1");
                    boolean bAspect1to1 = (Boolean) p.value;

                    if (bAspect1to1) {
                        tm = (TimeMarker) page.firstChildInstanceOf(
                        "gov.usgs.gr.tools.TimeMarker", false);

                        TextMarkup text = (TextMarkup) tm.firstChildInstanceOf(
                        "gov.usgs.gr.visual.TextMarkup", false);
                        text.setProperty("x", "0.5");
                        tm.setProperty("dateFormat",
                            "MMM'.' dd',' yyyy 'at' HH:mm:ss");
                        tm.setProperty("sleep", "" + m_tmSleep);

                        if (!m_bColorByTag) {
                            GlColorBar colorBar = new GlColorBar();
                            InterpolatingColorScheme ics =
                                (InterpolatingColorScheme) firstChildInstanceOf(
                                "gov.usgs.gr.visual.InterpolatingColorScheme",
                                true);
                            colorBar.add(ics);
                            colorBar.setProperty("screenFont",
                                "glutBitmapHelvetica12");
                            colorBar.setProperty("xCoordLevel",
                                "Frame Interior 0 to 1");
                            colorBar.setProperty("yCoordLevel",
                                "Frame Interior 0 to 1");
                            colorBar.setProperty("x", "0.5");
                            colorBar.setProperty("y", "0.2");
                            colorBar.setProperty("width", "0.1");
                            colorBar.setProperty("height", "0.6");
                            colorBar.setProperty("text", "Elevation in meters");
                            graph.add(colorBar);
                        }

                        for (GlParticleAnimation glp : glpList) {
                            graph.add(glp);

                            // Add the time series to the time marker.
                            TimeSeries tSeries =
                                (TimeSeries) glp.firstChildInstanceOf(
                            "gov.usgs.gr.data.TimeSeries", false);
                            if (tSeries != null) {
                                if (tm != null) {
                                    tm.add(tSeries);
                                }
                                GrHashList dims = new GrHashList();
                                dims.add(0, tSeries.getDimension(0), tSeries);
                                TimeSeries tempSeries = (TimeSeries)
                                    (DimensionOrganizer.combine(dims, false));
                                tempSeries.resample(m_animationTS, 0);
                                tempSeries.selectAll();
                                double halfMilli = 1d / 24d / 60d / 60d / 2000d;
                                tempSeries.modifySelectedPoints(1, 0,
                                    halfMilli);
                                tempSeries.deselectAll();
                                if (sharedSeries == null) {
                                    sharedSeries = tempSeries;
                                } else {
                                    sharedSeries.mergeWith(tempSeries, 0,
                                        false);
                                }
                            }
                        }

                        if (tm != null) {
                            tm.add(sharedSeries);
                        }
                        graph.add(text);

                        // Zoom all
                        graph.setSelected(true);
                        grCon.zoomAll();
                        graph.setSelected(false);
                    } else {
                        // Add a time series to the time graph showing entries
                        // into the system.
                        Object[] entryArray = m_entryTimes.toArray();
                        Arrays.sort(entryArray);
                        double[] entries = new double[entryArray.length * 2];
                        for (int i = 0; i < entryArray.length; i++) {
                            entries[i * 2] = (Double) entryArray[i];
                            entries[i * 2 + 1] = 1;
                        }
                        entrySeries = new TimeSeries(2);
                        entrySeries.setArray(entries);

                        Object[] exitArray = m_exitTimes.toArray();
                        Arrays.sort(exitArray);
                        double[] exits = new double[exitArray.length * 2];
                        for (int i = 0; i < exitArray.length; i++) {
                            exits[i * 2] = (Double) exitArray[i];
                            exits[i * 2 + 1] = -1;
                        }
                        TimeSeries exitSeries = new TimeSeries(2);
                        exitSeries.setArray(exits);
                        entrySeries.mergeWith(exitSeries, 0, true);
                        double[] array = entrySeries.getArray();
                        int numPoints = entrySeries.getCount();
                        int total = 0;
                        ArrayList<Double> valList = new ArrayList<Double>();
                        Double lastVal = Double.NEGATIVE_INFINITY;

                        // Make the series a step function with exactly two
                        // values at each time.
                        for (int i = 0; i < numPoints; i++) {
                            if (array[i * 2] != lastVal) {
                                valList.add(array[i * 2]);
                                valList.add(new Double(total));
                            } else {
                                valList.remove(valList.size() - 1);
                                valList.remove(valList.size() - 1);
                            }
                            total += array[i * 2 + 1];
                            valList.add(array[i * 2]);
                            valList.add(new Double(total));
                        }

                        entrySeries.setArray(valList);

                        graph.add(entrySeries);

                        graph.setTitle("Number of Active Fish");
                        graph.setShowMainTitle(true);
                        graph.getXAxis().setTitle("Julian Day");
                        graph.getXAxis().setShowTitle(true);

                        // Zoom all
                        graph.setSelected(true);
                        Zoom yZoom = graph.getYZoom();
                        yZoom.setProperty("autoSteps", "5, 10, 20, 50, 100");
                        grCon.zoomAll();
                        grCon.zoomAll();
                        graph.setSelected(false);
                        tZoom = graph.getXZoom();
                    }

                    if (tZoom != null) {
                        // Zoom all
                        tZoom.zoomAll();
                        tZoom.zoomAll();

                        // Pan to left
                        tZoom.pan(Double.POSITIVE_INFINITY);
                        tZoom.zoomAll();
                        tZoom.pan(Double.NEGATIVE_INFINITY);
                        grCon.zoomAll();

                        grCon.setStatusBarDateFormat(4);
                    }
                }

                // Set the reference year of the entry series to match the
                // others.
                if (sharedSeries != null) {
                    if (entrySeries != null) {
                        entrySeries.setReferenceYear(
                            sharedSeries.getReferenceYear());
                    }
                    if (tm != null) {
                        tm.setProperty("year",
                            "" + sharedSeries.getReferenceYear());
                    }

                    JMenuItem jmi = new JMenuItem();
                    jmi.setText("Sync Page to this Reference Year");
                    ActionEvent aEvent = new ActionEvent(jmi, 0,
                        "Sync Page to this Reference Year");
                    sharedSeries.actionPerformed(aEvent);
                }

                if (m_bShowOptionsPanel) {
                    ParticleOptionsEditor poe = (ParticleOptionsEditor)
                        firstChildInstanceOf(
                        "gov.usgs.crrl.data.ParticleOptionsEditor", false);
                    if (poe == null) {
                        poe = new ParticleOptionsEditor();
                        poe.setShowParticleTags(m_bShowParticleTags);
                        page.add(poe);
                    }
                }
                    
                grCon.setPage(page);
            }
        } catch (NullPointerException e) {
            return;
        }
    }

//------------------------------------------------------------------------------
/**
     * Shows a dialog for configuring the file.
     */
    public static boolean matchesFormat(String[] lines) {

        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        String dateDescriptor = "ddMMMyyyy:HH:mm:ss";

        for (int i = 1; i < lines.length; i++) {
            String[] tokens = AsciiIO.parseLine(lines[i], ",");


            if (tokens.length < 6) {
                return false;
            }
            try {
                String dateToken = tokens[2];

                // Create the date object.
                GregorianCalendar date = new GregorianCalendar(utcZone, locale);

                // Parse the current time
                SimpleDateFormat formatter =
                    new SimpleDateFormat(dateDescriptor);
                formatter.setTimeZone(utcZone);
                formatter.setLenient(true);
                date.clear();
                ParsePosition pos = new ParsePosition(0);
                date.setTime(formatter.parse(dateToken, pos));
            } catch (Exception e) {
                return false;
            }

            try {
                Integer.parseInt(tokens[0]);
                Double.parseDouble(tokens[1]);
                Double.parseDouble(tokens[3]);
                Double.parseDouble(tokens[4]);
                Double.parseDouble(tokens[5]);
            } catch (NumberFormatException ex) {
                return false;
            }
        }

        return true;
    }

//------------------------------------------------------------------------------
    /**
     * Whether this file class has an options dialog to display.
     * @return true if their is an options dialog available.
     */
    public boolean hasOptionsDialog() {

        return false;
    }

}
