/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 *                  California District, Sacramento
 *                  San Francisco Bay Hydrodynamics Project
 * @version         2008-10-14
 * Date started:    1998-06-15
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact Information
 * -------------------
 * John Donovan
 * jmd@usgs.gov
 * (916) 278-3120
 *
 * Placer Hall
 * 6000 J Street
 * Sacramento, CA  95819-6129
 *
 * The Gr web page is at:
 *
 * http://ca.water.usgs.gov/program/sfbay/gr/
 */

package gov.usgs.gr;

import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.gui.FileDropHandler;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import java.util.*;

//==============================================================================
/**
 * A graphing application in a frame.  Opens x-y data files and displays them
 * in multiple framed graphs.  The graphs can be drawn on the screen or
 * printed to a PostScript file.  On screen, graphs can be hidden and the
 * user can zoom in and pan around.
 */
public class Gr extends JFrame implements WindowListener {
    
    /** An object that creates a panel and handles commands. */
    protected GrController m_grCon;
    /** A static list for accessing any object to access the controller. */
    protected static List<GrController> m_grControllerList =
        new ArrayList<GrController>();
    /** True if we are using an OpenGL view. */
    protected boolean m_bUsingOpenGL = true;
    /** A list of file names to open from the command line. */
    protected List<String> m_filesToOpen = new ArrayList();
    
    // Some GL linetype definitions that other classes might use...
    public static final short SOLID_LINE_PATTERN        = 0x0000;
    public static final short TIGHT_DOTTED_LINE_PATTERN = 0x5555;
    public static final short DOTTED_LINE_PATTERN       = 0x1111;
    public static final short LOOSE_DOTTED_LINE_PATTERN = 0x0101;
    public static final short TIGHT_DASHED_LINE_PATTERN = 0x3333;
    public static final short DASHED_LINE_PATTERN       = 0x0F0F;
    public static final short LOOSE_DASHED_LINE_PATTERN = 0x00FF;
    public static final short DASH_DOT_LINE_PATTERN     = 0x060F;
    public static final short DASH_DOT_DASH_LINE_PATTERN = 0x1C47;
    
    /**
     * A list of objects to be used as default configurations.  The values
     * are read from disk.
     */
    public static GrHashList DEFAULT_OBJ_PROPS = new GrHashList();
    
//------------------------------------------------------------------------------
    /**
     * Main constructor for the Gr application.
     * @param title The string that will appear in the window's title bar.
     */
    public Gr(String title) {
        
        // Set up this window frame.
        super(title);
        addWindowListener(this);
    }
    
//------------------------------------------------------------------------------
    /**
     * Setup the contents of this window and read in data files that were given
     * on the command line.
     * <p><b> Pre Conditions </b><br>
     *      The member components have been instantiated.
     * <p><b> Post Conditions </b><br>
     *      This object still needs to be made visible.
     */
    public void setupAndShow() {
        
        // Get the Gr Home property and set the current file property.
        String grHome = System.getProperty("gov.usgs.gr.grhome");
        
        // If the property was not found, set it to a default value.
        if (grHome == null) {
            System.setProperty("gov.usgs.gr.grhome",
                System.getProperty("user.dir"));
            grHome = System.getProperty("gov.usgs.gr.grhome");
        }
        
        grHome.replace('\\', '/');
        System.setProperty("gov.usgs.gr.currentfile", grHome);
        
        // Look for Gr configuration files in XML format for default
        // object properties.  Create a list of files in the config directory
        // within the Gr home directory that begin with "gr_config" and end
        // with ".xml"
        String[] fileNames = new File(grHome + "/config").list(
            new FilenameFilter() {
            
            public boolean accept(File dir, String name) {
                name = name.toLowerCase();
                return (name.startsWith("gr_config") && name.endsWith(".xml"));
            }
        });
        
        // Read the defaults files.
        if (fileNames == null) fileNames = new String[0];
        for (int i = 0; i < fileNames.length; i++) {
            configureDefaults(grHome + "/config/" + fileNames[i]);
        }
        
        // Check that the main OpenGL canvas class is available.
        if (m_bUsingOpenGL) {
//            try {
//                Class.forName("javax.media.opengl.GLCanvas");
//            }
//            catch (ClassNotFoundException e) {
//                System.err.println(" *** Error *** ");
//                System.err.println(" JOGL not found");
//                System.err.println(" Check that OpenGL libraries were installed"
//                    + " to Java home from");
//                System.err.println(" JRE Add-ons zip file " + '\n');
//                System.err.println(" Press any key to continue...");
//                try {System.in.read();} catch (Exception e2){}
//                return;
//            }
        }
        
        // Create a controller for the program and set it to use OpenGL
        // for the view.
        m_grCon = new GrController();
        m_grCon.setUsingOpenGL(m_bUsingOpenGL);
        m_grCon.setupAndShow(this);
        m_grControllerList.add(m_grCon);
        
        // Read the Gr configuration file to set the data format.
        // Check the Gr directory and the user's home directory for
        // a file called gr_config.grs.
        File configFile = new File(grHome + "/gr_config.grs");
        if (configFile.canRead())
            m_grCon.openScript(configFile);
        
        // Open any files that were specified on the command line.
        for (String filename : m_filesToOpen)
            m_grCon.openFile(new File(filename));
        
        // Enable Drag and Drop
        FileDropHandler fdh = new FileDropHandler(this, m_grCon);
        
        // Let the user start off by opening a data file if none were
        // given on the command line.
//        if (m_filesToOpen.size() == 0) m_grCon.openFile();
    }
    
//------------------------------------------------------------------------------
/*
 * Reads the specified file to get default object attributes.
 * @param fileName  The name of the file to read.  It must be XML.
 */
    public void configureDefaults(String fileName) {
        
        File configFile = new File(fileName);
        if (configFile.canRead()) {
            ConfigXMLFile xmlConfig = new ConfigXMLFile();
            xmlConfig.setProperty("file", configFile.getAbsolutePath());
            try {
                // Read the file and get its class propery hash map.
                xmlConfig.read();
                GrHashList newClassMap = xmlConfig.getClassProps();
                
                // Look at each class in the new map.
                Iterator iter = newClassMap.listIterator();
                while (iter.hasNext()) {
                    String className = (String)iter.next();
                    
                    // Get the old and new propery maps for the class
                    GrHashList masterPropMap =
                        (GrHashList)DEFAULT_OBJ_PROPS.get(className);
                    Object testObj = newClassMap.get(className);
                    GrHashList newPropMap =
                        (GrHashList)newClassMap.get(className);
                    
                    if (masterPropMap == null) {
                        // Assign the new map if there wasn't an old one.
                        DEFAULT_OBJ_PROPS.put(className, newPropMap);
                    } else {
                        // Add the new properties to the old map.
                        Iterator propIter = newPropMap.listIterator();
                        while (propIter.hasNext()) {
                            String propName = (String)propIter.next();
                            masterPropMap.put(propName,
                                newPropMap.get(propName));
                        }
                    }
                }
            } catch (Exception e) {}
        }
    }
    
//------------------------------------------------------------------------------
/*
 * Parses the command line and opens each file that is listed.
 * @param args	An array of command line arguments.
 */
    public void parseCommandLine(String args[]) {
        
        // Look at each argument.
        for (int i = 0; i < args.length; i++) {
            // Check for the OpenGL option.
            if (args[i].compareToIgnoreCase("-gl") == 0) {
                m_bUsingOpenGL = true;
                continue;
            }
            
            // Check for the Java2D option.
            if (args[i].compareToIgnoreCase("-j2d") == 0) {
                m_bUsingOpenGL = false;
                continue;
            }
            
            // If the argument isn't an option, assume it is a file name
            // and add it to the list of files to be opened.
            m_filesToOpen.add(args[i]);
        }
    }
    
//------------------------------------------------------------------------------
/*
 *
 * @return
 */
    public static GrController getController() {
        
        if (m_grControllerList.size() == 0)
            return null;
        else
            return m_grControllerList.get(0);
        
    }
    
//------------------------------------------------------------------------------
/*
 * Returns a string describing this version of Gr.
 * @return	A string describing this version of Gr.
 */
    public static String getVersion() {
        
        return ("2008-10-14");
    }
    
//------------------------------------------------------------------------------
/*
 * When the user closes the window, destroy everything and close the program.
 */
    public void windowClosing( WindowEvent evt ) {
        
        dispose();
        System.exit( 0 );
    }
    
//------------------------------------------------------------------------------
// Window listener methods that must be defined...
    public void windowClosed( WindowEvent evt ) {}
    public void windowActivated( WindowEvent evt ) {}
    public void windowDeactivated( WindowEvent evt ) {}
    public void windowDeiconified( WindowEvent evt ) {}
    public void windowIconified( WindowEvent evt ) {}
    public void windowOpened( WindowEvent evt ) {}
    
//------------------------------------------------------------------------------
/*
 * Create an instance of the Gr program.
 */
    static public void main(String args[]) {
        
//System.getProperties().list(System.out);
        System.out.println("USGS Gr version " + getVersion());
        Gr gr = new Gr("Gr");
        gr.parseCommandLine(args);
        gr.setupAndShow();
    }
}











