/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-07-03
 * Date started:    1998-10-20
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */
package gov.usgs.gr;

import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.gl.*;
import gov.usgs.gr.visual.j2d.*;
import gov.usgs.gr.visual.ps.*;
import gov.usgs.gr.visual.gui.*;
import gov.usgs.sfhydro.data.*;
import gov.usgs.sfhydro.gr.formats.VariableSizeFormatter;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Rectangle;
import java.awt.event.*;
import java.awt.geom.*;
import java.io.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import java.lang.reflect.*;
import java.awt.EventQueue;

//==============================================================================
/**
 * A factory that creates either a Java2D or OpenGL drawing panel and handles
 * the GUI responsibilities for the application.
 */
public class GrController extends GrObject {

    /** Whether or not to allow the X axis to be moved or scaled. */
    public boolean m_bLockX = false;
    /** Whether or not to allow the Y axis to be moved or scaled. */
    public boolean m_bLockY = false;
    /** Whether to interpret mouse movement as zooming. */
    protected boolean m_bZoomMode = true;
    /** Whether to interpret mouse movement as panning. */
    protected boolean m_bPanMode = false;
    /** Whether to interpret mouse movement as selecting points. */
    protected boolean m_bSelectMode = false;
    /** Whether to interpret mouse movement as moving points. */
    protected boolean m_bDragMode = false;
    /** Whether to interpret mouse movement as drawing a new curve. */
    protected boolean m_bDrawMode = false;
    /** The curve to append to in draw mode. */
    protected Curve m_drawModeCurve = null;
    /** The graph containing the draw mode curve. */
    protected Graph m_drawModeGraph = null;
    /** The insertion point for draw mode. */
    protected int m_drawModeIndex = -1;
    /** The size of editable points. */
    protected float m_editablePointSize = 5;
    /**
     * Whether to save time by drawing a simpler representation of the
     * graphs.
     */
    protected boolean m_bQuickDraw = false;
    /** True if we are using an OpenGL view. */
    protected boolean m_bUsingOpenGL = true;
    /** Whether to add the children of files that are opened. */
    protected boolean m_bAddChildrenWhenOpening = true;
    /** The title of the current page.  This is written on the Gr window's
     *  title bar.
     */
    protected String m_pageTitle = "";
    /** The menu bar that lets the user give commands to the program. */
    protected GrMenuBar m_grMenuBar;
    /** The status bar object for sending information to the user. */
    protected GrStatusBar m_grStatusBar = null;
    /** The small version of the status bar. */
    protected GrStatusBar m_littleStatusBar = null;
    /** The big version of the status bar. */
    protected CurveStatusBar m_bigStatusBar = null;
    /** The data set object for loading files. */
    protected XyDataFile m_dataFile;
    /** The object containing options for the GUI. */
    protected GuiOptions m_guiOptions;
    /** The list of data series that have been cut or copied. */
    protected GrObjectContainer m_clipBoard = new GrObjectContainer();
    /** The last known x-coordinate of the mouse. */
    protected int m_mouseX = 0;
    /** The last known y-coordinate of the mouse. */
    protected int m_mouseY = 0;
    /** A page of graphs contained within this panel. */
    protected PageOfGraphs m_page;
    /** A template page that can be used with open data. */
    protected PageOfGraphs m_template = null;
    /** A panel to hold and draw all of the graphs. */
    protected PageProxy m_pageProxy;
    /** An alternate mouse listener. */
    protected GrMouseHandler m_mouseHandler = null;
    /** A list of data snap shots that can be restored. */
    protected GrObjectContainer m_undoList = new GrObjectContainer();
    /** For redoing something that was undone. */
    protected GrObjectContainer m_redoList = new GrObjectContainer();
    /** The type of XyDataFile to use when reading in data. */
    protected String m_dataFormatClass = "gov.usgs.sfhydro.data.GsDataFile";
    /** The path to the current working directory. */
    protected String m_currentPath = "";
    /** The name of the input or output file. */
    protected File m_currentFile = new File("gr.dat");
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GrController.gif";

//------------------------------------------------------------------------------
    /**
     * Create a GrController through the super class.
     */
    public GrController() {

        setDefaults("gov.usgs.gr.GrController");

        // Set up the undo/redo lists.
        m_undoList.addObserver(this);
        m_redoList.addObserver(this);
    }

//------------------------------------------------------------------------------
    /**
     * Create a GrController and wrap it around an existing page of graphs.
     * @param page	The page to place inside the new conroller.
     */
    public GrController(PageOfGraphs page) {

        setDefaults("gov.usgs.gr.GrController");

        m_undoList.addObserver(this);
        m_redoList.addObserver(this);

        m_page = page;
        add(page);
    }

//------------------------------------------------------------------------------
    /**
     * Create a GrController as a copy of another GrController.  Share the data.
     * @param	grCon	The template controller to use in creating this one.
     */
    public GrController(GrController grCon) {

        // Set up the undo/redo lists.
        m_undoList.addObserver(this);
        m_redoList.addObserver(this);

        // Copy state variables from the template controller.
        m_bLockX = grCon.getLockX();
        m_bLockY = grCon.getLockY();
        m_bZoomMode = grCon.getZoomMode();
        m_bPanMode = grCon.getPanMode();
        m_bSelectMode = grCon.getSelectMode();
        m_bQuickDraw = grCon.getQuickDraw();

        // Create a new page to match the page in the template.  If it was
        // OpenGL, create a new OpenGL page.  If was Java2D, use Java2D.
        if (grCon.getPage() instanceof PageOfGlGraphs) {
            m_page = new PageOfGlGraphs(grCon.getPage());
        } else {
            m_page = new PageOfJ2dGraphs(grCon.getPage());
        }

        add(m_page);
    }

//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.  Properties
     * include <p>
     * version      String  "Version"       no edit, no save    <p>
     * author       String  "Author"        no edit, no save    <p>
     * organization String  "Organization"  no edit, no save    <p>
     * location     String  "Location"      no edit, no save    <p>
     * @return A list of GrProperty objects describing this object's properties.
     */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("version", new GrProperty(gov.usgs.gr.Gr.getVersion(),
                "Version", false));
        props.put("author", new GrProperty("John M. Donovan",
                "Author", false));
        props.put("organization", new GrProperty(
                "U.S. Geological Survey, Water Resources Division",
                "Organization", false));
        props.put("location", new GrProperty(
                "Sacramento, California USA", "Location", false));

        return props;
    }

//------------------------------------------------------------------------------
    /**
     * Set the values of one or more of this object's properties.
     * @see gov.usgs.gr.GrController#getProperties getProperties() for a list
     * of accepted properties.
     * @param props A list of GrProperty objects specifying the new values for
     *              certain properties.
     */
    @Override
    public void setProperties(HashMap props) {

        // Let the super class handle any properties it recognizes.
        super.setProperties(props);

        GrProperty p;

        // For each property in the list.
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String) iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty) obj;

                // Handle string properties by passing them to the
                // setProperty() method.
                if (p.value instanceof String) {
                    setProperty(key, (String) p.value);
                    continue;
                }
            }
            // Handle string properties by passing them to the
            // setProperty() method.
            if (obj instanceof String) {
                setProperty(key, (String) obj);
            }
        }
    }

//------------------------------------------------------------------------------
    /**
     * Set a single property.
     * @see gov.usgs.gr.GrController#getProperties getProperties() for a list
     * of accepted properties.
     * @param key   The property name.
     * @param value The new value for the property.
     */
    @Override
    public void setProperty(String key, String value) {

        // Let the super class handle the property if it recognizes it.
        super.setProperty(key, value);

        if (key.equals("dataFormatClass")) {
            m_dataFormatClass = value;
        }
    }

//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    @Override
    public String toString() {

        String s = "Gr";
        return s;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An XML name of this element.
     */
    @Override
    public String getXmlElementName() {
        return "gov.usgs.gr";
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    @Override
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0) {
            setTreeIcon();
        }

        return (javax.swing.ImageIcon) TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    @Override
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null) {
            icon = super.setTreeIcon();
        }

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
    /**
     * Adds an object as a child.  Accepts all types of GrObject.  If it is an
     * instance of GuiOptions, it will replace the existing instance.  After
     * adding, the children are resorted so that the order is GUI Options,
     * Pages, all other objects.
     * @param gro   The object to add.
     * @return  Whether the object was successfully added.
     */
    @Override
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy) {
            gro = ((ObjectEffigy) gro).getSubject();
        }

        if (gro instanceof GrFile || gro instanceof PageOfGraphs ||
            gro instanceof DataSeries || gro instanceof GuiOptions) {

            // If the object is a GuiOptions, remove the existing GuiOptions
            // and place the new one first.
            if (gro instanceof GuiOptions) {
                Iterator iter = m_children.iterator();
                while (iter.hasNext()) {
                    Object o = iter.next();
                    if (o instanceof GuiOptions) {
                        iter.remove();
                    }
                }

                super.add(gro);
                m_children.remove(gro);
                m_children.add(0, gro);

                m_guiOptions = (GuiOptions) gro;
                update(gro, null);

                return true;
            }

            // If the object is a page, convert Java2D to OpenGL or vice versa
            // as needed.
            if (m_bUsingOpenGL && gro instanceof PageOfGraphs &&
                    !(gro instanceof PageOfGlGraphs)) {

                gro = new PageOfGlGraphs((PageOfGraphs) gro);
            }
            if (!m_bUsingOpenGL && gro instanceof PageOfGraphs &&
                    !(gro instanceof PageOfJ2dGraphs)) {

                gro = new PageOfJ2dGraphs((PageOfGraphs) gro);
            }

            // Add the object.
            super.add(gro);

            // If it is a page, set it as the current page.
            if (gro instanceof PageOfGraphs) {
                setPage((PageOfGraphs) gro);
            }

            // Sort the pages to the front.
            int lastPage = 0;
            for (int i = 0; i < m_children.size(); i++) {
                // Find pages.
                if (m_children.get(i) instanceof PageOfGraphs) {
                    // Make sure they are grouped together at the front.
                    if (lastPage < i) {
                        GrObject child = m_children.remove(i);
                        m_children.add(lastPage + 1, child);
                        lastPage++;
                    }
                }
            }

            return true;
        } // Just add it in anyway.
        else {
            super.add(gro);
            return true;
        }
    }

//------------------------------------------------------------------------------
    /**
     * Removes child at index i.  Doesn't allow the GuiOptions child be be
     * removed.  Reassigns the current page, in case it was removed.
     * @param i The index of the child to be removed.
     * @return  The object that was removed.
     */
    @Override
    public GrObject remove(int i) {

        // Don't allow the GUI Options to be removed.
        if (m_children.get(i) instanceof GuiOptions) {
            return null;
        }

        // Let the super class' method removed the object.
        GrObject gro = super.remove(i);

        //Reassigns the current page, in case it was removed.
        checkForPageRemoval(gro);

        return gro;
    }

//------------------------------------------------------------------------------
    /**
     * Removes the specified child.  Doesn't allow the GuiOptions child be be
     * removed.  Reassigns the current page, in case it was removed.
     * @param gro The child to be removed.
     * @return  Whether the object was successfully removed.
     */
    public boolean remove(GrObject gro) {

        // Don't allow the GUI Options to be removed.
        if (gro instanceof GuiOptions) {
            return false;
        }

        // Let the super class' method removed the object.
        boolean bFound = super.remove(gro);

        checkForPageRemoval(gro);

        return bFound;
    }

//------------------------------------------------------------------------------
    /**
     * Returns an XML version of this element and its children using the
     * Document Object Model.
     * @param doc   A document that can be used to create new DOM elements.
     * @return  A new DOM element containing this object and its children.
     */
    public org.w3c.dom.Element getDOMRepresentation(org.w3c.dom.Document doc) {

        org.w3c.dom.Element el = doc.createElement(getXmlElementName());
        return el;
    }

//------------------------------------------------------------------------------
    /**
     * Overrides default update() in GrObject.  Called by an observed GrObject's
     * notifyObservers() method to handle changes to the observed object.
     * May use the notifyObservers() method to propagate the event handling up
     * to this object's observers.
     * @param gro   The observed object that changed.
     * @param arg   Additional information that the updated object may choose to
     *              send.
     */
    @Override
    public void update(GrObject gro, Object arg) {

        // If the observed object is the current page...
        if (gro == m_page) {
            // If the page needs to be redrawn, update its content lists
            // and lay it out again and trigger a repaint.
            if (m_page.getNeedsRedraw() && !m_page.getDrawing()) {
                m_page.updateLists();
                m_page.getFormatter().layoutGraphs();
                if (!m_bDeferUpdates) {
                    m_pageProxy.repaint();
                }
            }

            // Look for an alternate mouse handler as a child of the current
            // page.
            boolean bFound = false;
            Enumeration enumeration = m_page.children();
            while (enumeration.hasMoreElements()) {
                GrObject child = (GrObject) enumeration.nextElement();
                if (child instanceof GrMouseHandler) {
                    m_mouseHandler = (GrMouseHandler) child;
                    m_mouseHandler.setControllerAndPage(this, m_page);
                    bFound = true;
                } else {
                    if (child instanceof GuiPanelProvider) {
                        GuiPanelProvider gpp = (GuiPanelProvider) child;
                        gpp.setControllerAndPage(this, m_page);
                    }
                }
            }
            if (!bFound) {
                m_mouseHandler = null;
            }

            updateWindowTitle("");
        }

        // If there is no current page, repaint the empty panel.
        if (m_page == null && m_pageProxy != null) {
            m_pageProxy.repaint();
        }

        // Handle changes to GUI options.
        if (gro instanceof gov.usgs.gr.visual.gui.GuiOptions) {
            // Update all components to the new look and feel.
            GrHashList props = gro.getProperties();
            GrProperty p = (GrProperty) props.get("lookAndFeel");

            try {
                UIManager.setLookAndFeel((String) p.value);
                SwingUtilities.updateComponentTreeUI(
                        getParentFrame(m_pageProxy.getComponent()));
                SwingUtilities.updateComponentTreeUI(m_bigStatusBar);
                SwingUtilities.updateComponentTreeUI(m_littleStatusBar);
                getParentFrame(m_pageProxy.getComponent()).pack();
                p.value = UIManager.getLookAndFeel().getClass().getName();
                props.clear();
                props.put("lookAndFeel", p);
                gro.setProperties(props);
            } catch (Exception e) {
            }

            // Shrink the undo list if it has grown bigger than the limit.
            if (m_guiOptions != null) {
                int undoLevels = m_guiOptions.getUndoLevels();
                while (m_undoList.getChildCount() > undoLevels) {
                    m_undoList.remove(0);
                }
            }
        }
    }

//------------------------------------------------------------------------------
    /**
     * Cause redraws and other effects of changing children to be deferred
     * until later.  Overridden from GrObject.
     */
    @Override
    public void deferUpdates() {

        m_bDeferUpdates = false;
        if (m_page == null) {
            return;
        }
        if (m_page.getNeedsRedraw()) {
            m_pageProxy.repaint();
        }
    }

//------------------------------------------------------------------------------
    /**
     * Receives a status bar for the panel to use in displaying messages.
     * @param statusBar The status bar object to be used by the panel.
     */
    public void setStatusBar(GrStatusBar statusBar) {

        m_grStatusBar = statusBar;
    }

//------------------------------------------------------------------------------
    /**
     * Setup the contents of this window.
     * @param frame	The frame to put the GUI and drawing panel in.
     */
    public void setupAndShow(JFrame frame) {

        // Find out the width and height of the screen in pixels.
        Dimension d = frame.getToolkit().getScreenSize();

        // Set the icon in the window to be the Gr icon.
        ImageIcon image = new javax.swing.ImageIcon(Thread.currentThread().
                getContextClassLoader().getResource("gr.gif"));
        frame.setIconImage(image.getImage());

        // Create a GUI-options object and make it a child of this controller.
        // If the GUI-options doesn't have a preferred look and feel, set it
        // to be the current look and feel.
        GrProperty p = new GrProperty(
                UIManager.getLookAndFeel().getClass().getName());
        GrHashList fallBackProps = new GrHashList();
        fallBackProps.put("lookAndFeel", p);
        GuiOptions guiOptions = new GuiOptions();
        GrHashList guiProps = guiOptions.getProperties();
        p = (GrProperty) guiProps.get("lookAndFeel");
        if (p.value == null) {
            guiOptions.setProperties(fallBackProps);
        }
        add(guiOptions);

        // Create the main graphics panel as either OpenGL or Java2D.
        if (GlGrPanel.openGlIsAvailable() && m_bUsingOpenGL) {
//        	m_pageProxy = (JPanel)new GlGrPanel();
            m_pageProxy = new GlGrPanel();
        } else {
            // Didn't or couldn't create an OpenGL panel, so use Java2D...
            m_bUsingOpenGL = false;
            m_pageProxy = new J2dGrPanel();
        }

        // Hook this controller to the panel.
        m_pageProxy.setController(this);

        // Set the size of the panel to be 70% of the width and height of
        // the screen.
        Component component = m_pageProxy.getComponent();
        component.setSize((int) (d.width * 0.7), (int) (d.height * 0.7));

        // Create the menu bar and status bar GUI components.
        m_grMenuBar = new GrMenuBar(frame, this);
        m_littleStatusBar = new GrStatusBar(this);
        m_bigStatusBar = new CurveStatusBar(this);
        setStatusBar(m_littleStatusBar);
        m_grStatusBar.modeLabel.setText("Zoom");

        // Add the status bar to the bottom of the frame.
        frame.getContentPane().setLayout(new BorderLayout());
        frame.getContentPane().add(m_grStatusBar, "South");

        // Create the button bar and add it to the left side of the frame.
        JPanel westPanel = m_grMenuBar.getButtonBar();
        JPanel westParentPanel = new JPanel();
        westParentPanel.setLayout(new BorderLayout());
        westParentPanel.add(westPanel, "North");
        JScrollPane buttonScrollPane = new JScrollPane();
        buttonScrollPane.setViewportView(westParentPanel);

        frame.getContentPane().add(buttonScrollPane, "West");

        // Add the main graphics panel to the center of the frame.
        frame.getContentPane().add(component, "Center");

        // Show olympic rings to represent an empty graphing area if the
        // spirit is right.
//        frame.getContentPane().add(new olympicCvs(150, 150, null, null),
//        "Center");
        frame.setJMenuBar(m_grMenuBar);

        // Set up the close operation, set the frame size, and show the window.
        frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        frame.setSize((int) (d.width * 0.7), (int) (d.height * 0.7));

        final JFrame frame_final = frame;
        javax.swing.SwingUtilities.invokeLater(new Runnable() {

            public void run() {
                frame_final.setVisible(true);
            }
        });
    }

//------------------------------------------------------------------------------
    /**
     * Shows a dialog box for the user to pick one or more file names, then
     * passes them to the openFile(File) method.
     */
    public void openFile() {

        // Ask the user for the file name or names.
        File[] files = showFileDialog(
                "Open File", true, m_currentFile);

        // Open each file that was specified.
        for (int i = 0; i < files.length; i++) {
            openFile(files[i]);
        }
    }

//------------------------------------------------------------------------------
    /**
     * Opens the file that is passed in.  If the file type can't be determined
     * by the extension (GS, GRS, GRT, or XML) the user is asked to confirm the
     * format.  Once the format is determined, the file is read and the contents
     * are added as data or graphics.
     * @param file  The file to open.
     */
    public void openFile(File file) {

        // Return if the file isn't valid.
        if (file == null) {
            return;
        }

        // Check for a configuration script in the current directory.
        String fileName = file.getAbsolutePath();
        fileName.replace('\\', '/');

        System.setProperty("gov.usgs.gr.currentfile", fileName);
        String currentPath = file.getAbsoluteFile().getParent();
        File configFile = new File(currentPath + "/gr_config.grs");
        if (configFile.canRead()) {
            openScript(configFile);
        }

        // Check if the file name ends with a known extension.
        boolean bShowDialog = true;
        boolean bOverlay = false;
        String bestGuessFormat = null;
        String upperName = file.getAbsolutePath().toUpperCase();
        if (upperName.endsWith(".GS")) {
            bestGuessFormat = "gov.usgs.sfhydro.data.GsDataFile";
            bShowDialog = false;
            bOverlay = true;
        }
        if (upperName.endsWith(".GRS")) {
            bestGuessFormat = "gov.usgs.gr.ScriptFile99";
            bShowDialog = false;
        }
        if (upperName.endsWith(".GRT")) {
            bestGuessFormat = "gov.usgs.gr.TemplateFile";
            bShowDialog = false;
        }
        if (upperName.endsWith(".XML")) {
            bestGuessFormat = "gov.usgs.gr.data.XMLProjectFile";
            bShowDialog = false;
        }
        if (upperName.endsWith(".VPV")) {
            bestGuessFormat = "gov.usgs.sfhydro.data.VPVDataFile";
            bShowDialog = false;
        }
        if (upperName.endsWith(".GIF") || upperName.endsWith(".PNG") ||
                upperName.endsWith(".JPG") || upperName.endsWith(".JPEG")) {
            bestGuessFormat = "gov.usgs.gr.ImageFile";
            bShowDialog = false;
        }
        if (bestGuessFormat == null) {
            bestGuessFormat = identifyFormat(file);

            if (bestGuessFormat != null) {
                // Handle special cases...
                if (bestGuessFormat.equals(
                        "gov.usgs.sfhydro.data.ListeningStationDataFile")) {

                    bOverlay = true;
                }

                // Determine if the class has an options dialog.
                try {
                    // Retrieve the hasOptionsDialog() method by reflection.
                    Class newClass = Class.forName(bestGuessFormat);
                    Class grFileClass = Class.forName("gov.usgs.gr.GrFile");
                    if (grFileClass.isAssignableFrom(newClass)) {
                        GrFile grFile = (GrFile) newClass.newInstance();
                        if (!grFile.hasOptionsDialog()) {
                            bShowDialog = false;
                            bOverlay = true;
                        }
                    }
                } catch (Exception e2) {
                }
            }
        }
// Not implemented yet...
        //if (upperName.endsWith(".ZIP"))
        //	bestGuessFormat = "gov.usgs.gr.CompressedProjectFile";
        if (bestGuessFormat == null && upperName.endsWith(".TXT")) {
            bestGuessFormat = m_dataFormatClass;
        }
        if (bestGuessFormat == null) {
            bestGuessFormat = m_dataFormatClass;
        }

        // Create and show a confirmation dialog which lets the user verify
        // the file format and whether it should be overlayed instead of
        // creating a new file.  Additional options may also be chosen for
        // certain formats.
        Frame frame = getParentFrame(m_pageProxy.getComponent());
        final OpenConfirmationDialog confirmDlg =
                new OpenConfirmationDialog(frame, "File Options", true);

        // Show the dialog if the format was not previously determined.
        if (bShowDialog) {
            confirmDlg.initialize(bestGuessFormat, bOverlay);

            if (EventQueue.isDispatchThread()) {
                confirmDlg.setVisible(true);
            } else {
                try {
                    javax.swing.SwingUtilities.invokeAndWait(new Runnable() {

                        public void run() {
                            confirmDlg.setVisible(true);
                        }
                    });
                } catch (InvocationTargetException ex) {
                } catch (InterruptedException ex) {
                }
            }

            if (!(confirmDlg.getOkPressed())) {
                return;
            }
            bOverlay = confirmDlg.getOverlay();

            // Get the best guess at the format from the dialog.
            bestGuessFormat = confirmDlg.getFileType();
        }

        // If the file was a script file, read the script and return.
        if (bestGuessFormat.equals("gov.usgs.gr.ScriptFile99")) {
            openScript(file);
            m_pageProxy.repaint();
            return;

        }

        // Set the current data format to the best guess.
        m_dataFormatClass = bestGuessFormat;

        // Check if the chosen format is derived from GrFile.
        Class grFileClass = null;
        Class formatClass = null;
        try {
            grFileClass = Class.forName("gov.usgs.gr.GrFile");
            formatClass = Class.forName(bestGuessFormat);
        } catch (ClassNotFoundException e) {
            return;
        }

        // If the file is not a type of GrFile, read it and return.
        if (!(grFileClass.isAssignableFrom(formatClass))) {
            // If overlay was selected, overay.  Otherwise, open on a new page.
            if (bOverlay) {
                overlayDataFile(file.getAbsolutePath());
            } else {
                openNewDataFile(file.getAbsolutePath());
            }

            // If the page title is blank, set it to the file name.
            updateWindowTitle(file.getName());

            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // When creating the new GrFile, get the properties from the template
        // GrFile that was created in the confirmation dialog.  The template
        // properties reflect the options chosen for this file by the user.
        GrFile grFileTemplate = confirmDlg.getGrFile();
        GrFile grFile = getGrFile(file.getAbsolutePath());
        if (grFileTemplate != null) {
            GrHashList props = grFileTemplate.getProperties();
            props.remove("file");
            grFile.setProperties(props);
        }

        // Read the GrFile.
        try {
            grFile.read();
        } catch (IOException e) {
            e.printStackTrace();

            JOptionPane.showMessageDialog(getParentFrame(
                    m_pageProxy.getComponent()),
                    "File " + file.getAbsolutePath() + " could not be loaded.",
                    "IO Exception", JOptionPane.ERROR_MESSAGE);
            return;
        }

        // Add the file as a child of the controller.
        add(grFile);

        // Declare new local variables.
        boolean bBareSeries = false;
        Enumeration enumeration;

        // Children of the file are usually added unless the file is a
        // template.
        if (m_bAddChildrenWhenOpening) {
            // Add certain children of the file to this controller.
            enumeration = grFile.children();
            while (enumeration.hasMoreElements()) {
                GrObject gro = (GrObject) enumeration.nextElement();
                while (gro instanceof ObjectEffigy) {
                    gro = ((ObjectEffigy) gro).getSubject();
                }

                // Child is a page...
                if (gro instanceof PageOfGraphs) {
                    // Add it as a child to this controller.
                    add(gro);

                    // If the page title is blank, set it equal to the
                    // file name.
                    PageOfGraphs page = (PageOfGraphs) gro;
                    String pageTitle = page.getTitle();
                    if (pageTitle == null || pageTitle.equals("")) {
                        pageTitle = file.getName();
                        page.setTitle(pageTitle);
                    }

                    // Set the page as the current page and log the open
                    // command
                    setPage(page);
                    logCommand("Open " + file.getAbsolutePath());


                    // Look for a TimeMarker object as a child of the page.
                    boolean bTime = false;
                    Enumeration pageChildren = page.children();
                    while (pageChildren.hasMoreElements()) {
                        GrObject pcGro = (GrObject) pageChildren.nextElement();
                        if (pcGro instanceof gov.usgs.gr.tools.TimeMarker) {
                            bTime = true;
                            break;
                        }
                    }

                    if (bTime) {
                        usePanMode(true);
                    }
                }

                // Child is a GUI Options object...
                if (gro instanceof GuiOptions) {
                    add(gro);
                }

                // Child is a data series...
                if (gro instanceof DataSeries) {
                    bBareSeries = true;
                }
            }
        }

        // Put all the bare series on graphs within a page.
        if (bBareSeries) {
            addData(grFile, bOverlay);

            // Command should be logged before reformatting in case
            // formatter displays the information.
            if (bOverlay) {
                logCommand("Overlay " + file.getAbsolutePath());
            } else {
                logCommand("Open " + file.getAbsolutePath());
            }
        }

        // Change the title of the window to the name of the file.
        updateWindowTitle(file.getName());

        // Update the curve list on the status bar.
        if (m_bigStatusBar != null) {
            m_bigStatusBar.updateCurveList();
        }

        // Trigger a repaint of the updated page contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     *
     */
    public String identifyFormat(File file) {


        BufferedReader br;
        String[] lines = new String[100];

        try {
            br = new BufferedReader(new FileReader(file.getAbsolutePath()));

            for (int i = 0; i < 100; i++) {
                lines[i] = br.readLine();
                if (lines[i] == null) {
                    break;
                }
            }
        } catch (Exception ex) {
            return null;
        }

        if (gov.usgs.sfhydro.data.ListeningStationDataFile.matchesFormat(lines)) {
            return "gov.usgs.sfhydro.data.ListeningStationDataFile";
        }

        if (gov.usgs.gr.data.TimeDataFile.matchesFormat(lines)) {
            return "gov.usgs.gr.data.TimeDataFile";
        }

        if (gov.usgs.gr.nwis.adaps.rdb.RdbFile.matchesFormat(lines)) {
            return "gov.usgs.gr.nwis.adaps.rdb.RdbFile";
        }

        if (gov.usgs.sfhydro.data.CR10DataFile.matchesFormat(lines)) {
            return "gov.usgs.sfhydro.data.CR10DataFile";
        }

        if (gov.usgs.sfhydro.data.CR1000DataFile.matchesFormat(lines)) {
            return "gov.usgs.sfhydro.data.CR1000DataFile";
        }

        if (gov.usgs.sfhydro.data.ArcGridAsciiFile.matchesFormat(lines)) {
            return "gov.usgs.sfhydro.data.ArcGridAsciiFile";
        }

        if (gov.usgs.sfhydro.data.Noaa2DataFile.matchesFormat(lines)) {
            return "gov.usgs.sfhydro.data.Noaa2DataFile";
        }

        if (gov.usgs.crrl.data.ParticleDataFile.matchesFormat(lines)) {
            return "gov.usgs.crrl.data.ParticleDataFile";
        }

        return null;
    }

//------------------------------------------------------------------------------
    /**
     *
     */
    public void addData(GrObject dataList, boolean bOverlay) {

        PageOfGraphs page = null;

        // Designate a page for receiving the data.  If there is an existing
        // page and the user chose to overlay data, use that page.
        // Otherwise, create a new page.
        if (bOverlay && m_page != null) {
            page = m_page;

            TimeMarker tm = (TimeMarker) page.firstChildInstanceOf(
                    "gov.usgs.gr.tools.TimeMarker", false);
            Graph mapGraph = null;
            List graphList = page.getGraphList();
            if (tm != null && graphList != null &&
                    graphList.size() > 0) {

                mapGraph = (Graph) graphList.get(graphList.size() - 1);
                GrHashList props = mapGraph.getProperties();
                GrProperty p = (GrProperty) props.get("aspect1to1");
                if (((Boolean) p.value) == false) {
                    mapGraph = null;
                }

                // Don't handle as animation unless a DoubleDataSeries is
                // in the data list.
                boolean bFound = false;
                Enumeration enum1 = dataList.children();
                while (enum1.hasMoreElements()) {
                    GrObject gro = (GrObject) enum1.nextElement();
                    while (gro instanceof gov.usgs.gr.ObjectEffigy) {
                        gro = ((ObjectEffigy) gro).getSubject();
                    }
                    if (gro instanceof DoubleDataSeries) {
                        DoubleDataSeries series = (DoubleDataSeries) gro;
                        int numDims = series.getNumDimensions();
                        if (numDims == 2) {
                            bFound = true;
                            break;
                        }
                    }
                }
                if (!bFound) {
                    mapGraph = null;
                }
            }
            if (mapGraph != null) {
                Frame frame = getParentFrame(m_pageProxy.getComponent());
                int val = JOptionPane.showConfirmDialog(frame, "Should this "
                        + "data be opened as a vector animation instead of a "
                        + "regular curve?", "Open as Vector?",
                        JOptionPane.YES_NO_CANCEL_OPTION);
                if (val == JOptionPane.CANCEL_OPTION) {
                    return;
                }
                if (val == JOptionPane.YES_OPTION) {
                    // Get a list of all TimeSeries objects on the page.
                    TimeSeries tSeries = (TimeSeries) m_page.
                            firstChildInstanceOf("gov.usgs.gr.data.TimeSeries",
                            true);

                    int refYear = -1;
                    if (tSeries != null) {
                        refYear = tSeries.getReferenceYear();
                    }

                    // Handle each element of the data list.
                    Enumeration childEnum = dataList.children();
                    while (childEnum.hasMoreElements()) {
                        GrObject gro = (GrObject) childEnum.nextElement();
                        while (gro instanceof ObjectEffigy) {
                            gro = ((ObjectEffigy) gro).getSubject();
                        }

                        // Sync the reference year, if it is a TimeSeries
                        if (refYear > 0 && gro instanceof TimeSeries) {
                            ((TimeSeries) gro).syncToReferenceYear(refYear);
                        }

                        if (gro instanceof DoubleDataSeries) {
                            DoubleDataSeries series = (DoubleDataSeries) gro;
                            tm.add(series);
                            GlVectorAnimation vec = new GlVectorAnimation();
                            vec.add(series);
                            vec.setProperty("symbolType", "Two Tone Arrow");
                            vec.setProperty("scale", "0.01");

                            List vecList = mapGraph.findChildInstancesOf(
                                    "gov.usgs.gr.visual.gl.GlVectorAnimation",
                                    false);
                            if (vecList.size() > 0) {
                                vec.setProperties(((GlVectorAnimation)vecList.
                                    get(vecList.size() - 1)).getProperties());
                            }

                            vec.setProperty("seriesNums",
                                    "1, 0, 0, 0, 0, 1, 0, 0");
                            vec.setProperty("dimNums",
                                    "0, =5, =10, -1, =30, 1, -1, -1");
                            mapGraph.add(vec);
                        }
                    }

                    return;
                }
            }

            Class cls = null;
            try {
                cls = Class.forName("gov.usgs.gr.data.DataSeries");
            } catch (Exception e) {
            }
            Enumeration childEnum = dataList.children();

            // Sync reference years if there are time series in the new file
            // and on the current page.
            int refYear = -1;

            // Find the reference year for time series on the current page
            // and sync the new series to it.
            if (m_page != null && cls != null) {
                // Get a list of all TimeSeries objects on the page.
                TimeSeries tSeries = (TimeSeries) m_page.firstChildInstanceOf(
                        "gov.usgs.gr.data.TimeSeries", true);

                if (tSeries != null) {
                    refYear = tSeries.getReferenceYear();
                }

                if (refYear > 0) {
                    // Sync all new time series to existing reference years
                    // of the time series on the page.
                    while (childEnum.hasMoreElements()) {
                        GrObject gro = (GrObject) childEnum.nextElement();
                        while (gro instanceof gov.usgs.gr.ObjectEffigy) {
                            gro = ((ObjectEffigy) gro).getSubject();
                        }
                        if (gro instanceof TimeSeries) {
                            ((TimeSeries) gro).syncToReferenceYear(
                                    refYear);
                        }
                    }
                }
            }
        } else {
            page = new PageOfGlGraphs();
        }

        // Remember how many graphs are on the page before adding this data.
        int oldNumGraphs = page.getShowList().size();

        // Keep track of whether the page is empty.
        boolean bPageEmpty = true;

        // Add each series from the file to the page.  The page will handle
        // putting the series within graphs.
        Enumeration childEnum = dataList.children();
        while (childEnum.hasMoreElements()) {
            GrObject gro = (GrObject) childEnum.nextElement();
            while (gro instanceof ObjectEffigy) {
                gro = ((ObjectEffigy) gro).getSubject();
            }

            if (gro instanceof DoubleDataSeries) {
                page.add((DoubleDataSeries) gro);
                bPageEmpty = false;
            }
        }

        // Don't add the page if there was nothing added to it.
        if (bPageEmpty) {
            return;
        }

        // If the page was just created, add it to this conroller.
        if (!bOverlay || m_page == null) {
            add(page);
            setPage(page);
        }

        // Have the page's formatter reinitialize so the new graphs will
        // be in the right places.
        m_page.getFormatter().initialize();

        // If the page was empty before adding the data, select all the
        // graphs and zoom to show all the data.
        if (oldNumGraphs == 0) {
            m_page.selectAllGraphs();
            zoomAll();
        }
    }

//------------------------------------------------------------------------------
    /**
     * Creates an animation so the page has a TimeMarker, time graph with
     * LineMarkup, and a map graph with 1 to 1 aspect ratio.
     */
    public void createAnimation() {

        // Create a page if necessary..
        if (m_page == null) {
            newPage();
        }
        if (m_page == null) {
            return;
        }

        m_page.deferUpdates();

        // Look for a VariableSizeFormatter
        VariableSizeFormatter vsf =
                (VariableSizeFormatter) m_page.firstChildInstanceOf(
                "gov.usgs.sfhydro.gr.formats.VariableSizeFormatter", false);
        if (vsf == null) {
            m_page.setProperty("formatter",
                    "gov.usgs.sfhydro.gr.formats.VariableSizeFormatter");
        }

        // Look for a TimeMarker object as a child of the current page.
        TimeMarker tm = (TimeMarker) m_page.firstChildInstanceOf(
                "gov.usgs.gr.tools.TimeMarker", false);
        if (tm == null) {
            tm = new TimeMarker();
            tm.setProperty("sleep", "45");
            m_page.add(tm);
        }

        // Look for a map graph and add one if it isn't already there.
        boolean bAddMapGraph = true;
        Graph mapGraph = null;
        List<Graph> graphList = m_page.getGraphList();
        if (graphList.size() > 0) {
            Graph graph = (Graph) graphList.get(graphList.size() - 1);
            GrHashList props = graph.getProperties();
            GrProperty p = (GrProperty) props.get("aspect1to1");
            if (((Boolean) p.value) == true) {
                bAddMapGraph = false;
                mapGraph = graph;
            }
        }
        if (bAddMapGraph) {
            newGraph();
            m_page.updateLists();
            graphList = m_page.getGraphList();
            mapGraph = (Graph) graphList.get(graphList.size() - 1);
            mapGraph.setProperty("aspect1to1", "true");
            LineMarkup lineMarkup = (LineMarkup) mapGraph.firstChildInstanceOf(
                    "gov.usgs.gr.visual.LineMarkup", false);
            mapGraph.remove(lineMarkup);
            Zoom xZoom = mapGraph.getXZoom();
            Axis xAxis = mapGraph.getXAxis();
            xAxis.add(xZoom.createCopy(false));
            Axis yAxis = mapGraph.getYAxis();
            mapGraph.setTitle("Map");
            xAxis.setProperty("showMajorTicks", "false");
            yAxis.setProperty("showMajorTicks", "false");
            xAxis.setProperty("showMinorTicks", "false");
            yAxis.setProperty("showMinorTicks", "false");
            xAxis.setProperty("showLabels", "false");
            yAxis.setProperty("showLabels", "false");
            xAxis.setProperty("showGridLines", "false");
            yAxis.setProperty("showGridLines", "false");
        }

        // Look for a text markup object as a child of the map graph and
        // TimeMarker.
        boolean bAddTextMarkup = true;
        TextMarkup markup = (TextMarkup) tm.firstChildInstanceOf(
                "gov.usgs.gr.visual.TextMarkup", false);
        if (markup != null) {
            List graphChildren = mapGraph.getChildList();
            if (graphChildren.contains(markup)) {
                bAddTextMarkup = false;
            }
        }

        if (bAddTextMarkup) {
            TextMarkup text = new TextMarkup();
            text.setProperty("screenFont", "glutBitmapTimesRoman24");
            text.setProperty("xCoordLevel", "Frame Interior 0 to 1");
            text.setProperty("yCoordLevel", "Frame Interior 0 to 1");
            text.setProperty("x", "0.05");
            text.setProperty("y", "0.05");
            tm.add(text);
            mapGraph.add(text);
        }

        // Set up each of the time graphs to have a line markup and share the
        // X Zoom with the TimeMarker.
        graphList = m_page.getGraphList();
        List tmChildren = tm.getChildList();
        boolean bTimeGraphFound = false;
        for (Graph graph : graphList) {
            GrHashList props = graph.getProperties();
            GrProperty p = (GrProperty) props.get("aspect1to1");
            if (((Boolean) p.value) == false) {
                // Make sure the X zoom matches that in the time marker.
                Zoom xZoom = graph.getXZoom();
                if (tmChildren.contains(xZoom)) {
                    tm.add(xZoom);
                }
                LineMarkup lineMarkup = (LineMarkup) graph.firstChildInstanceOf(
                        "gov.usgs.gr.visual.LineMarkup", false);
                if (lineMarkup == null) {
                    graph.add(new LineMarkup());
                }

                bTimeGraphFound = true;
            }
        }

        // Make sure there is at least one time graph.
        if (!bTimeGraphFound) {
            newGraph();
        }

        // Unshare the x zoom of the map graph.
        mapGraph.setXZoom((Zoom) mapGraph.getXZoom().createCopy(true));

        m_page.processDeferredUpdates();
        setPage(m_page);
    }

//------------------------------------------------------------------------------
    /**
     * Takes the page at the top of the list and pushes it to the bottom.  Sets
     * the current page to the new top page.
     */
    public void pushTop() {

        // Initialize local variables.
        int firstPageIndex = -1;
        int lastPageIndex = -1;
        int newTopIndex = -1;
        int index = 0;

        // Look through the children and find the index of the first and
        // last page and the index of the page that should be shown next.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject) iter.next();
            if (gro instanceof PageOfGraphs) {
                if (firstPageIndex < 0) {
                    firstPageIndex = index;
                }
                newTopIndex = lastPageIndex;
                lastPageIndex = index;
            }
            index++;
        }

        // If there aren't at least two pages, don't bother.
        if (lastPageIndex < 2) {
            return;
        }

        // Move the current page from the end of the list to the beginning.
        GrObject gro = m_children.remove(lastPageIndex);
        m_children.add(firstPageIndex, gro);

        // Set the new page.
        setPage((PageOfGraphs) m_children.get(newTopIndex));

        // Update the curve list on the status bar.
        m_bigStatusBar.updateCurveList();

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Takes the page at the bottom of the list and places it at the top.  Sets
     * the current page to the new top page.
     */
    public void popBottom() {

        // Initialize local variables.
        int firstPageIndex = -1;
        int lastPageIndex = -1;
        int index = 0;

        // Look through the children and find the index of the first and
        // last page and the index of the page that should be shown next.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject) iter.next();
            if (gro instanceof PageOfGraphs) {
                if (firstPageIndex < 0) {
                    firstPageIndex = index;
                }
                lastPageIndex = index;
            }
            index++;
        }

        // If there aren't at least two pages, don't bother.
        if (lastPageIndex < 2) {
            return;
        }

        // Move the current page from the end of the list to the beginning.
        GrObject gro = m_children.remove(firstPageIndex);
        m_children.add(lastPageIndex, gro);

        // Set the new page.
        setPage((PageOfGraphs) gro);

        // Update the curve list on the status bar.
        m_bigStatusBar.updateCurveList();

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Overrides clear in GrObject.  Sets the current page to null.
     */
    public void clear() {

        super.clear();

        setPage(null);
    }

//------------------------------------------------------------------------------
    /**
     * Reassigns the current page, in case it was removed.  Called by remove()
     * @param gro The child that was removed.
     */
    private void checkForPageRemoval(GrObject gro) {

        // Initialize local variables.
        PageOfGraphs newPage = null;
        GrFile associatedFile = null;

        // Reassign m_page or clear it if there are no others
        if (gro == m_page) {
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof PageOfGraphs) {
                    newPage = (PageOfGraphs) o;
                }
                // Find any file objects whose only child is the page
                // that is being removed.
                if (o instanceof GrFile) {
                    Enumeration enumeration = ((GrFile) o).children();
                    if (enumeration.hasMoreElements()) {
                        GrObject fileChild = (GrObject)
                                enumeration.nextElement();

                        while (fileChild instanceof gov.usgs.gr.ObjectEffigy) {
                            fileChild = ((ObjectEffigy) fileChild).getSubject();
                        }
                        if (fileChild == gro &&
                                !(enumeration.hasMoreElements())) {

                            associatedFile = (GrFile) o;
                        }
                    }
                }
            }
            // Reassign to the next page, or null if there was none.
            setPage(newPage);
        }

        // Remove the old page's associated file, if there was one.
        if (associatedFile != null) {
            super.remove(associatedFile);
        }

        // Update the curve list on the status bar.
        m_bigStatusBar.updateCurveList();
    }

//------------------------------------------------------------------------------
    /**
     * Open the specified file and add the data to the graph.  Works with files
     * that are derived from DimensionalDataFile rather than GrFile.
     * @param filename  The name of the file to open.
     */
    protected void openNewDataFile(String filename) {

        //Check if the file is actually a script file with a .grs extension.
        if (filename.toUpperCase().endsWith(".GRS")) {
            openScript(new File(filename));
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Load the specified file.
        XyDataFile dataFile = getDataFile(filename);
        List dataSeriesList = null;

        try {
            dataSeriesList = dataFile.load(filename);
        } catch (IOException e) {
            e.printStackTrace();

            // Return to the default arrow cursor.
            ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                    Cursor.getDefaultCursor());

            JOptionPane.showMessageDialog(getParentFrame(
                    m_pageProxy.getComponent()),
                    "File " + filename + " could not be loaded.", "IO Exception",
                    JOptionPane.ERROR_MESSAGE);
            return;
        }

        // If data were read, add it to the graph.
        if (dataSeriesList != null) {

            // Lock the X dimension of each data series.
            Iterator iter = dataSeriesList.iterator();
            while (iter.hasNext()) {
                DoubleDataSeries series = (DoubleDataSeries) iter.next();
                series.setDimensionLocked(0, true);
            }

            // Command should be logged before reformatting in case formatter
            // displays the information.
            logCommand("Open " + m_currentFile.getAbsolutePath());

            m_dataFile = dataFile;
            clearPage();
            // Create and add a new blank page if there is currently none.
            if (m_page == null) {
                m_page = m_pageProxy.setNewPage();
                add(m_page);
            }
            m_page.addData(dataSeriesList);
            String pageTitle = m_dataFile.getTitle();
            m_page.setTitle(pageTitle);
            m_page.getFormatter().initialize();
            m_page.selectAllGraphs();
            zoomAll();

            // Change the title of the window to the name of the file.
            updateWindowTitle(pageTitle);

            // Update the curve list on the status bar.
            m_bigStatusBar.updateCurveList();

            // Trigger a repaint of the updated contents.
            m_pageProxy.repaint();
        }
    }

//------------------------------------------------------------------------------
    /**
     * Clears the contents of the current page, as well as the current undo and
     * redo lists.  Resets the X and Y axis locking variables to unlocked.
     */
    public void clearPage() {

        // Return if there is no page open.
        if (m_page == null) {
            return;
        }

        // Clear the page and undo/redo lists.
        m_page.clear();
        m_undoList.clear();
        m_redoList.clear();

        // Turn X and Y axis locking off.
        m_bLockX = false;
        m_bLockY = false;
    }

//------------------------------------------------------------------------------
    /**
     * Gets a XyDataFile object that can be used to read the specified file in
     * the format determined by the m_dataFormatClass variable.  Checks for a
     * gr_config.grs file in the same directory and reads it if found.
     * @param filename  The name of the file to open.
     * @return  An object of type XyDataFile or a derived class.
     */
    protected XyDataFile getDataFile(String filename) {

        // Find the current window frame for displaying dialogs, if needed.
        Frame frame = getParentFrame(m_pageProxy.getComponent());

        // Find the data file path and name.
        File file = new File(filename);
        m_currentPath = file.getAbsoluteFile().getParent();
        m_currentFile = new File(filename);

        // Check for a configuration script in the current directory.
        File configFile = new File(m_currentPath + "/gr_config.grs");
        if (configFile.canRead()) {
            openScript(configFile);
        }

        // Initialize local variables.
        XyDataFile dataFile = null;
        Object o = null;
        Class newClass = null;

        // Find the class that reads the data format.
        // Show error dialog boxes if there was a problem...
        try {
            newClass = Class.forName(m_dataFormatClass);
        } catch (ClassNotFoundException e2) {
            e2.printStackTrace();
            JOptionPane.showMessageDialog(getParentFrame(
                    m_pageProxy.getComponent()),
                    "Could not find " + m_dataFormatClass
                    + ".  File could not be loaded.",
                    "Class Not Found Exception",
                    JOptionPane.ERROR_MESSAGE);
            return null;
        }

        // Try making an instance of the data format class.
        try {
            o = newClass.newInstance();
        } catch (InstantiationException e) {
            e.printStackTrace();
            JOptionPane.showMessageDialog(getParentFrame(
                    m_pageProxy.getComponent()),
                    "Could not instantiate " + m_dataFormatClass +
                    " to read data.  File could not be loaded.",
                    "Instantiation Exception",
                    JOptionPane.ERROR_MESSAGE);
            return null;
        } catch (IllegalAccessException e3) {
            e3.printStackTrace();
            JOptionPane.showMessageDialog(getParentFrame(
                    m_pageProxy.getComponent()),
                    "Format class " + m_dataFormatClass
                    + " is not publicly accessible.  File could not be loaded.",
                    "Illegal Access Exception",
                    JOptionPane.ERROR_MESSAGE);
            return null;
        }

        // Assign the new instance to the return variable if it is of the
        // proper type.
        if (o instanceof gov.usgs.gr.data.XyDataFile) {
            dataFile = (XyDataFile) o;
        } else {
            JOptionPane.showMessageDialog(getParentFrame(
                    m_pageProxy.getComponent()),
                    "Format class " + m_dataFormatClass
                    + " is not an XyDataFile.  File could not be loaded.",
                    "Improper Format Class",
                    JOptionPane.ERROR_MESSAGE);
            return null;
        }

        return dataFile;
    }

//------------------------------------------------------------------------------
    /**
     * Gets a GrFile object that can be used to read the specified file in
     * the format determined by the m_dataFormatClass variable.
     * @param filename  The name of the file to open.
     * @return  An object of type GrFile or a derived class.
     */
    protected GrFile getGrFile(String filename) {

        // Find the current window frame for displaying dialogs, if needed.
        Frame frame = getParentFrame(m_pageProxy.getComponent());

        // Find the data file path and name.
        File file = new File(filename);
        m_currentPath = file.getAbsoluteFile().getParent();
        m_currentFile = new File(filename);

        // Create an instance of the format class.
        GrFile grFile = (GrFile) createObject(m_dataFormatClass,
                "gov.usgs.GrFile", frame);

        // If the file object was successfully created, set its file name
        // property.
        if (grFile != null) {
            grFile.setProperty("file", filename);
        }

        return grFile;
    }

//------------------------------------------------------------------------------
    /**
     * Creates an object from the given class name.  Shows an error message
     * dialog if the object is not derived from the given super class name.
     * @param className The name of the class to instantiate.
     * @param superName The name of the super class for checking that the new
     *                  object will be properly assignable.
     * @param frame     The frame to which error dialogs should be attached.
     * @return  The newly created object, or null if either the class name could
     *          not be found and instantiated, or it was not assignable from the
     *          specified super class.
     */
    public static Object createObject(String className, String superName,
            java.awt.Frame frame) {

        // Initialize local variables.
        Object o = null;
        Class newClass = null;

        // Find the requested class.
        // Show error dialog boxes if there was a problem...
        try {
            newClass = Class.forName(className);
        } catch (ClassNotFoundException e2) {
            e2.printStackTrace();
            JOptionPane.showMessageDialog(frame,
                    "Could not find " + className
                    + ".  Operation could not be completed.",
                    "Class Not Found Exception",
                    JOptionPane.ERROR_MESSAGE);
            return null;
        }

        // Check if it is descended from class superName.
        try {
            if (superName != null) {
                if (!(Class.forName(superName).isAssignableFrom(newClass))) {
                    JOptionPane.showMessageDialog(frame,
                            "Class " + className + " is not a type of "
                            + superName + ".  Operation could not be completed.",
                            "Wrong Class Type",
                            JOptionPane.ERROR_MESSAGE);
                    return null;
                }
            }
        } catch (ClassNotFoundException cnfe) {
        }

        // Instantiate the class.
        try {
            o = newClass.newInstance();
        } catch (InstantiationException e) {
            e.printStackTrace();
            JOptionPane.showMessageDialog(frame,
                    "Could not instantiate " + className
                    + ".  Operation could not be completed.",
                    "Instantiation Exception",
                    JOptionPane.ERROR_MESSAGE);
            return null;
        } catch (IllegalAccessException e3) {
            e3.printStackTrace();
            JOptionPane.showMessageDialog(frame,
                    "Class " + className + " is not publicly accessible.  "
                    + "Operation could not be completed.",
                    "Illegal Access Exception",
                    JOptionPane.ERROR_MESSAGE);
            return null;
        }

        return o;
    }

//------------------------------------------------------------------------------
    /**
     * Open the named file and add the data to the graphs on the current page.
     * Creates graphs if there are none.  Creates a new page if there is none.
     * @param filename  The name of the file to open.
     */
    public void overlayDataFile(String filename) {

        //Check if the file is actually a script file with a .grs extension.
        if (filename.toUpperCase().indexOf(".GRS") == filename.length() - 4) {
            openScript(new File(filename));
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Save the page so changes can be undone later.
        savePageSnapShot();

        // Start with a blank page data storage object.
        XyDataFile dataFile = getDataFile(filename);
        List dataSeriesList = null;

        // Load the specified file.
        try {
            dataSeriesList = dataFile.load(filename);
        } catch (IOException e) {
            e.printStackTrace();

            // Return to the default arrow cursor.
            ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                    Cursor.getDefaultCursor());

            // Show a dialog box with an error message.
            JOptionPane.showMessageDialog(getParentFrame(
                    m_pageProxy.getComponent()),
                    "File " + filename + " could not be loaded.",
                    "IO Exception",
                    JOptionPane.ERROR_MESSAGE);
            return;
        }

        // Make this the data file to save, if it is GS format.
        if (dataFile instanceof GsDataFile) {
            m_dataFile = dataFile;
            m_currentFile = new File(filename);

            // Change the title of the window to the name of the file.
            updateWindowTitle(m_currentFile.getName());
        }

        // If the data are time referenced, set up the reference year.
        if (dataFile instanceof TimeSeriesDataFile &&
                m_dataFile instanceof TimeSeriesDataFile) {

            ((TimeSeriesDataFile) dataFile).syncToReferenceYear(
                    ((TimeSeriesDataFile) m_dataFile).getReferenceYear());
        }

        // If data were read, add them to the graph.
        if (dataSeriesList != null) {
            // Lock the X dimension of each data series.
            Iterator iter = dataSeriesList.iterator();
            while (iter.hasNext()) {
                DoubleDataSeries series = (DoubleDataSeries) iter.next();
                series.setDimensionLocked(0, true);
            }

            // Create and add a new blank page if there is currently none.
            if (m_page == null) {
                m_page = m_pageProxy.setNewPage();
                add(m_page);
            }

            // If the page doesn't have a title, set it to match the file name
            updateWindowTitle(m_currentFile.getName());

            int oldNumGraphs = m_page.getGraphList().size();


            TimeMarker tm = (TimeMarker) m_page.firstChildInstanceOf(
                    "gov.usgs.gr.tools.TimeMarker", false);
            Graph mapGraph = null;
            List graphList = m_page.getGraphList();
            if (tm != null && graphList != null &&
                    graphList.size() > 0) {

                mapGraph = (Graph) graphList.get(graphList.size() - 1);
                GrHashList props = mapGraph.getProperties();
                GrProperty p = (GrProperty) props.get("aspect1to1");
                if (((Boolean) p.value) == false) {
                    mapGraph = null;
                }

                // Don't handle as animation unless a DoubleDataSeries is
                // in the data list.
                boolean bFound = false;
                Iterator iter1 = dataSeriesList.iterator();
                while (iter1.hasNext()) {
                    GrObject gro = (GrObject) iter1.next();
                    while (gro instanceof gov.usgs.gr.ObjectEffigy) {
                        gro = ((ObjectEffigy) gro).getSubject();
                    }
                    if (gro instanceof DoubleDataSeries) {
                        DoubleDataSeries series = (DoubleDataSeries) gro;
                        int numDims = series.getNumDimensions();
                        if (numDims == 2) {
                            bFound = true;
                            break;
                        }
                    }
                }
                if (!bFound) {
                    mapGraph = null;
                }
            }
            if (mapGraph != null) {
                Frame frame = getParentFrame(m_pageProxy.getComponent());
                int val = JOptionPane.showConfirmDialog(frame, "Should this "
                        + "data be opened as a vector animation instead of a "
                        + "regular curve?", "Open as Vector?",
                        JOptionPane.YES_NO_CANCEL_OPTION);
                if (val == JOptionPane.CANCEL_OPTION) {
                    return;
                }
                if (val == JOptionPane.YES_OPTION) {
                    // Get a list of all TimeSeries objects on the page.
                    TimeSeries tSeries = (TimeSeries) m_page.
                            firstChildInstanceOf("gov.usgs.gr.data.TimeSeries",
                            true);

                    int refYear = -1;
                    if (tSeries != null) {
                        refYear = tSeries.getReferenceYear();
                    }

                    // Handle each element of the data list.
                    Iterator childIter = dataSeriesList.iterator();
                    while (childIter.hasNext()) {
                        GrObject gro = (GrObject) childIter.next();
                        while (gro instanceof ObjectEffigy) {
                            gro = ((ObjectEffigy) gro).getSubject();
                        }

                        // Sync the reference year, if it is a TimeSeries
                        if (refYear > 0 && gro instanceof TimeSeries) {
                            ((TimeSeries) gro).syncToReferenceYear(refYear);
                        }

                        if (gro instanceof DoubleDataSeries) {
                            DoubleDataSeries series = (DoubleDataSeries) gro;
                            tm.add(series);
                            GlVectorAnimation vec = new GlVectorAnimation();
                            vec.add(series);
                            vec.setProperty("symbolType", "Two Tone Arrow");
                            vec.setProperty("scale", "0.01");

                            List vecList = mapGraph.findChildInstancesOf(
                                    "gov.usgs.gr.visual.gl.GlVectorAnimation",
                                    false);
                            if (vecList.size() > 0) {
                                vec.setProperties(((GlVectorAnimation)vecList.
                                    get(vecList.size() - 1)).getProperties());
                            }

                            vec.setProperty("seriesNums",
                                    "1, 0, 0, 0, 0, 1, 0, 0");
                            vec.setProperty("dimNums",
                                    "0, =5, =10, -1, =30, 1, -1, -1");
                            mapGraph.add(vec);
                        }
                    }
                } else {
                    // Let the page decide where to put the data series.
                    m_page.addData(dataSeriesList);
                }
            } else {
                // Let the page decide where to put the data series.
                m_page.addData(dataSeriesList);
            }

            m_page.getFormatter().initialize();

            // If there were previously no graphs on the page, select all of the
            // new graphs in case the user want to zoom all.
            if (oldNumGraphs == 0) {
                m_page.selectAllGraphs();
                zoomAll();
            }

            // Log this overlay operation in the command list.
            logCommand("Overlay " + filename);

            // Update the curve list on the status bar.
            m_bigStatusBar.updateCurveList();

            // Trigger a repaint of the updated contents.
            m_pageProxy.repaint();
        }
    }

//------------------------------------------------------------------------------
    /**
     * Ask the user for a file name or names and overlay it on the current page.
     */
    public void overlayDataFile() {

        // Ask the user for the filename(s).
        Frame frame = getParentFrame(m_pageProxy.getComponent());

        File[] files = showFileDialog("Overlay Data File", true,
                m_currentFile);

        // Overlay each of the files.
        for (int i = 0; i < files.length; i++) {
            overlayDataFile(files[i].getAbsolutePath());
        }
    }

//------------------------------------------------------------------------------
    /**
     * Wrapper for main showFileDialog method.  Attempts to find the frame using
     * getParentFrame().  Sets the m_currentPath member variable to match the
     * first selected file.
     * @param title The title of the dialog box.
     * @param bOpen Whether the dialog is for opening or saving files.
     * @param file  The default file to display when the dialog first opens.
     * @return An array of File objects that the user chose.
     */
    public File[] showFileDialog(String title, boolean bOpen,
            File file) {

        // Show the dialog and store the list of selected files.
        File[] files = showFileDialog(title, bOpen, file,
                getParentFrame(m_pageProxy.getComponent()));

        // Set the current path variable to match the first selected file.
        if (files.length > 0) {
            m_currentPath = files[0].getAbsoluteFile().getParent();
        }

        return files;
    }

//------------------------------------------------------------------------------
    /**
     * Shows an open-file dialog and lets the user choose files.  Multiple files
     * can be selected if opening, but only one if saving.
     * <p><b> Post Conditions </b><br>
     *      Check the variable m_currentPath for the path of the file.
     * @param title	The title of the dialog box.
     * @param bOpen Whether the dialog is for opening or saving files.
     * @param file	The default file to display when the dialog first opens.
     * @param frame The frame to which the dialog should be attached.
     * @return An array of File objects that the user chose.
     */
    public static File[] showFileDialog(String title, boolean bOpen,
            File file, Frame frame) {

        // Set up the file chooser dialog.
        JFileChooser fc = new JFileChooser();
        fc.setDialogTitle(title);
        fc.setCurrentDirectory(file);
        fc.setSelectedFile(file);

        // Show the open-file dialog and let the user choose the file.
        int returnVal = 0;
        if (bOpen) {
            // Allow multiple file selection if the files are to be opened.
            fc.setMultiSelectionEnabled(true);
            returnVal = fc.showOpenDialog(frame);
        } else {
            fc.setDialogType(JFileChooser.SAVE_DIALOG);

            // Let the user choose the file.
            returnVal = fc.showSaveDialog(frame);

        // If the file is not writeable, show a message, then try anyway.
// Was causing false warnings...
//            if (returnVal == JFileChooser.APPROVE_OPTION &&
//                ! (fc.getSelectedFile().canWrite())) {
//
//                JOptionPane.showMessageDialog(frame,
//                    "File may not be writeable.", "Write access problem",
//                    JOptionPane.WARNING_MESSAGE);
//            }
        }
        // Create an empty file array to return if the user canceled the dialog.
        File[] files = new File[0];

        // If the user pressed OK...
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            // Get the list of files.
            if (bOpen) {
                files = fc.getSelectedFiles();
            } else {
                files = new File[1];
                files[0] = fc.getSelectedFile();

                // If the file exists, check if the user wants to overwrite.
                if (files[0].exists()) {
                    int choice = JOptionPane.showConfirmDialog(frame,
                            "Replace the existing file?", "Confirm Overwrite",
                            JOptionPane.YES_NO_OPTION,
                            JOptionPane.QUESTION_MESSAGE);

                    if (choice == JOptionPane.NO_OPTION) {
                        files = new File[0];
                    }
                }
            }
        }

        return files;
    }

//------------------------------------------------------------------------------
    /**
     * Shows a dialog to the user to choose a filename and then save the current
     * data file to it.
     */
    public void saveToDataFile() {

        // Ask the user for the filename and save to that file.
        File[] files = showFileDialog("Save Data File", false,
                m_currentFile);

        // Quit if no file was selected.
        if (files.length <= 0) {
            return;
        }

        // Save to the selected file.
        File file = files[0];
        saveToDataFile(file.getAbsolutePath());
    }

//------------------------------------------------------------------------------
    /**
     * Save the current data file to the specfied file.
     * @param filename  The name of the file to write to.
     */
    public void saveToDataFile(String filename) {

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Write the data to the file.
        try {
            m_dataFile.save(filename);
        } catch (IOException e) {
            e.printStackTrace();

            // Return to the default arrow cursor.
            ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                    Cursor.getDefaultCursor());

            // If there was an error, show an error dialog.
            JOptionPane.showMessageDialog(
                    getParentFrame(m_pageProxy.getComponent()),
                    "File " + filename + " could not be saved.",
                    "IO Exception",
                    JOptionPane.ERROR_MESSAGE);
            return;
        }

        // Update the current path and filename variables.
        File file = new File(filename);
        m_currentPath = file.getAbsoluteFile().getParent();
        m_currentFile = file;

        // If the page doesn't have a title, set it to match the file name
        updateWindowTitle(m_currentFile.getName());

        // Return to the default arrow cursor.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                Cursor.getDefaultCursor());
    }

//------------------------------------------------------------------------------
    /**
     * Saves all curves on the current page that don't belong to any file to an
     * XyDataFile file called xydata_gr_out.txt in the current directory.
     */
    public void saveOrphanCurves() {

        // Make sure there is a current page before proceeding.
        if (m_page == null) {
            return;
        }

        // Create a new empty data file.
        XyDataFile dataFile = new XyDataFile();

        // Look through the graphs for curves whose data series don't belong
        // to the main output file.  Add these series to the above file object.
        synchronized (m_page) {
            List graphList = m_page.getGraphList();
            Iterator graphIter = m_page.getGraphList().iterator();
            while (graphIter.hasNext()) {
                // Look through each graph on the page.
                Graph graph = (Graph) graphIter.next();
                List curveList = graph.getCurves();
                Iterator curveIter = curveList.iterator();
                while (curveIter.hasNext()) {
                    // Look through each curve in the graph.
                    Curve curve = (Curve) curveIter.next();
                    DataSeries series = curve.getSeries();
                    // Check if the main data file contains this series.
                    if (!m_dataFile.containsSeries(series)) {
                        dataFile.addSeries(series);
                    }
                }
            }
        }

        // Save the orphans file.
        if (dataFile.numChannels() > 0) {
            try {
                // Write a status message.
                System.out.println("Saving orphans to " + m_currentPath
                        + "/xydata_gr_out.txt");
                dataFile.save(m_currentPath + "/xydata_gr_out.txt");
            } catch (IOException e) {
                // If there was an error, show a message dialog.
                e.printStackTrace();
                JOptionPane.showMessageDialog(
                        getParentFrame(m_pageProxy.getComponent()),
                        "Orphan file could not be saved.",
                        "IO Exception",
                        JOptionPane.ERROR_MESSAGE);
                return;
            }
        }
    }

//------------------------------------------------------------------------------
    /**
     * Update the window title to match the current page title.  If the page
     * doesn't have a title, use the suggested title.  The window title is
     * trimmed to 80 charactersand then has " - Gr" appended to it.
     * If there is no current page, use "USGS Gr" as the title.
     * @param suggestedTitle    The title to use if the page doesn't have one.
     */
    public void updateWindowTitle(String suggestedTitle) {

        // If there is no current page, use "USGS Gr" as the title.
        if (m_page == null) {
            getParentFrame(m_pageProxy.getComponent()).setTitle("USGS Gr");
            return;
        }

        // If the page doesn't have a title, use the suggested title.
        String pageTitle = m_page.getTitle();
        if (pageTitle == null || pageTitle.equals("")) {
            pageTitle = suggestedTitle;
            m_page.setTitle(pageTitle);
        }

        // Trim the page title to 80 characters.
        if (pageTitle.length() > 80) {
            pageTitle = pageTitle.substring(0, 80);
        }

        // If the page title doesn't match the window title, update the
        // window.
        if (!m_pageTitle.equals(pageTitle)) {
            m_pageTitle = pageTitle;
            getParentFrame(m_pageProxy.getComponent()).
                    setTitle(pageTitle + " - Gr");
        }
    }

//------------------------------------------------------------------------------
    /**
     * Undoes the last modification.
     */
    public void undoModification() {

        // Check that there was a previous state saved that can be restored.
        int childCount = m_undoList.getChildCount();
        if (childCount == 0 || m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Swap the current array with the saved undo array and save
        // the current state for later redo.
        GrObject snapShot = m_undoList.remove(childCount - 1);

        // Restore the snapshot.
        if (snapShot instanceof DataSnapShot) {
            // Data snapshot...
            DataSnapShot redoShot = ((DataSnapShot) snapShot).restore();
            m_redoList.add(redoShot);
        }
        if (snapShot instanceof PageOfGraphs) {
            // Page snapshot...
            m_redoList.add(m_page);
            setPage((PageOfGraphs) snapShot);
        }

        // Update the curve list on the status bar.
        m_bigStatusBar.updateCurveList();

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Redoes the last undo.
     */
    public void redoModification() {

        // Check that there was a previous undo saved that can be redone.
        int childCount = m_redoList.getChildCount();
        if (childCount == 0 || m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Swap the current array with the saved undo array and save
        // the current state for undoing the redo.
        GrObject snapShot = m_redoList.remove(childCount - 1);

        // Restore the snapshot.
        if (snapShot instanceof DataSnapShot) {
            // Data snapshot...
            DataSnapShot undoShot = ((DataSnapShot) snapShot).restore();
            m_undoList.add(undoShot);
        }
        if (snapShot instanceof PageOfGraphs) {
            // Page snapshot...
            m_undoList.add(m_page);
            setPage((PageOfGraphs) snapShot);
        }

        // Update the curve list on the status bar.
        m_bigStatusBar.updateCurveList();

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Select all points on all curves on the current page that currently have at
     * least one point selected
     */
    public void selectAll() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Look at each series that is at least partially selected and
        // select all the points in that series.
        Iterator iter = m_page.getSelectedSeriesList().iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries) iter.next();
            series.selectAll();
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Deselect all points on the currently selected curves on the current page.
     */
    public void deselectAll() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Look at each series that is at least partially selected and
        // deselect all the points in that series.
        Iterator iter = m_page.getSelectedSeriesList().iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries) iter.next();
            series.deselectAll();
        }
        m_page.updateSelectedSeriesList();

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Remove all series on the current page that have any points selected from the
     * graphs they are on and add them to the program's clipboard.
     */
    public void cutSelected() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        synchronized (m_page) {
            // Don't proceed if there are no selected series.
            if (m_page.getSelectedSeriesList().size() <= 0) {
                return;
            }

            // Show an hourglass until the job is complete.
            ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                    new Cursor(Cursor.WAIT_CURSOR));

            // Save the current page arrangement so it can be restored later.
            savePageSnapShot();

            // Empty the clipboard of previously stored items.
            m_clipBoard.clear();

            // Iterate through the selected series.
            List selSeriesList = m_page.getSelectedSeriesList();
            Iterator iter = selSeriesList.iterator();
            while (iter.hasNext()) {
                DoubleDataSeries series = (DoubleDataSeries) iter.next();
                // Save the series in the clipboard.
                m_clipBoard.add(series);

                // Hunt through all the graphs and delete the curve that
                // was cut.
                List graphList = m_page.getGraphList();
                Iterator graphIter = graphList.iterator();
                while (graphIter.hasNext()) {
                    Graph graph = (Graph) graphIter.next();
                    List curveList = new ArrayList(graph.getCurves());
                    Iterator curveIter = curveList.iterator();
                    int curveNum = -1;
//                    int i = 0;
                    while (curveIter.hasNext()) {
                        Curve curve = (Curve) curveIter.next();
                        if (curve.getSeries() == series) {
                            // The series in this curve matches the
                            // cut series, so remove the curve.
                            curveIter.remove();
                            graph.remove(curve);
                        }
//                        i++;
                    }

                    // If the graph was emptied of curves, set its
                    // title to show that.
                    if (curveList.size() == 0) {
                        graph.setTitle("Empty");
                    }
                }
            }

            // Update the selected series list.
            m_page.updateSelectedSeriesList();

            // Update the curve list on the status bar.
            m_bigStatusBar.updateCurveList();
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Add all series on the current page that have any points selected to the
     * program's clipboard.
     */
    public void copySelected() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Don't proceed if there are no selected series.
        if (m_page.getSelectedSeriesList().size() <= 0) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Empty the clipboard of previously stored items.
        m_clipBoard.clear();

        // Iterate through the selected series.
        Iterator iter = m_page.getSelectedSeriesList().iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries) iter.next();
            // Save the series in the clipboard.
            m_clipBoard.add(series.createCopy(true));
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Paste all the curves in the clipboard to each of the selected graphs.
     */
    public void pasteSelected() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        synchronized (m_page) {

            // Don't proceed if there are no items to paste or no selected graphs.
            if (m_clipBoard.getChildCount() <= 0 ||
                    m_page.getSelectedGraphs().size() <= 0) {

                return;
            }

            // Show an hourglass until the job is complete.
            ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                    new Cursor(Cursor.WAIT_CURSOR));

            // Save the current page arrangement so it can be restored later.
            savePageSnapShot();

            // Keep track of any time series that are pasted.
            boolean m_bSyncReferenceYears = false;

            // For each item in the clipboard...
            Enumeration seriesEnum = m_clipBoard.children();
            while (seriesEnum.hasMoreElements()) {
                DoubleDataSeries series =
                        (DoubleDataSeries) seriesEnum.nextElement();

                // Check for time series that need to be synched.
                if (series instanceof TimeSeries) {
                    m_bSyncReferenceYears = true;
                }

                // Paste a copy of it into each selected graph.
                List selGraphs = m_page.getSelectedGraphs();
                Iterator graphIter = selGraphs.iterator();
                while (graphIter.hasNext()) {
                    Graph graph = (Graph) graphIter.next();
                    graph.add((DoubleDataSeries) series.createCopy(true));
                }
            }


            // Find the reference year for time series on the current page
            // and sync the new series to it.
            if (m_bSyncReferenceYears) {
                TimeSeries tSeries = (TimeSeries) m_page.firstChildInstanceOf(
                        "gov.usgs.gr.data.TimeSeries", true);

                // Get a list of all TimeSeries objects on the page.
                int refYear = 0;
                if (tSeries != null) {
                    refYear = tSeries.getReferenceYear();
                }

                if (refYear > 0) {
                    // Sync all new time series to existing reference years
                    // of the time series on the page.
                    List tsList = m_page.findChildInstancesOf(
                            "gov.usgs.gr.data.TimeSeries", true);
                    Iterator tsIter = tsList.iterator();
                    while (tsIter.hasNext()) {
                        TimeSeries series = (TimeSeries) tsIter.next();
                        series.syncToReferenceYear(refYear);
                    }
                }
            }

            // Update the selected series list.
            m_page.updateSelectedSeriesList();

            // Update the curve list on the status bar.
            m_bigStatusBar.updateCurveList();
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Paste copies of each series in a list to the same graph as the original
     * series.
     * @param seriesList	A list of series that should have copies made.
     * @return A list of copies that were created and added to the page.
     */
    public List pasteToSameGraph(List seriesList) {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return (new ArrayList());
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Make a list of the new curves that are created.
        List newSeriesList = new ArrayList();

        // Copy each series in the list.
        Iterator seriesEnum = seriesList.iterator();
        while (seriesEnum.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries) seriesEnum.next();

            // Look for the series in each graph on the page.
            synchronized (m_page) {
                List selGraphs = m_page.getSelectedGraphs();
                Iterator graphIter = selGraphs.iterator();
                graphOuter:
                while (graphIter.hasNext()) {
                    Graph graph = (Graph) graphIter.next();
                    List curveList = graph.getCurves();
                    Iterator curveIter = curveList.iterator();
                    while (curveIter.hasNext()) {
                        Curve curve = (Curve) curveIter.next();
                        // If this graph contains the series, add a copy of the
                        // series to the graph.
                        if (curve.getSeries() == series) {
                            DoubleDataSeries newSeries =
                                    (DoubleDataSeries) series.createCopy(true);
                            graph.add(newSeries);
                            curve.setEditing(true);
                            newSeriesList.add(newSeries);
                            break graphOuter;
                        }
                    }
                }
            }
        }

        // Update the selected series list.
        m_page.updateSelectedSeriesList();

        // Update the curve list on the status bar.
        m_bigStatusBar.updateCurveList();

        return newSeriesList;
    }

//------------------------------------------------------------------------------
    /**
     * Delete all selected points on the current page or, if there are no selected
     * points, prompt the user to delete the selected graphs.
     */
    public void deleteSelected() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        synchronized (m_page) {
            // Show an hourglass until the job is complete.
            ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                    new Cursor(Cursor.WAIT_CURSOR));

            // Create an empty snapshot object for storing selected series
            // for undoing later.
            DataSnapShot snapShot = new DataSnapShot();

            // For each selected series...
            List selectedSeriesList = m_page.getSelectedSeriesList();
            Iterator iter = m_page.getSelectedSeriesList().iterator();
            if (!iter.hasNext()) {
                // Quit the routine if no graphs are selected.
                if ((m_page.getSelectedGraphs()).size() == 0) {
                    // Return to the default arrow cursor.
                    ((Component) getParentFrame(m_pageProxy.getComponent())).
                            setCursor(Cursor.getDefaultCursor());
                    return;
                }

                // Have the user confirm whether selected graphs should be
                // deleted.
                int choice = JOptionPane.showConfirmDialog(
                        getParentFrame(m_pageProxy.getComponent()),
                        "Delete selected graphs?", "Confirm Delete",
                        JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
                if (choice == JOptionPane.YES_OPTION) {
                    // Save the page arrangement so it can be restored
                    // later.
                    savePageSnapShot();

                    // Delete the selected graphs.
                    m_page.deleteSelectedGraphs();

                    // Update the curve list on the status bar.
                    m_bigStatusBar.updateCurveList();
                }
            } else {
                // If there are selected series...
                while (iter.hasNext()) {
                    DoubleDataSeries series = (DoubleDataSeries) iter.next();

                    // Add the series to the data snapshot.
                    snapShot.store(series);

                    // Delete the selected points from the series.
                    series.deleteSelectedPoints();
                }

                // Save the data snapshot for undoing later.
                saveDataSnapShot(snapShot);
            }

            // Trigger a repaint of the updated contents.
            m_pageProxy.repaint();
        }
    }

//------------------------------------------------------------------------------
    /**
     * Adds a command to the page command list.
     * @param command   The command to log.
     */
    public void logCommand(String command) {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Add the command.
        List list = m_page.getCommandList();
        list.add(command);
    }

//------------------------------------------------------------------------------
    /**
     * Applies a format to the page that affects the screen or printed appearance.
     * @param formatter The formatter to use.
     */
    public void applyFormat(PageFormatter formatter) {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Save the page so changes can be undone later.
        savePageSnapShot();

        // Hook the formatter to the page and vice versa.
        formatter.setPage(m_page);
        m_page.setFormatter(formatter);

        // Initialize the formatter.
        formatter.initialize();

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Open a classic format template and apply it to the data that is open.
     */
    public void openTemplate() {

        // Find the frame for attaching the dialog.
        Frame frame = getParentFrame(m_pageProxy.getComponent());

        // Let the user choose the file name.
        File[] files = showFileDialog("Open Template", true,
                new File(m_currentFile.getParent() + "/t1.grt"));

        // Open each template specified by the user.
        for (int i = 0; i < files.length; i++) {
            openTemplate(files[i].getAbsolutePath());
        }
    }

//------------------------------------------------------------------------------
    /**
     * Opens the specified classic-format template file and applies it to the data.
     * @param filename  The name of the file to open.
     */
    public void openTemplate(String filename) {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        synchronized (m_page) {
            // Save the page so changes can be undone later.
            savePageSnapShot();

            // Initialize local variables.
            Graph graph = null;
            Curve curve = null;

            try {
                // Open the file.
                BufferedReader br = new BufferedReader(new FileReader(filename));

                // Set up the stream tokenizer for parsing input.
                StreamTokenizer st = new StreamTokenizer(br);
                st.eolIsSignificant(true);
                st.slashSlashComments(true);
                st.slashStarComments(true);

                // Initialize the lists.
                List attribs = new ArrayList();
                List showList = m_page.getShowList();
                List curveList = null;
                int curveNum = 0;

                // Process each token.
                String line = null;
                st.nextToken();
                Iterator graphIter = showList.iterator();
                // Proceed until the end of the file.
                while (st.ttype != StreamTokenizer.TT_EOF) {
                    // Clear the attributes list and read the current line.
                    attribs.clear();
                    // Load up the attribs list.
                    do {
                        if (st.ttype == StreamTokenizer.TT_NUMBER) {
                            attribs.add(new Float(st.nval));
                        } else {
                            if (st.sval != null) {
                                attribs.add(st.sval);
                            }
                        }

                        st.nextToken();
                    } while (st.ttype != StreamTokenizer.TT_EOL && st.ttype
                            != StreamTokenizer.TT_EOF);

                    // Skip this line of input if there were no attributes
                    // or the first token was not a String.
                    if (attribs.size() == 0) {
                        continue;
                    }
                    if (!(attribs.get(0) instanceof String)) {
                        continue;
                    }

                    // Look for the key as the first token.
                    String key = ((String) attribs.get(0)).toUpperCase();
                    attribs.remove(0);

                    // Process the line according to key.
                    if (key.equals("PAGE")) {
                        // Set the page's title.
                        if (attribs.size() > 0) {
                            m_page.setTitle((String) attribs.get(0));
                        }
                    }
                    if (key.equals("GRAPH")) {
                        // Get the graph that will be used from now until the
                        // next graph key is encountered in the file.
                        if (graphIter.hasNext()) {
                            graph = (Graph) graphIter.next();
                            curveList = graph.getCurves();
                        } else {
                            continue;
                        }

                        // Set the graph's title.
                        if (attribs.size() > 0) {
                            graph.setTitle((String) attribs.get(0));
                        }
                    }
                    if (key.equals("XAXIS")) {
                        // Set all axis attributes and update the opposing axis.
                        graph.getXAxis().setAttributes(attribs);
                        graph.getXAxis().update();
                        graph.getYAxis().update();
                    }
                    if (key.equals("YAXIS")) {
                        // Set all axis attributes and update the opposing axis.
                        graph.getYAxis().setAttributes(attribs);
                        graph.getXAxis().update();
                        graph.getYAxis().update();
                    }
                    if (key.equals("CURVE")) {
                        // Get the curve that will be used from now until the
                        // next curve key is encountered in the file.
                        if (curveList.size() > (curveNum + 1)) {
                            curve = (Curve) curveList.get(curveNum);
                            curveNum++;
                        } else {
                            continue;
                        }
                        // Set all axis attributes and update the opposing axis.
                        curve.setAttributes(attribs);
                    }
                }

                // Close the file.
                br.close();
            } catch (IOException e) {
                // If there was an error, show a message dialog.
                e.printStackTrace();
                JOptionPane.showMessageDialog(getParentFrame(
                        m_pageProxy.getComponent()),
                        "File " + filename + " could not be opened.",
                        "IO Exception", JOptionPane.ERROR_MESSAGE);
            }
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Let the user choose an XML file to open.  Applies the properties of its
     * first page in the file to the current page.
     */
    public void openXMLTemplate() {

        openXMLTemplate(null);
    }

//------------------------------------------------------------------------------
    /**
     * Open an XML file and applies the properties of its first page to the
     * current page.
     * @param file  The file to open.
     */
    public void openXMLTemplate(File file) {

        // Search for the last file added, to prevent using it as the template
        // by mistake.
        GrFile oldLastFile = null;
        Enumeration enumeration = children();
        while (enumeration.hasMoreElements()) {
            GrObject gro = (GrObject) enumeration.nextElement();
            while (gro instanceof ObjectEffigy) {
                gro = ((ObjectEffigy) gro).getSubject();
            }

            if (gro instanceof GrFile) {
                oldLastFile = (GrFile) gro;
            }
        }

        // Have the user choose the XML file to open, but save the window
        // title and reset it after opening, so the window title will reflect
        // the original data instead of the template.
        String windowTitle = getParentFrame(m_pageProxy.getComponent()).
                getTitle();
        m_bAddChildrenWhenOpening = false;
        String curFile = System.getProperty("gov.usgs.gr.currentfile");
        if (file == null) {
            openFile();
        } else {
            openFile(file);
        }
        m_bAddChildrenWhenOpening = true;
        curFile.replace('\\', '/');
        System.setProperty("gov.usgs.gr.currentfile", curFile);

        // Search for the last file added, and assume it is the template file.
        GrFile templateFile = null;
        enumeration = children();
        while (enumeration.hasMoreElements()) {
            GrObject gro = (GrObject) enumeration.nextElement();
            while (gro instanceof ObjectEffigy) {
                gro = ((ObjectEffigy) gro).getSubject();
            }

            if (gro instanceof GrFile) {
                templateFile = (GrFile) gro;
            }
        }

        // Don't proceed if the last file is the same as it was before
        // attempting to open the template file.
        if (templateFile == oldLastFile) {
            return;
        }

        // Remove the template file from the list of children.
        remove(templateFile);

        // Search for the first page in the file and apply its properties
        // to the current page.
        enumeration = templateFile.children();
        while (enumeration.hasMoreElements()) {
            GrObject gro = (GrObject) enumeration.nextElement();
            while (gro instanceof ObjectEffigy) {
                gro = ((ObjectEffigy) gro).getSubject();
            }

            if (gro instanceof PageOfGraphs) {
                m_template = (PageOfGraphs) gro;
                if (m_page != null) {
                    m_page.applyProperties(m_template);
                }
                break;
            }
        }

        // Reset the window title.
        updateWindowTitle("");
    }

//------------------------------------------------------------------------------
    /**
     * Applies the current XML template to the data that is open.
     */
    public void applyXMLTemplate() {

        // Open a template if one isn't already open.
        if (m_template == null) {
            openXMLTemplate();
        } else {
            // Apply the existing template.
            m_page.applyProperties(m_template);
        }
    }

//------------------------------------------------------------------------------
    /**
     * Saves the current page as an XML file.
     */
    public void savePageAs() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Ask the user for the filename and save to that file.
        File[] files = showFileDialog("Save Page As", false,
                m_currentFile);

        // Don't proceed if no files were chosen.
        if (files.length <= 0) {
            return;
        }

        // Use the first file in the list of chosen files.
        File file = files[0];

        // Create an XML project file containing the page.
        XMLProjectFile projFile = new XMLProjectFile();
        projFile.setProperty("file", file.getAbsolutePath());
        projFile.add(m_page);

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Write the file to disk.
        try {
            projFile.write();
        } catch (java.io.IOException ioe) {
            ioe.printStackTrace();
        }

        // Return to the default arrow cursor.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                Cursor.getDefaultCursor());
    }

//------------------------------------------------------------------------------
    /**
     * Opens the specified legacy script file and applies it to the data.
     * @param file  The file to open.
     */
    public void openScript(File file) {

        // Create a script file object to parse the data.
        ScriptFile99 sf99 = new ScriptFile99();
        sf99.setProperty("file", file.getAbsolutePath());

        // Set up some additional variables.
        sf99.m_grCon = this;
        sf99.m_currentPath = m_currentPath;
        sf99.m_parentFrame = getParentFrame(m_pageProxy.getComponent());

        // Read and execute the script.
        try {
            sf99.read();
        } catch (IOException e) {
            // If there was an error reading the file, print a stack trace
            // and show a message dialog.
            e.printStackTrace();

            JOptionPane.showMessageDialog(
                    getParentFrame(m_pageProxy.getComponent()),
                    "File " + file.getAbsolutePath() + " could not be loaded.",
                    "IO Exception", JOptionPane.ERROR_MESSAGE);
            return;
        }
    }

//------------------------------------------------------------------------------
    /**
     * Turn select mode on or off.  When in select mode, points can be picked
     * for editing.
     * @param val   True if turning select mode on, false otherwise.
     */
    public void useSelectMode(boolean val) {

        // Set select mode on or off and set pan and zoom to the opposite value.
        m_bSelectMode = val;
        m_bPanMode = !val;
        m_bZoomMode = !val;

        // If select mode is on, show symbols on the curves so they can be
        // more easily selected.  Otherwise, draw the curves normally.
        showEditingSymbols(val);

        // Notify the status and menu bars of the change.
        m_grStatusBar.showSelectMode(val);
        m_grMenuBar.showSelectMode(val);
    }

//------------------------------------------------------------------------------
    /**
     * Turn draw mode on or off.  When in draw mode, mouse clicks are
     * interpreted as points in a new line.
     * @param val   True if turning draw mode on, false otherwise.
     */
    public void useDrawMode(boolean val) {

        // Set draw mode on or off and set select mode to the opposite value.
        // Turn pan and zoom modes off.
        m_bSelectMode = !val;
        m_bPanMode = false;
        m_bZoomMode = false;
        m_bDrawMode = val;

        // Remove the drawing curve if it has fewer than two points.
        if (m_drawModeCurve != null) {
            if (m_drawModeCurve.getSeries().getCount() < 2) {
                GrObject gro = m_page.findParentOf(m_drawModeCurve);
                if (gro != null) {
                    gro.remove(m_drawModeCurve);
                }
            }
        }
        m_drawModeCurve = null;

        // If draw mode is on, show symbols on the curves so the points can be
        // seen more easily.  Otherwise, draw the curves normally.
        showEditingSymbols(val);

        // Notify the status and menu bars of the change.
        m_grStatusBar.showDrawMode(val);
        m_grMenuBar.showSelectMode(!val);
    }

//------------------------------------------------------------------------------
    /**
     * Turn zoom mode on or off.  When in zoom mode, the mouse can be dragged
     * to define a rectangle that will become the new view area.
     * @param val   True if turning zoom mode on, false otherwise.
     */
    public void useZoomMode(boolean val) {

        // Set zoom mode on or off and set select mode and pan mode to the
        // opposite value.
        m_bZoomMode = val;
        m_bSelectMode = !val;
        m_bPanMode = !val;

        // If zoom mode is on, draw the curves normally.  Otherwise, show
        // symbols on the curves.
        showEditingSymbols(!val);

        // Notify the status and menu bars of the change.
        m_grStatusBar.showZoomMode(val);
        m_grMenuBar.showZoomMode(val);
    }

//------------------------------------------------------------------------------
    /**
     * Turn pan mode on or off.  When in pan mode, the mouse can be dragged
     * to pan the view area.
     * @param val   True if turning pan mode on, false otherwise.
     */
    public void usePanMode(boolean val) {

        // Set pan mode on or off and set select mode and zoom mode to the
        // opposite value.
        m_bPanMode = val;
        m_bZoomMode = !val;
        m_bSelectMode = !val;

        // If pan mode is on, draw the curves normally.  Otherwise, show
        // symbols on the curves.
        showEditingSymbols(!val);

        // Notify the status and menu bars of the change.
        m_grStatusBar.showPanMode(val);
        m_grMenuBar.showPanMode(val);
    }

//------------------------------------------------------------------------------
    /**
     * Pan all graphs one major increment to the right.
     * @param e The user interface event that triggered this action.
     * @param direction A multiplier for how far to pan measured in zoom
     *                  increments.
     */
    public void panAllHorz(ActionEvent e, int direction) {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Find out if the shift key was held down.
        boolean bShift = ((e.getModifiers() & e.SHIFT_MASK) != 0);

        // Look at each unique X zoom on the page.
        Iterator iter = m_page.getAllXZooms().iterator();
        while (iter.hasNext()) {
            Zoom zoom = (Zoom) iter.next();

            // If Shift was held down, pan one graph width.
            double inc = 1;
            if (bShift) {
                inc = (zoom.getViewMax() - zoom.getViewMin()) * 0.5;
            } else {
                inc = zoom.getMajorInc();
            }
            inc *= (double) direction;

            // Pan the zoom.
            zoom.pan(inc);
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Pan all graphs one major increment down.
     */
    public void panAllDown() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Pan each unique y zoom on the page.
        Iterator iter = m_page.getAllYZooms().iterator();
        while (iter.hasNext()) {
            Zoom zoom = (Zoom) iter.next();
            zoom.pan(-zoom.getMajorInc());
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Pan all graphs one major increment up.
     */
    public void panAllUp() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Pan each unique y zoom on the page.
        Iterator iter = m_page.getAllYZooms().iterator();
        while (iter.hasNext()) {
            Zoom zoom = (Zoom) iter.next();
            zoom.pan(zoom.getMajorInc());
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Revert to the last zoom.
     */
    public void zoomPrevious() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // X zooms...
        Iterator iter = m_page.getSelectedXZooms().iterator();
        while (iter.hasNext()) {
            Zoom zoom = (Zoom) iter.next();
            zoom.zoomPrevious();
        }

        // Y zooms...
        iter = m_page.getSelectedYZooms().iterator();
        while (iter.hasNext()) {
            Zoom zoom = (Zoom) iter.next();
            zoom.zoomPrevious();
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Zoom in on a smaller area of the zooms on all selected graphs.
     */
    public void zoomIn() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

//        // Save the page so changes can be undone later.
//        savePageSnapShot();

        // Only change the x-axis if it is unlocked.
        if (!m_bLockX) {
            // Change the x-axis for all currently selected graphs.
            Iterator iter = m_page.getSelectedXZooms().iterator();
            while (iter.hasNext()) {
                Zoom zoom = (Zoom) iter.next();
                zoom.zoomIn();
                // Autoscale...
                int numColumns = 1;
                zoom.autoScale(m_page.getWidth() / numColumns / 80);
            }
        }

        // Only change the y-axis if it is unlocked.
        if (!m_bLockY) {
            // Change the y-axis for all currently selected graphs.
            Iterator iter = m_page.getSelectedYZooms().iterator();
            while (iter.hasNext()) {
                Zoom zoom = (Zoom) iter.next();
                zoom.zoomIn();
                // Autoscale...
                int numRows = m_page.getShowList().size();
                zoom.autoScale(m_page.getHeight() / numRows / 60);
            }
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Zoom out to show a larger area of the graph.
     */
    public void zoomOut() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

//        // Save the page so changes can be undone later.
//        savePageSnapShot();

        // Only change the x-axis if it is unlocked.
        if (!m_bLockX) {
            // Change the x-axis for all currently selected graphs.
            Iterator iter = m_page.getSelectedXZooms().iterator();
            while (iter.hasNext()) {
                Zoom zoom = (Zoom) iter.next();
                zoom.zoomOut();
                // Autoscale...
                int numColumns = 1;
                zoom.autoScale(m_page.getWidth() / numColumns / 80);
            }
        }

        // Only change the y-axis if it is unlocked.
        if (!m_bLockY) {
            // Change the y-axis for all currently selected graphs.
            Iterator iter = m_page.getSelectedYZooms().iterator();
            while (iter.hasNext()) {
                Zoom zoom = (Zoom) iter.next();
                zoom.zoomOut();
                // Autoscale...
                int numRows = m_page.getShowList().size();
                zoom.autoScale(m_page.getHeight() / numRows / 60);
            }
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Zoom out to show all the data in the graph.
     */
    public void zoomAll() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

//        // Save the page so changes can be undone later.
//        savePageSnapShot();
        m_page.updateZoomContents();
        // Only change the x-axis if it is unlocked.
        if (!m_bLockX) {
            // Change the x-axis for all currently selected graphs.
            Iterator iter = m_page.getSelectedXZooms().iterator();
            while (iter.hasNext()) {
                Zoom zoom = (Zoom) iter.next();
                // Autoscale...
                int numColumns = 1;
//                zoom.zoomAll();
//                zoom.autoScale(m_page.getWidth() / numColumns / 80);
                zoom.zoomAll(m_page.getWidth() / numColumns / 80);
//                zoom.zoomAll(m_page.getWidth() / numColumns / 80);
            }
        }

        // Only change the y-axis if it is unlocked.
        if (!m_bLockY) {
            // Change the y-axis for all currently selected graphs.
            Iterator iter = m_page.getSelectedYZooms().iterator();
            while (iter.hasNext()) {
                Zoom zoom = (Zoom) iter.next();
                // Autoscale...
                int numRows = m_page.getShowList().size();
//                zoom.zoomAll();
//                zoom.autoScale(m_page.getHeight() / numRows / 80);
                zoom.zoomAll(m_page.getHeight() / numRows / 80);
//                zoom.zoomAll(m_page.getHeight() / numRows / 80);
            }
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Lock the x-axis so it won't be affected by pans or zooms.  Acts as a
     * toggle if the axis is already locked.
     */
    public void lockX() {

        // Switch the value of the boolean lock variable.
        m_bLockX = (m_bLockX == true) ? false : true;

        // Display the new lock status on the status bar.
        if (m_bLockX) {
            m_grStatusBar.axis1LockButton.setSelected(true);
        } else {
            m_grStatusBar.axis1LockButton.setSelected(false);
        }

        // Trigger a repaint of the status bar.
        m_grStatusBar.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Lock the y-axis so it won't be affected by pans or zooms.  Acts as a
     * toggle if the axis is already locked.
     */
    public void lockY() {

        // Switch the value of the boolean lock variable.
        m_bLockY = (m_bLockY == true) ? false : true;

        // Display the new lock status on the status bar.
        if (m_bLockX) {
            m_grStatusBar.axis2LockButton.setSelected(true);
        } else {
            m_grStatusBar.axis2LockButton.setSelected(false);
        }

        // Trigger a repaint of the status bar.
        m_grStatusBar.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Add a new blank page and make it the current page.
     */
    public void newPage() {

        add(m_pageProxy.setNewPage());

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Add a new blank graph to the page.
     */
    public void newGraph() {

        // Create and add a new blank page if there is currently none.
        if (m_page == null) {
            m_page = m_pageProxy.setNewPage();
            add(m_page);
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        synchronized (m_page) {
            // Save the page so changes can be undone later.
            savePageSnapShot();

            // Have a new graph added to the page.
            m_page.addNewGraph();
            // Put the new graph in the showing list.
            m_page.updateLists();
            // Cause the new graph to be shown on the screen.
            m_page.getFormatter().layoutGraphs();
        }

        // Trigger a repaint in this panel.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Hides the currently selected graph or graphs.
     * @see gov.usgs.gr.visual.PageOfGraphs#hideGraph PageOfGraphs
     * hideGraph() for more details.
     */
    public void hideGraph() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Save the page so changes can be undone later.
        savePageSnapShot();

        // Hide the selected graphs.
        m_page.hideGraph();

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Trigger a repaint in this panel.
     */
    public void triggerRepaint() {

        // Cause the panel to redraw the current page.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Show the first graph in the list of hidden graphs.
     * @see gov.usgs.gr.visual.PageOfGraphs#showMoreGraphs PageOfGraphs
     * showMoreGraphs() for more details.
     */
    public void showMoreGraphs() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Save the page so changes can be undone later.
        savePageSnapShot();

        // Show the next hidden graph.
        m_page.showMoreGraphs();

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Switch between the regular status bar and the curve properties bar.  Acts
     * as a toggle.
     */
    public void switchStatusBar() {

        // Show an hourglass until the job is complete.
        JFrame frame = (JFrame) getParentFrame(m_pageProxy.getComponent());

        // Remove the current status bar from the main Gr window.
        frame.getContentPane().remove(m_grStatusBar);
        GrStatusBar oldBar = m_grStatusBar;

        // Switc the status bar variable to the other type of status bar.
        if (m_grStatusBar == m_littleStatusBar) {
            m_grStatusBar = m_bigStatusBar;
        } else {
            m_grStatusBar = m_littleStatusBar;
        }

        // Add the the new status bar to the bottom of the Gr window.
        frame.getContentPane().add(m_grStatusBar, "South");

        // Bring the new bar up to date on axis locking and mode status.
        m_grStatusBar.axis1LockButton.setSelected(
                oldBar.axis1LockButton.isSelected());
        m_grStatusBar.axis2LockButton.setSelected(
                oldBar.axis2LockButton.isSelected());
        m_grStatusBar.modeLabel.setText(oldBar.modeLabel.getText());

        // Trigger a repaint of the updated contents.
        frame.pack();
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Create printed output of the page on the screen.  Ask the user for the
     * file name first, and optionally ask if the file should be send directly
     * to the printer.
     */
    public void printPage() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Determine the default path and print file name to show in the
        // file selection dialog.
        String fileName = m_page.m_psOptions.getPsFileName();
        File oldFile = new File(fileName);
        if (!oldFile.exists()) {
            fileName = oldFile.getName();
        }
        if ((fileName.indexOf('/') < 0 && fileName.indexOf('\\') < 0)) {
            String currentFileName =
                    System.getProperty("gov.usgs.gr.currentfile");
            File file = new File(currentFileName);
            fileName = file.getAbsoluteFile().getParent() + "/" + fileName;
        }

        // Ask the user for the filename and save to that file.
        File psFile = new File(fileName);
        File[] files = showFileDialog("Save PS File", false, psFile);
        if (files.length > 0) {
            printPage(files[0].getAbsolutePath());

            // If the PostScript options say to, show the print dialog so the
            // user has the option of sending the file to the printer.
            GrHashList props = m_page.m_psOptions.getProperties();
            GrProperty p = (GrProperty) props.get("showPrintDialog");
            if ((Boolean) p.value == true) {
                // Send the file to the printer.
                final PrintPsFileDialog sendToPrinterDialog =
                        new PrintPsFileDialog(getParentFrame(
                        m_pageProxy.getComponent()), "Print PostScript File",
                        true, files[0]);
                javax.swing.SwingUtilities.invokeLater(new Runnable() {

                    public void run() {
                        sendToPrinterDialog.setVisible(true);
                    }
                });

            }
        }
    }

//------------------------------------------------------------------------------
    /**
     * Write printable version of the page on the screen to a file.  Performs
     * point reduction as specified in the PS options object.
     * @param filename  The name of the file to which print commands will be
     *                  written.
     */
    public void printPage(String filename) {

        // Show an hourglass until the job is complete.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                new Cursor(Cursor.WAIT_CURSOR));

        // Find out if information will be written out for debugging and
        // describing the point reduction method.
        boolean debugPR = m_page.getDebugPointReduction();

        // If debugging point reduction, save the current page arrangement so
        // it can be restored later.
        if (debugPR) {
            savePageSnapShot();
        }

        printPage(filename, m_page);

        // Return to the default arrow cursor.
        ((Component) getParentFrame(m_pageProxy.getComponent())).setCursor(
                Cursor.getDefaultCursor());
    }

//------------------------------------------------------------------------------
    /**
     * Write printable version of the page on the screen to a file.  Performs
     * point reduction as specified in the PS options object.
     * @param filename  The name of the file to which print commands will be
     *                  written.
     */
    public static void printPage(String filename, PageOfGraphs page) {

        // Ask the user for the filename and save to that file.
        File file = new File(filename);

        // Don't proceed if the file handle couldn't be obtained.
        if (file == null) {
            return;
        }

        // Set the PS options output name to the name of the file that is
        // about to be written to.
        page.m_psOptions.setPsFileName(file.getAbsolutePath());

        // Make a shallow copy of the page.
        PageOfPsGraphs psPage = new PageOfPsGraphs(page);
        psPage.getFormatter().reformat();

        // Find out if information will be written out for debugging and
        // describing the point reduction method.
        boolean debugPR = page.getDebugPointReduction();

        // Create point reduction tools for using later.
        LinePointReducer lpr1 = new LinePointReducer(1, 1, 1);
        SymbolPointReducer spr = new SymbolPointReducer(1, 1, 1);

        // Perform point reduction on the curves on the PS page.
        Iterator psGraphIter = psPage.getShowList().iterator();
        Iterator screenGraphIter = page.getShowList().iterator();
        while (psGraphIter.hasNext()) {
            Graph psGraph = (Graph) psGraphIter.next();
            Graph screenGraph = (Graph) screenGraphIter.next();

            // Calculate the scale and width parameters for the
            // point reducer.
            Rectangle2D.Double frame = psGraph.getFrameRect();
            if (psGraph instanceof PsGraph) {
                frame = ((PsGraph) psGraph).getMediaFrameRect();
            }
            double xScale =
                    (double) (frame.width / psGraph.getXZoom().getViewLength());
            double yScale =
                    (double) (frame.height / psGraph.getYZoom().getViewLength());

            // Create a new point reducer for this graph, because each
            // graph may have its own X and Y scales.
            lpr1.setXScale(xScale);
            lpr1.setYScale(yScale);
            spr.setXScale(xScale);
            spr.setYScale(yScale);

            // Reduce the points of each curve on the PS page.
            Iterator psCurveEnum = psGraph.getCurves().iterator();
            Iterator screenCurveEnum =
                    screenGraph.getCurves().iterator();
            while (psCurveEnum.hasNext()) {
                PsCurve curve = (PsCurve) psCurveEnum.next();
                Curve screenCurve = (Curve) screenCurveEnum.next();

                // Skip hidden curves.
                if (!curve.getShow()) {
                    continue;
                }

                // Set debug state for the curve.
                curve.setDebugPointReduction(debugPR);
                screenCurve.setDebugPointReduction(debugPR);

                // If drawing a line between points and LPR is on...
                if (curve.getShowLine() &&
                        page.m_psOptions.getUseLinePointReduction1()) {

                    // Create a copy of the data series to use for reducing
                    // points and drawing.
                    DoubleDataSeries series =
                            (DoubleDataSeries) (curve.getSeries().createCopy(true));
                    curve.setReducedLineSeries(series);
                    screenCurve.setReducedLineSeries(series);

                    // Set the width here, because each curve may have its
                    // own width.
                    double width = (psPage.m_psOptions.getPsUniformWidth() ?
                        0.5 : (0.25 * curve.getLineWidth()));
                    width *= psPage.m_psOptions.getPsLineWidth();
                    width *= psPage.m_psOptions.getLPR1WidthScale();

                    // Divide the width in half.  This lets the user specify
                    // the number of line widths the thinned-out series
                    // can be off from the original without requiring them
                    // to know that the LPR algorithm is accurate to within
                    // twice the width it is given.
                    width /= 2;
                    lpr1.setWidth(width);

                    // Reduce points on the copy of the data series.
                    lpr1.clear();
                    lpr1.add(series);
                    lpr1.recalculate();
                } else {
                    // If not drawing a line between points with LPR, set the
                    // reduced line series to null.
                    curve.setReducedLineSeries(null);
                    screenCurve.setReducedLineSeries(null);
                }

                // If drawing a symbol on each point and SPR is on...
                if (curve.getShowSymbols() && page.m_psOptions.
                        getUseSymbolPointReduction()) {

                    // Create a copy of the data series to use for reducing
                    // points and drawing.
                    DoubleDataSeries series =
                            (DoubleDataSeries) (curve.getSeries().
                            createCopy(true));
                    curve.setReducedSymbolSeries(series);
                    screenCurve.setReducedSymbolSeries(series);

                    // Set the width here, because each curve may have its
                    // own width.
                    float width = curve.getSymbolSize();
                    width *= psPage.m_psOptions.getSPRWidthScale();
                    spr.setWidth(width);

                    // Reduce points on the copy of the data series.
                    List seriesList = new ArrayList();
                    seriesList.add(series);
                    spr.performAction(seriesList);
                } else {
                    // If not drawing a symbol on each point with SPR, set the
                    // reduced symbol series to null.
                    screenCurve.setReducedSymbolSeries(null);
                    curve.setReducedSymbolSeries(null);
                }
            }
        }

        // Have the PostScript page write its drawing commands to the file.
        psPage.drawFromThread();
    }

//------------------------------------------------------------------------------
    /**
     * Toggle quick drawing so speed can be traded for quality.
     */
    public void toggleQuickDraw() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Set the quick draw member variable.
        m_bQuickDraw = (m_bQuickDraw == true ? false : true);

        // Propagate the new setting to each of the graphs.
        List graphList = m_page.getGraphList();
        Iterator iter = m_page.getGraphList().iterator();
        while (iter.hasNext()) {
            Graph graph = (Graph) iter.next();
            graph.setQuickDraw(m_bQuickDraw);
        }

        // Trigger a repaint of the updated contents.
        m_pageProxy.repaint();
    }

//------------------------------------------------------------------------------
    /**
     * Format the X-coordinate to be written in the status bar for the given
     * mouse position.
     * @param x The horizontal position of the mouse.
     * @param graph The graph that the mouse is over.
     * @return  The formatted X-coordinate.
     */
    protected String formatXCoord(double x, Graph graph) {

        // Initialize the output text string and the formatter.
        String xCoord = "";
        DecimalFormat form = new DecimalFormat(m_bigStatusBar.getFormat());

        // Format as a regular decimal number.
        xCoord = form.format(x);

        // Find out if the nearest curve is a TimeSeries.
        Curve curve = graph.getNearestCurve();
        if (curve != null) {
            DataSeries series = curve.getSeries();
            if (series instanceof TimeSeries) {
                // Use a date format instead of decimal.
                // Add half a minute because seconds will be truncated.
                // Assume that time units are in days.
                String dateFormat = m_grStatusBar.dateFormat;
                if (dateFormat.length() > 0) {
                    if (dateFormat.indexOf('0') < 0) {
                        // It is a date format...
                        if (dateFormat.indexOf('s') < 0) {
                            x += 0.0003472222222222222;
                        }
                        GregorianCalendar date = new GregorianCalendar(
                                TimeZone.getTimeZone("UTC"));
                        date = TimeDataFile.yearDayToGregCal(
                            ((TimeSeries) series).getReferenceYear(), x, date);
                        SimpleDateFormat formatter =
                                new SimpleDateFormat(dateFormat);
                        formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
                        StringBuffer buf = formatter.format(date.getTime(),
                                new StringBuffer(), new FieldPosition(0));
                        xCoord = buf.toString();
                    }
                }
            }
        }

        return xCoord;
    }

//------------------------------------------------------------------------------
    /**
     * Handle mouse movement within the page area.  Updates nearest point
     * coordinates on the status bar and shows and hides the side bar of the
     * graphs.
     * @param e     The mouse event object containing the mouse position.
     */
    public void mouseMoved(MouseEvent e) {

//        System.out.println("Mouse moved to " + e.getX() + ", " + e.getY());
        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }
        synchronized (m_page) {
            if (m_mouseHandler != null) {
                m_mouseHandler.mouseMoved(e);
            }

            // Initialize the X-coordinate text string and the formatter.
            String xCoord = "";
            DecimalFormat form = new DecimalFormat(m_bigStatusBar.getFormat());
            Point2D.Double p = null;

            // Find the graph nearest to the mouse.
            Graph graph = m_page.getNearestChild(e.getX(), e.getY());

            // Check which coordinate mode the status bar is in.
            if (m_grStatusBar.m_coordLevelIndex == 0) {
                // Look for the nearest point on a curve to the mouse.
                if (graph != null) {
                    p = graph.getNearestPoint(e.getX(),
                            Math.round(m_page.getHeight()) - e.getY());
                }
            } else {
                // Convert the screen pixel coordinates to user space
                // coordinates.
                Rectangle screenRect = new Rectangle(e.getX(), e.getY(), 0, 0);
                Rectangle2D.Double userRect =
                        m_page.screenToUser(screenRect, graph);

                if (userRect != null) {
                    p = new Point2D.Double(userRect.x, userRect.y);
                }
            }

            // Write zero for the X-coordinate if there was no nearest point.
            if (p == null) {
                p = new Point2D.Double(0, 0);
                xCoord = form.format(p.x);
            } else {
                xCoord = formatXCoord(p.x, graph);
            }

            // Choose the curve if the status bar is set to update its curve
            // whenever the mouse moves.
            if (m_grStatusBar == m_bigStatusBar && !m_bDrawMode) {
                // Big status bar is currently showing...
                if (!m_bigStatusBar.m_bSelectWithClick) {

                    m_grStatusBar.nearestXLabel.setText(xCoord);
                    m_grStatusBar.nearestYLabel.setText(form.format(p.y));

                    m_bigStatusBar.setCurve(m_page.getNearestCurve(
                            e.getX(), e.getY()));
                }
            } else {
                // Little status bar is currently showing...
                m_grStatusBar.nearestXLabel.setText(xCoord);
                m_grStatusBar.nearestYLabel.setText(form.format(p.y));
            }

            // Set the graphs' side bars on or off.
            boolean bSideBarChanged = m_page.setSideBar(e.getX(), e.getY(),
                    false);
            if (bSideBarChanged) {
                m_pageProxy.repaint();
            }

            // Trigger a repaint of the updated contents.
            m_grStatusBar.repaint();
        }
    }

//------------------------------------------------------------------------------
    /**
     * Handle mouse dragging within the page area.  Dragging can be interpreted
     * as panning, defining a zoom rectangle, or moving a data point.
     * @param e     The mouse event object containing the mouse position.
     */
    public void mouseDragged(MouseEvent e) {

//System.out.println("Mouse dragged to " + e.getX() + ", " + e.getY());
        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }
        synchronized (m_page) {

            if (m_mouseHandler != null) {
                m_mouseHandler.mouseDragged(e);
            }

            Graph graph0 = (Graph) m_page.getSelectedGraphs().get(0);
            // If the graph is a 3D graph, let it handle this event.
            if (graph0 instanceof Gl3dGraph) {
                ((Gl3dGraph) graph0).mouseDragged(e);

                // Trigger a repaint in this panel.
                m_pageProxy.repaint();
                return;
            }

            // Declare local variables.
            int xDiff, yDiff;

            // Check if the Shift or right mouse were down.
            boolean bShift = ((e.getModifiers() & e.SHIFT_MASK) != 0);
            boolean bButton3Down = ((e.getModifiers() & e.BUTTON3_MASK) != 0);

            // If currently moving a graph, then interpret the drag as further
            // graph movement.
            if (m_page.getMovingGraph() != null) {
                m_page.moveGraph(e.getX() - m_mouseX, m_mouseY - e.getY());

                // Save the current mouse position for future use.
                m_mouseX = e.getX();
                m_mouseY = e.getY();

                // Trigger a repaint.
                m_pageProxy.repaint();
                return;
            }

            // If in pan mode with the right mouse button down...
            if (m_bPanMode || bButton3Down) {

                m_bDragMode = false;

                // Get the difference in mouse position since the last call.
                // Make the difference zero if the axis is locked.
                xDiff = m_bLockX ? 0 : m_mouseX - e.getX();
                yDiff = m_bLockY ? 0 : e.getY() - m_mouseY;

                if (bShift) {
                    // Zoom if shift is held down.
                    m_bZoomMode = true;
                    m_page.setZoomWindow(m_mouseX, m_mouseY, e.getX(), e.getY());
                } else {
                    // Send the pan message to the selected graphs.
                    m_page.pan(xDiff, yDiff);

                    // Save the current mouse position for future use.
                    m_mouseX = e.getX();
                    m_mouseY = e.getY();
                }
            } else {
                // If in zoom or select mode and not in the middle of dragging
                // a point already, update the zoom rectangle window.
                if ((m_bZoomMode || m_bSelectMode) && !m_bDragMode) {
                    m_page.setZoomWindow(m_mouseX, m_mouseY, e.getX(), e.getY());
                }
            }

            // If currently dragging a point...
            if (m_bDragMode) {
                // Set the status bar to say drag mode.
                m_grStatusBar.modeLabel.setText("Drag");

                // Find how far the mouse was dragged.
                xDiff = e.getX() - m_mouseX;
                yDiff = e.getY() - m_mouseY;

                // Keep the x-y position of the rect in the rect's bottom left.
                if (xDiff < 0) {
                    m_mouseX = e.getX();
                }
                if (yDiff < 0) {
                    m_mouseY = e.getY();
                }

                Graph selectedGraph = (Graph) m_page.getSelectedGraphs().get(0);

                // Find the rectangle that describes the motion in the graph's
                // coordinate system.
                Rectangle2D.Double shift = m_page.screenToUser(new Rectangle(
                        m_mouseX, m_mouseY, Math.abs(xDiff), Math.abs(yDiff)),
                        selectedGraph);
                double xAmount = 0, yAmount = 0;
                if (shift != null) {
                    xAmount = shift.width;
                    yAmount = shift.height;
                }
                if (xDiff < 0) {
                    xAmount *= -1;
                }
                if (yDiff > 0) {
                    yAmount *= -1;
                }

                // Check if the x and y movement was big enough to overcome the
                // snap of at least one data series.
                boolean xTooSmall = true;
                boolean yTooSmall = true;

                // For each selected curve...
                Iterator snapIter = m_page.getSelectedSeriesList().iterator();
                while (snapIter.hasNext()) {
                    DoubleDataSeries series = (DoubleDataSeries) snapIter.next();
                    if (series.getNumDimensions() != 2) {
                        continue;
                    }

                    // If the movement was not enough to jump to the next snap
                    // increment, let it qualify for big enough if it was at
                    // least half an increment.
                    if (xTooSmall) {
                        if (series.getSnap(0) < 2 * Math.abs(xAmount)) {
                            xTooSmall = false;
                        }
                    }
                    if (yTooSmall) {
                        if (series.getSnap(1) < 2 * Math.abs(yAmount)) {
                            yTooSmall = false;
                        }
                    }
                }

                // Update the positions of all selected points that are currently
                // being dragged.
                Iterator iter = m_page.getSelectedSeriesList().iterator();
                while (iter.hasNext()) {
                    DoubleDataSeries series = (DoubleDataSeries) iter.next();

                    // Shift the value of the selected points as long as the
                    // distance was far enough to snap to the next increment
                    // and the series is unlocked in that direction.
                    if (!xTooSmall) {
                        if (!series.getDimensionLocked(0)) {
                            series.modifySelectedPoints(1, 0, xAmount);
                        }
                    }
                    if (!yTooSmall) {
                        if (!series.getDimensionLocked(1)) {
                            series.modifySelectedPoints(1, 1, yAmount);
                        }
                    }
                }

                // Update the coordinates of the nearest point to the mouse.
                String xCoord = "";
                DecimalFormat form = new DecimalFormat(m_bigStatusBar.getFormat());

                // Look for the nearest point on a curve to the mouse.
                Graph graph = m_page.getNearestChild(e.getX(), e.getY());
                Point2D.Double p = null;
                if (graph != null) {
                    p = graph.getNearestPoint(e.getX(),
                            Math.round(m_page.getHeight()) - e.getY());
                }

                // Write zero on the status bar if there was no nearest point.
                if (p == null) {
                    p = new Point2D.Double(0, 0);
                    xCoord = form.format(p.x);
                } else {
                    xCoord = formatXCoord(p.x, graph);
                }

                // Send the newly determined nearest coordinates to the status bar.
                m_grStatusBar.nearestXLabel.setText(xCoord);
                m_grStatusBar.nearestYLabel.setText(form.format(p.y));

                // Save the current mouse position for future use.
                // If the modification was rejected because the movement was too
                // small for the snap increment, don't update the mouse position.
                // That way, changes can accumulate until they are large enough
                // to snap to the next increment.
                if (!xTooSmall) {
                    m_mouseX = e.getX();
                }
                if (!yTooSmall) {
                    m_mouseY = e.getY();
                }
            }

            // Trigger a repaint of the updated contents.
            m_pageProxy.repaint();
        }
    }

//------------------------------------------------------------------------------
    /**
     * Handle mouse presses within the page area.  Set up variables for
     * comparison during later mouse drag and release events.
     * @param e     The mouse event object containing the mouse position.
     */
    public void mousePressed(MouseEvent e) {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }
        synchronized (m_page) {

            if (m_mouseHandler != null) {
                m_mouseHandler.mousePressed(e);
            }

            // Save the mouse position for future comparisons.
            m_mouseX = e.getX();
            m_mouseY = e.getY();

            // Find the graph nearest to the mouse.
            Graph graph0 = m_page.getNearestChild(e.getX(), e.getY());
            if (graph0 instanceof Gl3dGraph) {
                ((Gl3dGraph) graph0).mousePressed(e);
                m_page.selectGraph(m_mouseX, m_mouseY, true, true);
                return;
            }

            // Set the graphs' side bars on or off.
            boolean bSideBarChanged = m_page.setSideBar(e.getX(), e.getY(), true);
            if (bSideBarChanged) {
                m_pageProxy.repaint();
            }

            // Check if the Shift, Ctrl, Alt, or right mouse were down.
            boolean bCtrl = ((e.getModifiers() & e.CTRL_MASK) != 0);
            boolean bShift = ((e.getModifiers() & e.SHIFT_MASK) != 0);
            boolean bAlt = ((e.getModifiers() & e.ALT_MASK) != 0);
            boolean bButton3 = ((e.getModifiers() & e.BUTTON3_MASK) != 0);

            // If the status bar is set to show the last curve clicked on...
            if (m_bigStatusBar.m_bSelectWithClick &&
                    m_grStatusBar == m_bigStatusBar) {

                // Initialize the X-coordinate text string and the formatter.
                String xCoord = "";
                DecimalFormat form =
                        new DecimalFormat(m_bigStatusBar.getFormat());
                Point2D.Double p = null;

                // Find the graph nearest to the mouse.
                Graph graph = m_page.getNearestChild(e.getX(), e.getY());

                // Check which coordinate mode the status bar is in.
                if (m_grStatusBar.m_coordLevelIndex == 0) {
                    // Look for the nearest point on a curve to the mouse.
                    if (graph != null) {
                        p = graph.getNearestPoint(e.getX(),
                                Math.round(m_page.getHeight()) - e.getY());
                    }
                } else {
                    // Convert the screen pixel coordinates to user space
                    // coordinates.
                    Rectangle screenRect = new Rectangle(e.getX(), e.getY(), 0, 0);
                    Rectangle2D.Double userRect =
                            m_page.screenToUser(screenRect, graph);

                    if (userRect != null) {
                        p = new Point2D.Double(userRect.x, userRect.y);
                    }
                }

                // Write zero for the X-coordinate if there was no nearest point.
                if (p == null) {
                    p = new Point2D.Double(0, 0);
                    xCoord = form.format(p.x);
                } else {
                    xCoord = formatXCoord(p.x, graph);
                }

                m_grStatusBar.nearestXLabel.setText(xCoord);
                m_grStatusBar.nearestYLabel.setText(form.format(p.y));

                // Display the nearest curve on the status bar.
                m_bigStatusBar.setCurve(m_page.getNearestCurve(
                        e.getX(), e.getY()));
            }

            // If selecting points to modify...
            if (!m_bDrawMode && !m_bPanMode && !m_bZoomMode && !bButton3) {
                // Update the selection rectangle that is drawn over the page.
                selectWindow(e.getX(), e.getY(), bCtrl, bShift, bAlt);

                // Select the graph under the mouse and show points on all the
                // curves on that graph.
                m_page.selectGraph(m_mouseX, m_mouseY, true, true);
                showEditingSymbols(true);

                // If a point was under the mouse when pressed...
                if (m_page.containsSelectedPoint(new Rectangle(e.getX() - 2,
                        e.getY() - 2, 5, 5), false)) {

                    // Prepare to drag selected points with the mouse on later
                    // drag events.
                    m_bDragMode = true;

                    // Before letting the points be modified, save the selected
                    // curves for possible undo later.
                    DataSnapShot snapShot = new DataSnapShot();
                    Iterator iter = m_page.getSelectedSeriesList().iterator();
                    while (iter.hasNext()) {
                        DataSeries series = (DataSeries) iter.next();
                        snapShot.store(series);
                    }
                    saveDataSnapShot(snapShot);

                    // Set the status bar mode for the upcoming drag events.
                    m_grStatusBar.modeLabel.setText("Drag");
                }
            } else {
                if (bShift && m_bPanMode == true) {
                    // Zooming in on a rectangular area...
                    m_bZoomMode = true;
                    m_bPanMode = false;
                    m_grStatusBar.modeLabel.setText("Zoom");
                }
                if (bCtrl && m_bPanMode == true) {
                    // Multiple graph selections...
                    // Tell the page to select the proper graph.
                    m_page.selectGraph(m_mouseX, m_mouseY, false, !bAlt);
                } else {
                    // Adding a new point in draw mode...
                    m_page.selectGraph(m_mouseX, m_mouseY, true, !bAlt);
                    if (m_bSelectMode) {
                        showEditingSymbols(true);
                    }

                    // Add a point to a new curve that is being drawn by the user.
                    if (m_bDrawMode) {
                        // If this is the first click inside this graph, set up
                        // new curve for the graph.
                        if (m_drawModeCurve == null) {
                            m_drawModeIndex = -1;
                            // Determine if an existing curve should be used.
                            List selectedList = m_page.getSelectedSeriesList();
                            if (selectedList.size() == 1 &&
                                    DoubleDataSeries.countDims(
                                    selectedList.get(0), 2)) {

                                DoubleDataSeries series =
                                        (DoubleDataSeries) selectedList.get(0);
                                int numPoints = series.getNumDimensions();
                                int numSelected = series.getNumSelected();
                                if (numSelected == 1 || numSelected == 2) {
                                    int[] selIndexes =
                                            series.getSelectedIndexes();
                                    boolean bAppend = false;
                                    if (numSelected == 1) {
                                        if (selIndexes[0] == 0) {
                                            m_drawModeIndex = 0;
                                        }
                                        if (selIndexes[0] == series.getNumPoints() - 1) {
                                            bAppend = true;
                                        }
                                    }
                                    if (numSelected == 2) {
                                        m_drawModeIndex = selIndexes[1] * numPoints;
                                    }

                                    if (m_drawModeIndex >= 0 ||
                                            selIndexes[0] == series.getNumPoints() - 1) {

                                        List observers = series.getObservers();
                                        Iterator iter = observers.iterator();
                                        while (iter.hasNext()) {
                                            GrObject gro = (GrObject) iter.next();
                                            if (gro instanceof GlCurve) {
                                                m_drawModeCurve = (GlCurve) gro;
                                                break;
                                            }
                                        }
                                        if (m_drawModeCurve != null) {
                                            observers = m_drawModeCurve.getObservers();
                                            iter = observers.iterator();
                                            while (iter.hasNext()) {
                                                GrObject gro = (GrObject) iter.next();
                                                if (gro instanceof GlGraph) {
                                                    m_drawModeGraph = (GlGraph) gro;
                                                    break;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            if (m_drawModeCurve == null) {
                                // Allow the new series to be edited in both directions.
                                DoubleDataSeries newSeries =
                                        new DoubleDataSeries(2);
                                newSeries.getDimension(0).setDimensionLocked(false);
                                newSeries.getDimension(1).setDimensionLocked(false);
                                m_drawModeCurve = new GlCurve(newSeries);

                                // Check each graph that is showing to see if the mouse
                                // is over it.
                                m_drawModeGraph = null;
                                List showingGraphs = m_page.getShowList();
                                Iterator iter = showingGraphs.iterator();
                                while (iter.hasNext()) {
                                    Graph graph = (Graph) iter.next();
                                    if (graph.isMouseOver(m_mouseX,
                                            (m_page.getHeight() - m_mouseY))) {

                                        m_drawModeGraph = graph;
                                        break;
                                    }
                                }

                                // Make sure the curve that is being drawn is added
                                // to the graph.
                                if (m_drawModeGraph != null) {
                                    m_drawModeGraph.add(m_drawModeCurve);
                                }
                            }
                        }

                        if (m_drawModeGraph != null) {
                            // Get the data series for the draw curve.
                            DoubleDataSeries series =
                                    m_drawModeCurve.getSeries();
                            double[] oldArray = series.getArray();
                            int oldSize = oldArray.length;
                            int numDims = series.getNumDimensions();

                            // Create a new array to hold the curve data.
                            // Copy the old data over and leave an X and Y
                            // opening at the end for the new point.
                            double[] newArray = new double[oldSize + 2];
                            for (int i = 0; i < m_drawModeIndex; i++) {
                                newArray[i] = oldArray[i];
                            }

                            int shiftAmount = 0;
                            int insertIndex = m_drawModeIndex;
                            if (m_drawModeIndex < 0) {
                                insertIndex = 0;
                            } else {
                                shiftAmount = numDims;
                            }
                            for (int i = insertIndex; i < oldSize; i++) {
                                newArray[i + shiftAmount] = oldArray[i];
                            }
                            if (m_drawModeIndex < 0) {
                                insertIndex = oldSize;
                            }

                            // Calculate the position of the graph user area.
                            Rectangle rect = m_drawModeGraph.getDisplayRect();
                            int yPos = m_page.getHeight() - m_mouseY;

                            // If the mouse was pressed in the same graph as where
                            // the curve started being drawn...
                            if (m_drawModeGraph.isMouseOver(m_mouseX, yPos)) {
                                if (oldSize > 0) {
                                    // Save the current line for later undo.
                                    DataSnapShot snapShot = new DataSnapShot();
                                    snapShot.store(series);
                                    saveDataSnapShot(snapShot);
                                }

                                // Append the most recent point to the end of the
                                // curve's series' data array.
                                Point2D.Double p = m_drawModeGraph.getCoords(
                                        m_mouseX - rect.x, yPos);
                                newArray[insertIndex] = p.x;
                                newArray[insertIndex + 1] = p.y;

                                if (m_drawModeIndex > 0) {
                                    m_drawModeIndex += 2;
                                }

                                // Pass the new array to the series as long as it
                                // is not a duplicate of the previous point.  This
                                // is meant to filter out multiple mouse click
                                // events.
                                if (oldSize < 2 ||
                                        (newArray[oldSize] != newArray[oldSize - 2]) ||
                                        (newArray[oldSize + 1] != newArray[oldSize - 1])) {
                                    series.setArray(newArray);
                                }
                            } // If the mouse was pressed in a new graph...
                            else {
                                // Remove the existing draw curve if it has fewer
                                // than two points.
                                if (m_drawModeCurve.getSeries().getCount() < 2) {
                                    GrObject gro =
                                            m_page.findParentOf(m_drawModeCurve);
                                    if (gro != null) {
                                        gro.remove(m_drawModeCurve);
                                    }
                                }
                                m_drawModeCurve = null;
                            }
                        } // The draw curve is not attached to any graph...
                        else {
                            // Remove the curve if it has fewer than two points.
                            if (m_drawModeCurve.getSeries().getCount() < 2) {
                                GrObject gro = m_page.findParentOf(m_drawModeCurve);
                                if (gro != null) {
                                    gro.remove(m_drawModeCurve);
                                }
                            }
                            m_drawModeCurve = null;
                        }
                    }
                }
            }

            // Set the cursor.
            if (m_bPanMode == true) {
                m_pageProxy.getComponent().
                        setCursor(new Cursor(Cursor.MOVE_CURSOR));
            }

//        if (m_bPanMode || m_bZoomMode) {
//        	// Save the page so changes can be undone later.
//        	savePageSnapShot();
//        }

            // Trigger a repaint of the updated contents.
            m_pageProxy.repaint();
        }
    }

//------------------------------------------------------------------------------
    /**
     * Handle mouse being released within the page area.  Look at variables that
     * were set during earlier mouse press and drag events.
     * @param e     The mouse event.
     */
    public void mouseReleased(MouseEvent e) {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }
        synchronized (m_page) {

            if (m_mouseHandler != null) {
                m_mouseHandler.mouseReleased(e);
            }

            // Set the graphs' side bars on or off.
            boolean bSideBarChanged = m_page.setSideBar(e.getX(), e.getY(), false);
            if (bSideBarChanged) {
                m_pageProxy.repaint();
            }

            // Check if the Shift, Ctrl, Alt, or right mouse were down.
            boolean bCtrl = ((e.getModifiers() & e.CTRL_MASK) != 0);
            boolean bShift = ((e.getModifiers() & e.SHIFT_MASK) != 0);
            boolean bAlt = ((e.getModifiers() & e.ALT_MASK) != 0);
            boolean bButton3 = ((e.getModifiers() & e.BUTTON3_MASK) != 0);

            // If not in zoom or select mode.
            if (m_bZoomMode == false && m_bSelectMode == false) {
                // Set the cursor back to the default.
                m_pageProxy.getComponent().setCursor(new Cursor(
                        Cursor.DEFAULT_CURSOR));
            } else {
                // If in zoom mode...
                if (m_bZoomMode == true) {
                    // Perform the zoom operation that was defined by the previous
                    // mouse press and drag events.
                    m_page.zoomWindow();
                    m_bZoomMode = false;
                    if (m_bSelectMode == false) {
                        // Turn zoom mode off and revert to the earlier mode.
                        m_bPanMode = true;
                        m_grStatusBar.showPanMode(true);
                        m_grMenuBar.showPanMode(true);
                    } else {
                        m_grStatusBar.modeLabel.setText("Select");
                    }
                    // Trigger a repaint of the updated contents.
                    m_pageProxy.repaint();
                } // If in select mode...
                else {
                    if ((m_bDragMode && !bAlt) || bButton3) {
                        // Drag mode is on and not deselecting.  Update status bar.
                        m_grStatusBar.modeLabel.setText("Select");
                    } else {
                        if (!m_bDrawMode) {
                            // If not in draw mode, select the points that was
                            // defined by the previous mouse press and drag events.
                            selectWindow(e.getX(), e.getY(), bCtrl, bShift, bAlt);
                        }
                    }
                    // Trigger a repaint of the updated contents.
// let dataseries selectArea send update instead ???--
                    m_pageProxy.repaint();
                }
            }

            // Turn off drag mode until the next mouse press.
            m_bDragMode = false;
        }
    }

//------------------------------------------------------------------------------
    /**
     * Handle the mouse leaving the drawing area.
     * @param e     The mouse event.
     */
    public void mouseExited(MouseEvent e) {

//        System.out.println("Mouse exited at " + e.getX() + ", " + e.getY());
        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }
        synchronized (m_page) {

            if (m_mouseHandler != null) {
                m_mouseHandler.mouseExited(e);
            }

            // Set the graphs' side bars on or off.
            boolean bSideBarChanged = m_page.setSideBar(e.getX(), e.getY(), false);
            if (bSideBarChanged) {
                m_page.getFormatter().layoutGraphs();
                m_pageProxy.repaint();
            }
        }
    }

//------------------------------------------------------------------------------
    /**
     * Selects points on the page based on the coordinates and key modifiers.
     * Updates the selection rectangle for the next page redraw.
     * @param xMax	The new x coordinate of the corner of the selection window.
     * @param yMax	The new y coordinate of the corner of the selection window.
     * @param bCtrl	True if the Control key is held down on the key board.
     * @param bShift	True if the Shift key is held down on the key board.
     * @param bAlt	True if the Alt key is held down on the key board.
     */
    public void selectWindow(int xMax, int yMax, boolean bCtrl, boolean bShift,
            boolean bAlt) {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // If the mouse hasn't moved since it was pressed, select an
        // area (default 6x6) centered on the mouse location.
        int mouseX = m_mouseX;
        int mouseY = m_mouseY;

        int halfSpan = (int) Math.ceil(m_editablePointSize / 2);

        if (xMax == mouseX) {
            mouseX -= halfSpan;
            mouseY -= halfSpan;
            xMax += halfSpan;
            yMax += halfSpan;
        }

        m_page.deferUpdates();

        // Update the selection rectangle.
        m_page.setZoomWindow(mouseX, mouseY, xMax, yMax);

        // Select the points within the selection region.
        m_page.selectWindow(!bAlt, !bShift, !(bAlt || bCtrl));

        m_page.processDeferredUpdates();
    }

//------------------------------------------------------------------------------
    /**
     * Finds the frame that is ultimately the parent of the given component.
     * @param component The component to begin searching upward from.
     * @return  The frame that was eventually found, or a new frame if none.
     */
    public static Frame getParentFrame(Component component) {

        // Initialize the frame to null.
        Frame frame = null;

        // Return null if the component is null;
        if (component == null) {
            return new Frame();
        }

        // Ascend the hierarchy of containers until the parent frame is found.
        Container container = component.getParent();
        while (container != null && !(container instanceof Frame)) {
            container = container.getParent();
            if (container instanceof Frame) {
                frame =
                        (Frame) container;
            }
        }

        // If no frame was found, return a new frame.
        if (frame == null) {
            return new Frame();
        } else {
            return frame;
        }
    }

//------------------------------------------------------------------------------
    /**
     * Turns editing symbols on or off for the curves on the selected graphs.
     * @param bShow Whether to show (true) or hide (false) the symbols.
     */
    public void showEditingSymbols(boolean bShow) {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Make the selected graph show symbols, all others don't.
        Iterator graphIter = m_page.getGraphList().iterator();
        while (graphIter.hasNext()) {
            Graph graph = (Graph) graphIter.next();

            // Change each curve in the graph.
            List curveList = graph.getCurves();
            Iterator curveIter = curveList.iterator();
            while (curveIter.hasNext()) {
                Curve curve = (Curve) curveIter.next();
                if (graph.getSelected() && bShow) {
                    curve.setEditing(true);
                } else {
                    curve.setEditing(false);
                }
            }
        }
    }

//------------------------------------------------------------------------------
    /**
     * Save the current state of the data so it can be restored later.
     * @param shot	The snapshot to be stored.
     */
    public void saveDataSnapShot(DataSnapShot shot) {

        // Find out how many levels of undo are specified in the user prefs.
        int undoLevels = m_guiOptions.getUndoLevels();

        // If undo is enabled...
        if (undoLevels > 0) {
            // If the list is already holding as many levels of undo as
            // allowed, delete the oldest one.
            if (m_undoList.getChildCount() >= undoLevels) {
                m_undoList.remove(0);
            }

            // Store the latest snapshot.
            m_undoList.add(shot);
        }

        // Clear the redo list.
        m_redoList.clear();
    }

//------------------------------------------------------------------------------
    /**
     * Save the current page so it can be restored later.
     */
    public void savePageSnapShot() {

        // Don't proceed if there is no current page.
        if (m_page == null) {
            return;
        }

        // Find out how many levels of undo are specified in the user prefs.
        int undoLevels = m_guiOptions.getUndoLevels();

        // If undo is enabled...
        if (undoLevels > 0) {
            // If the list is already holding as many levels of undo as
            // allowed, delete the oldest one.
            if (m_undoList.getChildCount() >= undoLevels) {
                m_undoList.remove(0);
            }

            // Store a copy of the page.
            if (m_page instanceof PageOfGlGraphs) {
                synchronized (m_page) {
                    m_undoList.add(new PageOfGlGraphs(m_page));
                }
            }
//        	else {
///j2d        		m_undoList.addElement(new PageOfJ2dGraphs(m_page));
//	        }
        }

        // Clear the redo list.
        m_redoList.clear();
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return Whether the X-axis is locked.
     */
    public boolean getLockX() {

        return m_bLockX;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return Whether the Y-axis is locked.
     */
    public boolean getLockY() {

        return m_bLockY;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return Whether pan mode is on.
     */
    public boolean getPanMode() {

        return m_bPanMode;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return Whether quick draw mode is on.
     */
    public boolean getQuickDraw() {

        return m_bQuickDraw;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return Whether the select mode is on.
     */
    public boolean getSelectMode() {

        return m_bSelectMode;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return Whether zoom mode is on.
     */
    public boolean getZoomMode() {

        return m_bZoomMode;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return Whether drawing should be done with OpenGL.
     */
    public boolean getUsingOpenGL() {

        return m_bUsingOpenGL;
    }

//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param value	True if drawing should be done with OpenGL.
     */
    public void setUsingOpenGL(boolean value) {

        m_bUsingOpenGL = value;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The page of graphs object.
     */
    public PageOfGraphs getPage() {

        return m_page;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The class to use for reading data files.
     */
    public String getDataFormatClass() {

        return m_dataFormatClass;
    }

//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param value	The class to use for reading data files.
     */
    public void setDataFormatClass(String value) {

        m_dataFormatClass = value;
    }

//------------------------------------------------------------------------------
    /**
     */
    public void setStatusBarDateFormat(int dateFormatIndex) {

        m_grStatusBar.setDateFormatIndex(dateFormatIndex);
    }

//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param page	A new page to use for drawing.
     */
    public void setPage(PageOfGraphs page) {

        // If the drawing panel for the page is null, check that everything
        // is cleaned up and return.
        if (m_pageProxy == null) {
            if (m_grMenuBar != null) {
                m_grMenuBar.updateDynamicPanel();
                getParentFrame(m_grMenuBar).setTitle("USGS Gr");
            }
            return;
        }

        // If the old and new pages are not null, stop drawing on the old page.
        if (m_page != null && page != null) {
            page.stopDrawing();
        }

        // Set the page.
        m_pageProxy.setPage(page);
        m_page = page;

        // If it is an OpenGL page, reshape it to the window size.
        if (page instanceof PageOfGlGraphs && m_pageProxy instanceof GlGrPanel) {

            m_page.reshape(m_pageProxy.getComponent().getWidth(),
                    m_pageProxy.getComponent().getHeight());
        }

        // Check if the animation buttons should be shown.
        if (m_grMenuBar != null) {
            m_grMenuBar.updateDynamicPanel();
        }

        // Look for an alternate mouse handler as a child of the current page.
        boolean bFound = false;
        if (m_page != null && page != null) {
            Enumeration enumeration = page.children();
            while (enumeration.hasMoreElements()) {
                GrObject gro = (GrObject) enumeration.nextElement();
                if (gro instanceof GrMouseHandler) {
                    m_mouseHandler = (GrMouseHandler) gro;
                    m_mouseHandler.setControllerAndPage(this, m_page);
                    bFound = true;
                } else {
                    if (gro instanceof GuiPanelProvider) {
                        GuiPanelProvider gpp = (GuiPanelProvider) gro;
                        gpp.setControllerAndPage(this, m_page);
                    }
                }
            }
        }
        if (!bFound) {
            m_mouseHandler = null;
        }

        // Set the name of the window to match the name of the page.
        updateWindowTitle("");
    }

//------------------------------------------------------------------------------
    /**
     * Overrides the GrObject method to add code that causes a redraw of the
     * current page if it needs one.
     */
    public void processDeferredUpdates() {

        super.processDeferredUpdates();

        if (m_page != null) {
            if (m_page.getNeedsRedraw()) {
                // Trigger a reformat and repaint of the updated contents.
                m_page.updateLists();
                m_page.getFormatter().layoutGraphs();
                m_pageProxy.repaint();
            }
        }
    }
}


