/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 *                  California District, Sacramento
 *                  San Francisco Bay Hydrodynamics Project
 * @version         2008-04-24
 * Date started:    2006-03-17
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact Information
 * -------------------
 * John Donovan
 * jmd@usgs.gov
 * (916) 278-3120
 *
 * Placer Hall
 * 6000 J Street
 * Sacramento, CA  95819-6129
 *
 * The Gr web page is at:
 *
 * http://ca.water.usgs.gov/program/sfbay/gr/
 */
package gov.usgs.gr;

import gov.usgs.gr.data.*;
import java.io.*;
import java.util.*;
import gov.usgs.gr.nwis.adaps.rdb.RdbFile;
import java.text.SimpleDateFormat;
import java.text.FieldPosition;
import gov.usgs.sfhydro.gr.tools.GodinFilter;

//==============================================================================
import gov.usgs.sfhydro.gr.tools.TidalFilterAnalyzer;
/**
 * A filtering application that may optionally display a dialog.
 */
public class GrFilt {

    /** A list of files to open from the command line. */
    protected List m_filesToOpen = new ArrayList();
    /** The maximum allowable gap length for the input series, in days. */
    protected double m_allowableGap = 1d / 24d;
    /** Whether to analyze the series instead of filtering. */
    protected boolean m_bAnalyze = false;
    /**
     * A list of objects to be used as default configurations.  The values
     * are read from disk.
     */
    public static GrHashList DEFAULT_OBJ_PROPS = new GrHashList();
    /** The exit code for the process. */
    public int m_exitCode = 0;
    /** Whether to print stack traces when exceptions occur. */
    public boolean m_bPrintStackTraces = false;

//------------------------------------------------------------------------------
    /**
     * Main constructor for the Gr application.
     */
    public GrFilt() {
    }

//------------------------------------------------------------------------------
    /**
     * Setup and run the filter on the data files that were given on the
     * command line.
     */
    public void runFilter() {

        // Get the Gr Home property and set the current file property.
// Uncomment for Netbeans        System.setProperty("gov.usgs.gr.grhome", "c:\\Program Files\\USGS\\Gr");
        String grHome = System.getProperty("gov.usgs.gr.grhome");

        // If the property was not found, set it to a default value.
        if (grHome == null) {
            System.setProperty("gov.usgs.gr.grhome",
                    System.getProperty("user.dir"));
            grHome = System.getProperty("gov.usgs.gr.grhome");
        }

        grHome.replace('\\', '/');
        System.setProperty("gov.usgs.gr.currentfile", grHome);

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        GregorianCalendar date = new GregorianCalendar(utcZone, locale);
        SimpleDateFormat formatter = new SimpleDateFormat("yyyy/MM/dd HH:mm");
        formatter.setTimeZone(utcZone);

        // Define time interval constants.
        double sec30 = 1d / 24d / 60d / 2d;
        double oneHour = 1d / 24d;

        // Declare the filter outside the loop, so it can be reused between
        // iterations.
        GodinFilter gf = null;

        // Look for Gr configuration files in XML format for default
        // object properties.  Create a list of files in the config directory
        // within the Gr home directory that begin with "gr_config" and end
        // with ".xml"
        String[] fileNames = new File(grHome + "/config").list(
                new FilenameFilter() {

                    public boolean accept(File dir, String name) {
                        name = name.toLowerCase();
                        return (name.startsWith("gr_config") && name.endsWith(".xml"));
                    }
                });

        // Read the defaults files.
        if (fileNames == null) {
            fileNames = new String[0];
        }
        for (int i = 0; i < fileNames.length; i++) {
            configureDefaults(grHome + "/config/" + fileNames[i]);
        }

        // Open any files that were specified on the command line.
        Iterator iter = m_filesToOpen.iterator();
        while (iter.hasNext()) {
            String filename = (String) iter.next();

            // Assume the file is in RDB format.
            RdbFile rdbFile = new RdbFile();
            rdbFile.setProperty("file", filename);
            rdbFile.setProperty("requireUTC", "true");

            try {
                // Read the file.
                System.out.println("\nReading RDB input...");
                try {
                    rdbFile.read();

                    if (rdbFile.getError()) {
                        m_exitCode = 1;
                        System.err.println("Error reading RDB file" + filename);
                        return;
                    }
                } catch (Exception e) {
                    m_exitCode = 1;
                    System.err.println("Error reading RDB file" + filename);
                    if (m_bPrintStackTraces) {
                        e.printStackTrace();
                    }
                    return;
                }

                List seriesList = rdbFile.findChildInstancesOf(
                        "gov.usgs.gr.data.DoubleDataSeries", false);
                
                if (m_bAnalyze) {
                    TidalFilterAnalyzer analyzer = new TidalFilterAnalyzer();
                    if (seriesList.size() < 0) {
                        System.err.println("Error retrieving data series from RDB file" + filename);
                        System.err.println("Could not perform analysis");
                    }
                    else {               
                        System.out.println("Analyzing tidal data series...");
                        String stationLabel = rdbFile.getStationNumber()
                                + "  " + rdbFile.getStationName();
                        analyzer.analyze((TimeSeries)seriesList.get(0),
                                stationLabel, filename + ".ps");
                    }
                }
                
                System.out.println("Identifying gaps...");

                // Generate a list of segments between gaps for each series.
                List segmentList = generateSegmentList(seriesList);

                // Look at each series that was read and filter it.
                Iterator seriesIter = seriesList.iterator();
                Iterator segmentIter = segmentList.iterator();
                while (seriesIter.hasNext()) {
                    // Set up the basics for this series.
                    DoubleDataSeries series = (DoubleDataSeries) seriesIter.next();
                    double[] array = series.getArray();
                    int numDims = series.getNumDimensions();
                    int refYear = TimeReference.getReferenceYear(series);
                    List segments = (List) segmentIter.next();
                    ArrayList segmentSeriesList = new ArrayList();
                    int startIndex = 0;
                    int endIndex = 0;
                    int index = 0;

                    // Process each segment individually.
                    Iterator indexIter = segments.iterator();
                    while (indexIter.hasNext()) {
                        // Print gap information on all except the first iter.
                        if (startIndex > 0) {
                            // Determine where the last input gap was.
                            double gapStart =
                                    array[startIndex - numDims] + sec30;
                            double gapEnd = array[startIndex] + sec30;
                            writeGapMessage(refYear, date, gapStart, gapEnd,
                                    formatter);
                        }

                        // Determine the end of the current segment.
                        endIndex = ((Integer) indexIter.next()) * numDims;

                        // Determine the length of the segment.
                        int size = endIndex - startIndex;
                        double timeSpan =
                                array[endIndex - numDims] - array[startIndex];

                        // Only filter it if the time span is at least 71 hours.
                        if (timeSpan > (71d / 24d)) {
                            // Copy the segment to another array.
                            double[] segmentArray = new double[size];

                            for (int i = startIndex; i < endIndex; i++) {
                                segmentArray[i - startIndex] = array[i];
                            }

                            // Make a new data series out of the array for
                            // this segment.
                            DoubleDataSeries segmentSeries =
                                    new DoubleDataSeries(segmentArray, numDims);
                            segmentSeriesList.add(segmentSeries);
                        } else {
                            // If the segment was too short to filter, write
                            // out a message.
                            double startTime = array[startIndex] + sec30;
                            double endTime = array[endIndex - numDims] + sec30;
                            TimeReference.yearDayToGregCal(
                                    refYear, startTime, date);
                            StringBuffer buf = formatter.format(date.getTime(),
                                    new StringBuffer(), new FieldPosition(0));
                            String lineString = "Cannot filter segment of less " + "than 71 hours between ";
                            lineString += buf.toString();
                            lineString += " and ";
                            TimeReference.yearDayToGregCal(
                                    refYear, endTime, date);
                            buf = formatter.format(date.getTime(),
                                    new StringBuffer(), new FieldPosition(0));
                            lineString += buf.toString();
                            System.out.println(lineString);
                        }

                        // Get ready for the segment on the next iteration.
                        startIndex = endIndex;
                        index++;
                    }

                    // Run the Godin filter separately on all the segments of
                    // the series.
                    System.out.println("Running Godin Filter...");
                    if (gf == null) {
                        gf = new GodinFilter();
                    }

                    try {
                        gf.performAction(segmentSeriesList);
                    } catch (Exception e) {
                        m_exitCode = 2;
                        System.err.println("Error running filter.");
                        if (m_bPrintStackTraces) {
                            e.printStackTrace();
                        }
                        return;
                    }

                    // Reform the series using the filtered segments.
                    // The gaps will have increased after being run through
                    // the filter.  Retrieve the data array for each segment
                    // and calculate the size needed for the new array to store
                    // all the filtered segments.
                    Iterator filteredSegmentIter = segmentSeriesList.iterator();
                    double[][] filtSegArray =
                            new double[segmentSeriesList.size()][];
                    index = 0;
                    int totalSize = 0;
                    while (filteredSegmentIter.hasNext()) {
                        DoubleDataSeries filtSeg =
                                (DoubleDataSeries) filteredSegmentIter.next();
                        filtSegArray[index] = filtSeg.getArray();

                        totalSize += filtSegArray[index].length;
                        index++;
                    }

                    // Concatenate the filtered segments into one array.
                    double[] filtConcat = new double[totalSize];
                    index = 0;
                    for (int i = 0; i < filtSegArray.length; i++) {
                        double[] segArray = filtSegArray[i];
                        for (int j = 0; j < segArray.length; j++) {
                            filtConcat[index] = segArray[j];
                            index++;
                        }
                    }

                    // Put the contatenated array into the series.
                    series.setArray(filtConcat);
                }

                // Write the output to a file.
                System.out.println("Writing output...");
                try {
                    rdbFile.setProperty("file", filename + ".grfilt");
                    rdbFile.setProperty("quoteToWrite", "\"");
                    rdbFile.write();
                } catch (Exception e) {
                    m_exitCode = 3;
                    System.err.println("Error writing output");
                    if (m_bPrintStackTraces) {
                        e.printStackTrace();
                    }
                    return;
                }

                System.out.println("Complete.");
            } catch (Exception e) {
                System.err.println("Error in GrFilt");
                if (m_bPrintStackTraces) {
                    e.printStackTrace();
                }
                m_exitCode = 4;
                return;
            }
        }
    }

//------------------------------------------------------------------------------
    /**
     * Write a message describing the gap in the input series, as well as the gap
     * it caused in the output series.
     * @param refYear   The reference year.
     * @param date      A GregorianCalendar object to use for Julian day conversion.
     * @param gapStart  The start of the gap as a Julian decimal.
     * @param gapEnd    The end of the gap as a Julian decimal.
     * @param formatter The formatter to use for writing a date.
     */
    public void writeGapMessage(int refYear, GregorianCalendar date,
            double gapStart, double gapEnd, SimpleDateFormat formatter) {

        // Define constants.
        double sec30 = 1d / 24d / 60d / 2d;
        double oneHour = 1d / 24d;

        // Convert to a date string and write out.
        TimeReference.yearDayToGregCal(
                refYear, gapStart, date);
        StringBuffer buf = formatter.format(date.getTime(),
                new StringBuffer(), new FieldPosition(0));
        String lineString = "-- gap in  input from ";
        lineString += buf.toString();
        lineString += " to ";
        TimeReference.yearDayToGregCal(
                refYear, gapEnd, date);
        buf = formatter.format(date.getTime(),
                new StringBuffer(), new FieldPosition(0));
        lineString += buf.toString();
        lineString += " will cause";
        System.out.println(lineString);

        // Determine where the last output gap was.
        gapStart -= (35d / 24d);
        gapEnd += (35d / 24d);
        gapEnd += (oneHour + sec30);

        // Convert to a date string.
        TimeReference.yearDayToGregCal(
                refYear, gapStart, date);
        date.set(GregorianCalendar.MINUTE, 0);
        date.set(GregorianCalendar.SECOND, 0);
        date.set(GregorianCalendar.MILLISECOND, 0);
        buf = formatter.format(date.getTime(),
                new StringBuffer(), new FieldPosition(0));
        lineString = "   gap in output from ";
        lineString += buf.toString();
        lineString += " to ";
        TimeReference.yearDayToGregCal(
                refYear, gapEnd, date);
        date.set(GregorianCalendar.MINUTE, 0);
        date.set(GregorianCalendar.SECOND, 0);
        date.set(GregorianCalendar.MILLISECOND, 0);
        buf = formatter.format(date.getTime(),
                new StringBuffer(), new FieldPosition(0));
        lineString += buf.toString();

        // Append the gap length and write out.
        gapEnd -= (gapEnd % oneHour);
        gapStart -= (gapStart % oneHour);
        double gapLength = gapEnd - gapStart + sec30;
        int numDays = (int) gapLength;
        int numHours =
                (int) ((gapLength - (double) numDays) * 24);
        lineString += " (";
        lineString += ("" + numDays + " days");
        if (numHours > 0) {
            lineString += (", " + numHours + " hour");
            if (numHours != 1) {
                lineString += "s";
            }
        }
        lineString += ")\n";
        System.out.println(lineString);
    }

//------------------------------------------------------------------------------
/*
     * Generate a list of gaps for each series in the list.  Each gap is marked by
     * the index of the point that follows it.
     * @param seriesList    The list of input series to be searched for gaps.
     * @return  A list of lists, each with Integer values indicating the indexes of
    points preceding gaps.
     */
    public List<List<Integer>> generateSegmentList(List seriesList) {

        // By default, define a gap as 1% more than the allowable gap.
        double gapPlusBuffer = m_allowableGap * 1.01d;

        // The Godin interval of one hour.
        double godinInterval = 1d / 24d;

        // Time differences below 1/10 millisecond won't be considered.
        double tinyInc = 1d / 24d / 60d / 60d / 10000d;

        // Create a list to hold a list of segments for each series.
        ArrayList<List<Integer>> segmentList = new ArrayList<List<Integer>>();

        // Look at each series in the input list.
        Iterator iter = seriesList.iterator();
        while (iter.hasNext()) {
            // Get the data array and point count.
            DoubleDataSeries series = (DoubleDataSeries) iter.next();
            double[] array = series.getArray();
            int numPoints = series.getCount();
            int numDims = series.getNumDimensions();

            // Create a list to store segment ranges.
            ArrayList<Integer> segments = new ArrayList<Integer>();

            // Initialize the lastTime variable to the first Godin point after
            // the start of the series.
            double firstTimeValue = 0;
            if (numPoints > 0) {
                firstTimeValue = array[0];
            }
            double lastTime = firstTimeValue -
                    Math.abs(firstTimeValue % godinInterval) + godinInterval;

            for (int i = 1; i < numPoints; i++) {
                // Look at each time.
                double time = array[i * numDims];
                double timeDiff = time - lastTime;
                // If enough time was skipped to be considered a gap, mark
                // the end of the last segment (beginning of gap).
                if (timeDiff > gapPlusBuffer) {
                    // Look back at the previous Godin point.
                    double amountPastPreviousGodinPoint =
                            Math.abs(time % godinInterval);
                    if (amountPastPreviousGodinPoint < tinyInc) {
                        amountPastPreviousGodinPoint = godinInterval;
                    }
                    double gTime = time - amountPastPreviousGodinPoint;

                    // Only mark it if the gap spanned an hour point, which is
                    // what Godin uses for input.
                    if (gTime > lastTime && gTime < time) {
                        segments.add(i);
                    }
                }

                // The next lastTime becomes the current time.
                lastTime = time;
            }

            // Mark the end of the series as the end of the last segment.
            segments.add(numPoints - 1);

            // Put the segments into the list of segment lists.
            segmentList.add(segments);
        }

        return segmentList;
    }

//------------------------------------------------------------------------------
/*
     * Reads the specified file to get default object attributes.
     * @param fileName  The name of the file to read.  It must be XML.
     */
    public void configureDefaults(String fileName) {

        File configFile = new File(fileName);
        if (configFile.canRead()) {
            ConfigXMLFile xmlConfig = new ConfigXMLFile();
            xmlConfig.setProperty("file", configFile.getAbsolutePath());
            try {
                // Read the file and get its class propery hash map.
                xmlConfig.read();
                GrHashList newClassMap = xmlConfig.getClassProps();

                // Look at each class in the new map.
                Iterator iter = newClassMap.keySet().iterator();
                while (iter.hasNext()) {
                    String className = (String) iter.next();

                    // Get the old and new propery maps for the class
                    GrHashList masterPropMap =
                            (GrHashList) DEFAULT_OBJ_PROPS.get(className);
                    Object testObj = newClassMap.get(className);
                    GrHashList newPropMap =
                            (GrHashList) newClassMap.get(className);

                    if (masterPropMap == null) {
                        // Assign the new map if there wasn't an old one.
                        DEFAULT_OBJ_PROPS.put(className, newPropMap);
                    } else {
                        // Add the new properties to the old map.
                        Iterator propIter = newPropMap.keySet().iterator();
                        while (propIter.hasNext()) {
                            String propName = (String) propIter.next();
                            masterPropMap.put(propName,
                                    newPropMap.get(propName));
                        }
                    }
                }
            } catch (Exception e) {
                if (m_bPrintStackTraces) {
                    e.printStackTrace();
                }
            }
        }
    }

//------------------------------------------------------------------------------
/*
     * Parses the command line and opens each file that is listed.
     * @param args	An array of command line arguments.
     */
    public void parseCommandLine(String args[]) {

        // Look at each argument.
        for (int i = 0; i < args.length; i++) {
            // Check for the OpenGL option.
            String lowerArg = args[i].toLowerCase();

            // Look for an argument setting the allowable gap size in days.
            if (lowerArg.startsWith("-gap=")) {
                try {
                    String gapString = lowerArg.substring(5);
                    m_allowableGap = TimeReference.convertToDecimalDay(
                            gapString);
                    continue;
                } catch (Exception e) {
                    System.err.println("Bad argument value: " + args[i]);
                    if (m_bPrintStackTraces) {
                        e.printStackTrace();
                    }
                }
            }
            else if (lowerArg.startsWith("-printstack")) {
                m_bPrintStackTraces = true;
            }
            else if (lowerArg.startsWith("-a")) {
                m_bAnalyze = true;
            }
            else {
            // If the argument isn't an option, assume it is a file name
            // and add it to the list of files to be opened.
            m_filesToOpen.add(args[i]);
            }
        }
    }

//------------------------------------------------------------------------------
/*
     * Return the exit code.
     * @return The exit code as an integer.
     */
    public int exitCode() {

        return m_exitCode;
    }

//------------------------------------------------------------------------------
/*
     * Returns a string describing this version of GrFilt.
     * @return	A string describing this version of GrFilt.
     */
    public static String getVersion() {
        return ("2008-04-24");
    }

//------------------------------------------------------------------------------
/*
     * Create an instance of the GrFilt program and run the filter.
     */
    static public void main(String args[]) {

        GrFilt grFilt = new GrFilt();
        
        try {
            for (int i = 0; i < args.length; i++) {
                // Check for the OpenGL option.
                String lowerArg = args[i].toLowerCase();

                // Look for an argument for writing the version info.
                if (lowerArg.startsWith("-v")) {
                    System.out.println("USGS Gr version " + Gr.getVersion());
                    System.out.println("Gr Time Series Filter version "
                            + GrFilt.getVersion());
                    break;
                }
            }
                
            grFilt.parseCommandLine(args);
            grFilt.runFilter();

            int exitCode = grFilt.exitCode();
            if (exitCode != 0) {
                System.err.println("GrFilt exited before completion.");
            }

            System.exit(exitCode);
        } catch (Exception e) {
            //if (grFilt.m_bPrintStackTraces)
                e.printStackTrace();

            System.exit(100);
        }
    }
}











