/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 *                  California District, Sacramento
 *                  San Francisco Bay Hydrodynamics Project
 * @version         2006-06-09
 * Date started:    2006-04-25
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact Information
 * -------------------
 * John Donovan
 * jmd@usgs.gov
 * (916) 278-3120
 *
 * Placer Hall
 * 6000 J Street
 * Sacramento, CA  95819-6129
 *
 * The Gr web page is at:
 *
 * http://ca.water.usgs.gov/program/sfbay/gr/
*/

package gov.usgs.gr;

import java.io.*;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import java.util.jar.JarFile;

//==============================================================================
/**
 * Typically loaded by Java Web Start to install the Gr application.  Asks the
 * user where to install the files, then creates the directories and copies
 * files from a jar resource.  Automatically generates a batch file customized
 * for the location of Gr and the JRE.
 *
 * Limitations and Shortcomings:
 *
 * Can't put an icon on the desktop.  JWS can create an icon, but as of
 * 1.5.0_07, it doesn't pass arguments to the application, specifically the
 * names of files dropped on it.  If JWS can do this in the future, the
 * installer could run as a special case, called by Gr the first time it was
 * run.  This was submitted to Sun as a bug on 2006/05/24.  At JavaOne,
 * Andy Herrick told me to submit it because it might be an easy fix and could
 * potentially be back-ported from Mustang to Tiger.
 *
 * If the installer is run subsequent times to apply updates or reinstall, it
 * doesn't remember where the first installation is.  Need to find out if there
 * is a way to store this through JWS so it can be retrieved during the update.
 * Currently, the installation location reverts to the default, and the user
 * must type in the actual path.
 *
 * Config files are not copied.  If they are, the user will need to be prompted
 * for how to overwrite existing ones.  The X-noX system makes it hard to
 * overwrite.  For example, say Xfeature.xml exists, but the new distribution
 * contains a new and improved version in feature.xml that is worth being
 * enabled.  We don't know if the old was turned off because the feature wasn't
 * wanted, or because it wasn't good enough.  If the enabled-state differs
 * between old and new, ask the user whether to enable.  Example:
 * [ ] disabled (existing state)
 * [x] enabled (default state)
 *
 * If existing config file is as new or newer than its counterpart in the
 * distribution, don't overwrite and don't ask about it.  The existing system
 * requires the user to know something about the config process, which may not
 * be desireable to all users.  A redesign may be needed.
 *
 * If JWS is used without the installer, a location for config files would need
 * to be designated.  Perhaps $USER_HOME/.gr/config ?  If it does not exist at
 * startup, offer to copy default config from jar.
 *
 * If JWS is used without the installer, grfilt and other batch files can't
 * access the jars.  RFE was filed with Sun on 2006/06/30 to allow JWS-installed
 * jars to be accessed through -cp or -Djava.ext.dirs= options on the java
 * command line.  RFE tracking system under Bug Id: 6447147
 *
 * If Gr needs to dynamically load classes from jars not specified on the
 * command line, the code example is in
 * C:\jmd\USGS\Gr\Notes\Sun_Solutions\DynamicClassLoading
 */
public class GrInstaller {

    /** The configuration dialog shown to the user. */
    protected JDialog m_dialog = null;
    /** The Gr home directory, where files will be installed. */
    protected File m_installDir = new File("/Gr");
    /** The label that shows the user the chosen installation directory. */
    protected JTextField m_installDirTextField = new JTextField("/");
    /** Whether the installation finished. */
    protected boolean m_bFinished = false;
    /** Whether the installation experienced errors. */
    protected boolean m_bErrors = false;

//------------------------------------------------------------------------------
/**
 * Main constructor for the Gr installer.
 */
    public GrInstaller () {

        // Determine a default installation location based on the OS.
        if (System.getProperty("os.name").toLowerCase().indexOf("windows")
            >= 0) {

            m_installDir = new File("C:\\Program Files\\USGS\\Gr");
        }
        else {
            m_installDir = new File("/opt/Gr");
        }

        // Show the options dialog.
        m_dialog = new JDialog();
        createDialog();
    }

//------------------------------------------------------------------------------
/**
 * Setup the contents of the dialog and show it.
 */
    public void setupAndShow() {

        // Get the Gr Home property and set the current file property.
        String grHome = "/";

        // If the grhome property is not found, set it to a default value.
        if (grHome == null) {
            System.setProperty("gov.usgs.gr.grhome",
                "c:\\Program Files\\USGS\\Gr");
            grHome = System.getProperty("gov.usgs.gr.grhome");
        }

        grHome.replace('\\', '/');

        // Look for Gr configuration files in XML format for default
        // object properties.  Create a list of files in the config directory
        // within the Gr home directory that begin with "gr_config" and end
        // with ".xml"
        String[] fileNames = new File(grHome + "/config").list(
            new FilenameFilter() {

			public boolean accept(File dir, String name) {
                name = name.toLowerCase();
				return (name.startsWith("gr_config") && name.endsWith(".xml"));
            }
        });

        // Show the dialog.
//    	Dimension d = m_dialog.getToolkit().getScreenSize();
        m_dialog.setMinimumSize(new Dimension(600, 400));
        m_dialog.setLocationRelativeTo(null);
        m_dialog.setVisible(true);
    }

//------------------------------------------------------------------------------
/**
 * Show a dialog telling the user whether installation was successful or not.
 */
    public void showFinishDialog() {

        // Create the message string based on installation status.
        String message = "";
        if (m_bFinished && !m_bErrors)
            message = "Installation finished successfully.";
        if (m_bFinished && m_bErrors)
            message = "Installation finished with errors.";
        if (!m_bFinished && m_bErrors)
            message = "Installation could not finished because of errors.";
        if (!m_bFinished && !m_bErrors)
            message = "Installation could not finish.";

        message += "\n                                                       ";

        // Show the dialog.
        javax.swing.JOptionPane.showMessageDialog(null, message,
            "Installation Report",
            javax.swing.JOptionPane.INFORMATION_MESSAGE);
    }

//------------------------------------------------------------------------------
/**
 * Create the panel that will be displayed to the user.
 */
    public void createDialog() {

        // Define a method to handle the Browse button being clicked.
        java.awt.event.ActionListener browseAL =
            new java.awt.event.ActionListener() {

            public void actionPerformed(ActionEvent e) {
                // Set up the file chooser dialog.
                JFileChooser fc = new JFileChooser();
                fc.setDialogTitle("Choose installation location");
                fc.setCurrentDirectory(m_installDir);
                fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
                fc.setControlButtonsAreShown(false);
                JDialog fcDialog = new JDialog(m_dialog, "Browse location");
                fcDialog.getContentPane().add(fc, "Center");

                JPanel tfPanel = new JPanel();
                tfPanel.setLayout(new BoxLayout(tfPanel,
                    BoxLayout.Y_AXIS));
                tfPanel.add(new JLabel("Create new directory"));
                tfPanel.add(Box.createRigidArea(new Dimension(0,10)));
                JTextField textField = new JTextField();
                tfPanel.add(textField);
                tfPanel.add(Box.createRigidArea(new Dimension(0,20)));
                JPanel okCancelPanel = new JPanel();
                okCancelPanel.setLayout(new BoxLayout(okCancelPanel,
                    BoxLayout.X_AXIS));
                tfPanel.add(Box.createRigidArea(new Dimension(50,0)));

                JButton okButton = new JButton();
                tfPanel.add(okButton);
                tfPanel.add(Box.createRigidArea(new Dimension(20,0)));
                JButton cancelButton = new JButton();
                tfPanel.add(cancelButton);
                tfPanel.add(Box.createRigidArea(new Dimension(20,0)));

                JPanel bottomPanel = new JPanel();
                bottomPanel.setLayout(new BoxLayout(bottomPanel,
                    BoxLayout.X_AXIS));
                bottomPanel.add(Box.createRigidArea(new Dimension(50,0)));
                bottomPanel.add(tfPanel);
                bottomPanel.add(Box.createRigidArea(new Dimension(50,0)));

                fcDialog.getContentPane().add(bottomPanel, "South");
                fcDialog.pack();
                fcDialog.setLocationRelativeTo(null);
                fcDialog.setVisible(true);
//                int returnVal = fc.showOpenDialog(m_dialog);
//                if (returnVal == JFileChooser.APPROVE_OPTION) {
//                    m_installDir = fc.getSelectedFile();
//                    m_installDirTextField.setText(
//                        m_installDir.getAbsolutePath());
//                }
            }
        };

        // General dialog features.
        m_dialog.setTitle("Gr Installation");
        m_dialog.setModal(true);
        m_dialog.setResizable(true);
        m_dialog.getRootPane().setMinimumSize(new Dimension(500, 250));

        // Create a panel to show the path title.
        JPanel installDirPanel = new JPanel();
//        workingDirPanel.setBorder(BorderFactory.createTitledBorder(
//            "Choose Installation Location"));
        installDirPanel.setLayout(new BoxLayout(installDirPanel,
            BoxLayout.Y_AXIS));
        installDirPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
        JLabel descriptionLabel = new JLabel("Choose Installation Location");
        descriptionLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
        installDirPanel.add(Box.createRigidArea(new Dimension(0, 10)));
        installDirPanel.add(descriptionLabel);
        installDirPanel.add(Box.createRigidArea(new Dimension(0, 10)));

        JPanel tfPanel = new JPanel();
        tfPanel.setLayout(new BoxLayout(tfPanel,
            BoxLayout.X_AXIS));
        m_installDirTextField.setText(m_installDir.getAbsolutePath());
        m_installDirTextField.setAlignmentX(Component.TOP_ALIGNMENT);
        m_installDirTextField.setMinimumSize(new Dimension(400, 20));
        m_installDirTextField.setPreferredSize(new Dimension(400, 20));
        tfPanel.add(m_installDirTextField);
        tfPanel.add(Box.createRigidArea(new Dimension(5, 0)));
//
//        JButton browseButton = new JButton("Browse...");
//        browseButton.addActionListener(browseAL);
//        browseButton.setToolTipText("Browse...");
//        dirButtonPanel.add(browseButton);
//        dirButtonPanel.add(Box.createRigidArea(new Dimension(10,0)));

        tfPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
        installDirPanel.add(tfPanel);
        installDirPanel.add(Box.createRigidArea(new Dimension(0, 5)));

        // Create the Install button.
        JButton installButton = new JButton("Install");
        installButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                m_installDir = new File(m_installDirTextField.getText());
                m_dialog.dispose();
            }
        });

        // Add the Install button.
        JPanel bottomPanel = new JPanel();
        bottomPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
        installButton.setAlignmentX(Component.LEFT_ALIGNMENT);
        bottomPanel.add(installButton);

        // Create the Cancel button.
        JButton cancelButton = new JButton("Cancel");
        cancelButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                m_installDir = null;
                m_dialog.dispose();
            }
        });

        // Add the Cancel button.
        cancelButton.setAlignmentX(Component.LEFT_ALIGNMENT);
        bottomPanel.add(Box.createRigidArea(new Dimension(20, 0)));
        bottomPanel.add(cancelButton);
        bottomPanel.add(Box.createRigidArea(new Dimension(10, 0)));

        // Create and populate the main panel
        JPanel mainPanel = new JPanel();
        mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
        mainPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
        mainPanel.add(Box.createRigidArea(new Dimension(0, 20)));
        mainPanel.add(installDirPanel);
        mainPanel.add(Box.createRigidArea(new Dimension(0, 30)));
        mainPanel.add(bottomPanel);
        mainPanel.add(Box.createRigidArea(new Dimension(0, 20)));

        // Create and populate the aboveMain panel.
        JPanel aboveMainPanel = new JPanel();
        aboveMainPanel.setLayout(new BoxLayout(aboveMainPanel,
            BoxLayout.X_AXIS));
        aboveMainPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
        aboveMainPanel.add(Box.createRigidArea(new Dimension(30, 0)));
        aboveMainPanel.add(mainPanel);
        aboveMainPanel.add(Box.createRigidArea(new Dimension(30, 0)));

        // Pack the dialog.
        m_dialog.getContentPane().add(aboveMainPanel);
        m_dialog.pack();
        System.out.println("Size: " + m_dialog.getSize());
    }

//------------------------------------------------------------------------------
/**
 * Run the installation procedure.
 */
    public void install() {

        // Don't proceed if the installation directory is not defined.
        if (m_installDir == null)
            return;

        // Copy resources from the jar to the disk.
        m_installDir.mkdirs();
        copyResourceToFile("gr64X256.ico", m_installDir);
        copyResourceToFile("relnotes.txt", m_installDir);
        File jarDir = new File (m_installDir.getAbsolutePath() + File.separator
            + "lib" + File.separator + "jar");
        jarDir.mkdirs();
        copyResourceToFile("gr.jar", jarDir);
        copyResourceToFile("grres.jar", jarDir);
        copyResourceToFile("jogl.jar", jarDir);
        File nativeDir = new File (m_installDir.getAbsolutePath()
            + File.separator + "lib" + File.separator + "native");
        nativeDir.mkdirs();
        copyResourceToFile("jogl.dll", nativeDir);
        copyResourceToFile("jogl_awt.dll", nativeDir);
        copyResourceToFile("jogl_cg.dll", nativeDir);
        extractJarResourceToFile("ext.jar", m_installDir);

        // Generate the custom batch file.
        try {
            PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(
                m_installDir.getAbsolutePath() + File.separator + "gr.bat")));
            generateBatchFile(pw);
            pw.close();
        }
        catch (Exception ex) {}

        m_bFinished = true;

        // Perform checks.
        checkForResFolder();
        checkForConflictingFiles();
    }

//------------------------------------------------------------------------------
/**
 * Check for JOGL files installed in the wrong place, such as Java ext.
 * Ask the user for permission to delete them.
 */
    public void checkForConflictingFiles() {

        // Look for conflicting files.  Add each one found to the list.
        ArrayList filesFound = new ArrayList();
        File joglFile = new File(System.getProperty("java.home")
            + File.separator + "bin" + File.separator + "jogl.dll");
        if (joglFile.exists())
            filesFound.add(joglFile);
        joglFile = new File(System.getProperty("java.home")
            + File.separator + "bin" + File.separator + "jogl_cg.dll");
        if (joglFile.exists())
            filesFound.add(joglFile);
        joglFile = new File(System.getProperty("java.home")
            + File.separator + "bin" + File.separator + "jogl_awt.dll");
        if (joglFile.exists())
            filesFound.add(joglFile);
        joglFile = new File(System.getProperty("java.home")
            + File.separator + "lib" + File.separator + "ext" + File.separator
            + "jogl.jar");
        if (joglFile.exists())
            filesFound.add(joglFile);

        // If any of the files were found, inform the user.
        if (filesFound.size() > 0) {
            // Create the dialog message.
            String message = "Other Java OpenGL libraries were found on your "
                + "system.\n"
                + "It is recommended that these be removed to prevent a "
                + "conflict.\n";
            Iterator iter = filesFound.iterator();
            while (iter.hasNext()) {
                File file = (File) iter.next();
                message += ("\n" + file.getAbsolutePath());
            }
            message += "\n\nRemove other libraries now?\n";

            // Show the dialog box.
            int choice = JOptionPane.showConfirmDialog(m_dialog,
                message, "Other Libraries Found",
                JOptionPane.YES_NO_OPTION,JOptionPane.QUESTION_MESSAGE);

            ArrayList deniedList = new ArrayList();
            if (choice == JOptionPane.YES_OPTION) {
                // Attempt to delete each incompatable file.
                iter = filesFound.iterator();
                while (iter.hasNext()) {
                    File file = (File) iter.next();
                    try {
                        boolean bSuccess = file.delete();

                        // For any files that failed to be delted, request that
                        // they be deleted upon exit, and save the name so the
                        // user can be informed.
                        if (!bSuccess) {
                            deniedList.add(file);
                            file.deleteOnExit();
                        }
                    }
                    catch (SecurityException se) {
                        deniedList.add(file);
                    }
                }

                // Inform the user of any files that could not be deleted.
                if (deniedList.size() > 0) {
                    // Create the dialog message.
                    String str = "files";
                    if (deniedList.size() == 1)
                        str = "file";
                    message = "Access was denied to the following " + str
                        + ":\n";
                    Iterator deniedIter = deniedList.iterator();
                    while (deniedIter.hasNext()) {
                        File file = (File)deniedIter.next();
                        message += ("\n" + file.getAbsolutePath());
                    }

                    str = "these files";
                    if (deniedList.size() == 1)
                        str = "this file";
                    message += ("\n\nYou may need to delete " + str + " after"
                        + " installation is complete.\n");

                    // Show the dialog box.
                    JOptionPane.showMessageDialog(m_dialog, message,
                        "Access Denied", JOptionPane.ERROR_MESSAGE);
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Check for the resource folder on the disk.
 */
    public void checkForResFolder() {

        // Look for conflicting files.  Add each one found to the list.
        ArrayList filesFound = new ArrayList();
        File resFolder = new File(m_installDir.getAbsolutePath()
            + File.separator + "res");

        // If the folder was found, inform the user.
        if (resFolder.exists()) {
            filesFound.add(resFolder);
            // Create the dialog message.
            String message = "Incompatible files were found from a previous Gr "
                + "installation.\n"
                + "It is recommended that these be removed to prevent a "
                + "conflict.\n";

            message += ("\n" + resFolder.getAbsolutePath());
            message += "\n\nRemove old Gr files now?\n";

            // Show the dialog box.
            int choice = JOptionPane.showConfirmDialog(m_dialog,
                message, "Old Gr Files Found",
                JOptionPane.YES_NO_OPTION,JOptionPane.QUESTION_MESSAGE);

            ArrayList deniedList = new ArrayList();
            if (choice == JOptionPane.YES_OPTION) {
                // Attempt to delete each incompatable file.
                Iterator iter = filesFound.iterator();
                while (iter.hasNext()) {
                    File file = (File) iter.next();
                    try {
                        File returnFile = deleteDir(file);
                        if (returnFile != null) {
                            deniedList.add(returnFile);
                            deniedList.add(file);
                        }
                    }
                    catch (SecurityException se) {
                        deniedList.add(file);
                    }
                }

                // Inform the user of any files that could not be deleted.
                if (deniedList.size() > 0) {
                    // Create the dialog message.
                    String str = "files";
                    if (deniedList.size() == 1)
                        str = "file";
                    message = "Access was denied to the following " + str
                        + ":\n";

                    int counter = 0;
                    Iterator deniedIter = deniedList.iterator();
                    while (deniedIter.hasNext()) {
                        if (counter > 20) {
                            message += ("\n...");
                            break;
                        }
                        File file = (File)deniedIter.next();
                        message += ("\n" + file.getAbsolutePath());
                        counter++;
                    }

                    str = "these files";
                    if (deniedList.size() == 1)
                        str = "this file";
                    message += ("\n\nYou will need to delete " + str + " after"
                        + " installation is complete.\n");

                    // Show the dialog box.
                    JOptionPane.showMessageDialog(m_dialog, message,
                        "Access Denied", JOptionPane.ERROR_MESSAGE);
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * From Java Almanac examples.
 * Deletes all files and subdirectories under dir.
 * If a deletion fails, the method stops attempting to delete and returns the
 * file that could not be deleted.
 * @param dir   The directory to delete
 * @return null if all deletions were successful.
 */
    public static File deleteDir(File dir) {

        if (dir.isDirectory()) {
            File[] children = dir.listFiles();
            if (children == null)
                return dir;
            for (int i = 0; i < children.length; i++) {
                File file = deleteDir(children[i]);
                if (file != null) {
                    return file;
                }
            }
        }

        // The directory is now empty so delete it
        if (dir.delete())
            return null;
        else
            return dir;
    }

//------------------------------------------------------------------------------
/**
 * Copy a resource from the jar to a file on the disk.
 * @param resourceName The name of the resource in the jar file
 * @param destDir The name of the file on the disk.
 */
    public void copyResourceToFile(String resourceName, File destDir) {

        // Get the resource.
        InputStream inStream = this.getClass().getClassLoader().
            getResourceAsStream(resourceName);
        if (inStream == null) {
            System.out.println("Could not open resource: " + resourceName);
        }
        try {
            // Get the output file.
            FileOutputStream outStream = new FileOutputStream(
                destDir.getAbsolutePath() + File.separator + resourceName);

            // Transfer bytes from in to out
            byte[] buffer = new byte[10240];
            int numBytes;
            while ((numBytes = inStream.read(buffer)) > 0) {
                outStream.write(buffer, 0, numBytes);
            }
            inStream.close();
            outStream.close();
        }
        catch (FileNotFoundException ex) {
            m_bErrors = true;
        }
        catch (IOException ex) {
            m_bErrors = true;
        }
    }

//------------------------------------------------------------------------------
/**
 * Get a resource from the jar and put it in a directory on the disk.
 * @param resourceName The name of the resource.
 * @param destDir The destination dierctory.
 */
    public void extractJarResourceToFile(String resourceName, File destDir) {

        // Get the resource as input.
        InputStream inStream = this.getClass().getClassLoader().
            getResourceAsStream(resourceName);
        if (inStream == null) {
            return;
        }

        // Copy the resource to a file on the disk.
        File parentFile = destDir.getParentFile();
        copyResourceToFile(resourceName, parentFile);
        String jarFileName =
            parentFile.getAbsolutePath() + File.separator + resourceName;
        try {
            JarFile jarFile = new JarFile(jarFileName);
            jarFile.close();
        }
        catch (IOException e) {

        }

        // Delete the jar file.
        (new File(jarFileName)).delete();
    }

//------------------------------------------------------------------------------
/**
 * Write a custom batch file to the specified output channel.  The batch file
 * is customized based on the Gr location and JRE location.
 * @param pw The output channel.
 */
    public void generateBatchFile(PrintWriter pw) {

        // Write the contents of the batch file.
        pw.println("set GRHOME=" + m_installDir);
        pw.println("set GRJARHOME=" + m_installDir + File.separator
            + "lib" + File.separator + "jar");
        pw.println("set GRNATIVEHOME=" + m_installDir + File.separator
            + "lib" + File.separator + "native");
        pw.println("set JRE=\"" + System.getProperty("java.home")
            + File.separator + "bin" + File.separator + "java\"");
        pw.println("set PATH=%PATH%;%GRNATIVEHOME%;.");
        pw.println("%JRE% -showversion -Xms32m "
            + "-Xmx1100m -cp \"%GRJARHOME%" + File.separator
            + "gr.jar\";\"%GRJARHOME%" + File.separator
            + "grres.jar\";\"%GRJARHOME%" + File.separator
            + "jogl.jar\" gov.usgs.gr.Gr %*");
    }

//------------------------------------------------------------------------------
/**
 * The main method to start and run the installer.
 *
 * @param args Command line arguments.
 */
    static public void main (String args[]) {

//        String s = "";
//        for (int i = 0; i < args.length; i++) {
//            s += args[i];
//            s += '\n';
//        }
//        javax.swing.JOptionPane.showMessageDialog(null, s,
//            "Command Line Arguments",
//            javax.swing.JOptionPane.INFORMATION_MESSAGE);

//System.getProperties().list(System.out);

        // Create an installer object.
        GrInstaller grInst = new GrInstaller();

        // Show the installer's options dialog and then proceed with
        // installation.
        try {
            grInst.setupAndShow();
            grInst.install();
        } catch (Exception e) {}

        // Tell the user whether the installation was successful.
        grInst.showFinishDialog();

        // Exit the program.
        System.exit( 0 );
    }
}











