/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-03
 * Date started:    2000-06-06
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr;

import java.io.*;
import java.util.*;
import javax.swing.tree.*;
import org.w3c.dom.*;
import javax.swing.ImageIcon;

//==============================================================================
/**
 * An object that has properties and contains other objects in a hierarchy.
 * It can be displayed in a tree or written as XML.
 */
public abstract class GrObject implements javax.swing.tree.TreeNode {
    
    static String XML_ELEMENT_NAME = "obj";
    
    /** The parent element of this region. */
    protected TreeNode m_parent = null;
    /** The child element of this element. */
    protected List<GrObject> m_children =
        Collections.synchronizedList(new LinkedList());
    /** The list of observers for implementing the Observable methods. */
    protected ArrayList<GrObject> m_observers = new ArrayList();
    /** Whether this object has changed and Observers need to be notified. */
    protected boolean m_bHasChanged = false;
    /** Whether to defer updates to be handled later. */
    protected boolean m_bDeferUpdates = false;
    /** Whether to write the ID of this object when saving in XML format. */
    public boolean m_bSaveIdToXml = false;
    /** Whether to write the ID of this object when saving in XML format. */
    public boolean m_bSaveIdOnlyToXml = false;
    /** The tree icon to use for this class. */
    public static ArrayList<ImageIcon> TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GrObject.gif";
    
//------------------------------------------------------------------------------
    /**
     * Construct an empty GrObject.
     */
    public GrObject() {
        
    }
    
//------------------------------------------------------------------------------
    /**
     * Construct an copy of the parameter object.
     */
    public GrObject(GrObject gro) {
        
        gro.copyInto(this, true);
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a copy of this object.
     */
    public synchronized void setDefaults(String className) {
        
        if (!className.equals(this.getClass().getName())) {
            return;
        }
        
        // Create a list of this class' super classes starting at the top
        // and ending with this one.
        Class groClass = null;
        try {
            groClass = Class.forName("gov.usgs.gr.GrObject");
        } catch (ClassNotFoundException e){return;}
        
        Class currentClass = this.getClass();
        List<String> classList = new ArrayList();
        do {
            classList.add(0, currentClass.getName());
            currentClass = currentClass.getSuperclass();
        }
        while (currentClass.isAssignableFrom(groClass) == false);
        
        // Set the defaults for each of those classes.
        for (String propClassName : classList) {
//            setDefaultProps(propClassName);
            GrHashList propMap =
                (GrHashList)Gr.DEFAULT_OBJ_PROPS.get(propClassName);
            if (propMap == null) continue;
            setProperties(propMap);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a copy of this object.
     */
    public GrObject createCopy(boolean bDeep) {
        
        try {
            GrObject gro = (GrObject)(this.getClass().newInstance());
            copyInto(gro, bDeep);
            
            return gro;
        } catch (IllegalAccessException iae) { return null;} catch (InstantiationException ie) {return null;}
    }
    
//------------------------------------------------------------------------------
    /**
     * Copy the properties and children of this object into the parameter object.
     */
    public void copyInto(GrObject gro, boolean bDeep) {
        
        gro.setProperties(getProperties());
        for (GrObject child : m_children) {
            if (bDeep)
                gro.add(child.createCopy(false));
            else
                gro.add(child);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     *
     * Apply the properties of the parameter object and its children to this
     * object and its children.
     */
    public synchronized void applyProperties(GrObject gro) {
        
        while (gro instanceof ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();
        
        GrHashList props = gro.getProperties();
        
        // If the template title is blank, don't apply it.  Use the existing
        // title instead.  If a blank title is desired, use a space for the
        // template's title.
        GrProperty p = (GrProperty)props.get("title");
        if (p != null) {
            if (p.value == "") props.remove("title");
        }
        setProperties(props);
        
        // Apply properties of the template's children to this
        // object's children.
        Enumeration templateEnum = gro.children();
        if (!(templateEnum.hasMoreElements())) return;
        
        GrObject templateChild = null;
        for (GrObject child : m_children) {
            GrObject firstTemplateChild = null;
            while (true) {
                if (templateEnum.hasMoreElements()) {
                    // Get the next template child.
                    templateChild = (GrObject)templateEnum.nextElement();
                } else {
                    // Start over at the first template child.
                    templateEnum = gro.children();
                    templateChild = (GrObject)templateEnum.nextElement();
                }
                
                // Exit the loop if all the children have been examined
                // and we have come back to the starting point.
                if (firstTemplateChild == templateChild) break;
                
                // Remember the first template child passed.
                if (firstTemplateChild == null)
                    firstTemplateChild = templateChild;
                
                // Apply the properties of the template child to this
                // object's child
                if (child.isSameTypeAs(templateChild)) {
                    child.applyProperties(templateChild);
                    break;
                }
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Tell whether the parameter object is the same type as this object so
     * that its properties should be applied.
     */
    public boolean isSameTypeAs(GrObject gro) {
        
        if (getClass().isInstance(gro))
            return true;
        else
            return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public GrHashList getProperties() {
        
        GrHashList props = new GrHashList();
        
        props.put("class", new GrProperty(this.getClass().getName(),
            "Class Name", false, true, false));
        // Write the object ID in case other elements need to reference
        // this one.
        props.put("id", new GrProperty(this.hashCode(), "ID",
            false, m_bSaveIdToXml, false));
        
        return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public synchronized void setProperties(HashMap props) {
        
        m_bHasChanged = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public void setProperty(String key, String value) {
        
        m_bHasChanged = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of menu items that execute commands on this object.
     * @return  A menu item that can be placed on a menu and send an ActionEvent
     *          back to this object when it is chosen.
     */
    public javax.swing.JMenuItem getMenuCommands() {
        
        return null;
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides the comparison method so that only the exact same object will be
     * considered equal.
     */
    @Override
    public boolean equals(Object o) {
        
        return (o == this);
    }
    
//------------------------------------------------------------------------------
    /**
     * Returns an XML version of this element and its children.
     */
    public synchronized org.w3c.dom.Element getDOMRepresentation(
        org.w3c.dom.Document doc, boolean bIncludeChildren) {
        
        org.w3c.dom.Element el = doc.createElement(this.getXmlElementName());
        
        GrHashList props = getProperties();
        
        // Just write the id and return if this object has been written out
        // to the file before.
        if (m_bSaveIdOnlyToXml) {
            GrProperty p = (GrProperty)props.get("id");
            el.setAttribute("id", p.value.toString());
            p = (GrProperty)props.get("class");
            el.setAttribute("class", p.value.toString());
            
            return el;
        }
        
        ListIterator iter = props.listIterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            GrProperty p = (GrProperty)props.get(key);
            
            if (p.save == true)
                el.setAttribute(key, p.value.toString());
        }
        
        if (bIncludeChildren) {
            for (GrObject child : m_children)
                el.appendChild(child.getDOMRepresentation(doc, true));
        }
        
        // If the id was saved then make it so the next time this object
        // is written, only the id will be written.
        if (m_bSaveIdToXml)
            m_bSaveIdOnlyToXml = true;
        
        return el;
    }
    
//------------------------------------------------------------------------------
    /**
     * Read an XML version of this element and its children.
     */
    public synchronized void readFromDOM(org.w3c.dom.Node node,
        Map allObjects, List unlinkedObjects) {
        
        org.w3c.dom.NamedNodeMap map = node.getAttributes();
        int mapLength = map.getLength();
        GrHashList hashMap = new GrHashList();
        for (int i = 0; i < mapLength; i++) {
            org.w3c.dom.Node attr = map.item(i);
            hashMap.put(attr.getNodeName(), attr.getNodeValue());
        }
        
        setProperties(hashMap);
        
        NodeList nodeList = node.getChildNodes();
        int numChildren = nodeList.getLength();
        for (int i = 0; i < numChildren; i++) {
            Node childNode = nodeList.item(i);
            org.w3c.dom.NamedNodeMap attrMap = childNode.getAttributes();
            if (attrMap == null) continue;
            Node numDimNode = attrMap.getNamedItem("numDimensions");
            int numDimensions = 0;
            if (numDimNode != null)
                numDimensions = Integer.parseInt(numDimNode.getNodeValue());
            Node classNode = attrMap.getNamedItem("class");
            String className = "";
            if (classNode != null)
                className = classNode.getNodeValue();
            
            GrObject childObj = null;
            
            if (childNode.getNodeName().equals("dataseries")) {
		className = "gov.usgs.gr.data.DoubleDataSeries";
		childObj =
		    new gov.usgs.gr.data.DoubleDataSeries(numDimensions);
            }
            
            if (childObj == null)
                childObj = (GrObject)getObjectInstance(className);
            
            // If the node has an id attribute, add it to the hashmap of
            // all named objects.
            Node idNode = attrMap.getNamedItem("id");
            if (idNode != null) allObjects.put(idNode.getNodeValue(), childObj);
            
            childObj.readFromDOM(childNode, allObjects, unlinkedObjects);
            add(childObj);
        }
        
        m_bHasChanged = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accepts character data from a SAX parser.
     */
    public void receiveCharsFromSAX(char buf [], int offset, int len) {
    }
    
//------------------------------------------------------------------------------
    /**
     * Perform tasks that must wait for the entire XML element to be processed.
     */
    public void endSAXElement() {
    }
    
//------------------------------------------------------------------------------
    /**
     * Perform tasks that had to wait until all XML input had been read.
     */
    public void finishSAXInput(Map idObjects) {
        
        finishSAXForChildren(idObjects);
    }
    
//------------------------------------------------------------------------------
    /**
     * Finish building the children first.
     */
    public void finishSAXForChildren(Map idObjects) {
        
        // Finish processing for all children.
        for (GrObject gro : m_children)
            gro.finishSAXInput(idObjects);
    }
    
//------------------------------------------------------------------------------
    /**
     * Read an XML version of this element and its children.
     */
    public Object getObjectInstance(String className) {
        
        Class newClass = null;
        Object o = null;
        
        // Try making an instance of the class that reads the data format.
        // Show error dialog boxes if there was a problem...
        try {
            newClass = Class.forName(className);
        } catch (ClassNotFoundException e2) {
            e2.printStackTrace();
            return null;
        }
        try {
            o = newClass.newInstance();
        } catch (InstantiationException e) {
            e.printStackTrace();
            return null;
        } catch (IllegalAccessException e3) {
            e3.printStackTrace();
            return null;
        }
        
        return o;
    }
    
//------------------------------------------------------------------------------
    /**
     * Returns an XML version of this element and its children.
     */
    public String getXmlElementName() {
        return "obj";
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
        
        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();
        
        return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon loadTreeIcon(String imageFileLocation) {
        
        javax.swing.ImageIcon icon = null;
        try {
            icon = new javax.swing.ImageIcon(Thread.currentThread().
                getContextClassLoader().getResource(imageFileLocation));
        } catch (Exception e) {}
        
        return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected ImageIcon setTreeIcon() {
        
        ImageIcon icon = loadTreeIcon(m_iconFile);
        
        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);
        
        return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Converts newline characters to a backslash followed by an 'n'.
     */
    public String convertNewlineToBackslashN(String s) {
        
        String newString = "";
        int index1 = 0;
        
        while (index1 < s.length()) {
            int index2 = s.indexOf("" + '\n', index1);
            if (index2 < 0) index2 = s.length();
            
            newString += s.substring(index1, index2);
            index1 = index2 + 1;
            if (index1 < s.length()) newString += ("\\" + "n");
        }
        
        return newString;
    }
    
//------------------------------------------------------------------------------
    /**
     * Converts a backslash followed by an 'n' to newline characters.
     */
    public String convertBackslashNToNewline(String s) {
        
        String newString = "";
        int index1 = 0;
        
        while (index1 < s.length()) {
            int index2 = s.indexOf("\\n", index1);
            if (index2 < 0) index2 = s.length();
            
            newString += s.substring(index1, index2);
            index1 = index2 + 2;
            if (index1 < s.length()) newString += '\n';
        }
        
        return newString;
    }
    
//------------------------------------------------------------------------------
    /**
     * Wrapper for adding items from the list of children.
     */
    public synchronized boolean add(GrObject gro) {
        
        if (m_children.contains(gro)) return false;
        if (gro == this) return false;
        
        m_children.add(gro);
        gro.addObserver(this);
        
        m_bHasChanged = true;
        
        return true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Wrapper for clearing the list of children.
     */
    public synchronized void clear() {
        
        // First deregister this object as an observer of each child.
        for (GrObject child : m_children)
            child.deleteObserver(this);
        
        m_children.clear();
        
        m_bHasChanged = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Wrapper for removing items from the list of children.
     */
    public synchronized GrObject remove(int i) {
        
        // First deregister this object as an observer of the child.
        GrObject child = (GrObject)m_children.get(i);
        child.deleteObserver(this);
        
        m_bHasChanged = true;
        
        return ((GrObject)m_children.remove(i));
    }
    
//------------------------------------------------------------------------------
    /**
     * Wrapper for removing items from the list of children.
     */
    public synchronized boolean remove(GrObject gro) {
        
        // First deregister this object as an observer of the child.
        boolean bFound = m_children.remove(gro);
        if (bFound) gro.deleteObserver(this);
        
        m_bHasChanged = true;
        
        return (bFound);
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds the first instance of class within the children of this object.
     * @param className The name of the class to be matched.
     * @return  The first child object that is an instance of cls.
     */
    public synchronized GrObject firstChildInstanceOf(String className,
        boolean bRecursive) {
        
        List<? extends GrObject> list = new LinkedList();
        
        try {
            Class cls = Class.forName(className);
            list = findChildInstancesOf(cls, bRecursive, true);
        } catch (Exception e) {}

        if (list.size() == 0)
            return null;
        else
            return list.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds all instances of class within the children of this object.
     * @param className The name of the class to be matched.
     * @return  A list of child objects that are instances of cls.
     */
    public synchronized List<? extends GrObject> findChildInstancesOf(
        String className, boolean bRecursive) {
        
        List list = new LinkedList();
        
        try {
            Class cls = Class.forName(className);
            list = findChildInstancesOf(cls, bRecursive, false);
        } catch (Exception e) {}
        
        return list;
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds all instances of class within the children of this object.
     * @param cls The class to be matched.
     * @return  A list of child objects that are instances of cls.
     */
    public synchronized List<? extends GrObject> findChildInstancesOf(Class cls,
        boolean bRecursive, boolean bQuitAtFirst) {
        
        List list = new LinkedList();
        
        // Ignore ObjectEffigy, unless it is the class being searched for.
        boolean bIgnoreEffigies = true;
        if (cls.getName().indexOf("ObjectEffigy") >= 0)
            bIgnoreEffigies = false;
        
        // For each child, add the child to the list if it is an instance of
        // cls.  Append any of its matching children to the list.
        for (GrObject child : m_children) {
            if (bIgnoreEffigies) {
                while (child instanceof ObjectEffigy)
                    child = ((ObjectEffigy)child).getSubject();
            }
            if (cls.isInstance(child)) {
                list.add(child);
                if (bQuitAtFirst)
                    return list;
            }
            
            if (bRecursive)
                list.addAll(child.findChildInstancesOf(cls, true,
                    bQuitAtFirst));
        }
        
        return list;
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds first instance of class within the parents of this object.
     * @param className The name of the class to be matched.
     * @return  The first parent object that is an instances of cls.
     */
    public synchronized GrObject firstParentInstanceOf(String className,
        boolean bRecursive) {
        
        List<? extends GrObject> list = new LinkedList();
        
        try {
            Class cls = Class.forName(className);
            list = findParentInstancesOf(cls, bRecursive, true);
        } catch (Exception e) {}

        if (list.size() == 0)
            return null;
        else
            return list.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds all instances of class within the parents of this object.
     * @param className The name of the class to be matched.
     * @return  A list of parent objects that are instances of cls.
     */
    public synchronized List<? extends GrObject> findParentInstancesOf(
        String className, boolean bRecursive) {
        
        List list = new LinkedList();
        
        try {
            Class cls = Class.forName(className);
            list = findParentInstancesOf(cls, bRecursive, false);
        } catch (Exception e) {}
        
        return list;
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds all instances of class within the parents of this object.
     * @param cls The class to be matched.
     * @return  A list of parent objects that are instances of cls.
     */
    public synchronized List<? extends GrObject> findParentInstancesOf(
        Class cls, boolean bRecursive, boolean bQuitAtFirst) {
        
        List list = new LinkedList();
        
        // Ignore ObjectEffigy, unless it is the class being searched for.
        boolean bIgnoreEffigies = true;
        if (cls.getName().indexOf("ObjectEffigy") >= 0)
            bIgnoreEffigies = false;
        
        // For each child, add the child to the list if it is an instance of
        // cls.  Append any of its matching children to the list.
        for (GrObject parent : m_observers) {
            if (bIgnoreEffigies) {
                while (parent instanceof ObjectEffigy)
                    parent = ((ObjectEffigy)parent).getSubject();
            }
            if (cls.isInstance(parent)) {
                list.add(parent);
                if (bQuitAtFirst)
                    return list;
            }
            
            if (bRecursive)
                list.addAll(parent.findParentInstancesOf(cls, true,
                    bQuitAtFirst));
        }
        
        return list;
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds a child within the tree structure and removes it.
     */
    public synchronized boolean findAndRemove(GrObject gro) {
        
        // Try to remove it from the first generation.
        boolean bFound = remove(gro);
        if (bFound) {
            return true;
        } else {
            // Try to remove it from within the tree.
            for (GrObject child : m_children) {
                bFound = child.findAndRemove(gro);
                
                if (bFound)
                    return true;
            }
        }
        
        // The object was not found as a descendent of this node.
        return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds a child within the tree structure and returns its parent.
     * null otherwise
     */
    public GrObject findParentOf(GrObject gro) {
        
        boolean bFound = false;
        // Look for it in the first generation.
        for (GrObject child : m_children) {
            if (child == gro)
                return this;
        }
        
        // Look for it in the descendents.
        for (GrObject child : m_children) {
            GrObject parent = child.findParentOf(gro);
            if (parent != null)
                return parent;
        }
        
        // The object was not found as a descendent of this node.
        return null;
    }
    
//------------------------------------------------------------------------------
    /**
     * Returns a copy of the child list.
     */
    public List<GrObject> getChildList() {
        return new ArrayList(m_children);
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the TreeNode interface.
     */
    public Enumeration children() {
        return Collections.enumeration((List)m_children);
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the TreeNode interface.
     */
    public boolean getAllowsChildren() {
        return true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the TreeNode interface.
     */
    public TreeNode getChildAt(int childIndex) {
        return (TreeNode)m_children.get(childIndex);
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the TreeNode interface.
     */
    public int getChildCount() {
        return m_children.size();
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the TreeNode interface.
     */
    public int getIndex(TreeNode node) {
        
        return m_children.indexOf(node);
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the TreeNode interface.
     */
    public TreeNode getParent() {
        return m_parent;
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the TreeNode interface.
     */
    public boolean isLeaf() {
        return (m_children.size() == 0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Called by an observed GrObject's notifyObservers() method to handle changes
     * to the observed object.  May use the notifyObservers() method to propagate
     * the event handling up to this object's observers.
     * @param gro   The observed object that changed.
     * @param arg   Additional information that the updated object may choose to
     *              send.
     */
    public synchronized void update(GrObject gro, Object arg) {
        
        m_bHasChanged = true;
        if (arg == null)
            notifyObservers();
        else
            notifyObservers(arg);
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the Observable methods.
     */
    public synchronized void addObserver(GrObject o) {
        
        if (o == null) {
            System.err.println("Attempt to add null observer to " + this);
            return;
        }
        
        if (m_observers.indexOf(o) < 0)
            m_observers.add(o);
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the Observable methods.
     */
    public int countObservers() {
        
        return m_observers.size();
    }
    
//------------------------------------------------------------------------------
    /**
     * Returns a clone of the list of observers.
     */
    public List<GrObject> getObservers() {
        
        return (List<GrObject>)m_observers.clone();
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the Observable methods.
     */
    public synchronized void deleteObserver(GrObject o) {
        
        int index = m_observers.indexOf(o);
        if (index >= 0)
            m_observers.remove(index);
//        if (m_observers.size() == 0)
//            clear();
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the Observable methods.
     */
    public void deleteObservers() {
        
        m_observers.clear();
        clear();
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the Observable methods.
     */
    public boolean hasChanged() {
        
        return m_bHasChanged;
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the Observable methods.
     */
    public void notifyObservers() {
        
        if (!m_bHasChanged || m_bDeferUpdates) return;
        
        synchronized (this) {
            for (GrObject gro : m_observers)
                gro.update( (GrObject)this, (Object)null);
            
            clearChanged();
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the Observable methods.
     * @arg arg An argument discribing source of the notification.  It can be
     *          null, but is preferably a list containing a stack of GrObject
     *          instances that generated the notification.  The first value in
     *          the list can either be the source GrObject or an Integer
     *          describing the event with the source object as the second
     *          element.  The Integer should match up with a predefined public
     *          constant in the source object.
     */
    public synchronized void notifyObservers(Object arg) {
        
        if (!m_bHasChanged || m_bDeferUpdates) return;
        
        for (GrObject gro : m_observers) {
            // Pass null if arg is null.
            if (arg == null) {
                gro.update((GrObject)this, null);
            } else {
                // Pass a list of objects to the update method that includes
                // this object as the last element.  Create the list if arg is
                // an Integer.
                if (arg instanceof List) {
                    ((List)arg).add(this);
                    gro.update((GrObject)this, arg);
                } else {
                    if (arg instanceof Integer) {
                        List list = new ArrayList();
                        list.add(arg);
                        list.add(this);
                        gro.update((GrObject)this, list);
                    }
                }
            }
        }
        
        clearChanged();
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the Observable methods.
     */
    protected void clearChanged() {
        
        m_bHasChanged = false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Required for implementing the Observable methods.
     */
    protected void setChanged() {
        
        m_bHasChanged = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Cause redraws and other effects of changing children to be deferred
     * until later.
     */
    public void deferUpdates() {
        
        m_bDeferUpdates = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Cause redraws and other effects of changing children to be deferred
     * until later.
     */
    public void processDeferredUpdates() {
        
        m_bDeferUpdates = false;
        
        notifyObservers();
    }
}




