/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2000-08-18
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import java.io.*;
import java.util.*;

//==============================================================================
/**
 * A class full of utility methods for ASCII I/O.
 */
public class AsciiIO {

//------------------------------------------------------------------------------
/**
 * Creates an instance of this class.
 */
    public AsciiIO () {
    }

//------------------------------------------------------------------------------
/**
 * Pads a string by adding spaces to the left until it is the specified length.
 * If s is already longer than targetLength, it will be returned unchanged.
 * @param s		The string that may need to be padded.
 * @param targetLength	The length of the string after padding.
 * @return A string that is at least targetLength long.
 */
    public static String padLeft (String s, int targetLength) {

    	int currentLength = s.length();

        // Put together the required number of leading spaces.
        String spaces = "";
        for (int i = 0; i < (targetLength - currentLength); i++) {
        	spaces += " ";
        }

        String newString = spaces + s;

        return newString;
    }

//------------------------------------------------------------------------------
/**
 * Read a line from the buffered reader.  Break it up into tokens and store
 * them in a double array.
 * @param line    The String object to use for reading.
 * @param inputDelimiters The characters that should be interpreted as
 *                        delimiters.
 * @return An array of doubles containing tokens from the line, null if none.
 */
    public static double[] parseLineAsDoubles(String line,
        String inputDelimiters) {

        // Separate the String line into String tokens.
        // Return null if there were no tokens.
        String[] tokens = parseLine(line, inputDelimiters);
        if (tokens == null) return null;

        // Convert the String tokens to doubles in an array.
        double[] vals = new double[tokens.length];
        for (int i = 0; i < tokens.length; i++) {
            // If the token is not a double, store it as 0.
            try {
                vals[i] = Double.parseDouble(tokens[i].trim());
            }
            catch (NumberFormatException e) {
                vals[i] = -Double.MAX_VALUE;
            }
        }

        return vals;
    }

//------------------------------------------------------------------------------
/**
 * Read a line from the buffered reader.  Break it up into tokens and store
 * them in a float array.
 * @param line    The String object to use for reading.
 * @param inputDelimiters The characters that should be interpreted as
 *                        delimiters.
 * @return An array of floats containing tokens from the line, null if none.
 */
    public static float[] parseLineAsFloats(String line,
    	String inputDelimiters) {

        // Separate the String line into String tokens.
        // Return null if there were no tokens.
        String[] tokens = parseLine(line, inputDelimiters);
        if (tokens == null) return null;

        // Convert the String tokens to floats in an array.
        float[] vals = new float[tokens.length];
        for (int i = 0; i < tokens.length; i++) {
        	// If the token is not a float, store it as 0.
        	try {
				vals[i] = Float.parseFloat(tokens[i].trim());
            }
            catch (NumberFormatException e) {
            	vals[i] = -Float.MAX_VALUE;
            }
        }

        return vals;
    }

//------------------------------------------------------------------------------
/**
 * Read a line from the buffered reader.  Break it up into tokens and store
 * them in an int array.
 * @param line    The String object to use for reading.
 * @param inputDelimiters The characters that should be interpreted as
 *                        delimiters.
 * @return		An array of ints containing tokens from the line, null if none.
 */
    public static int[] parseLineAsInts(String line, String inputDelimiters) {

        // Separate the String line into String tokens.
        // Return null if there were no tokens.
        String[] tokens = parseLine(line, inputDelimiters);
        if (tokens == null) return null;

        // Convert the String tokens to ints in an array.
        int[] vals = new int[tokens.length];
        for (int i = 0; i < tokens.length; i++) {
        	// If the token is not an int, store it as 0.
        	try {
				vals[i] = Integer.parseInt(tokens[i].trim());
            }
            catch (NumberFormatException e) {
            	vals[i] = 0;
            }
        }

        return vals;
    }

//------------------------------------------------------------------------------
/**
 * Break a string up into tokens and store them in a String array.
 * @param line    The String object to use for reading.
 * @param inputDelimiters The characters that should be interpreted as
 *                        delimiters.
 * @return		An array of strings containing tokens from the line
 */
    public static String[] parseLine(String line, String inputDelimiters) {

        // Return null if the string was null.
        if (line == null) return null;

        // Parse the line into a String array.
        StringTokenizer st = new StringTokenizer(line, inputDelimiters);
        String[] tokens = new String[st.countTokens()];

        int i = 0;
        while (st.hasMoreTokens()) {
        	tokens[i++] = st.nextToken();
        }

        return tokens;
    }

//------------------------------------------------------------------------------
/**
 * Read a line from the buffered reader.  Break it up into tokens and store
 * them in an int array.
 * @param line    The String object to use for reading.
 * @param startCharIndex The index where parsing should begin.  Must be within
 *                       bounds of the string and less than endCharIndex.
 * @param endCharIndex  The index where parsing should end.  Must be within
 *                      bounds of the string and greater than startCharIndex.
 * @param stepInc   The length of each token on the line.
 * @return		An array of ints containing tokens from the line, null if none.
 */
    public static int[] parseLineAsInts(String line,
        int startCharIndex, int endCharIndex, int stepInc) {

        // Separate the String line into String tokens.
        // Return null if there were no tokens.
        String[] tokens = parseLine(line, startCharIndex, endCharIndex, stepInc);
        if (tokens == null) return null;

        // Convert the String tokens to ints in an array.
        int[] vals = new int[tokens.length];
        for (int i = 0; i < tokens.length; i++) {
            // If the token is not an int, store it as 0.
            try {
                vals[i] = Integer.parseInt(tokens[i].trim());
            }
            catch (NumberFormatException e) {
                vals[i] = 0;
            }
        }

        return vals;
    }

//------------------------------------------------------------------------------
/**
 * Break a string up into tokens and store them in a String array.
 * @param line    The String object to use for reading.
 * @param startCharIndex The index where parsing should begin.  Must be within
 *                       bounds of the string and less than endCharIndex.
 * @param endCharIndex  The index where parsing should end.  Must be within
 *                      bounds of the string and greater than startCharIndex.
 * @param stepInc   The length of each token on the line.
 * @return		An array of strings containing tokens from the line
 */
    public static String[] parseLine(String line,
        int startCharIndex, int endCharIndex, int stepInc) {

        // Return null if the string was null.
        if (line == null) return null;

        // Perform range checking on the character indexes.
/*
        if (startCharIndex < 0)
            startCharIndex = 0;
        if (startCharIndex >= line.length())
            return new String[0];

        if (endCharIndex >= line.length())
            endCharIndex = line.length();
        if (endCharIndex < startCharIndex)
            endCharIndex = startCharIndex;
*/
        // Create an array to hold the tokens.
        int numTokens = (endCharIndex - startCharIndex + 1) / stepInc;
        String[] tokens = new String[numTokens];

        // Break the tokens up into an array of Strings.
        int j = 0;
        for (int i = startCharIndex; i <= (endCharIndex - stepInc + 1);
            i += stepInc) {

            tokens[j++] = line.substring(i, i + stepInc);
        }

        return tokens;
    }

//------------------------------------------------------------------------------
/**
 * Read a line from the buffered reader.  A line is a string ending with
 * a linefeed character.  Carriage-return characters are stripped out and
 * the linefeed is not returned.
 * @param br    The buffered reader object to use for reading.
 * @return The next line from the buffered reader minus CR or LF.
 * @throws java.io.IOException
 */
    public static String readLine(BufferedReader br)
    	throws java.io.IOException {

        String line = "";
        char c;

        while ((c = (char)br.read()) != '\n') {
        	if (c != '\r') line += c;
                
        }

        return line;
    }
}
