/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-03-20
 * Date started:    2000-08-15
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import gov.usgs.gr.*;
import java.util.*;

//==============================================================================
/**
 * A dimension within a DataSeries.
 */
public class DataDimension extends gov.usgs.gr.GrObject {

    static String XML_ELEMENT_NAME = "doubledim";

    /** The number of this dimension within a DataSeries. */
    protected int m_num = 0;
    /**
     * The name of this dimension.  If a program is searching for a particular
     * dimension, it should check this parameter first.
     */
    protected String m_name = "";
    /** The descriptive title of this dimension. */
    protected String m_title = "";
    /**
     * The units that the dimension is stored in, such as meters, m/s, seconds,
     * celcius.
     */
    protected String m_units = "none";
	/**
     * Modifies the units parameter.  For example, kilometers would have a
     * units of meters and a unitscale of 1000.
     */
    protected double m_unitScale = 1;
    /**
     * A scalar for each value in the dimension.  Each value will be multiplied
     * by the scale when it is read in.
     */
    protected double m_scale = 1;
    /**
     * An offset for each value in the dimension.  Each value will have the
     * offset added to it when it is read in.  Applied after the scale.
     */
    protected double m_offset = 0;
    /**
     * The suggested storage type for this dimension.  Other choices could be
     * int, long, or double.  This attribute is only a hint and may be ignored.
     */
    protected String m_store = "double";
    /**
     * The type of data stored in the dimension such as distance, pressure,
     * temperature, concentration, etc.
     */
    protected String m_type = "none";
    /**
     * The starting index of the array.  If the starting and ending indeces of
     * each dimension are given, the storage arrays can be preallocated.
     */
    protected int m_start = 0;
    /**
     * The last index of the array.  End is ignored for the last dimension
     * because it must have a fixed size of 1.
	 */
    protected int m_end = 0;
    /**
     * Used with the stepSize attribute to automatically advance the value at
     * each point in this dimension.  If the value is looping, it will start
     * over at startStep.
     */
    protected double m_startStep = 0;
    /**
     * Used with the stepSize attribute to automatically advance the value at
     * each point in this dimension.  If the value is looping, it will advance
     * to stepUntil and then start over at startStep.
     */
    protected double m_stepUntil = 0;
    /**
     * When set to a number other than zero, automatically advances through
     * values at each point.  When the dimension is stepping automatically,
     * it's values should not be written out explicitly with the other values.
     * For example, in a 2D data series, as long as the step size of the first
	 * dimension is non-zero, then it's values are automatically generated and
     * all the values written are assumed to belong to the second dimension.
     */
    protected double m_stepSize = 0;
    /**
     * The number of points that should be read in before value is advanced by
     * stepSize.  This is used in data series with three or more dimensions.
     * For example, in a 3D series of size 100 X 50 X 1, the second dimension
	 * would step every point, but the first dimension would step every 50th
     * point.
     */
    protected double m_stepEvery = 1;
    /**
	 * If start and end are used to give the total size of the series, then any
     * points that are not given explicitly will default to this value.
     */
    protected double m_default = 0;
    /**
     * Used to set the value to a constant until changed later.  Values for
     * this dimension are assumed to be absent until value is set back to
     * "none".  If set to "next", it is advanced to the next step and the point
     * count is also advanced so that stepEvery will still work for lower
     * dimensions.
     */
    protected double m_value = 0;
    /**
     * Same as inserting the value the given number of times and then resetting
     * the value back to "none".  Only works if this dimension is the only one
     * not being automatically advanced.
     */
    protected int m_repeat = 0;
    /** The lowest value for this dimension. */
    public double m_min = 0;
    /** The largest value for this dimension. */
    public double m_max = 0;
    /**
     * Tells whether the dimension's data is sorted from smallest
     * to largest.
     */
    public boolean m_sorted = false;
    /**
     * Indicates whether this dimension should be locked in place, or if
     * it should be editable.  This state isn't enforced.
     */
    public boolean m_locked = false;
    /** The smallest increment between points. */
    public double m_snap = 0;
    /** The offset for the snap. */
    public double m_snapOffset = 0;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/DataDimension.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a generic dimension.
 */
    public DataDimension() {

    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a certain number dimension.
 * @param num The dimension number.
 */
    public DataDimension(int num) {

		m_num = num;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.  Properties include:
 * <p>
 * num          Integer "Dimension Number"  no edit             <p>
 * title        String  "Title"                                 <p>
 * snap         Double   "Snap"                                  <p>
 * snapOffset   Double   "Snap Offset"                           <p>
 * locked       Boolean "Locked"                                <p>
 * sorted       Boolean "Sorted"            no edit, no save    <p>
 * min          Double   "Min"               no edit, no save    <p>
 * max          Double   "Max"               no edit, no save    <p>
 * unitScale    Double   "Unit Scale"        no edit, no save    <p>
 * @return A list of GrProperty objects describing this object's properties.
 */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("num", new GrProperty(m_num, "Dimension Number", false));
        props.put("title", new GrProperty(m_title, "Title"));
        props.put("snap", new GrProperty(m_snap, "Snap"));
        props.put("snapOffset", new GrProperty(m_snapOffset, "Snap Offset"));
        props.put("locked", new GrProperty(m_locked, "Locked"));
        props.put("sorted", new GrProperty(m_sorted, "Sorted", false, false));
        props.put("min", new GrProperty(m_min, "Min", false, false));
        props.put("max", new GrProperty(m_max, "Max", false, false));
        boolean bSaveUnitScale = (m_unitScale == 1 ? false : true);
        props.put("unitScale", new GrProperty(m_unitScale, "Unit Scale", false,
            bSaveUnitScale));
//        props.put("scale", new GrProperty(m_scale,
//            "Scale", false, false));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.gr.data.DataDimension#getProperties getProperties() for a list
 * of accepted properties.
 * @param props  A list of GrProperty objects specifying the new values for
 *               certain properties.
 */
    public void setProperties(HashMap props) {

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (key.equals("num"))
                    m_num = ((Integer)p.value);
                if (key.equals("title")) m_title = (String)p.value;
                if (key.startsWith("snapOffset")) {
                    m_snapOffset = (Double)p.value;
                }
                else {
                    if (key.startsWith("snap")) {
                        m_snap = (Double)p.value;
                    }
                }
                if (key.startsWith("locked")) {
                    m_locked = (Boolean)p.value;
                }
                if (key.startsWith("unitScale")) {
                    m_unitScale = (Double)p.value;
                }
//        		if (key.startsWith("scale")) {
//                    m_scale = ((Double)p.value).doubleValue();
//                }
            }
            if (obj instanceof String) {
                String value = (String)obj;

                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.gr.data.DataDimension#getProperties getProperties() for a list
 * of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

        if (key.equals("num"))
            m_num = Integer.parseInt(value);

        if (key.equals("title")) m_title = value;

        if (key.startsWith("snapOffset")) {
            m_snapOffset = Double.parseDouble((String)value);
        }
        else {
            if (key.startsWith("snap")) {
                m_snap = Double.parseDouble((String)value);
            }
        }
        if (key.startsWith("locked")) {
            m_locked = Boolean.parseBoolean(value);
        }
        if (key.startsWith("unitScale")) {
            m_unitScale = Double.parseDouble((String)value);
        }
//        if (key.startsWith("scale")) {
//            m_scale = Double.parseDouble((String)value);
//        }
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  A string that describes this object.
 */
    public String toString() {

        return ("Dimension " + m_num + ": " + m_title);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An XML name of this element.
 */
    public String getXmlElementName() {
        return "doubledim";
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.  Prevents children from being added.
 * @param gro The object to add.
 * @return Whether the object was successfully added.  Always false.
 */
    public boolean add(GrObject gro) {

    	return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides method from GrObject to only allow one DataSeries observer
     * at a time.
     */
    public synchronized void addObserver(GrObject gro) {
        
        if (gro == null) {
            System.err.println("Attempt to add null observer to " + this);
            return;
        }
        
        if (m_observers.indexOf(gro) < 0) {
            if (gro instanceof DataSeries) {
                int index = 0;
                while (index < m_observers.size()) {
                    if (m_observers.get(index) instanceof DataSeries)
                        m_observers.remove(index);
                    else
                        index++;
                }
            }
            m_observers.add(gro);
        }
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The DataSeries object that contains this dimension object.
 */
    public DataSeries getAssociatedSeries() {

        Iterator iter = m_observers.listIterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof DataSeries)
                return (DataSeries)gro;
        }

        return null;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The number of this dimension.
 */
    public int getNum() {

        return m_num;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return Whether the dimension is locked.
 */
    public boolean getDimensionLocked() {

        return m_locked;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param val	The new lock status for the dimension.
 */
    public void setDimensionLocked(boolean val) {

        m_locked = val;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The snap value for the dimension.
 */
    public double getSnap() {

        return m_snap;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The snap offsetvalue for the dimension.
 */
    public double getSnapOffset() {

        return m_snapOffset;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The minimum value for the dimension.
 */
    public double getMin() {

        return m_min;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The maximum value for the dimension.
 */
    public double getMax() {

        return m_max;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return True if the dimension is sorted.
 */
    public boolean getSorted() {

        return m_sorted;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The title string that describes this series.
 */
    public String getTitle() {

        return new String(m_title);
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param title The new title.
 */
    public void setTitle(String title) {

        m_title = title;
    }
}
