/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-01
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import gov.usgs.gr.*;
import java.util.*;

//==============================================================================
/**
 * An interleaved, multidimensional data object that stores its data as a single
 * one-dimensional array.  The data in the buffer's array can represent any
 * number of dimensions by interleaving the values at each dimension.  For
 * example, a three-dimensional series (X, Y, and Z) with two points would be
 * stored as (X1, Y1, Z1, X2, Y2, Z2).  The DataSeries keeps track of its
 * beginning, end, min, max, and title.
 */
public abstract class DataSeries extends gov.usgs.gr.GrObject
    implements java.awt.event.ActionListener {

    /** The index of the first point. */
    protected int m_begin = 0;
    /** The index of the last point. */
    protected int m_end = 0;
    /** The number of points in the series. */
    protected int m_numPoints = 0;
    /** A numeric code describing the type of data stored. */
    protected int m_numericType;
    /** The number of dimensions that the data describes. */
    protected int m_numDimensions;
    /** A title describing the data object. */
    protected String m_title = "";
    /** An array of objects describing each of this series' dimensions. */
    protected DataDimension[] m_dims = null;
    /**
     * An interleaved array for holding indexes into the array of
     * selected points.
     */
    protected int[] m_selectedIndexes;
    /** The number of selected points. */
    protected int m_numSelected = 0;
    /** A place to store values as they are read in. */
    protected ArrayList m_inputTokens = new ArrayList();
    /** A buffer holding the most recent, unparsed characters read from XML. */
    protected StringBuffer m_recentSAXChars = new StringBuffer();
    /** 
     * Whether to use parabolic interpolation as the default type of
     *  interpolation.
     */
    public boolean m_bParabolicInterpolation = true;
    /** Whether to use interpolation when sampling. */
    public boolean m_bInterpolateSamples = true;
    /** The name of the class to be used for editing this series. */
    public String m_editorClassName = null;
    /** 
     * Update code for changing the selected points, but not selecting all or
     * deselecting all.
     */
    public static int SELECTION_CHANGED = 0;
    /** 
     * Update code for selecting all.
     */
    public static int SELECTED_ALL = 1;
    /** 
     * Update code for deselecting all.
     */
    public static int DESELECTED_ALL = 2;
    /** 
     * Update code for modified data.
     */
    public static int DATA_MODIFIED = 3;
    /** 
     * Update code for a new array.
     */
    public static int NEW_ARRAY = 4;
    /** 
     * Update code for setIndexRange.
     */
    public static int SET_INDEX_RANGE = 5;
    /** 
     * Update code for a change to the list of deleted points.
     */
    public static int DELETED_LIST_CHANGED = 6;
    
    public abstract GrObject createCopy(boolean bDeep);
    public abstract void copyFromDataSeries(DataSeries series, boolean bDeep);
    public abstract boolean containsSelectedPoint(double[] min, double[] max,
        boolean bAddIfDeselected);
    public abstract int getCount();

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * title            String  "Title"                                         <p>
 * numDimensions    Integer "Number of Dimensions"  no edit                 <p>
 * numPoints        Integer "Number of Points"      no edit                 <p>
 * numSelected      Integer "Number of Points Selected" no edit, no save    <p>
 * editorClassName  String  "Editor Class Name"     edit, save, no display  <p>
 * numericType      Integer "Numeric Type"          edit, save, no display  <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

        // First add properties from the super class.
        GrHashList props = super.getProperties();

        props.put("title", new GrProperty(m_title, "Title"));
        props.put("numDimensions", new GrProperty(m_numDimensions,
            "Number of Dimensions", false));
        props.put("numPoints", new GrProperty(getCount(), "Number of Points",
            false));
        props.put("numSelected", new GrProperty(m_numSelected,
            "Number of Points Selected", false, false));
        if (m_editorClassName != null)
            props.put("editorClassName", new GrProperty(m_editorClassName,
                "Editor Class Name", true, true, false));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.gr.data.DoubleDataSeries#getProperties getProperties() for a list
 * of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

        GrProperty p;

        // For each property with a string value, pass it to the setProperty()
        // method to be handled there.
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                // Handle string properties by passing them to the
                // setProperty() method.
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }

                if (key.equals("numericType"))
                    m_numericType = ((Integer)p.value);
            }
            // Handle string properties by passing them to the
            // setProperty() method.
            if (obj instanceof String) {
                setProperty(key, (String)obj);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.gr.data.DoubleDataSeries#getProperties getProperties() for a list
 * of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

        if (key.equals("title"))
            m_title = value;
        if (key.equals("editorClassName"))
            m_editorClassName = value;
        if (key.equals("numericType"))
            m_numericType = Integer.parseInt(value);
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

        // Use the classname of this object.
        String fullClassName = getClass().getName();
        String name =
            fullClassName.substring(fullClassName.lastIndexOf('.') + 1);

        return (name + ": " + m_title);
    }

//------------------------------------------------------------------------------
/**
 * Adds an object as a child, but only if it is an instance of DoubleDataDimension
 * or Location.  Removes the old child and replaces it with the new one.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public synchronized boolean add(GrObject gro) {

        boolean bAdded = false;

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        if (gro instanceof Location) {
            // Remove existing Location and place the new one first in the list.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof Location) iter.remove();
            }

            super.add(gro);
            m_children.remove(gro);
            m_children.add(0, gro);

            return true;
        }

        if (gro instanceof DataDimension) {
            // Perform range checking on the dimension number.
            DataDimension dataDim = (DataDimension)gro;
            int dimNum = dataDim.getNum();
            if (dimNum < 0 || dimNum >= m_numDimensions) return false;
            m_dims[dimNum] = dataDim;

            // When adding dimensions, replace existing ones.
            boolean bSet = false;
            ListIterator iter = m_children.listIterator();
            int i = 0;
            while (iter.hasNext()) {
                GrObject oldGro = (GrObject)iter.next();
                if (oldGro instanceof DataDimension) {
                    if (((DataDimension)oldGro).getNum() == dimNum) {
                        bSet = true;

                        // First deregister this object as an observer of
                        // the old child.  Then register this object as an
                        // observer of the new child.
                        oldGro.deleteObserver(this);
                        dataDim.addObserver(this);
                        iter.set(dataDim);
                        m_bHasChanged = true;
                        bAdded = true;
                        break;
                    }
                }
            }

            // If an existing dimension was found that matched the dimension
            // number of gro, then add gro to replace the old dimension.
            if (!bSet) {
                super.add(dataDim);
                bAdded = true;
            }
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Overrides GrObject.remove() to keep objects from being removed.
 * Replacing existing Location or DoubleDataDimension children is supported through
 * the add(GrObject) method.
 * @param i     The index of the object to remove.
 * @return      The object that was removed.  Always null.
 */
    public GrObject remove(int i) {

        return null;
    }

//------------------------------------------------------------------------------
/**
 * Overrides GrObject.remove() to keep objects from being removed.
 * Replacing existing Location or DoubleDataDimension children is supported through
 * the add(GrObject) method.
 * @param gro   The object to be removed.
 * @return      Whether the object was removed.  Always false.
 */
    public boolean remove(GrObject gro) {

        return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accepts character data from a SAX parser.  This generally provides faster
     * XML input than DOM.
     * @param buf   A character buffer from which to read.
     * @param offset    The offset into the buffer.
     * @param len       The length of the string to read from the buffer.
     */
    public void receiveCharsFromSAX(char buf [], int offset, int len) {
        
        m_recentSAXChars.append(buf, offset, len);
        int start = 0;
        int end = 0;
        while (end < m_recentSAXChars.length()) {
            boolean bMatch = false;
            char c0 = m_recentSAXChars.charAt(end);
            if (c0 == ' ')  bMatch = true;
            if (c0 == '\t') bMatch = true;
            if (c0 == '\n') bMatch = true;
            if (c0 == '\r') bMatch = true;
            if (c0 == '\f') bMatch = true;

            if (bMatch) {
                if (end <= start) {
                    start++;
                }
                else {
                    String s0 = m_recentSAXChars.substring(start, end);
                    tokenizeAndStore(s0);
                    start = end + 1;
                }
            }
            end++;
        }
        
        m_recentSAXChars.delete(0, start);
    }
    
//------------------------------------------------------------------------------
    /**
     * Perform tasks that must wait for the entire XML element to be processed.
     */
    public void endSAXElement() {
        
        if (m_recentSAXChars.length() > 0) {
            //Tokenize the left-over buffer.
            char[] buf = {' '};
            receiveCharsFromSAX(buf, 0, 1);
        }
        
        // Set the list of input tokens to be the new main data array.
        setArray(m_inputTokens);
        
        m_inputTokens = new ArrayList();
    }
    
//------------------------------------------------------------------------------
    /**
     * Aids the receiveCharsFromSAX() method.  Reads tokens from the buffer and
     * stores them in the m_inputTokens list.  They can be converted to an
     * interleaved array later.
     * @param buf   A character buffer from which to read.
     * @param offset    The offset into the buffer.
     * @param len       The length of the string to read from the buffer.
     */
    protected abstract void tokenizeAndStore(String s0);
    
//------------------------------------------------------------------------------
    /**
     * Checks if the object argument is a DataSeries with the specified
     * number of dimensions.
     * @param obj	The object to be checked
     * @param numDims	The number of dimensions the series should have.
     * @return Whether the object was a DS with the right number of dimensions.
     */
    public static boolean countDims(Object obj, int numDims) {
	
	if (obj instanceof DataSeries) {
	    DataSeries series = (DataSeries)obj;
	    if (series.getNumDimensions() == numDims)
		return true;
	}
	
	return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Creates a double array from the ArrayList and calls setArray(double[]).
     * @param list    An ArrayList of Double objects.
     */
    public abstract void setArray(ArrayList list);
 
//------------------------------------------------------------------------------
/**
 * Creates a DoubleDataDimension child object for each dimension of the data seres
 * for storing dimension information.  The number of dimensions is stored in
 * the m_numDimensions member variable, which is set through the constructor.
 */
    protected void initBounds(){

        for (int i = 0; i < m_numDimensions; i++) {
            add(new DataDimension(i));
        }
    }

//------------------------------------------------------------------------------
/**
 * Set the range of indexes to be considered active by the series.
 * @param newBegin  The index to start at.
 * @param newEnd    The index to end at.
 */
    public synchronized void setIndexRange(int newBegin, int newEnd) {

        // Perform range checking.
        if (newBegin < 0) newBegin = 0;
        if (newBegin >= m_numPoints) newBegin = m_numPoints - 1;
        if (newEnd < newBegin) newEnd = newBegin;
        if (newEnd >= m_numPoints) newEnd = m_numPoints - 1;

        // Set the variables.
        m_begin = newBegin;
        m_end = newEnd;

        // Notify observers of the change to this object.
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param i	The index of the dimension.
 * @return Whether the specified dimension is locked.
 */
    public boolean getDimensionLocked(int i) {

        return m_dims[i].m_locked;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param i	The index of the dimension.
 * @param val	The new lock status for that dimension.
 */
    public synchronized void setDimensionLocked(int i, boolean val) {

        m_dims[i].m_locked = val;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param num The number of the dimension to return.
 * @return The dimension object whose number matches the parameter.
 */
    public DataDimension getDimension(int num) {

        if (num >= m_dims.length || num < 0)
            return null;
        else
            return m_dims[num];
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The numeric code describing the data.
 */
    public int getType() {

        return m_numericType;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param val The numeric code describing the data.
 */
    public void setType(int val) {

        m_numericType = val;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The beginning index of the data.
 */
    public int getBegin() {

        return m_begin;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The ending index of the first dimension of data.
 */
    public int getEnd(){

        return m_end;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The number of dimensions in the interleaved array.
 */
    public int getNumDimensions(){

        return m_numDimensions;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The number of n-tuple points in the series.
 */
    public int getNumPoints(){

        return m_numPoints;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The number of selected points.
 */
    public int getNumSelected(){

        return m_numSelected;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param dimension The dimension to report the snap offset for.
 * @return The snap value for the specified dimension.
 */
    public double getSnap(int dimension) {

        return m_dims[dimension].m_snap;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param dimension The dimension to report the snap offset for.
 * @return The snap offset value for the specified dimension.
 */
    public double getSnapOffset(int dimension) {

        return m_dims[dimension].m_snapOffset;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param dimension The dimension to report the minimum for.  If it is 0, it
 *                  will report the min x value.  If 1, it will report the min
 *                  y value.
 * @return The minimum value for the specified dimension.
 */
    public double getMin(int dimension) {

        if (dimension < 0 || dimension >= m_dims.length)
            return 0;
        else
            return m_dims[dimension].m_min;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param dimension The dimension to report the maximum for.  If it is 0, it
 *                  will report the max x value.  If 1, it will report the max
 *                  y value.
 * @return The maximum value for the specified dimension.
 */
    public double getMax(int dimension) {

        if (dimension < 0 || dimension >= m_dims.length)
            return 0;
        else
            return m_dims[dimension].m_max;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param dimension The dimension to report the sorted status for.
 * @return True if the dimension is sorted.
 */
    public boolean getSorted(int dimension) {

        return m_dims[dimension].m_sorted;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The title string that describes this series.
 */
    public String getTitle() {

        return new String(m_title);
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param title The new title.
 */
    public void setTitle(String title) {

        m_title = title;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The name of the editor for this series.
 */
    public String getEditorClassName() {

        if (m_editorClassName == null)
            return null;
        else
            return new String(m_editorClassName);
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * Menu items include: <p>
 * "Select Area..."  <br>
 * "Select All"  <br>
 * "Deselect All"  <br>
 * "Delete Selected Points"  <p>
 * @return A menu full of commands applicable to this class.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
            "Data Series");
        javax.swing.JMenuItem item = new javax.swing.JMenuItem
            ("Select Area...");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Select All");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Deselect All");
        item.addActionListener(this);
        subMenu.add(item);
        subMenu.addSeparator();
        item = new javax.swing.JMenuItem("Delete Selected Points");
        item.addActionListener(this);
        subMenu.add(item);
//        item = new javax.swing.JMenuItem("Modify Selected Points...");
//        item.addActionListener(this);
//        subMenu.add(item);
//        item = new javax.swing.JMenuItem("Resample...");
//        item.addActionListener(this);
//        subMenu.add(item);

        return subMenu;
    }

}

