/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2002-10-15
 * Date started:    1999-04-15
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import gov.usgs.gr.*;
import java.util.*;

//==============================================================================
/**
 * A snap shot of a DataSeries that can be restored later in an undo operation.
 */
public class DataSnapShot extends GrObject {

    /** A place to save the series that held the arrays. */
    protected GrObjectContainer m_copies = new GrObjectContainer();
    /** The sources of the data to save. */
    protected GrObjectContainer m_sources = new GrObjectContainer();

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data snap shot.
 */
    public DataSnapShot() {

    }

//------------------------------------------------------------------------------
/**
 * Constructor for storing a list of DataSeries objects.
 * @param dataList	A list of series to store in this snap shot.
 */
    public DataSnapShot(List dataList) {

    	// Store each series individually.
        Iterator iter = dataList.iterator();
        while (iter.hasNext()) {
        	DataSeries series = (DataSeries)iter.next();
            store(series);
        }
    }

//------------------------------------------------------------------------------
/**
 * Stores a copy of the series so it can be restored later.
 * @param series	The DataSeries to be copied and stored.
 */
	public void store(DataSeries series) {

    	// Add the series to the list sources.
        m_sources.add(series);
        series.deleteObserver(m_sources);

        // Store a deep copy of the series.
        DataSeries copy = (DataSeries)series.createCopy(true);

        // This keeps "Copy of" from being prepended...
        copy.setProperties(series.getProperties());

        // Add the copy as a child of this object.
        m_copies.add(copy);
	}

//------------------------------------------------------------------------------
/**
 * Restores all stored attributes and data to the DataSeries objects from
 * which they came.
 * @return	A snapshot of the data before restoration, so that restoration
 *			can be undone.
 */
	public DataSnapShot restore() {

        // If this is an undo operation, the new snapshot will enable
        // a redo operation.
        DataSnapShot snapShot = new DataSnapShot();

    	// Traverse the source and copy lists.
        Enumeration sourceEnum = m_sources.children();
    	Enumeration copyEnum = m_copies.children();
        while (sourceEnum.hasMoreElements() && copyEnum.hasMoreElements()) {
        	DataSeries source = (DataSeries)sourceEnum.nextElement();
            DataSeries copy = (DataSeries)copyEnum.nextElement();

            // Save the current data from the source before restoring
            // the old data to the source.
            snapShot.store(source);
        	source.copyFromDataSeries(copy, true);

        	// This keeps "Copy of" from being prepended.
        	source.setProperties(copy.getProperties());
        }

        return snapShot;
	}
}
