/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-06-19
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */
package gov.usgs.gr.data;

import gov.usgs.gr.*;
import gov.usgs.gr.tools.*;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.*;
import java.io.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import java.awt.EventQueue;
import java.lang.reflect.InvocationTargetException;

//==============================================================================
/**
 * Read, stores, and writes data to/from files that stores data in columns
 * delimited by characters such as spaces or commas.
 */
public class DelimitedDataFile extends gov.usgs.gr.GrFile {

    /** The name of the file. */
    protected String m_filename = "";
    /** A list of lines that are read or written at the beginning of the file.*/
    protected List m_headerLines = new ArrayList();
    /** The number of header lines that should be read. */
    protected int m_numHeaderLines = 0;
    /** The description of how to assign columns to DataSeries. */
    private String m_colAssignments = "xyxy";
    /**
     * A String containing the characters which will be treated as delimiters
     * between data fields in the input file.
     */
    protected String m_inputDelimiters = " \t\n\r\f,";
    /**
     * A list of data series which which were read from or will be written
     * to the file.
     */
    public List m_seriesList = new ArrayList();
    /** An array containing all the series' data arrays. */
    protected double[][] m_seriesArrays;
    /** The number of data columns in the file. */
    protected int m_numChannels = 1;
    /** The number of dimensions in each data series. */
    protected int m_numDimensions = 2;
    /** The bad indicator string. */
    protected String m_badIndicator = "NaN";
    /** Whether to write deleted markers as bad value in output. */
    protected boolean m_bWriteDeletedAsBad = false;
    /** The output format string, see java.text.DecimalFormatter. */
    protected String m_outputDecimalFormat = "";
    /** The data source buffered reader. */
    protected BufferedReader m_br = null;
    /** The data destination buffered writer. */
    protected BufferedWriter m_bw = null;
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/DelimitedDataFile.gif";

//------------------------------------------------------------------------------
    /**
     * Creates an empty data file.
     */
    public DelimitedDataFile() {

        super();
        m_title = "Delimited Data File";
    }

//------------------------------------------------------------------------------
    @Override
    public boolean add(GrObject gro) {

        if (gro instanceof ObjectEffigy) {
            if (((ObjectEffigy) gro).depictProperties == false) {
                m_children.add(gro);
            } else {
                m_children.add(new ObjectEffigy(gro));
            }
        } else {
            m_children.add(new ObjectEffigy(gro));
        }

        return true;
    }

//------------------------------------------------------------------------------
    @Override
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("numHeaderLines", new GrProperty(m_numHeaderLines,
                "Number of Header Lines", false, false, false));
        props.put("outputDecimalFormat", new GrProperty(m_outputDecimalFormat,
                "Output Decimal Format"));
        props.put("writeDeletedAsBad", new GrProperty(m_bWriteDeletedAsBad,
                "Write Deleted As Bad"));
        props.put("badIndicator", new GrProperty(m_badIndicator,
                "Bad Indicator"));

        return props;
    }

//------------------------------------------------------------------------------
    @Override
    public void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String) iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty) obj;

                if (key.equals("numHeaderLines")) {
                    m_numHeaderLines = ((Integer) p.value);
                }
                if (key.equals("outputDecimalFormat")) {
                    m_outputDecimalFormat = ((String) p.value);
                }
                if (key.equals("writeDeletedAsBad")) {
                    m_bWriteDeletedAsBad = ((Boolean) p.value);
                }
                if (key.equals("badIndicator")) {
                    m_badIndicator = ((String) p.value);
                }
            }
            if (obj instanceof String) {
                String value = (String) obj;

                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
    @Override
    /**
     * @see gov.usgs.gr.data.DelimitedDataFile#getProperties getProperties()
     * for a list of inherited properties
     */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("numHeaderLines")) {
            m_numHeaderLines = Integer.parseInt(value);
        }
        if (key.equals("outputDecimalFormat")) {
            m_outputDecimalFormat = value;
        }
        if (key.equals("writeDeletedAsBad")) {
            m_bWriteDeletedAsBad = Boolean.parseBoolean(value);
        }
        if (key.equals("badIndicator")) {
            m_badIndicator = value;
        }
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0) {
            setTreeIcon();
        }

        return (javax.swing.ImageIcon) TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null) {
            icon = super.setTreeIcon();
        }

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
    @Override
    public void read() throws IOException {

        // Open the file.
        m_br = new BufferedReader(new FileReader(
                m_file.getAbsolutePath()));
        // Read the header.
        readHeader();
        // Read the data.
        readData();
        // Close the file.
        m_br.close();
    }

//------------------------------------------------------------------------------
    @Override
    public void write() throws IOException {

        // Open the file.
        m_bw = new BufferedWriter(new FileWriter(
                m_file.getAbsolutePath()));

        // Read the header.
        writeHeader();
        // Read the data.
        writeData();
        // Close the file.
        m_bw.close();
    }

//------------------------------------------------------------------------------
    /**
     * Read the header information from an open file attached to a reader.
     * Called by read().
     * @throws IOException
     * @see gov.usgs.gr.data.DelimitedDataFile#setHeaderLines setHeaderLines()
     */
    protected void readHeader() throws IOException {

        for (int i = 0; i < m_numHeaderLines; i++) {
            m_headerLines.add(AsciiIO.readLine(m_br));
        }
    }

//------------------------------------------------------------------------------
    /**
     * Read the data through a reader attached to an open file.  Called by read()
     * <p><b> Pre Conditions </b><br>
     *      The header has been read with readHeader()
     * <p><b> Post Conditions </b><br>
     *      The file needs to be closed.
     * @throws IOException
     */
    protected void readData() throws IOException {

        // Empty any previous contents before loading new ones.
        clear();

        // Initialize local variables.
        ArrayList lineList = new ArrayList();
        double[] vals;
        int seriesIndex = 0;
        int dimIndex = 0;

        // Read the line in as a double array, and store it in a list.
        while ((vals = AsciiIO.parseLineAsDoubles(m_br.readLine(),
                m_inputDelimiters)) != null) {
            lineList.add(vals);
        }

        // Transfer the doubles from the list to a big array of doubles.
        m_numChannels = ((double[]) lineList.get(0)).length;
        int numLines = lineList.size();
        double[] bigArray = new double[numLines * m_numChannels];
        int offset = 0;
        while (lineList.size() > 0) {
            vals = (double[]) lineList.get(0);
            for (int i = 0; i < vals.length; i++) {
                bigArray[offset + i] = vals[i];
            }
            offset += m_numChannels;
            lineList.remove(0);
        }

        // Create an array to put in the data series.
        double[] array = new double[numLines * 2];
        for (int i = 0; i < numLines; i++) {
            array[i * 2] = bigArray[i * m_numChannels];
            array[i * 2 + 1] = bigArray[i * m_numChannels + 1];
        }

        // Create a new data series from the array.
        DataSeries series = null;
        series = new DoubleDataSeries(array, 2);
        add(series);
    }

//------------------------------------------------------------------------------
    /**
     * Write the header information to an open file attached to a writer. Called by
     * write().  Writes the same lines that were read in with readHeader().
     * @throws IOException
     */
    protected void writeHeader() throws IOException {

        // Write out the lines that were read in from the file originally.
        ListIterator iter = m_headerLines.listIterator();
        while (iter.hasNext()) {
            String lineString = (String) iter.next() + "\n";
            m_bw.write(lineString, 0, lineString.length());
        }
    }

//------------------------------------------------------------------------------
    /**
     * Write the time series back to a file.  Called by write().
     * <p><b> Pre Conditions </b><br>
     *      The header has been written with writeHeader()
     * <p><b> Post Conditions </b><br>
     *      The file needs to be closed.
     * @throws IOException
     */
    protected void writeData() throws IOException {

        // Initialize local variables.  Use up to five decimal places on output.
        String lineString = "";
        java.text.NumberFormat form = null;
        if (m_outputDecimalFormat.length() > 0) {
            try {
                form = new java.text.DecimalFormat(m_outputDecimalFormat);
            } catch (IllegalArgumentException e) {
            }
        }
        if (form == null) {
            form = new DecimalFormat("0.#####");
        }

        // Find and compile a list of data arrays contained in child DataSeries.
        double[][] arrays = new double[m_children.size()][];
        int[] numDims = new int[m_children.size()];
        int numSeries = 0;
        int longest = 0;
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            Object o = iter.next();

            while (o instanceof ObjectEffigy) {
                o = ((ObjectEffigy) o).getSubject();
            }

            if (o instanceof DoubleDataSeries) {
                DoubleDataSeries series = (DoubleDataSeries) o;

                // Write deleted points as bad, if requested.
                if (m_bWriteDeletedAsBad &&
                        DoubleDataSeries.countDims(series, 2)) {

                    series = (DoubleDataSeries) series.createCopy(true);
                    series.selectAll();
                    SeriesGapFiller.fill((DoubleDataSeries) series,
                            new ConstantValueProvider(Double.NaN));
                }

                if (series.getNumPoints() > longest) {
                    longest = series.getNumPoints();
                }
                arrays[numSeries] = series.getArray();
                numDims[numSeries] = series.getNumDimensions();

                numSeries++;
            }
        }

        // Write each line of the arrays.
        for (int i = 0; i < longest; i++) {
            lineString = "";

            // Write the current value for each series.
            for (int j = 0; j < numSeries; j++) {
                for (int k = 0; k < numDims[j]; k++) {
                    String s = "";

                    if (arrays[j].length <= (i * numDims[j] + k)) // Write a 0 if past the end of the array.
                    {
                        s = "0";
                    } else // Normally write the number with the preset format.
                    {
                        s = form.format(arrays[j][i * numDims[j] + k]);
                    }

                    // Padd to 10 column width.
                    lineString += AsciiIO.padLeft(s, 10);

                    // Add a space delimiter between columns.
                    lineString += " ";
                }
            }

            // Add a newline character and send the line to the file.
            lineString += "\n";
            m_bw.write(lineString, 0, lineString.length());
        }
    }

//------------------------------------------------------------------------------
    /**
     * Allows other objects to see how many channels have been loaded.
     * @return the number of channels that have been loaded.
     */
    public int numChannels() {
        return m_numChannels;
    }

//------------------------------------------------------------------------------
    /**
     * Removes the specified series from this file.
     * @param series	The series that was removed.  Null if it wasn't found.
     */
    public void removeSeries(DataSeries series) {

        // Find the series' location in the list.
        Iterator iter = m_seriesList.iterator();
        int index = 0;
        int location = -1;
        while (iter.hasNext()) {
            DataSeries thisSeries = (DataSeries) iter.next();
            if (thisSeries == series) {
                location = index;
            }
            index++;
        }

        // Remove the series if it was found.
        if (location >= 0) {
            m_seriesList.remove(location);
            m_numChannels--;
        }
    }

//------------------------------------------------------------------------------
    /**
     * Adds the specified series to this file.
     * @param series	The series to add.
     */
    public void addSeries(DataSeries series) {

        m_seriesList.add(series);
        m_numChannels++;
    }

//------------------------------------------------------------------------------
    /**
     * Checks if this file contains the series.
     * @param series	The series to look for.
     * @return	True if the specified series is in this file.
     */
    public boolean containsSeries(DataSeries series) {

        boolean bFound = false;

        Iterator iter = m_seriesList.iterator();
        while (iter.hasNext()) {
            DataSeries thisSeries = (DataSeries) iter.next();
            if (thisSeries == series) {
                bFound = true;
            }
        }
        return bFound;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The name of the last read data file.
     */
    public String getFilename() {

        return m_filename;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The title of the data set.
     */
    public String getTitle() {
        return m_title;
    }

//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param value	The new number of header lines to be read or written.
     */
    public void setNumHeaderLines(int value) {
        m_numHeaderLines = value;
    }

//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param value	The characters to be used as input delimiters.
     */
    public void setInputDelimiters(String value) {
        m_inputDelimiters = value;
    }

//------------------------------------------------------------------------------
    @Override
    public boolean hasOptionsDialog() {

        return true;
    }

//------------------------------------------------------------------------------
    @Override
    public void showOptionsDialog(Frame frame) {

        final DDFConfigDialog dialog = new DDFConfigDialog(frame,
                "Input Config", true);

        // Show the configuration dialog.
        dialog.setConfig(m_numHeaderLines, m_colAssignments);

        if (EventQueue.isDispatchThread()) {
            dialog.setVisible(true);
        } else {
            try {
                javax.swing.SwingUtilities.invokeAndWait(new Runnable() {

                    public void run() {
                        dialog.setVisible(true);
                    }
                });
            } catch (InvocationTargetException ex) {
            } catch (InterruptedException ex) {
            }
        }

        // Don't make changes if Cancel was pressed.
        if (!(dialog.m_bOkPressed)) {
            return;
        }

        // Apply dialog settings to this object.
        m_numHeaderLines = dialog.getNumHeaderLines();
        m_colAssignments = dialog.getColAssignments();
    }

//------------------------------------------------------------------------------
    /**
     * A dialog that allows the user to configure input options.
     */
    public class DDFConfigDialog extends JDialog {

        /** Whether the user pressed the OK button as opposed to Cancel. */
        public boolean m_bOkPressed = false;
        /**
         * A text field for displaying and editing the number of header lines.
         */
        private JTextField numHeadersField = new JTextField();
        /**
         * A text field for displaying and editing the way columns are
         * assigned to data series.
         */
        private JTextField columnField = new JTextField();

        //----------------------------------------------------------------------
        /**
         * Creates the dialog and lays out the components.
         * @param frame The frame to which the dialog should be attached.
         * @param title The title of the dialog.
         * @param modal Whether the dialog is modal (must be closed before
         *              continuing).
         */
        public DDFConfigDialog(Frame frame, String title, boolean modal) {

            super(frame, title, modal);

            try {
                jbInit();
                pack();
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }

        //----------------------------------------------------------------------
        /**
         * Creates the dialog and lays out the components.
         * Calls constructor with parameters (null, "", false)
         */
        public DDFConfigDialog() {

            this(null, "", false);
        }

        //----------------------------------------------------------------------
        /**
         * Lay out components in the dialog.
         * @throws Exception
         */
        void jbInit() throws Exception {

            // Initialize components.
            JPanel mainPanel = new JPanel();
            JPanel topPanel = new JPanel();
            JPanel lowerPanel = new JPanel();
            JButton okButton = new JButton("OK");
            JButton cancelButton = new JButton("Cancel");
            JPanel headerPanel = new JPanel();
            JPanel columnPanel = new JPanel();

            // Set up button listeners.
            okButton.addActionListener(new java.awt.event.ActionListener() {

                public void actionPerformed(ActionEvent e) {
                    okButton_actionPerformed(e);
                }
            });
            cancelButton.addActionListener(new java.awt.event.ActionListener() {

                public void actionPerformed(ActionEvent e) {
                    cancelButton_actionPerformed(e);
                }
            });

            // Set the layout managers for the panels and compose the items
            // within the containers.
            mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
            mainPanel.add(Box.createRigidArea(new Dimension(350, 5)));

            topPanel.setLayout(new BoxLayout(topPanel, BoxLayout.Y_AXIS));

            headerPanel.setLayout(new BoxLayout(headerPanel, BoxLayout.X_AXIS));
            headerPanel.add(new JLabel("Number of Header Lines"));
            headerPanel.add(Box.createRigidArea(new Dimension(5, 0)));

            headerPanel.add(numHeadersField);
            headerPanel.add(Box.createRigidArea(new Dimension(0, 5)));

            mainPanel.add(headerPanel);
            mainPanel.add(Box.createRigidArea(new Dimension(0, 5)));
            /*
            columnField.setEditable(false);
            columnPanel.setLayout(new BoxLayout(columnPanel, BoxLayout.X_AXIS));
            columnPanel.add(new JLabel("Column Assignments"));
            columnPanel.add(Box.createRigidArea(new Dimension(5,0)));
            columnPanel.add(Box.createHorizontalGlue());
            columnPanel.add(columnField);
            mainPanel.add(columnPanel);
            mainPanel.add(Box.createRigidArea(new Dimension(0,5)));
             */
            mainPanel.add(Box.createRigidArea(new Dimension(0, 5)));
            lowerPanel.setLayout(new BoxLayout(lowerPanel, BoxLayout.X_AXIS));

            lowerPanel.add(okButton);
            lowerPanel.add(Box.createRigidArea(new Dimension(5, 0)));
            lowerPanel.add(cancelButton);

            mainPanel.add(lowerPanel);

            getContentPane().add(mainPanel);
        }

        //----------------------------------------------------------------------
        /**
         * Configure the dialog.
         * @param numHeaderLines The number of header lines to use as the
         *                       default when the dialog is presented.
         * @param colAssignments The column assignments to use as the
         *                       default when the dialog is presented.
         */
        void setConfig(int numHeaderLines, String colAssignments) {

            numHeadersField.setText("" + numHeaderLines);
            columnField.setText(colAssignments);
        }

        //----------------------------------------------------------------------
        /**
         * Accessor method.
         * @return The number of header lines specified in the dialog.
         */
        int getNumHeaderLines() {

            return Integer.parseInt(numHeadersField.getText());
        }

        //----------------------------------------------------------------------
        /**
         * Accessor method.
         * @return The column assignments specified in the dialog.
         */
        String getColAssignments() {

            return columnField.getText();
        }

        //----------------------------------------------------------------------
        /**
         * Handles the event of the OK button being pressed.  Closes the dialog.
         * @param e The event that occurred.
         */
        void okButton_actionPerformed(ActionEvent e) {

            m_bOkPressed = true;

            dispose();
        }

        //----------------------------------------------------------------------
        /**
         * Handles the event of the Cancel button being pressed.
         * Closes the dialog.
         * @param e The event that occurred.
         */
        void cancelButton_actionPerformed(ActionEvent e) {

            m_bOkPressed = false;
            dispose();
        }
    }
}
