/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2005-11-28
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import java.io.*;
import java.util.*;

//==============================================================================
/**
 * Reads and writes from/to a file and stores an array of data in one or more
 * dimensions.
 */
public abstract class DimensionalDataFile {

    /** The title of the data set. */
    protected String m_title = "";
    /** The name of the file. */
    protected String m_filename = "";
    /** A list of lines that are read or written at the beginning of the file.*/
    protected List m_headerLines =
        Collections.synchronizedList(new ArrayList());
    /** The number of header lines that should be read. */
    protected int m_numHeaderLines = 0;
    /**
     * A String containing the characters which will be treated as delimiters
     * between data fields in the input file.
     */
    protected String m_inputDelimiters = " \t\n\r\f,";
    /** A flexible list of data series which comprise the set. */
    public List m_seriesList = Collections.synchronizedList(new ArrayList());
    /** An array containing all the data set's data arrays. */
    protected double[][] m_seriesArrays;
    /** The number of data arrays in the set. */
    protected int m_numChannels;
    /** An array holding the minimum values for each dimension. */
    protected int[] m_min;
    /** An array holding the maximum values for each dimension. */
    protected int[] m_max;
    /** The number of dimensions in each data series. */
    protected int m_numDimensions;
    /** The data source buffered reader. */
    protected BufferedReader m_br = null;
    /** The data destination buffered writer. */
    protected BufferedWriter m_bw = null;

//------------------------------------------------------------------------------
/**
 * Creates a data set with the specified number of dimensions.
 * @param numDimensions The number of dimensions (x, y, z,...) of the data set.
 */
    public DimensionalDataFile (int numDimensions) {

        m_numDimensions = numDimensions;
        m_min = new int[m_numDimensions];
        m_max = new int[m_numDimensions];
    }

//------------------------------------------------------------------------------
/**
 * Load a file from disk.
 * <p><b> Pre Conditions </b><br>
 *      The filename is not null.
 * @param filename  The name of the file to load.
 * @return A list of DataSeries objects that hold the data from the file.
 * @throws IOException
 */
    public List load (String filename) throws IOException {

    	// Open the file.
        m_br = new BufferedReader(new FileReader(filename));
        // Read the header.
        readHeader();
        // Read the data.
        readData();
        // Close the file.
        m_br.close();

        m_filename = filename;

        // Return the data list with whatever was put in it.
        return m_seriesList;
    }

//------------------------------------------------------------------------------
/**
 * Read the header information from an open file attached to a reader.
 * Called by load().
 * @throws IOException
 * @see gov.usgs.gr.data.DelimitedDataFile#setHeaderLines setHeaderLines()
 */
    protected void readHeader () throws IOException {

    	for (int i = 0; i < m_numHeaderLines; i++) {
        	m_headerLines.add(AsciiIO.readLine(m_br));
        }
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.  Called by load().
 * Should be overridden to actually read a certain format.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 * @throws IOException
 */
    protected void readData ()  throws IOException {
    }

//------------------------------------------------------------------------------
/**
 * Saves the data to a file.
 * <p><b> Pre Conditions </b><br>
 *      The filename is not null.
 * @param filename  The name of the file to save to.
 * @throws IOException
 */
    public void save(String filename) throws IOException {

    	// Don't do anything if there is no data to write.
    	if (m_seriesList.size() == 0) return;

    	// Open the file.
        m_bw = new BufferedWriter(new FileWriter(filename));

        // Read the header.
        writeHeader();
        // Read the data.
        writeData();
        // Close the file.
        m_bw.close();

        m_filename = filename;
    }

//------------------------------------------------------------------------------
/**
 * Write the header information to an open file attached to a reader.
 * Called by save();  Should be overridden to actually write a certain format.
 * @throws IOException
 */
    private void writeHeader () throws IOException {

    }

//------------------------------------------------------------------------------
/**
 * Write the data back to a file.  Called by save().  Should be overridden to
 * actually write a certain format.
 * <p><b> Pre Conditions </b><br>
 *      The header has been written with writeHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 * @throws IOException
 */
    protected void writeData () throws IOException {

    }

//------------------------------------------------------------------------------
/**
 * Allows other objects to see how many channels have been loaded.
 * @return the number of channels that have been loaded.
 */
    public int numChannels () {

        return m_numChannels;
    }

//------------------------------------------------------------------------------
/**
 * Removes the specified series from this file.
 * @param series	The series to remove.
 */
    public void removeSeries(DataSeries series) {

            // Find the series' location in the list.
            Iterator iter = m_seriesList.iterator();
            int index = 0;
            int location = -1;
            while (iter.hasNext()) {
                DataSeries thisSeries = (DataSeries)iter.next();
                if (thisSeries == series) location = index;
                index++;
            }

            // Remove the series if it was found.
            if (location >= 0) {
                m_seriesList.remove(location);
                m_numChannels--;
            }
    }

//------------------------------------------------------------------------------
/**
 * Adds the specified series to this file.
 * @param series	The series to add.
 */
    public void addSeries(DataSeries series) {

    	m_seriesList.add(series);
        m_numChannels++;
    }

//------------------------------------------------------------------------------
/**
 * Checks if this file contains the series.
 * @param series	The series to look for.
 * @return	True if the specified series is in this file.
 */
    public boolean containsSeries(DataSeries series) {

    	boolean bFound = false;

        // Check each series in the list to see if it matches.
        Iterator iter = m_seriesList.iterator();
        while (iter.hasNext()) {
            DataSeries thisSeries = (DataSeries)iter.next();
            if (thisSeries == series) bFound = true;
        }

        return bFound;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The name of the last read data file.
 */
    public String getFilename(){

        return m_filename;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The title of the data set.
 */
    public String getTitle() {

    	return m_title;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	The new number of header lines to be read or written.
 */
    public void setNumHeaderLines(int value) {

    	m_numHeaderLines = value;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	The characters to be used as input delimiters.
 */
    public void setInputDelimiters(String value) {

    	m_inputDelimiters = value;
    }
}
