/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-06-28
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import gov.usgs.gr.*;
import gov.usgs.gr.visual.gui.*;
import java.nio.*;
import java.util.*;
import org.w3c.dom.*;

//==============================================================================
/**
 * An interleaved, multidimensional data object that stores its data as a single
 * one-dimensional float array.  The data in the array can represent any number
 * of dimensions by interleaving the values at each dimension.  For example, a
 * three-dimensional series (X, Y, and Z) with two points would be stored as
 * (X1, Y1, Z1, X2, Y2, Z2).  The DataSeries keeps track of its beginning, end,
 * min, max, and title.
 */
public class FloatDataSeries extends gov.usgs.gr.data.DataSeries {
    
    /**
     * The XML element name that will identify this class and classes that
     * are extended from it.
     */
    static String XML_ELEMENT_NAME = "floatdataseries";
    
    /** An interleaved array for holding all the data. */
    protected float[] m_array = new float[0];
    /** An interleaved array for holding all selected points. */
    protected float[] m_selected;
    /** A place to save the points that have been deleted. */
    protected List<Float> m_deleted =
        Collections.synchronizedList(new ArrayList());
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/DataSeries.gif";
    
//------------------------------------------------------------------------------
    /**
     * Constructor for creating a blank data series.
     * @param numDimensions The number of dimensions to be held in the object.
     */
    public FloatDataSeries(int numDimensions) {
        
        // Set up the object according to the number of dimensions.
        m_numDimensions = numDimensions;
        m_dims = new DataDimension[m_numDimensions];
        initBounds();
    }
    
//------------------------------------------------------------------------------
    /**
     * Constructor for creating a data series with a data type and title.
     * @param numericType   A code describing the type of data stored.
     * @param title         A string to identify the data series.
     * @param numDimensions The number of dimensions to be held in the object.
     */
    public FloatDataSeries(int numericType, String title, int numDimensions) {
        
        // Store the parameters as member variables.
        m_numericType = numericType;
        m_title = title;
        
        // Set up the object according to the number of dimensions.
        m_numDimensions = numDimensions;
        m_dims = new DataDimension[m_numDimensions];
        
        initBounds();
    }
    
//------------------------------------------------------------------------------
    /**
     * Constructor for creating a data series from an interleaved array of data.
     * @param array        An interleaved array to use as the data for this object.
     * @param numDimensions The number of dimensions to be held in the object.
     */
    public FloatDataSeries(float[] array, int numDimensions) {
        
        // Set up the object according to the number of dimensions.
        m_numDimensions = numDimensions;
        m_dims = new DataDimension[m_numDimensions];
        initBounds();
        
        setArray(array);
    }
    
//------------------------------------------------------------------------------
    /**
     * Creates another DataSeries with the same data and properties as this one.
     * @param bDeep		True if a deep copy (unshared arrays) is to be made.
     * @return    A copy of this data series.
     */
    public GrObject createCopy(boolean bDeep) {
        
        // Create a new series with the same number of dimensions as this one.
        FloatDataSeries series = new FloatDataSeries(m_numDimensions);
        
        // Copy all properties from this series to the new one.
        copyInto(series, bDeep);
        
        // Copy the arrays containing data, selected points, and deleted points.
        series.copyFromDataSeries(this, bDeep);
        
        return series;
    }
    
//------------------------------------------------------------------------------
    /**
     * Copies everything from another data series to this one.
     * @param series    A data series.
     * @param bDeep		True if a deep copy (unshared arrays) is to be made.
     */
    public void copyFromDataSeries(DataSeries ds, boolean bDeep) {
        
        FloatDataSeries series;
        if (ds instanceof FloatDataSeries)
            series = (FloatDataSeries)ds;
        else
            return;
        
        // Get a copy or reference to the main series.
        float[] array;
        if (bDeep) array = series.getArrayCopy();
        else array = series.getArray();
        
        // Copy the main attributes, in case they hadn't been already.
        m_numericType = series.getType();
        m_title = new String(series.getTitle());
        m_numSelected = series.getNumSelected();
        
        // Get references to the selected and deleted arrays.
        float[] selectedArray = series.getSelectedArray();
        int[] selectedIndexes = series.getSelectedIndexes();
        List<Float> deletedList = series.getDeletedList();
        
        if (bDeep) { // Deep copy (by value)...
            // Make a copy of the selected array.
            m_selected = new float[selectedArray.length];
            for (int i = 0; i < selectedArray.length; i++) {
                m_selected[i] = selectedArray[i];
            }
            
            // Make a copy of the selected indexes array.
            m_selectedIndexes = new int[selectedIndexes.length];
            for (int i = 0; i < selectedIndexes.length; i++) {
                m_selectedIndexes[i] = selectedIndexes[i];
            }
            
            // Make a copy of the deleted list.
            m_deleted.clear();
            for (float fVal : deletedList)
                m_deleted.add(fVal);
        } else { // Shallow copy (by reference)...
            // Point this DataSeries' arrays at the other's arrays.
            m_selected = selectedArray;
            m_selectedIndexes = selectedIndexes;
            m_deleted = deletedList;
        }
        
        // Setup variables linked to the arrays.
        setArray(array);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An XML name of this element.
     */
    public String getXmlElementName() {
        return "dataseries";
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
        
        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();
        
        return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {
        
        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);
        
        if (icon == null)
            icon = super.setTreeIcon();
        
        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);
        
        return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Returns an XML version of this element and its children using the
     * Document Object Model.
     * @param doc   A document that can be used to create new DOM elements.
     * @param bIncludeChildren  Whether to include this object's children.
     *                          Ignored here.
     * @return  A new DOM element containing this object and its children.
     */
    public org.w3c.dom.Element getDOMRepresentation(org.w3c.dom.Document doc,
        boolean bIncludeChildren) {
        
        org.w3c.dom.Element el = doc.createElement(getXmlElementName());
        
        GrHashList props = getProperties();
        
        // Just write the id and return if this object has been written out
        // to the file before.
        if (m_bSaveIdOnlyToXml) {
            GrProperty p = (GrProperty)props.get("id");
            el.setAttribute("id", p.value.toString());
            p = (GrProperty)props.get("class");
            el.setAttribute("class", p.value.toString());
            
            return el;
        }
        
        // Make an attribute in the new element for each property of this
        // object that was set to be saved.
        ListIterator iter = props.listIterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            GrProperty p = (GrProperty)props.get(key);
            
            if (p.save == true)
                el.setAttribute(key, p.value.toString());
        }
        
        // Write out the child nodes at the beginning.
        ListIterator iterator = m_children.listIterator(0);
        while (iterator.hasNext()) {
            GrObject child = (GrObject)iterator.next();
            el.appendChild(child.getDOMRepresentation(doc, true));
        }
        
        // Format the number for output.
        java.text.NumberFormat form = null;
//        java.text.NumberFormat[] forms = new java.text.NumberFormat[];
//        java.text.NumberFormat form0 = new java.text.DecimalFormat("0");
//        java.text.NumberFormat form1 = new java.text.DecimalFormat("0.#");
//        java.text.NumberFormat form2 = new java.text.DecimalFormat("0.##");
//        java.text.NumberFormat form3 = new java.text.DecimalFormat("0.###");
//        java.text.NumberFormat form4 = new java.text.DecimalFormat("0.####");
//        java.text.NumberFormat form5 = new java.text.DecimalFormat("0.#####");
//        java.text.NumberFormat form6 = new java.text.DecimalFormat("0.######");
        java.text.NumberFormat formD = new java.text.DecimalFormat(
            "0.##########");
        java.text.NumberFormat formE = new java.text.DecimalFormat(
            "0.########E0#######");
        
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < m_array.length; i++) {
            // Choose the approriate format depending on the size of the
            // number.
/*            if (m_array[i] > 10000000 || m_array[i] < -10000000) form = form7;
            else {
                if (m_array[i] > 16777216 || m_array[i] < -16777216)
                    form = form0;
                else {
                    if (m_array[i] > 8388608 || m_array[i] < -8388608)
                        form = form1;
                    else {
                        if (m_array[i] > 4194304 || m_array[i] < -4194304)
                            form = form2;
                        else {
                            if (m_array[i] > 2097152 || m_array[i] < -2097152)
                                form = form3;
                            else {
                                if (m_array[i] > 100 || m_array[i] < -100)
                                    form = form3;
                                else {
                                    if (m_array[i] > 10 || m_array[i] < -10)
                                        form = form4;
                                    else form = form5;
                                }
                            }
                        }
                    }
                }
            }
 */
            
            if (m_array[i] > 100000000 || m_array[i] < -100000000 ||
                (m_array[i] < 0.001 && m_array[i] > -0.001)) {
                
                form = formE;
            } else {
                form = formD;
            }
            
            // Break the output line after each point.
            if (i % m_numDimensions == 0) sb.append(" " + '\n');
            
            // Chop off trailing zeros, since the DecimalFormatter can't
            // be counted on to do it (at least in JDK 1.2).
            String s = "0";
            if (m_array[i] != 0) {
                if (Float.isNaN(m_array[i]))
                    s = "NaN";
                else
                    s = form.format(m_array[i]);
                int decimalIndex = s.indexOf('.');
                int lastIndex = s.length() - 1;
                if (decimalIndex < 0) decimalIndex = lastIndex + 1;
                char lastChar = s.charAt(lastIndex);
                while ((lastIndex > decimalIndex) && (s.charAt(lastIndex) == '0')) {
                    s = s.substring(0, lastIndex);
                    lastIndex--;
                    lastChar = s.charAt(lastIndex);
                }
            }
            
            //Add the number to the output buffer.
            sb.append(" " + s);
        }
        Text text = doc.createTextNode(sb.toString());
        
        el.appendChild(text);
        
        // If the id was saved then make it so the next time this object
        // is written, only the id will be written.
        if (m_bSaveIdToXml)
            m_bSaveIdOnlyToXml = true;
        
        return el;
    }
    
//------------------------------------------------------------------------------
    /**
     * Read an XML version of this element and its children using the Document
     * Object Model.
     * @param node  The DOM node containing this object's description.
     * @param allObjects A list of of other objects created from the same
     *                  document.  Ignored here.
     * @param unlinkedObjects   Ignored here.
     */
    public void readFromDOM(org.w3c.dom.Node node,
        Map allObjects, List unlinkedObjects) {
        
        // Apply each attribute in the XML element as a property to this object.
        org.w3c.dom.NamedNodeMap map = node.getAttributes();
        int mapLength = map.getLength();
        for (int i = 0; i < mapLength; i++) {
            org.w3c.dom.Node attr = map.item(i);
            setProperty(attr.getNodeName(), attr.getNodeValue());
        }
        
        // Add the child elements as children of this object.
        ArrayList tokenList = new ArrayList();
        NodeList nodeList = node.getChildNodes();
        int numChildren = nodeList.getLength();
        for (int i = 0; i < numChildren; i++) {
            // Look for character data inside the element and interpret it as
            // column data describing the main float data array.
            Node childNode = nodeList.item(i);
            if (childNode instanceof org.w3c.dom.Comment) continue;
            if (!(childNode instanceof org.w3c.dom.CharacterData)) continue;
            String s = ((org.w3c.dom.CharacterData)childNode).getData();
            
            // Parse the line into a String array.
            String inputDelimiters = " \t\n\r\f,";
            StringTokenizer st = new StringTokenizer(s, inputDelimiters);
            
            while (st.hasMoreTokens()) {
                String token = st.nextToken();
                tokenList.add(token);
            }
        }
        
        // Create a new main array and fill it with the data that was parsed.
        float[] newArray = new float[tokenList.size()];
        ListIterator tokenIter = tokenList.listIterator();
        int i = 0;
        while (tokenIter.hasNext()) {
            String token = (String)tokenIter.next();
            int eIndex = token.toUpperCase().indexOf('E');
            if (eIndex < 0) {
                newArray[i++] = Float.parseFloat(token);
            } else {
                // The token was of the form 0.0E0 and requires special parsing.
                double sig = Double.parseDouble(token.substring(0, eIndex));
                double exp = Double.parseDouble(token.substring(eIndex + 1));
                newArray[i++] = (float)(sig * Math.pow(10, exp));
            }
        }
        
        // Set the main array to be the newly created array.
        setArray(newArray);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accepts character data from a SAX parser.  This generally provides faster
     * XML input than DOM.
     * @param buf   A character buffer from which to read.
     * @param offset    The offset into the buffer.
     * @param len       The length of the string to read from the buffer.
     */
    
//------------------------------------------------------------------------------
    /**
     * Aids the receiveCharsFromSAX() method.  Reads tokens from the buffer and
     * stores them in the m_inputTokens list.  They can be converted to an
     * interleaved array later.
     * @param buf   A character buffer from which to read.
     * @param offset    The offset into the buffer.
     * @param len       The length of the string to read from the buffer.
     */
    protected void tokenizeAndStore(String s0) {
        
        Float f0 = Float.parseFloat(s0);
        m_inputTokens.add(f0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Set the range of x-values to be considered active by the series.  Used as
     * an alternate to setIndexRange() when x-values are known, but indexes are not.
     * Finds the new indexes by starting at their old values and moving them.
     * This speeds the common case of gradual changes to the range.
     * @param startVal  The x value to start at.
     * @param endVal    The x value to end at.
     */
    public synchronized void setLimits(float startVal, float endVal) {
        
        // If the beginning is in range...
        if (m_begin > 0 && m_begin < (m_array.length - m_numDimensions + 1)) {
            // If the beginning needs to be moved to a later position...
            if (startVal > m_array[m_numDimensions * m_begin]) {
                // Move the beginning until it is after startVal.
                while (startVal > m_array[m_numDimensions * m_begin]
                    && m_begin < m_end) m_begin += m_numDimensions;
            } else { // Beginning might need to be moved to an earlier position.
                // Move the beginning until it is before endVal.
                while (endVal < m_array[m_numDimensions * m_begin]
                    && m_begin >= 0) m_begin -= m_numDimensions;
            }
        }
        
        // If the end is in range...
        if (m_end > m_begin && m_end < m_numPoints) {
            // If the end needs to be moved to a later position...
            if (endVal > m_array[m_numDimensions * m_end]) {
                // Move the end until is is after endVal.
                while (endVal > m_array[m_numDimensions * m_end] && m_end <
                    (m_numPoints - 1)) {
                    
                    m_end += 1;
                }
            } else { // End might need to be moved to an earlier position.
                // Move the end until it is before endVal.
                while (endVal < m_array[m_numDimensions * m_end]
                    && m_end > m_begin) m_end -= 1;
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Check if any selected points are within the specified area.
     * @param min	Array containing the minimum value in the selected area
     *              for each dimension.
     * @param max	Array containing the maximum value in the selected area
     *              for each dimension.
     * @param bAddIfDeselected	True if unselected points should be selected and
     *							then counted as being contained inside the area.
     * @return	True if a selected point was found within the area.
     */
    public boolean containsSelectedPoint(double[] min, double[] max,
        boolean bAddIfDeselected) {
        
        // Make sure the min and max arrays don't have more elements
        // than this series has dimensions.
        int minLength = min.length;
        if (minLength > m_numDimensions) minLength = m_numDimensions;
        int maxLength = max.length;
        if (maxLength > m_numDimensions) maxLength = m_numDimensions;
        
        // Traverse the array and check if each point is inside the area.
        for (int i = m_begin; i <= m_end; i++) {
            boolean bInBounds = true;
            // Check the minimums for each dimension.
            for (int minIndex = 0; minIndex < minLength; minIndex++) {
                if (m_array[i * m_numDimensions + minIndex] < min[minIndex])
                    bInBounds = false;
            }
            // Check the maximums for each dimension.
            for (int maxIndex = 0; maxIndex < maxLength; maxIndex++) {
                if (m_array[i * m_numDimensions + maxIndex] > max[maxIndex])
                    bInBounds = false;
            }
            
            if (bInBounds) {
                // Check if the point is already in the selected list.
                for (int j = 0; j < m_numSelected; j++) {
                    if (m_selectedIndexes[j] == i) {
                        // Selected point was found to be already stored here.
                        return true;
                    }
                }
                // Either select the point and return true or just return false.
                if (bAddIfDeselected) {
                    selectPoint(i);
                    return true;
                }
            }
        }
        
        // No selected points were found inside the area.
        return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Select all the points in the given area.
     * @param min	Array containing the minimum value in the selected area
     *              for each dimension.
     * @param max	Array containing the maximum value in the selected area
     *              for each dimension.
     * @param bAdding	True if selecting points, false if deselecting.
     */
    public synchronized void selectArea(float[] min, float[] max,
        boolean bAdding) {
        
        // If deselecting, use the optimized version of this method.
        if (!bAdding) {
            deselectArea(min, max);
            return;
        }
        
        // Make sure the min and max arrays don't have more elements
        // than this series has dimensions.
        int minLength = min.length;
        if (minLength > m_numDimensions) minLength = m_numDimensions;
        int maxLength = max.length;
        if (maxLength > m_numDimensions) maxLength = m_numDimensions;
        
        // Try to reduce the area to be checked with a quick search of any
        // sorted dimensions.
        int lowerBoundsIndex = 0;
        int upperBoundsIndex = m_numPoints;
        for (int j = 0; j < minLength; j++) {
            if (m_dims[j].m_sorted) {
                float searchVal;
                if (min[j] > 0)
                    searchVal = min[j] * 0.9999f;
                else
                    searchVal = min[j] * 1.0001f;
                if (Double.isNaN(searchVal))
                    continue;

                int minIndex = binarySearch(searchVal, j) - 1;
                if (minIndex > lowerBoundsIndex)
                    lowerBoundsIndex = minIndex;
                if (max[j] > 0)
                    searchVal = max[j] * 1.0001f;
                else
                    searchVal = max[j] * 0.9999f;
                int maxIndex = binarySearch(searchVal, j) + 1;
                if (maxIndex < upperBoundsIndex)
                    upperBoundsIndex = maxIndex;
            }
        }
        
        // Traverse the array and check if each point is inside the area.
        for (int i = lowerBoundsIndex; i < upperBoundsIndex; i++) {
            boolean bInBounds = true;
            // Check the value against the range for each dimension.
            for (int j = 0; j < minLength; j++) {
                float val = m_array[i * m_numDimensions + j];
                // If bounds contain NaN, regular numbers are out of area.
                if (Float.isNaN(min[j]) || Float.isNaN(max[j])) {
                    if (!Float.isNaN(val)) {
                        bInBounds = false;
                        break;
                    }
                } else {
                    if (val < min[j] || val > max[j]) {
                        bInBounds = false;
                        break;
                    }
                }
            }
            
            if (bInBounds) {
                // Check if the point is already in the selected list.
                int selectedIndex = -1;
                for (int j = 0; j < m_numSelected; j++) {
                    if (m_selectedIndexes[j] == i) {
                        // Selected point was found to be already stored here.
                        selectedIndex = j;
                        break;
                    }
                }
                
                // Add the point if it wasn't found in the search above.
                if (selectedIndex < 0)
                    selectPoint(i);
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Remove the points in the given area from the current selection.
     * Works like selectArea, but optimized for deselecting.
     * @param min	Array containing the minimum value in the selected area
     *              for each dimension.
     * @param max	Array containing the maximum value in the selected area
     *              for each dimension.
     */
    public synchronized void deselectArea(float[] min, float[] max) {
        
        // Make sure the min and max arrays don't have more elements
        // than this series has dimensions.
        int minLength = min.length;
        if (minLength > m_numDimensions) minLength = m_numDimensions;
        int maxLength = max.length;
        if (maxLength > m_numDimensions) maxLength = m_numDimensions;
        
        // Traverse the selected array and check if each point is inside the
        // area.
        int n = 0;
        while (n < m_numSelected) {
            boolean bKept = false;
            // Let i be the index into the data array.
            int i = m_selectedIndexes[n];
            // Check the values for each dimension.
            for (int j = 0; j < minLength; j++) {
                float val = m_array[i * m_numDimensions + j];
                if (val < min[j] || val > max[j]) {
                    bKept = true;
                    break;
                }
            }
            if (bKept)
                n++;
            else
                deselectPoint(n);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Select a single point, but only if it is in bounds and unselected.
     * @param index	The index of the point in the main series.
     */
    public synchronized void selectPointSafely(int index) {
        
        if (index >= 0 && index < m_selectedIndexes.length) {
            // Check if the point is already in the selected list.
            int selectedIndex = -1;
            for (int j = 0; j < m_numSelected; j++) {
                if (m_selectedIndexes[j] == index) {
                    // Selected point was found to be already stored here.
                    selectedIndex = j;
                    break;
                }
            }
            
            // Add the point if it wasn't found in the search above.
            if (selectedIndex < 0) {
                try {
                    selectPoint(index);
                } catch (Exception e) {
                    e.printStackTrace();
                }
                m_bHasChanged = true;
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Select a single point.
     * @param index	The index of the point in the main series.
     */
    public synchronized void selectPoint(int index) {
        
        // Add the new index to the list.
        m_selectedIndexes[m_numSelected] = index;
        
        // Copy the value for each dimension of this point from
        // the main series to the selected array.
        if (m_numDimensions == 2) {
            for (int k = 0; k < m_numDimensions; k++) {
                m_selected[m_numSelected * m_numDimensions + k]
                    = m_array[index * m_numDimensions + k];
            }
        }
        
        m_numSelected++;
    }
    
//------------------------------------------------------------------------------
    /**
     * Deselect a single point.
     * @param index	The index of the point in the array of selected points.
     */
    public synchronized void deselectPoint(int index) {
        
        // Move the last selected point to cover up this one.
        m_selectedIndexes[index] = m_selectedIndexes[m_numSelected - 1];
        
        // Copy the value for each dimension of this point from
        // the end of the array to cover up this point.
        if (m_numDimensions == 2) {
            for (int k = 0; k < m_numDimensions; k++) {
                m_selected[index * m_numDimensions + k]
                    = m_array[m_selectedIndexes[(m_numSelected - 1)]
                    * m_numDimensions + k];
            }
        }
        
        m_numSelected--;
    }
    
//------------------------------------------------------------------------------
    /**
     * Delete all selected points from the data series.  Copies points that were
     * not selected to a new array and sets that array as the new array for the
     * DataSeries.  List of selected point is emptied.  Deleted point values for
     * each dimension are added to the deleted list.
     */
    public synchronized void deleteSelectedPoints() {
        
        // Make a new array to hold the unselected points.
        int newLen = (m_numPoints - m_numSelected) * m_numDimensions;
        float[] newArray = new float[newLen];
        
        // Traverse the main array and copy all unselected points to the
        // new array.
        int index = 0;
        for (int i = 0; i < m_numPoints; i++) {
            boolean bShouldDelete = false;
            
            // Traverse the selected index array to see if this point is
            // selected.
            for (int j = 0; j < m_numSelected; j++) {
                
                // Check if this index is selected.
                if (m_selectedIndexes[j] == i) {
                    bShouldDelete = true;
                    // Move the value at the end to the spot that matched and
                    // shrink the array by one place.
                    m_selectedIndexes[j] = m_selectedIndexes[m_numSelected - 1];
                    m_numSelected--;
                    
                    // Quit looking to see if it was selected
                    break;
                }
            }
            
            // Let this point live if it wasn't selected.
            if (bShouldDelete) {
                if (m_numDimensions == 2) {
                    for (int k = 0; k < m_numDimensions; k++) {
                        m_deleted.add(m_array[i * m_numDimensions + k]);
                    }
                }
            } else {
                for (int k = 0; k < m_numDimensions; k++) {
                    newArray[index * m_numDimensions + k]
                        = m_array[i * m_numDimensions + k];
                }
                index++;
            }
        }
        
        // Make the new array current.
        setArray(newArray);
        
        // For X-Y array, don't allow all points to be deleted, or else there
        // will be nothing left to select.
        if (m_numDimensions == 2) {
            restoreIfAllDeleted();
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Modify all selected points in some way using the value parameter.
     * Changes are applied to the original array to minimize memory usage.
     * For 2D series, updates the selected list to match the new data.
     * @param modificationType	The type of modification this method should
     *			perform.  0=set equal, 1=add, 2=subtract, 3=multiply,
     *                  4=divide, 5=absolute value, 6=power, 7=log10.
     * @param dimension	The dimension number to modify.
     * @param value	The amount to move each point
     */
    public synchronized void modifySelectedPoints(int modificationType,
        int dimension, float value) {
        
        // Don't proceed if the dimension number isn't valid.
        if (dimension < 0 || dimension >= m_numDimensions)
            return;
        
	// Traverse the array and modify the specified dimension using value.
        switch (modificationType) {
            case 0: // Set
                for (int i = 0; i < m_numSelected; i++) {
                    int index = m_selectedIndexes[i] * m_numDimensions
                        + dimension;
                    m_array[index] = value;
                }
                break;
            case 1: // Add
                for (int i = 0; i < m_numSelected; i++) {
                    int index = m_selectedIndexes[i] * m_numDimensions
                        + dimension;
                    m_array[index] += value;
                }
                break;
            case 2: // Subtract
                for (int i = 0; i < m_numSelected; i++) {
                    int index = m_selectedIndexes[i] * m_numDimensions
                        + dimension;
                    m_array[index] -= value;
                }
                break;
            case 3: // Multiply
                 for (int i = 0; i < m_numSelected; i++) {
                    int index = m_selectedIndexes[i] * m_numDimensions
                        + dimension;
                    m_array[index] *= value;
                 }
                break;
            case 4: // Divide
                for (int i = 0; i < m_numSelected; i++) {
                    int index = m_selectedIndexes[i] * m_numDimensions
                        + dimension;
                    m_array[index] /= value;
                }
                break;
            case 5: // Power
                for (int i = 0; i < m_numSelected; i++) {
                    int index = m_selectedIndexes[i] * m_numDimensions
                        + dimension;
                    m_array[index] = (float)Math.pow(m_array[index], value);
                }
                break;
            case 6: // Log10
                for (int i = 0; i < m_numSelected; i++) {
                    int index = m_selectedIndexes[i] * m_numDimensions
                        + dimension;
                    m_array[index] = (float)Math.log10(m_array[index]);
                }
                break;
            case 7: // Absolute Value
                for (int i = 0; i < m_numSelected; i++) {
                    int index = m_selectedIndexes[i] * m_numDimensions
                        + dimension;
                    m_array[index] = Math.abs(m_array[index]);
                }
                break;
        }

        // Snap to the nearest snap value;
        if (m_dims[dimension].m_snap > 0) {
            for (int i = 0; i < m_numSelected; i++) {
                int index = m_selectedIndexes[i] * m_numDimensions
                    + dimension;
                m_array[index] = snap(m_array[index], dimension);
            }
        }

        if (m_numDimensions == 2) {
            for (int i = 0; i < m_numSelected; i++) {
                int index = m_selectedIndexes[i] * m_numDimensions
                    + dimension;
                m_selected[i * m_numDimensions + dimension] =
                    m_array[index];
            }
        }
        
        // Update the min and max variables that describe the series.
        updateMinMax();
        
        // Notify observers of the change to this object.
        m_bHasChanged = true;
        notifyObservers();
    }
    
//------------------------------------------------------------------------------
    /**
     * Modify all selected points in some way with regard to a second data series.
     * Changes are applied to the original array to minimize memory usage.
     * For 2D series, updates the selected list to match the new data.
     * @param modificationType	The type of modification this method should
     *			perform.  0=set equal, 1=add, 2=subtract, 3=multiply, 4=divide
     * @param dimension	The dimension number to modify.
     * @param series	The data series used to modify this series.
     */
    public synchronized void modifySelectedPoints(int modificationType,
        int dimension, FloatDataSeries series) {
        
        // Don't proceed if dimension number is out of range.
        if (dimension < 0 || dimension >= m_numDimensions) return;
        
        // Initialize local variables.
        float xVal;
        boolean bTrim = false;
        float trimMin = (float)m_dims[0].m_min;
        float trimMax = (float)m_dims[0].m_max;
        
        // Traverse the array and shift the last dimension by the value
        // of the second series at each point.
        // Modify the specified dimension.
        for (int i = 0; i < m_numSelected; i++) {
            xVal = m_array[m_selectedIndexes[i] * m_numDimensions];
            // Check that the second series is defined over the same range
            // and that no point is modified outside of the intersection of
            // the two series' ranges.
            if (xVal < series.getMin(0)) {
                bTrim = true;
                trimMin = (float)series.getMin(0);
            }
            if (xVal > (float)series.getMax(0)) {
                bTrim = true;
                trimMax = (float)series.getMax(0);
            }
            
            // Find the corresponding value from the other series and use it to
            // modify this one.
            float value = series.getInterpolatedValue(xVal);
            int index = m_selectedIndexes[i] * m_numDimensions + dimension;
            switch (modificationType) {
                case 0: // Set
                    m_array[index] = value;
                    break;
                case 1: // Add
                    m_array[index] += value;
                    break;
                case 2: // Subtract
                    m_array[index] -= value;
                    break;
                case 3: // Multiply
                    m_array[index] *= value;
                    break;
                case 4: // Divide
                    m_array[index] /= value;
                    break;
            }
            
            // For X-Y data series, keep track of the point so it can still
            // remain selected after the modification.
            if (m_numDimensions == 2) {
                m_selected[i * m_numDimensions + dimension] = m_array[index];
            }
        }
        
        // Update the min and max variables that describe the series.
        updateMinMax();
        
        if (bTrim) {
            // Trim off the points whose x values don't lie within
            // the intersection of this series and the other series.
            trimDimension(0, trimMin, trimMax);
        }
        
        // Notify observers of the change to this object.
        m_bHasChanged = true;
        notifyObservers();
    }
    
//------------------------------------------------------------------------------
    /**
     * Delete all points who's value in the specified dimension fall outside the
     * specified min and max.  Creates a new array equal in size to the original
     * to hold points that were within range.  Creates a third array just large
     * enough to hold points that were within range.  Updates selected indexes list
     * for points that were kept.
     * @param dimension The number of the dimension that determines the range.
     * @param min	The new minimum range value for the dimension.
     * @param max	The new maximum range value for the dimension.
     */
    public synchronized void trimDimension(int dimension, float min,
        float max) {
        
        // Don't proceed if the dimension number is out of range.
        if (dimension < 0 || dimension >= m_numDimensions) return;
        
        // Make a new array to hold the points that are within range.
        float[] newArray = new float[m_array.length];
        int[] oldSelectedIndexes = m_selectedIndexes;
        m_selectedIndexes = new int[m_array.length];
        int oldNumSelected = m_numSelected;
        m_numSelected = 0;
        
        // Traverse the main array and copy all points that are within range.
        int index = 0;
        for (int i = m_begin; i <= m_end; i++) {
            boolean bShouldDelete = false;
            if (m_array[i * m_numDimensions + dimension] < min
                || m_array[i * m_numDimensions + dimension] > max) {
                
                bShouldDelete = true;
            }
            
            // Traverse the selected index array to see if the point that was
            // deleted was listed as selected.  Remove it from the selected
            // list if it was.
            if (bShouldDelete) {
                for (int j = 0; j < oldNumSelected; j++) {
                    
                    // Check if this index is selected.
                    if (oldSelectedIndexes[j] == i) {
                        // Move the value at the end to the spot that matched
                        // and shrink the array by one place.
                        
                        oldSelectedIndexes[j]
                            = oldSelectedIndexes[oldNumSelected - 1];
                        oldNumSelected--;
                        // Quit looking to see if it was selected
                        break;
                    }
                }
            } else {
                // Save the point into the new array.
                for (int k = 0; k < m_numDimensions; k++) {
                    newArray[index * m_numDimensions + k]
                        = m_array[i * m_numDimensions + k];
                }
                
                // If this point was selected, put it in the new selected
                // indexes.
                for (int j = 0; j < oldNumSelected; j++) {
                    // Check if this index points to an index higher than the
                    // one that was deleted.
                    if (oldSelectedIndexes[j] == i) {
                        // Decrement the index it points to.
                        m_selectedIndexes[m_numSelected] = index;
                        m_numSelected++;
                        break;
                    }
                }
                
                // Increment the number of points that haven't been deleted
                // and copied to the new array.
                index++;
            }
        }
        
        // Check if the new array is too small for it's size
        // (elements were trimmed).
        if ((index * m_numDimensions) < newArray.length) {
            // Copy the array to one that fits it.
            float[] newArray2 = new float[index * m_numDimensions];
            for (int i = 0; i < index; i++) {
                for (int k = 0; k < m_numDimensions; k++) {
                    newArray2[i * m_numDimensions + k]
                        = newArray[i * m_numDimensions + k];
                }
            }
            
            // Check that there are at least two points left.
            if (index >= (2 * m_numDimensions))
                setArray(newArray2);
        } else {
            // Make the new array current.
            if (index >= (2 * m_numDimensions))
                setArray(newArray);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the y value of a point along the curve.
     * @param xVal	The x coordinate of the interpolated point.
     * @return		The y coordinate of the interpolated point.
     */
    public float getInterpolatedValue(float xVal) {
        
        return getInterpolatedValue(xVal, 0, 1);
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the y value of a point along the curve.  Uses the method
     * parabolicallyInterpolate() if the m_bParabolicInterpolation variable is set.
     * Otherwise, uses the linearlyInterpolate() method.
     * @param xVal	The x coordinate of the interpolated point.
     * @param xDim  The dimension to use as the X dimension.
     * @param yDim  The dimension to use as the Y dimension.
     * @return		The y coordinate of the interpolated point.
     */
    public float getInterpolatedValue(float xVal, int xDim, int yDim) {
        
        if (m_bParabolicInterpolation) {
            return parabolicallyInterpolate(xVal, xDim, yDim);
        } else {
            return linearlyInterpolate(xVal, xDim, yDim);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the value of the Y dimension at the given coordinates by using
     * linear interpolation.  Calls linearlyInterpolate(xVal, 0, 1).
     * @param xVal	The X (first dimension) coordinate of the interpolated point.
     * @return The interpolated Y (second dimension) value.
     */
    public float linearlyInterpolate(float xVal) {
        
        return linearlyInterpolate(xVal, 0, 1);
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the value of a dimension at given coordinates by linearly interpolating
     * between points.
     * @param xVal	The X coordinate of the interpolated point.
     * @param xDim  The dimension to use as the X dimension.
     * @param yDim  The dimension to use as the Y dimension.
     * @return The interpolated Y value.
     */
    public float linearlyInterpolate(float xVal, int xDim, int yDim) {
        
        // Find the existing points before and after the interpolated point.
        int nearestIndex = binarySearch(xVal, xDim);
        if (nearestIndex > (getCount() - 1)) nearestIndex = getCount() - 1;
        if (nearestIndex < 0) nearestIndex = 0;
        int index1 = 0;
        if (m_array[nearestIndex * m_numDimensions + xDim] > xVal) {
            index1 = nearestIndex - 1;
        } else {
            index1 = nearestIndex;
        }
        
        return linearlyInterpolate(index1, xVal, xDim, yDim);
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the value of a dimension at given coordinates by linearly interpolating
     * between the point at index1 and the one following it.
     * @param index1 The index preceding xVal.
     * @param xVal	The X coordinate of the interpolated point.
     * @param xDim  The dimension to use as the X dimension.
     * @param yDim  The dimension to use as the Y dimension.
     * @return The interpolated Y value.
     */
    public float linearlyInterpolate(int index1, float xVal, int xDim,
        int yDim) {
        
        // Set the following index to be the one after index 1;
        int index2 = index1 + 1;
        
        // No need to interpolate if it is past the start or end of the series.
        if (index1 < 0) index1 = 0;
        if (index2 > (getCount() - 1)) index2 = getCount() - 1;
        if (index1 == index2) {
            float interpolatedValue =
                m_array[index1 * m_numDimensions + yDim];
            return interpolatedValue;
        }
        
        // Measure the distance between the points.
        float dist1to2 = Math.abs(m_array[index2 * m_numDimensions + xDim]
            - m_array[index1 * m_numDimensions + xDim]);
        float dist1 = Math.abs(xVal - m_array[index1 * m_numDimensions + xDim]);
        float dist2 = Math.abs(m_array[index2 * m_numDimensions + xDim] - xVal);
        float val1 = m_array[index1 * m_numDimensions + yDim];
        float val2 = m_array[index2 * m_numDimensions + yDim];
        
        // Return the exact original value if the points before and after
        // are equal.
        if (val1 == val2) return val1;
        
        // Linearly interpolate between the points.
        float interpolatedValue = (val1 * (dist2 / dist1to2))
        + (val2 * (dist1 / dist1to2));
        
        return interpolatedValue;
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the value of dimension 1 at the given coordinates along dimension 0 by
     * interpolating between points.  Creates a parabola through the preceding
     * two points and and following point and another parabola through the
     * preceding point and the two following points.  The result is the weighted
     * average of these two at the specified X value.
     * Calls parabolicallyInterpolate(xVal, 0, 1).
     * @param xVal	The x-coordinate of the interpolated point.
     * @return  The weighted average of two parabolas at the given X value.
     */
    public float parabolicallyInterpolate(float xVal) {
        
        return parabolicallyInterpolate(xVal, 0, 1);
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the value of a dimension at the given coordinates by
     * interpolating between points.  Creates a parabola through the preceding
     * two points and and following point and another parabola through the
     * preceding point and the two following points.  The result is the weighted
     * average of these two at the specified X value.
     * @param xVal	The x-coordinate of the interpolated point.
     * @param xDim  The dimension to use as the X dimension.
     * @param yDim  The dimension to use as the Y dimension.
     * @return  The weighted average of two parabolas at the given X value.
     */
    public float parabolicallyInterpolate(float xVal, int xDim, int yDim) {
        
        // Return the nearest y value, if the x value is out of bounds.
        float xMin = m_array[xDim];
        float yMin = m_array[yDim];
        if (xVal < xMin) return yMin;
        float xMax = m_array[(getCount() - 1) * m_numDimensions + xDim];
        float yMax = m_array[(getCount() - 1) * m_numDimensions + yDim];
        if (xVal > xMax) return yMax;
        
        // Find the existing points that bound the interpolated point.
        int nearestIndex = binarySearch(xVal);
        if (nearestIndex < 0) nearestIndex = 0;
        if (nearestIndex > (getCount() - 1)) nearestIndex = getCount() - 1;
        int index0 = 0;
        int index1 = 0;
        int index2 = 0;
        int index3 = 0;
        if (m_array[nearestIndex * m_numDimensions + xDim] > xVal) {
            index1 = nearestIndex - 1;
            index2 = nearestIndex;
        } else {
            index1 = nearestIndex;
            index2 = nearestIndex + 1;
        }
        index0 = index1 - 1;
        index3 = index2 + 1;
        
        // No need to interpolate if it is past the start or end of the series.
        if (index1 < 0) index1 = index2;
        if (index1 == 0) {
            index0 = 0;
            index1 = 0;
            index2 = 1;
            index3 = 2;
        }
        if (index3 > (getCount() - 1)) index3 = getCount() - 1;
        if (index1 > index2 || index2 > index3) {
            if (index2 > (getCount() - 1)) index2 = getCount() - 1;
            float interpolatedValue =
                m_array[index2 * m_numDimensions + yDim];
            return interpolatedValue;
        }
        
        // Find the middle coefficients of the parabola.
        float x2 = m_array[index1 * m_numDimensions + xDim];
        float y2 = m_array[index1 * m_numDimensions + yDim];
        float x3 = m_array[index2 * m_numDimensions + xDim];
        float y3 = m_array[index2 * m_numDimensions + yDim];
        
        // Interpolate the left and right parabolas for a final value.
        
        // Interpolate the value using the two preceding points and the
        // following point.
        float leftValue = 0;
        if (index0 == index1) {
            // Use linear interpolation for the first segment of the series.
            leftValue = linearlyInterpolate(index1, xVal, xDim, yDim);
        } else {
            float x1 = m_array[index0 * m_numDimensions + xDim];
            float y1 = m_array[index0 * m_numDimensions + yDim];
            double[] coeffs1 = fitParabola(x1, y1, x2, y2, x3, y3);
            
            // Use the parabola to interpolate the y value.
            leftValue = (float)(coeffs1[0] * xVal * xVal
                + coeffs1[1] * xVal + coeffs1[2]);
        }
        
        // Interpolate the value using the preceding point and the
        // two following points.
        float rightValue = 0;
        if (index2 == index3) {
            // Use linear interpolation for the last segment of the series.
            rightValue = linearlyInterpolate(index1, xVal, xDim, yDim);
        } else {
            float x4 = m_array[index3 * m_numDimensions + xDim];
            float y4 = m_array[index3 * m_numDimensions + yDim];
            double[] coeffs2 = fitParabola(x2, y2, x3, y3, x4, y4);
            
            // Use the parabola to interpolate the y value.
            rightValue = (float)(coeffs2[0] * xVal * xVal
                + coeffs2[1] * xVal + coeffs2[2]);
        }
        
        // Use a weighted average between the left and right values.
        float diff = x3 - x2;
        leftValue *= (diff - (xVal - x2));
        leftValue /= diff;
        rightValue *= (xVal - x2);
        rightValue /= diff;
        
        float interpolatedValue = leftValue + rightValue;
        
        return interpolatedValue;
    }
    
//------------------------------------------------------------------------------
    /**
     * Given three points (x1, y1), (x2, y2), (x3, y3), find the equation for the
     * parabola that goes through those points.  The equation is of the form
     * ax^2 + bx + c = y.
     * @param x1	The x coordinate of the first point.
     * @param y1	The y coordinate of the first point.
     * @param x2	The x coordinate of the second point.
     * @param y2	The y coordinate of the second point.
     * @param x3	The x coordinate of the third point.
     * @param y3	The y coordinate of the third point.
     * @return		An array of floats containing the equation coefficients a,
     *				b, and c.
     */
    public static final double[] fitParabola(double x1, double y1, double x2,
        double y2, double x3, double y3) {
        
        double coeffs[] = {0, 0, 0};
        
        // Solve for a.
        double aNumerator = (y3 - y1) * (x2 - x1) - (x3 - x1) * (y2 - y1);
        double aDenominator = x1 * (x2 * (x2 - x1) + x3 * (x1 - x3))
        + x2 * x3 * (x3 - x2);
        if (aDenominator == 0) coeffs[0] = 0;
        else coeffs[0] = aNumerator / aDenominator;
        
        // Solve for b.
        double bNumerator = coeffs[0] * (x1 + x2) * (x1 - x2) + y2 - y1;
        double bDenominator = x2 - x1;
        if (bDenominator == 0) coeffs[1] = 0;
        else coeffs[1] = bNumerator / bDenominator;
        
        // Solve for c.
        coeffs[2] = y1 - x1 * (coeffs[0] * x1 + coeffs[1]);
        
        return coeffs;
    }
    
//------------------------------------------------------------------------------
    /**
     * Search the interleaved array and return the nearest index along the
     * first dimension.  Calls binarySearch(key, 0).
     * @param key	The value that is being searched for.
     * @return The index or nearest index that holds a value matching the key.
     */
    protected int binarySearch(float key) {
        
        return binarySearch(key, 0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Search the interleaved array and return the nearest index along the
     * specified dimension whose value is less than key.
     * Modified from binary search example in _C++ Object-Oriented Data Structures_
     * by Sengupta and Korobkin.
     * <p><b> Pre Conditions </b><br>
     *      The "X" dimension of the interleaved array is sorted from lowest
     *		to highest.
     * @param key	The value that is being searched for.
     * @param xDim  The dimension to use as the X dimension.
     * @return The index that holds a value matching the key or, if no match is
     *         found, the nearest index less than key.
     */
    protected int binarySearch(float key, int xDim) {
        
        if (m_numPoints == 0) return -1;
        
        int end = getCount() - 1;
        int first = 0;
        int last = end - 1;
        int middle = 0;
        
        // Return right away if key is less than the lowest value.
        if (key < m_array[xDim]) return 0;
        // Return right away if key is greater than the largest value.
        if (key >= m_array[end * m_numDimensions + xDim]) return end;
        
        // Continue to break the array in half until a match is found or it
        // is broken down to a single value.
        while (first <= last) {
            // Calculate the midpoint of the current range.
            middle = (first + last) / 2 + (first + last) % 2;
            
            // Determine the range between this point and the next.
            float upper;
            if (middle > end) upper = Float.MAX_VALUE;
            else upper = m_array[(middle + 1) * m_numDimensions + xDim];
            
            if (key >= m_array[middle * m_numDimensions + xDim]
                && key < upper) {
                
                // Return if an exact match was found.
                return (middle);
            } else {
                // Reset the range to either the left half or the right half.
                if (key > m_array[middle * m_numDimensions + xDim])
                    first = middle + 1;
                else last = middle - 1;
            }
        }
        
        return (middle);
    }
    
//------------------------------------------------------------------------------
    /**
     * Resample the data to the specified interval along the 0 dimension, and
     * beginning at the specifed offset.  Calls resample(interval, offset, 0).
     * <p><b> Post Conditions </b><br>
     *      This series may no longer match the sampling of other series from its
     *		data file.
     * @param interval	The spacing between points.
     * @param offset	The amount to offset the points from zero.
     */
    public void resample(float interval, float offset) {
        
        resample(interval, offset, 0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Resample the data to the specified interval along the X dimension, and
     * beginning at the specifed offset.
     * <p><b> Post Conditions </b><br>
     *      This series may no longer match the sampling of other series from its
     *		file.
     * @param interval	The spacing between points.
     * @param offset	The amount to offset the points from zero.
     * @param xDim  The dimension to use as the X dimension.
     */
    public synchronized void resample(float interval, float offset, int xDim) {
        
        // Use the absolute value of the supplied interval.
        interval = Math.abs(interval);
        float halfInterval = interval / 2;
        
        // Include a little more area (20% of the interval above and below)
        // just so rounding doesn't keep any points on the boundary from being
        // included in the interpolation.
        float min = (float)(getMin(xDim) - (interval / 20));
        float max = (float)(getMax(xDim) + (interval / 20));
        float startOffset = min - offset;
        float startIntervals = (int)(startOffset / interval);
        if (startOffset > 0) startIntervals++;
        float start = (interval * startIntervals) + offset;
        int size = (int)((max - start) / interval) + 1;
        
        float[] newArray = new float[size * m_numDimensions];
        for (int i = 0; i < newArray.length; i += m_numDimensions) {
            // Set the coordinates.
            newArray[i + xDim] = start + ((i / m_numDimensions) * interval);
            
            for (int j = 0; j < m_numDimensions; j++) {
                float val = 0;
                
                if (j != xDim) {
                    // Find the value for the new point.
                    if (m_bInterpolateSamples) {
                        val = getInterpolatedValue(newArray[i + xDim], xDim, j);
                    } else {
                        float spanMin = newArray[i + xDim] - halfInterval;
                        float spanMax = newArray[i + xDim] + halfInterval;
                        val = findSpanAverage(spanMin, spanMax, -1, xDim, j);
                    }
                    
                    // Interpolate the last dimension's value at this point.
                    newArray[i + j] = val;
                }
            }
        }
        
        // The old list of deleted points was probably invalidated by this
        // operation, so wipe it out.
        m_deleted.clear();
        setArray(newArray);
    }
    
//------------------------------------------------------------------------------
    /**
     * Resample the data to match the supplied series.  When completed, this
     * series should have the same number of points as the other series, and
     * they should have the same X coordinates.
     * @param series	The series whose sampling should be matched.
     */
    public void resample(FloatDataSeries series) {
	
	resample(series, 0);
    }
//------------------------------------------------------------------------------
    /**
     * Resample the data to match the supplied series.  When completed, this
     * series should have the same number of points as the other series, and
     * they should have the same coordinates.
     * @param series	The series whose sampling should be matched.
     * @param xDim	The dimension to use as the X dimension.
     */
    public void resample(FloatDataSeries series, int xDim) {
	
	// Create an array with the same X values as series.  The Y values
	// will be interpolated.
	int newNumPoints = series.getNumPoints();
	float[] newArray = new float[newNumPoints * m_numDimensions];
	float[] otherArray = series.getArray();
	int otherNumDims = series.getNumDimensions();
	
	// Calculate an interpolated value for each new X point.
	for (int j = 0; j < m_numDimensions; j++) {
	    for (int i = 0; i < newNumPoints; i++) {
		int otherIndex = i * otherNumDims;
		float otherX = otherArray[otherIndex];
		if (j == xDim) {
		    // Copy the coordinates from the other series.
		    newArray[i * otherNumDims] = otherX;
		    continue;
		}
		
		// Interpolate the last dimension's value at this point.
		int prevI = i - 1;
		if (prevI < 0)
		    prevI = i;
		int nextI = i + 1;
		if (nextI >= newNumPoints)
		    nextI = i;
		
		// Find the Y value for the new point.
		float val;
		int otherPrevIndex = prevI * otherNumDims;
		float otherPrevX = otherArray[otherPrevIndex];
		int otherNextIndex = nextI * otherNumDims;
		float otherNextX = otherArray[otherNextIndex];
		if (m_bInterpolateSamples) {
		    val = getInterpolatedValue(otherX, xDim, j);
		} else {
		    float spanMin = (otherX + otherPrevX) / 2f;
		    float spanMax = (otherX + otherNextX) / 2f;
		    val = findSpanAverage(spanMin, spanMax, -1, xDim, j);
		}
		newArray[otherIndex + j] = val;
	    }
	}
	
	// Make the new array current.
	setArray(newArray);
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the average across a given span.  Look at the ParabolicInterpolation
     * member variable and the number of points across the span to determine
     * whether to use trapezoidal rule, or equally spaced samples.
     * Uses X dimension = 0 and Y dimension = last dimension.
     * @param spanMin	The lower boundary of the area that should be averaged.
     * @param spanMax	The upper boundary of the area that should be averaged.
     * @param numDivs	The number of samples to make across the span.
     *					If negative, use twice as many samples as there are points.
     * @return The average y value across the span.  See getInterpolatedValue().
     */
    public float findSpanAverage(float spanMin, float spanMax, int numDivs) {
        
        return
            findSpanAverage(spanMin, spanMax, numDivs, 0, m_numDimensions - 1);
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the average across a given span.  Look at the ParabolicInterpolation
     * member variable and the number of points across the span to determine
     * whether to use trapezoidal rule, or equally spaced samples.
     * @param spanMin	The lower boundary of the area that should be averaged.
     * @param spanMax	The upper boundary of the area that should be averaged.
     * @param numDivs	The number of samples to make across the span.
     *					If negative, use twice as many samples as there are points.
     * @param xDim  The dimension to use as the X dimension.
     * @param yDim  The dimension to use as the Y dimension.
     * @return The average y value across the span.  See getInterpolatedValue().
     */
    public float findSpanAverage(float spanMin, float spanMax, int numDivs,
        int xDim, int yDim) {
        
        // Find out how many points are in the span.
        int numPoints = findSpanCount(spanMin, spanMax, xDim);
        
        // Check for special cases.
        float span = Math.abs(spanMax - spanMin);
        if (span <= 0)
            return getInterpolatedValue(spanMin, xDim, yDim);
        
        float avg = 0;
        if (m_bParabolicInterpolation || numPoints <= 1) {
            // Use interpolation at fixed intervals to determine the average.
            // Use approximately twice as many divisons as there are points
            // across the span.
            if (numDivs < 0) numDivs = numPoints * 2;
            if (numDivs <= 0) numDivs = 1;
            
            // Calculate the step size.
            float stepSize = span / (float)numDivs;
            
            // Start at the min.
            float total = 0;
            for (int j = 0; j <= numDivs; j++) {
                // Sum the values, giving the end points half weighting.
                float value = getInterpolatedValue(spanMin, xDim, yDim);
                if (j == 0 || j == numDivs) {
                    total += (value / 2);
                } else {
                    total += value;
                }
                spanMin += stepSize;
            }
            // Calculate the average from the sum.
            avg = total / (float)numDivs;
        } else {
            // Use trapezoidal rule to sum the area under the trapezoids
            // formed by the points and the span boundaries.
            float prevX = spanMin;
            float prevY = getInterpolatedValue(spanMin, xDim, yDim);
            float trapTotal = 0;
            float dx, trapSeg;
            
            // Assign the leftmost index.
            int index = binarySearch(spanMin);
            if (m_array[index * m_numDimensions] < spanMin) index++;
            if (index < 0) index = 0;
            if (index > (getCount() - 1)) index = getCount() - 1;
            
            // Sum the area under each trapezoid except the last.
            while (index < getCount() && m_array[index * 2] < spanMax) {
                dx = m_array[index * m_numDimensions] - prevX;
                trapSeg = 0.5f * (prevY + m_array
                    [index * m_numDimensions + yDim]) * dx;
                trapTotal += trapSeg;
                prevX = m_array[index * m_numDimensions];
                prevY = m_array[index * m_numDimensions + yDim];
                
                index++;
            }
            
            // Sum the area under the last trapezoid.
            dx = spanMax - prevX;
            trapSeg = 0.5f * (prevY +
                getInterpolatedValue(spanMax, xDim, yDim)) * dx;
            trapTotal += trapSeg;
            
            // Find the average.
            avg = trapTotal / span;
        }
        
        return avg;
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds the number of points between min and max along the specified
     * X dimension.
     * @param min  The lower bounds of the span along the axis.
     * @param max  The upper bounds of the span along the axis.
     * @param xDim  The dimension to use as the X dimension.
     * @return The number of points along the span.
     */
    public int findSpanCount(float min, float max, int xDim) {
        
        // Find the nearest points preceding the min and max points.
        int p1 = binarySearch(min, xDim);
        int p2 = binarySearch(max, xDim);
        
        // Perform range checking so the nearest indexes.
        if (p1 < m_begin) p1 = m_begin;
        if (p2 >= m_numPoints) p2 = m_numPoints - 1;
        if (m_array[p1 * m_numDimensions + xDim] < min) p1++;
        if (m_array[p2 * m_numDimensions + xDim] > max) p2--;
        
        // Return 0 if the first nearest index was after the end of the array,
        // if the second nearest index was before the beginning of the array,
        // or if the first was greater than or equal to the second.
        if (p1 >= m_numPoints || p2 < 0 || p1 >= p2)
            return 0;
        
        // Return the difference in index numbers plus one.
        int numPoints = p2 - p1 + 1;
        return numPoints;
    }
    
//------------------------------------------------------------------------------
    /**
     * Determines what a value should snap to depending on that dimension's
     * snap increment.
     * @param oldValue  The old, unsnapped value.
     * @param dimension	The dimension (x=0, y=0, etc.) whose snap increment will
     *                  be used to calculate the new value.
     * @return  The value that the old value would snap to.
     */
    public float snap(float oldValue, int dimension) {
        
        // Determine how much oldValue would have to be shifted down to
        // be at a snap value.
        float temp = oldValue - (float)m_dims[dimension].m_snapOffset;
        float modVal = temp % (float)m_dims[dimension].m_snap;
        float shift;
        
        // If shifting down to a snap value is closer than shifting up,
        // shift down, otherwise shift up.
        if (modVal < (m_dims[dimension].m_snap / 2))
            shift = (float)m_dims[dimension].m_snapOffset - modVal;
        else
            shift = -modVal;
        float newValue = oldValue + shift;
        
        return newValue;
    }
    
//------------------------------------------------------------------------------
    /**
     * Creates a copy of the buffer.
     * @return A copy of the buffer.
     */
    public float[] getArrayCopy() {
        
        // Create a new float array the same size as the existing one.
        FloatBuffer newBuffer = newBuffer(m_array.length);
        
        // Copy every point from the old array to the new one.
        newBuffer.put(m_array);
        
        return newBuffer.array();
    }
    
//------------------------------------------------------------------------------
    /**
     * Copy each value from the main array to the selected array.  If the
     * m_selectedIndexes array doesn't match the size of the main array, or
     * if one of the indexes points outside the main array, all points are
     * deselected, and the method exits.
     * <p><b> Pre Conditions </b><br>
     *      m_array and m_selectedIndexes are not null.  All numbers held in the
     *		m_selectedIndexes array are greater than zero and less than the length
     *		of m_array.
     */
    protected void syncSelectedPoints() {
        
        // Return if the selected array doesn't exist.
        if (m_selectedIndexes == null) {
            deselectAll();
            return;
        } else {
            // Return if the selected array doesn't match the main array in
            // size.  If things don't add up, just deselect everything to
            // be safe.
            if (m_selectedIndexes.length < getCount()
            || m_numSelected > getCount()
            || m_array.length != m_selected.length) {
                
                deselectAll();
                return;
            }
        }
        
        int i = 0, k = 0;
        try {
            // Copy each value from the main array to each selected point.
            for (i = 0; i < m_numSelected; i++) {
                // Copy the value for each dimension of this point from
                // the main series to the selected array.
                if (m_numDimensions == 2) {
                    for (k = 0; k < m_numDimensions; k++) {
                        m_selected[i * m_numDimensions + k]
                            = m_array[m_selectedIndexes[i] * m_numDimensions
                            + k];
                    }
                }
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            deselectAll();
            return;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Select all points in the series.  Stores indexes of selected points
     * internally.  Copies values of selected points if this is a two-dimensional
     * series.
     */
    public synchronized void selectAll() {
        
        // If this is a 2D series, create an array to store data values.
        if (m_numDimensions == 2) {
            m_selected = new float[m_array.length];
        } else {
            m_selected = new float[1];
        }
        m_selectedIndexes = new int[getCount()];
        
        // Make the selected indexes advance from zero to the number of points.
        m_numSelected = getCount();
        for (int i = 0; i < m_numSelected; i++) {
            m_selectedIndexes[i] = i;
            
            // If this is a 2D series, copy the value for each dimension of
            // this point from the main series to the selected array.
            // This allows selected points to be easily drawn.  If there are
            // more than two dimensions, the number of selected points could
            // be very large, so storing the values of selected points in
            // a separate array requires too much memory.
            if (m_numDimensions == 2) {
                for (int k = 0; k < m_numDimensions; k++) {
                    m_selected[i * m_numDimensions + k]
                        = m_array[i * m_numDimensions + k];
                }
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Deselect all points in the series.
     */
    public synchronized void deselectAll() {
        
        // Create new blank arrays to replace the existing selected arrays.
        if (m_numDimensions == 2) {
            m_selected = new float[m_array.length];
        } else {
            m_selected = new float[1];
        }
        m_selectedIndexes = new int[m_array.length];
        
        // Set the number of selected points to zero.
        m_numSelected = 0;
    }
    
//------------------------------------------------------------------------------
    /**
     * Updates member variables when the data in the array are modified.
     * Updates the min and max for each dimension of the data.
     */
    public synchronized void arrayChanged(){
        
        arrayChanged(DATA_MODIFIED);
    }
    
//------------------------------------------------------------------------------
    /**
     * Updates variables when the array changes.
     * Updates the min and max for each dimension of the data.
     */
    public synchronized void arrayChanged(int reason){
        
        // Update the min and max descriptors.
        updateMinMax();
        
        // Make sure the selected points don't point out of bounds.
        syncSelectedPoints();
        
        // Notify observers that this series has changed.
        m_bHasChanged = true;
        notifyObservers();
    }
    
//------------------------------------------------------------------------------
    /**
     * Creates a float array from the ArrayList and calls setArray(float[]).
     * @param list    An ArrayList of Float objects.
     */
    public synchronized void setArray(ArrayList list){
        
        float[] newArray = new float[list.size()];
        
        int i = 0;
        ArrayList<Float> fList = (ArrayList<Float>)list;
        for (Float fVal : fList)
            newArray[i++] = fVal;
        
        setArray(newArray);
    }
    
//------------------------------------------------------------------------------
    /**
     * Updates the min and max for each dimension of the data.
     */
    public synchronized void updateMinMax() {
        
        // Initialize the array for keeping track of whether each dimension
        // is sorted.
        float[] lastVal = new float[m_numDimensions];
        for (int i = 0; i < m_numDimensions; i++) {
            lastVal[i] = Float.NEGATIVE_INFINITY;
        }
        
        // Initialize the mins and maxes.
        for (int j = 0; j < m_numDimensions; j++) {
            m_dims[j].m_min = Float.POSITIVE_INFINITY;
            m_dims[j].m_max = Float.NEGATIVE_INFINITY;
        }
        
        // Traverse the array.
        for (int i = 0; i < m_numPoints; i++) {
            for (int j = 0; j < m_numDimensions; j++) {
                int index = i * m_numDimensions + j;
                // Update the min and max for each dimension.
                if (m_array[index] < m_dims[j].m_min)
                    m_dims[j].m_min = m_array[index];
                if (m_array[index] > m_dims[j].m_max)
                    m_dims[j].m_max = m_array[index];
                
                // Check if the dimension is sorted.
                if (m_array[index] < lastVal[j])
                    lastVal[j] = Float.POSITIVE_INFINITY;
                else lastVal[j] = m_array[index];
            }
        }
        
        // Set the values of the sorted array to tell whether each
        // dimension is sorted.
        for (int i = 0; i < m_numDimensions; i++) {
            m_dims[i].m_sorted
                = (lastVal[i] == Float.POSITIVE_INFINITY ? false : true);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Check if all the points are deleted.  If so, restore the first and last.
     */
    public synchronized void restoreIfAllDeleted() {
        
        // Only proceed if there are no points in this series.
        if (m_numPoints != 0) return;
        
        // Make a new series with two values to hold the lowest and highest
        // deleted points.
        float[] newArray = new float[2 * m_numDimensions];
        
        // Determine the indexes of lowest and highest deleted points.
        int lowIndex = 0, highIndex = 0;
        float lowValue = Float.POSITIVE_INFINITY;
        float highValue = Float.NEGATIVE_INFINITY;
        int index = 0;
        
        for (Float fObj : m_deleted) {
            // Skip to the next X value.
            if ((index % m_numDimensions) != 0) {
                index++;
                continue;
            }
            
            float fVal = fObj;
            
            // Mark the index if this is the lowest so far.
            if (fVal < lowValue) {
                lowValue = fVal;
                lowIndex = index;
            }
            
            // Mark the index if this is the highest so far.
            if (fVal > highValue) {
                highValue = fVal;
                highIndex = index;
            }
            
            index++;
        }
        
        // Copy the values from the deleted list to the new array.
        for (int k = 0; k < m_numDimensions; k++) {
            newArray[k] = m_deleted.get(lowIndex + k);
            newArray[m_numDimensions + k] = m_deleted.get(highIndex + k);
        }
        
        // Remove the two points from the deleted list.
        for (int k = 0; k < m_numDimensions; k++)
            m_deleted.remove(lowIndex);
        
        if (highIndex > lowIndex) highIndex -= m_numDimensions;
        for (int k = 0; k < m_numDimensions; k++) {
            if (highIndex < m_deleted.size())
                m_deleted.remove(highIndex);
        }
        
        // Make the new array current.
        setArray(newArray);
    }
    
//------------------------------------------------------------------------------
    /**
     * Search the interleaved array and return the index of the nearest point
     * to key along the first (0) dimension.
     * @param key	The value that is being searched for.
     * @return The index or nearest index that holds a value matching the key.
     */
    public int findNearestIndex(float key) {
        
        // Search for the closest index whose value is lower than key.
        int index = search(key, 0);
        
        // Find the points above and below the key.
        float upperIndexX, lowerIndexX;
        lowerIndexX = m_array[index * m_numDimensions];
        if (index > (m_numPoints - 2))
            upperIndexX = lowerIndexX;
        else upperIndexX =
            m_array[(index + 1) * m_numDimensions];
        
        // Find the difference between the bounding points.
        float upperDiff = Math.abs(upperIndexX - key);
        float lowerDiff = Math.abs(lowerIndexX - key);
        
        // Return the index of the point that was closer along the X dimension.
        if (upperDiff < lowerDiff &&
            index < (m_numPoints - 1)) {
            
            index++;
        }
        
        return index;
    }
    
//------------------------------------------------------------------------------
    /**
     * Search the interleaved array and return the nearest index along the first
     * dimension.
     * @param key	The value that is being searched for.
     * @return The index or nearest index that holds a value matching the key.
     */
    public int search(float key) {
        
        return search(key, 0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Search the interleaved array and return the nearest index along the specified
     * dimension.  Uses the binarySearch() method if the dimension is sorted,
     * otherwise uses the sequentialSearch() method.
     * @param key	The value that is being searched for.
     * @param xDim  The dimension to use as the X dimension.
     * @return The index or nearest index that holds a value matching the key.
     */
    public int search(float key, int xDim) {
        
        // Choose between the two searching methods.
        if (m_dims[xDim].m_sorted) return binarySearch(key, xDim);
        else return sequentialSearch(key, xDim);
    }
    
//------------------------------------------------------------------------------
    /**
     * Search the interleaved array and return the nearest index along the first
     * dimension.  Calls sequentialSearch(key, 0).
     * @param key	The value that is being searched for.
     * @return The index or nearest index that holds a value matching the key.
     */
    protected int sequentialSearch(float key) {
        
        return sequentialSearch(key, 0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Search the interleaved array and return the nearest index along the specified
     * dimension.  Searches sequentially for an exact match.  Returns the last
     * index if none is found.  Works with unsorted arrays.
     * @param key	The value that is being searched for.
     * @param xDim  The dimension to use as the X dimension.
     * @return The index or nearest index that holds a value matching the key.
     */
    protected int sequentialSearch(float key, int xDim) {
        
        for (int i = 0; i <= getCount() - 1; i++) {
            if (key == m_array[i * m_numDimensions + xDim]) return i;
        }
        
        return (getEnd());
    }
    
//------------------------------------------------------------------------------
    /**
     * Allocates a new FloatBuffer of the indicated size.
     * @param numElements   The size of the buffer (number of float values).
     * @return A direct FloatBuffer of the size indicated.
     */
    public static FloatBuffer newBuffer(int numElements) {
        
        FloatBuffer buffer = FloatBuffer.allocate(numElements);
        
        return buffer;
    }
    
//------------------------------------------------------------------------------
    /**
     * Allocates a new FloatBuffer of the indicated size.
     * @param numElements   The size of the buffer (number of float values).
     * @return A direct FloatBuffer of the size indicated.
     */
    public static FloatBuffer newDirectBuffer(int numElements) {
        
        FloatBuffer buffer = ByteBuffer.allocateDirect(numElements * 4)
        .order(ByteOrder.nativeOrder()).asFloatBuffer();
        
        return buffer;
    }
    
//------------------------------------------------------------------------------
    /**
     * Allocates a new FloatBuffer of the indicated size.
     * @param numElements   The size of the buffer (number of float values).
     * @return A direct FloatBuffer of the size indicated.
     */
    public static FloatBuffer newDirectBuffer(float[] array) {
        
        FloatBuffer buffer = newDirectBuffer(array.length);
        buffer.put(array);
        
        return buffer;
    }
    
//------------------------------------------------------------------------------
    /**
     * Checks if the object argument is a FloatDataSeries with the specified
     * number of dimensions.
     * @param obj	The object to be checked
     * @param numDims	The number of dimensions the series should have.
     * @return Whether the object was a FDS with the right number of dimensions.
     */
    public static boolean countDims(Object obj, int numDims) {
	
	if (obj instanceof FloatDataSeries) {
	    FloatDataSeries series = (FloatDataSeries)obj;
	    if (series.getNumDimensions() == numDims)
		return true;
	}
	
	return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The interleaved array of data in this object.
     */
    public float[] getArray(){
        
        return m_array;
    }
    
//------------------------------------------------------------------------------
    /**
     * Set the data for this object to an interleaved array of X-Y data.  Updates
     * the min and max for each dimension of the data.
     * @param newBuffer    The new buffer to use for this series.
     */
    public synchronized void setArray(float[] newArray){
        
        // Set the array and update meta data.
        m_array = newArray;
        
        m_numPoints = m_array.length;
        
        if (m_numDimensions > 1)
            m_numPoints /= m_numDimensions;
        m_end = m_numPoints - 1;
        
        arrayChanged();
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The interleaved array of selected points.
     */
    public float[] getSelectedArray(){
        
        return m_selected;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The array of indexes of selected points.
     */
    public int[] getSelectedIndexes(){
        
        return m_selectedIndexes;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The interleaved list of deleted points.
     */
    public List<Float> getDeletedList(){
        
        return m_deleted;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param list The new interleaved list of deleted points.
     */
    public synchronized void setDeletedList(List list) {
        
        m_deleted = list;
        if (m_numDimensions == 2) {
            restoreIfAllDeleted();
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The number of n-tuples in the interleaved array.
     */
    public int getCount(){
        
        return (m_array.length / m_numDimensions);
    }
    
//------------------------------------------------------------------------------
    /**
     * Act on menu items being selected.
     * @param e The event to act on.
     */
    public void actionPerformed(java.awt.event.ActionEvent e) {
        
        javax.swing.JMenuItem source =
            (javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Select Area...")) {
            java.awt.Frame frame = GrController.getParentFrame(source);
            selectAreaWithDialog(frame);
        }
        if (srcText.equals("Select All")) selectAll();
        if (srcText.equals("Deselect All")) deselectAll();
        if (srcText.equals("Delete Selected Points")) deleteSelectedPoints();
//        if (srcText.equals("Modify Selected Points...")) modify();
//        if (srcText.equals("Resample...")) resample();
    }
    
//------------------------------------------------------------------------------
    /**
     * Shows a dialog box which allows the user to select points in the series.
     * @param frame The frame to which the dialog should be attached.
     */
    public synchronized void selectAreaWithDialog(java.awt.Frame frame) {
        
        // Instantiate and show the dialog box.
        final DataSelectionDialog dialog =
            new DataSelectionDialog(frame, "Select Area", true);
        
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                dialog.setVisible(true);
            }
        });
        
        if (!(dialog.getOkPressed())) return;
        
        // Create arrays for min and max that would cause nothing to
        // be selected.
        float[] min = new float[m_numDimensions];
        float[] max = new float[m_numDimensions];
        for (int i = 0; i < m_numDimensions; i++) {
            min[i] = Float.POSITIVE_INFINITY;
            max[i] = Float.NEGATIVE_INFINITY;
        }
        
        int[] dims = dialog.getDims();
        double[] mins = dialog.getMins();
        double[] maxes = dialog.getMaxes();
        
        // Assign the mins and maxes from the dialog to the selection arrays.
        for (int i = 0; i < dims.length; i++) {
            min[dims[i]] = (float)mins[i];
            max[dims[i]] = (float)maxes[i];
        }
        
        boolean bAdding = dialog.getAdding();
        
        // Select the area specified by the user.
        selectArea(min, max, bAdding);
    }
}

