/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-01
 * Date started:    2003-06-18
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import gov.usgs.gr.*;
import gov.usgs.gr.visual.gui.*;
import java.nio.*;
import java.util.*;
import org.w3c.dom.*;

//==============================================================================
/**
 * An interleaved, multidimensional data object that stores its data as a single
 * one-dimensional int array.  The data in the array can represent any number
 * of dimensions by interleaving the values at each dimension.  For example, a
 * three-dimensional series (X, Y, and Z) with two points would be stored as
 * (X1, Y1, Z1, X2, Y2, Z2).  The DataSeries keeps track of its beginning, end,
 * min, max, and title.
 */
public class IntDataSeries extends gov.usgs.gr.data.DataSeries {

    /**
     * The XML element name that will identify this class and classes that
     * are extended from it.
     */
    static String XML_ELEMENT_NAME = "intdataseries";

    /** An interleaved array for holding all the data. */
    protected int[] m_array = new int[0];
    /** An interleaved array for holding all selected points. */
    protected int[] m_selected;
    /** A place to save the points that have been deleted. */
    protected List<Integer> m_deleted =
        Collections.synchronizedList(new ArrayList());
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/IntDataSeries.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 * @param numDimensions The number of dimensions to be held in the object.
 */
    public IntDataSeries(int numDimensions) {

        // Set up the object according to the number of dimensions.
        m_numDimensions = numDimensions;
        m_dims = new DataDimension[m_numDimensions];
        initBounds();
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a data series with a data type and title.
 * @param numericType   A code describing the type of data stored.
 * @param title         A string to identify the data series.
 * @param numDimensions The number of dimensions to be held in the object.
 */
    public IntDataSeries(int numericType, String title, int numDimensions) {

        // Store the parameters as member variables.
        m_numericType = numericType;
        m_title = title;

        // Set up the object according to the number of dimensions.
        m_numDimensions = numDimensions;
        m_dims = new DataDimension[m_numDimensions];

        initBounds();
   }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a data series from an interleaved array of data.
 * @param array        An interleaved array to use as the data for this object.
 * @param numDimensions The number of dimensions to be held in the object.
 */
    public IntDataSeries(int[] array, int numDimensions) {

        // Set up the object according to the number of dimensions.
        m_numDimensions = numDimensions;
        m_dims = new DataDimension[m_numDimensions];
        initBounds();

        setArray(array);
    }

//------------------------------------------------------------------------------
/**
 * Copies everything from another data series to this one.
 * @param bDeep		True if a deep copy (unshared arrays) is to be made.
 * @return    A copy of this data series.
 */
    public GrObject createCopy(boolean bDeep) {

    	// Create a new series with the same number of dimensions as this one.
        IntDataSeries series = new IntDataSeries(m_numDimensions);

        // Copy all properties from this series to the new one.
        copyInto(series, bDeep);

        // Copy the arrays containing data, selected points, and deleted points.
        series.copyFromDataSeries(this, bDeep);

        return series;
	}

//------------------------------------------------------------------------------
/**
 * Copies everything from another data series to this one.
 * @param series    A data series.
 * @param bDeep		True if a deep copy (unshared arrays) is to be made.
 */
    public void copyFromDataSeries(DataSeries ds, boolean bDeep) {

        IntDataSeries series;
        if (ds instanceof IntDataSeries)
            series = (IntDataSeries)ds;
        else
            return;

        // Get a copy or reference to the main series.
        int[] array;
        if (bDeep) array = series.getArrayCopy();
        else array = series.getArray();

        // Copy the main attributes, in case they hadn't been already.
        m_numericType = series.getType();
        m_title = new String(series.getTitle());
        m_numSelected = series.getNumSelected();

        // Get references to the selected and deleted arrays.
        int[] selectedArray = series.getSelectedArray();
        int[] selectedIndexes = series.getSelectedIndexes();
        List<Integer> deletedList = series.getDeletedList();

        if (bDeep) { // Deep copy (by value)...
        	// Make a copy of the selected array.
        	m_selected = new int[selectedArray.length];
        	for (int i = 0; i < selectedArray.length; i++) {
        		m_selected[i] = selectedArray[i];
        	}

            // Make a copy of the selected indexes array.
        	m_selectedIndexes = new int[selectedIndexes.length];
        	for (int i = 0; i < selectedIndexes.length; i++) {
        		m_selectedIndexes[i] = selectedIndexes[i];
        	}

            // Make a copy of the deleted list.
            m_deleted.clear();
            for (int iVal : deletedList)
                m_deleted.add(iVal);
        }
        else { // Shallow copy (by reference)...
        	// Point this IntDataSeries' arrays at the other's arrays.
        	m_selected = selectedArray;
            m_selectedIndexes = selectedIndexes;
            m_deleted = deletedList;
        }

        // Setup variables linked to the arrays.
        setArray(array);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An XML name of this element.
 */
    public String getXmlElementName() {
        return "intdataseries";
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Returns an XML version of this element and its children using the
 * Document Object Model.
 * @param doc   A document that can be used to create new DOM elements.
 * @param bIncludeChildren  Whether to include this object's children.
 *                          Ignored here.
 * @return  A new DOM element containing this object and its children.
 */
    public org.w3c.dom.Element getDOMRepresentation(org.w3c.dom.Document doc,
        boolean bIncludeChildren) {

        org.w3c.dom.Element el = doc.createElement(getXmlElementName());

        GrHashList props = getProperties();

        // Just write the id and return if this object has been written out
        // to the file before.
        if (m_bSaveIdOnlyToXml) {
            GrProperty p = (GrProperty)props.get("id");
            el.setAttribute("id", p.value.toString());
            p = (GrProperty)props.get("class");
            el.setAttribute("class", p.value.toString());

            return el;
        }

        // Make an attribute in the new element for each property of this
        // object that was set to be saved.
        ListIterator iter = props.listIterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            GrProperty p = (GrProperty)props.get(key);

            if (p.save == true)
                el.setAttribute(key, p.value.toString());
        }

        // Write out the child nodes at the beginning.
        ListIterator iterator = m_children.listIterator(0);
        while (iterator.hasNext()) {
            GrObject child = (GrObject)iterator.next();
            el.appendChild(child.getDOMRepresentation(doc, true));
        }

        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < m_array.length; i++) {
            // Break the output line after each point.
            if (i % m_numDimensions == 0) sb.append(" " + '\n');

            //Add the number to the output buffer.
            sb.append(" " + m_array[i]);
        }
        Text text = doc.createTextNode(sb.toString());

        el.appendChild(text);

        // If the id was saved then make it so the next time this object
        // is written, only the id will be written.
        if (m_bSaveIdToXml)
            m_bSaveIdOnlyToXml = true;

        return el;
    }

//------------------------------------------------------------------------------
/**
 * Read an XML version of this element and its children using the Document
 * Object Model.
 * @param node  The DOM node containing this object's description.
 * @param allObjects A list of of other objects created from the same
 *                  document.  Ignored here.
 * @param unlinkedObjects   Ignored here.
 */
    public void readFromDOM(org.w3c.dom.Node node,
        Map allObjects, List unlinkedObjects) {

        // Apply each attribute in the XML element as a property to this object.
        org.w3c.dom.NamedNodeMap map = node.getAttributes();
        int mapLength = map.getLength();
        for (int i = 0; i < mapLength; i++) {
            org.w3c.dom.Node attr = map.item(i);
            setProperty(attr.getNodeName(), attr.getNodeValue());
        }

        // Add the child elements as children of this object.
        ArrayList<String> tokenList = new ArrayList();
        NodeList nodeList = node.getChildNodes();
        int numChildren = nodeList.getLength();
        for (int i = 0; i < numChildren; i++) {
            // Look for character data inside the element and interpret it as
            // column data describing the main int data array.
            Node childNode = nodeList.item(i);
            if (childNode instanceof org.w3c.dom.Comment) continue;
            if (!(childNode instanceof org.w3c.dom.CharacterData)) continue;
            String s = ((org.w3c.dom.CharacterData)childNode).getData();

            // Parse the line into a String array.
            String inputDelimiters = " \t\n\r\f,";
            StringTokenizer st = new StringTokenizer(s, inputDelimiters);

            while (st.hasMoreTokens()) {
                tokenList.add(st.nextToken());
            }
        }

        // Create a new main array and fill it with the data that was parsed.
        int[] newArray = new int[tokenList.size()];
        int i = 0;
        for (String s : tokenList)
            newArray[i++] = Integer.parseInt(s);

        // Set the main array to be the newly created array.
        setArray(newArray);
    }

//------------------------------------------------------------------------------
    /**
     * Aids the receiveCharsFromSAX() method.  Reads tokens from the buffer and
     * stores them in the m_inputTokens list.  They can be converted to an
     * interleaved array later.
     * @param buf   A character buffer from which to read.
     * @param offset    The offset into the buffer.
     * @param len       The length of the string to read from the buffer.
     */
    protected void tokenizeAndStore(String s0) {
        
        Integer i0 = Integer.parseInt(s0);
        m_inputTokens.add(i0);
    }

//------------------------------------------------------------------------------
/**
 * Check if any selected points are within the specified area.
 * @param min	Array containing the minimum value in the selected area
 *              for each dimension.
 * @param max	Array containing the maximum value in the selected area
 *              for each dimension.
 * @param bAddIfDeselected	True if unselected points should be selected and
 *							then counted as being contained inside the area.
 * @return	True if a selected point was found within the area.
 */
    public boolean containsSelectedPoint(double[] min, double[] max,
    	boolean bAddIfDeselected) {

        // Make sure the min and max arrays don't have more elements
        // than this series has dimensions.
        int minLength = min.length;
        if (minLength > m_numDimensions) minLength = m_numDimensions;
        int maxLength = max.length;
        if (maxLength > m_numDimensions) maxLength = m_numDimensions;

        // Traverse the array and check if each point is inside the area.
        for (int i = m_begin; i <= m_end; i++) {
            boolean bInBounds = true;
            // Check the minimums for each dimension.
            for (int minIndex = 0; minIndex < minLength; minIndex++) {
                if (m_array[i * m_numDimensions + minIndex] < min[minIndex])
                    bInBounds = false;
            }
            // Check the maximums for each dimension.
            for (int maxIndex = 0; maxIndex < maxLength; maxIndex++) {
                if (m_array[i * m_numDimensions + maxIndex] > max[maxIndex])
                    bInBounds = false;
            }

        	if (bInBounds) {
                // Check if the point is already in the selected list.
                for (int j = 0; j < m_numSelected; j++) {
                	if (m_selectedIndexes[j] == i) {
                    	// Selected point was found to be already stored here.
        				return true;
                    }
                }
                // Either select the point and return true or just return false.
                if (bAddIfDeselected) {
                	selectPoint(i);
                	return true;
                }
            }
        }

        // No selected points were found inside the area.
        return false;
 	}

//------------------------------------------------------------------------------
/**
 * Select all the points in the given area.
 * @param min	Array containing the minimum value in the selected area
 *              for each dimension.
 * @param max	Array containing the maximum value in the selected area
 *              for each dimension.
 * @param bAdding	True if selecting points, false if deselecting.
 */
    public synchronized void selectArea(int[] min, int[] max,
        boolean bAdding) {

        // Make sure the min and max arrays don't have more elements
        // than this series has dimensions.
        int minLength = min.length;
        if (minLength > m_numDimensions) minLength = m_numDimensions;
        int maxLength = max.length;
        if (maxLength > m_numDimensions) maxLength = m_numDimensions;

        // Try to reduce the area to be checked with a quick search of any
        // sorted dimensions.
        int lowerBoundsIndex = 0;
        int upperBoundsIndex = m_numPoints;
        for (int j = 0; j < minLength; j++) {
            if (m_dims[j].m_sorted) {
                int minIndex = binarySearch(min[j] - 1, j) - 1;
                if (minIndex > lowerBoundsIndex)
                    lowerBoundsIndex = minIndex;
                int maxIndex = binarySearch(max[j] + 1, j) + 1;
                if (maxIndex < upperBoundsIndex)
                    upperBoundsIndex = maxIndex;
            }
        }

        // Traverse the array and check if each point is inside the area.
        for (int i = lowerBoundsIndex; i < upperBoundsIndex; i++) {
            boolean bInBounds = true;
            // Check the minimums for each dimension.
            for (int j = 0; j < minLength; j++) {
            	int val = m_array[i * m_numDimensions + j];
                if (val < min[j] || val > max[j]) {
                    bInBounds = false;
                    break;
                }
            }

        	if (bInBounds) {
                // Check if the point is already in the selected list.
                int selectedIndex = -1;
                for (int j = 0; j < m_numSelected; j++) {
                	if (m_selectedIndexes[j] == i) {
                    	// Selected point was found to be already stored here.
                    	selectedIndex = j;
                        break;
                    }
                }

                // If adding points to the selection list...
                if (bAdding) {
    				// Add the point if it wasn't found in the search above.
        			if (selectedIndex < 0)
                    	selectPoint(i);
                }
                else {
                	// Deselect the point if it was already selected.
                	if (selectedIndex >= 0)
                    	deselectPoint(selectedIndex);
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Select a single point, but only if it is in bounds and unselected.
 * @param index	The index of the point in the main series.
 */
    public synchronized void selectPointSafely(int index) {

        if (index >= 0 && index < m_selectedIndexes.length) {
            // Check if the point is already in the selected list.
            int selectedIndex = -1;
            for (int j = 0; j < m_numSelected; j++) {
                    if (m_selectedIndexes[j] == index) {
                    // Selected point was found to be already stored here.
                    selectedIndex = j;
                    break;
                }
            }

            // Add the point if it wasn't found in the search above.
            if (selectedIndex < 0) {
                try {
                    selectPoint(index);
                }
                catch (Exception e) {
                    e.printStackTrace();
                }
                m_bHasChanged = true;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Select a single point.
 * @param index	The index of the point in the main series.
 */
    public synchronized void selectPoint(int index) {

        // Add the new index to the list.
        m_selectedIndexes[m_numSelected] = index;

        // Copy the value for each dimension of this point from
        // the main series to the selected array.
        if (m_numDimensions == 2) {
            for (int k = 0; k < m_numDimensions; k++) {
                m_selected[m_numSelected * m_numDimensions + k]
                    = m_array[index * m_numDimensions + k];
            }
        }

        m_numSelected++;
    }

//------------------------------------------------------------------------------
/**
 * Deselect a single point.
 * @param index	The index of the point in the array of selected points.
 */
    public synchronized void deselectPoint(int index) {

        // Move the last selected point to cover up this one.
        m_selectedIndexes[index] = m_selectedIndexes[m_numSelected - 1];

        // Copy the value for each dimension of this point from
        // the end of the array to cover up this point.
        if (m_numDimensions == 2) {
            for (int k = 0; k < m_numDimensions; k++) {
                m_selected[index * m_numDimensions + k]
                    = m_array[m_selectedIndexes[(m_numSelected - 1)]
                    * m_numDimensions + k];
            }
        }

        m_numSelected--;
    }

//------------------------------------------------------------------------------
/**
 * Delete all selected points from the data series.  Copies points that were
 * not selected to a new array and sets that array as the new array for the
 * IntDataSeries.  List of selected point is emptied.  Deleted point values for
 * each dimension are added to the deleted list.
 */
    public synchronized void deleteSelectedPoints() {

        // Make a new array to hold the unselected points.
        int newLen = (m_numPoints - m_numSelected) * m_numDimensions;
        int[] newArray = new int[newLen];

        // Traverse the main array and copy all unselected points to the
        // new array.
        int index = 0;
        for (int i = 0; i < m_numPoints; i++) {
        	boolean bShouldDelete = false;

            // Traverse the selected index array to see if this point is
            // selected.
        	for (int j = 0; j < m_numSelected; j++) {

            	// Check if this index is selected.
            	if (m_selectedIndexes[j] == i) {
                	bShouldDelete = true;
                    // Move the value at the end to the spot that matched and
                    // shrink the array by one place.
                    m_selectedIndexes[j] = m_selectedIndexes[m_numSelected - 1];
                    m_numSelected--;

                    // Quit looking to see if it was selected
                    break;
                }
            }

            // Let this point live if it wasn't selected.
        	if (bShouldDelete) {
                if (m_numDimensions == 2) {
                    for (int k = 0; k < m_numDimensions; k++) {
                        m_deleted.add(m_array[i * m_numDimensions + k]);
                    }
                }
            }
            else {
            	for (int k = 0; k < m_numDimensions; k++) {
            		newArray[index * m_numDimensions + k]
                		= m_array[i * m_numDimensions + k];
            	}
            	index++;
            }
        }

        // Make the new array current.
        setArray(newArray);
 	}

//------------------------------------------------------------------------------
/**
 * Modify all selected points in some way using the value parameter.
 * Changes are applied to the original array to minimize memory usage.
 * For 2D series, updates the selected list to match the new data.
 * @param modificationType	The type of modification this method should
 *			perform.  0=set equal, 1=add, 2=subtract, 3=multiply, 4=divide
 * @param dimension	The dimension number to modify.
 * @param value	The amount to move each point
 */
    public synchronized void modifySelectedPoints(int modificationType,
    	int dimension, float value) {

        // Don't proceed if the dimension number isn't valid.
    	if (dimension < 0 || dimension >= m_numDimensions) return;

        // Traverse the array and shift the last dimension by value.
        // Modify the specified dimension.
        for (int i = 0; i < m_numSelected; i++) {
            int index = m_selectedIndexes[i] * m_numDimensions + dimension;
        	switch (modificationType) {
            	case 0: // Set
                	m_array[index] = Math.round(value);
                	break;
                case 1: // Add
                	m_array[index] = Math.round((float)m_array[index] + value);
                	break;
                case 2: // Subtract
                	m_array[index] = Math.round((float)m_array[index] - value);
                	break;
                case 3: // Multiply
                	m_array[index] = Math.round((float)m_array[index] * value);
                	break;
                case 4: // Divide
                	m_array[index] = Math.round((float)m_array[index] / value);
                	break;
            }

			// Snap to the nearest snap value;
            if (m_dims[dimension].m_snap > 0) {
                m_array[index] = snap(m_array[index], dimension);
            }

            if (m_numDimensions == 2) {
                m_selected[i * m_numDimensions + dimension] = m_array[index];
            }
        }

        // Update the min and max variables that describe the series.
        updateMinMax();

        // Notify observers of the change to this object.
        m_bHasChanged = true;
        notifyObservers();
 	}

//------------------------------------------------------------------------------
/**
 * Delete all points who's value in the specified dimension fall outside the
 * specified min and max.  Creates a new array equal in size to the original
 * to hold points that were within range.  Creates a third array just large
 * enough to hold points that were within range.  Updates selected indexes list
 * for points that were kept.
 * @param dimension The number of the dimension that determines the range.
 * @param min	The new minimum range value for the dimension.
 * @param max	The new maximum range value for the dimension.
 */
    public synchronized void trimDimension(int dimension, int min,
        int max) {

        // Don't proceed if the dimension number is out of range.
        if (dimension < 0 || dimension >= m_numDimensions) return;

        // Make a new array to hold the points that are within range.
        int[] newArray = new int[m_array.length];
        int[] oldSelectedIndexes = m_selectedIndexes;
        m_selectedIndexes = new int[m_array.length];
        int oldNumSelected = m_numSelected;
        m_numSelected = 0;

        // Traverse the main array and copy all points that are within range.
        int index = 0;
        for (int i = m_begin; i <= m_end; i++) {
        	boolean bShouldDelete = false;
        	if (m_array[i * m_numDimensions + dimension] < min
            	|| m_array[i * m_numDimensions + dimension] > max) {

                bShouldDelete = true;
            }

            // Traverse the selected index array to see if the point that was
            // deleted was listed as selected.  Remove it from the selected
            // list if it was.
        	if (bShouldDelete) {
            	for (int j = 0; j < oldNumSelected; j++) {

            		// Check if this index is selected.
            		if (oldSelectedIndexes[j] == i) {
                    	// Move the value at the end to the spot that matched
                        // and shrink the array by one place.

                    	oldSelectedIndexes[j]
                        	= oldSelectedIndexes[oldNumSelected - 1];
                    	oldNumSelected--;
                    	// Quit looking to see if it was selected
                    	break;
                	}
                }
            }
            else {
                // Save the point into the new array.
            	for (int k = 0; k < m_numDimensions; k++) {
            		newArray[index * m_numDimensions + k]
                		= m_array[i * m_numDimensions + k];
            	}

                // If this point was selected, put it in the new selected
                // indexes.
            	for (int j = 0; j < oldNumSelected; j++) {
            		// Check if this index points to an index higher than the
                    // one that was deleted.
            		if (oldSelectedIndexes[j] == i) {
                    	// Decrement the index it points to.
                    	m_selectedIndexes[m_numSelected] = index;
                        m_numSelected++;
                        break;
                	}
                }

                // Increment the number of points that haven't been deleted
                // and copied to the new array.
            	index++;
            }
        }

        // Check if the new array is too small for it's size
        // (elements were trimmed).
        if ((index * m_numDimensions) < newArray.length) {
        	// Copy the array to one that fits it.
        	int[] newArray2 = new int[index * m_numDimensions];
        	for (int i = 0; i < index; i++) {
            	for (int k = 0; k < m_numDimensions; k++) {
            		newArray2[i * m_numDimensions + k]
                		= newArray[i * m_numDimensions + k];
            	}
        	}

            // Check that there are at least two points left.
        	if (index >= (2 * m_numDimensions))
            	setArray(newArray2);
        }
        else {
        	// Make the new array current.
        	if (index >= (2 * m_numDimensions))
        		setArray(newArray);
        }
    }

//------------------------------------------------------------------------------
/**
 * Find the y value of a point along the curve.
 * @param xVal	The x coordinate of the interpolated point.
 * @return		The y coordinate of the interpolated point.
 */
    public int getInterpolatedValue(int xVal) {

    	return getInterpolatedValue(xVal, 0, 1);
 	}

//------------------------------------------------------------------------------
/**
 * Find the y value of a point along the curve.
 * Otherwise, uses the linearlyInterpolate() method.
 * @param xVal	The x coordinate of the interpolated point.
 * @param xDim  The dimension to use as the X dimension.
 * @param yDim  The dimension to use as the Y dimension.
 * @return		The y coordinate of the interpolated point.
 */
    public int getInterpolatedValue(int xVal, int xDim, int yDim) {

        return linearlyInterpolate(xVal, xDim, yDim);
 	}

//------------------------------------------------------------------------------
/**
 * Find the value of the Y dimension at the given coordinates by using
 * linear interpolation.  Calls linearlyInterpolate(xVal, 0, 1).
 * @param xVal	The X (first dimension) coordinate of the interpolated point.
 * @return The interpolated Y (second dimension) value.
 */
    public int linearlyInterpolate(int xVal) {

        return linearlyInterpolate(xVal, 0, 1);
    }

//------------------------------------------------------------------------------
/**
 * Find the value of a dimension at given coordinates by linearly interpolating
 * between points.
 * @param xVal	The X coordinate of the interpolated point.
 * @param xDim  The dimension to use as the X dimension.
 * @param yDim  The dimension to use as the Y dimension.
 * @return The interpolated Y value.
 */
    public int linearlyInterpolate(int xVal, int xDim, int yDim) {

       	// Find the existing points before and after the interpolated point.
		int nearestIndex = binarySearch(xVal, xDim);
        if (nearestIndex > (getCount() - 1)) nearestIndex = getCount() - 1;
        if (nearestIndex < 0) nearestIndex = 0;
        int index1 = 0;
        if (m_array[nearestIndex * m_numDimensions + xDim] > xVal) {
        	index1 = nearestIndex - 1;
        }
        else {
        	index1 = nearestIndex;
        }

        return linearlyInterpolate(index1, xVal, xDim, yDim);
    }

//------------------------------------------------------------------------------
/**
 * Find the value of a dimension at given coordinates by linearly interpolating
 * between the point at index1 and the one following it.
 * @param index1 The index preceding xVal.
 * @param xVal	The X coordinate of the interpolated point.
 * @param xDim  The dimension to use as the X dimension.
 * @param yDim  The dimension to use as the Y dimension.
 * @return The interpolated Y value.
 */
    public int linearlyInterpolate(int index1, int xVal, int xDim,
        int yDim) {

        // Set the following index to be the one after index 1;
        int index2 = index1 + 1;

        // No need to interpolate if it is past the start or end of the series.
        if (index1 < 0) index1 = 0;
        if (index2 > (getCount() - 1)) index2 = getCount() - 1;
        if (index1 == index2) {
        	int interpolatedValue =
            	m_array[index1 * m_numDimensions + yDim];
    		return interpolatedValue;
        }

        // Measure the distance between the points.
        float dist1to2 = Math.abs(m_array[index2 * m_numDimensions + xDim]
        	- m_array[index1 * m_numDimensions + xDim]);
        float dist1 = Math.abs(xVal - m_array[index1 * m_numDimensions + xDim]);
        float dist2 = Math.abs(m_array[index2 * m_numDimensions + xDim] - xVal);
        float val1 = m_array[index1 * m_numDimensions + yDim];
        float val2 = m_array[index2 * m_numDimensions + yDim];

        // Return the exact original value if the points before and after
        // are equal.
        if (val1 == val2) return Math.round(val1);

        // Linearly interpolate between the points.
        int interpolatedValue = Math.round((val1 * (dist2 / dist1to2))
        	+ (val2 * (dist1 / dist1to2)));

    	return interpolatedValue;
 	}

//------------------------------------------------------------------------------
/**
 * Search the interleaved array and return the nearest index along the
 * first dimension.  Calls binarySearch(key, 0).
 * @param key	The value that is being searched for.
 * @return The index or nearest index that holds a value matching the key.
 */
    protected int binarySearch(int key) {

        return binarySearch(key, 0);
    }

//------------------------------------------------------------------------------
/**
 * Search the interleaved array and return the nearest index along the
 * specified dimension whose value is less than key.
 * Modified from binary search example in _C++ Object-Oriented Data Structures_
 * by Sengupta and Korobkin.
 * <p><b> Pre Conditions </b><br>
 *      The "X" dimension of the interleaved array is sorted from lowest
 *		to highest.
 * @param key	The value that is being searched for.
 * @param xDim  The dimension to use as the X dimension.
 * @return The index that holds a value matching the key or, if no match is
 *         found, the nearest index less than key.
 */
    protected int binarySearch(int key, int xDim) {

        if (m_numPoints == 0) return -1;

        int end = getCount() - 1;
    	int first = 0;
        int last = end - 1;
        int middle = 0;

        // Return right away if key is less than the lowest value.
        if (key < m_array[xDim]) return 0;
        // Return right away if key is greater than the largest value.
        if (key >= m_array[end * m_numDimensions + xDim]) return end;

        // Continue to break the array in half until a match is found or it
        // is broken down to a single value.
        while (first <= last) {
        	// Calculate the midpoint of the current range.
        	middle = (first + last) / 2 + (first + last) % 2;

            // Determine the range between this point and the next.
            int upper;
            if (middle > end) upper = Integer.MAX_VALUE;
            else upper = m_array[(middle + 1) * m_numDimensions + xDim];

            if (key >= m_array[middle * m_numDimensions + xDim]
                && key < upper) {

                // Return if an exact match was found.
            	return (middle);
            }
            else {
            	// Reset the range to either the left half or the right half.
            	if (key > m_array[middle * m_numDimensions + xDim])
                	first = middle + 1;
                else last = middle - 1;
            }
        }

        return (middle);
    }

//------------------------------------------------------------------------------
/**
 * Finds the number of points between min and max along the specified
 * X dimension.
 * @param min  The lower bounds of the span along the axis.
 * @param max  The upper bounds of the span along the axis.
 * @param xDim  The dimension to use as the X dimension.
 * @return The number of points along the span.
 */
    public int findSpanCount(int min, int max, int xDim) {

        // Find the nearest points preceding the min and max points.
        int p1 = binarySearch(min, xDim);
        int p2 = binarySearch(max, xDim);

        // Perform range checking so the nearest indexes.
        if (p1 < m_begin) p1 = m_begin;
        if (p2 >= m_numPoints) p2 = m_numPoints - 1;
        if (m_array[p1 * m_numDimensions + xDim] < min) p1++;
        if (m_array[p2 * m_numDimensions + xDim] > max) p2--;

        // Return 0 if the first nearest index was after the end of the array,
        // if the second nearest index was before the beginning of the array,
        // or if the first was greater than or equal to the second.
        if (p1 >= m_numPoints || p2 < 0 || p1 >= p2)
        	return 0;

        // Return the difference in index numbers plus one.
        int numPoints = p2 - p1 + 1;
    	return numPoints;
    }

//------------------------------------------------------------------------------
/**
 * Determines what a value should snap to depending on that dimension's
 * snap increment.
 * @param oldValue  The old, unsnapped value.
 * @param dimension	The dimension (x=0, y=0, etc.) whose snap increment will
 *                  be used to calculate the new value.
 * @return  The value that the old value would snap to.
 */
    public int snap(int oldValue, int dimension) {

    	// Determine how much oldValue would have to be shifted down to
        // be at a snap value.
        double temp = (double)oldValue - m_dims[dimension].m_snapOffset;
        double modVal = temp % m_dims[dimension].m_snap;
        double shift;

        // If shifting down to a snap value is closer than shifting up,
        // shift down, otherwise shift up.
        if (modVal < (m_dims[dimension].m_snap / 2))
        	shift = m_dims[dimension].m_snapOffset - modVal;
        else
        	shift = -modVal;
        int newValue = (int)Math.round((double)oldValue + shift);

        return newValue;
    }

//------------------------------------------------------------------------------
/**
 * Creates a copy of the buffer.
 * @return A copy of the buffer.
 */
    public int[] getArrayCopy() {

        // Create a new float array the same size as the existing one.
        IntBuffer newBuffer = newBuffer(m_array.length);

        // Copy every point from the old array to the new one.
        newBuffer.put(m_array);

        return newBuffer.array();
    }

//------------------------------------------------------------------------------
/**
 * Copy each value from the main array to the selected array.  If the
 * m_selectedIndexes array doesn't match the size of the main array, or
 * if one of the indexes points outside the main array, all points are
 * deselected, and the method exits.
 * <p><b> Pre Conditions </b><br>
 *      m_array and m_selectedIndexes are not null.  All numbers held in the
 *		m_selectedIndexes array are greater than zero and less than the length
 *		of m_array.
 */
    protected void syncSelectedPoints() {

    	// Return if the selected array doesn't exist.
    	if (m_selectedIndexes == null) {
        	deselectAll();
            return;
        }
        else {
        	// Return if the selected array doesn't match the main array in
            // size.  If things don't add up, just deselect everything to
            // be safe.
        	if (m_selectedIndexes.length < getCount()
            	|| m_numSelected > getCount()
                || m_array.length != m_selected.length) {

                deselectAll();
                return;
            }
        }

    	int i = 0, k = 0;
        try {
        	// Copy each value from the main array to each selected point.
        	for (i = 0; i < m_numSelected; i++) {
            	// Copy the value for each dimension of this point from
            	// the main series to the selected array.
                if (m_numDimensions == 2) {
                    for (k = 0; k < m_numDimensions; k++) {
                        m_selected[i * m_numDimensions + k]
                            = m_array[m_selectedIndexes[i] * m_numDimensions
                            + k];
                    }
                }
        	}
        }
        catch (ArrayIndexOutOfBoundsException e) {
        		deselectAll();
            	return;
        }
 	}

//------------------------------------------------------------------------------
/**
 * Select all points in the series.  Stores indexes of selected points
 * internally.  Copies values of selected points if this is a two-dimensional
 * series.
 */
    public synchronized void selectAll() {

        // If this is a 2D series, create an array to store data values.
        if (m_numDimensions == 2) {
            m_selected = new int[m_array.length];
        }
        else {
            m_selected = new int[1];
        }
    	m_selectedIndexes = new int[getCount()];

        // Make the selected indexes advance from zero to the number of points.
        m_numSelected = getCount();
        for (int i = 0; i < m_numSelected; i++) {
        	m_selectedIndexes[i] = i;

            // If this is a 2D series, copy the value for each dimension of
            // this point from the main series to the selected array.
            // This allows selected points to be easily drawn.  If there are
            // more than two dimensions, the number of selected points could
            // be very large, so storing the values of selected points in
            // a separate array requires too much memory.
            if (m_numDimensions == 2) {
                for (int k = 0; k < m_numDimensions; k++) {
                    m_selected[i * m_numDimensions + k]
                        = m_array[i * m_numDimensions + k];
                }
            }
        }
 	}

//------------------------------------------------------------------------------
/**
 * Deselect all points in the series.
 */
    public synchronized void deselectAll() {

        // Create new blank arrays to replace the existing selected arrays.
        if (m_numDimensions == 2) {
    	    m_selected = new int[m_array.length];
        }
        else {
            m_selected = new int[1];
        }
    	m_selectedIndexes = new int[m_array.length];

        // Set the number of selected points to zero.
        m_numSelected = 0;
        
    }
    
//------------------------------------------------------------------------------
    /**
     * Updates member variables when the data in the array are modified.
     * Updates the min and max for each dimension of the data.
     */
    public synchronized void arrayChanged(){
        
        arrayChanged(DATA_MODIFIED);
    }
    
//------------------------------------------------------------------------------
    /**
     * Updates variables when the array changes.
     * Updates the min and max for each dimension of the data.
     */
    public synchronized void arrayChanged(int reason){

        // Update the min and max descriptors.
        updateMinMax();

        // Make sure the selected points don't point out of bounds.
        syncSelectedPoints();

        // Notify observers that this series has changed.
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Creates an int array from the ArrayList and calls setArray(int[]).
 * @param list    An ArrayList of Integer objects.
 */
    public synchronized void setArray(ArrayList list){

        int[] newArray = new int[list.size()];
        int i = 0;
        ArrayList<Integer> iList = (ArrayList<Integer>)list;
        for (Integer iVal : iList)
        	newArray[i++] = iVal;

        setArray(newArray);
    }

//------------------------------------------------------------------------------
/**
 * Updates the min and max for each dimension of the data.
 */
    public synchronized void updateMinMax() {

    	// Initialize the array for keeping track of whether each dimension
        // is sorted.
        int[] lastVal = new int[m_numDimensions];
        for (int i = 0; i < m_numDimensions; i++) {
        	lastVal[i] = Integer.MIN_VALUE;
        }

        // Initialize the mins and maxes.
        for (int j = 0; j < m_numDimensions; j++) {
            m_dims[j].m_min = Integer.MAX_VALUE;
            m_dims[j].m_max = Integer.MIN_VALUE;
        }

        // Traverse the array.
        for (int i = 0; i < m_numPoints; i++) {
            for (int j = 0; j < m_numDimensions; j++) {
                int index = i * m_numDimensions + j;
                // Update the min and max for each dimension.
                if (m_array[index] < m_dims[j].m_min)
                	m_dims[j].m_min = m_array[index];
                if (m_array[index] > m_dims[j].m_max)
                	m_dims[j].m_max = m_array[index];

                // Check if the dimension is sorted.
                if (m_array[index] < lastVal[j])
                	lastVal[j] = Integer.MAX_VALUE;
                else lastVal[j] = m_array[index];
            }
        }

        // Set the values of the sorted array to tell whether each
        // dimension is sorted.
        for (int i = 0; i < m_numDimensions; i++) {
        	m_dims[i].m_sorted
            	= (lastVal[i] == Integer.MAX_VALUE ? false : true);
        }
	}

//------------------------------------------------------------------------------
/**
 * Search the interleaved array and return the nearest index along the first
 * dimension.
 * @param key	The value that is being searched for.
 * @return The index or nearest index that holds a value matching the key.
 */
    public int search(int key) {

    	return search(key, 0);
    }

//------------------------------------------------------------------------------
/**
 * Search the interleaved array and return the nearest index along the specified
 * dimension.  Uses the binarySearch() method if the dimension is sorted,
 * otherwise uses the sequentialSearch() method.
 * @param key	The value that is being searched for.
 * @param xDim  The dimension to use as the X dimension.
 * @return The index or nearest index that holds a value matching the key.
 */
    public int search(int key, int xDim) {

    	// Choose between the two searching methods.
        if (m_dims[xDim].m_sorted) return binarySearch(key, xDim);
        else return sequentialSearch(key, xDim);
    }

//------------------------------------------------------------------------------
/**
 * Search the interleaved array and return the nearest index along the first
 * dimension.  Calls sequentialSearch(key, 0).
 * @param key	The value that is being searched for.
 * @return The index or nearest index that holds a value matching the key.
 */
    protected int sequentialSearch(int key) {

        return sequentialSearch(key, 0);
    }

//------------------------------------------------------------------------------
/**
 * Search the interleaved array and return the nearest index along the specified
 * dimension.  Searches sequentially for an exact match.  Returns the last
 * index if none is found.  Works with unsorted arrays.
 * @param key	The value that is being searched for.
 * @param xDim  The dimension to use as the X dimension.
 * @return The index or nearest index that holds a value matching the key.
 */
    protected int sequentialSearch(int key, int xDim) {

    	for (int i = 0; i <= getCount() -1; i++) {
        	if (key == m_array[i * m_numDimensions + xDim]) return i;
        }

    	return (getEnd());
    }

//------------------------------------------------------------------------------
/**
 * Allocates a new IntBuffer of the indicated size.
 * @param numElements   The size of the buffer (number of Int values).
 * @return A direct IntBuffer of the size indicated.
 */
    public static IntBuffer newBuffer(int numElements) {

        IntBuffer buffer = IntBuffer.allocate(numElements);

        return buffer;
    }

//------------------------------------------------------------------------------
/**
 * Allocates a new IntBuffer of the indicated size.
 * @param numElements   The size of the buffer (number of Int values).
 * @return A direct IntBuffer of the size indicated.
 */
    public static IntBuffer newDirectBuffer(int numElements) {

        IntBuffer buffer = ByteBuffer.allocateDirect(numElements * 4)
            .order(ByteOrder.nativeOrder()).asIntBuffer();

        return buffer;
    }

//------------------------------------------------------------------------------
/**
 * Allocates a new IntBuffer of the indicated size.
 * @param numElements   The size of the buffer (number of int values).
 * @return A direct IntBuffer of the size indicated.
 */
    public static IntBuffer newDirectBuffer(int[] array) {

        IntBuffer buffer = newDirectBuffer(array.length);
        buffer.put(array);

        return buffer;
    }
    
//------------------------------------------------------------------------------
    /**
     * Checks if the object argument is a IntDataSeries with the specified
     * number of dimensions.
     * @param obj	The object to be checked
     * @param numDims	The number of dimensions the series should have.
     * @return Whether the object was a IDS with the right number of dimensions.
     */
    public static boolean countDims(Object obj, int numDims) {
	
	if (obj instanceof IntDataSeries) {
	    IntDataSeries series = (IntDataSeries)obj;
	    if (series.getNumDimensions() == numDims)
		return true;
	}
	
	return false;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The interleaved array of data in this object.
 */
    public int[] getArray(){

        return m_array;
    }

//------------------------------------------------------------------------------
/**
 * Set the data for this object to an interleaved array of X-Y data.  Updates
 * the min and max for each dimension of the data.
 * @param newBuffer    The new buffer to use for this series.
 */
    public synchronized void setArray(int[] newArray){

        // Set the array and update meta data.
        m_array = newArray;

        m_numPoints = m_array.length;

        if (m_numDimensions > 1)
            m_numPoints /= m_numDimensions;
        m_end = m_numPoints - 1;

        arrayChanged();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The interleaved array of selected points.
 */
    public int[] getSelectedArray(){

        return m_selected;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The array of indexes of selected points.
 */
    public int[] getSelectedIndexes(){

        return m_selectedIndexes;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The interleaved list of deleted points.
 */
    public List<Integer> getDeletedList(){

        return m_deleted;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param list The new interleaved list of deleted points.
 */
    public synchronized void setDeletedList(List list) {

        m_deleted = list;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The number of n-tuples in the interleaved array.
 */
    public int getCount(){

        return (m_array.length / m_numDimensions);
    }

//------------------------------------------------------------------------------
/**
 * Act on menu items being selected.
 * @param e The event to act on.
 */
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Select Area...")) {
            java.awt.Frame frame = GrController.getParentFrame(source);
        	selectAreaWithDialog(frame);
        }
        if (srcText.equals("Select All")) selectAll();
        if (srcText.equals("Deselect All")) deselectAll();
        if (srcText.equals("Delete Selected Points")) deleteSelectedPoints();
    }

//------------------------------------------------------------------------------
/**
 * Shows a dialog box which allows the user to select points in the series.
 * @param frame The frame to which the dialog should be attached.
 */
    public synchronized void selectAreaWithDialog (java.awt.Frame frame) {

        // Instantiate and show the dialog box.
        final DataSelectionDialog dialog =
            new DataSelectionDialog(frame, "Select Area", true);

            javax.swing.SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    dialog.setVisible(true);
                }
            });

        if (!(dialog.getOkPressed())) return;

        // Create arrays for min and max that would cause nothing to
        // be selected.
        int[] min = new int[m_numDimensions];
        int[] max = new int[m_numDimensions];
        for (int i = 0; i < m_numDimensions; i++) {
            min[i] = Integer.MAX_VALUE;
            max[i] = Integer.MIN_VALUE;
        }

        int[] dims = dialog.getDims();
        double[] mins = dialog.getMins();
        double[] maxes = dialog.getMaxes();

        // Assign the mins and maxes from the dialog to the selection arrays.
        for (int i = 0; i < dims.length; i++) {
            min[dims[i]] = (int)Math.round(mins[i]);
            max[dims[i]] = (int)Math.round(maxes[i]);
        }

        boolean bAdding = dialog.getAdding();

        // Select the area specified by the user.
        selectArea(min, max, bAdding);
    }
}

