/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2003-02-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import gov.usgs.gr.*;
import gov.usgs.gr.tools.*;
import java.io.*;
import java.util.*;
import org.w3c.dom.Text;

//==============================================================================
/**
 *
 */
public class RafDataSeries extends gov.usgs.gr.data.DataSeries
    implements java.awt.event.ActionListener, gov.usgs.gr.tools.ThreadWorker {

    /** The name of the data set's file, if there are no points in the array. */
    protected File m_file = new java.io.File("/");
    // The file object for reading from the disk file.
    protected RandomAccessFile m_raf = null;

    protected float[][] m_cacheSegmentsF = new float[0][];
    protected double[][] m_cacheSegmentsD = new double[0][];
    protected long[] m_cacheOffsets = new long[0];
    protected int m_numCacheSegments = 0;
    protected int m_leadingCacheSegment = 0;
    protected boolean m_bManageCache = false;

    protected byte[][] m_byteCacheSegments = new byte[0][];
    protected long[] m_byteCacheOffsets = new long[0];
    protected int m_numByteCacheSegments = 0;

    protected Object m_readLock = new Object();
    protected byte[] m_readByteArray = new byte[0];
    protected int m_readIndex = 0;
    protected int m_readSize = 0;
    protected int m_readSizePerThread = 0;
    protected int m_readBlockOffset = 0;
    protected float[] m_readDestF = null;
    protected double[] m_readDestD = null;

    protected Object m_writeLock = new Object();
    protected byte[] m_writeByteArray = null;
    protected int m_writeIndex = 0;
    protected int m_writeSize = 0;
    protected int m_writeSizePerThread = 0;
    protected int m_writeBlockOffset = 0;
    protected float[] m_writeSrcF = null;
    protected double[] m_writeSrcD = null;
    protected boolean[] m_bThreadStates = new boolean[0];
    protected int m_numThreads = 1;

    protected String m_errorString = null;

    public final static int SHORT = 0;
    public final static int INT = 1;
    public final static int LONG = 2;
    public final static int FLOAT = 3;
    public final static int DOUBLE = 4;

    protected int m_memType = FLOAT;
    protected int m_memTypeLen = 4;
    protected int m_diskType = SHORT;
    protected int m_diskTypeLen = 2;
    protected boolean m_bBigEndian = false;

    protected long m_readHits = 0;
    protected long m_readTotal = 0;
    protected long m_readMin = Long.MAX_VALUE;
    protected long m_readMax = Long.MIN_VALUE;

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 * @param numDimensions The number of dimensions to be held in the object.
 */
    public RafDataSeries(int numDimensions) {

        // Set up the object according to the number of dimensions.
        m_numDimensions = numDimensions;
        m_dims = new DataDimension[m_numDimensions];
        initBounds();
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a data series with a data type and title.
 * @param numericType   A code describing the type of data stored.
 * @param title         A string to identify the data series.
 * @param numDimensions The number of dimensions to be held in the object.
 */
    public RafDataSeries(int numericType, String title, int numDimensions) {

        // Store the parameters as member variables.
        m_numericType = numericType;
        m_title = title;

        // Set up the object according to the number of dimensions.
        m_numDimensions = numDimensions;
        m_dims = new DataDimension[m_numDimensions];

        initBounds();
   }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * file            String  "File"                                         <p>
 * numThreads      Integer "Number of Threads"                            <p>
 * bigEndian       Boolean "Big Endian Disk Format"                       <p>
 * diskType        String  "Disk Type"                                     <p>
 * memType         String  "Mem Type"                                      <p>
 * manageCache     Boolean "Manage Cache Automatically"                       <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

        // First add properties from the super class.
        GrHashList props = super.getProperties();

        props.put("file", new GrProperty(m_file.getAbsolutePath(), "File"));
        props.put("numThreads", new GrProperty(m_numThreads,
            "Number of Threads"));
        props.put("bigEndian", new GrProperty(m_bBigEndian,
            "Big Endian Disk Format", false, true, true));

        String[] types = {"SHORT", "INTEGER", "LONG", "FLOAT", "DOUBLE"};

        props.put("diskType", new GrProperty(types[m_diskType],
            "Disk Type", false, true, true));
        props.put("memType", new GrProperty(types[m_memType],
            "Memory Type", false, false, true));
        props.put("manageCache", new GrProperty(m_bManageCache,
            "Manage Cache Automatically", false, true, true));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.gr.data.DataSeries#getProperties getProperties() for a list
 * of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

        GrProperty p;
        m_errorString = null;

        // For each property with a string value, pass it to the setProperty()
        // method to be handled there.
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                // Handle string properties by passing them to the
                // setProperty() method.
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                if (key.equals("numThreads")) {
                    int numThreads = (Integer)p.value;
                    if (numThreads > 0 && numThreads < 1000)
                        m_numThreads = numThreads;
                }
                if (key.equals("bigEndian")) {
                    m_bBigEndian = (Boolean)p.value;
                }
                if (key.equals("manageCache")) {
                    m_bManageCache = (Boolean)p.value;
                }
            }
            // Handle string properties by passing them to the
            // setProperty() method.
            if (obj instanceof String) {
                setProperty(key, (String)obj);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.gr.data.DataSeries#getProperties getProperties() for a list
 * of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

        m_errorString = null;

        if (key.equals("file")) {
            m_file = new java.io.File(value);
            m_raf = null;
        }
        if (key.equals("numThreads")) {
            int numThreads = Integer.parseInt(value);
            if (numThreads > 0 && numThreads < 1000)
                m_numThreads = numThreads;
        }
        if (key.equals("bigEndian")) {
            m_bBigEndian = Boolean.parseBoolean(value);
        }
        if (key.equals("diskType")) {
            String[] types = {"SHORT", "INTEGER", "LONG", "FLOAT", "DOUBLE"};
            int[] lengths = {2, 4, 8, 4, 8};
            // Check each name.
            for (int i = 0; i < types.length; i++) {
                String name = value.trim().toUpperCase();
                if (name.equals(types[i])) {
                    m_diskType = i;
                    m_diskTypeLen = lengths[i];
                    break;
                }
            }
        }
        if (key.equals("memType")) {
            String[] types = {"SHORT", "INTEGER", "LONG", "FLOAT", "DOUBLE"};
            // Check each name.
            for (int i = 0; i < types.length; i++) {
                String name = value.trim().toUpperCase();
                if (name.equals(types[i])) {
                    m_memType = i;
                    break;
                }
            }
        }
        if (key.equals("manageCache")) {
            m_bManageCache = Boolean.parseBoolean(value);
        }
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An XML name of this element.
 */
    public String getXmlElementName() {
        return "dataseries";
    }

//------------------------------------------------------------------------------
/**
 * Check if any selected points are within the specified area.
 * @param min	Array containing the minimum value in the selected area
 *              for each dimension.
 * @param max	Array containing the maximum value in the selected area
 *              for each dimension.
 * @param bAddIfDeselected	True if unselected points should be selected and
 *							then counted as being contained inside the area.
 * @return	True if a selected point was found within the area.
 */
    public boolean containsSelectedPoint(double[] min, double[] max,
    	boolean bAddIfDeselected) {

        // No selected points were found inside the area.
        return false;
 	}

//------------------------------------------------------------------------------
/**
 * Copies everything from another data series to this one.
 * @param series    A data series.
 * @param bDeep		True if a deep copy (unshared arrays) is to be made.
 */
    public void copyFromDataSeries(DataSeries ds, boolean bDeep) {

        RafDataSeries series;
        if (ds instanceof RafDataSeries)
            series = (RafDataSeries)ds;
        else
            return;

        // Copy the main attributes, in case they hadn't been already.
        setProperties(series.getProperties());
    }

//------------------------------------------------------------------------------
/**
 * Creates another DataSeries with the same data and properties as this one.
 * @param bDeep		True if a deep copy (unshared arrays) is to be made.
 * @return    A copy of this data series.
 */
    public GrObject createCopy(boolean bDeep) {

    	// Create a new series with the same number of dimensions as this one.
        RafDataSeries series = new RafDataSeries(m_numDimensions);

        // Copy all properties from this series to the new one.
        copyInto(series, bDeep);

        // Copy the arrays containing data, selected points, and deleted points.
        series.copyFromDataSeries(this, bDeep);

        return series;
	}

//------------------------------------------------------------------------------
/**
 * Returns an XML version of this element and its children using the
 * Document Object Model.
 * @param doc   A document that can be used to create new DOM elements.
 * @param bIncludeChildren  Whether to include this object's children.
 *                          Ignored here.
 * @return  A new DOM element containing this object and its children.
 */
    public org.w3c.dom.Element getDOMRepresentation(org.w3c.dom.Document doc,
        boolean bIncludeChildren) {

        org.w3c.dom.Element el = doc.createElement(getXmlElementName());

        GrHashList props = getProperties();

        // Just write the id and return if this object has been written out
        // to the file before.
        if (m_bSaveIdOnlyToXml) {
            GrProperty p = (GrProperty)props.get("id");
            el.setAttribute("id", p.value.toString());
            p = (GrProperty)props.get("class");
            el.setAttribute("class", p.value.toString());

            return el;
        }

        // Make an attribute in the new element for each property of this
        // object that was set to be saved.
        ListIterator iter = props.listIterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            GrProperty p = (GrProperty)props.get(key);

            if (p.save == true)
                el.setAttribute(key, p.value.toString());
        }

        // Write out the child nodes at the beginning.
        ListIterator iterator = m_children.listIterator(0);
        while (iterator.hasNext()) {
            GrObject child = (GrObject)iterator.next();
            el.appendChild(child.getDOMRepresentation(doc, true));
        }

        // If the id was saved then make it so the next time this object
        // is written, only the id will be written.
        if (m_bSaveIdToXml)
            m_bSaveIdOnlyToXml = true;

        return el;
    }

//------------------------------------------------------------------------------
/**
 * Create an array of cache arrays that can hold data from various segments
 * of the file.
 */
    public boolean createCacheSegments(int numSegments, int size) {

        if (size < 0 || numSegments < 0)
            return false;

        m_numCacheSegments = numSegments;

        // Create the arrays that hold the segments and their offsets.
        if (m_memType == FLOAT)
            m_cacheSegmentsF = new float[numSegments][];
        if (m_memType == DOUBLE)
            m_cacheSegmentsD = new double[numSegments][];
        m_cacheOffsets = new long[numSegments];

        // Allocate the array for each segment.
        for (int i = 0; i < m_numCacheSegments; i++) {
            try {
                if (m_memType == FLOAT)
                    m_cacheSegmentsF[i] = new float[size];
                if (m_memType == DOUBLE)
                    m_cacheSegmentsD[i] = new double[size];
                m_cacheOffsets[i] = 0;
            }
            catch (OutOfMemoryError e) {
                m_cacheOffsets[i] = 0;
                m_cacheSegmentsF[i] = new float[0];
                m_cacheSegmentsD[i] = new double[0];

                return false;
            }
        }

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Read the values starting at offset from the file into the specified
 * cache segment.
 */
    public boolean fillCacheSegment(int segment, long offset) {

        // Check that the segment index is within range.
        if (segment < 0 || segment >= m_numCacheSegments)
            return false;

        // Read the values starting at offset from the file into the
        // specified cache segment.
        if (m_memType == FLOAT) {
            float[] cacheF = m_cacheSegmentsF[segment];
            getValuesInto(cacheF, 0, offset, cacheF.length);
        }
        if (m_memType == DOUBLE) {
            double[] cacheD = m_cacheSegmentsD[segment];
            getValuesInto(cacheD, 0, offset, cacheD.length);
        }

        // Store the offset of this segment.
        m_cacheOffsets[segment] = offset;

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Get a copy of the cache offsets array.
 */
    public long[] getCacheOffsets() {

        return (long[])m_cacheOffsets.clone();
    }

//------------------------------------------------------------------------------
/**
 * Get an array showing the sized of the cache segments.
 */
    public int[] getCacheSizes() {

        int[] cacheSizes = null;

        if (m_memType == FLOAT) {
            cacheSizes = new int[m_cacheSegmentsF.length];

            for (int i = 0; i < m_cacheSegmentsF.length; i++) {
                cacheSizes[i] = m_cacheSegmentsF[i].length;
            }
        }

        if (m_memType == DOUBLE) {
            cacheSizes = new int[m_cacheSegmentsD.length];

            for (int i = 0; i < m_cacheSegmentsD.length; i++) {
                cacheSizes[i] = m_cacheSegmentsD[i].length;
            }
        }

        return cacheSizes;
    }

//------------------------------------------------------------------------------
/**
 * Create an array of cache byte arrays that can hold data from various segments
 * of the file.
 */
    public boolean createByteCacheSegments(int numSegments, int size) {

        if (size < 0 || numSegments < 0)
            return false;

        m_numByteCacheSegments = numSegments;

        // Create the arrays that hold the segments and their offsets.
        m_byteCacheSegments = new byte[numSegments][];
        m_byteCacheOffsets = new long[numSegments];

        // Allocate the array for each segment.
        for (int i = 0; i < m_numByteCacheSegments; i++) {
            try {
                m_byteCacheSegments[i] = new byte[size * m_diskTypeLen];
                m_byteCacheOffsets[i] = 0;
            }
            catch (OutOfMemoryError e) {
                m_byteCacheOffsets[i] = 0;
                m_byteCacheSegments[i] = new byte[0];

                return false;
            }
        }

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Read the values starting at offset from the file into the specified
 * cache segment.
 */
    public boolean fillByteCacheSegment(int segment, long offset) {

        // Check that the segment index is within range.
        if (segment < 0 || segment >= m_numByteCacheSegments)
            return false;

        // Read the values starting at offset from the file into the
        // specified cache segment.
        byte[] cache = m_byteCacheSegments[segment];
        fillByteCache(cache, offset, cache.length);

        // Store the offset of this segment.
        m_byteCacheOffsets[segment] = offset * m_diskTypeLen;

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Create an array of cache arrays that can hold data from various segments
 * of the file, fill them with the beginning of the file, and set the
 * managed cache flag to true.
 */
    public boolean createManagedCache(int numSegments, int size) {

        boolean bSuccess = createCacheSegments(numSegments, size);
        if (!bSuccess)
            return false;

        long offset = 0;
        for (int i = 0; i < numSegments; i++) {
            fillCacheSegment(i, offset);
            offset += size;
        }

        m_bManageCache = true;

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Read from a random access file.
 */
    public float getValueAt(long offset) {

        return getFloatValueAt(offset);
    }

//------------------------------------------------------------------------------
/**
 * Read from a random access file.
 */
    public float getFloatValueAt(long offset) {

        // Check if the file exists.
        if (m_raf == null) {
            boolean success = openRAF("r");
            if (!success)
                return 0;
        }

        // Keep track of stats for min and max offset read.
        if (m_readMin > offset)
            m_readMin = offset;
        if (m_readMax < offset)
            m_readMax = offset;
        m_readTotal++;

        for (int i = 0; i < m_numCacheSegments; i++) {
            float[] cache = m_cacheSegmentsF[i];
            long cacheOffset = m_cacheOffsets[i];
            if (offset >= cacheOffset &&
                offset < (cacheOffset + cache.length)) {
                m_readHits++;
                return cache[(int)(offset - cacheOffset)];
            }
        }

        if (m_bManageCache && (m_numCacheSegments > 0)) {
            long cacheSize = m_cacheSegmentsF[0].length;
            long cacheOffset = (offset / cacheSize) * cacheSize;
            // Leapfrog the read cache as soon as the first one is finished.
            fillCacheSegment(m_leadingCacheSegment, cacheOffset);
            float val = m_cacheSegmentsF[m_leadingCacheSegment]
                [(int)(offset - cacheOffset)];

            m_readHits++;

            m_leadingCacheSegment++;
            if (m_leadingCacheSegment >= m_numCacheSegments)
                m_leadingCacheSegment = 0;

            return val;
        }

        try {
            offset *= m_diskTypeLen;

            float val;

            switch (m_diskType) {
                case SHORT:
                default:
                    if (m_bBigEndian) {
                        synchronized (m_readLock) {
                            m_raf.seek(offset);
                            return ( (float) m_raf.readShort());
                        }
                    }
                    else {
                        int lowByteAsInt = Integer.MAX_VALUE;
                        int highByteAsInt = Integer.MAX_VALUE;
                        for (int i = 0; i < m_numByteCacheSegments; i++) {
                            byte[] byteCache = m_byteCacheSegments[i];
                            long byteCacheOffset = m_byteCacheOffsets[i];
                            if (offset >= byteCacheOffset &&
                                offset < (byteCacheOffset + byteCache.length)) {

                                int byteOffset = (int)(offset -byteCacheOffset);
                                lowByteAsInt = ((int)byteCache[
                                    byteOffset]) & 255;
                                highByteAsInt = ((int)byteCache[
                                    byteOffset + 1]);

                                m_readHits++;
                                break;
                            }
                        }
                        if (lowByteAsInt == Integer.MAX_VALUE) {
                            // Read the two bytes from the file.
                            synchronized (m_readLock) {
                                m_raf.seek(offset);
                                lowByteAsInt = ( (int) m_raf.readByte()) & 255;
                                highByteAsInt = ( (int) m_raf.readByte());
                            }
                        }

                        // Reverse their order.
                        val = (float) ( (highByteAsInt << 8) | lowByteAsInt);

                        return val;
                    }
                case FLOAT:
                    synchronized (m_readLock) {
                        m_raf.seek(offset);
                        if (m_bBigEndian) {
                            return (m_raf.readFloat());
                        }
                        else {
                            // Little endian version not implemented.
                            return 0;
                        }
                    }
                case DOUBLE:
                    synchronized (m_readLock) {
                        m_raf.seek(offset);
                        if (m_bBigEndian) {
                            return ((float)m_raf.readDouble());
                        }
                        else {
                            // Little endian version not implemented.
                            return 0;
                        }
                    }
                case INT:
                    synchronized (m_readLock) {
                        m_raf.seek(offset);
                        if (m_bBigEndian) {
                            return ((float)m_raf.readInt());
                        }
                        else {
                            // Little endian version not implemented.
                            return 0;
                        }
                    }
                case LONG:
                    synchronized (m_readLock) {
                        m_raf.seek(offset);
                        if (m_bBigEndian) {
                            return ((float)m_raf.readLong());
                        }
                        else {
                            // Little endian version not implemented.
                            return 0;
                        }
                    }
            }
        }
        catch (IOException e) {
           e.printStackTrace();
        }

        return 0;
    }

//------------------------------------------------------------------------------
/**
 * Read from a random access file.
 */
    public double getDoubleValueAt(long offset) {

        // Check if the file exists.
        if (m_raf == null) {
            boolean success = openRAF("r");
            if (!success)
                return 0;
        }

        // Keep track of stats for min and max offset read.
        if (m_readMin > offset)
            m_readMin = offset;
        if (m_readMax < offset)
            m_readMax = offset;
        m_readTotal++;

        for (int i = 0; i < m_numCacheSegments; i++) {
            double[] cache = m_cacheSegmentsD[i];
            long cacheOffset = m_cacheOffsets[i];
            if (offset >= cacheOffset &&
                offset < (cacheOffset + cache.length)) {
                m_readHits++;
                return cache[(int)(offset - cacheOffset)];
            }
        }

        if (m_bManageCache && (m_numCacheSegments > 0)) {
            long cacheSize = m_cacheSegmentsD[0].length;
            long cacheOffset = (offset / cacheSize) * cacheSize;
            // Leapfrog the read cache as soon as the first one is finished.
            fillCacheSegment(m_leadingCacheSegment, cacheOffset);
            double val = m_cacheSegmentsD[m_leadingCacheSegment]
                [(int)(offset - cacheOffset)];

            m_readHits++;

            m_leadingCacheSegment++;
            if (m_leadingCacheSegment >= m_numCacheSegments)
                m_leadingCacheSegment = 0;

            return val;
        }

        try {
            offset *= m_diskTypeLen;

            double val;

            switch (m_diskType) {
                case SHORT:
                default:
                    if (m_bBigEndian) {
                        synchronized (m_readLock) {
                            m_raf.seek(offset);
                            return ( (double) m_raf.readShort());
                        }
                    }
                    else {
                        int lowByteAsInt = Integer.MAX_VALUE;
                        int highByteAsInt = Integer.MAX_VALUE;
                        for (int i = 0; i < m_numByteCacheSegments; i++) {
                            byte[] byteCache = m_byteCacheSegments[i];
                            long byteCacheOffset = m_byteCacheOffsets[i];
                            if (offset >= byteCacheOffset &&
                                offset < (byteCacheOffset + byteCache.length)) {

                                int byteOffset = (int)(offset -byteCacheOffset);
                                lowByteAsInt = ((int)byteCache[
                                    byteOffset]) & 255;
                                highByteAsInt = ((int)byteCache[
                                    byteOffset + 1]);

                                m_readHits++;
                                break;
                            }
                        }
                        if (lowByteAsInt == Integer.MAX_VALUE) {
                            // Read the two bytes from the file.
                            synchronized (m_readLock) {
                                m_raf.seek(offset);
                                lowByteAsInt = ( (int) m_raf.readByte()) & 255;
                                highByteAsInt = ( (int) m_raf.readByte());
                            }
                        }

                        // Reverse their order.
                        val = ( (highByteAsInt << 8) | lowByteAsInt);

                        return val;
                    }
                case FLOAT:
                    synchronized (m_readLock) {
                        m_raf.seek(offset);
                        if (m_bBigEndian) {
                            return m_raf.readFloat();
                        }
                        else {
                            // Little endian version not implemented.
                            return 0;
                        }
                    }
                case DOUBLE:
                    synchronized (m_readLock) {
                        m_raf.seek(offset);
                        if (m_bBigEndian) {
                            return m_raf.readDouble();
                        }
                        else {
                            // Little endian version not implemented.
                            return 0;
                        }
                    }
                case INT:
                    synchronized (m_readLock) {
                        m_raf.seek(offset);
                        if (m_bBigEndian) {
                            return ((double)m_raf.readInt());
                        }
                        else {
                            // Little endian version not implemented.
                            return 0;
                        }
                    }
                case LONG:
                    synchronized (m_readLock) {
                        m_raf.seek(offset);
                        if (m_bBigEndian) {
                            return ((double)m_raf.readLong());
                        }
                        else {
                            // Little endian version not implemented.
                            return 0;
                        }
                    }
            }
        }
        catch (IOException e) {
           e.printStackTrace();
        }

        return 0;
    }

//------------------------------------------------------------------------------
/**
 * Read from a random access file.
 */
    public void getValuesIntoReadDest(int arrayOffset,
        long fileOffset, int numberOfValues) {

        // Check if the file exists.
        if (m_raf == null) {
            boolean success = openRAF("r");
            if (!success)
                return;
        }

        if (m_bThreadStates.length != 0)
            throw new ConcurrentModificationException();

        fileOffset *= m_diskTypeLen;

        synchronized (m_readLock) {
            m_readSize = numberOfValues;
            if (m_readSize > 16000)
                m_readSize /= 16;
            boolean bSuccess = false;
            while (!bSuccess) {
                try {
                    m_readByteArray = new byte[m_readSize * m_diskTypeLen];
                    bSuccess = true;
                }
                catch (OutOfMemoryError e) {
                    m_readSize /= 2;
                    if (m_readSize < 1)
                        return;
                }
            }
        }

        int numBlocks =
            (int)Math.ceil((float)numberOfValues / (float)m_readSize);
        for (int n = 0; n < numBlocks; n++) {
            synchronized (m_readLock) {
                try {
                    m_raf.seek(fileOffset + (n * m_readSize * m_diskTypeLen));
                    m_raf.read(m_readByteArray);
                }
                catch (IOException e) {
                    e.printStackTrace();
                }

                m_readIndex = 0;
                m_readBlockOffset = arrayOffset + (n * m_readSize);
                m_readSizePerThread = m_readSize / m_numThreads + m_numThreads;
            }

            // Convert the data from byte to float using
            // parallel processing.
            MultiThreader multiThreader =
                new MultiThreader(this, m_numThreads, "RAF_Read");
            multiThreader.runThreads();
        }

        synchronized (m_readLock) {
            m_readIndex = 0;
            m_readSize = 0;
            m_bThreadStates = new boolean[0];
            m_readByteArray = new byte[0];
            m_readDestF = null;
            m_readDestD = null;
        }

        return;
    }

//------------------------------------------------------------------------------
/**
 * Read from a random access file.
 */
    public void getValuesInto(float[] dest, int arrayOffset,
        long fileOffset, int numberOfValues) {

        synchronized (m_readLock) {
            m_readDestF = dest;
        }

        getValuesIntoReadDest(arrayOffset, fileOffset, numberOfValues);
    }

//------------------------------------------------------------------------------
/**
 * Read from a random access file.
 */
    public void getValuesInto(double[] dest, int arrayOffset,
        long fileOffset, int numberOfValues) {

        synchronized (m_readLock) {
            m_readDestD = dest;
        }

        getValuesIntoReadDest(arrayOffset, fileOffset, numberOfValues);
    }

//------------------------------------------------------------------------------
/**
 * Read from a random access file into a byte array.
 */
    public void fillByteCache(byte[] dest, long fileOffset,
        int numberOfValues) {

        // Check if the file exists.
        if (m_raf == null) {
            boolean success = openRAF("r");
            if (!success)
                return;
        }

        fileOffset *= m_diskTypeLen;
        int numBytesPerRead = numberOfValues;
        if (numBytesPerRead > 32000)
            numBytesPerRead /= 32;
        int destOffset = 0;

        try {
            while (destOffset < numberOfValues) {
                if ((destOffset + numBytesPerRead) > dest.length)
                    numBytesPerRead = dest.length - destOffset;

                // Read from the file.
                m_raf.seek(fileOffset);
                m_raf.read(dest, destOffset, numBytesPerRead);

                // Advance to the next read.
                destOffset += numBytesPerRead;
                fileOffset += numBytesPerRead;
            }
        }
        catch (IOException e) {
            e.printStackTrace();
        }
        catch (Exception e) {
            e.printStackTrace();
        }

        return;
    }

//------------------------------------------------------------------------------
/**
 * Read from a random access file.
 */
    public void writeValueAt(double value, long offset) {

        // Check if the file exists.
        if (m_raf == null) {
            openRAF("rwd");
        }

        try {
            offset *= m_diskTypeLen;

            m_raf.seek(offset);

            switch (m_diskType) {
                case SHORT:
                default:
                    if (m_bBigEndian) {
                        m_raf.writeShort((short)value);
                        break;
                    }
                    else {
                        // Read the two bytes and reverse their order.
                        int lowByteAsInt = (Math.round((float)value)) & 255;
                        int highByteAsInt = (Math.round((float)value)) >> 8;

                        m_raf.writeByte(lowByteAsInt);
                        m_raf.writeByte(highByteAsInt);
                        break;
                    }
                case FLOAT:
                    if (m_bBigEndian) {
                        m_raf.writeFloat((float)value);
                        break;
                    }
                    else {
                        // Little endian version not implemented.
                        break;
                    }
                case DOUBLE:
                    if (m_bBigEndian) {
                        m_raf.writeDouble(value);
                        break;
                    }
                    else {
                        // Little endian version not implemented.
                        break;
                    }
                case INT:
                    if (m_bBigEndian) {
                        m_raf.writeInt((int)value);
                        break;
                    }
                    else {
                        // Little endian version not implemented.
                        break;
                    }
                case LONG:
                    if (m_bBigEndian) {
                        m_raf.writeLong((long)value);
                        break;
                    }
                    else {
                        // Little endian version not implemented.
                        break;
                    }
            }
        }
        catch (IOException e) {
            e.printStackTrace();
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void writeValuesFrom(float[] src, int arrayOffset,
        long fileOffset, int numberOfValues) {

        m_writeSrcF = src;
        writeValuesFromWriteSrc(arrayOffset, fileOffset, numberOfValues);
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void writeValuesFrom(double[] src, int arrayOffset,
        long fileOffset, int numberOfValues) {

        m_writeSrcD = src;
        writeValuesFromWriteSrc(arrayOffset, fileOffset, numberOfValues);
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void writeValuesFromWriteSrc(int arrayOffset,
        long fileOffset, int numberOfValues) {

        // Check if the file exists.
        if (m_raf == null) {
            openRAF("rwd");
        }

        if (m_bThreadStates.length != 0)
            throw new ConcurrentModificationException();

        try {
            fileOffset *= m_diskTypeLen;

            m_writeSize = numberOfValues;
            boolean bSuccess = false;
            while (!bSuccess) {
                try {
                    m_writeByteArray = new byte[m_writeSize * m_diskTypeLen];
                    bSuccess = true;
                }
                catch (OutOfMemoryError e) {
                    m_writeSize /= 2;
                    if (m_writeSize < 1)
                        return;
                }
            }

            if (m_writeSize < 1)
                return;

            for (int n = 0; n < (numberOfValues / m_writeSize); n++) {
                m_raf.seek(fileOffset + (n * m_writeSize * m_diskTypeLen));

                m_writeIndex = 0;
                m_writeBlockOffset = arrayOffset + (n * m_writeSize);
                m_writeSizePerThread =
                    m_writeSize / m_numThreads + m_numThreads;

                // Convert the data from float to short using
                // parallel processing.
                MultiThreader multiThreader =
                    new MultiThreader(this, m_numThreads, "RAF_Write");
                multiThreader.runThreads();

                m_raf.write(m_writeByteArray);
            }
        }
        catch (IOException e) {
            e.printStackTrace();
        }
        m_writeIndex = 0;
        m_writeSize = 0;
        m_bThreadStates = new boolean[0];
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public boolean fillWriteByteArray() {

        int iMin = 0;
        int iMax = 0;

        synchronized(m_writeLock) {
            iMin = m_writeIndex;
            iMax = iMin + m_writeSizePerThread;
            if (iMax > m_writeSize)
                iMax = m_writeSize;
            m_writeIndex += m_writeSizePerThread;
        }

        byte[] bytes = new byte[m_diskTypeLen];
        long iVal = 0;

        for (int i = iMin; i < iMax; i++) {
            // Create the two bytes in reverse order.
            int srcOffset = m_writeBlockOffset + i;

            switch (m_diskType) {
                case SHORT:
                case INT:
                case LONG:
                default:
                    if (m_memType == FLOAT)
                        iVal = Math.round(m_writeSrcF[srcOffset]);
                    if (m_memType == DOUBLE)
                        iVal = Math.round(m_writeSrcD[srcOffset]);
                    break;
                case FLOAT:
                case DOUBLE:
                    if (m_memType == FLOAT)
                        iVal = Float.floatToIntBits(m_writeSrcF[srcOffset]);
                    if (m_memType == DOUBLE)
                        iVal = Double.doubleToLongBits(m_writeSrcD[srcOffset]);
                    break;
            }

            // Break up the int value and store it in a big-endian temp array.
            for (int n = 0; n < m_diskTypeLen; n++) {
                bytes[n] = (byte)(iVal & 255);
                iVal >>= 8;
            }

            // Transfer the bytes from the big-endian temp array to the
            // output array.
            int byteOffset = i * m_diskTypeLen;
            if (m_bBigEndian) {
                for (int n = 0; n < m_diskTypeLen; n++) {
                    m_writeByteArray[byteOffset + n] =
                        bytes[m_diskTypeLen - n - 1];
                }
            }
            else {
                for (int n = 0; n < m_diskTypeLen; n++) {
                    m_writeByteArray[byteOffset + n] = bytes[n];
                }
            }

        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public boolean fillReadArray() {

        int iMin = 0;
        int iMax = 0;

        // Determine how much processing is left to do and grab a block
        // to be done now.
        synchronized(m_readLock) {
            iMin = m_readIndex;
            iMax = iMin + m_readSizePerThread;
            if (iMax > m_readSize)
                iMax = m_readSize;
            m_readIndex += m_readSizePerThread;
        }

        int readDestLength = 0;
        if (m_memType == FLOAT)
            readDestLength = m_readDestF.length;
        if (m_memType == DOUBLE)
            readDestLength = m_readDestD.length;

        for (int i = iMin; i < iMax; i++) {
            long iVal;
            if (m_bBigEndian) {
                iVal = 0;
                for (int n = 0; n < m_diskTypeLen; n++) {
                    iVal <<= 8;
                    int byteAsInt = (int)m_readByteArray[i * m_diskTypeLen + n];
                    byteAsInt &= 0xff;
                    iVal |= byteAsInt;
                }
            }
            else {
                iVal = (long)m_readByteArray[(i + 1) * m_diskTypeLen - 1];
                for (int n = (m_diskTypeLen - 2); n >= 0; n--) {
                    iVal <<= 8;
                    int byteAsInt = (int)m_readByteArray[i * m_diskTypeLen + n];
                    byteAsInt &= 0xff;
                    iVal |= byteAsInt;
                }
            }

            int destOffset = m_readBlockOffset + i;
            if (destOffset >= readDestLength)
                return false;

            switch (m_diskType) {
                case SHORT:
                case INT:
                case LONG:
                default:
                    if (m_memType == FLOAT)
                        m_readDestF[destOffset] = (float)iVal;
                    if (m_memType == DOUBLE)
                        m_readDestD[destOffset] = (double)iVal;
                    break;
                case FLOAT:
                    if (m_memType == FLOAT)
                        m_readDestF[destOffset] =
                            Float.intBitsToFloat((int)iVal);
                    if (m_memType == DOUBLE)
                        m_readDestD[destOffset] = Double.longBitsToDouble(iVal);
                    break;
                case DOUBLE:
                    if (m_memType == FLOAT)
                        m_readDestF[destOffset] =
                            (float)(Double.longBitsToDouble(iVal));
                    if (m_memType == DOUBLE)
                        m_readDestD[destOffset] = Double.longBitsToDouble(iVal);
                    break;
            }
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public boolean performThreadWork () {

        if (m_writeIndex < m_writeSize) {
            return fillWriteByteArray();
        }
        else {
            synchronized (m_readLock) {
                if (m_readIndex < m_readSize) {
                    return fillReadArray();
                }
                else {
                    return false;
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public boolean openRAF(String mode) {

        // Check if the file exists.
        if (m_raf == null) {
            if (m_file == null)
                return false;

            if (!m_file.exists()) {
                // Try using the same filename within the current directory.
                String currentFileName =
                    System.getProperty("gov.usgs.gr.currentfile");
                File file = new File(currentFileName);
                String parentString = file.getAbsoluteFile().getParent();
                parentString.replace('\\', '/');
                file = new File(file.getAbsoluteFile().getParent() + "/"
                    + m_file.getName());

                if (!file.exists() && !(mode.indexOf('w') >= 0)) {
                    // Write an error message if the file can't be found.
                    System.out.println("Could not find Random Access File "
                    + m_file.getAbsolutePath());
                }
                else {
                    if (file.exists() && mode.indexOf('w') >= 0) {
                        // Write an error message if the file already exists.
                        System.out.println(
                            "Warning -- Random Access File already exists "
                            + m_file.getAbsolutePath());
                        m_file = file;
                    }
                    else {
                        m_file = file;
                    }
                }
            }

            try {
                m_raf = new RandomAccessFile(
                    m_file, mode);
            }
            catch (FileNotFoundException fnfe) {
                m_errorString = " -- Error -- File not found";
                return false;
            }
        }

        return true;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void closeRAF() {

        if (m_raf != null) {
            try {
                m_raf.close();
            }
            catch (IOException e) {
                e.printStackTrace();
            }

            m_raf = null;

        }
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The number of n-tuples in the interleaved array.
 */
    public int getCount(){

        // Check if the file exists.
        if (m_raf == null) {
            boolean success = openRAF("r");
            if (!success) {
                m_errorString = " -- Error -- File could not be opened";
                return 0;
            }
        }

        int count = 0;
        try {
// stream I/O
            // The length() method calls seek() internally and repositions
            // the file pointer, so it must be in a synchronized block.
            synchronized (m_readLock) {
                count = (((int)m_raf.length()) / m_diskTypeLen / m_numDimensions);
            }
///// Fortran records
///            count = (((int)m_raf.length()) / 2 / (m_numDimensions + 5));
        }
        catch (IOException e) {
            e.printStackTrace();
        }

        return count;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return
 */
    public long getNumReads() {

        return m_readTotal;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return
 */
    public long getNumReadHits() {

        return m_readHits;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return
 */
    public long getReadMin() {

        return m_readMin;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return
 */
    public long getReadMax() {

        return m_readMax;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void resetCacheStats() {

        m_readHits = 0;
        m_readTotal = 0;
        m_readMin = Long.MAX_VALUE;
        m_readMax = Long.MIN_VALUE;
    }
    
//------------------------------------------------------------------------------
    /**
     * Implements required stub from DataSeries.
     * @param list    An ArrayList of objects.
     */
    public synchronized void setArray(ArrayList list){
        
    }

//------------------------------------------------------------------------------
    /**
     * Implements required stub from DataSeries.
     * @param buf   A character buffer from which to read.
     * @param offset    The offset into the buffer.
     * @param len       The length of the string to read from the buffer.
     */
    protected void tokenizeAndStore(String s0) {
        
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * Menu items include: <p>
 * "Select Area..."  <br>
 * "Select All"  <br>
 * "Deselect All"  <br>
 * "Delete Selected Points"  <p>
 * @return A menu full of commands applicable to this class.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
            "RAF");
        javax.swing.JMenuItem item = new javax.swing.JMenuItem
            ("Read");
        item.addActionListener(this);
        subMenu.add(item);

        return subMenu;
    }

//------------------------------------------------------------------------------
/**
 * Act on menu items being selected.
 * @param e The event to act on.
 */
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
            (javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Read")) {
            float val = getValueAt(0);
            System.out.println("val at offset 0 = " + val);
        }
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

        // Use the classname of this object.
        String fullClassName = getClass().getName();
        String name =
            fullClassName.substring(fullClassName.lastIndexOf('.') + 1);

        String s = name;

        if (m_errorString != null)
            s += m_errorString;
        else
            s += (": " + m_title);

        return (s);
    }
}

