/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-06-28
 * Date started:    2001-04-19
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import java.io.*;
import java.text.*;
import java.util.*;

//==============================================================================
/**
 * A data file object for reading, writing, and storing time series data
 * to/from a file.
 */
public class TimeDataFile extends gov.usgs.gr.data.DelimitedDataFile {
    
    /**
     * The pivot year to use for determining the century from two digit years.
     * If the year is greater than the pivot year, the century is assumed to
     * be 1900, otherwise 2000.
     */
    protected static int m_pivotYear = 60;
    /**
     * The year that all dates should be referenced to.  If it differs from
     * the year associated with the date, the data should be shifted until
     * they match.
     */
    protected int m_referenceYear = Integer.MIN_VALUE;
    /** The start date for data in the file. */
    protected GregorianCalendar m_startDate = new GregorianCalendar(0, 0, 0);
    /** The quote to write around the date and time in output.
     *  Usually single quote or double quote.
     */
    protected String m_quoteToWrite = "\'";
    /** The date separator to use in output.  Usually a slash or a dash. */
    protected String m_dateSeparatorToWrite = "/";
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/TimeDataFile.gif";
    
//------------------------------------------------------------------------------
    /**
     * Create an empty data file object.
     */
    public TimeDataFile() {
        
        super();
        // Set the title for use by the dialog box.
        m_title = "Time Data File";
        m_numHeaderLines = 0;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
        
        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();
        
        return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {
        
        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);
        
        if (icon == null)
            icon = super.setTreeIcon();
        
        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);
        
        return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public GrHashList getProperties() {
        
        GrHashList props = super.getProperties();
        
        props.put("dateSeparatorToWrite", new GrProperty(
            m_dateSeparatorToWrite, "Date Separator To Write"));
        props.put("quoteToWrite", new GrProperty(m_quoteToWrite,
            "Quote To Write"));
        
        return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Set the values of one or more of this object's properties.
     * @param props  A list of GrProperty objects specifying the new values for
     *               certain properties.
     */
    public void setProperties(HashMap props) {
        
        super.setProperties(props);
        
        GrProperty p;
        
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;
                
                if (key.equals("quoteToWrite"))
                    m_quoteToWrite = ((String)p.value);
                if (key.equals("dateSeparatorToWrite"))
                    m_dateSeparatorToWrite = ((String)p.value);
            }
            if (obj instanceof String) {
                String value = (String)obj;
                
                setProperty(key, value);
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Set a single property.
     * @see gov.usgs.gr.data.TimeDataFile#getProperties getProperties() for a list
     * of accepted properties.
     * @param key   The property name.
     * @param value The new value for the property.
     */
    public void setProperty(String key, String value) {
        
        super.setProperty(key, value);
        
        if (key.equals("quoteToWrite"))
            m_quoteToWrite = value;
        if (key.equals("dateSeparatorToWrite"))
            m_dateSeparatorToWrite = value;
    }
    
//------------------------------------------------------------------------------
/**
 * 
 */
    public static boolean matchesFormat(String[] lines) {

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();

        int numVals = -1;
        String inputDelimiters = " \t\n\r\f,";
        
        try {
            // Read each line of the data file.
            for (int i = 0; i < lines.length; i++) {
                if (lines[i] == null)
                    break;
                
                String line = lines[i];
                // Ignore lines that begin with a slash-slash as comment lines.
                if (line.indexOf("//") == 0)
                    continue;

                // Look for the quotes that separate the date from the rest
                // of the line.
                int quote1Index = line.indexOf('"');
                int quote2Index = 0;
                if (quote1Index >= 0) {
                    quote2Index = line.indexOf('"', quote1Index + 1);
                } else {
                    // Try single quotes if there were no double quotes.
                    quote1Index = line.indexOf('\'');
                    if (quote1Index >= 0) {
                        quote2Index = line.indexOf('\'', quote1Index + 1);
                    }
                }

                // Skip this line if there were no quotes.
                if (quote1Index < 0 || quote2Index < 0) continue;

                // Parse the date and add it to the date list.
                String dateString = line.substring(quote1Index + 1, quote2Index);
                SimpleDateFormat formatter =
                    new SimpleDateFormat("yyyy/MM/dd HH:mm", locale);
                formatter.setTimeZone(utcZone);
                GregorianCalendar date = new GregorianCalendar(utcZone, locale);
                date.clear();
                ParsePosition pos = new ParsePosition(0);
                date.setTime(formatter.parse(dateString, pos));

                // Parse the rest of the line as doubles.
                String remainderOfLine = line.substring(quote2Index + 1);

                double[] vals = AsciiIO.parseLineAsDoubles(remainderOfLine,
                    inputDelimiters);
                
                if (numVals < 0) {
                    numVals = vals.length;
                }
                else {
                    if (numVals != vals.length)
                        return false;
                }
            }
        }
        catch (Exception e) {
            return false;
        }
        
        if (numVals < 0)
            return false;
        
        return true;
    }
       
//------------------------------------------------------------------------------
    /**
     * Read the data through a reader attached to an open file.
     * The date is of the format 'yyyy/MM/dd HH:mm'.
     * <p><b> Pre Conditions </b><br>
     *      The header has been read with readHeader()
     * <p><b> Post Conditions </b><br>
     *      The file needs to be closed.
     * @throws IOException
     */
            protected void readData() throws IOException {
        
        // Remove any data that were children of this object.
        clear();
        
        // Make a list of column lists, one for each series.
        ArrayList columnList = new ArrayList();
        for (int i = 0; i < m_numChannels; i++) {
            columnList.add(new ArrayList());
        }
        
        // Fill the column list with dates and values.
        acquireData(columnList);
        
        // Get a handle to the date list in the first column.
        List dateList = (List)columnList.get(0);
        
        // Find missing points so they can be shown as deleted points.
        long mostCommonInc = medianIncrement(dateList);
        List deletedList = markMissingPoints(dateList, mostCommonInc);
        
        // Create a list containing a column of dates and columns for the other
        // data so they can be interleaved into individual series.
        m_startDate = (GregorianCalendar)
        ((GregorianCalendar)dateList.get(0)).clone();
        GregorianCalendar lastDate = (GregorianCalendar)
        ((GregorianCalendar)dateList.get(dateList.size() - 1)).clone();
        m_referenceYear = m_startDate.get(GregorianCalendar.YEAR);
        
        setupInterleavedSeries(columnList, deletedList);
    }
    
//------------------------------------------------------------------------------
    /**
     * Read the data from the data source and store them in the column list.
     * The first column is a list of GregorianCalendar date objects.  The others are
     * lists of Double objects.
     * @param columnList    This list of data values.
     */
    protected void acquireData(List columnList) throws IOException {
        
        String line;
        List dateList = (List)columnList.get(0);
        
        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        
        // Keep track of whether we have passed any // comments at the start.
        boolean bPastHeader = false;
        
        // Read each line of the data file.
        while ((line = m_br.readLine()) != null) {
            // Ignore lines that begin with a slash-slash as comment lines.
            if (line.indexOf("//") == 0) {
                m_headerLines.add(line);
                continue;
            }
            
            bPastHeader = true;
            
            // Look for the quotes that separate the date from the rest
            // of the line.
            int quote1Index = line.indexOf('"');
            int quote2Index = 0;
            if (quote1Index >= 0) {
                quote2Index = line.indexOf('"', quote1Index + 1);
            } else {
                // Try single quotes if there were no double quotes.
                quote1Index = line.indexOf('\'');
                if (quote1Index >= 0) {
                    quote2Index = line.indexOf('\'', quote1Index + 1);
                }
            }
            
            // Skip this line if there were no quotes.
            if (quote1Index < 0 || quote2Index < 0) continue;
            
            // Parse the date and add it to the date list.
            String dateString = line.substring(quote1Index + 1, quote2Index);
            SimpleDateFormat formatter =
                new SimpleDateFormat("yyyy/MM/dd HH:mm", locale);
            formatter.setTimeZone(utcZone);
            GregorianCalendar date = new GregorianCalendar(utcZone, locale);
            date.clear();
            ParsePosition pos = new ParsePosition(0);
            date.setTime(formatter.parse(dateString, pos));
            dateList.add(date);
            
            // Parse the rest of the line as doubles.
///            int firstCommaIndex = line.indexOf(',', quote2Index + 1);
///            String remainderOfLine = line.substring(firstCommaIndex + 1);
            String remainderOfLine = line.substring(quote2Index + 1);
            
            double[] vals = AsciiIO.parseLineAsDoubles(remainderOfLine,
                m_inputDelimiters);
            
            // Add each number on the line to the appropriate column list.
            for (int i = 0; i < vals.length; i++) {
                List column = null;
                // Add a new column to the list if there are not enough.
                if (columnList.size() < (i + 2)) {
                    column = new ArrayList();
                    columnList.add(column);
                } else {
                    column = (List)columnList.get(i + 1);
                }
                
                // Add the number to the list.
                column.add(vals[i]);
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Sets up the interleaved series from the provided list and adds them as
     * children of this object.
     * @param columnList    The list of data.
     * @param deletedList   The list of deleted points.
     */
    protected void setupInterleavedSeries(List<List> columnList,
        List<Double> deletedList) {
        
        // Interleave the series.
        List<TimeSeries> timeSeriesList = interleaveArrays(columnList);
        
        // Perform additional setup tasks on each new series.
        int index = 0;
        Iterator seriesIter = timeSeriesList.iterator();
        while (seriesIter.hasNext()) {
            TimeSeries series = (TimeSeries)seriesIter.next();
            
            // Set the type and title to a different number for each series.
            series.setType(index);
            series.setTitle("Series " + (index + 1));
            
            // Add a copy of the deleted list to the series.
            ArrayList<Double> copyOfDeletedList = new ArrayList<Double>();
            for (double dVal : deletedList)
                copyOfDeletedList.add(dVal);
            
            series.setDeletedList(copyOfDeletedList);
            
            // Delete points with a y value beyond +- 999980
            double[] min = {-Double.MAX_VALUE, -Double.MAX_VALUE};
            double[] max = {Double.MAX_VALUE, -999980};
            series.selectArea(min, max, true);
            min[0] = -Double.MAX_VALUE;
            min[1] = 999980;
            max[0] = Double.MAX_VALUE;
            max[1] = Double.MAX_VALUE;
            series.selectArea(min, max, true);
            
            series.deleteSelectedPoints();
            
            // Add the series as a child of this file.
            add(series);
            
            index++;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Write the time series back to a file.  Called by write().
     * The date is of the format "yyyy/MM/dd HH:mm".
     * <p><b> Pre Conditions </b><br>
     *      The header has been written with writeHeader()
     * <p><b> Post Conditions </b><br>
     *      The file needs to be closed.
     * @throws  IOException
     */
    protected void writeData() throws IOException {
        
        // Initialize local variables.  Use up to five decimal places on output.
        String lineString = "";
        
        // Store all time series children in a list.
        ArrayList seriesList = new ArrayList();
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            Object o = iter.next();
            
            while (o instanceof ObjectEffigy)
                o = ((ObjectEffigy)o).getSubject();
            
            if (o instanceof TimeSeries) {
                TimeSeries series = (TimeSeries)o;
                if (m_bWriteDeletedAsBad) {
                    series = (TimeSeries)series.createCopy(true);
                    series.selectAll();
                    SeriesGapFiller.fill(series,
                        new ConstantValueProvider(Double.NaN));
                }
                seriesList.add(series);
            }
        }
        
        // Create an array of indexes, initialized to zero, to keep track
        // of how many items from that series have been written.
        int[] indexes = new int[seriesList.size()];
        for (int i = 0; i < indexes.length; i++) indexes[i] = 0;
        
        double[][] arrays = new double[seriesList.size()][];
        int[] referenceYears = new int[seriesList.size()];
        for (int i = 0; i < arrays.length; i++) {
            TimeSeries series = (TimeSeries)seriesList.get(i);
            arrays[i] = series.getArray();
            referenceYears[i] = series.getReferenceYear();
        }
        
        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        
        // The format for the number's output.
        java.text.NumberFormat form = null;
        java.text.NumberFormat formU = null;
        if (m_outputDecimalFormat.length() > 0) {
            try {
                formU = new java.text.DecimalFormat(m_outputDecimalFormat);
            } catch(IllegalArgumentException e) {}
        }
            
        java.text.NumberFormat formD = new java.text.DecimalFormat(
            "0.####################");
        java.text.NumberFormat formE = new java.text.DecimalFormat(
            "0.####################E0#######");
        
        int referenceYear = 0;
        while (true) {
            // Look for the lowest X value that hasn't been written out yet.
            double currentX = Double.MAX_VALUE;
            for (int i = 0; i < arrays.length; i++) {
                if (indexes[i] < arrays[i].length &&
                    currentX > arrays[i][indexes[i]]) {
                    
                    currentX = arrays[i][indexes[i]];
                    referenceYear = referenceYears[i];
                }
            }
            
            if (currentX == Double.MAX_VALUE) break;
            
            // Write the date to the output string.
            lineString = m_quoteToWrite;
            // Since seconds won't be written out, add 0.5 minutes so that
            // after truncating the seconds, it will be written to the
            // nearest minute.
            currentX += 0.0003472222222222222;
            GregorianCalendar date = new GregorianCalendar(utcZone, locale);
            date = yearDayToGregCal(referenceYear, currentX, date);
            SimpleDateFormat formatter =
                new SimpleDateFormat("yyyy" + m_dateSeparatorToWrite + "MM"
                 + m_dateSeparatorToWrite + "dd HH:mm", locale);
            formatter.setTimeZone(utcZone);
            StringBuffer buf = formatter.format(date.getTime(),
                new StringBuffer(), new FieldPosition(0));
            lineString += buf.toString();
            lineString += m_quoteToWrite;
            
            // Write each series that is no more than 17 seconds beyond the
            // current time (0.0002 days).
            for (int i = 0; i < arrays.length; i++) {
                TimeSeries series = (TimeSeries)seriesList.get(i);
                if (indexes[i] < arrays[i].length &&
                    currentX > (arrays[i][indexes[i]] - 0.0002)) {
                    
                    lineString += ", ";
                    
                    // Decide on the format to use for output.
                    double yVal = arrays[i][indexes[i] + 1];
                    if (formU != null) {
                        form = formU;
                    }
                    else {
                        if (yVal > 100000000 || yVal < -100000000 ||
                            (yVal < 0.001 && yVal > -0.001)) {

                            form = formE;
                        } else {
                            form = formD;
                        }
                    }
                    
                    if (yVal == 0 && formU == null) {
                        lineString += "0";
                    } else {
                        if (Double.isNaN(yVal)) {
                            lineString += m_badIndicator;
                        }else {
                            String s = form.format(yVal);
                            lineString += s;}
                    }
                    indexes[i] += 2;
                } else {
                    lineString += (", " + m_badIndicator);
                }
            }
            
            // Add a newline character and send the line to the file.
            lineString += "\n";
            m_bw.write(lineString, 0, lineString.length());
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Transfers data from the input data lists to interleaved arrays.
     * Creates new TimeSeries using the arrays and passes them back in a new
     * ArrayList.  The first date in the first list is the "start date".
     * <p><b> Pre Conditions </b><br>
     *      The first column in inputList contains GregorianCalendar dates
     *		which are to be used as the x coordinate of each point.
     *		All of the other columns contain Double objects to be used as y values.
     * @param inputList	A List full of Lists, holding the date
     *                  series and each of the y series.
     * @return  An List of TimeSeries.
     */
    public static List<TimeSeries> interleaveArrays(List<List> inputList) {
        
        // Make a new array for holding interleaved data.
        List<TimeSeries> tsList = new ArrayList<TimeSeries>();
        List<GregorianCalendar> dateList = inputList.get(0);
        int numDates = dateList.size();
        int numColumns = inputList.size() - 1;
        double[][] seriesArrays = new double[numColumns][];
        
        // Define the start and end dates.
        GregorianCalendar startDate =
            (GregorianCalendar)dateList.get(0).clone();
        GregorianCalendar lastDate =
            (GregorianCalendar)dateList.get(numDates - 1).clone();
        int referenceYear = startDate.get(GregorianCalendar.YEAR);
        
        // Add an interleaved x/y array series for each y channel
        for (int column = 0; column < numColumns; column++) {
            // Create an array to hold the interleaved x-y points.
            seriesArrays[column] = new double[2 * numDates];
        }
        
        // Interleave the date and data in each array.
        for (int i = 0; i < numDates; i++) {
            GregorianCalendar date = (GregorianCalendar)dateList.get(i).clone();
            
            double jDay = julianDecimal(date, referenceYear);
            
            for (int column = 0; column < numColumns; column++) {
                double[] seriesArray = seriesArrays[column];
                double yVal = 0;
                try {
                    Double numObj = (Double)inputList.get(column + 1).get(i);
                    if (numObj != null)
                        yVal = numObj;
                } catch (IndexOutOfBoundsException e) { }
                
                seriesArray[(2 * i)] = jDay;
                seriesArray[(2 * i) + 1] = yVal;
            }
        }
        
        for (int column = 0; column < numColumns; column++) {
            // Create a time series from the interleaved array and add it to the
            // series list.
            TimeSeries series = createTimeSeriesInstance(seriesArrays[column]);
            series.setStartDate(startDate);
            
            // Shift the data to be referenced from the most appropriate year.
            series.syncToReferenceYear(findAppropriateReferenceYear(startDate,
                lastDate));
            tsList.add(series);
        }
        
        return tsList;
    }
    
//------------------------------------------------------------------------------
    /**
     * Creates an instance of the appropriate time series class to add as a chile
     * of this object.
     * @param array  The array to pass to the new time series.
     * @return  The newly created time series.
     */
    public static TimeSeries createTimeSeriesInstance(double[] array) {
        
        return new TimeSeries(array, 2);
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds the most common increment between dates in a list.
     * @param dateList  A sorted List of GregorianCalendar objects.
     * @return  The most common increment, measured in seconds.
     */
    public static long medianIncrement(List dateList) {
        
        // Initialize the objects for keeping track of the median increment.
        long lastMillis = 0;
        HashMap<Long, Long> diffHash = new HashMap<Long, Long>();
        
        // Run through the columns and combine them into a list of dates.
        Iterator iter = dateList.iterator();
        while (iter.hasNext()) {
            GregorianCalendar date = (GregorianCalendar)iter.next();
            
            // Find the number of milliseconds since 1970 for this date.
            long numMillis = date.getTime().getTime();
            long diffMillis = numMillis - lastMillis;
            double lowerPrecisionDiff = ((double)diffMillis) / 1000;
            Long key = new Long(Math.round(lowerPrecisionDiff));
            
            // Look for this difference in the hash map.
            Long numSameDiffs = diffHash.get(key);
            if (numSameDiffs == null) {
                diffHash.put(key, 1l);
            } else {
                diffHash.put(key, numSameDiffs + 1l);
            }
            lastMillis = numMillis;
        }
        
        // Look through the hash table for the most common increment.
        long mostCommonInc = -Long.MAX_VALUE;
        long mostCommonNum = -Long.MAX_VALUE;
        Set<Long> keySet = diffHash.keySet();
        for (long key : keySet) {
            long num = diffHash.get(key);
            if (num > mostCommonNum) {
                mostCommonInc = key;
                mostCommonNum = num;
            }
        }
        
        return mostCommonInc;
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds the most common increment between dates in a list.
     * @param dateList  A sorted List of GregorianCalendar objects.
     * @param mostCommonInc  The most common increment, measured in seconds.
     * @return  An List full of X and Y values marking the locations of
     *          missing points.  The Y values are set to 0.
     */
    public static List<Double> markMissingPoints(
        List<GregorianCalendar> dateList, long mostCommonInc) {
        
        // Create local variables to hold important date information.
        GregorianCalendar startDate = (GregorianCalendar)
        ((GregorianCalendar)dateList.get(0)).clone();
        GregorianCalendar lastDate = (GregorianCalendar)
        ((GregorianCalendar)dateList.get(dateList.size() - 1)).clone();
        int referenceYear = findAppropriateReferenceYear(startDate, lastDate);
        
        // Create a list of deleted point by looking through the date list
        // for missing points.
        long incMilli = mostCommonInc * 1000;
        long nextMilli = 0;
        long thisMilli = 0;
        ArrayList<Double> deletedList = new ArrayList<Double>();
        
        // Bail out if the increment is too small.
        long firstMilli = startDate.getTime().getTime();
        long lastMilli = lastDate.getTime().getTime();
        long diffMilli = lastMilli - firstMilli;
        if (incMilli <= 0 || ((diffMilli / incMilli) > Long.MAX_VALUE))
            return new ArrayList();
        
        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        
        Iterator iter = dateList.iterator();
        while (iter.hasNext()) {
            GregorianCalendar date = (GregorianCalendar)iter.next();
            thisMilli = date.getTime().getTime();
            if (nextMilli == 0) nextMilli = thisMilli;
            
            // Only insert new points if the next point is more than a
            // second past the missing point.
            while (nextMilli < (thisMilli - 1000)) {
                GregorianCalendar missingDate = new GregorianCalendar(
                    utcZone, locale);
                missingDate.clear();
                missingDate.setTime(new Date(nextMilli));
                double xVal = julianDecimal(missingDate, referenceYear);
                deletedList.add(xVal);
                deletedList.add(0d);
                
                nextMilli += incMilli;
            }
            
            nextMilli = thisMilli + incMilli;
        }
        
        return deletedList;
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds the year that would be most approriate to use for a given span of data.
     * Chooses the following year if the span starts after October 1st and extends
     * more than 90 days into the following year.
     * @param startDate The start of the data span.
     * @param lastDate  The end of the data span.
     * @return  The appropriate reference year in four digit format.
     */
    public static int findAppropriateReferenceYear(GregorianCalendar startDate,
        GregorianCalendar lastDate) {
        
        // Choose the following year if the first date is after October 1st,
        // and it goes more than 90 days into the following year.
        int referenceYear = startDate.get(GregorianCalendar.YEAR);
        GregorianCalendar waterYearStart = new GregorianCalendar(
            referenceYear, 9, 1);
        waterYearStart.setTimeZone(TimeZone.getTimeZone("UTC"));
        
        if (startDate.get(Calendar.DAY_OF_YEAR) >=
            waterYearStart.get(Calendar.DAY_OF_YEAR) &&
            ((lastDate.get(Calendar.DAY_OF_YEAR) > 90 &&
            lastDate.get(Calendar.YEAR) > startDate.get(Calendar.YEAR)) ||
            lastDate.get(Calendar.YEAR) > startDate.get(Calendar.YEAR) + 1)) {
            
            return (referenceYear + 1);
        } else {
            return referenceYear;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Convert a Calendar date to a julian day with decimal for fractions of a day.
     * @param date  The date that will be converted.
     * @param referenceYear	A two or four digit year from which to reference the
     *                      Julian day.
     * @return		A decimal number representing the number of Julian days
     * 				from December 31st at 00:00 of the previous year.  Fractions of
     *				a day are represented as a decimal.
     */
    public static double julianDecimal(GregorianCalendar date,
        int referenceYear) {
        
        // Allign the data offset to the reference year.
        int julianOffset = 0;
        while (date.get(Calendar.YEAR) < referenceYear) {
            if (date.isLeapYear(referenceYear - 1))
                julianOffset -= 366;
            else
                julianOffset -= 365;
            
            referenceYear--;
        }
        while (date.get(Calendar.YEAR) > referenceYear) {
            if (date.isLeapYear(referenceYear))
                julianOffset += 366;
            else
                julianOffset += 365;
            
            referenceYear++;
        }
        
        // Convert the date to a julian day plus a fraction.
        double julianDay = (double)date.get(Calendar.DAY_OF_YEAR)
        + (double)julianOffset
            + (double)date.get(Calendar.HOUR_OF_DAY) / 24
            + (double)date.get(Calendar.MINUTE) / 1440
            + (double)date.get(Calendar.SECOND) / 86400;
        
        return julianDay;
    }
    
//------------------------------------------------------------------------------
    /**
     * Convert a julian day with decimal to a date object.
     * @param referenceYear     A four digit year from which the day is
     *                          referenced.
     * @param decimalDay        The day of the year with the hour, minute, and
     *                          second represented by a decimal.
     * @param date              The date object to store the new date.
     * @return		A new GregorianCalendar object for the specified date.
     */
    public static GregorianCalendar yearDayToGregCal(int referenceYear,
        double decimalDay, GregorianCalendar date) {
        
        date.clear();
        int dayOfYear = (int)decimalDay;
        double decimalTime = decimalDay - dayOfYear;
        int hour = (int)(decimalTime * 24);
        int minute = (int)((decimalTime * 24 - hour) * 60);
        int second = (int)(((decimalTime * 24 - hour) * 60 - minute) * 60);
        
        date.set(Calendar.YEAR, referenceYear);
        date.set(Calendar.DAY_OF_YEAR, (int)dayOfYear);
        date.set(Calendar.HOUR_OF_DAY, hour);
        date.set(Calendar.MINUTE, minute);
        date.set(Calendar.SECOND, second);
        
        return date;
    }
    
//------------------------------------------------------------------------------
    /**
     * Creates a four digit year from a two digit year using the pivot year.
     * Adds either 1900 or 2000 to year if year is less than 100.
     * @param year	A two or four digit year.
     * @return The four digit version of what was passed in.
     */
    public static int fourDigitYear(int year) {
        
        // Check that year is not four digits already.
        if (year < 100) {
            if (year > m_pivotYear) year += 1900;
            else year += 2000;
        }
        
        return year;
    }
}

