/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    2006-03-13
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import gov.usgs.gr.*;
import java.util.*;

//==============================================================================
/**
 * A dimension within a DataSeries.
 */
public class TimeReference extends gov.usgs.gr.GrObject {

    public static final int MILLISECONDS = 0;
    public static final int SECONDS = 1;
    public static final int MINUTES = 2;
    public static final int HOURS = 3;
    public static final int DAYS = 4;
    public static final int WEEKS = 5;

    /**
     * The pivot year to use for determining the century from two digit years.
     * If the year is greater than the pivot year, the century is assumed to
     * be 1900, otherwise 2000.
     */
    protected static int m_pivotYear = 60;
	/** The start date for data in the series. */
    protected GregorianCalendar m_referenceDate =
        new GregorianCalendar(0, 0, 0);
	/** The year the data are reference from. */
    protected int m_referenceYear = Integer.MIN_VALUE;
    /** The units of the time data in the series. */
    protected int m_timeUnits = DAYS;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/DataDimension.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a generic dimension.
 */
    public TimeReference() {

    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a TimeReference to a certain year.
 * @param year The reference year.
 */
    public TimeReference(int year) {

		m_referenceYear = year;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.  Properties include:
 * <p>
 * @return A list of GrProperty objects describing this object's properties.
 */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

    	props.put("refYear", new GrProperty(m_referenceYear,
            "Refererence Year"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.gr.data.TimeReference#getProperties getProperties() for a list
 * of accepted properties.
 * @param props  A list of GrProperty objects specifying the new values for
 *               certain properties.
 */
    public void setProperties(HashMap props) {

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

            	if (key.equals("refYear"))
                    m_referenceYear = (Integer)p.value;
            }
            if (obj instanceof String) {
                String value = (String)obj;

                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.gr.data.TimeReference#getProperties getProperties() for a list
 * of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

    	if (key.equals("refYear"))
            m_referenceYear = Integer.parseInt(value);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 * @return The icon representing this class.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.  Prevents children from being added.
 * @param gro The object to add.
 * @return Whether the object was successfully added.  Always false.
 */
    public boolean add(GrObject gro) {

    	return false;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The DataSeries object that contains this dimension object.
 */
    public DataSeries getAssociatedSeries() {

        Iterator iter = m_observers.listIterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof DataSeries)
                return (DataSeries)gro;
        }

        return null;
    }

//------------------------------------------------------------------------------
/**
 * Shifts all data so they are referenced from the reference year.
 * @param newReferenceYear  The year that data should be referenced from.
 */
    public void syncToReferenceYear (int newReferenceYear) {

        // If it is already using the correct reference year, return.
        if (m_referenceYear == newReferenceYear) return;

        // If the reference year is uninitialized, set it to the start year.
        if (m_referenceYear <= Integer.MIN_VALUE)
        	m_referenceYear = m_referenceDate.get(Calendar.YEAR);

        // Calculate the difference between the current and new ref years.
        int yearDiff = m_referenceYear - newReferenceYear;

		// January 1st at 00:00 is 1.0 on the axis.
        int shift = 365 * yearDiff;
        int inc = 0;

        // Determine how much to adjust the shift if leap years were
        // encountered between the current start year and the reference.
        inc = (yearDiff > 0) ? 1 : -1;
        int lower = (yearDiff > 0) ? newReferenceYear : m_referenceYear;
        int upper = (yearDiff > 0) ? m_referenceYear : newReferenceYear;

        for (int i = lower; i < upper; i++) {
        	if (m_referenceDate.isLeapYear(i)) {
            	shift += inc;
            }
        }

        // Update the reference year.
        m_referenceYear = newReferenceYear;

/// Account for this...
        // Shift all the x data points.
//        for (int i = 0; i < m_array.length; i += 2) {
//        	m_array[i] += shift;
//        }

        // Shift all the x points in the deleted list.
//        ArrayList newDeletedList = new ArrayList();
//        Iterator iter = m_deleted.iterator();
//        while (iter.hasNext()) {
//        	Double xVal = (Double)iter.next();
//            Double yVal = (Double)iter.next();
//            newDeletedList.add(new Double(xVal + shift));
//            newDeletedList.add(yVal);
//        }
//
//		setDeletedList(newDeletedList);
//
//        // Update the min and max value for each dimension.
//        updateMinMax();
    }

//------------------------------------------------------------------------------
/**
 * Access the TimeReference object associated with the given data series.
 * @param series The DataSeries from which to retrieve the TimeReference.
 * @return The TimeReference object associated with the given data series.
 */
    public static TimeReference getTimeReference (DataSeries series) {

        Enumeration enum1 = series.children();
        while (enum1.hasMoreElements()) {
            GrObject gro = (GrObject)enum1.nextElement();
            if (gro instanceof TimeReference)
                return (TimeReference)gro;
        }

        return null;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The start date of the series.
 */
    public GregorianCalendar getStartDate() {
        return (GregorianCalendar)m_referenceDate.clone();
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param date The new start date of the series.
 */
    public void setStartDate(GregorianCalendar date) {

        m_referenceDate = date;
        if (m_referenceYear <= Integer.MIN_VALUE)
            m_referenceYear = m_referenceDate.get(Calendar.YEAR);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The reference year from which the data are measured.
 */
    public int getReferenceYear () {

    	return m_referenceYear;
    }

//------------------------------------------------------------------------------
/**
 * Get the reference year from a given data series.
 * @param series    The series from which to get the time reference.
 * @return The reference year from which the data are measured.
 */
    public static int getReferenceYear (DataSeries series) {

        if (series instanceof TimeSeries)
            return ((TimeSeries)series).getReferenceYear();

        TimeReference tr = getTimeReference(series);

        if (tr == null)
            return Integer.MIN_VALUE;

    	return tr.getReferenceYear();
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.  Does not shift data from the old year to
 * the new one.
 * @param value The new reference year.
 */
    public void setReferenceYear(int value) {
        m_referenceYear = value;
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  A string that describes this object.
 */
    public String toString() {

        return ("Time Reference: " + m_referenceYear);
    }

//------------------------------------------------------------------------------
/**
 * Finds the year that would be most approriate to use for a given span of data.
 * Chooses the following year if the span starts after October 1st and extends
 * more than 90 days into the following year.
 * @param startDate The start of the data span.
 * @param lastDate  The end of the data span.
 * @return  The appropriate reference year in four digit format.
 */
    public static int findAppropriateReferenceYear(GregorianCalendar startDate,
        GregorianCalendar lastDate) {

        // Choose the following year if the first date is after October 1st,
        // and it goes more than 90 days into the following year.
        int referenceYear = startDate.get(GregorianCalendar.YEAR);
        GregorianCalendar waterYearStart = new GregorianCalendar(
        	referenceYear, 9, 1);
        waterYearStart.setTimeZone(TimeZone.getTimeZone("UTC"));

        if (startDate.get(Calendar.DAY_OF_YEAR) >=
        	waterYearStart.get(Calendar.DAY_OF_YEAR) &&
            (lastDate.get(Calendar.DAY_OF_YEAR) > 90 &&
            lastDate.get(Calendar.YEAR) > startDate.get(Calendar.YEAR))) {

            return (referenceYear + 1);
        }
        else {
        	return referenceYear;
        }
    }

//------------------------------------------------------------------------------
/**
 * Convert a Calendar date to a julian day with decimal for fractions of a day.
 * @param date  The date that will be converted.
 * @param referenceYear	A two or four digit year from which to reference the
 *                      Julian day.
 * @return		A decimal number representing the number of Julian days
 * 				from December 31st at 00:00 of the previous year.  Fractions of
 *				a day are represented as a decimal.
 */
    public static double julianDecimal (GregorianCalendar date,
        int referenceYear) {

    	// Allign the data offset to the reference year.
        int julianOffset = 0;
        while (date.get(Calendar.YEAR) < referenceYear) {
        	if (date.isLeapYear(referenceYear - 1))
            	julianOffset -= 366;
            else
            	julianOffset -= 365;

            referenceYear--;
        }
        while (date.get(Calendar.YEAR) > referenceYear) {
        	if (date.isLeapYear(referenceYear))
            	julianOffset += 366;
            else
            	julianOffset += 365;

            referenceYear++;
        }

        // Convert the date to a julian day plus a fraction.
        double julianDay = (double)date.get(Calendar.DAY_OF_YEAR)
        	+ (double)julianOffset
            + (double)date.get(Calendar.HOUR_OF_DAY) / 24
            + (double)date.get(Calendar.MINUTE) / 1440
            + (double)date.get(Calendar.SECOND) / 86400;

        return julianDay;
    }

//------------------------------------------------------------------------------
/**
 * Convert a julian day with decimal to a date object.
 * @param referenceYear     A four digit year from which the day is
 *                          referenced.
 * @param decimalDay        The day of the year with the hour, minute, and
 *                          second represented by a decimal.
 * @param date              The date object to store the new date.
 * @return		A new GregorianCalendar object for the specified date.
 */
    public static GregorianCalendar yearDayToGregCal (int referenceYear,
        double decimalDay, GregorianCalendar date) {

        date.clear();
        int dayOfYear = (int)decimalDay;
        double decimalTime = decimalDay - dayOfYear;
        int hour = (int)(decimalTime * 24);
        int minute = (int)((decimalTime * 24 - hour) * 60);
        int second = (int)(((decimalTime * 24 - hour) * 60 - minute) * 60);

        date.set(Calendar.YEAR, referenceYear);
        date.set(Calendar.DAY_OF_YEAR, (int)dayOfYear);
        date.set(Calendar.HOUR_OF_DAY, hour);
        date.set(Calendar.MINUTE, minute);
        date.set(Calendar.SECOND, second);

        return date;
    }

//------------------------------------------------------------------------------
/**
 * Creates a four digit year from a two digit year using the pivot year.
 * Adds either 1900 or 2000 to year if year is less than 100.
 * @param year	A two or four digit year.
 * @return The four digit version of what was passed in.
 */
    public static int fourDigitYear (int year) {

    	// Check that year is not four digits already.
    	if (year < 100) {
        	if (year > m_pivotYear) year += 1900;
            else year += 2000;
        }

        return year;
    }

//------------------------------------------------------------------------------
/**
 * Looks for codes in the string that indicate it should be converted from
 * hours, minutes, or seconds to days.  The codes are 'd', 'h', 'm', and 's'.
 * If they are not present, simply parse the string as a double.  Example, if
 * the string is "12h", it will be converted from hours to days and 0.5 will
 * be returned.  "1d6h5.5m" would be one day, six hours, five minutes, and
 * thirty seconds.
 * @param s The string to parse.
 * @return  The double parsed from the string, accounting for 'd', 'h', 'm', or
 *          's' codes.
 */
    public static double convertToDecimalDay (String s) {

        // Initialize the cumulative value to zero.
        double val = 0;

        // Parse any portion at the beginning of the string ending in 'd' as
        // days, and then remove it.
        s = s.trim().toLowerCase();
        int index = s.indexOf("d");
        if (index >= 0) {
            val = Double.parseDouble(s.substring(0, index));
            s = s.substring(index + 1);
        }

        // Parse any portion at the beginning of the string ending in 'h' as
        // hours, and then remove it.
        index = s.indexOf("h");
        if (index >= 0) {
            val += (Double.parseDouble(s.substring(0, index)) / 24);
            s = s.substring(index + 1);
        }

        // Parse any portion at the beginning of the string ending in 'm' as
        // minutes, and then remove it.
        index = s.indexOf("m");
        if (index >= 0) {
            val += (Double.parseDouble(s.substring(0, index)) / 1440);
            s = s.substring(index + 1);
        }

        // Parse any portion at the beginning of the string ending in 's' as
        // seconds, and then remove it.
        index = s.indexOf("s");
        if (index >= 0) {
            val +=(Double.parseDouble(s.substring(0, index)) / 86400);
            s = s.substring(index + 1);
        }

        // Parse any remaining portion of the string as days.
        if (s.length() > 0) {
            val += Double.parseDouble(s);
        }

        return val;
    }
}
