/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-07-20
 * Date started:    1999-05-24
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import gov.usgs.gr.*;
import gov.usgs.gr.visual.PageOfGraphs;
import java.util.*;

//==============================================================================
/**
 * A two dimensional data series where the first dimension is time measured
 * in days.  A reference year is associated with the data.
 */
public class TimeSeries extends DoubleDataSeries {

	/** The start date for data in the file. */
    protected GregorianCalendar m_startDate = new GregorianCalendar(0, 0, 0);
	/** The reference year that is currently used. */
    protected TimeReference m_timeReference = new TimeReference();
    /** The dimension of the associated data series that contains time data. */
    protected int m_timeDimension = 0;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/TimeSeries.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.  Passes 2 to the DataSeries
 * constructor as the number of dimensions for this object.
 */
    public TimeSeries(int numDims) {

        super(numDims);

        m_dims[0].setProperty("locked", "true");
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a data series with a data type and title.
 * Locks editing on the X (0) dimension.
 * @param numericType	The numeric code describing the data held here.
 * @param title			The title of this data series.
 */
    public TimeSeries(int numericType, String title, int numDims) {

        super(numericType, title, numDims);

        m_dims[0].setProperty("locked", "true");
   }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a data series from an interleaved array of X-Y data.
 * Locks editing on the X (0) dimension.
 * @param series	The array of data for this object.
 */
    public TimeSeries(double[] array, int numDims) {

        super(array, numDims);

        m_dims[0].setProperty("locked", "true");
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Creates another TimeSeries with the same data and properties as this one.
 * @param bDeep		True if a deep copy (unshared arrays) is to be made.
 * @return    A copy of this data series.
 */
    public GrObject createCopy(boolean bDeep) {

    	TimeSeries series = new TimeSeries(m_numDimensions);
        series.copyFromDataSeries(this, bDeep);
        series.setStartDate((GregorianCalendar)m_startDate.clone());
        series.setReferenceYear(m_timeReference.getReferenceYear());

        return series;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * refYear Integer "Refererence Year" <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

    	props.put("refYear", new GrProperty(m_timeReference.getReferenceYear(),
            "Refererence Year"));
    	props.put("timeDim", new GrProperty(m_timeDimension,
            "Time Dimension"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.gr.data.TimeSeries#getProperties getProperties() for a list
 * of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("refYear"))
                	m_timeReference.setReferenceYear((Integer)p.value);
            }
            if (obj instanceof String) {
                setProperty(key, (String)obj);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.gr.data.TimeSeries#getProperties getProperties() for a list
 * of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

    	if (key.equals("refYear"))
        	m_timeReference.setReferenceYear(Integer.parseInt(value));
    }

//------------------------------------------------------------------------------
/**
 * Shifts all data so they are referenced from the reference year.
 * @param newReferenceYear  The year that data should be referenced from.
 */
    public void syncToReferenceYear (int newReferenceYear) {

        // If it is already using the correct reference year, return.
        int referenceYear = m_timeReference.getReferenceYear();
        if (referenceYear == newReferenceYear) return;

        // If the reference year is uninitialized, set it to the start year.
        if (referenceYear <= Integer.MIN_VALUE)
        	referenceYear = m_startDate.get(Calendar.YEAR);

        // Calculate the difference between the current and new ref years.
        int yearDiff = referenceYear - newReferenceYear;

		// January 1st at 00:00 is 1.0 on the axis.
        int shift = 365 * yearDiff;
        int inc = 0;

        // Determine how much to adjust the shift if leap years were
        // encountered between the current start year and the reference.
        inc = (yearDiff > 0) ? 1 : -1;
        int lower = (yearDiff > 0) ? newReferenceYear : referenceYear;
        int upper = (yearDiff > 0) ? referenceYear : newReferenceYear;

        for (int i = lower; i < upper; i++) {
        	if (m_startDate.isLeapYear(i)) {
            	shift += inc;
            }
        }

        // Update the reference year.
        m_timeReference.setReferenceYear(newReferenceYear);

        // Shift all the x data points.
        for (int i = 0; i < m_array.length; i += 2) {
        	m_array[i] += shift;
        }

        // Shift all the x points in the deleted list.
        List<Double> newDeletedList = new ArrayList<Double>();
        Iterator<Double> iter = m_deleted.iterator();
        while (iter.hasNext()) {
        	Double xVal = iter.next();
            Double yVal = iter.next();
            newDeletedList.add(xVal + shift);
            newDeletedList.add(yVal);
        }

		setDeletedList(newDeletedList);

        // Update the min and max value for each dimension.
        updateMinMax();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The start date of the series.
 */
    public GregorianCalendar getStartDate() {
        return (GregorianCalendar)m_startDate.clone();
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param date The new start date of the series.
 */
    public void setStartDate(GregorianCalendar date) {

        m_startDate = date;
        int referenceYear = m_timeReference.getReferenceYear();
        if (referenceYear <= Integer.MIN_VALUE)
            referenceYear = m_startDate.get(Calendar.YEAR);

        m_timeReference.setReferenceYear(referenceYear);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The reference year from which the data are measured.
 */
    public int getReferenceYear () {

        int referenceYear = m_timeReference.getReferenceYear();
        if (referenceYear <= Integer.MIN_VALUE)
        	referenceYear = m_startDate.get(Calendar.YEAR);

        m_timeReference.setReferenceYear(referenceYear);

    	return referenceYear;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.  Does not shift data from the old year to
 * the new one.
 * @param value The new reference year.
 */
    public void setReferenceYear(int value) {
        m_timeReference.setReferenceYear(value);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The reference year from which the data are measured.
 */
    public TimeReference getTimeReference () {

    	return m_timeReference;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The time dimension.
 */
    public int getTimeDimension () {

    	return m_timeDimension;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value The time dimension.
 */
    public void setTimeDimension(int value) {
	
	if (value >= 0 && value < m_numDimensions)	    
	    m_timeDimension = value;
    }
    
//------------------------------------------------------------------------------
    /**
     * Checks if the object argument is a TimeSeries with the specified
     * number of dimensions.
     * @param obj	The object to be checked
     * @param numDims	The number of dimensions the series should have.
     * @return Whether the object was a TS with the right number of dimensions.
     */
    public static boolean countDims(Object obj, int numDims) {
	
	if (obj instanceof TimeSeries) {
	    TimeSeries series = (TimeSeries)obj;
	    if (series.getNumDimensions() == numDims)
		return true;
	}
	
	return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Act on menu items being selected.
     * @param e The event to act on.
     */
    public void actionPerformed(java.awt.event.ActionEvent e) {
	
	super.actionPerformed(e);
        javax.swing.JMenuItem source =
	    (javax.swing.JMenuItem)(e.getSource());
	String srcText = source.getText();
	if (srcText.equals("Sync Reference Year to Page")) {
            PageOfGraphs page = (PageOfGraphs)firstParentInstanceOf(
                "gov.usgs.gr.visual.PageOfGraphs", true);
            if (page != null) {
                // Sync reference years if there are time series in the
                // current page.
                int refYear = -1;
            
                // Get the first TimeSeries object on the page.
                TimeSeries tSeries = (TimeSeries)page.firstChildInstanceOf(
                        "gov.usgs.gr.data.TimeSeries", true);

                if (tSeries != null)
                    refYear = tSeries.getReferenceYear();

                if (refYear > 0)
                    syncToReferenceYear(refYear);
            }
	}
	if (srcText.equals("Sync Page to this Reference Year")) {
            int refYear = m_timeReference.getReferenceYear();
            PageOfGraphs page = (PageOfGraphs)firstParentInstanceOf(
                "gov.usgs.gr.visual.PageOfGraphs", true);
            if (page != null && refYear > 0) {         
                // Get a list of all TimeSeries objects on the page.
                List tsList = page.findChildInstancesOf(
                        "gov.usgs.gr.data.TimeSeries", true);

                Iterator iter = tsList.iterator();
                while (iter.hasNext()) {
                    TimeSeries series = (TimeSeries)iter.next();
                    series.syncToReferenceYear(refYear);
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * Menu items include those from DataSeries, as well as: <p>
 * "Sync Reference Year to Page"  <br>
 * "Sync Page to this Reference Year"  <br>
 * @return A menu full of commands applicable to this class.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenuItem subMenu = super.getMenuCommands();
        subMenu.setText("Time Series");
        javax.swing.JMenuItem item = new javax.swing.JMenuItem
            ("Sync Reference Year to Page");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Sync Page to this Reference Year");
        item.addActionListener(this);
        subMenu.add(item);

        return subMenu;
    }
}





