/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    1999-05-24
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import java.util.*;

//==============================================================================
/**
 * Reads and writes time series data to a file.
 */
public abstract class TimeSeriesDataFile extends XyDataFile {
    /**
     * The pivot year to use for determining the century from two digit years.
     * If the year is greater than the pivot year, the century is assumed to
     * be 1900, otherwise 2000.
     */
    protected static int m_pivotYear = 60;
    /**
     * The year that all dates should be referenced to.  If it differs from
     * the year associated with the date, the data should be shifted until
     * they match.
     */
    protected int m_referenceYear = Integer.MIN_VALUE;
	/** The start date for data in the file. */
    protected GregorianCalendar m_startDate = new GregorianCalendar(0, 0, 0);

//------------------------------------------------------------------------------
/**
 * Creates an empty time series data file.
 */
    public TimeSeriesDataFile () {

        super();
    }

//------------------------------------------------------------------------------
/**
 * Creates a data series object to store the data.
 * Overrides method in gov.usgs.gr.data.XyDataFile.
 * @return	A new DataSeries.
 */
    protected DoubleDataSeries newSeries() {

    	return new TimeSeries(2);
    }

//------------------------------------------------------------------------------
/**
 * Creates an X-Y data series object to store the data.
 * @param array The interleaved data array for the new series.
 * @return	A new DataSeries.
 */
    protected DoubleDataSeries newSeries(double[] array) {

    	return new TimeSeries(array, 2);
    }

//------------------------------------------------------------------------------
/**
 * Creates an X-Y data series object to store the data.
 * @param numericType   A numeric code describing the data.
 * @param title The title of the data series.
 * @return	A new X-Y data series with no data, but numeric type and title set.
 */
    protected DoubleDataSeries newSeries(int numericType, String title) {

    	return new TimeSeries(numericType, title, 2);
    }

//------------------------------------------------------------------------------
/**
 * Creates a four digit year from a two digit year using the pivot year.
 * Adds either 1900 or 2000 to year if year is less than 100.
 * @param year	A two or four digit year.
 * @return The four digit version of what was passed in.
 */
    public static int fourDigitYear (int year) {

    	// Check that year is not four digits already.
    	if (year < 100) {
        	if (year > m_pivotYear) year += 1900;
            else year += 2000;
        }

        return year;
    }

//------------------------------------------------------------------------------
/**
 * Convert a Calendar date to a Julian day with decimal for fractions of a day.
 * <p><b> Pre Conditions </b><br>
 *      m_startDate and m_referenceYear must be set.
 * @param date  The date to be converted.
 * @param referenceYear	A two or four digit year to reference the new Julian
 *                      day from.
 * @return		A decimal number representing the number of Julian days
 * 				from December 31st at 00:00 the year preceding the reference
 *              year.  Hours are represented as a decimal.
 */
    public static double julianDecimal (GregorianCalendar date,
        int referenceYear) {

    	// Assign an offset to the day if it is in a year following
        // the start of the data.
        int julianOffset = 0;
        while (date.get(Calendar.YEAR) > referenceYear) {
            if (date.isLeapYear(referenceYear))
                julianOffset += 366;
            else
                julianOffset += 365;

            referenceYear++;
        }

        // Convert the date to a julian day plus a fraction.
        double julianDay = (double)date.get(Calendar.DAY_OF_YEAR)
        	+ julianOffset
            + (double)date.get(Calendar.HOUR_OF_DAY) / 24
            + (double)date.get(Calendar.MINUTE) / 1440
            + (double)date.get(Calendar.SECOND) / 86400;

        return julianDay;
    }

//------------------------------------------------------------------------------
/**
 * Transfers data from the input data lists to interleaved arrays.
 * Creates a new TimeSeries using the arrays and the supplied title and
 * adds the series to the m_seriesList.  The first date in the list is
 * the "start date".
 * <p><b> Pre Conditions </b><br>
 *      The first column in m_columnList contains GregorianCalendar dates
 *		which are to be used as the x coordinate of each point.
 *		All of the other columns contain Double objects to be used as y values.
 * @param titles	An array of titles to use for each series created.
 * @param deletedGroup	A list full of Lists, each of which contains
 *						interleaved x-y positions describing deleted points
 *						for each of the series.
 */
    protected void interleaveColumns(String[] titles, List deletedGroup) {

    	// Officially set the start date to the first date in the list.
        setStartDate((GregorianCalendar)
        	((List)m_columnList.get(0)).get(0));

        // Make a new array for holding interleaved data.
        int numColumns = m_columnList.size() - 1;
        m_seriesArrays = new double[numColumns][];

        // Add an interleaved x/y array series for each y channel
        for (int column = 0; column < numColumns; column++) {
        	// Create an array to hold the interleaved x-y points.
    		m_seriesArrays[column] = new double[2 *
        		((List)m_columnList.get(column + 1)).size()];
            double[] array = m_seriesArrays[column];

    		// Grab the source lists.
        	int columnOffset = 1;
        	int index = 0;
        	Iterator xIter = ((List)m_columnList.get(0)).iterator();
        	Iterator yIter = ((List)m_columnList
        		.get(column + columnOffset)).iterator();

        	// Combine the date column with the specified y column and
            // transfer them to the interleaved array.
        	while (xIter.hasNext()) {
        		GregorianCalendar date = (GregorianCalendar)xIter.next();

            	array[2 * index] =
                    julianDecimal(date, m_referenceYear);
            	double realY = ((Number)yIter.next()).doubleValue();
            	array[(2 * index) + 1] = realY;

            	index++;
        	}

            // Create a time series from the interleaved array and add it to the
            // series list.
        	TimeSeries series =
        		(TimeSeries)newSeries(column, titles[column]);
        	series.setStartDate(m_startDate);
        	series.setArray(m_seriesArrays[column]);
            if (deletedGroup != null) {
            	series.setDeletedList(
                    (List)deletedGroup.get(column));
            }
        	m_seriesList.add(series);
        }

    	// Shift the data to be referenced from the most appropriate year.
        syncToReferenceYear(findAppropriateReferenceYear());
    }

//------------------------------------------------------------------------------
/**
 * Shifts all data until they are referenced from the new reference year.
 * <p><b> Pre Conditions </b><br>
 *      The current reference year should have been set and the data
 *		should have been loaded.
 * @param newReferenceYear  The new reference year to use.
 */
    public synchronized void syncToReferenceYear (int newReferenceYear) {

        // Have each time series sync to the reference year.
        Iterator iter = m_seriesList.iterator();
        while (iter.hasNext()) {
            TimeSeries series = (TimeSeries)iter.next();
            series.syncToReferenceYear(newReferenceYear);
        }
        m_referenceYear = newReferenceYear;
    }

//------------------------------------------------------------------------------
/**
 * Sets the reference year to whatever would be most appropriate for the data
 * in the file.  Causes the data to be referenced from the following year if it
 * starts after October 1st and extends more than 90 days into the following
 * year.
 * <p><b> Pre Conditions </b><br>
 *      The reference year should have been set and the data
 *		should have been loaded.
 * @return  The reference year that was determined and used.
 */
    public int findAppropriateReferenceYear () {

        // Change the reference year if the first date is after October 1st,
        // and it goes more than 90 days into the following year.
        GregorianCalendar waterYearStart = new GregorianCalendar(
        	m_referenceYear, 9, 1);
        waterYearStart.setTimeZone(TimeZone.getTimeZone("UTC"));

        // Find the last date in the data.
        List dateList = (List)m_columnList.get(0);
        GregorianCalendar lastDate = (GregorianCalendar)dateList
            .get(dateList.size() - 1);

        // Use the year following the start date if the start date is after
        // Oct. 1 and the last date is after day 90 of the year.
        if (m_startDate.get(Calendar.DAY_OF_YEAR) >=
        	waterYearStart.get(Calendar.DAY_OF_YEAR) &&
            (lastDate.get(Calendar.DAY_OF_YEAR) > 90 &&
            lastDate.get(Calendar.YEAR) > m_startDate.get(Calendar.YEAR))) {

            return (m_referenceYear + 1);
        }
        else {
        	return m_referenceYear;
        }
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The reference year.
 */
    public int getReferenceYear () {

        // If the reference year has not been set, set it to the current year.
        if (m_referenceYear <= Integer.MIN_VALUE)
        	m_referenceYear = m_startDate.get(Calendar.YEAR);

    	return m_referenceYear;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value The reference year.
 */
    public synchronized void setReferenceYear(int value) {

        m_referenceYear = value;

        // Set each time series to the same reference year.
        Iterator iter = m_seriesList.iterator();
        while (iter.hasNext()) {
            TimeSeries series = (TimeSeries)iter.next();
            series.setReferenceYear(value);
        }
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return A copy of the start date of the series.
 */

	public GregorianCalendar getStartDate() {

    	return (GregorianCalendar)m_startDate.clone();
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param date The new start date of the series.
 */
    public synchronized void setStartDate(GregorianCalendar date) {

        m_startDate = date;

        // Set each time series to the same start date.
        Iterator iter = m_seriesList.iterator();
        while (iter.hasNext()) {
            TimeSeries series = (TimeSeries)iter.next();
            series.setStartDate(date);
        }

        // Set the reference year if it isn't already set.
        if (m_referenceYear <= Integer.MIN_VALUE)
        	m_referenceYear = m_startDate.get(Calendar.YEAR);
    }
}
