/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-04-06
 * Date started:    2000-07-06
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

///xml -- import com.sun.xml.tree.*;
import gov.usgs.gr.*;
import java.io.*;
import java.util.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import org.w3c.dom.*;
import org.xml.sax.*;

//==============================================================================
/**
 * Reads and writes instances of GrObject in a file using XML syntax.
 */
public class XMLProjectFile extends gov.usgs.gr.GrFile {

	/** The list of objects that are stored to be written out later. */
    List m_allObjects = Collections.synchronizedList(new ArrayList());
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/XmlProjectFile.gif";

//------------------------------------------------------------------------------
/**
 * Creates an empty file object.
 */
    public XMLProjectFile () {

    	super();
        m_title = "XML Project File";
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.  Adds everything with an ObjectEffigy wrapper
 * around it.
 * @param   The object to add to this file.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro isn't already the right kind of ObjectEffigy, wrap it with
        // one.
        if (gro instanceof ObjectEffigy) {
        	if (((ObjectEffigy)gro).depictProperties == false) {
        		m_children.add(gro);
            }
        	else {
        		m_children.add(new ObjectEffigy(gro));
        	}
        }
        else {
        	m_children.add(new ObjectEffigy(gro));
        }

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Reads the file from disk using the Simple API for XML parsing (SAX).
 * <p><b> Pre Conditions </b><br>
 *      The filename is not null.
 * @param filename  The name of the file to load.
 * @throws IOException
 */
    public void read () throws IOException {

//		readWithDOM();
        readWithSAX();
    }

//------------------------------------------------------------------------------
/**
 * Reads the file from disk using the Document Object Model (DOM) parser.
 * readWithSAX() is faster.
 * @throws IOException
 */
    public void readWithDOM () throws IOException {

		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        Document document = null;

		// Get a Parser and Parse the File
    	try {
			DocumentBuilder builder = factory.newDocumentBuilder();
			document = builder.parse(m_file);
            Node grNode = document.getElementsByTagName("gov.usgs.gr").item(0);
            NodeList nodeList = grNode.getChildNodes();

            // Set up lists to hold the child elements.
            GrHashList allObjects = new GrHashList();
            ArrayList unlinkedObjects = new ArrayList();

            // Read the file.
            readFromDOM(grNode, allObjects, unlinkedObjects);
		}
        catch (Exception e) {
        	e.printStackTrace();
        }
    }

//------------------------------------------------------------------------------
/**
 * Reads the file from disk using the Simple API for XML (SAX) parser.  Uses
 * the GrSAXHandler class and relies on the SAX input methods within the
 * classes that are read in.
 * @throws IOException
 */
    public void readWithSAX () throws IOException {

    	// Read the child elements from the file into a list.
        GrSAXHandler gsh = new GrSAXHandler();
        List newChildren = gsh.read(m_file.getAbsolutePath());

        // Add each element in the list as a child of this file object.
        Iterator iter = newChildren.iterator();
        while (iter.hasNext()) {
        	add((GrObject)iter.next());
        }
    }

//------------------------------------------------------------------------------
/**
 * Saves the objects and their children to a file.  Uses DOM to build the
 * XML document.
 * <p><b> Pre Conditions </b><br>
 *      The filename is not null.
 * @throws IOException
 */
    public synchronized void write() throws IOException {

    	// Have the object effigies redepict their subjects in case it wasn't
        // done previously.
        Enumeration enumeration = children();
        while (enumeration.hasMoreElements()) {
        	GrObject gro = (GrObject)enumeration.nextElement();
            if (gro instanceof ObjectEffigy) ((ObjectEffigy)gro).redepict();
        }

    	// Set the Id variables for all output objects before beginning
        // to create XML output for any of them.
        m_allObjects.clear();
        setSaveIds(this);

        // Create a new XML Document and fill it with DOM elements.
        DocumentBuilderFactory docBuilderFactory =
            DocumentBuilderFactory.newInstance();
        DocumentBuilder docBuilder = null;
        try {
            docBuilder = docBuilderFactory.newDocumentBuilder();
        }
        catch (ParserConfigurationException e) {
            e.printStackTrace();
            return;
        }
        Document doc = docBuilder.newDocument();
/// Newer XML...
///        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
///        DocumentBuilder builder = factory.newDocumentBuilder();
///        Document doc = builder.newDocument();

        // Create the top element.
        doc.appendChild(doc.createComment(
        	"USGS Gr Version " + gov.usgs.gr.Gr.getVersion()));
        Element grElement = doc.createElement("gov.usgs.gr");
        doc.appendChild(grElement);

        // Add the rest of the elements in the tree.
        enumeration = children();
        while (enumeration.hasMoreElements()) {
            GrObject gro = (GrObject)enumeration.nextElement();

            Element childElement = gro.getDOMRepresentation(doc, true);
            grElement.appendChild(childElement);
        }

        try {
// Non-zip-file code...

///        	BufferedWriter xmlbw =
///            	new BufferedWriter(new FileWriter(m_file.getAbsolutePath()));
            TransformerFactory tfac = TransformerFactory.newInstance();
            Transformer transformer = null;
            try {
                transformer = tfac.newTransformer();
            }
            catch (TransformerConfigurationException e) {
                e.printStackTrace();
                return;
            }
            DOMSource source = new DOMSource(doc);
            FileOutputStream fos = new FileOutputStream(m_file);
            StreamResult output = new StreamResult(fos);
            try {
                transformer.transform(source, output);
            }
            catch (TransformerException e) {
                e.printStackTrace();
            }
///xml -        	doc.write(xmlbw);
///            xmlbw.close();


//Code for creating a zip file with internal files...
/*			if (!(filename.toLowerCase().endsWith(".zip"))) filename += ".zip";
            ZipOutputStream zos = new ZipOutputStream(
            	new FileOutputStream(filename));
            zos.setLevel(9);
            zos.putNextEntry (new ZipEntry ("gr.xml"));
        	doc.write(zos);
            zos.close();
*/
        }
        catch (IOException e) {
        	e.printStackTrace();
        }
 	}

//------------------------------------------------------------------------------
/**
 * Sets the saveId variables appropriately on every child of the given node.
 * If the same object is found multiple times, its m_bSaveIdOnlyToXml variable
 * is set to false and its m_bSaveIdToXml is set to true.
 * All other objects will have both variables set to false.<p>
 * This method calls itself recursively.<p>
 * The purpose of this method is to set up the objects in the tree so that
 * later, on XML output, the first time the topNode object provides its XML
 * representation, it should provide the full data and then set its own
 * m_bSaveIdOnlyToXml variable to true so that only the ID version will be
 * provided for the remainder of the file.
 * @param topNode	The parent of the nodes that will be checked.
 */
    protected void setSaveIds(GrObject topNode) {

        // Each object that has been looked at is added to the m_allObjects
        // member variable list, if it wasn't already there.

        // Iterate through each child of the node.
        Enumeration enumeration = topNode.children();
        while (enumeration.hasMoreElements()) {
        	GrObject gro = (GrObject)enumeration.nextElement();
            GrObject originalGro = gro;

        	// If gro is an object effigy, get the eventual subject and try
        	// adding that.
        	while (gro instanceof gov.usgs.gr.ObjectEffigy)
        		gro = ((ObjectEffigy)gro).getSubject();

            // Look for gro in the list of all objects.
            GrObject oldGro = null;
            int oldIndex = m_allObjects.indexOf(gro);
            if (oldIndex >= 0)
            	oldGro = (GrObject)m_allObjects.get(oldIndex);

            // Modify the object's saveId variables based on whether it
            // was found already.
            if (oldGro == null) {
                gro.m_bSaveIdToXml = false;

                // Add gro to the list, because it wasn't already there.
                m_allObjects.add(gro);
            }
            else {
                gro.m_bSaveIdToXml = true;
            }

            // Reset this because it could have been changed if this
            // object was previously saved.
            gro.m_bSaveIdOnlyToXml = false;

            // Recursively call this method on the child so that its
            // children will be handled.  Use the original object,
            // in case it was an effigy that was not supposed to write
            // out all the children of it's subject.
            setSaveIds(originalGro);
        }
    }
}



