/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    1998-10-19
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.data;

import java.io.*;
import java.text.*;
import java.util.*;

//==============================================================================
/**
 * Reads and writes X-Y series data from/to a file.  The format of each line
 * is X-Y1-Y2-Y3-...-Yn.
 */
public class XyDataFile extends DimensionalDataFile {

	/** Instatiate the list for holding the data. */
    protected List<List<Number>> m_columnList = new ArrayList();

//------------------------------------------------------------------------------
/**
 * Passes 2 to the DataSet constructor as the number of dimensions for this
 * object.
 */
    public XyDataFile () {

        super(2);
        m_title = "X-Y Data";
    }

//------------------------------------------------------------------------------
/**
 * Creates a data series object to store the data.
 * @return	A new DataSeries.
 */
    protected DoubleDataSeries newSeries() {

    	return new DoubleDataSeries(2);
    }

//------------------------------------------------------------------------------
/**
 * Creates a data series object to store the data.
 * @param array The interleaved data array for the new series.
 * @return	A new DataSeries.
 */
    protected DoubleDataSeries newSeries(double[] array) {

    	return new DoubleDataSeries(array, 2);
    }

//------------------------------------------------------------------------------
/**
 * Creates an X-Y data series object to store the data.
 * @param numericType   A numeric code describing the data.
 * @param title The title of the data series.
 * @return	A new X-Y data series with no data, but numeric type and title set.
 */
    protected DoubleDataSeries newSeries(int numericType, String title) {

    	return new DoubleDataSeries(numericType, title, 2);
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.
 * Overrides readData() from gov.usgs.data.DimensionalDataFile.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 * @throws IOException
 */
    protected void readData () throws IOException {

    	double[] vals;

        // Read each token on the line into an array of doubles.
        while ((vals = AsciiIO.parseLineAsDoubles(m_br.readLine(),
        	m_inputDelimiters)) != null) {

            // Add a data list for each token that didn't already have one.
        	while (vals.length > m_columnList.size()) {
            	m_columnList.add(new ArrayList());
            }

            // Store each double token as the most recent item in its list.
            for (int i = 0; i < vals.length; i++) {
            	List column = (List)m_columnList.get(i);
                column.add(vals[i]);
            }
        }

        // Make a new array for holding interleaved data.
        int numColumns = m_columnList.size() - 1;
        m_seriesArrays = new double[numColumns][];

        // Add an interleaved x/y array for each y channel
        for (int column = 0; column < numColumns; column++) {
            m_seriesArrays[column] =
                new double[2 * ((List)m_columnList.get(column + 1)).size()];

            // Populate the interleaved array with data from the column list.
            int index = 0;
			Iterator<Number> xIter = m_columnList.get(0).iterator();
			Iterator<Number> yIter = m_columnList.get(column + 1).iterator();
            double[] array = m_seriesArrays[column];
			while (xIter.hasNext()) {
                double realX = xIter.next().doubleValue();
                double realY = yIter.next().doubleValue();

                array[2 * index] = realX;
                array[(2 * index) + 1] = realY;
                index++;
            }

            // Add the interleaved array to the data series list.
            DoubleDataSeries series =
                newSeries(column, "Data Column " + column);
            series.setArray(m_seriesArrays[column]);
            m_seriesList.add(series);
        }
    }

//------------------------------------------------------------------------------
/**
 * Write the header information to an open file attached to a writer.
 */
    private void writeHeader () {

    }

//------------------------------------------------------------------------------
/**
 * Write the time series back to a file.
 * <p><b> Pre Conditions </b><br>
 *      The header has been written with writeHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 * @throws IOException
 */
    protected void writeData () throws IOException {

        // Initialize local variables.
        String lineString;
        NumberFormat form = new DecimalFormat("0.00");

        // Create a string for each line of the file and then write it out.
        int arrayLength = ((DataSeries)m_seriesList.get(0)).getCount();
        for (int i = 0; i < arrayLength; i++) {
    		Iterator seriesIter = m_seriesList.iterator();

            // Begin the string with the x series.
            DoubleDataSeries series = (DoubleDataSeries)m_seriesList.get(0);
            double[] dataArray = series.getArray();
            lineString = AsciiIO.padLeft(form.format(dataArray[i * 2]), 10);

            // Append each of the y series.
            while (seriesIter.hasNext()) {
        		series = (DoubleDataSeries)seriesIter.next();
            	dataArray = series.getArray();
                lineString += " ";
            	lineString += AsciiIO.padLeft(form.format(
                	dataArray[i * 2 + 1]), 10);
            }
            lineString += "\n";

            // Write the string to the file.
            m_bw.write(lineString, 0, lineString.length());
        }
    }
}
