/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-06-26
 * Date started:    2006-03-13
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.nwis.adaps.rdb;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import java.lang.reflect.*;
import java.awt.EventQueue;

//==============================================================================
/**
 * A data file object for reading data in RDB format.
 */
public class RdbFile extends gov.usgs.gr.data.TimeDataFile {
    
    /** The indexes of the year, day, and time columns. */
    protected int[] m_dateColumns = {1, 2, 3};
    /** The indexes of the variables to keep. */
    protected int[] m_keepers;
    /** The numeric type of each time series. */
    protected int[] m_types;
    /** The title of each time series. */
    protected String[] m_titles;
    /** The names of the columns in the file. */
    protected String[] m_columnNames = null;
    /** Whether the file is daily values or unit values. */
    protected boolean m_bDailyValues = false;
    /** Whether there was an error during processing. */
    protected boolean m_bError = false;
    
    protected int m_dateColumn = 0;
    protected int m_timeColumn = 0;
    protected int m_tzcdColumn = 0;
    protected int m_valueColumn = 0;
    protected int m_qaColumn = 0;
    
    protected long m_numLinesRead = 0;
    
    protected boolean m_bRequireUTC = false;
    
    /** The station name, indicated in the header, if any. */
    protected String m_stationName = "Station Unspecified";
    /** The station name, indicated in the header, if any. */
    protected String m_stationNumber = "";
    
    /**
     * Whether to show a dialog when giving notification messages to
     * the user.
     */
    protected boolean m_bUseDialogNotification = true;
    /**
     * The index into the list of preConfigs that describes the
     * current config.
     */
    protected RdbOptions m_selectedConfig = new RdbOptions();
    /** A list of CR10Options objects from which to choose the input options. */
    protected List m_preConfigs = new ArrayList();
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/CR10DataFile.gif";
    
    /** An inner class for holding configuration options. */
    protected class RdbOptions extends GrObject {
        
        /** The title of the CR10 file. */
        public String title = "";
        /** The number of channels in the CR10 file. */
        public int numChannels = 0;
        /** The columns that hold the year, day, and time.  Comma delimited. */
        public String dateColumns = "1,2,3";
        /** The channels to keep when reading.  Comma delimited. */
        public String keepers = "";
        /** The data types for each keeper channel.  Comma delimited. */
        public String types = "";
        /** The title for each keeper channel.  Comma delimited. */
        public String titles = "";
        
        //----------------------------------------------------------------------
        /**
         * Provide the list of properties with the latest values.
         * @return a list of GrProperty objects containing the current values.
         */
        public GrHashList getProperties() {
            
            GrHashList props = super.getProperties();
            
            props.put("title", new GrProperty(title, "Title"));
            props.put("numChannels", new GrProperty(numChannels,
                "Total Number of Channels"));
            props.put("dateColumns", new GrProperty(dateColumns,
                "Year, Day, and Time Columns"));
            props.put("keepers", new GrProperty(keepers, "Channels to Keep"));
            props.put("types", new GrProperty(types, "Channel Types"));
            props.put("titles", new GrProperty(titles, "Channel Titles"));
            
            return props;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Create an empty data file object.
     */
    public RdbFile() {
        
        super();
        
        m_numHeaderLines = 2;
        
        m_inputDelimiters = "\t";
        
        // Set up a default options object.
        RdbOptions standardOptions = new RdbOptions();
        standardOptions.title = "Standard";
        standardOptions.numChannels = 14;
        standardOptions.dateColumns = "2,3,4";
        standardOptions.keepers = "5, 10, 11, 12";
        standardOptions.types = "1, 2, 3, 4";
        standardOptions.titles = "Vel,iStage,aStage,aVel";
        
        // Add the new object to the list of preConfigs.
        m_preConfigs.add(standardOptions);
        setConfig("Standard");
        
        // Check if any default CR10 options were set at start up from a
        // config file in the Gr/config directory.
        setDefaults("gov.usgs.sfhydro.data.CR10DataFile");
    }
    
//------------------------------------------------------------------------------
    /**
     * Looks for the preset configuration with the specified name and sets the
     * member variables to match those values.
     * @param name  The name of the preConfig to use.
     */
    public void setConfig(String name) {
        
        Iterator iter = m_preConfigs.iterator();
        while(iter.hasNext()) {
            RdbOptions config = (RdbOptions)iter.next();
            if (config.title.equals(name)) {
                m_selectedConfig = config;
                m_title = m_selectedConfig.title;
                m_dateColumns = AsciiIO.parseLineAsInts(
                    m_selectedConfig.dateColumns, ", \t");
                m_keepers = AsciiIO.parseLineAsInts(
                    m_selectedConfig.keepers, ", \t");
                m_types = AsciiIO.parseLineAsInts(
                    m_selectedConfig.types, ", \t");
                m_titles = AsciiIO.parseLine(m_selectedConfig.titles, ",");
                
                break;
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Tells whether this file options has an options dialog.
     * @return true if the showOptionsDialog() method actually does something
     *              in this file class.
     */
    public boolean hasOptionsDialog() {
        
        return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Shows a dialog for configuring this file.
     * @param frame The frame to attach the dialog box to.  Usually the frame of
     *              the window that is controlling opening the file.
     */
    public void showOptionsDialog(Frame frame) {
        
        // Create an instance of the inner dialog class that presents the
        // user with a GUI for modifying file options.
        final CR10ConfigDialog dialog = new CR10ConfigDialog(frame,
            "CR10 Config", true);
        
        // Find the index of the currently selected preConfig.
        int index = 0;
        Iterator iter = m_preConfigs.iterator();
        while (iter.hasNext()) {
            RdbOptions config = (RdbOptions)iter.next();
            if (config == m_selectedConfig) break;
            index++;
        }
        if (index >= m_preConfigs.size()) index = 0;
        
        // Show the configuration dialog.
        dialog.setConfigs(m_preConfigs, index);
        
        if (EventQueue.isDispatchThread()) {
            dialog.setVisible(true);
        } else {
            try {
                javax.swing.SwingUtilities.invokeAndWait(new Runnable() {
                    public void run() {
                        dialog.setVisible(true);
                    }
                });
            } catch (InvocationTargetException ex) {
            } catch (InterruptedException ex) {
            }
        }
        
        if (!(dialog.m_bOkPressed)) return;
        
        // Find out which preset the user chose.
        index = dialog.presetCombo.getSelectedIndex();
        RdbOptions comboConfig = (RdbOptions)m_preConfigs.get(index);
        
        // Copy the value from the dialog box.
        String dateText = dialog.dateField.getText();
        String keepText = dialog.keepField.getText();
        String typesText = dialog.typesField.getText();
        String titlesText = dialog.titlesField.getText();
        
        // Compare the values in the dialog to the currently selected preConfig.
        if (dateText.equals(comboConfig.dateColumns) &&
            keepText.equals(comboConfig.keepers) &&
            typesText.equals(comboConfig.types) &&
            titlesText.equals(comboConfig.titles)) {
            
            // The chosen preConfig was not customized, so apply it as is.
            setConfig(comboConfig.title);
        } else {
            // The user modified one of the fields, so add a
            // custom configuration and apply it.
            RdbOptions customConfig = new RdbOptions();
            customConfig.title = "Custom";
            customConfig.dateColumns = dateText;
            customConfig.keepers = keepText;
            customConfig.types = typesText;
            titlesText = titlesText.replace('(', '[');
            titlesText = titlesText.replace(')', ']');
            customConfig.titles = titlesText;
            m_preConfigs.add(customConfig);
            
            setConfig(customConfig.title);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     * @return  The ordered list of GrProperty objects.
     */
    public GrHashList getProperties() {
        
        GrHashList props = super.getProperties();
        
        // The value of each property is a comma delimited string.
        String s = "";
        for (int i = 0; i < m_dateColumns.length; i++) {
            if (i > 0) s += ", ";
            s += m_dateColumns[i];
        }
        props.put("dateColumns",
            new GrProperty(s, "Year, Day, and Time Columns"));
        
        s = "";
        for (int i = 0; i < m_keepers.length; i++) {
            if (i > 0) s += ", ";
            s += m_keepers[i];
        }
        props.put("keepers", new GrProperty(s, "Channels to Keep"));
        
        s = "";
        for (int i = 0; i < m_types.length; i++) {
            if (i > 0) s += ", ";
            s += m_types[i];
        }
        props.put("types", new GrProperty(s, "Channel Types"));
        
        s = "";
        for (int i = 0; i < m_titles.length; i++) {
            if (i > 0) s += ", ";
            s += m_titles[i];
        }
        props.put("titles", new GrProperty(s, "Channel Titles"));
        props.put("requireUTC", new GrProperty(m_bRequireUTC, "Require UTC"));
        
        return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Set properties from a list of values.
     * @param props The list of GrProperty objects.
     */
    public void setProperties(HashMap props) {
        
        super.setProperties(props);
        
        GrProperty p;
        
        // Iterate through the list of properties.
        Iterator iter = props.keySet().iterator();
        if (props instanceof GrHashList) {
            iter = ((GrHashList)props).listIterator();
        }
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            
            // Find the String value let the setProperty method handle it.
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;
                
                if (key.equals("requireUTC"))
                    m_bRequireUTC = (Boolean)p.value;

                
                if (p.value instanceof String)
                    setProperty(key, (String)p.value);
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Set a single property.
     * @param key   The name of the property being set.
     * @param value The value of the property being set.
     */
    public void setProperty(String key, String value) {
        
        super.setProperty(key, value);
        
        if (key.equals("dateColumns"))
            m_dateColumns = AsciiIO.parseLineAsInts(value, ", \t");
        if (key.equals("keepers"))
            m_keepers = AsciiIO.parseLineAsInts(value, ", \t");
        if (key.equals("types"))
            m_types = AsciiIO.parseLineAsInts(value, ", \t");
        if (key.equals("titles"))
            m_titles = AsciiIO.parseLine(value, ",");
        if (key.startsWith("preConfig") || key.startsWith("setConfig")) {
            int index1 = value.indexOf('(');
            int index2 = value.indexOf(')', index1 + 1);
            RdbOptions preset = new RdbOptions();
            preset.title = value.substring(index1 + 1, index2);
            if (preset.title.equals("CLEAR")) {
                m_preConfigs.clear();
                return;
            }
            index1 = value.indexOf('(', index2 + 1);
            index2 = value.indexOf(')', index1 + 1);
            preset.dateColumns = value.substring(index1 + 1, index2);
            index1 = value.indexOf('(', index2 + 1);
            index2 = value.indexOf(')', index1 + 1);
            preset.keepers = value.substring(index1 + 1, index2);
            index1 = value.indexOf('(', index2 + 1);
            index2 = value.indexOf(')', index1 + 1);
            preset.types = value.substring(index1 + 1, index2);
            index1 = value.indexOf('(', index2 + 1);
            index2 = value.indexOf(')', index1 + 1);
            preset.titles = value.substring(index1 + 1, index2);
            m_preConfigs.add(preset);
            
            if (key.startsWith("setConfig"))
                setConfig(preset.title);
        }
        if (key.startsWith("useDialogNotification"))
            m_bUseDialogNotification = Boolean.parseBoolean(value);
        if (key.equals("requireUTC"))
            m_bRequireUTC = Boolean.parseBoolean(value);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
        
        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();
        
        return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {
        
        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);
        
        if (icon == null)
            icon = super.setTreeIcon();
        
        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);
        
        return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Process a single line of the header.
     */
    protected void processHeaderComment(String lineString) {
        
        lineString = lineString.substring(4);
        String upString = lineString.toUpperCase();
        upString = upString.replace('-', ' ');
        if (upString.startsWith("FILE TYPE") &&
            upString.contains("DAILY VALUES")) {
            
            m_bDailyValues = true;
        }
        if (lineString.contains("STATION NAME=\"")) {
            try {
                int index0 = lineString.indexOf("STATION NAME=\"");
                int index1 = index0 + 14;
                int index2 = lineString.indexOf("\"", index1);
                
                m_stationName = lineString.substring(index1, index2).trim();
            }
            catch (Exception e) {
                m_stationName = "Station ???";
            }
        }
        if (lineString.contains("NUMBER=\"")) {
            try {
                int index0 = lineString.indexOf("NUMBER=\"");
                int index1 = index0 + 8;
                int index2 = lineString.indexOf("\"", index1);
                
                m_stationNumber = lineString.substring(index1, index2).trim();
            }
            catch (Exception e) {
                m_stationNumber = "";
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Read the header information from an open file attached to a reader.
     * Called by read().
     * @throws IOException
     * @see gov.usgs.gr.data.DelimitedDataFile#setHeaderLines setHeaderLines()
     */
    protected void readHeader() throws IOException {
        
        m_numLinesRead = 0;
        
        while (m_headerLines.size() < m_numHeaderLines) {
            String lineString = m_br.readLine();
            m_numLinesRead++;
            
            if (lineString.startsWith("# //")) {
                processHeaderComment(lineString);
                continue;
            }
            
            m_headerLines.add(lineString);
        }
        
        String[] m_columnNames =
            AsciiIO.parseLine((String)m_headerLines.get(0), m_inputDelimiters);
        
        for (int i = 0; i < m_columnNames.length; i++) {
            if (m_columnNames[i].equals("DATE")) {
                m_dateColumn = i;
                continue;
            }
            if (m_columnNames[i].equals("TIME")) {
                m_timeColumn = i;
                continue;
            }
            if (m_columnNames[i].equals("TZCD")) {
                m_tzcdColumn = i;
                continue;
            }
            if (m_columnNames[i].equals("VALUE")) {
                m_valueColumn = i;
                continue;
            }
            if (m_columnNames[i].equals("QA")) {
                m_qaColumn = i;
                continue;
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Read the data through a reader attached to an open file.
     * <p><b> Pre Conditions </b><br>
     *      The header has been read with readHeader()
     * <p><b> Post Conditions </b><br>
     *      The file needs to be closed.
     * @param br    The buffered reader object to use for reading.
     * @throws IOException When file I/O errors occur.
     */
    protected void readData() throws IOException {
        
        ArrayList dateList = new ArrayList();
        ArrayList valueList = new ArrayList();
        
        int year = 0;
        int month = 0;
        int day = 0;
        
        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        
        String tzString = null;
        
        // Read each line of the data file.
        String lineString = null;
        while ((lineString = m_br.readLine()) != null) {
            m_numLinesRead++;
            int fieldIndex = 0;
            int tabIndex = 0;
            int startIndex = 0;
            
            int hour = 12;
            int minute = 0;
            int second = 0;
            
            while (startIndex < lineString.length()) {
                tabIndex = lineString.indexOf('\t', startIndex);
                if (tabIndex < 0)
                    tabIndex = lineString.length();
                
                String token = lineString.substring(startIndex, tabIndex);
                
                if (fieldIndex == m_dateColumn) {
                    year = Integer.parseInt(token.substring(0, 4));
                    month = Integer.parseInt(token.substring(4, 6));
                    day = Integer.parseInt(token.substring(6, 8));
                }
                if (fieldIndex == m_timeColumn) {
                    if (token.length() >= 4)
                        hour = Integer.parseInt(token.substring(0, 2));
                    if (token.length() >= 6)
                        minute = Integer.parseInt(token.substring(2, 4));
                    if (token.length() >= 8)
                        second = Integer.parseInt(token.substring(4, 6));
                }
                if (!m_bDailyValues && fieldIndex == m_tzcdColumn) {
                    if (tzString == null)
                        tzString = token;
                    else {
                        if (!token.equals(tzString)) {
                            System.err.println(
                                "Multiple time zones encountered (" + tzString
                                + " and " + token + ") at line "
                                + m_numLinesRead);
                        
                        m_bError = true;
                        
                        return;

                        }
                    }
                    if (m_bRequireUTC && !token.equals("UTC")) {
                        System.err.println(
                            "Non-UTC time zone encountered at line "
                            + m_numLinesRead);
                        
                        m_bError = true;
                        
                        return;
                    }
                }
                if (fieldIndex == m_valueColumn) {
                    double dVal = Double.NaN;
                    try {
                        dVal = Double.parseDouble(token);
                    } catch (NumberFormatException ex) {}
                    
                    valueList.add(dVal);
                }
                if (fieldIndex == m_qaColumn) {
                    // Store QA value if needed
                }
                
                startIndex = tabIndex + 1;
                fieldIndex++;
            }
            
            // Create a single date object from the data.
            GregorianCalendar date = new GregorianCalendar(utcZone, locale);
            date.clear();
            date.set(year, month - 1, day, hour, minute, second);
            
            // Add the date to the list.
            dateList.add(date);
        }
        
        
        // Create a list containing a column of dates and columns for the other
        // data so they can be interleaved into individual series.
        m_startDate = (GregorianCalendar)
        ((GregorianCalendar)dateList.get(0)).clone();
        GregorianCalendar lastDate = (GregorianCalendar)
        ((GregorianCalendar)dateList.get(dateList.size() - 1)).clone();
        m_referenceYear = findAppropriateReferenceYear(m_startDate, lastDate);
        
        ArrayList inputList = new ArrayList();
        inputList.add(dateList);
        inputList.add(valueList);
        
        // Interleave the arrays.
        List<? extends TimeSeries> timeSeriesList =
            TimeDataFile.interleaveArrays(inputList);
        // Add the time series as children of this file.
        int index = 0;
        Iterator iter = timeSeriesList.iterator();
        while (iter.hasNext()) {
            TimeSeries series = (TimeSeries)iter.next();
//            series.setType(m_types[index]);
//            series.setTitle(m_titles[index]);
            double[] min = {Double.NEGATIVE_INFINITY, Double.NaN};
            double[] max = {Double.POSITIVE_INFINITY, Double.NaN};
            series.selectArea(min, max, true);
            series.deleteSelectedPoints();
            
            // Add the series as a child of this file.
            add(series);
            
            index++;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Override the default output.
     * @throws IOException
     */
    protected void writeHeader() throws IOException {
    }
    
//------------------------------------------------------------------------------
/**
 * Shows a dialog for configuring the file.
 */
    public static boolean matchesFormat(String[] lines) {
        
        int numLinesRead = 0;
        boolean bNwis = false;
        
        while (numLinesRead < lines.length && lines[numLinesRead] != null &&
            lines[numLinesRead].startsWith("# //")) {
            
            if (lines[numLinesRead].startsWith("# //"))
                if (lines[numLinesRead].indexOf(
                    "NATIONAL WATER INFORMATION SYSTEM") >= 0) {
                
                    bNwis = true;
                }
                
            numLinesRead++;
        }
        
        if (!bNwis)
            return false;
        
        if (lines[numLinesRead].indexOf("DATE") < 0)
            return false;
    
        return true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of menu items that execute commands on this object.
     * @return  A menu item that can be placed on a menu and send an ActionEvent
     *          back to this object when it is chosen.
     */
    public javax.swing.JMenuItem getMenuCommands() {
        
        // Create the menu item as a sub menu with other items.
        javax.swing.JMenu subMenu = new javax.swing.JMenu("CR10 Data File");
        
        // Set up menu items that use this object as their action listeners.
        javax.swing.JMenuItem saveAsItem = new javax.swing.JMenuItem("Save As");
        saveAsItem.addActionListener(this);
        subMenu.add(saveAsItem);
        javax.swing.JMenuItem resamp15Item = new javax.swing.JMenuItem(
            "Eliminate Non-15-Minute Data");
        resamp15Item.addActionListener(this);
        subMenu.add(resamp15Item);
        
        return subMenu;
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {
        
        String s = "RDB File (" + m_selectedConfig.title + ")";
        
        return (s + ": " + m_filename);
    }
 
//------------------------------------------------------------------------------
    /**
     * Gets the error status.
     * @return Whether there was an error during processing.
     */
    public boolean getError() {
        
        return m_bError;
    }
 
//------------------------------------------------------------------------------
    /**
     * Gets the station name.
     * @return The station name.
     */
    public String getStationName() {
        
        return m_stationName;
    }
 
//------------------------------------------------------------------------------
    /**
     * Gets the station number.
     * @return The station number.
     */
    public String getStationNumber() {
        
        return m_stationNumber;
    }
    
//------------------------------------------------------------------------------
    /**
     * Resets the error state to false.
     */
    public void resetError() {
        
        m_bError = false;
    }
    
//------------------------------------------------------------------------------
    /**
     * A dialog that allows the user to configure input options.
     */
    public class CR10ConfigDialog extends JDialog {
        
        // Initialize GUI components.
        public boolean m_bOkPressed = false;
        public List m_preConfigs = new ArrayList();
        
        private JComboBox presetCombo = new JComboBox();
        private JTextField dateField = new JTextField();
        private JTextField keepField = new JTextField();
        private JTextField typesField = new JTextField();
        private JTextField titlesField = new JTextField();
        
        
        //------------------------------------------------------------------------------
        /**
         * Create a new dialog using the give frame, title, and modal setting.
         * @param frame The frame to which this dialog should be attached.
         * @param title The title of the dialog.
         * @param modal Whether the dialog should be modal (blocks input to other
         *              windows until closed.)
         */
        public CR10ConfigDialog(Frame frame, String title, boolean modal) {
            
            super(frame, title, modal);
            
            try  {
                jbInit();
                pack();
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
        
        //------------------------------------------------------------------------------
        /**
         * Create a dialog with no title and attached to no frame.
         */
        public CR10ConfigDialog() {
            this(null, "", false);
        }
        
        //------------------------------------------------------------------------------
        /**
         * Set up the dialog and it's components.
         * @throws Exception
         */
        void jbInit() throws Exception {
            
            // Declare and initialize GUI components.
            JPanel mainPanel = new JPanel();
            JPanel topPanel = new JPanel();
            JPanel lowerPanel = new JPanel();
            JPanel channelPanel = new JPanel();
            JButton okButton = new JButton("OK");
            JButton cancelButton = new JButton("Cancel");
            JPanel presetPanel = new JPanel();
            JPanel datePanel = new JPanel();
            JPanel keepPanel = new JPanel();
            JPanel typesPanel = new JPanel();
            JPanel titlesPanel = new JPanel();
            
            // Set up button listeners.
            okButton.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    okButton_actionPerformed(e);
                }
            });
            cancelButton.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    cancelButton_actionPerformed(e);
                }
            });
            presetCombo.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    presetCombo_actionPerformed(e);
                }
            });
            
            // Set the layout managers for the panels and insert components.
            mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
            mainPanel.add(Box.createRigidArea(new Dimension(350,5)));
            
            topPanel.setLayout(new BoxLayout(topPanel, BoxLayout.Y_AXIS));
            
            presetPanel.setLayout(new BoxLayout(presetPanel, BoxLayout.X_AXIS));
            presetPanel.add(new JLabel("Preset"));
            presetPanel.add(Box.createRigidArea(new Dimension(5,0)));
            
            presetPanel.add(presetCombo);
            presetPanel.add(Box.createRigidArea(new Dimension(0,5)));
            
            mainPanel.add(presetPanel);
            mainPanel.add(Box.createRigidArea(new Dimension(0,5)));
            
            channelPanel.setBorder(BorderFactory.createTitledBorder(
                "Channel Configuration"));
            channelPanel.setLayout(new BoxLayout(channelPanel,
                BoxLayout.Y_AXIS));
            
            datePanel.setLayout(new BoxLayout(datePanel, BoxLayout.X_AXIS));
            datePanel.add(new JLabel("Year, Day, Time"));
            datePanel.add(Box.createRigidArea(new Dimension(5,0)));
            datePanel.add(Box.createHorizontalGlue());
            datePanel.add(dateField);
            channelPanel.add(datePanel);
            channelPanel.add(Box.createRigidArea(new Dimension(0,5)));
            
            keepPanel.setLayout(new BoxLayout(keepPanel, BoxLayout.X_AXIS));
            keepPanel.add(new JLabel("Keep   "));
            keepPanel.add(Box.createRigidArea(new Dimension(5,0)));
            keepPanel.add(Box.createHorizontalGlue());
            keepPanel.add(keepField);
            channelPanel.add(keepPanel);
            channelPanel.add(Box.createRigidArea(new Dimension(0,5)));
            
            typesPanel.setLayout(new BoxLayout(typesPanel, BoxLayout.X_AXIS));
            typesPanel.add(new JLabel("Types "));
            typesPanel.add(Box.createRigidArea(new Dimension(5,0)));
            typesPanel.add(Box.createHorizontalGlue());
            typesPanel.add(typesField);
            channelPanel.add(typesPanel);
            channelPanel.add(Box.createRigidArea(new Dimension(0,5)));
            
            titlesPanel.setLayout(new BoxLayout(titlesPanel, BoxLayout.X_AXIS));
            titlesPanel.add(new JLabel("Titles "));
            titlesPanel.add(Box.createRigidArea(new Dimension(5,0)));
            titlesPanel.add(Box.createHorizontalGlue());
            titlesPanel.add(titlesField);
            channelPanel.add(titlesPanel);
            channelPanel.add(Box.createRigidArea(new Dimension(0,5)));
            
            mainPanel.add(channelPanel);
            mainPanel.add(Box.createRigidArea(new Dimension(0,5)));
            lowerPanel.setLayout(new BoxLayout(lowerPanel, BoxLayout.X_AXIS));
            
            lowerPanel.add(okButton);
            lowerPanel.add(Box.createRigidArea(new Dimension(5,0)));
            lowerPanel.add(cancelButton);
            
            mainPanel.add(lowerPanel);
            
            getContentPane().add(mainPanel);
        }
        
        //------------------------------------------------------------------------------
        /**
         * Set the list of preConfigs that will be offered to the user.
         * @param list  The list of preConfigs to use.
         * @param index The list index of the currently selected preConfig.
         */
        void setConfigs(List list, int index) {
            
            m_preConfigs = list;
            
            // Fill the preset configurations combo box.
            Iterator iter = m_preConfigs.iterator();
            while (iter.hasNext()) {
                RdbOptions config = (RdbOptions)iter.next();
                presetCombo.addItem(config.title);
            }
            
            presetCombo.setSelectedIndex(index);
        }
        
        //--------------------------------------------------------------------------
        /**
         * Close the window and set a flag indicating the OK button was pressed.
         * @param e The action event to be handled.
         */
        void okButton_actionPerformed(ActionEvent e) {
            
            m_bOkPressed = true;
            dispose();
        }
        
        //--------------------------------------------------------------------------
        /**
         * Close the window and set a flag indicating the Cancel button was pressed.
         * @param e The action event to be handled.
         */
        void cancelButton_actionPerformed(ActionEvent e) {
            
            m_bOkPressed = false;
            dispose();
        }
        
        //--------------------------------------------------------------------------
        /**
         * Change the information displayed in the text fields because the user
         * chose a different preConfig in the presetCombo.
         * @param e The action event to be handled.
         */
        void presetCombo_actionPerformed(ActionEvent e) {
            
            // Find which presetCombo item was selected.
            int index = presetCombo.getSelectedIndex();
            if (index < m_preConfigs.size()) {
                // Copy the values of the newly selected preConfig into
                // the text fields so the user can see and edit them.
                RdbOptions config = (RdbOptions)m_preConfigs.get(index);
                dateField.setText(config.dateColumns);
                keepField.setText(config.keepers);
                typesField.setText(config.types);
                titlesField.setText(config.titles);
            }
        }
    }
}

