/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    2002-07-25
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.tools;

import java.util.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;

//==============================================================================
/**
 * Combines DataDimension objects in any order from any number of DataSeries
 * into a new DataSeries.  Dimension objects should be added, using add(),
 * as children of this object in the order in which they will be combined.
 * Combination is perfomed by createCombinationSeries() and combine();
 */

public class DimensionOrganizer extends GrObject implements
    java.awt.event.ActionListener {

    /** Whether to add an index counter as the first dimension of the
     *  the generated series.
     */
    boolean m_bAddIndexDimension = false;
    /** The series that results from combining the dimensions. */
    protected DataSeries m_series = null;

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public DimensionOrganizer () {

        // Check for default configuration settings.
        setDefaults("gov.usgs.gr.tools.DimensionOrganizer");
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * addIndexDim     Boolean  "Add Index Dimension"  <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("addIndexDim", new GrProperty(
            m_bAddIndexDimension, "Add Index Dimension"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.gr.tools.DimensionOrganizer#getProperties getProperties() for
 * a list of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("addIndexDim"))
            		m_bAddIndexDimension = (Boolean)p.value;
            }
            // For each property with a string value, pass it to the
            // setProperty() method to be handled there.
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.gr.data.DimensionOrganizer#getProperties getProperties() for
 * a list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("addIndexDim"))
            m_bAddIndexDimension = Boolean.parseBoolean(value);
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

        String s = "Dimension Organizer";

        return s;
     }

//------------------------------------------------------------------------------
/**
 * Adds DataDimension and DataSeries objects.  If the parameter is an
 * ObjectEffigy object, its subject will be added if it is a DataDimension or
 * DataSeries.  Only one DataSeries child is allowed at a time, so adding a
 * DataSeries replaces the old one.
 * @param gro   The GrObject to be added.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        // Don't add the object if it is already a child of this object.
        if (m_children.contains(gro)) return false;

        boolean bAdded = false;

        // Add DataDimensions.
        if (gro instanceof DataDimension) {
            m_children.add(gro);
            bAdded = true;
        }

        // If gro is a DataSeries, use it to replace the current series.
        if (gro instanceof DataSeries) {
            remove(m_series);
            super.add(gro);
            m_series = (DataSeries)gro;
            bAdded = false;
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * Menu items include: <p>
 * "Combine into Data Series"  <p>
 * @return A menu full of commands applicable to this class.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"Dimension Organizer");
        javax.swing.JMenuItem saveItem = new javax.swing.JMenuItem(
            "Combine into Data Series");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
    	return subMenu;
    }

//------------------------------------------------------------------------------
/**
 * Act on menu items being selected.
 * @param e The event to act on.
 */
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Combine into Data Series")) {
        	createCombinationSeries();
        }
    }

//------------------------------------------------------------------------------
/**
 * Combines the child dimensions into a new DataSeries and adds it as a child.
 * Calls the combine() method.
 * <p><b> Pre Conditions </b><br>
 *      Dimensions to be combined have been added as children of this object
 *      in the order they will appear in the new series.
 *      Since data is stored in DataSeries and not DataDimension, only
 *      DataDimension objects with an observing DataSeries can be combined.
 *      Otherwise, the data for the dimension can't be determined.
 * @return  The new DataSeries resulting from the combining action.
 */
    public DataSeries createCombinationSeries () {

        // Put each child dimension that has an observing DataSeries
        // into a hash list along with it's observing DataSeries.
        // Count the minimum number of points.
        GrHashList validDims = new GrHashList();
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof DataDimension) {
                DataSeries associatedSeries =
                    ((DataDimension)gro).getAssociatedSeries();
                if (associatedSeries != null) {
                    validDims.put(gro, associatedSeries);
                }
            }
        }

        // Return if no valid dimensions were found.
        if (validDims.size() == 0)
           return null;

        // Perform the combination and add the resulting series.
        DataSeries newSeries = combine(validDims, m_bAddIndexDimension);
        add(newSeries);

        return newSeries;
    }

//------------------------------------------------------------------------------
/**
 * Combines a list of dimensions into a new DataSeries.  The data for the
 * dimensions will be arranged in the new series in the same order as they
 * appear in the list.  The resulting series will be a TimeSeries if the first
 * (X) Dimension comes from a TimeSeries.
 * @param validDims     A list of dimensions that each has a DataSeries
 *                      observer.
 * @param bAddIndexDimension    Whether to prepend an additional dimension whose
 *                              values are indexes counting upward from zero.
 * @return  The new DataSeries resulting from the combining action.
 */
    public static DataSeries combine (GrHashList validDims,
        boolean bAddIndexDimension) {

        // Initialize local variables.
        int numPoints = Integer.MAX_VALUE;

        // Check the number of points in each series and use the longest as
        // the length of the new series.
        Iterator iter = validDims.listIterator();
        while (iter.hasNext()) {
            DataDimension dim = (DataDimension)iter.next();
            DataSeries associatedSeries = dim.getAssociatedSeries();
            if (associatedSeries != null) {
                int seriesNumPoints = associatedSeries.getNumPoints();
                if (seriesNumPoints < numPoints)
                    numPoints = seriesNumPoints;
            }
        }

        // Add a dimension at the front of the list with values equal to their
        // indexes in the array.
        if (bAddIndexDimension) {
            // Create an array containing advancing index numbers.
            double[] indexArray = new double[numPoints];
            for (int i = 0; i < numPoints; i++) {
                indexArray[i] = i;
            }

            // Store the data in a series and put the 0 dimension from the
            // series at the beginning of the list of dimensions to combine.
            DataSeries indexSeries = new DoubleDataSeries(indexArray, 1);
            DataDimension indexDim = indexSeries.getDimension(0);
            validDims.add(0, indexDim, indexSeries);
        }

        // Create the new DataSeries.  Use a TimeSeries if the first (X)
	// Dimension comes from a TimeSeries.
        int newNumDims = validDims.size();
        DoubleDataSeries newSeries = null;
	DataSeries xSeries = (DataSeries)validDims.getValue(0);
	if (xSeries instanceof TimeSeries) {
	    newSeries = new TimeSeries(newNumDims);
	    newSeries.setProperties(xSeries.getProperties());
	}
        else {
            newSeries = new DoubleDataSeries(newNumDims);
        }

        // Combine the data from each dimension into a new DataSeries.
        double[] newArray = new double[newNumDims * numPoints];
        for (int j = 0; j < newNumDims; j++) {
            DataDimension dim = (DataDimension)validDims.getKey(j);
            int dimNum = dim.getNum();
            DoubleDataSeries series = (DoubleDataSeries)validDims.getValue(j);
            int numDims = series.getNumDimensions();
            double[] array = series.getArray();
            for (int i = 0; i < numPoints; i++) {
                newArray[i * newNumDims + j] = array[i * numDims + dimNum];
            }

            // Transfer properties from the old dimension to the new.
            GrHashList props = dim.getProperties();
            props.remove("num");
            newSeries.getDimension(j).setProperties(props);
        }

        // Add the new combination series as a child of this Organizer.
        newSeries.setArray(newArray);
        newSeries.setTitle("New Series");

        return newSeries;
    }
}


