/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-09-14
 * Date started:    2000-02-08
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.tools;

import java.awt.geom.*;
import java.util.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;

//==============================================================================
/**
 * Reduces the number of points in a two-dimensional DataSeries object so that
 * it looks the same when plotted at a given scale and using a given line width.
 * Points are not added or moved.  Extra points along each straight line
 * segment are eliminated.  The algorithm is accurate to within twice the value
 * of the given width parameter.
 */
public class LinePointReducer extends GrObject implements
	gov.usgs.gr.tools.DataManipulator, java.awt.event.ActionListener  {

	/** The scale factor to use in the X direction. */
    protected double m_xScale = 1;
    /** The scale factor to use in the Y direction. */
    protected double m_yScale = 1;
    /**
     * The distance that points may vary from the new line in order to be
     * eliminated.
     */
    protected double m_width = 1;
    /** Temporary array storage that can be used between methods. */
    protected double[] m_tempArray = null;
    /** The array that is being thinned. */
    protected double[] m_array = null;
    /** The number of dimensions of the series. */
    protected int m_numDimensions = 2;
    /** The number of points kept in the thinned array. */
    protected int m_numKept = 0;
    /** Keeps track of the last index plotted. */
    protected int m_lastPlottedIndex = -1;
    /** Keeps track of the distance from the last point plotted. */
    protected double m_lastPlottedDist = 0;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/LinePointReducer.gif";

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public LinePointReducer () {

        // Check for default configuration settings.
        setDefaults("gov.usgs.gr.tools.LinePointReducer");
    }

//------------------------------------------------------------------------------
/**
 * Creates an empty object and sets the scales and widths to the specified
 * values.
 * @param xScale    The scale factor to use in the X direction.
 * @param yScale    The scale factor to use in the Y direction.
 * @param width     The distance that points may vary from the new line in
 *                  order to be eliminated.
 */
    public LinePointReducer (double xScale, double yScale, double width) {

        // Check for default configuration settings.
        setDefaults("gov.usgs.gr.tools.LinePointReducer");

    	m_xScale = xScale;
        m_yScale = yScale;
        m_width = width;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * xScale   Double  "X Scale"  <p>
 * yScale   Double  "Y Scale"  <p>
 * width    Double  Line Width"  <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("xScale", new GrProperty(m_xScale, "X Scale"));
        props.put("yScale", new GrProperty(m_yScale, "Y Scale"));
        props.put("width", new GrProperty(m_width, "Line Width"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.gr.tools.LinePointReducer#getProperties getProperties() for
 * a list of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("xScale"))
            		m_xScale = (Double)p.value;
            	if (key.equals("yScale"))
            		m_yScale = (Double)p.value;
            	if (key.equals("width"))
            		m_width = (Double)p.value;
            }
            // For each property with a string value, pass it to the
            // setProperty() method to be handled there.
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.gr.data.LinePointReducer#getProperties getProperties() for
 * a list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("xScale"))
            m_xScale = Double.parseDouble(value);
        if (key.equals("yScale"))
            m_yScale = Double.parseDouble(value);
        if (key.equals("width"))
            m_width = Double.parseDouble(value);
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

        String s = "Line Point Reducer";

        return s;
     }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Adds objects as children, but only if they are instances of DataSeries.
 * Objects are traversed to look for decendents that are DataSeries so those
 * series can be added.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // Don't proceed if gro is already a child of this object.
        if (m_children.contains(gro)) return false;

        boolean bAdded = false;

        // Add gro if it is a two-dimensional DataSeries.
        if (gro instanceof DataSeries) {
            DataSeries ds = (DataSeries)gro;

            // Only add 2D series.
            if (ds.getNumDimensions() == 2) {
                m_children.add(ds);
                bAdded = true;
            }
        }
        else {
            // Try adding all of the descendants in case they hold a
            // data series.  This allows curves to be traversed and
            // have their data series added.
            Enumeration enumeration = gro.children();
            while (enumeration.hasMoreElements()) {
                if (add((GrObject)enumeration.nextElement()))
                    bAdded = true;
            }
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Uses a smart point reduction algorithm to thin out the array.  The main array
 * for this series is not modified, but is the starting point for the algorithm.
 * The algorithm considers the line that would be drawn through the points on
 * a graph.  If several consecutive points lie within the width of the line,
 * only the endpoints of the line will be added to the thin array.
 */
    public void recalculate () {

        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries)iter.next();
            synchronized (series) {
                m_array = series.getArray();
                double[] newArray = getThinBuffer();
                series.setArray(newArray);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Creates an instance of LinePointReducer and adds the series objects from
 * dataList to it and calls its recalculate() method.
 * <p><b> Pre Conditions </b><br>
 *      The list contains DoubleDataSeries objects with xDim 0 and yDim 1.
 * @param dataList  A list of DoubleDataSeries objects.
 */
    public void performAction (List dataList) {

        // Create a clone of this object but with children from the dataList.
        LinePointReducer lpr = new LinePointReducer();
        lpr.setProperties(getProperties());
        Iterator iter = dataList.iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries)iter.next();
            lpr.add(series);
        }

        lpr.recalculate();
    }

//------------------------------------------------------------------------------
/**
 * Creates a new array that is a subset of the data in the current series,
 * the data for which is stored in m_array.
 * @return The new array.
 */
    protected double[] getThinBuffer() {

        // Proceed only if there is at least on X-Y point in the array.
        if (m_array.length < 2) return new double[0];

    	// Create a temporary array with enough room to potentially hold a
        // copy of all points in the main array.
        m_tempArray = new double[
            m_array.length + m_numDimensions];

        // Reset place holders.
        int p0Index = 0;
        m_numKept = 0;

        // Always keep the first point.
        keepPoint(0, 0);

        // Find the indexes of all the successive kept points.
        while (p0Index < m_array.length) {
        	p0Index = handleNewP0(p0Index);
        }

        // Always keep the last point.
        int lastIndex = m_array.length - m_numDimensions;
        if (m_lastPlottedIndex < lastIndex) {
        	keepPoint(lastIndex, 0);
        }

        // Make a new array that is only as long as the new set of points
        // and copy the points into it.
        double[] newArray = new double[m_numKept * m_numDimensions];
        for (int i = 0; i < (m_numKept * m_numDimensions); i++) {
        	newArray[i] = m_tempArray[i];
        }

        // Write a message to the console telling how many points were removed.
        System.out.println('\n'+ "Thinned array from " + (m_array.length / 2)
            + " to " + (newArray.length / 2) + " points using width of "
            + m_width);

    	return newArray;
    }

//------------------------------------------------------------------------------
/**
 * Applies the point reduction algorithm starting from some point P0 in the
 * array.  This method returns when it reaches the next point that should be
 * considered P0.  Any points that should be kept in between are added to the
 * member array m_tempArray.
 * @param p0Index	The index to use for P0.
 * @return The index of the next P0.
 */
    protected int handleNewP0(int p0Index) {

    	// Check that p0Index is within the bounds of the array and return
        // the index of the last point if it is not.
        if ((p0Index < 0) ||
        	(p0Index > (m_array.length - (2 * m_numDimensions)))) {

        	return m_array.length;
        }

        // P0 is always a point that has already been added to the new set
        // of points.  Get the x and y coordinates of the P0 point.
        double x_0 = 0;
        double y_0 = 0;
        int negativeFarthest = p0Index;
        double negFarDist = 0;

		// P1 is always the point following P0 in the original set of points.
		// Save the current coordinates for later comparison.
        int p1Index = p0Index + m_numDimensions;
        int positiveFarthest = p1Index;
        double x_1 = m_xScale * (m_array[p1Index] - m_array[p0Index]);
        double y_1 = m_yScale * (m_array[p1Index + 1] - m_array[p0Index + 1]);

        // Calculate the legs (a, b, and c) of the right triangle
        // through points P0 and P1.  c is the distance from P0 to P1.
        double a_0 = x_1 - x_0;
        double b_0 = y_1 - y_0;
        double c_z = Math.sqrt(a_0 * a_0 + b_0 * b_0);
        double posFarDist = c_z;

        // Set whether travelling along the line from P0 to P1
        // increases or decreases with X.
        boolean posIncreaseWithX = false;
        if (x_1 > x_0) posIncreaseWithX = true;

        // Assign the slope, after checking if it is undefined.
        // m will be used for all subsequent points until P0 is reassigned.
        double m = 0;
        if (a_0 != 0) m = b_0 / a_0;

        int prevIndex = p1Index;
        double prevDist = posFarDist;
        
        // Determine the NaN status of P0 and P1.
        boolean bP0NaN = Double.isNaN(m_array[p0Index]) ||
            Double.isNaN(m_array[p0Index + 1]);
        boolean bP1NaN = Double.isNaN(m_array[p1Index]) ||
            Double.isNaN(m_array[p1Index + 1]);
        boolean bPnNaN = false;

        int pnIndex = p1Index + m_numDimensions;
        while (pnIndex < m_array.length) {
            // Handle if NaN values.
            bPnNaN = Double.isNaN(m_array[pnIndex]) ||
                Double.isNaN(m_array[pnIndex + 1]);
            
            if (!bP0NaN && !bP1NaN && bPnNaN) {
                break;
            }
            if (!bP0NaN && bP1NaN) {
                posFarDist = Double.POSITIVE_INFINITY;
                break;
            }
            if (bP0NaN && bP1NaN && bPnNaN) {
                prevIndex = pnIndex;
                pnIndex += m_numDimensions;
                continue;
            }
            if (bP0NaN && bP1NaN && !bPnNaN) {
                positiveFarthest = pnIndex;
                posFarDist = Double.POSITIVE_INFINITY;
                break;
            }
            if (bP0NaN && !bP1NaN) {
                posFarDist = Double.POSITIVE_INFINITY;
                break;
            }
            
            // This point is Pn, where n is greater than 1 and measured
            // from the most recently assigned P0.
			// Get the x and y coordinates of the current point.
        	double x_n = m_xScale * (m_array[pnIndex] - m_array[p0Index]);
        	double y_n = m_yScale *
            	(m_array[pnIndex + 1] - m_array[p0Index + 1]);

			// Find the line through this point that is perpendicular to the
            // line through P0 and P1.  The point Pz is the intersection of
            // the two lines.  It is the point along the line P0-P1 that is
            // closest to this point.
        	Point2D.Double p_z = new Point2D.Double();
            if (a_0 != 0) {
				p_z = getPerpendicularIntersection(m, x_0, y_0, x_n, y_n);
			}
            else {
            	// Special case of a vertical line.
                p_z.x = x_0;
                p_z.y = y_n;
            }

            // Calculate the legs (a, b, and c) of the right triangle
            // through points P0 and Pz.  c_z is the distance from P0 to Pz.
            double a_z = p_z.x - x_0;
            double b_z = p_z.y - y_0;
            c_z = Math.sqrt(a_z * a_z + b_z * b_z);

            // Calculate the legs (a, b, and c) of the right triangle
            // through points Pn and Pz.  c_n is the distance from Pn to Pz.
            double b_n = p_z.x - x_n;
            double a_n = p_z.y - y_n;
            double c_n = Math.sqrt(a_n * a_n + b_n * b_n);

			if (c_n > m_width) {
            	break;
            }
            else {
            	// Check if this point is a farthest point.
				// Determine whether it is in the positive direction.
            	if ((a_z > 0) == posIncreaseWithX) {
            		if (c_z > posFarDist) {
                		positiveFarthest = pnIndex;
                        posFarDist = c_z;
					}
				}
				else {
					if (c_z > negFarDist) {
            			negativeFarthest = pnIndex;
                		negFarDist = c_z;
					}
				}
			}

            // Advance to the next point and iteration of the loop.
            prevDist = c_z;
            prevIndex = pnIndex;
            pnIndex += m_numDimensions;
        }

		// Consider the points farthest along the line P0-P1 in the positive
        // direction and in the negative direction for inclusion in the new
        // set of points.  This takes care of lines that double back on
        // themselves one or more times within the width of the line.

        // To maintain the order of points, first check whether the farthest
        // point in the positive direction comes after the farthest point in
        // the negative direction.
        if (positiveFarthest < negativeFarthest) {
        	// If the farthest positive point is sufficiently far from P0,
            // add it to the new set.
            if (posFarDist > m_width)
            	keepPoint(positiveFarthest, posFarDist);

			// If the farthest negative point is sufficiently far from the
            // last plotted point, add it to the new set.
        	if ((m_lastPlottedDist + negFarDist) > m_width)
				keepPoint(negativeFarthest, negFarDist);
        }
        else {
			// If the farthest negative point is sufficiently far from P0,
            // add it to the new set.
            if (negFarDist > m_width)
            	keepPoint(negativeFarthest, negFarDist);

            // If the farthest positive point is sufficiently far from the
            // last plotted point, add it to the new set.
            if ((posFarDist + m_lastPlottedDist) > m_width)
				keepPoint(positiveFarthest, posFarDist);
		}

		// Check if the previous point was already plotted as one of the
        // farthest points. If it was not, and the current point is not past
        // the end of the array, consider plotting the previous point.
        if ((m_lastPlottedIndex < prevIndex) && (pnIndex < m_array.length)) {
			// Calculate the legs (a, b, and c) of the right triangle
            // through points Pn and P(n-1).  c_p is the distance
            // from Pn to P(n-1).
            double a_p = m_array[pnIndex] - m_array[prevIndex];
            double b_p = m_array[pnIndex + 1] - m_array[prevIndex + 1];
            double c_p = Math.sqrt(a_p * a_p + b_p * b_p);

            // Check if the previous point is far enough along the P0-p1
            // line from the last plotted point and if it is far enough from
            // the current point. Add it to the new set if it is.
            if ((Math.abs(prevDist - m_lastPlottedDist) > m_width
            	&& c_p > m_width) || (m_lastPlottedIndex == p0Index)) {

				keepPoint(prevIndex, posFarDist);
			}
		}

        // Keep the previous point if it is the last in the array.
        if (pnIndex >= m_array.length) {
        	keepPoint(prevIndex, posFarDist);
        }

        // The last point plotted will become the next p0.
        return m_lastPlottedIndex;
    }

//------------------------------------------------------------------------------
/**
 * Adds the point at the specified index to the temporary array of kept points.
 * Both the X and Y values are copied to the temporary array.
 * @param xIndex	The index in the m_array of the point to keep.
 * @param dist		The distance to be stored as the last plotted distance.
 */
    protected void keepPoint(int xIndex, double dist) {

        // Don't add it if it is the last point plotted since that point
        // was already kept.
        if (xIndex == m_lastPlottedIndex) return;

        // Copy the X and Y values to the temporary storage.
    	m_tempArray[m_numKept * m_numDimensions] = m_array[xIndex];
        m_tempArray[m_numKept * m_numDimensions + 1] = m_array[xIndex + 1];

        // Advance the place holder variables.
        m_numKept++;
        m_lastPlottedIndex = xIndex;
        m_lastPlottedDist = dist;
    }

//------------------------------------------------------------------------------
/**
 * Finds the perpendicular line to a line and through one point on the line and
 * one point off the line.  Returns the intersection of the line with its
 * perpendicular.
 * @param m		The slope of the line.
 * @param x_0	The x coordinate of some point p_0 on the line.
 * @param y_0	The y coordinate of some point p_0 on the line.
 * @param x_n	The x coordinate of some point p_n, through which the
 *				perpendicular line will be constructed.
 * @param y_n	The y coordinate of some point p_n, through which the
 *				perpendicular line will be constructed.
 * @return The point of intersection of the given line and its perpendicular.
 */
    static public Point2D.Double getPerpendicularIntersection(double m,
    	double x_0, double y_0, double x_n, double y_n) {

        // Initialize the return variable.
        Point2D.Double p = new Point2D.Double(0, 0);

        // Calculate the intersection for the case of a non-horizontal line.
        if (m != 0) {
            double a_n = (m * (x_n - x_0) + (y_0 - y_n)) / (1 + m * m);
            double b_n = -m * a_n;
            p.x = x_n + b_n;
            p.y = y_n + a_n;
		}
        else {
        	// For a horizontal line.
            p.x = x_n;
            p.y = y_0;
        }

     	return p;
	}

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param val	The new xScale to use in the point reduction algorithm.
 */
    public void setXScale(double val) {

	    m_xScale = val;
	}

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param val	The new yScale to use in the point reduction algorithm.
 */
    public void setYScale(double val) {

	    m_yScale = val;
	}

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param width	The new width to use in the point reduction algorithm.
 */
    public void setWidth(double width) {

	    m_width = width;
	}

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * Menu items include: <p>
 * "Calculate"  <p>
 * @return A menu full of commands applicable to this class.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"Line Point Reducer");
        javax.swing.JMenuItem saveItem = new javax.swing.JMenuItem("Calculate");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
    	return subMenu;
    }

//------------------------------------------------------------------------------
/**
 * Act on menu items being selected.
 * @param e The event to act on.
 */
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Calculate")) {
        	recalculate();
        }
    }
}

