/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-01
 * Date started:    1999-05-03
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.tools;

import java.util.*;
import gov.usgs.gr.data.*;

//==============================================================================
/**
 * For each DataSeries in the list passed to this class, Calculates the means of a group of data series.
 */
public class MeanCalculator implements gov.usgs.gr.tools.DataManipulator {

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public MeanCalculator () {
    }

//------------------------------------------------------------------------------
/**
 * For each series in the list, calculate the mean of the last dimension
 * of its array.  Replace the array in the series with two points whose
 * last dimension values are the calculated mean and whose other dimension
 * values are copied from the first and last points of the original series.
 * Calls findMean() to do the work.
 * @param dataList  A list of DoubleDataSeries objects.
 */
    public void performAction (List dataList) {

    	Iterator iter = dataList.iterator();
        while (iter.hasNext()) {
        	DoubleDataSeries series = (DoubleDataSeries)iter.next();
            synchronized (series) {
                findMean(series);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Calculates the mean of the data series and replaces the data series with a
 * straight line between the end points.  The mean is calculated for the last
 * dimension (Y for X-Y data).
 * @param series    The series to modify.
 */
    public void findMean(DoubleDataSeries series) {

        double mean = calculateMean(series);

        // Create a new array with two points spanning between the endpoints
        // with the value of the last dimension being equal to the mean.
        double[] array = series.getArray();
        int numDimensions = series.getNumDimensions();
        double[] newArray = new double[numDimensions * 2];
        for (int i = 0; i < (numDimensions - 1); i++) {
        	newArray[i] = array[i];
            newArray[i + numDimensions] = array[array.length - numDimensions];
        }
        newArray[numDimensions - 1] = mean;
        newArray[(numDimensions * 2) - 1] = mean;

        // Replace the old array with the new mean.
        series.setArray(newArray);
	}

//------------------------------------------------------------------------------
/**
 * Calculates the mean of the data series.  The mean is calculated for the last
 * dimension (Y for X-Y data).
 * @param series    The series to read.
 */
    public static double calculateMean(DoubleDataSeries series) {

    	// Get the data and number of dimensions from the series.
        double[] array = series.getArray();
        double total = 0;
        int numDimensions = series.getNumDimensions();

        // Calculate the mean.
        for (int i = 0; i < series.getCount(); i++) {
        	total += array[(i + 1) * numDimensions - 1];
        }

        double mean = total / series.getCount();

        return mean;
	}
}

