/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-01
 * Date started:    2007-05-01
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.tools;

import java.util.*;
import gov.usgs.gr.data.*;

//==============================================================================
/**
 * For each DataSeries in the list passed to this class, Calculates the means of a group of data series.
 */
public class MovingAverageCalculator
    implements gov.usgs.gr.tools.DataManipulator {
    
    /** The span of the moving average. */
    protected double m_span = 1;
    /** the number of divisions for sampling. */
    protected int m_numDivs = 1;
    
//------------------------------------------------------------------------------
    /**
     * Creates an empty object.
     */
    public MovingAverageCalculator() {
    }
    
//------------------------------------------------------------------------------
    /**
     * Creates an object, and sets the span and numDivs values.
     */
    public MovingAverageCalculator(double span, int numDivs) {
	
	m_span = span;
	m_numDivs = numDivs;
    }
    
//------------------------------------------------------------------------------
    /**
     * For each series in the list, calculate the mean of the last dimension
     * of its array.  Replace the array in the series with two points whose
     * last dimension values are the calculated mean and whose other dimension
     * values are copied from the first and last points of the original series.
     * Calls findMean() to do the work.
     * @param dataList  A list of DoubleDataSeries objects.
     */
    public void performAction(List dataList) {
	
	Iterator iter = dataList.iterator();
	while (iter.hasNext()) {
	    DoubleDataSeries series = (DoubleDataSeries)iter.next();
	    synchronized (series) {
		applyMovingAverage(series, m_span, m_numDivs);
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Perform a moving average on all selected points.
     * @param span	The span that should be averaged.  If span=n, then the
     *			average at point p would be based on the data between
     *			p-(n/2) and p+(n/2).
     * @param numDivs	The number of sampling divisions across the span.
     */
    public static void applyMovingAverage(DoubleDataSeries series,
	double span, int numDivs) {
	
	// Don't bother proceeding if there aren't at least two points.
	if (series.getNumPoints() < 2 ) return;
	
	// Return if the span or number of divisions is out of range.
	span = Math.abs(span);
	if (span <= 0) return;
	
	// If number of divisions is 0, change it to -1, which is the signal
	// for using the number of points * 2.
	if (numDivs == 0)
	    numDivs = -1;
	
	// Calculate step size.
	double stepSize = span / (double)numDivs;
	
	// Traverse the array and write the calculated average for the Y value.
	int begin = series.getBegin();
	int end = series.getEnd();
	double[] array = series.getArray();
	int numDimensions = series.getNumDimensions();
	for (int i = begin; i <= end; i++) {
	    double curPoint = array[i * numDimensions];
	    double halfSpan = span / 2;
	    double spanMin = curPoint - halfSpan;
	    double spanMax = curPoint + halfSpan;
	    array[i * numDimensions + 1] =
		series.findSpanAverage(spanMin, spanMax, numDivs);
	}
	
	// Prepare to chop off the edges.
	series.deselectAll();
	double[] min = new double[2];
	double[] max = new double[2];
	// Select the left edge...
	min[0] = array[begin] - 1;
	max[0] = array[begin] + (span / 2);
	min[1] = -Double.MAX_VALUE;
	max[1] = Double.MAX_VALUE;
	series.selectArea(min, max, true);
	// Except for the point that is exactly half the span from the front...
	min[0] = array[begin] + (span / 2);
	max[0] = array[begin] + (span / 2);
	min[1] = -Double.MAX_VALUE;
	max[1] = Double.MAX_VALUE;
	series.selectArea(min, max, false);
	// Select the right edge...
	min[0] = array[end * numDimensions] - (span / 2);
	max[0] = array[end * numDimensions] + 1;
	min[1] = -Double.MAX_VALUE;
	max[1] = Double.MAX_VALUE;
	series.selectArea(min, max, true);
	// Except for the point that is exactly half the span from the end...
	min[0] = array[end * numDimensions] - (span / 2);
	max[0] = array[end * numDimensions] - (span / 2);
	min[1] = -Double.MAX_VALUE;
	max[1] = Double.MAX_VALUE;
	series.selectArea(min, max, false);
	
	// Chop the edges.
	series.deleteSelectedPoints();
	
	// Update the series descriptors.
	series.updateMinMax();
    }
}

