/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-01
 * Date started:    2007-03-05
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.tools;

import java.util.*;
import gov.usgs.gr.data.*;

//==============================================================================
/**
 * Fills gaps in DataSeries with a constant.  The gaps are
 * identified as any point from the series' list of deleted points that falls
 * between the series' first and last selected points.
 */
public class SeriesGapFiller implements gov.usgs.gr.tools.DataManipulator {
    
//------------------------------------------------------------------------------
    /**
     * Creates an empty object.
     */
    public SeriesGapFiller() {
    }
    
//------------------------------------------------------------------------------
    /**
     * Fills gaps between selected points in each data series in a list.
     * Calls the fill() method to do the work.
     * @param dataList  A list of DoubleDataSeries objects with gaps to fill.
     */
    public void performAction(List dataList) {
        
        Iterator iter = dataList.iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries)iter.next();
            fill(series);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Fills gaps between selected points in each data series in a list.
     * Calls the fill() method to do the work.
     * @param dataList  A list of DoubleDataSeries objects with gaps to fill.
     */
    public void performAction(List dataList, SeriesValueProvider svp) {
        
        Iterator iter = dataList.iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries)iter.next();
            fill(series, svp);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Fill gaps that exist within the bounds of the selected points.  Fill the
     * gaps linearly.
     * @param series    The series to fill.
     */
    static public void fill(DoubleDataSeries series) {
        
        fill(series, new InterpolatedValueProvider());
    }
    
//------------------------------------------------------------------------------
    /**
     * Fill gaps that exist within the bounds of the selected points.  Fill the
     * gaps linearly.
     * @param series    The series to fill.
     */
    static public void fill(DoubleDataSeries series, SeriesValueProvider svp) {
        
        synchronized (series) {
            
            // Initialize local variables.
            int minIndex = series.getEnd();
            int maxIndex = 0;
            int[] selectedIndexes = series.getSelectedIndexes();
            double[] array = series.getArray();
            
            // Determine a range inside of which gaps should be filled.
            // This is determined by the first and last selected points.
            for (int j = 0; j < series.getNumSelected(); j++) {
                if (selectedIndexes[j] < minIndex)
                    minIndex = selectedIndexes[j];
                if (selectedIndexes[j] > maxIndex)
                    maxIndex = selectedIndexes[j];
            }
            
            // Use the first and last selected indexes to find the range along
            // the X axis.
            double minVal, maxVal;
            if (minIndex == 0) minVal = -Double.MAX_VALUE;
            else minVal = array[minIndex * 2];
            if (maxIndex == series.getEnd()) maxVal = Double.MAX_VALUE;
            else maxVal = array[maxIndex * 2];
            
            // Get the list of deleted points and make a new list to hold points
            // that are to be restored from the deleted list.
            List<Double> deleted = series.getDeletedList();
            List<Double> restored = new ArrayList();
            
            // Look through the deleted list for points within the range.
            Iterator<Double> iter = deleted.iterator();
            while (iter.hasNext()) {
                Double xObj = (Double)iter.next();
                Double yObj = (Double)iter.next();
                double xVal = xObj;
                if (xVal > minVal && xVal < maxVal) {
                    // Place the point in the restored list in order from
                    // lowest to highest x value.
                    int index = restored.size();
                    for (int i = 0; i < restored.size(); i += 2) {
                        double val = restored.get(i);
                        if (val > xVal) {
                            index = i;
                            break;
                        }
                    }
                    restored.add(index, xObj);
                    restored.add(index + 1, yObj);
                } else {
                }
            }
            
            // Remove the restored points from the deleted list.
            iter = restored.iterator();
            while (iter.hasNext()) {
                Double real = (Double)iter.next();
                
                boolean bNotDone = true;
                int index = 0;
                while (bNotDone) {
                    if (deleted.get(index) == real) {
                        deleted.remove(index);
                    } else {
                        index++;
                    }
                    if (deleted.size() <= index) bNotDone = false;
                }
            }
            
            // Add the points that were deleted into the data array.
            restore(series, restored, svp);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Restores points from the list to a new array and set it as the current
     * array for the series.
     * @param series    The series to which the data should be added.
     * @param restored  The list of points to be restored stored as interleaved
     *                  X and Y values.
     */
    static public void restore(DoubleDataSeries series, List<Double> restored,
        SeriesValueProvider svp) {
        
        // Get the data array from the series.
        double[] array = series.getArray();
        
        // Create a new, longer array to hold the old array plus the restored
        // points.
        int newArraySize = (series.getCount() * 2) + restored.size();
        double[] newArray = new double[newArraySize];
        int oldIndex = 0;
        boolean bOldArrayExhausted = false;
        
        // Fill the new array, inserting restored points along the way.
        for (int i = 0; i < (newArray.length - 1); i += 2) {
            double restoreVal = Double.NaN;
            
            // Get the X value of the next point to restore.
            if (restored.size() > 0)
                restoreVal = restored.get(0);
            
            // Check whether the next point in the new array should come from
            // the old array or be restored as an interpolated point.
            if (restoreVal < array[oldIndex] || bOldArrayExhausted) {
                newArray[i] = restoreVal;
                restored.remove(0);
                
                double yValue = svp.getValueAt(series, restoreVal);
                
                // Assign the value and remove it from the restore list.
                newArray[i + 1] = yValue;
                restored.remove(0);
            } else {
                // The next point going into the new array should come from
                // the old array because the next point to restore is still
                // further down the line.
                // Copy the existing point from the old array.
                newArray[i] = array[oldIndex];
                newArray[i + 1] = array[oldIndex + 1];
                if (oldIndex < (array.length - 2)) oldIndex += 2;
                else bOldArrayExhausted = true;
            }
        }
        
        // Make the new array active.
        series.setArray(newArray);
        series.deselectAll();
    }
        
//------------------------------------------------------------------------------
    /**
     * Calculates the fill value at a certain point.
     */
    static public double getNewFillValueAt(DoubleDataSeries series,
	double xPoint) {
    
        double fillValue = series.getInterpolatedValue(xPoint);
        
        return fillValue;
    }
}

