/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-09-14
 * Date started:    2000-02-10
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.tools;

import java.awt.geom.*;
import java.util.*;
import gov.usgs.gr.data.*;

//==============================================================================
/**
 * Reduces the number of points in a two-dimensional DataSeries object so that
 * it looks the same when plotted at a given scale and using a given symbol
 * size.  No points are moved or added.  The algorithm attempts to eliminate as
 * many points as possible without affecting the appearance.  Points that are
 * within a certain width radius of an already-plotted point are eliminated.
 * The width parameter is serves as the acceptable radius.
 */
public class SymbolPointReducer extends gov.usgs.gr.tools.LinePointReducer {

    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/SymbolPointReducer.gif";

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public SymbolPointReducer () {

        // Check for default configuration settings.
        setDefaults("gov.usgs.gr.tools.SymbolPointReducer");
    }

//------------------------------------------------------------------------------
/**
 * Creates an empty object and sets the scales and widths to the specified
 * values.
 * @param xScale    The scale factor to use in the X direction.
 * @param yScale    The scale factor to use in the Y direction.
 * @param width     The distance that points may vary from a plotted point in
 *                  order to be eliminated.
 */
    public SymbolPointReducer (double xScale, double yScale, double width) {

        // Check for default configuration settings.
        setDefaults("gov.usgs.gr.tools.SymbolPointReducer");

    	m_xScale = xScale;
        m_yScale = yScale;
        m_width = width;
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

        String s = "Symbol Point Reducer";

        return s;
     }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Applies the point reduction algorithm starting from some point P0 in the
 * array.  This method returns when it reaches the next point that should be
 * considered P0.  Any points that should be kept in between are added to the
 * member array m_tempArray.
 * @param p0Index	The index to use for P0.
 * @return The index of the next P0.
 */
    public int handleNewP0(int p0Index) {

    	// Check that p0Index is within the bounds of the array.
        if ((p0Index < 0) ||
        	(p0Index > (m_array.length - (2 * m_numDimensions)))) {

        	return m_array.length;
        }

        // P0 is always a point that has already been added to the new set
        // of points.  Get the x and y coordinates of the P0 point.
        double x_0 = 0;
        double y_0 = 0;

        // Determine the NaN status of P0 and P1.
        boolean bP0NaN = Double.isNaN(m_array[p0Index]) ||
            Double.isNaN(m_array[p0Index + 1]);

        int pnIndex = p0Index + m_numDimensions;
        while (pnIndex < m_array.length) {
            // Handle if NaN values.
            boolean bPnNaN = Double.isNaN(m_array[pnIndex]) ||
                Double.isNaN(m_array[pnIndex + 1]);
            
            if (!bP0NaN && bPnNaN) {
                break;
            }
            if (bP0NaN && bPnNaN) {
                pnIndex += m_numDimensions;
                continue;
            }
            if (bP0NaN && !bPnNaN) {
                break;
            }
            
            // This point is Pn, where n is greater than 1 and measured
            // from the most recently assigned P0.
			// Get the x and y coordinates of the current point.
        	double x_n = m_xScale * (m_array[pnIndex] - m_array[p0Index]);
        	double y_n = m_yScale *
            	(m_array[pnIndex + 1] - m_array[p0Index + 1]);

            // Calculate the legs (a, b, and c) of the right triangle
            // through points Pn and Pz.  c_n is the distance from Pn to Pz.
            double b_n = x_n - x_0;
            double a_n = y_n - y_0;
            double c_n = Math.sqrt(a_n * a_n + b_n * b_n);

            // Keep the point if it is outside the specified radius.
			if (c_n > m_width) {
            	break;
            }

            // Advance to the next point and iteration of the loop.
            pnIndex += m_numDimensions;
        }

        if (pnIndex < m_array.length) keepPoint(pnIndex, 0);

        // Keep the previous point if it is the last in the array.
        if (pnIndex >= m_array.length) {
        	keepPoint(pnIndex - m_numDimensions, 0);
        }

        // The last point plotted will become the next p0.
        return m_lastPlottedIndex;
    }
}

