/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-07-20
 * Date started:    2001-02-05
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.tools;

import java.text.*;
import java.util.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
//==============================================================================
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JPanel;

/**
 * Keeps track of the current point along a time dimension of DataSeries
 * objects.  Capable of automatically advancing the current time of the series
 * on its own thread.  Observers of the series can then act on the changes.
 * The auto-advance increment is determined by the spacing between times
 * in the last DataSeries object added as a child.  Changes are sent to the
 * Zoom child.  Changes to the Zoom child are observed by this class and cause
 * the current time to be adjusted.  The zoom should only appear within one
 * page.  If the zoom or this object are moved to (become children of) a
 * different page, the zoom should be readded.  This object is usually the
 * child of a page.
 */
public class TimeMarker extends gov.usgs.gr.GrObject implements Runnable, gov.usgs.gr.visual.gui.GuiPanelProvider {

    /** The ID that is used for linking to a zoom when reading from XML. */
    public Object m_oldZoomID = null;
    /** The DataSeries that is driving the animation. */
    public DoubleDataSeries m_currentSeries = null;
    /** The current time value. */
    protected double m_currentValue = 0;
    /** The time step size. */
    protected int m_stepSize = 1;
    /** The default year to use for writing the date and time. */
    protected int m_year = 0;
    /**
     * A zoom object that allows this object to be linked to an axis on
     * a graph for visual input and output.
     */
    protected Zoom m_zoom;
    /** A description of how dates should be written. */
    protected String m_dateFormatString = "MMM'.' dd',' yyyy 'at' hh:mm a";
    /**
     * The page to watch for thread synchronization when running an
     * animation.
     */
    protected PageOfGraphs m_page;
    /** The thread priority. */
    protected int m_priority = Thread.NORM_PRIORITY;
    /** The offset of the current time from the left edge of the zoom. */
    protected double m_zoomOffset = 0.3;
    /** Whether the thread is currently running. */
    protected boolean m_bRunning = false;
    /**
     * The number of milliseconds for the thread to sleep before stepping
     * the next increment.
     */
    protected int m_sleepDuration = 100;
    /** The separate thread that allows this object to seamlessly animate. */
    protected Thread aThread = null;
    /** The frames per second calculated for the last run period. */
    protected float m_framesPerSecond = 0;
    /** The previous view length of the zoom. */
    protected double m_oldViewLength = 0;
    /** The minimum value of the current time series. */
    protected double m_currentSeriesMin = 0;
    /** The maximum value of the current time series. */
    protected double m_currentSeriesMax = 0;
    /** Whether to lock the time when zooming. */
    protected boolean m_bLockTimeWhenZooming = false;
    /** An instance of the controller associated with this object. */
    protected GrController m_grCon = null;
    /** The panel created by this class for displaying on a toolbar. */
    protected JPanel m_panel = new JPanel();
    /** The play button on the panel. */
    protected JButton m_animPlayButton = new JButton();
    /** The step forward button on the panel. */
    protected JButton m_animForwardButton = new JButton();
    /** The step backward button on the panel. */
    protected JButton m_animBackButton = new JButton();
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/TimeMarker.gif";

//------------------------------------------------------------------------------
/**
     * Create a new time marker.
     */
    public TimeMarker() {

        // Check for default configuration settings.
        setDefaults("gov.usgs.gr.tools.TimeMarker");

        // Start out with one and only one zoom as a child.
        add(new Zoom());

        // Create the panel once so it can be passed out later as needed.
        createPanel();
    }

//------------------------------------------------------------------------------
/**
     * Copy the properties and children of this object into the parameter object.
     * Turn the running property off so no new thread will be created by default.
     * @param gro   The object to copy properties and children into.
     * @param bDeep Whether data arrays should be copied or shared.
     */
    public void copyInto(GrObject gro, boolean bDeep) {

        // Copy the properties.
        GrHashList props = getProperties();
        GrProperty p = (GrProperty) props.get("run");
        p.value = false;
        gro.setProperties(props);

        // Copy the children.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject child = (GrObject) iter.next();
            if (bDeep) {
                gro.add(child.createCopy(false));
            } else {
                gro.add(child);
            }
        }
    }

//------------------------------------------------------------------------------
/**
     * Provide the list of properties with the current values for this object.
     * Properties include: <p>
     * val        Float     "Current Value"  <p>
     * stepSize   Integer   "Step Size"  <p>
     * run        Boolean   "Run Animation"  <p>
     * sleep      Integer   "Sleep Duration"  <p>
     * priority   String    "Priority"  values="Minimum", "Normal", "Maximum" <p>
     * dateFormat String    "Date Format"  <p>
     * year       Integer   "Reference Year"  <p>
     * @return  A list of GrProperty objects.
     */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("val", new GrProperty(m_currentValue, "Current Value"));
        props.put("stepSize", new GrProperty(m_stepSize, "Step Size"));
        props.put("run", new GrProperty(m_bRunning, "Run Animation"));
        props.put("sleep", new GrProperty(m_sleepDuration, "Sleep Duration"));

        // Give the user some priority choices...
        GrProperty p = new GrProperty("", "Priority");
        javax.swing.JComboBox jcb = new javax.swing.JComboBox();

        String[] priorityTypes = {"Minimum", "Normal", "Maximum"};

        // Fill the thread priority combo box.
        for (int i = 0; i < priorityTypes.length; i++) {
            String name = priorityTypes[i];
            jcb.addItem(name);
        }
        if (m_priority == Thread.MIN_PRIORITY) {
            jcb.setSelectedIndex(0);
            p.value = priorityTypes[0];
        }
        if (m_priority == Thread.NORM_PRIORITY) {
            jcb.setSelectedIndex(1);
            p.value = priorityTypes[1];
        }
        if (m_priority == Thread.MAX_PRIORITY) {
            jcb.setSelectedIndex(2);
            p.value = priorityTypes[2];
        }

        p.component = jcb;
        props.put("priority", p);

        props.put("dateFormat", new GrProperty(new String(m_dateFormatString), "Date Format"));
        props.put("year", new GrProperty(m_year, "Reference Year"));
        props.put("fps", new GrProperty(m_framesPerSecond, "Frames Per Second", false, false));

        return props;
    }

//------------------------------------------------------------------------------
/**
     * Set the values of one or more of this object's properties.
     * @see gov.usgs.gr.tools.TimeMarker#getProperties getProperties() for
     * a list of accepted properties.
     * @param props A list of GrProperty objects specifying the new values for
     *              certain properties.
     */
    public void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        String[] priorityTypes = {"minimum", "normal", "maximum"};

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String) iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty) obj;

                if (p.value instanceof String) {
                    setProperty(key, (String) p.value);
                    continue;
                }
                if (key.equals("val")) {
                    m_currentValue = (Double) p.value;
                }
                if (key.equals("run")) {
                    setRunning((Boolean) p.value);
                }
                if (key.equals("sleep")) {
                    m_sleepDuration = (Integer) p.value;
                }
                if (key.equals("year")) {
                    m_stepSize = (Integer) p.value;
                    if (m_stepSize == 0) {
                        m_stepSize = 1;
                    }
                }
                if (key.equals("year")) {
                    m_year = (Integer) p.value;
                }
            }
            if (obj instanceof String) {
                setProperty(key, (String) obj);
            }
        }
    }

//------------------------------------------------------------------------------
/**
     * Set a single property.
     * @see gov.usgs.gr.data.TimeMarker#getProperties getProperties() for
     * a list of accepted properties.
     * @param key   The property name.
     * @param value The new value for the property.
     */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        String[] priorityTypes = {"minimum", "normal", "maximum"};

        if (key.equals("val")) {
            m_currentValue = Double.parseDouble(value);
        }
        if (key.equals("run")) {
            setRunning(Boolean.parseBoolean(value));
        }
        if (key.equals("sleep")) {
            m_sleepDuration = Integer.parseInt(value);
        }
        if (key.equals("stepSize")) {
            m_stepSize = Integer.parseInt(value);
            if (m_stepSize == 0) {
                m_stepSize = 1;
            }
        }
        if (key.equals("priority")) {
            // Check each pre-set priority name.
            String name = value.trim().toLowerCase();
            if (name.equals(priorityTypes[0])) {
                m_priority = Thread.MIN_PRIORITY;
            }
            if (name.equals(priorityTypes[1])) {
                m_priority = Thread.NORM_PRIORITY;
            }
            if (name.equals(priorityTypes[2])) {
                m_priority = Thread.MAX_PRIORITY;
            }
            if (aThread != null && aThread.isAlive()) {
                aThread.setPriority(m_priority);
            }
        }
        if (key.equals("dateFormat")) {
            m_dateFormatString = value;
        }
        if (key.equals("year")) {
            m_year = Integer.parseInt(value);
        }
    }

//------------------------------------------------------------------------------
/**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {

        String s = "Time Marker";

        return s;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0) {
            setTreeIcon();
        }
        return (javax.swing.ImageIcon) TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null) {
            icon = super.setTreeIcon();
        }
        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
     * Adds Zoom, DataSeries, Page, and TextMarkup objects.  There can only be one
     * Zoom object at a time, so adding one replaces the old one.  The last
     * DataSeries added is used to drive the animation step increment.  Adding a
     * page always is successful, but rather than adding it as a child, it sets
     * the page member variable to point at the new page.
     * @param gro   The GrObject to be added.
     * @return  Whether the object was successfully added.
     */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy) {
            gro = ((ObjectEffigy) gro).getSubject();
        }
        if (gro instanceof Zoom) {
            // Allow one zoom at a time.
            remove(m_zoom);
            super.add(gro);
            m_zoom = (Zoom) gro;

            // Update the current time based on the new zoom.
            if (m_currentSeries != null) {
                double min = m_currentSeries.getMin(0);
                double max = m_currentSeries.getMin(0);
                m_zoom.setAllowableContentLimits(min, max);
            }

            m_bLockTimeWhenZooming = false;
            update(m_zoom, null);
            m_bLockTimeWhenZooming = true;

            findParentPageOfZoom();

            return true;
        }
        if (gro instanceof DoubleDataSeries) {
            // Add the DataSeries.
            super.add(gro);

            // Use this most recent series to drive the animation.  Update.
            m_currentSeries = (DoubleDataSeries) gro;
            update(m_currentSeries, null);
            m_bLockTimeWhenZooming = false;
            update(m_zoom, null);
            m_bLockTimeWhenZooming = true;

            return true;
        }
        if (gro instanceof PageOfGraphs) {
            // Set the page member variable .
            m_page = (PageOfGraphs) gro;

            // Start the thread in case there wasn't one running already.
            if (m_bRunning) {
                if (aThread == null) {
                    aThread = new Thread(this, "Animation Thread");
                }
                if (aThread.isAlive() == false) {
                    System.out.println("Creating a new animation thread...");
                    aThread = new Thread(this, "Animation Thread");
                }
                try {
                    aThread.start();
                } catch (IllegalThreadStateException e) {
                }
                aThread.setPriority(m_priority);
            }

            return true;
        }
        if (gro instanceof TextMarkup) {
            // Add the text markup so it can be updated with a text form of
            // the date and time value of this object.
            super.add(gro);

            // Allow the user to see and change the markup's location.
            ((TextMarkup) gro).setUserChangeable(true);

            // Set the text of the markup to the current date.
            ArrayList markupList = new ArrayList();
            markupList.add(gro);
            setDateText(markupList);

            return true;
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
     * Wrapper for removing items from the list of children.
     */
    public synchronized GrObject remove(int i) {

        GrObject gro = super.remove(i);
        if (gro == m_currentSeries) {
            m_currentSeries = null;
        }
        return gro;
    }

//------------------------------------------------------------------------------
/**
     * Wrapper for removing items from the list of children.
     */
    public synchronized boolean remove(GrObject gro) {

        boolean bFound = super.remove(gro);
        if (gro == m_currentSeries) {
            m_currentSeries = null;
        }
        return bFound;
    }

//------------------------------------------------------------------------------
/**
     * Wrapper for clearing the list of children.
     */
    public synchronized void clear() {

        if (m_zoom != null) {
            m_zoom.setAllowableContentLimits(Double.POSITIVE_INFINITY, Double.NEGATIVE_INFINITY);
        }
        super.clear();
    }

//------------------------------------------------------------------------------
/**
     * Perform tasks that had to wait until all XML input had been read.
     * @param idObjects All the objects that were read in a an object ID, such as a
     *                  String or Integer.
     */
    public void finishSAXInput(Map idObjects) {

        // Add the zoom object to the id objects, if it isn't already there.
        if (m_oldZoomID != null) {
            if (idObjects.get(m_oldZoomID) == null) {
                idObjects.put(m_oldZoomID, m_zoom);
            }
        }

        // All the children of this object to receive the finish signal.
        finishSAXForChildren(idObjects);

        // If the zoom ID has been set...
        if (m_oldZoomID != null) {
            // Look through the list of ID objects for a zoom that matches the
            // old zoom ID, and use it if found.
            Iterator iter = idObjects.keySet().iterator();
            while (iter.hasNext()) {
                Object key = iter.next();
                Object val = idObjects.get(key);
                Zoom zoom = null;
                if (val instanceof Zoom) {
                    zoom = (Zoom) val;
                    if (zoom == m_zoom) {
                        continue;
                    }
                    if (m_oldZoomID.equals(key)) {
                        m_zoom = zoom;
                        break;
                    }
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
     * Overrides default update() in GrObject.  Called by an observed GrObject's
     * notifyObservers() method to handle changes to the observed object.
     * May use the notifyObservers() method to propagate the event handling up to
     * this object's observers.
     * @param gro   The observed object that changed.
     * @param arg   Additional information that the updated object may choose to
     *              send.
     */
    public void update(GrObject gro, Object arg) {

        if (m_bDeferUpdates) {
            return;
        }
        // If the zoom changes, change the current mark value.
        if (gro instanceof Zoom && gro == m_zoom) {
            synchronized (m_zoom) {
                double viewLength = m_zoom.getViewLength();
                if (viewLength == 0) {
                    return;
                }
                m_bHasChanged = true;

                double viewMin = m_zoom.getViewMin();
                if (Math.abs((viewLength - m_oldViewLength) / viewLength) < 1E-14) {

                    m_currentValue = viewMin + (m_zoomOffset * viewLength);

                    m_bDeferUpdates = true;
                    processPanEvent();
                    m_bDeferUpdates = false;

                    double offsetValue = viewMin + (m_zoomOffset * viewLength);
                    double seriesMin = offsetValue;
                    double seriesMax = offsetValue;
                    if (m_currentSeries != null) {
                        seriesMin = m_currentSeries.getMin(0);
                        seriesMax = m_currentSeries.getMax(0);
                    }
                    double viewMax = m_zoom.getViewMax();
                    double lowerTravel = seriesMin - offsetValue;
                    double newOuterMin = viewMin;
                    if (lowerTravel < 0) {
                        newOuterMin += lowerTravel;
                    }
                    double upperTravel = seriesMax - offsetValue;
                    double newOuterMax = viewMax;
                    if (upperTravel > 0) {
                        newOuterMax += upperTravel;
                    }
                    m_zoom.setOuterLimits(newOuterMin, newOuterMax);
                } else {
                    m_oldViewLength = viewLength;
                    if (m_bLockTimeWhenZooming) {
                        // For zoom in, out, or all.
                        setZoom(m_currentValue);
                    } else {
                        // For zoom window or adding a new zoom.
                        m_currentValue = viewMin + (m_zoomOffset * viewLength);
                        m_bDeferUpdates = true;
                        setZoom(m_currentValue);
                        m_bDeferUpdates = false;
                        m_oldViewLength = m_zoom.getViewLength();
                        update(m_zoom, null);
                    }
                }
            }
        }
        if (gro == m_currentSeries) {
            if (m_zoom != null) {
                double min = m_currentSeries.getMin(0);
                double max = m_currentSeries.getMin(0);
                m_zoom.setAllowableContentLimits(min, max);
            }

            int tDim = 0;
            double[] array = m_currentSeries.getArray();
            // Get the range of times for the series.
            m_currentSeriesMin = array[tDim];
            m_currentSeriesMax = array[(m_currentSeries.getCount() - 1) * m_currentSeries.getNumDimensions() + tDim];
        }
    }

//------------------------------------------------------------------------------
/**
     * Handle the event of the Zoom being panned to change the current time.
     */
    public synchronized void processPanEvent() {

        // Keep the page from redrawing for every graphical element
        // that is about to be updated.
        if (m_page != null) {
            m_page.deferUpdates();
        }
        ArrayList markupList = new ArrayList();

        // Propogate the current time through the child series objects.
        for (int childIndex = 0; childIndex < m_children.size(); childIndex++) {

            GrObject child = m_children.get(childIndex);
            if (child instanceof DoubleDataSeries) {
                DoubleDataSeries series = (DoubleDataSeries) child;

                // Get the data and active region of the series.
                int oldBegin = series.getBegin();
                int tDim = 0;
                int numDims = series.getNumDimensions();
                double[] array = series.getArray();

                if (array.length < numDims) {
                    continue;
                }
                // Keep track of the last value seen and the direction we are
                // progressing.
                double lastT = 0;
                int lastI = oldBegin;
                int direction = 0;

                int i = oldBegin;
                while (true) {
                    // Quit if the index is out of bounds.
                    if (i < 0) {
                        i = 0;
                        break;
                    }
                    if (i > series.getCount() - 1) {
                        i = series.getCount() - 1;
                        break;
                    }

                    // Move the index up or down in the direction of
                    // the current time.
                    int offset = i * numDims;
                    double t = array[offset + tDim];

                    // Once the index has moved at least once and has passed
                    // the target, determine if the current value or last value
                    // are closer to the target.  Then break.
                    if (direction == 0 && i != lastI) {
                        double lastDiff = Math.abs(m_currentValue - lastT);
                        double currentDiff = Math.abs(m_currentValue - t);
                        if (lastDiff < currentDiff) {
                            i = lastI;
                        }
                        break;
                    }

                    lastT = t;
                    lastI = i;

                    if (m_currentValue >= t) {
                        i++;
                        if (direction < 0) {
                            direction = 0;
                        } else {
                            direction = 1;
                        }
                    } else {
                        i--;
                        if (direction > 0) {
                            direction = 0;
                        } else {
                            direction = -1;
                        }
                    }
                }

                // Set begin and end equal to the index found.
                int begin = i;
                int end = i;
                double currentT = array[i * numDims + tDim];

                // Expand the end upward.
                for (int j = i; j < series.getCount(); j++) {
                    int offset = j * numDims;
                    double t = array[offset + tDim];
                    if (t == currentT) {
                        end = j;
                    }
                    if (t > currentT) {
                        break;
                    }
                }

                // Expand the beginning downward
                for (int j = i; j >= 0; j--) {
                    int offset = j * numDims;
                    double t = array[offset + tDim];
                    if (t == currentT) {
                        begin = j;
                    }
                    if (t < currentT) {
                        break;
                    }
                }

                // Lock in the new range.
                series.setIndexRange(begin, end);
            }

            // Keep a list of text markups so they can be updated.
            if (child instanceof TextMarkup) {
                markupList.add(child);
            }
        }

        // Update the text markup children that were found with the
        // new time.
        if (markupList.size() > 0) {
            setDateText(markupList);
        }
        // Notify observers to the change that was made.
        m_bHasChanged = true;
        notifyObservers();

        if (m_page != null) {
            m_page.processDeferredUpdates();
        }
    }

//------------------------------------------------------------------------------
/**
     * The method that is called by a newly created thread that is set to run
     * this object.
     */
    public void run() {

        long startTime = System.currentTimeMillis();
        long numFramesDrawn = 0;

        // Loop through work and sleep until the thread is turned off.
        while (m_bRunning) {
            // The automatic stepping must be synchronized on the page
            // to keep from deadlocking with other threads that are
            // modifying the page or its contents.
            if (m_page == null) {
                findParentPage();
            }
            if (m_page == null) {
                return;
            }
            synchronized (m_page) {
                step(m_stepSize);
            }


            try {
                // Sleep for the specified length of time before looping again.
                if (m_sleepDuration > 0) {
                    Thread.currentThread().sleep(m_sleepDuration);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }

            numFramesDrawn++;
        }

        long endTime = System.currentTimeMillis();

        m_framesPerSecond = 1000 * (float) (numFramesDrawn) / (float) (endTime - startTime);
    }

//------------------------------------------------------------------------------
/**
     * Step forward or back.  The first dimension of the last DataSeries
     * added to this object is used to determine the step increment.
     * @param stepAmount    How many increments to step.
     */
    public synchronized void step(int stepAmount) {

        // Find the last series that is a child of this marker.
        // Also keep a list of all date text markups.
        ArrayList markupList = new ArrayList();
        m_currentSeries = null;

        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject child = (GrObject) iter.next();
            if (child instanceof DataSeries) {
                m_currentSeries = (DoubleDataSeries) child;
            }
            if (child instanceof TextMarkup) {
                markupList.add(child);
            }
        }

        // If a series was found, proceed.
        if (m_currentSeries != null) {
            // Find the current active range of the series.
            int index = 0;
            if (stepAmount > 0) {
                index = m_currentSeries.getEnd();
            } else {
                index = m_currentSeries.getBegin();
            }
            boolean bRollOver = false;
            // Move the index the specified amount.
            index += stepAmount;
            if (index >= m_currentSeries.getCount()) {
                index = 0;
                bRollOver = true;
            }
            if (index < 0) {
                index = m_currentSeries.getCount() - 1;
            }
            // Find the time value at the new index value.
            int tDim = 0;
            double[] array = m_currentSeries.getArray();
            double newValue = array[index * m_currentSeries.getNumDimensions() + tDim];
            // Get the range of times for the series.
            m_currentSeriesMin = array[tDim];
            m_currentSeriesMax = array[(m_currentSeries.getCount() - 1) * m_currentSeries.getNumDimensions() + tDim];

            setZoom(newValue);

            // Update the date text.
            setDateText(markupList);
        }
    }

//------------------------------------------------------------------------------
/**
     * Set the zoom limits and update the outer limits if needed.
     * @param minValue The lower axis value.
     * @param maxValue The upper axis value.
     * @param newValue The new time value.
     */
    public void setZoom(double newValue) {

        // Synchronize on the Zoom so other threads can't trigger an
        // update concurrently.
        if (m_zoom != null) {
            synchronized (m_zoom) {
                // Pan the axis so that the range stays the same but its
                // time marker line is centered on the new time.
                double viewLength = m_zoom.getViewLength();
                double viewMin = newValue - m_zoomOffset * viewLength;
                double viewMax = newValue + (1.0 - m_zoomOffset) * viewLength;

                double seriesMin = newValue;
                double seriesMax = newValue;
                if (m_currentSeries != null) {
                    seriesMin = m_currentSeries.getMin(0);
                    seriesMax = m_currentSeries.getMax(0);
                }
                double lowerTravel = seriesMin - newValue;
                double newOuterMin = viewMin;
                if (lowerTravel < 0) {
                    newOuterMin += lowerTravel;
                }
                double upperTravel = seriesMax - newValue;
                double newOuterMax = viewMax;
                if (upperTravel > 0) {
                    newOuterMax += upperTravel;
                }
                // Feed the new calculated axis span into the zoom
                // This will trigger an update to the page and this
                // object.
                m_zoom.setOuterLimits(newOuterMin, newOuterMax);
                m_zoom.setViewLimits(viewMin, viewMax);
            }
        }
    }

//------------------------------------------------------------------------------
/**
     * Toggle between play and pause.
     */
    public void toggleRunning() {

        setRunning(!m_bRunning);
        m_animPlayButton.setSelected(m_bRunning);
    }

//------------------------------------------------------------------------------
/**
     * Step forward.
     */
    public void stepForward() {

        step(m_stepSize);
    }

//------------------------------------------------------------------------------
/**
     * Step back.
     */
    public void stepBackward() {

        step(-m_stepSize);
    }

//------------------------------------------------------------------------------
/**
     * Finds the PageofGraphs that contains m_zoom.
     */
    public void findParentPage() {

        //Initially try to use the parent page of the zoom, if there is one.
        findParentPageOfZoom();

        if (m_page == null) {
            boolean bPageFound = false;
            List tmObservers = getObservers();
            Iterator obIter = tmObservers.iterator();
            while (obIter.hasNext()) {
                // Look for a page parent of the graph.
                GrObject gro = (GrObject) obIter.next();
                if (gro instanceof PageOfGraphs) {
                    m_page = (PageOfGraphs) gro;
                    bPageFound = true;
                    break;
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
     * Finds the PageofGraphs that contains m_zoom.
     */
    public void findParentPageOfZoom() {

        if (m_zoom == null) {
            return;
        }
        // Start with the zoom.
        boolean bPageFound = false;
        List zoomObservers = m_zoom.getObservers();
        Iterator zoomIter = zoomObservers.iterator();
        outer:
        while (zoomIter.hasNext() && !bPageFound) {
            // Look for an axis parent of the zoom.
            GrObject gro = (GrObject) zoomIter.next();
            if (gro instanceof Axis) {
                List axisObservers = gro.getObservers();
                Iterator axisIter = axisObservers.iterator();
                while (axisIter.hasNext()) {
                    // Look for a graph parent of the axis.
                    GrObject gro2 = (GrObject) axisIter.next();
                    if (gro2 instanceof Graph) {
                        List graphObservers = gro2.getObservers();
                        Iterator graphIter = graphObservers.iterator();
                        while (graphIter.hasNext()) {
                            // Look for a page parent of the graph.
                            GrObject gro3 = (GrObject) graphIter.next();
                            if (gro3 instanceof PageOfGraphs) {
                                m_page = (PageOfGraphs) gro3;
                                bPageFound = true;
                                break outer;
                            }
                        }
                    }
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
     * Updates the dates for all the markups in the list.
     * @param markupList    The list of markups to update.
     */
    public void setDateText(ArrayList markupList) {

        String dateString = "---";

        if (m_currentSeries != null) {
            // Create a string representing the date.
            if (m_currentSeries instanceof TimeSeries) {
                m_year = ((TimeSeries) m_currentSeries).getReferenceYear();
            }
            GregorianCalendar cal = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
            cal = TimeDataFile.yearDayToGregCal(m_year, m_currentValue, cal);
            Date date = cal.getTime();
            SimpleDateFormat formatter = new SimpleDateFormat(m_dateFormatString);
            formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
            dateString = formatter.format(date);
        }

        // Update all the markups in the list.
        Iterator iter = markupList.iterator();
        while (iter.hasNext()) {
            TextMarkup tm = (TextMarkup) iter.next();
            tm.setText(dateString);
        }
    }

//------------------------------------------------------------------------------
/**
     * Updates the run state.
     * @param bRun  Whether the animation will be running.
     */
    public void setRunning(boolean bRun) {
        
        m_bRunning = bRun;
        if (m_bRunning) {
            if (aThread == null) {
                aThread = new Thread(this, "Animation Thread");
            }
            // Create and start a new thread if one is not already running.
            if (aThread.isAlive() == false) {
                System.out.println("Creating a new animation thread...");
                aThread = new Thread(this, "Animation Thread");
            }
            try {
                aThread.start();
            } catch (IllegalThreadStateException e) {
            }
            aThread.setPriority(m_priority);
        }
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The zoom for this axis.
     */

    public Zoom getZoom() {
        return m_zoom;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The current value.
     */

    public double getCurrentValue() {
        return m_currentValue;
    }

//------------------------------------------------------------------------------
/**
     * Attribute modification method.
     * @param value	True if the time should be locked when zooming.
     */
    public void setLockTimeWhenZooming(boolean value) {
        m_bLockTimeWhenZooming = value;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The panel created by this class at initialization time.
     */
    public JPanel getPanel() {

        return m_panel;
    }

//------------------------------------------------------------------------------
/**
     * Create the panel that will be displayed on the tool bar.
     */
    public void createPanel() {

        // Create the play/pause action listener.
        java.awt.event.ActionListener toggleRunningAL =
            new java.awt.event.ActionListener() {

            public void actionPerformed(ActionEvent e) {
                // Toggle the state of the run property.
                toggleRunning();
            }
        };

        // Create the step forward action listener.
        java.awt.event.ActionListener stepForwardAL =
            new java.awt.event.ActionListener() {

            public void actionPerformed(ActionEvent e) {
                // Toggle the state of the run property.
                stepForward();
            }
        };

        // Create the step backward action listener.
        java.awt.event.ActionListener stepBackwardAL =
            new java.awt.event.ActionListener() {

            public void actionPerformed(ActionEvent e) {
                // Toggle the state of the run property.
                stepBackward();
            }
        };

        // Set up the play/pause button.
        m_animPlayButton.setRolloverEnabled(false);

        ImageIcon animPlayIcon =
            new ImageIcon(Thread.currentThread().getContextClassLoader().getResource("GUI/animPlay.gif"));
        ImageIcon animPauseIcon =
            new ImageIcon(Thread.currentThread().getContextClassLoader().getResource("GUI/animPause.gif"));
        if (animPlayIcon.getIconWidth() <= 0) {
            m_animPlayButton.setText("Play");
        } else {
            m_animPlayButton.setIcon(animPlayIcon);
        }
        m_animPlayButton.setSelected(false);
        if (animPauseIcon.getIconWidth() > 0) {
            m_animPlayButton.setSelectedIcon(animPauseIcon);
        }
        m_animPlayButton.addActionListener(toggleRunningAL);
        
        // Set up the step forward button.
        ImageIcon animForwardIcon = new ImageIcon(
            Thread.currentThread().getContextClassLoader().getResource(
            "GUI/animForward.gif"));
        m_animForwardButton = animForwardIcon.getIconWidth() <= 0 ?
            new JButton("Forward") : new JButton(animForwardIcon);
        m_animForwardButton.setRolloverEnabled(false);
        m_animForwardButton.addActionListener(stepForwardAL);
        
        // Set up the step backward button.
        ImageIcon animBackIcon = new ImageIcon(
            Thread.currentThread().getContextClassLoader().getResource(
            "GUI/animBack.gif"));
        m_animBackButton = animBackIcon.getIconWidth() <= 0 ?
            new JButton("Back") : new JButton(animBackIcon);
        m_animBackButton.setRolloverEnabled(false);
        m_animBackButton.addActionListener(stepBackwardAL);
        if (animPlayIcon.getIconWidth() <= 0 ||
            animBackIcon.getIconWidth() <= 0 ||
            animForwardIcon.getIconWidth() <= 0) {

            m_panel.setLayout(new GridLayout(3, 1));
        } else {
            m_panel.setLayout(new GridLayout(1, 3));
        }
        
        // Layout the panel with the buttons on it.
        m_panel.setBorder(BorderFactory.createTitledBorder("Animation"));
        m_animPlayButton.setMargin(new Insets(0, 0, 0, 0));
        m_animForwardButton.setMargin(new Insets(0, 0, 0, 0));
        m_animBackButton.setMargin(new Insets(0, 0, 0, 0));
        m_panel.add(m_animBackButton);
        m_panel.add(m_animPlayButton);
        m_panel.add(m_animForwardButton);
    }

//------------------------------------------------------------------------------
/**
     * Tells this interface what the current controller and page are.
     * @param grCon   The GrController to use.
     * @param page    The PageOfGraphs to use.
     */
    public void setControllerAndPage(GrController grCon, PageOfGraphs page) {

        m_grCon = grCon;
        m_page = page;
    }
}
