/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import java.util.*;
import gov.usgs.gr.*;

//==============================================================================
/**
 * A generic axis with ticks and labels for an X-Y graph.
*/
public abstract class Axis extends gov.usgs.gr.visual.GraphicalElement {
    /** The ID that is used for linking when reading from XML. */
    public Object m_oldZoomID = null;
    /** The dimension that this axis is associated with. */
    protected int m_dimension = 0;
    /** Whether to show minor ticks. */
    protected boolean m_bShowMinorTicks = true;
    /** Whether to show major ticks. */
    protected boolean m_bShowMajorTicks = true;
    /** Whether to show labels beneath each major tick. */
    protected boolean m_bShowText = true;
    /** Whether to show grid lines. */
    protected boolean m_bShowGridLines = true;
    /** The color of the ticks, in an RGB array. */
    protected float[] m_tickColor = {1, 1, 0};
    /** The color of the grid, in an RGB array. */
    protected float[] m_gridColor = {0.5f, 0.5f, 0.5f};
    /**
     * Whether to save time by drawing a simpler representation of the
     * graph.
     */
    protected boolean m_bQuickDraw = false;
    /** The relative size of the ticks. */
    protected double m_tickSize = 3;
    /** The length of the major ticks. */
    protected double m_majorTickLen = 2;
    /** The length of the minor ticks as a proportion of major ticks. */
    protected double m_minorTickProportion = 0.5;
    /** The number of sections along the axis separated by major ticks. */
    protected double m_numMajorDivs = 5;
    /** The number of sections between major ticks separated by minor ticks. */
    protected double m_numMinorDivs = 4;
    /** An object describing the limits of this axis. */
    protected Zoom m_zoom;
    /** An object describing the limits of the perpendicular axis. */
    protected Zoom m_opposingZoom;
    /** The lowest point shown on this axis. */
    protected double m_min;
    /** The highest point shown on this axis. */
    protected double m_max;
    /** The space between major ticks in axis units. */
    protected double m_majorTickInc = 50;
    /** The space between minor ticks in axis units. */
    protected double m_minorTickInc = 10;
    /** The lowest point shown on the perpindicular axis. */
    protected double m_oppMin;
    /** The highest point shown on the perpindicular axis. */
    protected double m_oppMax;
    /** The lowest point on this axis that has a major tick mark. */
    protected double m_majorStart;
    /**
     * For scaling ticks so they always appear the same size regardless of
     * the size and shape of the window, or the current zoom.
     */
    protected double m_displayScale;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/Axis.gif";

//------------------------------------------------------------------------------
/**
 * Create a new empty axis.
 */
    public Axis () {

    	add(new Zoom());

        setDefaults("gov.usgs.gr.visual.gl.Axis");
    }

//------------------------------------------------------------------------------
/**
 * Create a new axis and associates it with a set of limits for itself and
 * its perpendicular counterpart.
 * <p><b> Pre Conditions </b><br>
 *      Axis limits exist for this axis and its perpendicular counterpart.
 * <p><b> Post Conditions </b><br>
 *      None.
 */
    public Axis (Zoom zoom, Zoom opposingZoom) {

        add(zoom);
        m_opposingZoom = opposingZoom;

        setDefaults("gov.usgs.gr.visual.gl.Axis");
    }

//------------------------------------------------------------------------------
/**
 * Create a new axis by copying an existing one.
 * @param axis	The axis to be copied.
 */
    public Axis (Axis axis) {

        Zoom zoom = axis.getZoom();
		add(new Zoom(zoom));
        m_zoom.setProperties(zoom.getProperties());

        zoom = axis.getOpposingZoom();
    	m_opposingZoom = new Zoom(zoom);
        m_opposingZoom.setProperties(zoom.getProperties());

        setProperties(axis.getProperties());

        // Synchronize with the new zoom objects.
		update();
    }

//------------------------------------------------------------------------------
/**
 * Make the member variables match the limits objects.
 */
    public void update() {

        // Store the most recent axis limits.
        m_min = m_zoom.getViewMin();
        m_max = m_zoom.getViewMax();
        m_majorTickInc = m_zoom.getMajorInc();
        m_minorTickInc = m_zoom.getMinorInc();
        if (m_opposingZoom != null) {
        	m_oppMin = m_opposingZoom.getViewMin();
        	m_oppMax = m_opposingZoom.getViewMax();
        }
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Axis";

        return s;
 	}

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.remove("show");

        props.put("dimension", new GrProperty(m_dimension, "Dimension", false));
        props.put("showMajorTicks", new GrProperty(m_bShowMajorTicks,
            "Show Major Ticks"));
        props.put("showMinorTicks", new GrProperty(m_bShowMinorTicks,
            "Show Minor Ticks"));
        props.put("showLabels", new GrProperty(m_bShowText, "Show Labels"));
        props.put("showGridLines", new GrProperty(m_bShowGridLines,
            "Show Grid Lines"));

        ColorProperty cp = new ColorProperty();
        cp.description = "Grid Color";
        cp.setColor(m_gridColor);
        cp.display = false;
        cp.save = false;
        props.put("gridColor", cp);

        cp = new ColorProperty();
        cp.description = "Tick Color";
        cp.setColor(m_tickColor);
        cp.display = false;
        cp.save = false;
        props.put("tickColor", cp);

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("zoomID"))
            		m_oldZoomID = p.value;
            	if (key.equals("dimension"))
            		m_dimension = ((Integer)p.value);
            	if (key.equals("showMinorTicks"))
            		m_bShowMinorTicks = (Boolean)p.value;
            	if (key.equals("showMajorTicks"))
            		m_bShowMajorTicks = (Boolean)p.value;
            	if (key.equals("showLabels"))
            		m_bShowText = (Boolean)p.value;
            	if (key.equals("showGridLines"))
            		m_bShowGridLines = (Boolean)p.value;
            	if (key.equals("gridColor")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_gridColor = cp.getColorArray();
                }
            	if (key.equals("tickColor")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_tickColor = cp.getColorArray();
                }
            }
            if (obj instanceof String) {
				String value = (String)obj;

        		setProperty(key, value);

/*
            	if (key.equals("zoomID"))
            		m_oldZoomID = value;
        		if (key.equals("dimension"))
        			m_dimension = Integer.parseInt(value);
        		if (key.equals("showMinorTicks"))
        			m_bShowMinorTicks = Boolean.parseBoolean(value);
        		if (key.equals("showMajorTicks"))
        			m_bShowMajorTicks = Boolean.parseBoolean(value);
        		if (key.equals("showLabels"))
        			m_bShowText = Boolean.parseBoolean(value);
        		if (key.equals("showGridLines"))
        			m_bShowGridLines = Boolean.parseBoolean(value);
*/
            }
        }

        m_zoom.setProperties(props);
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("zoomID"))
        	m_oldZoomID = value;
        if (key.equals("dimension"))
        	m_dimension = Integer.parseInt(value);
        if (key.equals("showMinorTicks"))
        	m_bShowMinorTicks = Boolean.parseBoolean(value);
        if (key.equals("showMajorTicks"))
        	m_bShowMajorTicks = Boolean.parseBoolean(value);
        if (key.equals("showLabels"))
        	m_bShowText = Boolean.parseBoolean(value);
        if (key.equals("showGridLines"))
        	m_bShowGridLines = Boolean.parseBoolean(value);
        if (key.equals("unitScale"))
        	m_zoom.setUnitScale(Double.parseDouble(value));
        if (key.equals("labelFormat"))
        	m_zoom.setFormatString(value);
        if (key.equals("gridColor")) {
            m_gridColor = ColorProperty.getColorArray(value);
        }
        if (key.equals("tickColor")) {
            m_tickColor = ColorProperty.getColorArray(value);
        }

    	m_zoom.setProperty(key, value);
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.  Don't add any children.
 */
    public boolean add(GrObject gro) {

    	if (gro instanceof Axis) {
        	add(((Axis)gro).getZoom());
        }
    	if (gro instanceof Zoom) {
        	remove(m_zoom);
            super.add(gro);
            m_zoom = (Zoom)gro;

            return true;
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Read an XML version of this element and its children.
 */
    public void readFromDOM(org.w3c.dom.Node node,
    	Map allObjects, List unlinkedObjects) {

    	super.readFromDOM(node, allObjects, unlinkedObjects);

        // Look for a zoom in the allObjects list that has the same ID.
		// If one is found, share with it.  Otherwise add this axis' zoom.
    	org.w3c.dom.NamedNodeMap map = node.getAttributes();
        org.w3c.dom.Node idNode = map.getNamedItem("zoomID");
        if (idNode != null) {
        	Object id = idNode.getNodeValue();
        	boolean bFound = false;

        	Iterator iter = allObjects.keySet().iterator();
        	while (iter.hasNext()) {
        		Object key = iter.next();
        		if (id.equals(key)) {
            		m_zoom = (Zoom)allObjects.get(key);
                	bFound = true;
                	break;
            	}
        	}

        	if (!bFound) allObjects.put(idNode.getNodeValue(), m_zoom);
        }
    }

//------------------------------------------------------------------------------
/**
 * Perform tasks that had to wait until all XML input had been read.
 */
    public void finishSAXInput(Map idObjects) {

        // Add the zoom object to the id objects, if it isn't already there.
        if (m_oldZoomID != null) {
    		if (idObjects.get(m_oldZoomID) == null)
        		idObjects.put(m_oldZoomID, m_zoom);
        }

    	finishSAXForChildren(idObjects);

        // Return if the zoom ID hasn't been set.
        if (m_oldZoomID == null) return;

        // Look through the list of ID objects for a zoom that matches the
        // old zoom ID, and use it if found.
    	Iterator iter = idObjects.keySet().iterator();
        while (iter.hasNext()) {
        	Object key = iter.next();
            Object val = idObjects.get(key);
            Zoom zoom = null;
            if (val instanceof Zoom) zoom = (Zoom)val;
            else continue;
            if (zoom == m_zoom) continue;
            if (m_oldZoomID.equals(key)) {
            	m_zoom = zoom;
                break;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Read the attribute settings from a values stored in a list.
 * <p><b> Pre Conditions </b><br>
 *      None.
 * <p><b> Post Conditions </b><br>
 *      update() needs to be called on this axis and the opposing axis.
 * @param attribs	A list of attribute values.
 */
    public void setAttributes(List attribs) {

        if (attribs.size() != 8) return;

        m_zoom.setAttributes(attribs);
      	setTitle((String)attribs.get(7));
 	}

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Draw a light line across the graph for each major tick.  Line and color
 * attributes are saved and then restored when drawing is complete.
 */
    protected void drawGridLines() {
    }

//------------------------------------------------------------------------------
/**
 * Draw one or more small tick marks between the large tick marks.
 */
    protected void drawMinorTicks() {
    }

//------------------------------------------------------------------------------
/**
 * Draw one or more large tick marks along the axis.
 */
    protected void drawMajorTicks() {
	}

//------------------------------------------------------------------------------
/**
 * Draw line on both sides of the graph to represent the axis.  Saves
 * existing line attributes and restores them after drawing.
 */
    protected void drawAxisLine() {
    }

//------------------------------------------------------------------------------
/**
 * Draws labels in magenta beside each major tick mark.
 */
    public void drawText() {
    }

//------------------------------------------------------------------------------
/**
 * Draw all the elements that make up a time series graph.  This method should
 * be overridden to provide output to a certain device.
 */
    public void draw(){
        // Get the latest changes from the axis limits objects.
        update();

        // Calculate the start for the major tick marks, grid lines, and labels.
        // Calculate the length of a tick for this axis.
	    // Save the calling function's GL settings.

        // Set the line style and color.

        // Draw the various parts of the  graph.
        drawGridLines();
        drawMinorTicks();
        drawMajorTicks();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The dimension for this axis.
 */

	public int getDimension() {
    	return m_dimension;
    }

//------------------------------------------------------------------------------
/**
 * Tell whether the parameter object is the same type as this object so
 * that its properties should be applied.  Overrides GraphicalElement to keep
 * from checking "show".
 */
    public boolean isSameTypeAs(GrObject gro) {

        if (getClass().isInstance(gro))
            return true;
        else
            return false;
    }

//------------------------------------------------------------------------------
/**
 * Set the spacing between minor ticks marks.
 * @param inc   The increment in axis units.
 */
    public void setMinorTickInc(double inc) {
        if (inc > 0) {
            m_minorTickInc = inc;
            m_bShowMinorTicks = true;
        }
        else {
            m_bShowMinorTicks = false;
        }
    }

//------------------------------------------------------------------------------
/**
 * Set the spacing between major ticks marks.
 * @param inc   The increment in axis units.
 */
    public void setMajorTickInc(double inc) {
        if (inc > 0) {
            m_majorTickInc = inc;
            m_bShowMajorTicks = true;
        }
        else {
            m_bShowMajorTicks = false;
        }
    }

//------------------------------------------------------------------------------
/**
 * Set the screen length of the opposing axis.
 * @param scale The scale to be used for making uniform ticks.
 */
    public void setDisplayScale(double scale) {

        m_displayScale = scale;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return true if the grid lines are to be shown.
 */
	public boolean getShowGridLines() {
    	return m_bShowGridLines;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bShowGridLines	True if the grid lines are to be shown.
 */
    public void setShowGridLines(boolean bShowGridLines) {
        m_bShowGridLines = bShowGridLines;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return true if minor tick marks are to be shown.
 */
	public boolean getShowMinorTicks() {
    	return m_bShowMinorTicks;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bSShowMinorTicks	True if the minor ticks are to be shown.
 */
    public void setShowMinorTicks(boolean bShowMinorTicks) {
        m_bShowMinorTicks = bShowMinorTicks;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return true if major tick marks are to be shown.
 */

	public boolean getShowMajorTicks() {
    	return m_bShowMajorTicks;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bShowGridLines	True if the major ticks are to be shown.
 */
    public void setShowMajorTicks(boolean bShowMajorTicks) {
        m_bShowMajorTicks = bShowMajorTicks;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return true if text is to be shown.
 */

	public boolean getShowText() {
    	return m_bShowText;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bShowText	True if the axis labels are to be shown.
 */
    public void setShowText(boolean bShowText) {
        m_bShowText = bShowText;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return true if title is to be shown.
 */

	public boolean getShowTitle() {
    	return m_bShowTitle;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bShowTitle	True if the axis title is to be shown.
 */
    public void setShowTitle(boolean bShowTitle) {
        m_bShowTitle = bShowTitle;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The title of the graph.
 */
    public String getTitle() {

    	String title = "";
        int index1 = 0;

        while (index1 < m_title.length()) {
        	int index2 = m_title.indexOf("" + '\n', index1);
            if (index2 < 0) index2 = m_title.length();

        	title += m_title.substring(index1, index2);
            index1 = index2 + 1;
            if (index1 < m_title.length()) title += ("\\" + "n");
		}

        return title;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param title	The new title of the graph.
 */
    public void setTitle(String title) {

    	m_title = "";
        int index1 = 0;

        while (index1 < title.length()) {
        	int index2 = title.indexOf("\\n", index1);
            if (index2 < 0) index2 = title.length();

        	m_title += title.substring(index1, index2);
            index1 = index2 + 2;
            if (index1 < title.length()) m_title += '\n';
		}
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The length of major tick marks.
 */

	public double getMajorTickLen() {
    	return m_majorTickLen;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The tick scalar.
 */

	public double getTickSize() {
    	return m_tickSize;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param size	The new scalar to be used for tick length.
 */
    public void setTickSize(double size) {
        m_tickSize = size;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The proportion of major tick length to minor tick length.
 */

	public double getMinorTickProportion() {
    	return m_minorTickProportion;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return the number of major divisions.
 */

	public double getNumMajorDivs() {
    	return m_numMajorDivs;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return the number of minor divisions.
 */

	public double getNumMinorDivs() {
    	return m_numMinorDivs;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The zoom for this axis.
 */

	public Zoom getZoom() {
    	return m_zoom;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param limits The zoom for this axis.
 */

	public void setZoom(Zoom zoom) {

    	add(zoom);
        update();
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param limits The zoom for this axis.
 * @param oppLimits The zoom for the opposing axis.
 */

	public void setZoom(Zoom zoom, Zoom opposingZoom) {

    	add(zoom);
        m_opposingZoom = opposingZoom;
        update();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The zoom for the opposing axis.
 */

	public Zoom getOpposingZoom() {
    	return m_opposingZoom;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @parameter limits The new object to use for the opposing axis' limits.
 */

	public void setOpposingZoom(Zoom zoom) {

    	m_opposingZoom = zoom;
        update();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return a True if this object is being drawn for speed instead of looks.
 */
    public boolean getQuickDraw() {
    	return m_bQuickDraw;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bQuickDraw	True if the graph is to be drawn for speed at the
 *						expense of looks and accuracy.
 */
    public void setQuickDraw(boolean bQuickDraw) {
        m_bQuickDraw = bQuickDraw;
    }
}



