/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2002-02-04
 * Date started:    2001-02-16
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import gov.usgs.gr.data.*;
import gov.usgs.gr.*;
import java.awt.*;
import java.util.*;

//==============================================================================
/**
 * A property that provides a drop down choice of colors and a color chooser
 * dialog.
 */
public class ColorProperty extends gov.usgs.gr.GrProperty {

    protected float[] m_color = {1, 1, 1};
    /**
     * Predefined colors stored as RGB triples in float arrays.
     * The order is green, magenta, cyan, yellow, white, red.
     */
    protected float[][] m_colors = {{0, 1, 0}, {1, 0, 1}, {0, 1, 1}, {1, 1, 0},
    	{1, 1, 1}, {1, 0, 0}};
    /** Whether to draw the combo box in the color that is selected. */
    public boolean m_bUseRepresentativeColor = true;

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void createComponent() {

        // Give the user some color choices...
        final javax.swing.JComboBox colorCombo = new javax.swing.JComboBox();
		colorCombo.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent e) {
                float[] rgb = colorCombo.getBackground().getRGBComponents(null);
                String selection = (String)colorCombo.getSelectedItem();
				if (selection.equals("Other...")) {
					java.awt.Color c = javax.swing.JColorChooser.showDialog(
                    	GrController.getParentFrame(colorCombo),
                        "Choose a Color", new Color(rgb[0], rgb[1], rgb[2]));
                    if (c != null) {
                        rgb = c.getRGBComponents(null);
                    }

                    int index = colorCombo.getItemCount() - 1;
                    String s = "RGB (" + Math.round(rgb[0] * 255) + ", "
                    	+ Math.round(rgb[1] * 255) +  ", "
                        + Math.round(rgb[2] * 255) + ")";
                    colorCombo.insertItemAt(s, index);
					colorCombo.setSelectedIndex(index);
                }
                else {
                    rgb = getColorArray(selection);
                }

                // Make the combo box the same color as the selected color.
                if (m_bUseRepresentativeColor) {
                    colorCombo.setBackground(
                        new Color(rgb[0], rgb[1], rgb[2]));
                    if ((0.5 * rgb[0] + 1.4 * rgb[1] + 0.3 * rgb[2]) > 1)
                        colorCombo.setForeground(Color.black);
                    else
                        colorCombo.setForeground(Color.white);
                }
			}
		});

        String[] colorNames = {
        	"Green", "Magenta", "Cyan", "Yellow", "White", "Red"
        };

        // Fill the color combo box.
        boolean bFoundColor = false;
        for (int i = 0; i < colorNames.length; i++) {
        	String name = colorNames[i];
        	colorCombo.addItem(name);
            if (m_color[0] == m_colors[i][0] && m_color[1] == m_colors[i][1]
            	&& m_color[2] == m_colors[i][2]) {
				colorCombo.setSelectedIndex(i);
                this.value = colorNames[i];
                bFoundColor = true;
            }
        }

        if (!bFoundColor) {
            String s = "RGB (" + Math.round(m_color[0] * 255) + ", "
            	+ Math.round(m_color[1] * 255) +  ", "
                + Math.round(m_color[2] * 255) + ")";
            colorCombo.addItem(s);
        	colorCombo.setSelectedIndex(colorNames.length);
            this.value = s;
        }

        colorCombo.addItem("Other...");

        this.component = colorCombo;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public float[] getColorArray() {

        String s = ((String)this.value).trim().toLowerCase();
        return getColorArray(s);
    }

//------------------------------------------------------------------------------
/**
 * Translate a color name to a color RGB array.
 */
    public static float[] getColorArray(String value) {

        String[] colorNames = {
        	"black", "blue", "cyan", "darkgray", "gray", "green", "lightgray",
            "magenta", "orange", "pink", "red", "white", "yellow"
        };
        float[][] colorVals = {
        	{0f, 0f, 0f}, {0f, 0f, 1f}, {0f, 1f, 1f}, {0.25f, 0.25f, 0.25f},
            {0.5f, 0.5f, 0.5f}, {0f, 1f, 0f}, {0.75f, 0.75f, 0.75f}, {1, 0, 1},
            {1f, 0.5f, 0f}, {1f, 0.75f, 0.75f}, {1, 0, 0}, {1, 1, 1}, {1, 1, 0}
        };

        String s = value.trim().toLowerCase();
        // Check if it is a defined color name.
        boolean bColorFound = false;
        for (int i = 0; i < colorNames.length; i++) {
            if (s.equals(colorNames[i])) {
                return (colorVals[i]);
            }
        }

        // Parse it as (R, G, B) in the range 0-255.
        int i0 = s.indexOf('(');
        int i1 = s.indexOf(',', i0 + 1);
        int i2 = s.indexOf(',', i1 + 1);
        int i3 = s.indexOf(')', i2 + 1);
        if (i0 < 0) {
            float[] rgb = {1, 1, 1};
            return rgb;
        }

        int r = Integer.parseInt(
            s.substring(i0 + 1, i1).trim());
        int g = Integer.parseInt(
            s.substring(i1 + 1, i2).trim());
        int b = Integer.parseInt(
            s.substring(i2 + 1, i3).trim());
        Color c = new Color(r, g, b);
        if (c != null)
            return (c.getRGBComponents(null));
        else
            return (Color.white.getRGBComponents(null));
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param index	The index to use from the predefined colors array.
 */
    public void setColor(int index){

        if (index < 0 || index >= m_colors.length) index = 0;
        m_color = m_colors[index];

        createComponent();
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param red   The red components value between 0.0 and 1.0
 * @param green The green components value between 0.0 and 1.0
 * @param blue  The blue components value between 0.0 and 1.0
 */
    public void setColor(float red, float green, float blue){

    	m_color = new float[3];
        m_color[0] = red;
        m_color[1] = green;
        m_color[2] = blue;

        createComponent();
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param color An array of floats decribing the red, green, and blue.
 */
    public void setColor(float[] color){

        setColor(color[0], color[1], color[2]);

        createComponent();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return a copy of the array describing the RGB color of this curve.
 */
    public float[] getColor() {

    	// Create a new array and copy this curve's color into it.
    	float color[] = new float[3];
        color[0] = m_color[0];
        color[1] = m_color[1];
        color[2] = m_color[2];

    	return color;
    }
}






