/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-01
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import gov.usgs.gr.data.*;
import gov.usgs.gr.*;
import java.awt.geom.*;
import java.util.*;

//==============================================================================
/**
 * A single strip of data points that can draw itself.
 */
public abstract class Curve extends gov.usgs.gr.visual.GraphicalElement {

    /** The data that is visualized by this curve. */
    protected DoubleDataSeries m_series = null;
    /** A smaller version of m_series, optimized for drawing lines. */
    protected DoubleDataSeries m_reducedLineSeries = null;
    /** A smaller version of m_series, optimized for drawing symbols. */
    protected DoubleDataSeries m_reducedSymbolSeries = null;
    /**
     * An array of begin and end pairs describing where line segments should
     * begin and end.
     */
    protected int[] m_breaks = null;
    /** True if this curve should be drawn. */
//    protected boolean m_bShow = true;
    /** The color of this curve line in an RGB array. */
    protected float[] m_color = {0, 1, 0};
    /** The color to draw after applying the bleach multiplier. */
    protected float[] m_drawColor = {0, 1, 0};
    /**
     * Predefined colors stored as RGB triples in float arrays.
     * The order is green, magenta, cyan, yellow, white, red.
     */
    protected float[][] m_colors = {{0, 1, 0}, {1, 0, 1}, {0, 1, 1}, {1, 1, 0},
    	{1, 1, 1}, {1, 0, 0}};
    /** Predefined line patterns to use. */
    protected short[] m_linePatterns = {Gr.SOLID_LINE_PATTERN,
    	Gr.DASHED_LINE_PATTERN, Gr.DOTTED_LINE_PATTERN,
        Gr.DASH_DOT_LINE_PATTERN, Gr.LOOSE_DASHED_LINE_PATTERN,
        Gr.LOOSE_DOTTED_LINE_PATTERN, Gr.DASH_DOT_DASH_LINE_PATTERN,
        Gr.TIGHT_DASHED_LINE_PATTERN, Gr.TIGHT_DOTTED_LINE_PATTERN};
    /** The dash pattern of the line. */
    protected short m_linePattern = Gr.SOLID_LINE_PATTERN;
    /** The bleach factor for making the curve lighter or darker. */
    protected float m_bleach = 0;
    /** Whether this curve is being edited. */
    protected boolean m_bEditing = false;
    /** Whether to draw symbols for each point. */
    protected boolean m_bShowSymbols = false;
    /** Whether to draw a line connecting the points. */
    protected boolean m_bShowLine = true;
    /** Whether to draw the curve as a solid line instead of a dash pattern. */
    protected boolean m_bSolid = false;
    /**
     * The numeric code of the symbol to be used on this curve.
     *  0 = single pixel.
     *  1 = plus sign.
     *  2 = X
     *  3 = tick mark
     *  4 = hollow circle
     *  5 = hollow square
     *  6 = hollow upward-pointing triangle
     *  7 = hollow downward-pointing triangle
     *  8 = hollow diamond
     *  9 = filled circle
     * 10 = filled square
     * 11 = filled upward-pointing triangle
     * 12 = filled downward-pointing triangle
     * 13 = filled diamond
     */
    protected int m_symbolIndex = 0;
    /** The size of the symbol. */
    protected float m_symbolSize = 5;
    /** The size of editable points. */
    protected float m_editablePointSize = 5;
    /**
     * Whether to save time by drawing a simpler representation of the
     * graph.
     */
    protected boolean m_bQuickDraw = false;
    /** The thickness of the line in pixels. */
    protected float m_lineWidth = 1;
    /**
     * Whether the regular and reduced points series should be drawn over
     * each other to help debug.
     */
	protected boolean m_bDebugPointReduction = false;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/Curve.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public Curve(){

        setDefaults("gov.usgs.gr.visual.Curve");
    }

//------------------------------------------------------------------------------
/**
 * Copy an existing curve, but share the data.
 */
    public Curve(Curve curve){

    	// Copy the other curve's attributes.
    	m_series = curve.getSeries();

        m_children.add(m_series);

    	setProperties(curve.getProperties());


        m_reducedLineSeries = curve.getReducedLineSeries();
    	m_reducedSymbolSeries = curve.getReducedSymbolSeries();
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a curve from a data series.
 * @param series    The data that this curve will use.
 */
    public Curve(DoubleDataSeries series){

        add(series);
   }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.remove("showTitle");

        GrProperty p = (GrProperty)props.get("title");
        p.value = getTitle();

        props.put("showLine", new GrProperty(m_bShowLine, "Show Line"));
        props.put("lineWidth", new GrProperty(m_lineWidth, "Line Width"));

        // Give the user some line type choices...
        p = new GrProperty("", "Line Pattern");
        javax.swing.JComboBox jcb = new javax.swing.JComboBox();
        String[] patterns = {
        	"Solid", "Dashed", "Dotted", "Dash Dot", "Loose Dashed",
        	"Loose Dotted", "Dash Dot Dash", "Tight Dashed", "Tight Dotted"
            };

        // Fill the line pattern combo box.
        for (int i = 0; i < patterns.length; i++) {
        	String name = patterns[i];
        	jcb.addItem(name);
            if (m_linePatterns[i] == m_linePattern) {
				jcb.setSelectedIndex(i);
                p.value = patterns[i];
            }
        }

        p.component = jcb;
		props.put("linePattern", p);

        props.put("showSymbols", new GrProperty(m_bShowSymbols,
            "Show Symbols"));
        props.put("symbolSize", new GrProperty(m_symbolSize, "Symbol Size"));

        // Give the user some symbol choices...
        p = new GrProperty("", "Symbol Type");
        jcb = new javax.swing.JComboBox();

        String[] symbolTypes = {
        	"Single Pixel", "Plus Sign", "X", "Tick Mark",
            "Hollow Circle", "Hollow Square", "Hollow Triangle Pointing Up",
            "Hollow Triangle Pointing Down", "Hollow Diamond",
            "Filled Circle", "Filled Square", "Filled Triangle Pointing Up",
            "Filled Triangle Pointing Down", "Filled Diamond"
            };

        // Fill the symbol type combo box.
        for (int i = 0; i < symbolTypes.length; i++) {
        	String name = symbolTypes[i];
        	jcb.addItem(name);
            if (m_symbolIndex == i) {
				jcb.setSelectedIndex(i);
                p.value = symbolTypes[i];
            }
        }

        p.component = jcb;
		props.put("symbolType", p);

        ColorProperty cp = new ColorProperty();
        cp.description = "Color";
        cp.setColor(m_color);
        props.put("color", cp);

        p = new GrProperty(m_bleach, "Bleach");
        p.display = false;
        p.save = false;
        props.put("bleach", p);

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);

            try {

            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }

            	if (key.equals("showLine"))
            		m_bShowLine = (Boolean)p.value;
            	if (key.equals("lineWidth"))
            		m_lineWidth = (Float)p.value;
            	if (key.equals("showSymbols"))
            		m_bShowSymbols = (Boolean)p.value;
            	if (key.equals("symbolSize"))
            		m_symbolSize = (Float)p.value;
            	if (key.equals("color")) {
                    ColorProperty cp = (ColorProperty)p;
                    setColor(cp.getColorArray());
                }
            	if (key.equals("bleach")) {
            		m_bleach = (Float)p.value;
                    updateDrawColor();
                }
            }
            if (obj instanceof String) {
            	String value = (String)obj;
                setProperty(key, value);
			}
            }
            catch (Exception e) {}
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("showLine"))
        	m_bShowLine = Boolean.parseBoolean(value);
        if (key.equals("lineWidth"))
        	m_lineWidth = Float.parseFloat(value);

        if (key.equals("linePattern")) {
            String[] patterns = {
                "solid", "dashed", "dotted", "dash dot",
                "loose dashed", "loose dotted", "dash dot dash",
                "tight dashed", "tight dotted"
            };

            // Check each name.
            for (int i = 0; i < patterns.length; i++) {
                String name = value.trim().toLowerCase();
                if (name.equals(patterns[i])) {
                    m_linePattern = m_linePatterns[i];
                    break;
                }
            }
        }

        if (key.equals("title")) {
            if (m_series != null)
                m_series.setProperty("title", value);
        }
    	if (key.equals("showSymbols"))
        	m_bShowSymbols = Boolean.parseBoolean(value);
        if (key.equals("symbolSize"))
        	m_symbolSize = Float.parseFloat(value);
        if (key.equals("editablePointSize"))
        	m_editablePointSize = Float.parseFloat(value);

        if (key.equals("symbolType")) {
        	String[] symbolTypes = {
        		"single pixel", "plus sign", "x", "tick mark", "hollow circle",
            	"hollow square", "hollow triangle pointing up",
            	"hollow triangle pointing down", "hollow diamond",
                "filled circle", "filled square", "filled triangle pointing up",
            	"filled triangle pointing down", "filled diamond"
        	};
        	// Check each name.
            for (int i = 0; i < symbolTypes.length; i++) {
            	String name = value.trim().toLowerCase();
                if (name.equals(symbolTypes[i])) {
                	m_symbolIndex = i;
                    break;
                }
            }
        }
        if (key.equals("color")) {
            setColor(ColorProperty.getColorArray(value));
        }
        if (key.equals("bleach")) {
        	m_bleach = Float.parseFloat(value);
            updateDrawColor();
        }

        m_bNeedsRedraw = true;
    }

//------------------------------------------------------------------------------
/**
 * Add an object to the curve.  Only accepts two-dimensional DoubleDataSeries. 
 * Only holds one at a time.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

    	if (DoubleDataSeries.countDims(gro, 2)) {
        	clear();
        	super.add(gro);
			setSeries((DoubleDataSeries)gro);

            return true;
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param index	The index to use from the predefined colors array.
 */
    public void setColor(int index){

        if (index < 0 || index >= m_colors.length) index = 0;
        m_color = m_colors[index];

        // Update the draw color by bleaching the new main color.
        updateDrawColor();
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param index	The index to use from the predefined line patterns array.
 */
    public void setLinePattern(int index){

        if (index < 0 || index >= m_linePatterns.length) index = 0;
        m_linePattern = m_linePatterns[index];
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bShow    True if this curve should be drawn.
 */
    public void setShow(boolean bShow){
        m_bShow = bShow;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bShow    True if symbols should be drawn for each point.
 */
    public void setShowSymbols(boolean bShowSymbols){
        m_bShowSymbols = bShowSymbols;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return True if the curve is being edited.
 */
    public boolean getEditing(){
        return m_bEditing;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bEditing    True if the curve is being edited.
 */
    public void setEditing(boolean bEditing){
        m_bEditing = bEditing;
    }

//------------------------------------------------------------------------------
/**
 * Draw the data as a line.  This method should be overridden to create
 * output for a specific device.
 * <p><b> Pre Conditions </b><br>
 *      The origin has been moved to (0, 0), scaling has occurred, and
 *      line type is solid.
 */
    public void draw(){
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param startVal  The x value to start at.
 * @param endVal    The x value to end at.
 */
    public void setLimits(double startVal, double endVal) {
        m_series.setLimits(startVal, endVal);
	}

//------------------------------------------------------------------------------
/**
 * Update the draw color by bleaching the main color.
 */
    public void updateDrawColor(){

        if (m_bleach == 0) {
            m_drawColor[0] = m_color[0];
            m_drawColor[1] = m_color[1];
            m_drawColor[2] = m_color[2];
        }
        if (m_bleach > 0) {
        	for (int i = 0; i < 3; i++) {
        		m_drawColor[i] = (1 - m_color[i]) * m_bleach + m_color[i];
        	}
        }
        if (m_bleach < 0) {
        	for (int i = 0; i < 3; i++) {
        		m_drawColor[i] = m_color[i] * m_bleach + m_color[i];
        	}
        }
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param red   The red components value between 0.0 and 1.0
 * @param green The green components value between 0.0 and 1.0
 * @param blue  The blue components value between 0.0 and 1.0
 */
    public void setColor(float red, float green, float blue){

    	m_color = new float[3];
        m_color[0] = red;
        m_color[1] = green;
        m_color[2] = blue;

        // Update the draw color by bleaching the new main color.
        updateDrawColor();
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param color An array of floats decribing the red, green, and blue.
 */
    public void setColor(float[] color){
        setColor(color[0], color[1], color[2]);
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param linePattern   A predefined number representing a GL line pattern.
 */
    public void setLinePattern(short linePattern){
        m_linePattern = linePattern;
        if (linePattern == Gr.SOLID_LINE_PATTERN) m_bSolid = true;
        else m_bSolid = false;
    }

//------------------------------------------------------------------------------
/**
 * Read the attribute settings from a values stored in a list.
 * <p><b> Post Conditions </b><br>
 *      update() needs to be called on this axis and the opposing axis.
 * @param attribs	A list of attribute values.
 */
    public void setAttributes(List attribs) {

        if (attribs.size() != 10) return;

        // Color...
        m_color[0] = (Float)attribs.get(0);
        m_color[1] = (Float)attribs.get(1);
        m_color[2] = (Float)attribs.get(2);
        // Clamp the value between 0 and 1.
        for (int i = 0; i < 3; i++) {
        	if (m_color[i] < 0) m_color[i] = 0;
            if (m_color[i] > 1) m_color[i] = 1;
        }

        // Lines...
        m_bSolid = Boolean.parseBoolean((String)attribs.get(3));
        m_bShowLine = Boolean.parseBoolean((String)attribs.get(4));
        int linePatternIndex = ((Float)attribs.get(5)).intValue();
        if (linePatternIndex >= m_linePatterns.length) {
			linePatternIndex = 0;
        }
        m_linePattern = m_linePatterns[linePatternIndex];
        m_lineWidth = (Float)attribs.get(6);

        // Symbols...
        m_bShowSymbols = Boolean.parseBoolean((String)attribs.get(7));
        if (!m_bShowLine && !m_bShowSymbols) m_bShowLine = true;
        m_symbolIndex = ((Float)attribs.get(8)).intValue();
        m_symbolSize = (Float)attribs.get(9);

 	}

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = null;
        String title = null;
        if (m_series != null)
            title = m_series.getTitle();
        if (m_bShow)
            s = "Curve: " + title;
        else
            s = "(Curve): " + title;

        return s;
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum value in the data series for the specified dimension.
 * @param dimension The dimension to report the minimum.
 */
    public double getMin(int dimension){

        return m_series.getMin(dimension);
    }

//------------------------------------------------------------------------------
/**
 * Find the maximum value in the data series for the specified dimension.
 * @param dimension The dimension to report the minimum.
 */
    public double getMax(int dimension){

        return m_series.getMax(dimension);
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.  null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        if (!m_bShow || m_series == null) return null;

        Point2D.Double p = new Point2D.Double(m_series.getMin(dimension),
            m_series.getMax(dimension));

        return p;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The title string that describes this series.
 */
    public String getTitle(){
        if (m_series != null) return m_series.getTitle();
        else return " ";
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param The new title of the data series.
 */
    public void setTitle(String title){
        if (m_series != null) m_series.setTitle(title);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return the data series for this curve.
 */
    public DoubleDataSeries getSeries() {
    	return m_series;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param the data series for this curve.
 */
    public void setSeries(DoubleDataSeries series) {
    	m_series = series;
		m_reducedLineSeries = m_series;
		m_reducedSymbolSeries = m_series;

        m_children.clear();
        m_children.add(series);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return the reduced line data series for this curve.
 */
    public DoubleDataSeries getReducedLineSeries() {
    	return m_reducedLineSeries;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param the reduced line data series for this curve.
 */
    public void setReducedLineSeries(DoubleDataSeries series) {
		m_reducedLineSeries = series;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return the reduced symbol data series for this curve.
 */
    public DoubleDataSeries getReducedSymbolSeries() {
    	return m_reducedSymbolSeries;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param the reduced symbol data series for this curve.
 */
    public void setReducedSymbolSeries(DoubleDataSeries series) {
		m_reducedSymbolSeries = series;
    }

//------------------------------------------------------------------------------
/**
 * Sets array describing the beginning and ending points of the line segments.
 * @param m_breaks  The array of begin and end pairs describing line segments.
 */
    public void setSegmentIndexes(int[] breaks) {

        m_breaks = breaks;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return true if this curve should be shown.
 */
    public boolean getShow() {
    	return m_bShow;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return true if a line will be drawn through the points in the curve.
 */
    public boolean getShowLine() {
    	return m_bShowLine;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bShow    True if line should be drawn between each point.
 */
    public void setShowLine(boolean val){
        m_bShowLine = val;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return true if symbols will be drawn for the points in the curve.
 */
    public boolean getShowSymbols() {
    	return m_bShowSymbols;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The index of the type of symbol to be used.
 */
    public int getSymbolIndex() {
    	return m_symbolIndex;
    }
//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param index	The index of the symbol to use.
 */
    public void setSymbolIndex(int index) {
    	m_symbolIndex = index;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The relative size of the sybol to be used.
 */
    public float getSymbolSize() {
    	if (m_symbolIndex == 0) return 1;
        else return m_symbolSize;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param val	The new symbol size.
 */
    public void setSymbolSize(float val) {
        m_symbolSize = val;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return a copy of the array describing the RGB color of this curve.
 */
    public float[] getColor() {

    	// Create a new array and copy this curve's color into it.
    	float color[] = new float[3];
        color[0] = m_color[0];
        color[1] = m_color[1];
        color[2] = m_color[2];

    	return color;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return the line pattern for this curve.
 */
    public short getLinePattern() {
    	return m_linePattern;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return the data series for this curve.
 */
    public boolean getSolid() {
    	return m_bSolid;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return the data series for this curve.
 */
    public float getLineWidth() {
    	return m_lineWidth;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param val	The new line width.
 */
    public void setLineWidth(float val) {
        m_lineWidth = val;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return a True if this object is being drawn for speed instead of looks.
 */
    public boolean getQuickDraw() {
    	return m_bQuickDraw;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bQuickDraw	True if the graph is to be drawn for speed at the
 *						expense of looks and accuracy.
 */
    public void setQuickDraw(boolean bQuickDraw) {
        m_bQuickDraw = bQuickDraw;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	Whether curves should be drawn to help debug point reduction.
 */
    public void setDebugPointReduction(boolean value) {
        m_bDebugPointReduction = value;
    }

}






