/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-29
 * Date started:    2001-12-17
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.*;
import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import javax.media.opengl.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class DataAnimation
    extends gov.usgs.gr.visual.GraphicalElement {

    protected double[][] m_dataArrays;
    protected int[] m_seriesNums;
    protected int[] m_dimNums;
    protected int[] m_seriesNumDims;
    protected String m_dimNames = "T,X,Y,Z,U,V,W,C";
    protected double[] m_dimScales;
    protected DoubleDataSeries[] m_seriesArray = null;
    protected DoubleDataSeries m_constantSeries = null;

    protected int m_numDimsAllowed = 0;

    protected String[] m_symbolTypes;

    protected double m_scale = 1;
    /** The color to use when drawing. */
    protected float[] m_color = {1, 1, 1};
    protected ColorScheme m_colorScheme = null;
    // How to scale the vertical locations.
    protected double m_zScale = 0;
    // How much to vertically exagerate objects.
    protected double m_zStretch = 1;
    // Whether to draw the width as a constant.  <=Zero indicates variable.
    protected double m_constantWidth = 0;
    /**
     * The numeric code of the symbol to be used.
     */
    protected int m_symbolIndex = -1;
    /**
     * Holds any error descriptions that should be shown to the user in
     * the properties dialog.
     */
    protected String m_errorString = null;

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public DataAnimation(){

        setDefaults("gov.usgs.gr.visual.DataAnimation");
    }

//------------------------------------------------------------------------------
    /**
     * Constructor for creating a curve from a data series.
     *
     * @param series DataSeries
     */
    public DataAnimation(DataSeries series){

        setDefaults("gov.usgs.gr.visual.DataAnimation");
   }

//------------------------------------------------------------------------------
    /**
     * Add a child object. Accepts DataSeries, TimeMarker. Only holds one DS at
     * a time.
     *
     * @param gro GrObject
     * @return boolean
     */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        if (m_children.contains(gro)) return false;

        if (gro instanceof DoubleDataSeries) {
            super.add(gro);

            resetDataDims();

            return true;
        }
        if (gro instanceof IntDataSeries) {
            super.add(gro);

            return true;
        }
        if (gro instanceof ColorScheme) {
            if (m_colorScheme != null)
                remove((GrObject)m_colorScheme);
            super.add(gro);
            m_colorScheme = (ColorScheme)gro;

            return true;
        }
        if (gro instanceof Markup) {
            super.add(gro);

            return true;
        }

        return false;
    }

//------------------------------------------------------------------------------
    /**
     * Wrapper for removing items from the list of children and resetting data
     * dimensions afterward.
     *
     * @param gro GrObject
     * @return boolean
     */
    public boolean remove(GrObject gro) {

        // Let the super class handle removal.
        boolean bFound = super.remove(gro);
        resetDataDims();

        return bFound;
    }

//------------------------------------------------------------------------------
    /**
     * Wrapper for removing items from the list of children and resetting data
     * dimensions afterward.
     *
     * @param i int
     * @return GrObject
     */
    public GrObject remove(int i) {

        // Let the super class handle removal.
        GrObject child = super.remove(i);
        resetDataDims();

        return child;
    }

//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values. The numeric code
     * of the symbol to be used. 0 = simple arrow. 1 = arrow with black inner
     * area. 2 = bar
     *
     * @return GrHashList
     */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("scale", new GrProperty(m_scale, "Scale"));
        props.put("zScale", new GrProperty(m_zScale, "Z Scale"));
        props.put("zStretch", new GrProperty(m_zStretch, "Z Stretch"));
        props.put("constantWidth", new GrProperty(m_constantWidth,
            "Constant Width"));

        String s = "";
        for (int i = 0; i < m_seriesNums.length; i++) {
            if (i > 0) s += ", ";
            s += m_seriesNums[i];
        }
        props.put("seriesNums", new GrProperty(s,
            "Series Numbers for " + m_dimNames));

        s = "";
        for (int i = 0; i < m_dimNums.length; i++) {
            if (i > 0) s += ", ";

            // If a constant is defined for this dimension, write it instead.
            if (m_dimNums[i] >= 0 && m_seriesNums[i] >= 0 &&
		m_seriesNums[i] < m_seriesNums.length &&
                m_seriesArray[m_seriesNums[i]] == m_constantSeries) {

                s += "=";
                double[] array = m_constantSeries.getArray();
                s += array[i];
            }
            else {
                s += m_dimNums[i];
            }
        }
        props.put("dimNums", new GrProperty(s,
            "Dimension Numbers for " + m_dimNames));

        if (m_symbolTypes != null) {
            // Give the user some symbol choices...
            GrProperty p = new GrProperty("", "Symbol Type");
            javax.swing.JComboBox jcb = new javax.swing.JComboBox();

            // Fill the symbol type combo box.
            for (int i = 0; i < m_symbolTypes.length; i++) {
                String name = m_symbolTypes[i];
                jcb.addItem(name);
                if (m_symbolIndex == i) {
                    jcb.setSelectedIndex(i);
                    p.value = m_symbolTypes[i];
                }
            }

            p.component = jcb;
            props.put("symbolType", p);
        }

        ColorProperty cp = new ColorProperty();
        cp.description = "Color";
        cp.setColor(m_color);
        props.put("color", cp);

        return props;
    }

//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     *
     * @param props HashMap
     */
    public void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        // If the series and dim nums follow the old convention,
        // convert them to the new convention.
        convertIfOldConvention(props);

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }

                if (key.equals("scale"))
                    m_scale = (Double)p.value;
                if (key.equals("zScale"))
                    m_zScale = (Double)p.value;
                if (key.equals("zStretch"))
                    m_zStretch = (Double)p.value;
                if (key.equals("constantWidth"))
                    m_constantWidth = (Double)p.value;
                if (key.equals("symbolType")) {
                    if (m_symbolTypes == null)
                        break;

                    // Check each name.
                    for (int i = 0; i < m_symbolTypes.length; i++) {
                        String name = ((String)p.value).trim().toLowerCase();
                        if (name.equals(m_symbolTypes[i].toLowerCase())) {
                            m_symbolIndex = i;
                            break;
                        }
                    }
                }
                if (key.equals("color")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_color = cp.getColorArray();
                }
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
    /**
     * Set a single property.
     *
     * @param key String
     * @param value String
     */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("scale"))
            m_scale = Double.parseDouble(value);
        if (key.equals("zScale"))
            m_zScale = Double.parseDouble(value);
        if (key.equals("zStretch"))
            m_zStretch = Double.parseDouble(value);
        if (key.equals("constantWidth"))
            m_constantWidth = Double.parseDouble(value);
        if (key.equals("seriesNums")) {
            int[] nums = AsciiIO.parseLineAsInts(value, ", \t");
            if (nums.length == m_numDimsAllowed)
                m_seriesNums = nums;
            else {
                if (m_seriesNums.length != m_numDimsAllowed)
                    m_seriesNums = new int[m_numDimsAllowed];
                for (int i = 0; i < m_numDimsAllowed; i++) {
                    if (i < nums.length && nums[i] <= m_numDimsAllowed)
                        m_seriesNums[i] = nums[i];
                    else
                        m_seriesNums[i] = 0;
                }
            }
        }
        if (key.equals("dimNums")) {
            // Make sure the series num array is the correct length.
            if (m_dimNums.length != m_numDimsAllowed)
                m_dimNums = new int[m_numDimsAllowed];

            // Process each token.
            String[] tokens = AsciiIO.parseLine(value, ", \t");
            for (int i = 0; i < tokens.length; i++) {
                if (i >= m_numDimsAllowed)
                    break;

                // Check if a constant is defined.
                int startIndex = tokens[i].indexOf('=', 0);
                if (startIndex >= 0) {
                    String constString = tokens[i].substring(startIndex + 1);

                    double val = 0;
                    try {
                        val = Double.parseDouble(constString);
                    }
                    catch (NumberFormatException e) {}

                    m_constantSeries.selectPoint(0);
                    m_constantSeries.modifySelectedPoints(0, i, val);
                    m_constantSeries.deselectAll();
                    m_dimNums[i] = i;
                }
                else {
                    // Set the series num to the integer value in the token.
                    int iVal = -1;
                    try {
                        iVal = Integer.parseInt(tokens[i]);
                    }
                    catch (NumberFormatException e) {}

                    m_dimNums[i] = iVal;
                }
            }

            // Zero out all series nums past the last token.
            for (int i = tokens.length; i < m_numDimsAllowed; i++) {
                m_dimNums[i] = -1;
            }





        }
        if (key.equals("symbolType")) {
            if (m_symbolTypes == null)
                return;

            // Check each name.
            for (int i = 0; i < m_symbolTypes.length; i++) {
                String name = value.trim().toLowerCase();
                if (name.equals(m_symbolTypes[i].toLowerCase())) {
                    m_symbolIndex = i;
                    break;
                }
            }
        }
        if (key.equals("color")) {
            m_color = ColorProperty.getColorArray(value);
        }
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        if (!m_bShow) return null;

        double min = Double.POSITIVE_INFINITY;
        double max = Double.NEGATIVE_INFINITY;
        try {
            if (dimension >= 0 && dimension <= 2) {
                int dim = dimension + 1;
                double locMin = m_seriesArray[m_seriesNums[dim]].
                    getMin(m_dimNums[dim]) * m_dimScales[dim];
                double locMax = m_seriesArray[m_seriesNums[dim]].
                    getMax(m_dimNums[dim]) * m_dimScales[dim];

                min = Math.min(min, locMin);
                max = Math.max(max, locMax);
            }
            else {
                return null;
            }
        }
        catch (NullPointerException e) {}
        catch (ArrayIndexOutOfBoundsException e) {}

        if (min == Double.POSITIVE_INFINITY|| max == Double.NEGATIVE_INFINITY)
            return null;

        Point2D.Double p = new Point2D.Double(min, max);

        return p;
    }

//------------------------------------------------------------------------------
/**
 * Reread the DataSeries children and set the DataSeries and dimension for
 * Time, X, Y, Z, U, V, and W.
 */
    public void resetDataDims() {

        // Add the constant series in the 0 slot.
        m_dataArrays[0] = m_constantSeries.getArray();
        m_seriesNumDims[0] = m_constantSeries.getNumDimensions();
        m_seriesArray[0] = m_constantSeries;

        // Locate child data series and store their info in the
        // member arrays.
        int dsIndex = 1;
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            if (dsIndex > m_numDimsAllowed) break;
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof DoubleDataSeries) {
                DoubleDataSeries series = (DoubleDataSeries)gro;
                m_dataArrays[dsIndex] = series.getArray();
                m_seriesNumDims[dsIndex] = series.getNumDimensions();
                m_seriesArray[dsIndex] = series;

                dsIndex++;
            }
        }

        // Blank out the remaining values.
        for (int i = dsIndex; i < m_numDimsAllowed; i++) {
            m_dataArrays[i] = null;
            m_seriesNumDims[i] = 0;
            m_seriesArray[i] = null;
            m_dimScales[i] = 1;
            dsIndex++;
        }

        // Set the dimension scales.
        for (int i = 0; i < m_numDimsAllowed; i++) {
            // Skip if the series is null.
            DataSeries series = null;
            if (m_seriesNums[i] >= 0 && m_seriesNums[i] < m_seriesArray.length)
                series = m_seriesArray[m_seriesNums[i]];
            if (series == null) {
                m_dimScales[i] = 1;
                continue;
            }

            // Skip if the dimension is null.
            DataDimension dim = series.getDimension(m_dimNums[i]);
            if (dim == null) {
                m_dimScales[i] = 1;
                continue;
            }

            // Store the scale factor for the dimension.
            GrHashList dimProps = dim.getProperties();
            GrProperty p = (GrProperty)dimProps.get("unitScale");
            double unitScale = (Double)p.value;
            m_dimScales[i] = unitScale;
        }

    }

//------------------------------------------------------------------------------
/**
 * Hide or show this object.
 * @param  bShow    True if this object should be drawn.
 */
    public void show(boolean bShow){
        m_bShow = bShow;
    }

//------------------------------------------------------------------------------
    /**
     *
     * @param seriesNum int
     * @return int
     */
    protected int pointCount(int seriesNum) {

        try {
            return (m_seriesArray[m_seriesNums[seriesNum]].getCount());
        }
        catch (IndexOutOfBoundsException e) {}

        return 0;
    }

//------------------------------------------------------------------------------
    /**
     *
     * @param seriesNum1 int
     * @param seriesNum2 int
     * @return int
     */
    protected int compareCounts(int seriesNum1, int seriesNum2) {

        return (pointCount(seriesNum1) - pointCount(seriesNum2));
    }

//------------------------------------------------------------------------------
    /**
     *
     * @param timeNum int
     * @param dimension int
     * @return double
     */
    protected double getValueAt(int timeNum, int dimension) {

        double val = 0;
        int seriesNum = m_seriesNums[dimension];
        double[] dataArray = m_dataArrays[seriesNum];
        int numPoints = dataArray.length / m_seriesNumDims[seriesNum];

        if (timeNum >= numPoints)
            timeNum = timeNum % numPoints;

        try {
            int numDims = m_seriesNumDims[seriesNum];
            int dimNum = m_dimNums[dimension];
            val = dataArray[timeNum * numDims + dimNum];
            val *= m_dimScales[dimension];
        }
        catch(Exception e) {}

        return val;
    }

//------------------------------------------------------------------------------
    /**
     *
     * @param timeNum int
     * @param dimension int
     * @param dimNum int
     * @return double
     */
    protected double getValueAt(int timeNum, int dimension, int dimNum) {

        double val = 0;
        int seriesNum = m_seriesNums[dimension];

        int numDims = m_seriesNumDims[seriesNum];
        int numPoints = m_dataArrays[seriesNum].length / numDims;

        if (timeNum >= numPoints)
            timeNum = timeNum % numPoints;

        try {
            val = m_dataArrays[seriesNum][timeNum * numDims + dimNum];
            val *= m_dimScales[dimension];
        }
        catch(Exception e) {}

        return val;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        // Check if the object is a graphical element that needs to be redrawn.
        if (gro instanceof DataSeries) {
            m_bNeedsRedraw = true;
            resetDataDims();
        }

        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Change the color of the curve.
 * @param red   The red components value between 0.0 and 1.0
 * @param green The green components value between 0.0 and 1.0
 * @param blue  The blue components value between 0.0 and 1.0
 */
    public void setColor(float red, float green, float blue){
        m_color[0] = red;
        m_color[1] = green;
        m_color[2] = blue;
    }

//------------------------------------------------------------------------------
/**
 * Change the color of the curve (see other setColor routine).
 * @param color An array of floats decribing the red, green, and blue.
 */
    public void setColor(float[] color){
        setColor(color[0], color[1], color[2]);
    }

//------------------------------------------------------------------------------
    /**
     *
     * @param props HashMap
     */
    public void convertIfOldConvention(HashMap props){

        GrProperty p;
        String seriesString = null;
        String dimString = null;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            if (key.equals("seriesNums")) {
                Object obj = props.get(key);
                if (obj instanceof String)
                    seriesString = (String)obj;
            }
            if (key.equals("dimNums")) {
                Object obj = props.get(key);
                if (obj instanceof String)
                    dimString = (String)obj;
            }
        }

        if (seriesString == null)
            return;

        int[] seriesInts = AsciiIO.parseLineAsInts(seriesString, ", \t");
        String[] dimTokens;

        if (dimString == null) {
            dimTokens = new String[seriesInts.length];
            for (int i = 0; i < dimTokens.length; i++)
                dimTokens[i] = "0";
        }
        else {
            dimTokens= AsciiIO.parseLine(dimString, ", \t");
        }

        // Determine if conversion needed.
        boolean bConversionNeeded = false;
        for (int i = 0; i < seriesInts.length; i++) {
            if (i >= dimTokens.length)
                break;

            if (seriesInts[i] == 0) {
                // Check if a constant is defined.
                int startIndex = dimTokens[i].indexOf('=', 0);
                if (startIndex < 0) {
                    int iVal = -1;
                    try {
                        iVal = Integer.parseInt(dimTokens[i]);
                    }
                    catch (NumberFormatException e) {}

                    if (iVal >= 0) {
                        bConversionNeeded = true;
                        break;
                    }
                }
            }
        }

        if (!bConversionNeeded)
            return;

        for (int i = 0; i < seriesInts.length; i++) {
            if (i >= dimTokens.length)
                break;

            if (seriesInts[i] >= 0) {
                int iVal = -1;
                try {
                    iVal = Integer.parseInt(dimTokens[i]);
                }
                catch (NumberFormatException e) {}

                if (iVal < 0)
                    seriesInts[i] = 0;
                else
                    seriesInts[i]++;
            }
        }

        String s = "";
        for (int i = 0; i < seriesInts.length; i++) {
            if (i > 0) s += ", ";
            s += seriesInts[i];
        }

        props.put("seriesNums", s);
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){
        String s = "Data Animation: ";

        if (m_errorString != null)
            s += m_errorString;
        else
            s += m_title;

        return s;
    }
}

