/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2000-09-15
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import gov.usgs.gr.*;
import java.util.*;

//==============================================================================
/**
 * Calculates a color from a value based on the formula
 * colorComponent[i] = (value - offset[i]) / range[i]
 * for each of the color components Red, Green, and Blue.
 */
public class FormulaColorScheme extends gov.usgs.gr.GrObject
	implements gov.usgs.gr.visual.ColorScheme {

    /** The input min value array for Red, Green, and Blue. */
    protected float[] m_offset = {0, 0, 0};
    /** The input range array for Red, Green, and Blue. */
    protected float[] m_range = {100, 100, 100};
    /** The amount to bleach the color toward white. */
    protected float m_bleach = 0;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/FormulaColorScheme.gif";

//------------------------------------------------------------------------------
/**
 * Construct an empty object.
 */
    public FormulaColorScheme () {

        setDefaults("gov.usgs.gr.visual.FormulaColorScheme");
    }

//------------------------------------------------------------------------------
/**
 * Copy Constructor.
 */
    public FormulaColorScheme (FormulaColorScheme fcs) {

    	setProperties(fcs.getProperties());
    }

//------------------------------------------------------------------------------
/**
 * Implements getColor from ColorScheme interface.
 */
	public float[] getColor(double val) {

    	/** The output color array for Red, Green, and Blue. */
    	float[] color = new float[3];

        for (int i = 0; i < 3; i++) {
        	color[i] = ((float)val - m_offset[i]) / m_range[i];
        }

        if (m_bleach > 0) {
        	for (int i = 0; i < 3; i++) {
        		color[i] = (1 - color[i]) * m_bleach + color[i];
        	}
        }

        return color;
	}

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Formula Color Scheme";
        return s;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("redOffset", new GrProperty(m_offset[0], "Red Offset"));
        props.put("redRange", new GrProperty(m_range[0], "Red Range"));
        props.put("greenOffset", new GrProperty(m_offset[1], "Green Offset"));
        props.put("greenRange", new GrProperty(m_range[1], "Green Range"));
        props.put("blueOffset", new GrProperty(m_offset[2], "Blue Offset"));
        props.put("blueRange", new GrProperty(m_range[2], "Blue Range"));
        props.put("bleach", new GrProperty(m_bleach, "Bleach (0-1)"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("redOffset"))
            		m_offset[0] = (Float)p.value;
            	if (key.equals("greenOffset"))
            		m_offset[1] = (Float)p.value;
            	if (key.equals("blueOffset"))
            		m_offset[2] = (Float)p.value;
            	if (key.equals("redRange"))
            		m_range[0] = (Float)p.value;
            	if (key.equals("greenRange"))
            		m_range[1] = (Float)p.value;
            	if (key.equals("blueRange"))
            		m_range[2] = (Float)p.value;
            	if (key.equals("bleach"))
            		m_bleach = (Float)p.value;
			}
            if (obj instanceof String) {
            	String value = (String)obj;

            	if (key.equals("redOffset"))
            		m_offset[0] = Float.parseFloat(value);
            	if (key.equals("greenOffset"))
            		m_offset[1] = Float.parseFloat(value);
            	if (key.equals("blueOffset"))
            		m_offset[2] = Float.parseFloat(value);
            	if (key.equals("redRange"))
            		m_range[0] = Float.parseFloat(value);
            	if (key.equals("greenRange"))
            		m_range[1] = Float.parseFloat(value);
            	if (key.equals("blueRange"))
            		m_range[2] = Float.parseFloat(value);
            	if (key.equals("bleach"))
            		m_bleach = Float.parseFloat(value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("redOffset"))
        	m_offset[0] = Float.parseFloat(value);
        if (key.equals("greenOffset"))
        	m_offset[1] = Float.parseFloat(value);
        if (key.equals("blueOffset"))
        	m_offset[2] = Float.parseFloat(value);
        if (key.equals("redRange"))
        	m_range[0] = Float.parseFloat(value);
        if (key.equals("greenRange"))
        	m_range[1] = Float.parseFloat(value);
        if (key.equals("blueRange"))
        	m_range[2] = Float.parseFloat(value);
        if (key.equals("bleach"))
        	m_bleach = Float.parseFloat(value);
    }
}




