/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-01
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import gov.usgs.gr.data.*;
import gov.usgs.gr.*;
import java.awt.Rectangle;
import java.awt.geom.*;
import java.util.*;

//==============================================================================
/**
 * An x-y graph with ticks, labels, a grid, and multiple curves drawn inside.
 * This class and the components withing in it should be extended for drawing
 * to a particular output device.
 */
public abstract class Graph extends gov.usgs.gr.visual.GraphicalElement {
    
    /** A list of curves to be drawn in this graph. */
    protected List<Curve> m_curves =
	Collections.synchronizedList(new ArrayList());
    /** A list of data series in this graph that are selected. */
    protected List m_selectedSeriesList =
	Collections.synchronizedList(new ArrayList());
    /** A list of miscellaneous graphical objects that are drawn on the page. */
    protected List m_markupList = Collections.synchronizedList(new ArrayList());
    /** Whether to show labels on the x-axis. */
    protected boolean m_bShowXtext = false;
    /** Whether to show a bar along the side. */
    protected boolean m_bShowSideBar = false;
    /** Whether to draw the side bar down. */
    protected boolean m_bSideBarDown = false;
    /** The scale to use to make ticks the right lenghth on the output. */
    protected double m_xTickScale = 1;
    /** The scale to use to make ticks the right lenghth on the output. */
    protected double m_yTickScale = 1;
    /** The y-axis object, which can draw itself. */
    protected Axis m_yAxis;
    /** The x-axis object, which can draw itself. */
    protected Axis m_xAxis;
    /**
     * The axis limits descriptor for the x-axis.  This is shared with other
     * Graphs.
     */
    public Zoom m_xZoom;
    /**
     * The  axis limits descriptor for the y-axis.   It is only used by
     * this graph.
     */
    protected Zoom m_yZoom = new Zoom();
    /** Whether to keep a 1:1 aspect ratio between the X and Y axes. */
    protected boolean m_bAspect1to1 = false;
    /** The color of the frame around the graph, in an RGB array. */
    protected float[] m_frameColor = {1, 1, 0};
    /** The color of the text on the graph, in an RGB array. */
    protected float[] m_textColor = {1f, 0.5f, 1f};
    /** The color of the bottom background of the graph, in an RGB array. */
    protected float[] m_topBgColor = {0f, 0f, 0f};
    /** The color of the bottom background of the graph, in an RGB array. */
    protected float[] m_bottomBgColor = {0f, 0f, 0.4f};
    /** Whether this graph area is currently selected. */
    protected boolean m_bSelected = false;
    /**
     * Whether deleted points should be drawn outside the graph (true)
     * or over the curve (false).
     */
    protected boolean m_bDrawDeletedOutside = true;
    /**
     * Whether to save time by drawing a simpler representation of the
     * graph.
     */
    protected boolean m_bQuickDraw = false;
    /**
     * Describes the size of the graph area after it has been transformed
     * to a regular shape.
     */
    protected Rectangle m_local = new Rectangle(0, 0, 1, 1);
    /**
     * Describes the size of the display area and the location of the graph
     * in that area.  Given in screen or page coordinates.
     */
    protected Rectangle m_display = new Rectangle(0, 0, 1, 1);
    /**
     * Describes the rectangle containing the graph frame as a proportion
     * of the total width and height of the graph area.  Between 0 and 1.
     */
    protected Rectangle2D.Double m_frame = new Rectangle2D.Double(
	0.08, 0.1, 0.9, 0.8);
    /**
     * The curve that was recently calculated as being nearest to the cursor.
     */
    protected Curve m_nearestCurve = null;
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/Graph.gif";
    
//------------------------------------------------------------------------------
    /**
     * Construct a time series graph and have it create its own x-axis zoom.
     *
     */
    public Graph() {
	
	// Let the other constructor do the work of setting up.
	this(new Zoom());
	
	setDefaults("gov.usgs.gr.visual.Graph");
    }
    
//------------------------------------------------------------------------------
    /**
     * Construct a time series graph and use the specified x-axis limits.
     * @param xZoom The x-axis limits object to use for this graph.
     */
    public Graph(Zoom xZoom) {
	
	super();
	
	m_xAxis = newXAxis();
	m_yAxis = newYAxis();
	
	super.add(m_xAxis);
	super.add(m_yAxis);
	
	// Save the zoom parameter for future use.
	setXZoom(xZoom);
	
	m_title = "Empty";
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a graph cloned from an existing one.
     */
    public Graph(Graph graph) {
	
	super();
	
	graph.copyInto(this, true);
	
	m_xTickScale = graph.getXTickScale();
	m_yTickScale = graph.getYTickScale();
	m_bQuickDraw = graph.getQuickDraw();
	copyAxes(graph);
	
	// Copy the other graph area's attributes.
	m_bSelected = graph.getSelected();
	m_bShow = graph.getShow();
	m_display = graph.getDisplayRect();
	m_local = graph.getLocalRect();
	m_frame = graph.getFrameRect();
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides add in GrObject.  Replace existing axes so that there are exactly
     * one of each (dim = 0, dim = 1).  Makes X axis child 0 and Y axis child 1.
     */
    public boolean add(GrObject gro) {
	
	boolean bAdded = false;
	
	// If gro is an object effigy, get the eventual subject and try
	// adding that.
	while (gro instanceof gov.usgs.gr.ObjectEffigy)
	    gro = ((ObjectEffigy)gro).getSubject();
	
	// Don't allow pages to be added to graphs.
	if (gro instanceof PageOfGraphs) return false;
	
	if (gro instanceof Axis) {
	    gro = GraphicalElement.appropriateVersionOf(this, gro);
	    int groDim = ((Axis)gro).getDimension();
	    if (groDim == 0) m_xAxis = (Axis)gro;
	    else if (groDim == 1) m_yAxis = (Axis)gro;
	    
	    // When adding axes, only replace existing ones.
	    boolean bSet = false;
	    String oClassName = gro.getClass().getName();
	    ListIterator iter = m_children.listIterator();
	    int i = 0;
	    while (iter.hasNext()) {
		GrObject oldGro = (GrObject)iter.next();
		if (oldGro instanceof Axis) {
		    if (((Axis)oldGro).getDimension() == groDim) {
			bSet = true;
			iter.set(gro);
			
			m_xAxis.setOpposingZoom(m_yAxis.getZoom());
			m_yAxis.setOpposingZoom(m_xAxis.getZoom());
			m_xAxis.update();
			m_yAxis.update();
			
			m_xZoom = (Zoom)m_xAxis.getZoom();
			m_yZoom = (Zoom)m_yAxis.getZoom();
			
			gro.addObserver(this);
			
			break;
		    }
		}
	    }
	    
	    if (!bSet) {
		super.add(gro);
		bAdded = true;
	    }
	} else {
	    // When adding curves, use the first one as the title.
	    if (gro instanceof Curve || DoubleDataSeries.countDims(gro, 2)) {
		ListIterator iter = m_children.listIterator();
		boolean bWasEmpty = true;
		while (iter.hasNext()) {
		    GrObject oldGro = (GrObject)iter.next();
		    if (oldGro instanceof Curve) {
			bWasEmpty = false;
			break;
		    }
		}
		
		if (bWasEmpty && (m_title.equals("") ||
		    m_title.equals("Empty"))) {
		    
		    if (gro instanceof Curve)
			m_title = ((Curve)gro).getTitle();
		    if (DoubleDataSeries.countDims(gro, 2))
			m_title = ((DoubleDataSeries)gro).getTitle();
		}
	    }
	    if (gro instanceof GraphicalElement) {
		// Add it.
		super.add(gro);
		bAdded = true;
		
		if (gro instanceof Curve) {
		    m_curves.add((Curve)gro);
		}
		if (gro instanceof Markup) {
		    m_markupList.add((Markup)gro);
		}
	    } else {
		if (DoubleDataSeries.countDims(gro, 2)) {
		    Curve curve = newCurve((DoubleDataSeries)gro);
		    curve.setColor(m_curves.size());
		    curve.setLinePattern(m_curves.size());
		    
		    // Add it.
		    super.add(curve);
		    bAdded = true;
		    m_curves.add(curve);
		    
		    // Rescale the axes to fit the current collection of curves.
		    m_xZoom.autoScale(10);
		    
		    m_yZoom.autoScale(6);
		    
		    // Update the axes in case the limits changed from autoscaling.
		    m_xAxis.update();
		    m_yAxis.update();
		}
	    }
	}
	
	if (bAdded) gro.addObserver(this);
	
	return bAdded;
    }
    
//------------------------------------------------------------------------------
    /**
     * Removes all children from the graph except the axes.
     */
    public void clear() {
	
	// Remove everything except the axes.
	ArrayList removables = new ArrayList();
	Iterator iter = m_children.iterator();
	while (iter.hasNext()) {
	    Object o = iter.next();
	    if (!(o instanceof Axis)) {
		removables.add(o);
	    }
	}
	
	iter = removables.iterator();
	while (iter.hasNext()) {
	    remove((GrObject)iter.next());
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Removes child i from the graph.  Does not remove axes.
     */
    public GrObject remove(int i) {
	
	if (!(m_children.get(i) instanceof Axis)) {
	    GrObject gro = super.remove(i);
	    if (gro instanceof Curve) {
		m_curves.remove((Curve)gro);
	    }
	    if (gro instanceof Markup) {
		m_markupList.remove((Markup)gro);
	    }
	    return gro;
	} else {
	    return null;
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Removes child gro from the graph.  Does not remove axes.
     */
    public boolean remove(GrObject gro) {
	
	
	
	
	
	
	
	
	
	
	
	if (!(gro instanceof Axis)) {
	    if (gro instanceof Curve) {
		m_curves.remove((Curve)gro);
	    }
	    if (gro instanceof Markup) {
		m_markupList.remove((Markup)gro);
	    }
	    return super.remove(gro);
	} else {
	    return false;
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Read an XML version of this element and its children.
     */
    public void readFromDOM(org.w3c.dom.Node node,
	Map allObjects, List unlinkedObjects) {
	
	super.readFromDOM(node, allObjects, unlinkedObjects);
	
	m_xAxis.setOpposingZoom(m_yAxis.getZoom());
	m_yAxis.setOpposingZoom(m_xAxis.getZoom());
	m_xAxis.update();
	m_yAxis.update();
	
	m_xZoom = (Zoom)m_xAxis.getZoom();
	m_yZoom = (Zoom)m_yAxis.getZoom();
    }
    
//------------------------------------------------------------------------------
    /**
     * Perform tasks that had to wait until all XML input had been read.
     */
    public void finishSAXInput(Map idObjects) {
	
	finishSAXForChildren(idObjects);
	
	m_xAxis.setOpposingZoom(m_yAxis.getZoom());
	m_yAxis.setOpposingZoom(m_xAxis.getZoom());
	m_xAxis.update();
	m_yAxis.update();
	
	m_xZoom = (Zoom)m_xAxis.getZoom();
	m_yZoom = (Zoom)m_yAxis.getZoom();
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
	
	if (TREE_ICON_HOLDER.size() <= 0)
	    setTreeIcon();
	
	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {
	
	javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);
	
	if (icon == null)
	    icon = super.setTreeIcon();
	
	TREE_ICON_HOLDER.clear();
	TREE_ICON_HOLDER.add(icon);
	
	return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides default update() in GrObject.  Called by an observed GrObject's
     * notifyObservers() method to handle changes to the observed object.
     * May use the notifyObservers() method to propagate the event handling up to
     * this object's observers.
     * @param gro   The observed object that changed.
     * @param arg   Additional information that the updated object may choose to
     *              send.
     */
    public void update(GrObject gro, Object arg) {
	
	if (gro == m_xAxis) {
	    m_xZoom = (Zoom)m_xAxis.getZoom();
	    m_yAxis.setOpposingZoom(m_xZoom);
	    m_xAxis.update();
	}
	if (gro == m_yAxis) {
	    m_yZoom = (Zoom)m_yAxis.getZoom();
	    m_xAxis.setOpposingZoom(m_yZoom);
	    m_yAxis.update();
	}
	
	// Check if the object is a graphical element that needs to be redrawn.
	if (gro instanceof GraphicalElement) {
	    if (((GraphicalElement)gro).getNeedsRedraw()) {
		if (m_bShow) m_bNeedsRedraw = true;
	    }
	}
	
	m_bHasChanged = true;
	notifyObservers();
    }
    
//------------------------------------------------------------------------------
    /**
     * Make the aspect ratio 1 to 1 (X to Y) if set to do so.
     */
    public void makeAspect1to1() {
	
	if (!m_bAspect1to1)
	    return;
	
	// Find the length of each side of the graph frame in pixels.
	double framePixelWidth = m_display.width * m_frame.width;
	double framePixelHeight = m_display.height * m_frame.height;
	double pixelMin = Math.min(framePixelWidth, framePixelHeight);
	if (pixelMin < 5)
	    return;
	
	// Find the length of each axis in local coordinates.
	double xLength = m_xZoom.getViewLength();
	double yLength = m_yZoom.getViewLength();
	
	// The X and Y ratios should be equal to within 1 pixel.
	double xRatio = xLength / framePixelWidth;
	double xMinRatioDiff = (xLength / (framePixelWidth - 1)) - xRatio;
	xMinRatioDiff *= 1.0000000000001;
	double yRatio = yLength / framePixelHeight;
	double yMinRatioDiff = (yLength / (framePixelHeight - 1)) - yRatio;
	yMinRatioDiff *= 1.0000000000001;
	double minRatioDiff = Math.max(xMinRatioDiff, yMinRatioDiff);
	if (xRatio < (yRatio - minRatioDiff) ||
	    xRatio > (yRatio + minRatioDiff)) {
	    
	    if (xRatio < yRatio) {
		double newXLength = yRatio * framePixelWidth;
		resizeZoom1to1(m_xZoom, newXLength);
	    } else {
		double newYLength = xRatio * framePixelHeight;
		resizeZoom1to1(m_yZoom, newYLength);
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Resize an axis zoom to the given length.
     */
    protected void resizeZoom1to1(Zoom zoom, double newLength) {
	
	double viewMin = zoom.getViewMin();
	double viewMax = zoom.getViewMax();
	double outerMin = zoom.getOuterMin();
	double outerMax = zoom.getOuterMax();
	double oldLength = zoom.getViewLength();
	double diff = newLength - oldLength;
	if ((viewMax + diff) > outerMax) {
	    double upperDiff = outerMax - viewMax;
	    viewMax = outerMax;
	    diff -= upperDiff;
	    if ((viewMin - diff) < outerMin) {
		double lowerDiff = viewMin - outerMin;
		viewMin = outerMin;
		diff -= lowerDiff;
		viewMax += diff;
		outerMax += diff;
	    } else {
		viewMin -= diff;
	    }
	    zoom.setOuterLimits(outerMin, outerMax);
	} else {
	    viewMax = viewMin + newLength;
	}
	zoom.setViewLimits(viewMin, viewMax);
    }
    
//------------------------------------------------------------------------------
    /**
     * Should return a new x axis object.
     */
    protected abstract Axis newXAxis();
    
//------------------------------------------------------------------------------
    /**
     * Should return a new y axis object.
     */
    protected abstract Axis newYAxis();
    
//------------------------------------------------------------------------------
    /**
     * Should return a new Curve object.
     */
    protected abstract Curve newCurve();
    
//------------------------------------------------------------------------------
    /**
     * Should return a new Curve object based on the data series parameter.
     */
    protected abstract Curve newCurve(DoubleDataSeries series);
    
//------------------------------------------------------------------------------
    /**
     * Should return a new Curve object copied from an existing curve.
     */
    protected abstract Curve newCurve(Curve curve);
    
//------------------------------------------------------------------------------
    /**
     * Should make new copies of the axis objects associated with graph.
     * @param graph	The graph with the template axes.
     */
    protected abstract void copyAxes(Graph graph);
    
//------------------------------------------------------------------------------
    /**
     * Draw a background behind the curves and axes.
     * <p><b> Pre Conditions </b><br>
     *      The viewing transformation has been set up so that the graphing
     *		area is between 0 and 1 on both axes.
     * <p><b> Post Conditions </b><br>
     *      The current drawing color has been modified.
     */
    protected void drawBackground() {
    }
    
//------------------------------------------------------------------------------
    /**
     * Draw a thick yellow frame around the graph.
     * <p><b> Pre Conditions </b><br>
     *      The viewing transformation has been set up so that the graphing
     *		area is between 0 and 1 on both axes.
     * <p><b> Post Conditions </b><br>
     *      None.
     */
    protected abstract void drawFrame();
    
//------------------------------------------------------------------------------
    /**
     * Define clipping planes around the graph.
     * <p><b> Pre Conditions </b><br>
     *      The viewing transformation has been set up so that the graphing
     *		area is in graph coordinates.
     * <p><b> Post Conditions </b><br>
     *      None.
     */
    protected void setupClipping() {
	
    }
    
//------------------------------------------------------------------------------
    /**
     * Disable all clipping planes.
     */
    protected void disableClipping() {
	
    }
    
//------------------------------------------------------------------------------
    /**
     * Mark this graph on the screen as being selected.
     * <p><b> Pre Conditions </b><br>
     *      Viewing transformation has been set so this graph's section of the
     *		page is between 0 and 1 in both x and y directions.
     * <p><b> Post Conditions </b><br>
     *      The current drawing color has been changed.
     */
    private void drawAsSelected() {
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public GrHashList getProperties() {
	
	GrHashList props = super.getProperties();
	
	props.put("showXtext", new GrProperty(m_bShowXtext, "Show X Labels"));
	props.put("aspect1to1", new GrProperty(m_bAspect1to1,
	    "Aspect Ratio 1:1"));
	
	ColorProperty cp = new ColorProperty();
	cp.description = "Frame Color";
	cp.setColor(m_frameColor);
	cp.display = false;
	cp.save = false;
	props.put("frameColor", cp);
	
	cp = new ColorProperty();
	cp.description = "Text Color";
	cp.setColor(m_textColor);
	cp.display = false;
	cp.save = false;
	props.put("textColor", cp);
	
	cp = new ColorProperty();
	cp.description = "Top Background Color";
	cp.setColor(m_topBgColor);
	cp.display = false;
	cp.save = false;
	props.put("topBgColor", cp);
	
	cp = new ColorProperty();
	cp.description = "Bottom Background Color";
	cp.setColor(m_bottomBgColor);
	cp.display = false;
	cp.save = false;
	props.put("bottomBgColor", cp);
	
	return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public void setProperties(HashMap props) {
	
	boolean bOldShow = m_bShow;
	
	super.setProperties(props);
	
	GrProperty p;
	
	Iterator iter = props.keySet().iterator();
	while (iter.hasNext()) {
	    String key = (String)iter.next();
	    Object obj = props.get(key);
	    if (obj instanceof GrProperty) {
		p = (GrProperty)obj;
		
		if (key.equals("showXtext"))
		    m_bShowXtext = (Boolean)p.value;
		if (key.equals("aspect1to1"))
		    m_bAspect1to1 = (Boolean)p.value;
		if (key.equals("frameColor")) {
		    ColorProperty cp = (ColorProperty)p;
		    m_frameColor = cp.getColorArray();
		}
		if (key.equals("textColor")) {
		    ColorProperty cp = (ColorProperty)p;
		    m_textColor = cp.getColorArray();
		}
		if (key.equals("topBgColor")) {
		    ColorProperty cp = (ColorProperty)p;
		    m_topBgColor = cp.getColorArray();
		}
		if (key.equals("bottomBgColor")) {
		    ColorProperty cp = (ColorProperty)p;
		    m_bottomBgColor = cp.getColorArray();
		}
	    }
	    if (obj instanceof String) {
		String value = (String)obj;
		setProperty(key, value);
	    }
	}
	
	if (m_bShow || bOldShow) m_bNeedsRedraw = true;
	m_bHasChanged = true;
	notifyObservers();
    }
    
//------------------------------------------------------------------------------
    /**
     * Set a single property.
     */
    public void setProperty(String key, String value) {
	
	super.setProperty(key, value);
	
	if (key.equals("showXtext"))
	    m_bShowXtext = Boolean.parseBoolean(value);
	if (key.equals("aspect1to1"))
	    m_bAspect1to1 = Boolean.parseBoolean(value);
	if (key.equals("frameColor")) {
	    m_frameColor = ColorProperty.getColorArray(value);
	}
	if (key.equals("textColor")) {
	    m_textColor = ColorProperty.getColorArray(value);
	}
	if (key.equals("topBgColor")) {
	    m_topBgColor = ColorProperty.getColorArray(value);
	}
	if (key.equals("bottomBgColor")) {
	    m_bottomBgColor = ColorProperty.getColorArray(value);
	}
	
	m_bNeedsRedraw = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Changes how much area this graph has on the page.
     * @param width	 The new width of the graph in the units of the output device.
     * @param height The new width of the graph in the units of the output device.
     */
    public void setSize(int width, int height) {
	
	setWidth(width);
	setHeight(height);
    }
    
//------------------------------------------------------------------------------
    /**
     * Tells whether the mouse is over this graph, based on the mouse's position
     * the saved position and height values.
     * @param mouseX	The horizontal position of the mouse in screen coordinates.
     * @param mouseY	The vertical position of the mouse in screen coordinates.
     * @return True if the mouse is over this graph.
     */
    public boolean isMouseOver(int mouseX, int mouseY) {
	
	if ((mouseY > m_display.y && mouseY < (m_display.y + m_display.height))
	&& (mouseX > m_display.x && mouseX < (m_display.x
	    + m_display.width)))
	    return true;
	else return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Tells whether the graph has any curves in it.
     * @return True if the graph is empty.
     */
    public boolean isEmpty() {
	
	return (m_curves.size() == 0 ? true : false);
    }
    
//------------------------------------------------------------------------------
    /**
     * Return the nearest point on a curve to the screen point.  Point is
     * measured from (0, 0) in the bottom left corner.
     * @param i	The x position of the point on the screen.
     * @param j	The y position of the point on the screen.
     */
    public Point2D.Double getNearestPoint(int i, int j) {
	
	Point2D.Double p = getCoords(i, j);
	
	boolean bAtLeastOneSorted = false;
	double upperIndexX, lowerIndexX;
	double closestX = Double.MAX_VALUE;
	double closestY = Double.MAX_VALUE;
	
	// Make a variable to keep track of whether the nearest curve is
	// actually over or under the mouse, or if it is away from the mouse.
	boolean bClosestIsAway = true;
	
	// Check each curve.
	Iterator iter = m_curves.iterator();
	while (iter.hasNext()) {
	    Curve curve = (Curve)iter.next();
	    // Skip the curve if it is supposed to be hidden.
	    if (!curve.getShow()) continue;
	    
	    DoubleDataSeries series = curve.getSeries();
	    if (series == null)
		continue;
	    
	    // Check if the series is sorted.
	    if (series.getSorted(0)) bAtLeastOneSorted = true;
	    else continue;
	    
//hb            double[] array = series.getArray();
	    double[] array = series.getArray();
	    int index = series.search(p.x) * series.getNumDimensions();
	    if (index < 0) continue;
	    
	    // Use this point or the one before it, which ever is closest
	    // to the mouse.
	    lowerIndexX = array[index];
	    if (index > (array.length - 4)) upperIndexX = lowerIndexX;
	    else upperIndexX = array[index + 2];
	    
	    double upperDiff = Math.abs(upperIndexX - p.x);
	    double lowerDiff = Math.abs(lowerIndexX - p.x);
	    if (upperDiff < lowerDiff && index < (array.length - 3))
		index += 2;
	    
	    // Find out whether the mouse is outside the X range of
	    // the curve.
	    boolean bAwayFromCurve = false;
	    if (p.x < lowerIndexX || p.x > upperIndexX)
		bAwayFromCurve = true;
	    
	    // Find the difference between the current point and each
	    // curve.  Go with the nearest curve.  Compare X values if the
	    // curve isn't directly over the mouse, Y values otherwise.
	    double currentDiff = 0, newDiff = 1;
	    if (bAwayFromCurve) {
		currentDiff = Math.abs(closestX - p.x);
		newDiff = Math.abs(array[index] - p.x);
	    }
	    if (!bAwayFromCurve || closestX == array[index]) {
		currentDiff = Math.abs(closestY - p.y);
		newDiff = Math.abs(array[index + 1] - p.y);
	    }
	    
	    // Don't consider this curve if it is away from the mouse and a
	    // curve that isn't has already been found.
	    if (!bClosestIsAway && bAwayFromCurve) continue;
	    
	    // Make this the new closest point if its Y value is closer or
	    // if this is the first curve to be directly over or under the
	    // mouse.
	    if (newDiff < currentDiff ||
		(bClosestIsAway && !bAwayFromCurve)) {
		
		closestX = array[index];
		closestY = array[index + 1];
		m_nearestCurve = curve;
		
		// Once a curve has been found that is directly over or
		// under the mouse, don't consider any curves that aren't.
		if (!bAwayFromCurve) bClosestIsAway = false;
	    }
	}
	
	if (bAtLeastOneSorted) return new Point2D.Double(closestX, closestY);
	else return new Point2D.Double(p.x, p.y);
    }
    
//------------------------------------------------------------------------------
    /**
     * Return the nearest curve to the cursor.
     */
    public Curve getNearestCurve() {
	
	return m_nearestCurve;
    }
    
//------------------------------------------------------------------------------
    /**
     * Convert from screen coordinates to user coordinates.
     * @param i	The x position of the point on the screen.
     * @param j	The y position of the point on the screen.
     */
    public Rectangle2D.Double screenToUser(Rectangle screenRect) {
	
	Point2D.Double p1 = getCoords(screenRect.x, screenRect.y);
	Point2D.Double p2 = getCoords(screenRect.x + screenRect.width,
	    screenRect.y + screenRect.height);
	Rectangle2D.Double r = new Rectangle2D.Double(p1.x, p1.y, p2.x - p1.x,
	    p2.y - p1.y);
	
	return r;
    }
    
//------------------------------------------------------------------------------
    /**
     * Return the graph coordinates of the screen point.  Point is measured from
     * (0, 0) in the bottom left corner.
     * @param i	The x position of the point on the screen.
     * @param j	The y position of the point on the screen.
     */
    public Point2D.Double getCoords(int i, int j) {
	
//    	String[] coords = {"0", "0"};
	Point2D.Double p = new Point2D.Double(i, j);
	
	// Convert the x position to graph coordinates.
	p.x -= (double)m_display.x;
	p.x /= (double)m_display.width;
	p.x -= m_frame.x;
	p.x /= m_frame.width;
	p.x /= m_xZoom.getScale();
	p.x += m_xZoom.getViewMin();
	
//        DecimalFormat form = new DecimalFormat(m_xZoom.getFormatString());
//        coords[0] = form.format(p.x);
	
	// Convert the y position to graph coordinates.
	p.y -= (double)m_display.y;
	p.y /= (double)m_display.height;
	p.y -= m_frame.y;
	p.y /= m_frame.height;
	p.y /= m_yZoom.getScale();
	p.y += m_yZoom.getViewMin();
	
//        form = new DecimalFormat(m_yZoom.getFormatString());
//        coords[1] = form.format(p.y);
	
	return p;
    }
    
//------------------------------------------------------------------------------
    /**
     * Move the zoom window by the specified x and y amount.
     * <p><b> Pre Conditions </b><br>
     *      None.
     * <p><b> Post Conditions </b><br>
     *      Changes won't be seen until next redraw.
     * @param xChange	The amount that the zoom window is to move horizontally
					in screen pixels.
     * @param yChange	The amount that the zoom window is to move vertically
					in screen pixels.
     */
    public void pan(double xChange, double yChange) {
	
	// Convert from screen pixels to graph coordinates and send the
	// message to the zoom objects.
	synchronized (m_xZoom) {
	    synchronized (m_yZoom) {
		xChange = xChange / m_display.width / (double)m_frame.width;
		m_xZoom.panPercent(xChange);
		yChange = yChange / m_display.height / (double)m_frame.height;
		m_yZoom.panPercent(yChange);
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Zooms in on the specified area if the area is contained completely within
     * this graph.
     * <p><b> Post Conditions </b><br>
     *      Changes won't be seen until next redraw.
     * @param zoomRect	The rectangle describing the zoom area in display
     *					coordinates.
     * @return	True if the zoom operation was completed on this graph.
     */
    public boolean zoomWindow(Rectangle zoomRect) {
	
	// Return false if the zoom window is not inside this graph.
	if (!m_display.contains(zoomRect)) return false;
	
	zoomRect.translate(-m_display.x, -m_display.y);
	
	double min, max, min1, max1;
	
	// Calculate the x-axis coordinates of the window
	min1 = ((double)zoomRect.x / (double)m_display.width) - m_frame.x;
	min = (min1 / m_frame.width * m_xZoom.getViewLength())
	+ m_xZoom.getViewMin();
	max1 = ((double)(zoomRect.x + zoomRect.width) / (double)m_display.width)
	- m_frame.x;
	max = (max1 / m_frame.width * m_xZoom.getViewLength())
	+ m_xZoom.getViewMin();
	// Zoom in on the x-axis.
	m_xZoom.setViewLimits(min, max);
	m_xZoom.autoScale(-1);
	
	// Calculate the y-axis coordinates of the window
	min1 = ((double)zoomRect.y / (double)m_display.height) - m_frame.y;
	min = (min1 / m_frame.height * m_yZoom.getViewLength())
	+ m_yZoom.getViewMin();
	max1 = ((double)(zoomRect.y + zoomRect.height) / (double)m_display.height)
	- m_frame.y;
	max = (max1 / m_frame.height * m_yZoom.getViewLength())
	+ m_yZoom.getViewMin();
	// Zoom in on the y-axis.
	m_yZoom.setViewLimits(min, max, true);
	m_yZoom.autoScale(-1);
	
	return true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Selects or deselects all points within the zoom window.
     * <p><b> Post Conditions </b><br>
     *      Changes won't be seen until next redraw.
     * @param zoomRect	The rectangle describing the zoom area in display
     *					coordinates.
     * @param bAdding	True if selecting points, false if deselecting.
     * @param bExclusive	True if only one curve should be selected or,
     *						if multiple curves are already selected, only
     *						the selected curves can have more points selected.
     * @param bDeselectFirst	True if old selection is deselected before selecting
     *							new points.
     * @return	An unmodifiable list of selected data series in this graph.
     */
    public List selectWindow(Rectangle zoomRect, boolean bAdding,
	boolean bExclusive, boolean bDeselectFirst) {
	
	// Save the number of items currently selected.
	int listSize = m_selectedSeriesList.size();
	
	// Make the zoom rect relative to this graph's position.
	zoomRect.translate(-m_display.x, -m_display.y);
	
	double xMin, xMax, yMin, yMax, xMin1, xMax1, yMin1, yMax1;
	
	// Calculate the x-axis coordinates of the window
	xMin1 = ((double)zoomRect.x / (double)m_display.width) - m_frame.x;
	xMin = (xMin1 / m_frame.width * m_xZoom.getViewLength())
	+ m_xZoom.getViewMin();
	xMax1 = ((double)(zoomRect.x + zoomRect.width) / (double)m_display.width)
	- m_frame.x;
	xMax = (xMax1 / m_frame.width * m_xZoom.getViewLength())
	+ m_xZoom.getViewMin();
	
	// Calculate the y-axis coordinates of the window
	yMin1 = ((double)zoomRect.y / (double)m_display.height) - m_frame.y;
	yMin = (yMin1 / m_frame.height * m_yZoom.getViewLength())
	+ m_yZoom.getViewMin();
	yMax1 = ((double)(zoomRect.y + zoomRect.height) /
	    (double)m_display.height) - m_frame.y;
	yMax = (yMax1 / m_frame.height * m_yZoom.getViewLength())
	+ m_yZoom.getViewMin();
	
	// Determine whether the window is completely outside the graph frame.
	boolean bInsideFrame;
	if ((xMin1 < 0 && xMax1 < 0)|| (xMin1 > m_frame.width &&
	    xMax1 > m_frame.width) || (yMin1 < 0 && yMax1 < 0) ||
	    (yMin1 > m_frame.height && yMax1 > m_frame.height)) {
	    
	    bInsideFrame = false;
	    m_selectedSeriesList.clear();
	    return new ArrayList(0);
	} else {
	    bInsideFrame = true;
	}
	
	m_selectedSeriesList.clear();
	// Perform the operation on each series in the graph.
	Iterator iter = m_children.iterator();
	while (iter.hasNext()) {
	    GrObject gro = (GrObject)iter.next();
	    if (!(gro instanceof Curve)) {
		if (gro instanceof SelectableGE) {
		    Point2D.Double selectRectMin =
			new Point2D.Double(xMin, yMin);
		    Point2D.Double selectRectMax =
			new Point2D.Double(xMax, yMax);
		    List selectedList =
			((SelectableGE)gro).selectWindow(selectRectMin,
			selectRectMax, bAdding, bExclusive, bDeselectFirst);
		    
		    // Add selected series to the selected series list.
		    if (selectedList != null) {
			Iterator selIter = selectedList.iterator();
			while (selIter.hasNext()) {
			    Object obj = selIter.next();
			    if (obj instanceof DataSeries)
				m_selectedSeriesList.add(obj);
			}
		    }
		}
		
		continue;
	    }
	    
	    Curve curve = (Curve)gro;
	    DoubleDataSeries series = curve.getSeries();
	    
	    // Only select if the series is already selected, it is the first
	    // series to be selected, or the user enable multiple selections.
	    if (!bExclusive || (series.getNumSelected() > 0) || listSize == 0) {
		if (bDeselectFirst) series.deselectAll();
		if (!curve.getShow()) continue;
		// Select or deselect the points in the window.
		double[] min = new double[2];
		double[] max = new double[2];
		min[0] = (double)xMin;
		max[0] = (double)xMax;
		min[1] = (double)yMin;
		max[1] = (double)yMax;
		series.selectArea(min, max, bAdding);
		if (series.getNumSelected() > 0) {
		    m_selectedSeriesList.add(series);
		    // Make sure the list size now shows non-zero.
		    listSize = 1;
		}
	    }
	}
	
	return Collections.unmodifiableList(m_selectedSeriesList);
    }
    
//------------------------------------------------------------------------------
    /**
     * Selects or deselects all points within the zoom window.
     * <p><b> Post Conditions </b><br>
     *      Changes won't be seen until next redraw.
     * @param curveNum	The index of the curve to select.
     * @param bExclusive	True if only one curve should be selected at a time.
     * @return	The list of selected data series in this graph.
     */
    public List selectCurve(int curveNum, boolean bExclusive) {
	
	// Deselect all curves first, and check for bounds.
	m_selectedSeriesList.clear();
	if (curveNum < 0 || curveNum > m_curves.size())
	    return m_selectedSeriesList;
	
	// Run through the list and select the right curve.
	int index = 0;
	Iterator iter = m_curves.iterator();
	while (iter.hasNext()) {
	    Curve curve = (Curve)iter.next();
	    DoubleDataSeries series = curve.getSeries();
	    
	    if (index == curveNum) {
		series.selectAll();
		m_selectedSeriesList.add(series);
	    } else {
		if (bExclusive) series.deselectAll();
	    }
	    index++;
	}
	
	return m_selectedSeriesList;
    }
    
//------------------------------------------------------------------------------
    /**
     * Delete all markup objects from the list.
     */
    public void deleteAllMarkups() {
	
	m_markupList.clear();
	Iterator iter = m_children.iterator();
	while (iter.hasNext()) {
	    GraphicalElement ge = (GraphicalElement)iter.next();
	    if (ge instanceof gov.usgs.gr.visual.Markup) iter.remove();
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {
	
	String s = null;
	
	if (m_bShow)
	    s = "Graph: " + m_title;
	else
	    s = "(Graph): " + m_title;
	
	return s;
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the minimum value between all the curves in this graph for the
     * specified dimension.
     * @param dimension The dimension to report the minimum for.  If it is 0, it
     *                  will report the min x value.  If 1, it will report the min
     *                  y value.
     * @return The mimimum value for the data along the specified dimension.
     */
    public double getMin(int dimension){
	
	// Start out with a min that is larger than any other.
	double min = Double.POSITIVE_INFINITY;
	
	// Check each curve for the lowest min.
	Iterator iter = m_curves.iterator();
	if (!iter.hasNext()) return 0;
	
	while (iter.hasNext()) {
	    double curveMin = ((Curve)iter.next()).getMin(dimension);
	    if (curveMin < min) min = curveMin;
	}
	
	return min;
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the maximum value between all the curves in this graph for the
     * specified dimension.
     * @param dimension The dimension to report the minimum.  If it is 0, it
     *                  will report the max x value.  If 1, it will report the max
     *                  y value.
     * @return The maximum value for the data along the specified dimension.
     */
    public double getMax(int dimension){
	
	// Start out with a min that is larger than any other.
	double max = Double.NEGATIVE_INFINITY;
	
	// Check each curve for the lowest min.
	Iterator iter = m_curves.iterator();
	if (!iter.hasNext()) return 100;
	
	while (iter.hasNext()) {
	    double curveMax = ((Curve)iter.next()).getMax(dimension);
	    if (curveMax > max) max = curveMax;
	}
	
	return max;
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the minimum and maximum values between all the graphical objects shown
     * in this graph for the specified dimension.
     * @param dimension The dimension to report on.  X=0, Y=1.
     * @return The minimum (stored as point.x) and maximum (stored as point.y)
     *         values for the graph contents.
     */
    public Point2D.Double getBounds(int dimension){
	
	// Start out with a min that is larger than any other.
	Point2D.Double p = new Point2D.Double(Double.POSITIVE_INFINITY,
	    Double.NEGATIVE_INFINITY);
	
	// Check each curve for the lowest min.
	// Get the bounds from each child and compare to the most negative
	// and most positive so far.  If it is beyond the current bounds,
	// set it as the current bounds.
	Iterator iter = m_children.iterator();
	while (iter.hasNext()) {
	    GrObject gro = (GrObject)iter.next();
	    if (gro instanceof GraphicalElement){
		GraphicalElement gel = (GraphicalElement)gro;
		Point2D.Double gelP = gel.getBounds(dimension);
		if (gelP != null) {
		    if (gelP.x < p.x) p.x = gelP.x;
		    if (gelP.y > p.y) p.y = gelP.y;
		}
	    }
	}
	
	// If the bounds were not reduced from the initial value, send back
	// a simple 0-100 range.
	if (p.x == Double.POSITIVE_INFINITY) {
	    p.x = 5;
	    p.y = 95;
	}
	
	return p;
    }
    
//------------------------------------------------------------------------------
    /**
     * Reproportion the frame so it fits with any titles that are being displayed.
     */
    public void layoutFrame(){
	
	// Resize the horizontal margins.
	if (m_yAxis.getShowTitle() && m_yAxis.getShowText()) {
	    m_frame.x = 0.08;
	    m_frame.width = 0.9;
	} else {
	    m_frame.x = 0.05;
	    m_frame.width = 0.92;
	}
	
	// Resize the vertical margins.
	if (m_bShowXtext) {
	    // Make a bigger lower margin to hold the labels.
	    if (m_xAxis.getShowTitle()) {
		m_frame.y = 0.13;
		if (m_bShowTitle) m_frame.height = 0.77;
		else m_frame.height = 0.81;
	    } else {
		m_frame.y = 0.1;
		if (m_bShowTitle) m_frame.height = 0.8;
		else m_frame.height = 0.84;
	    }
	} else {
	    // Make a bigger frame in place of the labels.
	    m_frame.y = 0.01;
	    if (m_bShowTitle) m_frame.height = 0.89;
	    else m_frame.height = 0.93;
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return A list full of Curve ojects.
     */
    public List<Curve> getCurves() {
	
	return m_curves;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return True if text is to be shown along the x-axis.
     */
    public boolean getShowXtext() {
	return m_bShowXtext;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param value true if labels are to be shown, false otherwise.
     */
    public void setShowXtext(boolean value) {
	
	m_bShowXtext = value;
	m_xAxis.setShowText(m_bShowXtext);
	layoutFrame();
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return True if the main title is to be shown.
     */
    public boolean getShowMainTitle() {
	return m_bShowTitle;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param value true if labels are to be shown, false otherwise.
     */
    public void setShowMainTitle(boolean value) {
	
	m_bShowTitle = value;
	layoutFrame();
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return True if the side bar is to be shown.
     */
    public boolean getShowSideBar() {
	
	return m_bShowSideBar;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param bShowSideBar	True if the side bar is to be shown.
     */
    public void setShowSideBar(boolean bShowSideBar) {
	
	m_bShowSideBar = bShowSideBar;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return True if the side bar is drawn down.
     */
    public boolean getSideBarDown() {
	
	return m_bSideBarDown;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param bDown			True if the bar is to be drawn down.
     */
    public void setSideBarDown(boolean bDown) {
	
	m_bSideBarDown = bDown;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return the tick scale.
     */
    public double getXTickScale() {
	
	return m_xTickScale;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param value	True if the graph should now be selected.
     */
    public void setXTickScale(double scale) {
	
	m_xTickScale = scale;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return the tick scale.
     */
    public double getYTickScale() {
	
	return m_yTickScale;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param value	True if the graph should now be selected.
     */
    public void setYTickScale(double scale) {
	
	m_yTickScale = scale;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return the y-axis object.
     */
    public Axis getYAxis() {
	return m_yAxis;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return the x-axis object.
     */
    public Axis getXAxis() {
	return m_xAxis;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return the x-axis Zoom object.
     */
    public Zoom getXZoom() {
	return m_xZoom;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param xZoom The x-axis limits object to use for this graph.
     */
    public void setXZoom(Zoom xZoom) {
	m_xZoom = xZoom;
	m_xAxis.setZoom(m_xZoom, m_yZoom);
	m_yAxis.setZoom(m_yZoom, m_xZoom);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The y-axis Zoom object.
     */
    public Zoom getYZoom() {
	return m_yZoom;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The title of the graph.
     */
    public String getTitle() {
	return m_title;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param title	The new title of the graph.
     */
    public void setTitle(String title) {
	
	m_title = title;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return True if this graph is selected.
     */public boolean getSelected() {
	 return m_bSelected;
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param value	True if the graph should now be selected.
      */
     public void setSelected(boolean value) {
	 
	 m_bSelected = value;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return True if this graph should be shown.
      */
     public boolean getShow() {
	 return m_bShow;
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param value	True if the graph should be drawn.
      */
     public void setShow(boolean value) {
	 m_bShow = value;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return True if symbols should be drawn outside.
      */
     public boolean getDrawDeletedOutside() {
	 return m_bDrawDeletedOutside;
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param value	True if symbols should be drawn outside.
      */
     public void setDrawDeletedOutside(boolean value) {
	 m_bDrawDeletedOutside = value;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return the width of the graph in display units.
      */
     public int getWidth() {
	 return m_display.width;
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param width	 The new width of the graph in display units.
      */
     public void setWidth(int width) {
	 m_display.width = width;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return the height of the graph in display units.
      */
     public int getHeight() {
	 return m_display.height;
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param height The new height of the graph in display units.
      */
     public void setHeight(int height) {
	 m_display.height = height;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return the vertical position of the bottom of this graph measured from
      * 			the bottom of the page in display units.
      */
     public int getXPosition() {
	 return m_display.x;
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param xPosition	The new bottom position of the graph measured from the
      *					bottom of the page in display units.
      */
     public void setXPosition(int xPosition) {
	 m_display.x = xPosition;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return the horizontal position of the bottom of this graph measured from
      * 			the left of the page in display units.
      */
     public int getYPosition() {
	 return m_display.y;
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param yPosition	The new bottom position of the graph measured from the
      *					bottom of the page in display units.
      */
     public void setYPosition(int yPosition) {
	 m_display.y = yPosition;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return a copy of the Rectangle discribing the dimensions of the output area.
      */
     public Rectangle getDisplayRect() {
	 return new Rectangle(m_display);
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param display Rectangle discribing the dimensions of the output area.
      */
     public void setDisplayRect(Rectangle display) {
	 m_display = display;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return a copy of the Rectangle discribing the dimensions of the frame area.
      */
     public Rectangle2D.Double getFrameRect() {
	 return new Rectangle2D.Double(m_frame.x, m_frame.y, m_frame.width,
	     m_frame.height);
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param frame Rectangle discribing the dimensions of the frame area.
      */
     public void setFrameRect(Rectangle2D.Double frame) {
	 m_frame = frame;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return a copy of the Rectangle discribing the size of the graph area in
      *			local coordinates.
      */
     public Rectangle getLocalRect() {
	 return new Rectangle(m_local);
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param local Rectangle discribing the size of the graph area in local
      * 				coordinates.
      */
     public void setLocalRect(Rectangle local) {
	 m_local = local;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return a True if this object is being drawn for speed instead of looks.
      */
     public boolean getQuickDraw() {
	 return m_bQuickDraw;
     }
     
//------------------------------------------------------------------------------
     /**
      * Accessor method.
      * @return a True if this graph should have a 1 to 1 aspect ratio (X to Y).
      */
     public boolean getAspect1to1() {
	 return m_bAspect1to1;
     }
     
//------------------------------------------------------------------------------
     /**
      * Attribute modification method.
      * @param bQuickDraw	True if the graph is to be drawn for speed at the
      *						expense of looks and accuracy.
      */
     public void setQuickDraw(boolean bQuickDraw) {
	 
	 m_bQuickDraw = bQuickDraw;
	 
	 // Set quick draw for the elements contained within this graph.
	 m_xAxis.setQuickDraw(m_bQuickDraw);
	 m_yAxis.setQuickDraw(m_bQuickDraw);
	 
	 Iterator iter = m_curves.iterator();
	 while (iter.hasNext()) {
	     Curve curve = (Curve)iter.next();
	     curve.setQuickDraw(m_bQuickDraw);
	 }
     }
}







