/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2000-03-13
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import gov.usgs.gr.*;
import gov.usgs.gr.visual.gl.*;
import gov.usgs.gr.visual.j2d.*;
import gov.usgs.gr.visual.ps.*;
import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import java.util.*;
import javax.swing.tree.*;

//==============================================================================
/**
 *
 */
public abstract class GraphicalElement extends gov.usgs.gr.GrObject {

	/** The title at the top of the page. */
    protected String m_title = "";
    /** Whether to show a title for the axis. */
    protected boolean m_bShowTitle = false;
    /** Whether this markup should be drawn. */
    protected boolean m_bShow = true;
    /** Whether this object has changed and needs to be drawn again. */
    protected boolean m_bNeedsRedraw = true;
    /**
     * The root transformation that describes the display area.
     */
    protected Rectangle2D.Double m_rootTrans =
    	new Rectangle2D.Double(0, 0, 1, 1);
    /**
     * Describes the parent transformation.
     */
    protected Rectangle2D.Double m_parentTrans =
    	new Rectangle2D.Double(0, 0, 1, 1);
    /**
     * Describes the user coordinate system used in the parent element.
     */
    protected Rectangle2D.Double m_contentTrans =
    	new Rectangle2D.Double(0, 0, 1, 1);
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GraphicalElement.gif";

//------------------------------------------------------------------------------
/**
 * Construct an empty graphical element.
 */
    public GraphicalElement () {

    }

//------------------------------------------------------------------------------
/**
 * Tell whether the parameter object is the same type as this object so
 * that its properties should be applied.
 */
    public boolean isSameTypeAs(GrObject gro) {

        if (getClass().isInstance(gro)) {
            GrHashList props = gro.getProperties();
            GrProperty p = (GrProperty)props.get("show");
            if (p == null)
                return false;
            if ((Boolean)p.value == m_bShow)
                return true;
            else
                return false;
        }
        else
            return false;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

    	props.put("title", new GrProperty(
        	convertNewlineToBackslashN(m_title), "Title"));
    	props.put("showTitle", new GrProperty(m_bShowTitle, "Show Title"));
    	props.put("show", new GrProperty(m_bShow, "Showing"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

    	super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

               	if (key.equals("title"))
                    m_title = convertBackslashNToNewline((String)p.value);
            	if (key.equals("showTitle"))
                    m_bShowTitle = (Boolean)p.value;
            	if (key.equals("show"))
                    m_bShow = (Boolean)p.value;
            }
            if (obj instanceof String) {
            	String value = (String)obj;

                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

    	if (key.equals("title"))
            m_title = convertBackslashNToNewline(value);
        if (key.equals("showTitle"))
            m_bShowTitle = Boolean.parseBoolean(value);
        if (key.equals("show"))
            m_bShow = Boolean.parseBoolean(value);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Add an object as a child of this one.  Overrides add() in GrObject.
 * Only add elements that are meant for the same type of drawing context.
 */
    public boolean add(GrObject gro) {

        GrObject newObj = appropriateVersionOf(gro);
        if (newObj != null)
        	return super.add(newObj);
        else
        	return false;
    }

//------------------------------------------------------------------------------
/**
 * Copy the properties and children of this object into the parameter object.
 */
    public void copyInto(GrObject gro, boolean bDeep) {

    	// Copy the properties to the subject.
		gro.setProperties(getProperties());

        // Add an appropriate copy of each child.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
        	GrObject child = (GrObject)iter.next();
        	if (bDeep) {
                if (gro instanceof GraphicalElement) {
                	// Get the correct version if it is a GraphicalElement.
                	GrObject newChild = ((GraphicalElement)gro).
                    	appropriateVersionOf(child);
        			if (newChild != null) {
                		if (newChild == child)
							gro.add(child.createCopy(false));
                		else
							gro.add(newChild);
                	}
                }
                else {
                	gro.add(child.createCopy(false));
                }
            }
        	else {
            	gro.add(child);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.
 */
    public Point2D.Double getBounds(int dimension){

        return null;
    }

//------------------------------------------------------------------------------
/**
 * Determines if gro is meant for the same type of drawing context.
 * If the object doesn't match, an attempt is made to convert it.  For example,
 * if this object is a PsGraph and gro is a GlCurve, a new PsCurve will be
 * created from gro and that curve will be added.
 */
    public GrObject appropriateVersionOf(GrObject gro) {

        return appropriateVersionOf(this, gro);
    }

//------------------------------------------------------------------------------
/**
 * Determines if gro is meant for the same type of drawing context.
 * If the object doesn't match, an attempt is made to convert it.  For example,
 * if this object is a PsGraph and gro is a GlCurve, a new PsCurve will be
 * created from gro and that curve will be added.
 */
    public static GrObject appropriateVersionOf(GrObject target, GrObject gro) {

    	// Add non-GraphicalElements without further checking.
        if (!(gro instanceof GraphicalElement)) {
        	return gro;
        }

        String fromKey = "";
        if (gro instanceof GlElement)
            fromKey = "Gl";
        if (gro instanceof J2dElement)
            fromKey = "J2d";
        if (gro instanceof PsElement)
            fromKey = "Ps";

        // If this is a GlElement...
        if (target instanceof GlElement) {
        	// Allow other GlElements to be added.
    		if (gro instanceof GlElement) {
            	return gro;
        	}
            else {
            	// Check if it is one of the known objects which have GL
                // equivalents.
            	if (gro instanceof PageOfGraphs)
                	return new PageOfGlGraphs((PageOfGraphs)gro);
            	if (gro instanceof Axis) {
                	Axis axis = (Axis)gro;
                    if (axis.getDimension() == 0)
                		return new GlXAxis(axis);
            		else
                    	return new GlYAxis(axis);
                }
            	if (gro instanceof Graph)
                	return new GlGraph((Graph)gro);
            	if (gro instanceof Curve)
                	return new GlCurve((Curve)gro);

             GrObject newGro = searchForOtherVersion(gro, fromKey, "Gl");
             if (newGro != null)
                 return newGro;
             else
                 // Can't identify it, so add it as it is.
                 return gro;
            }
        }

        // If this is a PsElement...
        if (target instanceof PsElement) {
        	// Allow other PsElements to be added.
    		if (gro instanceof PsElement) {
            	return gro;
        	}
            else {
            	// Check if it is one of the known objects which have Ps
                // equivalents.
            	if (gro instanceof PageOfGraphs)
                	return new PageOfPsGraphs((PageOfGraphs)gro);
            	if (gro instanceof Axis) {
                	Axis axis = (Axis)gro;
                    if (axis.getDimension() == 0)
                		return new PsXAxis(axis);
            		else
                    	return new PsYAxis(axis);
                }
            	if (gro instanceof Graph)
                	return new PsGraph((Graph)gro);
            	if (gro instanceof Curve)
                	return new PsCurve((Curve)gro);

                GrObject newGro = searchForOtherVersion(gro, fromKey, "Ps");
                if (newGro != null)
                    return newGro;
                else
                    // Can't identify it, so add it as it is.
                    return gro;
            }
        }


        // If this is a J2dElement...
        if (target instanceof J2dElement) {
        	// Allow other J2dElements to be added.
    		if (gro instanceof J2dElement) {
            	return gro;
        	}
            else {
            	// Check if it is one of the known objects which have Ps
                // equivalents.
            	if (gro instanceof PageOfGraphs)
                	return new PageOfJ2dGraphs((PageOfGraphs)gro);
            	if (gro instanceof Axis) {
                	Axis axis = (Axis)gro;
                    if (axis.getDimension() == 0)
                		return new J2dXAxis(axis);
            		else
                    	return new J2dYAxis(axis);
                }
            	if (gro instanceof Graph)
                	return new J2dGraph((Graph)gro);
            	if (gro instanceof Curve)
                	return new J2dCurve((Curve)gro);

             GrObject newGro = searchForOtherVersion(gro, fromKey, "J2d");
             if (newGro != null)
                 return newGro;
             else
                 // Can't identify it, so add it as it is.
                 return gro;
            }
        }
        return null;
    }

//------------------------------------------------------------------------------
/**
 * Convert one graphical element type to another.
 */
    public static GrObject searchForOtherVersion(GrObject target,
        String fromKey, String toKey) {

        Class fromClass = target.getClass();
        String fromName = fromClass.getName();
        String toName = fromName.replaceAll(fromKey, toKey);
        toName = toName.replaceAll(fromKey.toLowerCase(), toKey.toLowerCase());
        toName = toName.replaceAll(fromKey.toUpperCase(), toKey.toUpperCase());

        try {
            Class toClass = Class.forName(toName);
            Class[] parameterTypes = new Class[1];
            parameterTypes[0] = fromClass;
            java.lang.reflect.Constructor constructor =
                toClass.getConstructor(parameterTypes);
            Object[] initargs = new Object[1];
            initargs[0] = target;
            return (GrObject)constructor.newInstance(initargs);
        }
        catch (Exception e) {

        }

        return null;
    }

//------------------------------------------------------------------------------
/**
 * Should draw this element's content.
 */
    public void draw() {

    	ListIterator iterator = m_children.listIterator(0);
        while (iterator.hasNext()) {
        	GraphicalElement child = (GraphicalElement)iterator.next();
            child.draw();
        }
    }

//------------------------------------------------------------------------------
/**
 * Sets the rectangles that will be used for transfomations.
 */
    public void setRects(Rectangle2D.Double rect1, Rectangle2D.Double rect2,
    	Rectangle2D.Double rect3) {

    	m_rootTrans = rect1;
        m_parentTrans = rect2;
        m_contentTrans = rect3;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return whether this object has changed and needs to be redrawn.
 */
    public boolean getNeedsRedraw() {
    	return m_bNeedsRedraw;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject o, Object arg) {

    	m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }
}




