/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2000-09-18
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import gov.usgs.gr.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

//==============================================================================
/**
 * Calculates a color from a value based on the formula
 * colorComponent[i] = (value - offset[i]) / range[i]
 * for each of the color components Red, Green, and Blue.
 */
public class InterpolatingColorScheme extends gov.usgs.gr.GrObject
	implements gov.usgs.gr.visual.ColorScheme {

    public class ColorAtValue {
    	public ColorAtValue () {
        }
        public ColorAtValue(float r, float g, float b, double v) {
        	red = r;
            green = g;
            blue = b;
            value = v;
        }
    	public double value = 0;
        public float red = 0;
        public float green = 0;
        public float blue = 0;
        public boolean bRemove = false;
    }

    /** A list of ColorAtValue objects. */
    protected ArrayList m_colorValues = new ArrayList();

    /** The minimum number of values passed out in getProperties. */
    protected int m_numVals = 1;

    /** The input range array for Red, Green, and Blue. */
    protected float[] m_range = {100, 100, 100};
    /** The amount to bleach the color toward white. */
    protected float m_bleach = 0;
    /** Whether to interpolate or use discrete color. */
    protected boolean m_bDiscrete = false;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/InterpolatingColorScheme.gif";

//------------------------------------------------------------------------------
/**
 * Construct an empty object.
 */
    public InterpolatingColorScheme () {

        setDefaults("gov.usgs.gr.visual.InterpolatingColorScheme");
    }

//------------------------------------------------------------------------------
/**
 * Copy constructor.
 */
    public InterpolatingColorScheme (InterpolatingColorScheme ics) {

    	setProperties(ics.getProperties());
    }

//------------------------------------------------------------------------------
/**
 * Implements getColor from ColorScheme interface.
 */
	public float[] getColor(double val) {

    	/** The output color array for Red, Green, and Blue. */
    	float[] color = new float[3];

        // Return 50% gray if there are no entries.
        if (m_colorValues.size() == 0) {
        	color[0] = 0.5f;
            color[1] = 0.5f;
            color[2] = 0.5f;
            return color;
        }

        ColorAtValue lesserCav  = (ColorAtValue)m_colorValues.get(0);
        ColorAtValue greaterCav = (ColorAtValue)m_colorValues.get(0);
        for (int i = 1; i < m_colorValues.size(); i++) {
        	greaterCav = (ColorAtValue)m_colorValues.get(i);
            if (greaterCav.value > val) break;
            lesserCav = greaterCav;
        }

        if (greaterCav == lesserCav || lesserCav.value > val || m_bDiscrete) {
        	color[0] = lesserCav.red;
        	color[1] = lesserCav.green;
        	color[2] = lesserCav.blue;
        }
        else {
        	double totalDiff = greaterCav.value - lesserCav.value;
        	double valDiff = val - lesserCav.value;
            double greaterRatio = valDiff / totalDiff;
            double lesserRatio = 1 - greaterRatio;
        	color[0] = (float)
            	(lesserCav.red * lesserRatio + greaterCav.red * greaterRatio);
        	color[1] = (float)
            	(lesserCav.green * lesserRatio + greaterCav.green
                * greaterRatio);
        	color[2] = (float)
            	(lesserCav.blue * lesserRatio + greaterCav.blue * greaterRatio);
        }

        // Bleach the color, if the bleach factor is greater than zero.
        if (m_bleach > 0) {
        	for (int i = 0; i < 3; i++) {
        		color[i] = (1 - color[i]) * m_bleach + color[i];
        	}
        }

        return color;
	}

//------------------------------------------------------------------------------
/**
 *
 */
    public ArrayList getColorValues() {

        return new ArrayList(m_colorValues);
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Interpolating Color Scheme";
        return s;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("numVals", new GrProperty(m_numVals, "Number of Values"));

        int valNum = 0;

        Iterator iter = m_colorValues.iterator();
        while (iter.hasNext()) {
        	ColorAtValue cav = (ColorAtValue)iter.next();

            // Give the user some color choices...
        	GrProperty p = new GrProperty("(" + Math.round(cav.red * 255) + ","
            	+ Math.round(cav.green * 255) + ","
                + Math.round(cav.blue * 255)
                + ") " + cav.value, "Color At Value");

        	p.component = new ColorButton(cav);
			props.put("colorValue" + valNum, p);

            valNum++;
        }

        while (valNum < m_numVals) {
        	ColorAtValue cav = new ColorAtValue(1, 1, 1, 0);

            // Give the user some color choices...
        	GrProperty p = new GrProperty("(1,1,1)0", "Color At Value");

        	p.component = new ColorButton(cav);
            p.save = false;
			props.put("colorValue" + valNum, p);

            valNum++;
        }

        props.put("discrete", new GrProperty(m_bDiscrete, "Discrete Colors"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

                if (p.value instanceof String &&
                    !(p.component instanceof ColorButton)) {

                    setProperty(key, (String)p.value);
                    continue;
                }

            	if (key.equals("numVals")) {
            		m_numVals = (Integer)p.value;
                    if (m_numVals < 1 || m_numVals > 50)
                    	m_numVals = 1;
                }
            	if (key.startsWith("colorValue")) {
                	if (p.component instanceof ColorButton) {
                    	setColorAtValue(((ColorButton)p.component).m_cav);
                    }
                    else {
                    	setColorAtValue((String)p.value);
                    }
                }
                if (key.equals("bleach"))
                    m_bleach = (Float)p.value;
                if (key.equals("discreet") || key.equals("discrete"))
                    m_bDiscrete = (Boolean)p.value;
			}
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("numVals")) {
        	m_numVals = Math.round(Float.parseFloat(value));
            if (m_numVals < 1 || m_numVals > 50)
            	m_numVals = 1;
        }
        if (key.startsWith("colorValue")) {
        	setColorAtValue(value);
        }
        if (key.equals("bleach"))
            m_bleach = Float.parseFloat(value);
        if (key.equals("discreet") || key.equals("discrete"))
            m_bDiscrete = Boolean.parseBoolean(value);
    }

//------------------------------------------------------------------------------
/**
 * Converts a string of the form "(R, G, B) V" to a ColorAtValue and
 * inserts it in the m_colorValues list sorted by value.
 */
    public void setColorAtValue(String s) {

        // Provide a way to clear all existing values.
		if (s.toLowerCase().equals("clear")) {
        	m_colorValues.clear();
            return;
        }

    	// Parse it as "(R, G, B) V" in the range 0-255.
        int i0 = s.indexOf('(');
        int i1 = s.indexOf(',', i0 + 1);
        int i2 = s.indexOf(',', i1 + 1);
        int i3 = s.indexOf(')', i2 + 1);

        float r = 0, g = 0, b = 0, val = 0;
        try {
        	r = Float.parseFloat(
            	s.substring(i0 + 1, i1).trim()) / 255;
            g = Float.parseFloat(
            	s.substring(i1 + 1, i2).trim()) / 255;
            b = Float.parseFloat(
            	s.substring(i2 + 1, i3).trim()) / 255;
            val = Float.parseFloat(
            	s.substring(i3 + 1).trim());
        }
        catch (NumberFormatException e) {
        	return;
        }

        setColorAtValue(new ColorAtValue(r, g, b, val));

        m_bHasChanged = true;
	}

//------------------------------------------------------------------------------
/**
 * Inserts the cav at the proper place in the color value list.
 */
    public void setColorAtValue(ColorAtValue newCav) {

        boolean bFound = false;
        for (int i = 0; i < m_colorValues.size(); i++) {
        	ColorAtValue cav = (ColorAtValue)m_colorValues.get(i);
            if (newCav.value == cav.value) {
            	m_colorValues.remove(i);
            	if (newCav.bRemove == true)
            		return;
            }
            if (newCav.value <= cav.value && !newCav.bRemove) {
            	m_colorValues.add(i, newCav);
                bFound = true;
                break;
            }
        }

        if (!bFound && !newCav.bRemove) {
        	m_colorValues.add(m_colorValues.size(), newCav);
        }
    }

//------------------------------------------------------------------------------
    public void regenerateColorScheme(double min, double max,
        double coefficient, float[][] colors) {

        m_colorValues.clear();

        // Use the parabola to interpolate the y value.
        double[] coeffs = gov.usgs.gr.data.DoubleDataSeries.fitParabola(
            0, 0, 0.5, coefficient, 1, 1);
        double exponent = 1;
        if (coefficient > 0 && coefficient < 0.99999)
            exponent = 1 / (1 - coefficient);
        if (coefficient < 0 && coefficient > -0.99999)
            exponent = 1 + coefficient;


        double range = max - min;

        // Format the number for output.
        java.text.NumberFormat form = null;
        java.text.NumberFormat form0 = new java.text.DecimalFormat("0");
        java.text.NumberFormat form1 = new java.text.DecimalFormat("0.#");
        java.text.NumberFormat form2 = new java.text.DecimalFormat("0.##");
        java.text.NumberFormat form3 = new java.text.DecimalFormat("0.###");
        java.text.NumberFormat form4 = new java.text.DecimalFormat("0.####");
        java.text.NumberFormat form5 = new java.text.DecimalFormat("0.#####");
        java.text.NumberFormat form6 = new java.text.DecimalFormat("0.######");

        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < 10; i++) {
            float val = (float)(Math.pow(0.1 * (float)i, exponent));
            val *= range;
            val += min;

            // Choose the approriate format depending on the size of the
            // number.
            if (val > 1000000)  form = form0;
            else {
                if (val > 100000) form = form0;
                else {
                    if (val > 10000)  form = form1;
                    else {
                        if (val > 1000) form = form2;
                        else {
                            if (val > 100) form = form3;
                            else {
                                if (val > 10) form = form4;
                                else form = form5;
                            }
                        }
                    }
                }
            }

            setColorAtValue("(" + colors[i][0] + "," + colors[i][1] + ","
                + colors[i][2] + ")" + val);
        }

        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return true if discrete color should be shown.
 */
    public boolean getDiscrete() {
        return m_bDiscrete;
    }

//------------------------------------------------------------------------------
/**
 * Creates a button that brings up a dialog to edit the color and value.
 */
//    public JComponent createColorButton(ColorAtValue oldCav) {
public class ColorButton extends JButton {

	ColorAtValue m_cav = null;
    ColorValueDialog m_dialog = null;

	public ColorButton(ColorAtValue oldCav) {

        super("" + oldCav.value);

        // Give the user a file name button...
        m_cav = new ColorAtValue(oldCav.red, oldCav.green,
        	oldCav.blue, oldCav.value);
        setBackground(new Color(m_cav.red, m_cav.green, m_cav.blue));
        if ((0.5 * m_cav.red + 1.4 * m_cav.green + 0.3 * m_cav.blue) > 1)
        	setForeground(Color.black);
        else
        	setForeground(Color.white);

        Frame frame = GrController.getParentFrame(this);
        if (frame == null)
            return;

        m_dialog = new ColorValueDialog(frame, m_cav);

		addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent e) {

//                showDialog(colorButton, cav);
                m_dialog.m_cav = m_cav;
                javax.swing.SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        m_dialog.setVisible(true);
                    }
                });

                if (m_dialog.m_cav != null) {
                	m_cav = m_dialog.m_cav;
                }

                String buttonText = "";
                if (m_cav.bRemove == true)
                	buttonText += "Remove ";
                buttonText += m_cav.value;
                setText(buttonText);

        		setBackground(
                	new Color(m_cav.red, m_cav.green, m_cav.blue));

        		if ((0.5 * m_cav.red + 1.4 * m_cav.green + 0.3 * m_cav.blue)
                	> 1)

        			setForeground(Color.black);
        		else
        			setForeground(Color.white);
			}
		});

    }
}

//------------------------------------------------------------------------------
/**
 * Shows a file dialog and returns the file selected.
 */
//    public void showDialog(java.awt.Container container, ColorAtValue cav) {
public class ColorValueDialog extends JDialog {

    public ColorAtValue m_cav = null;

	public ColorValueDialog (java.awt.Container container, ColorAtValue cav) {

        super(GrController.getParentFrame(container), "Edit", true);

        final JDialog dialog = this;
        final Frame parentFrame = GrController.getParentFrame(this);
        m_cav = new ColorAtValue(cav.red, cav.green, cav.blue, cav.value);
        final JTextField valueTextField = new JTextField("" + cav.value);

//		java.awt.Frame frame = GrController.getParentFrame(container);

        // Show the dialog.
//        final JDialog dialog = new JDialog(frame,
//        	"Choose Color and Value");

    	// Set up button listeners.
		JButton okButton = new JButton("OK");
		okButton.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(ActionEvent e) {
            	try {
                	m_cav.value = Float.parseFloat(valueTextField.getText());
            		m_cav.bRemove = false;
                }
                catch (NumberFormatException nfe) {
                	m_cav = null;
                }
				dialog.dispose();
			}
		});
		JButton cancelButton = new JButton("Cancel");
		cancelButton.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(ActionEvent e) {
            	m_cav = null;
				dialog.dispose();
			}
		});
		JButton removeButton = new JButton("Remove");
		removeButton.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(ActionEvent e) {
            	m_cav.bRemove = true;
				dialog.dispose();
			}
		});

        // Set the layout managers for the panels.
		JPanel mainPanel = new JPanel();
        mainPanel.setLayout(new BorderLayout());
		JPanel topPanel = new JPanel();
		topPanel.setLayout(new BorderLayout());
		JPanel lowerPanel = new JPanel();
		lowerPanel.setLayout(new BorderLayout());

        // Set the text in each component.

		dialog.getContentPane().add(mainPanel);
		mainPanel.add(topPanel, BorderLayout.NORTH);
		mainPanel.add(lowerPanel, BorderLayout.SOUTH);

		topPanel.add(valueTextField, BorderLayout.NORTH);
        valueTextField.setEditable(true);

        JButton colorButton = new JButton("Change Color");
		colorButton.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(ActionEvent e) {
					java.awt.Color c = javax.swing.JColorChooser.showDialog(
                    	parentFrame, "Choose a Color",
                        new Color(m_cav.red, m_cav.green, m_cav.blue));
                    if (c == null) return;
                    float[] rgb = c.getRGBComponents(null);
                    m_cav.red = rgb[0];
                    m_cav.green = rgb[1];
                    m_cav.blue = rgb[2];
			}
		});
		topPanel.add(colorButton, BorderLayout.SOUTH);
//        topPanel.setBorder(BorderFactory.createTitledBorder(
//        	"Data Format Class"));

//        topPanel.setBorder(BorderFactory.createTitledBorder("Levels of Undo"));
		lowerPanel.add(okButton, BorderLayout.WEST);
		lowerPanel.add(removeButton, BorderLayout.CENTER);
		lowerPanel.add(cancelButton, BorderLayout.EAST);

        dialog.getContentPane().add(mainPanel);
        dialog.setResizable(false);
        dialog.pack();
//        dialog.setVisible(true);
    }
}

}




