/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2008-10-14
 * Date started:    1999-05-19
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import com.sun.opengl.util.GLUT;
import gov.usgs.gr.GrHashList;
import gov.usgs.gr.GrProperty;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.glu.GLU;

/**
 * An x-y graph with ticks, labels, a grid, and multiple curves drawn inside.
 * This class and the components withing in it should be extended for drawing
 * to a particular output device.
 */
public abstract class Markup extends gov.usgs.gr.visual.GraphicalElement
    implements gov.usgs.gr.visual.gl.GlElement {

    /** The x coordinate of this mark up. */
    public double x = 0;
    /** The y coordinate of this mark up. */
    public double y = 0;
    /** Whether coordinates should be displaed to the user and saved out
     *  to XML files. */
    protected boolean m_bUserChangeable = true;
    /** The array of coordinate level names. */
    protected String[] m_levelChoices = {"Graph Screen Pixels", "Graph 0 to 1",
        "Frame Interior 0 to 1", "Frame Interior User Space"};
    /**
     * Whether to save time by drawing a simpler representation of the
     * markup.
     */
    protected boolean m_bQuickDraw = false;
    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;
    /** The OpenGL utility toolkit class with all the glut methods. */
    protected GLUT glut = null;
    /**
     * Tells whether the x coordinate applies to m_outer (0), m_inner (1),
     * or m_innerUser (2) coordinate systems.
     */
    protected int m_xCoordinateLevel = 0;
    /** The y version of m_xCoordinateLevel. */
    protected int m_yCoordinateLevel = 0;
    /**
     * Describes the size of the outer region.
     */
    protected Rectangle2D.Double m_outer =
    	new Rectangle2D.Double(0, 0, 1, 1);
    /**
     * Describes the size of the inner region.
     */
    protected Rectangle2D.Double m_inner =
    	new Rectangle2D.Double(0, 0, 1, 1);
    /**
     * Describes the user coordinate system used in the inner region.
     */
    protected Rectangle2D.Double m_innerUser =
    	new Rectangle2D.Double(0, 0, 1, 1);
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/Markup.gif";

//------------------------------------------------------------------------------
/**
 * Construct an empty markup.
 */
    public Markup () {
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.remove("title");
        props.remove("showTitle");

        // Give the user some X Coordinate Level choices...
        GrProperty p = new GrProperty("", "X Coordinate Transformation",
            true, m_bUserChangeable, m_bUserChangeable);
        javax.swing.JComboBox jcb = new javax.swing.JComboBox();

        // Fill the X level combo box.
        for (int i = 0; i < m_levelChoices.length; i++) {
        	jcb.addItem(m_levelChoices[i]);
        }
        jcb.setSelectedIndex(m_xCoordinateLevel);

        p.value = jcb.getItemAt(m_xCoordinateLevel);
        p.component = jcb;
		props.put("xCoordLevel", p);

        // Give the user some Y Coordinate Level choices...
        p = new GrProperty("", "Y Coordinate Transformation",
            true, m_bUserChangeable, m_bUserChangeable);
        jcb = new javax.swing.JComboBox();

        // Fill the Y level combo box.
        for (int i = 0; i < m_levelChoices.length; i++) {
        	jcb.addItem(m_levelChoices[i]);
        }
        jcb.setSelectedIndex(m_yCoordinateLevel);

        p.value = jcb.getItemAt(m_yCoordinateLevel);
        p.component = jcb;
		props.put("yCoordLevel", p);

        props.put("x", new GrProperty(x, "X", true, m_bUserChangeable,
            m_bUserChangeable));
        props.put("y", new GrProperty(y, "Y", true, m_bUserChangeable,
            m_bUserChangeable));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);

            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("xCoordLevel")) {
        			// Check each name.
        			for (int i = 0; i < m_levelChoices.length; i++) {
        				String name = (String)p.value;
                        if (name.equals(m_levelChoices[i])) {
                            m_xCoordinateLevel = i;
                        }
                    }
                }
            	if (key.equals("yCoordLevel")) {
        			// Check each name.
        			for (int i = 0; i < m_levelChoices.length; i++) {
        				String name = (String)p.value;
                        if (name.equals(m_levelChoices[i])) {
                            m_yCoordinateLevel = i;
                        }
                    }
                }
            	if (key.equals("x")) {
            		x = (Double)p.value;
            	}
            	if (key.equals("y")) {
            		y = (Double)p.value;
            	}
            }
            if (obj instanceof String) {
            	String value = (String)obj;

                setProperty(key, value);
			}
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("xCoordLevel")) {
            // Check each name.
            for (int i = 0; i < m_levelChoices.length; i++) {
                String name = value;
                if (name.equals(m_levelChoices[i])) {
                    m_xCoordinateLevel = i;
                }
            }
        }
        if (key.equals("yCoordLevel")) {
            // Check each name.
            for (int i = 0; i < m_levelChoices.length; i++) {
                String name = value;
                if (name.equals(m_levelChoices[i])) {
                    m_yCoordinateLevel = i;
                }
            }
        }
        if (key.equals("x")) {
            x = Double.parseDouble(value);
        }
        if (key.equals("y")) {
            y = Double.parseDouble(value);
        }

        m_bNeedsRedraw = true;
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        if (!m_bShow) return null;

        if (dimension == 0 && m_xCoordinateLevel == 3) {
            Point2D.Double p = new Point2D.Double(x, x);
            return p;
        }
        if (dimension == 1 && m_yCoordinateLevel == 3) {
            Point2D.Double p = new Point2D.Double(y, y);
            return p;
        }

        return null;
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {

        if (glut == null || newgl != gl || newglu != glu)
            glut = new GLUT();

        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 */
    public abstract void drawGL();

//------------------------------------------------------------------------------
/**
 * Should draw this mark up to a PostScript file..
 */
    public abstract void drawPS(PrintWriter ps);

//------------------------------------------------------------------------------
/**
 * Sets the rectangles that will be used for transfomations.
 */
    public void setRects(Rectangle2D.Double rect1, Rectangle2D.Double rect2,
    	Rectangle2D.Double rect3) {

    	m_outer = rect1;
        m_inner = rect2;
        m_innerUser = rect3;
    }

//------------------------------------------------------------------------------
/**
 * Sets the transformation to be used for the x and y coordinates.
 */
    public void setCoordinateLevels(int xLevel, int yLevel) {

    	m_xCoordinateLevel = xLevel;
    	m_yCoordinateLevel = yLevel;
    }

//------------------------------------------------------------------------------
/**
 * Sets whether the user can change and save the location.
 */
    public void setUserChangeable(boolean val) {

    	m_bUserChangeable = val;
    }

//------------------------------------------------------------------------------
/**
 * Converts the x coordinate to display coordinates.
 */
    public double transformX(double x) {

    	if (m_xCoordinateLevel >= 3) {
            if (m_innerUser == null) return 0;

        	x -= m_innerUser.x;
            x /= m_innerUser.width;
        }
    	if (m_xCoordinateLevel >= 2) {
            if (m_inner == null) return 0;

            x *= m_inner.width;
        	x += m_inner.x;
        }
    	if (m_xCoordinateLevel >= 1) {
            if (m_outer == null) return 0;

            x *= m_outer.width;
        	x += m_outer.x;
        }

        return x;
    }

//------------------------------------------------------------------------------
/**
 * Converts the x coordinate to display coordinates.
 */
    public double transformY(double y) {

    	if (m_yCoordinateLevel >= 3) {
            if (m_innerUser == null) return 0;

        	y -= m_innerUser.y;
            y /= m_innerUser.height;
        }
    	if (m_yCoordinateLevel >= 2) {
            if (m_inner == null) return 0;

            y *= m_inner.height;
        	y += m_inner.y;
        }
    	if (m_yCoordinateLevel >= 1) {
            if (m_outer == null) return 0;

            y *= m_outer.height;
        	y += m_outer.y;
        }

    	return y;
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

        String s = null;
        if (m_bShow)
            s = "Markup";
        else
            s = "(Markup)";

        return s;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }
}




