/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-04-06
 * Date started:    1999-05-14
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.awt.Rectangle;
import java.awt.geom.*;
import java.util.*;

//==============================================================================
/**
 * Applies formatting to a PageOfGraphs.
 */
public class PageFormatter extends GrObject {

	/** The title at the top of the page. */
    protected PageOfGraphs m_page;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/PageFormatter.gif";

//------------------------------------------------------------------------------
/**
 * Create an empty formatter.
 */
    public PageFormatter () {
    }

//------------------------------------------------------------------------------
/**
 * Create a page formatter and give it a page to format.
 */
    public PageFormatter (PageOfGraphs page) {

    	m_page = page;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Sets the page that will be formatted.
 */
    public void setPage(PageOfGraphs page) {

    	m_page = page;
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Page Formatter";

        return s;
 	}

//------------------------------------------------------------------------------
/**
 * Makes a copy of this formatter.
 * @return	A copy of this formatter.
 */
    public PageFormatter getCopy() {

    	return new PageFormatter();
    }

//------------------------------------------------------------------------------
/**
 * Applies the format to the page as well as making some one time changes.
 */
    public void initialize() {

        synchronized(m_page) {
            m_page.deleteAllMarkups();

            List graphList = m_page.getGraphList();
            Iterator iter = graphList.iterator();
            while (iter.hasNext()) {
                Graph graph = (Graph)iter.next();

                graph.deleteAllMarkups();
                graph.getXAxis().setShowGridLines(true);
                graph.getYAxis().setShowGridLines(true);
                graph.getXAxis().setShowMinorTicks(true);
                graph.getYAxis().setShowMinorTicks(true);
                graph.setShowMainTitle(true);
            }
        }

    	reformat();
    }

//------------------------------------------------------------------------------
/**
 * Applies the format to the page.
 */
    public void reformat() {

        layoutGraphs();

        m_page.m_bNeedsReformat = false;
    }

//------------------------------------------------------------------------------
/**
 * Layout all showing graphs and reproportion each graph within the panel.
 * Triggers a redraw.
 */
    public void layoutGraphs() {

        if (m_page instanceof gov.usgs.gr.visual.ps.PageOfPsGraphs)
            layoutPsGraphs();
        else
            layoutScreenGraphs();
    }

//------------------------------------------------------------------------------
/**
 * Layout all showing graphs and reproportion each graph within the panel.
 * Makes all the graph frames the same height. Triggers a redraw.
 */
    protected void layoutScreenGraphs() {

        if (m_page == null)
            return;

        synchronized (m_page) {
            // Create a list containing the frame height of each graph expressed
            // as a proportion of the total graph height.
            List showList = m_page.getShowList();
            float[] yFrames = new float[showList.size()];
            float yFramesProduct = 1;
            Iterator iter = showList.iterator();
            int i = 0;
            while (iter.hasNext()) {
                Graph graph = (Graph)iter.next();

                // Make sure only the bottom showing graph has x-axis labels.
                if (i == 0) graph.setShowXtext(true);
                else graph.setShowXtext(false);

                graph.layoutFrame();
                Rectangle2D.Double graphFrame = graph.getFrameRect();

                yFrames[i] = (float)graphFrame.height;

                // Create a variable to store the product of all the frame
                // proportions.
                yFramesProduct *= yFrames[i];

                i++;
            }

            // Knowing the total page height and the frame height proportion of
            // each graph, with the goal of making all frame heights equal, the
            // height of each graph can be calculated and set using the
            // following formulas.
            // The sum of all graph heights equals the page height.
            // The frame ratio of a graph, which is known and constant,
            // multiplied by the new height of the graph equals the new frame
            // height, which is the same for all graphs.
            // Solving for the height of each graph gives
            // The new graph height equals the page height multiplied by the
            // product of all frame proportions divided by the frame proportion
            // of this graph, all divided by the sum of all combinations of
            // products of n - 1 of the n the frame proportions.
            // For example, for a three graph page with page height p, and frame
            // proportions r1, r2, and r3, the new graph height h1 would be:
            // h1 = (p * r2 * r3) / (r1 * r2 + r1 * r3 + r2 * r3)

            // Sum up all the combinations of products for the denominator.
            float sumOfProducts = 0;
            for (i = 0; i < yFrames.length; i++) {
                float allExceptThisProduct = 1;
                for (int j = 0; j < yFrames.length; j++) {
                    if (i != j)
                        allExceptThisProduct *= yFrames[j];
                }
                sumOfProducts += allExceptThisProduct;
            }

            // Create a multiplier that only needs to be divided by a given
            // graph's frame proportion to yield the graph's new height.
            float totalHeight = (float)m_page.getContentRect().height;
            float scalar = totalHeight * yFramesProduct / sumOfProducts;

            // Reproportion each graph within the panel.
            int xPos = 0, yPos = 0;
            i = 0;
            iter = showList.iterator();
            while (iter.hasNext()) {
                Graph graph = (Graph)iter.next();

                // Calculate how much to scale the graph so its frame will be
                // the same size as all the others.
                Rectangle2D.Double graphFrame = graph.getFrameRect();

                // Set the new size and position for the graph.  The height of
                // the last graph is whatever is left over, since pixel and
                // point positions are int numbers and the sum of the
                // rounded-off heights might not exactly equal the page height.
                int newWidth = m_page.getContentRect().width;
                int newHeight = 0;
                if (iter.hasNext())
                    newHeight = Math.round(scalar / yFrames[i]);
                else
                    newHeight = m_page.getContentRect().height - yPos;

                // Only change the size and position if the graph is not
                // being dragged to a new location.
                if (graph != m_page.getMovingGraph()) {
                    graph.setSize(newWidth, newHeight);
                    graph.setXPosition(xPos);
                    graph.setYPosition(yPos);
                }

                // The next Y positions is at the top of this graph.
                yPos += newHeight;

                // Set the tick scale so both axes have the same size ticks.
                float xTickScale = 1;
                if (newHeight != 0) xTickScale = m_page.getContentRect().height
                    / (float)newHeight;
                graph.setXTickScale(xTickScale);
                float yTickScale = 1;
                if (newWidth != 0) yTickScale = m_page.getContentRect().height
                    / (float)newWidth;
                graph.setYTickScale(yTickScale);

                // PS graphs need to be layed out again because their local
                // sizes may have changed.
                graph.layoutFrame();

                graph.setFrameRect(graphFrame);

                i++;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Layout all showing graphs and reproportion each graph within the panel.
 * Triggers a redraw.
 */
    protected void layoutPsGraphs() {

        layoutScreenGraphs();
    }
}
