/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-03
 * Date started:    1998-10-20
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import gov.usgs.gr.data.*;
import gov.usgs.gr.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.ps.*;
import gov.usgs.sfhydro.gr.formats.VariableSizeFormatter;
import java.awt.Rectangle;
import java.awt.geom.*;
import java.util.*;

//==============================================================================
/**
 * A group of graphs layed out on a page.  This class should be extended to
 * output to various devices such as a PostScript file, OpenGL context, or
 * Java2D context.
 */
public abstract class PageOfGraphs extends gov.usgs.gr.visual.GraphicalElement
    implements Runnable {
    
    static String XML_ELEMENT_NAME = "page";
    
    /** The object that lays out and formats this page. */
    protected PageFormatter m_formatter = null;
    /** The currently selected curve's data series, if any. */
    protected List m_selectedSeriesList =
	Collections.synchronizedList(new ArrayList());
    /** A list of Graph objects. */
//    protected List m_graphList = Collections.synchronizedList(new ArrayList());
    protected List<Graph> m_graphList = new ArrayList();
    /** A list of showing graphs. */
    protected List<Graph> m_showList =
	Collections.synchronizedList(new ArrayList());
    /** A list of selected graphs. */
    protected List m_selectedGraphList =
	Collections.synchronizedList(new ArrayList());
    /** A list of miscellaneous graphical objects that are drawn on the page. */
    protected List m_markupList =
	Collections.synchronizedList(new ArrayList());
    /** A list of commands used to create this page. */
    protected List m_commandList =
	Collections.synchronizedList(new ArrayList());
    /** The color of the background, in an RGB array. */
    protected float[] m_bgColor = {0, 0, 0};
    /**
     * Describes the size of the display area and the location of the graph
     * in that area.  Given in screen or page coordinates.
     */
    protected Rectangle m_display = new Rectangle(0, 0, 200, 200);
    /**
     * Describes the size of the display area and the location of the graph
     * in that area after margins.
     */
    protected Rectangle m_content = new Rectangle(0, 0, 200, 200);
    /**
     * Describes the window that the user draws for zooming in.
     */
    protected Rectangle m_zoomRect = new Rectangle(0, 0, 0, 0);
    /** The thread that is started for rendering. */
    protected Thread m_thread = new Thread(this);
    /** Tells whether a draw operation is unfolding. */
    protected boolean m_bDrawing = false;
    /** The graph that is being moved, or null if none are being moved. */
    protected Graph m_movingGraph = null;
    
    /** A GrObject that holds all of the PostScript options for the page. */
    public PsOptions m_psOptions = null;
    /**
     * Whether to draw thinned and unthinned curves over each other to help
     * debug.
     */
    protected boolean m_bDebugPointReduction = false;
    /** Signals a reformat at the next redraw. */
    public boolean m_bNeedsReformat = false;
    
    /** Disables drawing temporarily. */
    protected boolean m_bDisableRedraw = false;
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/PageOfGraphs.gif";
    
//------------------------------------------------------------------------------
    /**
     * Create a blank page for graphs.
     */
    public PageOfGraphs() {
	
	super();
	
	add(new gov.usgs.gr.visual.ps.PsOptions());
	add(new gov.usgs.gr.visual.PageFormatter());
	
	setDefaults("gov.usgs.gr.visual.PageOfGraphs");
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a page of graphs cloned from an existing one.   Data is shared.
     */
    public PageOfGraphs(PageOfGraphs page) {
	
	super();
	
	page.copyInto(this, true);
	
	// Copy the other attributes from the old page.
	m_display.width = page.getWidth();
	m_display.height = page.getHeight();
	Rectangle oldContent = page.getContentRect();
	m_content.x = oldContent.x;
	m_content.y = oldContent.y;
	m_content.width = oldContent.width;
	m_content.height = oldContent.height;
	
	// Make a copy of the command list.
	List oldCommandList = page.getCommandList();
	Iterator iter = oldCommandList.iterator();
	while (iter.hasNext()) {
	    m_commandList.add(new String((String)iter.next()));
	}
	
	// Update the showing and selected lists.
	updateLists();
	
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public GrHashList getProperties() {
	
	GrHashList props = super.getProperties();
	
	props.remove("show");
	
	// Give the user some formatter choices...
	String currentFormatterName = m_formatter.getClass().getName();
	GrProperty p = new GrProperty(currentFormatterName, "Formatter");
	javax.swing.JComboBox jcb = new javax.swing.JComboBox();
	jcb.setEditable(true);
	boolean bFormatterFound = false;
	String[] formatterPossibilities = {
	    "gov.usgs.gr.visual.PageFormatter",
	    "gov.usgs.sfhydro.gr.formats.WorkingFormatter",
	    "gov.usgs.sfhydro.gr.formats.ReportFormatter",
	    "gov.usgs.sfhydro.gr.formats.VariableSizeFormatter",
	    "gov.usgs.sfhydro.gr.formats.VariableShapeFormatter"
	};
	
	for (int i = 0; i < formatterPossibilities.length; i++) {
	    String name = formatterPossibilities[i];
	    jcb.addItem(name);
	    if (name.equals(currentFormatterName)) {
		jcb.setSelectedIndex(i);
		bFormatterFound = true;
	    }
	}
	
	if (!bFormatterFound) {
	    jcb.addItem(currentFormatterName);
	    jcb.setSelectedIndex(formatterPossibilities.length);
	}
	
	p.component = jcb;
	p.save = false;
	props.put("formatter", p);
	
	ColorProperty cp = new ColorProperty();
	cp.description = "Background Color";
	cp.setColor(m_bgColor);
	cp.display = false;
	cp.save = false;
	props.put("bgColor", cp);
	
	return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public void setProperties(HashMap props) {
	
	super.setProperties(props);
	
	GrProperty p;
	
	Iterator iter = props.keySet().iterator();
	while (iter.hasNext()) {
	    String key = (String)iter.next();
	    Object obj = props.get(key);
	    if (obj instanceof GrProperty) {
		p = (GrProperty)obj;
		
		if (key.equals("formatter")) {
		    Object o = GrController.createObject((String)p.value,
			"gov.usgs.gr.visual.PageFormatter", null);
		    if (o != null) {
			add((PageFormatter)o);
		    }
		    
		    m_bNeedsRedraw = true;
		}
		if (key.equals("bgColor")) {
		    ColorProperty cp = (ColorProperty)p;
		    m_bgColor = cp.getColorArray();
		}
	    }
	    if (obj instanceof String) {
		String value = (String)obj;
		
		setProperty(key, value);
	    }
	}
	
	m_bHasChanged = true;
	notifyObservers();
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public void setProperty(String key, String value) {
	
	super.setProperty(key, value);
	
	if (key.equals("formatter")) {
	    Object o = GrController.createObject(value,
		"gov.usgs.gr.visual.PageFormatter", null);
	    if (o != null) {
		add((PageFormatter)o);
	    }
	    
	    m_bNeedsRedraw = true;
	}
	if (key.equals("bgColor")) {
	    m_bgColor = ColorProperty.getColorArray(value);
	    m_bNeedsRedraw = true;
	}
	if (key.startsWith("gui")) {
	    try {
		GrObject newGro =
		    (GrObject)Class.forName(value).newInstance();
		
		add(newGro);
	    } catch (Exception ex) {
		ex.printStackTrace();
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Returns an XML name of this element.
     */
    public String getXmlElementName() {
	return "page";
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
	
	if (TREE_ICON_HOLDER.size() <= 0)
	    setTreeIcon();
	
	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {
	
	javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);
	
	if (icon == null)
	    icon = super.setTreeIcon();
	
	TREE_ICON_HOLDER.clear();
	TREE_ICON_HOLDER.add(icon);
	
	return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param formatter	The new page formatting object.
     */
    public void setFormatter(PageFormatter formatter) {
	
	add(formatter);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The formatting object for this page.
     */
    public PageFormatter getFormatter() {
	return m_formatter;
    }
    
//------------------------------------------------------------------------------
    /**
     * Changes the size and shape of the panel's associated drawing area and the
     * objects that are drawn on it.
     * @param width     The new width for the drawing area.
     * @param heicht    The new height for the drawing area.
     */
    public synchronized void reshape(int width, int height ) {
	
	if (width == m_display.width && height == m_display.height) return;
	
	m_display.width = width;
	m_display.height = height;
	m_content.width = width;
	m_content.height = height;
	
	m_formatter.layoutGraphs();
    }
    
//------------------------------------------------------------------------------
    /**
     * Removes all children from the page except the formatter.
     */
    public synchronized void clear() {
	
	// Remove everything except the axes.
	ArrayList removables = new ArrayList();
	Iterator iter = m_children.iterator();
	while (iter.hasNext()) {
	    Object o = iter.next();
	    if (!(o instanceof PageFormatter)) {
		removables.add(o);
	    }
	}
	
	iter = removables.iterator();
	while (iter.hasNext()) {
	    remove((GrObject)iter.next());
	}
	
	m_title = "";
	
	updateLists();
    }
    
//------------------------------------------------------------------------------
    /**
     * Removes child at index i.
     */
    public synchronized GrObject remove(int i) {
	
	if (m_children.get(i) == m_formatter) return null;
	
	GrObject gro = super.remove(i);
	m_graphList.remove(gro);
	updateLists();
	updateSelectedSeriesList();
	
	m_bNeedsReformat = true;
	
	return gro;
    }
    
//------------------------------------------------------------------------------
    /**
     * Removes child gro.
     */
    public synchronized boolean remove(GrObject gro) {
	
	if (gro == m_formatter) return false;
	
	boolean bFound = super.remove(gro);
	m_graphList.remove(gro);
	updateLists();
	updateSelectedSeriesList();
	
	m_bNeedsReformat = true;
	
	return bFound;
    }
    
//------------------------------------------------------------------------------
    /**
     * Clears the drawing area and redraws the contents.
     */
    public synchronized void draw() {
	
	// Check if this method is being executed by this object's own thread.
	// If it isn't, split a new thread off to do the work.
	if (!m_bDrawing) {
	    start();
	    return;
	} else {
	    
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Makes copies of all the data series pointed to by objects on the page,
     * and uses those so that no other pages will be sharing this page's data.
     */
    public synchronized void unshareData() {
	
	// Go through each curve on each graph and replace its data series
	// with a copy of its data series.
	Iterator graphIter = m_graphList.iterator();
	while (graphIter.hasNext()) {
	    Graph graph = (Graph)graphIter.next();
	    List curveList = graph.getCurves();
	    Iterator curveIter = curveList.iterator();
	    while (curveIter.hasNext()) {
		Curve curve = (Curve)curveIter.next();
		DoubleDataSeries newSeries =
		    (DoubleDataSeries)curve.getSeries().createCopy(true);
		curve.setSeries(newSeries);
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Clears the drawing area and redraws the contents.
     */
    public abstract void drawFromThread();
    
//------------------------------------------------------------------------------
    /**
     * Cause the selected graphs to pan based on the mouse being dragged.
     * @param xDiff	The amount to pan in the x direction.
     * @param yDiff	The amount to pan in the y direction.
     */
    public synchronized void pan(double xDiff, double yDiff) {
	
	// Select the graph the mouse is over and send it a pan command.
	// Since all the graphs share a common X AxisLimits object, they
	// will all follow if the X axis moves.
	Iterator iter = m_selectedGraphList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    graph.pan(xDiff, yDiff);
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Cause one of the graphs to zoom into the area specified by m_zoomRect.
     */
    public synchronized void zoomWindow() {
	
	List tmList =
            findChildInstancesOf("gov.usgs.gr.tools.TimeMarker", false);
	Iterator tmIter = tmList.iterator();
	while (tmIter.hasNext()) {
	    TimeMarker tm = (TimeMarker)tmIter.next();
	    tm.setLockTimeWhenZooming(false);
	}
	
	Iterator iter = m_showList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    
	    boolean bSuccess = graph.zoomWindow(m_zoomRect);
	    if (bSuccess) break;
	}
	
	// Reset the zoom window to nothing.
	m_zoomRect.width = 0;
	m_zoomRect.height = 0;
	
	tmIter = tmList.iterator();
	while (tmIter.hasNext()) {
	    TimeMarker tm = (TimeMarker)tmIter.next();
	    tm.setLockTimeWhenZooming(true);
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Select or deselect all of the points inside the zoom window
     * @param bAdding	True if selecting points, false if deselecting.
     * @param bExclusive	True if only one curve should be selected or,
     *						if multiple curves are already selected, only
     *						the selected curves can have more points selected.
     * @param bDeselectFirst	True if old selection is deselected before selecting
     *							new points.
     */
    public synchronized void selectWindow(boolean bAdding, boolean bExclusive,
	boolean bDeselectFirst) {
	
	m_selectedSeriesList.clear();
	
	Iterator iter = m_showList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    
	    // Only select more if nothing is yet selected, or multiple
	    // selection is enabled.
	    if ((m_selectedSeriesList.size() == 0) || !bExclusive) {
		List seriesList = graph.selectWindow(
		    new Rectangle(m_zoomRect), bAdding, bExclusive,
		    bDeselectFirst);
		
		// Copy any elements from the graphs's selected series list to
		// the page's selected series list.
		Iterator iter2 = seriesList.iterator();
		while (iter2.hasNext()) {
		    m_selectedSeriesList.add(iter2.next());
		}
	    }
	}
	
	// Reset the zoom window to nothing.
	m_zoomRect.width = 0;
	m_zoomRect.height = 0;
	
    }
    
//------------------------------------------------------------------------------
    /**
     * Initiate panning when the mouse is pressed.
     * @param mouseX	An x coordinate that is above the selected graph.
     * @param mouseY    A y coordinate that is above the selected graph.
     * @param bExclusive	True if only the specified graphs should be selected
     *						and all others deselected.
     * @param bAdding	True if selecting, false if deselecting.
     */
    public synchronized void selectShowingGraph(int graphNum,
	boolean bExclusive, boolean bAdding) {
	
	if (graphNum < 0 || graphNum > m_showList.size()) return;
	
	// Check each graph to see if the mouse is over it.  If it is,
	// tell the graph it is selected, otherwise deselect it.
	int index = 0;
	Iterator iter = m_showList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    if (index == graphNum) {
		graph.setSelected(bAdding);
	    } else {
		if (bExclusive) graph.setSelected(!bAdding);
	    }
	    index++;
	}
	
	// Put this graph in the selected list.
	updateLists();
    }
    
//------------------------------------------------------------------------------
    /**
     * Initiate panning when the mouse is pressed.
     * @param mouseX	An x coordinate that is above the selected graph.
     * @param mouseY    A y coordinate that is above the selected graph.
     * @param bExclusive	True if only the specified graphs should be selected
     *						and all others deselected.
     * @param bAdding	True if selecting, false if deselecting.
     */
    public synchronized void selectGraph(int mouseX, int mouseY,
	boolean bExclusive, boolean bAdding) {
	
	// Check each graph to see if the mouse is over it.  If it is,
	// tell the graph it is selected, otherwise deselect it.
	Iterator iter = m_graphList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    if (graph.isMouseOver(mouseX, ((int)m_display.height - mouseY))) {
		graph.setSelected(bAdding);
	    } else {
		if (bExclusive) graph.setSelected(!bAdding);
	    }
	}
	
	// Put this graph in the selected list.
	updateLists();
    }
    
//------------------------------------------------------------------------------
    /**
     * Select all of the graphs.
     */
    public synchronized void selectAllGraphs() {
	
	// Select or deselect all graphs.
	Iterator iter = m_graphList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    graph.setSelected(true);
	}
	
	// Put this graph in the selected list.
	updateLists();
    }
    
//------------------------------------------------------------------------------
    /**
     * Select all of the showing graphs.
     */
    public synchronized void selectAllShowingGraphs() {
	
	// Check each graph to see if the mouse is over it.  If it is,
	// tell the graph it is selected, otherwise deselect it.
	Iterator iter = m_showList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    graph.setSelected(true);
	}
	
	// Put this graph in the selected list.
	updateLists();
    }
    
//------------------------------------------------------------------------------
    /**
     * Move the movingGraph by the specified number of display units.
     * @param xDiff	The amount to move the graph in the x direction.
     * @param yDiff	The amount to move the graph in the y direction.
     */
    public synchronized void moveGraph(int xDiff, int yDiff) {
	
	if (m_movingGraph == null) return;
	
	m_movingGraph.setXPosition(m_movingGraph.getXPosition() + xDiff);
	m_movingGraph.setYPosition(m_movingGraph.getYPosition() + yDiff);
	orderGraphs();
    }
    
//------------------------------------------------------------------------------
    /**
     * Check if the mouse is over the side bar area.  If it is, set the bar to
     * be drawn.
     * @param mouseX	An x coordinate that is above the selected graph.
     * @param mouseY    A y coordinate that is above the selected graph.
     * @param bDown		True if the mouse button is down.
     * @return	True if the new state is different from the old.
     */
    public synchronized boolean setSideBar(int mouseX, int mouseY,
	boolean bDown) {
	
	boolean bChanged = false;
	m_movingGraph = null;
	
	if (m_showList.size() == 1)
	    bDown = false;
	
	// Check each graph to see if the mouse is over it.  If it is,
	// tell the graph it is selected, otherwise deselect it.
	Iterator iter = m_showList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    Rectangle rect = graph.getDisplayRect();
	    
	    // Check if the mouse is over the side bar.
	    if (graph.isMouseOver(mouseX, ((int)m_display.height - mouseY))
	    && mouseX > rect.x && mouseX < (rect.x + 10)
	    && m_showList.size() > 1) {
		
		// Draw the side bar if it was not already drawn.
		boolean bShowing = graph.getShowSideBar();
		boolean bWasDown = graph.getSideBarDown();
		if (!bShowing || (bWasDown != bDown)) {
		    graph.setShowSideBar(true);
		    graph.setSideBarDown(bDown);
		    if (bWasDown) m_formatter.layoutGraphs();
		    bChanged = true;
		}
		if (bDown) m_movingGraph = graph;
	    } else {
		// Turn off the side bar if it was previously on.
		boolean bShowing = graph.getShowSideBar();
		boolean bWasDown = graph.getSideBarDown();
		if (bShowing || (bWasDown != bDown)) {
		    graph.setShowSideBar(false);
		    graph.setSideBarDown(bDown);
		    bChanged = true;
		}
	    }
	}
	
//        if (bChanged && !bDown) {
//        	layoutGraphs();
//        }
	
	return bChanged;
    }
    
//------------------------------------------------------------------------------
    /**
     * Should return a new Graph object.
     */
    protected abstract Graph newGraph();
    
//------------------------------------------------------------------------------
    /**
     * Should return a new Graph object.
     */
    protected abstract Graph newGraph(Graph graph);
    
//------------------------------------------------------------------------------
    /**
     * Delete all markup objects from the list.
     */
    public synchronized void deleteAllMarkups() {
	
	m_markupList.clear();
	
	Iterator iter = m_children.iterator();
	while (iter.hasNext()) {
	    GrObject gro = (GrObject)iter.next();
	    if (gro instanceof Markup) iter.remove();
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Add new time series data to the view, creating new graphs as needed.
     * @param timeSeriesList    A list of two-dimensional DoubleDataSeries to
     *				be drawn.
     */
    public synchronized void addData(List dataSeriesList) {
	
	Iterator iter, iter2;
	
	// Create a list of uniqe data types in the series list.
	ArrayList dataTypeList = new ArrayList();
	
	iter = dataSeriesList.iterator();
	while (iter.hasNext()) {
	    Object o = iter.next();
	    if (!(DoubleDataSeries.countDims(o, 2)))
		continue;
	    DoubleDataSeries series = (DoubleDataSeries)o;
	    Graph matchingGraph = null;
	    
	    int currentType = series.getType();
	    
	    // Check if the current type is in an existing graph.
	    iter2 = m_graphList.iterator();
	    while (iter2.hasNext()) {
		Graph graph = (Graph)iter2.next();
		List curves = graph.getCurves();
		if (graph.isEmpty()) {
		    matchingGraph = graph;
		    break;
		}
		int graphType =
		    ((Curve)curves.get(0)).getSeries().getType();
		if (graphType == currentType) {
		    matchingGraph = graph;
		    break;
		}
	    }
	    
	    if (matchingGraph == null) {
		matchingGraph = addNewGraph();
	    }
	    matchingGraph.add(series);
	}
	// Put the new graph in the showing list.
	updateLists();
	// Update the content ranges.
	updateZoomContents();
	
	// Cause the new graph to be shown on the screen.
	m_formatter.layoutGraphs();
    }
    
//------------------------------------------------------------------------------
    /**
     * Add an object to the page.
     */
    public synchronized boolean add(GrObject gro) {
	
	if (m_children.contains(gro)) return false;
	
	if (gro instanceof GraphicalElement && !(gro instanceof PageOfGraphs)) {
	    super.add(gro);
	    if (gro instanceof Graph) {
		m_graphList.add((Graph)gro);
	    } else {
		if (gro instanceof Markup) {
		    m_markupList.add(gro);
		}
	    }
	    
	    m_bNeedsReformat = true;
	    
	    return true;
	} else {
	    if (gro instanceof DataSeries) {
		ArrayList al = new ArrayList();
		al.add(gro);
		addData(al);
		
		return true;
	    }
	    // If PageFormatter, should remove existing formatter and place it
	    // first.
	    if (gro instanceof PageFormatter) {
		// Look for the old formatter and remove it.
		PageFormatter formatter = (PageFormatter)gro;
		Iterator iter = m_children.iterator();
		while (iter.hasNext()) {
		    Object o = iter.next();
		    if (o instanceof PageFormatter) iter.remove();
		}
		
		// Add the new formatter at the beginning of the list.
		super.add(formatter);
		m_children.remove(formatter);
		m_children.add(0, formatter);
		
		// Only reinitialize if it is a new formatter.
		if (m_formatter != formatter) {
		    m_formatter = formatter;
		    formatter.setPage(this);
		    m_formatter.initialize();
		}
		
		return true;
	    }
	    // If PsOptions, should remove existing formatter and place it
	    // second.
	    if (gro instanceof PsOptions) {
		// Look for the old formatter and remove it.
		PsOptions pso = (PsOptions)gro;
		Iterator iter = m_children.iterator();
		while (iter.hasNext()) {
		    Object o = iter.next();
		    if (o instanceof PsOptions) iter.remove();
		}
		
		// Add the new ps options at the beginning of the list
		// after the formatter.
		super.add(pso);
		m_children.remove(pso);
		int formatterIndex = m_children.indexOf(m_formatter);
		if (formatterIndex < -1) formatterIndex = -1;
		m_children.add(formatterIndex + 1, pso);
		
		m_psOptions = pso;
		
		return true;
	    }
	}
	
	// If gro is a panel provider, remove any instances of the same class.
	if (gro instanceof gov.usgs.gr.visual.gui.GuiPanelProvider) {
	    List panelProviderList =
                findChildInstancesOf(gro.getClass(), false, false);
	    Iterator iter = panelProviderList.iterator();
	    while (iter.hasNext()) {
		GrObject panelProvider = (GrObject) iter.next();
		remove(panelProvider);
	    }
	}
	
	super.add(gro);
	return true;
	
//        return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Read an XML version of this element and its children.
     */
    public synchronized void readFromDOM(org.w3c.dom.Node node,
	Map allObjects, List unlinkedObjects) {
	
	super.readFromDOM(node, allObjects, unlinkedObjects);
	
	// Put the new graph in the showing list.
	updateLists();
	// Update the content ranges.
	updateZoomContents();
	
	// Cause the new graph to be shown on the screen.
	m_formatter.layoutGraphs();
    }
    
//------------------------------------------------------------------------------
    /**
     * Perform tasks that had to wait until all XML input had been read.
     */
    public synchronized void finishSAXInput(Map idObjects) {
	
	finishSAXForChildren(idObjects);
	
	// Put the new graph in the showing list.
	updateLists();
	// Update the content ranges.
	updateZoomContents();
	
	// Cause the new graph to be shown on the screen.
	m_formatter.layoutGraphs();
    }
    
//------------------------------------------------------------------------------
    /**
     * Add a new blank graph to the page.
     * @return The newly created graph.
     */
    public synchronized Graph addNewGraph() {
	
	Graph graph = newGraph();
	Zoom zoom = null;
	
	if (m_graphList.size() == 0) {
	    // The first graph created makes its own X zoom.
	    graph.setShowXtext(true);
	} else {
	    // All graphs after the first share the X Zoom with the first.
	    zoom = ((Graph)m_graphList.get(0)).getXZoom();
	    graph.setXZoom(zoom);
            graph.setShowXtext(false);
        }
	
	// Check for a time marker on the page.
	Graph mapGraph = null;
	TimeMarker tm = (TimeMarker)firstChildInstanceOf(
            "gov.usgs.gr.tools.TimeMarker", false);
	if (tm != null) {          
            graph.add(new LineMarkup());
            tm.add(graph.getXZoom());

            if (m_graphList.size() > 0 && m_formatter instanceof
		gov.usgs.sfhydro.gr.formats.VariableSizeFormatter) {

		mapGraph = (Graph)m_graphList.get(m_graphList.size() - 1);
                GrHashList props = mapGraph.getProperties();
                GrProperty p = (GrProperty)props.get("aspect1to1");
                if (((Boolean)p.value) == false)
                    mapGraph = null;
	    }
	}

	deferUpdates();
	
	add(graph);
	
	processDeferredUpdates();

        if (mapGraph != null) {
            if (m_formatter instanceof
                gov.usgs.sfhydro.gr.formats.VariableSizeFormatter) {

                m_formatter.setProperty("numVals", ""
                    + (m_graphList.size() - 1));
                GrHashList props = m_formatter.getProperties();
                GrProperty p = (GrProperty)props.get("size0");
                float size = (Float)p.value;
                m_formatter.setProperty("size" + (m_graphList.size() - 2),
                    "" + size);
            }
            
	    m_children.remove(mapGraph);
	    m_children.add(mapGraph);
            updateLists();
	    mapGraph.setShowXtext(false);
	    ((Graph)m_graphList.get(0)).setShowXtext(true);
	    mapGraph.setXZoom((Zoom)(mapGraph.getXZoom().createCopy(true)));
	}
	
	return graph;
    }
    
//------------------------------------------------------------------------------
    /**
     * Hides the currently selected graph or graphs.
     */
    public synchronized void hideGraph() {
	
	// Hide all selected graphs.
	Iterator iter = m_selectedGraphList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    // Deselect all points in the graph.
	    graph.selectWindow(new Rectangle(-1, -1, -1, -1), false, false,
		true);
	    m_selectedSeriesList.clear();
	    graph.setShow(false);
	}
	
	// Redo the layout and redraw.
	updateLists();
	
	m_bNeedsReformat = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Deletes the currently selected graph or graphs.
     */
    public synchronized void deleteSelectedGraphs() {
	
	// Delete all selected graphs from the main graph list.
	ArrayList selectedList = new ArrayList(m_selectedGraphList);
	Iterator iter = selectedList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    
	    remove(graph);
	}
	
	
	// Update lists, redo the layout and redraw.
	m_bNeedsReformat = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Lock the y-axis so it won't be affected by pans or zooms.  Acts as a toggle
     * if the axis is already locked.
     */
    public synchronized void showMoreGraphs() {
	
	// Check each graph to see if it is selected.
	Iterator iter = m_graphList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    // Show the first hidden graph, this break out of this loop.
	    if (!graph.getShow()) {
		graph.setShow(true);
		break;
	    }
	}
	
	// Redo the layout and redraw.
	updateLists();
	
	m_bNeedsReformat = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Order the graphs according to their current x and y values.
     */
    public synchronized void orderGraphs() {
	
	int previousHeight = Integer.MIN_VALUE;
	int previousIndex = 0;
	
	// Loop through each graph in the list.
	int index = 0;
	while (index < m_children.size()) {
	    GrObject gro = (GrObject)m_children.get(index);
	    if (!(gro instanceof Graph)) {
		index++;
		continue;
	    }
	    
	    Graph graph = (Graph)gro;
	    int yPosition = graph.getYPosition();
	    // Check if the graph is showing.
	    if (graph.getShow()) {
		if (yPosition < previousHeight) {
		    m_children.remove(index);
		    m_children.add(previousIndex, graph);
		    
		    previousHeight = yPosition;
		    index = previousIndex + 1;
		} else {
		    previousHeight = yPosition;
		    previousIndex = index;
		    index++;
		}
	    } else {
		index++;
	    }
	}
	
	// Redo the layout and redraw.
	updateLists();
	
	m_bNeedsReformat = true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {
	
	String s = "Page of Graphs: " + m_title;
	return s;
    }
    
//------------------------------------------------------------------------------
    /**
     * Implements the start method for Runnable.
     */
    public void start() {
	
/// System.out.println("Starting new draw thread...");
	m_bDrawing = true;
//		m_thread = new Thread(this);	// Uncomment this for multithreading
//		m_thread.start();				// Uncomment this for multithreading
///    	drawFromThread();   			// Comment this out for multithreading
///		m_bDrawing = false;				// Comment this out for multithreading
    }
    
//------------------------------------------------------------------------------
    /**
     * Implements the run method for Runnable.
     */
    public synchronized void run() {
	drawFromThread();
	m_bDrawing = false;
///System.out.println("...Finished thread.  Set m_bDrawing = " + m_bDrawing);
    }
    
//------------------------------------------------------------------------------
    /**
     * Resets the drawing variable.
     */
    public void stopDrawing() {
	
	m_bDrawing = false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Make sure the show list and selected list reflect the current state.
     */
    public synchronized void updateLists() {
	
	// Repopulate the graph list.
	m_graphList.clear();
	Enumeration enumeration = children();
	while (enumeration.hasMoreElements()) {
	    GrObject gro = (GrObject)enumeration.nextElement();
	    if (gro instanceof Graph) m_graphList.add((Graph)gro);
	}
	
	m_showList.clear();
	m_selectedGraphList.clear();
	
	// Calculate the number of graphs showing in the panel.
	// Make sure no hidden graphs are selected, and that at
	// least one showing graph is selected.
	Iterator iter = m_graphList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    
	    // Add the graph to the show list if it is showing.
	    if (graph.getShow()) {
		m_showList.add(graph);
		
		// Make sure at least one showing graph is selected.
		if (graph.getSelected()) {
		    m_selectedGraphList.add(graph);
		}
	    } else {
		// Make sure hidden graphs aren't selected.
		graph.setSelected(false);
	    }
	}
	
	// If nothing is selected, select the first showing graph.
	if (m_selectedGraphList.size() == 0) {
	    Graph graph = null;
	    if (m_showList.size() != 0)
		graph = (Graph)m_showList.get(0);
	    if (graph != null) {
		graph.setSelected(true);
		m_selectedGraphList.add(graph);
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Make sure the selected series list reflect the current state.
     */
    public synchronized void updateSelectedSeriesList() {
	
	m_selectedSeriesList.clear();
	
	Iterator graphIter = m_showList.iterator();
	while (graphIter.hasNext()) {
	    Graph graph = (Graph)graphIter.next();
	    List curveList = graph.getCurves();
	    Iterator curveIter = curveList.iterator();
	    while (curveIter.hasNext()) {
		Curve curve = (Curve)curveIter.next();
		if (curve.getShow()) {
		    DoubleDataSeries series = curve.getSeries();
		    if (series.getNumSelected() > 0) {
			m_selectedSeriesList.add(series);
		    }
		}
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the minimum value between all showing graphs on this page for the
     * specified dimension.
     * @param dimension The dimension to report the minimum for.  If it is 0, it
     *                  will report the min x value.  If 1, it will report the min
     *                  y value.
     * @return The mimimum value for the data along the specified dimension.
     */
    public synchronized double getMin(int dimension){
	
	// Start out with a min that is larger than any other.
	double min = Double.POSITIVE_INFINITY;
	
	// Check each curve for the lowest min.
	Iterator iter = m_showList.iterator();
	while (iter.hasNext()) {
	    double graphMin = ((Graph)iter.next()).getMin(dimension);
	    if (graphMin < min) min = graphMin;
	}
	
	return min;
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the maximum value between all showing graphs on this page for the
     * specified dimension.
     * @param dimension The dimension to report the minimum.  If it is 0, it
     *                  will report the max x value.  If 1, it will report the max
     *                  y value.
     * @return The maximum value for the data along the specified dimension.
     */
    public synchronized double getMax(int dimension){
	
	// Start out with a min that is larger than any other.
	double max = Double.NEGATIVE_INFINITY;
	
	// Check each curve for the lowest min.
	Iterator iter = m_showList.iterator();
	while (iter.hasNext()) {
	    double graphMax = ((Graph)iter.next()).getMax(dimension);
	    if (graphMax > max) max = graphMax;
	}
	
	return max;
    }
    
//------------------------------------------------------------------------------
    /**
     * Update the min and max content for each zoom on the page.
     */
    public synchronized void updateZoomContents(){
	
	ArrayList updatedZooms = new ArrayList();
	
	// Run through each graph's x and y zoom.
	Graph graph = null;
	int dimension = 0;
	Iterator graphIter = m_showList.iterator();
	while (graphIter.hasNext() || dimension == 1) {
	    Zoom zoom;
	    graph = (Graph)graphIter.next();
	    for (dimension = 0; dimension <= 1; dimension++) {
		if (dimension == 0) { // Doing X zoom for the current graph.
		    zoom = graph.getXZoom();
		} else { // Doing Y zoom for the current graph.
		    zoom = graph.getYZoom();
		}
		
		// Check if this zoom was already updated.
		boolean bAlreadyUpdated = false;
		Iterator updatedIter = updatedZooms.iterator();
		while (updatedIter.hasNext()) {
		    Zoom updatedZoom = (Zoom)updatedIter.next();
		    if (zoom == updatedZoom) {
			bAlreadyUpdated = true;
		    }
		}
		
		// Update the zoom if it wasn't already.
		if (!bAlreadyUpdated) {
		    double min = Double.POSITIVE_INFINITY;
		    double max = Double.NEGATIVE_INFINITY;
		    
		    Iterator iter2 = m_showList.iterator();
		    while (iter2.hasNext()) {
			Graph graph2 = (Graph)iter2.next();
			Zoom zoom2;
			if (dimension == 0) zoom2 = graph2.getXZoom();
			else zoom2 = graph2.getYZoom();
			
			if (zoom == zoom2) {
			    Point2D.Double p = graph2.getBounds(dimension);
			    min = Math.min(min, p.x);
			    max = Math.max(max, p.y);
			}
			// Else no match.
		    }
		    
		    // Adjust the min and max to allow a small padding area
		    // around the content.
		    double extra = 0;
		    if ((max - min) < 0.0001) {
			// Allow flat line curves to be shown in the middle of
			// the graph.
			extra = (Math.abs(min) + 1) * 0.0001;
		    } else {
			extra = (max - min) * 0.02;
		    }
		    
		    // Pass the range of content on to the zoom.
		    zoom.setContentLimits(min - extra, max + extra);
		    
		    // Add this zoom to the list of updated zooms.
		    updatedZooms.add(zoom);
		}
		// else Zoom Already Updated
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Return the nearest point along a curve to the screen point.
     * @param i	The x position of the point on the screen.
     * @param j	The y position of the point on the screen.
     */
    public synchronized Graph getNearestChild(int i, int j) {
	
	Point2D.Double p = new Point2D.Double(0, 0);
	
	Iterator iter = m_showList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    if (graph.isMouseOver(i, m_display.height - j)) {
		return graph;
	    }
	}
	
	return null;
    }
    
//------------------------------------------------------------------------------
    /**
     * Return the nearest point along a curve to the screen point.
     * @param i	The x position of the point on the screen.
     * @param j	The y position of the point on the screen.
     */
    public synchronized Point2D.Double getNearestPoint(int i, int j) {
	
	Point2D.Double p = new Point2D.Double(0, 0);
	
	Iterator iter = m_showList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    if (graph.isMouseOver(i, m_display.height - j)) {
		p = graph.getNearestPoint(i, m_display.height - j);
		break;
	    }
	}
	
	return p;
    }
    
//------------------------------------------------------------------------------
    /**
     * Return the nearest curve to the screen point.
     * @param i	The x position of the point on the screen.
     * @param j	The y position of the point on the screen.
     */
    public synchronized Curve getNearestCurve(int i, int j) {
	
	Curve c = null;
	
	Iterator iter = m_showList.iterator();
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    if (graph.isMouseOver(i, m_display.height - j)) {
		c = graph.getNearestCurve();
		break;
	    }
	}
	
	return c;
    }
    
//------------------------------------------------------------------------------
    /**
     * Converts screen coordinates to user coordinates.
     * @param screenRect	A rectangle described in screen coordinates.
     * return	The user coordinates describing screenRect.
     */
    public Rectangle2D.Double screenToUser(Rectangle screenRect, Graph graph) {
	
	if (graph == null) return null;
	if (!m_children.contains(graph)) return null;
	if (!graph.getShow()) return null;
	
	double oldY = screenRect.y;
	
	screenRect.y = m_display.height - (screenRect.y + screenRect.height);
	
	return graph.screenToUser(screenRect);
    }
    
//------------------------------------------------------------------------------
    /**
     * Checks if any selected points lies within the bounds of the screenRect.
     * @param screenRect	A rectangle described in screen coordinates.
     * @param bAddIfDeselected	True if unselected points should be selected and
     *							then counted as being contained inside the area.
     * return	True if a selected point lies with the rectangle
     */
    public synchronized boolean containsSelectedPoint(Rectangle screenRect,
	boolean bAddIfDeselected) {
	
	if (m_selectedSeriesList.size() == 0) return false;
	
	Graph graph = (Graph)m_selectedGraphList.get(0);
	
	Rectangle2D.Double r = screenToUser(screenRect, graph);
	
	double[] min = new double[2];
	double[] max = new double[2];
	min[0] = r.x;
	min[1] = r.y;
	max[0] = r.x + r.width;
	max[1] = r.y + r.height;
	
	// Return true if any of the selected series contains the point.
	Iterator iter = m_selectedSeriesList.iterator();
	while (iter.hasNext()) {
	    DataSeries series = (DataSeries)iter.next();
	    if (series.containsSelectedPoint(min, max, bAddIfDeselected))
		return true;
	}
	
	return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @param enum	A list graphs to peruse for unique zooms.
     * @return The list of x zooms used by the graphs in the list.
     */
    public synchronized List getXZooms(List graphList) {
	
	// Create a list for unique zoom objects.
	List zoomList = new ArrayList();
	
	Iterator iter = graphList.iterator();
	// Look for zooms in each selected graph.
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    
	    // Check that this graph's zoom hasn't already been counted.
	    boolean bAlreadyOnList = false;
	    Iterator iter2 = zoomList.iterator();
	    while (iter2.hasNext()) {
		Zoom zoom = (Zoom)iter2.next();
		if (zoom == graph.getXZoom()) bAlreadyOnList = true;
	    }
	    
	    // Add the zoom to the list, if it isn't already there.
	    if (!bAlreadyOnList) {
		zoomList.add(graph.getXZoom());
	    }
	}
	
	return zoomList;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The list of x zooms used by the selected graphs.
     */
    public List getSelectedXZooms() {
	
	return getXZooms(m_selectedGraphList);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The list of x zooms used by the selected graphs.
     */
    public List getAllXZooms() {
	
	return getXZooms(m_graphList);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @param enum	A list of graphs to peruse for unique zooms.
     * @return The list of y zooms used by the graphs in the list.
     */
    public synchronized List getYZooms(List graphList) {
	
	// Create a list for unique zoom objects.
	List zoomList = new ArrayList();
	
	Iterator iter = graphList.iterator();
	// Look for zooms in each selected graph.
	while (iter.hasNext()) {
	    Graph graph = (Graph)iter.next();
	    
	    // Check that this graph's zoom hasn't already been counted.
	    boolean bAlreadyOnList = false;
	    Iterator iter2 = zoomList.iterator();
	    while (iter2.hasNext()) {
		Zoom zoom = (Zoom)iter2.next();
		if (zoom == graph.getYZoom()) bAlreadyOnList = true;
	    }
	    
	    // Add the zoom to the list, if it isn't already there.
	    if (!bAlreadyOnList) {
		zoomList.add(graph.getYZoom());
	    }
	}
	
	return zoomList;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The list of y zooms used by the selected graphs.
     */
    public List getSelectedYZooms() {
	
	return getYZooms(m_selectedGraphList);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The series (curve) with selected points.
     */
    public List<DataSeries> getSelectedSeriesList() {
	
	return m_selectedSeriesList;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The list of y zooms used by the selected graphs.
     */
    public List getAllYZooms() {
	
	return getYZooms(m_graphList);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The list of selected graphs for this page.
     */
    public List getSelectedGraphs() {
	
	return m_selectedGraphList;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The list of showing graphs for this page.
     */
    public List getShowList() {
	
	return m_showList;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The list of graphs for this page.
     */
    public List<Graph> getGraphList() {
	
	return new ArrayList(m_graphList);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The width of the page.
     */
    public int getWidth() {
	
	return m_display.width;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The height of the page.
     */
    public int getHeight() {
	
	return m_display.height;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The content area of the page.
     */
    public Rectangle getContentRect() {
	
	return m_content;
    }
    
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param rect The new content area of the page.
     */
    public void setContentRect(Rectangle rect) {
	
	m_content = rect;
    }
    
//------------------------------------------------------------------------------
    /**
     * Changes the window that is drawn when the user is zooming to a certain area.
     * @param x1	The x-coordinate of the first corner of the rectangle.
     * @param y1	The y-coordinate of the first corner of the rectangle.
     * @param x2	The x-coordinate of the second corner of the rectangle.
     * @param y2	The y-coordinate of the second corner of the rectangle.
     */
    public synchronized void setZoomWindow(int x1, int y1, int x2, int y2) {
	
	if (x2 > x1) {
	    m_zoomRect.x = x1;
	    m_zoomRect.width = x2 - x1;
	} else {
	    m_zoomRect.x = x2;
	    m_zoomRect.width = x1 - x2;
	}
	
	y1 = m_display.height - y1;
	y2 = m_display.height - y2;
	if (y2 > y1) {
	    m_zoomRect.y = y1;
	    m_zoomRect.height = y2 - y1;
	} else {
	    m_zoomRect.y = y2;
	    m_zoomRect.height = y1 - y2;
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return Whether to point reduction debugging on.
     */
    public boolean getDebugPointReduction() {
	return m_bDebugPointReduction;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param val Whether to point reduction debugging on.
     */
    
    public synchronized void setDebugPointReduction(boolean val) {
	
	if (val == m_bDebugPointReduction) return;
	
	m_bDebugPointReduction = val;
	
	// Go through each curve on each graph and set it's debug value
	// to false.
	Iterator graphIter = m_graphList.iterator();
	while (graphIter.hasNext()) {
	    Graph graph = (Graph)graphIter.next();
	    List curveList = graph.getCurves();
	    Iterator curveIter = curveList.iterator();
	    while (curveIter.hasNext()) {
		Curve curve = (Curve)curveIter.next();
		curve.setDebugPointReduction(val);
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides default update() in GrObject.  Called by an observed GrObject's
     * notifyObservers() method to handle changes to the observed object.
     * May use the notifyObservers() method to propagate the event handling up to
     * this object's observers.
     * @param gro   The observed object that changed.
     * @param arg   Additional information that the updated object may choose to
     *              send.
     */
    public void update(GrObject gro, Object arg) {
	
	// Handle PS options.
	if (gro instanceof gov.usgs.gr.visual.ps.PsOptions) {
	    setDebugPointReduction(((PsOptions)gro).m_bDebugPointReduction);
	}
	
	// Check if the object is a graphical element that needs to be redrawn.
	if (gro instanceof GraphicalElement) {
	    if (((GraphicalElement)gro).getNeedsRedraw()) {
		m_bNeedsRedraw = true;
	    }
	}
	
	m_bHasChanged = true;
	
	if (!m_bDrawing) {
	    notifyObservers();
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The graph that is being moved, or null if none.
     */
    public Graph getMovingGraph() {
	return m_movingGraph;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The list of commands.
     */
    public List getCommandList() {
	return m_commandList;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param list	The new command list.
     */
    public void setCommandList(List list) {
	m_commandList = list;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The title of the page.
     */
    public String getTitle() {
	return m_title;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param title	The new title of the page.
     */
    public void setTitle(String title) {
	m_title = title;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return The value of the m_bDrawing variable.
     */
    public boolean getDrawing() {
	return m_bDrawing;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param value The value of the m_bDrawing variable.
     */
    public void setDrawing(boolean value) {
	m_bDrawing = value;
    }
    
//------------------------------------------------------------------------------
    /**
     * Attribute modification method.
     * @param value The new value.
     */
    public void setDisableRedraw(boolean value) {
	m_bDisableRedraw = value;
    }
}


