/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual;

import gov.usgs.gr.*;
import gov.usgs.gr.data.AsciiIO;
import java.util.*;
import java.lang.Math;

//==============================================================================
/**
 * A Zoom keeps track of the current view along an axis, and provides a
 * mechanism for changing that view.  Each zoom object describes the view for
 * one dimension only.  In order to zoom in on an area on a 2D graph, two Zoom
 * objects should be employed:  one for the X-Axis, and one for the Y-Axis.
 * Each Zoom has an outer range, a view range, and a content range.
 * The view range describes the current view (for example 33 to 43).  The
 * outer range describes the boundary within which the view range can
 * exist (for example 0 to 100).  The view range can be shrunken, enlarged,
 * or moved one way or the other, as long as it remains completely within the
 * outer range.  The content range describes the minimum and maximum values
 * that need to be displayed by the Zoom.  When autoscaling, the outer range
 * is set to round numbers such that the content range is wholly contained
 * within the outer range.
 * The Zoom also provides hints to axes that need to draw tick marks and labels.
 * Appropriate tick increments are calculated, as is the appropriate decimal
 * format for numerical labels.
 */
public class Zoom extends gov.usgs.gr.GrObject {

    /**
     * The minimum value shown in the current zoom window.
     * It should always be >= m_outerMin and <= m_viewMax.
     */
    protected double m_viewMin = -Double.MIN_VALUE;
    /**
     * The maximum value shown in the current zoom window.
     * It should always be >= m_viewMin and <= m_outerMax.
     */
    protected double m_viewMax = Double.MIN_VALUE;
    /** The previous "view" min. */
    protected double m_prevMin = -Double.MIN_VALUE;
    /** The previous "view" max. */
    protected double m_prevMax = Double.MIN_VALUE;
    /**
     * The minimum value of any visible object show within this zoom.
     * After autoscaling, m_outerMin should be less than this value.
     */
    protected double m_contentMin = 0;
    /**
     * The maximum value of any visible object show within this zoom.
     * After autoscaling, m_outerMax should be greater than this value.
     */
    protected double m_contentMax = 100;
    /**
     * The lower bounds for the content range.  The content min cannot be
     * higher than this.
     */
    protected double m_highestContentMin = Double.POSITIVE_INFINITY;
    /**
     * The upper bounds for the content range.  The content max cannot be
     * lower than this.
     */
    protected double m_lowestContentMax = Double.NEGATIVE_INFINITY;
    /** The minimum value that could be shown in the zoom window. */
    protected double m_outerMin;
    /** The maximum value that could be shown in the zoom window. */
    protected double m_outerMax;
    /** The diffence between the outer min and max. */
    protected double m_outerLength;
    /** The smallest increment that can be displayed on the axis. */
    protected static double m_smallestInc = 1E-12;

    // Variables for calculating hints for drawing axes.
    /** The major increment that should be used by axes drawn with this zoom. */
    protected double m_majorInc;
    /** The increment that should be used between major increments. */
    protected double m_minorInc;
    /** The offset from which major increments should be measured. */
    protected double m_incOffset = 0;
    /** The format to use for outputting the axis labels. */
    protected String m_format = "0";
//    /**
//     * True if outer limits are fixed and the view limits must be within
//     * them.  False if outer limits grow to fit the view limits.
//     */
//    protected boolean m_bOuterFixed = true;
//    /** True if the view limits should not change, only the step. */
//    protected boolean m_bViewFixed = false;
    /** The preferred number of major divisions along the axis. */
    protected int m_prefNumMajDivs = 5;
    /** The preferred number of minor divisions along each major division. */
    protected int m_prefNumMinDivs = 4;
    /** The step increments to choose from when autoscaling. */
    protected static double[] m_autoSteps = {5, 10, 25, 50, 100};
    /**
     * The scale to use when displaying data.  Used for axes that don't have
     * the same scale for their labels as for the data they display.
     */
    protected double m_unitScale = 1;
    /**
     * Keeps track of an object that is updating this object until that object
     * is finished.  Set using setUpdater().
     */
    protected Object m_updater = null;
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/Zoom.gif";

//------------------------------------------------------------------------------
/**
     * Creates an uninitialized zoom.
     */
    public Zoom() {

        setOuterLimits(0, 100);
        zoomAll();

        setDefaults("gov.usgs.gr.visual.Zoom");
    }

//------------------------------------------------------------------------------
/**
     * Creates a zoom and specifies an outer range.  The initial zoom window
     * will be set to show the entire range.
     * @param min	The minimum value that a zoom window can have.
     * @param max	The maximum value that a zoom window can have.
     */
    public Zoom(double min, double max) {

        setDefaults("gov.usgs.gr.visual.Zoom");

        setContentLimits(min, max);
        setOuterLimits(min, max);
        zoomAll();
    }

//------------------------------------------------------------------------------
/**
     * Copies a zoom from an existing one.
     * @param zoom	The template zoom object to copy.
     */
    public Zoom(Zoom zoom) {

        setProperties(zoom.getProperties());
    }

//------------------------------------------------------------------------------
/**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {

        java.text.NumberFormat form = new java.text.DecimalFormat(m_format);
        String s = "Zoom: " + form.format(m_viewMin) + " --> " + form.format(m_viewMax);

        return s;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0) {
            setTreeIcon();
        }
        return (javax.swing.ImageIcon) TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null) {
            icon = super.setTreeIcon();
        }
        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
     * Provide the list of properties with the latest values.
     */
    public synchronized GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("viewMin", new GrProperty(m_viewMin, "View Min"));
        props.put("viewMax", new GrProperty(m_viewMax, "View Max"));
        props.put("outerMin", new GrProperty(m_outerMin, "Outer Min"));
        props.put("outerMax", new GrProperty(m_outerMax, "Outer Max"));
        props.put("majorTickInc", new GrProperty(m_majorInc, "Major Tick Inc"));
        props.put("minorTickInc", new GrProperty(m_minorInc, "Minor Tick Inc"));
        props.put("tickOffset", new GrProperty(m_incOffset, "Tick Offset"));
        props.put("unitScale", new GrProperty(m_unitScale, "Unit Scale"));
        props.put("labelFormat", new GrProperty(m_format, "Label Format"));
        
        // Write autoSteps, if it is not the default value.
        String sVal = "";
        for (int i = 0; i < m_autoSteps.length; i++) {
            if (i > 0)
                sVal += ", ";
            
            sVal += m_autoSteps[i];
        }
        if (!sVal.equals("5, 10, 25, 50, 100"))
            props.put("autoSteps", new GrProperty(sVal, "Auto Steps"));

        return props;
    }

//------------------------------------------------------------------------------
/**
     * Provide the list of properties with the latest values.
     */
    public synchronized void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String) iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty) obj;

                if (key.equals("majorTickInc")) {
                    m_majorInc = (Double) p.value;
                    m_prefNumMajDivs = 10;
                }
                if (key.equals("minorTickInc")) {
                    m_minorInc = (Double) p.value;
                }
                if (key.equals("viewMin")) {
                    m_viewMin = (Double) p.value;
                }
                if (key.equals("viewMax")) {
                    m_viewMax = (Double) p.value;
                }
                if (key.equals("outerMin")) {
                    m_outerMin = (Double) p.value;
                }
                if (key.equals("outerMax")) {
                    m_outerMax = (Double) p.value;
                }
                if (key.equals("tickOffset")) {
                    m_incOffset = (Double) p.value;
                }
                if (key.equals("unitScale")) {
                    m_unitScale = (Double) p.value;
                }
                if (key.equals("labelFormat")) {
                    try {
                        java.text.NumberFormat form = new java.text.DecimalFormat((String) p.value);
                    } catch (Exception e) {
                        continue;
                    }
                    m_format = (String) p.value;
                }
            }
            if (obj instanceof String) {
                String value = (String) obj;

                if (key.equals("majorTickInc")) {
                    m_majorInc = Double.parseDouble(value);
                }
                if (key.equals("minorTickInc")) {
                    m_minorInc = Double.parseDouble(value);
                }

                if (key.equals("viewMin")) {
                    m_viewMin = Double.parseDouble(value);
                }
                if (key.equals("viewMax")) {
                    m_viewMax = Double.parseDouble(value);
                }

                if (key.equals("outerMin")) {
                    m_outerMin = Double.parseDouble(value);
                }
                if (key.equals("outerMax")) {
                    m_outerMax = Double.parseDouble(value);
                }

                if (key.equals("tickOffset")) {
                    m_incOffset = Double.parseDouble(value);
                }
                if (key.equals("unitScale")) {
                    m_unitScale = Double.parseDouble(value);
                }
                if (key.equals("labelFormat")) {
                    m_format = value;
                }
                if (key.equals("autoSteps")) {
                    try {
                        m_autoSteps = AsciiIO.parseLineAsDoubles(value, ", ");
                    }
                    catch (NumberFormatException e) {
                        // Do nothing.
                    }
                }
            }
        }


        // Check validity of outer limits.
        if (m_outerMax < m_viewMax) {
            m_outerMax = m_viewMax;
        }
        if (m_outerMin > m_viewMin) {
            m_outerMin = m_viewMin;
        }
        if (m_outerMax < m_outerMin) {
            m_outerMax = m_outerMin + 1;
        }
        // Update the outer width and height.
        m_outerLength = m_outerMax - m_outerMin;

        // Check that the current view is within the new limits.
        checkCurrentBounds();
    }

//------------------------------------------------------------------------------
/**
     * Set a single property.
     */
    public synchronized void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("unitScale")) {
            m_unitScale = Double.parseDouble(value);
        }
        if (key.equals("labelFormat")) {
            m_format = value;
        }
        if (key.equals("autoSteps")) {
            try {
                m_autoSteps = AsciiIO.parseLineAsDoubles(value, ", ");
            }
            catch (NumberFormatException e) {
                // Do nothing.
            }
        }
    }

//------------------------------------------------------------------------------
/**
     * Overrides add in GrObject.  Don't add any children.
     */
    public boolean add(GrObject gro) {

        return false;
    }

//------------------------------------------------------------------------------
/**
     * Read the attribute settings from a values stored in a list.
     * Sets this object to match the attributes.  The content range is set to the
     * same as the outer range.
     * <p><b> Pre Conditions </b><br>
     *      None.
     * <p><b> Post Conditions </b><br>
     *      update() needs to be called on this axis and the opposing axis.
     * @param attribs	A list of attribute values containing the outer min and max,
     *					the view min and max, the major and minor incs, and the
     *					format string for numerical labels.
     */
    public synchronized void setAttributes(List attribs) {

        setOuterLimits((double) ( (Float) attribs.get(0)), (double) ( (Float) attribs.get(1)));
        setContentLimits(getOuterMin(), getOuterMax());
        setViewLimits((double) ( (Float) attribs.get(2)), (double) ( (Float) attribs.get(3)));
        setMajorInc((double) ( (Float) attribs.get(4)));
        setMinorInc((double) ( (Float) attribs.get(5)));
        setFormatString((String) attribs.get(6));
//        if (attribs.size() >= 8)
//        	m_unitScale = (double)((Float)attribs.elementAt(7)).doubleValue();
    }

//------------------------------------------------------------------------------
/**
     * Changes the outer limits.  Calls checkCurrentBounds afterwards.
     * <p><b> Pre Conditions </b><br>
     *      None.
     * <p><b> Post Conditions </b><br>
     *      Changes will not be seen until next redraw.
     * @param min	The new minimum value that can be seen.
     * @param max	The new maximum value that can be seen.
     */
    public synchronized void setOuterLimits(double min, double max) {

        // Save the parameters in member variables.
        m_outerMin = min;
        m_outerMax = max;

        // Check validity of outer limits.
        if (m_outerMax < m_outerMin) {
            m_outerMax = m_outerMin + 1;
        }
        // Update the outer width and height.
        m_outerLength = m_outerMax - m_outerMin;

        // Check that the current view is within the new limits.
        checkCurrentBounds();
    }

//------------------------------------------------------------------------------
/**
     * Tells this zoom object the range of values that need to be seen.
     * <p><b> Pre Conditions </b><br>
     *      None.
     * <p><b> Post Conditions </b><br>
     *      Changes will not be seen until next autoscale.
     * @param min	The new minimum value.
     * @param max	The new maximum value.
     */
    public synchronized void setContentLimits(double min, double max) {

        // First, respect the preset allowable content range.
        if (min > m_highestContentMin) {
            min = m_highestContentMin;
        }
        if (max < m_lowestContentMax) {
            max = m_lowestContentMax;
        }
        m_contentMin = min;
        m_contentMax = max;
    }

//------------------------------------------------------------------------------
/**
     * Tells this zoom object the allowabl range for content limits.  This is
     * useful when there is some overriding factor that requires the content
     * limits to be larger than they would otherwise be.  For example, a TimeMarker
     * object may need the current time to be within the content limits,
     * even if there is no curve within that range.
     * @param min	The new minimum value.
     * @param max	The new maximum value.
     */
    public synchronized void setAllowableContentLimits(double min, double max) {

        m_highestContentMin = min;
        m_lowestContentMax = max;
    }

//------------------------------------------------------------------------------
/**
     * Convenience method for setViewLimits(double, double, boolean false).
     * @param min	The new minimum value that will be seen.
     * @param max	The new maximum value that will be seen.
     */
    public synchronized void setViewLimits(double min, double max) {

        setViewLimits(min, max, false);
    }

//------------------------------------------------------------------------------
/**
     * Change the zoom window that is currently being used.  checkCurrentBounds is
     * called afterwards.
     * @param min	The new minimum value that will be seen.
     * @param max	The new maximum value that will be seen.
     * @param bSave True if the "previous" limits should be updated.
     */
    public synchronized void setViewLimits(double min, double max, boolean bSave) {

        // If save is true, save the min and max into the "previous" variables
        // before writing over them.
        if (bSave) {
            m_prevMin = m_viewMin;
            m_prevMax = m_viewMax;
        }

        // Save the parameters in member variables.
        m_viewMin = min;
        m_viewMax = max;

        // Check that the new view is within the outer limits.
        checkCurrentBounds();

        // Notify the observers that this zoom has changed.
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
     * Returns the inverse of the length of the current zoom.
     * @return The ratio to use for scaling from 1 to the length of this zoom.
     */
    public double getScale() {

        double diff = m_viewMax - m_viewMin;
        if (diff == 0) {
            return 1;
        } else {
            return 1 / diff;
        }
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return	The scale multiplier to make axis labels in the same units as
     *			the data.
     */
    public double getUnitScale() {
        return m_unitScale;
    }

//------------------------------------------------------------------------------
/**
     * Attribute modification method.
     * @param	The scale multiplier to make axis labels in the same units as
     *			the data.
     */
    public synchronized void setUnitScale(double val) {
        m_unitScale = val;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The minimum value shown in the current zoom window.
     */
    public double getOuterMin() {
        return m_outerMin;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The minimum content value.
     */
    public double getContentMin() {
        return m_contentMin;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The previous view min.
     */
    public double getPrevMin() {
        return m_prevMin;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The previous view max.
     */
    public double getPrevMax() {
        return m_prevMax;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The minimum value that could be shown in the zoom window.
     */
    public double getViewMin() {
        return m_viewMin;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The maximum value that could be shown in the zoom window.
     */
    public double getOuterMax() {
        return m_outerMax;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The minimum content value.
     */
    public double getViewMax() {
        return m_viewMax;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The maximum content value.
     */
    public double getContentMax() {
        return m_contentMax;
    }


//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The major increment offset.
     */
    public double getIncOffset() {
        return m_incOffset;
    }

//------------------------------------------------------------------------------
/**
     * Attribute modification method.
     * @param inc	The major increment offset.
     */
    public synchronized void setIncOffset(double inc) {
        m_incOffset = inc;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return The format string that goes with the step increment.
     */
    public String getFormatString() {
        return new String(m_format);
    }

//------------------------------------------------------------------------------
/**
     * Attribute modification method.
     * @param s	The new format string.
     */
    public synchronized void setFormatString(String s) {
        m_format = s;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * <p><b> Pre Conditions </b><br>
     *      The number of increments needs to be set.  autoScale() does this.
     * <p><b> Post Conditions </b><br>
     *      None.
     * @return the suggested number of grid increments to use with the axis.
     */
    public double getMajorInc() {
        return m_majorInc;
    }

//------------------------------------------------------------------------------
/**
     * Attribute modification method.
     * @param inc	The new major increment.
     */
    public synchronized void setMajorInc(double inc) {
        m_majorInc = inc;
    }

//------------------------------------------------------------------------------
/**
     * Access method.
     * <p><b> Pre Conditions </b><br>
     *      The number of increments needs to be set.  autoScale() does this.
     * <p><b> Post Conditions </b><br>
     *      None.
     * @return the suggested number of minor tick increments to use with the axis.
     */
    public double getMinorInc() {
        return m_minorInc;
    }

//------------------------------------------------------------------------------
/**
     * Attribute modification method.
     * @param inc	The new major increment.
     */
    public synchronized void setMinorInc(double inc) {
        m_minorInc = inc;
    }

//------------------------------------------------------------------------------
/**
     * Change the zoom window to span from the lowest outer value to the
     * highest outer value.
     */
    public void zoomAll() {

        m_viewMin = m_outerMin;
        m_viewMax = m_outerMax;

        checkIncrementMagnitude();

        // Notify the observers that this zoom has changed.
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
     * Change the zoom window to span from the lowest outer value to the
     * highest outer value.  Applies autoscaling.
     * @param prefNumDivs	The preferred number of divisions for autoscaling.
     */
    public synchronized void zoomAll(int prefNumDivs) {

        m_bDeferUpdates = true;
        autoScale(m_contentMin, m_contentMax, prefNumDivs);
        m_bDeferUpdates = false;
        zoomAll();
    }

//------------------------------------------------------------------------------
/**
     * Save the current zoom so it can be restored later.
     */
    public synchronized void saveZoom() {

        // Save the previous zoom limits.
        m_prevMin = m_viewMin;
        m_prevMax = m_viewMax;
    }

//------------------------------------------------------------------------------
/**
     * Restore the last saved zoom.
     */
    public synchronized void restoreSavedZoom() {

        // Save the previous zoom limits.
        m_viewMin = m_prevMin;
        m_viewMax = m_prevMax;
    }

//------------------------------------------------------------------------------
/**
     * Change the zoom window to previous values.
     */
    public synchronized void zoomPrevious() {

        double tempMin = m_viewMin;
        double tempMax = m_viewMax;
        m_viewMin = m_prevMin;
        m_viewMax = m_prevMax;
        m_prevMin = tempMin;
        m_prevMax = tempMax;

        checkCurrentBounds();
    }

//------------------------------------------------------------------------------
/**
     * Looks at the current zoom and shrinks it if it is out of bounds.  Makes sure
     * the following conditions are true.
     * view min >= outer min AND view max <= outer max AND view min <= view max
     */
    private synchronized void checkCurrentBounds() {

        // Perform range checking
        if (m_viewMin < -Double.MAX_VALUE || m_viewMin > Double.MAX_VALUE) {
            m_viewMin = 0;
        }
        if (m_viewMax < -Double.MAX_VALUE || m_viewMax > Double.MAX_VALUE) {
            m_viewMax = 100;
        }
        // Check validity of view limits within outer limits.
        if (m_viewMin < m_outerMin || m_viewMin > m_outerMax) {
            m_viewMin = m_outerMin;
        }
        if (m_viewMax < m_viewMin || m_viewMax > m_outerMax) {
            m_viewMax = m_outerMax;
        }
        if (m_viewMax < (m_viewMin + m_smallestInc)) {
            m_viewMax = m_viewMin + m_smallestInc;
        }
        checkIncrementMagnitude();
    }

//------------------------------------------------------------------------------
/**
     * Keeps the major and minor increments from being out of scale with the
     * view min and max by more than an order of magnitude.
     */
    private synchronized void checkIncrementMagnitude() {

        // Do some range checking on the tick increments.
        if (Math.abs(m_majorInc) < Float.MIN_VALUE || Math.abs(m_majorInc) > Float.MAX_VALUE || m_prefNumMajDivs == 0 || ((m_viewMax - m_viewMin) / m_majorInc / m_prefNumMajDivs) > 10) {

            m_majorInc = m_viewMax - m_viewMin;
        }
        if (m_majorInc < m_smallestInc) {
            m_majorInc = m_smallestInc;
        }
        if (Math.abs(m_minorInc) < Float.MIN_VALUE || Math.abs(m_minorInc) > Float.MAX_VALUE || m_prefNumMinDivs == 0 || (m_majorInc / m_minorInc / m_prefNumMinDivs) > 10) {

            m_minorInc = m_majorInc;
        }
        if (m_minorInc < m_smallestInc) {
            m_minorInc = m_smallestInc;
        }
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return the difference between the highest and lowest defined values.
     */
    public double getOuterLength() {
        return m_outerLength;
    }

//------------------------------------------------------------------------------
/**
     * Accessor method.
     * @return the difference between the current highest and lowest values.
     */
    public double getViewLength() {
        return m_viewMax - m_viewMin;
    }

//------------------------------------------------------------------------------
/**
     * Zoom in by a small amount, keeping the same center point.
     */
    public synchronized void zoomIn() {

        // Zoom in a percentage of the view range.
        double viewLength = m_viewMax - m_viewMin;

        // Save the previous zoom limits.
        m_prevMin = m_viewMin;
        m_prevMax = m_viewMax;
        double viewRange = m_viewMax - m_viewMin;
        if (Math.abs(m_viewMax) > m_smallestInc) {
            if ((viewRange / Math.abs(m_viewMax)) < (m_smallestInc * 10)) {
                return;
            }
        }

        autoScale(m_viewMin + m_majorInc, m_viewMax - m_majorInc, (int) Math.round(viewRange / m_majorInc));
    }

//------------------------------------------------------------------------------
/**
     * Zoom out by a small amount, keeping the same center point.
     */
    public synchronized void zoomOut() {

        // Zoom out a percentage of the view range.
        double viewLength = m_viewMax - m_viewMin;

        // Save the previous zoom limits.
        m_prevMin = m_viewMin;
        m_prevMax = m_viewMax;

        autoScale(m_viewMin - m_majorInc, m_viewMax + m_majorInc, (int) Math.round((m_viewMax - m_viewMin) / m_majorInc));
    }

//------------------------------------------------------------------------------
/**
     * Pan a specified amount.
     * @param change	The amount to move as a fraction of the view range.
     */
    public synchronized void panPercent(double change) {

        // Convert to graph coodinates
        change *= (m_viewMax - m_viewMin);

        // Move the zoom up or down the axis by change, but not past the
        // outer limits.
        if (change > 0) {
            change = Math.min(change, m_outerMax - m_viewMax);
        } else {
            change = Math.max(change, m_outerMin - m_viewMin);
        }
        setViewLimits(m_viewMin + change, m_viewMax + change);
    }

//------------------------------------------------------------------------------
/**
     * Pan a specified amount.
     * @param change	The amount to move in zoom coordinates.
     */
    public synchronized void pan(double change) {

        // Move the zoom up or down the axis by change, but not past the
        // outer limits.
        if (change > 0) {
            change = Math.min(change, m_outerMax - m_viewMax);
        } else {
            change = Math.max(change, m_outerMin - m_viewMin);
        }
        setViewLimits(m_viewMin + change, m_viewMax + change);
    }

//------------------------------------------------------------------------------
/**
     * Calls autoScale for the current view's min and max.
     * @param prefNumDivs  The preferred number of divisions along the axis.
     */
    public synchronized void autoScale(int prefNumDivs) {

        if (prefNumDivs < 1) {
            if (m_prefNumMajDivs < 1) {
                prefNumDivs = 1;
            } else {
                prefNumDivs = m_prefNumMajDivs;
            }
        }

        autoScale(m_viewMin, m_viewMax, prefNumDivs);
    }

//------------------------------------------------------------------------------
/**
     * Reset the bounds of the axis to evenly fit data that spans from min to max.
     * The bounds will be set to span a number of grid increments, where each
     * increment is a size that makes sense for the span of data being seen.
     * For example, a typical result would be from -25 to 100 in increments of 25.
     * @param min   The suggested lowest data value to be displayed on the axis.
     * @param max   The suggested highest data value to be displayed on the axis.
     * @param prefNumDivs  The preferred number of divisions along the axis.
     */
    public synchronized void autoScale(double min, double max, int prefNumDivs) {

        min /= m_unitScale;
        max /= m_unitScale;

        // Find the best step increment to use.
        double outerStep = (double) calcStep(m_contentMin, m_contentMax, prefNumDivs);
        if (outerStep != 0) {
            m_outerMin = (double) (Math.floor(m_contentMin / outerStep)) * outerStep;
            m_outerMax = (double) (Math.ceil(m_contentMax / outerStep)) * outerStep;
        } else {
            m_outerMin = 0;
            m_outerMax = 100;
        }

        // Check that the request is within the outer bounds.
        if (min < m_outerMin) {
            min = m_outerMin;
        }
        if (max > m_outerMax) {
            max = m_outerMax;
        }
        if ((max > min) && (max < (min + m_smallestInc))) {
            max = min + m_smallestInc;
        }
        if (min >= max) {
            min = m_outerMin;
            max = m_outerMax;
        }

        // Find the best step increment to use.
        m_majorInc = (double) calcStep(min, max, prefNumDivs);

        // Determine the format that would be best for this step.
//        double tempStep = m_majorInc;
        int formatPlaces = countPlaces(1 / m_smallestInc) - countPlaces(m_majorInc / m_smallestInc);
        m_format = "0";
        if (formatPlaces > 0) {
            m_format += ".";
            for (int i = 0; i < formatPlaces; i++) {
                m_format += "0";
            }
        }

        setViewLimits(min, max);
        m_minorInc = calcStep(m_majorInc, m_prefNumMinDivs);
        m_incOffset = 0;
        m_prefNumMajDivs = prefNumDivs;
        checkIncrementMagnitude();
    }

//------------------------------------------------------------------------------
/**
     * Find an increment (step) that makes sense for the span of data being seen.
     * The increments are 10, 25, and 50, or those number times some power of 10.
     * For example, a typical result would be from -25 to 100 in increments of 25.
     * The step may not divide evenly into the start of end values of the span.
     * @param min   The lowest data value to be displayed on the axis.
     * @param max   The highest data value to be displayed on the axis.
     * @param prefNumDivs  The preferred number of divisions along the axis.
     * @return The number of grid increments that should be on the axis.
     */
    public static double calcStep(double min, double max, int prefNumDivs) {

        // Perform range checking
        if (min < -Float.MAX_VALUE || min > Float.MAX_VALUE) {
            min = 0;
        }
        if (max < -Float.MAX_VALUE || max > Float.MAX_VALUE) {
            max = 100;
        }
        double[] numIncs = new double[m_autoSteps.length];
        double[] diffs = new double[m_autoSteps.length];

        // Find how much axis is spanned by the min-max range.
        double diff = max - min;

        // Calculate the order of magnitude to use.
        int power = 1;
        if (diff > 100) {
            // Large numbers...
            while (Math.pow(10, power) < diff) {
                power++;
            }
            power -= 3;
        } else {
            // Small numbers...
            if (diff > 0) {
                while (Math.pow(10, power) > diff) {
                    power--;
                }
            } else {
                power = 1;
            }
            power -= 2;
        }

        // Divide each step size into diff and find out which one provides
        // the number of increments closest to the preferred number.
        double diffMin = Double.POSITIVE_INFINITY;
        int lowestIndex = 0;
        for (int i = 0; i < m_autoSteps.length; i++) {
            numIncs[i] = diff / (m_autoSteps[i] * (double) Math.pow(10, power));
            diffs[i] = Math.abs(numIncs[i] - (double) prefNumDivs);
            if (diffMin > diffs[i]) {
                diffMin = diffs[i];
                lowestIndex = i;
            }
        }

        // Map the index found above to the corresponding step size.
        double step = m_autoSteps[lowestIndex] * (double) Math.pow(10, power);
        if (step < m_smallestInc) {
            step = m_smallestInc;
        }
        return step;
    }

//------------------------------------------------------------------------------
/**
     * Find an increment (step) that makes sense for the span of data being seen.
     * The increments are 10, 25, and 50, or those number times some power of 10.
     * For example, a typical result would be from -25 to 100 in increments of 25.
     * The step will divide evenly into the span.
     * @param min   The lowest data value to be displayed on the axis.
     * @param max   The highest data value to be displayed on the axis.
     * @param prefNumDivs  The preferred number of divisions along the axis.
     * @return The number of grid increments that should be on the axis.
     */
    public static double calcStep(double span, int prefNumDivs) {

        double step = calcStep(0, span, prefNumDivs);
        if (Math.abs(span % step) > m_smallestInc) {
            step = span / (double) prefNumDivs;
        }
        return step;
    }

//------------------------------------------------------------------------------
/**
     * Count the number of zero place holders to the left of the decimal in num.
     * @param num	The number to be counted.
     * @return The number of zero place holders in num.
     */
    public static int countPlaces(double num) {

        num = Math.round(num);
        long lastRounded = Math.round(num) / 10;
        num /= 10;
        int places = 0;

        while (true) {
            long rounded = Math.round(num);
            if (rounded != lastRounded || rounded == 0) {
                break;
            } else {
                num /= 10;
                lastRounded = rounded / 10;
            }
            places++;
        }

        return places;
    }

//------------------------------------------------------------------------------
/**
     * Keeps track of an object that is performing an update to this object.
     * @param updater   The object that is updating this one.  null to clear
     * @return True if no object was set as the updater prior to calling.
     */
/*
    public boolean setUpdater(Object updater) {
    if (m_updater == null) {
    m_updater = updater;
    return true;
    }
    else {
    if (updater == null)
    m_updater = null;
    return false;
    }
    }
     */
}
