/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2005-11-04
 * Date started:    2004-12-01
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.visual.*;
import java.awt.geom.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class Gl3dAxisGrid
	extends gov.usgs.gr.visual.GraphicalElement implements GlElement {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;

    /** The x-axis object, which can draw itself. */
    protected Axis m_xAxis;
    /** The y-axis object, which can draw itself. */
    protected Axis m_yAxis;
    /** The axis limits descriptor for the x-axis. */
    public Zoom m_xZoom = new Zoom();
    /** The axis limits descriptor for the y-axis. */
    protected Zoom m_yZoom = new Zoom();
    /** The color of the frame around the graph, in an RGB array. */
    protected float[] m_frameColor = {1, 1, 0};


//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public Gl3dAxisGrid(){

        setDefaults("gov.usgs.gr.visual.gl.Gl3dAxisGrid");

        m_xAxis = new GlXAxis();
        m_yAxis = new GlYAxis();
        super.add(m_xAxis);
        super.add(m_yAxis);
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

        super.setProperties(props);

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);
    }
//------------------------------------------------------------------------------
/**
 * Add an object to the curve.  Only accepts DataSeries.  Only holds one at
 * a time.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

        if (m_children.contains(gro)) return false;

        boolean bAdded = false;

        if (gro instanceof Axis) {
            gro = GraphicalElement.appropriateVersionOf(this, gro);
            int groDim = ((Axis)gro).getDimension();
            if (groDim == 0) m_xAxis = (Axis)gro;
            else if (groDim == 1) m_yAxis = (Axis)gro;

            // When adding axes, only replace existing ones.
            boolean bSet = false;
            String oClassName = gro.getClass().getName();
            ListIterator iter = m_children.listIterator();
            int i = 0;
            while (iter.hasNext()) {
                GrObject oldGro = (GrObject)iter.next();
                if (oldGro instanceof Axis) {
                    if (((Axis)oldGro).getDimension() == groDim) {
                        bSet = true;
                        iter.set(gro);

                        m_xAxis.setOpposingZoom(m_yAxis.getZoom());
                        m_yAxis.setOpposingZoom(m_xAxis.getZoom());
                        m_xAxis.update();
                        m_yAxis.update();

                        m_xZoom = (Zoom)m_xAxis.getZoom();
                        m_yZoom = (Zoom)m_yAxis.getZoom();

                        gro.addObserver(this);

                        break;
                    }
                }
            }
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        if (!m_bShow) return null;

        Point2D.Double p = null;

        if (dimension == 0) {
            p = new Point2D.Double((double)m_xZoom.getViewMin(),
                (double)m_xZoom.getViewMax());
        }
        if (dimension == 1) {
            p = new Point2D.Double((double)m_yZoom.getViewMin(),
                (double)m_yZoom.getViewMax());
        }
        if (dimension == 2) {
            p = new Point2D.Double(0, 1);
        }

        return p;
    }

//------------------------------------------------------------------------------
/**
 * Hide or show this series.
 * @param  bShow    True if this curve should be drawn.
 * @pre     none
 * @post    On the next redraw, this curve will reflect the new show value.
 */
    public void show(boolean bShow){
        m_bShow = bShow;
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {
        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 * @pre     The origin has been moved to (0, 0), scaling has occurred, and
 *          line type is solid.
 * @post    The curve has been drawn.
 */
    public void drawGL() {

        // Before drawing anything, check if the curve is supposed to be shown.
        if (!m_bShow)
            return;

        // Draw the axes and grid.
        gl.glColor3f(0, 0, 0.5f);
        gl.glBegin(gl.GL_QUADS);
        gl.glVertex3d(m_xZoom.getViewMin(), m_yZoom.getViewMin(), 0);
        gl.glVertex3d(m_xZoom.getViewMax(), m_yZoom.getViewMin(), 0);
        gl.glVertex3d(m_xZoom.getViewMax(), m_yZoom.getViewMax(), 0);
        gl.glVertex3d(m_xZoom.getViewMin(), m_yZoom.getViewMax(), 0);
        gl.glEnd();


        gl.glColor3fv(m_frameColor, 0);
        ((GlAxis)m_xAxis).setOpenGL(gld, gl, glu);
        ((GlXAxis)m_xAxis).drawGL();
        ((GlAxis)m_yAxis).setOpenGL(gld, gl, glu);
        ((GlYAxis)m_yAxis).drawGL();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return the x-axis Zoom object.
 */
    public Zoom getXZoom() {
        return m_xZoom;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return the y-axis Zoom object.
 */
    public Zoom getYZoom() {
        return m_yZoom;
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s = null;
        if (m_bShow)
            s = "GL 3D Axis Grid: " + m_title;
        else
            s = "(GL 3D Axis Grid): " + m_title;

        return s;
    }
}






