/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    1999-01-16
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import java.awt.Point;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import com.sun.opengl.util.GLUT;

//==============================================================================
/**
 * An x-y graph with ticks, labels, a grid, and multiple curves drawn inside.
 */
public class Gl3dGraph extends gov.usgs.gr.visual.gl.GlGraph {

    protected Gl3dViewParameters m_3dvp = new Gl3dViewParameters();
    protected double m_fieldOfView = 60;
    protected double m_nearFactor = 0.1;
    protected double m_farFactor = 2;
    protected double m_eyeX = 0f;
    protected double m_eyeY = 0f;
    protected double m_eyeZ = 1.5f;
    protected double m_centerX = -0.5f;
    protected double m_centerY = -0.5f;
    protected double m_centerZ = 0f;
    protected Trackball m_trackball = new Trackball();
    protected DoubleDataSeries m_matrixSeries =
        new DoubleDataSeries(new double[40], 2);
    protected double[] m_modelMatrix = new double[16];
    protected double[] m_projMatrix = new double[16];
    protected boolean m_bUpdatingMatrixSeries = false;
    protected boolean m_bUsePresetMatrix = false;
    protected Point m_mousePressed = new Point(0, 0);
    protected Point m_mouseLast = new Point(0, 0);
    protected boolean m_bDrawBoundingBox = true;
    protected float[] m_clearColor = {0f, 0f, 0f};

//------------------------------------------------------------------------------
/**
 * Construct a time series graph and have it create its own x-axis zoom.
 *
 */
    public Gl3dGraph () {

        setDefaults("gov.usgs.gr.visual.gl.Gl3dGraph");

        remove(m_xAxis);
        remove(m_yAxis);

        add(m_3dvp);
        m_matrixSeries.selectAll();
        m_matrixSeries.modifySelectedPoints(0, 0, 0);
        m_matrixSeries.modifySelectedPoints(0, 1, 0);
        m_matrixSeries.deselectAll();
        m_matrixSeries.selectPoint(19);
        m_matrixSeries.modifySelectedPoints(0, 0, 1);
        m_matrixSeries.modifySelectedPoints(0, 1, 1);
        m_matrixSeries.deselectAll();
        add(m_matrixSeries);
    }

//------------------------------------------------------------------------------
/**
 * Construct a time series graph and use the specified x-axis limits.
 * @param xZoom The x-axis limits object to use for this graph.
 */
    public Gl3dGraph (Zoom xZoom) {

        this();
    }

//------------------------------------------------------------------------------
/**
 * Create a graph cloned from an existing one.
 */
    public Gl3dGraph (Graph graph) {

        this();

        graph.copyInto(this, true);

        // Copy the other graph area's attributes.
        m_bSelected = graph.getSelected();
        m_bShow = graph.getShow();
        m_display = graph.getDisplayRect();
        m_local = graph.getLocalRect();
        m_frame = graph.getFrameRect();
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.
 */
    public boolean add(GrObject gro) {

        boolean bAdded = false;

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        // Don't allow pages to be added to graphs.
        if (gro instanceof PageOfGraphs) return false;
        if (gro == this) return false;

        // If params, should remove existing and place it
        // second.
        if (gro instanceof Gl3dViewParameters) {
            // Look for the old formatter and remove it.
            Gl3dViewParameters vp3d = (Gl3dViewParameters)gro;
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof Gl3dViewParameters) iter.remove();
            }

            // Add the new ps options at the beginning of the list
            // after the formatter.
            m_children.add(0, vp3d);

            m_3dvp = vp3d;
            update(m_3dvp, null);

            bAdded = true;
        }

        // If params, should remove existing and place it
        // second.
        if (gro instanceof DoubleDataSeries) {
            // Look for the old formatter and remove it.
            DoubleDataSeries ds = (DoubleDataSeries)gro;
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof DoubleDataSeries) iter.remove();
            }

            // Add the new obj at the beginning of the list.
            m_children.add(1, ds);

            m_matrixSeries = ds;
            update(m_matrixSeries, null);

            bAdded = true;
        }
        if (gro instanceof GraphicalElement) {
            // Add it.
            m_children.add(gro);
            bAdded = true;
        }

        if (bAdded) gro.addObserver(this);

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Wrapper for clearing the list of children.
 */
    public synchronized void clear() {

        // First deregister this object as an observer of each child.
        Iterator iter = m_children.listIterator();
        while (iter.hasNext()) {
            GrObject child = (GrObject)iter.next();
            child.deleteObserver(this);
        }

        m_children.clear();

        m_bHasChanged = true;
    }

//------------------------------------------------------------------------------
/**
 * Wrapper for removing items from the list of children.
 */
    public synchronized GrObject remove(int i) {

        // First deregister this object as an observer of the child.
        GrObject child = (GrObject)m_children.get(i);
        child.deleteObserver(this);

        m_bHasChanged = true;

        return ((GrObject)m_children.remove(i));
    }

//------------------------------------------------------------------------------
/**
 * Wrapper for removing items from the list of children.
 */
    public synchronized boolean remove(GrObject gro) {

        // First deregister this object as an observer of the child.
        boolean bFound = m_children.remove(gro);
        if (bFound) gro.deleteObserver(this);

        m_bHasChanged = true;

        return (bFound);
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("drawBoundingBox", new GrProperty(m_bDrawBoundingBox,
            "Draw Bounding Box"));

        ColorProperty cp = new ColorProperty();
        cp.description = "Clear Color";
        cp.setColor(m_clearColor);
        props.put("clearColor", cp);

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public synchronized void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (key.equals("drawBoundingBox"))
                    m_bDrawBoundingBox = (Boolean)p.value;
                 if (key.equals("clearColor")) {
                     ColorProperty cp = (ColorProperty)p;
                     m_clearColor = cp.getColorArray();
                 }
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public synchronized void setProperty(String key, String value) {

        super.setProperty(key, value);
        m_3dvp.setProperty(key, value);

        if (key.equals("fieldOfView"))
            m_fieldOfView = Double.parseDouble(value);
        if (key.equals("nearFactor"))
            m_nearFactor = Double.parseDouble(value);
        if (key.equals("farFactor"))
            m_farFactor = Double.parseDouble(value);
        if (key.equals("eyeX"))
            m_eyeX = Double.parseDouble(value);
        if (key.equals("eyeY"))
            m_eyeY = Double.parseDouble(value);
        if (key.equals("eyeZ"))
            m_eyeZ = Double.parseDouble(value);
        if (key.equals("centerX"))
            m_centerX = Double.parseDouble(value);
        if (key.equals("centerY"))
            m_centerY = Double.parseDouble(value);
        if (key.equals("centerZ"))
            m_centerZ = Double.parseDouble(value);
        if (key.equals("drawBoundingBox"))
            m_bDrawBoundingBox = Boolean.parseBoolean(value);
        if (key.equals("clearColor")) {
            m_clearColor = ColorProperty.getColorArray(value);
        }
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public double[] getProjectionMatrix() {

        return m_projMatrix;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public double[] getModelMatrix() {

        return m_modelMatrix;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setViewMatrices(double[] modelMatrix, double[] projMatrix) {

        if (modelMatrix != null && modelMatrix.length == 16) {
            for (int i = 0; i < 16; i++) {
                m_modelMatrix[i] = modelMatrix[i];
            }
        }
        if (projMatrix != null && projMatrix.length == 16) {
            for (int i = 0; i < 16; i++) {
                m_projMatrix[i] = projMatrix[i];
            }
        }

        updateMatrixSeries();

        m_bUsePresetMatrix = true;
        m_bHasChanged = true;
        m_bNeedsRedraw = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Draw this graph with all its parts on the screen.
 * <p><b> Pre Conditions </b><br>
 *      gl has been set to a valid OpenGL context.
 */
    public void drawGL() {

        // Don't bother to draw anything if the drawing area is too small.
        if (m_display.height <= 0) return;

        // Restrict the drawing area.
        gl.glEnable(gl.GL_SCISSOR_TEST);
        gl.glScissor(m_display.x, m_display.y, m_display.width,
                     m_display.height);

        // Save the previous viewport for the window.
        int[] params = new int[4];
        gl.glGetIntegerv(gl.GL_VIEWPORT, params, 0);

        // Set up the viewing transformation for this section of the graph.
        gl.glViewport(m_display.x, m_display.y, m_display.width,
                     m_display.height);

        gl.glMatrixMode(gl.GL_PROJECTION);
        gl.glPushMatrix();
        gl.glLoadIdentity();

        glu.gluPerspective(m_fieldOfView,
            (double)m_display.width / (double)m_display.height,
            m_nearFactor * Math.abs(m_eyeZ), m_farFactor * Math.abs(m_eyeZ));

        gl.glMatrixMode(gl.GL_MODELVIEW);
        gl.glPushMatrix();

        // Clear the window.
        gl.glClearColor(m_clearColor[0], m_clearColor[1], m_clearColor[2], 0);
        gl.glClear(gl.GL_COLOR_BUFFER_BIT);

        draw3d();

        gl.glPopMatrix();
        gl.glMatrixMode(gl.GL_PROJECTION);
        gl.glPopMatrix();

        // Restore the original viewport.
        gl.glViewport(params[0], params[1], params[2], params[3]);

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Draw this graph with all its parts on the screen.
 * <p><b> Pre Conditions </b><br>
 *      gl has been set to a valid OpenGL context.
 */
    public synchronized void draw3d() {

        if (glut == null)
            glut = new GLUT();

        gl.glMatrixMode(gl.GL_MODELVIEW);
        if (m_bUsePresetMatrix) {
            gl.glLoadMatrixd(m_modelMatrix, 0);
            gl.glMatrixMode(gl.GL_PROJECTION);
            gl.glLoadMatrixd(m_projMatrix, 0);
            gl.glMatrixMode(gl.GL_MODELVIEW);
            m_bUsePresetMatrix = false;
        }
        else {
            gl.glLoadIdentity();
            glu.gluLookAt(m_eyeX, m_eyeY, m_eyeZ, 0, 0, 0,
                0, 1, 0);
            gl.glMultMatrixd(m_trackball.getRotationMatrix(), 0);

            if (m_bDrawBoundingBox) {
                gl.glColor3f(1, 1, 1);
                glut.glutWireCube(1);
            }

            gl.glTranslated(m_centerX, m_centerY, m_centerZ);
            gl.glScalef(0.0001f, 0.0001f, 0.0001f);

            // Store matrices for later use
            gl.glGetDoublev(gl.GL_PROJECTION_MATRIX, m_projMatrix, 0);
            gl.glGetDoublev(gl.GL_MODELVIEW_MATRIX, m_modelMatrix, 0);
        }

        Rectangle2D.Double outerRect = new Rectangle2D.Double(0, 0, 0, 0);
        Rectangle2D.Double innerRect = new Rectangle2D.Double(0, 0, 0, 0);
        Rectangle2D.Double innerUser = new Rectangle2D.Double(0, 0, 0, 0);

        // Draw all child graphical elements other than curves, axes, and
        // markups.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof GraphicalElement && !(gro instanceof Curve)) {
                if (gro instanceof GlElement && !(gro instanceof GlAxis)) {
                    if (!(gro instanceof Markup)) {
                        GlElement glel = (GlElement)gro;
                        glel.setOpenGL(gld, gl, glu);
                        ((GraphicalElement)glel).setRects(
                            outerRect, innerRect, innerUser);

                        glel.drawGL();
                    }
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Initiate panning when the mouse is pressed.
 * @param e     The mouse event object containing the mouse position.
 */
    public synchronized void mousePressed(MouseEvent e) {

        m_mousePressed.x = e.getX() - m_display.x;
        m_mousePressed.y =
            e.getComponent().getHeight() - e.getY() - m_display.y;

        m_mouseLast.x = m_mousePressed.x;
        m_mouseLast.y = m_mousePressed.y;
    }

//------------------------------------------------------------------------------
/**
 * Initiate panning when the mouse is dragged.
 * @param e     The mouse event object containing the mouse position.
 */
    public synchronized void mouseDragged(MouseEvent e) {

        int mouseX = e.getX() - m_display.x;
        int mouseY =
            e.getComponent().getHeight() - e.getY() - m_display.y;
        boolean bButton3Down = ((e.getModifiers() & e.BUTTON3_MASK) != 0);
        boolean bShift = ((e.getModifiers() & e.SHIFT_MASK) != 0);
        boolean bAlt = ((e.getModifiers() & e.ALT_MASK) != 0);
        boolean bCtrl = ((e.getModifiers() & e.CTRL_MASK) != 0);

        double yShift = (double)(mouseY - m_mouseLast.y)
            / (double)m_display.height;

        if (bButton3Down) {
            if (bShift) {
                double yScale = (1 + yShift);
                m_eyeZ *= yScale;
            }
            else {
                double xShift = (double)(mouseX - m_mouseLast.x)
                    / (double)m_display.width;
                m_centerX += (xShift * m_eyeZ);
                if (bAlt) {
                    m_centerZ += (yShift * m_eyeZ);
                }
                else {
                    if (bCtrl) {
                        if (yShift < 0)
                            m_fieldOfView += (yShift * m_fieldOfView);
                        else
                            m_fieldOfView += (yShift * (180 - m_fieldOfView));
                    }
                    else {
                        m_centerY += (yShift * m_eyeZ);
                    }
                }
            }
        }
        else {
            if (bAlt) {
                if (yShift < 0)
                    m_nearFactor += (yShift * m_nearFactor);
                else
                    m_nearFactor += (yShift * (m_farFactor - m_nearFactor));
            }
            else {
                if (bCtrl) {
                    if (yShift < 0)
                        m_farFactor += (yShift * (m_farFactor - m_nearFactor));
                    else
                        m_farFactor += (yShift * m_farFactor);
                }
                else {
                    m_trackball.setBallSize(Math.abs(m_eyeZ));
                    m_trackball.setStepSize(
                        (double)m_mouseLast.x * m_eyeZ/(double)m_display.width,
                        (double)m_mouseLast.y * m_eyeZ/(double)m_display.height,
                        (double)mouseX * m_eyeZ / (double)m_display.width,
                        (double)mouseY * m_eyeZ / (double)m_display.height);
                    m_trackball.nextStep();
                }
            }
        }

        updateMatrixSeries();

        m_mouseLast.x = mouseX;
        m_mouseLast.y = mouseY;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public synchronized void updateMatrixSeries() {

        double[] array = new double[40];
        for (int i = 0; i < 16; i++) {
            array[i * 2] = m_modelMatrix[i];
            array[i * 2 + 1] = m_projMatrix[i];
        }
        for (int i = 0; i < 4; i++) {
            array[(i + 16) * 2] = m_trackball.m_currentQuaternion[i];
            array[(i + 16) * 2 + 1] = m_trackball.m_stepQuaternion[i];
        }

        m_bUpdatingMatrixSeries = true;
        m_matrixSeries.setArray(array);
        m_bUpdatingMatrixSeries = false;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public synchronized void update(GrObject gro, Object arg) {

        if (gro == m_3dvp) {
            m_fieldOfView = m_3dvp.m_fieldOfView;
            m_nearFactor = m_3dvp.m_nearFactor;
            m_farFactor = m_3dvp.m_farFactor;
            m_eyeX = m_3dvp.m_eyeX;
            m_eyeY = m_3dvp.m_eyeY;
            m_eyeZ = m_3dvp.m_eyeZ;
            m_centerX = m_3dvp.m_centerX;
            m_centerY = m_3dvp.m_centerY;
            m_centerZ = m_3dvp.m_centerZ;

            m_bHasChanged = true;
            m_bNeedsRedraw = true;
            notifyObservers();
        }
        if (gro == m_matrixSeries && !m_bUpdatingMatrixSeries) {
            double[] array = m_matrixSeries.getArray();
            if (array.length == 40) {
                for (int i = 0; i < 16; i++) {
                    m_modelMatrix[i] = array[i * 2];
                    m_projMatrix[i] = array[i * 2 + 1];
                }
                for (int i = 0; i < 4; i++) {
                    m_trackball.m_currentQuaternion[i] =
                        array[(i + 16) * 2];
                    m_trackball.m_stepQuaternion[i] =
                        array[(i + 16) * 2 + 1];
                }
            }

            m_bUsePresetMatrix = true;
            m_bHasChanged = true;
            m_bNeedsRedraw = true;
            notifyObservers();
        }

        super.update(gro, arg);
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "3D Graph";

        return s;
    }
}



