/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2003-05-07
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import java.awt.geom.*;
import java.nio.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class Gl3dRectangles
	extends gov.usgs.gr.visual.GraphicalElement implements GlElement {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;

    public DoubleDataSeries m_series;
    protected DoubleDataSeries m_softEdgeSeries = null;
    protected boolean m_bDrawSoftEdge = false;
    protected float[] m_color = {1, 1, 1};
    protected boolean m_bSolid = true;
    protected ColorScheme m_colorScheme = null;
    /** The spacing between points. */
    protected double m_gridWidth = 10;
    protected double m_zScale = 0;
    protected double m_xShift = 0;
    protected double m_yShift = 0;
    protected double[] m_vertexes = null;
    protected int[] m_indexes = null;
    protected int[] m_neighbors = null;
    protected double[] m_normals = null;
    protected float[] m_texCoords = null;
    protected GlTexture m_texture = null;
    protected int m_numRows = 0;
    protected int m_numColumns = 0;
    protected double m_pointReductionTolerance = 0;

    /** A buffer to hold the color array for faster drawing. */
    protected FloatBuffer m_colorBuffer = FloatDataSeries.newDirectBuffer(0);
    /** A buffer to hold the color array for faster drawing. */
    protected FloatBuffer m_texCoordBuffer = FloatDataSeries.newDirectBuffer(0);
    /** A buffer to hold the color array for faster drawing. */
    protected FloatBuffer m_vertexBuffer = FloatDataSeries.newDirectBuffer(0);
    /** A buffer to hold the color array for faster drawing. */
    protected FloatBuffer m_normalsBuffer = FloatDataSeries.newDirectBuffer(0);
    /** A buffer to hold the color array for faster drawing. */
    protected IntBuffer m_indexBuffer = IntDataSeries.newDirectBuffer(0);
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GlDiscontinuousTriFanSurface.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public Gl3dRectangles(){

    	add(getDefaultColorScheme());

        setDefaults("gov.usgs.gr.visual.gl.GlDiscontinuousTriFanSurface");
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a curve from a data series.
 */
    public Gl3dRectangles(DataSeries series){

        setDefaults("gov.usgs.gr.visual.gl.Gl3dRectangles");

        add(series);
    	add(getDefaultColorScheme());
   }

//------------------------------------------------------------------------------
/**
 * Creates a default color scheme.
 */
    public GrObject getDefaultColorScheme(){

    	GrObject cs = new FormulaColorScheme();
        cs.setProperty("redRange", "80");
        cs.setProperty("greenOffset", "60");
        cs.setProperty("greenRange", "-60");
        cs.setProperty("blueRange", "40");

        return cs;
	}


//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("xShift", new GrProperty(m_xShift, "X Shift"));
        props.put("yShift", new GrProperty(m_yShift, "Y Shift"));
        props.put("gridWidth", new GrProperty(m_gridWidth, "Grid Width"));
        props.put("zScale", new GrProperty(m_zScale, "Z Scale"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

                if (key.equals("xShift"))
                    m_xShift = (Double)p.value;
                if (key.equals("yShift"))
                    m_yShift = (Double)p.value;
            	if (key.equals("gridWidth"))
            		m_gridWidth = (Double)p.value;
            	if (key.equals("zScale"))
            		m_zScale = (Double)p.value;
                 if (key.equals("numRows"))
                     m_numRows = (Integer)p.value;
                 if (key.equals("numColumns"))
                     m_numColumns = (Integer)p.value;
            }
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("xShift"))
            m_xShift = Double.parseDouble(value);
        if (key.equals("yShift"))
            m_yShift = Double.parseDouble(value);
        if (key.equals("gridWidth"))
            m_gridWidth = Double.parseDouble(value);
        if (key.equals("zScale"))
            m_zScale = Double.parseDouble(value);
        if (key.equals("numRows"))
            m_numRows = Integer.parseInt(value);
        if (key.equals("numColumns"))
            m_numColumns = Integer.parseInt(value);
    }

//------------------------------------------------------------------------------
/**
 * Add an object to the curve.  Only accepts DataSeries.  Only holds one at
 * a time.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

    	if (gro instanceof DataSeries) {
        	if (((DataSeries)gro).getNumDimensions() == 1) {
        		clear();
        		super.add(gro);
				m_series = (DoubleDataSeries)gro;

                createVertexes();
                calculateIndexes(m_numRows - 1, m_numColumns - 1);
                if (m_pointReductionTolerance >= 0) {
                    createNeighbors(m_numRows - 1, m_numColumns - 1);
                    reducePoints();
                }

                if (m_colorScheme != null)
                	super.add((GrObject)m_colorScheme);

                m_colorBuffer = FloatDataSeries.newDirectBuffer(0);
                m_normalsBuffer = FloatDataSeries.newDirectBuffer(0);
                m_texCoordBuffer = FloatDataSeries.newDirectBuffer(0);

            	return true;
            }
        }
        else {
        	if (gro instanceof ColorScheme) {
            	if (m_colorScheme != null)
                	remove((GrObject)m_colorScheme);
                super.add(gro);
                m_colorScheme = (ColorScheme)gro;

                m_colorBuffer = FloatDataSeries.newDirectBuffer(0);

                return true;
        	}
            if (gro instanceof GlTexture) {
                if (m_texture != null)
                    remove((GrObject)m_texture);
                super.add(gro);
                m_texture = (GlTexture)gro;

                m_texCoordBuffer = FloatDataSeries.newDirectBuffer(0);
                m_colorBuffer = FloatDataSeries.newDirectBuffer(0);

                return true;
            }
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Removes all children from the graph except the axes.
 */
    public void clear() {

        super.clear();
        m_texture = null;
    }

//------------------------------------------------------------------------------
/**
 * Removes child i from the graph.  Does not remove axes.
 */
    public GrObject remove(int i) {

        if (m_children.get(i) == m_texture) {
            m_texture = null;
            m_texCoords = null;
        }

        return super.remove(i);
    }

//------------------------------------------------------------------------------
/**
 * Removes child gro from the graph.  Does not remove axes.
 */
    public boolean remove(GrObject gro) {

        if (gro == m_texture) {
            m_texture = null;
            m_texCoords = null;
        }

        return super.remove(gro);
	}

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        if (!m_bShow || m_series == null) return null;

        Point2D.Double p = null;

        if (dimension == 0) {
            p = new Point2D.Double(m_xShift,
                m_xShift + (m_numRows * m_gridWidth));
        }
        if (dimension == 1) {
            p = new Point2D.Double(m_yShift,
                m_yShift + (m_numColumns * m_gridWidth));
        }
        if (dimension == 2) {
            p = new Point2D.Double(m_series.getMin(0) * m_zScale,
                m_series.getMax(0) * m_zScale);
        }

        return p;
    }

//------------------------------------------------------------------------------
/**
 * Hide or show this series.
 * @param  bShow    True if this curve should be drawn.
 * @pre     none
 * @post    On the next redraw, this curve will reflect the new show value.
 */
    public void show(boolean bShow){
        m_bShow = bShow;
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {
        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 * @pre     The origin has been moved to (0, 0), scaling has occurred, and
 *          line type is solid.
 * @post    The curve has been drawn.
 */
    public void drawGL() {

        // Before drawing anything, check if the curve is supposed to be shown.
        if (!m_bShow)
            return;

        if (m_series == null || m_vertexes == null)
            return;

        // Save the previous line color and pattern.
        gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

        drawQuads();

        // Restore the previous line attributes.
        gl.glPopAttrib();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw quadrilaterals that skip the first
 * (center) of the five points.
 */
    public void drawQuads() {

        //		gl.polygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE);
        double z;

// lighting        if (m_normals == null)
// lighting            calculateNormals(1201, 1201, series);

        gl.glPushAttrib(gl.GL_POLYGON_BIT | gl.GL_ENABLE_BIT | gl.GL_HINT_BIT);
        gl.glPushMatrix();
        gl.glPolygonMode(gl.GL_FRONT, gl.GL_FILL);
        gl.glCullFace(gl.GL_BACK);
        gl.glEnable(gl.GL_CULL_FACE);

//lighting        turnOnLighting();
//lighting        gl.glEnable(gl.GL_COLOR_MATERIAL);
//lighting        gl.glColorMaterial(gl.GL_FRONT, gl.GL_AMBIENT_AND_DIFFUSE);

        if (m_texture == null) {
            gl.glDisableClientState(gl.GL_TEXTURE_COORD_ARRAY);
            if (m_colorBuffer.capacity() < m_vertexBuffer.capacity())
                calculateColors();
            gl.glEnableClientState(gl.GL_COLOR_ARRAY);
            m_colorBuffer.rewind();
            gl.glColorPointer(3, gl.GL_FLOAT, 0, m_colorBuffer);
        }
// *****
// *****
// -----  Convert these buffer calls to be like the color buffer.

        else {
            gl.glDisableClientState(gl.GL_COLOR_ARRAY);
            if (m_texCoords == null)
                calculateTexCoords();
            m_texture.setOpenGL(gld, gl, glu);
            if (m_texture.bindTexture()) {
                gl.glEnableClientState(gl.GL_TEXTURE_COORD_ARRAY);
                gl.glTexCoordPointer(2, gl.GL_FLOAT, 0,
                    FloatDataSeries.newDirectBuffer(m_texCoords));
            }
        }
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY);
        gl.glEnableClientState(gl.GL_INDEX_ARRAY);
//lighting        gl.glEnableClientState(gl.GL_NORMAL_ARRAY);
        gl.glVertexPointer(3, gl.GL_DOUBLE, 0, DoubleDataSeries.
            newDirectBuffer(m_vertexes));
        gl.glIndexPointer(gl.GL_INT, 0, IntDataSeries.
                        newDirectBuffer(m_indexes));
//lighting        gl.glNormalPointer(gl.GL_FLOAT, 0, m_normals);

        gl.glTranslated(m_xShift, m_yShift, 0);
        gl.glScaled(m_gridWidth, m_gridWidth, m_zScale);

        gl.glDrawElements(gl.GL_QUADS, m_indexes.length, gl.GL_UNSIGNED_INT,
            0);

//lighting        turnOffLighting();

        gl.glDisable(gl.GL_CULL_FACE);
        gl.glPopMatrix();
        gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void createVertexes() {

        synchronized (m_series) {
            double[] mainArray = m_series.getArray();
            m_vertexes = new double[m_numRows * m_numColumns * 3];

            for (int i = 0; i < m_numRows; i++) {
                for (int j = 0; j < m_numColumns; j++) {
                    int srcOffset = i * m_numRows + j;
                    int dstOffset = (i * m_numRows + j) * 3;
                    m_vertexes[dstOffset] = i;
                    m_vertexes[dstOffset + 1] = j;
                    m_vertexes[dstOffset + 2] = mainArray[srcOffset];
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void calculateColors() {


        double z;
        m_colorBuffer = FloatDataSeries.newDirectBuffer(m_vertexes.length);

        for (int i = 0; i < m_vertexes.length; i += 3) {
            z = m_vertexes[i + 2];
            float[] color = m_colorScheme.getColor(z);
            m_colorBuffer.put(i, 0);
            m_colorBuffer.put(i + 1, color[1]);
            m_colorBuffer.put(i + 2, color[2]);
        }

    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void calculateIndexes(int numI, int numJ) {

        m_indexes = new int[numI * numJ * 4];

        for (int i = 0; i < numI; i++) {
            for (int j = 0; j < numJ; j++) {
                m_indexes[(i * numJ + j) * 4] = i * (numJ + 1) + j;
                m_indexes[(i * numJ + j) * 4 + 1] = (i + 1) * (numJ + 1) + j;
                m_indexes[(i * numJ + j) * 4 + 2] =(i + 1) * (numJ + 1) + j + 1;
                m_indexes[(i * numJ + j) * 4 + 3] = i * (numJ + 1) + j + 1;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void createNeighbors(int numI, int numJ) {

        m_neighbors = new int[numI * numJ * 2];

        for (int i = 0; i < numI; i++) {
            for (int j = 0; j < numJ; j++) {
                if (i < (numI - 1))
                    m_neighbors[(i * numJ + j) * 2] = i + 1;
                else
                    m_neighbors[(i * numJ + j) * 2] = -1;
                if (j < (numJ - 1))
                    m_neighbors[(i * numJ + j) * 2 + 1] = j + 1;
                else
                    m_neighbors[(i * numJ + j) * 2 + 1] = -1;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void calculateTexCoords() {

        double[] texCorners = m_texture.getCorners();

        m_texCoords = new float[m_vertexes.length / 3 * 2];

        double xMax = m_numRows - 1;
        double yMax = m_numColumns - 1;

        int iStart = m_series.getBegin();
        int iEnd = m_series.getCount();
        for (int i = 0; i < iEnd; i++) {
//            float x = array[i * 3] * m_gridWidth + m_xShift;
//            float y = array[i * 3 + 1] * m_gridWidth + m_yShift;
//            float xPercentage = (float)((texCorners[2] - texCorners[0])
//                * (x - texCorners[0]));
//            float yPercentage = (float)((texCorners[7] - texCorners[1])
//                * (y - texCorners[1]));
//            float leftTexEdge = (float)(texCorners[6] - texCorners[0])
//                * yPercentage;
//            float bottomTexEdge = (float)(texCorners[6] - texCorners[0]);
            m_texCoords[i * 2] = (float)(m_vertexes[i * 3] * m_gridWidth
                + m_xShift - texCorners[0])
                / (float)(texCorners[2] - texCorners[0]);
            m_texCoords[i * 2 + 1] = (float)(m_vertexes[i * 3 + 1] * m_gridWidth
                + m_yShift - texCorners[1])
                / (float)(texCorners[2] - texCorners[0]);
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void calculateNormals(int numI, int numJ) {

        double z;
        m_normals = new double[m_vertexes.length];

        double u0 = 0, v0 = 0, w0 = 0, u1 = 0, v1 = 0, w1 = 0;

        for (int i = 0; i < numI; i++) {
            for (int j = 0; j < numJ; j++) {
                if (i < (numI - 1)) {
                    u0 = 1;
                    w0 = m_vertexes[((i + 1) * numJ + j) * 3 + 2] -
                        m_vertexes[(i * numJ + j) * 3 + 2];
                }
                else {
                    u0 = -1;
                    w0 = m_vertexes[((i - 1) * numJ + j) * 3 + 2] -
                        m_vertexes[(i * numJ + j) * 3 + 2];
                }
                if (j < (numJ - 1)) {
                    v1 = 1;
                    w1 = m_vertexes[(i * numJ + j + 1) * 3 + 2] -
                        m_vertexes[(i * numJ + j) * 3 + 2];
                }
                else {
                    v1 = -1;
                    w1 = m_vertexes[(i * numJ + j - 1) * 3 + 2] -
                        m_vertexes[(i * numJ + j) * 3 + 2];
                }

                // Calculate the cross product of the two vectors.
                double u_cross = (v0 * w1) - (v1 * w0);
                double v_cross = (u1 * w0) - (u0 * w1);
                double w_cross = (u0 * v1) - (u1 * v0);

                // Normalize the cross product.
                double length = (double)Math.sqrt((u_cross * u_cross)
                    + (v_cross * v_cross) + (w_cross * w_cross));

                u_cross /= length;
                v_cross /= length;
                w_cross /= length;

                m_normals[(i * numJ + j) * 3] = u_cross;
                m_normals[(i * numJ + j) * 3 + 1] = v_cross;
                m_normals[(i * numJ + j) * 3 + 2] = w_cross;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void reducePoints() {

        for (int i = 0; i < (m_neighbors.length - 1); i += 2) {
            if (m_neighbors[i] < 0 && m_neighbors[i + 1] < 0)
                continue;

            int currentI = (i / 2) / (m_numRows - 1);
            int currentJ = (i / 2) % (m_numRows - 1);
            int currentVertexOffset = (currentI * m_numRows
                            + currentJ) * 3;

            boolean bEastExhausted = false;
            boolean bNorthExhausted = false;

            while (!bEastExhausted || !bNorthExhausted) {
                if (currentJ > 1195) {
                    int top = 0;
                }
                int eastNeighbor = currentI + 1;
                if (m_neighbors[i] > 0)
                    eastNeighbor = m_neighbors[i];
                else
                    bEastExhausted = true;
                int eastVertexOffset = (eastNeighbor * m_numRows
                                + currentJ) * 3;
                int northNeighbor = currentJ + 1;
                if (m_neighbors[i + 1] > 0)
                    northNeighbor = m_neighbors[i + 1];
                else
                    bNorthExhausted = true;

                int northNeighborsNeighbor = -1;
                if (northNeighbor < (m_numRows - 1)) {
                    int northNeighborOffset = (currentI * (m_numRows - 1)
                        + northNeighbor) * 2;
                    northNeighborsNeighbor =
                        m_neighbors[northNeighborOffset + 1];
                }

                int northVertexOffset = (currentI * m_numRows
                                + northNeighbor) * 3;
                int farNorthVertexOffset = (currentI * m_numRows
                                + northNeighbor + 1) * 3;
                int eNorthVertexOffset = (eastNeighbor * m_numRows
                                + northNeighbor) * 3;
                int eFarNorthVertexOffset = (eastNeighbor * m_numRows
                                + northNeighbor + 1) * 3;

                if (!bEastExhausted) {
                    bEastExhausted = true;
                }
                if (!bNorthExhausted) {
                    try {
                    double[] lineArray1 = new double[6];
                    lineArray1[0] = m_vertexes[currentVertexOffset + 1];
                    lineArray1[1] = m_vertexes[currentVertexOffset + 2];
                    lineArray1[2] = m_vertexes[northVertexOffset + 1];
                    lineArray1[3] = m_vertexes[northVertexOffset + 2];
                    lineArray1[4] = m_vertexes[farNorthVertexOffset + 1];
                    lineArray1[5] = m_vertexes[farNorthVertexOffset + 2];
                    // if lineArray1 does not reduce, continue

                    double[] lineArray2 = new double[6];
                    lineArray1[0] = m_vertexes[eastVertexOffset + 1];
                    lineArray1[1] = m_vertexes[eastVertexOffset + 2];
                    lineArray1[2] = m_vertexes[eNorthVertexOffset + 1];
                    lineArray1[3] = m_vertexes[eNorthVertexOffset + 2];
                    lineArray1[4] = m_vertexes[eFarNorthVertexOffset + 1];
                    lineArray1[5] = m_vertexes[eFarNorthVertexOffset + 2];
                    // if lineArray2 does not reduce, continue
                    }
                    catch(Exception e) {
System.err.println("currentI = " + currentI + ", currentJ = " + currentJ
    + ", eastNeighbor = " + eastNeighbor + ", northNeighbor = " + northNeighbor);
                        e.printStackTrace();
                        System.exit(1);
                    }

                    if ((currentI + 1) < eastNeighbor) {
                        double[] lineArray3 = createLineArrayI(currentI,
                            eastNeighbor, northNeighbor);
                        // if lineArray3 does not reduce, continue

                        double[] lineArray4 = createLineArrayI(currentI,
                            eastNeighbor, northNeighbor + 1);
                        // if lineArray4 does not reduce, continue
                    }

                    m_neighbors[i + 1] = northNeighborsNeighbor;
                    // copy vertex far north to vertex north
//                    m_vertexes[northVertexOffset + 1] =
//                            m_vertexes[farNorthVertexOffset + 1];
//                    m_vertexes[northVertexOffset + 2] =
//                            m_vertexes[farNorthVertexOffset + 2];
//                    m_vertexes[eNorthVertexOffset + 1] =
//                            m_vertexes[eFarNorthVertexOffset + 1];
//                    m_vertexes[eNorthVertexOffset + 2] =
//                            m_vertexes[eFarNorthVertexOffset + 2];
                    m_indexes[i * 2 + 3] = m_indexes[(currentI
                        * (m_numRows - 1) + northNeighbor) * 4 + 3];
                    m_indexes[i * 2 + 2] = m_indexes[((eastNeighbor - 1)
                        * (m_numRows - 1) + northNeighbor) * 4 + 2];
                    // set neighbors of all cells along northern boundary to -2
                    removeNeighbors(currentI, eastNeighbor,
                                    northNeighbor, northNeighbor + 1);
                }
            }
        }

        cullRemovedPoints();
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public double[] createLineArrayI(int startI, int stopI, int j) {

        double[] array = new double[(stopI - startI) * 2];
        int index = 0;
        for (int i = startI; i < stopI; i++) {
            int offset = (i * m_numRows + j) * 3;
            array[index] = m_vertexes[offset];
            array[index + 1] = m_vertexes[offset + 2];
            index += 2;
        }

        return array;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public double[] createLineArrayJ(int startJ, int stopJ, int i) {

        double[] array = new double[(stopJ - startJ) * 2];
        int index = 0;
        for (int j = startJ; j < stopJ; j++) {
            int offset = (i * m_numRows + j) * 3;
            array[index] = m_vertexes[offset + 1];
            array[index + 1] = m_vertexes[offset + 2];
            index += 2;
        }

        return array;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void removeNeighbors(int startI, int stopI, int startJ, int stopJ) {

        for (int i = startI; i < stopI; i++) {
            for (int j = startJ; j < stopJ; j++) {
                int offset = (i * (m_numRows - 1) + j) * 2;
                try {
                m_neighbors[offset] = -2;
                m_neighbors[offset + 1] = -2;
                }
                catch(Exception e) {
                    e.printStackTrace();
                    return;
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void cullRemovedPoints() {

        int numRectsKept = 0;
        double[] newVerts = new double[m_vertexes.length];
        int currentVert = 0;
        int numOldVerts = m_vertexes.length / 3;
        for (int i = 0; i < numOldVerts; i++) {
            int currentI = i / m_numRows;
            int currentJ = i % m_numRows;

            boolean bSaveVertex = false;

            if (currentI > 0 && currentJ > 0) {
                int lowerLeftRectOffset =
                    (currentI - 1) * (m_numRows - 1) + (currentJ - 1);
                if (m_neighbors[lowerLeftRectOffset * 2] > -2) {
                    numRectsKept++;
                    if (m_indexes[lowerLeftRectOffset * 4 + 2] == i) {
                        bSaveVertex = true;
                        m_indexes[lowerLeftRectOffset * 4 + 2] = currentVert;
                    }
                }
            }

            if (currentI > 0 && currentJ < (m_numRows - 1)) {
                int upperLeftRectOffset =
                    (currentI - 1) * (m_numRows - 1) + currentJ;
                if (m_neighbors[upperLeftRectOffset * 2] > -2) {
                    if (m_indexes[upperLeftRectOffset * 4 + 1] == i) {
                        bSaveVertex = true;
                        m_indexes[upperLeftRectOffset * 4 + 1] =
                            currentVert;
                    }
                }
            }

            if (currentI < (m_numColumns - 1) && currentJ > 0) {
                int lowerRightRectOffset =
                        currentI * (m_numRows - 1) + currentJ - 1;
                if (m_neighbors[lowerRightRectOffset * 2] > -2) {
                    if (m_indexes[lowerRightRectOffset * 4 + 3] == i) {
                        bSaveVertex = true;
                        m_indexes[lowerRightRectOffset * 4 + 3] =
                            currentVert;
                    }
                }
            }

            if (currentI < (m_numColumns - 1) && currentJ < (m_numRows - 1)) {
                int upperRightRectOffset =
                        currentI * (m_numRows - 1) + currentJ;
                if (m_neighbors[upperRightRectOffset * 2] > -2) {
                    if (m_indexes[upperRightRectOffset * 4] == i) {
                        bSaveVertex = true;
                        m_indexes[upperRightRectOffset * 4] =
                            currentVert;
                    }
                }
            }

            if (bSaveVertex) {
               newVerts[currentVert * 3] = m_vertexes[i * 3];
               newVerts[currentVert * 3 + 1] = m_vertexes[i * 3 + 1];
               newVerts[currentVert * 3 + 2] = m_vertexes[i * 3 + 2];
               currentVert++;
            }
        }

        m_vertexes = new double[currentVert * 3];
        for (int i = 0; i < currentVert; i++) {
            m_vertexes[i * 3] = newVerts[i * 3];
            m_vertexes[i * 3 + 1] = newVerts[i * 3 + 1];
            m_vertexes[i * 3 + 2] = newVerts[i * 3 + 2];
        }

        int[] newIndexes = new int[numRectsKept * 4];
        int numOldRects = m_neighbors.length / 2;
        int currentKeptOffset = 0;
        for (int i = 0; i < numOldRects; i++) {
            if (m_neighbors[i * 2] > -2) {
                newIndexes[currentKeptOffset] = m_indexes[i * 4];
                newIndexes[currentKeptOffset + 1] = m_indexes[i * 4 + 1];
                newIndexes[currentKeptOffset + 2] = m_indexes[i * 4 + 2];
                newIndexes[currentKeptOffset + 3] = m_indexes[i * 4 + 3];
                currentKeptOffset += 4;
            }
            if ((currentKeptOffset / 4) >= numRectsKept) {
                int hey = 0;
//                System.err.println("currentKeptOffset = "
//                    + currentKeptOffset + ", numRectsKept = " + numRectsKept);
            }
        }

        System.out.println("Kept " + currentVert + " vertices of "
            + (m_numRows * m_numColumns));
        System.out.println("Kept " + numRectsKept + " rects of "
            + ((m_numRows - 1) * (m_numColumns - 1)));
    }

//------------------------------------------------------------------------------
/**
 * Set up lighting and turn it on.
 */
    public void turnOnLighting() {

        gl.glPushAttrib(gl.GL_POLYGON_BIT);
        gl.glPolygonMode(gl.GL_FRONT, gl.GL_FILL);
        gl.glCullFace(gl.GL_BACK);
        gl.glEnable(gl.GL_CULL_FACE);
//        gl.glEnable(gl.GL_DEPTH_TEST);

        float mat_specular[] = {0.1f, 0.1f, 0.1f, 1};
//        float shininess[] = {50};
        float light_position[] = {-1, -1, 0.2f, 0};
        float white_light[] = {0.01f, 0.01f, 0.01f, 1};
        float lmodel_ambient[] = {0.5f, 0.5f, 0.5f, 1};

        gl.glMaterialfv(gl.GL_FRONT, gl.GL_SPECULAR, mat_specular, 0);
        gl.glMaterialf(gl.GL_FRONT, gl.GL_SHININESS, 10);
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_POSITION, light_position, 0);
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_DIFFUSE, white_light, 0);
        gl.glLightfv(gl.GL_LIGHT0, gl.GL_SPECULAR, white_light, 0);
        gl.glLightModelfv(gl.GL_LIGHT_MODEL_AMBIENT, lmodel_ambient, 0);

        gl.glEnable(gl.GL_LIGHTING);
        gl.glEnable(gl.GL_LIGHT0);

//        gl.glClear(gl.GL_DEPTH_BUFFER_BIT);
    }

//------------------------------------------------------------------------------
/**
 * Turn off lighting.
 */
    public void turnOffLighting() {

        gl.glDisable(gl.GL_LIGHTING);
        gl.glDisable(gl.GL_LIGHT0);

        gl.glPopAttrib();
        gl.glDisable(gl.GL_CULL_FACE);
//        gl.glDisable(gl.GL_DEPTH_TEST);
    }

//------------------------------------------------------------------------------
/**
 * Set the range of x-values to be drawn by the time series.
 * @param startVal  The x value to start at.
 * @param endVal    The x value to end at.
 */
    public void setLimits(double startVal, double endVal) {
        m_series.setLimits(startVal, endVal);
	}

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s = null;
        if (m_bShow)
            s = "GL 3D Rectangles: " + m_title;
        else
            s = "(GL 3D Rectangles): " + m_title;

        return s;
    }
}






