/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2004-12-08
 * Date started:    1999-01-16
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import com.sun.opengl.util.*;
import gov.usgs.gr.Gr;
import gov.usgs.gr.visual.*;
import java.awt.*;
import java.text.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A generic axis with ticks and labels for an X-Y graph.
*/
public abstract class GlAxis extends gov.usgs.gr.visual.Axis
	implements GlElement {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;
    /** The OpenGL utility toolkit class with all the glut methods. */
    protected GLUT glut = null;

//------------------------------------------------------------------------------
/**
 * Empty constructor.
 */
    public GlAxis () {

        setDefaults("gov.usgs.gr.visual.gl.GlAxis");
    }

//------------------------------------------------------------------------------
/**
 * Create a new axis and associates it with a set of limits for itself and
 * its perpendicular counterpart.
 * <p><b> Pre Conditions </b><br>
 *      Axis limits exist for this axis and its perpendicular counterpart.
 */
    public GlAxis (Zoom zoom, Zoom opposingZoom) {

        setDefaults("gov.usgs.gr.visual.gl.GlAxis");

        m_zoom = zoom;
        m_opposingZoom = opposingZoom;
    }

//------------------------------------------------------------------------------
/**
 * Create a new axis by copying an existing one.
 * @param axis	The axis to be copied.
 */
    public GlAxis (Axis axis) {
    	super(axis);
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {

        if (glut == null || newgl != gl || newglu != glu)
            glut = new GLUT();

        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Draw all the elements that make up a time series graph.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context and the graph has been initialized.
 */
    public void drawGL() {
        // Get the latest changes from the axis limits objects.
        update();
        if (m_majorTickInc <= 0) return;

        // Calculate the start for the major tick marks, grid lines, and labels.
        double offset = m_zoom.getIncOffset();
        double startOffset = m_min - offset - 0.01 * m_majorTickInc;
        double startIntervals = (int)(startOffset / m_majorTickInc);
        if (startOffset > 0) startIntervals++;
        m_majorStart = (m_majorTickInc * startIntervals) + offset;

        // Calculate the length of a tick for this axis.
        double oppLen = (m_oppMax - m_oppMin);
        m_majorTickLen = 0.005 * oppLen * m_displayScale * m_tickSize;

	    // Save the calling function's GL settings.
	    gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

        // Set the line style and color.
	    gl.glColor3fv(m_tickColor, 0);

        // Draw the various parts of the  graph.
        if (!m_bQuickDraw) {
        	if (m_bShowGridLines) drawGridLines();
        	if (m_bShowMinorTicks) drawMinorTicks();
        }
        if (m_bShowMajorTicks) drawMajorTicks();

	    // Restore GL settings.
	    gl.glPopAttrib();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Draw a light line across the graph for each major tick.  Line and color
 * attributes are saved and then restored when drawing is complete.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context.
 */
    protected void drawGridLines() {

	    double x;

	    // Save the calling function's GL settings.
	    gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

	    if (true) {
			// Turn on a regular, dashed line pattern.
		    gl.glEnable(gl.GL_LINE_STIPPLE);
		    gl.glLineStipple(1, Gr.TIGHT_DOTTED_LINE_PATTERN);
		}

	    // Set the color to medium gray.
	    gl.glColor3fv(m_gridColor, 0);

	    gl.glBegin(gl.GL_LINES); {
		    // Vertical grid lines...
		    for (x = m_majorStart; x < m_max; x += m_majorTickInc) {
			    gl.glVertex2d(x, m_oppMin);
			    gl.glVertex2d(x, m_oppMax);
			}
		} gl.glEnd();

	    // Restore GL settings.
	    gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 * Draw one or more small tick marks between the large tick marks.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context, and the color and line width
 *      have been set.
 *		m_majorTickLen, and m_majorStart have been set appropriately.
 *		m_majorTickInc, and m_minorTickInc have been set appropriately.
 */
    protected void drawMinorTicks() {
	    double x, y;
        if (m_minorTickInc <= 0) return;

        int minsPerMax = (int)Math.round(m_majorTickInc / m_minorTickInc);
        double majorStart = m_majorStart - m_majorTickInc;

	    gl.glBegin(gl.GL_LINES); {
		    // Loop through and draw each tick...
            double len = m_majorTickLen * m_minorTickProportion;

		    for (x = majorStart; x < m_max; x += m_majorTickInc) {
            	for (int i = 1; i < minsPerMax; i++) {
                	double minX = x + i * m_minorTickInc;
                	if (minX > m_min && minX < m_max) {
			    		gl.glVertex2d(minX, m_oppMin);
			    		gl.glVertex2d(minX, m_oppMin + len);
			    		gl.glVertex2d(minX, m_oppMax);
			    		gl.glVertex2d(minX, m_oppMax - len);
                    }
                }
			}
		} gl.glEnd();
    }

//------------------------------------------------------------------------------
/**
 * Draw one or more large tick marks along the axis.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context, and the color and line width
 *      have been set.
 *		m_majorTickLen and m_majorStart have been set appropriately.
 */
    protected void drawMajorTicks() {

	    double x;

	    gl.glBegin(gl.GL_LINES); {

		    // Loop through and draw each tick...
            double len = m_majorTickLen;
		    for (x = m_majorStart; x < m_max; x += m_majorTickInc) {
			    gl.glVertex2d(x, m_oppMin);
			    gl.glVertex2d(x, m_oppMin + len);
			    gl.glVertex2d(x, m_oppMax);
			    gl.glVertex2d(x, m_oppMax - len);
			}
		} gl.glEnd();
    }

//------------------------------------------------------------------------------
/**
 * Draw line on both sides of the graph to represent the axis.  Saves
 * existing line attributes and restores them after drawing.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context and the color has been set.
 */
    protected void drawAxisLine() {
	    // Save the calling function's GL settings.
	    gl.glPushAttrib(gl.GL_LINE_BIT);

	    gl.glLineWidth(2.0f);

	    gl.glBegin(gl.GL_LINES); {
            gl.glVertex2d(m_min, m_oppMin);
            gl.glVertex2d(m_max, m_oppMin);
            gl.glVertex2d(m_min, m_oppMax);
            gl.glVertex2d(m_max, m_oppMax);
		} gl.glEnd();

	    // Restore GL settings.
	    gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 * Draws labels in magenta beside each major tick mark.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context.
 */
    public void drawText(double xOrig, double yOrig, double axisDisplayWidth,
    	int fontIndex) {
//    	GLSolidFont font) {

        if (!m_bShowText) return;
        if (m_majorTickInc <= 0) return;

        double x;

///        GLSolidFont solidFont = new GLSolidFont(
///            new java.awt.Font("Serif", java.awt.Font.PLAIN, 12), 0);
//        Graphics2D g = (Graphics2D)((GLComponent)gld).getGraphics();

        // Save the current position
        gl.glPushMatrix();

        yOrig -= 15;
        if (yOrig < 0) yOrig = 0;
        gl.glTranslated(xOrig, yOrig, 0);

        double displayScale = axisDisplayWidth / (m_max - m_min);

        NumberFormat form = new DecimalFormat(m_zoom.getFormatString());

        // Loop through and draw each label.
        for (x = m_majorStart; (x - m_max) <= (0.01 * m_majorTickInc);
        	x += m_majorTickInc) {

            if (Math.abs(x - 0) <= (0.01 * m_majorTickInc)) x = 0;
        	String output = form.format(x);
            gl.glRasterPos2d(((x - m_min) * displayScale)
            	- (2.5 * output.length()), 0);
//            float offset = (float)((x - m_min) * displayScale);
//            gl.translate(offset, 0, 0);
//            font.drawString(g, output);
///            font.drawString(output);
            glut.glutBitmapString(fontIndex, output);
//            gl.translate(-offset, 0, 0);
        }

        // Restore the previous position.
        gl.glPopMatrix();

        // Draw the axis title.
        if (m_bShowTitle) {
        	// Save the current position
        	gl.glPushMatrix();

        	yOrig -= 5;
        	if (yOrig < 0) yOrig = 0;
        	gl.glTranslated(xOrig, yOrig, 0);
            double yChange = 10;

            StringTokenizer st = new StringTokenizer(m_title, "" + '\n');
            while (st.hasMoreTokens()) {
        		yOrig -= yChange;
        		if (yOrig <= 0) yChange += yOrig;
        		gl.glTranslated(0, -yChange, 0);
                if (yOrig <= 0) yChange = 0;

            	gl.glRasterPos2d(((m_max - m_min) * displayScale / 2)
            		- (2.5 * m_title.length()), 0);
//                float offset = (float)(m_max - m_min);
//                gl.translate(offset, 0, 0);
//            	font.drawString(g, st.nextToken());
///            	font.drawString(st.nextToken());
                glut.glutBitmapString(fontIndex, st.nextToken());
//                gl.translate(-offset, 0, 0);
            }

        	// Restore the previous position.
        	gl.glPopMatrix();
        }
    }
}



