/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    2004-06-18
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import java.nio.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A single strip of data points that can draw itself.
 */
public class GlBars extends gov.usgs.gr.visual.Curve
    implements GlElement {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;
    /** The lower bounds of what needs to be drawn for the 0 dimension. */
    public double m_viewMin = Double.MAX_VALUE;
    /** The upper bounds of what needs to be drawn for the 0 dimension. */
    public double m_viewMax = Double.MIN_VALUE;
    /** The width of the bars. */
    protected double m_barWidth = 0;
    /** The y-value of the base of each bar. */
    protected double m_yBase = 0;
    /** A buffer to hold a copy of the series for faster drawing. */
    protected DoubleBuffer m_buffer = DoubleDataSeries.newDirectBuffer(0);
    /** A buffer to hold a copy of the selectedSeries for faster drawing. */
    protected DoubleBuffer m_selectedBuffer =
        DoubleDataSeries.newDirectBuffer(0);
    /** A buffer to hold a copy of the reducedLineSeries for faster drawing. */
    protected DoubleBuffer m_reducedLineBuffer =
        DoubleDataSeries.newDirectBuffer(0);
    /** A buffer to hold a copy of the reducedSymbolSeries for faster drawing.*/
    protected DoubleBuffer m_reducedSymbolBuffer =
        DoubleDataSeries.newDirectBuffer(0);

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public GlBars(){

        setDefaults("gov.usgs.gr.visual.gl.GlBars");
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.remove("showTitle");

        props.put("barWidth", new GrProperty(m_barWidth, "Bar Width"));
        props.put("yBase", new GrProperty(m_yBase, "Y Base"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);

            try {

            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }

                if (key.equals("barWidth"))
                    m_barWidth = (Double)p.value;
                if (key.equals("yBase"))
                    m_yBase = (Double)p.value;
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
            }
            catch (Exception e) {}
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("barWidth"))
            m_barWidth = Double.parseDouble(value);
        if (key.equals("yBase"))
            m_yBase = Double.parseDouble(value);

        m_bNeedsRedraw = true;
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {
        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 * <p><b> Pre Conditions </b><br>
 *      The origin has been moved to (0, 0), scaling has occurred, and
 *      line type is solid.  gl has been set to a valie OpenGL context.
 */
    public void drawGL(){

        // Before drawing anything, check if the curve is supposed to be shown.
        if (!m_bShow) return;

        synchronized (m_series) {
            if (m_buffer.capacity() == 0) {
                Thread t1 = new Thread(
                    new Runnable() {
                        public void run() {
                            update(m_series, null);
                        }
                    }
                    , "Gr Curve Series Update");
                t1.start();
                return;
            }

            // Draw a reduced version of the line underneath in cyan
            // with points if debug is turned on.
            if (m_bDebugPointReduction &&
                (m_reducedLineSeries != null || m_reducedSymbolSeries !=null)) {

                gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

                if (m_reducedLineSeries != null) {
                    // Draw the reduced line points in cyan.
                    gl.glColor3f(0, 1, 0.6f);
                    m_reducedLineBuffer.rewind();
                    gl.glVertexPointer(2, gl.GL_DOUBLE, 0, m_reducedLineBuffer);
                    // Draw the line.
                    gl.glDrawArrays(gl.GL_LINE_STRIP, 0,
                            (m_reducedLineSeries.getArray().length / 2));
                    // Draw the points.
                    gl.glPointSize(7);
                    gl.glDrawArrays(gl.GL_POINTS, 0,
                        (m_reducedLineSeries.getArray().length / 2));
                }

                if (m_reducedSymbolSeries != null) {
                    // Draw the reduced symbol points in yellow.
                    gl.glColor3f(0.6f, 1, 0);

                    m_reducedSymbolBuffer.rewind();
                    gl.glVertexPointer(2, gl.GL_DOUBLE, 0,
                        m_reducedSymbolBuffer);
                    // Draw the points.
                    gl.glPointSize(5);
                    gl.glDrawArrays(gl.GL_POINTS, 0,
                        (m_reducedSymbolSeries.getArray().length / 2));
                }

                // Draw the regular version as smaller magenta points.
                gl.glColor3f(1, 0, 1);
                gl.glPointSize(3);
                m_buffer.rewind();
                gl.glVertexPointer(2, gl.GL_DOUBLE, 0, m_buffer);

                gl.glDrawArrays(gl.GL_POINTS, m_series.getBegin(),
                    m_series.getEnd() + 1);

                gl.glPopAttrib();

                return;
            }

            // Save the previous line color and pattern.
            gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

            // Set the width of the line.
            gl.glLineWidth(m_lineWidth);

            // Point OpenGL to the interleaved x-y array.
            m_buffer.rewind();
            gl.glVertexPointer(2, gl.GL_DOUBLE, 0, m_buffer);

            // Set the color.
            gl.glColor3fv(m_drawColor, 0);

            // Find out if a smaller subset of the data array can be drawn.
            int startIndex = m_series.getBegin();
            int endIndex = m_series.getEnd();
            if (m_viewMin < m_viewMax && m_series.getSorted(0)) {
    //long startTime = System.currentTimeMillis();
                int temp = m_series.search(m_viewMin) - 1;
                if (temp > startIndex && temp < endIndex) startIndex = temp;
                temp = m_series.search(m_viewMax) + 1;
                if (temp < endIndex) endIndex = temp;
            }

            if (m_bShowLine) {
                // Set the line pattern.
                if (!m_bSolid && m_linePattern != Gr.SOLID_LINE_PATTERN) {
                    gl.glEnable(gl.GL_LINE_STIPPLE);
                    gl.glLineStipple(1, m_linePattern);
                }
                // Draw the line strip connecting the points.
    //          gl.drawArrays(GL.GL_LINE_STRIP, m_series.getBegin(),
    //                	m_series.getEnd() + 1);

                double[] array = m_series.getArray();
                int numDims = m_series.getNumDimensions();
                for (int i = startIndex; i < endIndex; i++) {
                    double xStart = array[i * numDims] - (m_barWidth / 2);
                    double xEnd = xStart + m_barWidth;
                    double yValue = array[i * numDims + 1];

                    if (m_bShowSymbols) {
                        gl.glRectd(xStart, m_yBase, xEnd, yValue);
                        gl.glEnd();
                    }

                    gl.glBegin(gl.GL_LINE_STRIP);
                    gl.glVertex2d(xStart, m_yBase);
                    gl.glVertex2d(xStart, yValue);
                    gl.glVertex2d(xEnd, yValue);
                    gl.glVertex2d(xEnd, m_yBase);
                    gl.glEnd();
                }
            }

            // Restore the previous line attributes.
            gl.glPopAttrib();
        }
        m_bNeedsRedraw = false;
	}

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param the reduced line data series for this curve.
 */
    public void setReducedLineSeries(DoubleDataSeries series) {

        super.setReducedLineSeries(series);

        if (m_reducedLineSeries == m_series) {
            m_reducedLineBuffer = m_buffer;
        }
        else {
            double[] reducedLineArray = m_reducedLineSeries.getArray();
            if (reducedLineArray.length > m_reducedLineBuffer.capacity()) {
                m_reducedLineBuffer = DoubleDataSeries.newDirectBuffer(
                    reducedLineArray.length * 2);
            }

            m_reducedLineBuffer.position(0);
            m_reducedLineBuffer.put(reducedLineArray);
        }
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param the reduced line data series for this curve.
 */
    public void setReducedSymbolSeries(DoubleDataSeries series) {

        super.setReducedSymbolSeries(series);

        if (m_reducedSymbolSeries == m_series) {
            m_reducedSymbolBuffer = m_buffer;
        }
        else {
            double[] reducedSymbolArray = m_reducedSymbolSeries.getArray();
            if (reducedSymbolArray.length > m_reducedSymbolBuffer.capacity()) {
                m_reducedSymbolBuffer = DoubleDataSeries.newDirectBuffer(
                    reducedSymbolArray.length * 2);
            }

            m_reducedSymbolBuffer.position(0);
            m_reducedSymbolBuffer.put(reducedSymbolArray);
        }
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        if (gro == m_series) {
            DoubleDataSeries series = (DoubleDataSeries)gro;
            double[] array = series.getArray();
            if (array.length > m_buffer.capacity()) {
                m_buffer = DoubleDataSeries.newDirectBuffer(array.length * 2);
            }

            m_buffer.position(0);
            m_buffer.put(array);

            double[] selectedArray = series.getSelectedArray();
            if (selectedArray.length > m_selectedBuffer.capacity()) {
                m_selectedBuffer = DoubleDataSeries.newDirectBuffer(
                    selectedArray.length * 2);
            }

            m_selectedBuffer.position(0);
            m_selectedBuffer.put(selectedArray);
        }

        super.update(gro, arg);
}

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

        String s = null;
        String title = null;
        if (m_series != null)
            title = m_series.getTitle();
        if (m_bShow)
            s = "GlBars: " + title;
        else
            s = "(GlBars): " + title;

        return s;
    }
}






