/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-10-19
 * Date started:    2004-02-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.*;
import java.awt.*;
import java.awt.geom.*;
import java.text.*;
import java.util.*;
import javax.media.opengl.*;
import com.sun.opengl.util.*;

//==============================================================================
/**
 * A color bar drawn as a vertical rectangle with color gradients from top to
 * bottom.  The gradients match those of the InterpolatingColorScheme child.
 * The value for each color is written next to the color.  There are properties
 * to control the position, width, and height.
 */
public class GlColorBar
	extends gov.usgs.gr.visual.TextMarkup implements GlElement {

    /** The x coordinate of this mark up. */
    public double m_width = 0;
    /** The y coordinate of this mark up. */
    public double m_height = 0;
    /** The format to use for outputting the axis labels. */
    protected String m_format = "0";
    /** The color scheme child whose values will be displayed. */
    protected InterpolatingColorScheme m_colorScheme = null;
    /** The array to hold the color values from the color scheme. */
    protected double[] m_vals = new double[0];
    /** The array to hold the RGB color arrays from the color scheme. */
    protected float[][] m_colors = new float[0][0];
    /**
     * The array to hold the formatted labels that are written next to the
     * color bar.
     */
    protected String[] m_labels = new String[0];
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/ColorBar.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank object.
 */
    public GlColorBar(){

        setDefaults("gov.usgs.gr.visual.gl.GlColorBar");
    }

//------------------------------------------------------------------------------
/**
 * Adds an object as a child, but only if it is an instance of
 * InterpolatingColorScheme.  Removes the old child, if a matching one can be
 * found, and replaces it with the new one.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        if (m_children.contains(gro)) return false;

        boolean bAdded = false;

        // If it is the right type, add it.
        if (gro instanceof InterpolatingColorScheme) {
            if (m_colorScheme != null)
                remove((GrObject)m_colorScheme);
            super.add(gro);
            m_colorScheme = (InterpolatingColorScheme)gro;
            compileColorInfo();

            return true;
        }

        return (super.add(gro));
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * width          Double   "Width"                  <p>
 * height         Double   "Height"                 <p>
 * labelFormat    String  "Label Format"           <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

        // First add properties from the super class.
        GrHashList props = super.getProperties();

        // Add the properties to the list.
        props.put("width", new GrProperty(m_width, "Width", true));
        props.put("height", new GrProperty(m_height, "Height", true));
        props.put("labelFormat", new GrProperty(m_format, "Label Format"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.gr.visual.gl.GlColorBar#getProperties getProperties()
 * for a list of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

        // Let the super class handle its properties.
        super.setProperties(props);

        GrProperty p;

        // Handle the property according to its key.
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);

            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (key.equals("width")) {
                    m_width = (Double)p.value;
                }
                if (key.equals("height")) {
                    m_height = (Double)p.value;
                }
                // Handle string properties by passing them to the
                // setProperty() method.
                if (p.value instanceof String) {
                    String value = (String)p.value;
                    setProperty(key, value);
                }
            }
            // Handle string properties by passing them to the
            // setProperty() method.
            if (obj instanceof String) {
                String value = (String)obj;

                setProperty(key, value);
            }
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.gr.visual.gl.GlColorBar#getProperties getProperties()
 * for a list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

        // Let the super class handle its properties.
        super.setProperty(key, value);

        // Handle the property according to its key.
        if (key.equals("width")) {
            m_width = Double.parseDouble(value);
        }
        if (key.equals("height")) {
            m_height = Double.parseDouble(value);
        }
        if (key.equals("labelFormat"))
            m_format = value;

        m_bNeedsRedraw = true;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 */
    public void drawGL() {

        // Before drawing anything, check if the curve is supposed to be shown.
        if (!m_bShow) return;

        // Don't proceed if there is not a color scheme child.
        int numVals = m_vals.length;
        if (m_colorScheme == null || numVals <= 0)
            return;

        boolean bDiscrete = m_colorScheme.getDiscrete();

        // Determine the total span of color values.
        double totalSpan = m_vals[numVals - 1] - m_vals[0];

        // Transform the position and size to use coordinates.
        double transX = transformX(x);
        double transY = transformY(y);
        double width = m_width * transX / x;
        double height = m_height * transY / y;

        // Save the current position and line attributes.
        gl.glPushMatrix();
        gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

        // Draw a quad strip that begins at the bottom of the bar, ends at
        // at the top, and has segments for each value in between.
        gl.glBegin(gl.GL_QUAD_STRIP);
        gl.glColor3fv(m_colors[0], 0);
        gl.glVertex2d(transX, transY);
        gl.glVertex2d(transX + width, transY);

        // Iterate over the inner values.
        double valY = 0;
        for (int i = 1; i < (numVals - 1); i++) {
            valY = height * ((m_vals[i] - m_vals[0]) / totalSpan);
            if (bDiscrete) {
                gl.glVertex2d(transX, transY + valY);
                gl.glVertex2d(transX + width, transY + valY);
            }
            gl.glColor3fv(m_colors[i], 0);
            gl.glVertex2d(transX, transY + valY);
            gl.glVertex2d(transX + width, transY + valY);
        }

        // Finish off at the top, using the last value, even if it is the
        // same as the first.
        if (bDiscrete)
            gl.glColor3fv(m_colors[m_colors.length - 2], 0);
        else
            gl.glColor3fv(m_colors[m_colors.length - 1], 0);
        gl.glVertex2d(transX, transY + height);
        gl.glVertex2d(transX + width, transY + height);
        gl.glEnd();

        // Draw a frame around the color bar using this object's color
        // property.
        gl.glLineWidth(2);
        gl.glColor3fv(m_color, 0);
        gl.glBegin(gl.GL_LINE_LOOP);
        gl.glVertex2d(transX, transY);
        gl.glVertex2d(transX + width, transY);
        gl.glVertex2d(transX + width, transY + height);
        gl.glVertex2d(transX, transY + height);
        gl.glEnd();

        // Draw tick marks.
        gl.glBegin(gl.GL_LINES);

        // Bottom tick mark.
        gl.glVertex2d(transX + width, transY);
        gl.glVertex2d(transX + (width * 1.2f), transY);

        // Iterator over the inner tick marks.
        for (int i = 1; i < (numVals - 1); i++) {
            valY = height * ((m_vals[i] - m_vals[0]) / totalSpan);
            double yCoord = transY + valY;

            gl.glVertex2d(transX + width, yCoord);
            gl.glVertex2d(transX + (width * 1.2), yCoord);
        }

        // Top tick mark.
        gl.glVertex2d(transX + width, transY + height);
        gl.glVertex2d(transX + (width * 1.2f), transY + height);
        gl.glEnd();

        // Draw text labels next to the ticks.

        // Define an offset to make the labels look more vertically centered.
        yScreenOffset = -4;

        // Format the decimal places according to the format property.
        NumberFormat form = new DecimalFormat(m_format);

        // Draw the bottom label.
        gl.glRasterPos2d(transX + (width * 1.4) + xScreenOffset,
            transY + yScreenOffset);
        glut.glutBitmapString(m_fontIndex, m_labels[0]);

        // Iterate over the inner labels.
        for (int i = 1; i < (numVals - 1); i++) {
            valY = height * ((m_vals[i] - m_vals[0]) / totalSpan);
            double yCoord = transY + valY;

            gl.glRasterPos2d(transX + (width * 1.4) + xScreenOffset,
                yCoord + yScreenOffset);
            glut.glutBitmapString(m_fontIndex, m_labels[i]);
        }

        // Draw the top label.
        gl.glRasterPos2d(transX + (width * 1.4) + xScreenOffset,
            transY + height + yScreenOffset);
        glut.glutBitmapString(m_fontIndex, m_labels[m_labels.length - 1]);

        // Draw the text property as a title above the color bar.
        gl.glRasterPos2d(transX + xScreenOffset,
            transY + (height * 1.05) - yScreenOffset);
        glut.glutBitmapString(m_fontIndex, m_text);

        // Restore the previous properties and position.
        gl.glPopAttrib();
        gl.glPopMatrix();
    }

//------------------------------------------------------------------------------
/**
 * Get color info from the color scheme child and compile it into arrays of
 * values, colors, and labels.
 */
    public void compileColorInfo() {

        // Zero out the arrays
        m_vals = new double[0];
        m_colors = new float[0][0];
        m_labels = new String[0];

        // Don't proceed if there is not a color scheme child.
        if (m_colorScheme == null) {
            return;
        }

        // Get the list of color value objects from the color scheme object.
        ArrayList colorValues = m_colorScheme.getColorValues();

        // Make sure there is at least one value.
        int numVals = colorValues.size();
        if (numVals <= 0)
            return;

        // Allocate the arrays to hold each color and value.
        m_vals = new double[numVals];
        m_colors = new float[numVals][3];
        m_labels = new String[numVals];

        // Create a decimal format for the text labels.
        NumberFormat form = new DecimalFormat(m_format);

        // Compile the data from the color-at-value list and store them in the
        // arrays for later use by the draw method.
        int index = 0;
        Iterator iter = colorValues.listIterator();
        while (iter.hasNext()) {
            InterpolatingColorScheme.ColorAtValue cav =
                (InterpolatingColorScheme.ColorAtValue)iter.next();
            m_vals[index] = cav.value;
            m_colors[index][0] = cav.red;
            m_colors[index][1] = cav.green;
            m_colors[index][2] = cav.blue;
            m_labels[index] = form.format(m_vals[index]);
            index++;
        }
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        // If the color scheme changed, recompile the arrays of colors, values,
        // and labels.
        if (gro == m_colorScheme)
            compileColorInfo();

        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s = null;
        if (m_bShow)
            s = "GL Color Bar: " + m_text;
        else
            s = "(GL Color Bar): " + m_text;

        return s;
    }
}






