/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-11-07
 * Date started:    1999-01-16
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import java.nio.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A single strip of data points that can draw itself.
 */
public class GlCurve extends gov.usgs.gr.visual.Curve implements GlElement {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;
    /** The lower bounds of what needs to be drawn for the 0 dimension. */
    public double m_viewMin = Double.MAX_VALUE;
    /** The upper bounds of what needs to be drawn for the 0 dimension. */
    public double m_viewMax = Double.MIN_VALUE;
    /**
     * An array of begin and end pairs describing where line segments should
     * begin and end.
     */
    protected int[] m_breaks = null;
    /** A buffer to hold a copy of the series for faster drawing. */
    protected DoubleBuffer m_buffer = DoubleDataSeries.newDirectBuffer(0);
    /** A buffer to hold a copy of the selectedSeries for faster drawing. */
    protected DoubleBuffer m_selectedBuffer =
        DoubleDataSeries.newDirectBuffer(0);
    /** A buffer to hold a copy of the reducedLineSeries for faster drawing. */
    protected DoubleBuffer m_reducedLineBuffer =
        DoubleDataSeries.newDirectBuffer(0);
    /** A buffer to hold a copy of the reducedSymbolSeries for faster drawing.*/
    protected DoubleBuffer m_reducedSymbolBuffer =
        DoubleDataSeries.newDirectBuffer(0);

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public GlCurve(){

        setDefaults("gov.usgs.gr.visual.gl.GlCurve");
    }

//------------------------------------------------------------------------------
/**
 * Constructor for cloning this curve from an existing one.
 */
    public GlCurve(Curve curve) {

		super(curve);
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a curve from a data series.
 * @param series    The data that this curve will use.
 */
    public GlCurve(DoubleDataSeries series) {
	
        super(series);

        setDefaults("gov.usgs.gr.visual.gl.GlCurve");
   }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {
        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 * <p><b> Pre Conditions </b><br>
 *      The origin has been moved to (0, 0), scaling has occurred, and
 *      line type is solid.  gl has been set to a valie OpenGL context.
 */
    public void drawGL(){

        // Before drawing anything, check if the curve is supposed to be shown.
        if (!m_bShow) return;

        synchronized (m_series) {
            if (m_buffer.capacity() == 0) {
                Thread t1 = new Thread(
                    new Runnable() {
                        public void run() {
                            update(m_series, null);
                        }
                    }
                    , "Gr Curve Series Update");
                t1.start();
                return;
            }

            // Draw a reduced version of the line underneath in cyan
            // with points if debug is turned on.
            if (m_bDebugPointReduction &&
                (m_reducedLineSeries != null || m_reducedSymbolSeries !=null)) {

                gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

                if (m_reducedLineSeries != null) {
                    // Draw the reduced line points in cyan.
                    gl.glColor3f(0f, 1f, 0.6f);
                    m_reducedLineBuffer.rewind();
                    gl.glVertexPointer(2, gl.GL_DOUBLE, 0, m_reducedLineBuffer);
                    // Draw the line.
                    gl.glDrawArrays(gl.GL_LINE_STRIP, 0,
                            (m_reducedLineSeries.getArray().length / 2));
                    // Draw the points.
                    gl.glPointSize(7);
                    gl.glDrawArrays(gl.GL_POINTS, 0,
                        (m_reducedLineSeries.getArray().length / 2));
                }

                if (m_reducedSymbolSeries != null) {
                    // Draw the reduced symbol points in yellow.
                    gl.glColor3f(0.6f, 1f, 0f);

                    m_reducedSymbolBuffer.rewind();
                    gl.glVertexPointer(2, gl.GL_DOUBLE, 0,
                        m_reducedSymbolBuffer);
                    // Draw the points.
                    gl.glPointSize(5);
                    gl.glDrawArrays(gl.GL_POINTS, 0,
                        (m_reducedSymbolSeries.getArray().length / 2));
                }

                // Draw the regular version as smaller magenta points.
                gl.glColor3f(1f, 0f, 1f);
                gl.glPointSize(3);
                m_buffer.rewind();
                gl.glVertexPointer(2, gl.GL_DOUBLE, 0, m_buffer);

                gl.glDrawArrays(gl.GL_POINTS, 0, m_series.getCount());

                gl.glPopAttrib();

                return;
            }

            // Save the previous line color and pattern.
            gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

            // Set the width of the line.
            gl.glLineWidth(m_lineWidth);

            // Point OpenGL to the interleaved x-y array.
            m_buffer.rewind();
            gl.glVertexPointer(2, gl.GL_DOUBLE, 0, m_buffer);

            // Set the color.
            gl.glColor3fv(m_drawColor, 0);

            // Find out if a smaller subset of the data array can be drawn.
            int startIndex = 0;
            int endIndex = m_series.getCount() - 1;
            if (m_viewMin < m_viewMax && m_series.getSorted(0)) {
    //long startTime = System.currentTimeMillis();
                int temp = m_series.search(m_viewMin) - 1;
                if (temp > startIndex && temp < endIndex) startIndex = temp;
                temp = m_series.search(m_viewMax) + 1;
                if (temp < endIndex) endIndex = temp;
            }

            if (m_bShowLine) {
                // Set the line pattern.
                if (!m_bSolid && m_linePattern != Gr.SOLID_LINE_PATTERN) {
                    gl.glEnable(gl.GL_LINE_STIPPLE);
                    gl.glLineStipple(1, m_linePattern);
                }
                // Draw the line strip connecting the points.
    //          gl.drawArrays(GL.GL_LINE_STRIP, m_series.getBegin(),
    //                	m_series.getEnd() + 1);

                // If there are no breaks, just draw the active section of the
                // data series.
                if (m_breaks == null) {
                    gl.glDrawArrays(gl.GL_LINE_STRIP, startIndex,
                        endIndex - startIndex + 1);
                }
                else {
                    // The breaks array has been set and holds pairs of begin
                    // and end indexes for the line segments.  Draw each segment
                    // separately.
                    int numPoints = m_series.getCount();
                    for (int i = 0; i < (m_breaks.length - 1); i += 2) {
                        if (m_breaks[i] >= 0 && m_breaks[i] < numPoints &&
                            m_breaks[i + 1] >= 0 && m_breaks[i + 1] <
                            numPoints) {

                            gl.glDrawArrays(gl.GL_LINE_STRIP, m_breaks[i],
                                m_breaks[i + 1] - m_breaks[i] + 1);
                        }
                    }
                }
            }

            if (m_bShowSymbols && !m_bShowLine && m_symbolIndex == 0
                && !m_bEditing) {

                gl.glPointSize(1);
            }
            else {
                if (m_bEditing)
                    gl.glPointSize(m_editablePointSize);
                else
                    gl.glPointSize(m_symbolSize);
            }

            if (m_series.getNumSelected() > 0 || m_bEditing || m_bShowSymbols) {
                // Draw all points underneath in the same color as the curve.
                gl.glColor3fv(m_drawColor, 0);
    //            	gl.drawArrays(GL.GL_POINTS, m_series.getBegin(),
    //                	m_series.getEnd() + 1);
                gl.glDrawArrays(gl.GL_POINTS, startIndex,
                    endIndex - startIndex + 1);
            }

            if (m_series.getNumSelected() > 0 || m_bEditing) {
                // Bleach the opposingColor.
                float[] opposingColor = new float[3];
                    opposingColor[0] = 1f - m_drawColor[0];
                    opposingColor[1] = 1f - m_drawColor[1];
                    opposingColor[2] = 0.5f;
                if (m_bleach > 0) {
                    opposingColor[0] = (1f - opposingColor[0]) * m_bleach
                        + opposingColor[0];
                    opposingColor[1] = (1f - opposingColor[1]) * m_bleach
                        + opposingColor[1];
                }
                if (m_bleach < 0) {
                    opposingColor[0] = opposingColor[0] * (m_bleach + 1f);
                    opposingColor[1] = opposingColor[1] * (m_bleach + 1f);
                }

                // Draw the selected points.
                gl.glColor3fv(opposingColor, 0);
                m_selectedBuffer.rewind();
                gl.glVertexPointer(2, gl.GL_DOUBLE, 0, m_selectedBuffer);
                gl.glDrawArrays(gl.GL_POINTS, 0, m_series.getNumSelected());
            }

            //drawPerpTest();
            //drawParabTest();

            // Restore the previous line attributes.
            gl.glPopAttrib();
        }
        m_bNeedsRedraw = false;
	}

//------------------------------------------------------------------------------
/**
 *
 */
    public void drawPerpTest() {

    ///// Perpendicular test
                double[] array = m_series.getArray();

                if (array.length >= 6) {
                    double x_0 = array[0];
                    double y_0 = array[1];
                    double x_1 = array[2];
                    double y_1 = array[3];
                    double x_n = array[4];
                    double y_n = array[5];
                    double a_0 = x_1 - x_0;
                    double b_0 = y_1 - y_0;
                    double x_z = 0, y_z = 0;

                    if (a_0 != 0 && b_0 != 0) {
                        double m = b_0 / a_0;
    ///                    float a_n = (a_0 * (b_0 * (x_n - x_0) +  *
    ///                        (y_0 - y_n)))
    ///                    	/ (a_0 * a_0 + b_0 * b_0);
    ///                    float b_n = -a_n * b_0 / a_0;
                        double a_n = (m * (x_n - x_0) + (y_0 - y_n)) /
                        (1 + m * m);
                        double b_n = -m * a_n;
                        x_z = x_n + b_n;
                        y_z = y_n + a_n;
    //System.out.println("p0 = (" + x_0 + ", " + y_0 + "), p1 = (" + x_1 + ",
    // + y_1
    //	+ "), p_n = (" + x_n + ", " + y_n + "), p_z = (" + x_z + ", " + y_z +
    // ")");
    //System.out.println("a_0 = " + a_0 + ", b_0 = " + b_0
    //	+ "a_n = " + a_n + ", b_n = " + b_n);
                    }
                    else {
                        if (a_0 == 0) {
                            // p0 and p1 form a vertical line.
                            x_z = x_0;
                            y_z = y_n;
                        }
                        else {
                            // p0 and p1 form a horizontal line.
                            x_z = x_n;
                            y_z = y_0;
                        }
                    }

                    gl.glBegin(gl.GL_LINES);
                    gl.glColor3f(0f, 0f, 0.8f);
                    gl.glVertex2d(x_0 - 10 * a_0, y_0 - 10 * b_0);
                    gl.glVertex2d(x_1 + 10 * a_0, y_1 + 10 * b_0);
                    gl.glColor3f(0.8f, 0f, 0f);
                    gl.glVertex2d(x_n, y_n);
                    gl.glVertex2d(x_z, y_z);
                    gl.glEnd();

                }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void drawParabTest() {
        ///// Parab test

                    double[] array = m_series.getArray();
                    if (array.length < 8) {
                        gl.glPopAttrib();
                        return;
                    }

                    double[] coeffs1 = m_series.fitParabola(array[0], array[1],
                        array[2], array[3], array[4], array[5]);
                    double[] coeffs2 = m_series.fitParabola(array[2], array[3],
                        array[4], array[5], array[6], array[7]);
        //            System.out.println("a = " + coeffs[0] + ", b = " + coeffs[1]
        //            	 + ", c = " + coeffs[2]);
        //System.out.println("(" + array[6] + ", " + array[7] + ")");
                    gl.glLineWidth(2);
                    gl.glColor3f(0f, 0f, 0.7f);
                    gl.glBegin(gl.GL_LINE_STRIP);
                    for (double i = 0; i < 16; i += 0.05) {
                        gl.glVertex2d(i,
                            coeffs1[0] * i * i + coeffs1[1] * i + coeffs1[2]);
                    }
                    gl.glEnd();

                    gl.glColor3f(0f, 0.7f, 0f);
                    gl.glBegin(gl.GL_LINE_STRIP);
                    for (double i = 0; i < 16; i += 0.05) {
                        gl.glVertex2d(i,
                            coeffs2[0] * i * i + coeffs2[1] * i + coeffs2[2]);
                    }
                    gl.glEnd();

                    gl.glColor3f(0.7f, 0f, 0f);
                    gl.glEnable(gl.GL_LINE_STIPPLE);
                    gl.glLineStipple(1, Gr.DASHED_LINE_PATTERN);

                    // Old averaging method.
        //            gl.begin(gl.GL_LINE_STRIP);
        //            for (float i = array[2]; i <= array[4];
        //                i += ((array[4] - array[2]) / 32)) {

        //            	gl.vertex(i, ((coeffs1[0] * i * i + coeffs1[1] * i
        //                    + coeffs1[2])
        //                    + (coeffs2[0] * i * i + coeffs2[1] * i + coeffs2[2]))
        //                    / 2);
        //            }
        //            gl.end();

                    // Average with weighting toward nearest one.
        //            float leftAvg = (array[0] + array[2]) / 2;
        //            float rightAvg = (array[4] + array[6]) / 2;
                    double rlDiff = array[4] - array[2];
                    gl.glBegin(gl.GL_LINE_STRIP);
                    for (double i = array[2]; i <= array[4];
                        i += ((array[4] - array[2]) / 32)) {

                        double leftY =
                            coeffs1[0] * i * i + coeffs1[1] * i + coeffs1[2];
                        leftY *= (rlDiff - (i - array[2]));
                        leftY /= rlDiff;

                        double rightY =
                            coeffs2[0] * i * i + coeffs2[1] * i + coeffs2[2];
                        rightY *= (i - array[2]);
                        rightY /= rlDiff;
                        gl.glVertex2d(i, leftY + rightY);
                    }
                    gl.glEnd();
    }

//------------------------------------------------------------------------------
/**
 * Sets array describing the beginning and ending points of the line segments.
 * @param m_breaks  The array of begin and end pairs describing line segments.
 */
    public void setSegmentIndexes(int[] breaks) {

        m_breaks = breaks;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param the reduced line data series for this curve.
 */
    public void setReducedLineSeries(DoubleDataSeries series) {

        super.setReducedLineSeries(series);

        if (series == null)
            return;

        if (m_reducedLineSeries == m_series) {
            m_reducedLineBuffer = m_buffer;
        }
        else {
            double[] reducedLineArray = m_reducedLineSeries.getArray();
            if (reducedLineArray.length > m_reducedLineBuffer.capacity()) {
                m_reducedLineBuffer = DoubleDataSeries.newDirectBuffer(
                    reducedLineArray.length * 2);
            }

            m_reducedLineBuffer.position(0);
            m_reducedLineBuffer.put(reducedLineArray);
        }
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @param the reduced line data series for this curve.
 */
    public void setReducedSymbolSeries(DoubleDataSeries series) {

        super.setReducedSymbolSeries(series);

        if (series == null)
            return;

        if (m_reducedSymbolSeries == m_series) {
            m_reducedSymbolBuffer = m_buffer;
        }
        else {
            double[] reducedSymbolArray = m_reducedSymbolSeries.getArray();
            if (reducedSymbolArray.length > m_reducedSymbolBuffer.capacity()) {
                m_reducedSymbolBuffer = DoubleDataSeries.newDirectBuffer(
                    reducedSymbolArray.length * 2);
            }

            m_reducedSymbolBuffer.position(0);
            m_reducedSymbolBuffer.put(reducedSymbolArray);
        }
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        if (gro == m_series) {
            DoubleDataSeries series = (DoubleDataSeries)gro;
            double[] array = series.getArray();
            if (array.length > m_buffer.capacity()) {
                m_buffer = DoubleDataSeries.newDirectBuffer(array.length * 2);
            }

            m_buffer.position(0);
            m_buffer.put(array);

            double[] selectedArray = series.getSelectedArray();
            if (selectedArray != null &&
                selectedArray.length > m_selectedBuffer.capacity()) {

                m_selectedBuffer = DoubleDataSeries.newDirectBuffer(
                    selectedArray.length * 2);
            }

            m_selectedBuffer.position(0);
            if (selectedArray != null)
                m_selectedBuffer.put(selectedArray);
        }

        super.update(gro, arg);
    }
}






