/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-05
 * Date started:    2000-08-21
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import java.awt.geom.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class GlDiscontinuousTriFanSurface
	extends gov.usgs.gr.visual.GraphicalElement implements GlElement {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;

    public DoubleDataSeries m_series;
    protected DoubleDataSeries m_softEdgeSeries = null;
    protected boolean m_bDrawSoftEdge = false;
    protected float[] m_color = {1, 1, 1};
    protected boolean m_bSolid = true;
    protected ColorScheme m_colorScheme = null;
    /** The spacing between points. */
    protected double m_gridWidth = 10;
    protected double m_zScale = 0;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GlDiscontinuousTriFanSurface.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public GlDiscontinuousTriFanSurface(){

    	add(getDefaultColorScheme());

        setDefaults("gov.usgs.gr.visual.gl.GlDiscontinuousTriFanSurface");
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a curve from a data series.
 */
    public GlDiscontinuousTriFanSurface(DoubleDataSeries series){

        setDefaults("gov.usgs.gr.visual.gl.GlDiscontinuousTriFanSurface");

        add(series);
    	add(getDefaultColorScheme());
   }

//------------------------------------------------------------------------------
/**
 * Creates a default color scheme.
 */
    public GrObject getDefaultColorScheme(){

    	GrObject cs = new FormulaColorScheme();
        cs.setProperty("redRange", "80");
        cs.setProperty("greenOffset", "60");
        cs.setProperty("greenRange", "-60");
        cs.setProperty("blueRange", "40");

        return cs;
	}


//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("gridWidth", new GrProperty(m_gridWidth, "Grid Width"));
        props.put("zScale", new GrProperty(m_zScale, "Z Scale"));
        props.put("drawSoftEdge", new GrProperty(m_bDrawSoftEdge,
            "Draw Soft Edge"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("gridWidth"))
            		m_gridWidth = (Double)p.value;
            	if (key.equals("zScale"))
            		m_zScale = (Double)p.value;
            	if (key.equals("drawSoftEdge"))
            		m_bDrawSoftEdge = (Boolean)p.value;
            }
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("gridWidth"))
            m_gridWidth = Double.parseDouble(value);
        if (key.equals("zScale"))
            m_zScale = Double.parseDouble(value);
        if (key.equals("drawSoftEdge"))
            m_bDrawSoftEdge = Boolean.parseBoolean(value);
    }

//------------------------------------------------------------------------------
/**
 * Add an object to the curve.  Only accepts DoubleDataSeries.  Only holds one at
 * a time.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

    	if (gro instanceof DoubleDataSeries) {
        	if (((DoubleDataSeries)gro).getNumDimensions() == 3) {
        		clear();
        		super.add(gro);
				m_series = (DoubleDataSeries)gro;

                if (m_colorScheme != null)
                	super.add((GrObject)m_colorScheme);

            	return true;
            }
        	if (((DoubleDataSeries)gro).getNumDimensions() == 7) {

            	DoubleDataSeries drawableSeries =
                    ds7toDs3((DoubleDataSeries)gro, m_gridWidth);

        		clear();
                add(drawableSeries);
        		super.add(gro);

                if (m_colorScheme != null)
                	super.add((GrObject)m_colorScheme);

            	return true;
            }
        }
        else {
        	if (gro instanceof ColorScheme) {
            	if (m_colorScheme != null)
                	remove((GrObject)m_colorScheme);
                super.add(gro);
                m_colorScheme = (ColorScheme)gro;

                return true;
        	}
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        if (!m_bShow || m_series == null) return null;

        Point2D.Double p = new Point2D.Double(m_series.getMin(dimension),
            m_series.getMax(dimension));

        return p;
    }

//------------------------------------------------------------------------------
/**
 * Hide or show this series.
 * @param  bShow    True if this curve should be drawn.
 * @pre     none
 * @post    On the next redraw, this curve will reflect the new show value.
 */
    public void show(boolean bShow){
        m_bShow = bShow;
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {
        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 * @pre     The origin has been moved to (0, 0), scaling has occurred, and
 *          line type is solid.
 * @post    The curve has been drawn.
 */
    public void drawGL() {

        // Before drawing anything, check if the curve is supposed to be shown.
        if (!m_bShow) return;

        synchronized (m_series) {
            // Save the previous line color and pattern.
            gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

//            drawFans(m_series);
            drawQuads(m_series);
            if (m_bDrawSoftEdge) {
                if (m_softEdgeSeries == null) calculateSoftEdge();
//                drawFans(m_softEdgeSeries);
                drawQuads(m_softEdgeSeries);
            }

            // Restore the previous line attributes.
            gl.glPopAttrib();
        }

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw triangle fans with five vertices.
 */
    public void drawFans(DoubleDataSeries series) {

        //		gl.polygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE);
        int numDims = series.getNumDimensions();
        double[] array = series.getArray();
        double z;

        gl.glPushAttrib(gl.GL_POLYGON_BIT);
        gl.glPolygonMode(gl.GL_FRONT, gl.GL_FILL);
        gl.glCullFace(gl.GL_BACK);
        gl.glEnable(gl.GL_CULL_FACE);

        for (int i = series.getBegin() * numDims;
            i < series.getCount() * numDims; i += (5 * numDims)) {

            gl.glBegin(gl.GL_TRIANGLE_FAN);
//            	gl.begin(GL.GL_LINE_STRIP);
            for (int j = 0; j < (5 * numDims); j += numDims) {
                z = array[i + j + 2];
                gl.glColor3fv(m_colorScheme.getColor(z), 0);
                gl.glVertex3d(array[i + j],
                    array[i + j + 1], z * m_zScale);
            }

            z = array[i + numDims + 2];
            gl.glColor3fv(m_colorScheme.getColor(z), 0);
            gl.glVertex3d(array[i + numDims],
                array[i + numDims + 1], z * m_zScale);
            gl.glEnd();
        }

        gl.glDisable(gl.GL_CULL_FACE);
        gl.glPopAttrib();

    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw quadrilaterals that skip the first
 * (center) of the five points.
 */
    public void drawQuads(DoubleDataSeries series) {

        //		gl.polygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE);
        int numDims = series.getNumDimensions();
        double[] array = series.getArray();
        double z;

        gl.glPushAttrib(gl.GL_POLYGON_BIT);
        gl.glPolygonMode(gl.GL_FRONT, gl.GL_FILL);
        gl.glCullFace(gl.GL_BACK);
        gl.glEnable(gl.GL_CULL_FACE);

        gl.glBegin(gl.GL_QUADS);
        int iStart = series.getBegin() * numDims;
        int iEnd = series.getCount() * numDims;
        for (int i = iStart; i < iEnd; i += (5 * numDims)) {
            for (int j = numDims; j < (5 * numDims); j += numDims) {
                z = array[i + j + 2];
                gl.glColor3fv(m_colorScheme.getColor(z), 0);
                gl.glVertex3d(array[i + j],
                    array[i + j + 1], z * m_zScale);
            }
        }

        gl.glEnd();

        gl.glDisable(gl.GL_CULL_FACE);
        gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public static DoubleDataSeries ds7toDs3(DoubleDataSeries ds7,
        double gridWidth) {

        double[] array7 = ds7.getArray();
        int numPoints = ds7.getCount();
        double[] array3 = new double[numPoints * 15];

        for (int i = 0; i < numPoints; i++) {
        	int offset = i * 15;
        	double xi = array7[i * 7];
            xi -= 1;
        	array3[offset] = xi * gridWidth;
        	array3[offset + 3] = (xi - 0.5) * gridWidth;
        	array3[offset + 6] = (xi + 0.5) * gridWidth;
        	array3[offset + 9] = (xi + 0.5) * gridWidth;
        	array3[offset + 12] = (xi - 0.5) * gridWidth;

        	double xj = array7[i * 7 + 1];
            xj -= 1;
        	array3[offset + 1] = xj * gridWidth;
        	array3[offset + 4] = (xj - 0.5) * gridWidth;
        	array3[offset + 7] = (xj - 0.5) * gridWidth;
        	array3[offset + 10] = (xj + 0.5) * gridWidth;
        	array3[offset + 13] = (xj + 0.5) * gridWidth;

        	double z_sw = array7[i * 7 + 3];
            double z_se = array7[i * 7 + 4];
            double z_ne = array7[i * 7 + 5];
            double z_nw = array7[i * 7 + 6];
        	array3[offset + 5] = z_sw;
        	array3[offset + 8] = z_se;
        	array3[offset + 11] = z_ne;
        	array3[offset + 14] = z_nw;
        	array3[offset + 2] = (z_sw + z_se + z_ne + z_nw) / 4;
        }

    	DoubleDataSeries ds3 = new DoubleDataSeries(3);
        ds3.setArray(array3);

        return ds3;
	}

//------------------------------------------------------------------------------
/**
 *
 */
    public void calculateSoftEdge() {

        // Find the number of wet points in the surface layer.
        double[] mainArray = m_series.getArray();
        int numWetPoints = mainArray.length / 15;
        int numI = (int)Math.round(m_series.getMax(0) / m_gridWidth) + 1;
        int numJ = (int)Math.round(m_series.getMax(1) / m_gridWidth) + 1;

        // Create an array that will hold indexes into the I-J-K index array.
        // The ranges are 2 to maxI, 2 to maxJ, and 1 to numLayers.
        int[] lookups = new int[numI * numJ];
        for (int i = 0; i < lookups.length; i++) {
            lookups[i] = -1;
        }

        for (int i = 0; i < numWetPoints; i++) {
            int indexOffset = i * 15;
            int iIndex = (int)Math.round(mainArray[indexOffset] / m_gridWidth);
            int jIndex = (int)Math.round(mainArray[indexOffset + 1]
                / m_gridWidth);
            lookups[iIndex + jIndex * numI] = i;
        }

        // For every dry point, append a trifan to the soft edge list.
        ArrayList softEdgeList = new ArrayList();
        for (int i = 0; i < numI; i++) {
            for (int j = 0; j < numJ; j++) {
                if (lookups[i + j * numI] == -1) {
                    // Determine the lookups for the neighboring cells.
                    int[] neighborLookups = {-1, -1, -1, -1, -1, -1, -1, -1};
                    if ((i - 1) >= 0 && (j - 1) >= 0)
                        neighborLookups[0] = lookups[(i - 1) + (j - 1) * numI];
                    if ((j - 1) >= 0)
                        neighborLookups[1] = lookups[i + (j - 1) * numI];
                    if ((i + 1) < numI && (j - 1) >= 0)
                        neighborLookups[2] = lookups[(i + 1) + (j - 1) * numI];
                    if ((i + 1) < numI)
                        neighborLookups[3] = lookups[(i + 1) + j * numI];
                    if ((i + 1) < numI && (j + 1) < numJ)
                        neighborLookups[4] = lookups[(i + 1) + (j + 1) * numI];
                    if ((j + 1) < numJ)
                        neighborLookups[5] = lookups[i + (j + 1) * numI];
                    if ((i - 1) >= 0 && (j + 1) < numJ)
                        neighborLookups[6] = lookups[(i - 1) + (j + 1) * numI];
                    if ((i - 1) >= 0)
                        neighborLookups[7] = lookups[(i - 1) + j * numI];

                    boolean bEdge = false;
                    for (int n = 0; n < 8; n++) {
                        if (neighborLookups[n] != -1) bEdge = true;
                    }

                    // Make the values of the trifan the average of the
                    // surrounding points.
                    double[] triFan = new double[15];

                    int sum1 = 0; triFan[5] = 0;
                    int sum2 = 0; triFan[8] = 0;
                    int sum3 = 0; triFan[11] = 0;
                    int sum4 = 0; triFan[14] = 0;

                    if (neighborLookups[0] >= 0) {
                        sum1 += 1;
                        triFan[5] += mainArray[neighborLookups[0] * 15 + 11];
                    }
                    if (neighborLookups[1] >= 0) {
                        sum1 += 1;
                        triFan[5] += mainArray[neighborLookups[1] * 15 + 14];
                        sum2 += 1;
                        triFan[8] += mainArray[neighborLookups[1] * 15 + 11];
                    }
                    if (neighborLookups[2] >= 0) {
                        sum2 += 1;
                        triFan[8] += mainArray[neighborLookups[2] * 15 + 14];
                    }
                    if (neighborLookups[3] >= 0) {
                        sum2 += 1;
                        triFan[8] += mainArray[neighborLookups[3] * 15 + 5];
                        sum3 += 1;
                        triFan[11] += mainArray[neighborLookups[3] * 15 + 14];
                    }
                    if (neighborLookups[4] >= 0) {
                        sum3 += 1;
                        triFan[11] += mainArray[neighborLookups[4] * 15 + 5];
                    }
                    if (neighborLookups[5] >= 0) {
                        sum3 += 1;
                        triFan[11] += mainArray[neighborLookups[5] * 15 + 8];
                        sum4 += 1;
                        triFan[14] += mainArray[neighborLookups[5] * 15 + 5];
                    }
                    if (neighborLookups[6] >= 0) {
                        sum4 += 1;
                        triFan[14] += mainArray[neighborLookups[6] * 15 + 8];
                    }
                    if (neighborLookups[7] >= 0) {
                        sum1 += 1;
                        triFan[5] += mainArray[neighborLookups[7] * 15 + 8];
                        sum4 += 1;
                        triFan[14] += mainArray[neighborLookups[7] * 15 + 11];
                    }

                    if (sum1 == 0)
                        triFan[5] = 0;
                    else
                        triFan[5] /= (double)sum1;
                    if (sum2 == 0)
                        triFan[8] = 0;
                    else
                        triFan[8] /= (double)sum2;
                    if (sum3 == 0)
                        triFan[11] = 0;
                    else
                        triFan[11] /= (double)sum3;
                    if (sum4 == 0)
                        triFan[14] = 0;
                    else
                        triFan[14] /= (double)sum4;

                    triFan[0] = i * m_gridWidth;
                    triFan[1] = j * m_gridWidth;
                    triFan[2] =
                        (triFan[5] + triFan[8] + triFan[11] + triFan[14] + 0)
                        / 4;
                    triFan[3] = ((double)i - 0.5) * m_gridWidth;
                    triFan[4] = ((double)j - 0.5) * m_gridWidth;
                    //triFan[5] = 100;
                    triFan[6] = ((double)i + 0.5) * m_gridWidth;
                    triFan[7] = ((double)j - 0.5) * m_gridWidth;
                    //triFan[8] = 100;
                    triFan[9] = ((double)i + 0.5) * m_gridWidth;
                    triFan[10] = ((double)j + 0.5) * m_gridWidth;
                    //triFan[11] = 100;
                    triFan[12] = ((double)i - 0.5) * m_gridWidth;
                    triFan[13] = ((double)j + 0.5) * m_gridWidth;
                    //triFan[14] = 100;

                    if (bEdge)
                        softEdgeList.add(triFan);
                }
            }
        }

        double[] softEdgeArray = new double[softEdgeList.size() * 15];
        int index = 0;
        Iterator iter = softEdgeList.iterator();
        while (iter.hasNext()) {
            double[] triFan = (double[])iter.next();
            for (int i = 0; i < 15; i++) {
                softEdgeArray[index + i] = triFan[i];
            }
            index += 15;
        }

        m_softEdgeSeries = new DoubleDataSeries(softEdgeArray, 3);
    }

//------------------------------------------------------------------------------
/**
 * Set the range of x-values to be drawn by the time series.
 * @param startVal  The x value to start at.
 * @param endVal    The x value to end at.
 * @pre     none
 * @post    The beginning and ending indices have been adjusted and will be
 *          reflected in the next draw.
 */
    public void setLimits(double startVal, double endVal) {
        m_series.setLimits(startVal, endVal);
	}

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s = null;
        if (m_bShow)
            s = "GL TriFan Surface: " + m_title;
        else
            s = "(GL TriFan Surface): " + m_title;

        return s;
    }
}






