/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    1999-01-16
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import com.sun.opengl.util.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import java.awt.geom.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * An x-y graph with ticks, labels, a grid, and multiple curves drawn inside.
 */
public class GlGraph extends gov.usgs.gr.visual.Graph implements GlElement {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /** The main OpenGL class with all the gl drawing methods. */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;
    /** The OpenGL utility toolkit class with all the glut methods. */
    protected GLUT glut = null;

//------------------------------------------------------------------------------
/**
 * Construct a time series graph and have it create its own x-axis zoom.
 *
 */
    public GlGraph () {

        super();

        setDefaults("gov.usgs.gr.visual.gl.GlGraph");
    }

//------------------------------------------------------------------------------
/**
 * Construct a time series graph and use the specified x-axis limits.
 * @param xZoom The x-axis limits object to use for this graph.
 */
    public GlGraph (Zoom xZoom) {

        super(xZoom);

        setDefaults("gov.usgs.gr.visual.gl.GlGraph");
    }

//------------------------------------------------------------------------------
/**
 * Create a graph cloned from an existing one.
 */
    public GlGraph (Graph graph) {
    	super(graph);
    }

//------------------------------------------------------------------------------
/**
 * Returns a new x axis object.
 */
    protected Axis newXAxis() {
    	return new GlXAxis();
    }

//------------------------------------------------------------------------------
/**
 * Returns a new y axis object.
 */
    protected Axis newYAxis() {
    	return new GlYAxis();
    }

//------------------------------------------------------------------------------
/**
 * Returns a new Curve object.
 */
    protected Curve newCurve() {
    	return new GlCurve();
    }

//------------------------------------------------------------------------------
/**
 * Returns a new Curve object based on the data series parameter.
 */
    protected Curve newCurve(DoubleDataSeries series) {
    	return new GlCurve(series);
    }

//------------------------------------------------------------------------------
/**
 * Returns a new Curve object copied from an existing curve.
 */
    protected Curve newCurve(Curve curve) {
    	return new GlCurve(curve);
    }

//------------------------------------------------------------------------------
/**
 * Makes new copies of the axis objects associated with graph.
 * @param graph	The graph with the template axes.
 */
    protected void copyAxes(Graph graph) {

		m_xAxis = new GlXAxis(graph.getXAxis());
		m_yAxis = new GlYAxis(graph.getYAxis());
        m_xAxis.setOpposingZoom(m_yAxis.getZoom());
        m_yAxis.setOpposingZoom(m_xAxis.getZoom());

        m_xZoom = (Zoom)m_xAxis.getZoom();
        m_yZoom = (Zoom)m_yAxis.getZoom();
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {

        if (glut == null || newgl != gl || newglu != glu)
            glut = new GLUT();

        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Draw this graph with all its parts on the screen.
 * <p><b> Pre Conditions </b><br>
 *      gl has been set to a valid OpenGL context.
 */
    public void drawGL() {

        gl.glScissor(m_display.x, m_display.y, m_display.width,
                     m_display.height);

    	// Move to the correct position on the display area for this graph.
    	gl.glMatrixMode(gl.GL_PROJECTION);
        gl.glTranslatef((float)m_display.x, (float)m_display.y, 0);

        gl.glMatrixMode(gl.GL_MODELVIEW);
        gl.glLoadIdentity();
        gl.glPushMatrix();
        gl.glScaled(m_display.width, m_display.height, 1);

        // If this graph is selected, mark it on the screen.
        if (m_bSelected) drawAsSelected();

        gl.glTranslated(m_frame.x, m_frame.y, 0);
        gl.glScaled(m_frame.width, m_frame.height, 1);

        // Paint a background for the curves.
        if (!m_bQuickDraw) drawBackground();

        // Change the transformation so the curves will fit inside the graph
        // using their coordinates.
        gl.glPushMatrix();
        gl.glScaled(m_xZoom.getScale(), m_yZoom.getScale(), 1);
        gl.glTranslated(-m_xZoom.getViewMin(), -m_yZoom.getViewMin(), 0);

        // Set the scale for the axes so the x and y ticks appear equally
        // long on the screen.
        m_xAxis.setDisplayScale(m_xTickScale);
        m_yAxis.setDisplayScale(m_yTickScale);

    	// Draw the axes and grid.
        gl.glColor3fv(m_frameColor, 0);
        ((GlAxis)m_xAxis).setOpenGL(gld, gl, glu);
        ((GlXAxis)m_xAxis).drawGL();
        ((GlAxis)m_yAxis).setOpenGL(gld, gl, glu);
        ((GlYAxis)m_yAxis).drawGL();

        // Draw the curves inside the graph.
///        setupClipping();
        int x = (int)Math.round(m_frame.x * (double)m_display.width
            + (double)m_display.x);
        int y = (int)Math.round(m_frame.y * (double)m_display.height
            + (double)m_display.y);
        int width = (int)Math.round(m_frame.width * (double)m_display.width);
        int height = (int)Math.round(m_frame.height * (double)m_display.height);
        gl.glScissor(x, y, width, height);

        // Scale the content if the axes have a unit scale.
        double xUnitScale = m_xZoom.getUnitScale();
        double yUnitScale = m_yZoom.getUnitScale();
        if (xUnitScale > 1.0001 || xUnitScale < 0.999
    		|| yUnitScale > 1.0001 || yUnitScale < 0.999) {

        	gl.glTranslated(m_xZoom.getViewMin(), m_yZoom.getViewMin(), 0);
        	gl.glScaled(1 / xUnitScale, 1 / yUnitScale, 1);
        	gl.glTranslated(-m_xZoom.getViewMin() * xUnitScale,
        	-m_yZoom.getViewMin() * yUnitScale, 0);
        }

        Rectangle2D.Double outerRect = new Rectangle2D.Double(
            0, 0, m_display.width, m_display.height);
        Rectangle2D.Double innerRect = new Rectangle2D.Double(
            m_frame.x, m_frame.y, m_frame.width, m_frame.height);
        Rectangle2D.Double innerUser = new Rectangle2D.Double(
            m_xZoom.getViewMin(), m_yZoom.getViewMin(),
            m_xZoom.getViewLength(), m_yZoom.getViewLength());

        // Draw all child graphical elements other than curves, axes, and
        // markups.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof GraphicalElement && !(gro instanceof Curve)){
                if (gro instanceof GlElement && !(gro instanceof GlAxis)) {
                    if (!(gro instanceof Markup)) {
                        GlElement glel = (GlElement)gro;
                        ((GraphicalElement)glel).setRects(
                            outerRect, innerRect, innerUser);
                        glel.setOpenGL(gld, gl, glu);
                        glel.drawGL();
                    }
                }
            }
        }

        gl.glEnableClientState(gl.GL_VERTEX_ARRAY);

        Iterator curveIter = m_curves.iterator();
        while (curveIter.hasNext()) {
            GlElement glel = (GlElement)curveIter.next();

            if (glel instanceof GlCurve) {
                ((GlCurve)glel).m_viewMin = m_xZoom.getViewMin();
                ((GlCurve)glel).m_viewMax = m_xZoom.getViewMax();
            }

            glel.setOpenGL(gld, gl, glu);
            glel.drawGL();
        }

        // Reset the drawing area so deleted markers can be drawn.
        int partialHeight = (int)Math.round((1.0 - m_frame.y)
            * (double)m_display.height);
        gl.glScissor(x, y, width, partialHeight);

        if (!m_bDrawDeletedOutside) drawDeletedMarkers();

        // Restore the previous transformation.
        gl.glPopMatrix();

//        disableClipping();
		// Disable the top clipping plane so the deleted point markers can
        // be drawn.
//	    gl.glDisable (gl.GL_CLIP_PLANE3);

        // Change the transformation so the x axis is in curve coordinates,
        // but the y axis goes from 0 to 1.  Draw markers for each deleted
        // point.
        gl.glPushMatrix();
        gl.glScaled(m_xZoom.getScale(), 1, 1);
        gl.glTranslated(-m_xZoom.getViewMin(), 0, 0);
		if (m_bDrawDeletedOutside) drawDeletedMarkers();
        gl.glPopMatrix();

		// Disable the remaining clipping planes.
//	    gl.glDisable(gl.GL_CLIP_PLANE0);
//	    gl.glDisable(gl.GL_CLIP_PLANE1);
//	    gl.glDisable(gl.GL_CLIP_PLANE2);
        // Reset the drawing area so labels and the frame can be drawn.
        gl.glScissor(m_display.x, m_display.y, m_display.width,
                     m_display.height);

        // Draw a frame around the plot.
        drawFrame();

        // Draw a bar along the side.
        gl.glPopMatrix();
        if (m_bShowSideBar) drawSideBar();

        // Reset the model transformation and revert to the projection trans.
        gl.glLoadIdentity();
    	gl.glMatrixMode(gl.GL_PROJECTION);

        // Pass in x, y, width, and height in screen coordinates.
        gl.glColor3fv(m_textColor, 0);
        int fontIndex = GLUT.BITMAP_HELVETICA_10;
//        GLSolidFont font = new GLSolidFont(
//            new java.awt.Font("SanSerif", java.awt.Font.PLAIN, 16), 0);
        ((GlXAxis)m_xAxis).drawText(m_frame.x * m_display.width,
        	m_frame.y * m_display.height, m_frame.width * m_display.width,
            fontIndex);
        ((GlYAxis)m_yAxis).drawText(m_frame.x * m_display.width,
        	m_frame.y * m_display.height, m_frame.height * m_display.height,
            fontIndex);

        // Add text on top of the graph.
        if (m_bShowTitle) drawText(fontIndex);

        // Draw all markups.
        Iterator iter2 = m_children.iterator();
        while (iter2.hasNext()) {
        	GrObject gro = (GrObject)iter2.next();
            if (gro instanceof Markup) {
//                Rectangle2D.Double outerRect = new Rectangle2D.Double(
//                    0, 0, m_display.width, m_display.height);
//                Rectangle2D.Double innerRect = new Rectangle2D.Double(
//                    m_frame.x, m_frame.y, m_frame.width, m_frame.height);
//                Rectangle2D.Double innerUser = new Rectangle2D.Double(
//                    m_xZoom.getViewMin(), m_yZoom.getViewMin(),
//                    m_xZoom.getViewLength(), m_yZoom.getViewLength());
                ((Markup)gro).setRects(outerRect, innerRect, innerUser);

                GlElement glel = (GlElement)gro;
                glel.setOpenGL(gld, gl, glu);
                glel.drawGL();
            }
        }

        gl.glTranslatef((float)-m_display.x, (float)-m_display.y, 0);

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Draw a background behind the curves and axes.
 * <p><b> Pre Conditions </b><br>
 *      The viewing transformation has been set up so that the graphing
 *		area is between 0 and 1 on both axes.
 * <p><b> Post Conditions </b><br>
 *      The current drawing color has been modified.
 */
    protected void drawBackground() {

	    // Draw black to blue fade rectangle to cover any objects in background.
	    gl.glBegin(gl.GL_QUADS); {
		    gl.glColor3fv(m_bottomBgColor, 0);
		    gl.glVertex2f(0f, 0f);
		    gl.glVertex2f(1f, 0f);
		    gl.glColor3fv(m_topBgColor, 0);
		    gl.glVertex2f(1f, 1f);
		    gl.glVertex2f(0f, 1f);
		} gl.glEnd();
    }

//------------------------------------------------------------------------------
/**
 * Draw a thick yellow frame around the graph.
 * <p><b> Pre Conditions </b><br>
 *      The viewing transformation has been set up so that the graphing
 *		area is between 0 and 1 on both axes.
 * <p><b> Post Conditions </b><br>
 *      None.
 */
    protected void drawFrame() {

	    // Save the calling function's GL settings.
	    gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

    	// Draw a thick yellow frame around the graph.
        gl.glColor3fv(m_frameColor, 0);
	    gl.glLineWidth(2.0f);
	    gl.glBegin(gl.GL_LINE_LOOP); {
		    gl.glVertex2f(0f, 0f);
		    gl.glVertex2f(1f, 0f);
		    gl.glVertex2f(1f, 1f);
		    gl.glVertex2f(0f, 1f);
		} gl.glEnd();

	    // Restore GL settings.
	    gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 * Draw a thick point above the frame to mark each deleted point.
 * <p><b> Pre Conditions </b><br>
 *      The viewing transformation has been set up so that the graphing
 *		area is between 0 and 1 on the y axes and in curve coordinates on
 *		the x axis.
 * <p><b> Post Conditions </b><br>
 *      None.
 */
    protected void drawDeletedMarkers() {

	    // Save the calling function's GL settings.
	    gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

        gl.glPointSize(5);
        // Draw deleted points.
        gl.glColor3f(0f, 1f, 1f);
        gl.glBegin(gl.GL_POINTS);

        Iterator iter = m_curves.iterator();
        while (iter.hasNext()) {
            // Check each curve for deleted points.
            GlCurve curve;
            try {
                curve = (GlCurve)iter.next();
            }
            catch (ClassCastException cce) {
                continue;
            }
            if (!curve.getShow()) continue;

            DoubleDataSeries series = curve.getSeries();
            if (series == null)
                return;

            synchronized (series) {
                series.m_bParabolicInterpolation = false;

                List<Double> delList = series.getDeletedList();
                Iterator<Double> delIter = delList.iterator();
                if (series.getNumSelected() > 0 || curve.getEditing()) {
                    // If the curve is selected and being edited, show the
                    // markers for its deleted points.
                    double yVal = 1.01;
                    while (delIter.hasNext()) {
                        double xVal = delIter.next();
                        delIter.next();

                        if (!m_bDrawDeletedOutside) {
                            yVal = series.getInterpolatedValue(xVal);
                        }

                        gl.glVertex2d(xVal, yVal);
                    }
                }
            }
        }
        gl.glEnd();

	    // Restore GL settings.
	    gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 * Draw a rectangle.
 * <p><b> Pre Conditions </b><br>
 *      The viewing transformation has been set up so that the graphing
 *		area is between 0 and 1 on both axes.
 */
    protected void drawSideBar() {

	    // Save the calling function's GL settings.
	    gl.glPushAttrib(gl.GL_POLYGON_BIT | gl.GL_CURRENT_BIT);

		double leftX = 0;
		double rightX = 10;
    	double bottomY = 0;
    	double topY = m_display.height;
		double triHeight = (topY - bottomY) * 0.02f + 0.5f;

        // Draw the grab bar along the side.
		gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL);

        if (m_bQuickDraw) {
    		if (m_bSideBarDown) gl.glColor3f(0.9f, 0.9f, 1f);
    		else gl.glColor3f(0.3f, 0.3f, 0.4f);

    		gl.glBegin(gl.GL_QUADS);
    		gl.glVertex2d(rightX, topY);
    		gl.glVertex2d(rightX, bottomY);
	    	gl.glVertex2d(leftX, bottomY);
    		gl.glVertex2d(leftX, topY);
    		gl.glEnd();

	    	// Restore GL settings.
	    	gl.glPopAttrib();
        	return;
        }

    	gl.glBegin(gl.GL_QUADS);

    	// Main part of raised bar...
    	if (m_bSideBarDown) gl.glColor3f(0.9f, 0.9f, 1f);
    	else gl.glColor3f(0.3f, 0.3f, 0.4f);
    	gl.glVertex2d(rightX, topY);
    	gl.glVertex2d(rightX, bottomY);

    	if (m_bSideBarDown) gl.glColor3f(0.3f, 0.3f, 0.4f);
    	else gl.glColor3f(0.9f, 0.9f, 1f);
    	gl.glVertex2d(leftX, bottomY);
    	gl.glVertex2d(leftX, topY);
    	gl.glEnd();

		gl.glBegin(gl.GL_TRIANGLES);

		// Left of raised bar...
    	if (m_bSideBarDown) gl.glColor3f(0.8f, 0.8f, 0.9f);
    	else gl.glColor3f(0.4f, 0.4f, 0.5f);
		gl.glVertex2d(leftX, bottomY);
		gl.glVertex2d(rightX, bottomY);
		gl.glVertex2d((leftX + rightX) / 2, bottomY + triHeight);

		// Right of raised bar...
    	if (m_bSideBarDown) gl.glColor3f(0.4f, 0.4f, 0.5f);
    	else gl.glColor3f(0.8f, 0.8f, 0.9f);
		gl.glVertex2d(rightX, topY);
		gl.glVertex2d(leftX, topY);
		gl.glVertex2d((leftX + rightX) / 2, topY - triHeight);
    	gl.glEnd();

	    // Restore GL settings.
	    gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 * Draws title in magenta above the graph.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context.
 */
    protected void drawText(int fontIndex) {
//    	GLSolidFont font) {

//        Graphics2D g = (Graphics2D)((GLComponent)gld).getGraphics();

        // Save the current position
        gl.glPushMatrix();
        double yPos =
        	(double)((m_frame.y + m_frame.height) * m_display.height + 8);
        if (yPos > m_display.height - 10) yPos = (float)m_display.height - 10;
        gl.glTranslated((double)(m_frame.x * m_display.width), yPos, 0);

        // Draw the text.
        gl.glRasterPos2f(0f, 0f);
//        font.drawString(g, m_title);
///        font.drawString(m_title);
        glut.glutBitmapString(fontIndex, m_title);

        // Restore the previous position.
        gl.glPopMatrix();
    }

//------------------------------------------------------------------------------
/**
 * Mark this graph on the screen as being selected.
 * <p><b> Pre Conditions </b><br>
 *      Viewing transformation has been set so this graph's section of the
 *		page is between 0 and 1 in both x and y directions.
 * <p><b> Post Conditions </b><br>
 *      The current drawing color has been changed.
 */
    protected void drawAsSelected() {

    	// Draw green corners at the left of the graph's area.
        gl.glColor3f(0.1f, 0.4f, 0.1f);
        gl.glBegin(gl.GL_TRIANGLES); {
            gl.glVertex2f(0f, 0f);
            gl.glVertex2f(0.02f, 0f);
            gl.glVertex2f(0f, 0.05f);
            gl.glVertex2f(0f, 1f);
            gl.glVertex2f(0.02f, 1f);
            gl.glVertex2f(0f, 0.95f);
        } gl.glEnd();
    }

//------------------------------------------------------------------------------
/**
 * Define clipping planes around the graph.
 * <p><b> Pre Conditions </b><br>
 *      The viewing transformation has been set up so that the graphing
 *		area is in graph coordinates.
 * <p><b> Post Conditions </b><br>
 *      None.
 * @param gl    The OpenGL object containing the drawing commands.
 */
    protected void setupClipping() {

        // Horizontal clipping plane.
	    double[] eqn0 = {0, 1, 0, 0};
        // Vertical clipping plane.
	    double[] eqn1 = {1, 0, 0, 0};

        // Cut off the region below the graph.
	    gl.glPushMatrix();
	    gl.glTranslated(0, m_yZoom.getViewMin(), 0);
	    gl.glClipPlane(gl.GL_CLIP_PLANE0, eqn0, 0);
	    gl.glPopMatrix();

        // Cut off the region to the left of the graph.
	    gl.glPushMatrix();
	    gl.glTranslated(m_xZoom.getViewMin(), 0, 0);
	    gl.glClipPlane (gl.GL_CLIP_PLANE1, eqn1, 0);
	    gl.glPopMatrix();

        // Cut off the region to the right of the graph.
	    gl.glPushMatrix();
	    gl.glTranslated(m_xZoom.getViewMax(), 0, 0);
	    gl.glRotatef(90, 0, 0, 1);
	    gl.glClipPlane (gl.GL_CLIP_PLANE2, eqn0, 0);
	    gl.glPopMatrix();

        // Cut off the region above the graph.
	    gl.glPushMatrix();
	    gl.glTranslated(0, m_yZoom.getViewMax(), 0);
	    gl.glRotatef(-90, 0, 0, 1);
	    gl.glClipPlane(gl.GL_CLIP_PLANE3, eqn1, 0);
	    gl.glPopMatrix();

	    // Enable each of the clipping regions defined above.
        gl.glEnable(gl.GL_CLIP_PLANE0);
	    gl.glEnable(gl.GL_CLIP_PLANE1);
	    gl.glEnable(gl.GL_CLIP_PLANE2);
	    gl.glEnable(gl.GL_CLIP_PLANE3);
	}

//------------------------------------------------------------------------------
/**
 * Disable all clipping planes.
 * @param gl    The OpenGL object containing the drawing commands.
 */
    protected void disableClipping() {

	    gl.glDisable(gl.GL_CLIP_PLANE0);
	    gl.glDisable(gl.GL_CLIP_PLANE1);
	    gl.glDisable(gl.GL_CLIP_PLANE2);
	    gl.glDisable(gl.GL_CLIP_PLANE3);
	}
}



