/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2001-07-19
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.*;
import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class GlRectAnimation
	extends gov.usgs.gr.visual.GraphicalElement implements GlElement {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;

    public DoubleDataSeries m_series = new DoubleDataSeries(4);
    protected double m_scale = 1;
    protected float[] m_color = {1, 0.2f, 0.2f};
    protected TimeMarker m_timeMarker = new TimeMarker();
    protected ColorScheme m_colorScheme = null;
    /**
     * The numeric code of the symbol to be used on this curve.
     *  0 = single pixel.
     *  1 = plus sign.
     *  2 = X
     *  3 = tick mark
     *  4 = hollow circle
     *  5 = hollow square
     *  6 = hollow upward-pointing triangle
     *  7 = hollow downward-pointing triangle
     *  8 = hollow diamond
     *  9 = filled circle
     * 10 = filled square
     * 11 = filled upward-pointing triangle
     * 12 = filled downward-pointing triangle
     * 13 = filled diamond
     */
    protected int m_symbolIndex = 0;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GlRectAnimation.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public GlRectAnimation(){

        setDefaults("gov.usgs.gr.visual.gl.GlRectAnimation");

        m_series.setArray(new double[0]);
    	add(getDefaultColorScheme());
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a curve from a data series.
 */
    public GlRectAnimation(DoubleDataSeries series){

        setDefaults("gov.usgs.gr.visual.gl.GlRectAnimation");

        add(series);
    	add(getDefaultColorScheme());
   }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Creates a default color scheme.
 */
    public GrObject getDefaultColorScheme(){

    	GrObject cs = new FormulaColorScheme();
        cs.setProperty("redRange", "80");
        cs.setProperty("greenOffset", "60");
        cs.setProperty("greenRange", "-60");
        cs.setProperty("blueRange", "40");

        return cs;
	}

//------------------------------------------------------------------------------
/**
 * Add a child object.  Accepts DoubleDataSeries, TimeMarker.  Only holds one DS at
 * a time.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

    	if (gro instanceof DoubleDataSeries) {
            remove(m_series);
            super.add(gro);
            m_series = (DoubleDataSeries)gro;

            if (m_colorScheme != null)
                super.add((GrObject)m_colorScheme);

            return true;
        }
        else {
        	if (gro instanceof ColorScheme) {
            	if (m_colorScheme != null)
                	remove((GrObject)m_colorScheme);
                super.add(gro);
                m_colorScheme = (ColorScheme)gro;

                return true;
        	}
        	if (gro instanceof TimeMarker) {
            	if (m_timeMarker != null)
                	remove(m_timeMarker);
                super.add(gro);
                m_timeMarker = (TimeMarker)gro;

                return true;
        	}
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("scale", new GrProperty(m_scale, "Scale"));

        // Give the user some symbol choices...
        GrProperty p = new GrProperty("", "Symbol Type");
        javax.swing.JComboBox jcb = new javax.swing.JComboBox();

        String[] symbolTypes = {
        	"Single Pixel", "Plus Sign", "X", "Tick Mark",
            "Hollow Circle", "Hollow Square", "Hollow Triangle Pointing Up",
            "Hollow Triangle Pointing Down", "Hollow Diamond",
            "Filled Circle", "Filled Square", "Filled Triangle Pointing Up",
            "Filled Triangle Pointing Down", "Filled Diamond"
            };

        // Fill the symbol type combo box.
        for (int i = 0; i < symbolTypes.length; i++) {
        	String name = symbolTypes[i];
        	jcb.addItem(name);
            if (m_symbolIndex == i) {
				jcb.setSelectedIndex(i);
                p.value = symbolTypes[i];
            }
        }

        p.component = jcb;
		props.put("symbolType", p);

        ColorProperty cp = new ColorProperty();
        cp.description = "Color";
        cp.setColor(m_color);
        props.put("color", cp);

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        String[] symbolTypes = {
        	"single pixel", "plus sign", "x", "tick mark", "hollow circle",
            "hollow square", "hollow triangle pointing up",
            "hollow triangle pointing down", "hollow diamond", "filled circle",
            "filled square", "filled triangle pointing up",
            "filled triangle pointing down", "filled diamond"
        };

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("scale"))
            		m_scale = (Double)p.value;
            	if (key.equals("symbolType")) {
        			// Check each name.
        			for (int i = 0; i < symbolTypes.length; i++) {
        				String name = ((String)p.value).trim().toLowerCase();
                        if (name.equals(symbolTypes[i])) {
                        	m_symbolIndex = i;
                        	break;
                        }
                    }
                }
            	if (key.equals("color")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_color = cp.getColorArray();
                }
            }
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("scale"))
            m_scale = Double.parseDouble(value);
        if (key.equals("symbolType")) {
        	String[] symbolTypes = {
        		"single pixel", "plus sign", "x", "tick mark", "hollow circle",
            	"hollow square", "hollow triangle pointing up",
            	"hollow triangle pointing down", "hollow diamond",
                "filled circle", "filled square", "filled triangle pointing up",
            	"filled triangle pointing down", "filled diamond"
        	};
        	// Check each name.
            for (int i = 0; i < symbolTypes.length; i++) {
            	String name = value.trim().toLowerCase();
                if (name.equals(symbolTypes[i])) {
                	m_symbolIndex = i;
                    break;
                }
            }
        }
        if (key.equals("color")) {
            m_color = ColorProperty.getColorArray(value);
        }
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        if (!m_bShow || m_series == null) return null;

        Point2D.Double p = new Point2D.Double(m_series.getMin(dimension),
            m_series.getMax(dimension));

        return p;
    }

//------------------------------------------------------------------------------
/**
 * Hide or show this series.
 * @param  bShow    True if this curve should be drawn.
 */
    public void show(boolean bShow){
        m_bShow = bShow;
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {
        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 */
    public void drawGL() {

        // Before drawing, check if this object is supposed to be shown.
        if (!m_bShow) return;

        synchronized (m_series) {

            // Save the previous line color and pattern.
            gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

            int xDim = 0;
            int yDim = 1;
            int zDim = 2;

            int numDims = m_series.getNumDimensions();
            int dimensionOffset = (numDims == 3 ? 0 : 1);

            if (numDims == 6) {
                dimensionOffset = 1;
            }

            double[] array = m_series.getArray();
            for (int i = m_series.getBegin(); i <= m_series.getEnd(); i++) {
                int offset = i * numDims;

                for (int j = dimensionOffset;
                    j < (m_series.getNumDimensions() - 1); j += 2) {

                    double x = array[offset + j + xDim];
                    double y = array[offset + j + yDim];
                    double z = array[offset + j + zDim];

                    gl.glPushMatrix();
                    gl.glTranslated(x, y, 0);
                    gl.glScaled(m_scale, m_scale, 1);

                    // Set the color.
                    //gl.color(m_color);
                    if (numDims == 6) {
                        if (j == 1) {
                            if (array[offset + j + zDim] >= 0) {
                                gl.glColor3d(array[offset + j + zDim],
                                             array[offset + j + zDim + 1],
                                             array[offset + j + zDim + 2]);
                                drawSquare();
                            }
                            else {
                                gl.glColor3d(1, 1, 1);
                                drawTwoToneSquare();
                            }
                        }
                    }
                    else {
                        gl.glColor3fv(m_colorScheme.getColor(z), 0);
                        drawSquare();
                    }


    	            gl.glPopMatrix();
                }
            }

            // Restore the previous line attributes.
            gl.glPopAttrib();
        }

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw one two-tonw square.
 */
    public void drawTwoToneSquare() {

        drawSquare();

        gl.glScalef(0.6f, 0.6f, 1);

        // Set the color.
        gl.glPushAttrib(gl.GL_CURRENT_BIT);
        gl.glColor3f(0, 0, 0);

        drawSquare();

        gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw one square.
 */
    public void drawSquare() {

        gl.glBegin(gl.GL_QUADS);
        gl.glVertex2f(-0.5f, -0.5f);
        gl.glVertex2f(-0.5f, 0.5f);
        gl.glVertex2f(0.5f, 0.5f);
        gl.glVertex2f(0.5f, -0.5f);
        gl.glEnd();
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        // Check if the object is a graphical element that needs to be redrawn.
        if (gro instanceof DoubleDataSeries) {
            m_bNeedsRedraw = true;
        }

    	m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Change the color of the curve.
 * @param red   The red components value between 0.0 and 1.0
 * @param green The green components value between 0.0 and 1.0
 * @param blue  The blue components value between 0.0 and 1.0
 */
    public void setColor(float red, float green, float blue){
        m_color[0] = red;
        m_color[1] = green;
        m_color[2] = blue;
    }

//------------------------------------------------------------------------------
/**
 * Change the color of the curve (see other setColor routine).
 * @param color An array of floats decribing the red, green, and blue.
 */
    public void setColor(float[] color){
        setColor(color[0], color[1], color[2]);
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s = null;
        if (m_bShow)
            s = "GL Rectangle Animation: " + m_title;
        else
            s = "(GL Rectangle Animation): " + m_title;

        return s;
    }
}






