/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-01-04
 * Date started:    2000-10-20
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.MultiThreader;
import gov.usgs.gr.visual.*;
import java.awt.geom.*;
import java.awt.image.BufferedImage;
import java.io.*;
import java.nio.ByteBuffer;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class GlTexture extends gov.usgs.gr.visual.GraphicalElement
    implements gov.usgs.gr.visual.gl.GlElement, java.lang.Runnable {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;
    /** The name of the data set's file. */
    protected File m_file = new File("/");

    public DoubleDataSeries m_series = new DoubleDataSeries(4);
    protected TextureLoader m_textureLoader = null;
    protected Texture m_texture = null;
    protected int[] m_textureNums = {1};
    protected boolean m_bTexGenerated = false;
    protected String m_errorString = null;
    protected String m_resourceName = "";

    protected BufferedImage m_bufferedImage = null;
    protected ByteBuffer m_byteBuffer = null;

    /**
     * Describes the location and size of the texture.
     */
    protected Point2D.Double m_bl = new Point2D.Double(0, 0);
    protected Point2D.Double m_br = new Point2D.Double(100, 0);
    protected Point2D.Double m_tr = new Point2D.Double(100, 100);
    protected Point2D.Double m_tl = new Point2D.Double(0, 100);
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GlTexture.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public GlTexture(){

        setDefaults("gov.usgs.gr.visual.gl.GlTexture");
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a curve from a data series.
 */
    public GlTexture(DataSeries series){

        setDefaults("gov.usgs.gr.visual.gl.GlTexture");

        add(series);
   }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("file", new GrProperty(m_file.getAbsolutePath(), "File"));
        props.put("bottomLeftX", new GrProperty(
            m_bl.x, "Bottom-Left X Coordinate"));
        props.put("bottomLeftY", new GrProperty(
            m_bl.y, "Bottom-Left Y Coordinate"));
        props.put("bottomRightX", new GrProperty(
            m_br.x, "Bottom-Right X Coordinate"));
        props.put("bottomRightY", new GrProperty(
            m_br.y, "Bottom-Right Y Coordinate"));
        props.put("topRightX", new GrProperty(
            m_tr.x, "Top-Right X Coordinate"));
        props.put("topRightY", new GrProperty(
            m_tr.y, "Top-Right Y Coordinate"));
        props.put("topLeftX", new GrProperty(
            m_tl.x, "Top-Left X Coordinate"));
        props.put("topLeftY", new GrProperty(
            m_tl.y, "Top-Left Y Coordinate"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("bottomLeftX"))
            		m_bl.x = (Double)p.value;
            	if (key.equals("bottomLeftY"))
            		m_bl.y = (Double)p.value;
            	if (key.equals("bottomRightX"))
            		m_br.x = (Double)p.value;
            	if (key.equals("bottomRightY"))
            		m_br.y = (Double)p.value;
            	if (key.equals("topRightX"))
            		m_tr.x = (Double)p.value;
            	if (key.equals("topRightY"))
            		m_tr.y = (Double)p.value;
            	if (key.equals("topLeftY"))
            		m_tl.y = (Double)p.value;
            	if (key.equals("topLeftX"))
            		m_tl.x = (Double)p.value;
            	if (key.equals("file")) {
            		m_file = new File((String)p.value);
                    m_textureLoader = null;
                }
            }
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("bottomLeftX"))
            m_bl.x = Double.parseDouble(value);
        if (key.equals("bottomLeftY"))
            m_bl.y = Double.parseDouble(value);
        if (key.equals("bottomRightX"))
            m_br.x = Double.parseDouble(value);
        if (key.equals("bottomRightY"))
            m_br.y = Double.parseDouble(value);
        if (key.equals("topRightX"))
            m_tr.x = Double.parseDouble(value);
        if (key.equals("topRightY"))
            m_tr.y = Double.parseDouble(value);
        if (key.equals("topLeftX"))
            m_tl.x = Double.parseDouble(value);
        if (key.equals("topLeftY"))
            m_tl.y = Double.parseDouble(value);
        if (key.equals("file")) {
            m_file = new File(value);
            m_textureLoader = null;
        }
    }

//------------------------------------------------------------------------------
/**
 * Add a child object.  Accepts DataSeries, TimeMarker.  Only holds one DS at
 * a time.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

    	if (gro instanceof DoubleDataSeries) {
            remove(m_series);
            super.add(gro);
            m_series = (DoubleDataSeries)gro;
            return true;
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public double[] getCorners() {

        double[] corners = new double[8];

        corners[0] = m_bl.x;
        corners[1] = m_bl.y;
        corners[2] = m_br.x;
        corners[3] = m_br.y;
        corners[4] = m_tr.x;
        corners[5] = m_tr.y;
        corners[6] = m_tl.x;
        corners[7] = m_tl.y;

        return corners;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void setCorners(double[] corners) {

        m_bl.x = corners[0];
        m_bl.y = corners[1];
        m_br.x = corners[2];
        m_br.y = corners[3];
        m_tr.x = corners[4];
        m_tr.y = corners[5];
        m_tl.x = corners[6];
        m_tl.y = corners[7];
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        if (!m_bShow || m_series == null) return null;

        double min = Double.POSITIVE_INFINITY;
        double max = Double.NEGATIVE_INFINITY;
        if (dimension == 0) {
            min = Math.min(min, m_bl.x);
            min = Math.min(min, m_br.x);
            min = Math.min(min, m_tr.x);
            min = Math.min(min, m_tl.x);
            max = Math.max(max, m_bl.x);
            max = Math.max(max, m_br.x);
            max = Math.max(max, m_tr.x);
            max = Math.max(max, m_tl.x);
        }
        if (dimension == 1) {
            min = Math.min(min, m_bl.y);
            min = Math.min(min, m_br.y);
            min = Math.min(min, m_tr.y);
            min = Math.min(min, m_tl.y);
            max = Math.max(max, m_bl.y);
            max = Math.max(max, m_br.y);
            max = Math.max(max, m_tr.y);
            max = Math.max(max, m_tl.y);
        }

        Point2D.Double p = new Point2D.Double(min, max);

        return p;
    }

//------------------------------------------------------------------------------
/**
 * Hide or show this series.
 * @param  bShow    True if this curve should be drawn.
 */
    public void show(boolean bShow){
        m_bShow = bShow;
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {
        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 */
    public void drawGL() {

        // Before drawing, check if this object is supposed to be shown.
        if (!m_bShow) return;

        drawImage();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw an image.
 */
    public void drawImage() {

        gl.glPushAttrib(gl.GL_ENABLE_BIT | gl.GL_HINT_BIT);

        gl.glTexEnvi(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_DECAL);
        gl.glEnable(gl.GL_TEXTURE_2D);
        gl.glHint(gl.GL_PERSPECTIVE_CORRECTION_HINT, gl.GL_NICEST);

        if (!bindTexture()) {
            gl.glPopAttrib();
            return;
        }

        gl.glBegin(gl.GL_QUADS);
            gl.glTexCoord2f(0, 1);
            gl.glVertex3d(m_bl.x, m_bl.y, 0);
            gl.glTexCoord2f(1, 1);
            gl.glVertex3d(m_br.x, m_br.y, 0);
            gl.glTexCoord2f(1, 0);
            gl.glVertex3d(m_tr.x,m_tr.y, 0);
            gl.glTexCoord2f(0, 0);
            gl.glVertex3d(m_tl.x, m_tl.y, 0);
        gl.glEnd();

        gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public boolean bindTexture() {
        //load the texture
        if (m_textureLoader == null) {
            m_textureLoader = new TextureLoader();
            m_bufferedImage = null;
            m_byteBuffer = null;
        }
        if (m_byteBuffer == null) {

            m_bTexGenerated = false;
            try {
                // If the file does not exist, try fixing the path.
                if (!m_file.exists()) {
                    // Try using the same filename within the current directory.
                    String currentFileName =
                        System.getProperty("gov.usgs.gr.currentfile");
                    File file = new File(currentFileName);
                    String parentString = file.getAbsoluteFile().getParent();
                    parentString.replace('\\', '/');
                    file = new File(file.getAbsoluteFile().getParent() + "/"
                        + m_file.getName());

                    if (file.exists())
                        m_file = file;
                }

                // If the file exists, try loading it into the texture.
                if (m_file.exists()) {
                        Thread t1 = new Thread(this, "TextureLoader");
                        t1.start();
                }
                else {
                        // Write an error message if the file can't be found.
                        System.out.println("Could not find texture image file "
                            + m_file.getAbsolutePath());
                        m_errorString = " -- Error -- file does not exist";
                }
            }
            catch ( Exception e ) {
                e.printStackTrace();
                return false;
            }
        }

        if (m_texture == null) {
            if (m_bufferedImage == null || m_byteBuffer == null)
                return false;

            m_texture =
                m_textureLoader.getTexture("gr", m_resourceName,
                    m_bufferedImage, m_byteBuffer);
       }

        m_texture.bind(gl);

        return true;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void preLoadImage() {
        
        if (m_textureLoader == null)
            m_textureLoader = new TextureLoader();
        
        try {
            m_resourceName = m_file.getAbsolutePath();
            m_bufferedImage = m_textureLoader.loadImage(m_resourceName);
            
            m_bl.x = 0;
            m_bl.y = 0;
            m_br.x = m_bufferedImage.getWidth();
            m_br.y = 0;
            m_tr.x = m_bufferedImage.getWidth();
            m_tr.y = m_bufferedImage.getHeight();
            m_tl.x = 0;
            m_tl.y = m_bufferedImage.getHeight();
        } catch (IOException ex) {
            ex.printStackTrace();
        }
    }
    
//------------------------------------------------------------------------------
/**
 * Loads the texture from disk.
 */
    public void run() {

        try {
            m_textureLoader.initializeDrawable(gld);
            if (m_bufferedImage == null) {
                m_resourceName = m_file.getAbsolutePath();
                m_bufferedImage = m_textureLoader.loadImage(m_resourceName);
            }
            m_byteBuffer = m_textureLoader.convertImageData(m_bufferedImage);

            m_bNeedsRedraw = true;
            m_bHasChanged = true;
            notifyObservers();
        }
        catch (IOException ioe) {
            m_errorString = "Texture not loaded...";
            return;
        }
        m_errorString = null;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        if (gro == m_series) {
            if (m_series != null && m_series.getNumDimensions() == 2 &&
                m_series.getNumPoints() >= 4) {

                double[] array = m_series.getArray();
                m_bl.x = array[0];
                m_bl.y = array[1];
                m_br.x = array[2];
                m_br.y = array[3];
                m_tr.x = array[4];
                m_tr.y = array[5];
                m_tl.x = array[6];
                m_tl.y = array[7];
            }
        }

    	m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s = null;
        if (m_bShow)
            s = "GL Texture: ";
        else
            s = "(GL Texture): ";

        if (m_errorString != null)
            s += m_errorString;
        else
            s += m_title;

        return s;
    }
}






