/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2005-11-04
 * Date started:    2000-10-20
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import com.sun.opengl.util.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.*;
import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class GlTransformation
	extends gov.usgs.gr.visual.GraphicalElement implements GlElement {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;

    protected GrObjectContainer m_transformationContainer = null;

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public GlTransformation(){

        setDefaults("gov.usgs.gr.visual.gl.GlTransformation");

        GrObjectContainer groc = new GrObjectContainer();
        groc.setName("Transformations");
        groc.setProperty("title", "Transformations");
        add(groc);
    }

//------------------------------------------------------------------------------
/**
 * Add a child object.  Accepts GraphicalElement and TransformationOperation.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        if (gro instanceof GraphicalElement) {
            super.add(gro);
            return true;
        }
        if (gro instanceof GrObjectContainer) {
            GrObjectContainer groc = (GrObjectContainer)gro;
            GrHashList props = groc.getProperties();
            GrProperty p = (GrProperty)props.get("title");
            if (p == null)
                return false;
            String title = (String)p.value;

            // Compare the title of groc to each container and replace the
            // current one if there is a match.
            if (title.equals("Transformations")) {
                GrObjectContainer oldGroc = null;

                Iterator iter = m_children.iterator();
                while (iter.hasNext()) {
                    GrObject child = (GrObject)iter.next();
                    if (child instanceof GrObjectContainer) {
                        GrObjectContainer oc = (GrObjectContainer)child;
                        GrHashList childProps = oc.getProperties();
                        GrProperty cp = (GrProperty)childProps.get("title");
                        if (cp.value.equals(title)) {
                            oldGroc = oc;
                            break;
                        }
                    }
                }

                int index = m_children.size();
                if (oldGroc != null) {
                    index = m_children.indexOf(oldGroc);
                    oldGroc.deleteObserver(this);
                    m_children.remove(oldGroc);
                }
                m_children.add(index, groc);
                groc.addObserver(this);
                groc.setName("");
                m_transformationContainer = groc;

                m_bHasChanged = true;
                return true;
            }
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        return null;
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {
        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 */
    public void drawGL() {

        // Before drawing, check if this object is supposed to be shown.
        if (!m_bShow) return;

        gl.glPushMatrix();

        double contX = m_contentTrans.getX();
        double contY = m_contentTrans.getY();
        double contWidth = m_contentTrans.getWidth();
        double contHeight = m_contentTrans.getHeight();

        Enumeration transformEnum = m_transformationContainer.children();
        while (transformEnum.hasMoreElements()) {
            GrObject gro = (GrObject)transformEnum.nextElement();
            if (gro instanceof TransformationOperation) {
                TransformationOperation to = (TransformationOperation)gro;

                switch (to.type) {
                    case 0:
                        gl.glTranslated(to.x, to.y, to.z);
                        contX -= to.x;
                        contY -= to.y;
                        break;
                    case 1:
                        gl.glScaled(to.x, to.y, to.z);
                        contX /= to.x;
                        contY /= to.y;
                        contWidth /= to.x;
                        contHeight /= to.y;
                        break;
                    case 2:
                        gl.glRotated(to.angle, to.x, to.y, to.z);
                        break;
                }
            }
        }

        Rectangle2D.Double newTrans =
            new Rectangle2D.Double(contX, contY, contWidth, contHeight);


        // Draw all child graphical elements other than curves, axes, and
        // markups.
        Iterator childIter = m_children.iterator();
        while (childIter.hasNext()) {
            GrObject gro = (GrObject)childIter.next();
            if (gro instanceof GraphicalElement && gro instanceof GlElement) {
                GlElement glel = (GlElement)gro;
                ((GraphicalElement)glel).setRects(
                    m_rootTrans, m_parentTrans, newTrans);
                glel.setOpenGL(gld, gl, glu);
                glel.drawGL();
            }
        }

        gl.glPopMatrix();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s = null;
        if (m_bShow)
            s = "GL Transformation: " + m_title;
        else
            s = "(GL Transformation): " + m_title;

        return s;
    }
}






