/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-11-15
 * Date started:    2001-02-01
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import java.awt.geom.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class GlVectorAnimation
	extends gov.usgs.gr.visual.DataAnimation implements GlElement,
        SelectableGE {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;

    /** Whether to interpret the third and fourth dims as U and V as
     *  opposed to direction and speed.
     *  */
    protected boolean m_bUsingUV = false;

    protected boolean m_bXZLine = false;
    protected String m_numberFormatString = "";
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GlVectorAnimation.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public GlVectorAnimation(){

        initialize();

        GrObjectContainer groc = new GrObjectContainer();
        groc.setName("Unrotated Source");
        add(groc);

        setDefaults("gov.usgs.gr.visual.gl.GlVectorAnimation");
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a curve from a data series.
 */
    public GlVectorAnimation(DataSeries series){

        initialize();

        GrObjectContainer groc = new GrObjectContainer();
        groc.setName("Unrotated Source");
        add(groc);

        add(series);

        setDefaults("gov.usgs.gr.visual.gl.GlVectorAnimation");
   }

//------------------------------------------------------------------------------
/**
 * Initialize all the member variables.
 */
    public void initialize() {

        m_numDimsAllowed = 8;
        m_symbolTypes = new String[] {"Simple Arrow", "Two Tone Arrow",
            "Narrow Arrow", "Line Arrow", "Simple Bar", "Line"};

        m_dataArrays = new double[m_numDimsAllowed][];
        m_seriesArray = new DoubleDataSeries[m_numDimsAllowed + 1];
        m_seriesNums = new int[] {1, 1, 1, 0, 1, 1, 0, 0};
        m_dimNums = new int[] {0, 1, 2, -1, 3, 4, -1, -1};
        m_seriesNumDims = new int[] {0, 0, 0, 0, 0, 0, 0, 0};
        m_dimScales = new double[] {1, 1, 1, 1, 1, 1, 1, 1};
        m_constantSeries = new DoubleDataSeries(new double[m_numDimsAllowed],
            m_numDimsAllowed);
        m_dataArrays[0] = m_constantSeries.getArray();
        m_seriesNumDims[0] = m_constantSeries.getNumDimensions();
        m_seriesArray[0] = m_constantSeries;

        m_color = new float[] {1f, 0.2f, 0.2f};
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 *
 *  The numeric code of the symbol to be used.
 *  0 = simple arrow.
 *  1 = arrow with black inner area.
 *  2 = bar
 *
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("usingUV", new GrProperty(
            m_bUsingUV, "U-V instead of Direction-Magnitude"));
        props.put("xzLine", new GrProperty(m_bXZLine, "Show XZ Lines"));
        props.put("numberFormat", new GrProperty(
        	new String(m_numberFormatString), "Number Format"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

                if (key.equals("usingUV"))
                    m_bUsingUV = (Boolean)(p.value);
                if (key.equals("xzLine"))
                    m_bXZLine = (Boolean)(p.value);
            }
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("usingUV"))
            m_bUsingUV = Boolean.parseBoolean(value);
        if (key.equals("xzLine"))
            m_bXZLine = Boolean.parseBoolean(value);
        if (key.equals("numberFormat"))
            m_numberFormatString = value;
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        Point2D.Double p = super.getBounds(dimension);
        if (p == null) return null;

        try {
            int dim = dimension + 4;
            double vecMin = m_seriesArray[m_seriesNums[dim]].
                getMin(m_dimNums[dim]) * m_dimScales[dim];
            if (vecMin < 0) p.x += vecMin;

            double vecMax = m_seriesArray[m_seriesNums[dim]].
                getMax(m_dimNums[dim]) * m_dimScales[dim];
            if (vecMax < 0) p.y += vecMax;
        }
        catch (NullPointerException e) {}

        return p;
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {
        gld = newgld;
      	gl = newgl;
		glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public boolean checkForValidData() {

        int tCount = pointCount(0);
        int xCount = pointCount(1);
        int yCount = pointCount(2);
        int zCount = pointCount(3);
        int uCount = pointCount(4);
        int vCount = pointCount(5);
        int wCount = pointCount(6);
        int cCount = pointCount(7);

        // Return false if any required fields are missing.
        if (tCount == 0 || xCount == 0 || yCount == 0 || uCount == 0 ||
            vCount == 0) {

            m_errorString = "No series specified (T, X, Y, U, V)";

            return false;
        }

        // Make sure the location fields have the same count or are constant.
        int maxLocCount = Math.max(Math.max(xCount, yCount), zCount);
        if (xCount != maxLocCount && xCount != 1) {
            m_errorString = "Error in X series point count";
            return false;
        }
        if (yCount != maxLocCount && yCount != 1) {
            m_errorString = "Error in Y series point count";
            return false;
        }
        if (zCount != maxLocCount && zCount <= 1 && m_dimNums[3] >= 0) {
            m_errorString = "Error in Z series point count";
            return false;
        }

        // Make sure the U-V-W fields have the same count or are constant.
        int maxUvwCount = Math.max(Math.max(uCount, vCount), wCount);
        if (uCount != maxUvwCount && uCount != 1) {
            m_errorString = "Error in U series point count";
            return false;
        }
        if (vCount != maxUvwCount && vCount != 1) {
            m_errorString = "Error in V series point count";
            return false;
        }
        if (wCount != maxUvwCount && wCount <= 1 && m_dimNums[6] >= 0) {
            m_errorString = "Error in W series point count";
            return false;
        }

        // Make sure T * Loc = Val
        if (tCount * maxLocCount != maxUvwCount && tCount != maxLocCount) {
            m_errorString = "Error -- T * Loc != Val";
            return false;
        }

        // Make sure C = Val
        if (cCount != maxUvwCount && m_dimNums[7] >= 0) {
            m_errorString = "Color count does not match value count";
            m_dimNums[7] = -1;
        }

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 */
    public void drawGL() {

        m_errorString = null;

        // Before drawing, check if this object is supposed to be shown.
        if (!m_bShow) return;

        // Save the previous line color and pattern.
        gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

        // Calculate the number of iterations for the
        // draw loop and the offsets for the T, Loc, and Vec arrays.
        int locOffset = 0;
        int vecOffset = 0;
        int endOffset = 0;
        int seriesNumTime = m_seriesNums[0];
        if (m_seriesArray[seriesNumTime] == null)
            return;

        DataSeries timeSeries = m_seriesArray[seriesNumTime];
        synchronized (timeSeries) {
            if (!checkForValidData()) return;

            int tCount = pointCount(0);
            int xCount = pointCount(1);
            int yCount = pointCount(2);
            int zCount = pointCount(3);
            int uCount = pointCount(4);
            int vCount = pointCount(5);
            int wCount = pointCount(6);
            int cCount = pointCount(7);
            int maxLocCount = Math.max(Math.max(xCount, yCount), zCount);
            int maxUvwCount = Math.max(Math.max(uCount, vCount), wCount);

            if (tCount != maxLocCount || maxLocCount != maxUvwCount) {

                // Not a one-to-one relationship between times, locations,
                // and vectors.
                vecOffset = timeSeries.getBegin() * maxLocCount;
                endOffset = maxLocCount - 1;
            }
            else {
                // One-to-one relationship between times, locations,
                // and vectors.
                locOffset = timeSeries.getBegin();
                vecOffset = 0;
                endOffset = timeSeries.getEnd();
            }
        }

        // Initialize variables for the XZ line.
        double lastX = Double.NaN;
        double lastY = Double.NaN;
        double lastZ = Double.NaN;
        double lastU = Double.NaN;
        double lastV = Double.NaN;
        double lastW = Double.NaN;
        boolean bConnectToLast = false;

        for (int i = locOffset; i <= endOffset; i++) {
            gl.glPushMatrix();

            double x = getValueAt(i, 1);
            double y = getValueAt(i, 2);
            double z = 0;
            if (m_dimNums[3] >= 0) {
                z = getValueAt(i, 3);
            }

            if (m_bXZLine && x == lastX && y == lastY)
                bConnectToLast = true;
            else
                bConnectToLast = false;
            if (bConnectToLast) {
                gl.glBegin(gl.GL_LINES);
                gl.glVertex3d(lastX, lastY, lastZ * m_zScale);
                gl.glVertex3d(x, y, z * m_zScale);
                gl.glEnd();
            }

            double u = 0;
            double v = 0;
            double w = 0;
            double uvTheta = 0;
            double uwTheta = 0;
            double len;
            if (m_bUsingUV) {
                u = getValueAt(i + vecOffset, 4);
                v = getValueAt(i + vecOffset, 5);
                if (m_dimNums[6] >= 0) {
                    w = getValueAt(i + vecOffset, 6);
                    w *= (m_zStretch);
                }

                uvTheta = Math.atan2(v, u);
                uvTheta *= (180 / Math.PI);
                double uvLen = Math.sqrt((u * u) + (v * v));
                if (m_dimNums[6] >= 0) {
                    uwTheta = Math.atan2(w, uvLen);
                    uwTheta *= (180 / Math.PI);
                    len = Math.sqrt((u * u) + (v * v) + (w * w));
                    double lenTest = Math.sqrt(uvLen * uvLen + w * w);
                    int hey = 0;
                }
                else {
                    len = uvLen;
                }
            }
            else {
                uvTheta = getValueAt(i + vecOffset, 4);
                len = getValueAt(i + vecOffset, 5);
                if (m_dimNums[6] >= 0) {
                    uwTheta = getValueAt(i + vecOffset, 6);
                }
                u = len * Math.cos(Math.toRadians(uvTheta));
                v = len * Math.sin(Math.toRadians(uvTheta));
            }

            if (bConnectToLast) {
                gl.glBegin(gl.GL_LINES);
                gl.glVertex3d(lastX + (lastU * m_scale), lastY
                    + (lastV * m_scale), (lastZ * m_zScale) + (lastW * m_scale));
                gl.glVertex3d(x + (u * m_scale), y + (v * m_scale),
                    (z * m_zScale) + (w * m_scale));
                gl.glEnd();

//                gl.glColor3f(0.1f, 0.1f, 0.1f);
//                gl.glBegin(gl.GL_QUADS);
//                gl.glVertex3d(lastX, lastY, lastZ * m_zScale);
//                gl.glVertex3d(x, y, z * m_zScale);
//                gl.glVertex3d(x + (u * m_scale), y + (v * m_scale),
//                    (z * m_zScale) + w);
//                gl.glVertex3d(lastX + (lastU * m_scale),
//                    lastY + (lastV * m_scale), (lastZ * m_zScale) + lastW);
//                gl.glEnd();
            }
            lastU = u;
            lastV = v;
            lastW = w;

            lastX = x;
            lastY = y;
            lastZ = z;

//            gl.glScaled(1, 1, m_zScale);
            gl.glTranslated(x, y, z * m_zScale);
            gl.glRotated(uvTheta, 0, 0, 1);
            if (m_dimNums[6] >= 0)
//                gl.glRotated(-45, 0, 1, 0);
                gl.glRotated(-uwTheta, 0, 1, 0);
            
            // Allow for a constant width.
            double yLen = len;
            double yScale = m_scale;
            if (m_constantWidth > 0) {
                yLen = m_constantWidth;
                yScale = 1;
            }

            gl.glScaled(len, yLen, len);
            gl.glScaled(m_scale, yScale, m_scale);

            // Set the color.
            if (m_colorScheme == null || m_dimNums[7] < 0) {
                gl.glColor3fv(m_color, 0);
            }
            else {
                int offset = i;
                // If there are the same number of points in the color array
                // as in the vector array, add the vector offset.
                if (m_dataArrays[m_seriesNums[7]].length >
                    (offset + vecOffset)) {

                    offset += vecOffset;
                }
                gl.glColor3fv(m_colorScheme.getColor(
                    getValueAt(offset, 7)), 0);
            }

            switch (m_symbolIndex) {
                case 0:
                    drawArrow();
                    break;
                case 1:
                    drawTwoToneArrow();
                    break;
                case 2:
                    drawNarrowArrow();
                    break;
                case 3:
                    // Draw line arrow.
                    gl.glBegin(gl.GL_LINES);
                    gl.glVertex2f(0, 0);
                    gl.glVertex2f(1, 0);
                    gl.glVertex2f(1, 0);
                    gl.glVertex2f(0.8f, 0.07f);
                    gl.glVertex2f(1, 0);
                    gl.glVertex2f(0.8f, -0.06f);
                    //gl.glVertex2(0.8, 0.06);
                    //gl.glVertex2(0.8, -0.06);
                    gl.glEnd();
                    break;
                case 4:
                    drawBar();
                    break;
                case 5:
                    gl.glBegin(gl.GL_LINES);
                    gl.glVertex2f(0, 0);
                    gl.glVertex2f(1, 0);
                    gl.glEnd();
                    break;
                default:
                    drawTwoToneArrow();
                    break;
            }

            gl.glPopMatrix();

            Iterator iter2 = m_children.iterator();
            while (iter2.hasNext()) {
                GrObject gro = (GrObject)iter2.next();
                if (gro instanceof Markup) {
                    Markup m = (Markup)gro;
                    if (m instanceof TextMarkup &&
                        m_numberFormatString.length() > 0) {

                        TextMarkup tm = (TextMarkup)m;

                        try {
                            // Format the current numeric value for the text.
                            java.text.NumberFormat formatter =
                                new java.text.DecimalFormat(
                                m_numberFormatString.trim());
                            String formattedNumber = formatter.format(len);
                            tm.setText(formattedNumber);
                        }
                        catch (Exception e) {}
                    }

                    gl.glPushMatrix();
                    gl.glTranslated(x, y, z * m_zScale);
                    m.setOpenGL(gld, gl, glu);
                    m.drawGL();
                    gl.glPopMatrix();
                }
            }
        }

        // Restore the previous line attributes.
        gl.glPopAttrib();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw one vector arrow.
 */
    public void drawTwoToneArrow() {

        drawArrow();

        gl.glPushMatrix();
        gl.glTranslatef(0.15f, 0, 0);
        gl.glScalef(0.8f, 0.6f, 1);

        // Set the color.
        gl.glPushAttrib(gl.GL_CURRENT_BIT);
        gl.glColor3f(0, 0, 0);

        drawArrow();

        gl.glPopMatrix();
        gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw one vector arrow.
 */
    public void drawArrow() {

        gl.glBegin(gl.GL_QUADS);
        gl.glVertex2f(0, -0.05f);
        gl.glVertex2f(0, 0.05f);
        gl.glVertex2f(0.5f, 0.05f);
        gl.glVertex2f(0.5f, -0.05f);
        gl.glEnd();

        gl.glBegin(gl.GL_TRIANGLES);
        gl.glVertex2f(0.5f, 0.2f);
        gl.glVertex2f(1, 0);
        gl.glVertex2f(0.5f, -0.2f);
        gl.glEnd();
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw one vector arrow.
 */
    public void drawNarrowArrow() {

        gl.glBegin(gl.GL_QUADS);
        gl.glVertex2f(0, -0.025f);
        gl.glVertex2f(0, 0.025f);
        gl.glVertex2f(0.5f, 0.025f);
        gl.glVertex2f(0.5f, -0.025f);
        gl.glEnd();

        gl.glBegin(gl.GL_TRIANGLES);
        gl.glVertex2f(0.5f, 0.1f);
        gl.glVertex2f(1, 0);
        gl.glVertex2f(0.5f, -0.1f);
        gl.glEnd();
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw a bar.
 */
    public void drawBar() {

        gl.glBegin(gl.GL_QUADS);
        gl.glVertex2f(0, -0.1f);
        gl.glVertex2f(0, 0.1f);
        gl.glVertex2f(1, 0.1f);
        gl.glVertex2f(1, -0.1f);
        gl.glEnd();
    }

//------------------------------------------------------------------------------
/**
 * Selects or deselects all points within the zoom window.
 * @param selectRectMin	The lower left corner of the rectangle describing the
 *                      selected area.
 * @param selectRectMax	The upper right corner of the rectangle describing the
 *                      selected area.
 * @param bAdding	True if selecting points, false if deselecting.
 * @param bExclusive	True if only one curve should be selected or,
 *						if multiple curves are already selected, only
 *						the selected curves can have more points selected.
 * @param bDeselectFirst	True if old selection is deselected before selecting
 *							new points.
 * @return	An unmodifiable list of selected data series.
 */
    public List selectWindow(Point2D.Double selectRectMin,
        Point2D.Double selectRectMax, boolean bAdding,
        boolean bExclusive, boolean bDeselectFirst) {

        List list = new ArrayList();
        DoubleDataSeries locSeries = null;

        try {
            DoubleDataSeries xSeries = m_seriesArray[m_seriesNums[1]];
            DoubleDataSeries ySeries = m_seriesArray[m_seriesNums[2]];
            if (xSeries == ySeries)
                locSeries = xSeries;

            if (locSeries == null)
                return null;

            boolean bXlocked =
                locSeries.getDimension(m_dimNums[1]).getDimensionLocked();
            boolean bYlocked =
                locSeries.getDimension(m_dimNums[2]).getDimensionLocked();
            if (!bXlocked || !bYlocked) {
                int numDims = locSeries.getNumDimensions();
                double[] min = new double[numDims];
                double[] max = new double[numDims];
                for (int i = 0; i < numDims; i++) {
                    min[i] = Double.NEGATIVE_INFINITY;
                    max[i] = Double.POSITIVE_INFINITY;
                }

                min[m_dimNums[1]] = (double)selectRectMin.x;
                max[m_dimNums[1]] = (double)selectRectMax.x;
                min[m_dimNums[2]] = (double)selectRectMin.y;
                max[m_dimNums[2]] = (double)selectRectMax.y;

                if (bDeselectFirst) locSeries.deselectAll();
                locSeries.selectArea(min, max, bAdding);
                if (locSeries.getNumSelected() > 0)
                    list.add(locSeries);
            }
        }
        catch (Exception e) {
            return null;
        }

        return Collections.unmodifiableList(list);
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        // Check if the object is a graphical element that needs to be redrawn.
        if (gro instanceof TextMarkup) {
            m_bNeedsRedraw = true;
        }

        super.update(gro, arg);
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s = null;
        if (m_bShow)
            s = "GL Vector: ";
        else
            s = "(GL Vector): ";

        if (m_errorString != null)
            s += m_errorString;
        else
            s += m_title;

        return s;
    }
}






