/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-03-27
 * Date started:    1998-10-20
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.visual.*;
import java.awt.Rectangle;
import java.awt.geom.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import java.nio.ByteBuffer;

//==============================================================================
/**
 * A group of graphs layed out on a page.  Draws itself to an OpenGL drawing
 * area.
 */
public class PageOfGlGraphs extends gov.usgs.gr.visual.PageOfGraphs
	implements GlElement {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /** The main OpenGL class with all the gl drawing methods. */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;
    /**
     * Describes the window that the user draws for zooming on the last
     * frame.  This rectangle must be erased before drawing the new one.
     */
    protected Rectangle m_oldZoomRect = null;
    protected ByteBuffer m_pixels = ByteBuffer.allocate(0);
    protected long m_lastDrawTime = 0;

    protected int[] m_imageBufferIds = {0};

//------------------------------------------------------------------------------
/**
 * Create a blank page for graphs.
 */
    public PageOfGlGraphs () {

        setDefaults("gov.usgs.gr.visual.gl.PageOfGlGraphs");
    }

//------------------------------------------------------------------------------
/**
 * Create a page of graphs cloned from an existing one.   Data is shared.
 */
    public PageOfGlGraphs (PageOfGraphs page) {
    	super (page);
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param gld   The OpenGL GLAutoDrawable component.
 * @param gl    The OpenGL object containing the drawing commands.
 * @param glu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {

    	if (!m_bDrawing) {
            gld = newgld;
        	gl = newgl;
        	glu = newglu;
        }
        else {
        }
    }

//------------------------------------------------------------------------------
/**
 * Clears the drawing area and redraws the contents.
 * <p><b> Pre Conditions </b><br>
 *      setOpenGL() should be called before this method.
 */
    public synchronized void draw() {

        // drawGL();
/*
    	// Check if this method is being executed by this object's own thread.
        // If it isn't, split a new thread off to do the work.
    	if (!m_bDrawing) {
        	start();
            return;
        }
*/
	}

//------------------------------------------------------------------------------
/**
 */
    public void drawGL() {

        if (m_bDrawing || m_bDisableRedraw)
            return;
///
        synchronized (this) {
            gl.glEnable(gl.GL_SCISSOR_TEST);
            gl.glScissor(m_content.x, m_content.y,
                         m_content.width, m_content.height);
            gl.glViewport(m_content.x, m_content.y,
                          m_content.width, m_content.height);

            gl.glMatrixMode(gl.GL_MODELVIEW);

            int size = m_content.width * m_content.height * 4;

if (m_pixels == null) {
    int hey = 0;
}
            if (m_bNeedsRedraw || m_pixels.limit() != size
                || (m_lastDrawTime <= 5000)) {
                long startTime = System.currentTimeMillis();

                // Next four lines work by themselves in this method after [if(draw)->return].
                m_bDrawing = true;
                m_bNeedsRedraw = false;
                try {
                    drawFromThread();
                }
                catch (Exception e) {
                    e.printStackTrace();
                }
                m_bDrawing = false;

                long endTime = System.currentTimeMillis();

                m_lastDrawTime = endTime - startTime;

                if (m_lastDrawTime > 5000) {
                    if (m_pixels.limit() != size) {
                        // Create a pixel buffer of the size necessary to store
                        // the pixel data in the window.
                        gl.glDeleteBuffers(1, m_imageBufferIds, 0);
                        gl.glGenBuffers(1, m_imageBufferIds, 0);
                        gl.glBufferData(gl.GL_PIXEL_PACK_BUFFER_ARB, size, null,
                            gl.GL_STREAM_READ);
                    }


                    // Copy the pixel data from the back drawing buffer to the
                    // storage buffer.
                    gl.glBindBuffer(gl.GL_PIXEL_PACK_BUFFER_ARB,
                        m_imageBufferIds[0]);
                    gl.glReadBuffer(gl.GL_BACK);
                    gl.glReadPixels(0, 0, m_content.width, m_content.height,
                                    gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, 0);

                    m_pixels = gl.glMapBuffer(gl.GL_PIXEL_PACK_BUFFER_ARB,
                                 gl.GL_READ_ONLY);

                    if (m_pixels == null)
                        m_pixels = ByteBuffer.allocate(0);

                    // Unmap the image buffers once copying is finished.
                    gl.glBindBuffer(gl.GL_PIXEL_PACK_BUFFER_ARB,
                        m_imageBufferIds[0]);
                    gl.glUnmapBuffer(gl.GL_PIXEL_PACK_BUFFER_ARB);
                }
            }
            else {
                // Copy the pixel data to the back drawing buffer from the
                // storage buffer.
                gl.glDrawBuffer(gl.GL_BACK);
                gl.glRasterPos2i(0, 0);
                gl.glDrawPixels(m_content.width, m_content.height,
                                gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, m_pixels);
            }

            gl.glMatrixMode(gl.GL_PROJECTION);

            if (m_bNeedsRedraw) {
                notifyObservers();
            }
        }
	}

//------------------------------------------------------------------------------
/**
 * Clears the drawing area and redraws the contents.
 * <p><b> Pre Conditions </b><br>
 *      setOpenGL() should be called before this method.
 */
    public synchronized void drawFromThread() {

        if (m_bNeedsReformat) m_formatter.reformat();

        // Don't bother to draw anything if the drawing area is too small.
        if (m_display.height <= 0) return;

        // Draw the rubber band if a region is being selected.
        if (m_zoomRect.width > 0) {
        	gl.glEnable(gl.GL_COLOR_LOGIC_OP);
            gl.glLogicOp(gl.GL_XOR);

            drawZoomWindow();
            m_oldZoomRect = new Rectangle(m_zoomRect);
            gl.glDisable(gl.GL_COLOR_LOGIC_OP);
        	return;
        }
        else {
        	m_oldZoomRect = null;
        }

        // Clear the window.
        gl.glClearColor(m_bgColor[0], m_bgColor[1], m_bgColor[2], 0);
	    gl.glClear(gl.GL_COLOR_BUFFER_BIT);

		// Calculate the ratio between screen hieght
        float displayScale = 1 / (float)m_showList.size();

        // Set up the viewing transformation for this section of the page.
        gl.glMatrixMode(gl.GL_PROJECTION);
        gl.glLoadIdentity();
        glu.gluOrtho2D(0, m_display.width, 0, m_display.height);
        gl.glTranslated(m_content.x, m_content.y, 0);

        List aspectZoomList = getXZooms(m_showList);
        aspectZoomList.addAll(getYZooms(m_showList));

        // Let each graph in the panel draw itself using the OpenGL
        // command objects.
        Iterator iter = m_graphList.iterator();
        while(iter.hasNext()) {
            GlGraph graph = (GlGraph) iter.next();
            if (graph.getShow() && graph != m_movingGraph) {
                if (graph.getAspect1to1()) {
                    Zoom xZoom = graph.getXZoom();
                    Zoom yZoom = graph.getXZoom();
                    if (aspectZoomList.contains(xZoom)) {
                        if (aspectZoomList.remove(yZoom)) {
                            aspectZoomList.remove(xZoom);
                            graph.makeAspect1to1();
                        }
                    }
                }

                graph.setOpenGL(gld, gl, glu);
                graph.drawGL();
            }
        }

        // Draw the moving graph over top of all the others.
        if (m_movingGraph != null) {
        	((GlGraph)m_movingGraph).setOpenGL(gld, gl, glu);
            ((GlElement)m_movingGraph).drawGL();
        }

        // Draw any additional markup objects.
        Rectangle2D.Double outerRect = new Rectangle2D.Double(0, 0,
        	m_display.width, m_display.height);
        Rectangle2D.Double innerRect = new Rectangle2D.Double(m_content.x,
        	m_content.y, m_content.width, m_content.height);
        Rectangle2D.Double innerUser = new Rectangle2D.Double(0,
        	0, m_content.width, m_content.height);
        iter = m_markupList.iterator();
        while (iter.hasNext()) {
            Markup markup = (Markup)iter.next();
            markup.setRects(outerRect, innerRect, innerUser);

            markup.setOpenGL(gld, gl, glu);
            markup.drawGL();
        }
    }

//------------------------------------------------------------------------------
/**
 * Draw a thick yellow frame around the graph.
 */
    protected void drawZoomWindow() {

	    // Save the calling function's GL settings.
	    gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

    	// Draw a dashed cyan box around the zoom area.
        gl.glColor3f(1f, 1f, 1f);
	    gl.glLineWidth(1.0f);
        gl.glEnable(gl.GL_LINE_STIPPLE);
        gl.glLineStipple(1, Gr.TIGHT_DASHED_LINE_PATTERN);

        if (m_oldZoomRect != null) {
        	gl.glBegin(gl.GL_LINE_LOOP); {
		    	gl.glVertex2i(m_oldZoomRect.x, m_oldZoomRect.y);
		    	gl.glVertex2i(m_oldZoomRect.x + m_oldZoomRect.width,
                	m_oldZoomRect.y);
		    	gl.glVertex2i(m_oldZoomRect.x + m_oldZoomRect.width,
            		m_oldZoomRect.y + m_oldZoomRect.height);
		    	gl.glVertex2i(m_oldZoomRect.x,
            		m_oldZoomRect.y + m_oldZoomRect.height);
        	} gl.glEnd();
        }

	    gl.glBegin(gl.GL_LINE_LOOP); {
		    gl.glVertex2i(m_zoomRect.x, m_zoomRect.y);
		    gl.glVertex2i(m_zoomRect.x + m_zoomRect.width, m_zoomRect.y);
		    gl.glVertex2i(m_zoomRect.x + m_zoomRect.width,
            	m_zoomRect.y + m_zoomRect.height);
		    gl.glVertex2i(m_zoomRect.x,
            	m_zoomRect.y + m_zoomRect.height);
		} gl.glEnd();

	    // Restore GL settings.
	    gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 * Create an OpenGL graph to be used in this page.
 */
    protected Graph newGraph() {
    	return new GlGraph();
    }

//------------------------------------------------------------------------------
/**
 * Create an OpenGL graph as a clone of an existing graph.
 * @param graph	A template graph to use in creating this one.
 */
    protected Graph newGraph(Graph graph) {
    	return new GlGraph(graph);
    }
}


