/*
Texure loading example

Below you will find an example for loading a Texture in JOGL.
It was created in the thread
http://www.JavaGaming.org/cgi-bin/JGNetForums/YaBB.cgi?board=jogl;action=display;num=1058674409 by kevglass.

Class TextureLoader
*/

package gov.usgs.gr.visual.gl;

import java.awt.Graphics;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.HashMap;
import java.util.Hashtable;
import javax.imageio.ImageIO;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.glu.GLU;

/**
 * A utility class to load textures for JOGL
 *
 * @author Kevin Glass
 */
public class TextureLoader {
    private HashMap table = new HashMap();
    private GL gl;
    private GLU glu;

    private ColorModel glAlphaColorModel;
    private ColorModel glColorModel;

    public TextureLoader() {
    
    }
    public void initializeDrawable(GLAutoDrawable drawable) {
        
        if (gl != null)
            return;
        
        this.gl = drawable.getGL();
        this.glu = new GLU();

        glAlphaColorModel = new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_sRGB),
                                            new int[] {8,8,8,8},
                                            true,
                                            false,
                                            ComponentColorModel.TRANSLUCENT,
                                            DataBuffer.TYPE_BYTE);

        glColorModel = new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_sRGB),
                                            new int[] {8,8,8,0},
                                            false,
                                            false,
                                            ComponentColorModel.OPAQUE,
                                            DataBuffer.TYPE_BYTE);
    }

    private int createTextureID()
    {
       int[] tmp = new int[1];
       gl.glGenTextures(1, tmp, 0);
       return tmp[0];
    }

    public Texture getTexture(String name,String resourceName,
        BufferedImage bufferedImage, ByteBuffer byteBuffer) {
        Texture tex = (Texture) table.get(resourceName);

        if (tex != null) {
            return tex;
        }

        tex = getTexture(name,resourceName, bufferedImage, byteBuffer,
                         GL.GL_TEXTURE_2D, // target
                         GL.GL_RGBA,     // dst pixel format
                         GL.GL_LINEAR, // min filter (unused)
                         GL.GL_LINEAR, // mag filter (unused)
                         true,  // wrap?
                         false); // mipmap?

        table.put(resourceName,tex);

        return tex;
    }

    public Texture getTexture(String name, String resourceName,
                              BufferedImage bufferedImage, ByteBuffer textureBuffer,
                              int target,
                              int dstPixelFormat,
                              int minFilter,
                              int magFilter,
                              boolean wrap,
                              boolean mipmapped)
    {
        int srcPixelFormat = 0;

        // create the texture ID for this texture
        int textureID = createTextureID();
        System.out.println("Loading texture "+resourceName+" to ID: "+textureID);
        Texture texture = new Texture(resourceName,target,textureID);

        int texWidth = get2Fold(bufferedImage.getWidth());
        int texHeight = get2Fold(bufferedImage.getHeight());

        texture.setBufferedImage(bufferedImage);
        texture.setTextureHeight(texHeight);
        texture.setTextureWidth(texWidth);

        // bind this texture
        gl.glBindTexture(target, textureID);

        if (bufferedImage.getColorModel().hasAlpha()) {
            srcPixelFormat = GL.GL_RGBA;
        } else {
            srcPixelFormat = GL.GL_RGB;
        }

        int wrapMode = wrap ? GL.GL_REPEAT : GL.GL_CLAMP;

        if (target == GL.GL_TEXTURE_2D)
        {
              gl.glTexParameteri(target, GL.GL_TEXTURE_WRAP_S, wrapMode);
              gl.glTexParameteri(target, GL.GL_TEXTURE_WRAP_T, wrapMode);
              gl.glTexParameteri(target, GL.GL_TEXTURE_MIN_FILTER, minFilter);
              gl.glTexParameteri(target, GL.GL_TEXTURE_MAG_FILTER, magFilter);
        }

        // create either a series of mipmaps of a single texture image based on what's loaded
        if (mipmapped)
        {
            textureBuffer.rewind();
            glu.gluBuild2DMipmaps(target,
                                  dstPixelFormat,
                                texWidth,
                          texHeight,
                                  srcPixelFormat,
                                  GL.GL_UNSIGNED_BYTE,
                                  textureBuffer);
        }
        else
        {
            textureBuffer.rewind();
            gl.glTexImage2D(target,
                          0,
                          dstPixelFormat,
                          texWidth,
                       texHeight,
                          0,
                          srcPixelFormat,
                          GL.GL_UNSIGNED_BYTE,
                          textureBuffer );
        }

        return texture;
    }

    public ByteBuffer createByteBuffer(BufferedImage bufferedImage) {

        // convert that image into a byte buffer of texture data
        int texWidth = get2Fold(bufferedImage.getWidth());
        int texHeight = get2Fold(bufferedImage.getHeight());
        ByteBuffer textureBuffer = convertImageData(bufferedImage);

        return textureBuffer;
    }

   private int get2Fold(int fold) {
      int ret = 2;
      while (ret < fold) {
         ret *= 2;
      }
      return ret;
   }

    public ByteBuffer convertImageData(BufferedImage bufferedImage)
    {
        ByteBuffer imageBuffer = null;
        WritableRaster raster;
        BufferedImage texImage;

        int texWidth = get2Fold(bufferedImage.getWidth());
        int texHeight = get2Fold(bufferedImage.getHeight());

        if (bufferedImage.getColorModel().hasAlpha()) {
            raster = Raster.createInterleavedRaster(DataBuffer.TYPE_BYTE,texWidth,texHeight,4,null);
            texImage = new BufferedImage(glAlphaColorModel,raster,false,new Hashtable());
        } else {
            raster = Raster.createInterleavedRaster(DataBuffer.TYPE_BYTE,texWidth,texHeight,3,null);
            texImage = new BufferedImage(glColorModel,raster,false,new Hashtable());
        }

// Added by jmd 2006-01-19 for non-power-of-2 image dimensions.
        java.awt.Image scaledImage = bufferedImage.getScaledInstance(texWidth, texHeight,
            java.awt.Image.SCALE_SMOOTH);

        Graphics g = texImage.getGraphics();
        g.drawImage(scaledImage,0,0,null);

        byte[] data = ((DataBufferByte) texImage.getRaster().getDataBuffer()).getData();

        imageBuffer = ByteBuffer.allocateDirect(data.length);
        imageBuffer.order(ByteOrder.nativeOrder());
        imageBuffer.put(data, 0, data.length);

        return imageBuffer;
    }

    public BufferedImage loadImage(String ref) throws IOException
    {
      File file = new File(ref);

        if (!file.isFile()) {
            throw new IOException("Cannot find: "+ref);
        }

        BufferedImage bufferedImage = ImageIO.read(new BufferedInputStream(
           new FileInputStream(file))
        );

        return bufferedImage;
    }
}
