/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-04-06
 * Date started:    2000-08-03
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.gui;

import gov.usgs.gr.*;
//import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;
//import org.w3c.dom.*;

//==============================================================================
/**
 * An object that has properties and contains other objects in a hierarchy.
 * It can be displayed in a tree or written as XML.
 */
public class GuiOptions extends gov.usgs.gr.GrObject {

    static String XML_ELEMENT_NAME = "gui";
	protected String m_lookAndFeel = null;
    /** Maximum number of undos that will be saved. */
    protected int m_undoLevels = 20;
    /** Possibilities to choose from for the number of levels of undo. */
    int[] m_undoPresets = {0, 5, 10, 20, 40, 80};
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GuiOptions.gif";

//------------------------------------------------------------------------------
/**
 * Construct an empty object.
 */
    public GuiOptions () {

        setDefaults("gov.usgs.gr.visual.gui.GuiOptions");
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        GrProperty p = new GrProperty(m_lookAndFeel, "GUI Look And Feel");
        JComboBox jcb = new JComboBox();
        UIManager.LookAndFeelInfo[] looks =
        	UIManager.getInstalledLookAndFeels();
        for (int i = 0; i < looks.length; i++) {
        	String name = looks[i].getClassName();
        	jcb.addItem(name);
            if (name.equals(m_lookAndFeel)) jcb.setSelectedIndex(i);
        }

        p.component = jcb;
		props.put("lookAndFeel", p);

        // Levels of undo.
        p = new GrProperty("" + m_undoLevels, "Levels of Undo");
        JComboBox undoComboBox = new JComboBox();
        for (int i = 0; i < m_undoPresets.length; i++) {
        	undoComboBox.addItem("" + m_undoPresets[i]);
        }
        undoComboBox.addItem("Unlimited");

        // Set the number of levels to one of the preset increments.
        if (m_undoLevels < 0) {
            undoComboBox.setSelectedIndex(0);
        }
        if (m_undoLevels > 80) {
            undoComboBox.setSelectedIndex(m_undoPresets.length);
        }

        // Select the correct undo combo box item.
        for (int i = 0; i < (m_undoPresets.length - 1); i++) {
        	if (m_undoLevels > m_undoPresets[i]
            	&& m_undoLevels <= m_undoPresets[i + 1]) {

            	undoComboBox.setSelectedIndex(i+1);
            }
        }

        p.component = undoComboBox;
		props.put("undoLevels", p);

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("lookAndFeel")) {
                	m_lookAndFeel = (String)p.value;
                }
            	if (key.equals("undoLevels")) {
                	if (p.component instanceof JComboBox) {
                    	// Read from the combo box, if it is present.
    					int selectedIndex =
                        	((JComboBox)p.component).getSelectedIndex();
        				if (selectedIndex == m_undoPresets.length)
        					m_undoLevels = Integer.MAX_VALUE;
        				else
        					m_undoLevels = m_undoPresets[selectedIndex];
                    }
                    else {
                		m_undoLevels = Integer.parseInt((String)p.value);
                    }
                }
            }
            if (obj instanceof String) {
            	String value = (String)obj;
                setProperty(key, value);
			}
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

    	if (key.equals("lookAndFeel")) {
        	m_lookAndFeel = value;
        }
        if (key.equals("undoLevels")) {
        	m_undoLevels = Integer.parseInt(value);
        }
    }

//------------------------------------------------------------------------------
/**
 * Returns an XML name of this element.
 */
    public String getXmlElementName() {
    	return XML_ELEMENT_NAME;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "GUI";
        return s;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return Then number of undo levels.
 */
    public int getUndoLevels() {
    	return m_undoLevels;
    }
}




