/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2002-01-22
 * Date started:    1999-01-16
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.j2d;

import java.awt.*;
import java.awt.geom.*;
import java.awt.font.TextLayout;
import java.text.*;
import gov.usgs.gr.Gr;
import gov.usgs.gr.visual.*;

//==============================================================================
/**
 * A generic axis with ticks and labels for an X-Y graph.
*/
public abstract class J2dAxis extends gov.usgs.gr.visual.Axis
	implements J2dElement {

	Graphics2D j2d;
    AffineTransform m_at;

//------------------------------------------------------------------------------
/**
 * Empty constructor.
 */
    public J2dAxis () {

        setDefaults("gov.usgs.gr.visual.j2d.J2dAxis");
    }

//------------------------------------------------------------------------------
/**
 * Create a new axis and associates it with a set of limits for itself and
 * its perpendicular counterpart.
 * <p><b> Pre Conditions </b><br>
 *      Axis limits exist for this axis and its perpendicular counterpart.
 */
    public J2dAxis (Zoom zoom, Zoom opposingZoom) {

        setDefaults("gov.usgs.gr.visual.j2d.J2dAxis");

        m_zoom = zoom;
        m_opposingZoom = opposingZoom;
    }

//------------------------------------------------------------------------------
/**
 * Create a new axis by copying an existing one.
 * @param axis	The axis to be copied.
 */
    public J2dAxis (Axis axis) {
    	super(axis);
    }

//------------------------------------------------------------------------------
/**
 * Sets the Graphics2D object to use for drawing.
 * @param g	The graphics context.
 */
    public void setGraphics2D(Graphics2D g) {
		j2d = g;
    }

//------------------------------------------------------------------------------
/**
 * Sets the affine transform to be used for drawing objects.
 * @param at The transform that should be used.
 */
    public void setTransform(AffineTransform at) {
		m_at = at;
    }

//------------------------------------------------------------------------------
/**
 * Draw all the elements that make up a time series graph.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context and the graph has been initialized.
 */
    public void drawJ2d(){

        // Get the latest changes from the axis limits objects.
        update();
        if (m_majorTickInc <= 0) return;

        // Calculate the start for the major tick marks, grid lines, and labels.
        double offset = (double)Math.abs(m_min - m_zoom.getOuterMin());
        m_majorStart = (double)Math.ceil(offset / (double)m_majorTickInc)
            * m_majorTickInc + m_zoom.getOuterMin();

        // Calculate the length of a tick for this axis.
        double oppLen = (m_oppMax - m_oppMin);
        m_majorTickLen = 0.005 * oppLen * m_displayScale * m_tickSize;

	    // Save the calling function's GL settings.

        // Set the line style and color.
	    j2d.setColor(new Color(1, 1, 0));

        // Draw the various parts of the  graph.
        if (!m_bQuickDraw) {
        	if (m_bShowGridLines) drawGridLines();
        	if (m_bShowMinorTicks) drawMinorTicks();
        }
        if (m_bShowMajorTicks) drawMajorTicks();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Draw a light line across the graph for each major tick.  Line and color
 * attributes are saved and then restored when drawing is complete.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context.
 */
    protected void drawGridLines() {

	    double x;

		// Save the old stroke, and change to dotted.
		Stroke oldStroke = j2d.getStroke();
        float[] dotted = {2.2f};
        BasicStroke thickStroke = new BasicStroke(1, BasicStroke.CAP_SQUARE,
        	BasicStroke.JOIN_MITER, 10, dotted, 0);
        j2d.setStroke(thickStroke);

	    // Set the color to medium gray.
	    j2d.setColor(new Color(0.5f, 0.5f, 0.5f));

        // Vertical grid lines...
        for (x = m_majorStart; x < m_max; x += m_majorTickInc) {
        	j2d.draw(m_at.createTransformedShape(
        		new Line2D.Double(x, m_oppMin, x, m_oppMax)));
        }

	    // Restore line settings.
		j2d.setStroke(oldStroke);

    }

//------------------------------------------------------------------------------
/**
 * Draw one or more small tick marks between the large tick marks.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context, and the color and line width
 *      have been set.
 */
    protected void drawMinorTicks() {
	    double x, y;

        // Calculate the location of the first tick mark.
		int tickNum = 0;
        int minsPerMax = (int)Math.round(m_majorTickInc / m_minorTickInc);
        double start = m_majorStart - m_majorTickInc;
        while (start <= m_min) {
        	start += m_minorTickInc;
            tickNum++;
        }
        if (tickNum > minsPerMax) tickNum -= minsPerMax;

        double len = m_majorTickLen * m_minorTickProportion;

	    // Set the color to yellow.
        j2d.setColor(new Color(1, 1, 0));

        // Loop through and draw each tick...
        for (x = start; x < m_max; x += m_minorTickInc) {
        	if (tickNum >= minsPerMax) {
            	// Skip drawing a tick here because there is a major tick.
                tickNum = 1;
            }
            else {
                    // Draw a minor tick on both sides of the graph.
        		j2d.draw(m_at.createTransformedShape(
        			new Line2D.Double(x, m_oppMin, x, m_oppMin + len)));
        		j2d.draw(m_at.createTransformedShape(
        			new Line2D.Double(x, m_oppMax, x, m_oppMax - len)));
                tickNum++;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Draw one or more large tick marks along the axis.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context, and the color and line width
 *      have been set.
 */
    protected void drawMajorTicks() {

	    double x;

        // Loop through and draw each tick...
        double len = m_majorTickLen;

	    j2d.setColor(new Color(1, 1, 0));

        for (x = m_majorStart; x < m_max; x += m_majorTickInc) {
        	j2d.draw(m_at.createTransformedShape(
        		new Line2D.Double(x, m_oppMin, x, m_oppMin + len)));
        	j2d.draw(m_at.createTransformedShape(
        		new Line2D.Double(x, m_oppMax, x, m_oppMax - len)));
        }
    }

//------------------------------------------------------------------------------
/**
 * Draw line on both sides of the graph to represent the axis.  Saves
 * existing line attributes and restores them after drawing.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context and the color has been set.
 */
    protected void drawAxisLine() {


	    // Save the calling function's GL settings.
//	    gl.pushAttrib(GL.GL_LINE_BIT);

//	    gl.lineWidth(2.0);

        j2d.draw(m_at.createTransformedShape(
        	new Line2D.Double(m_min, m_oppMin, m_max, m_oppMin)));
        j2d.draw(m_at.createTransformedShape(
        	new Line2D.Double(m_min, m_oppMax, m_max, m_oppMax)));
    }

//------------------------------------------------------------------------------
/**
 * Draws labels in magenta beside each major tick mark.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context.
 */
    public void drawText(double xOrig, double yOrig, double axisDisplayWidth,
    	Font font) {


        double x;

        if (!m_bShowText) return;
        if (m_majorTickInc <= 0) return;

		AffineTransform oldAt = j2d.getTransform();
        j2d.translate(xOrig, 0.8 * yOrig);
		j2d.scale(1, -1);
        double displayScale = axisDisplayWidth / (m_max - m_min);

        NumberFormat form = new DecimalFormat(m_zoom.getFormatString());

        // Loop through and draw each label.
        for (x = m_majorStart; (x - m_max) <= (0.01 * m_majorTickInc);
        	 x += m_majorTickInc) {

            if (Math.abs(x - 0) <= (0.01 * m_majorTickInc)) x = 0;
			TextLayout tl = new TextLayout(form.format(x), font,
            	j2d.getFontRenderContext());
            tl.draw(j2d, -0.5f * tl.getVisibleAdvance()
            	+ (float)((x - m_min) * displayScale), tl.getAscent());
        }

        // Restore the previous position.
        j2d.setTransform(oldAt);
    }
}



