/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-05-26
 * Date started:    1999-01-16
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.j2d;

import java.util.*;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;

//==============================================================================
/**
 * A single strip of data points that can draw itself.
 */
public class J2dCurve extends gov.usgs.gr.visual.Curve implements J2dElement {

	Graphics2D j2d;
    AffineTransform m_at;
    float m_xRes = 1, m_yRes = 1;
    ArrayList m_parentPathList = new ArrayList();
    ArrayList m_parentPointList = new ArrayList();
    float m_xRangeMin;
    float m_xRangeMax;

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public J2dCurve(){

        setDefaults("gov.usgs.gr.visual.j2d.J2dCurve");
    }

//------------------------------------------------------------------------------
/**
 * Constructor for cloning this curve from an existing one.
 */
    public J2dCurve(Curve curve) {

		super(curve);

        constructDrawList();
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a curve from a data series.
 * @param series    The data that this curve will use.
 */
    public J2dCurve(DoubleDataSeries series) {
	
        super(series);

        constructDrawList();
   }

//------------------------------------------------------------------------------
/**
 * Sets the Graphics2D object to use for drawing.
 * @param g	The graphics context.
 */
    public void setGraphics2D(Graphics2D g) {
		j2d = g;
    }

//------------------------------------------------------------------------------
/**
 * Sets the affine transform to be used for drawing objects.
 * @param at The transform that should be used.
 */
    public void setTransform(AffineTransform at) {
		m_at = at;
    }

//------------------------------------------------------------------------------
/**
 * Should create a list of paths that would be suitable for drawing at
 * various resolutions.
 */
    public void constructDrawList() {

        double xDiffTotal = 0;
        double yDiffTotal = 0;
        double array[] = m_series.getArray();

    	// Create a list of paths that contains every point in the series.
        ArrayList pointList = new ArrayList();
    	m_parentPathList.add(constructPaths(0, 0, pointList));
    	m_parentPointList.add(pointList);

        // Go through the entire series and find the average change in
        // x and y between adjacent points.
        for (int i = m_series.getBegin() + 1; i <= m_series.getEnd(); i++) {
        	// Get the x and y values for the current point.
        	double x = array[i * 2];
            double y = array[i * 2 + 1];
        	double lastX = array[(i - 1) * 2];
            double lastY = array[(i - 1) * 2 + 1];

            xDiffTotal += Math.abs(x - lastX);
            yDiffTotal += Math.abs(y - lastY);
        }

        // Set the intial resolutions to be the average difference.
        int numPoints = m_series.getEnd() - m_series.getBegin() + 1;
        if (numPoints == 0) numPoints = 1;
        double xRes = xDiffTotal / (double)numPoints;
        double yRes = yDiffTotal / (double)numPoints;

        // Add draw lists of varying sizes down to 1 element.
        int currentListSize = ((List)m_parentPathList.get(
            m_parentPathList.size() - 1)).size();
        int lastListSize = currentListSize;
        while (currentListSize > 1) {
        	pointList = new ArrayList();
        	ArrayList newPaths = constructPaths(xRes, yRes, pointList);
        	currentListSize = newPaths.size();
            if (currentListSize < 1) break;

            // Only add the path if it is sufficiently smaller than the
            // previous one.
            if ((lastListSize / currentListSize) > 1.6) {
    			m_parentPathList.add(newPaths);
    			m_parentPointList.add(pointList);
        		lastListSize = currentListSize;
            }

            xRes *= 1.6;
            yRes *= 1.6;
        }
    }

//------------------------------------------------------------------------------
/**
 * Should create the graphics associated with displaying the data series.
 */
    public ArrayList constructPaths(double xRes, double yRes,
        ArrayList pointList) {

        ArrayList paths = new ArrayList();
        double x, y;
        int lastI = 0;
        double lastY;
        int numPlotted = 0;

        GeneralPath path = new GeneralPath();

        double array[] = m_series.getArray();

        // Move to the first point on the path, line to each point after that.
        path.moveTo((float)array[0], (float)array[1]);
        lastY = array[1];

        // Loop through each point in the data series.
        for (int i = m_series.getBegin() + 1; i <= m_series.getEnd(); i++) {
        	// Get the x and y values for the current point.
        	x = array[i*2];
            y = array[i*2+1];

            // Put another point on the course curve if the difference from
            // the last plotted point is enough to be seen at the current
            // resolution.

            // Course curve...
            if (Math.abs(y - lastY) > yRes) {
                // Plot the previous point if it was far enough along in the
                // x direction to be resolved.
            	if (Math.abs(array[(i - 1) * 2] - array[lastI * 2]) >
                	xRes && (i - 1) > lastI) {

            		path.lineTo((float)array[(i - 1) * 2],
                    	(float)array[(i - 1) * 2 + 1]);
                }

                // Add the new point to the course path and update this
                // point to be the last plotted point.
            	path.lineTo((float)x, (float)y);
            	lastI = i;
                lastY = y;
                if (numPlotted == 0)
                	pointList.add(new Point2D.Double(x, y));
                numPlotted++;
            }

            // Break the curve up into groups of 100.
            if (numPlotted > 100) {
            	paths.add(path);
                path = new GeneralPath();
                path.moveTo((float)x, (float)y);
            	numPlotted = 0;
                pointList.add(new Point2D.Double(x, y));
//                System.out.print("-seg-");
            }
        }

        // Finish off the course path if it didn't go to the last point.
        // Course curve...
        if (numPlotted > 0) {
        	int i = m_series.getEnd();
        	x = array[i*2];
            y = array[i*2+1];
        	path.lineTo((float)x, (float)y);
            paths.add(path);
            pointList.add(new Point2D.Double(x, y));
//            System.out.print("-seg " + x + ", " + y + "-");
        }

        return paths;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 * <p><b> Pre Conditions </b><br>
 *      The origin has been moved to (0, 0), scaling has occurred, and
 *      line type is solid.  gl has been set to a valie OpenGL context.
 */
    public void drawJ2d() {

        // Before drawing anything, check if the curve is supposed to be shown.
        if (!m_bShow) return;

        synchronized (m_series) {
            float xRange = m_xRangeMax - m_xRangeMin;
            int numPoints = (int)(xRange / m_xRes);
            List pathList = (List)m_parentPathList.get(0);
            List pointList = (List)m_parentPointList.get(0);

            // Pick a courseness scale based on whether quickDraw is true.
            float courseScale;
            if (m_bQuickDraw) courseScale = 4;
            else courseScale = 1;

            // Pick a list of paths with sufficient level of detail.
            int i = 0;
            Iterator iter = m_parentPathList.iterator();
            while (iter.hasNext()) {
                List list = (List)iter.next();
                if ((list.size() * 100 * courseScale) > numPoints) {
                    pathList = list;
                    pointList = (List)m_parentPointList.get(i);
                }
                i++;
            }

            // Draw the paths.
            i = 0;
            iter = pathList.iterator();
            while (iter.hasNext()) {
                GeneralPath path = (GeneralPath)iter.next();

                // Weed out paths that aren't being shown right now.
                Point2D.Float p1 = (Point2D.Float)pointList.get(i + 1);
                Point2D.Float p2 = (Point2D.Float)pointList.get(i);
                if ((p1.x > m_xRangeMin && p1.x < m_xRangeMax)
                    || (p2.x > m_xRangeMin && p2.x < m_xRangeMax)
                    || (p1.x > m_xRangeMax && p2.x < m_xRangeMin)) {

                    // Draw the path.
    //                if ((i % 4) == 0) j2d.setColor(Color.magenta);
    //                else j2d.setColor(Color.yellow);
                    j2d.setColor(Color.green);
                    j2d.draw(m_at.createTransformedShape(path));

                }
                i += 2;
            }

    //        gl.pointSize(5);
            if (m_series.getNumSelected() > 0) {
                // Draw deleted points.
                j2d.setColor(new Color(m_color[0] / 2, m_color[1] / 2,
                    m_color[2] / 2));
    //           	gl.begin(GL.GL_POINTS);
                Iterator iter2 = m_series.getDeletedList().iterator();
                while (iter2.hasNext()) {
                    float xVal = (Float)iter2.next();
                    float yVal = (Float)iter2.next();
                    j2d.draw(m_at.createTransformedShape(new Rectangle2D.Float(
                        xVal - 2, yVal - 2, 5, 5)));
                }
            }
        }

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param bQuickDraw	True if the graph is to be drawn for speed at the
 *						expense of looks and accuracy.
 */
    public void setQuickDraw(boolean bQuickDraw) {

        m_bQuickDraw = bQuickDraw;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param xRes	The smallest number of x-units in user coordinates that can
 *				be resolved in output coordinates.
 * @param yRes	The smallest number of y-units in user coordinates that can
 *				be resolved in output coordinates.
 */
    public void setResolution(float xRes, float yRes) {

        m_xRes = xRes;
        m_yRes = yRes;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param min	The minimum x value to be shown.
 * @param max	The maximum x value to be shown.
 */
    public void setXRange(float min, float max) {

        m_xRangeMin = min;
        m_xRangeMax = max;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject o, Object arg) {

    	constructDrawList();
        super.update(o, arg);
    }
}






