/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    1999-01-16
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.j2d;

import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.font.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;

//==============================================================================
/**
 * An x-y graph with ticks, labels, a grid, and multiple curves drawn inside.
 */
public class J2dGraph extends gov.usgs.gr.visual.Graph implements J2dElement {

	Graphics2D j2d;
    AffineTransform m_at;
    GradientPaint m_bgPaint = new GradientPaint(0f, 0f, new Color(0f, 0f, 0.4f),
    	0, m_display.height, Color.black);
    Rectangle2D m_bgRect = new Rectangle2D.Float(0,0,1,1);

//------------------------------------------------------------------------------
/**
 * Construct a time series graph and have it create its own x-axis zoom.
 */
    public J2dGraph () {

        super();

        setDefaults("gov.usgs.gr.visual.j2d.J2dGraph");
    }

//------------------------------------------------------------------------------
/**
 * Construct a time series graph and use the specified x-axis limits.
 * @param xZoom The x-axis limits object to use for this graph.
 */
    public J2dGraph (Zoom xZoom) {

        super(xZoom);

        setDefaults("gov.usgs.gr.visual.j2d.J2dGraph");
    }

//------------------------------------------------------------------------------
/**
 * Create a graph cloned from an existing one.
 */
    public J2dGraph (Graph graph) {
    	super(graph);
    }

//------------------------------------------------------------------------------
/**
 * Returns a new x axis object.
 */
    protected Axis newXAxis() {
    	return new J2dXAxis();
    }

//------------------------------------------------------------------------------
/**
 * Returns a new y axis object.
 */
    protected Axis newYAxis() {
    	return new J2dYAxis();
    }

//------------------------------------------------------------------------------
/**
 * Returns a new Curve object.
 */
    protected Curve newCurve() {
	
    	return new J2dCurve();
    }

//------------------------------------------------------------------------------
/**
 * Returns a new Curve object based on the data series parameter.
 */
    protected Curve newCurve(DoubleDataSeries series) {
	
    	return new J2dCurve(series);
    }

//------------------------------------------------------------------------------
/**
 * Returns a new Curve object copied from an existing curve.
 */
    protected Curve newCurve(Curve curve) {
	
    	return new J2dCurve(curve);
    }

//------------------------------------------------------------------------------
/**
 * Makes new copies of the axis objects associated with graph.
 * @param graph	The graph with the template axes.
 */
    protected void copyAxes(Graph graph) {

		m_xAxis = new J2dXAxis(graph.getXAxis());
		m_yAxis = new J2dYAxis(graph.getYAxis());
        m_xAxis.setOpposingZoom(m_yAxis.getZoom());
        m_yAxis.setOpposingZoom(m_xAxis.getZoom());

        m_xZoom = (Zoom)m_xAxis.getZoom();
        m_yZoom = (Zoom)m_yAxis.getZoom();
    }

//------------------------------------------------------------------------------
/**
 * Sets the Graphics2D object to use for drawing.
 * @param g	The graphics context.
 */
    public void setGraphics2D(Graphics2D g) {
		j2d = g;
    }

//------------------------------------------------------------------------------
/**
 * Draw this graph with all its parts on the screen.
 * <p><b> Pre Conditions </b><br>
 *      gl has been set to a valid OpenGL context.
 */
    public void drawJ2d() {

    	// Move to the correct position on the display area for this graph.
        j2d.translate((float)m_display.x, (float)m_display.y);

        // Scale down to the area inside the axes.
        m_at = AffineTransform.getScaleInstance(m_display.width, m_display.height);

        // If this graph is selected, mark it on the screen.
        if (m_bSelected) drawAsSelected();

        m_at.translate(m_frame.x, m_frame.y);
        m_at.scale(m_frame.width, m_frame.height);

        // Paint a background for the curves.
        if (!m_bQuickDraw) drawBackground();

        // Change the transformation so the curves will fit inside the graph
        // using their coordinates.
        AffineTransform graphTransform = (AffineTransform)m_at.clone();
        m_at.scale(m_xZoom.getScale(), m_yZoom.getScale());
        m_at.translate(-m_xZoom.getViewMin(), -m_yZoom.getViewMin());

        // Set the scale for the axes so the x and y ticks appear equally
        // long on the screen.
        m_xAxis.setDisplayScale(m_xTickScale);
        m_yAxis.setDisplayScale(m_yTickScale);

    	// Draw the axes and grid.
        ((J2dAxis)m_xAxis).setGraphics2D(j2d);
        ((J2dAxis)m_xAxis).setTransform(m_at);
        ((J2dXAxis)m_xAxis).drawJ2d();
        ((J2dAxis)m_yAxis).setGraphics2D(j2d);
        ((J2dAxis)m_yAxis).setTransform(m_at);
        ((J2dYAxis)m_yAxis).drawJ2d();


        // Save the current unclipped state, and then turn on clipping.
        Rectangle2D.Float oldClip = (Rectangle2D.Float)j2d.getClip();
        j2d.setClip(graphTransform.createTransformedShape(m_bgRect));

        // Draw the curves inside the graph.
        Iterator iter = m_curves.iterator();
        while (iter.hasNext()) {
            J2dCurve curve = (J2dCurve)iter.next();

            curve.setResolution((float)(m_xZoom.getViewLength() /
                (m_display.width * m_frame.width)),
                (float)(m_yZoom.getViewLength() /
                (m_display.height) * m_frame.height));
            curve.setXRange((float)m_xZoom.getViewMin(),
                (float)m_xZoom.getViewMax());
            curve.setGraphics2D(j2d);
            curve.setTransform(m_at);
            curve.drawJ2d();
        }

        // Turn clipping back off.
        j2d.setClip(oldClip);

        // Draw a frame around the plot.
        m_at = graphTransform;
        drawFrame();

        // Pass in x, y, width, and height in screen coordinates.
        j2d.setColor(Color.magenta);
        int fontSize = (int)(0.06 * (float)m_display.height * m_frame.height);
        if (fontSize > 24) fontSize = 24;
        Font font = new Font("serif", Font.PLAIN, fontSize);
        ((J2dXAxis)m_xAxis).drawText(m_frame.x * m_display.width,
        	m_frame.y * m_display.height, m_frame.width * m_display.width, font);
        ((J2dYAxis)m_yAxis).drawText(m_frame.x * m_display.width,
        	m_frame.y * m_display.height, m_frame.height * m_display.height, font);

        // Add text on top of the graph.
        if (m_bShowTitle) drawText(font);

    	// Move back to the display origin.
        j2d.translate((float)-m_display.x, (float)-m_display.y);

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Draw a background behind the curves and axes.
 * <p><b> Pre Conditions </b><br>
 *      The viewing transformation has been set up so that the graphing
 *		area is between 0 and 1 on both axes.
 * <p><b> Post Conditions </b><br>
 *      The current drawing color has been modified.
 */
    protected void drawBackground() {

	    // Draw black to blue fade rectangle to cover any objects in background.
        Color deepBlue = new Color(0f, 0f, 0.4f);
        j2d.setPaint(new GradientPaint(0, 0, deepBlue, 0, m_display.height,
        	Color.black));
        j2d.fill(m_at.createTransformedShape(m_bgRect));
    }


//------------------------------------------------------------------------------
/**
 * Draw a thick yellow frame around the graph.
 * <p><b> Pre Conditions </b><br>
 *      The viewing transformation has been set up so that the graphing
 *		area is between 0 and 1 on both axes.
 */
    protected void drawFrame() {

    	// Draw a thick yellow frame around the graph.
        j2d.setColor(Color.yellow);
		Stroke oldStroke = j2d.getStroke();
        BasicStroke thickStroke = new BasicStroke(2);
        j2d.setStroke(thickStroke);

        j2d.draw(m_at.createTransformedShape(m_bgRect));
        j2d.setStroke(oldStroke);
    }

//------------------------------------------------------------------------------
/**
 * Draws title in magenta above the graph.
 * <p><b> Pre Conditions </b><br>
 *      j2d has been set.
 */
    protected void drawText(Font font) {

		AffineTransform oldAt = j2d.getTransform();
        j2d.translate(m_frame.x * m_display.width,
        	(m_frame.y + m_frame.height) * m_display.height);
        j2d.scale(1, -1);
        TextLayout tl = new TextLayout(m_title, font,
        	j2d.getFontRenderContext());
            tl.draw(j2d, 0, -2 * tl.getDescent());
        j2d.setTransform(oldAt);
    }

//------------------------------------------------------------------------------
/**
 * Mark this graph on the screen as being selected.
 * <p><b> Pre Conditions </b><br>
 *      Viewing transformation has been set so this graph's section of the
 *		page is between 0 and 1 in both x and y directions.
 * <p><b> Post Conditions </b><br>
 *      The current drawing color has been changed.
 */
    private void drawAsSelected() {

    	// Draw green corners at the left of the graph's area.
        j2d.setColor(new Color(0, 1, 0));

        j2d.draw(m_at.createTransformedShape(
        	new Line2D.Float(0f, 0.0001f, 0.02f, 0.0001f)));
        j2d.draw(m_at.createTransformedShape(
        	new Line2D.Float(0f, 0.0001f, 0f, 0.05f)));
        j2d.draw(m_at.createTransformedShape(
        	new Line2D.Float(0f, 1f, 0.02f, 1f)));
        j2d.draw(m_at.createTransformedShape(
        	new Line2D.Float(0f, 1f, 0f, 0.95f)));
    }

//------------------------------------------------------------------------------
/**
 * Define clipping planes around the graph.
 * <p><b> Pre Conditions </b><br>
 *      The viewing transformation has been set up so that the graphing
 *		area is in graph coordinates.
 */
    protected void setupClipping() {
	}

//------------------------------------------------------------------------------
/**
 * Disable all clipping planes.
 * @param gl    The OpenGL object containing the drawing commands.
 */
    protected void disableClipping() {
	}
}




