/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2005-11-04
 * Date started:    1999-01-16
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.j2d;

import java.awt.*;
import java.awt.geom.*;
import java.awt.font.TextLayout;
import java.text.*;
import java.util.*;
import gov.usgs.gr.visual.*;

//==============================================================================
/**
 * The Y version of the axis object.
 */
public class J2dYAxis extends J2dAxis implements J2dElement {

//------------------------------------------------------------------------------
/**
 * Empty constructor.
 */
    public J2dYAxis () {

    	m_dimension = 1;

        setDefaults("gov.usgs.gr.visual.j2d.J2dYAxis");
    }

//------------------------------------------------------------------------------
/**
 * Comment.
 * @param limits	The limits object telling what segment of the axis should
 *					be represented.
 * @param opposingLimits	The opposing axis' limits.  Used to determine the
 *							how the length of tick marks when drawing in
 *							graph coordinates.
 */
    public J2dYAxis (Zoom zoom, Zoom opposingZoom) {

        super(zoom, opposingZoom);
    	m_dimension = 1;

        setDefaults("gov.usgs.gr.visual.j2d.J2dYAxis");
    }

//------------------------------------------------------------------------------
/**
 * Create a new axis by copying an existing one.
 * @param axis	The axis to be copied.
 */
    public J2dYAxis (Axis axis) {

    	super(axis);
    	m_dimension = 1;
    }

//------------------------------------------------------------------------------
/**
 * Overrides the draw() function from Axis.  Rotates the axis so it will
 * be drawing vertically.
 * <p><b> Pre Conditions </b><br>
 *      gl has been set to a valid OpenGL drawing object.
 */
    public void drawJ2d() {

    	// Get the latest from the axis limits objects.
        update();

        // Save the current transformation and then move the axis to its
        // proper location rotated 90 degrees.
        AffineTransform tempAt = (AffineTransform)m_at.clone();
        m_at = tempAt;
        m_at.translate(m_oppMin, -m_min);
        m_at.rotate(Math.toRadians(90));
        m_at.translate(m_min, -m_oppMax);

        // Draw the axis restore the previous transformation.
        super.drawJ2d();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Draws labels in magenta beside each major tick mark.
 * <p><b> Pre Conditions </b><br>
 *      j2d has been set.
 */
    public void drawText(double xOrig, double yOrig, double axisDisplayWidth,
    	Font font) {

        double x;

        if (!m_bShowText) return;
        if (m_majorTickInc <= 0) return;

		AffineTransform oldAt = j2d.getTransform();
        j2d.translate(xOrig - (0.6 * (double)font.getSize()), yOrig);
		j2d.scale(1, -1);
        double displayScale = axisDisplayWidth / (m_max - m_min);

        NumberFormat form = new DecimalFormat(m_zoom.getFormatString());

        // Loop through and draw each label.
        for (x = m_majorStart; (x - m_max) <= (0.01 * m_majorTickInc);
        	 x += m_majorTickInc) {

            if (Math.abs(x - 0) <= (0.01 * m_majorTickInc)) x = 0;
			TextLayout tl = new TextLayout(form.format(x), font,
            	j2d.getFontRenderContext());
            tl.draw(j2d, -tl.getVisibleAdvance(),
            	0.5f * tl.getAscent() - (float)((x - m_min) * displayScale));
        }

        // Draw the axis title.
        if (m_bShowTitle && (m_title.length() > 0)) {
            j2d.rotate(-Math.PI / 2);
            StringTokenizer st = new StringTokenizer(m_title, "" + '\n');

            int numLines = st.countTokens();
            int lineNum = 0;
            while (st.hasMoreTokens()) {
                String s = st.nextToken();
                TextLayout tl = new TextLayout(s, font,
                    j2d.getFontRenderContext());
                tl.draw(j2d, ((float)axisDisplayWidth - tl.getVisibleAdvance())
                    / 2f, -1.1f * (numLines - lineNum + 1.5f)
                    * tl.getAscent());
                lineNum++;
            }
        }

        // Restore the previous position.
        j2d.setTransform(oldAt);
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Y Axis";

        return s;
 	}
}



