/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2003-08-18
 * Date started:    1999-01-21
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.ps;

import java.awt.Rectangle;
import java.awt.geom.*;
import java.text.*;
import java.util.*;
import java.io.*;
import gov.usgs.gr.visual.*;

//==============================================================================
/**
 * A group of graphs layed out on a page.  Draws itself to a PostScript file.
 */
public class PageOfPsGraphs extends gov.usgs.gr.visual.PageOfGraphs
	implements PsElement {

    /** The output file writer. */
    protected PrintWriter ps;

//------------------------------------------------------------------------------
/**
 * Create a blank page for graphs.
 */
    public PageOfPsGraphs () {

     	partitionAreas();
        setDefaults("gov.usgs.gr.visual.ps.PageOfPsGraphs");
    }

//------------------------------------------------------------------------------
/**
 * Create a page of graphs cloned from an existing one.   Data is shared.
 */
    public PageOfPsGraphs (PageOfGraphs page) {

    	super (page);

        synchronized (page) {

            // Save the display size from the original page.
            Rectangle oldContent = new Rectangle(m_display);

            partitionAreas();

            float contentWidthRatio = (float)m_content.width
                / (float)oldContent.width;
            float contentHeightRatio = (float)m_content.height
                / (float)oldContent.height;

            // Scale the included graphs to fit the new page size.
            Iterator iter = m_graphList.iterator();
            while (iter.hasNext()) {
                PsGraph graph = (PsGraph)iter.next();

                Rectangle graphLocal = graph.getLocalRect();
                Rectangle graphDisplay = graph.getDisplayRect();

                graphDisplay.x *= contentWidthRatio;
                graphDisplay.y *= contentHeightRatio;
                graphDisplay.width *= contentWidthRatio;
                graphDisplay.height *= contentHeightRatio;
                graph.setDisplayRect(graphDisplay);

                // Local needs to be the same as display so that lines will
                // be proportioned correctly width/height.
                graphLocal.x = graphDisplay.x;
                graphLocal.y = graphDisplay.y;
                graphLocal.width = graphDisplay.width;
                graphLocal.height = graphDisplay.height;
                graph.setLocalRect(graphLocal);

/*
                Rectangle2D.Double graphMediaFrame = graph.getFrameRect();
                graphMediaFrame.x *= graphLocal.width;
                graphMediaFrame.y *= graphLocal.height;
                graphMediaFrame.width *= graphLocal.width;
                graphMediaFrame.height *= graphLocal.height;
                graph.setMediaFrameRect(graphMediaFrame);
*/
                graph.layoutFrame();

                // Set printing preferences.
                graph.psOptions = m_psOptions;
    /*
                graph.setColor(m_psOptions.m_bPsColor);
                graph.setUniformWidth(m_psOptions.m_bPsUniformWidth);
                graph.setFont(m_psOptions.m_psFont);
                graph.setFontSize(m_psOptions.m_psFontSize);
    */
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Sets up the size of the display and content areas with margins.
 */
    public void partitionAreas() {

    	// Define the display area as either 11 X 8.5 inches or 8.5 X 11
        // inches.  Measurements are given in points (1/72 of an inch).
    	if (m_psOptions.m_bPsLandscape) {
    		m_display.width = 792;
        	m_display.height = 612;
        }
        else {
    		m_display.width = 612;
        	m_display.height = 792;
        }

        // Define the content area size.
        m_content.x = 36;
        m_content.y = 36;
        m_content.width = m_display.width - 72;
        m_content.height = m_display.height - 72;
    }

//------------------------------------------------------------------------------
/**
 * Create a PostScript graph to be used in this page.
 */
    protected Graph newGraph() {
    	return new PsGraph();
    }

//------------------------------------------------------------------------------
/**
 * Create an PostScript graph as a clone of an existing graph.
 * @param graph	A template graph to use in creating this one.
 */
    protected Graph newGraph(Graph graph) {
    	return new PsGraph(graph);
    }

//------------------------------------------------------------------------------
/**
 * Sets the PrintWriter to use for output.
 * @param pw    The PrintWriter object.
 */
    public void setPsFile(PrintWriter pw) {
    	ps = pw;
    }

//------------------------------------------------------------------------------
/**
 * Clears the drawing area and redraws the contents.
 */
    public synchronized void draw() {

        //drawPS();
/*
    	// Check if this method is being executed by this object's own thread.
        // If it isn't, split a new thread off to do the work.
    	if (!m_bDrawing) {
        	start();
            return;
        }
*/
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public synchronized void drawPS() {

        drawFromThread();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Clears the drawing area and redraws the contents.
 */
    public synchronized void drawFromThread() {

        // Create an output file.
        PrintWriter out = null;
        try {
            out = new PrintWriter(new BufferedWriter(new FileWriter(
                m_psOptions.m_psFileName)));
        }
        catch (IOException e) {
        	javax.swing.JOptionPane.showMessageDialog(null,
                "File " + m_psOptions.m_psFileName
                + " could not be opened for PS output.",
                "IO Exception", javax.swing.JOptionPane.ERROR_MESSAGE);
        }

        setPsFile(out);
        writeHeader();

        DecimalFormat form = new DecimalFormat("0.##");

        // Rotate the page if it is landscape.
        if (m_psOptions.m_bPsLandscape) {
        	ps.println("90 rotate");
        	ps.println("0 " + form.format(-m_display.height) + " translate");
        }

        if (m_psOptions.getPsColorBg()) {
            // Draw a color background.
        	ps.println("\n" + "gsave");
        	ps.println(m_bgColor[0] + " " + m_bgColor[1] + " "
                + m_bgColor[2] + " setrgbcolor");
        	ps.println("0 0 moveto");
        	ps.println(m_display.width + " 0 rlineto");
        	ps.println("0 " + m_display.height + " rlineto");
        	ps.println(-m_display.width + " 0 rlineto closepath fill");
            ps.println("grestore" + "\n");
        }

        ps.println(form.format(m_content.x) + " " + form.format(m_content.y)
        	+ " translate");

        // Get a list of zooms on the page for possible corrections for
        // 1:1 aspect ratio.
        List aspectZoomList = getXZooms(m_showList);
        aspectZoomList.addAll(getYZooms(m_showList));

        // Let each graph in the panel draw itself using the PostScript
        // commands sent to the output file.
        int i = 0;
        Iterator iter = m_graphList.iterator();
        while (iter.hasNext()) {
            PsGraph graph = (PsGraph) iter.next();
            graph.psOptions = m_psOptions;
            if (graph.getShow()) {
                if (graph.getAspect1to1()) {
                    Zoom xZoom = graph.getXZoom();
                    Zoom yZoom = graph.getXZoom();
                    if (aspectZoomList.contains(xZoom)) {
                        if (aspectZoomList.remove(yZoom)) {
                            aspectZoomList.remove(xZoom);
                            graph.makeAspect1to1();
                        }
                    }
                }

                ps.println("\n% Graph " + i);
                graph.setPsFile(out);
                graph.drawPS();
            }
            i++;
        }

        drawText(m_psOptions.m_psFont, m_psOptions.m_psFontSize);

        // Draw any additional markup objects.
        Rectangle2D.Double outerRect = new Rectangle2D.Double(0, 0,
        	m_display.width, m_display.height);
        Rectangle2D.Double innerRect = new Rectangle2D.Double(m_content.x,
        	m_content.y, m_content.width, m_content.height);
        Rectangle2D.Double innerUser = new Rectangle2D.Double(0,
        	0, m_content.width, m_content.height);

        iter = m_markupList.iterator();
        while (iter.hasNext()) {
            Markup markup = (Markup)iter.next();
            markup.setRects(outerRect, innerRect, innerUser);
            markup.drawPS(ps);
        }

        writeTrailer();
        out.close();
    }

//------------------------------------------------------------------------------
/**
 * Draws title of the page.
 */
    protected void drawText(String font, float size) {

        if (m_title == null) return;

        float x;
        DecimalFormat form = new DecimalFormat("0.##");

        // Set the font.
		ps.println("\n% Set the font for the graph title");
		ps.println("/" + font + " findfont " + form.format(size)
        	+ " scalefont setfont");

        // Locate and draw the text.
        float pageX = 0;
        float pageY = (size / 2) + m_content.height;
        if (m_bShowTitle) {
            String text = m_title.replaceAll("\\(", "\\\\050")
                .replaceAll("\\)", "\\\\051");
            ps.println("(" + text + ") " + form.format(pageX) + " "
                + form.format(pageY) + " leftshow");
        }
    }

//------------------------------------------------------------------------------
/**
 * Writes a header into the PostScript file.
 * <p><b> Pre Conditions </b><br>
 *      Output file is open.
 * <p><b> Post Conditions </b><br>
 *      Need to draw contents, write trailer, and close file.
 */
    public void writeHeader() {

        // Write opening comments.
        ps.println("%!PS-Adobe-3.0 EPSF-3.0");
        ps.println("%%Creator: USGS Gr");

        // Define the edges for the bounding boxes.
        int minx, miny, maxx, maxy, width, height;
        if (m_psOptions.m_bPsLandscape) {
        	minx = m_display.height - m_content.height - m_content.y;
            miny = m_content.x;
            maxx = m_display.height - m_content.y;
            maxy = m_content.x + m_content.width;
            width = m_content.height;
            height = m_content.width;
 		}
        else {
        	minx = m_content.x;
            miny = m_content.y;
            maxx = m_content.x + m_content.width;
            maxy = m_content.y + m_content.height;
            width = m_content.width;
            height = m_content.height;
        }

        // Write out the bounding box definitions.
        ps.println("%%BoundingBox: " + minx + " " + miny
        	+ " " + maxx + " " + maxy);
        ps.println("%%PageOrigin: " + minx + " " + miny);
        ps.println("%AI3_TemplateBox: " + minx + " "
        	+ width + " " + miny + " " + height);
        ps.println("%AI3_TileBox: " + minx + " " + miny
        	+ " " + width + " " + height);

        ps.println("\n%%BeginProlog");

        // Define functions.

        ps.println("/thick {" + m_psOptions.m_psLineWidth + " mul} def");
        ps.println("/l {lineto} def");
        ps.println("/np {newpath} def");
        ps.println("/st {stroke} def");
        ps.println("/mv {moveto} def");
        ps.println("/rl {rlineto} def");
        ps.println("/rgb {setrgbcolor} def");

        // Define Line Styles.
        ps.println("\n% Define Line Styles");
        ps.println("/solid      {[]             0 setdash} def");
        ps.println("/dash       {[4]            0 setdash} def");
        ps.println("/dot        {[.3333333 1]   0 setdash} def");
        ps.println("/loosedot   {[.5 2.5]   0 setdash} def");
        ps.println("/chaindash  {[7 3 3 3]      0 setdash} def");
        ps.println("/chaindot   {[7 4 1 4]      0 setdash} def");
        ps.println("/dashdotdot {[7 2 1 3 1 2]  0 setdash} def");
        ps.println("/vardash    {[7 2 3 2]      0 setdash} def");

        // Define text placement functions.
        ps.println("\n% Define text placement functions");
        ps.println("/centershow {moveto dup stringwidth pop");
        ps.println("             -2 div 0 rmoveto show     } def");
        ps.println("/rightshow  {moveto dup stringwidth pop");
        ps.println("             neg 0 rmoveto show        } def");
        ps.println("/yshow      {moveto 90 rotate dup");
        ps.println("             stringwidth pop -2 div 0");
        ps.println("             rmoveto show -90 rotate   } def");
        ps.println("/rotshow    {moveto dup rotate exch dup stringwidth pop");
        ps.println("             -2 div 0 rmoveto show -1 mul rotate  } def");
        ps.println("/leftshow   {moveto show               } def");

        // Define symbol drawing functions in the Postscript file.
        ps.println("\n% Define symbol drawing functions");
        ps.println("/radius  { " + m_psOptions.m_psSymbolSize + " } def");
        ps.println("/mradius { radius neg    } def");
        ps.println("/diamtr  { 2 radius mul  } def");
        ps.println("/mdiamtr { 2 mradius mul } def");
        ps.println("/plussign {moveto");
        ps.println("           radius  0.0     rmoveto");
        ps.println("           mdiamtr 0.0     rlineto");
        ps.println("           radius  radius  rmoveto");
        ps.println("           0.0     mdiamtr rlineto");
        ps.println("           solid   stroke} def");
        ps.println("/xsymbol  {moveto");
        ps.println("           radius  radius  rmoveto");
        ps.println("           mdiamtr mdiamtr rlineto");
        ps.println("           diamtr  0       rmoveto");
        ps.println("           mdiamtr diamtr  rlineto");
        ps.println("           solid   stroke} def");
        ps.println("/tickmark {moveto");
        ps.println("           0.0     radius  rmoveto");
        ps.println("           0.0     mdiamtr rlineto");
        ps.println("           solid   stroke} def");
        ps.println("/onepixel {moveto   0.25   -0.25    rmoveto");
        ps.println("           0.0     0.5     rlineto");
        ps.println("           0.5     0.0     rlineto");
        ps.println("           0.0    -0.5     rlineto");
        ps.println("           closepath solid stroke } def");
        ps.println("/ocircle  {radius");
        ps.println("           0 360   arc solid stroke}def");
        ps.println("/osquare {moveto");
        ps.println("           radius  mradius rmoveto");
        ps.println("           0.0     diamtr  rlineto");
        ps.println("           mdiamtr 0.0     rlineto");
        ps.println("           0.0     mdiamtr rlineto");
        ps.println("           closepath solid stroke } def");
        ps.println("/otriup   {moveto");
        ps.println("           0.0     radius  rmoveto");
        ps.println("           radius  mdiamtr rlineto");
        ps.println("           mdiamtr 0.0     rlineto");
        ps.println("           closepath solid stroke } def");
        ps.println("/otridown {moveto");
        ps.println("           0.0     mradius rmoveto");
        ps.println("           radius  diamtr  rlineto");
        ps.println("           mdiamtr 0.0     rlineto");
        ps.println("           closepath solid stroke } def");
        ps.println("/odiamond {moveto");
        ps.println("           0.0     mradius rmoveto");
        ps.println("           radius  radius  rlineto");
        ps.println("           mradius radius  rlineto");
        ps.println("           mradius mradius rlineto");
        ps.println("           closepath solid stroke } def");
        ps.println("/fcircle  {radius  0 360   arc fill} def");
        ps.println("/fsquare  {moveto");
        ps.println("           radius  mradius rmoveto");
        ps.println("           0.0     diamtr  rlineto");
        ps.println("           mdiamtr 0.0     rlineto");
        ps.println("           0.0     mdiamtr rlineto");
        ps.println("           closepath fill } def");
        ps.println("/ftriup   {moveto");
        ps.println("           0.0     radius  rmoveto");
        ps.println("           radius  mdiamtr rlineto");
        ps.println("           mdiamtr 0.0     rlineto");
        ps.println("           closepath fill } def");
        ps.println("/ftridown {moveto");
        ps.println("           0.0     mradius rmoveto");
        ps.println("           radius  diamtr  rlineto");
        ps.println("           mdiamtr 0.0     rlineto");
        ps.println("           closepath fill } def");
        ps.println("/fdiamond {moveto");
        ps.println("           0.0     mradius rmoveto");
        ps.println("           radius  radius  rlineto");
        ps.println("           mradius radius  rlineto");
        ps.println("           mradius mradius rlineto");
		ps.println("           closepath fill} def");

        // Define a routine for drawing arrows.
		ps.println("/arrowdict 14 dict def arrowdict begin");
		ps.println("/mtrx matrix def end");
		ps.println("/arrow {arrowdict begin");
		ps.println("        /headlength exch def");
		ps.println("        /halfheadthickness exch 2 div def");
		ps.println("        /halfthickness exch 2 div def");
		ps.println("        /tipy exch def /tipx exch def");
		ps.println("        /taily exch def /tailx exch def");
		ps.println("        /dx tipx tailx sub def");
		ps.println("        /dy tipy taily sub def");
		ps.println("/arrowlength dx dx mul dy dy mul add sqrt def");
		ps.println("        /angle dy dx atan def");
		ps.println("        /base arrowlength headlength sub def");
		ps.println("        /savematrix mtrx currentmatrix def");
		ps.println("        tailx taily translate");
		ps.println("        angle rotate");
		ps.println("        0 halfthickness neg moveto");
		ps.println("        base halfthickness neg lineto");
		ps.println("        base halfheadthickness neg lineto");
		ps.println("        arrowlength 0 lineto");
		ps.println("        base halfheadthickness lineto");
		ps.println("        base halfthickness lineto");
		ps.println("        0 halfthickness lineto");
		ps.println("        closepath savematrix setmatrix end} def");
		ps.println("/rcfl  {setrgbcolor newpath 4 -2 roll moveto exch dup 0");
		ps.println("       rlineto exch 0 exch rlineto neg 0 rlineto fill}def");


        // End the function definition section.  Set the page to landscape.
        ps.println("\n%%EndProlog");
        ps.println("\n%%BeginSetup");
    }

//------------------------------------------------------------------------------
/**
 * Writes a header into the PostScript file.
 * <p><b> Pre Conditions </b><br>
 *      File is open and all contents have been drawn.
 * <p><b> Post Conditions </b><br>
 *      Output file needs to be closed.
 */
    protected void writeTrailer() {

        // Define the number of copies and show the page.
        ps.println("\n% Define the number of copies and show the page.");

        ps.println("/#copies 1 def");
        ps.println("showpage");

        // Write the closing comments.
        ps.println("%%EndPage: \"1\" 1");
        ps.println("%%Trailer");
        ps.println("%%EOF");
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	True if orientation is to be landscape instead of portrait.
 */
    public void setPsLandscape(boolean value) {

        m_psOptions.m_bPsLandscape = value;
		partitionAreas();
    }
}


