/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2003-02-17
 * Date started:    1999-01-21
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.ps;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import java.text.*;
import java.util.*;
import gov.usgs.gr.Gr;
import gov.usgs.gr.visual.*;

//==============================================================================
/**
 * A generic axis with ticks and labels for an X-Y graph.
*/
public abstract class PsAxis extends gov.usgs.gr.visual.Axis
	implements PsElement {

    /** The writer for the output file. */
    protected PrintWriter ps;

    /**

     * Describes the size of the graph area after it has been transformed
     * to a regular shape.
     */
    protected Rectangle2D.Double m_local = new Rectangle2D.Double(0, 0, 1, 1);

    /** Whether all lines should be the same width. */
    protected boolean m_bUniformWidth = false;
    /** The PS Options object with formatting parameters */
    public PsOptions psOptions = null;

//------------------------------------------------------------------------------
/**
 * Empty constructor.
 */
    public PsAxis () {

        setDefaults("gov.usgs.gr.visual.ps.PsAxis");
    }

//------------------------------------------------------------------------------
/**
 * Create a new axis and associates it with a set of limits for itself and
 * its perpendicular counterpart.
 * <p><b> Pre Conditions </b><br>
 *      Axis limits exist for this axis and its perpendicular counterpart.
 * <p><b> Post Conditions </b><br>
 *      None.
 */
    public PsAxis (Zoom zoom, Zoom opposingZoom) {

        setDefaults("gov.usgs.gr.visual.ps.PsAxis");

        m_zoom = zoom;
        m_opposingZoom = opposingZoom;
    }

//------------------------------------------------------------------------------
/**
 * Create a new axis by copying an existing one.
 * @param axis	The axis to be copied.
 */
    public PsAxis (Axis axis) {
    	super (axis);
    }

//------------------------------------------------------------------------------
/**
 * Sets the PrintWriter to use for output.
 * @param pw    The PrintWriter object.
 */
    public void setPsFile(PrintWriter pw) {
    	ps = pw;
    }

//------------------------------------------------------------------------------
/**
 * Draw all the elements that make up a time series graph.
 * <p><b> Pre Conditions </b><br>
 *      ps is a valid output file and the graph has been initialized.
 * <p><b> Post Conditions </b><br>
 *      None.
 */
    public void drawPS(){

        // Get the latest changes from the axis limits objects.
        update();
        if (m_majorTickInc <= 0) return;

	    ps.println("\n% Draw axis");

        // Calculate the start for the major tick marks, grid lines, and labels.
        double offset = m_zoom.getIncOffset();
        double startOffset = m_min - offset - 0.01 * m_majorTickInc;
        double startIntervals = (int)(startOffset / m_majorTickInc);
        if (startOffset > 0) startIntervals++;
        m_majorStart = (m_majorTickInc * startIntervals) + offset;

        // Calculate the length of a tick for this axis.
        m_majorTickLen = m_displayScale * m_tickSize * 0.25;

        // Draw the various parts of the  graph.
        if (m_bShowGridLines) drawGridLines();
        if (m_bShowMinorTicks) drawMinorTicks();
        if (m_bShowMajorTicks) drawMajorTicks();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Draw a light line across the graph for each major tick.  Line and color
 * attributes are saved and then restored when drawing is complete.
 * <p><b> Pre Conditions </b><br>
 *      ps is a valid output file and the graph has been initialized.
 */
    protected void drawGridLines() {

	    double x;
        NumberFormat form = new DecimalFormat("0.##");

        double widthScale = (m_bUniformWidth ? 1 : 0.2);

        ps.println("\n% Draw grid lines");

        ps.println("gsave");
        if (psOptions.getPsColorBg())
            ps.println(m_tickColor[0] + " " + m_tickColor[1] + " "
                + m_tickColor[2] + " setrgbcolor");
        ps.println(form.format(widthScale) + " thick setlinewidth");

        // Turn on a regular, dotted line pattern.
        ps.println("\n" + "% Draw grid lines\n" + "loosedot");

        // Vertical grid lines...
        for (x = m_majorStart; x < m_max; x += m_majorTickInc) {
        	ps.print("np ");
            double pageX = (float)
            	(m_local.width * (x - m_min) / (m_max - m_min));
            ps.print(form.format(pageX) + " 0 mv ");
            ps.print("0 " + form.format(m_local.height) + " rl ");
            ps.println("st");
        }
        ps.println("solid");

        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 * Draw one or more small tick marks between the large tick marks.
 * <p><b> Pre Conditions </b><br>
 *      ps is a valid output file and the graph has been initialized.
 */
    protected void drawMinorTicks() {

        if (m_minorTickInc <= 0) return;

	    double x, y;
        DecimalFormat form = new DecimalFormat("0.##");

        ps.println("\n% Draw minor ticks");

        ps.println("gsave");
        if (psOptions.getPsColorBg())
            ps.println(m_tickColor[0] + " " + m_tickColor[1] + " "
                + m_tickColor[2] + " setrgbcolor");

        double widthScale = (m_bUniformWidth ? 1 : 1);
        ps.println(form.format(widthScale) + " thick setlinewidth");

        // Calculate the location of the first tick mark.
		int tickNum = 0;
        int minsPerMax = (int)Math.round(m_majorTickInc / m_minorTickInc);
        double start = m_majorStart - m_majorTickInc;
        while (start <= m_min) {
        	start += m_minorTickInc;
            tickNum++;
        }
        if (tickNum > minsPerMax) tickNum -= minsPerMax;

        double len = m_majorTickLen * m_minorTickProportion;

        // Loop through and draw each tick...
        for (x = start; x < m_max; x += m_minorTickInc) {
        	if (tickNum >= minsPerMax) {
            	// Skip drawing a tick here because there is a major tick.
                tickNum = 1;
            }
        	else {
        		// Draw a minor tick on both sides of the graph.
                // Near side...
        		ps.print("np ");
            	double pageX = (double)(m_local.width
                	* (x - m_min) / (m_max - m_min));
            	ps.print(form.format(pageX) + " 0 mv ");
            	ps.print("0 " + form.format(len) + " rl ");
            	ps.println("st");

                // Far side...
                ps.print("np ");
            	ps.print(form.format(pageX) + " " + form.format(m_local.height)
                	+ " mv ");
            	ps.print("0 " + form.format(-len) + " rl ");
            	ps.println("st");

            	tickNum++;
        	}
        }

        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 * Draw one or more large tick marks along the axis.
 * <p><b> Pre Conditions </b><br>
 *      ps is a valid output file and the graph has been initialized.
 */
    protected void drawMajorTicks() {

	    double x;
        DecimalFormat form = new DecimalFormat("0.##");

        // Loop through and draw each tick...
        double len = m_majorTickLen;

        ps.println("\n% Draw major ticks");

        ps.println("gsave");
        if (psOptions.getPsColorBg())
            ps.println(m_tickColor[0] + " " + m_tickColor[1] + " "
                + m_tickColor[2] + " setrgbcolor");

        double widthScale = (m_bUniformWidth ? 1 : 1.6);
        ps.println(form.format(widthScale) + " thick setlinewidth");

        form = new DecimalFormat("0.##");

        for (x = m_majorStart; x < m_max; x += m_majorTickInc) {
        	// Draw a minor tick on both sides of the graph.
            // Near side...
            double pageX =
            	(double)(m_local.width * (x - m_min) / (m_max - m_min));
            ps.print(form.format(pageX) + " 0 mv ");
            ps.print("0 " + form.format(len) + " rl ");
            ps.println("st");

            // Far side...
            ps.print("np ");
            ps.print(form.format(pageX) + " " + form.format(m_local.height)
            	+ " mv ");
            ps.print("0 " + form.format(-len) + " rl ");
            ps.println("st");

        }
        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 * Draws labels in magenta beside each major tick mark.
 * <p><b> Pre Conditions </b><br>
 *      ps is a valid output file and the graph has been initialized.
 */
    public void drawText(String font, double size, double xOrig, double yOrig,
    	double axisDisplayWidth) {

        // Return if the text shouldn't be shown.
        if (!m_bShowText) return;
        if (m_majorTickInc <= 0) return;

        DecimalFormat form = new DecimalFormat("0.##");
        double x;

        // Set the font with PostScript.
		ps.println("\n% Set the font for the axis titles");
		ps.println("/" + font + " findfont " + form.format(size)
        	+ " scalefont setfont");

        // Save the old location and move to the starting point.
        ps.println("gsave");
        ps.println(form.format(xOrig) + " "
        	+ form.format(yOrig - (size * 1.5)) + " translate");
        double displayScale = axisDisplayWidth / (m_max - m_min);

        NumberFormat charForm = new DecimalFormat(m_zoom.getFormatString());

        // Loop through and draw each label.
        for (x = m_majorStart; (x - m_max) <= (0.01 * m_majorTickInc);
        	 x += m_majorTickInc) {

            if (Math.abs(x - 0) <= (0.01 * m_majorTickInc)) x = 0;
            double pageX = (double)(displayScale * (x - m_min));
            ps.println("(" + charForm.format(x) + ") " + form.format(pageX)
            	+ " 0 centershow");
        }

        // Draw the axis title.
        if (m_bShowTitle) {
        	ps.println("0 " + form.format(-size * 0.4) + " translate");

            double pageX = (double)(displayScale * (m_max - m_min) / 2);

            StringTokenizer st = new StringTokenizer(m_title, "" + '\n');
            while (st.hasMoreTokens()) {
        		ps.println("0 " + (-size * 1.1) + " translate");

                String text = st.nextToken().replaceAll("\\(", "\\\\050")
                    .replaceAll("\\)", "\\\\051");
            	ps.println("(" + text + ") "
                	+ form.format(pageX) + " 0 centershow");
            }
        }

        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param local Rectangle discribing the size of the graph area in local
 * 				coordinates.
 */
    public void setLocalRect(Rectangle2D.Double local) {
        m_local = local;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	True if all lines are to be drawn the same width.
 */
    public void setUniformWidth(boolean value) {
        m_bUniformWidth = value;
    }
}



