/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2008-04-24
 * Date started:    1999-01-22
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.ps;

import gov.usgs.gr.Gr;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.*;
import java.awt.geom.*;
import java.util.*;
import java.text.*;
import java.io.*;

//==============================================================================
/**
 * A single strip of data points that can draw itself.
 */
public class PsCurve extends gov.usgs.gr.visual.Curve implements PsElement,
    PsDrawingObject {

    /** The writer for the output file. */
    protected PrintWriter ps;
    /** The lowest value on the x axis. */
    protected double m_minx = 0;
    /** The highest value on the x axis. */
    protected double m_maxx = 1;
    /** The lowest value on the y axis. */
    protected double m_miny = 0;
    /** The highest value on the y axis. */
    protected double m_maxy = 1;
    /** Predefined symbols to use. */
    protected String[] m_symbols = {"onepixel", "plussign", "xsymbol",
    	"tickmark", "ocircle", "osquare", "otriup", "otridown", "odiamond",
        "fcircle", "fsquare", "ftriup", "ftridown", "fdiamond"};
    /**
     * The resolution of the output in points.  Best choice is between
     * 0.1 (720 dpi) and 1.0 (72 dpi).  Points that are closer together
     * than what can be resolved on the page won't be plotted.
     */
    protected float m_resolution = 0.1f;
    /**
     * Describes the size of the graph area after it has been transformed
     * to a regular shape.
     */
    protected Rectangle2D.Double m_local = new Rectangle2D.Double(0, 0, 1, 1);
    /** Whether all lines should be the same width. */
    protected boolean m_bUniformWidth = false;
    /** The PS Options object with formatting parameters */
    public PsOptions psOptions = null;
    /**
     * Whether the color should be drawn as is without converting to black
     * and white or darkening the color.
     */
//    protected boolean m_bColorOverride = false;

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public PsCurve(){

        m_bleach = -0.5f;
        updateDrawColor();

        setDefaults("gov.usgs.gr.visual.ps.PsCurve");
    }

//------------------------------------------------------------------------------
/**
 * Constructor for cloning this curve from an existing one.
 */
    public PsCurve(Curve curve) {

		super(curve);
    }

//------------------------------------------------------------------------------
/**
 * Constructor for creating a curve from a data series.
 * @param series    The data that this curve will use.
 */
    public PsCurve(DoubleDataSeries series) {
	
        super(series);

        m_bleach = -0.5f;
        updateDrawColor();

        setDefaults("gov.usgs.gr.visual.ps.PsCurve");
   }

//------------------------------------------------------------------------------
/**
 * Sets the PrintWriter to use for output.
 * @param pw    The PrintWriter object.
 */
    public void setPsFile(PrintWriter pw) {
    	ps = pw;
    }

//------------------------------------------------------------------------------
/**
 * Range of the axis, so the curve can be scaled to fit.
 * @param pw    The PrintWriter object.
 */
    public void setAxisRange(double minx, double maxx, double miny,
        double maxy) {

    	m_minx = minx;
        m_maxx = maxx;
        m_miny = miny;
        m_maxy = maxy;
    }

//------------------------------------------------------------------------------
/**
 * Issue the PostScript commands to draw this series.
 * <p><b> Pre Conditions </b><br>
 *      The origin has been moved to (0, 0), scaling has occurred, and
 *      line type is solid.
 */
    public void drawPS(){

        if (!m_bShow) return;

        synchronized (m_series) {
    //        if (m_bColorOverride) {
                ps.println("\n" + "gsave");
                ps.println(m_drawColor[0] + " " + m_drawColor[1] + " "
                    + m_drawColor[2] + " setrgbcolor");
    //        }
            if (m_bShowLine) {
                // Draw an unreduced version of the line underneath in red
                // if debug is turned on.
                if (m_bDebugPointReduction && m_reducedLineSeries != null) {
                    ps.println("\n" + "gsave" + "\n" + "% Red Debug Curve...");
                    ps.println("1 0 0  setrgbcolor");
                    DoubleDataSeries tempSeries = m_reducedLineSeries;
                    m_reducedLineSeries = null;
                    drawLine();
                    m_reducedLineSeries = tempSeries;
                    ps.println("grestore" + "\n");
                }

                // Draw the line.
                drawLine();
            }

            if (m_bShowSymbols) {
                // Draw an unreduced version of the line underneath in red
                // if debug is turned on.
                if (m_bDebugPointReduction && m_reducedSymbolSeries != null) {
                    ps.println("\n" + "gsave" + "\n" + "% Red Debug Symbols...");
                    ps.println("1 0 0  setrgbcolor");
                    DoubleDataSeries tempSeries = m_reducedSymbolSeries;
                    m_reducedSymbolSeries = null;
                    drawSymbols();
                    m_reducedSymbolSeries = tempSeries;
                    ps.println("grestore" + "\n");
                }

                // Draw the symbols.
                drawSymbols();
            }
    //        if (m_bColorOverride) {
                ps.println("\n" + "grestore");
    //        }
        }

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Issue the PostScript commands to draw a line for this series.
 * <p><b> Pre Conditions </b><br>
 *      The origin has been moved to (0, 0), scaling has occurred, and
 *      line type is solid.
 */
    public void drawLine() {

    	// Get the array of data points to be drawn.
        double[] array = null;
        if (m_reducedLineSeries != null) {
        	array = m_reducedLineSeries.getArray();
            m_breaks = null;
        }
        else
        	array = m_series.getArray();

        // Don't draw anything if the array is empty.
        if (array.length == 0) return;

        double widthScale = (m_bUniformWidth ? 1.0 : m_lineWidth);

    	// Set the line thickness.
        ps.println("\n% Draw curve");

        if (m_bSolid || m_linePattern == Gr.SOLID_LINE_PATTERN) {
        	ps.println("solid");
        }
        else {
        	switch (m_linePattern) {
            	case (Gr.DASHED_LINE_PATTERN):
                	ps.println("dash");
                    break;
                case (Gr.DOTTED_LINE_PATTERN):
                	ps.println("dot");
                    break;
                case (Gr.LOOSE_DOTTED_LINE_PATTERN):
                	ps.println("loosedot");
                    break;
                case (Gr.DASH_DOT_DASH_LINE_PATTERN):
                	ps.println("chaindash");
                    break;
                case (Gr.DASH_DOT_LINE_PATTERN):
                	ps.println("chaindot");
                    break;
                case (Gr.LOOSE_DASHED_LINE_PATTERN):
                	ps.println("vardash");
                    break;
                default:
                	ps.println("solid");
                    break;
            }
        }

        NumberFormat form = new DecimalFormat("0.##");
        ps.println(form.format(widthScale) + " thick setlinewidth");

        // If there are no breaks, just draw the active section of the
        // data series.
        if (m_breaks == null) {
            drawLineSegment(array, 0, ((array.length - 1) / 2));
        }
        else {
            // The breaks array has been set and holds pairs of begin and
            // end indexes for the line segments.  Draw each segment
            // separately.
            int numPoints = m_series.getCount();
            for (int i = 0; i < (m_breaks.length - 1); i += 2) {
                if (m_breaks[i] >= 0 && m_breaks[i] < numPoints &&
                    m_breaks[i + 1] >= 0 && m_breaks[i + 1] < numPoints) {

                    drawLineSegment(array, m_breaks[i],
                        m_breaks[i + 1] - m_breaks[i]);
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Issue the PostScript commands to draw a line segment for this series.
 */
    public void drawLineSegment(double[] array, int begin, int end) {

        NumberFormat form = new DecimalFormat("0.##");
        NumberFormat largeForm = new DecimalFormat("#0.#E000");
        double xScale = (double)m_local.width / (m_maxx - m_minx);
        double yScale = (double)m_local.height / (m_maxy - m_miny);
       
        // Draw the curve itself using the lineto command.
        int nextI;
        int prevI;
        int numPlotted = 0;
        boolean bLineUnfinished = false;
        boolean bPlot = false;
        boolean bFirst = true;
        boolean bLivePath = false;
        Rectangle2D.Double bounds = new Rectangle2D.Double(m_minx, m_miny,
            (m_maxx - m_minx), (m_maxy - m_miny));

        for (int i = begin; i <= end; i++) {
        	// Perform bounds checking on next and previous indexes.
            nextI = i + 1;
            prevI = i - 1;
        	if (i >= end) nextI = i;
        	if (i <= begin) prevI = i;

            // Create a Line2D object from the current point to the next.
            double curX = array[i * 2];
            double curY = array[i * 2 + 1];
            double nextX = array[nextI * 2];
            double nextY = array[nextI * 2 + 1];
                        
            Line2D.Double curNextLine =
                new Line2D.Double(curX, curY, nextX, nextY);

            // Determine if this point should be plotted and if the line
            // from this point to the next intersects the plot boundary.
            // If it does, the next point should automatically be marked
            // to be plotted.
            if (bLineUnfinished) {
                bPlot = true;
                bLineUnfinished = curNextLine.intersects(bounds);
            }
            else {
                if (bounds.contains(curX, curY)) {
                    bPlot = true;
                    bLineUnfinished = true;
                }
                else {
                    if (bLineUnfinished = curNextLine.intersects(bounds)) {
                        bPlot = true;
                    }
                    else {
                        bPlot = false;
                    }
                }
            }
            
            // Don't draw segments that contain NaN.
            if (Double.isNaN(curX) || Double.isNaN(curY) ||
                Double.isNaN(nextX) || Double.isNaN(nextY)) {
                
                bPlot = false;

                // Stroke the line.
                if (bLivePath) {
                    ps.println("st");
                    bLivePath = false;
                    bFirst = true;
                }
            }

            if (bPlot) {
                // Start a new path for a line, if one is not already live.
                if (!bLivePath) {
                    ps.print("np ");
                    bLivePath = true;
                }

                // Calculate the page coordinates of the current point.
        	double pageX = xScale * (curX - m_minx);
            	double pageY = yScale * (curY - m_miny);
                
                String xForm = form.format(pageX);
                String yForm = form.format(pageY);
                if (pageX >= Double.MAX_VALUE) {
                    xForm = "1.0E37";
                }
                if (pageX <= -Double.MAX_VALUE) {
                    xForm = "-1.0E37";
                }
                if (pageY >= Double.MAX_VALUE) {
                    yForm = "1.0E37";
                }
                if (pageY <= -Double.MAX_VALUE) {
                    yForm = "-1.0E37";
                }

                ps.print(xForm + " " + yForm);
                numPlotted++;

                // Break the text line after every six points.
                if (bFirst) {
                    // The first coordinates given should be followed by
                    // the moveto command rather than lineto.
                    ps.print(" mv ");
                    bFirst = false;
                }
                else {
                    if ((numPlotted % 6) == 0) ps.println(" l");
                    else ps.print(" l ");
                }
            }
        }

        // Stroke the line.
        if (bLivePath) {
            ps.println("st");
        }

System.out.println("Plotted " + numPlotted + " of " + (array.length / 2)
	+ " line points.");
    }

//------------------------------------------------------------------------------
/**
 * Issue the PostScript commands to draw symbols for this series.
 * <p><b> Pre Conditions </b><br>
 *      The origin has been moved to (0, 0), scaling has occurred, and
 *      line type is solid.
 */
    public void drawSymbols() {


    	// Get the array of data points to be drawn.
        double[] array = null;
        if (m_reducedSymbolSeries != null)
        	array = m_reducedSymbolSeries.getArray();
        else
        	array = m_series.getArray();

        // Don't draw anything if the array is empty.
        if (array.length == 0) return;

    	double pageX, pageY;
        double xScale = (double)m_local.width / (m_maxx - m_minx);
        double yScale = (double)m_local.height / (m_maxy - m_miny);

        ps.println("\n% Draw symbols");

        ps.println("/radius  { " + m_symbolSize + " } def");
        ps.println("newpath");

        NumberFormat form = new DecimalFormat("0.##");

        // Draw the curve itself using the lineto command.
        int nextI;
        int prevI;
        int numPlotted = 0;

        int end = (array.length - 1) / 2;
        for (int i = 1; i <= end; i++) {
            // Don't plot NaN values.
            if (Double.isNaN(array[i * 2]) || Double.isNaN(array[i * 2 + 1]))
                continue;
            
        	// Set do bounds checking on next and previous indexes.
            nextI = i + 1;
            prevI = i - 1;
        	if (i >= end) nextI = i;
        	if (i <= 0) prevI = i;

        	// Only draw points whose preceding or following point are within
            // the clipping region.
        	if ((array[nextI * 2] > m_minx && array[nextI * 2] < m_maxx)
            	|| (array[prevI * 2] > m_minx && array[prevI * 2] < m_maxx)
                || nextI == i) {

                // Calculate the page coordinates of the current point.
        		pageX = xScale * (array[i * 2] - m_minx);
            	pageY = yScale * (array[i * 2 + 1] - m_miny);

                ps.print(form.format(pageX) + " " + form.format(pageY));
                ps.println(" " + m_symbols[m_symbolIndex]);
                numPlotted++;
            }
        }

System.out.println("Plotted " + (numPlotted + 1) + " of " + (array.length / 2)
	+ " symbol points.");
    }

//------------------------------------------------------------------------------
/**
 * Issue the PostScript commands to draw symbols for this series.
 * <p><b> Pre Conditions </b><br>
 *      The origin has been moved to (0, 0), scaling has occurred, and
 *      line type is solid.
 */
    public void drawDeletedMarkers() {

    	double pageX, pageY;

        NumberFormat form = new DecimalFormat("0.##");

        // Get the list of deleted points.
        List deletedList = m_series.getDeletedList();
        if (deletedList.size() <= 0) return;

        ps.println("\n% Draw deleted points");
        ps.println("newpath \ngsave \n0.25 setlinewidth \n/radius  { 2 } def");
        m_series.m_bParabolicInterpolation = false;

        // Copy the elements to an array.
        double[] array = new double[deletedList.size()];
        Iterator iter = deletedList.iterator();
        int index = 0;
        while (iter.hasNext()) {
            double x = (Double)iter.next();
            array[index++] = x;
            iter.next();
            double y = m_series.getInterpolatedValue(x);
            array[index++] = y;
        }

        // Thin out the array.
        DoubleDataSeries series = new DoubleDataSeries(array, 2);
        double xScale = (double)m_local.width / (m_maxx - m_minx);
        double yScale = (double)m_local.height / (m_maxy - m_miny);
        double widthScale = 1;
        SymbolPointReducer spr = new SymbolPointReducer(xScale, yScale,
        	widthScale);
        List seriesList = new ArrayList();
        seriesList.add(series);
        spr.performAction(seriesList);
        array = series.getArray();

        // Draw a marker for each point in the array.
        for (int i = 0; i < array.length; i += 2) {
            double xPos = array[i];
            double yPos = array[i + 1];

            if (xPos < m_minx || xPos > m_maxx
            	|| yPos < m_miny || yPos > m_maxy) continue;

            // Calculate the page coordinates of the current point.
            pageX = (double)m_local.width * (xPos - m_minx) / (m_maxx - m_minx);
            pageY = (double)(m_local.height * (yPos - m_miny)
                / (m_maxy - m_miny));

        	ps.println("1 1 1 setrgbcolor");
            ps.print(form.format(pageX) + " " + form.format(pageY));
        	ps.println(" fcircle");
        	ps.println("1 0 0 setrgbcolor");
            ps.print(form.format(pageX) + " " + form.format(pageY));
        	ps.println(" xsymbol");
        	ps.println("0 0 0 setrgbcolor");
            ps.print(form.format(pageX) + " " + form.format(pageY));
        	ps.println(" ocircle");
        }

        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param local Rectangle discribing the size of the graph area in local
 * 				coordinates.
 */
    public void setLocalRect(Rectangle2D.Double local) {
        m_local = local;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	True if all lines are to be drawn the same width.
 */
    public void setUniformWidth(boolean value) {
        m_bUniformWidth = value;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	The new resolution in points.
 */
    public void setResolution(float value) {
        m_resolution = value;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	Whether to draw the color as is no matter what.
 */
//    public void setColorOverride(boolean value) {
//        m_bColorOverride = value;
//    }
}







