/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    1999-01-21
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.ps;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import java.lang.reflect.*;
import java.text.*;
import java.util.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;

//==============================================================================
/**
 * An x-y graph with ticks, labels, a grid, and multiple curves drawn inside.
 */
public class PsGraph extends gov.usgs.gr.visual.Graph implements PsElement {

    /** The writer for the output file. */
    protected PrintWriter ps;

    // Print options.
    /** Whether to draw curves in color. */
    protected boolean m_bColorCurves = false;
    /** Whether to draw background elements in color. */
    protected boolean m_bColorBg = false;
    /** The font to use for printing. */
    protected String m_font = "Helvetica-Bold";
    /** The height of the print font. */
    protected float m_fontSize = 8;
    /** Whether all lines should be the same width. */
    protected boolean m_bUniformWidth = false;
    /** The PS Options object with formatting parameters */
    public PsOptions psOptions = null;
    /**
     * Describes the rectangle containing the graph frame in screen or
     * page coordinates.
     */
    protected Rectangle2D.Double m_mediaFrame = new Rectangle2D.Double(
        0.08, 0.1, 0.9, 0.8);

//------------------------------------------------------------------------------
/**
 * Construct a time series graph and have it create its own x-axis zoom.
 */
    public PsGraph () {

        super();

        setDefaults("gov.usgs.gr.visual.ps.PsGraph");
    }

//------------------------------------------------------------------------------
/**
 * Construct a time series graph and use the specified x-axis limits.
 * @param xZoom The x-axis limits object to use for this graph.
 */
    public PsGraph (Zoom xZoom) {

        super(xZoom);

        setDefaults("gov.usgs.gr.visual.ps.PsGraph");
    }

//------------------------------------------------------------------------------
/**
 * Create a graph cloned from an existing one.
 */
    public PsGraph (Graph graph) {
    	super(graph);
    }

//------------------------------------------------------------------------------
/**
 * Returns a new x axis object.
 */
    protected Axis newXAxis() {
    	return new PsXAxis();
    }

//------------------------------------------------------------------------------
/**
 * Returns a new y axis object.
 */
    protected Axis newYAxis() {
    	return new PsYAxis();
    }

//------------------------------------------------------------------------------
/**
 * Returns a new Curve object.
 */
    protected Curve newCurve() {
    	return new PsCurve();
    }

//------------------------------------------------------------------------------
/**
 * Returns a new Curve object based on the data series parameter.
 */
    protected Curve newCurve(DoubleDataSeries series) {
	
    	return new PsCurve(series);
    }

//------------------------------------------------------------------------------
/**
 * Returns a new Curve object copied from an existing curve.
 */
    protected Curve newCurve(Curve curve) {
    	return new PsCurve(curve);
    }

//------------------------------------------------------------------------------
/**
 * Makes new copies of the axis objects associated with graph.
 * @param graph	The graph with the template axes.
 */
    protected void copyAxes(Graph graph) {

		m_xAxis = new PsXAxis(graph.getXAxis());
		m_yAxis = new PsYAxis(graph.getYAxis());
        m_xAxis.setOpposingZoom(m_yAxis.getZoom());
        m_yAxis.setOpposingZoom(m_xAxis.getZoom());

        m_xZoom = m_xAxis.getZoom();
        m_yZoom = m_yAxis.getZoom();
    }

//------------------------------------------------------------------------------
/**
 * Sets the PrintWriter to use for output.
 * @param pw    The PrintWriter object.
 */
    public void setPsFile(PrintWriter pw) {
    	ps = pw;
    }

//------------------------------------------------------------------------------
/**
 * Draw this graph with all its parts on the screen.
 * <p><b> Pre Conditions </b><br>
 *      The output file, ps, has been set.
 */
    public void drawPS() {

        if (psOptions != null) {
            m_bColorCurves = psOptions.getPsColor();
            m_bColorBg = psOptions.getPsColorBg();
            m_font = psOptions.getPsFont();
            m_fontSize = psOptions.getPsFontSize();
            m_bUniformWidth = psOptions.getPsUniformWidth();
        }

        DecimalFormat form = new DecimalFormat("0.##");

 		ps.println("\n" + "gsave");
        ps.println(form.format(m_display.x) + " " + form.format(m_display.y)
        	+ " translate");

        // Recalculate the media frame to use the latest frame and display
        // values.
        getMediaFrameRect();

    	ps.println(form.format(m_mediaFrame.x) + " "
            + form.format(m_mediaFrame.y) + " translate");

 		ps.println("\n" + "gsave");
        enableClipping();

        // Set the scale for the axes so the x and y ticks appear equally
        // long on the screen.
        m_xAxis.setDisplayScale(m_local.height / 20);
        m_yAxis.setDisplayScale(m_local.height / 20);

        // Tell the axes what size they are.
        ((PsAxis)m_xAxis).setLocalRect(m_mediaFrame);

        Rectangle2D.Double yLocal = new Rectangle2D.Double();
        yLocal.width = m_mediaFrame.height;
        yLocal.height = m_mediaFrame.width;
        ((PsAxis)m_yAxis).setLocalRect(yLocal);

    	// Draw the axes and grid.
        if (m_bColorBg) ps.println("1 1 1 setrgbcolor");

        ((PsAxis)m_xAxis).psOptions = psOptions;
        ((PsAxis)m_xAxis).setPsFile(ps);
        ((PsXAxis)m_xAxis).drawPS();
        ((PsAxis)m_yAxis).psOptions = psOptions;
        ((PsAxis)m_yAxis).setPsFile(ps);
        ((PsYAxis)m_yAxis).drawPS();

        // Draw all child graphical elements other than curves, axes, and
        // markups.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof GraphicalElement && !(gro instanceof Curve)){
                if (gro instanceof PsElement && !(gro instanceof PsAxis)) {
                    if (!(gro instanceof Markup)) {
                        PsElement psel = (PsElement)gro;
                        if (psel instanceof PsDrawingObject) {
                            ((PsDrawingObject)psel).setLocalRect(m_mediaFrame);
                            ((PsDrawingObject)psel).setAxisRange(
                                m_xZoom.getViewMin(), m_xZoom.getViewMax(),
                                m_yZoom.getViewMin(), m_yZoom.getViewMax());
                        }
                        ((PsDrawingObject)psel).psOptions.setProperties(
                            psOptions.getProperties());
                        psel.setPsFile(ps);
                        psel.drawPS();
                    }
                }
            }
        }

        // Set the line joining method from mitered to rounded.
        ps.println("1 setlinejoin");

        // Draw the curves inside the graph.
        iter = m_curves.iterator();
        while (iter.hasNext()) {
            PsCurve curve = (PsCurve)iter.next();
            curve.psOptions = psOptions;
            if (!m_bColorBg) {
                curve.setProperty("bleach", "-0.5");
            }
            curve.setPsFile(ps);
            curve.setLocalRect(m_mediaFrame);
            curve.setAxisRange(
                m_xZoom.getViewMin(), m_xZoom.getViewMax(),
                m_yZoom.getViewMin(), m_yZoom.getViewMax());
/*
            if (m_bColor) {
                float[] color = curve.getColor();
                // Darken the color so it will show up on a white page.
                color[0] *= 0.5;
                color[1] *= 0.5;
                color[2] *= 0.5;
                ps.println(color[0] + " " + color[1] + " " + color[2]
                    + " setrgbcolor");
            }
*/
            if (!m_bColorCurves) curve.setProperty("bleach", "-1");
            curve.drawPS();
            if (!m_bDrawDeletedOutside) curve.drawDeletedMarkers();
        }


        // Set the line joining method back to mitered.
        ps.println("0 setlinejoin");
        // Turn off clipping.
   		ps.println("\n% Done drawing curves and axes.  "
        	+ "Disable clipping.");
   		ps.println("grestore");

        // Draw a frame border.
    	drawFrame();

        // Draw text and labels.
        ps.println("gsave");
        if (m_bColorBg)
            ps.println(m_textColor[0] + " " + m_textColor[1] + " "
                + m_textColor[2] + " setrgbcolor");

        if (m_bShowTitle) drawText(m_font, m_fontSize);
        ((PsXAxis)m_xAxis).drawText(m_font, m_fontSize,
        	0, 0, m_mediaFrame.width);
        ((PsYAxis)m_yAxis).drawText(m_font, m_fontSize,
        	0, 0, m_mediaFrame.height);

        ps.println("grestore");

        // Draw any additional markup objects.
    	ps.println(form.format(-m_mediaFrame.x) + " "
            + form.format(-m_mediaFrame.y) + " translate");
        Rectangle2D.Double outerRect = new Rectangle2D.Double(0, 0,
        	m_display.width, m_display.height);
        Rectangle2D.Double innerRect = new Rectangle2D.Double(
        	m_frame.x, m_frame.y, m_frame.width, m_frame.height);
//        Rectangle2D.Double innerRect = new Rectangle2D.Double(
//            m_mediaFrame.x, m_mediaFrame.y, m_mediaFrame.width,
//            m_mediaFrame.height);
        Rectangle2D.Double innerUser = new Rectangle2D.Double(m_xZoom.getViewMin(),
        	m_yZoom.getViewMin(), m_xZoom.getViewLength(),
            m_yZoom.getViewLength());

        iter = m_markupList.iterator();
        while (iter.hasNext()) {
            Markup markup = (Markup)iter.next();
            markup.setRects(outerRect, innerRect, innerUser);
            markup.drawPS(ps);
        }

        // Move back to the page content origin.
   		ps.println("\n% Move back to page content origin.");
   		ps.println("grestore");

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Mark this graph on the screen as being selected.
 */
    private void drawAsSelected() {
    }

//------------------------------------------------------------------------------
/**
 * Draw a background behind the curves and axes.
 */
    protected void drawBackground() {
    }


//------------------------------------------------------------------------------
/**
 * Draw a thick frame around the graph.
 * <p><b> Pre Conditions </b><br>
 *      The viewing transformation has been set up so that the graphing
 *		area is the same as the frame size.
 */
    protected void drawFrame(){

        DecimalFormat form = new DecimalFormat("0.##");

		// Define a path and stroke it.
        ps.println("\n% Draw a frame around the plot area");
        ps.println("gsave");
        if (m_bColorBg)
            ps.println(m_frameColor[0] + " " + m_frameColor[1] + " "
                + m_frameColor[2] + " setrgbcolor");
        ps.println("solid");

        double widthScale = (m_bUniformWidth ? 1 : 2.7);
        ps.println(widthScale + " thick setlinewidth");
        ps.println("newpath");
        ps.println("0 0  moveto");
        ps.println(form.format(m_mediaFrame.width) + " 0 rlineto");
        ps.println("0 " + form.format(m_mediaFrame.height) + " rlineto");
        ps.println(form.format(-m_mediaFrame.width) + " 0 rlineto");
        ps.println("closepath  stroke");

        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 * Define a clipping region around the edge of the graph..
 * <p><b> Pre Conditions </b><br>
 *      The viewing transformation has been set up so that the graphing
 *		area is the same as the frame size.
 */
    protected void enableClipping() {

        DecimalFormat form = new DecimalFormat("0.##");

		// Define a path and clip it.
        ps.println("\n% Define a clipping area.");
        ps.println("newpath");
        ps.println("0 0  moveto");
        ps.println(form.format(m_mediaFrame.width) + " 0 rlineto");
        ps.println("0 " + form.format(m_mediaFrame.height) + " rlineto");
        ps.println(form.format(-m_mediaFrame.width) + " 0 rlineto");
        ps.println("closepath  clip");
    }

//------------------------------------------------------------------------------
/**
 * Draws titles above the graph.
 */
    protected void drawText(String font, float size) {

        if (m_bShowTitle) {
            DecimalFormat form = new DecimalFormat("0.##");

            // Set the font.
            ps.println("\n% Set the font for the graph title");
            ps.println("/" + font + " findfont " + form.format(size)
                + " scalefont setfont");

            // Locate and draw the text.
            float pageX = 0;
            float pageY = (size / 2) + (float)m_mediaFrame.height;
            String text = m_title.replaceAll("\\(", "\\\\050")
                .replaceAll("\\)", "\\\\051");
            ps.println("(" + text + ") " + form.format(pageX) + " "
        	    + form.format(pageY) + " leftshow");
        }
    }

//------------------------------------------------------------------------------
/**
 * Reproportion the frame so it fits with any titles that are being displayed.
 */
    public void layoutFrame() {

    	super.layoutFrame();

        m_mediaFrame.x *= m_local.width;
        m_mediaFrame.y *= m_local.height;
        m_mediaFrame.width *= m_local.width;
        m_mediaFrame.height *= m_local.height;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param height The new height of the graph in display units.
 */
    public void setHeight(int height) {

        m_display.height = height;
        m_local.height = height;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param width	 The new width of the graph in display units.
 */
    public void setWidth(int width) {

        m_display.width = width;
        m_local.width = width;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return a copy of the Rectangle discribing the dimensions of the frame area.
 */
    public Rectangle2D.Double getMediaFrameRect() {

        m_mediaFrame.x = m_frame.x;
        m_mediaFrame.width = m_frame.width;
        m_mediaFrame.y = m_frame.y;
        m_mediaFrame.height = m_frame.height;
        if (m_frame.width <= 1 && m_frame.height <= 1) {
            m_mediaFrame.x *= (double)m_display.width;
            m_mediaFrame.width *= (double)m_display.width;
            m_mediaFrame.y *= (double)m_display.height;
            m_mediaFrame.height *= (double)m_display.height;
        }

        return new Rectangle2D.Double(m_mediaFrame.x, m_mediaFrame.y,
            m_mediaFrame.width, m_mediaFrame.height);
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param frame Rectangle discribing the dimensions of the media frame area.
 */
    public void setMediaFrameRect(Rectangle2D.Double frame) {

        m_mediaFrame.x = frame.x;
        m_mediaFrame.width = frame.width;
        m_mediaFrame.y = frame.y;
        m_mediaFrame.height = frame.height;

        // Update the frame proportions to fit the new dimensions.
        if (m_mediaFrame.width > 1 || m_mediaFrame.height > 1) {
            m_frame.x = m_mediaFrame.x / (float)m_display.width;
            m_frame.width = m_mediaFrame.width / (float)m_display.width;
            m_frame.y = m_mediaFrame.y / (float)m_display.height;
            m_frame.height = m_mediaFrame.height / (float)m_display.height;
        }
    }
}



