/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2000-10-30
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.ps;

import gov.usgs.gr.*;
import java.util.*;
import javax.swing.*;

//==============================================================================
/**
 * An object that has properties and contains other objects in a hierarchy.
 * It can be displayed in a tree or written as XML.
 */
public class PsOptions extends gov.usgs.gr.GrObject {

    // Print options.  Keeping them in the super class allows them to be
    // specified for a page while it is being displayed on the screen before
    // a PageOfPsGraphs objects is cloned.  They will stay in effect for
    // multiple prints.
    /** The output file name. */
    protected String m_psFileName = "gr.ps";
    /** Whether this graph should be drawn in landscape. */
    protected boolean m_bPsLandscape = true;
    /** The font to use for printing. */
    protected String m_psFont = "Helvetica-Bold";
    /** The height of the print font. */
    protected float m_psFontSize = 8;
    /** Whether the curves should be drawn in color. */
    protected boolean m_bPsColor = false;
    /** Whether the background should be drawn in color. */
    protected boolean m_bPsColorBg = false;
    /** The thickness of the basic line in points. */
    protected float m_psLineWidth = 0.5f;
    /** Whether all lines should be drawn the same width. */
    protected boolean m_bPsUniformWidth = false;
    /** The size of symbols in points. */
    protected int m_psSymbolSize = 5;

    /** Whether to use line point reduction algorithm 1 when printing. */
    protected boolean m_bUseLinePointReduction1 = true;
    /** The width scale to use with line point reduction algorithm 1. */
    protected double m_LPR1WidthScale = 0.25;
    /** Whether to use the symbol point reduction algorithm when printing. */
    protected boolean m_bUseSymbolPointReduction = true;
    /** The width scale to use with the symbol point reduction algorithm. */
    protected double m_SPRWidthScale = 0.25;
    /**
     * Whether to draw thinned and unthinned curves over each other to help
     * debug.
     */
    public boolean m_bDebugPointReduction = false;
    /** Whether to show the print dialog. */
    protected boolean m_bShowPrintDialog = true;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/PsOptions.gif";

//------------------------------------------------------------------------------
/**
 * Construct an empty object.
 */
    public PsOptions () {

        setDefaults("gov.usgs.gr.visual.ps.PsOptions");
    }

//------------------------------------------------------------------------------
/**
 * Write a description of the object.
 */
    public String toString() {

    	String s = "PostScript Options";

        return s;
 	}

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("psFileName", new GrProperty(m_psFileName,
        	"File Name"));
        props.put("psLandscape", new GrProperty(
            m_bPsLandscape, "Landscape Orientation"));

        // Give the user some font choices...
        GrProperty p = new GrProperty(m_psFont, "Font");
        JComboBox jcb = new javax.swing.JComboBox();
        jcb.setEditable(true);
        boolean bFontFound = false;
        String[] fontPossibilities = {
        	"Times-Roman", "Times-Italic", "Times-Bold", "Times-BoldItalic",
            "Helvetica", "Helvetica-Oblique", "Helvetica-Bold",
            "Helvetica-BoldOblique", "Courier", "Courier-Oblique",
            "Courier-Bold", "Courier-BoldOblique",
            };
        for (int i = 0; i < fontPossibilities.length; i++) {
        	String name = fontPossibilities[i];
        	jcb.addItem(name);
            if (name.equals(m_psFont)) {
				jcb.setSelectedIndex(i);
                bFontFound = true;
            }
        }

        if (!bFontFound) {
        	jcb.addItem(m_psFont);
            jcb.setSelectedIndex(fontPossibilities.length);
        }

        p.component = jcb;
		props.put("psFont", p);

        props.put("psFontSize", new GrProperty(m_psFontSize, "Font Size"));
        props.put("psColor", new GrProperty(m_bPsColor, "Color Curves"));
        props.put("psColorBg", new GrProperty(m_bPsColorBg, "Color Background"));
        props.put("psLineWidth", new GrProperty(m_psLineWidth, "Line Width"));
        props.put("psUniformWidth", new GrProperty(
            m_bPsUniformWidth, "Uniform Line Width"));
        props.put("psSymbolSize", new GrProperty(
        	new Float(m_psSymbolSize), "Symbol Size"));
        props.put("useLinePointReduction1", new GrProperty(
            m_bUseLinePointReduction1, "Reduce Points on Lines"));
        props.put("LPR1WidthScale", new GrProperty(
            m_LPR1WidthScale, "Line Reduction Tolerance"));
        props.put("useSymbolPointReduction", new GrProperty(
            m_bUseSymbolPointReduction, "Reduce Points on Symbols"));
        props.put("SPRWidthScale", new GrProperty(
            m_SPRWidthScale, "Symbol Reduction Tolerance"));
        props.put("debugPointReduction", new GrProperty(
            m_bDebugPointReduction, "Debug Point Reduction"));
        props.put("showPrintDialog", new GrProperty(
            m_bShowPrintDialog, "Show Print Dialog"));
        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

            	if (key.equals("psFileName"))
                    m_psFileName = (String)p.value;
            	if (key.equals("psLandscape"))
                    m_bPsLandscape =(Boolean)p.value;
            	if (key.equals("psFont"))
                    m_psFont = (String)p.value;
            	if (key.equals("psFontSize"))
                    m_psFontSize = (Float)p.value;
            	if (key.equals("psColor"))
                    m_bPsColor = (Boolean)p.value;
            	if (key.equals("psColorBg"))
                    m_bPsColorBg = (Boolean)p.value;
            	if (key.equals("psLineWidth"))
                    m_psLineWidth = (Float)p.value;
            	if (key.equals("psUniformWidth"))
                    m_bPsUniformWidth = (Boolean)p.value;
            	if (key.equals("psSymbolSize"))
                    m_psSymbolSize = Math.round((Float)p.value);
            	if (key.equals("useLinePointReduction1"))
            		m_bUseLinePointReduction1 = (Boolean)p.value;
            	if (key.equals("LPR1WidthScale"))
                    m_LPR1WidthScale = (Double)p.value;
            	if (key.equals("useSymbolPointReduction"))
            		m_bUseSymbolPointReduction = (Boolean)p.value;
            	if (key.equals("SPRWidthScale"))
                    m_SPRWidthScale = (Double)p.value;
            	if (key.equals("debugPointReduction"))
                    m_bDebugPointReduction = (Boolean)p.value;
                if (key.equals("showPrintDialog"))
                    m_bShowPrintDialog = (Boolean)p.value;
            }
            if (obj instanceof String) {
            	String value = (String)obj;

                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("psFileName"))
            m_psFileName = value;
        if (key.equals("psLandscape"))
            m_bPsLandscape = Boolean.parseBoolean(value);
        if (key.equals("psFont"))
            m_psFont = value;
        if (key.equals("psFontSize"))
            m_psFontSize = Float.parseFloat(value);
        if (key.equals("psColor"))
            m_bPsColor = Boolean.parseBoolean(value);
        if (key.equals("psColorBg"))
            m_bPsColorBg = Boolean.parseBoolean(value);
        if (key.equals("psLineWidth"))
            m_psLineWidth = Float.parseFloat(value);
        if (key.equals("psUniformWidth"))
            m_bPsUniformWidth = Boolean.parseBoolean(value);
        if (key.equals("psSymbolSize"))
            m_psSymbolSize = Math.round(Float.parseFloat(value));
        if (key.equals("useLinePointReduction1"))
        	m_bUseLinePointReduction1 = Boolean.parseBoolean(value);
        if (key.equals("LPR1WidthScale"))
            m_LPR1WidthScale = (Double.parseDouble(value));
        if (key.equals("useSymbolPointReduction"))
        	m_bUseSymbolPointReduction = Boolean.parseBoolean(value);
        if (key.equals("SPRWidthScale"))
            m_SPRWidthScale = (Double.parseDouble(value));
        if (key.equals("debugPointReduction"))
            m_bDebugPointReduction = Boolean.parseBoolean(value);
        if (key.equals("showPrintDialog"))
            m_bShowPrintDialog = Boolean.parseBoolean(value);
	}

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The name of the PostScript output file.
 */
    public String getPsFileName() {
    	return new String(m_psFileName);
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param name	The name of the PostScript output file.
 */
    public void setPsFileName(String name) {
    	m_psFileName = name;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return True if orientation is to be landscape instead of portrait.
 */
    public boolean getPsLandscape() {
    	return m_bPsLandscape;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	True if orientation is to be landscape instead of portrait.
 */
    public void setPsLandscape(boolean value) {
        m_bPsLandscape = value;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return True if PostScript output should be in color.
 */
    public boolean getPsColor() {
    	return m_bPsColor;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	True if PostScript output should be in color.
 */
    public void setPsColor(boolean value) {
        m_bPsColor = value;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return True if PostScript output should be in color.
 */
    public boolean getPsColorBg() {
    	return m_bPsColorBg;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	True if PostScript output should be in color.
 */
    public void setPsColorbg(boolean value) {
        m_bPsColorBg = value;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return True if all lines are to be drawn the same width.
 */
    public boolean getPsUniformWidth() {
    	return m_bPsUniformWidth;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	True if all lines are to be drawn the same width.
 */
    public void setPsUniformWidth(boolean value) {
        m_bPsUniformWidth = value;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The size of PostScript symbols in points.
 */
    public int getPsSymbolSize() {
    	return m_psSymbolSize;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	The size of PostScript symbols in points.
 */
    public void setPsSymbolSize(int value) {
        m_psSymbolSize = value;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The width of PostScript lines in points.
 */
    public float getPsLineWidth() {
    	return m_psLineWidth;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	The width of PostScript lines in points.
 */
    public void setPsLineWidth(float value) {
        m_psLineWidth = value;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The height of PostScript text in points.
 */
    public float getPsFontSize() {
    	return m_psFontSize;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param value	The height of PostScript text in points.
 */
    public void setPsFontSize(float value) {
        m_psFontSize = value;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The name of the PostScript font.
 */
    public String getPsFont() {
    	return new String(m_psFont);
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param name The name of the PostScript font.
 */
    public void setPsFont(String name) {
    	m_psFont = name;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return Whether to use LPR1.
 */
    public boolean getUseLinePointReduction1() {
    	return m_bUseLinePointReduction1;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param val Whether to use LPR1.
 */
    public void setUseLinePointReduction1(boolean val) {
    	m_bUseLinePointReduction1 = val;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return Whether to use SPR.
 */
    public boolean getUseSymbolPointReduction() {
    	return m_bUseLinePointReduction1;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param val Whether to use SPR.
 */
    public void setUseSymbolPointReduction(boolean val) {
    	m_bUseSymbolPointReduction = val;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The width scale for point LPR1.
 */
    public double getLPR1WidthScale() {
    	return m_LPR1WidthScale;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param val The width scale for point LPR1.
 */
    public void setLPR1WidthScale(double val) {
    	m_LPR1WidthScale = val;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The width scale for point SPR.
 */
    public double getSPRWidthScale() {
        return m_SPRWidthScale;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param val The width scale for point SPR.
 */
    public void setSPRWidthScale(double val) {
    	m_SPRWidthScale = val;
    }
}




