/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2005-11-04
 * Date started:    1999-01-21
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.gr.visual.ps;

import java.io.*;
import java.text.*;
import java.util.*;
import gov.usgs.gr.visual.*;

//==============================================================================
/**
 * The Y version of the axis object.
 */
public class PsYAxis extends PsAxis implements PsElement {


//------------------------------------------------------------------------------
/**
 * Empty constructor.
 */
    public PsYAxis () {

    	m_dimension = 1;

        setDefaults("gov.usgs.gr.visual.ps.PsYAxis");
    }

//------------------------------------------------------------------------------
/**
 * Comment.
 * @param limits	The limits object telling what segment of the axis should
 *					be represented.
 * @param opposingLimits	The opposing axis' limits.  Used to determine the
 *							how the length of tick marks when drawing in
 *							graph coordinates.
 */
    public PsYAxis (Zoom zoom, Zoom opposingZoom) {

        super(zoom, opposingZoom);
    	m_dimension = 1;

        setDefaults("gov.usgs.gr.visual.ps.PsYAxis");
    }

//------------------------------------------------------------------------------
/**
 * Create a new axis by copying an existing one.
 * @param axis	The axis to be copied.
 */
    public PsYAxis (Axis axis) {

    	super(axis);
    	m_dimension = 1;
    }

//------------------------------------------------------------------------------
/**
 * Overrides the draw() function from Axis.  Rotates the axis so it will
 * be drawing vertically.
 * <p><b> Pre Conditions </b><br>
 *      ps is a valid output file.
 * <p><b> Post Conditions </b><br>
 *      None.
 */
    public void drawPS() {

	    double x;
        DecimalFormat form = new DecimalFormat("0.##");

        // Save the current transformation and then move the axis to its
        // proper location rotated 90 degrees.
	    ps.println("\n% Set up for drawing y-axis");
	    ps.println("gsave");
	    ps.println("90 rotate");
	    ps.println("0 " + form.format(-m_local.height) + " translate");

        // Let the super class do the rest of the work.
        super.drawPS();

	    // Restore previous settings.
	    ps.println("\n% Done drawing y-axis.  Restore previous transformation");
	    ps.println("grestore");

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Draws labels in magenta beside each major tick mark.
 * <p><b> Pre Conditions </b><br>
 *      gl is a valid OpenGL context.
 * <p><b> Post Conditions </b><br>
 *      None.
 */
    public void drawText(String font, double size, double xOrig, double yOrig,
    	double axisDisplayWidth) {

        // Return if this text shouldn't be shown.
        if (!m_bShowText) return;
        if (m_majorTickInc <= 0) return;

        DecimalFormat form = new DecimalFormat("0.##");
        double x;

        // Set up the font.
		ps.println("\n% Set the font for the axis titles");
		ps.println("/" + font + " findfont " + form.format(size)
        	+ " scalefont setfont");

        // Save the previous location and move to the starting point.
        ps.println("gsave");
        ps.println(form.format(xOrig - (size * 0.6)) + " " + form.format(yOrig)
        	+ " translate");
        double displayScale = axisDisplayWidth / (m_max - m_min);

        // Get the decimal format that should be used for these labels.
        DecimalFormat charForm = new DecimalFormat(m_zoom.getFormatString());

        // Loop through and draw each label.
        for (x = m_majorStart; (x - m_max) <= (0.01 * m_majorTickInc);
        	 x += m_majorTickInc) {

            if (Math.abs(x - 0) <= (0.01 * m_majorTickInc)) x = 0;
            double pageY = (double)(displayScale * (x - m_min) - (0.35 * size));
            ps.println("(" + charForm.format(x) + ") 0 " + form.format(pageY)
            	+ " rightshow");
        }

        // Draw the axis title.
        if (m_bShowTitle) {
        	ps.println((-size * 3) + " 0 translate");
            double pageY = (double)(displayScale * (m_max - m_min) / 2);

            StringTokenizer st = new StringTokenizer(m_title, "" + '\n');
        	ps.println(form.format(-size * 1.1 * (double)st.countTokens())
            	+ " 0 translate");
            while (st.hasMoreTokens()) {
        		ps.println(form.format(size * 1.1) + " 0 translate");

                String text = st.nextToken().replaceAll("\\(", "\\\\050")
                    .replaceAll("\\)", "\\\\051");
            	ps.println("(" + text + ") 0 " + form.format(pageY)
                	+ " yshow");
            }
        }

        // Restore the old transformation.
        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Y Axis";

        return s;
 	}
}



