/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-07-06
 * Date started:    2007-07-03
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.data.*;
import gov.usgs.sfhydro.si3d.*;
import gov.usgs.sfhydro.si3d.gui.*;
import java.io.*;

//==============================================================================
/**
 * An array of data in one or more dimensions.
 * @see gov.usgs.data.XyData.
 */
public class ArcGridAsciiFile extends gov.usgs.sfhydro.data.Si3dBathymetryFile {
    
//------------------------------------------------------------------------------
    /**
     * Creates a data file.
     */
    public ArcGridAsciiFile() {
        
        super();
        m_title = "Arc Grid Ascii File";
        
        m_min = new int[m_numDimensions];
        m_max = new int[m_numDimensions];
        
        m_numHeaderLines = 6;
        m_inputDelimiters = " ";
        m_bFileInMeters = false;
        m_bDisplayInMeters = true;
        m_fileMultiplier = -1;
        m_displayMultiplier = 1;
        m_fileShift = 0;
        m_displayShift = 0;
        m_bEnableFileFields = true;
        m_bEnableDisplayFields = true;
    }
   
//------------------------------------------------------------------------------
    /**
     *
     */
    public static boolean matchesFormat(String[] lines) {
        
        boolean bNcolsFound = false;
        boolean bNrowsFound = false;
        boolean bNodataFound = false;
        
        // Keep track of whether we have passed the header.
        boolean bPastHeader = false;
        
        try {
            // Read each line of the data file.
            for (int i = 0; i < lines.length; i++) {
                if (i > 5)
                    break;
                
                if (lines[i].indexOf("ncols") >= 0)
                    bNcolsFound = true;
                if (lines[i].indexOf("nrows") >= 0)
                    bNrowsFound = true;
                if (lines[i].indexOf("NODATA_value") >= 0)
                    bNodataFound = true;
            }
        } catch (Exception e) {
            return false;
        }
        
        if (bNcolsFound && bNrowsFound && bNodataFound)
            return true;
        else
            return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Read the header information from an open file attached to a reader.
     */
    protected void readHeader() throws IOException {
        
        for (int i = 0; i < m_numHeaderLines; i++) {
            m_headerLines.add(AsciiIO.readLine(m_br));
        }
        
        String[] tokens = AsciiIO.parseLine(m_headerLines.get(0),
            m_inputDelimiters);
        m_numI = Integer.parseInt(tokens[1]);
        tokens = AsciiIO.parseLine(m_headerLines.get(1),
            m_inputDelimiters);
        m_numJ = Integer.parseInt(tokens[1]);
        tokens = AsciiIO.parseLine(m_headerLines.get(4),
            m_inputDelimiters);
        m_gridWidth = Double.parseDouble(tokens[1]);
        tokens = AsciiIO.parseLine(m_headerLines.get(5),
            m_inputDelimiters);
        m_noDataIndicator = tokens[1];
        
        m_headerLines.clear();
    }
    
//------------------------------------------------------------------------------
    /**
     * Read the data through a reader attached to an open file.  Assumes the
     * header has been read with readHeader() and the file needs to be closed.
     */
    protected void readData() throws IOException {
        
        clear();
        
        // Make a big data series for storage.
        DoubleDataSeries triFanSeries = new DoubleDataSeries(3);
        DoubleDataSeries rectSeries = new DoubleDataSeries(3);
        double[] array = new double[m_numI * m_numJ];
        double[] cornerDepths = new double[m_numI * m_numJ];
        
        int lineNum = 0;
        String[] vals;
        
        // Read each line of the data file.
        while (true) {
            String textLine = m_br.readLine();
            if (textLine == null) break;
            
            vals = AsciiIO.parseLine(textLine, m_inputDelimiters);
            if (vals == null) break;
            if (vals.length <= 0)
                continue;
            
            for (int i = 1; i < vals.length; i++) {
                int iOffset = i;
                int jOffset = m_numJ  - lineNum - 1;
                double depth = Double.parseDouble(vals[i]);
                depth -= m_fileShift;
                depth /= m_fileMultiplier;
                if (m_bFileInMeters && !m_bDisplayInMeters)
                    depth *= 3.2808399;
                if (!m_bFileInMeters && m_bDisplayInMeters)
                    depth *= 0.3048;
                depth *= m_displayMultiplier;
                depth += m_displayShift;
                
                if (vals[i].equals(m_noDataIndicator))
                    cornerDepths[iOffset + (jOffset * m_numI)] = Double.NaN;
                else
                    cornerDepths[iOffset + (jOffset * m_numI)] = depth;
            }
            
            lineNum++;
        }
        
        // Put the 1D corner depth series into a GlBathymetryGrid and add the
        // grid to this file.
        DoubleDataSeries cornerDepthSeries = new DoubleDataSeries(1);
        cornerDepthSeries.setArray(cornerDepths);
        RawBathymetry bath = new RawBathymetry();
        bath.setProperty("title", m_rawBathTitle);
        bath.setProperty("xSpacing", "" + m_gridWidth);
        bath.setProperty("ySpacing", "" + m_gridWidth);
        bath.setProperty("zSpacing", "2");
        bath.setProperty("numI", "" + m_numI);
        bath.setProperty("numJ", "" + m_numJ);
        bath.add(cornerDepthSeries);
        gov.usgs.sfhydro.si3d.gl.GlBathymetryGrid glGrid =
            new gov.usgs.sfhydro.si3d.gl.GlBathymetryGrid();
        glGrid.add(bath);
        glGrid.regenerateColorScheme();
        glGrid.setProperty("zScale", "0");
        gov.usgs.gr.visual.gl.GlGraph graph =
            new gov.usgs.gr.visual.gl.GlGraph();
        graph.add(glGrid);
        gov.usgs.gr.visual.gl.PageOfGlGraphs page =
            new gov.usgs.gr.visual.gl.PageOfGlGraphs();
        page.add(graph);
        page.updateZoomContents();
        page.showMoreGraphs();
        page.setProperty("formatter",
            "gov.usgs.sfhydro.gr.formats.VariableSizeFormatter");
        gov.usgs.gr.visual.Axis xAxis = graph.getXAxis();
        gov.usgs.gr.visual.Axis yAxis = graph.getYAxis();
        xAxis.setProperty("showMajorTicks", "false");
        yAxis.setProperty("showMajorTicks", "false");
        xAxis.setProperty("showMinorTicks", "false");
        yAxis.setProperty("showMinorTicks", "false");
        xAxis.setProperty("showLabels", "false");
        yAxis.setProperty("showLabels", "false");
        xAxis.setProperty("showGridLines", "false");
        yAxis.setProperty("showGridLines", "false");
        xAxis.getZoom().setOuterLimits(0, m_numI * m_gridWidth);
        yAxis.getZoom().setOuterLimits(0, m_numJ * m_gridWidth);
        xAxis.getZoom().zoomAll();
        yAxis.getZoom().zoomAll();
        xAxis.getZoom().setContentLimits(0, m_numI * m_gridWidth);
        yAxis.getZoom().setContentLimits(0, m_numJ * m_gridWidth);
        add(page);
        add(glGrid);
        add(bath);
        RawBathymetryCalculator calc = new RawBathymetryCalculator();
        calc.add(bath);
        add(calc);
        add(new BarrierEditorProvider());
    }
}
