/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    2000-08-22
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.data.*;
import java.io.*;
import java.util.*;

//==============================================================================
/**
 * Reads data in the format specified by the Bay-Delta Modeling Forum.
 */
public class BDMFDataFile extends gov.usgs.gr.data.DelimitedDataFile {

//------------------------------------------------------------------------------
/**
 * Creates an empty data file.
 */
    public BDMFDataFile () {

    	super();

        m_title = "BDMF Data File";
        m_numChannels = 2;
    }

//------------------------------------------------------------------------------
/**
 * Read the column data through a reader attached to an open file.
 * Overrides readData() from gov.usgs.data.DimensionalDataFile.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 * @throws IOException on IO error.
 */
    protected void readData () throws IOException {

    	clear();

        // Make a list of column lists, one for each series.
        List<List<Double>> columnList = new ArrayList();
        for (int i = 0; i < m_numChannels; i++) {
            columnList.add(new ArrayList<Double>());
        }

    	double[] vals;
        int seriesIndex = 0;

        List<Double> column0 = (List)columnList.get(0);
        List<Double> column1 = (List)columnList.get(1);

        // Read each line of the data file.
		boolean bLookingForTime = false;
		while (true) {
        	vals = AsciiIO.parseLineAsDoubles(m_br.readLine(),
                m_inputDelimiters);
            if (vals == null) break;

            // Set up to read a repitition of two lines with one value each
            // followed by multiple lines with multiple values each.  The second
            // single-value line is time.  When the first value of a multivalue
            // line is 50000, the third value on that line is stored as the
            // second dimension of the data series.
            if (bLookingForTime) {
            	if (vals.length == 1) {
                	column0.add(vals[0]);
                	column1.add(vals[0]);
                    bLookingForTime = false;
                }
            }
            else {
            	if (vals.length < 3) {
                    bLookingForTime = true;
                }
                else {
                	if (vals[0] == 50000) {
//                	if (vals[0] == 0)
                		column0.add(vals[2]);
                		column1.add(vals[4]);
					}
                }
            }
        }

        // Transfer the Doubles to an array of doubles and create data series.
        seriesIndex = 0;
        for (List<Double> column : columnList) {
            Object[] realObjects = column.toArray();
            vals = new double[realObjects.length];
            for (int i = 0; i < realObjects.length; i++) {
            	vals[i] = (Double)realObjects[i];
            }

            // Create a new data series from the array.
        	DoubleDataSeries series = null;
        	if (m_numDimensions == 2)
		    series = new DoubleDataSeries(vals, 2);
        	else
		    series = new DoubleDataSeries(vals, m_numDimensions);
            if (seriesIndex == 0) series.setTitle("Flow in CFS");
            if (seriesIndex == 1) series.setTitle("Cross Secional Area");
            add(series);

            seriesIndex++;
        }
    }

//------------------------------------------------------------------------------
/**
 * Output for this data format is not supported.  This stub overrides the
 * DelimitedDataFile output method.
 * @throws IOException on IO error.
 */
    protected void writeHeader () throws IOException {

    }

//------------------------------------------------------------------------------
/**
 * Output for this data format is not supported.  This stub overrides the
 * DelimitedDataFile output method.
 * @throws IOException on IO error.
 */
    protected void writeData () throws IOException {

    }
}
