/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-06-28
 * Date started:    2006-03-29
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */
package gov.usgs.sfhydro.data;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import java.lang.reflect.*;
import java.awt.EventQueue;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;

//==============================================================================
/**
 * A data file object for reading data in CR1000 format.
 */
public class CR1000DataFile extends gov.usgs.gr.data.TimeDataFile {

    /** The indexes of the year, day, and time columns. */
    protected int[] m_dateColumns = {1};
    /** The indexes of the variables to keep. */
    protected int[] m_keepers;
    /** The numeric type of each time series. */
    protected int[] m_types;
    /** The title of each time series. */
    protected String[] m_titles;
    /**
     * Whether to show a dialog when giving notification messages to
     * the user.
     */
    protected boolean m_bUseDialogNotification = true;
    /** A list of CR10Options objects from which to choose the input options. */
    protected List<CR1000Options> m_preConfigs = new ArrayList<CR1000Options>();
    /**
     * The index into the list of preConfigs that describes the
     * current config.
     */
    protected CR1000Options m_selectedConfig = new CR1000Options();
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/CR1000DataFile.gif";

    /** An inner class for holding configuration options. */
    protected class CR1000Options extends GrObject {

        /** The title of the CR1000 file. */
        public String title = "";
        /** The number of channels in the CR1000 file. */
        public int numChannels = 0;
        /** The columns that hold the year, day, and time.  Comma delimited. */
        public String dateColumns = "1";
        /** The channels to keep when reading.  Comma delimited. */
        public String keepers = "";
        /** The data types for each keeper channel.  Comma delimited. */
        public String types = "";
        /** The title for each keeper channel.  Comma delimited. */
        public String titles = "";

        //----------------------------------------------------------------------
        @Override
        public GrHashList getProperties() {

            GrHashList props = super.getProperties();

            props.put("title", new GrProperty(title, "Title"));
            props.put("numChannels", new GrProperty(numChannels,
                    "Total Number of Channels"));
            props.put("dateColumn", new GrProperty(dateColumns,
                    "Date Column"));
            props.put("keepers", new GrProperty(keepers, "Channels to Keep"));
            props.put("types", new GrProperty(types, "Channel Types"));
            props.put("titles", new GrProperty(titles, "Channel Titles"));

            return props;
        }
    }

//------------------------------------------------------------------------------
    /**
     * Create an empty data file object.
     */
    public CR1000DataFile() {

        super();

        m_numHeaderLines = 4;

        m_inputDelimiters = "\n\r\f,";

        // Set up a default options object.
        CR1000Options standardOptions = new CR1000Options();
        standardOptions.title = "Standard";
        standardOptions.numChannels = 14;
        standardOptions.dateColumns = "1";
        standardOptions.keepers = "2";
        standardOptions.types = "1";
        standardOptions.titles = "Index";

        // Add the new object to the list of preConfigs.
        m_preConfigs.add(standardOptions);
        setConfig("Standard");

        // Check if any default CR1000 options were set at start up from a
        // config file in the Gr/config directory.
        setDefaults("gov.usgs.sfhydro.data.CR1000DataFile");
        if (m_preConfigs.size() > 0) {
            setConfig(m_preConfigs.get(0).title);
        }
    }

//------------------------------------------------------------------------------
    /**
     * Looks for the preset configuration with the specified name and sets the
     * member variables to match those values.
     * @param name  The name of the preConfig to use.
     */
    public void setConfig(String name) {

        Iterator iter = m_preConfigs.iterator();
        while (iter.hasNext()) {
            CR1000Options config = (CR1000Options) iter.next();
            if (config.title.equals(name)) {
                m_selectedConfig = config;
                m_title = m_selectedConfig.title;
                m_dateColumns = AsciiIO.parseLineAsInts(
                        m_selectedConfig.dateColumns, ", \t");
                m_keepers = AsciiIO.parseLineAsInts(
                        m_selectedConfig.keepers, ", \t");
                m_types = AsciiIO.parseLineAsInts(
                        m_selectedConfig.types, ", \t");
                m_titles = AsciiIO.parseLine(m_selectedConfig.titles, ",");

                break;
            }
        }
    }

//------------------------------------------------------------------------------
    @Override
    public void showOptionsDialog(Frame frame) {

        // Create an instance of the inner dialog class that presents the
        // user with a GUI for modifying file options.
        final CR1000ConfigDialog dialog = new CR1000ConfigDialog(frame,
                "CR1000 Config", true);

        // Find the index of the currently selected preConfig.
        int index = 0;
        Iterator iter = m_preConfigs.iterator();
        while (iter.hasNext()) {
            CR1000Options config = (CR1000Options) iter.next();
            if (config == m_selectedConfig) {
                break;
            }
            index++;
        }
        if (index >= m_preConfigs.size()) {
            index = 0;
        }

        // Show the configuration dialog.
        dialog.setConfigs(m_preConfigs, index);

        if (EventQueue.isDispatchThread()) {
            dialog.setVisible(true);
        } else {
            try {
                javax.swing.SwingUtilities.invokeAndWait(new Runnable() {

                    public void run() {
                        dialog.setVisible(true);
                    }
                });
            } catch (InvocationTargetException ex) {
            } catch (InterruptedException ex) {
            }
        }

        if (!(dialog.m_bOkPressed)) {
            return;
        }

        // Find out which preset the user chose.
        index = dialog.presetCombo.getSelectedIndex();
        CR1000Options comboConfig = (CR1000Options) m_preConfigs.get(index);

        // Copy the value from the dialog box.
        String dateText = dialog.dateField.getText();
        String keepText = dialog.keepField.getText();
        String typesText = dialog.typesField.getText();
        String titlesText = dialog.titlesField.getText();

        // Compare the values in the dialog to the currently selected preConfig.
        if (dateText.equals(comboConfig.dateColumns) &&
                keepText.equals(comboConfig.keepers) &&
                typesText.equals(comboConfig.types) &&
                titlesText.equals(comboConfig.titles)) {

            // The chosen preConfig was not customized, so apply it as is.
            setConfig(comboConfig.title);
        } else {
            // The user modified one of the fields, so add a
            // custom configuration and apply it.
            CR1000Options customConfig = new CR1000Options();
            customConfig.title = "Custom";
            customConfig.dateColumns = dateText;
            customConfig.keepers = keepText;
            customConfig.types = typesText;
            titlesText = titlesText.replace('(', '[');
            titlesText = titlesText.replace(')', ']');
            customConfig.titles = titlesText;
            m_preConfigs.add(customConfig);

            setConfig(customConfig.title);
        }
    }

//------------------------------------------------------------------------------
    @Override
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        // The value of each property is a comma delimited string.
        String s = "";
        for (int i = 0; i < m_dateColumns.length; i++) {
            if (i > 0) {
                s += ", ";
            }
            s += m_dateColumns[i];
        }
        props.put("dateColumn",
                new GrProperty(s, "Date Column"));

        s = "";
        for (int i = 0; i < m_keepers.length; i++) {
            if (i > 0) {
                s += ", ";
            }
            s += m_keepers[i];
        }
        props.put("keepers", new GrProperty(s, "Channels to Keep"));

        s = "";
        for (int i = 0; i < m_types.length; i++) {
            if (i > 0) {
                s += ", ";
            }
            s += m_types[i];
        }
        props.put("types", new GrProperty(s, "Channel Types"));

        s = "";
        for (int i = 0; i < m_titles.length; i++) {
            if (i > 0) {
                s += ", ";
            }
            s += m_titles[i];
        }
        props.put("titles", new GrProperty(s, "Channel Titles"));

        return props;
    }

//------------------------------------------------------------------------------
    @Override
    public void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        // Iterate through the list of properties.
        Iterator iter = props.keySet().iterator();
        if (props instanceof GrHashList) {
            iter = ((GrHashList) props).listIterator();
        }
        while (iter.hasNext()) {
            String key = (String) iter.next();
            Object obj = props.get(key);

            // Find the String value let the setProperty method handle it.
            if (obj instanceof GrProperty) {
                p = (GrProperty) obj;

                if (p.value instanceof String) {
                    setProperty(key, (String) p.value);
                }
            }
            if (obj instanceof String) {
                String value = (String) obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
    @Override
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("dateColumn")) {
            m_dateColumns = AsciiIO.parseLineAsInts(value, ", \t");
        }
        if (key.equals("keepers")) {
            m_keepers = AsciiIO.parseLineAsInts(value, ", \t");
        }
        if (key.equals("types")) {
            m_types = AsciiIO.parseLineAsInts(value, ", \t");
        }
        if (key.equals("titles")) {
            m_titles = AsciiIO.parseLine(value, ",");
        }
        if (key.startsWith("preConfig") || key.startsWith("setConfig")) {
            int index1 = value.indexOf('(');
            int index2 = value.indexOf(')', index1 + 1);
            CR1000Options preset = new CR1000Options();
            preset.title = value.substring(index1 + 1, index2);
            if (preset.title.equals("CLEAR")) {
                m_preConfigs.clear();
                return;
            }
            index1 = value.indexOf('(', index2 + 1);
            index2 = value.indexOf(')', index1 + 1);
            preset.dateColumns = value.substring(index1 + 1, index2);
            index1 = value.indexOf('(', index2 + 1);
            index2 = value.indexOf(')', index1 + 1);
            preset.keepers = value.substring(index1 + 1, index2);
            index1 = value.indexOf('(', index2 + 1);
            index2 = value.indexOf(')', index1 + 1);
            preset.types = value.substring(index1 + 1, index2);
            index1 = value.indexOf('(', index2 + 1);
            index2 = value.indexOf(')', index1 + 1);
            preset.titles = value.substring(index1 + 1, index2);
            m_preConfigs.add(preset);

            if (key.startsWith("setConfig")) {
                setConfig(preset.title);
            }
        }
        if (key.startsWith("useDialogNotification")) {
            m_bUseDialogNotification = Boolean.parseBoolean(value);
        }
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0) {
            setTreeIcon();
        }

        return (javax.swing.ImageIcon) TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null) {
            icon = super.setTreeIcon();
        }

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
    @Override
    protected void readData() throws IOException {

        // Remove any data that were children of this object.
        clear();

        m_titles = new String[m_keepers.length];
        for (int i = 0; i < m_titles.length; i++) {
            m_titles[i] = "";
        }
        try {
            String titleLine = (String) m_headerLines.get(1);
            String[] tokens = AsciiIO.parseLine(titleLine, m_inputDelimiters);
            for (int i = 0; i < m_titles.length; i++) {
                String title = tokens[m_keepers[i] - 1];
                int quote1Index = title.indexOf('"');
                int quote2Index = 0;
                if (quote1Index >= 0) {
                    quote2Index = title.indexOf('"', quote1Index + 1);
                    if (quote2Index >= 0) {
                        title = title.substring(quote1Index + 1, quote2Index);
                    }
                }
                m_titles[i] = title;
            }
        } catch (Exception e) {
        }

        // Determine minimum number of channels to read.
        m_numChannels = m_keepers.length + 1;

        // Make a list of column lists, one for each series.
        List columnList = new ArrayList();
        for (int i = 0; i < m_numChannels; i++) {
            columnList.add(new ArrayList());
        }

        // Read Year, Day of Year, and Hour/Minute from the specified columns.
        List dateColumn = (List) columnList.get(m_dateColumns[0] - 1);

        // Declare and intialize variables.
        String[] tokens;
        long lastMillis = 0;
        int numPointsRemoved = 0;

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        String ymdhms = "yyyy/MM/dd HH:mm:ss";
        String ymdhm = "yyyy/MM/dd HH:mm";

        // Read each line of the data file.
        int lineNumber = m_numHeaderLines;
        int firstOutOfOrderLine = -1;
        try {
            while ((tokens = AsciiIO.parseLine(m_br.readLine(),
                    m_inputDelimiters)) != null) {
                lineNumber++;

                if (tokens.length == 0) {
                    continue;
                }

                String dateToken = tokens[m_dateColumns[0] - 1];

                int quote1Index = dateToken.indexOf('"');
                int quote2Index = 0;
                if (quote1Index >= 0) {
                    quote2Index = dateToken.indexOf('"', quote1Index + 1);
                } else {
                    // Try single quotes if there were no double quotes.
                    quote1Index = dateToken.indexOf('\'');
                    if (quote1Index >= 0) {
                        quote2Index = dateToken.indexOf('\'', quote1Index + 1);
                    }
                }

                // Skip this line if there were no quotes.
                if (quote1Index < 0 || quote2Index < 0) {
                    continue;
                }

                // Parse the date and add it to the date list.
                String dateString = dateToken.substring(quote1Index + 1, quote2Index);
                dateString = dateString.replace('-', '/');
                SimpleDateFormat formatter = null;
                if (dateString.length() > 16) {
                    formatter = new SimpleDateFormat(ymdhms);
                } else {
                    formatter = new SimpleDateFormat(ymdhm);
                }
                formatter.setTimeZone(utcZone);
                formatter.setLenient(true);
                GregorianCalendar date = new GregorianCalendar(utcZone, locale);
                date.clear();
                ParsePosition pos = new ParsePosition(0);
                date.setTime(formatter.parse(dateString, pos));
                long timeInMillis = date.getTimeInMillis();
                if (timeInMillis > lastMillis) {
                    dateColumn.add(date);
                    lastMillis = timeInMillis;
                } else {
                    numPointsRemoved++;
                    if (firstOutOfOrderLine < 1) {
                        firstOutOfOrderLine = lineNumber;
                    }
                    continue;
                }

                // Add each number on the line to the appropriate column list.
                for (int i = 0; i < m_keepers.length; i++) {
                    // Parse the token as a double value.
                    double val = Double.NEGATIVE_INFINITY;
                    if ((m_keepers[i] - 1) < tokens.length) {
                        String token = tokens[m_keepers[i] - 1];
                        if (!token.equals("-7999") && !token.equals("-999999")) {
                            try {
                                val = Double.parseDouble(token);
                            } catch (NumberFormatException e) {
                            }
                        }
                    } else {
                        val = 0;
                    }

                    // Append the new value to the appropriate column.
                    List column = (List) columnList.get(i + 1);
                    column.add(val);
                }
            }
        } catch (Exception e) {
            System.err.println("Could not read CR1000 line " + lineNumber);
            e.printStackTrace();
            return;
        }

        // If any points were removed for being out of order, notify the user.
        if (numPointsRemoved > 0) {
            String message = "Not all data points were sequential. " + numPointsRemoved + " points were removed.  See line " + firstOutOfOrderLine + ".";
            if (m_bUseDialogNotification) {
                javax.swing.JOptionPane.showMessageDialog(null, message,
                        "Notification",
                        javax.swing.JOptionPane.INFORMATION_MESSAGE);
            } else {
                System.out.println("CR1000 input: " + message);
            }
        }

        // Create a list containing a column of dates and columns for the other
        // data so they can be interleaved into individual series.
        m_startDate = (GregorianCalendar) ((GregorianCalendar) dateColumn.get(0)).clone();
        GregorianCalendar lastDate = (GregorianCalendar) ((GregorianCalendar) dateColumn.get(dateColumn.size() - 1)).clone();
        m_referenceYear = findAppropriateReferenceYear(m_startDate, lastDate);

        // Interleave the arrays.
        List<? extends TimeSeries> timeSeriesList =
                interleaveArrays(columnList);

        // Calculate the median spacing between data points.
        long mostCommonInc = medianIncrement(dateColumn);
        List<Double> deletedList = markMissingPoints(dateColumn, mostCommonInc);
        // Add the time series as children of this file.
        int index = 0;
        Iterator iter = timeSeriesList.iterator();
        while (iter.hasNext()) {
            TimeSeries series = (TimeSeries) iter.next();
            series.setType(m_types[index]);
            series.setTitle(m_titles[index]);

            // Add a copy of the deleted list to the series.
            ArrayList copyOfDeletedList = new ArrayList();
            for (double dVal : deletedList) {
                copyOfDeletedList.add(dVal);
            }

            series.setDeletedList(copyOfDeletedList);

            // Delete points with a y value of NEGATIVE_INFINITY
            double[] min = {Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY};
            double[] max = {Double.POSITIVE_INFINITY, Double.NEGATIVE_INFINITY};
            series.selectArea(min, max, true);

            series.deleteSelectedPoints();

            // Add the series as a child of this file.
            add(series);

            index++;
        }
    }

//------------------------------------------------------------------------------
    /**
     *
     */
    public static boolean matchesFormat(String[] lines) {

        int numDates = 0;
        String inputDelimiters = "\n\r\f,";

        try {
            // Read each line of the data file.
            for (int i = 0; i < lines.length; i++) {
                if (lines[i] == null) {
                    break;
                }

                String line = lines[i].trim();
                String[] tokens = AsciiIO.parseLine(line, inputDelimiters);

                if (tokens.length == 0) {
                    return false;
                }

                // After the header, try looking for tokens that resemble a date.
                for (int j = 0; j < tokens.length; j++) {
                    if (i < 4) {
                        if (tokens[j].charAt(0) != '"' ||
                                tokens[j].charAt(tokens[j].length() - 1) != '"') {

                            return false;
                        }
                    } else {
                        // Beyond header...
                        try {
                            if (tokens[j].charAt(0) == '"' &&
                                    tokens[j].charAt(tokens[j].length() - 1) == '"' &&
                                    (tokens[j].charAt(5) == '-' ||
                                    tokens[j].charAt(5) == '/') &&
                                    (tokens[j].charAt(8) == '-' ||
                                    tokens[j].charAt(8) == '/') &&
                                    tokens[j].charAt(14) == ':') {

                                numDates++;
                            }
                        } catch (Exception e2) {
                        }
                    }
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }

        if (numDates <= 0) {
            return false;
        } else {
            return true;
        }
    }

//------------------------------------------------------------------------------
    @Override
    public javax.swing.JMenuItem getMenuCommands() {

        // Create the menu item as a sub menu with other items.
        javax.swing.JMenu subMenu = new javax.swing.JMenu("CR1000 Data File");

        // Set up menu items that use this object as their action listeners.
        javax.swing.JMenuItem saveAsItem = new javax.swing.JMenuItem("Save As");
        saveAsItem.addActionListener(this);
        subMenu.add(saveAsItem);
        javax.swing.JMenuItem resamp15Item = new javax.swing.JMenuItem(
                "Eliminate Non-15-Minute Data");
        resamp15Item.addActionListener(this);
        subMenu.add(resamp15Item);

        return subMenu;
    }

//------------------------------------------------------------------------------
    @Override
    public String toString() {

        String s = "CR1000 Data File (" + m_selectedConfig.title + ")";

        return (s + ": " + m_filename);
    }

//------------------------------------------------------------------------------
    @Override
    public void actionPerformed(java.awt.event.ActionEvent e) {

        // Get the text from the menu item that was chosen.
        javax.swing.JMenuItem source =
                (javax.swing.JMenuItem) (e.getSource());
        String srcText = source.getText();

        // Show an hourglass until the job is complete.
        java.awt.Frame frame = GrController.getParentFrame(source);
        frame.setCursor(new Cursor(Cursor.WAIT_CURSOR));

        // Check if the command was to resample.
        if (srcText.equals("Eliminate Non-15-Minute Data")) {

            // Create a data snap shot of the series in this file.
            DataSnapShot snapShot = new DataSnapShot();
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                while (o instanceof ObjectEffigy) {
                    o = ((ObjectEffigy) o).getSubject();
                }

                if (o instanceof DataSeries) {
                    snapShot.store((DataSeries) o);
                }
            }

            // Attempt to save a snapshot, so this operation can be undone.
            iter = m_observers.listIterator();
            while (iter.hasNext()) {
                Object gro = iter.next();
                while (gro instanceof ObjectEffigy) {
                    gro = ((ObjectEffigy) gro).getSubject();
                }
                if (gro instanceof GrController) {
                    ((GrController) gro).saveDataSnapShot(snapShot);
                    break;
                }
            }

            eliminateNon15();
        } else {
            // Pass other commands to the parent class.
            super.actionPerformed(e);
        }

        // Return to the default arrow cursor.
        frame.setCursor(Cursor.getDefaultCursor());
    }

//------------------------------------------------------------------------------
    /**
     * Remove all points that aren't on a 15 minute interval.
     */
    protected void eliminateNon15() {

        // Create a constant to represent how close to 15 minute spacing
        // is close enough.  Use 30 seconds.
        double sec30 = 0.0003472222222222222;
        double tolerance = sec30;
        double min15 = 0.01041666666666667;
        double span = min15;

        // Find all children that are DataSeries.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            Object o = iter.next();

            while (o instanceof ObjectEffigy) {
                o = ((ObjectEffigy) o).getSubject();
            }

            if (o instanceof DataSeries) {
                // Make a new series and resample it to 15 minutes with
                // averaging.
                DoubleDataSeries series = (DoubleDataSeries) o;

                // Create a list for keeping track of points to be deleted.
                ArrayList IndexesToDelete = new ArrayList();

                // Get the arrays from the series.
                double[] oldArray = series.getArray();
                int numOldPoints = series.getNumPoints();
                int numOldDims = series.getNumDimensions();

                // Loop through the old array looking at each point to see if
                // it needs to be changed, deleted, or kept as is.
                for (int i = 0; i < numOldPoints; i++) {
                    double oldX = oldArray[i * numOldDims];

                    if ((oldX % span) > tolerance && (oldX % span) <
                            (span - tolerance)) {

                        IndexesToDelete.add(i);
                    }
                }

                int numToDelete = IndexesToDelete.size();
                if (numToDelete <= 0) {
                    return;
                } else {
                    String message = "" + numToDelete + " data points were removed because they were not on" + " a 15 minute inteval.";
                    if (m_bUseDialogNotification) {
                        javax.swing.JOptionPane.showMessageDialog(null, message,
                                "Notification",
                                javax.swing.JOptionPane.INFORMATION_MESSAGE);
                    } else {
                        System.out.println("CR10 input: " + message);
                    }
                }

                // Delete all points marked for deletion.
                int numNewPoints = numOldPoints - numToDelete;
                double[] newArray = new double[numNewPoints * 2];
                int nextIndexToDelete = -1;
                int oldIndex = 0;
                for (int i = 0; i < numNewPoints; i++) {
                    while (nextIndexToDelete <= oldIndex) {
                        if (nextIndexToDelete == oldIndex) {
                            oldIndex++;
                        }
                        if (IndexesToDelete.size() <= 0) {
                            nextIndexToDelete = Integer.MAX_VALUE;
                        } else {
                            nextIndexToDelete =
                                    ((Integer) IndexesToDelete.get(0));
                            IndexesToDelete.remove(0);
                        }
                    }
                    if (nextIndexToDelete > oldIndex) {
                        newArray[i * 2] = oldArray[oldIndex * 2];
                        newArray[i * 2 + 1] = oldArray[oldIndex * 2 + 1];
                    }
                    oldIndex++;
                }
                series.setArray(newArray);
            }
        }
    }

//------------------------------------------------------------------------------
    /**
     * A dialog that allows the user to configure input options.
     */
    public class CR1000ConfigDialog extends JDialog {

        // Initialize GUI components.
        public boolean m_bOkPressed = false;
        public List m_preConfigs = new ArrayList();
        private JComboBox presetCombo = new JComboBox();
        private JTextField dateField = new JTextField();
        private JTextField keepField = new JTextField();
        private JTextField typesField = new JTextField();
        private JTextField titlesField = new JTextField();


        //------------------------------------------------------------------------------
        /**
         * Create a new dialog using the give frame, title, and modal setting.
         * @param frame The frame to which this dialog should be attached.
         * @param title The title of the dialog.
         * @param modal Whether the dialog should be modal (blocks input to other
         *              windows until closed.)
         */
        public CR1000ConfigDialog(Frame frame, String title, boolean modal) {

            super(frame, title, modal);

            try {
                jbInit();
                pack();
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }

        //------------------------------------------------------------------------------
        /**
         * Create a dialog with no title and attached to no frame.
         */
        public CR1000ConfigDialog() {
            this(null, "", false);
        }

        //------------------------------------------------------------------------------
        /**
         * Set up the dialog and it's components.
         * @throws Exception
         */
        void jbInit() throws Exception {

            // Declare and initialize GUI components.
            JPanel mainPanel = new JPanel();
            JPanel topPanel = new JPanel();
            JPanel lowerPanel = new JPanel();
            JPanel channelPanel = new JPanel();
            JButton okButton = new JButton("OK");
            JButton cancelButton = new JButton("Cancel");
            JPanel presetPanel = new JPanel();
            JPanel datePanel = new JPanel();
            JPanel keepPanel = new JPanel();
            JPanel typesPanel = new JPanel();
            JPanel titlesPanel = new JPanel();

            // Set up button listeners.
            okButton.addActionListener(new java.awt.event.ActionListener() {

                public void actionPerformed(ActionEvent e) {
                    okButton_actionPerformed(e);
                }
            });
            cancelButton.addActionListener(new java.awt.event.ActionListener() {

                public void actionPerformed(ActionEvent e) {
                    cancelButton_actionPerformed(e);
                }
            });
            presetCombo.addActionListener(new java.awt.event.ActionListener() {

                public void actionPerformed(ActionEvent e) {
                    presetCombo_actionPerformed(e);
                }
            });

            // Set the layout managers for the panels and insert components.
            mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
            mainPanel.add(Box.createRigidArea(new Dimension(350, 5)));

            topPanel.setLayout(new BoxLayout(topPanel, BoxLayout.Y_AXIS));

            presetPanel.setLayout(new BoxLayout(presetPanel, BoxLayout.X_AXIS));
            presetPanel.add(new JLabel("Preset"));
            presetPanel.add(Box.createRigidArea(new Dimension(5, 0)));

            presetPanel.add(presetCombo);
            presetPanel.add(Box.createRigidArea(new Dimension(0, 5)));

            mainPanel.add(presetPanel);
            mainPanel.add(Box.createRigidArea(new Dimension(0, 5)));

            channelPanel.setBorder(BorderFactory.createTitledBorder(
                    "Channel Configuration"));
            channelPanel.setLayout(new BoxLayout(channelPanel,
                    BoxLayout.Y_AXIS));

            datePanel.setLayout(new BoxLayout(datePanel, BoxLayout.X_AXIS));
            datePanel.add(new JLabel("Date"));
            datePanel.add(Box.createRigidArea(new Dimension(5, 0)));
            datePanel.add(Box.createHorizontalGlue());
            datePanel.add(dateField);
            channelPanel.add(datePanel);
            channelPanel.add(Box.createRigidArea(new Dimension(0, 5)));

            keepPanel.setLayout(new BoxLayout(keepPanel, BoxLayout.X_AXIS));
            keepPanel.add(new JLabel("Keep   "));
            keepPanel.add(Box.createRigidArea(new Dimension(5, 0)));
            keepPanel.add(Box.createHorizontalGlue());
            keepPanel.add(keepField);
            channelPanel.add(keepPanel);
            channelPanel.add(Box.createRigidArea(new Dimension(0, 5)));

            typesPanel.setLayout(new BoxLayout(typesPanel, BoxLayout.X_AXIS));
            typesPanel.add(new JLabel("Types "));
            typesPanel.add(Box.createRigidArea(new Dimension(5, 0)));
            typesPanel.add(Box.createHorizontalGlue());
            typesPanel.add(typesField);
            channelPanel.add(typesPanel);
            channelPanel.add(Box.createRigidArea(new Dimension(0, 5)));

//            titlesPanel.setLayout(new BoxLayout(titlesPanel, BoxLayout.X_AXIS));
//            titlesPanel.add(new JLabel("Titles "));
//            titlesPanel.add(Box.createRigidArea(new Dimension(5,0)));
//            titlesPanel.add(Box.createHorizontalGlue());
//            titlesPanel.add(titlesField);
//            channelPanel.add(titlesPanel);
//            channelPanel.add(Box.createRigidArea(new Dimension(0,5)));

            mainPanel.add(channelPanel);
            mainPanel.add(Box.createRigidArea(new Dimension(0, 5)));
            lowerPanel.setLayout(new BoxLayout(lowerPanel, BoxLayout.X_AXIS));

            lowerPanel.add(okButton);
            lowerPanel.add(Box.createRigidArea(new Dimension(5, 0)));
            lowerPanel.add(cancelButton);

            mainPanel.add(lowerPanel);

            getContentPane().add(mainPanel);
        }

        //------------------------------------------------------------------------------
        /**
         * Set the list of preConfigs that will be offered to the user.
         * @param list  The list of preConfigs to use.
         * @param index The list index of the currently selected preConfig.
         */
        void setConfigs(List list, int index) {

            m_preConfigs = new ArrayList();
            Iterator iter = list.iterator();
            while (iter.hasNext()) {
                m_preConfigs.add(iter.next());
            }

            // Fill the preset configurations combo box.
            iter = m_preConfigs.iterator();
            while (iter.hasNext()) {
                CR1000Options config = (CR1000Options) iter.next();
                presetCombo.addItem(config.title);
            }

            presetCombo.setSelectedIndex(index);
        }

        //--------------------------------------------------------------------------
        /**
         * Close the window and set a flag indicating the OK button was pressed.
         * @param e The action event to be handled.
         */
        void okButton_actionPerformed(ActionEvent e) {

            m_bOkPressed = true;
            dispose();
        }

        //--------------------------------------------------------------------------
        /**
         * Close the window and set a flag indicating the Cancel button was pressed.
         * @param e The action event to be handled.
         */
        void cancelButton_actionPerformed(ActionEvent e) {

            m_bOkPressed = false;
            dispose();
        }

        //--------------------------------------------------------------------------
        /**
         * Change the information displayed in the text fields because the user
         * chose a different preConfig in the presetCombo.
         * @param e The action event to be handled.
         */
        void presetCombo_actionPerformed(ActionEvent e) {

            // Find which presetCombo item was selected.
            int index = presetCombo.getSelectedIndex();
            if (index < m_preConfigs.size()) {
                // Copy the values of the newly selected preConfig into
                // the text fields so the user can see and edit them.
                CR1000Options config = (CR1000Options) m_preConfigs.get(index);
                dateField.setText(config.dateColumns);
                keepField.setText(config.keepers);
                typesField.setText(config.types);
                titlesField.setText(config.titles);
            }
        }
    }
}

