/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-02-26
 * Date started:    2004-06-29
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.io.*;
import java.text.*;
import java.util.*;

//==============================================================================
/**
 * A data file object for reading, writing, and storing time series data
 * to/from a file.
 */
public class CdpdDataFile extends gov.usgs.gr.data.TimeDataFile {

    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/TimeDataFile.gif";

//------------------------------------------------------------------------------
/**
 * Create an empty data file object.
 */
    public CdpdDataFile () {

        super();
        // Set the title for use by the dialog box.
        m_title = "Bridge Data File";
        m_numHeaderLines = 0;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.
 * The date is of the format
 *     code,julian day,time,S,T,S,T,S,T,S,T,S,T,S,T,pressure
 * Pressure is in decibars at bottom probe.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 * @param br    The buffered reader object to use for reading.
 * @throws IOException
 */
    protected void readData () throws IOException {

        // Remove any data that were children of this object.
        clear();

        // Make a list of column lists, one for each series.
        ArrayList columnList = new ArrayList();
        for (int i = 0; i < m_numChannels; i++) {
            columnList.add(new ArrayList());
        }

        String line;
        List dateList = (List)columnList.get(0);

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();

        // Read each line of the data file.
        while ((line = m_br.readLine()) != null) {
            // Ignore lines that begin with a slash-slash as comment lines.
            if (line.indexOf("//") == 0)
                continue;

            // Parse the rest of the line as doubles.
            double[] vals = AsciiIO.parseLineAsDoubles(line,
                m_inputDelimiters);

            if (vals.length < 3)
                continue;

            int day = (int)Math.round(vals[1]);
            // Add 1.8 seconds before truncating for roundoff.
            double hundredHour = vals[2] + 0.0005;
            int hour = (int)Math.round(hundredHour / 100);
            int minute =
                (int)Math.round(((hundredHour / 100) - (double)hour) * 100);
            int year = 2004;
            if (day > 200)
                year = 2003;

            // Create a date object from the information above.
            GregorianCalendar date =
                new GregorianCalendar(utcZone, locale);
            date.clear();
            date.set(Calendar.YEAR, fourDigitYear(year));
            date.set(Calendar.DAY_OF_YEAR, day);
            date.set(Calendar.HOUR_OF_DAY, hour);
            date.set(Calendar.MINUTE, minute);

            // Parse the date and add it to the date list.
            dateList.add(date);

            // Add each number on the line to the appropriate column list.
            for (int i = 3; i < vals.length; i++) {
                List column = null;
                // Add a new column to the list if there are not enough.
                if (columnList.size() < (i - 1)) {
                    column = new ArrayList();
                    columnList.add(column);
                }
                else {
                    column = (List)columnList.get(i - 2);
                }

                // Add the number to the list.
                column.add(vals[i]);
            }
        }

        // Find missing points so they can be shown as deleted points.
        long mostCommonInc = medianIncrement(dateList);
        List<Double> deletedList = markMissingPoints(dateList, mostCommonInc);

        // Create a list containing a column of dates and columns for the other
        // data so they can be interleaved into individual series.
        m_startDate = (GregorianCalendar)
            ((GregorianCalendar)dateList.get(0)).clone();
        GregorianCalendar lastDate = (GregorianCalendar)
            ((GregorianCalendar)dateList.get(dateList.size() - 1)).clone();
        m_referenceYear = m_startDate.get(GregorianCalendar.YEAR);

        // Interleave the series.
        List<TimeSeries> timeSeriesList = interleaveArrays(columnList);

        // Perform additional setup tasks on each new series.
        int index = 0;
        Iterator seriesIter = timeSeriesList.iterator();
        while (seriesIter.hasNext()) {
            TimeSeries series = (TimeSeries)seriesIter.next();

            // Set the type and title to a different number for each series.
            int type = index % 2;
            if (!seriesIter.hasNext())
                type = 2;
            series.setType(type);
            series.setTitle("Series " + (index + 1));

            // Add a copy of the deleted list to the series.
            ArrayList copyOfDeletedList = new ArrayList();
            for (double dVal : deletedList)
                copyOfDeletedList.add(dVal);

            // Delete points with a y value beyond +- 999980
            double[] min = {-Double.MAX_VALUE, -Double.MAX_VALUE};
            double[] max = {Double.MAX_VALUE, -99998};
            series.selectArea(min, max, true);
            min[0] = -Double.MAX_VALUE;
            min[1] = 99998;
            max[0] = Double.MAX_VALUE;
            max[1] = Double.MAX_VALUE;
            series.selectArea(min, max, true);

            series.deleteSelectedPoints();

            // Add the series as a child of this file.
            add(series);

            index++;
        }
    }
}

