/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-02-26
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.data.*;
import java.io.*;
import java.util.*;

//==============================================================================
/**
 * A data file for reading and storing data at sub-daily intervals from the
 * 1990's SF Bay Hydrodynamics database.
 */
public class Db1DataFile extends gov.usgs.gr.data.TimeDataFile {

//------------------------------------------------------------------------------
/**
 * Create an empty data file.
 */
    public Db1DataFile () {

    	super();
        // Set the title for use by the dialog box.
        m_title = "DB1 Data File";
        m_numHeaderLines = 0;
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.
 * Overrides readData() from super class.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 * @throws IOException on IO error.
 */
    protected void readData () throws IOException {

    	double[] vals;

        // Determine column labels from the last header line.
        String columnHeader = "";
        while (columnHeader.indexOf("_") < 0) {
            m_headerLines.add(AsciiIO.readLine(m_br));
            m_numHeaderLines++;
            columnHeader =
                (String)m_headerLines.get(m_headerLines.size() - 1);
        }
        m_title = (String)m_headerLines.get(1);

        // Parse the line into a String array.
        StringTokenizer st = new StringTokenizer(columnHeader, "_");
        String[] columnLabels = new String[st.countTokens()];

        int i = 0;
        while (st.hasMoreTokens()) {
        	columnLabels[i++] = st.nextToken();
        }

        List<List> columnList = new ArrayList<List>();
        List<GregorianCalendar> dateList = new ArrayList<GregorianCalendar>();
        columnList.add(dateList);

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();

        // Read each line in the file, and store each line's parts in a list.
        boolean bFirstLine = true;
        while ((vals = AsciiIO.parseLineAsDoubles(m_br.readLine(),
        	m_inputDelimiters)) != null) {

        	if (vals.length < 5) continue;
        	if (bFirstLine) {
                while (vals.length > (columnList.size() + 3)) {
                    columnList.add(new ArrayList());
                }
                bFirstLine = false;
            }

            // Add 1.8 seconds before truncating for roundoff.
            double hundredHour = (vals[3] / 100) + 0.0005;
            int hour = (int)hundredHour;
            int minute = (int)((hundredHour - (double)hour) * 60);
            GregorianCalendar date =
            	new GregorianCalendar(utcZone, locale);
            date.clear();
            date.set(Calendar.YEAR, fourDigitYear((int)vals[0]));
            date.set(Calendar.MONTH, (int)(vals[1] - 1));
            date.set(Calendar.DAY_OF_MONTH, (int)vals[2]);
            date.set(Calendar.HOUR_OF_DAY, hour);
            date.set(Calendar.MINUTE, minute);

            // Store the date.
            List column = (List)columnList.get(0);
            column.add(date);

            // Store any other values on the line as Doubles.
            for (i = 4; i < vals.length; i++) {
            	column = (List)columnList.get(i - 3);
                column.add(vals[i]);
            }
        }

        // Find missing points so they can be shown as deleted points.
        long mostCommonInc = medianIncrement(dateList);
        List<Double> deletedList = markMissingPoints(dateList, mostCommonInc);

        // Create a list containing a column of dates and columns for the other
        // data so they can be interleaved into individual series.
        m_startDate = (GregorianCalendar)
            ((GregorianCalendar)dateList.get(0)).clone();
        GregorianCalendar lastDate = (GregorianCalendar)
            ((GregorianCalendar)dateList.get(dateList.size() - 1)).clone();
        m_referenceYear = m_startDate.get(GregorianCalendar.YEAR);

        // Interleave the series.
        List<TimeSeries> timeSeriesList = interleaveArrays(columnList);

        // Perform additional setup tasks on each new series.
        int index = 0;
        for (TimeSeries series : timeSeriesList) {
            // Set the type and title to a different number for each series.
            series.setType(index);
            series.setTitle(columnLabels[index + 4]);

            // Add a copy of the deleted list to the series.
            ArrayList copyOfDeletedList = new ArrayList();
            Iterator<Double> iter2 = deletedList.iterator();
            while (iter2.hasNext()) {
                double real = iter2.next();
                copyOfDeletedList.add(real);
            }
            series.setDeletedList(copyOfDeletedList);

            // Delete points with a y value beyond +- 999980
            double[] min = {-Double.MAX_VALUE, -Double.MAX_VALUE};
            double[] max = {Double.MAX_VALUE, -999980};
            series.selectArea(min, max, true);
            min[0] = -Double.MAX_VALUE;
            min[1] = 999980;
            max[0] = Double.MAX_VALUE;
            max[1] = Double.MAX_VALUE;
            series.selectArea(min, max, true);

            series.deleteSelectedPoints();

            // Add the series as a child of this file.
            add(series);

            index++;
        }

/*
        // Make a new array for holding interleaved data.
        int numColumns = columnList.size() - 1;

        // Make a list of titles, one for each series.
        String[] titles = new String[numColumns];
        for (i = 0; i < numColumns; i++) {
        	titles[i] = columnLabels[i + 4];
        }

        // Populate the series list by interleaving data from the columns.
        interleaveColumns(titles, null);
*/
    }
}

