/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.data.*;
import java.io.*;
import java.util.*;

//==============================================================================
/**
 * A data file object for reading and storing data in GS format.
 */
public class Db2DataFile extends gov.usgs.gr.data.TimeSeriesDataFile {

//------------------------------------------------------------------------------
/**
 * Create an empty data object.
 * <p><b> Pre Conditions </b><br>
 *      None.
 * <p><b> Post Conditions </b><br>
 *      Use load to fill this object.
 */
    public Db2DataFile () {
    	super();
        // Set the title for use by the dialog box.
        m_title = "DB2 Data File";
        m_numHeaderLines = 7;
    }

//------------------------------------------------------------------------------
/**
 * Read the header information from an open file attached to a reader.
 * Overrides same method in super class.
 * <p><b> Pre Conditions </b><br>
 *      None.
 * <p><b> Post Conditions </b><br>
 *      None.
 * @param br    The buffered reader object to use for reading.
 */
    protected void readHeader () throws IOException {

    	super.readHeader();
        m_title = (String)m_headerLines.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.
 * Overrides readData() from gov.usgs.gr.data.XyDataFile.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 * @param br    The buffered reader object to use for reading.
 */
    protected void readData () throws IOException {

    	double[] vals;
        ArrayList deletedGroup = new ArrayList();

        // Determine column labels.
        String columnHeader = (String)m_headerLines.get(
            m_headerLines.size() - 1);

        // Parse the line into a String array.
        StringTokenizer st = new StringTokenizer(columnHeader, "_");
        String[] columnLabels = new String[st.countTokens()];

        int i = 0;
        while (st.hasMoreTokens()) {
        	columnLabels[i++] = st.nextToken();
        }

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();

        // Read each line in the file, and store each line's parts in a list.
        boolean bFirstLine = true;
        while ((vals = AsciiIO.parseLineAsDoubles(m_br.readLine(),
        	m_inputDelimiters)) != null) {

        	if (vals.length < 2) continue;
        	while (vals.length > m_columnList.size()) {
            	m_columnList.add(new ArrayList());

        		// Make a list of lists for holding deleted points.
        		deletedGroup.add(new ArrayList());
            }

            vals[0] += 0.0001;
            int year = (int)(vals[0] / 10000);
            int month = (int)((vals[0] / 100) - ((float)year * 100));
            int day = (int)(vals[0] - ((float)month * 100)
            	- ((float)year * 10000));
            GregorianCalendar date = new GregorianCalendar(utcZone, locale);
            date.clear();
            date.set(Calendar.YEAR, fourDigitYear(year));
            date.set(Calendar.MONTH, month - 1);
            date.set(Calendar.DAY_OF_MONTH, day);
            date.set(Calendar.HOUR_OF_DAY, 12);
            date.set(Calendar.MINUTE, 0);

			// Must set start date and reference year if calling julianDecimal
            // routine.
            if (bFirstLine) {
            	setStartDate(date);
                bFirstLine = false;
            }

            // Store the date.
            List column = (List)m_columnList.get(0);
            column.add(date);

            // Store any other values on the line as Doubles.
            for (i = 1; i < vals.length; i++) {
            	column = (List)m_columnList.get(i);
                if (vals[i] < -999000) {
                	// Remove the date that was just added to the first column.
            		column = (List)m_columnList.get(0);
            		column.remove(column.size() - 1);

                    // Add this point to the deleted list.
                	List deleted = (List)deletedGroup.get(i - 1);
                    deleted.add(julianDecimal(date, m_referenceYear));
                    deleted.add(0d);
                }
                else {
                	// It is not a bad point, so add it as the y-value.
                    column.add(vals[i]);
                }
            }
        }

        // Make a new array for holding interleaved data.
        int numColumns = m_columnList.size() - 1;

        // Make a list of titles, one for each series.
        String[] titles = new String[numColumns];
        for (i = 0; i < numColumns; i++) {
        	titles[i] = columnLabels[i + 1];
        }

        // Populate the series list by interleaving data from the columns.
        interleaveColumns(titles, deletedGroup);
    }
}

