/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2002-01-18
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.data.*;
import java.io.*;
import java.util.*;

//==============================================================================
/**
 * An array of data in one or more dimensions.
 * @see gov.usgs.data.XyData.
 */
public class DssAsciiDataFile extends gov.usgs.gr.data.TimeDataFile {

    /** The title of each data column. */
    protected String[] m_titles;
    /** The data type of each data column. */
    protected String[] m_types;

//------------------------------------------------------------------------------
/**
 * Creates a data file.
 */
    public DssAsciiDataFile () {

    	super();
        m_title = "DSS ASCII Data File";
        m_numHeaderLines = 3;
    }

//------------------------------------------------------------------------------
/**
 * Read the header information from an open file attached to a reader.
 * Called by load().
 */
    protected void readHeader () throws IOException {

    	super.readHeader();

        if (m_numHeaderLines <= 0)
            return;

        // Read the main title.
        m_title = (String)m_headerLines.get(0);

        if (m_numHeaderLines <= 1)
            return;

        // Look for the colon that signals the beginning of the location and
        // type.  Combine the two parts into one title.
        String line = (String)m_headerLines.get(1);
        int colonIndex = line.indexOf(':');
        String types = line.substring(colonIndex + 1).trim();
        m_types = AsciiIO.parseLine(types, ",");

        if (m_numHeaderLines <= 2)
            return;

        // Read the titles.
        line = (String)m_headerLines.get(2);
        colonIndex = line.indexOf(':');
        String titles = line.substring(colonIndex + 1).trim();
        m_titles = AsciiIO.parseLine(titles, "\"");

        // Throw out the tokens between opening and closing pairs of quotes.
        String[] realTitles = new String[(m_titles.length / 2) + 1];
        for (int i = 0; i < m_titles.length; i += 2) {
            realTitles[i / 2] = m_titles[i];
        }
        m_titles = realTitles;
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.
 * Overrides readData() from gov.usgs.data.DimensionalDataFile.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 */
    protected void readData () throws IOException {

    	clear();

        // Make a list of column lists, one for each series.
        // This is not currently needed for this file format, because
        // it only has one column of data.
        List columnList = new ArrayList();
        for (int i = 0; i < m_numChannels; i++) {
            columnList.add(new ArrayList());
        }

        String line;
        List dateList = (List)columnList.get(0);

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();

        double[] dummyDouble = {0};

        // Read each line of the data file.
        while ((line = m_br.readLine()) != null) {
            // Look for the quotes that separate the date from the rest
            // of the line.
            int space1Index = line.indexOf(' ');
//            int space2Index = line.indexOf(' ', space1Index + 1);;

            if (space1Index < 0) continue;

            String dayString = line.substring(0, 2);
            String monthString = line.substring(2, 5);
            String yearString = line.substring(5, 9);
            String hourString =
                line.substring(space1Index + 1, space1Index + 3);
            String minuteString =
                line.substring(space1Index + 3, space1Index + 5);

            int month = 0;
            String[] monthNames = {"JAN", "FEB", "MAR", "APR", "MAY", "JUN",
                "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"};
            for (int i = 0; i < 12; i++) {
                if (monthString.equals(monthNames[i])) {
                    month = i;
                    break;
                }
            }

            // Parse the date and add it to the date list.
            GregorianCalendar date = new GregorianCalendar(utcZone, locale);
            date.clear();
            date.set(Calendar.YEAR,
                fourDigitYear(Integer.parseInt(yearString)));
            date.set(Calendar.MONTH, month);
            date.set(Calendar.DAY_OF_MONTH, Integer.parseInt(dayString));
            date.set(Calendar.HOUR_OF_DAY, Integer.parseInt(hourString));
            date.set(Calendar.MINUTE, Integer.parseInt(minuteString));

            dateList.add(date);

            // Parse the rest of the line (after the decimal date) as double.
            String remainderOfLine = line.substring(space1Index + 5);

            double[] vals = AsciiIO.parseLineAsDoubles(remainderOfLine,
                m_inputDelimiters);

            if (vals.length == 0) {
                vals = dummyDouble;
            }

            // Add each number on the line to the appropriate column list.
            for (int i = 0; i < vals.length; i++) {
                List column = null;
                // Add a new column to the list if there are not enough.
                if (columnList.size() < (i + 2)) {
                    column = new ArrayList();
                    columnList.add(column);
                }
                else {
            	    column = (List)columnList.get(i + 1);
                }

                // Add the number to the list.
                column.add(vals[i]);
            }
        }

        // Find missing points so they can be shown as deleted points.
        long mostCommonInc = medianIncrement(dateList);
        List<Double> deletedList = markMissingPoints(dateList, mostCommonInc);

        // Create a list containing column of dates and columns for the other
        // data so they can be interleaved into individual series.
        m_startDate = (GregorianCalendar)
            ((GregorianCalendar)dateList.get(0)).clone();
        GregorianCalendar lastDate = (GregorianCalendar)
            ((GregorianCalendar)dateList.get(dateList.size() - 1)).clone();
        m_referenceYear = m_startDate.get(GregorianCalendar.YEAR);

        List<TimeSeries> timeSeriesList = interleaveArrays(columnList);

        int index = 0;
        Iterator iter = timeSeriesList.iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries)iter.next();

            // Set the type and title to a different number for each series.
            if (m_types != null && m_types.length > 0)
                series.setType(Integer.parseInt(m_types[index]));
            if (m_titles != null && m_titles.length > 0)
                series.setTitle(m_titles[index]);

            // Add a copy of the deleted list to the series.
            List<Double> copyOfDeletedList = new ArrayList();
            for (double real : deletedList)
                copyOfDeletedList.add(real);

            series.setDeletedList(copyOfDeletedList);

            if (series instanceof TimeSeries) {
                // Delete points with a y value of -9998
                double[] min = {-Double.MAX_VALUE, -Double.MAX_VALUE};
                double[] max = {Double.MAX_VALUE, -9998};
                series.selectArea(min, max, true);

                series.deleteSelectedPoints();
            }

            // Add the series as a child of this file.
            add(series);

            index++;
        }
    }
}
