/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-02-26
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.data.*;
import java.io.*;
import java.util.*;

//==============================================================================
/**
 * A data file object for reading and storing data in GS format.
 */
public class Dwr1DataFile extends gov.usgs.gr.data.TimeSeriesDataFile {

//------------------------------------------------------------------------------
/**
 * Create an empty data object.
 * <p><b> Pre Conditions </b><br>
 *      None.
 * <p><b> Post Conditions </b><br>
 *      Use load to fill this object.
 */
    public Dwr1DataFile () {
    	super();
        // Set the title for use by the dialog box.
        m_title = "DWR Data File";
        m_numHeaderLines = 4;
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.
 * Overrides readData() from gov.usgs.gr.data.XyDataFile.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 */
    protected void readData () throws IOException {

    	String[] toks;
        String[] monthAbbrevs = {"JAN", "FEB", "MAR", "APR", "MAY",
        	"JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"};
        ArrayList deletedGroup = new ArrayList();

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();

        // Read each line in the file, and store each line's parts in a list.
        boolean bFirstLine = true;
        while ((toks = AsciiIO.parseLine(m_br.readLine(), m_inputDelimiters))
        	!= null) {

        	if (toks.length < 3) continue;
        	while (toks.length > (m_columnList.size() + 1)) {
            	m_columnList.add(new ArrayList());

        		// Make a list of lists for holding deleted points.
        		deletedGroup.add(new ArrayList());
            }

            // Determine the year, month, day, and time from what was read
            // from the first two columns.
            int year = Integer.parseInt(toks[0].substring(5, 9));
            String monthStr = toks[0].substring(2, 5);
            int month = 1;
            for (int i = 0; i < monthAbbrevs.length; i++) {
            	if (monthStr.equals(monthAbbrevs[i]))
                	month = i + 1;
            }
            int day = Integer.parseInt(toks[0].substring(0, 2));
            int hour = Integer.parseInt(toks[1].substring(0, 2));
            int minute = Integer.parseInt(toks[1].substring(2, 4));

            // Create a date object from the information above.
            GregorianCalendar date =
            	new GregorianCalendar(utcZone, locale);
            date.clear();
            date.set(Calendar.YEAR, fourDigitYear(year));
            date.set(Calendar.MONTH, month - 1);
            date.set(Calendar.DAY_OF_MONTH, day);
            date.set(Calendar.HOUR_OF_DAY, hour);
            date.set(Calendar.MINUTE, minute);

            if (bFirstLine) {
            	setStartDate(date);
                bFirstLine = false;
            }

            // Store the date.
            List column = (List)m_columnList.get(0);
            column.add(date);

            // Store any other values on the line as Doubles.
            for (int i = 2; i < toks.length; i++) {
            	column = (List)m_columnList.get(i - 1);

                if (Double.parseDouble(toks[i].trim()) < -900) {
                	// Remove the date that was just added to the first column.
            		column = (List)m_columnList.get(0);
            		column.remove(column.size() - 1);

                    // Add this point to the deleted list.
                	List deleted = (List)deletedGroup.get(i - 2);
                    deleted.add(julianDecimal(date, m_referenceYear));
                    deleted.add(0d);
                }
                else {
                	// It is not a bad point, so add it as the y-value.
                	column.add(new Double(toks[i]));
                }
            }
        }

        // Make a new array for holding interleaved data.
        int numColumns = m_columnList.size() - 1;

        // Make a list of titles, one for each series.
        String[] titles = new String[numColumns];
        for (int i = 0; i < numColumns; i++) {
        	titles[i] = "Data Column " + i;
        }

        // Populate the series list by interleaving data from the columns.
        interleaveColumns(titles, deletedGroup);
    }
}

