/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    1998-06-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.data.*;
import java.io.*;
import java.util.*;

//==============================================================================
/**
 * A data file object for reading and storing data in GS format.
 */
public class GrGsDataFile extends gov.usgs.gr.data.DelimitedDataFile {

  /** Hold the exponent of data that is read. */
    List<Number> m_exponentList = new ArrayList();
    /**
     * Whether to pad the deleted list with values that are missing from
     * the data file.
     */
    boolean m_bReadMissingValues = true;
    /** Whether to write out times that have no valid data points. */
    boolean m_bWriteMissingValues = false;
    /**
     * Whether to write out the year, month, day, and hour at the end
     * of each line.
     */
    boolean m_bWriteYMDH = true;
    /* The increment between lines of data. */
    double m_timeInc = 0;
    /* A list of columns read from the file. */
    List<List<Number>> m_columnList = new ArrayList();
    /**
     * The year that all dates should be reference to.  If it differs from
     * the year associated with the date, the data should be shifted until
     * they match.
     */
    protected int m_referenceYear = Integer.MIN_VALUE;
	/** The start date for data in the file. */
    protected GregorianCalendar m_startDate = new GregorianCalendar(0, 0, 0);
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GrGsDataFile.gif";

//------------------------------------------------------------------------------
/**
 * Create an empty data object.
 * <p><b> Post Conditions </b><br>
 *      Use load to fill this object.
 */
    public GrGsDataFile () {
    	super();
        // Set the title for use by the dialog box.
        m_title = "GS Data File";
        m_numHeaderLines = 6;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Reads the file from disk and use it to create DataSeries objects.  The
 * series are added to this object as children.
 * <p><b> Pre Conditions </b><br>
 *      The filename is not null.
 * @param filename  The name of the file to load.
 */
    public void read () throws IOException {

    	// Open the file.
        m_br = new BufferedReader(new FileReader(
        	m_file.getAbsolutePath()));
        // Read the header.
        readHeader();
        // Read the data.
        readData();
        // Close the file.
        m_br.close();
	}

//------------------------------------------------------------------------------
/**
 * Read the header information from an open file attached to a reader.
 * @param br    The buffered reader object to use for reading.
 */
    protected void readHeader () throws IOException {
        String line = "", title;

        // Read past three lines.
        line = AsciiIO.readLine(m_br);
        m_title = line.trim() + "  ";
        m_headerLines.add(line);
        m_headerLines.add(AsciiIO.readLine(m_br));
        m_headerLines.add(AsciiIO.readLine(m_br));

        // Read in fourth line and extract the start date.
        line = AsciiIO.readLine(m_br);
        String dateString = line.substring(0, 20);
        m_title += dateString;

        int startYear = (Integer.parseInt(
            dateString.substring(0, 9).trim()));
        int startMonth = (Integer.parseInt(
            dateString.substring(9, 12).trim()));
        int startDay = (Integer.parseInt(
            dateString.substring(12, 15).trim()));
        int startHour = (Integer.parseInt(
            dateString.substring(15, 20).trim()));
        startHour /= 100;
        int startMinute = (Integer.parseInt(
            dateString.substring(15, 20).trim()));
        m_referenceYear = TimeSeriesDataFile.fourDigitYear(startYear);
        GregorianCalendar m_startDate =
        	new GregorianCalendar(TimeZone.getTimeZone("UTC"));
        m_startDate.clear();
        m_startDate.set(Calendar.YEAR, TimeSeriesDataFile.fourDigitYear(startYear));
        m_startDate.set(Calendar.MONTH, startMonth - 1);
        m_startDate.set(Calendar.DAY_OF_MONTH, startDay);
        m_startDate.set(Calendar.HOUR_OF_DAY, startHour);
        m_startDate.set(Calendar.MINUTE, startMinute);

        startHour -= (startHour * 100);

        // Get the number of channels off the fourth line.
        String s = line.substring(40, 46).trim();
        m_numChannels = (Integer.parseInt(s));

        // Make sure the start year is written as four digits.
        line = AsciiIO.padLeft("" + TimeSeriesDataFile.fourDigitYear(startYear),
            9) + line.substring(9);
        m_headerLines.add(line);

        // Read the time increment, if it is present in the header.
        if (line.length() >= 40) {
            String subString = line.substring(30, 40).trim();
            m_timeInc = Double.parseDouble(subString);
            m_timeInc *= 100000;
        }

        // Read past two more lines.
        m_headerLines.add(AsciiIO.readLine(m_br));
        m_headerLines.add(AsciiIO.readLine(m_br));

        // Read the same number of lines as there are channels
        for (int i = 0; i < m_numChannels; i++) {
            line = AsciiIO.readLine(m_br);
            title = line.substring(2, 22).trim();
            m_headerLines.add(line);

            // Read the exponent to use for scaling each column of data.
            s = line.substring(22, 28).trim();
            Integer exp = Integer.parseInt(s);
            m_exponentList.add(exp);

            // Read the numeric code for each column of data.
            s = line.substring(28, 34).trim();
            int dType = Integer.parseInt(s);
            TimeSeries series = new TimeSeries(dType, title, 2);
            series.setStartDate((GregorianCalendar)m_startDate.clone());
            m_seriesList.add(series);
        }

        // Read past on more line.
        m_headerLines.add(AsciiIO.readLine(m_br));

        // For each channel, add a blank element to the column list.
        for (int i = 0; i < (m_numChannels + 1); i++) {
            m_columnList.add(new ArrayList());
        }
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.  Assumes the header
 * has been read with readHeader() and the file needs to be closed.
 */
    protected void readData () throws IOException {
        String line = "";

        double realX, realY;
        Integer time = Integer.MIN_VALUE;
        Integer yInt;

        // Just count the channels after the first (x).

        int numColumns = m_columnList.size() - 1;
        int[] lengths = new int[numColumns];

        // Make a list of lists for holding deleted points.
        List<List<Double>> deletedGroup = new ArrayList<List<Double>>();
        for (int i = 0; i < numColumns; i++) {
        	deletedGroup.add(new ArrayList<Double>());
        }

        // Used for calculating the increment and finding missing points...
        int firstLineTime = Integer.MIN_VALUE;
        int secondLineTime = Integer.MIN_VALUE;
        int thirdLineTime = Integer.MIN_VALUE;
        double lastTime = 0;
        double missingTime = 0;
// Read increment from header if it is there.  Other wise look at first one or
// two increments.  timeInc and lastTime should be doubles.

		// Read through the entire file to the end.
        do {
        	line = m_br.readLine();
            // Check if the end has been reached.
            if (line == null || line.length() <= 9) continue;

            // Read the x value.
            try {
            	time = Integer.parseInt(line.substring(0, 9).trim());
            }
            catch (NumberFormatException e) {
            	// Skip lines whose time can't be read as an integer.
            	continue;
            }

            m_columnList.get(0).add(time);

            // Use the first three lines to establish the increment.
            if (firstLineTime <= Integer.MIN_VALUE) {
            	firstLineTime = time;
                if (firstLineTime < 0) m_referenceYear++;
    		}
            else {
            	if (secondLineTime <= Integer.MIN_VALUE)
                	secondLineTime = time;
                else {
                	if (thirdLineTime <= Integer.MIN_VALUE) {
                    	thirdLineTime = time;
                        m_timeInc =
                        	(double)(thirdLineTime - firstLineTime) / 2;
                        lastTime = thirdLineTime;
                    }
                }
            }

            // Beginning at the third line on, check for missing times.
            if (thirdLineTime > Integer.MIN_VALUE) {
				// Add a deleted point to the deleted list for each
                // column for each line that was missed.
                missingTime = lastTime + m_timeInc;
                while (missingTime < (time - (0.9 * m_timeInc))) {

					for (int i = 0; i < numColumns; i++) {
                    	List<Double> deleted = deletedGroup.get(i);
                        deleted.add(((double)missingTime) / 100000d);
                        deleted.add(0d);
                    }
                    missingTime += m_timeInc;
				}
                lastTime = time;
			}

            // Read each of the y channel values.
            for (int i = 0; i < numColumns; i++) {
                boolean badPoint = false;
            	String sub = line.substring((i*6)+9, (i*6)+15).trim();
                List<Number> series = m_columnList.get(i+1);
                if (sub.equals("999999")) badPoint = true;
                try {
                	yInt = Integer.parseInt(sub);
                }
                catch (NumberFormatException e) {
                	badPoint = true;
                    yInt = 999999;
                }

                if (badPoint) {
                	List<Double> deleted = deletedGroup.get(i);
                    deleted.add(time / 100000d);
                    deleted.add(0d);
                }
                else {
                	lengths[i]++;
                }

                series.add(yInt);
            }

		} while (line != null);

        // Make a new array for holding interleaved data.
        m_seriesArrays = new double[numColumns][];
        // Create a single data series that will hold all the columns.
        DoubleDataSeries bigSeries = new DoubleDataSeries(numColumns + 1);
        // Determine the longest column.
        int tLength = ((List)m_columnList.get(0)).size();
        double[] bigArray = new double[(numColumns + 1) * tLength];

        // Add an interleaved x/y array for each y channel
        for (int column = 0; column < numColumns; column++) {
            m_seriesArrays[column] = new double[2 * lengths[column]];
            Number test = m_exponentList.get(column);
            double scale = Math.pow(10,
                m_exponentList.get(column).doubleValue());

            // Populate the interleaved array with data from the column list.
            int index = 0;
			Iterator<Number> xIter = m_columnList.get(0).iterator();
			Iterator<Number> yIter = m_columnList.get(column + 1).iterator();

            int bigIndex = 0;
			while (xIter.hasNext()) {
                realX = xIter.next().doubleValue();

                // Add the time value to the big array.
                if (column == 0)
                    bigArray[(numColumns + 1) * bigIndex] = realX / 100000;

                yInt = (Integer)yIter.next();
                if (yInt != 999999) {
                	// The x value needs the decimal point moved five
                    // places left.
                	m_seriesArrays[column][2 * index] = realX / 100000;
                	realY = yInt / scale;
                	m_seriesArrays[column][(2 * index) + 1] = realY;

                    // Add the same number to the big array.
                    bigArray[(numColumns + 1) * bigIndex + column + 1] = realY;

                	index++;
                }
                else {
                    // Add a bad number to the big array.
                    bigArray[(numColumns + 1) * bigIndex + column + 1]
                        = Double.MAX_VALUE;
                }

                // Advance the index on the big array.
                bigIndex++;
            }

            // Add the interleaved array to the data series list.
            TimeSeries series = (TimeSeries)m_seriesList.get(column);
            series.setReferenceYear(m_referenceYear);
            series.setArray(m_seriesArrays[column]);
            List list = (List)deletedGroup.get(column);
            series.setDeletedList(list);

            DataDimension dim = bigSeries.getDimension(column + 1);
            dim.setTitle(series.getTitle());

            // Add the X-Y series to this object as a child.
            add(series);
        }

        // Add the series containing all the columns.
        bigSeries.setArray(bigArray);
        bigSeries.setTitle(m_file.getName());
        DataDimension dim = bigSeries.getDimension(0);
        dim.setTitle("Time in Days");

        // Check for Easting/Northing coordinates in the file header.
        Iterator iter = m_headerLines.iterator();
        while (iter.hasNext()) {
            String s = (String)iter.next();
            s = s.toLowerCase();
            int eIndex = s.indexOf("easting:");
            int nIndex = s.indexOf("northing:");
            if (eIndex > -1 && nIndex > -1) {
                String easting = s.substring(eIndex + 8, nIndex);
                String northing = s.substring(nIndex + 9);

                Location location = new Location();
                location.setProperty("x", easting);
                location.setProperty("y", northing);
                bigSeries.add(location);
            }
        }

        add(bigSeries);
    }

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
/**
 * Saves the data to a file.
 * <p><b> Pre Conditions </b><br>
 *      The filename is not null.
 */
    public void write() throws IOException {

    	// Open the file.
        m_bw = new BufferedWriter(new FileWriter(
        	m_file.getAbsolutePath()));

        // Read the header.
        writeHeader();
        // Read the data.
        writeData();
        // Close the file.
        m_bw.close();
    }

//------------------------------------------------------------------------------
/**
 * Read the header information from an open file attached to a reader.
 */
    protected void writeHeader () {

        try {
			Iterator iter = m_headerLines.iterator();
			while (iter.hasNext()) {
            	String s = (String)iter.next();
                m_bw.write(s + "\n", 0, s.length() + 1);
            }
        }
        catch (IOException e) {
        	javax.swing.JOptionPane.showMessageDialog(null,
                "Error saving header of file " + m_filename, "IO Exception",
        		javax.swing.JOptionPane.ERROR_MESSAGE);
		}
    }

//------------------------------------------------------------------------------
/**
 * Write the time series back to a file.  The original times are stored as
 * integers so they can be written back at the exact same time steps that were
 * read in.  Missing values are written as 999999.  If one or more of the
 * channels has a point that is missing from the others, it is written at its
 * own time step and the others channels are written as 999999.  If all data
 * is missing at a time step, that time step is not written to the file.
 * <p><b> Pre Conditions </b><br>
 *      The header has been written with writeHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 */
    protected void writeData () throws IOException {

    	boolean timeBaseMismatchWarningGiven = false;

        int numColumns = m_columnList.size() - 1;
        if (numColumns < 1) return;

        int[] lineOffset = new int[numColumns];
        double[] scale = new double[numColumns];

        // Update the series arrays by copying from the series list.
        Iterator iter = m_seriesList.iterator();
        m_seriesArrays = new double[m_seriesList.size()][];
        for (int i = 0; i < m_seriesList.size(); i++) {
            m_seriesArrays[i] =
            	((DoubleDataSeries)m_seriesList.get(i)).getArray();

            // Put the scales in a convenience array.
        	scale[i] = Math.pow(10.0, (Integer)m_exponentList.get(i));
        }

        int xVal;
        int line = 0;
        int time = 0;
        int alreadySeenTime = Integer.MIN_VALUE;
        boolean bAddingTimeStep = false;

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();

        // Write out each line.
        iter = ((List)m_columnList.get(0)).iterator();
        lineLoop: while (iter.hasNext()) {
            // Grab the next time step from the stored list, unless a
            // new time step is being added or was just added.
            if (!bAddingTimeStep) {
                if (alreadySeenTime == Integer.MIN_VALUE) {
                    time = ((Integer)iter.next());
                }
                else {
                    time = alreadySeenTime;
                    alreadySeenTime = Integer.MIN_VALUE;
                }
            }
            String columnString = AsciiIO.padLeft("" + time, 9);
            String lineString = columnString;

            // The line can be set to write out no matter what, or only
            // if at least one of the columns below has valid data.
            boolean bWriteLine = m_bWriteMissingValues;

            // Check each column for a time value that is less than the
            // one from the columnList.  If one is found, the lineLoop
            // will be restarted using the newly inserted time step.
            for (int column = 0; column < numColumns; column++) {
                // Check that this column hasn't gotten to the end already.
                int columnXVal = Integer.MAX_VALUE;
                if ((line - lineOffset[column])
                    < (m_seriesArrays[column].length / 2)) {

                    // Calculate the column's time value.
                    columnXVal = (int)
                        Math.round(m_seriesArrays[column]
                        [(line - lineOffset[column]) * 2] * 100000);

                    // If this point is significantly before the current
                    // time step, restart the loop and insert it as a
                    // new time step.
                    if (columnXVal <= (time - 5)) {
                        alreadySeenTime = time;
                        time = columnXVal;
                        bAddingTimeStep = true;
                        continue lineLoop;
                    }
                    else {
                        bAddingTimeStep = false;
                    }
                }
            }

            // Repeat the loop and append each column's y value to the
            // output line.
            for (int column = 0; column < numColumns; column++) {
                // Check that this column hasn't gotten to the end already.
                int columnXVal = Integer.MAX_VALUE;
                if ((line - lineOffset[column])
                    < (m_seriesArrays[column].length / 2)) {

                    // Calculate the column's time value.
                    columnXVal = (int)
                        Math.round(m_seriesArrays[column]
                        [(line - lineOffset[column])* 2] * 100000);
                }

                int yVal;

                if (columnXVal < (time + 5) && columnXVal > (time - 5)) {
                    // This y value belongs with the time on the current
                    // line, so write it out.
                    yVal = (int)Math.round(m_seriesArrays[column]
                        [(line - lineOffset[column]) * 2 + 1]
                        * scale[column]);
                    bWriteLine = true;
                }
                else {
                    // The time on the current line doesn't belong with
                    // this y value, so write a non-data point instead.
                    yVal = 999999;
                    if (columnXVal > (time + 8)) {
                        lineOffset[column]++;
                    }
                    else {
                        // Give a warning that the time didn't line up.
                        System.out.println("Warning: timebase mismatch"
                            + ": value not written at " + columnXVal);
                        if (!timeBaseMismatchWarningGiven) {
                            javax.swing.JOptionPane.showMessageDialog(null,
                                "Data at " + columnXVal
                                + " was slightly offset and was not saved.",
                                "Time Base Mismatch",
                                javax.swing.JOptionPane.WARNING_MESSAGE);
                            timeBaseMismatchWarningGiven = true;
                        }
                    }
                }

                // Pad the value to the right number of spaces and add
                // it to the line.
                columnString = AsciiIO.padLeft("" + yVal, 6);
                lineString += columnString;
            }

            if (m_bWriteYMDH) {
                // Write the year, month, day, and hour at the end of
                // the line.
                int year = m_startDate.get(Calendar.YEAR);
                GregorianCalendar date =
                    new GregorianCalendar(utcZone, locale);
                date.clear();
                date.set(Calendar.YEAR, year);
                date.set(Calendar.MONTH, 0);
                date.set(Calendar.DAY_OF_MONTH, 1);
                GregorianCalendar m_startDate =
                    new GregorianCalendar(utcZone, locale);
                m_startDate.clear();

                int jDay = time / 100000;
                // Calculate the resulting year, month, and day resulting
                // from the start year and julian offset.
                date.add(Calendar.DAY_OF_YEAR, (jDay - 1));
                year = date.get(Calendar.YEAR);
                int month = date.get(Calendar.MONTH) + 1;
                int day = date.get(Calendar.DAY_OF_MONTH);

                // Calculate the hour * 100.
                double dayFraction = (double)(time - (jDay * 100000)) / 100000;
                if (dayFraction < 0)
                    dayFraction = 1 + dayFraction;
                int hour = (int)Math.round(dayFraction * 24 * 100);

                // Append the values to the line.
                lineString += AsciiIO.padLeft("" + year, 6);
                lineString += AsciiIO.padLeft("" + month, 6);
                lineString += AsciiIO.padLeft("" + day, 6);
                lineString += AsciiIO.padLeft("" + hour, 6);
            }

            // Add a newline to the end and write it to the file.
            lineString += "\n";
            // Only write the line if at least one column had valid data.
            if (bWriteLine) m_bw.write(lineString, 0, lineString.length());
            line++;
        }
    }
}

