/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-07-06
 * Date started:    2007-04-20
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.TimeMarker;
import gov.usgs.gr.visual.PageOfGraphs;
import gov.usgs.sfhydro.gr.tools.LsAnimCalculator;
import java.awt.Frame;
import java.io.*;
import java.util.*;
import javax.swing.JOptionPane;

//==============================================================================
/**
 * An array of data in one or more dimensions.
 * @see gov.usgs.data.XyData.
 */
public class ListeningStationDataFile extends gov.usgs.gr.data.TimeDataFile {
    
    /** The calculator for generating an animation. */
    protected LsAnimCalculator m_calc = new LsAnimCalculator();
    /** The time step for the particle animation. */
    protected double m_ts = 1d / 24d / 4d;
    /** An array holding the minimum values for each dimension. */
    protected int[] m_min;
    /** An array holding the maximum values for each dimension. */
    protected int[] m_max;
    /** The number of dimensions in each data series. */
    protected int m_numDimensions = 3;
    
//------------------------------------------------------------------------------
    /**
     * Creates a data file.
     */
    public ListeningStationDataFile() {
	
	super();
	m_title = "Listening Station Data File";
	m_numHeaderLines = 0;
	
	m_min = new int[m_numDimensions];
	m_max = new int[m_numDimensions];
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the current values for this object.
     * Properties include: <p>
     * timeStep         String  "Time Step"                                     <p>
     * @return  A list of GrProperty objects.
     */
    public GrHashList getProperties() {
	
	// First add properties from the super class.
	GrHashList props = super.getProperties();
	
	String val = null;
	props.put("timeStep", new GrProperty(m_calc.getTsString(),
	    "Time Step"));
	
	return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Set the values of one or more of this object's properties.
     * @see gov.usgs.sfhydro.gr.tools.LsAnimCalculator#getProperties getProperties()
     * for a list of accepted properties.
     * @param props A list of GrProperty objects specifying the new values for
     *              certain properties.
     */
    public void setProperties(HashMap props) {
	
	GrProperty p;
	
	// For each property with a string value, pass it to the setProperty()
	// method to be handled there.
	Iterator iter = props.keySet().iterator();
	while (iter.hasNext()) {
	    String key = (String)iter.next();
	    Object obj = props.get(key);
	    if (obj instanceof GrProperty) {
		p = (GrProperty)obj;
		
		// Handle string properties by passing them to the
		// setProperty() method.
		if (p.value instanceof String) {
		    setProperty(key, (String)p.value);
		    continue;
		}
	    }
	    // Handle string properties by passing them to the
	    // setProperty() method.
	    if (obj instanceof String) {
		setProperty(key, (String)obj);
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Set a single property.
     * @see gov.usgs.sfhydro.gr.tools.ListeningStationDataFile#getProperties getProperties()
     * for a list of accepted properties.
     * @param key   The property name.
     * @param value The new value for the property.
     */
    public void setProperty(String key, String value) {
	
	super.setProperty(key, value);
	
	if (key.equals("timeStep")) {
	    m_ts = TimeReference.convertToDecimalDay(value);
	    m_calc.setProperty(key, value);
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Read the data through a reader attached to an open file.
     * Overrides readData() from gov.usgs.data.DimensionalDataFile.
     * <p><b> Pre Conditions </b><br>
     *      The header has been read with readHeader()
     * <p><b> Post Conditions </b><br>
     *      The file needs to be closed.
     */
    protected void readData() throws IOException {
	
	// Remove any data that were children of this object.
	clear();
	
	// Make a list of column lists, one for each series.
	ArrayList columnList = new ArrayList();
	for (int i = 0; i < m_numChannels; i++) {
	    columnList.add(new ArrayList());
	}
	
	// Fill the column list with dates and values.
	acquireData(columnList);
	
	// Get a handle to the date list in the first column.
	List dateList = (List)columnList.get(0);
	columnList.remove(dateList);
	columnList.add(1, dateList);
	
	// Create a list containing a column of dates and columns for the other
	// data so they can be interleaved into individual series.
	m_startDate = (GregorianCalendar)
	((GregorianCalendar)dateList.get(0)).clone();
	GregorianCalendar lastDate = (GregorianCalendar)
	((GregorianCalendar)dateList.get(dateList.size() - 1)).clone();
	m_referenceYear = m_startDate.get(GregorianCalendar.YEAR);
	
	setupInterleavedSeries(columnList, null);
    }
    
//------------------------------------------------------------------------------
    /**
     * Sets up the interleaved series from the provided list and adds them as
     * children of this object.
     * @param columnList    The list of data.
     * @param deletedList   The list of deleted points.
     */
    protected void setupInterleavedSeries(List<List> columnList,
	List<Double> deletedList) {
	
	// Interleave the series.
	List<DoubleDataSeries> seriesList = interleaveArrays2(columnList);
	add(seriesList.get(0));
	
	GrController grCon = Gr.getController();
	if (grCon != null) {
	    PageOfGraphs page = grCon.getPage();
	    
	    if (page != null) {
		LsAnimCalculator calc = (LsAnimCalculator)page.
                    firstChildInstanceOf(
                    "gov.usgs.sfhydro.gr.tools.LsAnimCalculator", false);
		
		// Use the existing calculator if there is one.
		if (calc != null)
		    m_calc = calc;
		else
		    // Add the new calc to the page.
		    page.add(m_calc);		
	    }
	}
	
	m_calc.clear();
	m_calc.add(seriesList.get(0));
	m_calc.recalculate();
	add(m_calc);
    }
    
//------------------------------------------------------------------------------
    /**
     * Transfers data from the input data lists to interleaved arrays.
     * Creates new TimeSeries using the arrays and passes them back in a new
     * ArrayList.  The first date in the first list is the "start date".
     * <p><b> Pre Conditions </b><br>
     *      The first column in inputList contains GregorianCalendar dates
     *		which are to be used as the x coordinate of each point.
     *		All of the other columns contain Double objects to be used as y values.
     * @param inputList	A List full of Lists, holding the date
     *                  series and each of the y series.
     * @return  An List of TimeSeries.
     */
    public static List<DoubleDataSeries> interleaveArrays2(List<List> inputList) {
	
	// Make a new array for holding interleaved data.
	List<DoubleDataSeries> tsList = new ArrayList<DoubleDataSeries>();
	List<GregorianCalendar> dateList = inputList.get(1);
	int numDates = dateList.size();
	int numColumns = inputList.size();
	
	// Define the start and end dates.
	GregorianCalendar startDate =
	    (GregorianCalendar)dateList.get(0).clone();
	GregorianCalendar lastDate =
	    (GregorianCalendar)dateList.get(numDates - 1).clone();
	int referenceYear = startDate.get(GregorianCalendar.YEAR);
	
	// Add an interleaved x/y array series.
	double[] seriesArray = new double[numColumns * numDates];
	
	// Interleave the date and data in each array.
	for (int i = 0; i < numDates; i++) {
	    GregorianCalendar date = (GregorianCalendar)dateList.get(i).clone();
	    
	    double jDay = julianDecimal(date, referenceYear);
	    seriesArray[(numColumns * i) + 1] = jDay;
	    
	    for (int column = 0; column < numColumns; column++) {
		if (column == 1)
		    continue;
		
		double yVal = 0;
		try {
		    Double numObj = (Double)(inputList.get(column).get(i));
		    if (numObj != null)
			yVal = numObj;
		} catch (IndexOutOfBoundsException e) { }
		
		seriesArray[i * numColumns + column] = yVal;
	    }
	}
	
	// Create a time series from the interleaved array and add it to the
	// series list.
	TimeSeries series = new TimeSeries(seriesArray, numColumns);

        // Shift the data to be referenced from the most appropriate year.
	series.setStartDate(startDate);
        series.syncToReferenceYear(findAppropriateReferenceYear(startDate,
            lastDate));
	tsList.add(series);
	
	return tsList;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public static boolean matchesFormat(String[] lines) {
        
        try {
            // Read each line of the data file.
            for (int i = 0; i < lines.length; i++) {
                if (lines[i] == null)
                    break;
                
                if (lines[i].indexOf("//") == 0 && lines[i].indexOf(
                    "Listening Station Data File") >= 0) {
                    
                    return true;
                }
            }
        } catch (Exception e) {
        }
            
        return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Shows a dialog for configuring the file.
     * @param frame The frame to use as a parent for a modal dialog box.
     */
    public void showOptionsDialog(Frame frame) {
	
	String inputValue = JOptionPane.showInputDialog(frame, "Time Step",
	    m_calc.getTsString());
	if (inputValue != null) {
	    m_calc.setProperty("timeStep", inputValue);
	    m_ts = TimeReference.convertToDecimalDay(inputValue);
	}
    }
}
