/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-06-29
 * Date started:    2007-06-29
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.data.AsciiIO;
import gov.usgs.gr.data.TimeSeries;
import java.io.IOException;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;

//==============================================================================
/**
 * An array of data in one or more dimensions.
 * @see gov.usgs.data.XyData.
 */
public class Noaa2DataFile extends gov.usgs.gr.data.TimeDataFile {
    
//------------------------------------------------------------------------------
    /**
     * Creates a data file.
     */
    public Noaa2DataFile() {
        
        m_inputDelimiters = " \t\n\r\f";
    }
    
//------------------------------------------------------------------------------
    /**
     * Read the data from the data source and store them in the column list.
     * The first column is a list of GregorianCalendar date objects.  The others are
     * lists of Double objects.
     * @param columnList    This list of data values.
     */
    protected void acquireData(List columnList) throws IOException {
        
        String line;
        List dateList = (List)columnList.get(0);
        
        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        
        // Keep track of whether we have passed the header.
        boolean bPastHeader = false;
        
        // Read each line of the data file.
        while ((line = m_br.readLine()) != null) {
            // Check for header.
            if (!bPastHeader) {
                if (line.indexOf("--") == 0)
                    bPastHeader = true;  
                
                m_headerLines.add(line);
                
                continue;
            }
            
            String[] tokens = AsciiIO.parseLine(line, m_inputDelimiters);
            
            // Parse the date and add it to the date list.
            String dateString = tokens[1] + tokens[2];
            SimpleDateFormat formatter =
                new SimpleDateFormat("yyyyMMddHH:mm", locale);
            formatter.setTimeZone(utcZone);
            GregorianCalendar date = new GregorianCalendar(utcZone, locale);
            date.clear();
            ParsePosition pos = new ParsePosition(0);
            date.setTime(formatter.parse(dateString, pos));
            dateList.add(date);
            
            // Parse the rest of the line as doubles.
            
            // Add each number on the line to the appropriate column list.
            for (int i = 3; i < tokens.length; i++) {
                List column = null;
                // Add a new column to the list if there are not enough.
                if (columnList.size() < (i - 1)) {
                    column = new ArrayList();
                    columnList.add(column);
                } else {
                    column = (List)columnList.get(i - 2);
                }
                
                // Add the number to the list.
                column.add(Double.parseDouble(tokens[i]));
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Override the super method so that series type can be set.
     * @param columnList    The list of data.
     * @param deletedList   The list of deleted points.
     */
    protected void setupInterleavedSeries(List<List> columnList,
        List<Double> deletedList) {
        
        super.setupInterleavedSeries(columnList, deletedList);
        List<TimeSeries> seriesList = (List<TimeSeries>)findChildInstancesOf(
            "gov.usgs.gr.data.TimeSeries",true);
        
        // Set the series to have the same type so they will be drawn in the
        // same graph by default.
        for (TimeSeries series : seriesList) {
            series.setType(0);
        }
        
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public static boolean matchesFormat(String[] lines) {
        
        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        
        boolean bStationFound = false;
        boolean bMllwFound = false;
        boolean bPredFound = false;
        
        String inputDelimiters = " \t\n\r\f";
        
        // Keep track of whether we have passed the header.
        boolean bPastHeader = false;
        
        try {
            // Read each line of the data file.
            for (int i = 0; i < lines.length; i++) {
                if (lines[i] == null)
                    break;
                
                String line = lines[i];
                // Check for header.
                if (!bPastHeader) {
                    if (line.indexOf("--") == 0) {
                        bPastHeader = true;
                    }
                    else {
                        if (line.indexOf("Station Date") >= 0)
                            bStationFound = true;
                        if (line.indexOf("MLLW") >= 0)
                            bMllwFound = true;
                        if (line.indexOf("Pred") >= 0)
                            bPredFound = true;
                    }
                    
                    continue;
                }
                
                bPastHeader = true;
                
                String[] tokens = AsciiIO.parseLine(line, inputDelimiters);
                
                // Parse the date.
                String dateString = tokens[1] + tokens[2];
                SimpleDateFormat formatter =
                    new SimpleDateFormat("yyyyMMddHH:mm", locale);
                formatter.setTimeZone(utcZone);
                ParsePosition pos = new ParsePosition(0);
                formatter.parse(dateString, pos);
                
                // Parse each number on the line as a double.
                for (int j = 3; j < 4; j++) {
                    Double.parseDouble(tokens[j]);
                }
            }
        } catch (Exception e) {
            return false;
        }
        
        if (bStationFound && bMllwFound && bPredFound)
            return true;
        else
            return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Whether this file class has an options dialog to display.
     * @return true if their is an options dialog available.
     */
    public boolean hasOptionsDialog() {
        
        return false;
    }
}
