/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2002-01-18
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.io.*;
import java.text.*;
import java.util.*;

//==============================================================================
/**
 * An array of data in one or more dimensions.
 * @see gov.usgs.data.XyData.
 */
public class SFPortsDataFile extends gov.usgs.gr.data.TimeDataFile {

    /** The name of the series in the file. */
    protected String m_seriesTitle = "";

//------------------------------------------------------------------------------
/**
 * Creates a data file.
 */
    public SFPortsDataFile () {

    	super();
        m_title = "SFPorts Data File";
        m_numHeaderLines = 7;
    }

//------------------------------------------------------------------------------
/**
 * Read the header information from an open file attached to a reader.
 * Called by load().
 */
    protected void readHeader () throws IOException {

    	super.readHeader();

        // Look for the colon that signals the beginning of the location and
        // type.  Combine the two parts into one title.
        String line = (String)m_headerLines.get(0);
        int colonIndex = line.indexOf(':');
        String location = line.substring(colonIndex + 1).trim();

        line = (String)m_headerLines.get(1);
        colonIndex = line.indexOf(':');
        String type = line.substring(colonIndex + 1).trim();

        m_seriesTitle = type + " at " + location;
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.
 * Overrides readData() from gov.usgs.data.DimensionalDataFile.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 */
    protected void readData () throws IOException {

    	clear();

        // Make a list of column lists, one for each series.
        // This is not currently needed for this file format, because
        // it only has one column of data.
        List columnList = new ArrayList();
        for (int i = 0; i < m_numChannels; i++) {
            columnList.add(new ArrayList());
        }

        String line;
        List dateList = (List)columnList.get(0);

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();

        // Read each line of the data file.
        while ((line = m_br.readLine()) != null) {
            // Look for the quotes that separate the date from the rest
            // of the line.
            int paren1Index = line.indexOf('(');
            int paren2Index = 0;
            if (paren1Index >= 0) {
                paren2Index = line.indexOf(')', paren1Index + 1);
            }

            if (paren1Index < 0 || paren2Index < 0) continue;

            String dateString = line.substring(paren1Index + 1, paren2Index);

            // Parse the date and add it to the date list.
            SimpleDateFormat formatter =
                new SimpleDateFormat("M/d/yyyy H:m");
            formatter.setTimeZone(utcZone);
            GregorianCalendar date = new GregorianCalendar(utcZone, locale);
            date.clear();
            ParsePosition pos = new ParsePosition(0);
            Date testDate = formatter.parse(dateString, pos);
            date.setTime(testDate);
            dateList.add(date);

            // Parse the rest of the line (after the decimal date) as doubles.
            int firstSpaceIndex = line.indexOf(' ');
            String remainderOfLine = line.substring(firstSpaceIndex + 1,
                paren1Index);

            double[] vals = AsciiIO.parseLineAsDoubles(remainderOfLine,
                m_inputDelimiters);

            // Add each number on the line to the appropriate column list.
            for (int i = 0; i < vals.length; i++) {
                List column = null;
                // Add a new column to the list if there are not enough.
                if (columnList.size() < (i + 2)) {
                    column = new ArrayList();
                    columnList.add(column);
                }
                else {
            	    column = (List)columnList.get(i + 1);
                }

                // Add the number to the list.
                column.add(vals[i]);
            }
        }

        // Find missing points so they can be shown as deleted points.
        long mostCommonInc = medianIncrement(dateList);
        List<Double> deletedList = markMissingPoints(dateList, mostCommonInc);

        // Create a list containing column of dates and columns for the other
        // data so they can be interleaved into individual series.
        m_startDate = (GregorianCalendar)
            ((GregorianCalendar)dateList.get(0)).clone();
        GregorianCalendar lastDate = (GregorianCalendar)
            ((GregorianCalendar)dateList.get(dateList.size() - 1)).clone();
        m_referenceYear = m_startDate.get(GregorianCalendar.YEAR);

        List<? extends TimeSeries> timeSeriesList =
            interleaveArrays(columnList);

        int index = 0;
        Iterator iter = timeSeriesList.iterator();
        while (iter.hasNext()) {
            TimeSeries series = (TimeSeries)iter.next();

            // Set the type and title to a different number for each series.
            series.setType(index);
            series.setTitle(m_seriesTitle);
            String lowTitle = m_seriesTitle.toLowerCase();
            if (lowTitle.indexOf("sal") >= 0) series.setType(5);
            if (lowTitle.indexOf("temp") >= 0) series.setType(6);
            if (lowTitle.indexOf("cond") >= 0) series.setType(8);
            if (lowTitle.indexOf("speed") >= 0) series.setType(21);
            if (lowTitle.indexOf("stage") >= 0 ||
                lowTitle.indexOf("height") >= 0 ||
                lowTitle.indexOf("level") >= 0) {

                series.setType(23);
            }

            // Add a copy of the deleted list to the series.
            ArrayList<Double> copyOfDeletedList = new ArrayList<Double>();
            for (double dVal : deletedList)
            	copyOfDeletedList.add(dVal);

            series.setDeletedList(copyOfDeletedList);

            // Delete points with a y value of -9998
            double[] min = {-Double.MAX_VALUE, -Double.MAX_VALUE};
            double[] max = {Double.MAX_VALUE, -9998};
            series.selectArea(min, max, true);

            series.deleteSelectedPoints();

            // Add the series as a child of this file.
            add(series);

            index++;
        }
    }
}
