/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2005-11-17
 * Date started:    2003-07-03
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.sfhydro.si3d.*;
import java.io.*;
import java.util.*;

//==============================================================================
/**
 * Reads and writes the format for Si3d thin-wall barriers.
 */
public class Si3dBarrierFile extends gov.usgs.gr.data.DelimitedDataFile {

//------------------------------------------------------------------------------
/**
 * Creates a data file.
 */
    public Si3dBarrierFile () {

    	super();
        m_title = "Si3d Barrier File";
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 */
    protected void readData () throws IOException {

        clear();

        int[] uArray = null;
        int[] vArray = null;

        while (true) {
            String textLine = m_br.readLine();
            if (textLine == null) break;
            if (textLine.startsWith("//")) continue;
            if (textLine.startsWith("nbarru")) {
                int numStart = textLine.indexOf('=') + 1;
                int numBarriers = Integer.parseInt(
                    textLine.substring(numStart).trim());
                uArray = new int[numBarriers * 2];
                for (int i = 0; i < numBarriers; i++) {
                    int[] vals = AsciiIO.parseLineAsInts(m_br.readLine(),
                        " \t\n\r\f,");
                    int iVal = vals[0] - 1;
                    int jVal = vals[1] - 2;
                    uArray[i * 2] = iVal * 10000 + jVal;
                    uArray[i * 2 + 1] = iVal * 10000 + jVal + 1;
                }
            }
            if (textLine.startsWith("nbarrv")) {
                int numStart = textLine.indexOf('=') + 1;
                int numBarriers = Integer.parseInt(
                    textLine.substring(numStart).trim());
                vArray = new int[numBarriers * 2];
                for (int i = 0; i < numBarriers; i++) {
                    int[] vals = AsciiIO.parseLineAsInts(m_br.readLine(),
                        " \t\n\r\f,");
                    int iVal = vals[0] - 2;
                    int jVal = vals[1] - 1;
                    vArray[i * 2] = iVal * 10000 + jVal;
                    vArray[i * 2 + 1] = (iVal + 1) * 10000 + jVal;
                }
            }
            if (textLine.startsWith("ndrycl")) {
                int numStart = textLine.indexOf('=') + 1;
                int numBarriers = Integer.parseInt(
                    textLine.substring(numStart).trim());
                int[] array = new int[numBarriers];
                for (int i = 0; i < numBarriers; i++) {
                    int[] vals = AsciiIO.parseLineAsInts(m_br.readLine(),
                        " \t\n\r\f,");
                    array[i] = (vals[0] - 2) * 10000 + vals[1] - 2;
                }
                IntDataSeries dryCells = new IntDataSeries(array, 1);
                dryCells.setTitle("Dry Cells");
                add(dryCells);
            }
        }

        if (uArray != null || vArray != null) {
            int totalSize = 0;
            if (uArray != null)
                totalSize += uArray.length;
            if (vArray != null)
                totalSize += vArray.length;
            int[] twbArray = new int[totalSize];
            int uIndex = 0;
            int vIndex = 0;
            for (int i = 0; i < totalSize; i += 2) {
                int p1 = 0;
                int p2 = 0;
                int u1 = Integer.MAX_VALUE;
                int u2 = Integer.MAX_VALUE;
                int v1 = Integer.MAX_VALUE;
                int v2 = Integer.MAX_VALUE;
                if (uArray != null && uIndex < uArray.length) {
                    u1 = uArray[uIndex];
                    u2 = uArray[uIndex + 1];
                }
                if (vArray != null && vIndex < vArray.length) {
                    v1 = vArray[vIndex];
                    v2 = vArray[vIndex + 1];
                }
                if (u1 < v1) {
                    p1 = u1;
                    p2 = u2;
                    uIndex += 2;
                }
                else {
                    p1 = v1;
                    p2 = v2;
                    vIndex += 2;
                }
                twbArray[i] = p1;
                twbArray[i + 1] = p2;
            }

            IntDataSeries barriers = new IntDataSeries(twbArray, 2);
            barriers.setTitle("Thin Wall Barriers");
            add(barriers);
        }
	}

//------------------------------------------------------------------------------
/**
 * Write the time series back to a file.
 * <p><b> Pre Conditions </b><br>
 *      The header has been written with writeHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 */
    protected void writeData () throws IOException {


        RawBathymetry bath = null;

        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();

            // If gro is an object effigy, get the eventual subject and try
            // adding that.
            while (gro instanceof gov.usgs.gr.ObjectEffigy)
                gro = ((ObjectEffigy)gro).getSubject();

            if (gro instanceof RawBathymetry) {
                bath = (RawBathymetry)gro;
                break;
            }
        }

        if (bath == null)
            return;

        IntDataSeries twbSeries = bath.getBarrierSeries();

        // Write the title out as a comment on the first line.
        GrHashList bathProps = bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("title");
        String title = ((String)p.value);
        if (title.length() > 0) {
            m_bw.write("// " + title + "\n", 0, title.length() + 4);
            m_bw.write("//" + "\n", 0, 3);
        }

        if (twbSeries != null) {
            String line1 = "// INTERIOR THIN-WALL BARRIERS" + "\n";
            m_bw.write(line1, 0, line1.length());
            int numU = countU(twbSeries);
            String numString = AsciiIO.padLeft("" + numU, 6);
            String line2 = "nbarru=" + numString + "\n";
            m_bw.write(line2, 0, line2.length());
            writeBarriers(twbSeries, true);
            numString = AsciiIO.padLeft("" + (twbSeries.getCount() - numU), 6);
            String line3 = "nbarrv=" + numString + "\n";
            m_bw.write(line3, 0, line3.length());
            writeBarriers(twbSeries, false);
        }

        IntDataSeries drySeries = bath.getDryCellSeries();

        if (drySeries != null) {
            String line1 = "";
            if (twbSeries != null)
                line1 += ("//" + "\n");
            line1 += "// INTERIOR DRY CELLS" + "\n";
            m_bw.write(line1, 0, line1.length());
            String num = AsciiIO.padLeft("" + drySeries.getCount(), 6);
            String line2 = "ndrycl=" + num + "\n";
            m_bw.write(line2, 0, line2.length());
            writeBarriers(drySeries, true);
        }
    }

//------------------------------------------------------------------------------
/**
 * Write the series data to the file.
 * @param series    The series to write
 */
    protected void writeBarriers (IntDataSeries series,
        boolean bWriteU) throws IOException {

        int[] array = series.getArray();
        int numDims = series.getNumDimensions();
        String textLine = "";

        for (int n = 0; n < array.length; n++) {
            int i = array[n] / 10000;
            int j = array[n] - (i * 10000);

            if (numDims == 1) {
                i += 2;
                j += 2;
                textLine += AsciiIO.padLeft("" + i, 4);
                textLine += ", ";
                textLine += AsciiIO.padLeft("" + j, 4);

                m_bw.write(textLine + "\n", 0, textLine.length() + 1);
                textLine = "";
            }
            if (numDims == 2) {
                i += 1;
                j += 1;

                int i2 = array[n + 1] / 10000;
                int j2 = array[n + 1] - (i2 * 10000);
                i2 += 1;
                j2 += 1;
                textLine += AsciiIO.padLeft("" + i2, 4);
                textLine += ", ";
                textLine += AsciiIO.padLeft("" + j2, 4);
                if ((bWriteU && (i == i2)) || (!bWriteU && (j == j2)))
                    m_bw.write(textLine + "\n", 0, textLine.length() + 1);
                textLine = "";
                n++;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Count the number of barriers along the U sides of cells.
 * @param array
 */
    protected int countU (IntDataSeries series) {

        int[] array = series.getArray();
        int numDims = series.getNumDimensions();

        int numU = 0;
        for (int n = 0; n < array.length; n += 2) {
            int i = array[n] / 10000;
            int j = array[n] - (i * 10000);

                int i2 = array[n + 1] / 10000;
                int j2 = array[n + 1] - (i2 * 10000);
                if (i == i2)
                    numU++;
        }

        return numU;
    }
}
