/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-05
 * Date started:    2000-10-06
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.sfhydro.data.gui.ArcGridAsciiOptionsPanel;
import gov.usgs.sfhydro.si3d.*;
import gov.usgs.sfhydro.si3d.gui.*;
import java.awt.Frame;
import java.io.*;
import java.text.*;
import java.util.*;
import javax.swing.JDialog;
import javax.swing.JFrame;

//==============================================================================
/**
 * An array of data in one or more dimensions.
 * @see gov.usgs.data.XyData.
 */
public class Si3dBathymetryFile extends gov.usgs.gr.GrFile {
    
    /** The title of the data set. */
    protected String m_rawBathTitle = "";
    /** The name of the data set's file. */
    protected String m_filename = "";
    /** A list of lines that are read or written at the beginning of the file.*/
    protected List<String> m_headerLines = new ArrayList<String>();
    /** The number of header lines that should be read. */
    protected int m_numHeaderLines = 3;
    /**
     * A String containing the characters which will be treated as delimiters
     * between data fields in the input file.
     */
    protected String m_inputDelimiters = " \t\n\r\f,";
    /** The first header line read or written from the file. */
    protected String m_headerLine1Extra = "";
    /** The number of data arrays in the set. */
    protected int m_numChannels = 2;
    /**
     * The raw bathymetry that has been read in by this file object, or
     * is to be written out by this file object.
     */
    protected RawBathymetry m_rawBath = null;
    /** An array holding the minimum values for each dimension. */
    protected int[] m_min;
    /** An array holding the maximum values for each dimension. */
    protected int[] m_max;
    /** The number of dimensions in each data series. */
    protected int m_numDimensions = 2;
    /** The number of points in the I direction. */
    protected int m_numI = 0;
    /** The number of points in the J direction. */
    protected int m_numJ = 0;
    /** The number of values on each line. */
    protected int m_valuesPerLine = 18;
    /** The spacing between points. */
    protected double m_gridWidth = 10;
    /** The data source buffered reader. */
    protected BufferedReader m_br = null;
    /** The data destination buffered writer. */
    protected BufferedWriter m_bw = null;
    /** The value in the file that represents dry land. */
    final int DRY_LAND = -90;
    /** The indicator for dry land in the file. */
    protected String m_noDataIndicator = "-90";
    /** Whether the depths in the file are in meters. */
    protected boolean m_bFileInMeters = true;
    /** Whether the depths are displayed on the screen in meters. */
    protected boolean m_bDisplayInMeters = true;
    /** The number by which depths in the file are multiplied. */
    protected double m_fileMultiplier = 10;
    /** The amount depths on the screen are shifted. */
    protected double m_displayMultiplier = 1;
    /** The amount depths in the file are shifted. */
    protected double m_fileShift = 0;
    /** The amount depths on the screen are shifted. */
    protected double m_displayShift = 0;
    /** Whether to enable the file fields in the options dialog. */
    protected boolean m_bEnableFileFields = false;
    /** Whether to enable the display fields in the options dialog. */
    protected boolean m_bEnableDisplayFields = true;

//------------------------------------------------------------------------------
    /**
     * Creates a data file.
     */
    public Si3dBathymetryFile() {
        
        super();
        m_title = "Si3d Bathymetry File";
        
        m_min = new int[m_numDimensions];
        m_max = new int[m_numDimensions];
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     * @return  The ordered list of GrProperty objects.
     */
    public GrHashList getProperties() {
        
        GrHashList props = super.getProperties();
        
        props.put("fileInMeters", new GrProperty(m_bFileInMeters,
            "File in Meters"));
        props.put("displayInMeters", new GrProperty(m_bDisplayInMeters,
            "Display in Meters"));
        props.put("fileMultiplier", new GrProperty(m_fileMultiplier,
            "File Multiplier"));
        props.put("displayMultiplier", new GrProperty(m_displayMultiplier,
            "Display Multiplier"));
        props.put("fileShift", new GrProperty(m_fileShift,
            "File Shift"));
        props.put("displayMultiplier", new GrProperty(m_displayMultiplier,
            "Display Shift"));
        
        return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Set properties from a list of values.
     * @param props The list of GrProperty objects.
     */
    public void setProperties(HashMap props) {
        
        super.setProperties(props);
        
        GrProperty p;
        
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;
                
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                
                if (key.equals("fileInMeters"))
                    m_bFileInMeters = (Boolean)p.value;
                if (key.equals("displayInMeters"))
                    m_bDisplayInMeters = (Boolean)p.value;
                if (key.equals("fileMultiplier"))
                    m_fileMultiplier = (Double)p.value;
                if (key.equals("displayMultiplier"))
                    m_displayMultiplier = (Double)p.value;
                if (key.equals("fileShift"))
                    m_fileShift = (Double)p.value;
                if (key.equals("displayShift"))
                    m_displayShift = (Double)p.value;
                
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Set a single property.
     * @param key   The name of the property being set.
     * @param value The value of the property being set.
     */
    public void setProperty(String key, String value) {
        
        super.setProperty(key, value);
        
        if (key.equals("fileInMeters"))
            m_bFileInMeters = Boolean.parseBoolean(value);
        if (key.equals("displayInMeters"))
            m_bDisplayInMeters = Boolean.parseBoolean(value);
        if (key.equals("fileMultiplier"))
            m_fileMultiplier = Double.parseDouble(value);
        if (key.equals("displayMultiplier"))
            m_displayMultiplier = Double.parseDouble(value);
        if (key.equals("fileShift"))
            m_fileShift = Double.parseDouble(value);
        if (key.equals("displayShift"))
            m_displayShift = Double.parseDouble(value);
        
    }
     
//------------------------------------------------------------------------------
    /**
     * Overrides add in GrObject.  Adds everything as an effigy.
     */
    public boolean add(GrObject gro) {
        
        if (gro instanceof ObjectEffigy) {
            if (((ObjectEffigy)gro).depictProperties == false) {
                m_children.add(gro);
            } else {
                m_children.add(new ObjectEffigy(gro));
            }
        } else {
            m_children.add(new ObjectEffigy(gro));
        }
        
        return true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Updates the member variables to reflect any changes in the bathymetry
     * object child of this object.
     */
    public void checkForNewRawBath() {
        
        RawBathymetry bath = null;
        boolean bMatch = false;
        
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            
            // If gro is an object effigy, get the eventual subject and try
            // adding that.
            while (gro instanceof gov.usgs.gr.ObjectEffigy)
                gro = ((ObjectEffigy)gro).getSubject();
            
            if (gro instanceof RawBathymetry) {
                if (!bMatch) {
                    bath = (RawBathymetry)gro;
                    if (bath == m_rawBath)
                        bMatch = true;
                }
            }
        }
        
        if (bath != null) {
            m_rawBath = bath;
            GrHashList bathProps = m_rawBath.getProperties();
            GrProperty p = (GrProperty)bathProps.get("title");
            m_rawBathTitle = ((String)p.value);
            p = (GrProperty)bathProps.get("xSpacing");
            m_gridWidth = (Double)p.value;
            p = (GrProperty)bathProps.get("numI");
            m_numI = (Integer)p.value;
            p = (GrProperty)bathProps.get("numJ");
            m_numJ = (Integer)p.value;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Reads the file from disk.
     * <p><b> Pre Conditions </b><br>
     *      The filename is not null.
     * @param filename  The name of the file to load.
     */
    public void read() throws IOException {
        
        // Open the file.
        m_br = new BufferedReader(new FileReader(
                m_file.getAbsolutePath()));
        // Read the header.
        readHeader();
        // Read the data.
        readData();
        // Close the file.
        m_br.close();
    }
    
//------------------------------------------------------------------------------
    /**
     * Saves the data to a file.
     * <p><b> Pre Conditions </b><br>
     *      The filename is not null.
     * <p><b> Post Conditions </b><br>
     *      None.
     * @param filename  The name of the file to save to.
     */
    public void write() throws IOException {
        
        // Update member variables if there was a change in the bathymetry.
        checkForNewRawBath();
        
        // Open the file.
        m_bw = new BufferedWriter(new FileWriter(
                m_file.getAbsolutePath()));
        
        // Read the header.
        writeHeader();
        // Read the data.
        writeData();
        // Close the file.
        m_bw.close();
    }
    
//------------------------------------------------------------------------------
    /**
     * Read the header information from an open file attached to a reader.
     * <p><b> Pre Conditions </b><br>
     *      None.
     * <p><b> Post Conditions </b><br>
     *      None.
     */
    protected void readHeader() throws IOException {
        
        for (int i = 0; i < m_numHeaderLines; i++) {
            m_headerLines.add(AsciiIO.readLine(m_br));
        }
        
        // Pick the values of imx and jmx off the top line.  They should
        // be of a similar form to "imx = 200 jmx = 200"
        String topLine = (String)m_headerLines.get(0);
        String[] tokens = AsciiIO.parseLine(topLine, m_inputDelimiters);
        int index;
        for (int i = 0; i < tokens.length; i++) {
            index = tokens[i].indexOf("imx");
            if (index >= 0) {
                index = tokens[i].indexOf("=", index);
                if (index == (tokens[i].length() - 1)) {
                    m_numI = Integer.parseInt(tokens[i + 1]);
                } else {
                    index = tokens[i+1].indexOf("=");
                    if (index != (tokens[i+1].length() - 1)) {
                        m_numI = Integer.parseInt(
                                tokens[i + 1].substring(index + 1));
                    } else {
                        m_numI = Integer.parseInt(tokens[i + 2]);
                    }
                }
            }
            
            index = tokens[i].indexOf("jmx");
            if (index >= 0) {
                index = tokens[i].indexOf("=", index);
                if (index == (tokens[i].length() - 1)) {
                    m_numJ = Integer.parseInt(tokens[i + 1]);
                } else {
                    index = tokens[i+1].indexOf("=");
                    if (index != (tokens[i+1].length() - 1)) {
                        m_numJ = Integer.parseInt(
                                tokens[i + 1].substring(index + 1));
                    } else {
                        m_numJ = Integer.parseInt(tokens[i + 2]);
                    }
                }
            }
        }
        
        int titleEndIndex = topLine.indexOf("imx") - 1;
        ;
        if (titleEndIndex > 0) {
            m_rawBathTitle = topLine.substring(0, titleEndIndex + 1);
            m_rawBathTitle = m_rawBathTitle.trim();
            if (m_rawBathTitle.endsWith(","))
                m_rawBathTitle =
                        m_rawBathTitle.substring(0, m_rawBathTitle.length() - 1);
        }
        
//        m_iSize = Integer.parseInt(topLine.substring(37, 41).trim());
//        m_jSize = Integer.parseInt(topLine.substring(48, 52).trim());
    }
    
//------------------------------------------------------------------------------
    /**
     * Read the data through a reader attached to an open file.
     * <p><b> Pre Conditions </b><br>
     *      The header has been read with readHeader()
     * <p><b> Post Conditions </b><br>
     *      The file needs to be closed.
     */
    protected void readData() throws IOException {
        
        clear();
        
        // Make a big data series for storage.
        DoubleDataSeries triFanSeries = new DoubleDataSeries(3);
        DoubleDataSeries rectSeries = new DoubleDataSeries(3);
//        int[] array = new int[m_numI * m_numJ];
        double[] cornerDepths = new double[m_numI * m_numJ];
        
        int lineNum = 0;
        int[] vals;
        
        // Read each line of the data file.
        while (true) {
            String textLine = m_br.readLine();
            if (textLine == null) break;
            if (textLine.length() < 4) break;
            String statedLineText = textLine.substring(0, 4).trim();
            int statedLineNum = Integer.parseInt(statedLineText);
            vals = AsciiIO.parseLineAsInts(textLine, 1,
                    textLine.length() - 1, 4);
            if (vals == null) break;
            if (vals.length <= 0)
                continue;
            
            int calculatedJ = (lineNum % m_numJ) + 1;
            int jOffset = m_numJ - calculatedJ;
            if ((jOffset + 1) != statedLineNum)
                System.out.println("J value " + statedLineNum + " at line "
                        + lineNum + " does not match caculated value of "
                        + (jOffset + 1));
            
            for (int i = 1; i < vals.length; i++) {
                int iOffset = ((lineNum / m_numJ) * m_valuesPerLine) + i - 1;
                try {
//                    array[iOffset + (jOffset * m_numI)] = vals[i];
                    double depth = ((double)vals[i]) / 10;
                    if (!m_bDisplayInMeters)
                        depth *= 3.2808399;
                    depth *= m_displayMultiplier;
                    depth += m_displayShift;
                    cornerDepths[iOffset + (jOffset * m_numI)] = depth;
                } catch (Exception e) {
                    System.out.println("Exception (" + e + ")");
                    System.out.println("Line num = " + lineNum
                            + ", I Offset = "
                            + iOffset + ", J Offset = " + jOffset
                            + ", imx = " + m_numI + ", jmx = " + m_numJ);
                    System.out.println("Text line ->" + textLine + '\n');
                    break;
                }
            }
            
            lineNum++;
        }
        
        // Put the 1D corner depth series into a GlBathymetryGrid and add the
        // grid to this file.
        DoubleDataSeries cornerDepthSeries = new DoubleDataSeries(1);
        cornerDepthSeries.setArray(cornerDepths);
        RawBathymetry bath = new RawBathymetry();
        bath.setProperty("title", m_rawBathTitle);
        bath.setProperty("xSpacing", "" + m_gridWidth);
        bath.setProperty("ySpacing", "" + m_gridWidth);
        bath.setProperty("zSpacing", "2");
        bath.setProperty("numI", "" + m_numI);
        bath.setProperty("numJ", "" + m_numJ);
        bath.add(cornerDepthSeries);
        gov.usgs.sfhydro.si3d.gl.GlBathymetryGrid glGrid =
                new gov.usgs.sfhydro.si3d.gl.GlBathymetryGrid();
        glGrid.add(bath);
        glGrid.regenerateColorScheme();
        glGrid.setProperty("zScale", "0");
        gov.usgs.gr.visual.gl.GlGraph graph =
                new gov.usgs.gr.visual.gl.GlGraph();
        graph.add(glGrid);
        gov.usgs.gr.visual.gl.PageOfGlGraphs page =
                new gov.usgs.gr.visual.gl.PageOfGlGraphs();
        page.add(graph);
        page.updateZoomContents();
        page.showMoreGraphs();
        page.setProperty("formatter",
                "gov.usgs.sfhydro.gr.formats.VariableSizeFormatter");
        gov.usgs.gr.visual.Axis xAxis = graph.getXAxis();
        gov.usgs.gr.visual.Axis yAxis = graph.getYAxis();
        xAxis.setProperty("showMajorTicks", "false");
        yAxis.setProperty("showMajorTicks", "false");
        xAxis.setProperty("showMinorTicks", "false");
        yAxis.setProperty("showMinorTicks", "false");
        xAxis.setProperty("showLabels", "false");
        yAxis.setProperty("showLabels", "false");
        xAxis.setProperty("showGridLines", "false");
        yAxis.setProperty("showGridLines", "false");
        xAxis.getZoom().setOuterLimits(0, m_numI * m_gridWidth);
        yAxis.getZoom().setOuterLimits(0, m_numJ * m_gridWidth);
        xAxis.getZoom().zoomAll();
        yAxis.getZoom().zoomAll();
        xAxis.getZoom().setContentLimits(0, m_numI * m_gridWidth);
        yAxis.getZoom().setContentLimits(0, m_numJ * m_gridWidth);
        add(page);
        add(glGrid);
        add(bath);
        RawBathymetryCalculator calc = new RawBathymetryCalculator();
        calc.add(bath);
        add(calc);
        add(new BarrierEditorProvider());
    }
    
//------------------------------------------------------------------------------
    /**
     * Retrieves the value from a 1D array for a given (I, J) location.  Checks
     * if the indexes are out of bounds or if the point is DRY_LAND and returns
     * 0 if they are.
     * @param i		The i index of the point in the array.
     * @param j		The j index of the point in the array.
     * @param array	The array to retrieve the value from.
     * @param arrayWidth	The width of the I dimension in the array.
     * @return	The value at the specified location, or 0 if the location is
     * 			out of bounds or DRY_LAND.
     */
    private double getDepthAt(int i, int j, int[] array, int arrayWidth) {
        
        if (i < 0) i = 0;
        if (j < 0) j = 0;
        if (i >= arrayWidth)
            i = arrayWidth - 1;
        if (j >= (array.length / arrayWidth))
            j = (array.length / arrayWidth) - 1;
        
        double val = array[j * arrayWidth + i];
        if (val == DRY_LAND)
            return 0;
        else
            return (val / 10);
        
    }
    
//------------------------------------------------------------------------------
    /**
     * Write the header information to an open file attached to a reader.
     */
    private void writeHeader() throws IOException {
        
        try {
            String line1 = m_rawBathTitle;
            if (line1.length() > 31)
                line1 = line1.substring(0, 31);
            if (line1.length() > 0)
                line1 += ',';
            while (line1.length() < 32)
                line1 += ' ';
            line1 += "imx =";
            line1 += AsciiIO.padLeft("" + m_numI, 5);
            line1 += ",jmx =";
            line1 += AsciiIO.padLeft("" + m_numJ, 5);
            line1 += ",ncols =  18";
            String line2 = "HV       V   V   V   V   V   V   V   V   V   V   V"
                    + "   V   V   V   V   V   V";
            String line3 = "HU     U z U z U z U z U z U z U z U z U z U z U z"
                    + " U z U z U z U z U z U z U";
            
            m_bw.write(line1 + "\n", 0, line1.length() + 1);
            m_bw.write(line2 + "\n", 0, line2.length() + 1);
            m_bw.write(line3 + "\n", 0, line3.length() + 1);
            Iterator iter = m_headerLines.iterator();
            while (iter.hasNext()) {
                String s = (String)iter.next();
                m_bw.write(s + "\n", 0, s.length() + 1);
            }
        } catch (IOException e) {
            javax.swing.JOptionPane.showMessageDialog(null,
                    "Error saving header of file " + m_filename, "IO Exception",
                    javax.swing.JOptionPane.ERROR_MESSAGE);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Write the time series back to a file.
     * <p><b> Pre Conditions </b><br>
     *      The header has been written with writeHeader()
     * <p><b> Post Conditions </b><br>
     *      The file needs to be closed.
     */
    protected void writeData() throws IOException {
        
        DoubleDataSeries series = m_rawBath.getBathymetrySeries();
        int numPointsRemaining = series.getNumPoints();
        double[] array = series.getArray();
        
        // Read each line of the data file.
        int valuesOnThisLine = m_valuesPerLine;
        int lineNum = 0;
        while (numPointsRemaining > 0) {
            String textLine = "";
            NumberFormat form = new DecimalFormat("###0");
            
            int calculatedJ = (lineNum % m_numJ) + 1;
            int jOffset = m_numJ - calculatedJ;
            textLine += AsciiIO.padLeft(form.format(jOffset + 1), 4);
            textLine += " ";
            
            if ((lineNum % m_numJ) == 0) {
                if (numPointsRemaining < (m_numJ * m_valuesPerLine))
                    valuesOnThisLine = numPointsRemaining / m_numJ;
            }
            
            for (int i = 0; i < valuesOnThisLine; i++) {
                int iOffset = ((lineNum / m_numJ) * m_valuesPerLine) + i;
                double val = array[iOffset + (jOffset * m_numI)];
                if (m_displayShift != 0)
                    val -= m_displayShift;
                if (m_displayMultiplier != 1)
                    val /= m_displayMultiplier;
                if (!m_bDisplayInMeters)
                    val *= 0.3048;
                val = val * 10;
                if (val < -50 || Double.isNaN(val))
                    textLine += " -90";
//                    textLine += (" " + m_noDataIndicator);
                else
                    textLine +=
                            AsciiIO.padLeft(form.format(Math.round(val)), 4);
            }
            
            m_bw.write(textLine + "\n", 0, textLine.length() + 1);
            numPointsRemaining -= valuesOnThisLine;
            lineNum++;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Write thin-wall barrier and dry cell files if there are any in the
     * raw bathymetry.
     */
    protected void writeOtherFiles() throws IOException {
        
        RawBathymetry bath = null;
        
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            
            // If gro is an object effigy, get the eventual subject and try
            // adding that.
            while (gro instanceof gov.usgs.gr.ObjectEffigy)
                gro = ((ObjectEffigy)gro).getSubject();
            
            if (gro instanceof RawBathymetry) {
                bath = (RawBathymetry)gro;
                break;
            }
        }
        
        if (bath == null)
            return;
        
        Si3dBarrierFile barrierFile = new Si3dBarrierFile();
        barrierFile.add(bath);
        barrierFile.setProperty("file", "bar_" + m_file.getAbsolutePath());
        barrierFile.write();
    }
    
//------------------------------------------------------------------------------
    /**
     * Tells whether this file options has an options dialog.
     * @return true if the showOptionsDialog() method actually does something
     *              in this file class.
     */
    public boolean hasOptionsDialog() {
        
        return true;
    }
    
//------------------------------------------------------------------------------
    /**
     * Shows a dialog for configuring the file.
     * @param frame The frame to use as a parent for a modal dialog box.
     */
    public void showOptionsDialog(Frame frame) {
        
        JDialog dialog = new JDialog(new JFrame(), "Options");
        dialog.setModal(true);
        ArcGridAsciiOptionsPanel panel = new ArcGridAsciiOptionsPanel();
        panel.setFileUnits(m_bFileInMeters ? 0 : 1);
        panel.setDisplayUnits(m_bDisplayInMeters ? 0 : 1);
        panel.setFileMultiplier(m_fileMultiplier);
        panel.setDisplayMultiplier(m_displayMultiplier);
        panel.setFileShift(m_fileShift);
        panel.setDisplayShift(m_displayShift);
        panel.enableFileFields(m_bEnableFileFields);
        panel.enableDisplayFields(m_bEnableDisplayFields);
        dialog.add(panel);
        panel.setDialog(dialog);
        dialog.pack();
        dialog.setVisible(true);
        m_bFileInMeters = (panel.getFileUnits() == 0 ? true : false);
        m_bDisplayInMeters = (panel.getDisplayUnits() == 0 ? true : false);
        m_fileMultiplier = panel.getFileMultiplier();
        m_displayMultiplier = panel.getDisplayMultiplier();
        m_fileShift = panel.getFileShift();
        m_displayShift = panel.getDisplayShift();
    }
}
