/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    2000-08-22
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.io.*;
import java.util.*;

//==============================================================================
/**
 * Reads the Si3D output text space file.
 */
public class Si3dSpaceFile99 extends gov.usgs.gr.GrFile {

    /** The title of the data set. */
    protected String m_title = "";
    /** The name of the data set's file. */
    protected String m_filename = "";
    /** A list of lines that are read or written at the beginning of the file.*/
    protected List m_headerLines = new ArrayList();
    /** The number of header lines that should be read. */
    protected int m_numHeaderLines = 5;
    /**
     * A String containing the characters which will be treated as delimiters
     * between data fields in the input file.
     */
    protected String m_inputDelimiters = " \t\n\r\f,";
    /** An array containing all the data set's data arrays. */
    protected double[][] m_seriesArrays;
    /** The number of data arrays in the set. */
    protected int m_numChannels = 2;
    /** An array holding the minimum values for each dimension. */
    protected int[] m_min;
    /** An array holding the maximum values for each dimension. */
    protected int[] m_max;
    /** The number of dimensions in each data series. */
    protected int m_numDimensions = 2;
    /** The data source buffered reader. */
    protected BufferedReader m_br = null;
    /** The data destination buffered writer. */
    protected BufferedWriter m_bw = null;

//------------------------------------------------------------------------------
/**
 * Creates a data file.
 */
    public Si3dSpaceFile99 () {

    	super();
        m_title = "Delimited Data File";

        m_min = new int[m_numDimensions];
        m_max = new int[m_numDimensions];
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.  Adds everything as an effigy.
 */
    public boolean add(GrObject gro) {

        if (gro instanceof ObjectEffigy) {
        	if (((ObjectEffigy)gro).depictProperties == false) {
        		m_children.add(gro);
            }
        	else {
        		m_children.add(new ObjectEffigy(gro));
        	}
        }
        else {
        	m_children.add(new ObjectEffigy(gro));
        }

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Reads the file from disk.
 */
    public void read () throws IOException {

    	// Open the file.
        m_br = new BufferedReader(new FileReader(
        	m_file.getAbsolutePath()));
        // Read the header.
        readHeader();
        // Read the data.
        readData();
        // Close the file.
        m_br.close();
	}

//------------------------------------------------------------------------------
/**
 * Saves the data to a file.
 */
    public void write() throws IOException {

    	// Open the file.
        m_bw = new BufferedWriter(new FileWriter(
        	m_file.getAbsolutePath()));

        // Read the header.
        writeHeader();
        // Read the data.
        writeData();
        // Close the file.
        m_bw.close();
 	}

//------------------------------------------------------------------------------
/**
 * Read the header information from an open file attached to a reader.
 * <p><b> Pre Conditions </b><br>
 *      None.
 * <p><b> Post Conditions </b><br>
 *      None.
 */
    protected void readHeader () throws IOException {

    	for (int i = 0; i < m_numHeaderLines; i++) {
        	m_headerLines.add(AsciiIO.readLine(m_br));
        }
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.
 * Overrides readData() from gov.usgs.data.DimensionalDataFile.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 */
    protected void readData () throws IOException {

    	clear();

        // Make a couple of big data series for storage.
//        DataSeries channelTimeDS = new DataSeries(2);
//        DataSeries bigDS = new DataSeries(7);

        // Make a list of column lists, one for each series.
        List columnList = new ArrayList();
        for (int i = 0; i < m_numChannels; i++) {
            columnList.add(new ArrayList());
        }

    	double[] vals;
        int seriesIndex = 0;
        int dimIndex = 0;

        List column0 = (List)columnList.get(0);
        List column1 = (List)columnList.get(1);

        // Read each line of the data file.
		boolean bLookingForTime = false;
		while (true) {
        	vals = AsciiIO.parseLineAsDoubles(m_br.readLine(),
                m_inputDelimiters);
            if (vals == null) break;

            // Set up to read a repitition of two lines with one value each
            // followed by multiple lines with multiple values each.  The second
            // single-value line is time.  When the first value of a multivalue
            // line is 50000, the third value on that line is stored as the
            // second dimension of the data series.
            if (bLookingForTime) {
            	if (vals.length == 1) {
                    column0.add(vals[0]);
                    column1.add(vals[0]);
                    bLookingForTime = false;
                }
            }
            else {
            	if (vals.length < 3) {
                    bLookingForTime = true;
                }
                else {
                	if (vals[0] == 50000) {
//                	if (vals[0] == 0)
                        column0.add(vals[2]);
                        column1.add(vals[4]);
					}
                }
            }
        }

        // Transfer the Double to an array of doubles and create data series.
        seriesIndex = 0;
        Iterator iter = columnList.iterator();
        while (iter.hasNext()) {
        	List column = (List)iter.next();
            Object[] realObjects = column.toArray();
            vals = new double[realObjects.length];
            for (int i = 0; i < realObjects.length; i++) {
            	vals[i] = (Double)realObjects[i];
            }

            // Create a new data series from the array.
        	DataSeries series = null;
        	if (m_numDimensions == 2)
		    series = new DoubleDataSeries(vals, 2);
        	else
		    series = new DoubleDataSeries(vals, m_numDimensions);
            if (seriesIndex == 0) series.setTitle("Flow in CFS");
            if (seriesIndex == 1) series.setTitle("Cross Secional Area");
            add(series);

            seriesIndex++;
        }
    }

//------------------------------------------------------------------------------
/**
 * Write the header information to an open file attached to a reader.
 */
    private void writeHeader () throws IOException {

    }

//------------------------------------------------------------------------------
/**
 * Write the time series back to a file.
 * <p><b> Pre Conditions </b><br>
 *      The header has been written with writeHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 */
    protected void writeData () throws IOException {

    }
}
