/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-03
 * Date started:    2001-09-03
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.data;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.*;
import gov.usgs.gr.visual.gl.*;
import java.io.*;
import java.util.*;
import java.text.SimpleDateFormat;
import java.text.DecimalFormat;
import java.text.FieldPosition;
import java.text.NumberFormat;
import java.awt.EventQueue;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import javax.swing.JDialog;
import javax.swing.JTextField;
import javax.swing.BoxLayout;
import javax.swing.Box;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JButton;
import java.awt.Dimension;
import java.lang.reflect.InvocationTargetException;
import javax.swing.JCheckBox;
import java.awt.Cursor;

//==============================================================================
/**
 * Reads data in VPV (Velocity Profile Viewer) format and sets up display
 * objects.
 */
public class VPVDataFile extends gov.usgs.gr.data.TimeDataFile {

    /** An array containing all the data set's data arrays. */
    protected double[][] m_seriesArrays;
    /** The name of the collection station or location. */
    protected String m_stationName = "";
    /** The number of steps between data. */
    protected double m_timeStep = 0;
    /** The number of vertical bins in each time step. */
    protected int m_numBins = 0;
    /** The number of data arrays in the set. */
    protected int m_numChannels = 4;
    /** An array holding the minimum values for each dimension. */
    protected int[] m_min;
    /** An array holding the maximum values for each dimension. */
    protected int[] m_max;
    /** The number of dimensions in each data series. */
    protected int m_numDimensions = 2;
    /** The number of points in the I direction. */
    protected int m_iSize = 0;
    /** The number of points in the J direction. */
    protected int m_jSize = 0;
    protected double m_principalDirection = 90;
    protected boolean m_bCreateProfilePage = true;
    protected boolean m_bCreateUvPage = false;
    protected boolean m_bCreateDirMagPage = false;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/VPVDataFile.gif";

//------------------------------------------------------------------------------
/**
 * Creates a data file.
 */
    public VPVDataFile () {

    	super();
        m_title = "VPV Data File";
        m_numHeaderLines = 4;

        m_min = new int[m_numDimensions];
        m_max = new int[m_numDimensions];
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

		props.put("prinDir", new GrProperty(m_principalDirection,
            "Principal Direction", false, false, false));
		props.put("createProfilePage", new GrProperty(m_bCreateProfilePage,
            "Create 3D Profile Page", false, false, false));
		props.put("createUvPage", new GrProperty(m_bCreateUvPage,
            "Create U-V Page", false, false, false));
		props.put("createDirMagPage", new GrProperty(m_bCreateDirMagPage,
            "Create Direction-Magnitude Page", false, false, false));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @param props  A list of GrProperty objects specifying the new values for
 *               certain properties.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (key.equals("prinDir"))
                    m_principalDirection = (Double)p.value;
                if (key.equals("createProfilePage"))
                    m_bCreateProfilePage = (Boolean)p.value;
                if (key.equals("createUvPage"))
                    m_bCreateUvPage = (Boolean)p.value;
                if (key.equals("createDirMagPage"))
                    m_bCreateDirMagPage = (Boolean)p.value;
            }
            if (obj instanceof String) {
                String value = (String)obj;

                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.sfhydro.data.VPVDataFile#getProperties getProperties() for a
 * list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

		super.setProperty(key, value);

        if (key.equals("prinDir"))
            m_principalDirection = Double.parseDouble(value);
        if (key.equals("createProfilePage"))
            m_bCreateProfilePage = Boolean.parseBoolean(value);
        if (key.equals("createUvPage"))
            m_bCreateUvPage = Boolean.parseBoolean(value);
        if (key.equals("createDirMagPage"))
            m_bCreateDirMagPage = Boolean.parseBoolean(value);
    }
//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.  Adds everything as an effigy.
 */
    public boolean add(GrObject gro) {

        if (gro instanceof ObjectEffigy) {
        	if (((ObjectEffigy)gro).depictProperties == false) {
        		m_children.add(gro);
            }
        	else {
        		m_children.add(new ObjectEffigy(gro));
        	}
        }
        else {
        	m_children.add(new ObjectEffigy(gro));
        }

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Read the header information from an open file attached to a reader.
 */
    protected void readHeader () throws IOException {

    	for (int i = 0; i < m_numHeaderLines; i++) {
        	m_headerLines.add(AsciiIO.readLine(m_br));
        }

        int binLineNum = 2;

        // The station name is columns 0-40 on top line.
        m_stationName = (String)m_headerLines.get(0);
        int maxLength =
            (m_stationName.length() < 41) ? m_stationName.length() : 41;
        m_stationName = m_stationName.substring(0, maxLength).trim();

//        // Skip an extra line if this has Jon's jdate line as line 2.
//        try {
//            String s = ( (String) m_headerLines.get(1)).substring(13, 14);
//            if ( (s.compareTo("0") >= 0 && s.compareTo("9") <= 0) ||
//                ( (String) m_headerLines.get(2)).length() < 1) {

                m_headerLines.add(AsciiIO.readLine(m_br));
                binLineNum++;
//            }
//        }
//        catch (StringIndexOutOfBoundsException sioobe) {}

        // Parse the second to last header line.
        String binLine = (String)m_headerLines.get(binLineNum);
        String[] tokens = AsciiIO.parseLine(binLine, m_inputDelimiters);
        if (tokens.length >= 3) {
            int firstBin = Integer.parseInt(tokens[0]);
            int lastBin = Integer.parseInt(tokens[1]);
            m_timeStep = Double.parseDouble(tokens[2]) / 24;
            m_numBins = lastBin - firstBin + 1;
        }

//        String lastLine = (String)m_headerLines.get(binLineNum + 1);
//        tokens = AsciiIO.parseLine(lastLine, m_inputDelimiters);
    }

//------------------------------------------------------------------------------
/**
 * Read the data through a reader attached to an open file.
 * Overrides readData() from gov.usgs.data.DimensionalDataFile.
 * <p><b> Pre Conditions </b><br>
 *      The header has been read with readHeader()
 * <p><b> Post Conditions </b><br>
 *      The file needs to be closed.
 */
    protected void readData () throws IOException {

    	clear();

        // Make a big data series for storage.
        DataSeries waterLevelSeries = new TimeSeries(2);
        DataSeries vpSeries = new DoubleDataSeries(4);

    	clear();

        // Make a list of column lists, one for each series.
        List columnList = new ArrayList();
        for (int i = 0; i < m_numChannels; i++) {
            columnList.add(new ArrayList());
        }

        List dateList = (List)columnList.get(0);
        List zList = (List)columnList.get(1);
        List uList = (List)columnList.get(2);
        List vList = (List)columnList.get(3);

        List wlColumnList = new ArrayList();
        List wlDateList = new ArrayList();
        List wlList = new ArrayList();
        wlColumnList.add(wlDateList);
        wlColumnList.add(wlList);

        int currentBin = Integer.MAX_VALUE;
        String[] tokens;
        Double waterLevel = 0d;
        GregorianCalendar date = null;
        int referenceYear = Integer.MIN_VALUE;

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();

        // Read each line of the data file.
        while ((tokens = AsciiIO.parseLine(m_br.readLine(),
        	m_inputDelimiters)) != null) {

            if (tokens.length == 0) continue;

            if (currentBin >= m_numBins) {
                if (tokens.length > 4) {
                    waterLevel = Double.parseDouble(tokens[1]);
                }
                else
                    waterLevel = 0d;

                // Calculate the year, day, hour, and minute at this index.
                int year = Integer.parseInt(tokens[tokens.length - 4]);
                year = fourDigitYear(year);
                if (referenceYear == Integer.MIN_VALUE)
                    referenceYear = year;
                int month = Integer.parseInt(tokens[tokens.length - 3]) - 1;
                int dayOfMonth = Integer.parseInt(tokens[tokens.length - 2]);
                int hourMinute = Integer.parseInt(tokens[tokens.length - 1]);
                int hourOfDay = hourMinute / 100;
                int minute = hourMinute - (hourOfDay * 100);
                minute *= 0.6;

                // Create a single date object from the data.
                date = new GregorianCalendar(utcZone, locale);
                date.clear();
                date.set(GregorianCalendar.YEAR, year);
                date.set(GregorianCalendar.MONTH, month);
                date.set(GregorianCalendar.DAY_OF_MONTH, dayOfMonth);
                date.set(GregorianCalendar.HOUR_OF_DAY, hourOfDay);
                date.set(GregorianCalendar.MINUTE, minute);

                wlDateList.add(date);
                wlList.add(waterLevel);

                currentBin = 0;

                continue;
            }

            // Add the date to the list.
            dateList.add(date);

            double val = Double.parseDouble(tokens[0]);
            if (val > 9000) val = 0;
            zList.add(val);
            val = Double.parseDouble(tokens[1]);
            if (val > 9000) val = 0;
            uList.add(val);
            val = Double.parseDouble(tokens[2]);
            if (val > 9000) val = 0;
            vList.add(val);

            currentBin++;
        }

        // Create a list containing column of dates and columns for the other
        // data so they can be interleaved into individual series.
        m_startDate = (GregorianCalendar)
            ((GregorianCalendar)dateList.get(0)).clone();
        GregorianCalendar lastDate = (GregorianCalendar)
            ((GregorianCalendar)dateList.get(dateList.size() - 1)).clone();
        m_referenceYear = findAppropriateReferenceYear(m_startDate, lastDate);

        ArrayList inputList = new ArrayList();
        inputList.add(dateList);
        inputList.add(zList);

        List<? extends TimeSeries> timeSeriesList =
            interleaveArrays(inputList);

        // Add the U and V to the T-Z series.
        TimeSeries zSeries = (TimeSeries)timeSeriesList.get(0);
        double[] tzuvArray = new double[4 * uList.size()];
        double[] zArray = zSeries.getArray();

        // Also create time series for the U and V at each bin.
        double[][] uBinArrays = new double[m_numBins][];
        double[][] vBinArrays = new double[m_numBins][];
        double[][] dBinArrays = new double[m_numBins][];
        double[][] mBinArrays = new double[m_numBins][];
        int numTimes = uList.size() / m_numBins;
        for (int i = 0; i < m_numBins; i++) {
            uBinArrays[i] = new double[numTimes * 2];
            vBinArrays[i] = new double[numTimes * 2];
            dBinArrays[i] = new double[numTimes * 2];
            mBinArrays[i] = new double[numTimes * 2];
        }
        currentBin = 0;
        int timeIndex = 0;

        // Read the U and V values from the U and V lists.
        for (int i = 0; i < uList.size(); i++) {
            // Store the values in the TZUV array.
            tzuvArray[i * 4] = zArray[i * 2];
            tzuvArray[i * 4 + 1] = zArray[i * 2 + 1];
            tzuvArray[i * 4 + 2] = (Double)uList.get(i);
            tzuvArray[i * 4 + 3] = (Double)vList.get(i);

            // Store the values in the Bin time series arrays.
            double[] uArray = uBinArrays[currentBin];
            double[] vArray = vBinArrays[currentBin];
            double[] dArray = dBinArrays[currentBin];
            double[] mArray = mBinArrays[currentBin];

            double uVal = tzuvArray[i * 4 + 2];
            double vVal = tzuvArray[i * 4 + 3];
            double cosTheta =
                Math.cos(Math.toRadians(m_principalDirection - 90));
            double sinTheta =
                Math.sin(Math.toRadians(m_principalDirection - 90));
            double rotatedU = uVal * cosTheta - vVal * sinTheta;
            double rotatedV = uVal * sinTheta + vVal * cosTheta;
            uArray[timeIndex * 2] = tzuvArray[i * 4];
            uArray[timeIndex * 2 + 1] = rotatedU;
            vArray[timeIndex * 2] = tzuvArray[i * 4];
            vArray[timeIndex * 2 + 1] = rotatedV;
            dArray[timeIndex * 2] = tzuvArray[i * 4];

            double atanAngle = 90;
            if (rotatedU != 0) {
                atanAngle = Math.toDegrees(Math.atan(rotatedV / rotatedU));
            }
            else {
                if (rotatedV < 0)
                    atanAngle = -90;
            }

            // Handle the case of angles that should be in the upper-left or
            // lower-left quadrants, but aren't because of atan's limitations.
            if (rotatedV < 0 && rotatedU < 0)
                atanAngle -= 180;
            if (rotatedV > 0 && rotatedU < 0)
                atanAngle += 180;
            dArray[timeIndex * 2 + 1] = -atanAngle + 90;
            mArray[timeIndex * 2] = tzuvArray[i * 4];
            mArray[timeIndex * 2 + 1] = Math.sqrt((rotatedU * rotatedU) +
                (rotatedV * rotatedV));
if (mArray[timeIndex * 2 + 1] < 0) {
    System.out.println("Negative magnitude -- " + rotatedU + "  " + rotatedV);
}
            currentBin++;
            if (currentBin >= m_numBins) {
                currentBin = 0;
                timeIndex++;
            }
        }

        // Add the TZUV series as a child of this file.
        DoubleDataSeries ds = new DoubleDataSeries(tzuvArray, 4);
        GlVelocityProfile vp = new GlVelocityProfile();
        vp.setProperty("width", "1");
        vp.setProperty("height", "1");
        vp.setProperty("scale", "5000");
        vp.setProperty("drawBG", "false");
        vp.setProperty("useV", "true");
        vp.add(ds);
        InterpolatingColorScheme ics = new InterpolatingColorScheme();
        ics.setColorAtValue("(255,0,102)0");
        ics.setColorAtValue("(255,255,102)" + ds.getMax(1));
        vp.add(ics);
        Gl3dAxisGrid axisGrid = new Gl3dAxisGrid();
        vp.add(axisGrid);

        // Add the water level series as a child of this file.
        inputList.clear();
        inputList.add(wlDateList);
        inputList.add(wlList);

        timeSeriesList = interleaveArrays(inputList);

        gov.usgs.gr.visual.gl.GlGraph tsGraph =
            new gov.usgs.gr.visual.gl.GlGraph();
        tsGraph.setProperty("showXtext", "true");
        tsGraph.add(new gov.usgs.gr.visual.LineMarkup());

        TimeSeries wlSeries = (TimeSeries)timeSeriesList.get(0);
        if (wlSeries.getMin(1) != 0 || wlSeries.getMax(1) != 0) {
            wlSeries.setTitle("Water Level");
            tsGraph.add(wlSeries);
        }

        gov.usgs.gr.visual.gl.Gl3dGraph profileGraph =
            new gov.usgs.gr.visual.gl.Gl3dGraph();
        profileGraph.setProperty("drawBoundingBox", "false");
        profileGraph.add(vp);
        gov.usgs.gr.visual.Axis xAxis = tsGraph.getXAxis();
        gov.usgs.gr.visual.Axis yAxis = tsGraph.getYAxis();
        gov.usgs.gr.visual.gl.PageOfGlGraphs page =
            new gov.usgs.gr.visual.gl.PageOfGlGraphs();
        page.add(tsGraph);
        page.add(profileGraph);
        gov.usgs.gr.tools.TimeMarker tm = new gov.usgs.gr.tools.TimeMarker();
        xAxis.setProperty("showLabels", "true");
        Zoom xZoom = xAxis.getZoom();
        Zoom yZoom = yAxis.getZoom();
        xZoom.setContentLimits(wlSeries.getMin(0), wlSeries.getMax(0));
        yZoom.setContentLimits(wlSeries.getMin(1), wlSeries.getMax(1));
        xZoom.setOuterLimits(wlSeries.getMin(0), wlSeries.getMax(0));
        yZoom.setOuterLimits(wlSeries.getMin(1), wlSeries.getMax(1));
        xZoom.setViewLimits(wlSeries.getMin(0), wlSeries.getMax(0));
        yZoom.setViewLimits(wlSeries.getMin(1), wlSeries.getMax(1));
        xZoom.autoScale(6);
        yZoom.autoScale(4);
        page.updateZoomContents();
        tm.add(ds);
        tm.add(xAxis.getZoom());
        tm.setProperty("sleep", "40");
        tm.setProperty("year", "" + wlSeries.getReferenceYear());
        TextMarkup textMarkup = new TextMarkup();
        textMarkup.setCoordinateLevels(2, 2);
        textMarkup.setProperty("x", "0.6");
        textMarkup.setProperty("y", "0.07");
        textMarkup.setProperty("screenFont", "glutBitmapTimesRoman24");
        tm.add(textMarkup);
        tsGraph.add(textMarkup);
        page.add(tm);
        page.showMoreGraphs();
        page.setProperty("formatter",
            "gov.usgs.sfhydro.gr.formats.VariableSizeFormatter");

        double showInc = ((double)m_numBins - 1) / 3;
        int[] binsToShow = new int[4];
        for (int i = 0; i < 4; ++i)
            binsToShow[i] = (int)((double)i * showInc);

        gov.usgs.gr.visual.gl.PageOfGlGraphs uvPage =
            new gov.usgs.gr.visual.gl.PageOfGlGraphs();
        Zoom xZoomB1 = null;
        Zoom yZoomB1 = null;

        gov.usgs.gr.visual.gl.PageOfGlGraphs dmPage =
            new gov.usgs.gr.visual.gl.PageOfGlGraphs();
        Zoom xZoomD1 = null;
        Zoom yZoomD1 = null;

        for (int i = 0; i < m_numBins; i++) {
            // U-V
            gov.usgs.gr.visual.gl.GlGraph uvGraph =
                new gov.usgs.gr.visual.gl.GlGraph();
            TimeSeries uBinSeries = new TimeSeries(uBinArrays[i], 2);
            uBinSeries.setReferenceYear(referenceYear);
            uBinSeries.setTitle("U Velocity, Bin " + (i + 1));
            uvGraph.add(uBinSeries);
            TimeSeries vBinSeries = new TimeSeries(vBinArrays[i], 2);
            vBinSeries.setReferenceYear(referenceYear);
            vBinSeries.setTitle("V Velocity, Bin " + (i + 1));
            uvGraph.add(vBinSeries);
            uvPage.add(uvGraph);
            gov.usgs.gr.visual.Axis xAxisB = uvGraph.getXAxis();
            gov.usgs.gr.visual.Axis yAxisB = uvGraph.getYAxis();
            xAxisB.setShowGridLines(false);
            if (xZoomB1 == null)
                xZoomB1 = xAxisB.getZoom();
            else
                xAxisB.add(xZoomB1);

            if (yZoomB1 == null)
                yZoomB1 = yAxisB.getZoom();
            else
                yAxisB.add(yZoomB1);

            uvGraph.setTitle("Bin " + (i + 1));
            uvGraph.setShowMainTitle(true);

            // Direction-Magnitude
            gov.usgs.gr.visual.gl.GlGraph dmGraph =
                new gov.usgs.gr.visual.gl.GlGraph();
            TimeSeries dBinSeries = new TimeSeries(dBinArrays[i], 2);
            dBinSeries.setReferenceYear(referenceYear);
            dBinSeries.setTitle("Direction (Degrees Clockwise) from North, Bin "
                + (i + 1));
            dmGraph.add(dBinSeries);
            TimeSeries mBinSeries = new TimeSeries(mBinArrays[i], 2);
            mBinSeries.setReferenceYear(referenceYear);
            mBinSeries.setTitle("Velocity Magnitude, Bin " + (i + 1));
            dmGraph.add(mBinSeries);
            dmPage.add(dmGraph);
            gov.usgs.gr.visual.Axis xAxisD = dmGraph.getXAxis();
            gov.usgs.gr.visual.Axis yAxisD = dmGraph.getYAxis();
            xAxisD.setShowGridLines(false);
            if (xZoomD1 == null)
                xZoomD1 = xAxisD.getZoom();
            else
                xAxisD.add(xZoomD1);

            if (yZoomD1 == null)
                yZoomD1 = yAxisD.getZoom();
            else
                yAxisD.add(yZoomD1);

            dmGraph.setTitle("Bin " + (i + 1));
            dmGraph.setShowMainTitle(true);
        }

        List list = uvPage.findChildInstancesOf(
            "gov.usgs.gr.visual.Curve", true);
        List l2 = dmPage.findChildInstancesOf("gov.usgs.gr.visual.Curve", true);
        list.addAll(l2);
        Iterator iter = list.iterator();
        while (iter.hasNext()) {
            Curve curve = (Curve)iter.next();
            curve.setLinePattern(Gr.SOLID_LINE_PATTERN);
        }
        uvPage.updateLists();
        uvPage.updateZoomContents();
        dmPage.updateLists();
        dmPage.updateZoomContents();
        list = uvPage.findChildInstancesOf("gov.usgs.gr.visual.Zoom", true);
        l2 = dmPage.findChildInstancesOf("gov.usgs.gr.visual.Zoom", true);
        list.addAll(l2);
        iter = list.iterator();
        while (iter.hasNext()) {
            Zoom zoom = (Zoom)iter.next();
            zoom.zoomAll(4);
        }

        // Hide certain graphs so that the first and last are shown, along
        // with two intermediate graphs.
        list = uvPage.findChildInstancesOf("gov.usgs.gr.visual.Graph", false);
        l2 = dmPage.findChildInstancesOf("gov.usgs.gr.visual.Graph", false);
        for (int i = 0; i < list.size(); i++) {
            Graph uvGraph = (Graph)list.get(i);
            Graph dmGraph = (Graph)l2.get(i);
            boolean bShow = false;
            for (int j = 0; j < 4; j++) {
                if (binsToShow[j] == i)
                bShow = true;
            }
            uvGraph.setShow(bShow);
            dmGraph.setShow(bShow);
        }

        if (m_bCreateDirMagPage)
            add(dmPage);
        if (m_bCreateUvPage)
            add(uvPage);
        if (m_bCreateProfilePage)
            add(page);

	}

//------------------------------------------------------------------------------
/**
 * Write the header information to an open file attached to a reader.
 */
    protected void writeHeader () throws IOException {

        List list = findChildInstancesOf(
            "gov.usgs.gr.data.DoubleDataSeries", true);
        if (list.size() < 2)
            return;

        DoubleDataSeries timeSeries = (DoubleDataSeries)list.get(0);
        DoubleDataSeries uvSeries = (DoubleDataSeries)list.get(1);

        double[] timeArray = timeSeries.getArray();
        double binHeight = 1;
        int minBin = 1;
        int maxBin = uvSeries.getCount() / timeSeries.getCount();
        int maxWriteBin = 9;
        maxBin = Math.min(maxBin, maxWriteBin);
        double timeStep = (timeArray[1] - timeArray[0]) * 24;

        NumberFormat form = new DecimalFormat("0.00000");

        String str = uvSeries.getTitle() + "\n\n\n";
        m_bw.write(str, 0, str.length());
        str = "" + minBin + " " + maxBin + " " + form.format(timeStep)
            + " " + binHeight + "\n";
        m_bw.write(str, 0, str.length());

        int startYear = 2000;
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();

        GregorianCalendar date = new GregorianCalendar(utcZone, locale);
        date = yearDayToGregCal(startYear, timeArray[0], date);
        SimpleDateFormat formatter =
            new SimpleDateFormat("yyyy MM dd");
        formatter.setTimeZone(utcZone);
        StringBuffer buf = formatter.format(date.getTime(),
            new StringBuffer(), new FieldPosition(0));
        str = buf.toString() + " ";

        double lastTime = timeArray[timeArray.length - 1];
        date = yearDayToGregCal(startYear, lastTime, date);
        buf = formatter.format(date.getTime(),
            new StringBuffer(), new FieldPosition(0));
        str += buf.toString();
        str += "\n";

        m_bw.write(str, 0, str.length());
    }

//------------------------------------------------------------------------------
/**
 * Write the data to a file.
 */
    protected void writeData () throws IOException {

        List list = findChildInstancesOf(
            "gov.usgs.gr.data.DoubleDataSeries", true);
        if (list.size() < 2)
            return;

        DoubleDataSeries timeSeries = (DoubleDataSeries)list.get(0);
        DoubleDataSeries uvSeries = (DoubleDataSeries)list.get(1);

        double[] timeArray = timeSeries.getArray();
        double[] uvArray = uvSeries.getArray();

        // Get the UTC time zone and default locale in advance of the loop.
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        Locale locale = Locale.getDefault();
        NumberFormat form1 = new DecimalFormat("0");
        NumberFormat form2 = new DecimalFormat("0.##");
        NumberFormat form3 = new DecimalFormat("0.00000");
        NumberFormat form4 = new DecimalFormat("00");

        double startTime = timeArray[0];
        int startYear = 2000;
        int referenceYear = 0;
        int numBins = uvSeries.getCount() / timeSeries.getCount();
        int maxWriteBin = Math.min(9, numBins);
        for (int i = 0; i < timeArray.length; i++) {
            // Add 1/200 hour so the hour will truncate correctly.
            double currentX = timeArray[i] - startTime;
            String str = AsciiIO.padLeft(form3.format(currentX), 10) + " 0 ";
            GregorianCalendar date = new GregorianCalendar(utcZone, locale);
            date = yearDayToGregCal(startYear,
                timeArray[i] + 0.0002083333333333333, date);
            SimpleDateFormat formatter =
                new SimpleDateFormat("yyyy MM dd HH");
            formatter.setTimeZone(utcZone);
            StringBuffer buf = formatter.format(date.getTime(),
                new StringBuffer(), new FieldPosition(0));
            str += buf.toString();
            float minute = date.get(GregorianCalendar.MINUTE);
            str += form4.format(minute * 100f / 60f);
            str += "\n";

            m_bw.write(str, 0, str.length());

            int uvNumDims = uvSeries.getNumDimensions();
            for (int j = 0; j < maxWriteBin; j++) {
                str = form1.format(j + 1);
                String binStr = AsciiIO.padLeft(form1.format(j + 1), 6);
                int binNum = maxWriteBin - j - 1;
                int index = (numBins * i + binNum) * uvNumDims;
                String val1Str =
                    AsciiIO.padLeft(form2.format(-uvArray[index + 1] * 100), 9);
                String val2Str =
                    AsciiIO.padLeft(form2.format(uvArray[index] * 100), 9);
                str = binStr + " " + val1Str + " " + val2Str + "\n";
                m_bw.write(str, 0, str.length());
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Shows a dialog for configuring the file.
 * @param frame The the frame to use as a parent for a modal dialog box.
 */
    public void showOptionsDialog(Frame frame) {

        final VPVConfigDialog dialog = new VPVConfigDialog(frame,
            "Input Config", true);

        // Show the configuration dialog.
        dialog.setConfig(m_principalDirection, m_bCreateProfilePage,
            m_bCreateUvPage, m_bCreateDirMagPage);

        if (EventQueue.isDispatchThread()) {
            dialog.setVisible(true);
        }
        else {
            try {
                javax.swing.SwingUtilities.invokeAndWait(new Runnable() {
                    public void run() {
                        dialog.setVisible(true);
                    }
                });
            }
            catch (InvocationTargetException ex) {
            }
            catch (InterruptedException ex) {
            }
        }

        // Don't make changes if Cancel was pressed.
        if (!(dialog.m_bOkPressed)) return;

        // Apply dialog settings to this object.
        m_principalDirection = dialog.getPrinDir();
        m_bCreateProfilePage = dialog.getCreateProfilePage();
        m_bCreateUvPage = dialog.getCreateUvPage();
        m_bCreateDirMagPage = dialog.getCreateDirMagPage();
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void mergeChildPages() {

        List vpPages = new ArrayList();
        List uvPages = new ArrayList();
        List dmPages = new ArrayList();

        List pageList =
            findChildInstancesOf("gov.usgs.gr.visual.PageOfGraphs", true);
        Iterator pageIter = pageList.iterator();
        while (pageIter.hasNext()) {
            PageOfGraphs page = (PageOfGraphs)pageIter.next();
            GlVelocityProfile vp = (GlVelocityProfile)page.firstChildInstanceOf(
                "gov.usgs.gr.visual.gl.GlVelocityProfile", true);
            if (vp != null) {
                vpPages.add(page);
                continue;
            }

            TimeSeries series = (TimeSeries)page.firstChildInstanceOf(
                "gov.usgs.gr.data.TimeSeries", true);
            if (series != null) {
                if (series.getTitle().indexOf("U Velocity") >= 0) {
                    uvPages.add(page);
                    continue;
                }
                if (series.getTitle().indexOf("Direction") >= 0) {
                    dmPages.add(page);
                    continue;
                }
            }
        }

        // Put the page lists in a list so they can be iterated.
        List pageListList = new ArrayList();
        pageListList.add(uvPages);
        pageListList.add(dmPages);
        Iterator iter = pageListList.iterator();

        // Look at each group of similar pages.
        while (iter.hasNext()) {
            pageList = (List) iter.next();
            if (pageList.size() > 0) {
                // Determine the main page.
                PageOfGraphs mainPage = (PageOfGraphs) pageList.get(0);

                // Copy the curves from each respective graph on the other
                // page to the main page.
                List mainGraphs = mainPage.getGraphList();
                for (int i = 1; i < pageList.size(); i++) {
                    PageOfGraphs otherPage = (PageOfGraphs) pageList.get(i);
                    List otherGraphs = otherPage.getGraphList();
                    for (int j = 0; j < mainGraphs.size(); j++) {
                        Graph mainGraph = (Graph) mainGraphs.get(j);
                        Graph otherGraph = (Graph) otherGraphs.get(j);
                        List otherCurves = otherGraph.getCurves();

                        // Set the properties for the first curve.
                        Curve curve1 = (Curve) otherCurves.get(0);
                        curve1 = (Curve)curve1.createCopy(false);
                        curve1.setLinePattern(Gr.DASHED_LINE_PATTERN);
                        curve1.setColor(1, 1, 0);
                        mainGraph.add(curve1);

                        // Set the properties for the second curve.
                        Curve curve2 = (Curve) otherCurves.get(1);
                        curve2 = (Curve)curve2.createCopy(false);
                        curve2.setLinePattern(Gr.DASHED_LINE_PATTERN);
                        curve2.setColor(0, 1, 1);
                        mainGraph.add(curve2);
                    }
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * @return  A menu item that can be placed on a menu and send an ActionEvent
 *          back to this object when it is chosen.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        // Create the menu item as a sub menu with other items.
        javax.swing.JMenuItem subMenu = super.getMenuCommands();
        subMenu.setText("VPV Data File");

        // Set up menu items that use this object as their action listeners.
        javax.swing.JMenuItem mergeItem = new javax.swing.JMenuItem(
            "Merge Child Pages");
        mergeItem.addActionListener(this);
        subMenu.add(mergeItem);

    	return subMenu;
    }

//------------------------------------------------------------------------------
/**
 * Handle action events involving this object.
 * @param e The ActionEvent that needs to be handled.
 */
    public void actionPerformed(java.awt.event.ActionEvent e) {

        // Get the text from the menu item that was chosen.
        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();

    	// Show an hourglass until the job is complete.
        java.awt.Frame frame = GrController.getParentFrame(source);
        frame.setCursor(new Cursor(Cursor.WAIT_CURSOR));

        // Check if the command was to resample.
        if (srcText.equals("Merge Child Pages")) {
            mergeChildPages();
        }

        // Return to the default arrow cursor.
        frame.setCursor(Cursor.getDefaultCursor());
    }

//------------------------------------------------------------------------------
/**
 * A dialog that allows the user to configure input options.
 */
    public class VPVConfigDialog extends JDialog {

        /** Whether the user pressed the OK button as opposed to Cancel. */
        public boolean m_bOkPressed = false;
        /**
         * A text field for displaying and editing the principal direction.
         */
        private JTextField prinDirTextField = new JTextField();

        private JCheckBox createProfileCB = new JCheckBox();
        private JCheckBox createUvCB = new JCheckBox();
        private JCheckBox createDirMagCB = new JCheckBox();

        //----------------------------------------------------------------------
        /**
         * Creates the dialog and lays out the components.
         * @param frame The frame to which the dialog should be attached.
         * @param title The title of the dialog.
         * @param modal Whether the dialog is modal (must be closed before
         *              continuing).
         */
        public VPVConfigDialog(Frame frame, String title, boolean modal) {

            super(frame, title, modal);

            try  {
                jbInit();
                pack();
            }
            catch (Exception ex) {
                ex.printStackTrace();
            }
        }

        //----------------------------------------------------------------------
        /**
         * Creates the dialog and lays out the components.
         * Calls constructor with parameters (null, "", false)
         */
        public VPVConfigDialog() {

            this(null, "", false);
        }

        //----------------------------------------------------------------------
        /**
         * Lay out components in the dialog.
         * @throws Exception
         */
        void jbInit() throws Exception {

            // Initialize components.
            JPanel mainPanel = new JPanel();
            JPanel topPanel = new JPanel();
            JPanel lowerPanel = new JPanel();
            JButton okButton = new JButton("OK");
            JButton cancelButton = new JButton("Cancel");
            JPanel prinDirPanel = new JPanel();
            JPanel createProfPanel = new JPanel();
            JPanel createUvPanel = new JPanel();
            JPanel createDirMagPanel = new JPanel();

            // Set up button listeners.
            okButton.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    okButton_actionPerformed(e);
                }
            });
            cancelButton.addActionListener(new java.awt.event.ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    cancelButton_actionPerformed(e);
                }
            });

            // Set the layout managers for the panels and compose the items
            // within the containers.
            mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
            mainPanel.add(Box.createRigidArea(new Dimension(350,5)));

            topPanel.setLayout(new BoxLayout(topPanel, BoxLayout.Y_AXIS));

            // "Principal Direction" panel
            prinDirPanel.setLayout(new BoxLayout(prinDirPanel,
                BoxLayout.X_AXIS));
            prinDirPanel.add(new JLabel("Principal Direction"));
            prinDirPanel.add(Box.createRigidArea(new Dimension(5,0)));

            prinDirPanel.add(prinDirTextField);
            prinDirPanel.add(Box.createRigidArea(new Dimension(0,5)));

            mainPanel.add(prinDirPanel);
            mainPanel.add(Box.createRigidArea(new Dimension(0,5)));

            // "Create Profile" panel
            createProfPanel.setLayout(new BoxLayout(createProfPanel,
                BoxLayout.X_AXIS));
            createProfPanel.add(new JLabel("Create 3D Profile Page"));
            createProfPanel.add(Box.createRigidArea(new Dimension(5,0)));

            createProfPanel.add(createProfileCB);
            createProfPanel.add(Box.createRigidArea(new Dimension(0,5)));

            mainPanel.add(createProfPanel);
            mainPanel.add(Box.createRigidArea(new Dimension(0,5)));

            // "Create U-V" panel
            createUvPanel.setLayout(new BoxLayout(createUvPanel,
                BoxLayout.X_AXIS));
            createUvPanel.add(new JLabel("Create U-V Page"));
            createUvPanel.add(Box.createRigidArea(new Dimension(5,0)));

            createUvPanel.add(createUvCB);
            createUvPanel.add(Box.createRigidArea(new Dimension(0,5)));

            mainPanel.add(createUvPanel);
            mainPanel.add(Box.createRigidArea(new Dimension(0,5)));

            // "Create Dir-Mag" panel
            createDirMagPanel.setLayout(new BoxLayout(createDirMagPanel,
                BoxLayout.X_AXIS));
            createDirMagPanel.add(new JLabel("Create Direction-Magnitude Page"));
            createDirMagPanel.add(Box.createRigidArea(new Dimension(5,0)));

            createDirMagPanel.add(createDirMagCB);
            createDirMagPanel.add(Box.createRigidArea(new Dimension(0,5)));

            mainPanel.add(createDirMagPanel);
            mainPanel.add(Box.createRigidArea(new Dimension(0,5)));

            mainPanel.add(Box.createRigidArea(new Dimension(0,5)));
            lowerPanel.setLayout(new BoxLayout(lowerPanel, BoxLayout.X_AXIS));

            lowerPanel.add(okButton);
            lowerPanel.add(Box.createRigidArea(new Dimension(5,0)));
            lowerPanel.add(cancelButton);

            mainPanel.add(lowerPanel);

            getContentPane().add(mainPanel);
        }

        //----------------------------------------------------------------------
        /**
         * Configure the dialog.
         */
        void setConfig(double prinDir, boolean bCreateProfilePage,
            boolean bCreateUvPage, boolean bCreateDirMagPage) {

            prinDirTextField.setText("" + prinDir);
            createProfileCB.setSelected(bCreateProfilePage);
            createProfileCB.setSelected(bCreateProfilePage);
            createUvCB.setSelected(bCreateUvPage);
            createDirMagCB.setSelected(bCreateDirMagPage);
        }

        //----------------------------------------------------------------------
        /**
         * Accessor method.
         * @return
         */
        double getPrinDir() {

            return Double.parseDouble(prinDirTextField.getText());
        }

        //----------------------------------------------------------------------
        /**
         * Accessor method.
         * @return
         */
        boolean getCreateProfilePage() {

            return createProfileCB.isSelected();
        }

        //----------------------------------------------------------------------
        /**
         * Accessor method.
         * @return
         */
        boolean getCreateUvPage() {

            return createUvCB.isSelected();
        }

        //----------------------------------------------------------------------
        /**
         * Accessor method.
         * @return
         */
        boolean getCreateDirMagPage() {

            return createDirMagCB.isSelected();
        }

        //----------------------------------------------------------------------
        /**
         * Handles the event of the OK button being pressed.  Closes the dialog.
         * @param e The event that occurred.
         */
        void okButton_actionPerformed(ActionEvent e) {

            m_bOkPressed = true;

            dispose();
        }

        //----------------------------------------------------------------------
        /**
         * Handles the event of the Cancel button being pressed.
         * Closes the dialog.
         * @param e The event that occurred.
         */
        void cancelButton_actionPerformed(ActionEvent e) {

            m_bOkPressed = false;
            dispose();
        }
    }
}
