/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-03
 * Date started:    2001-02-19
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.gr.formats;

import java.awt.Rectangle;
import java.awt.geom.*;
import java.text.*;
import java.util.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.sfhydro.data.*;
import gov.usgs.gr.visual.*;
import gov.usgs.gr.visual.ps.*;

//==============================================================================
/**
 * Applies formatting to a PageOfGraphs.
 */
public class VariableSizeFormatter extends gov.usgs.gr.visual.PageFormatter {

    /** A list of sizes for graphs. */
    List<Float> m_sizeValues = new ArrayList<Float>();

    /** The minimum number of values passed out in getProperties. */
    protected int m_numVals = 1;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/VariableSizeFormatter.gif";

//------------------------------------------------------------------------------
/**
 * Create an empty formatter.
 */
    public VariableSizeFormatter () {

        setDefaults("gov.usgs.sfhydro.gr.formats.VariableSizeFormatter");

        m_sizeValues.add(20f);
    }

//------------------------------------------------------------------------------
/**
 * Create a page formatter and give it a page to format.
 */
    public VariableSizeFormatter (PageOfGraphs page) {

        setDefaults("gov.usgs.sfhydro.gr.formats.VariableSizeFormatter");

    	m_page = page;
        m_sizeValues.add(20f);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Variable Size Formatter";

        return s;
 	}

//------------------------------------------------------------------------------
/**
 * Makes a copy of this formatter.
 * @return	A copy of this formatter.
 */
    public PageFormatter getCopy() {

    	return new VariableSizeFormatter();
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("numVals", new GrProperty(m_numVals, "Number of Values"));

        int valNum = 0;

        for (Float size : m_sizeValues) {

        	GrProperty p = new GrProperty(size, "Pct. Size of Graph " + valNum);

			props.put("size" + valNum, p);

            valNum++;
        }

        while (valNum < m_numVals) {
            // Add additional entries if needed...
        	GrProperty p = new GrProperty(0f, "Pct. Size of Graph " + valNum);

            p.save = false;
			props.put("size" + valNum, p);

            valNum++;
        }


        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("numVals")) {
            		m_numVals = (Integer)p.value;
                    if (m_numVals < 1 || m_numVals > 50)
                    	m_numVals = 1;
                    while (m_sizeValues.size() > m_numVals)
                        m_sizeValues.remove(m_sizeValues.size() - 1);
                }
            	if (key.startsWith("size")) {
                    setSizeAtValue(Integer.parseInt(key.substring(4)),
                        (Float)p.value);
                }
			}
            if (obj instanceof String) {
            	String value = (String)obj;
            	setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("numVals")) {
        	m_numVals = (int)(Float.parseFloat(value));
            if (m_numVals < 1 || m_numVals > 50)
            	m_numVals = 1;
            while (m_sizeValues.size() > m_numVals)
                m_sizeValues.remove(m_sizeValues.size() - 1);
        }
        if (key.startsWith("size")) {
            setSizeAtValue(Integer.parseInt(key.substring(4)),
                Float.parseFloat(value));
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void setSizeAtValue(int index, Float size) {

        boolean bFound = false;

        for (int i = 0; i < m_sizeValues.size(); i++) {
            if (i == index) {
            	m_sizeValues.remove(i);
            	m_sizeValues.add(i, size);
                bFound = true;
                break;
            }
        }

        if (!bFound) {
            // Insert blank values
            while (m_sizeValues.size() < index)
                m_sizeValues.add(0f);

        	m_sizeValues.add(index, size);
        }
    }

//------------------------------------------------------------------------------
/**
 * Applies the format to the page as well as making some one time changes.
 */
    public void initialize() {

    	reformat();
    }

//------------------------------------------------------------------------------
/**
 * Applies the format to the page.
 */
    public void reformat() {

        layoutGraphs();

        m_page.m_bNeedsReformat = false;
    }

//------------------------------------------------------------------------------
/**
 * Layout all showing graphs and reproportion each graph within the panel.
 * Triggers a redraw.
 */
    protected void layoutScreenGraphs() {

        synchronized(m_page) {
            int xPos = 0, yPos = 0;
            List<Graph> showList = m_page.getShowList();
            float contentHeight = (float)m_page.getContentRect().height;

            // Find the sum of the frame sizes and use that to find the average.
            for (Graph graph : showList)
                graph.layoutFrame();

            // Reproportion each graph within the panel.
            Iterator<Float> sizeIter = m_sizeValues.iterator();
            float remainingSize = 100;
            int i = 0;
            for (Graph graph : showList) {
                float size = remainingSize;
                if (sizeIter.hasNext()) {
                    size = sizeIter.next();
                    remainingSize -= size;
                    if (remainingSize < 0) remainingSize = 0;
                }

                // Calculate how much to scale the graph so its frame will be
                // the same size as all the others.
                Rectangle2D.Double graphFrame = graph.getFrameRect();
                Rectangle graphLocal = graph.getLocalRect();

                // Set the new size and position for the graph.
                int newWidth = m_page.getContentRect().width;
                int newHeight = 0;
                if (size < remainingSize && i < (showList.size() - 1))
                    newHeight = (int)((size / 100) * contentHeight);
                else
                    newHeight = (int)(contentHeight - yPos);

                // Only change the size and position if the graph is not
                // being dragged to a new location.
                if (graph != m_page.getMovingGraph()) {
                    graph.setSize(newWidth, newHeight);
                    graph.setXPosition(xPos);
                    graph.setYPosition(yPos);
                }
                yPos += newHeight;

                // Set the tick scale so both axes have the same size ticks.
                float xTickScale = 1;
                if (newHeight != 0) xTickScale = m_page.getHeight()
                    / (float)newHeight;
                graph.setXTickScale(xTickScale);
                float yTickScale = 1;
                if (newWidth != 0) yTickScale = m_page.getHeight()
                    / (float)newWidth;
                graph.setYTickScale(yTickScale);

                // Enlarge the frame area if there are no axis labels.
                if (!(graph.getYAxis().getShowTitle()) &&
                    !(graph.getYAxis().getShowText())) {

                    if (showList.size() == 1) {
                        graphFrame.x = 0;
                        graphFrame.width = 1;
                    }
                    else {
                        graphFrame.x = 0.01;
                        graphFrame.width = 1.0 - graphFrame.x - 0.005;
                    }
                }

                if (!(graph.getShowMainTitle())) {
                    if (showList.size() == 1) {
                        graphFrame.y = 0;
                        graphFrame.height = 1;
                    }
                    else {
                        graphFrame.height = 1.0 - graphFrame.y - 0.005;
                    }
                }

                graph.setFrameRect(graphFrame);

                i++;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Layout all showing graphs and reproportion each graph within the panel.
 * Triggers a redraw.
 */
    protected void layoutPsGraphs() {

        layoutScreenGraphs();
    }

//------------------------------------------------------------------------------
/**
 * Layout all showing pages within the 3D page and reproportions each one.
 * Triggers a redraw.
 */
    protected void layout3DPage() {

        synchronized(m_page) {
            int xPos = 0, yPos = 0;
            List pageList = m_page.findChildInstancesOf(
                    "gov.usgs.gr.visual.PageOfGraphs", false);
            pageList.add(m_page);
            float displayHeight = (float)m_page.getHeight();

            // Reproportion each graph within the panel.
            Iterator iter = pageList.iterator();
            Iterator<Float> sizeIter = m_sizeValues.iterator();
            float remainingSize = 100;
            int i = 0;
            while (iter.hasNext()) {
                PageOfGraphs page = (PageOfGraphs)iter.next();

                float size = remainingSize;
                if (sizeIter.hasNext()) {
                    size = sizeIter.next();
                    remainingSize -= size;
                    if (remainingSize < 0) remainingSize = 0;
                }

                // Set the new size and position for the graph.
                int newWidth = m_page.getWidth();
                int newHeight = 0;
                if (iter.hasNext())
                    newHeight = (int)((size / 100) * displayHeight);
                else
                    newHeight = (int)(displayHeight - yPos);

                // Change the size and position.
                Rectangle rect = new Rectangle(xPos, yPos, newWidth, newHeight);
                page.setContentRect(rect);
                if (page != m_page)
                    page.getFormatter().layoutGraphs();
                yPos += newHeight;

                i++;
            }
        }
    }
}
