/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    2000-09-07
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.gr.tools;

import java.util.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.TextMarkup;
import gov.usgs.sfhydro.si3d.*;

//==============================================================================
/**
 * Calculates the cross sectional path where a vertical surface intersects
 * a surfaces of discontinuous squares.  When a surface data set and a cross
 * secion line XyDataSet are added as children, an output X-Z data set is
 * calculated and added as a third child.  The output series can be viewed
 * if it is copied and the original pasted into a graph.
 */

public class CrossSectionCalculator extends GrObject implements
    java.awt.event.ActionListener {
    
    protected RawBathymetry m_rawBath = null;
    protected DoubleDataSeries m_surfaceDs = null;
    protected DoubleDataSeries m_inputCsDs = null;
    protected DoubleDataSeries m_outputCsDs = new DoubleDataSeries(2);
    protected GrObjectContainer m_outputContainer = new GrObjectContainer();
    protected int m_exitWall = 0;
    protected TextMarkup m_textMarkup = null;
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/CrossSectionCalculator.gif";
    
//------------------------------------------------------------------------------
    /**
     * Creates an empty object.
     */
    public CrossSectionCalculator() {
        
        m_outputContainer.setName("Output");
        m_outputContainer.setProperty("title", "Output");
        m_outputContainer.add(m_outputCsDs);
        add(m_outputContainer);
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides add in GrObject.
     */
    public boolean add(GrObject gro) {
        
        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();
        
        if (m_children.contains(gro)) return false;
        
        boolean bAdded = false;
        
        if (gro instanceof DoubleDataSeries) {
            DoubleDataSeries ds = (DoubleDataSeries)gro;
            
            if (ds.getNumDimensions() == 2 || ds.getNumDimensions() == 3) {
                // Remove all old series with the same number of dims.
                Iterator iter = m_children.iterator();
                while (iter.hasNext()) {
                    GrObject child = (GrObject)iter.next();
                    if (child instanceof DoubleDataSeries) {
                        DoubleDataSeries oldDs = (DoubleDataSeries) child;
                        if (oldDs.getNumDimensions() == ds.getNumDimensions()
                        && oldDs != m_outputCsDs) {
                            
                            iter.remove();
                        }
                    }
                }
                
                m_children.add(ds);
                ds.addObserver(this);
                bAdded = true;
            }
            
            // If it is a surface series...
            if (ds.getNumDimensions() == 3)
                m_surfaceDs = ds;
            
            // If it is the input cross section...
            if (ds.getNumDimensions() == 2)
                m_inputCsDs = ds;
        }
        if (gro instanceof RawBathymetry) {
            if (m_rawBath != null)
                remove(m_rawBath);
            
            super.add(gro);
            
            m_rawBath = (RawBathymetry)gro;
            
            return true;
        }
        if (gro instanceof TextMarkup) {
            // Add the text markup so it can be updated with a text form of
            // the date and time value of this object.
            super.add(gro);
            
            // Allow the user to see and change the markup's location.
            ((TextMarkup)gro).setUserChangeable(true);
            
            m_textMarkup = (TextMarkup)gro;
            
            return true;
        }
        if (gro instanceof GrObjectContainer) {
            GrObjectContainer groc = (GrObjectContainer)gro;
            GrHashList props = groc.getProperties();
            GrProperty p = (GrProperty)props.get("title");
            if (p == null)
                return false;
            String title = (String)p.value;
            
            // Compare the title of groc to each container and replace the
            // current one if there is a match.
            if (title.equals("Output")) {
                GrObjectContainer oldGroc = m_outputContainer;
                
                int index = m_children.size();
                if (oldGroc != null) {
                    index = m_children.indexOf(oldGroc);
                    oldGroc.deleteObserver(this);
                    m_children.remove(oldGroc);
                    if (index < 0)
                        index = 0;
                }
                m_children.add(index, groc);
                groc.addObserver(this);
                groc.setName("");
                
                m_outputContainer = groc;
                
                // Look for DoubleDataSeries within the output container
                // and use them for output if they are there.
                Enumeration enumeration = m_outputContainer.children();
                while (enumeration.hasMoreElements()) {
                    GrObject child = (GrObject)enumeration.nextElement();
                    if (DoubleDataSeries.countDims(child, 2)) {
                        m_outputCsDs = (DoubleDataSeries)child;
                    }
                }
                
                // Make sure there are no other output series
                // in the output container besided those defined.
                m_outputContainer.clear();
                m_outputContainer.add(m_outputCsDs);
                
                m_outputCsDs.getDimension(0).
                    setProperty("locked", "true");
                m_outputCsDs.getDimension(1).
                    setProperty("locked", "true");
                m_outputCsDs.setTitle(
                    "Cross Section Output");
                
                m_bHasChanged = true;
                return true;
            }
        }
        
        recalculate();
        
        return bAdded;
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides clear in GrObject to do nothing.
     */
    public void clear() {
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides remove in GrObject to do nothing.
     */
    public GrObject remove(int i) {
        
        return null;
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides remove in GrObject to do nothing.
     */
    public boolean remove(GrObject gro) {
        
        return false;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
        
        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();
        
        return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {
        
        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);
        
        if (icon == null)
            icon = super.setTreeIcon();
        
        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);
        
        return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public void recalculate() {
        
        if (m_inputCsDs == null) return;
        if (m_surfaceDs == null) return;
        
        if (m_outputCsDs == null) {
            m_outputCsDs = new DoubleDataSeries(2);
            m_outputCsDs.setTitle("Output Cross Section");
            m_children.add(m_outputCsDs);
        }
        
        synchronized (m_inputCsDs) {
            synchronized (m_surfaceDs) {
                synchronized (m_outputCsDs) {
                    ArrayList outputPoints = getOutputPoints();
                    
                    m_outputCsDs.setArray(outputPoints);
                    setAreaUnderCurve(m_outputCsDs.getArray());
                }
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public ArrayList getOutputPoints() {
        
        ArrayList outputPoints = new ArrayList();
        
        // First X point will always be 0;
        outputPoints.add(0d);
        
        double[] inputArray = m_inputCsDs.getArray();
        for (int i = 0; i < inputArray.length - 2; i += 2) {
            if (i > 0)
                outputPoints.remove(outputPoints.size() - 1);
            double[] startPoint = new double[2];
            double[] endPoint = new double[2];
            startPoint[0] = inputArray[i];
            startPoint[1] = inputArray[i + 1];
            endPoint[0] = inputArray[i + 2];
            endPoint[1] = inputArray[i + 3];
            
            travelLineSegment(outputPoints, startPoint, endPoint);
        }
        
        return outputPoints;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public void travelLineSegment(ArrayList<Double> outputPoints,
        double[] startPoint, double[] endPoint) {
        
        double[] currentPoint = new double[2];
        currentPoint[0] = startPoint[0];
        currentPoint[1] = startPoint[1];
        double[] currentCell = getCell(currentPoint);
        double[] endCell = getCell(endPoint);
        double newOutputX = 0;
        
        boolean bThisIsEndCell = false;
        while (!bThisIsEndCell) {
            // Determine if this is the cell containing the end point.
            if (currentPoint[0] > endCell[3] && currentPoint[0] < endCell[6] &&
                currentPoint[1] > endCell[4] && currentPoint[1] < endCell[10]) {
                
                bThisIsEndCell = true;
            }
            if (currentPoint[0] == endPoint[0] &&
                currentPoint[1] == endPoint[1]) {
                
                bThisIsEndCell = true;
            }
            if (Double.isNaN(currentPoint[0]) ||
                Double.isNaN(currentPoint[1])) {
                
                bThisIsEndCell = true;
            }
            
            // Output the Z value at the current point.
            double z = getZatPointInCell(currentPoint, currentCell);
            outputPoints.add(-z);
            
            // Find the exit point.
            double[] exitPoint;
            if (bThisIsEndCell) {
                exitPoint = endPoint;
            } else {
                exitPoint = determineExitPoint(currentPoint, endPoint,
                    currentCell);
            }
            
            // Calculate the distance traveled along the segment between the
            // current point and the exit point.
            double xDiff = exitPoint[0] - currentPoint[0];
            double yDiff = exitPoint[1] - currentPoint[1];
            double segDist = Math.sqrt(xDiff * xDiff + yDiff * yDiff);
            
            
            // Output the X (distance travelled) value at the exit point.
            newOutputX = outputPoints.get(outputPoints.size() - 2);
            newOutputX += segDist;
            outputPoints.add(newOutputX);
            
            // Output the Z value at the exit point.
            z = getZatPointInCell(exitPoint, currentCell);
            outputPoints.add(-z);
            // Add the same X again because of the step function.
            outputPoints.add(newOutputX);
            
            // Determine the next cell.
            if (!bThisIsEndCell) {
                double[] nextCenter = new double[2];
                
                switch (m_exitWall) {
                    case 0:
                        nextCenter[0] = currentCell[0];
                        nextCenter[1] = 2 * currentCell[4] - currentCell[1];
                        break;
                    case 1:
                        nextCenter[1] = currentCell[1];
                        nextCenter[0] = 2 * currentCell[6] - currentCell[0];
                        break;
                    case 2:
                        nextCenter[0] = currentCell[0];
                        nextCenter[1] = 2 * currentCell[10] - currentCell[1];
                        break;
                    case 3:
                        nextCenter[1] = currentCell[1];
                        nextCenter[0] = 2 * currentCell[3] - currentCell[0];
                        break;
                    default:
                        nextCenter[0] = endPoint[0];
                        nextCenter[1] = endPoint[1];
                }
                
                currentCell = getCell(nextCenter);
            } else {
                // This is the last cell, so just write out the end point Z.
                z = getZatPointInCell(endPoint, currentCell);
                outputPoints.add(-z);
            }
            
            // Advance to the next point.  The exit point of this cell
            // becomes the entry point of the next.
            currentPoint = exitPoint;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public double[] getCell(double[] currentPoint) {
        
        if (m_rawBath != null)
            return getCellFromRawBath(currentPoint);
        
        double[] cell = null;
        
        // Run through the array looking for the cell containing the point.
        double[] surfaceArray = m_surfaceDs.getArray();
        int numPoints = m_surfaceDs.getNumPoints() / 5;
        for (int i = 0; i < numPoints; i++) {
            int offset = i * 15;
            if (currentPoint[1] > surfaceArray[offset + 4]) {
                if (currentPoint[1] < surfaceArray[offset + 10]) {
                    if (currentPoint[0] > surfaceArray[offset + 3]) {
                        if (currentPoint[0] < surfaceArray[offset + 6]) {
                            // Copy these values for the cell.
                            cell = new double[15];
                            for (int j = 0; j < 15; j++) {
                                cell[j] = surfaceArray[offset + j];
                            }
                            return cell;
                        }
                    }
                }
            } else {
                // Passed it up, so fall through and return a blank cell.
                break;
            }
        }
        
        // The cell was absent from the surface array.
        // Make up a fictitios cell for the absent one that resides where
        // the absent one should have been.
        // Set Z to 0.
        double xInc = surfaceArray[6] - surfaceArray[3];
        double yInc = surfaceArray[10] - surfaceArray[4];
        double xOffset = surfaceArray[3] % xInc;
        double yOffset = surfaceArray[4] % yInc;
        double xDiff = Math.abs((currentPoint[0] - xOffset) % xInc);
        double yDiff = Math.abs((currentPoint[1] - yOffset) % yInc);
        double blX = currentPoint[0] - xDiff;
        double blY = currentPoint[1] - yDiff;
        
        cell = new double[15];
        cell[0] = blX + 0.5 * xInc;
        cell[1] = blY + 0.5 * yInc;
        cell[2] = 0;
        cell[3] = blX;
        cell[4] = blY;
        cell[5] = 0;
        cell[6] = blX + xInc;
        cell[7] = blY;
        cell[8] = 0;
        cell[9] = blX + xInc;
        cell[10] = blY + yInc;
        cell[11] = 0;
        cell[12] = blX;
        cell[13] = blY + yInc;
        cell[14] = 0;
        
        return cell;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public double[] getCellFromRawBath(double[] currentPoint) {
        
        double[] cell = null;
        
        // Run through the array looking for the cell containing the point.
        double[] depthGrid = m_rawBath.getBathymetrySeries().getArray();
        GrHashList bathProps = m_rawBath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        double gridWidth = (Double)p.value;
        p = (GrProperty)bathProps.get("numI");
        int numI = (Integer)p.value;
        p = (GrProperty)bathProps.get("numJ");
        double numJ = (Integer)p.value;
        
        int currentI = (int)(currentPoint[0] / gridWidth);
        int currentJ = (int)(currentPoint[1] / gridWidth);
        
        // The cell was absent from the surface array.
        // Make up a fictitios cell for the absent one that resides where
        // the absent one should have been.
        // Set Z to 0.
        double xOffset = 0;//surfaceArray[3] % xInc;
        double yOffset = 0;//surfaceArray[4] % yInc;
        double xDiff = (currentPoint[0] - xOffset) % gridWidth;
        double yDiff = (currentPoint[1] - yOffset) % gridWidth;
        double blX = currentPoint[0] - xDiff;
        double blY = currentPoint[1] - yDiff;
        
        cell = new double[15];
        cell[0] = blX + 0.5 * gridWidth;
        cell[1] = blY + 0.5 * gridWidth;
        cell[2] = 0;
        cell[3] = blX;
        cell[4] = blY;
        cell[5] = 0;
        cell[6] = blX + gridWidth;
        cell[7] = blY;
        cell[8] = 0;
        cell[9] = blX + gridWidth;
        cell[10] = blY + gridWidth;
        cell[11] = 0;
        cell[12] = blX;
        cell[13] = blY + gridWidth;
        cell[14] = 0;
        
        if (currentI >= 0 && currentJ >= 0 &&
            currentI < numI && currentJ < numJ) {
            
            cell[5] = depthGrid[currentJ * numI + currentI];
            if ((currentI + 1) < numI) {
                cell[8] = depthGrid[currentJ * numI + currentI + 1];
                if ((currentJ + 1) < numJ)
                    cell[11] = depthGrid[ (currentJ + 1) * numI + currentI + 1];
            }
            if ((currentJ + 1) < numJ)
                cell[14] = depthGrid[ (currentJ + 1) * numI + currentI];
            
            // Clamp land points to the lowest land value.
            double lowestLand = 0;
            double landSubstitute = 0;
            if (cell[5] < lowestLand)
                cell[5] = landSubstitute;
            if (cell[8] < lowestLand)
                cell[8] = landSubstitute;
            if (cell[11] < lowestLand)
                cell[11] = landSubstitute;
            if (cell[14] < lowestLand)
                cell[14] = landSubstitute;
            
            cell[2] = (cell[5] + cell[8] + cell[11] + cell[14]) / 4;
        }
        
        return cell;
    }
    
//------------------------------------------------------------------------------
    /**
     * Find the Z value at a certain point within the cell.
     */
    public double getZatPointInCell(double[] currentPoint, double[] currentCell) {
        
        double x1Diff = Math.abs(currentPoint[0] - currentCell[3]);
        double x2Diff = Math.abs(currentPoint[0] - currentCell[6]);
        double x3Diff = Math.abs(currentPoint[0] - currentCell[9]);
        double x4Diff = Math.abs(currentPoint[0] - currentCell[12]);
        double y1Diff = Math.abs(currentPoint[1] - currentCell[4]);
        double y2Diff = Math.abs(currentPoint[1] - currentCell[7]);
        double y3Diff = Math.abs(currentPoint[1] - currentCell[10]);
        double y4Diff = Math.abs(currentPoint[1] - currentCell[13]);
        double a1 = x1Diff * y1Diff;
        double a2 = x2Diff * y2Diff;
        double a3 = x3Diff * y3Diff;
        double a4 = x4Diff * y4Diff;
        double aTotal = a1 + a2 + a3 + a4;
        double z1 = currentCell[5];
        double z2 = currentCell[8];
        double z3 = currentCell[11];
        double z4 = currentCell[14];
        double z = (a3 * z1 + a4 * z2 + a1 * z3 + a2 * z4) / aTotal;
        
        return z;
    }
    
//------------------------------------------------------------------------------
    /**
     * Determine coordinates of the intersection of the line between the current
     * point and the end point with the wall of the cell.
     */
    public double[] determineExitPoint(double[] currentPoint, double[] endPoint,
        double[] currentCell) {
        
        double[] exitPoint = new double[2];
        
        if (endPoint[0] > currentCell[3] && endPoint[0] < currentCell[6] &&
            endPoint[1] > currentCell[4] && endPoint[1] < currentCell[10]) {
            
            // If the destination cell has been reached, just return the
            // end point.
            exitPoint[0] = endPoint[0];
            exitPoint[1] = endPoint[1];
            m_exitWall = -1;
            return exitPoint;
        }
        
        double xEndDiff = endPoint[0] - currentPoint[0];
        double yEndDiff = endPoint[1] - currentPoint[1];
        boolean bRightWall = false;
        boolean bTopWall = false;
        if (xEndDiff > 0) bRightWall = true;
        if (yEndDiff > 0) bTopWall = true;
        
        if (bRightWall) {
            double xWallDiff = currentCell[6] - currentPoint[0];
            double yWallDiff = yEndDiff * (xWallDiff / xEndDiff);
            double yPos = currentPoint[1] + yWallDiff;
            if (yPos >= currentCell[7] && yPos <= currentCell[10]) {
                exitPoint[0] = currentCell[6];
                exitPoint[1] = yPos;
                m_exitWall = 1;
                return exitPoint;
            }
        }
        
        if (!bRightWall) {
            double xWallDiff = currentCell[3] - currentPoint[0];
            double yWallDiff = yEndDiff * (xWallDiff / xEndDiff);
            double yPos = currentPoint[1] + yWallDiff;
            if (yPos >= currentCell[7] && yPos <= currentCell[10]) {
                exitPoint[0] = currentCell[3];
                exitPoint[1] = yPos;
                m_exitWall = 3;
                return exitPoint;
            }
        }
        
        if (bTopWall) {
            double yWallDiff = currentCell[10] - currentPoint[1];
            double xWallDiff = xEndDiff * (yWallDiff / yEndDiff);
            double xPos = currentPoint[0] + xWallDiff;
            
            exitPoint[0] = xPos;
            exitPoint[1] = currentCell[10];
            m_exitWall = 2;
            return exitPoint;
        }
        
        if (!bTopWall) {
            double yWallDiff = currentCell[4] - currentPoint[1];
            double xWallDiff = xEndDiff * (yWallDiff / yEndDiff);
            double xPos = currentPoint[0] + xWallDiff;
            
            exitPoint[0] = xPos;
            exitPoint[1] = currentCell[4];
            m_exitWall = 0;
            return exitPoint;
        }
        
        // It should never get to this point, but just in case...
        exitPoint[0] = endPoint[0];
        exitPoint[1] = endPoint[1];
        m_exitWall = -1;
        
        return exitPoint;
    }
    
//------------------------------------------------------------------------------
    /**
     * Calculate the area under the curve in the interleaved array.  Write the
     * value in the text markup child of this object.
     */
    public void setAreaUnderCurve(double[] array) {
        
        if (m_textMarkup == null)
            return;
        
        double totalArea = 0;
        
        if (array.length < 4) {
            m_textMarkup.setText("0");
            return;
        }
        
        double lastX = array[0];
        double lastY = array[1];
        for (int i = 2; i < (array.length - 1); i += 2) {
            double xDist = array[i] - lastX;
            double avgY = (array[i + 1] + lastY) / 2;
            double area = xDist * avgY;
            totalArea += area;
            lastX = array[i];
            lastY = array[i + 1];
        }
        
        m_textMarkup.setText("" + totalArea);
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides default update() in GrObject.  Called by an observed GrObject's
     * notifyObservers() method to handle changes to the observed object.
     * May use the notifyObservers() method to propagate the event handling up to
     * this object's observers.
     * @param gro   The observed object that changed.
     * @param arg   Additional information that the updated object may choose to
     *              send.
     */
    public void update(GrObject gro, Object arg) {
        
        if (gro == m_surfaceDs || gro == m_inputCsDs)
            recalculate();
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of menu items that execute commands on this object.
     * @return  A menu item that can be placed on a menu and send an ActionEvent
     *          back to this object when it is chosen.
     */
    public javax.swing.JMenuItem getMenuCommands() {
        
        javax.swing.JMenu subMenu = new javax.swing.JMenu(
            "Cross Section Calculator");
        javax.swing.JMenuItem saveItem = new javax.swing.JMenuItem("Calculate");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
        return subMenu;
    }
    
//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {
        
        javax.swing.JMenuItem source =
            (javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Calculate")) {
            recalculate();
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {
        
        String s = "Cross Section Calculator";
        
        return s;
    }
}

