/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-02-27
 * Date started:    1999-08-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.gr.tools;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.util.*;


//==============================================================================
/**
 */

public class DigitalFilter extends GrObject implements
    gov.usgs.gr.tools.DataManipulator, java.awt.event.ActionListener {
    
    int m_tDim = 0;
    int[] m_fDims = {1};
    double m_outputInterval = 0;
    double m_timeScale = 1;
    boolean m_bParabolicInterpolation = false;
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GodinFilter.gif";
    
//------------------------------------------------------------------------------
    /**
     * Creates an empty filter object.
     */
    public DigitalFilter() {
        
        setDefaults("gov.usgs.sfhydro.gr.tools.DigitalFilter");
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides add in GrObject.
     */
    public boolean add(GrObject gro) {
        
        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();
        
        if (m_children.contains(gro)) return false;
        
        boolean bAdded = false;
        
        if (gro instanceof DoubleDataSeries) {
            DoubleDataSeries ds = (DoubleDataSeries)gro;
            
            m_children.add(ds);
            bAdded = true;
        } else {
            // Try adding all of the descendants in case they hold a
            // data series.  This allows curves to be traversed and
            // have their data series added.
            Enumeration enumeration = gro.children();
            while (enumeration.hasMoreElements()) {
                if (add((GrObject)enumeration.nextElement()))
                    bAdded = true;
            }
        }
        
        return bAdded;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public GrHashList getProperties() {
        
        GrHashList props = super.getProperties();
        
        props.put("tDim", new GrProperty(m_tDim, "Time Dimension"));
        
        // Construct a string listing the dimensions to be filtered.
        // First find out how many dimensions are in the first data series.
        int numDims = Integer.MAX_VALUE;
        try {
            DoubleDataSeries series = (DoubleDataSeries)m_children.get(0);
            numDims = series.getNumDimensions();
        } catch (Exception e) {}
        
        // Only write the dimension values that are less than the number
        // of dimensions for the first series.  Other values may be in the
        // list from the last time performAction() was called if there
        // are multiple points at each time value.
        String s = "" + m_fDims[0];
        for (int i = 1; i < m_fDims.length; i++) {
            if (i > numDims) break;
            
            s += ", ";
            s += m_fDims[i];
        }
        props.put("fDims", new GrProperty(s, "Dimensions to be Filtered"));
        
        props.put("interval", new GrProperty(m_outputInterval,
            "Output Interval"));
        props.put("timeScale", new GrProperty(m_timeScale,
            "Time Scale in Days"));
        props.put("parabolicInterpolation", new GrProperty(
            m_bParabolicInterpolation, "Parabolic Interpolation"));
        
        return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public void setProperties(HashMap props) {
        
        super.setProperties(props);
        
        GrProperty p;
        
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;
                
                if (key.equals("tDim"))
                    m_tDim = ((Integer)p.value);
                if (key.equals("fDims")) {
                    m_fDims = AsciiIO.parseLineAsInts((String)p.value,
                        " \t\n\r\f,");
                    Arrays.sort(m_fDims);
                }
                if (key.equals("interval"))
                    m_outputInterval = (Double)p.value;
                if (key.equals("timeScale"))
                    m_timeScale = (Double)p.value;
                if (key.equals("parabolicInterpolation"))
                    m_bParabolicInterpolation = (Boolean)p.value;
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Set a single property.
     */
    public void setProperty(String key, String value) {
        
        super.setProperty(key, value);
        
        if (key.equals("tDim"))
            m_tDim = Integer.parseInt(value);
        if (key.equals("fDims")) {
            m_fDims = AsciiIO.parseLineAsInts(value, " \t\n\r\f,");
            Arrays.sort(m_fDims);
        }
        if (key.equals("interval"))
            m_outputInterval = Double.parseDouble(value);
        if (key.equals("timeScale"))
            m_timeScale = Double.parseDouble(value);
        if (key.equals("parabolicInterpolation"))
            m_bParabolicInterpolation = Boolean.parseBoolean(value);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
        
        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();
        
        return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {
        
        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);
        
        if (icon == null)
            icon = super.setTreeIcon();
        
        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);
        
        return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Fills gaps between selected points in each data series.
     * <p><b> Pre Conditions </b><br>
     *      The data list contains DoubleDataSeries objects.
     *		The units of the time dimension is in days.
     * @param dataList  A list of DoubleDataSeries objects that will be filtered.
     */
    public void performAction(List dataList) {
        
        // Loop through each series in the data list.
        Iterator iter = dataList.iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries)iter.next();
            series = checkForStaticTime(series);
            if (series != null) performAction(series);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Perform the action on a data series.  Assume the units of the time
     * dimension are in days.
     * @param series  The DoubleDataSeries that will be modified.
     */
    public void performAction(DoubleDataSeries series) {
        
        synchronized (series) {
            
            int numDims = series.getNumDimensions();
            series.m_bParabolicInterpolation = m_bParabolicInterpolation;
            
            
            // Retrieve the interleaved x-y array from the series.
            double[] oldArray = series.getArray();
            
            // show dialog
            final gov.ca.dwr.des.digitalfilter.DigitalFilterDialog dialog =
                new gov.ca.dwr.des.digitalfilter.DigitalFilterDialog(
                null,oldArray,numDims,0,1,0.25);
            dialog.pack();
            
            dialog.setVisible(true);
            
            if(dialog.filterComplete()){
                
                series.setArray(dialog.getDigitalFilterIO().getInterleavedData());
                series.selectAll();
                
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Check if the data series has a time dimension that advances slowly with
     * multiple points at the exact same time, followed by multiple points at
     * another time.
     * @param series    The series to check.
     */
    public DoubleDataSeries checkForStaticTime(DoubleDataSeries series) {
        
        double[] array = series.getArray();
        int seriesNumDims = series.getNumDimensions();
        int[] numDims = {0, 0, 0};
        int currentIndex = 0;
        
        // Check the first three time steps and count how many points are at
        // each one.
        for (int i = 0; i < 3; i++) {
            if (currentIndex >= array.length) return series;
            
            double currentT = array[currentIndex + m_tDim];
            while (array[currentIndex + m_tDim] == currentT) {
                currentIndex += seriesNumDims;
                numDims[i]++;
                if (currentIndex >= array.length) break;
            }
        }
        
        int dimsPerTime = numDims[0] * seriesNumDims;
        
        // If there was more than one point at each time step, and there
        // were the same number of points at each of the first three time
        // steps, return a new series with more dimensions.
        if (numDims[0] == numDims[1] && numDims[0] == numDims[2]) {
            if (numDims[0] > seriesNumDims) {
                
                // Multiply the time dimension by a scale factor,
                // if one was given.
                if (m_timeScale != 1) {
                    series.selectAll();
                    series.modifySelectedPoints(3, m_tDim, m_timeScale);
                    series.deselectAll();
                    array = series.getArray();
                }
                
                DoubleDataSeries newSeries = new DoubleDataSeries(2);
                double[] newBigArray = null;
                
                // Fill in the time dimension...
                int numTimeSteps = array.length / dimsPerTime;
                double[] array2 = new double[2 * numTimeSteps];
                for (int i = 0; i < numTimeSteps; i++) {
                    array2[i * 2] = array[i * dimsPerTime + m_tDim];
                }
                
                // Loop through the dimensions to be filtered.
                int fTimeSteps = 0;
                for (int k = 0; k < numDims[0]; k++) {
                    for (int j = 0; j < m_fDims.length; j++) {
                        // Fill in the values for the filtered dimension
                        for (int i = 0; i < numTimeSteps; i++) {
                            array2[i * 2 + 1] = array[i * dimsPerTime
                                + k * seriesNumDims + m_fDims[j]];
                        }
                        
                        // Filter the series.
                        newSeries.setArray(array2);
                        performAction(newSeries);
                        if (m_outputInterval > 0)
                            newSeries.resample(m_outputInterval, 0, m_tDim);
                        double[] fArray = newSeries.getArray();
                        
                        // Create a new series for reassembling the filtered
                        // series if it hasn't been created already.
                        if (newBigArray == null) {
                            fTimeSteps = newSeries.getCount();
                            newBigArray = new double[fTimeSteps * dimsPerTime];
                            for (int m = 0; m < numDims[0]; m++) {
                                for (int i = 0; i < fTimeSteps; i++) {
                                    newBigArray[i * dimsPerTime
                                        + m * seriesNumDims + m_tDim] =
                                        fArray[i * 2];
                                }
                            }
                        }
                        
                        // Add the newly filtered dimension to the big array.
                        for (int i = 0; i < fTimeSteps; i++) {
                            newBigArray[i * dimsPerTime + k * seriesNumDims
                                + m_fDims[j]] = fArray[i * 2 + 1];
                        }
                    }
                    
                    // Loop through all the dimensions and fill in the ones
                    // that were not filtered.
                    for (int j = 0; j < seriesNumDims; j++) {
                        int fDimIndex = Arrays.binarySearch(m_fDims, j);
                        
                        // Skip the time dimension and the dimensions that
                        // have already been added through filtering...
                        if (j == m_tDim || fDimIndex >= 0) continue;
                        
                        // Fill in the values for the unfiltered dimension
                        for (int i = 0; i < numTimeSteps; i++) {
                            array2[i * 2 + 1] = array[i * dimsPerTime
                                + k * seriesNumDims + j];
                        }
                        
                        // Put the new X-Y array into a series and resample it.
                        DoubleDataSeries ufSeries =
                            new DoubleDataSeries(array2, 2);
                        ufSeries.resample(newSeries);
                        double[] ufArray = ufSeries.getArray();
                        
                        // Transfer the interpolated, unfiltered values to
                        // the big array.
                        for (int i = 0; i < fTimeSteps; i++) {
                            newBigArray[i * dimsPerTime
                                + k * seriesNumDims + j] = ufArray[i * 2 + 1];
                        }
                    }
                }
                
                series.setArray(newBigArray);
                
                return null;
            }
        }
        
        return series;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of menu items that execute commands on this object.
     * @return  A menu item that can be placed on a menu and send an ActionEvent
     *          back to this object when it is chosen.
     */
    public javax.swing.JMenuItem getMenuCommands() {
        
        javax.swing.JMenu subMenu = new javax.swing.JMenu(
            "Digital Filter");
        javax.swing.JMenuItem saveItem = new javax.swing.JMenuItem("Calculate");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
        return subMenu;
    }
    
//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {
        
        javax.swing.JMenuItem source =
            (javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Calculate")) {
            recalculate();
        }
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public void recalculate() {
        
        performAction(new ArrayList(m_children));
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {
        
        String s = "Digital Filter";
        
        return s;
    }
}

