/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-03-15
 * Date started:    2007-03-15
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.gr.tools;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.util.*;

//==============================================================================
/**
 */

public class LowPassFilter extends GrObject implements
    gov.usgs.gr.tools.DataManipulator, java.awt.event.ActionListener {

    int m_tDim = 0;
    int[] m_fDims = {1};
    double m_outputInterval = 0;
    double m_timeScale = 1;
    boolean m_bParabolicInterpolation = false;
    
    protected double m_dataInterval = 0.041666666666666666666666666666667;
    protected int m_n = 24;
    protected int m_n_minusOne = 23;
    protected int m_n_plusOne = 25;
    protected int m_half_n = 12;
    protected int m_half_n_minusOne = 11;
    
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);

//------------------------------------------------------------------------------
/**
 * Creates an empty filter object.
 */
    public LowPassFilter () {

        setDefaults("gov.usgs.sfhydro.gr.tools.LowPassFilter");
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

		boolean bAdded = false;

        if (gro instanceof DataSeries) {
        	DataSeries ds = (DataSeries)gro;

            m_children.add(ds);
            bAdded = true;
        }
        else {
            // Try adding all of the descendants in case they hold a
            // data series.  This allows curves to be traversed and
            // have their data series added.
            Enumeration enumeration = gro.children();
            while (enumeration.hasMoreElements()) {
                if (add((GrObject)enumeration.nextElement()))
                    bAdded = true;
            }
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("tDim", new GrProperty(m_tDim, "Time Dimension"));

        // Construct a string listing the dimensions to be filtered.
        // First find out how many dimensions are in the first data series.
        int numDims = Integer.MAX_VALUE;
        try {
            DataSeries series = (DataSeries)m_children.get(0);
            numDims = series.getNumDimensions();
        }
        catch (Exception e) {}

        // Only write the dimension values that are less than the number
        // of dimensions for the first series.  Other values may be in the
        // list from the last time performAction() was called if there
        // are multiple points at each time value.
        String s = "" + m_fDims[0];
        for (int i = 1; i < m_fDims.length; i++) {
            if (i > numDims) break;

            s += ", ";
            s += m_fDims[i];
        }
        props.put("fDims", new GrProperty(s, "Dimensions to be Filtered"));

        props.put("interval", new GrProperty(m_outputInterval,
            "Output Interval"));
        props.put("n", new GrProperty(m_n, "n"));
        props.put("dataInterval", new GrProperty("" + m_dataInterval,
            "Data Interval"));
        props.put("timeScale", new GrProperty(m_timeScale,
            "Time Scale in Days"));
        props.put("parabolicInterpolation", new GrProperty(
            m_bParabolicInterpolation, "Parabolic Interpolation"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("tDim"))
                    m_tDim = ((Integer)p.value);
            	if (key.equals("fDims")) {
                    m_fDims = AsciiIO.parseLineAsInts((String)p.value,
                        " \t\n\r\f,");
                    Arrays.sort(m_fDims);
                }
            	if (key.equals("dataInterval"))
                    m_dataInterval = TimeReference.convertToDecimalDay(
                        (String)p.value);
            	if (key.equals("n")) {
            		m_n = (Integer)p.value;
                        m_n += (m_n % 2);
                        m_n_minusOne = m_n - 1;
                        m_n_plusOne = m_n + 1;
                        m_half_n = m_n / 2;
                        m_half_n_minusOne = (m_n / 2) - 1;
                }
            	if (key.equals("interval"))
            		m_outputInterval = (Double)p.value;
            	if (key.equals("timeScale"))
            		m_timeScale = (Double)p.value;
            	if (key.equals("parabolicInterpolation"))
            		m_bParabolicInterpolation = (Boolean)p.value;
            }
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("tDim"))
            m_tDim = Integer.parseInt(value);
        if (key.equals("fDims")) {
            m_fDims = AsciiIO.parseLineAsInts(value, " \t\n\r\f,");
            Arrays.sort(m_fDims);
        }
        if (key.equals("dataInterval"))
            m_dataInterval = TimeReference.convertToDecimalDay(value);
        if (key.equals("n")) {
            m_n = Integer.parseInt(value);
            m_n += (m_n % 2);
            m_n_minusOne = m_n - 1;
            m_n_plusOne = m_n + 1;
            m_half_n = m_n / 2;
            m_half_n_minusOne = (m_n / 2) - 1;
        }
        if (key.equals("interval"))
            m_outputInterval = Double.parseDouble(value);
        if (key.equals("timeScale"))
            m_timeScale = Double.parseDouble(value);
        if (key.equals("parabolicInterpolation"))
            m_bParabolicInterpolation = Boolean.parseBoolean(value);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Fills gaps between selected points in each data series.
 * <p><b> Pre Conditions </b><br>
 *      The data list contains DoubleDataSeries objects.
 *		The units of the time dimension is in days.
 * @param dataList  A list of DoubleDataSeries objects that will be filtered.
 */
    public void performAction (List dataList) {

    	// Loop through each series in the data list.
    	Iterator iter = dataList.iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries)iter.next();
            if (series != null) performAction(series);
        }
    }

//------------------------------------------------------------------------------
/**
 * Performs the action on a data series.  Assumes the units of the time
 * dimension are in days.
 * @param series  A DoubleDataSeries to use as input.
 */
    public void performAction (DoubleDataSeries series) {

        synchronized (series) {
            // Loop through each series in the data list.
            int numDims = series.getNumDimensions();
            series.m_bParabolicInterpolation = m_bParabolicInterpolation;

            // Resample to data interval, assuming units in days.
            double offset = 0;
            series.resample(m_dataInterval, offset, m_tDim);

            // Retrieve the interleaved x-y array from the series.
            double[] oldArray = series.getArray();

            // Perform a moving average of the previous half_n and following
            // half_n_minusOne points for each point.
            int newSize = oldArray.length - (m_n_minusOne * numDims);
            if (newSize < 0) return;
            double[] array_pass1 = new double[newSize];
            for (int i = 0; i < (array_pass1.length / numDims); i++) {
                for (int k = 0; k < numDims; k++) {
    //                    int fDimIndex = Arrays.binarySearch(m_fDims, k);

                    // If this is the X dimension...
    //                    if (k == m_tDim || fDimIndex < 0) {
                    if (k == m_tDim) {
                        array_pass1[i * numDims + k] =
                            oldArray[(i + m_half_n) * numDims + k];
                    }
                    else {
                        // If this dimension is to be filtered...
                        double total = 0;
                        for (int j = 0; j < m_n; j++) {
                            total += oldArray [(i + j) * numDims + k];
                        }
                        array_pass1[i * numDims + k] = total / m_n;
                    }
                }
            }

            // Perform a moving average of the previous half_n_minusOne and
            // following half_n points for each point.
            newSize = array_pass1.length - (m_n_minusOne * numDims);
            if (newSize < 0) return;
            double[] array_pass2 =
                new double[newSize];
            for (int i = 0; i < (array_pass2.length / numDims); i++) {
                for (int k = 0; k < numDims; k++) {
    //                    int fDimIndex = Arrays.binarySearch(m_fDims, k);

                    // If this is the X dimension...
    //                    if (k == m_tDim || fDimIndex < 0) {
                    if (k == m_tDim) {
                        array_pass2[i * numDims + k] =
                            array_pass1[(i + m_half_n_minusOne) * numDims + k];
                    }
                    else {
                        // If this dimension is to be filtered...
                        double total = 0;
                        for (int j = 0; j < m_n; j++) {
                            total += array_pass1[(i + j) * numDims + k];
                        }
                        array_pass2[i * numDims + k] = total / m_n;
                    }
                }
            }
            array_pass1 = null;

            // Perform a moving average of the previous half_n and following
            // half_n points for each point.
            newSize = array_pass2.length - (m_n * numDims);
            if (newSize < 0) return;
            double[] array_pass3 = new double[newSize];
            for (int i = 0; i < (array_pass3.length / 2); i++) {
                for (int k = 0; k < numDims; k++) {
    //                    int fDimIndex = Arrays.binarySearch(m_fDims, k);

                    // If this is the X dimension...
    //                    if (k == m_tDim || fDimIndex < 0) {
                    if (k == m_tDim) {
                        array_pass3[i * numDims + k] =
                            array_pass2[(i + m_half_n) * numDims + k];
                    }
                    else {
                        // If this dimension is to be filtered...
                        double total = 0;
                        for (int j = 0; j < m_n_plusOne; j++) {
                            total += array_pass2[(i + j) * numDims + k];
                        }
                        array_pass3[i * numDims + k] = total / m_n_plusOne;
                    }
                }
            }

            // Replace the old array in the series with the new filtered one.
            series.setArray(array_pass3);
            series.selectAll();
        }
    }


//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * @return  A menu item that can be placed on a menu and send an ActionEvent
 *          back to this object when it is chosen.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"Low Pass Filter");
        javax.swing.JMenuItem saveItem = new javax.swing.JMenuItem("Calculate");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
    	return subMenu;
    }

//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Calculate")) {
        	recalculate();
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void recalculate () {

        performAction(new ArrayList(m_children));
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Low Pass Filter";

        return s;
 	}
}

