/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-21
 * Date started:    2007-04-20
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.gr.tools;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.TimeMarker;
import gov.usgs.gr.visual.*;
import gov.usgs.gr.visual.gl.*;
import java.nio.DoubleBuffer;
import java.util.*;

//==============================================================================
/**
 *
 */

public class LsAnimCalculator extends GrObject implements
    java.awt.event.ActionListener {
    
    /** The time step for the particle animation. */
    protected double m_ts = 1d / 24d / 4d;
    /** The particle animation itself. */
    protected GlParticleAnimation m_glp = null;
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/ParticleAnimCalculator.gif";
    
//------------------------------------------------------------------------------
    /**
     * Creates an empty object.
     */
    public LsAnimCalculator() {
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides add in GrObject.  Only allows DataDimensions to be added.
     * @param gro   The object to add.
     * @return  Whether the object was successfully added.
     */
    public boolean add(GrObject gro) {
	
	// If gro is an object effigy, get the eventual subject and try
	// adding that.
	while (gro instanceof gov.usgs.gr.ObjectEffigy)
	    gro = ((ObjectEffigy)gro).getSubject();
	
	return super.add(gro);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
	
	if (TREE_ICON_HOLDER.size() <= 0)
	    setTreeIcon();
	
	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {
	
	javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);
	
	if (icon == null)
	    icon = super.setTreeIcon();
	
	TREE_ICON_HOLDER.clear();
	TREE_ICON_HOLDER.add(icon);
	
	return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the current values for this object.
     * Properties include: <p>
     * timeStep         String  "Time Step"                                     <p>
     * @return  A list of GrProperty objects.
     */
    public GrHashList getProperties() {
	
	// First add properties from the super class.
	GrHashList props = super.getProperties();
	
	String val = null;
	props.put("timeStep", new GrProperty(getTsString(), "Time Step"));
	
	return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Set the values of one or more of this object's properties.
     * @see gov.usgs.sfhydro.gr.tools.LsAnimCalculator#getProperties getProperties()
     * for a list of accepted properties.
     * @param props A list of GrProperty objects specifying the new values for
     *              certain properties.
     */
    public void setProperties(HashMap props) {
	
	GrProperty p;
	
	// For each property with a string value, pass it to the setProperty()
	// method to be handled there.
	Iterator iter = props.keySet().iterator();
	while (iter.hasNext()) {
	    String key = (String)iter.next();
	    Object obj = props.get(key);
	    if (obj instanceof GrProperty) {
		p = (GrProperty)obj;
		
		// Handle string properties by passing them to the
		// setProperty() method.
		if (p.value instanceof String) {
		    setProperty(key, (String)p.value);
		    continue;
		}
	    }
	    // Handle string properties by passing them to the
	    // setProperty() method.
	    if (obj instanceof String) {
		setProperty(key, (String)obj);
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Set a single property.
     * @see gov.usgs.sfhydro.gr.tools.LsAnimCalculator#getProperties getProperties()
     * for a list of accepted properties.
     * @param key   The property name.
     * @param value The new value for the property.
     */
    public void setProperty(String key, String value) {
	
	super.setProperty(key, value);
	
	if (key.equals("timeStep"))
	    m_ts = TimeReference.convertToDecimalDay(value);
    }
    
//------------------------------------------------------------------------------
    public String getTsString() {
	
	String val = null;
	if (m_ts < 1) {
	    double ts = m_ts * 24d;
	    if (ts < 1) {
		ts *= 60d;
		if (ts < 1) {
		    ts *= 60d;
		    val = "" + ts + "s";
		} else {
		    val = "" + ts + "m";
		}
	    } else {
		val = "" + ts + "h";
	    }
	} else {
	    val = "" + m_ts + "d";
	}
	
	return val;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public void recalculate() {
	
	GrController grCon = Gr.getController();
	TimeSeries series = null;
	
	Iterator iter = m_children.iterator();
	while (iter.hasNext()) {
	    GrObject gro = (GrObject)iter.next();
	    if (gro instanceof TimeSeries)
		series = (TimeSeries)gro;
	}
	
	if (grCon == null || series == null)
	    return;
	try {
	    // Use the page that observes this calculator, if there is one.
	    // Otherwise, use the current page, or create one.
	    PageOfGraphs page = null;
	    List<GrObject> observers = getObservers();
	    for (GrObject gro : observers) {
		if (gro instanceof PageOfGraphs) {
		    page = (PageOfGraphs)gro;
		    break;
		}
	    }
	    if (page == null)
		page = grCon.getPage();
	    if (page == null) {
		grCon.createAnimation();
		page = grCon.getPage();
	    }
		
	    List<Graph> list = page.getGraphList();
	    
	    for (Graph graph : list) {
		GrHashList props = graph.getProperties();
		GrProperty p = (GrProperty) props.get("aspect1to1");
		boolean bAspect1to1 = (Boolean)p.value;
		
		if (bAspect1to1) {
		    List<Curve> curveList = graph.getCurves();
		    List<List> pathList = createPathList(curveList);
		    
		    // Create and add a particle animation to the graph.
		    GlParticleAnimation glp =
			createParticleAnimation(series, pathList);
		    graph.add(glp);
		    
		    // Add the time series to the time marker.
		    TimeSeries tSeries = (TimeSeries)glp.firstChildInstanceOf(
			"gov.usgs.gr.data.TimeSeries", false);
		    if (tSeries != null) {
			TimeMarker tm = (TimeMarker)page.firstChildInstanceOf(
			    "gov.usgs.gr.tools.TimeMarker", false);
			
			// Create a new animation if the time marker doesn't
			// exist.
			if (tm == null)
			    grCon.createAnimation();
			tm = (TimeMarker)page.firstChildInstanceOf(
			    "gov.usgs.gr.tools.TimeMarker", false);
			if (tm != null)
			    tm.add(tSeries);
		    }
		    
		    break;
		}
	    }
	} catch (NullPointerException e) {
	    return;
	}
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    protected List<List> createPathList(List<Curve> curveList) {
	
	List<List> pathList = new ArrayList();
	
	for (Curve curve : curveList) {
	    String title = curve.getTitle();
	    title = title.toUpperCase();
	    int index1 = title.indexOf("LS");
	    if (index1 < 0)
		continue;
	    index1 += 2;
	    int index2 = title.indexOf("LS", index1);
	    if (index2 < 0)
		continue;
	    String str1 = title.substring(index1, index2).trim();
	    index2 += 2;
	    String str2 = title.substring(index2).trim();
	    
	    try {
		int startNum = Integer.parseInt(str1);
                // Assign MAX_VALUE if the ending station number is "end".
		int endNum = Integer.MAX_VALUE;
                if (!str2.toLowerCase().equals("end"))
                    endNum = Integer.parseInt(str2);
		
		double[] pathArray = curve.getSeries().getArray();
		int numPoints = pathArray.length / 2;
		double[] ratioArray = new double[numPoints];
		ratioArray[0] = 0;
		double totalDist = 0;
		for (int i = 1; i < numPoints; i++) {
		    double xDiff = pathArray[i * 2] - pathArray[(i - 1) * 2];
		    double yDiff =
			pathArray[i * 2 + 1] - pathArray[(i - 1) * 2 + 1];
		    double dist = Math.sqrt((xDiff * xDiff) + (yDiff * yDiff));
		    totalDist += dist;
		    ratioArray[i] = totalDist;
		}
		for (int i = 1; i < numPoints; i++) {
		    ratioArray[i] /= totalDist;
		}
		ratioArray[numPoints - 1] = 1;
		
		List pathItems = new ArrayList();
		pathItems.add(startNum);
		pathItems.add(endNum);
		pathItems.add(DoubleBuffer.wrap(ratioArray));
		pathItems.add(DoubleBuffer.wrap(pathArray));
		pathItems.add(totalDist);
		
		pathList.add(pathItems);
	    } catch (NumberFormatException ex) {
		continue;
	    }
	}
	
	return pathList;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    protected GlParticleAnimation createParticleAnimation(
	TimeSeries ptsSeries, List<List> pathList) {
	
	// Initialize variables.
	double[] ptsArray = ptsSeries.getArray();
	int numPtsDims = ptsSeries.getNumDimensions();
	int numPtsPoints = ptsSeries.getNumPoints();
	int lastPartNum = Integer.MAX_VALUE;
	int numPrts = 0;
	double minTime = Double.MAX_VALUE;
	double maxTime = Double.MIN_VALUE;
	List<Integer> missingPaths = new ArrayList<Integer>();
	
	// Find the number of particles and time range in the pts array.
	for (int i = 0; i < ptsArray.length; i += numPtsDims) {
	    int partNum = (int)Math.round(ptsArray[i]);
	    if (partNum != lastPartNum) {
		numPrts++;
		lastPartNum = partNum;
	    }
	    
	    double jDay = ptsArray[i + 1];
	    if (jDay > maxTime)
		maxTime = jDay;
	    if (jDay < minTime)
		minTime = jDay;
	}
	
	// Create a time array from the start time to the end time at the
	// specified time interval.
	double[] timeArray = new double[2];
	timeArray[0] = minTime;
	timeArray[1] = maxTime;
	TimeSeries tSeries = new TimeSeries(timeArray, 1);
	tSeries.setStartDate(ptsSeries.getStartDate());
	tSeries.resample(m_ts, 0);
	tSeries.deselectAll();
	tSeries.selectPoint(0);
	tSeries.selectPoint(tSeries.getNumPoints() - 1);
	tSeries.deleteSelectedPoints();
	tSeries.deselectAll();
	timeArray = tSeries.getArray();
	int numTimes = tSeries.getNumPoints();
        List<List> endPlots = new ArrayList<List>();
        List endPlot = new ArrayList();
	
	// Create the time and X-Y-Viability arrays for the particle animation.
	int numOutDims = 3;
	double[] xyvArray = new double[numTimes * numPrts * numOutDims];
	Arrays.fill(xyvArray, Double.NaN);
	
	// Initialize particle index values.
	int partStart = 0;
	int partEnd = -1;
	int partNum = (int)Math.round(ptsArray[0]);
	int partIndex = 0;
	
	// Look at each line in the pts array and process each particle, once
	// the end of its data has been reached.
	for (int i = 1; i <= numPtsPoints; i++) {
            // Mark when a new particle has been found, so the last particle
	    // can be processed.
	    if (i == numPtsPoints ||
		partNum != (int)Math.round(ptsArray[i * numPtsDims])) {
		
		if (partEnd < partStart)
		    partEnd = i - 1;
		else
		    partStart = i;
	    }
	    
	    double xVal = Double.NaN;
	    double yVal = Double.NaN;
	    double viability = 1;
            
            if (endPlot.size() > 0) {
                endPlots.add(endPlot);
                endPlot = new ArrayList();
            }
            
	    // Process the particle if its last line in the pts array has
	    // been found.
	    if (partEnd >= partStart) {
		// Determine the time range that will be used for this particle
		// in the output array.  Get the indexes for those times.
		double firstStationTime = ptsArray[partStart * numPtsDims + 1];
		int firstValidTimeIndex =
		    tSeries.findNearestIndex(firstStationTime);
		if (timeArray[firstValidTimeIndex] < firstStationTime)
		    firstValidTimeIndex++;
		double lastStationTime = ptsArray[partEnd * numPtsDims + 1];
		int lastValidTimeIndex =
		    tSeries.findNearestIndex(lastStationTime);
		if (lastValidTimeIndex >= numTimes ||
                    timeArray[lastValidTimeIndex] > lastStationTime) {
                    
		    lastValidTimeIndex--;
                }
                int lastTimeIndex = tSeries.getCount() - 1;
		
		// Determine the particle position at each output time.
		int ptsIndex = partStart;
		for (int j = firstValidTimeIndex; j <= lastTimeIndex; j++) {
		    double particleTime = timeArray[j];
		    double ptsTime1 = Double.NEGATIVE_INFINITY;
		    int ptsStation1 = (int)Math.round(
                        ptsArray[(ptsIndex + 1) * numPtsDims + 2]);
		    double ptsTime2 = Double.POSITIVE_INFINITY;
		    int ptsStation2 = Integer.MAX_VALUE;

                    if (j <= lastValidTimeIndex) {
                        // Determine the particle output time and the time of the
                        // pts array before and after.
                        ptsTime1 = ptsArray[ptsIndex * numPtsDims + 1];
                        ptsStation1 =
                            (int)Math.round(ptsArray[ptsIndex * numPtsDims + 2]);
                        ptsTime2 = ptsArray[(ptsIndex + 1) * numPtsDims + 1];
                        ptsStation2 = (int)Math.round(
                            ptsArray[(ptsIndex + 1) * numPtsDims + 2]);
                        while (particleTime > ptsTime2) {
                            ptsIndex++;
                            ptsTime1 = ptsTime2;
                            ptsStation1 = ptsStation2;
                            if ((ptsIndex + 1) <= partEnd) {
                                ptsTime2 =
                                    ptsArray[(ptsIndex + 1) * numPtsDims + 1];
                                ptsStation2 = (int)Math.round(
                                    ptsArray[(ptsIndex + 1) * numPtsDims + 2]);
                            }
                            else
                                break;
                        }
                    }
                    else {
                        viability = 0;
                    }
		    
                    int xyvIndex = (j * numPrts + partIndex) * numOutDims;
		    
		    // Find the path between the stations.
		    int match = 0;
		    for (List pathItems : pathList) {
			double ptsTimeDiff = ptsTime2 - ptsTime1;
			if (ptsTimeDiff <= 0)
			    break;
			
			int pathStation1 = (Integer)pathItems.get(0);
			int pathStation2 = (Integer)pathItems.get(1);
			
			// If the path matches, mark it as 1 if the particle is
			// travelling the path forward and -1 if backward.
			if (ptsStation1 == pathStation1) {
			    if (ptsStation2 == pathStation2 ||
				ptsStation1 == ptsStation2)
				match = 1;
			}
			if (ptsStation1 == pathStation2) {
			    if (ptsStation2 == pathStation1 ||
				ptsStation1 == ptsStation2)
				match = -1;
			}
			
			// If this path matched the stations...
			if (match != 0) {
			    double timeRatio =
				(particleTime - ptsTime1) / ptsTimeDiff;
			    if (ptsStation1 == ptsStation2)
				timeRatio = 0;
			    if (match < 0)
				timeRatio = 1 - timeRatio;
                            
                            double[] xAndY =
                                getPositionByRatio(pathItems, timeRatio);
                            if (xAndY == null)
                                break;

			    xVal = xAndY[0];
			    yVal = xAndY[1];

                            if (viability == 0) {
                                if (endPlot.size() == 0) {
                                    endPlot.add(pathItems);
                                    endPlot.add(particleTime);
                                }
                                
                                endPlot.add(xyvIndex);
                            }
			    
			    break;
			}
		    }
		    
		    xyvArray[xyvIndex] = xVal;
		    xyvArray[xyvIndex + 1] = yVal;
		    xyvArray[xyvIndex + 2] = viability;

		    if (match == 0) {
			// Check if the missing paths list already contains this
			// pair.
			boolean bFound = false;
			Iterator<Integer> iter = missingPaths.iterator();
			while (iter.hasNext()) {
			    int p1 = iter.next();
			    int p2 = iter.next();
			    if (ptsStation1 == p1 && ptsStation2 == p2) {
				bFound = true;
				break;
			    }
			}
			
			if (!bFound) {
			    missingPaths.add(ptsStation1);
			    missingPaths.add(ptsStation2);
			}
		    }
		}
		
		if (i < numPtsPoints)
		    partNum = (int)Math.round(ptsArray[i * numPtsDims]);
		partIndex++;
		partStart = i;
	    }
	}

                    
        if (endPlot.size() > 0)
            endPlots.add(endPlot);
        
        // Arrange the particles on the end paths so they don't overlap and
        // are positioned in the order they arrive.
        while (endPlots.size() > 0) {
            List<List> samePlots = new ArrayList<List>();
            List pathItems = null;
            
            Iterator<List> iter = endPlots.iterator();
            while (iter.hasNext()) {
                endPlot = iter.next();
                if (pathItems == null)
                    pathItems = (List)endPlot.get(0);
                
                if (pathItems == (List)endPlot.get(0)) {
                    samePlots.add(endPlot);
                    iter.remove();
                }
            }
            
            List firstPlot = samePlots.get(0);
            double spacing = 1d / (double)(samePlots.size());
            if (spacing > 0.3)
                spacing = 0.3;
                
            // Sort samePlots by exit time.
            
            double pos = Double.MIN_VALUE;
            for (List plot : samePlots) {
                pathItems = (List)plot.get(0);
                double[] xAndY =
                    getPositionByRatio(pathItems, pos);
                if (xAndY == null)
                    System.out.println("getPositionByRatio returned null");

                for (int j = 3; j < plot.size(); j++) {
                    int xyvIndex = (Integer)plot.get(j);
                    xyvArray[xyvIndex] = xAndY[0];
                    xyvArray[xyvIndex + 1] = xAndY[1];
                }

                pos += spacing;
            }
        }

	// Write a message for each path that was missing.
	if (missingPaths.size() > 0) {
	    System.out.println("\nMissing paths:");
	    boolean bStation1 = true;
	    for (int station : missingPaths) {
                if (station == Integer.MAX_VALUE)
                    System.out.print("end");
                else
                    System.out.print("" + station);
		if (bStation1) {
		    System.out.print(" --> ");
		    bStation1 = false;
		} else {
		    System.out.println("");
		    bStation1 = true;
		}
	    }
	    System.out.println("\n");
	}
	
	if (m_glp == null) {
	    m_glp = new GlParticleAnimation();
	
	    m_glp.setProperty("title", "Between Stations");
	    m_glp.setProperty("scale", "20");
	    m_glp.setProperty("numTrails", "0");
	    m_glp.setProperty("seriesNums", "1, 2, 2, -1, 2");
	    m_glp.setProperty("dimNums", "0, 0, 1, -1, 2");
	    m_glp.setProperty("symbolType", "Bull's Eye");

            InterpolatingColorScheme ics = new InterpolatingColorScheme();
            ics.setProperty("discrete", "true");
            ics.setColorAtValue("(255, 255, 255) 1");
            ics.setColorAtValue("(255, 0, 0) 0");
            m_glp.add(ics);
        }

	List dsList = m_glp.findChildInstancesOf(
	    "gov.usgs.gr.data.DoubleDataSeries", false);

	// Put the data series into the particle animation.  Reuse the existing
	// series, if they are available and match the number of dimensions.
	if (dsList.size() == 2) {
	    DoubleDataSeries existingTimeSeries =
		(DoubleDataSeries)dsList.get(0);
	    DoubleDataSeries existingXyvSeries =
		(DoubleDataSeries)dsList.get(1);
	    existingTimeSeries.setArray(tSeries.getArray());
	    if (existingXyvSeries.getNumDimensions() == (numOutDims * numPrts))
		existingXyvSeries.setArray(xyvArray);
	    else {
                m_glp.remove(existingXyvSeries);
		DoubleDataSeries xyvSeries =
		    new DoubleDataSeries(xyvArray, numOutDims * numPrts);
		m_glp.add(xyvSeries);	    
	    }
	}
	else {
	    // Particle animation needs new series.
	    m_glp.add(tSeries);
	    DoubleDataSeries xyvSeries =
		new DoubleDataSeries(xyvArray, numOutDims * numPrts);
	    m_glp.add(xyvSeries);
	}
	
	return m_glp;
	
    }
    
//------------------------------------------------------------------------------
    /**
     * 
     */
    public double[] getPositionByRatio(List pathItems, double placeRatio) {
        
        double[] ratioArray =
            ((DoubleBuffer)pathItems.get(2)).array();
        int ip = Arrays.binarySearch(ratioArray, placeRatio);
        if (ip < 0)
            ip = -(ip + 1);
        if (ip == 0)
            return null;
        if (ip >= ratioArray.length)
            ip = ratioArray.length - 1;

        double r1 = ratioArray[ip - 1];
        double r2 = ratioArray[ip];
        double rDiff = r2 - r1;
        double posRatio = (placeRatio - r1) / rDiff;
        if (Double.isNaN(posRatio))
            posRatio = 0;

        double[] xyPathArray =
            ((DoubleBuffer)pathItems.get(3)).array();
        double x1 = xyPathArray[(ip - 1) * 2];
        double y1 = xyPathArray[(ip - 1) * 2 + 1];
        double x2 = xyPathArray[ip * 2];
        double y2 = xyPathArray[ip * 2 + 1];
        double xDiff = x2 - x1;
        double yDiff = y2 - y1;
        
        double[] xAndY = new double[2];
        xAndY[0] = x1 + (posRatio * xDiff);
        xAndY[1] = y1 + (posRatio * yDiff);
        
        return xAndY;
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides default update() in GrObject.  Called by an observed GrObject's
     * notifyObservers() method to handle changes to the observed object.
     * May use the notifyObservers() method to propagate the event handling up to
     * this object's observers.
     * @param gro   The observed object that changed.
     * @param arg   Additional information that the updated object may choose to
     *              send.
     */
    public void update(GrObject gro, Object arg) {
	
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of menu items that execute commands on this object.
     * @return  A menu item that can be placed on a menu and send an ActionEvent
     *          back to this object when it is chosen.
     */
    public javax.swing.JMenuItem getMenuCommands() {
	
	javax.swing.JMenu subMenu = new javax.swing.JMenu(
	    "Listening Station Animation Calculator");
	javax.swing.JMenuItem item = new javax.swing.JMenuItem("Calculate");
	item.addActionListener(this);
	subMenu.add(item);
	
	return subMenu;
    }
    
//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {
	
	javax.swing.JMenuItem source =
	    (javax.swing.JMenuItem)(e.getSource());
	String srcText = source.getText();
	if (srcText.equals("Calculate")) {
	    recalculate();
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {
	
	String s = "Listening Station Animation Calculator";
	
	return s;
    }
}

