/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2004-06-25
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.gr.tools;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.util.*;

//==============================================================================
/**
 */

public class RmsCalculator extends GrObject implements
    gov.usgs.gr.tools.DataManipulator, java.awt.event.ActionListener {

    int m_tDim = 0;
    int[] m_fDims = {1};
    double m_outputInterval = 0;
    double m_timeScale = 1;
    boolean m_bParabolicInterpolation = false;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/RmsCalculator.gif";

//------------------------------------------------------------------------------
/**
 * Creates an empty filter object.
 */
    public RmsCalculator () {

        setDefaults("gov.usgs.sfhydro.gr.tools.RmsCalculator");
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

		boolean bAdded = false;

        if (gro instanceof DoubleDataSeries) {
        	DoubleDataSeries ds = (DoubleDataSeries)gro;

            m_children.add(ds);
            bAdded = true;
        }
        else {
            // Try adding all of the descendants in case they hold a
            // data series.  This allows curves to be traversed and
            // have their data series added.
            Enumeration enumeration = gro.children();
            while (enumeration.hasMoreElements()) {
                if (add((GrObject)enumeration.nextElement()))
                    bAdded = true;
            }
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("tDim", new GrProperty(m_tDim, "Time Dimension"));

        // Construct a string listing the dimensions to be filtered.
        // First find out how many dimensions are in the first data series.
        int numDims = Integer.MAX_VALUE;
        try {
            DoubleDataSeries series = (DoubleDataSeries)m_children.get(0);
            numDims = series.getNumDimensions();
        }
        catch (Exception e) {}

        // Only write the dimension values that are less than the number
        // of dimensions for the first series.  Other values may be in the
        // list from the last time performAction() was called if there
        // are multiple points at each time value.
        String s = "" + m_fDims[0];
        for (int i = 1; i < m_fDims.length; i++) {
            if (i > numDims) break;

            s += ", ";
            s += m_fDims[i];
        }
        props.put("fDims", new GrProperty(s, "Dimensions to be Filtered"));

        props.put("interval", new GrProperty(m_outputInterval,
            "Output Interval"));
        props.put("timeScale", new GrProperty(m_timeScale,
            "Time Scale in Days"));
        props.put("parabolicInterpolation", new GrProperty(
            m_bParabolicInterpolation, "Parabolic Interpolation"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("tDim"))
            		m_tDim = (Integer)p.value;
            	if (key.equals("fDims")) {
                    m_fDims = AsciiIO.parseLineAsInts((String)p.value,
                        " \t\n\r\f,");
                    Arrays.sort(m_fDims);
                }
            	if (key.equals("interval"))
            		m_outputInterval = (Double)p.value;
            	if (key.equals("timeScale"))
            		m_timeScale = (Double)p.value;
            	if (key.equals("parabolicInterpolation"))
            		m_bParabolicInterpolation = (Boolean)p.value;
            }
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("tDim"))
            m_tDim = Integer.parseInt(value);
        if (key.equals("fDims")) {
            m_fDims = AsciiIO.parseLineAsInts(value, " \t\n\r\f,");
            Arrays.sort(m_fDims);
        }
        if (key.equals("interval"))
            m_outputInterval = Double.parseDouble(value);
        if (key.equals("timeScale"))
            m_timeScale = Double.parseDouble(value);
        if (key.equals("parabolicInterpolation"))
            m_bParabolicInterpolation = Boolean.parseBoolean(value);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Fills gaps between selected points in each data series.
 * <p><b> Pre Conditions </b><br>
 *      The data list contains DoubleDataSeries objects.
 *		The units of the time dimension is in days.
 * @param dataList  A list of DoubleDataSeries objects that will be filtered.
 */
    public void performAction (List dataList) {

    	// Loop through each series in the data list.
    	Iterator iter = dataList.iterator();
        while (iter.hasNext()) {
        	DoubleDataSeries series = (DoubleDataSeries)iter.next();
            if (series != null) performAction(series);
        }
    }

//------------------------------------------------------------------------------
/**
 * Performs the action on a data series.  Assumes the units of the time
 * dimension are in days.
 * @param series  A DoubleDataSeries to use as input.
 */
    public void performAction (DoubleDataSeries series) {

        synchronized (series) {
            // Square the series.
            series.selectAll();
            series.modifySelectedPoints(3, 1, series);

            // Filter the series.
            List list = new ArrayList();
            list.add(series);
            GodinFilter gf = new GodinFilter();
            gf.performAction(list);

            // Take the square root of the series.
            double[] array = series.getArray();
            int numPoints = series.getCount();
            int numDims = series.getNumDimensions();
            for (int i = 0; i < numPoints; i++) {
                double val = array[i * numDims + 1];
                array[i * numDims + 1] = Math.sqrt(array[i * numDims + 1]);

                if (array[i * numDims + 1] < 0)
                    System.out.println("sqrt(" + val + ") = "
                        + array[i * numDims + 1]);
            }

            // Reinitialize the series.
            series.arrayChanged();

            series.selectAll();
        }
    }
    
//------------------------------------------------------------------------------
/**
 * Performs action on a data series.  Assumes the units of the time dimension
 * are in days.
 * @param series  A DoubleDataSeries to use as input.
 */
    public void performRmsDiffAction (DoubleDataSeries series) {

        synchronized (series) {
            // Create a new series that is the filtered version of the
            // original.
            DoubleDataSeries series2 =
                (DoubleDataSeries)series.createCopy(true);
            List list = new ArrayList();
            list.add(series2);
            GodinFilter gf = new GodinFilter();
            gf.performAction(list);

            // Subtract the filtered series off of the original.
            series.selectAll();
            series.modifySelectedPoints(2, 1, series2);

            // Square the series.
            series.selectAll();
            series.modifySelectedPoints(3, 1, series);

            // Filter the series.
            list.clear();
            list.add(series);
            gf.performAction(list);

            // Take the square root of the series.
            double[] array = series.getArray();
            int numPoints = series.getCount();
            int numDims = series.getNumDimensions();
            for (int i = 0; i < numPoints; i++) {
                double val = array[i * numDims + 1];
                array[i * numDims + 1] = Math.sqrt(array[i * numDims + 1]);

                if (array[i * numDims + 1] < 0)
                    System.out.println("sqrt(" + val + ") = "
                        + array[i * numDims + 1]);
            }

            // Reinitialize the series.
            series.arrayChanged();

            series.selectAll();
        }
    }

//------------------------------------------------------------------------------
/**
 * 
 */
    public DoubleDataSeries createRmsSeries (DoubleDataSeries series1,
        DoubleDataSeries series2) {

        DoubleDataSeries newSeries = (DoubleDataSeries)series1.createCopy(true);

        // Subtract the filtered series off of the original.
        newSeries.selectAll();
        newSeries.modifySelectedPoints(2, 1, series2);

        // Square the series.
        newSeries.selectAll();
        newSeries.modifySelectedPoints(3, 1, series1);

        // Take the square root of the series.
        double[] array = newSeries.getArray();
        int numPoints = newSeries.getCount();
        int numDims = newSeries.getNumDimensions();
        for (int i = 0; i < numPoints; i++) {
            double val = array[i * numDims + 1];
            array[i * numDims + 1] = Math.sqrt(array[i * numDims + 1]);

            if (array[i * numDims + 1] < 0)
                System.out.println("sqrt(" + val + ") = "
                    + array[i * numDims + 1]);
        }

        // Reinitialize the series.
        newSeries.arrayChanged();

        return newSeries;      
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * @return  A menu item that can be placed on a menu and send an ActionEvent
 *          back to this object when it is chosen.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"RMS Average");
        javax.swing.JMenuItem saveItem = new javax.swing.JMenuItem("Calculate");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
    	return subMenu;
    }

//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Calculate")) {
        	recalculate();
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void recalculate () {

        performAction(new ArrayList(m_children));
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "RMS Average";

        return s;
 	}
}

