/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2008-04-24
 * Date started:    2008-04-07
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */
package gov.usgs.sfhydro.gr.tools;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.*;
import gov.usgs.gr.visual.ps.PageOfPsGraphs;
import gov.usgs.gr.visual.ps.PsCurve;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import java.text.DecimalFormat;
import java.text.FieldPosition;
import java.text.SimpleDateFormat;

/**
 * Analyzes a tidal time series to help determine if it should be filtered
 * when calculating net flows.
 */
public class TidalFilterAnalyzer {

//------------------------------------------------------------------------------
    /**
     * Constructor for creating a default object.
     */
    public TidalFilterAnalyzer() {
    }

//------------------------------------------------------------------------------
    /**
     * Analyze the time series and determine its suitability for tidal filtering.
     * Plot graphs to a PostScript file and report statistics to the user.
     * @param sourceTS  The source time series
     * @param stationLabel  The graph label that describes the station.
     * @param filename  The filename base, which is used as the base for 
     *                  PostScript output.
     */
    public void analyze(TimeSeries sourceTS, String stationLabel,
            String filename) {

        // Create a page with graphs for the output plots.
       
        PageOfPsGraphs page = new PageOfPsGraphs();
        page.m_psOptions.setPsFontSize(7);
        page.m_psOptions.setPsColor(true);

        Graph graph1 = page.addNewGraph();
        Graph graph2 = page.addNewGraph();
        Graph graph3 = page.addNewGraph();
        Graph graph4 = page.addNewGraph();
        Graph graph5 = page.addNewGraph();

        // Cause the new graphs to be shown on the page.
        page.updateLists();
        page.getFormatter().layoutGraphs();

        // Set up some time constants.
        double interval24 = TimeReference.convertToDecimalDay("24h");
        double offset = TimeReference.convertToDecimalDay("12h");
        double outputInterval = TimeReference.convertToDecimalDay("1d");

        // Determine what segments (minus gaps) are available for analysis
        // using the input series.
        List<List<Double>> segmentList = createSegmentList(sourceTS);
        List<Curve> curves = createCurveList(segmentList, sourceTS, 0, 0, 0);

        // Show the input curve as segments with no straight-lining across gaps.
        for (Curve curve : curves) {
            curve.setColor(0, 0, 0);
            graph1.add(curve);
        }

        // Find the tidal high-high and add it to the graph.
        TimeSeries hhSeries =
                (TimeSeries) sourceTS.createCopy(true);
        List hhList = new ArrayList();
        hhList.add(hhSeries);
        gov.usgs.sfhydro.gr.tools.TidalHighFinder thh =
                new gov.usgs.sfhydro.gr.tools.TidalHighFinder();
        thh.setProperty("span", "" + (30.8d / 24d));
        thh.setProperty("subSpan", "" + (6d / 24d));
        thh.performAction(hhList);
        curves = createCurveList(segmentList, hhSeries, 12d / 24d, 1, 0);
        for (Curve curve : curves) {
            curve.setLineWidth(3);
            graph1.add(curve);
        }

        // Find the tidal low-low and add it to the graph.
        TimeSeries llSeries =
                (TimeSeries) sourceTS.createCopy(true);
        List llList = new ArrayList();
        llList.add(llSeries);
        gov.usgs.sfhydro.gr.tools.TidalLowFinder tll =
                new gov.usgs.sfhydro.gr.tools.TidalLowFinder();
        tll.setProperty("span", "" + (30.8d / 24d));
        tll.setProperty("subSpan", "" + (6d / 24d));
        tll.performAction(llList);
        curves = createCurveList(segmentList, llSeries, 12d / 24d, 2, 0);
        for (Curve curve : curves) {
            curve.setLineWidth(3);
            graph1.add(curve);
        }

        // Create a daily-averaged Godin series and add it to the graph.
        TimeSeries godinDailySeries =
                (TimeSeries) sourceTS.createCopy(true);
        List godinList = new ArrayList();
        godinList.add(godinDailySeries);
        gov.usgs.sfhydro.gr.tools.GodinFilter gf =
                new gov.usgs.sfhydro.gr.tools.GodinFilter();
        gf.performAction(godinList);
        godinDailySeries.deselectAll();
        average(godinDailySeries, interval24, offset, outputInterval);
        curves = createCurveList(segmentList, godinDailySeries, 47.5d / 24d, 0, 0);
        for (Curve curve : curves) {
            graph2.add(curve);
        }

        // Create a daily-averaged series and add it to the graph.
        TimeSeries dailySeries =
                (TimeSeries) sourceTS.createCopy(true);
        average(dailySeries, interval24, offset, outputInterval);
        curves = createCurveList(segmentList, dailySeries, 47.5d / 24d, 1, 1);
        for (Curve curve : curves) {
            graph2.add(curve);
        }
        TimeSeries godinDailyDiffSeries =
                (TimeSeries) godinDailySeries.createCopy(true);
        godinDailyDiffSeries.selectAll();
        godinDailyDiffSeries.modifySelectedPoints(DoubleDataSeries.SUBTRACT, 1, dailySeries);
        godinDailyDiffSeries.deselectAll();
        curves = createCurveList(segmentList, godinDailyDiffSeries, 47.5d / 24d, 0, 0);
        for (Curve curve : curves) {
            curve.setColor(0, 0, 0);
            graph3.add(curve);
        }

        // Calculate statistics for the difference.
        double graph3Median = 0;
        double graph3Max = 0;
        double graph3Min = Double.POSITIVE_INFINITY;
        int numTotal3 = 0;
        for (Curve curve : curves) {
            numTotal3 += curve.getSeries().getNumPoints();
        }
        double[] sortArray3 = new double[numTotal3];
        int index3 = 0;
        for (Curve curve : curves) {
            double[] array = curve.getSeries().getArray();
            int numPoints = curve.getSeries().getNumPoints();
            for (int i = 0; i < numPoints; i++) {
                sortArray3[index3++] = Math.abs(array[i * 2 + 1]);
            }
            graph3Max = Math.max(graph3Max, curve.getSeries().getMax(1));
            graph3Min = Math.min(graph3Min, curve.getSeries().getMin(1));
        }
        Arrays.sort(sortArray3);
        graph3Median = sortArray3[numTotal3 / 2];
        String graph3StatText = "Median Absolute Value = ";
        graph3StatText += new DecimalFormat("0.##").format(graph3Median);
        graph3StatText += ", Minimum = ";
        graph3StatText += new DecimalFormat("0.##").format(graph3Min);
        graph3StatText += ", Maximum = ";
        graph3StatText += new DecimalFormat("0.##").format(graph3Max);
        graph3StatText += "";

        // Create a tidal range series.
        TimeSeries tidalRangeSeries =
                (TimeSeries) hhSeries.createCopy(true);
        tidalRangeSeries.selectAll();
        tidalRangeSeries.modifySelectedPoints(DoubleDataSeries.SUBTRACT, 1, llSeries);
        tidalRangeSeries.deselectAll();

        // Create a percent difference series and add it to the graph.
        TimeSeries percentDiffSeries =
                (TimeSeries) godinDailySeries.createCopy(true);
        percentDiffSeries.selectAll();
        percentDiffSeries.modifySelectedPoints(DoubleDataSeries.SUBTRACT, 1, dailySeries);
        percentDiffSeries.selectAll();
        percentDiffSeries.modifySelectedPoints(DoubleDataSeries.DIVIDE, 1, godinDailySeries);
        percentDiffSeries.selectAll();
        percentDiffSeries.modifySelectedPoints(DoubleDataSeries.ABSOLUTE_VALUE, 1, 0);
        percentDiffSeries.selectAll();
        percentDiffSeries.modifySelectedPoints(DoubleDataSeries.MULTIPLY, 1, 100);
        percentDiffSeries.deselectAll();
        curves = createCurveList(segmentList, percentDiffSeries, 47.5d / 24d, 0, 0);

        // Calculate statistics for the percent difference.
        double graph4Median = 0;
        double graph4Max = 0;
        double graph4Min = Double.POSITIVE_INFINITY;
        int numTotal4 = 0;
        for (Curve curve : curves) {
            numTotal4 += curve.getSeries().getNumPoints();
        }
        double[] sortArray = new double[numTotal4];
        int index4 = 0;
        for (Curve curve : curves) {
            double[] array = curve.getSeries().getArray();
            int numPoints = curve.getSeries().getNumPoints();
            for (int i = 0; i < numPoints; i++) {
                sortArray[index4++] = array[i * 2 + 1];
            }
            graph4Max = Math.max(graph4Max, curve.getSeries().getMax(1));
            graph4Min = Math.min(graph4Min, curve.getSeries().getMin(1));

            curve.setColor(0, 0, 0);
            graph4.add(curve);
        }
        Arrays.sort(sortArray);
        graph4Median = sortArray[numTotal4 / 2];
        String graph4StatText = "Median = ";
        graph4StatText += new DecimalFormat("0.##").format(graph4Median);
        graph4StatText += "%, Minimum = ";
        graph4StatText += new DecimalFormat("0.##").format(graph4Min);
        graph4StatText += "%, Maximum = ";
        graph4StatText += new DecimalFormat("0.##").format(graph4Max);
        graph4StatText += "%";

        // Create the "Tidal Range to Filtered Discharge" ratio series and add
        // it to the graph.
        TimeSeries ratioSeries = (TimeSeries) godinDailySeries.createCopy(true);
        ratioSeries.selectAll();
        ratioSeries.modifySelectedPoints(DoubleDataSeries.DIVIDE, 1, tidalRangeSeries);
        ratioSeries.selectAll();
        ratioSeries.modifySelectedPoints(DoubleDataSeries.ABSOLUTE_VALUE, 1, 0);
        //ratioSeries.selectAll();
        //ratioSeries.modifySelectedPoints(DoubleDataSeries.MULTIPLY, 1, percentDiffSeries);
        //ratioSeries.selectAll();
        //ratioSeries.modifySelectedPoints(DoubleDataSeries.ABSOLUTE_VALUE, 1, 0);
        ratioSeries.deselectAll();
        curves = createCurveList(segmentList, ratioSeries, 47.5d / 24d, 0, 0);

        // Calculate the ratio statistics.
        int numTotal = 0;
        int numBelow1 = 0;
        int numBetween1and3 = 0;
        int numAbove3 = 0;
        double graph5Max = 0;
        for (Curve curve : curves) {
            double[] array = curve.getSeries().getArray();
            int numPoints = curve.getSeries().getNumPoints();
            for (int i = 0; i < numPoints; i++) {
                double value = array[i * 2 + 1];
                numTotal++;
                if (value < 1) {
                    numBelow1++;
                } else if (value > 3) {
                    numAbove3++;
                } else {
                    numBetween1and3++;
                }
            }
            graph5Max = Math.max(graph5Max, curve.getSeries().getMax(1));
            curve.setColor(0, 0, 0);
            graph5.add(curve);
        }
        
        // Add lines at Y = 1 and Y = 3.
        double[] at1Array = {sourceTS.getMin(0), 1, sourceTS.getMax(0), 1};
        double[] at3Array = {sourceTS.getMin(0), 3, sourceTS.getMax(0), 3};
        DoubleDataSeries seriesAt1 = new DoubleDataSeries(at1Array, 2);
        DoubleDataSeries seriesAt3 = new DoubleDataSeries(at3Array, 2);
        PsCurve curveAt1 = new PsCurve(seriesAt1);
        curveAt1.setColor(2);
        curveAt1.setLinePattern(2);
        curveAt1.setLineWidth(3);
        PsCurve curveAt3 = new PsCurve(seriesAt3);
        curveAt3.setColor(2);
        curveAt3.setLinePattern(2);
        curveAt3.setLineWidth(3);
        graph5.add(curveAt1);
        graph5.add(curveAt3);
        
        // Create text strings for the ratio statistics.
        String graph5StatText1 = "" + numTotal + " Points Analyzed.  ";
        String graph5StatText2 = numAbove3 + " (";
        long percentAbove3 =
                Math.round(100d * (double) numAbove3 / (double) numTotal);
        graph5StatText2 += percentAbove3 + "%) above 3.  ";
        String graph5StatText3 = numBetween1and3 + " (";
        long percentBetween1and3 =
                Math.round(100d * (double) numBetween1and3 / (double) numTotal);
        graph5StatText3 += percentBetween1and3 + "%) between 1 and 3.  ";
        String graph5StatText4 = numBelow1 + " (";
        long percentBelow1 =
                Math.round(100d * (double) numBelow1 / (double) numTotal);
        graph5StatText4 += percentBelow1 + "%) below 1.";
        String graph5StatText = graph5StatText1 + graph5StatText2
                + graph5StatText3 + graph5StatText4;

        // Look at each series that is at least partially selected and
        // deselect all the points in that series.
        Iterator iter = page.getSelectedSeriesList().iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries) iter.next();
            series.deselectAll();
        }
        page.updateSelectedSeriesList();

        graph1.setSelected(true);
        graph2.setSelected(true);
        graph3.setSelected(true);
        graph4.setSelected(true);
        graph5.setSelected(true);
        page.updateLists();
        page.updateZoomContents();
        
        // Set the Y range for graph 4.
        Zoom yZoom4 = graph4.getYZoom();
        double yMin = yZoom4.getContentMin();
        double yMax = yZoom4.getContentMax();
        if (yMax < 30)
            yMax = 30;
        if (yMax > 400)
            yMax = 400;
        yMin = 0;
 	yZoom4.setContentLimits(yMin, yMax);
        
        // Set the Y range for graph 5.
        Zoom yZoom5 = graph5.getYZoom();
        yMin = yZoom5.getContentMin();
        yMax = yZoom5.getContentMax();
        if (yMax < 1)
            yMax = 1;
        if (yMax > 9)
            yMax = 9;
        yMin = 0;
 	yZoom5.setContentLimits(yMin, yMax);
        
        // Zoom all on all graphs.
        zoomAll(page);
        
        // Make a little more room at the bottom of graphs 4 and 5.
        yMax = yZoom4.getViewMax();
        yZoom4.setOuterLimits(-5, yMax);
        yZoom4.setViewLimits(-5, yMax);
        yMax = yZoom5.getViewMax();
        yZoom5.setOuterLimits(-0.5, yMax);
        yZoom5.setViewLimits(-0.5, yMax);
        double xMin = graph5.getXZoom().getViewMin();
        double xMax = graph5.getXZoom().getViewMax();
        
        // Update the X range for the Y = 1 and Y = 3 lines on graph 5.
        at1Array[0] = xMin;
        at1Array[2] = xMax;
        at3Array[0] = xMin;
        at3Array[2] = xMax;
        seriesAt1.arrayChanged();
        seriesAt3.arrayChanged();
        
        // Set the graph titles.
        graph1.setSelected(false);
        graph1.setShowMainTitle(true);
        graph1.setShowXtext(true);
        graph1.getXAxis().setShowTitle(true);
        String xTitle = stationLabel + "  --  Julian Day  (";
        xTitle += createDateString(sourceTS, sourceTS.getMin(0));
        xTitle += " to ";
        xTitle += createDateString(sourceTS, sourceTS.getMax(0));
        xTitle += ")";
        graph1.getXAxis().setTitle(xTitle);
        graph1.getYAxis().setShowTitle(true);
        graph1.getYAxis().setTitle("Discharge, in cfs\\n ");
        graph1.setTitle("Unit Value Discharge (Middle); " + "Higher High Tidal Range (Top); " + "Lower Low Tidal Range (Bottom)");
        graph2.setSelected(false);
        graph2.setShowMainTitle(true);
        graph2.setTitle("Daily Godin Discharge (Solid); " + "Daily Mean Discharge (Dashed)");
        graph2.getYAxis().setShowTitle(true);
        graph2.getYAxis().setTitle("Discharge, in cfs\\n ");
        graph3.setSelected(false);
        graph5.setSelected(false);
        graph3.setShowMainTitle(true);
        graph3.setTitle("Difference (Daily Godin Discharge - Daily Mean Discharge)    " + graph3StatText);
        graph3.getYAxis().setShowTitle(true);
        graph3.getYAxis().setTitle("Discharge, in cfs\\n ");
        graph4.setShowMainTitle(true);
        graph4.setTitle("Percent Difference Absolute Value ((Daily Godin Discharge - Daily Mean Discharge) / Daily Godin Discharge)    " + graph4StatText);
        graph4.getYAxis().setShowTitle(true);
        graph4.getYAxis().setTitle("Percent Difference\\n ");
        graph5.setShowMainTitle(true);
        graph5.setTitle("Ratio of \"Daily Godin Discharge / Tidal Range\" " + "(above 3 --> do not filter; " + "below 1 --> filter)    " + graph5StatText);
        graph5.getYAxis().setShowTitle(true);
        graph5.getYAxis().setTitle("Godin / Tidal Range\\n ");
        page.updateLists();
        
        // Write the statistics to the console.
        System.out.println("\nRatio of \"Filtered Discharge / Tidal Range\" calculated.");
        System.out.println("If above 3, do not filter.  If below 1, filter.");
        System.out.println(graph5StatText1);
        System.out.println(graph5StatText2);
        System.out.println(graph5StatText3);
        System.out.println(graph5StatText4);
        String recString = "Undetermined";
        String conString = null;
        if (numTotal >= 300) {
            if (numBelow1 > numBetween1and3 && percentAbove3 < 3) {
                recString = "Filter";
                if (percentBelow1 > 95 && numAbove3 == 0)
                    conString = "High";
                else if (percentBelow1 > 50 && percentAbove3 < 5)
                    conString = "Medium";
                else
                    conString = "Low";
            }
            else if (numAbove3 > numBetween1and3 && percentBelow1 < 3) {
                recString = "Do not filter";
                if (percentAbove3 > 95 && numBelow1 == 0)
                    conString = "High";
                else if (percentAbove3 > 50 && percentBelow1 < 5)
                    conString = "Medium";
                else
                    conString = "Low";
            }
        }
        else {
            recString += ", fewer than 300 data points available";
        }
        System.out.println("\nRecommendation:  " + recString);
        if (conString != null)
            System.out.println("Confidence:  " + conString);
        System.out.println("\nWriting analysis graphs to PostScript file...\n");
        
        // Write the recommendation on the plot.
        TextMarkup titleMarkup = new TextMarkup();
        String pageTitle = "Tidal Filter Analysis Recommendation:  "
                + recString;
        if (conString != null)
            pageTitle += ",   Confidence:  " + conString;
        titleMarkup.setText(pageTitle);
        titleMarkup.setPsFont("Helvetica-Oblique");
        titleMarkup.setPsFontSize(10);
        titleMarkup.x = 0;
        titleMarkup.y = 1;
        titleMarkup.xPointOffset = 50;
        titleMarkup.yPointOffset = -65;
        titleMarkup.setCoordinateLevels(1, 1);
        titleMarkup.xJustification = 0;
        titleMarkup.setUserChangeable(false);
        page.add(titleMarkup);

        // Write the graphs to a PostScript file, but divert the LPR messages
        // so the user won't see them.
        PrintStream outStream = System.out;
        System.setOut(new PrintStream(new PipedOutputStream()));
        GrController.printPage(filename, page);
        System.setOut(outStream);
    }
    
//------------------------------------------------------------------------------
    /**
     * Convert a julian day with decimal to a date object.
     * @param referenceYear     A four digit year from which the day is
     *                          referenced.
     * @param decimalDay        The day of the year with the hour, minute, and
     *                          second represented by a decimal.
     * @param date              The date object to store the new date.
     * @return		A new GregorianCalendar object for the specified date.
     */
    public static String createDateString(TimeSeries series, double decimalDay) {
        
        TimeZone utcZone = TimeZone.getTimeZone("UTC");
        GregorianCalendar date = new GregorianCalendar(utcZone);
        date.clear();
        int dayOfYear = (int)decimalDay;
        double decimalTime = decimalDay - dayOfYear;
        int hour = (int)(decimalTime * 24);
        int minute = (int)((decimalTime * 24 - hour) * 60);
        int second = (int)(((decimalTime * 24 - hour) * 60 - minute) * 60);
        
        date.set(Calendar.YEAR, series.getReferenceYear());
        date.set(Calendar.DAY_OF_YEAR, (int)dayOfYear);
        date.set(Calendar.HOUR_OF_DAY, hour);
        date.set(Calendar.MINUTE, minute);
        date.set(Calendar.SECOND, second);
        
        SimpleDateFormat formatter =
            new SimpleDateFormat("MM/dd/yyyy", Locale.getDefault());
        formatter.setTimeZone(utcZone);
        StringBuffer buf = formatter.format(date.getTime(),
            new StringBuffer(), new FieldPosition(0));
        
        return buf.toString();
    }

//------------------------------------------------------------------------------
    /**
     * Zoom out to show all the data in the graphs.
     * @param page The page containing the graphs to zoom on.
     */
    public void zoomAll(PageOfGraphs page) {

       // Change the x-axis for all currently selected graphs.
        Iterator iter = page.getSelectedXZooms().iterator();
        while (iter.hasNext()) {
            Zoom zoom = (Zoom) iter.next();
            // Autoscale...
            int numColumns = 1;
            zoom.zoomAll(page.getWidth() / numColumns / 80);
        }

        // Change the y-axis for all currently selected graphs.
        iter = page.getSelectedYZooms().iterator();
        while (iter.hasNext()) {
            Zoom zoom = (Zoom) iter.next();
            // Autoscale...
            int numRows = page.getShowList().size();
            zoom.zoomAll(page.getHeight() / numRows / 30);
        }
    }

//------------------------------------------------------------------------------
    /**
     * Replace a data series with its box car average at a given interval and
     * starting at a given offset.  Output can be at a different interval than
     * the averaging interval.
     * @param series            The data series that will be modifoed.
     * @param interval          The span for determining each averaged point.
     * @param offset            The amount to offset the points from zero.
     * @param outputInterval    The spacing of the output.
     */
    public synchronized void average(TimeSeries series, double interval,
            double offset, double outputInterval) {

        int xDim = 0;

        // Use the absolute value of the supplied interval.
        interval = Math.abs(interval);
        double halfInterval = interval / 2;

        // Include a little more area (20% of the interval above and below)
        // just so rounding doesn't keep any points on the boundary from being
        // included in the interpolation.
        double min = series.getMin(xDim) - (interval / 20);
        double max = series.getMax(xDim) + (interval / 20);
        double startOffset = min - offset;
        double startIntervals = (int) (startOffset / interval);
        if (startOffset > 0) {
            startIntervals++;
        }
        double start = (interval * startIntervals) + offset;
        int size = (int) ((max - start) / outputInterval) + 1;

        int numDims = series.getNumDimensions();
        double[] newArray = new double[size * numDims];
        for (int i = 0; i < newArray.length; i += numDims) {
            // Set the coordinates.
            newArray[i + xDim] = start + ((i / numDims) * outputInterval);

            for (int j = 0; j < numDims; j++) {
                double val = 0;

                if (j != xDim) {
                    // Find the value for the new point.
                    double spanMin = newArray[i + xDim] - halfInterval;
                    double spanMax = newArray[i + xDim] + halfInterval;
                    val = series.findSpanAverage(spanMin, spanMax, -1, xDim, j);

                    // Interpolate the last dimension's value at this point.
                    newArray[i + j] = val;
                }
            }
        }

        // The old list of deleted points was probably invalidated by this
        // operation, so wipe it out.
        series.setArray(newArray);
    }

//------------------------------------------------------------------------------
    /**
     * Create a list of segments between gaps in the series.
     * @param The input series.
     * @return A list of segments.
     */
    protected List<List<Double>> createSegmentList(TimeSeries series) {

        // By default, define a gap as 1% more than the allowable gap.
        double gapPlusBuffer = (1d / 24d) * 1.01d;

        // Time differences below 1/10 millisecond won't be considered.
        double tinyInc = 1d / 24d / 60d / 60d / 10000d;

        // Create a list to hold a list of segments for each series.
        List<List<Double>> segmentList = new ArrayList<List<Double>>();

        double[] array = series.getArray();
        int numPoints = series.getCount();
        int numDims = series.getNumDimensions();

        // Initialize the lastTime variable to the first Godin point after
        // the start of the series.
        double firstTimeValue = 0;
        if (numPoints > 0) {
            firstTimeValue = array[0];
        }
        double lastTime = firstTimeValue;

        // Create a list to store segment ranges.
        List<Double> segment = new ArrayList<Double>();
        segment.add(firstTimeValue - tinyInc);

        for (int i = 1; i < numPoints; i++) {
            // Look at each time.
            double time = array[i * numDims];
            double timeDiff = time - lastTime;

            // Start a new segment if the last one was finished.
            if (segment == null) {
                segment = new ArrayList<Double>();
                segment.add(time - tinyInc);

            }

            // If enough time was skipped to be considered a gap, mark
            // the end of the last segment (beginning of gap).
            if (timeDiff > gapPlusBuffer) {
                segment.add(lastTime + tinyInc);
                segmentList.add(segment);
                segment = null;
            }

            // The next lastTime becomes the current time.
            lastTime = time;
        }

        // Mark the end of the series as the end of the last segment.
        segment.add(array[(numPoints - 1) * numDims] - tinyInc);

        // Put the last segment into the list of segments.
        segmentList.add(segment);

        return segmentList;
    }

//------------------------------------------------------------------------------
    /**
     * Create a list of curves created by breaking up the series with the
     * segment list.  The segments are shrunk on both ends by gap expansion.
     * The final curve uses the color and line pattern indexes.
     * @param segmentList       The input segment list.
     * @param series            The input series.
     * @param gapExpansion      The gap expansion.
     * @param lineColorIndex    The line color index.
     * @param linePatternIndex  The line pattern index.
     * @return A list of curves.
     */
    protected List<Curve> createCurveList(List<List<Double>> segmentList,
            TimeSeries series, double gapExpansion, int lineColorIndex,
            int linePatternIndex) {

        ArrayList<TimeSeries> segmentSeriesList = new ArrayList<TimeSeries>();
        double[] array = series.getArray();
        int numDims = 2;
        int startIndex = 0;
        int endIndex = 0;
        int lastIndex = -1;
        double tinyInc = 1d / 24d / 60d / 60d / 10000d;

        // Process each segment individually.
        for (List<Double> segment : segmentList) {
            // Determine the range of the current segment.
            double startTime = segment.get(0);
            startIndex = series.findNearestIndex(startTime + gapExpansion);
            if (array[startIndex * numDims] < startTime) {
                startIndex++;
            }
            double endTime = segment.get(1);
            endIndex = series.findNearestIndex(endTime - gapExpansion);
            if (array[endIndex * numDims] > endTime &&
                    (endIndex - 1) >= 0) {

                endIndex--;
            }

            if (startIndex <= lastIndex || startIndex >= endIndex) {
                continue;
            }

            // Determine the length of the segment.
            int size = (endIndex - startIndex + 1) * numDims;

            // Copy the segment to another array.
            double[] segmentArray = new double[size];

            int first = startIndex * numDims;
            int past = (endIndex + 1) * numDims;
            for (int i = first; i < past; i++) {
                segmentArray[i - first] = array[i];
            }

            // this segment.
            TimeSeries segmentSeries = (TimeSeries) series.createCopy(true);
            segmentSeries.setArray(segmentArray);
            segmentSeriesList.add(segmentSeries);

            lastIndex = endIndex;
        }

        List<Curve> curves = new ArrayList<Curve>();
        for (TimeSeries ts : segmentSeriesList) {
            PsCurve curve = new PsCurve(ts);
            curve.setColor(lineColorIndex);
            curve.setLinePattern(linePatternIndex);
            curves.add(curve);
        }

        return curves;
    }

//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {

        String s = "Tidal Filter Analyzer";

        return s;
    }
}

