/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-24
 * Date started:    2007-05-23
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.gr.tools;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.util.*;

//==============================================================================
/**
 */

public class TidalHighFinder extends GrObject implements
    gov.usgs.gr.tools.DataManipulator, java.awt.event.ActionListener {
    
    protected double m_span = 24.84d / 24d;
    protected double m_subSpan = 6d / 24d;
    
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    
//------------------------------------------------------------------------------
    /**
     * Creates an empty object.
     */
    public TidalHighFinder() {
        
        setDefaults("gov.usgs.sfhydro.gr.tools.TidalHighFinder");
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides add in GrObject.
     */
    public boolean add(GrObject gro) {
        
        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();
        
        if (m_children.contains(gro)) return false;
        
        boolean bAdded = false;
        
        if (gro instanceof DataSeries) {
            DataSeries ds = (DataSeries)gro;
            
            m_children.add(ds);
            bAdded = true;
        } else {
            // Try adding all of the descendants in case they hold a
            // data series.  This allows curves to be traversed and
            // have their data series added.
            Enumeration enumeration = gro.children();
            while (enumeration.hasMoreElements()) {
                if (add((GrObject)enumeration.nextElement()))
                    bAdded = true;
            }
        }
        
        return bAdded;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public GrHashList getProperties() {
        
        GrHashList props = super.getProperties();
        
        props.put("span", new GrProperty(m_span, "Span"));
        props.put("subSpan", new GrProperty(m_subSpan, "Sub Span"));
        
        return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public void setProperties(HashMap props) {
        
        super.setProperties(props);
        
        GrProperty p;
        
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;
                
                if (key.equals("span"))
                    m_span = (Double)p.value;
                if (key.equals("subSpan"))
                    m_subSpan = (Double)p.value;
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Set a single property.
     */
    public void setProperty(String key, String value) {
        
        super.setProperty(key, value);
        
        if (key.equals("span"))
            m_span = Double.parseDouble(value);
        if (key.equals("subspan"))
            m_subSpan = Double.parseDouble(value);
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
        
        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();
        
        return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {
        
        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);
        
        if (icon == null)
            icon = super.setTreeIcon();
        
        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);
        
        return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Fills gaps between selected points in each data series.
     * <p><b> Pre Conditions </b><br>
     *      The data list contains DoubleDataSeries objects.
     *		The units of the time dimension is in days.
     * @param dataList  A list of DoubleDataSeries objects that will be filtered.
     */
    public void performAction(List dataList) {
        
        // Loop through each series in the data list.
        Iterator iter = dataList.iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries)iter.next();
            if (series != null) performAction(series);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Perform the action on a data series.  Assumes the units of the time
     * dimension are in days.
     * @param series  A DoubleDataSeries to use as input.
     */
    public void performAction(DoubleDataSeries series) {
        
        synchronized (series) {
            // Loop through each series in the data list.
            int numDims = series.getNumDimensions();
            int numPoints = series.getCount();
            double[] array = series.getArray();
            List<Integer> indexList = new ArrayList<Integer>();
            
            for (int i = 0; i < numPoints; i++) {
                double tVal = array[i * numDims];
                int localHighIndex = highestInNextSubSpan(array, numDims,
                    numPoints, i);
                if (localHighIndex >= 0)
                    indexList.add(localHighIndex);
                
                i = nextHighAfterSubSpan(array, numDims, numPoints,
                    localHighIndex) - 1;
            }
            
            indexList.remove(0);
            double[] newArray = new double[indexList.size() * numDims];
            int newIndex = 0;
            for (int index : indexList) {
                for (int j = 0; j < numDims; j++) {
                    newArray[newIndex * numDims + j] =
                        array[index * numDims + j];
                }
                newIndex++;
            }
            
            series.setArray(newArray);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     */
    public int highestInNextSubSpan(double[] array, int numDims, int numPoints,
        int start) {
        
        double startVal = array[start * numDims];
        double highestVal = array[start * numDims + 1];
        int highestIndex = start;
        
        for (int i = (start + 1); i < numPoints; i++) {
            double tVal = array[i * numDims];
            double tDiff = tVal - startVal;
            if (tDiff > m_subSpan)
                return highestIndex;
            
            double yVal = array[i * numDims + 1];
            if (yVal > highestVal) {
                highestVal = yVal;
                highestIndex = i;
            }
        }
        
        return -1;
    }
    
//------------------------------------------------------------------------------
    /**
     */
    public int nextHighAfterSubSpan(double[] array, int numDims, int numPoints,
        int start) {
        
        double startVal = array[start * numDims];
        double highestVal = Double.NEGATIVE_INFINITY;
        int highestIndex = start + 1;
        
        for (int i = (start + 1); i < numPoints; i++) {
            double tVal = array[i * numDims];
            double tDiff = tVal - startVal;
            
            if (tDiff <= m_subSpan)
                continue;
            
            if (tDiff > m_span)
                return highestIndex;
            
            double yVal = array[i * numDims + 1];
            if (yVal > highestVal) {
                highestVal = yVal;
                highestIndex = i;
            }
        }
        
        return Integer.MAX_VALUE;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of menu items that execute commands on this object.
     * @return  A menu item that can be placed on a menu and send an ActionEvent
     *          back to this object when it is chosen.
     */
    public javax.swing.JMenuItem getMenuCommands() {
        
        javax.swing.JMenu subMenu = new javax.swing.JMenu(
            "Tidal High Finder");
        javax.swing.JMenuItem saveItem = new javax.swing.JMenuItem("Calculate");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
        return subMenu;
    }
    
//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {
        
        javax.swing.JMenuItem source =
            (javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Calculate")) {
            recalculate();
        }
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public void recalculate() {
        
        performAction(new ArrayList(m_children));
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {
        
        String s = "Tidal High Finder";
        
        return s;
    }
}

