/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-24
 * Date started:    2007-05-23
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.gr.tools;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.util.*;

//==============================================================================
/**
 */

public class TidalLowFinder extends gov.usgs.sfhydro.gr.tools.TidalHighFinder {
    
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public TidalLowFinder () {

        setDefaults("gov.usgs.sfhydro.gr.tools.TidalLowFinder");
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Finds the tidal low.  Assumes the data list contains DoubleDataSeries
 * objects.  The units of the time dimension is in days.
 * @param series  A DoubleDataSeries to use as input.
 */
    public void performAction (DoubleDataSeries series) {

        synchronized (series) {
            // Loop through each series in the data list.
            int numDims = series.getNumDimensions();
	    int numPoints = series.getCount();
	    double[] array = series.getArray();
	    List<Integer> indexList = new ArrayList<Integer>();
	    
	    for (int i = 0; i < numPoints; i++) {
		double tVal = array[i * numDims];
		int localLowIndex = lowestInNextSubSpan(array, numDims,
		    numPoints, i);
		if (localLowIndex >= 0) {
		    indexList.add(localLowIndex);
		
                    i = nextLowAfterSubSpan(array, numDims, numPoints,
                        localLowIndex) - 1;
                }
                else {
                    i++;
                }
	    }
	    
	    indexList.remove(0);
	    double[] newArray = new double[indexList.size() * numDims];
	    int newIndex = 0;
	    for (int index : indexList) {
		for (int j = 0; j < numDims; j++) {
		    newArray[newIndex * numDims + j] =
			array[index * numDims + j];
		}
		newIndex++;
	    }
	    
	    series.setArray(newArray);
	}
    }

//------------------------------------------------------------------------------
/**
 */
    public int lowestInNextSubSpan(double[] array, int numDims, int numPoints,
	int start) {
	
	double startVal = array[start * numDims];
	double lowestVal = array[start * numDims + 1];
	int lowestIndex = start;
	
	for (int i = (start + 1); i < numPoints; i++) {
	    double tVal = array[i * numDims];
	    double tDiff = tVal - startVal;
	    if (tDiff > m_subSpan)
		return lowestIndex;
	    
	    double yVal = array[i * numDims + 1];
	    if (yVal < lowestVal) {
		lowestVal = yVal;
		lowestIndex = i;
	    }
	}	
	
	return -1;
    }

//------------------------------------------------------------------------------
/**
 */
    public int nextLowAfterSubSpan(double[] array, int numDims, int numPoints,
	int start) {
	
	double startVal = array[start * numDims];
	double lowestVal = Double.POSITIVE_INFINITY;
	int lowestIndex = start + 1;
	
	for (int i = (start + 1); i < numPoints; i++) {
	    double tVal = array[i * numDims];
	    double tDiff = tVal - startVal;
	    
	    if (tDiff <= m_subSpan)
		continue;
	    
	    if (tDiff > m_span)
		return lowestIndex;
	    
	    double yVal = array[i * numDims + 1];
	    if (yVal < lowestVal) {
		lowestVal = yVal;
		lowestIndex = i;
	    }
	}	
	
	return Integer.MAX_VALUE;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * @return  A menu item that can be placed on a menu and send an ActionEvent
 *          back to this object when it is chosen.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"Tidal Low Finder");
        javax.swing.JMenuItem saveItem = new javax.swing.JMenuItem("Calculate");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
    	return subMenu;
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Tidal Low Finder";

        return s;
 	}
}

