/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    1999-08-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.gr.tools;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.util.*;

//==============================================================================
/**
 *
 */

public class UV2DirSpeed extends GrObject implements
    gov.usgs.gr.tools.DataManipulator, java.awt.event.ActionListener {

    int m_uDim = 1;
    int m_vDim = 2;
    boolean m_b2DirSpeed = true;

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public UV2DirSpeed () {

        setDefaults("gov.usgs.sfhydro.gr.tools.UV2DirSpeed");
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

		boolean bAdded = false;

        if (gro instanceof DoubleDataSeries) {
        	DoubleDataSeries ds = (DoubleDataSeries)gro;

            m_children.add(ds);
            bAdded = true;
        }
        else {
            // Try adding all of the descendants in case they hold a
            // data series.  This allows curves to be traversed and
            // have their data series added.
            Enumeration enumeration = gro.children();
            while (enumeration.hasMoreElements()) {
                if (add((GrObject)enumeration.nextElement()))
                    bAdded = true;
            }
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("uDim", new GrProperty(m_uDim, "U Dimension"));
        props.put("vDim", new GrProperty(m_vDim, "V Dimension"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("uDim"))
            		m_uDim = (Integer)p.value;
            	if (key.equals("vDim"))
            		m_vDim = (Integer)p.value;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("uDim"))
            m_uDim = Integer.parseInt(value);
        if (key.equals("vDim"))
            m_vDim = Integer.parseInt(value);
    }

//------------------------------------------------------------------------------
/**
 * Fills gaps between selected points in each data series.
 * <p><b> Pre Conditions </b><br>
 *      The data list contains DoubleDataSeries objects.
 *		The units of the time dimension is in days.
 * @param dataList  A list of DoubleDataSeries objects that will be filtered.
 */
    public void performAction (List dataList) {

    	// Loop through each series in the data list.
    	Iterator iter = dataList.iterator();
        while (iter.hasNext()) {
        	DoubleDataSeries series = (DoubleDataSeries)iter.next();
            if (series != null) performAction(series);
        }
    }

//------------------------------------------------------------------------------
/**
 * Fills gaps between selected points in each data series.  Assumes the data
 * list contains DoubleDataSeries objects.  The units of the time dimension is
 * in days.
 * @param series  A DoubleDataSeries to use as input.
 */
    public void performAction (DoubleDataSeries series) {

    	double[] array = series.getArray();
        int numPoints = series.getNumPoints();
        int numDims = series.getNumDimensions();

        if (m_b2DirSpeed) {
            for (int i = 0; i < numPoints; i++) {
                double u = array[i * numDims + m_uDim];
                double v = array[i * numDims + m_vDim];
                double theta = Math.atan2(v, u);
                theta *= (180 / Math.PI);
                double len = Math.sqrt((u * u) + (v * v));
                array[i * numDims + m_uDim] = theta;
                array[i * numDims + m_vDim] = len;
            }
        }
        else {
            for (int i = 0; i < numPoints; i++) {
                double theta = array[i * numDims + m_uDim] * Math.PI / 180;
                double len = array[i * numDims + m_vDim];
                double x = Math.cos(theta) * len;
                double y = Math.sin(theta) * len;
                array[i * numDims + m_uDim] = x;
                array[i * numDims + m_vDim] = y;
            }
        }

        series.arrayChanged();
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * @return  A menu item that can be placed on a menu and send an ActionEvent
 *          back to this object when it is chosen.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"U-V to Dir-Speed Converter");
        javax.swing.JMenuItem saveItem =
            new javax.swing.JMenuItem("Convert to U-V");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
        saveItem = new javax.swing.JMenuItem("Convert to Dir-Speed");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
    	return subMenu;
    }

//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Convert to U-V")) {
            m_b2DirSpeed = false;
        	recalculate();
        }
        if (srcText.equals("Convert to Dir-Speed")) {
            m_b2DirSpeed = true;
        	recalculate();
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void recalculate () {

        performAction(new ArrayList(m_children));
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "U-V to Dir-Speed Converter";

        return s;
 	}
}

