/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2001-02-08
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.gr.tools;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.gl.*;
import java.util.*;

//==============================================================================
/**
 * Calculates the cross sectional path where a vertical surface intersects
 * a surfaces of discontinuous squares.  When a surface data set and a cross
 * secion line XyDataSet are added as children, an output X-Z data set is
 * calculated and added as a third child.  The output series can be viewed
 * if it is copied and the original pasted into a graph.
 */

public class VectAnimCalculator extends GrObject implements
	java.awt.event.ActionListener {

    /** The channel in the input to use as direction. */
    protected int m_dirChannel = 1;
    /** the channel in the input to use as speed. */
    protected int m_speedChannel = 2;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/VectorAnimCalculator.gif";

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public VectAnimCalculator () {

        setDefaults("gov.usgs.sfhydro.gr.tools.VectAnimCalculator");
    }


//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.  Only allows DataDimensions to be added.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

        return super.add(gro);
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("dirChan", new GrProperty(m_dirChannel, "Direction Channel"));
        props.put("speedChan", new GrProperty(m_speedChannel, "Speed Channel"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

            	if (key.equals("dirChan"))
            		m_dirChannel = (Integer)p.value;
            	if (key.equals("speedChan"))
            		m_speedChannel = (Integer)p.value;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("dirChan"))
            m_dirChannel = Integer.parseInt(value);
        if (key.equals("speedChan"))
            m_speedChannel = Integer.parseInt(value);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void recalculate () {

        int newNumDims = 5;
        DoubleDataSeries fieldSeries = new DoubleDataSeries(newNumDims);
        fieldSeries.getDimension(0).setDimensionLocked(true);
        fieldSeries.getDimension(1).setDimensionLocked(false);
        fieldSeries.getDimension(2).setDimensionLocked(false);
        fieldSeries.getDimension(3).setDimensionLocked(true);
        fieldSeries.getDimension(4).setDimensionLocked(true);

        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof DataSeries) {
                DoubleDataSeries oldSeries = (DoubleDataSeries)gro;

                // Check for a location object associated with the DataSeries.
                double x = 0;
                double y = 0;
                Enumeration oldEnum = oldSeries.children();
                while (oldEnum.hasMoreElements()) {
                    GrObject oldChild = (GrObject)oldEnum.nextElement();
                    if (oldChild instanceof Location) {
                        Location location = (Location)oldChild;
                        GrHashList props = location.getProperties();
                        GrProperty p = (GrProperty)props.get("x");
                        x = (Double)p.value;
                        x -= 600000;
                        p = (GrProperty)props.get("y");
                        y = (Double)p.value;
                        y -= 4200000;
                        break;
                    }
                }

                int numPoints = oldSeries.getCount();
                int oldNumDims = oldSeries.getNumDimensions();
                double[] oldArray = oldSeries.getArray();
                double[] fieldArray = new double[numPoints * newNumDims];
                for (int i = 0; i < numPoints; i++) {
                    fieldArray[i * newNumDims] = oldArray[i * oldNumDims];
                    fieldArray[i * newNumDims + 1] = (double)x;
                    fieldArray[i * newNumDims + 2] = (double)y;
                    fieldArray[i * newNumDims + 3] =
                        -(oldArray[i * oldNumDims + m_dirChannel] - 90);
                    fieldArray[i * newNumDims + 4] =
                        oldArray[i * oldNumDims + m_speedChannel];
                }

                fieldSeries.setArray(fieldArray);
                fieldSeries.setTitle(oldSeries.getTitle());
            }
        }

        GlVectorAnimation glv = null;
        Iterator iter2 = m_children.iterator();
        while (iter2.hasNext()) {
            GrObject gro2 = (GrObject)iter2.next();
            if (gro2 instanceof GlVectorAnimation) {
                glv = (GlVectorAnimation)gro2;
                break;
            }
        }
        if (glv == null) {
            glv = new GlVectorAnimation();
            add(glv);
        }
        glv.add(fieldSeries);
        glv.setProperty("title", fieldSeries.getTitle());

        TimeMarker timeMarker = null;
        iter2 = m_children.iterator();
        while (iter2.hasNext()) {
            GrObject gro2 = (GrObject)iter2.next();
            if (gro2 instanceof TimeMarker) {
                timeMarker = (TimeMarker)gro2;
                break;
            }
        }
        if (timeMarker == null) {
            timeMarker = new TimeMarker();
            add(timeMarker);
        }
        timeMarker.add(fieldSeries);
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

//    	if (o == m_surfaceDs || o == m_inputCsDs)
//        	recalculate();
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * @return  A menu item that can be placed on a menu and send an ActionEvent
 *          back to this object when it is chosen.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"Vector Field Animation Calculator");
        javax.swing.JMenuItem saveItem = new javax.swing.JMenuItem("Calculate");
        saveItem.addActionListener(this);
        subMenu.add(saveItem);
    	return subMenu;
    }

//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Calculate")) {
        	recalculate();
        }
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Vector Field Animation Calculator";

        return s;
 	}
}

