/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-03
 * Date started:    2004-01-12
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.model1d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import gov.usgs.sfhydro.model1d.gl.*;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

//==============================================================================
/**
 * An instance of this class can be placed as a child of a PageOfGraphs object
 * to cause an editing panel to be shown on the toolbar.  It looks for curves
 * in a graph to use as definitions for paths.  The curves specify cross
 * sections that will eventually be drawn as quad strips by a GlPathAnimation
 * object.  They should be specified from right to left going down the channel
 * from start to finish.  The editor panel supplied by this class allows the
 * user to work with the curves and specify start and end node info.  An
 * instance of this class should contain four child objects.
 * 1 - A graph containing curves which define paths between nodes.  The graph
 *     should also contain a GlPathAnimation object.
 * 2 - An IntDataSeries with two dimensions specifying remapping information
 *     (aliases) for node numbers.  Dimension 0 contains user-specified node
 *     numbers and dimension 1 contains the array index for each node number.
 * 3 - An IntDataSeries with three dimensions for output.  It should be
 *     shared with a GlPathAnimation object.  Dimension 0 contains the start
 *     node number for each path.  Dimension 1 contains the end node number for
 *     each path.  Dimension 2 contains the number of points in each path.
 * 4 - A DataSeries with three dimensions for output.  It should be shared with
 *     a GlPathAnimation object.  Dimension 0 contains the X location of each
 *     point on each path.  Dimension 1 contains the Y location of each point.
 *     Dimension 2 contains the distance of the point's cross section along
 *     the centerline from the start node to the end node.  This value is
 *     normalized between 0 (start) and 1 (end).
 */
public class PathEditor extends gov.usgs.gr.GrObject
    implements gov.usgs.gr.visual.gui.GuiPanelProvider {

    /** An instance of the controller associated with this object. */
    GrController m_grCon = null;
    /** An instance of the page associated with this object. */
    PageOfGraphs m_page = null;
    /** The panel created by this class for displaying on a toolbar. */
    protected JPanel m_panel = new JPanel();
    /** The graph containing the curves and path animation. */
    protected Graph m_graph = null;
    /**
     * An ordered hash map to store and look up the 2D DataSeries from the
     * path definition curves.  Stores and array containing start node and end
     * node for each series.
     */
    protected GrHashList m_xySeriesHashList = new GrHashList();
    /** The series containing the X-Y-Distance output data. */
    protected DoubleDataSeries m_xydOutputSeries = null;
    /** The series containing the Start Node - End Node - Num Points data. */
    protected IntDataSeries m_intOutputSeries = null;
    /**
     * The series that holds the lookup array for node numbers.  The series has
     * two dimensions.  The first is the node number entered by the user.  The
     * second is the alias, or index into the scalar data series used by the
     * path animation object.  For example, node 79 may be mapped to index 1
     * if it is the first node written in the scalar data series.  The alias
     * series may contain zero entries or a partial list of node aliases.
     * Nodes without aliases will have their entered number written to the
     * IntOutputSeries instead of a remapped number.  Performance will improve
     * if the first dimension is sorted, but it is not required.
     */
    protected IntDataSeries m_nodeAliasSeries = null;
    /** The number of the currently selected series, negative if none. */
    protected int m_currentSeriesNum = -1;
    /** Whether the graph has changed since the last recalc. */
    protected boolean m_bNeedRecalc = true;
    /** The text field containing the title of the selected path. */
    protected JTextField m_titleTextField = new JTextField();
    /** The text field containing the start node of the selected path. */
    protected JTextField m_startNodeTextField = new JTextField();
    /** The text field containing the end node of the selected path. */
    protected JTextField m_endNodeTextField = new JTextField();
    /**
     * The checkbox for whether cross sections should be drawn by the
     * path animation object.
     */
    protected JCheckBox m_crossSectionCB = new JCheckBox("Show Cross Sections",
        null, true);
    /**
     * The checkbox for whether the centerline should be drawn by the
     * path animation object.
     */
    protected JCheckBox m_centerLineCB = new JCheckBox("Show Center Lines",
        null, true);
    /**
     * The checkbox for whether the filled quad strip should be drawn by the
     * path animation object.
     */
    protected JCheckBox m_pathStripCB = new JCheckBox("Show Path Strip",
        null, true);
    /** The checkbox for whether the curves should be drawn in the graph. */
    protected JCheckBox m_curvesCB = new JCheckBox("Show Curves", null, true);
    /** The JSspinner model for the cycling between selected curves. */
    protected AbstractSpinnerModel m_spinnerModel = null;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/PathEditor.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a default object.
 */
    public PathEditor() {

        // Create the editing panel once so it can be passed out later
        // as needed.
        createPanel();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Adds an object as a child, but only if it is an instance of Graph, an
 * IntDataSeries with two or three dimensions, or a DoubleDataSeries with three
 * dimensions.  Removes the old child, if a matching one can be found, and
 * replaces it with the new one.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        if (m_children.contains(gro)) return false;

        boolean bAdded = false;

        if (gro instanceof Graph) {
            // Look for the old one and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof Graph) iter.remove();
            }

            // Add it and store it in the member variable.
            super.add(gro);
            m_graph = (Graph)gro;
            getViewSettings();

            // As soon as the int output series, node alias series, and graph
            // are all present, update the series hash list.
            if (m_intOutputSeries != null && m_nodeAliasSeries != null) {
                updateSeriesHashList();
                m_spinnerModel.setValue(1);
            }

            return true;
        }

        if (gro instanceof IntDataSeries) {
            IntDataSeries ids = (IntDataSeries)gro;
            if (ids.getNumDimensions() == 2) {
                // Look for the old one and remove it.
                Iterator iter = m_children.iterator();
                while (iter.hasNext()) {
                    Object o = iter.next();
                    if (o instanceof IntDataSeries) {
                        if (((IntDataSeries)o).getNumDimensions() == 2)
                            iter.remove();
                    }
                }

                // Add it and store it in the member variable.
                super.add(gro);
                m_nodeAliasSeries = (IntDataSeries)gro;

                // As soon as the int output series, node alias series, and graph
                // are all present, update the series hash list.
                if (m_intOutputSeries != null && m_graph != null) {
                    updateSeriesHashList();
                    m_spinnerModel.setValue(1);
                }

                return true;
            }
            if (ids.getNumDimensions() == 3) {
                // Look for the old one and remove it.
                Iterator iter = m_children.iterator();
                while (iter.hasNext()) {
                    Object o = iter.next();
                    if (o instanceof IntDataSeries) {
                        if (((IntDataSeries)o).getNumDimensions() == 3)
                            iter.remove();
                    }
                }

                // Add it and store it in the member variable.
                super.add(gro);
                m_intOutputSeries = (IntDataSeries)gro;

                // As soon as the int output series, node alias series, and graph
                // are all present, update the series hash list.
                if (m_graph != null && m_nodeAliasSeries != null) {
                    updateSeriesHashList();
                    m_spinnerModel.setValue(1);
                }

                return true;
            }
        }

        if (gro instanceof DoubleDataSeries) {
            DoubleDataSeries series = (DoubleDataSeries)gro;
            if (series.getNumDimensions() == 3) {
                // Look for the old one and remove it.
                Iterator iter = m_children.iterator();
                while (iter.hasNext()) {
                    Object o = iter.next();
                    if (o instanceof DoubleDataSeries) iter.remove();
                }

                // Add it and store it in the member variable.
                super.add(gro);
                m_xydOutputSeries = (DoubleDataSeries)gro;
                return true;
            }
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Overrides GrObject.remove() so member variable references to children can be
 * made null when they are removed.
 * @param i     The index of the object to remove.
 * @return      The object that was removed.
 */
    public GrObject remove(int i) {

        if (m_children.get(i) instanceof GrObjectContainer)
            return null;

        // Let the super class handle removal.
        GrObject child = super.remove(i);

        // Make the reference null.
        if (child == m_graph) m_graph = null;
        if (child == m_intOutputSeries) m_intOutputSeries = null;
        if (child == m_xydOutputSeries) m_xydOutputSeries = null;

        return child;
    }

//------------------------------------------------------------------------------
/**
 * Overrides GrObject.remove() so member variable references to children can be
 * made null when they are removed.
 * @param gro   The object to be removed.
 * @return      Whether the object was removed.
 */
    public boolean remove(GrObject gro) {

        // Make the reference null.
        if (gro == m_graph) m_graph = null;
        if (gro == m_intOutputSeries) m_intOutputSeries = null;
        if (gro == m_xydOutputSeries) m_xydOutputSeries = null;

        return super.remove(gro);
    }

//------------------------------------------------------------------------------
/**
 * Get the view settings of the graphical objects inside the graph and set
 * the editor panel's check boxes to reflect those settings.
 */
    public void getViewSettings() {

        // Don't proceed without a graph. */
        if (m_graph == null)
            return;

        // Find out if the first curve in the graph is showing.  That will
        // be the value shown in the check box for all the curves.
        Curve curve = (Curve)m_graph.firstChildInstanceOf(
            "gov.usgs.gr.visual.Curve", false);
        if (curve != null) {
            m_curvesCB.setSelected(curve.getShow());
        }

        // Look at the path animation object and set the checkboxes to reflect
        // its view properties.
        GlPathAnimation pathAnimation = (GlPathAnimation)m_graph.
            firstChildInstanceOf(
            "gov.usgs.sfhydro.model1d.gl.GlPathAnimation", false);
        if (pathAnimation != null) {
            GrHashList props = pathAnimation.getProperties();
            GrProperty p = (GrProperty)props.get("drawCrossSections");
            m_crossSectionCB.setSelected((Boolean)p.value);
            p = (GrProperty)props.get("drawCenterLines");
            m_centerLineCB.setSelected((Boolean)p.value);
            p = (GrProperty)props.get("drawPathStrips");
            m_pathStripCB.setSelected((Boolean)p.value);
        }
    }

//------------------------------------------------------------------------------
/**
 * Set the view settings of the graphical objects inside the graph to match
 * one of the editor panel's check boxes.
 * @param cb    The checkbox that changed.
 */
    public void setViewSettings(JCheckBox cb) {

        // Don't proceed without a graph. */
        if (m_graph == null)
            return;

        // Find all of the curves in the graph and set their Show property
        // to match the checkbox.
        if (cb == m_curvesCB) {
            List curveList = m_graph.findChildInstancesOf(
                "gov.usgs.gr.visual.Curve", false);
            Iterator iter = curveList.iterator();
            while (iter.hasNext()) {
                Curve curve = (Curve)iter.next();
                DoubleDataSeries series = curve.getSeries();
                series.deselectAll();
                curve.setShow(m_curvesCB.isSelected());
            }
        }

        // Find the path animation in the graph and set the property that
        // corresponds to the checkbox to match the checkbox value.
        if (cb == m_crossSectionCB || cb == m_centerLineCB ||
            cb == m_pathStripCB) {
            GlPathAnimation pathAnimation = (GlPathAnimation)m_graph.
                firstChildInstanceOf(
                "gov.usgs.sfhydro.model1d.gl.GlPathAnimation", false);
            if (pathAnimation != null) {
                if (cb == m_crossSectionCB)
                pathAnimation.setProperty("drawCrossSections",
                    "" + m_crossSectionCB.isSelected());
                if (cb == m_centerLineCB)
                pathAnimation.setProperty("drawCenterLines",
                    "" + m_centerLineCB.isSelected());
                if (cb == m_pathStripCB)
                pathAnimation.setProperty("drawPathStrips",
                    "" + m_pathStripCB.isSelected());
            }
        }

        // Cause the page to redraw.
        if (m_grCon != null)
            m_grCon.triggerRepaint();
    }

//------------------------------------------------------------------------------
/**
 * Create the editor panel that will be displayed on the tool bar.
 */
    public void createPanel() {

        // Create an instance of the spinner model.
        m_spinnerModel = new PathNumSpinnerModel();

        // Define a change listener for the view options check boxes.
        javax.swing.event.ChangeListener cbCL =
            new javax.swing.event.ChangeListener() {

            public void stateChanged(javax.swing.event.ChangeEvent e) {
                // Let the setViewSettings method handle the work.
                setViewSettings((JCheckBox)e.getSource());
            }
        };

        // Define a method to handle the Add button being clicked.
        java.awt.event.ActionListener addAL =
            new java.awt.event.ActionListener() {

            public void actionPerformed(ActionEvent e) {
                // Call the addPoints method on each series in the graph that
                // has some points selected.
                List ddsSeriesList = m_graph.findChildInstancesOf(
                    "gov.usgs.gr.data.DoubleDataSeries", true);
		List xySeriesList = new ArrayList();
		Iterator ddsIter = ddsSeriesList.iterator();
		while (ddsIter.hasNext()) {
		    DoubleDataSeries series = (DoubleDataSeries)ddsIter.next();
		    if (DoubleDataSeries.countDims(series, 2))
			xySeriesList.add(series);
		}
                Iterator iter = xySeriesList.iterator();
                while (iter.hasNext()) {
                    DoubleDataSeries series = (DoubleDataSeries)iter.next();
                    if (DoubleDataSeries.countDims(series, 2) &&
			series.getNumSelected() > 0) {
			
                        addPoints(series);
		    }
                }

                // Request a redraw.
                if (m_grCon != null)
                    m_grCon.triggerRepaint();
            }
        };

        // Define a method to handle the Snap button being clicked.
        java.awt.event.ActionListener snapAL =
            new java.awt.event.ActionListener() {

            public void actionPerformed(ActionEvent e) {
                List xySeriesList = m_graph.findChildInstancesOf(
                    "gov.usgs.gr.data.DoubleDataSeries", true);
                boolean bInitialized = false;
                double xVal = 0;
                double yVal = 0;

                // Look at each XY series that has a point selected.
                Iterator iter = xySeriesList.iterator();
                while (iter.hasNext()) {
                    DoubleDataSeries xySeries = (DoubleDataSeries)iter.next();
                    if (DoubleDataSeries.countDims(xySeries, 2) &&
			xySeries.getNumSelected() > 0) {
			
                        int[] selectedIndexes = xySeries.getSelectedIndexes();
                        int index = selectedIndexes[0];
                        if (bInitialized) {
                            // Set the selected points to the same location
                            // as that of the first series in the list.
                            xySeries.modifySelectedPoints(0, 0, xVal);
                            xySeries.modifySelectedPoints(0, 1,yVal);
                        }
                        else {
                            // Store the point location of the first series
                            // encountered.
                            double[] array = xySeries.getArray();
                            xVal = array[index * 2];
                            yVal = array[index * 2 + 1];
                            bInitialized = true;
                        }
                    }
                }

                // Request a redraw.
                if (m_grCon != null)
                    m_grCon.triggerRepaint();
            }
        };

        // Set up the panel's border and layout.
        m_panel.setBorder(BorderFactory.createTitledBorder("Path"));
        m_panel.setLayout(new GridLayout(10, 1));
//        m_panel.setLayout(new BoxLayout(m_panel, BoxLayout.Y_AXIS));

        // Create a panel to show the path num spinner.
        JPanel pathNumPanel = new JPanel();
        pathNumPanel.setLayout(new BoxLayout(pathNumPanel, BoxLayout.X_AXIS));
        JLabel pathNumLabel = new JLabel("Path Num");
        pathNumPanel.add(pathNumLabel);
        JSpinner pathNumSpinner = new JSpinner(m_spinnerModel);
        m_spinnerModel.setValue(1);
        pathNumLabel.setLabelFor(pathNumSpinner);
        pathNumPanel.add(pathNumSpinner);
        m_panel.add(pathNumPanel);

        // Create a panel to show the path title.
        JPanel titlePanel = new JPanel();
        titlePanel.setLayout(new BoxLayout(titlePanel,
            BoxLayout.X_AXIS));
        JLabel titleLabel = new JLabel("Title");
        titlePanel.add(titleLabel);
        titleLabel.setLabelFor(m_titleTextField);
        titlePanel.add(m_titleTextField);
        m_panel.add(titlePanel);

        // Create a panel to show the path's start node number.
        JPanel startNodePanel = new JPanel();
        startNodePanel.setLayout(new BoxLayout(startNodePanel,
            BoxLayout.X_AXIS));
        JLabel startNodeLabel = new JLabel("Start Node");
        startNodePanel.add(startNodeLabel);
        startNodeLabel.setLabelFor(m_startNodeTextField);
        startNodePanel.add(m_startNodeTextField);
        m_panel.add(startNodePanel);

        // Create a panel to show the path's end node number.
        JPanel endNodePanel = new JPanel();
        endNodePanel.setLayout(new BoxLayout(endNodePanel,
            BoxLayout.X_AXIS));
        JLabel endNodeLabel = new JLabel("End Node");
        endNodePanel.add(endNodeLabel);
        endNodeLabel.setLabelFor(m_endNodeTextField);
        endNodePanel.add(m_endNodeTextField);
        m_panel.add(endNodePanel);

        // Create check boxes to turn view settings on and off.
        // Add them to the main panel below the other components.
        m_crossSectionCB.addChangeListener(cbCL);
        m_centerLineCB.addChangeListener(cbCL);
        m_pathStripCB.addChangeListener(cbCL);
        m_curvesCB.addChangeListener(cbCL);
        m_crossSectionCB.setMargin(new Insets(0, 0, 0, 0));
        m_centerLineCB.setMargin(new Insets(0, 0, 0, 0));
        m_pathStripCB.setMargin(new Insets(0, 0, 0, 0));
        m_curvesCB.setMargin(new Insets(0, 0, 0, 0));
        m_panel.add(m_crossSectionCB);
        m_panel.add(m_centerLineCB);
        m_panel.add(m_pathStripCB);
        m_panel.add(m_curvesCB);

        // Create an "Add Points" button.
        // Add it to the main panel below the other components.
        JButton addButton = new JButton("Add Points");
        addButton.addActionListener(addAL);
        addButton.setActionCommand("Add");
        addButton.setToolTipText("Add");
        byte[] byteArray = {0};
        addButton.setIcon(new ImageIcon(byteArray));
        addButton.setBorderPainted(true);
        addButton.setMargin(new Insets(0, 0, 0, 0));
        m_panel.add(addButton);

        // Create a "Snap Together" button.
        // Add it to the main panel below the other components.
        JButton snapButton = new JButton("Snap Together");
        snapButton.addActionListener(snapAL);
        snapButton.setActionCommand("Snap");
        snapButton.setToolTipText("Snap Selected Points Together");
        snapButton.setIcon(new ImageIcon(byteArray));
        snapButton.setBorderPainted(true);
        snapButton.setMargin(new Insets(0, 0, 0, 0));
        m_panel.add(snapButton);

        m_panel.setMaximumSize(new java.awt.Dimension(400, 400));

    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The editor panel created by this class at initialization time.
 */
    public JPanel getPanel() {

        return m_panel;
    }

//------------------------------------------------------------------------------
/**
 * Tells this interface what the current controller and page are.
 * @param grCon   The GrController to use.
 * @param page    The PageOfGraphs to use.
 */
    public void setControllerAndPage(GrController grCon, PageOfGraphs page) {

        m_grCon = grCon;
        m_page = page;
    }

//------------------------------------------------------------------------------
/**
 * Look through the graph and its children for 2D DataSeries objects.
 * Synchronize the hash list so it is in the same order and has the same
 * series as the graph.  If a series moved in the graph, find it and move it
 * in the hash list so its information is not lost. The hash list stores a
 * link to the series and the start and end nodes for the path.
 * @return  Whether the ordering of the series in the graph changed.
 */
    public boolean updateSeriesHashList() {

        // Do not proceed if there is no graph.
        if (m_graph == null || m_nodeAliasSeries == null)
            return true;

        // Get a list of DoubleDataSeries objects within the graph.
        List ddsSeriesList = m_graph.findChildInstancesOf(
            "gov.usgs.gr.data.DoubleDataSeries", true);
	List xySeriesList = new ArrayList();
	Iterator ddsIter = ddsSeriesList.iterator();
	while (ddsIter.hasNext()) {
	    DoubleDataSeries series = (DoubleDataSeries)ddsIter.next();
	    if (DoubleDataSeries.countDims(series, 2))
		xySeriesList.add(series);
	}

        // If the Integer output series for Start Node, End Node, and Num Points
        // does not exist, create it.
        if (m_intOutputSeries == null)
            add(new IntDataSeries(3));

        // Get the data arrays and related info.
        int numIntOutputPoints = m_intOutputSeries.getCount();
        int numIntDims = m_intOutputSeries.getNumDimensions();
        int[] intOutputArray = m_intOutputSeries.getArray();
        int[] nodeAliasArray = m_nodeAliasSeries.getArray();

        // Make sure the int output series has the same number of points as
        // there are 2D DataSeries within the graph.
        int numPaths = xySeriesList.size();

        // For each index in the list of series in the graph, compare the series
        // in the graph list to the series at the same index in the hash list.
        // If they don't match, add, remove, or shuffle as necessary.
        boolean bChanged = false;
        int hashIndex = 0;
        int graphIndex = 0;
        while (graphIndex < xySeriesList.size()) {
            DoubleDataSeries graphSeries =
                (DoubleDataSeries)xySeriesList.get(graphIndex);
            if (hashIndex < m_xySeriesHashList.size()) {
                // Neither the hash list nor graph list has been exhausted...
                DoubleDataSeries hashSeries =
                    (DoubleDataSeries)m_xySeriesHashList.getKey(hashIndex);
                Object[] objArray =
                    (Object[])m_xySeriesHashList.get(graphSeries);
                if (graphSeries != hashSeries) {
                    int firstIndex = xySeriesList.indexOf(graphSeries);
                    int lastIndex = xySeriesList.lastIndexOf(graphSeries);
                    if (firstIndex != lastIndex) {
                        xySeriesList.remove(lastIndex);
                        System.err.println("Duplicate series at index "
                            + firstIndex + " and " + lastIndex + ": "
                            + graphSeries.getTitle());
                        continue;
                    }

                    // The graph series and hash series don't match...
                    if (objArray == null) {
                        // The graph series was not found elsewhere in the hash
                        // list, so add it as a new entry in the hash list.
                        objArray = new Object[2];
                        objArray[0] = 1;
                        objArray[1] = 1;
                        m_xySeriesHashList.add(graphIndex, graphSeries,
                            objArray);

                        graphIndex++;
                        hashIndex++;
                    }
                    else {
                        // The current graph series was found later in the
                        // hash list, so the current hash series is out of
                        // place.  Remove the hash series from the hash list
                        // and insert it at the end of the hash list.
                        objArray =
                            (Object[])m_xySeriesHashList.remove(hashSeries);
                        m_xySeriesHashList.add(m_xySeriesHashList.size(),
                            hashSeries, objArray);

                        // Cycle and compare the same indexes again with
                        // the new series at that position in the hash list.
                    }

                    // Reordering has occurred, so the old int output series
                    // should be ignored from here on.
                    bChanged = true;
                }
                else {
                    // The series match, so proceed to the next index.
                    graphIndex++;
                    hashIndex++;
                }
            }
            else {
                int firstIndex = xySeriesList.indexOf(graphSeries);
                int lastIndex = xySeriesList.lastIndexOf(graphSeries);
                if (firstIndex != lastIndex) {
                    xySeriesList.remove(lastIndex);
                    System.err.println("Duplicate series at index "
                        + firstIndex + " and " + lastIndex + ": "
                        + graphSeries.getTitle());
                    continue;
                }
                // The hash list has been exhausted...
                int startNodeNum = -1;
                int endNodeNum = -1;
                if (!bChanged) {
                    // No reordering happened for the previous hash list
                    // elements, so get the start and end node nums from the
                    // int output series, if the index is in range.
                    // This is used when recreating this object from an XML
                    // representation.
                    if (numIntOutputPoints > graphIndex) {
                        startNodeNum = intOutputArray[graphIndex * numIntDims];
                        endNodeNum =
                            intOutputArray[graphIndex * numIntDims + 1];

                        // If there is an entry in the node alias array
                        // for these node numbers, translate them back to
                        // the original values.
                        int index = m_nodeAliasSeries.search(startNodeNum, 1);
                        if (index >= 0 &&
                            nodeAliasArray[index * 2 + 1] == startNodeNum) {

                            startNodeNum = nodeAliasArray[index * 2];
                        }
                        index = m_nodeAliasSeries.search(endNodeNum, 1);
                        if (index >= 0 &&
                            nodeAliasArray[index * 2 + 1] == endNodeNum) {

                            endNodeNum = nodeAliasArray[index * 2];
                        }
                    }
                }
                // Add the graph series as a new entry in the hash list.
                Object[] objArray = new Object[2];
                objArray[0] = startNodeNum;
                objArray[1] = endNodeNum;
//                if (graphIndex == 343) {
//                    System.out.println("graphIndex == 343");
//                }
//                try {
                m_xySeriesHashList.add(graphIndex, graphSeries, objArray);
//                }
//                catch (Exception e) {
//                    e.printStackTrace();
//                }

//                graphIndex = m_xySeriesHashList.size();
                graphIndex++;
                hashIndex++;
            }
        }

        // Remove all extra elements at the end of the hash list.
        while (m_xySeriesHashList.size() > xySeriesList.size()) {
            Object lastKey = m_xySeriesHashList.getKey(
                m_xySeriesHashList.size() - 1);
            m_xySeriesHashList.remove(lastKey);
        }

        return bChanged;
    }

//------------------------------------------------------------------------------
/**
 * Look through the graph and its children for 2D DataSeries objects.
 * @return  Whether the ordering of the series in the graph changed.
 */
    public boolean calculate() {

        // Do not proceed if there is no graph.
        if (m_graph == null || m_nodeAliasSeries == null)
            return true;

        // Get the series list up to date.
        boolean bChanged = updateSeriesHashList();

        // Get arrays and related info.
        int numIntOutputPoints = m_intOutputSeries.getCount();
        int numIntDims = m_intOutputSeries.getNumDimensions();
        int[] intOutputArray = m_intOutputSeries.getArray();
        if (numIntOutputPoints != m_xySeriesHashList.size())
            intOutputArray = new int[m_xySeriesHashList.size() * numIntDims];

        // Get the lookup array of node aliases.
        int[] nodeAliasArray = m_nodeAliasSeries.getArray();

        // Put the latest values of start node num, end node num, and series
        // count for each path into the IntOutputSeries.
        for (int i = 0; i < m_xySeriesHashList.size(); i++) {
            Object[] objArray = (Object[])m_xySeriesHashList.getValue(i);

            // Node nums are stored in the series hash list.
            int startNodeNum = (Integer)objArray[0];
            int endNodeNum = (Integer)objArray[1];

            // Check for node aliases.
            int index = m_nodeAliasSeries.search(startNodeNum, 0);
            if (index >= 0 &&
                nodeAliasArray[index * 2] == startNodeNum) {

                // Remap the node number.
                startNodeNum = nodeAliasArray[index * 2 + 1];
            }

            index = m_nodeAliasSeries.search(endNodeNum, 0);
            if (index >= 0 &&
                nodeAliasArray[index * 2] == endNodeNum) {

                // Remap the node number.
                endNodeNum = nodeAliasArray[index * 2 + 1];
            }

            // Write the data into the int output array.
            intOutputArray[i * numIntDims] = startNodeNum;
            intOutputArray[i * numIntDims + 1] = endNodeNum;

            // Get the title and length from the series object itself.
            DoubleDataSeries xySeries =
		(DoubleDataSeries)m_xySeriesHashList.getKey(i);
            intOutputArray[i * numIntDims + 2] = xySeries.getCount();
        }

        // Tell the int output series that its data has changed.
        m_intOutputSeries.setArray(intOutputArray);

        // Get a list of 2D DataSeries objects within the graph.
        List ddsSeriesList = m_graph.findChildInstancesOf(
            "gov.usgs.gr.data.DoubleDataSeries", true);
	List xySeriesList = new ArrayList();
	Iterator ddsIter = ddsSeriesList.iterator();
	while (ddsIter.hasNext()) {
	    DoubleDataSeries series = (DoubleDataSeries)ddsIter.next();
	    if (DoubleDataSeries.countDims(series, 2))
		xySeriesList.add(series);
	}

        // Make sure the int output series has the same number of points as
        // there are 2D DataSeries within the graph.
        int numPaths = xySeriesList.size();

        // Create an array to hold the processed path arrays.
        double[][] xydArrays = new double[xySeriesList.size()][];

        // Look at each path series from the graph.
        int totalSize = 0;
        int i = 0;
        Iterator iter = xySeriesList.iterator();
        while (iter.hasNext()) {
            DoubleDataSeries series = (DoubleDataSeries)iter.next();

            // Process the path, store the resulting array, and add the length
            // variable.
            double[] xydArray = null;
            xydArray = processPath(series);
            totalSize += xydArray.length;
            xydArrays[i] = xydArray;
            i++;
        }

        // Create a new X-Y-Distance array.
        double[] outArray = new double[totalSize];

        // Append each of the processed path series into the combined array.
        int outIndex = 0;
        for (i = 0; i < xydArrays.length; i++) {
            double[] xydArray = xydArrays[i];
            for (int j = 0; j < xydArray.length; j++) {
                outArray[outIndex + j] = xydArray[j];
            }
            outIndex += xydArray.length;
        }

        // If the X-Y-Distance output series does not exist, create it.
        if (m_xydOutputSeries == null)
            add(new DoubleDataSeries(3));

        // Put the array into the output series.
        m_xydOutputSeries.setArray(outArray);

        m_bNeedRecalc = false;

        return bChanged;
    }

//------------------------------------------------------------------------------
/**
 * Take 2D DataSeries and create a 3D interleaved array containing the data
 * from the series as the first two dimensions.  Each pair of points in the
 * Xy series are assumed to be end points of a cross section defining a path
 * of variable width.  A center line is calculated through the centers of all
 * the cross sections.  The third dimension of the new array is the normalized
 * distance from the center of the first cross section along the center line
 * to the current cross section.  The same value is written for both points
 * on the cross section.  The distance for the first two points is 0.0 and
 * 1.0 for the last two points.
 * @param xyInputSeries The series to process
 * @return  The interleaved X-Y-Distance array.
 */
    public double[] processPath(DoubleDataSeries xyInputSeries) {

        // Get the data array and info.
        double[] xyArray = xyInputSeries.getArray();
        int numPathPts = xyInputSeries.getCount();
        double[] newArray = new double[numPathPts * 3];

        // Initialize counters.
        double lastCenterX = 0;
        double lastCenterY = 0;
        double totalLength = 0;
        int numCS = numPathPts / 2;

        // Look at each cross section in the array.
        for (int i = 0; i < numCS; i++) {
            // Calculate the center location.
            double csCenterX = (xyArray[i * 4] + xyArray[i * 4 + 2]) / 2;
            double csCenterY = (xyArray[i * 4 + 1] + xyArray[i * 4 + 3]) / 2;

            if (i != 0) {
                // Calculate the distance from the last center point.
                double xDiff = csCenterX - lastCenterX;
                double yDiff = csCenterY - lastCenterY;
                double segmentLength = (double)
                    Math.sqrt((xDiff * xDiff) + (yDiff * yDiff));

                // Add the length to the running total of center line length.
                totalLength += segmentLength;
            }

            // Set this center as the previous center for the next iteration.
            lastCenterX = csCenterX;
            lastCenterY = csCenterY;

            // Write the X and Y coordinates along with the unnormalized
            // distances to the interleaved array.
            newArray[i * 6] = xyArray[i * 4];
            newArray[i * 6 + 1] = xyArray[i * 4 + 1];
            newArray[i * 6 + 3] = xyArray[i * 4 + 2];
            newArray[i * 6 + 4] = xyArray[i * 4 + 3];
            newArray[i * 6 + 2] = totalLength;
            newArray[i * 6 + 5] = totalLength;
        }

        // Normalize the distances to the total center line length.
        for (int i = 0; i < numCS; i++) {
            newArray[i * 6 + 2] /= totalLength;
            newArray[i * 6 + 5] /= totalLength;
        }

        return newArray;
    }

//------------------------------------------------------------------------------
/**
 * Add a pair of points to an 2D DataSeries between every cross section that has
 * a point selected.  The new pair of points will form a cross section midway
 * between the two bounding cross sections.
 * @param series    The series to modify.
 */
    static public void addPoints(DoubleDataSeries series) {

        synchronized (series) {
            // Initialize local variables.
            int oldNumPoints = series.getCount();
            int numDims = series.getNumDimensions();
            int minIndex = series.getEnd();
            int maxIndex = 0;
            int[] selectedIndexes = series.getSelectedIndexes();
            double[] array = series.getArray();

            // Determine a range inside of which gaps should be filled.
            // This is determined by the first and last selected points.
            for (int j = 0; j < series.getNumSelected(); j++) {
                if (selectedIndexes[j] < minIndex)
                    minIndex = selectedIndexes[j];
                if (selectedIndexes[j] > maxIndex)
                    maxIndex = selectedIndexes[j];
            }

            // Determine how many new points will be added.
            int numNewPoints = ((maxIndex / 2) - (minIndex / 2)) * 2;

            // Create the new array and start filling it with all the values
            // from the old array.
            double[] newArray =
                new double[(oldNumPoints + numNewPoints) * numDims];
            int offset = 0;
            for (int i = 0; i < oldNumPoints; i++) {
                // Add the X and Y coordinates of the old point.
                newArray[(i + offset) * numDims] = array[i * numDims];
                newArray[(i + offset) * numDims + 1] = array[i * numDims + 1];


                if (i >= minIndex && i < maxIndex && i % 2 != 0) {
                    // This index is within the range of selected points and
                    // in between cross sections, so add a new cross section
                    // here.
                    int index1 = i - 1;
                    int index2 = i + 1;
                    if (index1 < 0)
                        index1 = 0;

                    // The new coordinates are the average of the cross sections
                    // on either side.
                    double newX = (array[index1 * numDims]
                        + array[index2 * numDims]) / 2;
                    double newY = (array[index1 * numDims + 1]
                        + array[index2 * numDims + 1]) / 2;
                    newArray[(i + offset + 1) * numDims] = newX;
                    newArray[(i + offset + 1) * numDims + 1] = newY;

                    // Repeat for the second point in the cross section.
                    index1 = i;
                    index2 = i + 2;
                    if (index2 >= oldNumPoints)
                        index2 = oldNumPoints - 1;
                    newX = (array[index1 * numDims]
                        + array[index2 * numDims]) / 2;
                    newY = (array[index1 * numDims + 1]
                        + array[index2 * numDims + 1]) / 2;
                    newArray[(i + offset + 2) * numDims] = newX;
                    newArray[(i + offset + 2) * numDims + 1] = newY;
                    offset += 2;
                }
            }

            // Update the data series with the new array.
            series.setArray(newArray);
            series.deselectAll();
        }
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        // If the graph changed, it could indicate a change to the curves inside it.
        if (gro == m_graph)
            m_bNeedRecalc = true;

        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

        String s = "Path Editor";

        return s;
     }

//==============================================================================
/**
 * A class to handle choosing a new path number with the spinner control.
 */
     class PathNumSpinnerModel extends javax.swing.AbstractSpinnerModel {

         //------------------------------------------------------------------------------
         /**
          * Tell the caller what the next value would be.
          * @return  An Integer containing the number of the next path.
          */
        public Object getNextValue() {
            return m_currentSeriesNum + 1;
        }

        //------------------------------------------------------------------------------
        /**
         * Tell the caller what the previous value would be.
         * @return  An Integer containing the number of the previous path.
         */
        public Object getPreviousValue() {
            return m_currentSeriesNum - 1;
        }

        //------------------------------------------------------------------------------
        /**
         * Tell the caller what the current value is.  Use "NONE" if the index
         * is out of range.
         * @return  An Integer containing the number of the current path.
         */
        public Object getValue() {
            if (m_currentSeriesNum < 1)
                return "NONE";
            else {
                return m_currentSeriesNum;
            }
        }

        //------------------------------------------------------------------------------
        /**
         * Handle a new path number being chosen.
         * @param value An Integer containing the number of the new path.
         */
        public void setValue(Object value) {

            // The value must be an Integer object.
            if (!(value instanceof Integer))
                return;

            // There must be a valid graph and int output series defined to
            // complete the operation.
            if (m_graph == null || m_intOutputSeries == null)
                return;

            // Update the list of 2D DataSeries objects within the graph.
            boolean bChanged = false;
            if (m_bNeedRecalc)
                bChanged = calculate();
            m_bNeedRecalc = false;

            // Perform bounds checking on the number.  The index rolls from the
            // last value back to the first and vice versa.
            int num = (Integer)value;
            int numPaths = m_xySeriesHashList.size();
            if (num > numPaths)
                num = 1;
            if (num < 1)
                num = numPaths;
            if (numPaths < 1)
                num = -1;

            // Don't set the value again if it is already correct.
            if (num != m_currentSeriesNum) {
                // Don't set the value if the calculate() method determined
                // the curves had changed independently.  Otherwise,
                // changes made in the properties dialog or in the graph could
                // cause the series hash list to become corrupted.
                if (!bChanged) {
                    // Before moving to another series, take the values
                    // from the text fields and apply them to the current
                    // series.
                    // First, node numbers...
                    try {
                        int startNodeNum = Integer.parseInt(
                            m_startNodeTextField.getText());
                        int endNodeNum = Integer.parseInt(
                            m_endNodeTextField.getText());

                        Object[] objArray =
                            (Object[])m_xySeriesHashList.getValue(
                            m_currentSeriesNum - 1);

                        // Update the list with the user-entered value
                        // and cause a recalc, if value was changed.
                        if ((Integer)objArray[0] != startNodeNum ||
                            (Integer)objArray[1] != endNodeNum) {

                            objArray[0] = startNodeNum;
                            objArray[1] = endNodeNum;
                            m_bNeedRecalc = true;
                        }

                        // Then, the path title...
                        String title = m_titleTextField.getText();
                        DoubleDataSeries xySeries = (DoubleDataSeries)
                            m_xySeriesHashList.getKey(
                            m_currentSeriesNum - 1);
                        if (!(xySeries.getTitle().equals(title))) {
                            m_bNeedRecalc = true;
                            xySeries.setTitle(title);
                        }
                    }
                    catch (Exception ex) {}
                }

                // Recalculate if one was requested.
                if (m_bNeedRecalc)
                    calculate();

                // Update the series number and fire a change event.
                m_currentSeriesNum = num;
                fireStateChanged();
            }

            // Put the values of the new path in the text fields.
            if (num > 0) {
                Object[] objArray =
                    (Object[])m_xySeriesHashList.getValue(num - 1);
                int startNodeNum = (Integer)objArray[0];
                int endNodeNum = (Integer)objArray[1];
                DoubleDataSeries xySeries = (DoubleDataSeries)
                    m_xySeriesHashList.getKey(num - 1);
                m_titleTextField.setText(xySeries.getTitle());
                m_startNodeTextField.setText("" + startNodeNum);
                m_endNodeTextField.setText("" + endNodeNum);
            }

            // Set the index of the highlighted path so it will show up
            // graphically.
            GlPathAnimation pathAnimation = (GlPathAnimation)m_graph.
                firstChildInstanceOf(
                "gov.usgs.sfhydro.model1d.gl.GlPathAnimation", false);
            if (pathAnimation != null)
                pathAnimation.setSelectedPathIndex(num - 1);

            // Cause a repaint.
            if (m_grCon != null)
                m_grCon.triggerRepaint();
        }
    }
}

