/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2004-01-12
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.model1d.gl;

import com.sun.opengl.util.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import gov.usgs.gr.visual.gl.*;
import java.awt.geom.*;
import java.nio.*;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * Draws an OpenGL object representing a path of variable width.  The path is
 * defined by cross sections and drawn as a stip of connected quadrilaterals.
 */
public class GlPathAnimation
         extends gov.usgs.gr.visual.DataAnimation implements GlElement {

    /** The OpenGL component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;
    /** The OpenGL utility toolkit class with all the glut methods. */
    protected GLUT glut = null;

    /** The thickness in pixels of lines that are drawn. */
    protected float m_thickness = 3;
    /** Whether to draw the cross sections. */
    protected boolean m_bDrawCrossSections = true;
    /** Whether to draw the center lines. */
    protected boolean m_bDrawCenterLines = true;
    /** Whether to draw the path quad strips. */
    protected boolean m_bDrawPathStrips = true;
    /** The path to highlight as selected. */
    protected int m_selectedPathIndex = -1;
    /** The array offset for the selected path. */
    protected int m_selectedPathOffset = -1;
    /**
     * The 3D data series holding the quadstip vertexes and the normalized
     * distances for calculating colors.
     */
    protected DoubleDataSeries m_xydSeries = null;
    /**
     * The 3D int data series holding the start node index, end node index, and
     * number of points for each path.
     */
    protected IntDataSeries m_intSeries = null;
    /** A buffer to hold a copy of the xydSeries for faster drawing. */
    protected DoubleBuffer m_xydBuffer = DoubleDataSeries.newDirectBuffer(0);

    protected String m_errorString = null;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GlPathAnimation.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating an empty object.
 */
    public GlPathAnimation(){

        // Set up member variables.
        initialize();

        // Read in default values.
        setDefaults("gov.usgs.gr.visual.gl.GlPathAnimation");
    }

//------------------------------------------------------------------------------
/**
 * Initialize all the member variables.
 */
    public void initialize() {

        // Set up the arrays to hold two 1D series for times and scalars.
        m_numDimsAllowed = 2;

        m_dataArrays = new double[m_numDimsAllowed + 1][];
        m_seriesArray = new DoubleDataSeries[m_numDimsAllowed + 1];
        m_seriesNums = new int[] {1, 2};
        m_dimNums = new int[] {1, 1};
        m_seriesNumDims = new int[] {0, 0, 0};
        m_dimScales = new double[] {1, 1};
        m_constantSeries = new DoubleDataSeries(
            new double[m_numDimsAllowed], m_numDimsAllowed);

        m_dimNames = "Times, Scalars";

        m_color = new float[] {1, 0.2f, 0.2f};
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * thickness          Integer    "Thickness"                  <p>
 * drawCrossSections  Boolean    "Draw Cross Sections"        <p>
 * drawCenterLines    Boolean    "Draw Center Lines"          <p>
 * drawPathStrips     Boolean    "Draw Path Strips"           <p>
 * Removes properties scale, zScale, zStretch, symbolType, showTitle,
 * and dimNums from super class property list.
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

        // First add properties from the super class.
        GrHashList props = super.getProperties();

        // Remove properties from super classes that don't need to be shown.
        props.remove("scale");
        props.remove("zScale");
        props.remove("zStretch");
        props.remove("symbolType");
        props.remove("showTitle");
        props.remove("dimNums");

        // Add the properties to the list.
        props.put("thickness", new GrProperty(m_thickness, "Thickness"));
        props.put("drawCrossSections", new GrProperty(
            m_bDrawCrossSections, "Draw Cross Sections"));
        props.put("drawCenterLines", new GrProperty(
            m_bDrawCenterLines, "Draw Center Lines"));
        props.put("drawPathStrips", new GrProperty(
            m_bDrawPathStrips, "Draw Path Strips"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.sfhydro.model1d.gl.GlPathAnimation#getProperties getProperties()
 * for a list of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

        // Let the super class handle its properties.
        super.setProperties(props);

             GrProperty p;

        // For each property with a string value, pass it to the setProperty()
        // method to be handled there.
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
                 String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (key.equals("thickness")) {
                    m_thickness = (Float)p.value;
                }
                if (key.equals("drawCrossSections")) {
                    m_bDrawCrossSections = (Boolean)p.value;
                    if (!m_bDrawCrossSections)
                        m_selectedPathIndex = -1;
                }
                if (key.equals("drawCenterLines")) {
                    m_bDrawCenterLines = (Boolean)p.value;
                    if (!m_bDrawCenterLines)
                        m_selectedPathIndex = -1;
                }
                if (key.equals("drawPathStrips")) {
                    m_bDrawPathStrips = (Boolean)p.value;
                    if (!m_bDrawPathStrips)
                        m_selectedPathIndex = -1;
                }
                // Handle string properties by passing them to the
                // setProperty() method.
                if (p.value instanceof String) {
                    String value = (String)p.value;
                    setProperty(key, value);
                }
            }
            // Handle string properties by passing them to the
            // setProperty() method.
            if (obj instanceof String) {
                                    String value = (String)obj;
                setProperty(key, value);
            }
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.sfhydro.model1d.gl.GlPathAnimation#getProperties getProperties()
 * for a list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

        // Let the super class handle its properties.
        super.setProperty(key, value);

        // Handle the property according to its key.
        if (key.equals("thickness")) {
            m_thickness = Float.parseFloat(value);
        }
        if (key.equals("drawCrossSections")) {
            m_bDrawCrossSections = Boolean.parseBoolean(value);
            if (!m_bDrawCrossSections)
                m_selectedPathIndex = -1;
        }
        if (key.equals("drawCenterLines")) {
            m_bDrawCenterLines = Boolean.parseBoolean(value);
            if (!m_bDrawCenterLines)
                m_selectedPathIndex = -1;
        }
        if (key.equals("drawPathStrips")) {
            m_bDrawPathStrips = Boolean.parseBoolean(value);
            if (!m_bDrawPathStrips)
                m_selectedPathIndex = -1;
        }
    }

//------------------------------------------------------------------------------
/**
 * Adds an object as a child, but only if it is an instance of IntDataSeries
 * with three dimensions, or a DoubleDataSeries with one or three
 * dimensions.  Removes the old child, if a matching one can be found, and
 * replaces it with the new one.  The exception is for 1D series, which are
 * added and not replaced.  The Times and Scalars series are specified by the
 * series num index property.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        if (m_children.contains(gro)) return false;

        boolean bAdded = false;

        if (gro instanceof IntDataSeries) {
            // Look for the old one and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof IntDataSeries) iter.remove();
            }

            // Add it and store it in the member variable.
            super.add(gro);
            m_intSeries = (IntDataSeries)gro;
            return true;
        }

        if (gro instanceof DoubleDataSeries) {
            DoubleDataSeries series = (DoubleDataSeries)gro;
            if (series.getNumDimensions() == 3) {
                // Look for the old one and remove it.
                Iterator iter = m_children.iterator();
                while (iter.hasNext()) {
                    Object o = iter.next();
                    if (o instanceof DoubleDataSeries) {
                        DoubleDataSeries oldSeries = (DoubleDataSeries)o;
                        if (oldSeries.getNumDimensions() == 3)
                            iter.remove();
                    }
                }

                // Add it and store it in the member variable.
                super.add(gro);
                m_xydSeries = (DoubleDataSeries)gro;
                return true;
            }
            // For the Times or Scalars...
            if (series.getNumDimensions() == 1) {
                super.add(gro);
                return true;
            }
        }

        return (super.add(gro));
    }

//------------------------------------------------------------------------------
/**
 * Overrides clear() in GrObject.
 */
    public synchronized void clear() {

        // Reset member variables to null.
        m_intSeries = null;
        m_xydSeries = null;
        m_colorScheme = null;

        super.clear();
    }

//------------------------------------------------------------------------------
/**
 * Overrides GrObject.remove() so member variable references to children can be
 * made null when they are removed.
 * @param i     The index of the object to remove.
 * @return      The object that was removed.
 */
    public GrObject remove(int i) {

        // Let the super class handle removal.
        GrObject child = super.remove(i);

        // Reset any associated member variable to null.
        if (child == m_intSeries) m_intSeries = null;
        if (child == m_xydSeries) m_xydSeries = null;
        if (child == m_colorScheme) m_colorScheme = null;

        return child;
    }

//------------------------------------------------------------------------------
/**
 * Overrides GrObject.remove() so member variable references to children can be
 * made null when they are removed.
 * @param gro   The object to be removed.
 * @return      Whether the object was removed.
 */
    public boolean remove(GrObject gro) {

        // Reset any associated member variable to null.
        if (gro == m_intSeries) m_intSeries = null;
        if (gro == m_xydSeries) m_xydSeries = null;
        if (gro == m_colorScheme) m_colorScheme = null;

        // Let the super class handle removal.
        return super.remove(gro);
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Sets the OpenGL objects to use for accessing drawing commands.
 * @param newgld   The OpenGL GLAutoDrawable component.
 * @param newgl    The OpenGL object containing the drawing commands.
 * @param newglu   The OpenGL utilites object.
 */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {

        if (glut == null || newgl != gl || newglu != glu)
            glut = new GLUT();

        gld = newgld;
        gl = newgl;
        glu = newglu;
    }

//------------------------------------------------------------------------------
/**
 * Check that there are valid data series for the times and scalars at the
 * indexes specified by the series nums.
 * @return Whether the data series are valid and ready to be drawn.
 */
    public boolean checkForValidData() {

        // Make sure the series indexes are within range.
        if (m_seriesNums[0] < 0 || m_seriesNums[1] < 0
            || m_seriesNums[0] == m_seriesNums[1]) {

            return false;
        }

        if (m_seriesArray[m_seriesNums[0]] == null
            || m_seriesArray[m_seriesNums[1]] == null) {

            return false;
        }

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 */
    public void drawGL() {

        // Before drawing, check if this object is supposed to be shown.
        if (!m_bShow) return;

        m_errorString = null;

        // Save the previous line thickness and color.
        gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

        // If there aren't valid times and scalars, just draw the cross sections
        // and center lines.
        if (!checkForValidData()) {
            if (m_bDrawCrossSections)
                drawCrossSections();
            if (m_bDrawCenterLines)
                drawCenterLines();

            m_errorString = "Data not valid";

            return;
        }

        DoubleDataSeries timeSeries = m_seriesArray[m_seriesNums[0]];
        synchronized (timeSeries) {
            // Determine the number of nodes to draw and what time step to use.
            int numNodes = pointCount(1) / pointCount(0);
            int timeOffset = timeSeries.getBegin() * numNodes;

            // Draw each of the components that are turned on.
            if (m_bDrawPathStrips)
                drawQuadStrip(timeOffset, numNodes);
            if (m_bDrawCrossSections)
                drawCrossSections();
            if (m_bDrawCenterLines)
                drawCenterLines();
            // Only draw the selection marker if at least one of the other
            // components is also turned on.
            if ((m_bDrawPathStrips || m_bDrawCrossSections
                || m_bDrawCenterLines) && (m_selectedPathIndex >= 0)) {

                drawSelectionMarker();
            }
        }

        // Restore the previous line attributes.
        gl.glPopAttrib();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw a quad strip for each path.
 * @param timeOffset  The time step number.  This number must be multiplied
 *                    by the number of nodes to determine the scalar array
 *                    offset.
 * @param numNodes    The number of nodes in the scalar array at each time step.
 */
    public void drawQuadStrip(int timeOffset, int numNodes) {

        // Don't proceed if there aren't data series available.
        if (m_xydSeries == null || m_intSeries == null)
            return;

        // Get the data arrays and related info.
        int numPaths = m_intSeries.getCount();
        int numIntDims = m_intSeries.getNumDimensions();
        int[] intArray = m_intSeries.getArray();
        int numXydDims = m_xydSeries.getNumDimensions();
        double[] xydArray = m_xydSeries.getArray();

        // Save the previous line color settings.
        gl.glPushAttrib(gl.GL_CURRENT_BIT);

        int offset = 0;
        for (int i = 0; i < numPaths; i++) {
            // Mark the offset of the selected path so the drawSelectionPath
            // method doesn't have to do it.
            if (i == m_selectedPathIndex)
                m_selectedPathOffset = offset;

            // Retrieve the start node, end node, and num points numbers.
            int startNodeNum = intArray[i * numIntDims] - 1;
            int endNodeNum = intArray[i * numIntDims + 1] - 1;
            int numPoints = intArray[i * numIntDims + 2];

            // Initialize local variables.
            double startVal = 0;
            double endVal = 0;
            double diffVal = 0;
            boolean bColorsDefined = false;

            if (startNodeNum < 0 || startNodeNum >= numNodes ||
                endNodeNum < 0 || startNodeNum >= numNodes) {

                // The end nodes of this path are undefined, so treat it as
                // as land area.  Set the color to the user specified color.
                gl.glColor3fv(m_color, 0);
            }
            else {
                if (m_colorScheme != null && timeOffset >= 0) {
                    // Set up the variables to use the color scheme and the
                    // proportional value between the node values.
                    startVal = getValueAt(timeOffset + startNodeNum, 1);
                    endVal = getValueAt(timeOffset + endNodeNum, 1);
                    diffVal = endVal - startVal;
                    bColorsDefined = true;
                }
                else {
                    // This path is not a land area, but the time series or
                    // color scheme are not defined, so draw it wet.
                    // Use light gray unless the land color is very light.
                    // In that case, use medium gray.
                    if ((m_color[0] + m_color[1] + m_color[2]) < 2)
                        gl.glColor3f(0.8f, 0.8f, 0.8f);
                    else
                        gl.glColor3f(0.4f, 0.4f, 0.4f);
                }
            }

            // Draw the quad strip.
            gl.glBegin(gl.GL_QUAD_STRIP);
            for (int j = 1; j < numPoints; j += 2) {
                if (bColorsDefined) {
                    // The end points were defined previously, so use the
                    // color scheme to determine the color for each vertex.
                    // The color should fade from the start value to the end
                    // value across the length of the quad strip.
                    double midValue = startVal + (diffVal * xydArray[
                    (offset + j) * numXydDims + 2]);
                    float[] color = m_colorScheme.getColor(midValue);
///System.out.println("" + color[0] + ", " + color[1] + ", " + color[2]);
// Caused a crash
//                    gl.glColor3fv(color);
                    gl.glColor3f(color[0], color[1], color[2]);
                }
                // Specify the X-Y locations of the vertexes.
                gl.glVertex2d(xydArray[ (offset + j - 1) * numXydDims],
                              xydArray[ (offset + j - 1) * numXydDims + 1]);
                gl.glVertex2d(xydArray[ (offset + j) * numXydDims],
                              xydArray[ (offset + j) * numXydDims + 1]);
            }
            gl.glEnd();

            // Jump to the next path offset.
            offset += numPoints;
        }

        // Restore the previous line attributes.
        gl.glPopAttrib();

    }

//------------------------------------------------------------------------------
/**
 * Draw a marker around the currently selected path to highlight it.
 */
    public void drawSelectionMarker() {

        // Don't proceed if there aren't data series available.
        if (m_xydSeries == null || m_intSeries == null)
            return;

        // Don't proceed if the selected path index isn't within range.
        if (m_selectedPathIndex >= m_intSeries.getCount())
            return;

        // Get the data arrays and related info.
        int numIntDims = m_intSeries.getNumDimensions();
        int[] intArray = m_intSeries.getArray();
        int numPoints = intArray[m_selectedPathIndex * numIntDims + 2];
        int numXydDims = m_xydSeries.getNumDimensions();
        double[] xydArray = m_xydSeries.getArray();

        // Don't proceed if the offset is past the end of the X-Y-D series.
        if ((m_selectedPathOffset + numPoints) > m_xydSeries.getCount() ||
            m_selectedPathOffset < 0) {

            return;
        }

        // Save the previous line pattern and color.
        gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT | gl.GL_ENABLE_BIT);

        // Set the width of the line.
        gl.glLineWidth(m_thickness + 2);

        // Draw a white line around the edge of the path.
        gl.glColor3f(1, 1, 1);

        gl.glBegin(gl.GL_LINE_LOOP);
        for (int j = 0; j < (numPoints - 1); j += 2) {
            gl.glVertex2d(xydArray[(m_selectedPathOffset + j) * numXydDims],
                          xydArray[(m_selectedPathOffset + j) * numXydDims + 1]);
        }
        for (int j = (numPoints - 1); j > 0; j -= 2) {
            gl.glVertex2d(xydArray[(m_selectedPathOffset + j) * numXydDims],
                          xydArray[(m_selectedPathOffset + j) * numXydDims + 1]);
        }
        gl.glEnd();

        // Draw a dashed red line around the edge of the path to create a
        // red and white striped effect.
        gl.glEnable(gl.GL_LINE_STIPPLE);
        gl.glLineStipple(1, Gr.DASHED_LINE_PATTERN);

        gl.glBegin(gl.GL_LINE_STRIP);
        for (int j = 0; j < (numPoints - 1); j += 2) {
            double val = xydArray[(m_selectedPathOffset + j) * numXydDims + 2];
            gl.glColor3f(1, 1 - (float)val, 1 - (float)val);
            gl.glVertex2d(xydArray[(m_selectedPathOffset + j) * numXydDims],
                          xydArray[(m_selectedPathOffset + j) * numXydDims + 1]);
        }
        for (int j = (numPoints - 1); j > 0; j -= 2) {
            double val = xydArray[(m_selectedPathOffset + j) * numXydDims + 2];
            gl.glColor3f(1, 1 - (float)val, 1 - (float)val);
            gl.glVertex2d(xydArray[(m_selectedPathOffset + j) * numXydDims],
                          xydArray[(m_selectedPathOffset + j) * numXydDims + 1]);
        }
        gl.glEnd();

        // Restore the previous line attributes.
        gl.glPopAttrib();

    }

//------------------------------------------------------------------------------
/**
 * Draw a line to show the location of every cross section on every path.
 */
    public void drawCrossSections() {

        // Don't proceed if there aren't data series available.
        if (m_xydSeries == null || m_intSeries == null)
            return;

        // Get the data arrays and related info.
        int numPaths = m_intSeries.getCount();
        int numIntDims = m_intSeries.getNumDimensions();
        int[] intArray = m_intSeries.getArray();
        int numXydDims = m_xydSeries.getNumDimensions();
        double[] xydArray = m_xydSeries.getArray();

        // Save the previous line pattern and color.
        gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

        // Set the width of the line.
        gl.glLineWidth(m_thickness);

        // Set the color.
        gl.glColor3f(1, 1, 0);

        int sizeOfDouble = 8;
        int offset = 0;
        for (int i = 0; i < numPaths; i++) {
            // Mark the offset of the selected path so the drawSelectionPath
            // method doesn't have to do it.
            if (i == m_selectedPathIndex)
                m_selectedPathOffset = offset;

            // Retrieve the start node, end node, and num points info.
            int startNodeNum = intArray[i * numIntDims];
            int endNodeNum = intArray[i * numIntDims + 1];
            int numPoints = intArray[i * numIntDims + 2];
            int stride = numPoints * numXydDims;
            if (startNodeNum == endNodeNum) {
                offset += numPoints;
                continue;

            }

            // Point OpenGL to the interleaved x-y array.
            m_xydBuffer.rewind();
            gl.glVertexPointer(2, gl.GL_DOUBLE, numXydDims * sizeOfDouble,
                m_xydBuffer);

            // Tell OpenGL to draw lines using the array values as indexes.
            gl.glDrawArrays(gl.GL_LINES, offset, numPoints);

            // Advance to the next path.
            offset += numPoints;
        }

        // Restore the previous line attributes.
        gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 * Draw a line to show the location of every path's center line.  The center
 * line passes through the mid point of each cross section.
 */
    public void drawCenterLines() {

        // Don't proceed if there aren't data series available.
        if (m_xydSeries == null || m_intSeries == null)
            return;

        // Get the data arrays and related info.
        int numPaths = m_intSeries.getCount();
        int numIntDims = m_intSeries.getNumDimensions();
        int[] intArray = m_intSeries.getArray();
        int numXydDims = m_xydSeries.getNumDimensions();
        double[] xydArray = m_xydSeries.getArray();

        // Save the previous line pattern and color.
        gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

        // Set the width of the line.
        gl.glLineWidth(m_thickness);

        // Set the color.
        gl.glColor3f(1, 1, 1);

        int offset = 0;
        for (int i = 0; i < numPaths; i++) {
            // Mark the offset of the selected path so the drawSelectionPath
            // method doesn't have to do it.
            if (i == m_selectedPathIndex)
                m_selectedPathOffset = offset;

            // Retrieve the start node, end node, and num points info.
            int startNodeNum = intArray[i * numIntDims];
            int endNodeNum = intArray[i * numIntDims + 1];
            int numPoints = intArray[i * numIntDims + 2];
            if (startNodeNum == endNodeNum) {
                offset += numPoints;
                continue;
            }

            // Draw the line strip down the center of the path.
            gl.glBegin(gl.GL_LINE_STRIP);
            for (int j = 1; j < numPoints; j += 2) {
                double x1 = xydArray[(offset + j - 1) * numXydDims];
                double y1 = xydArray[(offset + j - 1) * numXydDims + 1];
                double x2 = xydArray[(offset + j) * numXydDims];
                double y2 = xydArray[(offset + j) * numXydDims + 1];
                double midX = (x1 + x2) / 2;
                double midY = (y1 + y2) / 2;
                gl.glVertex2d(midX, midY);
            }
            gl.glEnd();

            // Advance to the next path.
            offset += numPoints;
        }

        // Restore the previous line attributes.
        gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        // There are no bounds without an X-Y-D series.
        if (!m_bShow || m_xydSeries == null) return null;

        Point2D.Double p = new Point2D.Double(m_xydSeries.getMin(dimension),
            m_xydSeries.getMax(dimension));

        return p;
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param index The index of the path to be highlighted, -1 if none.
 */
    public void setSelectedPathIndex(int index) {

        m_selectedPathIndex = index;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        if (gro == m_xydSeries) {
            DoubleDataSeries series = (DoubleDataSeries)gro;
            double[] array = series.getArray();
            if (array.length > m_xydBuffer.capacity()) {
                m_xydBuffer = DoubleDataSeries.newDirectBuffer(
                    array.length * 2);
            }

            m_xydBuffer.position(0);
            m_xydBuffer.put(array);
        }

        super.update(gro, arg);
}

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s = null;
        if (m_bShow)
            s = "GL Path Animation: ";
        else
            s = "(GL Path Animation): ";

        if (m_errorString != null)
            s = "<html>" + s + "<font color=\"#FF0000\">" + m_errorString
                + "</font></html>";
        else
            s += m_title;

        return s;
    }
}
