/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    2002-04-24
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.util.*;

//==============================================================================
/**
 * Bathymetry for the SI3D model, including barriers and boundaries.
 */

public class Bathymetry extends GrObject {

    protected DoubleDataSeries m_bathSeries = null;
    protected String m_runTitle = "";
    protected String m_runNumber = "";
    protected double m_xSpacing = 100;
    protected double m_ySpacing = 100;
    protected double m_zSpacing = 1;
    protected double m_rotation = 0;
    protected int[] m_lookups2D = null;
    protected int[] m_lookups3D = null;
    protected int[] m_index2D = null;
    protected int[] m_index3D = null;
    protected double[] m_avgDepths = null;
//------------------------------------------------------------------------------
/**
 * Creates an empty filter object.
 */
    public Bathymetry () {

        setDefaults("gov.usgs.sfhydro.si3d.Bathymetry");
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

		boolean bAdded = false;

        if (gro instanceof DoubleDataSeries) {
        	m_bathSeries = (DoubleDataSeries)gro;

            m_lookups2D = null;
            m_lookups3D = null;
            m_index2D = null;
            m_index3D = null;

            return super.add(gro);
        }
        if (gro instanceof GrObjectContainer) {
            ((GrObjectContainer)gro).setName("");

            return super.add(gro);
        }
        if (gro instanceof Location) {

            return super.add(gro);
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("runTitle", new GrProperty(m_runTitle, "Run Title"));
        props.put("runNumber", new GrProperty(m_runNumber, "Run Number"));
        props.put("xSpacing", new GrProperty(m_xSpacing, "X Spacing"));
        props.put("ySpacing", new GrProperty(m_ySpacing, "Y Spacing"));
        props.put("zSpacing", new GrProperty(m_zSpacing, "Z Spacing"));
        props.put("rotation", new GrProperty(m_rotation, "Rotation"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }

            	if (key.equals("xSpacing"))
            		m_xSpacing = (Double)p.value;
            	if (key.equals("ySpacing"))
            		m_ySpacing = (Double)p.value;
            	if (key.equals("zSpacing"))
            		m_zSpacing = (Double)p.value;
            	if (key.equals("rotation"))
            		m_rotation = (Double)p.value;
            }
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("runTitle"))
            m_runTitle = value;
        if (key.equals("runNumber"))
            m_runNumber = value;
        if (key.equals("xSpacing"))
            m_xSpacing = Double.parseDouble(value);
        if (key.equals("ySpacing"))
            m_ySpacing = Double.parseDouble(value);
        if (key.equals("zSpacing"))
            m_zSpacing = Double.parseDouble(value);
        if (key.equals("rotation"))
            m_rotation = Double.parseDouble(value);
    }

//------------------------------------------------------------------------------
/**
 * Create a regular array of integer indexes that locate the corresponding
 * point in the wet index array.
 */
    public int[] get2DLookupArray(int[] indexArray) {

        if (m_lookups2D != null) return m_lookups2D;

        // Find the number of wet points in the surface layer.
        int numIndexPoints = indexArray.length / 2;
        int numI = (int)Math.round(m_bathSeries.getMax(0)) + 1;
        int numJ = (int)Math.round(m_bathSeries.getMax(1)) + 1;

        // Create an array that will hold indexes into the I-J index array.
        // The ranges are 2 to maxI and 2 to maxJ.
        m_lookups2D = new int[numI * numJ];
        for (int i = 0; i < m_lookups2D.length; i++) {
            m_lookups2D[i] = -1;
        }

        for (int i = 0; i < numIndexPoints; i++) {
            int indexOffset = i * 2;
            int iIndex = indexArray[indexOffset];
            int jIndex = indexArray[indexOffset + 1];
            m_lookups2D[iIndex + jIndex * numI] = i;
        }

    	return m_lookups2D;
 	}

//------------------------------------------------------------------------------
/**
 * Create a regular array of integer indexes that locate the corresponding
 * point in the wet index array.
 */
    public int[] get3DLookupArray(int[] indexArray) {

        if (m_lookups3D != null) return m_lookups3D;

        // Find the number of wet points in the surface layer.
        int numIndexPoints = indexArray.length / 3;
        int numI = (int)Math.round(m_bathSeries.getMax(0)) + 1;
        int numJ = (int)Math.round(m_bathSeries.getMax(1)) + 1;
        int numLayers = (int)Math.round(m_bathSeries.getMax(2));

        // Create an array that will hold indexes into the I-J-K index array.
        // The ranges are 2 to maxI, 2 to maxJ, and 1 to numLayers.
        m_lookups3D = new int[numI * numJ * numLayers];
        for (int i = 0; i < m_lookups3D.length; i++) {
            m_lookups3D[i] = -1;
        }

        for (int i = 0; i < numIndexPoints; i++) {
            int indexOffset = i * 3;
            int iIndex = indexArray[indexOffset];
            int jIndex = indexArray[indexOffset + 1];
            int kIndex = indexArray[indexOffset + 2];

            m_lookups3D[iIndex * numLayers + jIndex * numLayers * numI + kIndex]
                = i;
        }

    	return m_lookups3D;
 	}

//------------------------------------------------------------------------------
/**
 * Convert the integer indexes to locations using the spacing variables.
 * @return the locations in a new data series.
 */
    public DoubleDataSeries get2DPositionSeries(int[] indexArray) {

        double[] positionArray = new double[indexArray.length];
        int numIndexPoints = indexArray.length / 2;

        for (int i = 0; i < numIndexPoints; i++) {
            int indexOffset = i * 2;
            positionArray[indexOffset] =
                (double)(indexArray[indexOffset]) * m_xSpacing;
            positionArray[indexOffset + 1] =
                (double)(indexArray[indexOffset + 1]) * m_ySpacing;
        }

        DoubleDataSeries series = new DoubleDataSeries(positionArray, 2);
        series.setTitle("2D Wet Cell Locations");
        series.getDimension(0).setTitle("X");
        series.getDimension(1).setTitle("Y");

    	return series;
 	}

//------------------------------------------------------------------------------
/**
 * Convert the integer indexes to locations using the spacing variables.
 * @return the locations in a new data series.
 */
    public DoubleDataSeries get3DPositionSeries(int[] indexArray) {

        double[] positionArray = new double[indexArray.length];
        int numIndexPoints = indexArray.length / 3;

        for (int i = 0; i < numIndexPoints; i++) {
            int indexOffset = i * 3;
            positionArray[indexOffset] =
                (double)(indexArray[indexOffset]) * m_xSpacing;
            positionArray[indexOffset + 1] =
                (double)(indexArray[indexOffset + 1]) * m_ySpacing;
            positionArray[indexOffset + 2] =
                (double)(indexArray[indexOffset + 2]) * m_zSpacing;
        }

        DoubleDataSeries series = new DoubleDataSeries(positionArray, 3);
        series.setTitle("3D Wet Cell Locations");
        series.getDimension(0).setTitle("X");
        series.getDimension(1).setTitle("Y");
        series.getDimension(2).setTitle("Z");

    	return series;
 	}

//------------------------------------------------------------------------------
/**
 * Create an array of integer indexes giving the location of each wet
 * vertical I-J column.
 * This will be used as a look-up for locations of data in Si3dOutput objects.
 */
    public int[] get2DIndexArray() {

        if (m_index2D != null) return m_index2D;

        // Find the number of wet points in the surface layer.
        double[] bath = m_bathSeries.getArray();
        int numPoints = m_bathSeries.getCount();
        int numBathDims = m_bathSeries.getNumDimensions();

        // Create an array that will hold I-J-K triples for every surface wet
        // point.  In the new array, I advances most quickly, followed by J.
        m_index2D = new int[numPoints * 2];
        for (int i = 0; i < numPoints; i++) {
            int bathOffset = i * numBathDims;
            int indexOffset = i * 2;

            // Put an I and J in the new array for each surface wet cell.
            m_index2D[indexOffset] = (int)Math.round(bath[bathOffset]);
            m_index2D[indexOffset + 1] = (int)Math.round(bath[bathOffset + 1]);
        }

    	return m_index2D;
 	}

//------------------------------------------------------------------------------
/**
 * Create an array of integer indexes giving the location of each wet cell.
 * This will be used as a look-up for locations of data in Si3dOutput objects.
 */
    public int[] get3DIndexArray() {

        if (m_index3D != null) return m_index3D;

        // Count the number of layers in the field of wet points.
        double[] bath = m_bathSeries.getArray();
        int numPoints = m_bathSeries.getCount();
        int numBathDims = m_bathSeries.getNumDimensions();
        int nwlSum = 0;
        for (int i = 0; i < numPoints; i++) {
            int bathOffset = i * numBathDims;
            nwlSum += Math.round(bath[bathOffset + 2]);
        }

        // Create an array that will hold I-J-K triples for every 3D wet point.
        // In the new array, K advances most quickly, followed by I and then J.
        m_index3D = new int[3 * nwlSum];
        int indexOffset = 0;
        for (int wet2dIndex = 0; wet2dIndex < numPoints; wet2dIndex++) {
            int bathOffset = wet2dIndex * numBathDims;
            int iVal = (int)Math.round(bath[bathOffset]);
            int jVal = (int)Math.round(bath[bathOffset + 1]);
            int nwlVal = (int)Math.round(bath[bathOffset + 2]);

            // Put an I-J-K triple in the new array for each wet cell.
            for (int k = 0; k < nwlVal; k++) {
                m_index3D[indexOffset] = iVal;
                m_index3D[indexOffset + 1] = jVal;
                m_index3D[indexOffset + 2] = k + 1;
                indexOffset += 3;
            }
        }


    	return m_index3D;
 	}

//------------------------------------------------------------------------------
/**
 * .
 */
    public double[] getAverageDepthArray() {

        if (m_avgDepths != null) return m_avgDepths;

        double[] bath = m_bathSeries.getArray();
        int numBathDims = m_bathSeries.getNumDimensions();
        m_avgDepths = new double[bath.length / 7];
        for (int i = 0; i < m_avgDepths.length; i++) {
            int offset = i * numBathDims;
            m_avgDepths[i] = (bath[offset + 3] + bath[offset + 4]
                + bath[offset + 5] + bath[offset + 6]) / 4;
        }

    	return m_avgDepths;
 	}

//------------------------------------------------------------------------------
/**
 * .
 */
    public DoubleDataSeries getBathymetrySeries() {

    	return m_bathSeries;
 	}

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 */
/*
public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"Bathymetry");

    	return subMenu;
    }
*/

//------------------------------------------------------------------------------
/*
public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
    }
*/

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Si3D Bathymetry";

        return s;
 	}
}

