/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-03
 * Date started:    2005-09-25
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import java.util.*;
import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.io.File;
import java.awt.Frame;
import gov.usgs.gr.visual.gl.GlVectorAnimation;
import gov.usgs.gr.visual.gl.GlTransformation;
import gov.usgs.gr.visual.TransformationOperation;
import gov.usgs.gr.visual.Curve;
import gov.usgs.gr.visual.gl.GlCurve;
import java.awt.Cursor;

//==============================================================================
/**
 * Calculates interpolated velocity vectors along an arbitrary cross sectional
 * path and rotates the vectors into the plane of the cross section.
 */

public class CrossSectionAnimator extends GrObject implements
	java.awt.event.ActionListener {

    /**
     * Holds any error descriptions that should be shown to the user in
     * the properties dialog.
     */
    protected String m_errorString = null;
    /** The name of the data set's file. */
    protected String m_filename = "tracks.bin";

    /** The working directory and ultimate output file. */
    protected File m_file =
        new File(System.getProperty("gov.usgs.gr.currentfile"));
    /** The default starting file name. */
    protected String m_startFileName = "";
    /** A handle to frame to use for showing dialogs. */
    protected Frame m_frame = null;
    /** Si3D bathymetry object. */
    protected Bathymetry m_bath = null;
    /** The X spacing of the Si3D bathymetry grid. */
    protected double m_xSpacing = 1;
    /** The Y spacing of the Si3D bathymetry grid. */
    protected double m_ySpacing = 1;
    /** The Z spacing of the Si3D bathymetry grid. */
    protected double m_zSpacing = 1;
    /** The number of columns on the Si3D grid. */
    protected int m_numI = 0;
    /** The number of rows on the Si3D grid. */
    protected int m_numJ = 0;
    /** The number of layers on the Si3D grid. */
    protected int m_numK = 0;
    /** The I-J indexes for each water column. */
    protected int[] m_2DIndexArray = null;
    /** The indexes for retrieving 2D variables from the wet-only 2D array. */
    protected int[] m_2DLookupArray = null;
    /** The I-J-K indexes for each wet cell. */
    protected int[] m_3DIndexArray = null;
    /** The index into the wet array for each cell on the grid.  -1 if dry. */
    protected int[] m_3DLookupArray = null;
    /** The number of wet columns on the grid. */
    protected int m_numWetPoints2D = 0;
    /** Si3D run time step increment. */
    protected double m_timeStepInc = 1;
    /** The number of time steps in the Si3D output. */
    protected int m_numTimeSteps = 0;
    /** The number of wet cells in the grid. */
    protected int m_numWetPoints3D = 0;
    /** The number of dimensions for the 3D output. */
    protected int m_num3DVarsDims = 0;
    /** The data series containing data types that vary by time, X, Y and Z. */
    protected RafDataSeries m_3DVarsSeries = null;
    /** The array stored in the 3D Vars Series. */
    protected float[] m_3DVarsArray = null;
    /**
     * The scales used for multiplying the U, V, and W values when they are
     * read from the data file.
     */
    protected double[] m_unitScales = {1, 1, 1};
    /** Time of the start of the Si3D run. */
    protected double m_si3dRunStartTime = 0;
    /** The data series containing data types that vary by time, X, and Y. */
    protected RafDataSeries m_2DVarsSeries = null;
    /** The array stored in the 2D Vars Series. */
    protected float[] m_2DVarsArray = null;
    /** The general output from the Si3D model. */
    protected Si3dOutput m_si3dOut = null;
    /** The Water Surface Elevation scale */
    protected double m_wseScale = 1;
    /** The cross section grid that describes the locations of the vectors. */
    protected CrossSectionGrid m_csg = null;
    /**
     * The start time of the vector output, in Julian decimal days referenced to
     * the same year as the Si3D run.
     */
    protected double m_startTime = 0;
    /**
     * The start time of the vector output, in Julian decimal days referenced to
     * the same year as the Si3D run.
     */
    protected double m_endTime = -1;
    /** The scaling applied to the W components of the vectors. */
    protected double m_wScale = 1;
    /** The scaling applied to the Z locations of the vectors. */
    protected double m_zScale = 1;
    /** The size of the scale arrow that is drawn next to the grid. */
    protected double m_scaleArrowSize = 0;
    /** Stores the most recently used cross section input array so it can be
     * compared during subsequent updates to determine if it has changed.
     */
    protected double[] m_lastInputCsArray = new double[0];
    /** The output container where the output objects are placed. */
    protected GrObjectContainer m_outputContainer = new GrObjectContainer();
    /** Whether the user requested a principal direction adjustment. */
    protected boolean m_bUserReqestedPrinDirAdjustment = false;
    /** Whether the user requested a forced recalculation. */
    protected boolean m_bForceRecalc = false;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/CrossSectionCalculator.gif";

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public CrossSectionAnimator () {

        // Set the output container.
        m_outputContainer.setName("Output");
        m_outputContainer.setProperty("title", "Output");
        add(m_outputContainer);
    }

//------------------------------------------------------------------------------
/**
 * Adds an object as a child, but only if it is an instance of Bathymetry,
 * CrossSectionGrid, DoubleDataSeries, GlVectorAnimation, or GrObjectContainer.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

    	if (m_children.contains(gro)) return false;

        boolean bAdded = false;

        if (gro instanceof DoubleDataSeries) {
        	DoubleDataSeries ds = (DoubleDataSeries)gro;

            if (ds.getNumDimensions() == 2 || ds.getNumDimensions() == 3) {
                return super.add(gro);
            }
        }
    	if (gro instanceof CrossSectionGrid) {
            return super.add(gro);
        }
    	if (gro instanceof GlVectorAnimation) {
            return super.add(gro);
        }
        if (gro instanceof GrObjectContainer) {
            GrObjectContainer groc = (GrObjectContainer) gro;
            GrHashList props = groc.getProperties();
            GrProperty p = (GrProperty) props.get("title");
            if (p == null)
                return false;
            String title = (String) p.value;

            // Compare the title of groc to each container and replace the
            // current one if there is a match.
            if (title.equals("Output")) {
                GrObjectContainer oldGroc = m_outputContainer;

                int index = m_children.size();
                if (oldGroc != null) {
                    index = m_children.indexOf(oldGroc);
                    oldGroc.deleteObserver(this);
                    m_children.remove(oldGroc);
                    if (index < 0)
                        index = 0;
                }
                m_children.add(index, groc);
                groc.addObserver(this);
                groc.setName("");

                m_outputContainer = groc;

                m_bHasChanged = true;
                return true;
            }
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * startTime       Double  "Start Time"                                    <p>
 * endTime         Double  "End Time"                                      <p>
 * wScale          Double  "W Scale"                                       <p>
 * zScale          Double  "Z Scale"                                       <p>
 * scaleArrowSize  Double  "Scale Arrow Size"                              <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("startTime", new GrProperty(m_startTime, "Start Time"));
        props.put("endTime", new GrProperty(m_endTime, "End Time"));
        props.put("wScale", new GrProperty(m_wScale, "W Scale"));
        props.put("zScale", new GrProperty(m_zScale, "Z Scale"));
        props.put("scaleArrowSize", new GrProperty(
            m_scaleArrowSize, "Scale Arrow Size"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.sfhydro.si3d.CrossSectionAnimator#getProperties getProperties()
 * for a list of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }

            	if (key.equals("startTime"))
            		m_startTime = (Double)p.value;
            	if (key.equals("endTime"))
            		m_endTime = (Double)p.value;
            	if (key.equals("wScale"))
            		m_wScale = (Double)p.value;
            	if (key.equals("zScale"))
            		m_zScale = (Double)p.value;
            	if (key.equals("scaleArrowSize"))
            		m_scaleArrowSize = (Double)p.value;
            }
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.sfhydro.si3d.CrossSectionGrid#getProperties getProperties()
 * for a list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("startTime")) {
            m_startTime = Double.parseDouble(value);
        }
        if (key.equals("endTime"))
            m_endTime = Double.parseDouble(value);
        if (key.equals("wScale"))
            m_wScale = Double.parseDouble(value);
        if (key.equals("zScale"))
            m_zScale = Double.parseDouble(value);
        if (key.equals("scaleArrowSize"))
            m_scaleArrowSize = Double.parseDouble(value);
    }

//------------------------------------------------------------------------------
/**
 * Recreate the vector animation by recalculating the interpolated vectors
 * along the cross section grid.
 */
    public void recalculate () {

        // Use the latest variables for cross section, bathymetry, etc.
        readVariables();

        // Look for the cross section grid child.
        CrossSectionGrid csg = (CrossSectionGrid)firstChildInstanceOf(
            "gov.usgs.sfhydro.si3d.CrossSectionGrid", false);

        // Quit if no cross section grid was found.
        if (csg == null)
            return;

        // Recreate the cross section grid in case anything has changed.
        csg.createCrossSectionGrid();

        // Get the output of the recreated cross section grid.
        IntDataSeries ikSeries = csg.getIKSeries();
        int[] ikArray = ikSeries.getArray();
        double[] xyzLocations = csg.getXyz();

        // Quit if the line series was not found.
        if (xyzLocations == null)
            return;

        int numXyzPoints = xyzLocations.length / 3;
        double csgSpacingX = csg.getSpacingX();
        double csgSpacingZ = csg.getSpacingZ();

        // Find the cross section line.
        DoubleDataSeries csSeries = (DoubleDataSeries)csg.firstChildInstanceOf(
            "gov.usgs.gr.data.DoubleDataSeries", false);

        // Quit if the line series was not found.
        if (csSeries == null)
            return;

        // Determine if the cross section or output location has changed.
        // Don't proceed unless the csInput array has changed.
        boolean bSameCS = true;
        boolean bSameGraphLocation = true;
        boolean bCS2Changed = false;
        double[] csArray = csSeries.getArray();

        // Initialize the last array if it hasn't been.
        if (m_lastInputCsArray.length == 0) {
            m_lastInputCsArray = (double[])csArray.clone();
        }

        // Look for changes in each point.
        if (csArray.length == m_lastInputCsArray.length) {
            for (int i = 0; i < csArray.length; i++) {
                if (csArray[i] != m_lastInputCsArray[i]) {
                    if (i < 4)
                        bSameCS = false;
                    if (i >= 4 && i >= (csArray.length - 2))
                        bSameGraphLocation = false;
                    if (i == 2 || i == 3)
                        bCS2Changed = true;
                }
            }
        }
        else {
            // If the current array is not the same length as the last array,
            // consider it changed.
            bSameCS = false;
            bSameGraphLocation = false;
        }

        // If this array is not the same as the last one, set this one to be
        // the last one for next time.
        if (!bSameCS || !bSameGraphLocation) {
            m_lastInputCsArray = (double[])csArray.clone();
        }

        // Calculate cosTheta and sinTheta for rotation vectors later.
        double x1 = csArray[0];
        double y1 = csArray[1];
        double x2 = csArray[2];
        double y2 = csArray[3];
        double xDiff = x2 - x1;
        double yDiff = y2 - y1;
        double length = Math.sqrt((xDiff * xDiff) + (yDiff * yDiff));
        double cosTheta = xDiff / length;
        double sinTheta = -yDiff / length;

        // Get output series and vector animation.
        DoubleDataSeries timeSeries = null;
        DoubleDataSeries locSeries = null;
        DoubleDataSeries vecSeries = null;
        double[] timeArray = null;
        double[] locArray = null;
        double[] vecArray = null;

        // Calculate the start and end time step.
        int startTimeStep =
            (int)((m_startTime - m_si3dRunStartTime) / m_timeStepInc);
        startTimeStep = Math.max(0, startTimeStep);
        int endTimeStep =
            (int)((m_endTime - m_si3dRunStartTime) / m_timeStepInc);
        endTimeStep = Math.max(startTimeStep, endTimeStep);
        endTimeStep = Math.min(m_numTimeSteps - 1, endTimeStep);
        int numTimes = endTimeStep - startTimeStep + 1;

        // Quit if there are no times defined.
        if (numTimes < 0)
            return;

        // Get the child output vector animation.
        GlVectorAnimation vecAnim = (GlVectorAnimation)firstChildInstanceOf(
            "gov.usgs.gr.visual.gl.GlVectorAnimation", false);

        // If no child animation was found, create and add one.
        if (vecAnim == null) {
            vecAnim = new GlVectorAnimation();
            add(vecAnim);
        }
        
        if (!bSameGraphLocation) {
            // Set the location of the graphics within the graph according to
            // the last point in the input series.
            List vecObservers = vecAnim.getObservers();
            Iterator observerIter = vecObservers.iterator();
            while (observerIter.hasNext()) {
                GrObject gro = (GrObject) observerIter.next();
                if (gro instanceof GlTransformation) {
                    List opList = gro.findChildInstancesOf(
                        "gov.usgs.gr.visual.TransformationOperation", true);
                    Iterator opIter = opList.iterator();
                    while (opIter.hasNext()) {
                        TransformationOperation op =
                            (TransformationOperation) opIter.next();
                        if (op.type == 0) {
                            op.x = csArray[csArray.length - 2];
                            op.y = csArray[csArray.length - 1];
                            break;
                        }
                    }
                }
            }
        }

        // The graphs has been moved, if the last point of the input series
        // changed, and that may be all that is required.
        // Quit if the cross section hasn't changed since previous recalc.
        if (bSameCS && !m_bForceRecalc)
            return;

        // Get the child time series, location series, and vector series from
        // the vector animation.  If any are missing, create and add them to
        // the vector animation.
        List childList = vecAnim.findChildInstancesOf(
            "gov.usgs.gr.data.DoubleDataSeries", false);
        if (childList.size() < 1) {
            timeSeries = new DoubleDataSeries(1);
            vecAnim.add(timeSeries);
        }
        else {
            timeSeries = (DoubleDataSeries)childList.get(0);
        }
        if (childList.size() < 2) {
            locSeries = new DoubleDataSeries(2);
            vecAnim.add(locSeries);
        }
        else {
            locSeries = (DoubleDataSeries)childList.get(1);
        }
        if (childList.size() < 3) {
            vecSeries = new DoubleDataSeries(3);
            vecAnim.add(vecSeries);
        }
        else {
            vecSeries = (DoubleDataSeries)childList.get(2);
        }

        // Get the time, location, and vector arrays, or allocate them if they
        // are not present in the vector animation.
        if (timeSeries.getCount() == numTimes) {
            timeArray = timeSeries.getArray();
        }
        else {
            timeArray = new double[numTimes];
            timeSeries.setArray(timeArray);
        }
        if (locSeries.getCount() == (numXyzPoints + 1)) {
            locArray = locSeries.getArray();
        }
        else {
            locArray = new double[(numXyzPoints + 1) * 2];
            locSeries.setArray(locArray);
        }
        if (vecSeries.getCount() == ((numXyzPoints + 1) * numTimes)) {
            vecArray = vecSeries.getArray();
        }
        else {
            vecArray = new double[(numXyzPoints + 1) * numTimes * 3];
            vecSeries.setArray(vecArray);
        }

        // Set up disk caching if the 3D series is in a random access file.
/*
        synchronized (m_3DVarsSeries) {
        int numWetPoints = m_3DIndexArray.length / 3;
        int numDims = m_3DVarsSeries.getNumDimensions();
        int[] existingCacheSizes = ((RafDataSeries)m_3DVarsSeries).
            getCacheSizes();
        int rafCacheSize = numWetPoints * numDims;
        if (existingCacheSizes.length == 0 ||
            rafCacheSize != existingCacheSizes[0]) {

            ((RafDataSeries)m_3DVarsSeries).createCacheSegments(1,
                rafCacheSize);
        }
*/

        // Initialize principal direction array and variables.
        double[] prinDiffArray = new double[timeArray.length * 2 + 4];
        int uvTotalNum = 0;
        double uTotal = 0;
        double vTotal = 0;
        double csPerpAngle = Math.toDegrees(Math.atan(yDiff / xDiff)) - 90;

        // Calculate the vectors in the cross section for each time step.
        for (int i = startTimeStep; i <= endTimeStep; i++) {
            int timeOffset = i - startTimeStep;
/*
            // Fill the cache with values from this time step.
            long vecOffset = i * m_numWetPoints3D;
            long rafCacheOffset = vecOffset * numDims;
            long[] existingCacheOffsets = ((RafDataSeries)m_3DVarsSeries).
                getCacheOffsets();
            if (existingCacheOffsets.length == 0 ||
                rafCacheOffset != existingCacheOffsets[0]) {

                ((RafDataSeries)m_3DVarsSeries).fillCacheSegment(0,
                    rafCacheOffset);
            }
*/
            // Write the times out.
            timeArray[timeOffset] = (double)i *
                m_timeStepInc + m_si3dRunStartTime;
            prinDiffArray[timeOffset * 2 + 2] = timeArray[timeOffset];

            // Write output for each point in the cross section.
            for (int j = 0; j < numXyzPoints; j++) {
                // Find the X-Y-Z location for this point.
                double x = xyzLocations[j * 3] + (1.5 * m_xSpacing);
                double y = xyzLocations[j * 3 + 1] + (1.5 * m_ySpacing);
                double z = xyzLocations[j * 3 + 2];
                if (timeOffset == 0) {
                    // One time only, write the X-Y-Z locations to the
                    // location array.
                    locArray[j * 2] =
                        (double)ikArray[j * 2] * csgSpacingX;
                    locArray[j * 2 + 1] = (double)ikArray[j * 2 + 1]
                        * csgSpacingZ  * m_zScale * -1;
                }

                // Calculate the interpolated vector at this time and location.
                int index = (timeOffset * (numXyzPoints + 1) + j) * 3;
                double[] vec = getVecAtPoint(i, x, y, z);

                // Rotate the vector into the plane of the cross section.
                double rotatedU = vec[0] * cosTheta - vec[1] * sinTheta;
                double rotatedV = vec[0] * sinTheta + vec[1] * cosTheta;
//test                double lenUn = Math.sqrt(vec[0] * vec[0] + vec[1] * vec[1]);
//test                double lenRot = Math.sqrt(rotatedU * rotatedU + rotatedV * rotatedV);
//test                double vRatio = (vec[0] * sinTheta) / (vec[1] * cosTheta);
//if (i == 32 && ikArray[j * 2 + 1] > 2) {
//    System.out.println("" + ikArray[j * 2] + ", " + ikArray[j * 2 + 1] + ", "
//        + vec[0] + ", " + vec[1] + ", " + rotatedU + ", " + rotatedV);
//}
                // Write the vector to the vector output array.
                vecArray[index] = rotatedU;
                vecArray[index + 1] = rotatedV;
                vecArray[index + 2] = vec[2] * m_wScale;

                // Add to the U and V totals for calculating principal
                // direction later.
                if (vec[0] != 0 || vec[1] != 0) {
                    uTotal += vec[0];
                    vTotal += vec[1];
                    uvTotalNum ++;
                }
            }

            // Calculate the principal direction.
            double avgU = uTotal / (double)uvTotalNum;
            double avgV = vTotal / (double)uvTotalNum;
            double prinAngle = Math.toDegrees(Math.atan(avgV / avgU));
            prinDiffArray[timeOffset * 2 + 3] =
                csPerpAngle - prinAngle;
            uTotal = 0;
            vTotal = 0;
            uvTotalNum = 0;

            // Add the scale arrow.
            int index = (timeOffset + 1) * (numXyzPoints + 1) * 3;
            vecArray[index - 3] = m_scaleArrowSize;
            vecArray[index - 2] = 0;
            vecArray[index - 1] = 0;
        }
//        } // Synchronized block

        // Add the scale arrow location.
        locArray[locArray.length - 2] = 0;
        locArray[locArray.length - 1] = 2 * csgSpacingZ  * m_zScale;

        // Extend the prin diff line to the start and end of its time series.
        prinDiffArray[0] = m_si3dRunStartTime;
        prinDiffArray[1] = prinDiffArray[3];
        prinDiffArray[prinDiffArray.length - 2] =
            m_si3dRunStartTime + (m_numTimeSteps * m_timeStepInc);
        prinDiffArray[prinDiffArray.length - 1] =
            prinDiffArray[prinDiffArray.length - 3];

        // Cause an update for the vector series.
        timeSeries.arrayChanged();
        locSeries.arrayChanged();
        vecSeries.arrayChanged();
//        vecAnim.setProperty("seriesNums", "1,2,2,0,3,3,3,0");
//        vecAnim.setProperty("dimNums", "0,0,1,-1,0,2,-1,-1");

        // Look for the output object container.
        GrObjectContainer groc =
            GrObjectContainer.getChildContainer(this, "Output");

        // If no container was found, create and add one as a child.
        if (groc == null) {
            groc = new GrObjectContainer();
            groc.setName("Output");
            groc.setProperty("title", "Output");
            add(groc);
        }

        // Create a list of the X-Y locations of the points in the cross
        // section, in the
        double lastX = xyzLocations[0];
        double lastY = xyzLocations[1];
        ArrayList<Double> xyList = new ArrayList();
        for (int j = 0; j < numXyzPoints; j++) {
            double x = xyzLocations[j * 3];
            double y = xyzLocations[j * 3 + 1];
            if (x != lastX && y != lastY) {
                xyList.add(x);
                xyList.add(y);
                lastX = x;
                lastY = y;
            }
        }

        // Find the output series with X-Y locations.
        DoubleDataSeries xyOutputSeries = (DoubleDataSeries)groc.
            firstChildInstanceOf("gov.usgs.gr.data.DoubleDataSeries", false);

        // Create and add the series if it was not found.
        if (xyOutputSeries == null) {
            xyOutputSeries = new DoubleDataSeries(2);
            groc.add(xyOutputSeries);
        }

        // Put the X-Y locations in the array for the X-Y location series.
        double[] xyArray = new double[xyList.size()];
        int index = 0;
        for (Double dVal : xyList)
            xyArray[index++] = dVal;

        xyOutputSeries.setArray(xyArray);

        // Look for the principal direction output curve.
        Curve prinDiffCurve = (Curve)groc.firstChildInstanceOf(
            "gov.usgs.gr.visual.Curve", false);

        // If the curve was not found, create and add a new one.
        if (prinDiffCurve == null) {
            prinDiffCurve = new GlCurve(new DoubleDataSeries(2));
            groc.add(prinDiffCurve);
        }

        // Put the recently calculated principal direction array into the
        // principal direction series.
        DoubleDataSeries prinDiffSeries = prinDiffCurve.getSeries();
        prinDiffSeries.setArray(prinDiffArray);

        // If the second cross section point was not changed, make an automatic
        // adjustment to it so it lines up with the principal direction.
        if ((m_bForceRecalc && m_bUserReqestedPrinDirAdjustment) ||
            (!m_bForceRecalc && !bCS2Changed)) {

            // Use the principal direction at the current time.
            int timeIndex = -1;
            if (timeSeries.getCount() == 1) {
                timeIndex = 0;
            }
            else {
                timeIndex = timeSeries.getBegin();
            }
            if (timeIndex >= 0) {
                int diffIndex = timeIndex * 2 + 3;
                if (diffIndex >= 0 && diffIndex < prinDiffArray.length) {
                    double diffAngle = prinDiffArray[diffIndex];
                    if (diffAngle > 90)
                        diffAngle -= 180;
                    if (diffAngle < -90)
                        diffAngle += 180;
                    if (diffAngle > 30)
                        diffAngle = 30;
                    if (diffAngle < -30)
                        diffAngle = -30;

                    if (m_bUserReqestedPrinDirAdjustment) {
                        System.out.println(
                            "Difference angle from principal direction = "
                            + diffAngle);
                    }

                    // Determine how much the angle needs to be adjusted.
                    double neededChange =
                        length * Math.asin(Math.toRadians(diffAngle));
                    if (Double.isNaN(neededChange)) {
                        System.out.println("Error - NaN");
                        return;
                    }

                    // Move the second end point of the cross section.
                    // This will cause this routine to be called recursively,
                    // but it re-execute this code because it will see that
                    // the second point was most recently modified.
                    csArray[2] -= neededChange * sinTheta;
                    csArray[3] -= neededChange * cosTheta;
                    csSeries.arrayChanged();
                }
            }
        }

        m_bUserReqestedPrinDirAdjustment = false;
        m_bForceRecalc = false;
    }

//------------------------------------------------------------------------------
/**
 * Interpolates the vector at an arbitrary point using the the nearest eight
 * values from the staggered Si3D grid.  The points from the staggered grid
 * make up the eight corners of a cube surrounding the point.
 * @param timeStep The time step of the vector.
 * @param x        The X location of the vector, on the Si3D grid.
 * @param y        The Y location of the vector, on the Si3D grid.
 * @param z        The Z location of the vector, on the Si3D grid.
 * @return An three-element array containing the U-V-W vector.
 */
    public double[] getVecAtPoint (int timeStep, double x, double y, double z) {

        // Clamp the depth below the zero line.
        if (z < 0)
            z = 0;

        // Determine the I, J, and K coordinates to use as the southwest-top
        // corner of the surrounding cube.  It will be different for U, V, and
        // W because of the staggered grid.
        int uI = (int)(x / m_xSpacing - 1);
        int uJ = (int)(y / m_ySpacing - 0.5);
        int uK = (int)(z / m_zSpacing + 0.5) + 1;

        int vI = (int)(x / m_xSpacing - 0.5);
        int vJ = (int)(y / m_ySpacing - 1);
        int vK = (int)(z / m_zSpacing + 0.5) + 1;

        int wI = (int)(x / m_xSpacing - 0.5);
        int wJ = (int)(y / m_ySpacing + 0.5);
        int wK = (int)(z / m_zSpacing) + 2;

        // Calculate the U component...
        // Determine the offset from the southwest-top corner.
        double cellX = x % m_xSpacing;
        double cellY = (y + (0.5 * m_ySpacing)) % m_ySpacing;
        double cellZ = (z + (0.5 * m_zSpacing)) % m_zSpacing;

        // Retrieve the U value at the eight corners of the cube.
        double swtU = getValueAt(timeStep, uI, uJ, uK, 0);
        double nwtU = getValueAt(timeStep, uI, uJ + 1, uK, 0);
        double setU = getValueAt(timeStep, uI + 1, uJ, uK, 0);
        double netU = getValueAt(timeStep, uI + 1, uJ + 1, uK, 0);
        double swbU = getValueAt(timeStep, uI, uJ, uK + 1, 0);
        double nwbU = getValueAt(timeStep, uI, uJ + 1, uK + 1, 0);
        double sebU = getValueAt(timeStep, uI + 1, uJ, uK + 1, 0);
        double nebU = getValueAt(timeStep, uI + 1, uJ + 1, uK + 1, 0);

        // Interpolate the value at (X, Y, Z).
        double stU = findValueBetween(swtU, setU, cellX, m_xSpacing);
        double ntU = findValueBetween(nwtU, netU, cellX, m_xSpacing);
        double sbU = findValueBetween(swbU, sebU, cellX, m_xSpacing);
        double nbU = findValueBetween(nwbU, nebU, cellX, m_xSpacing);
        double tU = findValueBetween(stU, ntU, cellY, m_ySpacing);
        double bU = findValueBetween(sbU, nbU, cellY, m_ySpacing);
        double xyzU = findValueBetween(tU, bU, cellZ, m_zSpacing);

        // Calculate the V component...
        // Determine the offset from the southwest-top corner.
        cellX = (x + (0.5 * m_xSpacing)) % m_xSpacing;
        cellY = y % m_ySpacing;
        cellZ = (z + (0.5 * m_zSpacing)) % m_zSpacing;

        // Retrieve the V value at the eight corners of the cube.
        double swtV = getValueAt(timeStep, vI, vJ, vK, 1);
        double nwtV = getValueAt(timeStep, vI, vJ + 1, vK, 1);
        double setV = getValueAt(timeStep, vI + 1, vJ, vK, 1);
        double netV = getValueAt(timeStep, vI + 1, vJ + 1, vK, 1);
        double swbV = getValueAt(timeStep, vI, vJ, vK + 1, 1);
        double nwbV = getValueAt(timeStep, vI, vJ + 1, vK + 1, 1);
        double sebV = getValueAt(timeStep, vI + 1, vJ, vK + 1, 1);
        double nebV = getValueAt(timeStep, vI + 1, vJ + 1, vK + 1, 1);

        // Interpolate the value at (X, Y, Z).
        double wtV = findValueBetween(swtV, nwtV, cellY, m_ySpacing);
        double etV = findValueBetween(setV, netV, cellY, m_ySpacing);
        double wbV = findValueBetween(swbV, nwbV, cellY, m_ySpacing);
        double ebV = findValueBetween(sebV, nebV, cellY, m_ySpacing);
        double tV = findValueBetween(wtV, etV, cellX, m_xSpacing);
        double bV = findValueBetween(wbV, ebV, cellX, m_xSpacing);
        double xyzV = findValueBetween(tV, bV, cellZ, m_zSpacing);

        // Calculate the W component...
        // Determine the offset from the southwest-top corner.
        cellX = (x + (0.5 * m_xSpacing)) % m_xSpacing;
        cellY = (y + (0.5 * m_ySpacing)) % m_ySpacing;
        cellZ = z % m_zSpacing;

        // Retrieve the W value at the eight corners of the cube.
        double swtW = getValueAt(timeStep, wI, wJ, wK, 2);
        double nwtW = getValueAt(timeStep, wI, wJ + 1, wK, 2);
        double setW = getValueAt(timeStep, wI + 1, wJ, wK, 2);
        double netW = getValueAt(timeStep, wI + 1, wJ + 1, wK, 2);
        double swbW = getValueAt(timeStep, wI, wJ, wK + 1, 2);
        double nwbW = getValueAt(timeStep, wI, wJ + 1, wK + 1, 2);
        double sebW = getValueAt(timeStep, wI + 1, wJ, wK + 1, 2);
        double nebW = getValueAt(timeStep, wI + 1, wJ + 1, wK + 1, 2);

        // Interpolate the value at (X, Y, Z).
        double swW = findValueBetween(swtW, swbW, cellZ, m_zSpacing);
        double nwW = findValueBetween(nwtW, nwbW, cellZ, m_zSpacing);
        double seW = findValueBetween(setW, sebW, cellZ, m_zSpacing);
        double neW = findValueBetween(netW, nebW, cellZ, m_zSpacing);
        double sW = findValueBetween(swW, seW, cellX, m_xSpacing);
        double nW = findValueBetween(nwW, neW, cellX, m_xSpacing);
        double xyzW = findValueBetween(sW, nW, cellY, m_ySpacing);

        // Return the U, V, and W values in a three-element array.
        double[] vec = new double[3];
        vec[0] = xyzU;
        vec[1] = xyzV;
        vec[2] = xyzW;

        return vec;
    }

//------------------------------------------------------------------------------
/**
 * Calculate a value on a line between two known values using interpolation.
 * @param val0     The first known value.
 * @param val1     The second known value.
 * @param distance The distance from first know value to the unknown value.
 * @param spacing  The distance between the two known values.
 * @return The interpolated value.
 */
    public double findValueBetween(double val0, double val1, double distance,
    double spacing) {

        // If the two know values are both zero, the interpolated value
        // must be zero.
        if (val0 == 0 && val1 == 0) {
            return 0;
        }

        // Calculate a weighted average.
        double proportion = distance / spacing;
        double newVal = val0 * (1 - proportion) + val1 * proportion;

        return newVal;
    }

//------------------------------------------------------------------------------
/**
 * Read the Si3D output variable at the specified time, location, and
 * dimension.
 * @param timeStep  The time step of the value.
 * @param locationI The I coordinate of the value.
 * @param locationJ The J coordinate of the value.
 * @param layer     The K coordinate of the value.
 * @param dimNum    The dimension to read.
 * @return The value that was retrieved.
 */
    public double getValueAt(int timeStep, int locationI, int locationJ,
    int layer, int dimNum) {

        // Calculate the offset of the time and location needed.
        int currentLookup = getLookupIndex(locationI, locationJ,
            layer, m_numI, m_numJ, m_numK, m_num3DVarsDims,
            timeStep, m_numWetPoints3D, m_3DLookupArray);

        return getVar3D(currentLookup, dimNum);
	}

//------------------------------------------------------------------------------
/**
 * Returns the U, V, or W components of the specified cell at the specified
 * time from the Si3D output or array.
 * @param lookupIndex The offset into the file for the value.
 * @param dimNum      The column to read at the given offset.
 * @return The value that was retrieved.
 */
    protected double getVar3D(int lookupIndex, int dimNum) {

        double var = 0;

        if (m_3DVarsSeries instanceof RafDataSeries) {
            // Read the variable from the file, if it is an RAF series.
            if (lookupIndex >= 0) {
                var = ((RafDataSeries)m_3DVarsSeries).getValueAt(
                    lookupIndex + dimNum);
                var *= m_unitScales[dimNum];
            }
        }
        else {
            // Read the variable from the array, if it is a regular series.
            if (lookupIndex >= 0 &&
                (lookupIndex + dimNum) < m_3DVarsArray.length) {

                var = m_3DVarsArray[lookupIndex + dimNum];
                var *= m_unitScales[dimNum];
            }
        }

        return var;
    }

//------------------------------------------------------------------------------
/**
 * Returns the lookup index of the specified cell at the specified time for
 * the Si3D output.
 * @param i            The I coordinate of the cell.
 * @param j            The J coordinate of the cell.
 * @param k            The K coordinate of the cell.
 * @param numI         The number of I columns in the grid.
 * @param numJ         The number of J rows in the grid.
 * @param numK         The number of K layers in the grid.
 * @param numVarDims   The number of variable dimensions at each time and
 *                     location.
 * @param ts           The time step.
 * @param numWetPoints The number of wet points written at each time.
 * @param lookupArray  The array containing the index of each wet point.
 * @return The index into the output array for the specified value.
 */
    protected int getLookupIndex(int i, int j, int k, int numI, int numJ,
    int numK, int numVarDims, int ts, int numWetPoints, int[] lookupArray) {

        // The default is the code for a dry cell.
        int lookupIndex = -1;

        // Get the bottom value from the lower neighbor cell.
        if (i >= 0 && i < numI && j >= 0 && j < numJ && k >= 0 && k < numK) {

            lookupIndex = lookupArray[i * numK + j * numK * numI + k];
            if (lookupIndex >= 0)
                lookupIndex = (lookupIndex + ts * numWetPoints) * numVarDims;
        }

        return lookupIndex;
    }

//------------------------------------------------------------------------------
/**
 * Initialize member variables by reading values from the Bathymetry and
 * Si3dOutput child objects.
 *
 * @return 0 if successful, -1 if child object was missing.
 */
    public synchronized int readVariables() {

        // Look for the child objects.
        CrossSectionGrid csg = (CrossSectionGrid)firstChildInstanceOf(
            "gov.usgs.sfhydro.si3d.CrossSectionGrid", true);

        if (csg == null)
            return -1;

        m_csg = csg;

        Bathymetry bath = (Bathymetry)m_csg.firstChildInstanceOf(
            "gov.usgs.sfhydro.si3d.Bathymetry", true);

        if (bath == null)
            return -1;

        m_bath = bath;

        Si3dOutput si3dOutput = (Si3dOutput)m_csg.firstChildInstanceOf(
            "gov.usgs.sfhydro.si3d.Si3dOutput", true);

        if (si3dOutput == null)
            return -1;

        m_si3dOut = si3dOutput;


        // Read variables...
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        m_xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        m_ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        m_zSpacing = (Double)p.value;

        m_si3dRunStartTime = m_si3dOut.getJulianStartTime();
        GrHashList si3dOutProps = m_si3dOut.getProperties();
        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long)p.value;
        m_timeStepInc = ((double)tsInSec) / 86400;

        m_2DIndexArray = m_bath.get2DIndexArray();
        m_2DLookupArray = m_bath.get2DLookupArray(m_2DIndexArray);
        m_3DIndexArray = m_bath.get3DIndexArray();
        m_3DLookupArray = m_bath.get3DLookupArray(m_3DIndexArray);
        DataSeries bathSeries = m_bath.getBathymetrySeries();
        m_3DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("3-D Variables");
//        if (m_3DVarsSeries != null)
//            m_3DVarsArray = m_3DVarsSeries.getArray();
        m_2DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("2-D x-y Variables");
        if (m_2DVarsSeries != null) {
//            m_2DVarsArray = m_2DVarsSeries.getArray();

            // Store the Water Surface Elevation scale.
            DataDimension dim = m_2DVarsSeries.getDimension(0);
            if (dim != null) {
                // Store the scale factor for the dimension.
                GrHashList dimProps = dim.getProperties();
                p = (GrProperty)dimProps.get("unitScale");
                m_wseScale = (Double)p.value;
            }
        }

        m_numWetPoints2D = m_2DIndexArray.length / 2;

        m_numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        m_numJ = (int)Math.round(bathSeries.getMax(1)) + 1;
        m_numK = (int)Math.round(bathSeries.getMax(2));

        m_numWetPoints3D = m_3DIndexArray.length / 3;
        m_num3DVarsDims = m_3DVarsSeries.getNumDimensions();
        m_numTimeSteps = m_3DVarsSeries.getCount() / m_numWetPoints3D;

        // Find the unit scales for the X, Y, and Z dimensions.
        GrProperty unitProp = (GrProperty)
            (m_3DVarsSeries.getDimension(0).getProperties().get("unitScale"));
        m_unitScales[0] = (Double)unitProp.value;
        unitProp = (GrProperty)
            (m_3DVarsSeries.getDimension(1).getProperties().get("unitScale"));
        m_unitScales[1] = (Double)unitProp.value;
        unitProp = (GrProperty)
            (m_3DVarsSeries.getDimension(2).getProperties().get("unitScale"));
        m_unitScales[2] = (Double)unitProp.value;

        return 0;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

   	    if (gro instanceof CrossSectionGrid)
        	recalculate();
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * @return  A menu item that can be placed on a menu and send an ActionEvent
 *          back to this object when it is chosen.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        // Create the menu item as a sub menu with other items.
        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"Cross Section Animator");

        // Set up menu items that use this object as their action listeners.
        javax.swing.JMenuItem menuItem =
            new javax.swing.JMenuItem("Create Vector Animation");
        menuItem.addActionListener(this);
        subMenu.add(menuItem);
        menuItem = new javax.swing.JMenuItem(
            "Create Vector Animation with Angle Adjustment");
        menuItem.addActionListener(this);
        subMenu.add(menuItem);
    	return subMenu;
    }

//------------------------------------------------------------------------------
/**
 * Handle action events involving this object.
 * @param e The ActionEvent that needs to be handled.
 */
    public void actionPerformed(java.awt.event.ActionEvent e) {

        // Get the text from the menu item that was chosen.
        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());

        // Show an hourglass until the job is complete.
        java.awt.Frame frame = GrController.getParentFrame(source);
        frame.setCursor(new Cursor(Cursor.WAIT_CURSOR));

        String srcText = source.getText();
        if (srcText.equals("Create Zone Animation")) {
            m_frame = GrController.getParentFrame(source);
            File[] files = GrController.showFileDialog("Choose File", true,
                m_file, m_frame);
            if (files != null && files.length > 0) {
                m_file = files[0];
                m_filename = m_file.getName();


                int status = readVariables();
//                if (status >= 0)
//                    createAnimation();
            }
        }
        if (srcText.equals("Create Vector Animation")) {
            m_lastInputCsArray = new double[0];
            m_bUserReqestedPrinDirAdjustment = false;
            m_bForceRecalc = true;
            recalculate();
        }
        if (srcText.equals("Create Vector Animation with Angle Adjustment")) {
            m_lastInputCsArray = new double[0];
            m_bUserReqestedPrinDirAdjustment = true;
            m_bForceRecalc = true;
            recalculate();
        }

        // Return to the default arrow cursor.
        frame.setCursor(Cursor.getDefaultCursor());
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Cross Section Animator";

        return s;
 	}
}

