/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    2005-10-04
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.util.*;
import gov.usgs.gr.visual.gl.GlDiscontinuousTriFanSurface;
import gov.usgs.sfhydro.gr.tools.CrossSectionCalculator;

//==============================================================================
/**
 * Cross section grid for the SI3D model.  Contains methods for creating a grid
 * along a given line from (X1, Y1) to (X2, Y2) extending vertically to the
 * bottom.
 */

public class CrossSectionGrid extends GrObject {

    /** Si3D bathymetry object. */
    protected Bathymetry m_bath = null;
    /** The X spacing of the Si3D bathymetry grid. */
    protected double m_xSpacing = 1;
    /** The Y spacing of the Si3D bathymetry grid. */
    protected double m_ySpacing = 1;
    /** The Z spacing of the Si3D bathymetry grid. */
    protected double m_zSpacing = 1;
    /** The number of columns on the Si3D grid. */
    protected int m_numI = 0;
    /** The number of rows on the Si3D grid. */
    protected int m_numJ = 0;
    /** The horizontal spacing of the interpolated cross section grid. */
    protected double m_csgSpacingX = 0;
    /** The vertical spacing of the interpolated cross section grid. */
    protected double m_csgSpacingZ = 0;
    /**
     * A series of interleaved I-K coordinates describing the locations of
     * possible wet points on the grid.
     */
    protected IntDataSeries m_csgIkSeries = new IntDataSeries(2);
    /** A series for looking up a cell's index in the wet series using it's
     * I-K location.  Stores -1 for dry cells.
     */
    protected IntDataSeries m_csgIndexSeries = new IntDataSeries(1);
    /** The number of columns on the cross section grid. */
    protected int m_csgNumI = 0;
    /** The number of rows on the cross section grid. */
    protected int m_csgNumK = 0;

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public CrossSectionGrid () {

        setDefaults("gov.usgs.sfhydro.si3d.CrossSectionGrid");
    }

//------------------------------------------------------------------------------
/**
 * Adds an object as a child, but only if it is an instance of Bathymetry,
 * Si3dOutput, DoubleDataSeries, or GrObjectContainer.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

        // Don't add it again if it is there already.
    	if (m_children.contains(gro)) return false;

		boolean bAdded = false;

        // Add a bathymetry object.
        if (gro instanceof Bathymetry)
            return super.add(gro);
        // Add Si3dOutput object
        if (gro instanceof Si3dOutput) {
            return super.add(gro);
        }
        // Add the data series that will contain the endpoints of the
        // cross section.
        if (gro instanceof DoubleDataSeries) {
            return super.add(gro);
        }
        // Add container object.
        if (gro instanceof GrObjectContainer) {
            ((GrObjectContainer)gro).setName("");

            return super.add(gro);
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * xSpacing    Double  "X Spacing"                                         <p>
 * zSpacing    Double  "Z Spacing"                                         <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("xSpacing", new GrProperty(m_csgSpacingX, "X Spacing"));
        props.put("zSpacing", new GrProperty(m_csgSpacingZ, "Z Spacing"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.sfhydro.si3d.CrossSectionGrid#getProperties getProperties()
 * for a list of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }

            	if (key.equals("xSpacing")) {
            		m_csgSpacingX = (Double)p.value;
            	}
            	if (key.equals("zSpacing"))
            		m_csgSpacingZ = (Double)p.value;
            }
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.sfhydro.si3d.CrossSectionGrid#getProperties getProperties()
 * for a list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("xSpacing")) {
            m_csgSpacingX = Double.parseDouble(value);
        }
        if (key.equals("zSpacing"))
            m_csgSpacingZ = Double.parseDouble(value);
    }

//------------------------------------------------------------------------------
/**
 * Create the I-K index array for the grid along the cross section.
 */
    public void createCrossSectionGrid () {

        // Read variables from child objects.
        if (readVariables() != 0)
            return;

        // Look for the input cross section data series.  Quit if none found.
        DoubleDataSeries csSeries = (DoubleDataSeries)firstChildInstanceOf(
            "gov.usgs.gr.data.DoubleDataSeries", false);

        if (csSeries == null)
            return;

        // Make a copy of the series to work with.  Quit if there aren't at
        // least two end points.
        csSeries = (DoubleDataSeries)csSeries.createCopy(true);
        if (csSeries.getCount() < 2)
            return;

        // Initialize local variables.
        List coordList = new ArrayList();
        double[] csArray = csSeries.getArray();
        double samplingRange = 0;
        double halfSpacing = samplingRange * 0.5;
        double xDiff = csArray[2] - csArray[0];
        double yDiff = csArray[3] - csArray[1];
        double xChange = 0;
        double yChange = 0;
        double xRatio = 0;
        double mPerpindicular = 0;

        // Handle the case of zero slope or infinite slope.
        if (xDiff != 0 && yDiff != 0) {
            mPerpindicular = -xDiff / yDiff;
            xRatio = 1 / Math.sqrt((1 + (mPerpindicular * mPerpindicular)));
        }

        // Run a cross section across parallel to the main one, but upstream and
        // downstream at regular intervals.  The spacing of the intervals is
        // equal to the Si3D grid spacing.
        int numCS = 1;
        if (samplingRange > 0)
            numCS = (int)(samplingRange / m_xSpacing) + 2;
        double dist = 0;

        // Calculate the endpoints of a parallel cross section at each
        // interval.
        for (int i = 0; i < numCS; i++) {
            double gap;
            // Increment the location by the Si3D grid spacing amount until the
            // last cross section.
            if (i == (numCS - 1))
                gap = halfSpacing;
            else
                gap = dist - halfSpacing;

            dist += m_xSpacing;

            // Calculate how much to change each endpoint from original cross
            // section.
            xChange = 0;
            yChange = 0;
            // Set the change for infinite slope.
            if (xDiff == 0) {
                if (yDiff > 0)
                    xChange = -gap;
                else
                    xChange = gap;
            }
            // Set the change for zero slope.
            if (yDiff == 0) {
                if (xDiff > 0)
                    yChange = -gap;
                else
                    yChange = -gap;
            }
            // Set the change for all other slopes.
            if (xDiff != 0 && yDiff != 0) {
                xChange = gap * xRatio;
                yChange = xChange * mPerpindicular;
            }

            // Store the coordinate of the end points in a list.
            coordList.add(csArray[0] + xChange);
            coordList.add(csArray[1] + yChange);
            coordList.add(csArray[2] + xChange);
            coordList.add(csArray[3] + yChange);
        }

        // Use a tri fan surface object to create seven-dimensional series
        // of the bathymetry surface.
        DoubleDataSeries ds7Series = m_bath.getBathymetrySeries();
        DoubleDataSeries ds3Series = gov.usgs.gr.visual.gl.
            GlDiscontinuousTriFanSurface.ds7toDs3(ds7Series, m_xSpacing);

        // Use a cross section calculator to calculate the line that passes
        // along the bottom of the cross section.
        CrossSectionCalculator csc = new CrossSectionCalculator();
        csc.add(csSeries);
        csc.add(ds3Series);
        csc.recalculate();
        GrObjectContainer groc =
            GrObjectContainer.getChildContainer(csc, "Output");
        DoubleDataSeries lineSeries = (DoubleDataSeries)groc.
            firstChildInstanceOf("gov.usgs.gr.data.DoubleDataSeries", false);
        if (lineSeries == null)
            return;

        // Get the output line that the calculator made.
        double[] lineArray = lineSeries.getArray();

        // The last point should be just past the far bank.  Find the index
        // of the far bank by starting at the last point and working toward
        // the near bank until a non-zero depth is found.
        int farBankIndex = lineArray.length - 2;
        while (farBankIndex > 0 && lineArray[farBankIndex - 1] >= 0) {
            farBankIndex -= 2;
        }

        // Determine the effective length of the cross section and add some
        // padding around the sides.
        double lineLength = lineArray[farBankIndex] - lineArray[0];
        int numHorizontalPads = (int)(m_xSpacing / m_csgSpacingX) + 1;
        int numPadLayers = (int)(m_zSpacing / m_csgSpacingZ) + 1;

        // Keep a count of the number of layers in each column across the
        // channel.
        int[] layerCounts = new int[(int)(lineLength / m_csgSpacingX) + 1
            + numHorizontalPads];
        Arrays.fill(layerCounts, Integer.MIN_VALUE);

        // Initialize variables for comparing to the last I and layer values.
        int lastIVal = 0;
        int lastNumLayers = 0;

        // Check the parallel lines at evenly spaced intervals.
        Iterator iter = coordList.iterator();
        while (iter.hasNext()) {
            // Get the end points of a parallel cross section and calculate
            // its bottom profile.
            csArray[0] = (Double)iter.next();
            csArray[1] = (Double)iter.next();
            csArray[2] = (Double)iter.next();
            csArray[3] = (Double)iter.next();
            csSeries.arrayChanged();
            csc.recalculate();

            // Look at each point along the cross section profile.
            lineArray = lineSeries.getArray();
            int numPoints = lineSeries.getCount();
            for (int i = 0; i < numPoints; i++) {
                // Convert the X location to I column index.
                int iVal = (int)(lineArray[i * 2] / m_csgSpacingX);
                if (iVal >= layerCounts.length)
                    break;

                // If there are more layers on this column than on any previous
                // column, set the maxNumLayers.
                double depth = -lineArray[i * 2 + 1] / m_csgSpacingZ;
                int numLayers = (int)Math.ceil(depth) + numPadLayers;
                int maxNumLayers = Math.max(numLayers, lastNumLayers);

                // Since the profile can skip columns, fill in any intermediate
                // columns with the same number of layers.
                for (int j = lastIVal; j <= iVal; j++) {
                    if (layerCounts[j] < maxNumLayers)
                        layerCounts[j] = maxNumLayers;
                }

                // Set the next "last value" to be the current value.
                lastIVal = iVal;
                lastNumLayers = numLayers;
            }
        }

        // Fill in the corners along the bottom.
        int[] layerCountCopy = (int[])layerCounts.clone();
        // To the left and right of each column fill at least as much as one
        // Si3D grid cell would occupy.
        for (int i = 1; i < (layerCounts.length - 1); i++) {
            for (int j = 0; j < numHorizontalPads; j++) {
                if ((i - j) >= 0) {
                    if (layerCounts[i] < layerCountCopy[i - j])
                        layerCounts[i] = layerCountCopy[i - j];
                }
                if ((i + j) < layerCounts.length) {
                    if (layerCounts[i] < layerCountCopy[i + j])
                        layerCounts[i] = layerCountCopy[i + j];
                }
            }
        }

        // Set the columns from the initial value to zero.
        for (int i = 0; i < layerCounts.length; i++) {
            if (layerCounts[i] == Integer.MIN_VALUE)
                layerCounts[i] = 0;
        }

        // Initialize variables for creating I-K location series.
        List columnList = new ArrayList();
        int lowestI = 0;
        int highestI = -1;

        // Look for wet water columns across the line.
        for (int i = 0; i < layerCounts.length; i++) {
            // Create an I-K location for each wet cell.
            for (int k = 0; k < layerCounts[i]; k++) {
                if (lowestI == 0)
                    lowestI = i;
                columnList.add(i - lowestI);
                columnList.add(k);
                highestI = i;
            }

            // Keep track of the largest number of layers encountered.
            if (m_csgNumK < layerCounts[i])
                m_csgNumK = layerCounts[i];
        }

        // Copy the I-K values from the list to an int array.
        int[] csgIkLocations = new int[columnList.size()];
        for (int i = 0; i < csgIkLocations.length; i++) {
            csgIkLocations[i] = ((Integer)columnList.get(i));
        }

        // Update the I-K series with the array.
        m_csgIkSeries.setArray(csgIkLocations);

        // Determine the I range.
        m_csgNumI = highestI - lowestI + 1;

        // Don't create an index array if there are no wet points.
        if (m_csgNumI < 1)
            return;

        // Create the I-K index array.
        int[] csgIndexArray = new int[m_csgNumI * m_csgNumK];
        Arrays.fill(csgIndexArray, -1);
        int numPoints = csgIkLocations.length / 2;
        for (int i = 0; i < numPoints; i++) {
            int iVal = csgIkLocations[i * 2];
            int kVal = csgIkLocations[i * 2 + 1];
            int index = iVal * m_csgNumK + kVal;
            csgIndexArray[index] = i;
        }

        // Update the index series with the array.
        m_csgIndexSeries.setArray(csgIndexArray);
    }

//------------------------------------------------------------------------------
/**
 * Initialize member variables by reading values from the Bathymetry and
 * Si3dOutput child objects.
 *
 * @return int, 0 if successful, -1 if child object was missing.
 */
    public synchronized int readVariables() {

        boolean bChanged = false;
        int status = 0;

        // Look for the child objects.
        Bathymetry bath = (Bathymetry)firstChildInstanceOf(
            "gov.usgs.sfhydro.si3d.Bathymetry", true);

        if (bath == null)
            return -1;

        m_bath = bath;

        Si3dOutput si3dOutput = (Si3dOutput)firstChildInstanceOf(
            "gov.usgs.sfhydro.si3d.Si3dOutput", true);

        if (si3dOutput == null)
            return -1;

//        m_si3dOut = (Si3dOutput)list.get(0);

        // Read variables...
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        m_xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        m_ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        m_zSpacing = (Double)p.value;

//        double startTime = m_si3dOut.getJulianStartTime();
//        GrHashList si3dOutProps = m_si3dOut.getProperties();
//        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
//        Long tsInSec = (Long)p.value;
//        double timeStepInc = ((double)tsInSec) / 86400;

//        m_2DIndexArray = m_bath.get2DIndexArray();
//        m_2DLookupArray = m_bath.get2DLookupArray(m_3DIndexArray);
//        m_3DIndexArray = m_bath.get3DIndexArray();
//        m_3DLookupArray = m_bath.get3DLookupArray(m_3DIndexArray);
        DataSeries bathSeries = m_bath.getBathymetrySeries();
//        m_2DVarsSeries = (FloatDataSeries)
//            m_si3dOut.getContainerSeries("2-D x-y Variables");
//        if (m_2DVarsSeries != null) {
//            m_2DVarsArray = m_2DVarsSeries.getArray();
//
//            // Store the Water Surface Elevation scale.
//            DataDimension dim = m_2DVarsSeries.getDimension(0);
//            if (dim != null) {
//                // Store the scale factor for the dimension.
//                GrHashList dimProps = dim.getProperties();
//                p = (GrProperty)dimProps.get("unitScale");
//                m_wseScale = ((Double)p.value).doubleValue();
//            }
//        }
//
//        m_numWetPoints2D = m_2DIndexArray.length / 2;

        m_numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        m_numJ = (int)Math.round(bathSeries.getMax(1)) + 1;
//        m_numK = (int)Math.round(bathSeries.getMax(2));

        return 0;
    }

//------------------------------------------------------------------------------
/**
 * Get an instance of the I-K location series for wet points in the cross
 * section.
 * @return  The I-K location IntDataSeries object.
 */
    public IntDataSeries getIKSeries() {

    	return m_csgIkSeries;
 	}

//------------------------------------------------------------------------------
/**
 * Get an instance of the I-K index series for all points in the cross
 * section, using -1 for dry points.
 * @return  The I-K index IntDataSeries object.
 */
    public IntDataSeries getIndexSeries() {

    	return m_csgIndexSeries;
 	}

//------------------------------------------------------------------------------
/**
 * Data access method.
 * @return The horizontal spacing of the interpolated cross section grid.
 */
    public double getSpacingX() {

    	return m_csgSpacingX;
 	}

//------------------------------------------------------------------------------
/**
 * Data access method.
 * @return The vertical spacing of the interpolated cross section grid.
 */
    public double getSpacingZ() {

    	return m_csgSpacingZ;
 	}

//------------------------------------------------------------------------------
/**
 * Create an array of X-Y-Z points.  Each point represents the location of a
 * wet point on the cross section grid, after moving it to the Si3D grid's
 * coordinate system.
 * @return  An array containing the X-Y-Z location of each wet point on the
 *          cross section grid.
 */
    public double[] getXyz() {


        // Read variables from child objects.
        if (readVariables() != 0)
            return null;

        // Look for the input cross section data series.
        DoubleDataSeries csSeries = (DoubleDataSeries)firstChildInstanceOf(
            "gov.usgs.gr.data.DoubleDataSeries", false);

        // Quit if there is no input cross section.
        if (csSeries == null)
            return null;

        // Make a working copy of the cross section.  Quit if there is none.
        csSeries = (DoubleDataSeries)csSeries.createCopy(true);
        if (csSeries.getCount() < 2)
            return null;

        // Find the endpoints of the input cross section.
        List coordList = new ArrayList();
        double[] csArray = csSeries.getArray();
        double x1 = csArray[0];
        double y1 = csArray[1];
        double x2 = csArray[2];
        double y2 = csArray[3];

        // Calculate the length, sine, and cosine of the line.
        double xDiff = x2 - x1;
        double yDiff = y2 - y1;
        double length = Math.sqrt((xDiff * xDiff) + (yDiff * yDiff));
        double cosTheta = xDiff / length;
        double sinTheta = yDiff / length;

        // Work from the I-K grid.  Create it if it hasn't been already.
        if (m_csgIkSeries.getCount() == 0)
            createCrossSectionGrid();

        // Look at each point in the cross section grid.
        int[] m_csgIkArray = m_csgIkSeries.getArray();
        int numPoints = m_csgIkSeries.getCount();
        double[] newArray = new double[numPoints * 3];
        for (int i = 0; i < numPoints; i++) {
            // Rotate and shift the point into the Si3D grid's coordiate
            // system and store the value in the new array.
            double csgX = m_csgIkArray[i * 2] * m_csgSpacingX;
            double csgY = 0;
            double csgZ = m_csgIkArray[i * 2 + 1] * m_csgSpacingZ;

            double rotatedX = csgX * cosTheta - csgY * sinTheta;
            double rotatedY = csgX * sinTheta - csgY * cosTheta;
            double shiftedX = rotatedX + x1;
            double shiftedY = rotatedY + y1;

            newArray[i * 3] = shiftedX;
            newArray[i * 3 + 1] = shiftedY;
            newArray[i * 3 + 2] = csgZ;
        }

        return newArray;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        // Recreate the grid if the input cross section changed.
        if (gro instanceof DoubleDataSeries)
            createCrossSectionGrid();

        super.update(gro, arg);
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Cross Section Grid";

        return s;
 	}
}

