/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-07-13
 * Date started:    2003-05-22
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.sfhydro.si3d.*;
import java.util.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class FluxCalculator extends GrObject {

    protected double m_nearZero = 0.00000000000001;
    protected double[] m_unitScales = {1, 1, 1};
    protected double m_wseScale = 1;

    protected RafDataSeries m_2DVarsSeries = null;
    protected int[] m_2DLookupArray = null;
    protected float[] m_2DVarsArray = null;

    protected RafDataSeries m_3DVarsSeries = null;
    protected float[] m_3DVarsArray = null;

    protected Bathymetry m_bath = null;
    protected Si3dOutput m_si3dOut = null;
    protected GrObjectContainer m_outputContainer = new GrObjectContainer();
    protected DoubleDataSeries m_crossSectionInputSeries =
	new DoubleDataSeries(2);
    protected DoubleDataSeries m_crossSectionOutputSeries =
	new DoubleDataSeries(2);
    protected TimeSeries m_tfOutputSeries = new TimeSeries(2);
///    protected DataSeries m_txyzuvOutputSeries = new DataSeries(6);
///    protected DataSeries m_xyzdOutputSeries = new DataSeries(4);

    protected double m_xSpacing = 1;
    protected double m_ySpacing = 1;
    protected double m_zSpacing = 1;
    protected double m_si3dRunStartTime = 0;
    protected int m_si3dRunStartYear = 0;
    protected double m_timeStepInc = 1;
    protected int m_numTimeSteps = 0;
    protected int m_numI = -1;
    protected int m_numJ = -1;
    protected int m_numLayers = -1;
    protected int m_numWet2Dpoints = 0;
    protected double[] m_avgDepthArray = null;
    protected int[] m_3DIndexArray = null;
    protected int[] m_3DLookupArray = null;
    protected int m_num3DVarsDims = 0;
    protected int[] m_2DIndexArray = null;
    protected int m_num3DWetPoints = 0;

    protected boolean m_bInitialized = false;
    protected boolean m_bBathInitialized = false;
    protected boolean m_bAutoRecalc = false;
    protected boolean m_bGodinFilter = false;
    protected ArrayList m_segmentList = new ArrayList();
    protected double m_fluxScale = 1;
    protected double m_uScale = 1;
    protected double m_vScale = 1;

    protected class CrossSectionSegment {
        public int i = 0;
        public int j = 0;
        public boolean bUFace = true;
        public int numLayers = 0;
        public double bottomLayerHeight = 0;
    }

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public FluxCalculator () {

        m_outputContainer.setName("Output");
        m_outputContainer.setProperty("title", "Output");
        m_outputContainer.add(m_tfOutputSeries);
///        m_outputContainer.add(m_xyzdOutputSeries);
        add(m_outputContainer);
    }

//------------------------------------------------------------------------------
/**
 * Overrides     in GrObject.  Only allows DataDimensions to be added.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        if (gro instanceof Bathymetry) {
            // Look for the old bathymetry and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof Bathymetry) iter.remove();
            }

            // Add it at the beginning of the list.
            super.add(gro);
            m_children.remove(gro);
            m_children.add(0, gro);
            m_bath = (Bathymetry)(gro);
            m_bInitialized = false;
            return true;
        }
        if (gro instanceof Si3dOutput) {
            // Look for the old one and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof Si3dOutput) iter.remove();
            }

            // Add it at the beginning of the list.
            super.add(gro);
            m_children.remove(gro);
            m_children.add(0, gro);
            m_si3dOut = (Si3dOutput)gro;
            m_bInitialized = false;
            return true;
        }
        if (DoubleDataSeries.countDims(gro, 2)) {
            // Look for the old one and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (DoubleDataSeries.countDims(o, 2))
		    iter.remove();
            }

            // Add it at the beginning of the list.
            super.add(gro);
            m_children.remove(gro);
            m_children.add(0, gro);
            m_crossSectionInputSeries = (DoubleDataSeries)gro;
            return true;
        }
        if (gro instanceof GrObjectContainer) {
            GrObjectContainer groc = (GrObjectContainer)gro;
            GrHashList props = groc.getProperties();
            GrProperty p = (GrProperty)props.get("title");
            if (p == null)
                return false;
            String title = (String)p.value;

            // Compare the title of groc to each container and replace the
            // current one if there is a match.
            if (title.equals("Output")) {
                GrObjectContainer oldGroc = m_outputContainer;

                int index = m_children.size();
                if (oldGroc != null) {
                    index = m_children.indexOf(oldGroc);
                    oldGroc.deleteObserver(this);
                    m_children.remove(oldGroc);
                    if (index < 0)
                        index = 0;
                }
                m_children.add(index, groc);
                groc.addObserver(this);
                groc.setName("");

                m_outputContainer = groc;

                // Look for DataSeries within the output container
                // and use them for output if they are there.
                Enumeration enumeration = m_outputContainer.children();
                while (enumeration.hasMoreElements()) {
                    GrObject child = (GrObject)enumeration.nextElement();
                    if (child instanceof DataSeries) {
                        if (DoubleDataSeries.countDims(child, 2)  &&
                            !(TimeSeries.countDims(child, 2))) {

                            m_crossSectionOutputSeries =
                                (DoubleDataSeries)child;
                        }
                        if (child instanceof TimeSeries) {
                            m_tfOutputSeries = (TimeSeries)child;
                        }
/*
                        if (((DataSeries)child).getNumDimensions() == 6) {
                            m_txyzuvOutputSeries = (DataSeries)child;
                        }
                        if (((DataSeries)child).getNumDimensions() == 4) {
                            m_xyzdOutputSeries = (DataSeries)child;
                        }
*/
                    }
                }

                // Make sure there are no other output series
                // in the output container besided those defined.
                m_outputContainer.clear();
                m_outputContainer.add(m_crossSectionOutputSeries);
                m_outputContainer.add(m_tfOutputSeries);
///                m_outputContainer.add(m_xyzdOutputSeries);

                m_crossSectionOutputSeries.getDimension(0).
                        setProperty("locked", "true");
                m_crossSectionOutputSeries.getDimension(1).
                        setProperty("locked", "true");
                m_tfOutputSeries.getDimension(0).
                        setProperty("locked", "true");
                m_tfOutputSeries.getDimension(1).
                        setProperty("locked", "true");
/*
                for (int dim = 0; dim < 6; dim++) {
                    m_txyzuvOutputSeries.getDimension(dim).
                        setProperty("locked", "true");
                }
                for (int dim = 0; dim < 4; dim++) {
                    m_xyzdOutputSeries.getDimension(dim).
                        setProperty("locked", "true");
                }
*/
                m_crossSectionOutputSeries.setTitle(
                        "Discrete Cross Section");
                m_tfOutputSeries.setTitle(
                        "Time-Flux");
/*
                m_txyzuvOutputSeries.setTitle(
                        "Time-X-Y-Z-U-V");
                m_xyzdOutputSeries.setTitle(
                        "X-Y-Z-Direction");
*/
                m_bHasChanged = true;
                return true;
            }
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Wrapper for removing items from the list of children.
 */
    public GrObject remove(int i) {

        if (m_children.get(i) instanceof GrObjectContainer)
            return null;

        // Let the super class handle removal.
        GrObject child = super.remove(i);

        if (child == m_bath) m_bath = null;
        if (child == m_si3dOut) m_si3dOut = null;
        if (child == m_crossSectionInputSeries)
            m_crossSectionInputSeries = null;

        return child;
    }

//------------------------------------------------------------------------------
/**
 * Removes child gro from the graph.  Does not remove object containers.
 */
    public boolean remove(GrObject gro) {

        if (gro == m_bath) m_bath = null;
        if (gro == m_si3dOut) m_si3dOut = null;
        if (gro == m_crossSectionInputSeries)
            m_crossSectionInputSeries = null;

        if (gro instanceof GrObjectContainer) {
            return false;
        }
        else {
            return super.remove(gro);
        }
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("autoRecalc", new GrProperty(
            m_bAutoRecalc, "Automatic Recalculation"));
        props.put("fluxScale", new GrProperty(m_fluxScale, "Flux Scale"));
        props.put("godinFilter", new GrProperty(
            m_bGodinFilter, "Apply Godin Filter"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

        super.setProperties(props);
        
        boolean bNeedsUpdate = false;

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                if (key.equals("autoRecalc")) {
                    m_bAutoRecalc = (Boolean)p.value;
                    bNeedsUpdate = true;
                }
                if (key.equals("fluxScale")) {
                    m_fluxScale = (Double)p.value;
                    bNeedsUpdate = true;
                }
                if (key.equals("godinFilter")) {
                    m_bGodinFilter = (Boolean)p.value;
                    bNeedsUpdate = true;
                }
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }

        m_bHasChanged = true;
        if (bNeedsUpdate)
            update(this, null);
        
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("autoRecalc"))
            m_bAutoRecalc = Boolean.parseBoolean(value);
        if (key.equals("fluxScale"))
            m_fluxScale = Double.parseDouble(value);
        if (key.equals("godinFilter"))
            m_bGodinFilter = Boolean.parseBoolean(value);
    }

//------------------------------------------------------------------------------
/**
 * Return the DataSeries within the specified ObjectContainer within this
 * object.
 */
    public DataSeries getContainerSeries(String containerTitle) {

        // Look through the children of this object for an ObjectContainer
        // with a matching title.
        GrObjectContainer groc =
            GrObjectContainer.getChildContainer(this, containerTitle);

        if (groc != null) {
            // Look through the contents of the ObjectContainer for
            // a DataSeries.
            Enumeration enumeration = groc.children();
            while (enumeration.hasMoreElements()) {
                GrObject oc_child = (GrObject)enumeration.nextElement();
                if (oc_child instanceof DataSeries)
                    return (DataSeries)oc_child;
            }
        }

        // Could not find a variables series under a container with the
        // given title.
        return null;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void createDiscreteCrossSection () {

        // Initialize the bathymetry member variables, if they haven't been
        // already.
//        if (!m_bBathInitialized) {
//            initializeVariables();
//            if (!m_bBathInitialized)
//                return;
//        }

        if (m_crossSectionInputSeries == null)
            return;

        double[] crossSectionInputArray = m_crossSectionInputSeries.getArray();

        if (crossSectionInputArray.length < 4)
            return;

        // Set the positive U and V directions based on the orientation of
        // the input line.
        if (crossSectionInputArray[3] > crossSectionInputArray[1])
            m_uScale = 1;
        else
            m_uScale = -1;

        if (crossSectionInputArray[2] > crossSectionInputArray[0])
            m_vScale = -1;
        else
            m_vScale = 1;

        double x1, x2, y1, y2;
        if (crossSectionInputArray[0] < crossSectionInputArray[2] ||
            (crossSectionInputArray[0] == crossSectionInputArray[2] &&
            crossSectionInputArray[1] < crossSectionInputArray[3])) {
            x1 = crossSectionInputArray[0] / m_xSpacing;
            y1 = crossSectionInputArray[1] / m_ySpacing;
            x2 = crossSectionInputArray[2] / m_xSpacing;
            y2 = crossSectionInputArray[3] / m_ySpacing;
        }
        else {
            x2 = crossSectionInputArray[0] / m_xSpacing;
            y2 = crossSectionInputArray[1] / m_ySpacing;
            x1 = crossSectionInputArray[2] / m_xSpacing;
            y1 = crossSectionInputArray[3] / m_ySpacing;
        }

        double slope = Double.POSITIVE_INFINITY;
        if (x1 != x2)
            slope = (y2 - y1) / (x2 - x1);
        double b = y1 - (slope * x1);
        int currentI = (int)Math.round(x1);
        int currentJ = (int)Math.round(y1);
        int endI = (int)Math.round(x2);
        int endJ = (int)Math.round(y2);

        double[] crossSectionOutputArray = new double[((endI - currentI)
            + (int)Math.abs(endJ - currentJ) + 1) * 2];
        crossSectionOutputArray[0] = currentI * m_xSpacing;
        crossSectionOutputArray[1] = currentJ * m_ySpacing;
        int outputIndex = 1;
        boolean bNewSegmentFound = false;

        m_segmentList.clear();

        while (currentI != endI || currentJ != endJ) {
            double yAtNextX = slope * ((float)(currentI) + 0.5) + b;
            if (slope > 0 &&
                (yAtNextX > ((float)currentJ + 0.5) || x1 == x2)) {

                CrossSectionSegment segment =
                    createCrossSectionSegment(currentI, currentJ, true);
                if (segment != null)
                    m_segmentList.add(segment);
                currentJ++;
                bNewSegmentFound = true;
            }
            else {
                if (slope < 0 && yAtNextX < ((float)currentJ - 0.5)) {
                    currentJ--;
                    CrossSectionSegment segment =
                        createCrossSectionSegment(currentI, currentJ, true);
                    if (segment != null)
                        m_segmentList.add(segment);
                    bNewSegmentFound = true;
                }
                else {
                    CrossSectionSegment segment =
                        createCrossSectionSegment(currentI, currentJ, false);
                    if (segment != null)
                        m_segmentList.add(segment);
                    currentI++;
                    bNewSegmentFound = true;
                }
            }

            // Add the next point in the cross section output array.
            if (bNewSegmentFound) {
                crossSectionOutputArray[outputIndex * 2] =
                    currentI * m_xSpacing;
                crossSectionOutputArray[outputIndex * 2 + 1] =
                    currentJ * m_ySpacing;
                outputIndex++;
                bNewSegmentFound = false;
            }
        }

        m_crossSectionOutputSeries.setArray(crossSectionOutputArray);
    }
    
//------------------------------------------------------------------------------
/**
 *
 */
    public CrossSectionSegment createCrossSectionSegment (int currentI,
        int currentJ, boolean bUFace) {

        int i1 = currentI;
        int j1 = currentJ;
        int i2 = currentI;
        int j2 = currentJ;
        if (bUFace) {
            i2 = currentI - 1;
        }
        else {
            j2 = currentJ - 1;
        }

        // Find the lookup index for surface and bottom heights.
        int lookup2D1 = -1;
        int lookup2D2 = -1;
        int ijIndex1 = i1 + (j1 * m_numI);
        int ijIndex2 = i2 + (j2 * m_numI);

        // Return null if the coordinate is out of bounds.
        if (ijIndex1 < 0 || m_2DLookupArray == null ||
            ijIndex1 > m_2DLookupArray.length) {
            
            ijIndex1 = -1;
        }
        if (ijIndex2 < 0 || m_2DLookupArray == null ||
            ijIndex2 > m_2DLookupArray.length) {
            
            ijIndex2 = -1;
        }

        if (ijIndex1 < 0 || ijIndex2 < 0) {

             return null;
        }

        // Check that the data is valid and the index is a wet point.
        if (m_2DLookupArray != null) {
            lookup2D1 = m_2DLookupArray[ijIndex1];
            lookup2D2 = m_2DLookupArray[ijIndex2];
        }

        // Return null if the coordinate is land.
        if (lookup2D1 < 0 || lookup2D2 < 0) {
            return null;
        }

        // Create the segment.
        CrossSectionSegment segment = new CrossSectionSegment();
        segment.i = i2;
        segment.j = j2;
        segment.bUFace = bUFace;

        DoubleDataSeries bathSeries = m_bath.getBathymetrySeries();
        int numBathDims = bathSeries.getNumDimensions();
        double[] bathArray = bathSeries.getArray();

        // Calculate the number of layers.
//        segment.numLayers = ((int)((depth - 0.00001) / m_zSpacing)) + 1;
        int numLayers1 =
            (int)Math.round(bathArray[lookup2D1 * numBathDims + 2]);
        int numLayers2 =
            (int)Math.round(bathArray[lookup2D2 * numBathDims + 2]);
        segment.numLayers = numLayers1 < numLayers2 ? numLayers1 : numLayers2;

        // Retrieve the depth at the current location from the array.
        int index1a = 3;
        int index1b = 6;
        int index2a = 4;
        int index2b = 5;
        if (!bUFace) {
            index1a = 3;
            index1b = 4;
            index2a = 5;
            index2b = 6;
        }
        double depth1 = (bathArray[lookup2D1 * numBathDims + index1a] +
            bathArray[lookup2D1 * numBathDims + index1b]) / 2;
        double depth2 = (bathArray[lookup2D2 * numBathDims + index2a] +
            bathArray[lookup2D1 * numBathDims + index2b]) / 2;
        double depth = depth1 < depth2 ? depth1 : depth2;
//        float depth = m_avgDepthArray[lookup2D1] < m_avgDepthArray[lookup2D2] ?
//            m_avgDepthArray[lookup2D1] : m_avgDepthArray[lookup2D2];

        segment.bottomLayerHeight =
            depth - ((segment.numLayers - 1) * m_zSpacing);

        return segment;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void calculateFlux () {

        double[] tfArray = new double[m_numTimeSteps * 2];
        double uFaceArea = m_ySpacing * m_zSpacing;
        double vFaceArea = m_xSpacing * m_zSpacing;
        double vel = 0;
        double layerFlux = 0;

        for (int timeStep = 0; timeStep < m_numTimeSteps; timeStep++) {
            double flux = 0;
            double time = (double)timeStep * m_timeStepInc + m_si3dRunStartTime;
            Iterator iter = m_segmentList.iterator();
            while (iter.hasNext()) {
                CrossSectionSegment segment = (CrossSectionSegment)iter.next();

                for (int layer = 1; layer <= segment.numLayers; layer++) {
                    int currentLookup = getLookupIndex(segment.i, segment.j,
                        layer, m_numI, m_numJ, m_numLayers, m_num3DVarsDims,
                        timeStep, m_num3DWetPoints, m_3DLookupArray);

                    double layerHeight = m_zSpacing;
                    if (layer == segment.numLayers)
                        layerHeight = segment.bottomLayerHeight;
                    if (layer == 1) {

                        double wse1 = getWaterSurfaceElevation(
                            segment.i, segment.j, timeStep);
                        double wse2 = 0;
                        if (segment.bUFace)
                            wse2 = getWaterSurfaceElevation(
                                segment.i + 1, segment.j, timeStep);
                        else
                            wse2 = getWaterSurfaceElevation(
                                segment.i, segment.j + 1, timeStep);

                        layerHeight += ((wse1 + wse2) / 2);
                    }

                    if (segment.bUFace) {
                        vel = getVar3D(currentLookup, 0);
                        vel *= m_unitScales[0];
                        vel *= m_uScale;
                        double faceArea = uFaceArea;
                        if (layer == 1 || layer == segment.numLayers)
                            faceArea = m_ySpacing * layerHeight;
                        layerFlux = vel * faceArea;
                    }
                    else {
                        vel = getVar3D(currentLookup, 1);
                        vel *= m_unitScales[1];
                        vel *= m_vScale;
                        double faceArea = vFaceArea;
                        if (layer == 1 || layer == segment.numLayers)
                            faceArea = m_xSpacing * layerHeight;
                        layerFlux = vel * faceArea;
                    }
                    layerFlux *= m_fluxScale;
                    flux += layerFlux;
                }
            }
            tfArray[timeStep * 2] = time;
            tfArray[timeStep * 2 + 1] = flux;
        }

        m_tfOutputSeries.setReferenceYear(m_si3dRunStartYear);
        m_tfOutputSeries.setArray(tfArray);
//        if (tfArray[0] > 275 && tfArray[tfArray.length - 2] > 355)
//            m_tfOutputSeries.syncToReferenceYear(m_si3dRunStartYear + 1);

        if (m_bGodinFilter) {
            ArrayList dataList = new ArrayList();
            dataList.add(m_tfOutputSeries);
            gov.usgs.sfhydro.gr.tools.GodinFilter gf =
                new gov.usgs.sfhydro.gr.tools.GodinFilter();
            gf.performAction(dataList);
            m_tfOutputSeries.deselectAll();
        }

/*
        // Get the middle segment.
        float[] xyzdArray = new float[4];
        if (m_segmentList.size() > 0) {
            CrossSectionSegment firstSegment = (CrossSectionSegment)m_segmentList.
                get(0);
            CrossSectionSegment midSegment = (CrossSectionSegment)m_segmentList.
                get(m_segmentList.size() / 2);
            CrossSectionSegment lastSegment = (CrossSectionSegment)m_segmentList.
                get(m_segmentList.size() - 1);
            int iDiff = Math.abs(lastSegment.i - firstSegment.i);
            int jDiff = Math.abs(lastSegment.j - firstSegment.j);

            float x = (float)midSegment.i + 0.5;
            float y = (float)midSegment.j + 0.5;
            if (midSegment.bUFace)
                x += 0.5;
            else
                y += 0.5;

            xyzdArray[0] = x * m_xSpacing;
            xyzdArray[1] = y * m_ySpacing;
            xyzdArray[2] = 0;
            xyzdArray[3] = (float)Math.toDegrees(Math.atan2(iDiff * m_vScale,
                jDiff * m_uScale));
        }

        m_xyzdOutputSeries.setArray(xyzdArray);
*/
    }

//------------------------------------------------------------------------------
/**
 *
 */
    protected double getWaterSurfaceElevation(int i, int j, int timeStep) {

        int lookup2D = -1;
        int ijIndex = i + (j * m_numI);
        if (ijIndex < 0 || ijIndex > m_2DLookupArray.length) {
            return 0;
        }

        // Check that the data is valid and the index is a
        // wet point.
        if (m_2DLookupArray != null)
            lookup2D = m_2DLookupArray[ijIndex];

        if (lookup2D < 0)
            return 0;

        double waterSurfaceElevation = getVar2D(lookup2D
            + (timeStep * m_numWet2Dpoints)) * m_wseScale;

        return waterSurfaceElevation;
    }

//------------------------------------------------------------------------------
/**
 * Returns the lookup index of the specified cell for the vars3D array.
 */
    protected int getLookupIndex(int i, int j, int k, int numI, int numJ,
        int numK, int numVarDims, int ts, int numWetPoints, int[] lookupArray) {

        int lookupIndex = -1;

        // Get the bottom value from the lower neighbor cell.
        if (i >= 0 && i < numI && j >= 0 && j < numJ && k >= 0 && k < numK) {

            lookupIndex = lookupArray[i * numK + j * numK * numI + k];
            if (lookupIndex >= 0)
                lookupIndex = (lookupIndex + ts * numWetPoints) * numVarDims;
        }

        return lookupIndex;
    }

//------------------------------------------------------------------------------
/**
 * Returns the U, V, or W components of the specified cell.
 */
    protected float getVar3D(int lookupIndex, int dimNum) {

        float var = 0;

        if (m_3DVarsSeries instanceof RafDataSeries) {
            if (lookupIndex >= 0) {
                var = ((RafDataSeries)m_3DVarsSeries).getValueAt(
                    lookupIndex + dimNum);
            }
        }
        else {
            if (lookupIndex >= 0 &&
                (lookupIndex + dimNum) < m_3DVarsArray.length) {

                var = m_3DVarsArray[lookupIndex + dimNum];
            }
        }

        return var;
    }

//------------------------------------------------------------------------------
/**
 * Returns the 2D variable at the specified offset.
 */
    protected float getVar2D(int offset) {

        float var = 0;

        if (m_2DVarsSeries instanceof RafDataSeries) {
            var = ((RafDataSeries)m_2DVarsSeries).getValueAt(offset);
        }
        else {
            var = m_2DVarsArray[offset];
        }

        return var;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    protected boolean initializeVariables() {

        // Make sure the Bathymetry and Si3dOutput objects are defined.
        if (m_bath == null || m_si3dOut == null) {
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                GrObject gro = (GrObject)iter.next();
                if (gro instanceof Bathymetry) m_bath = (Bathymetry)(gro);
                if (gro instanceof Si3dOutput) m_si3dOut = (Si3dOutput)gro;
            }
        }

        if (m_bath == null)
            return false;

        // Get grid spacing information from the Bathymetry object.
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        m_xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        m_ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        m_zSpacing = (Double)p.value;

        // Find the dimensions of the grid and the number of wet 2D points.
        DataSeries bathSeries = m_bath.getBathymetrySeries();
        m_numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        m_numJ = (int)Math.round(bathSeries.getMax(1)) + 1;
        m_numLayers = (int)Math.round(bathSeries.getMax(2));
        m_numWet2Dpoints = bathSeries.getNumPoints();
        m_avgDepthArray = m_bath.getAverageDepthArray();

        // Get array indexing information from the Bathymetry object.
        m_3DIndexArray = m_bath.get3DIndexArray();
        // Get the rectangular array that holds -1 at dry points and the
        // index into the variables array at wet points.
        m_3DLookupArray = m_bath.get3DLookupArray(m_3DIndexArray);

        m_2DIndexArray = m_bath.get2DIndexArray();
        m_2DLookupArray = m_bath.get2DLookupArray(m_2DIndexArray);
        
        m_bBathInitialized = true;
        
        if (m_si3dOut == null) return false;

        // Get time step information from the Si3dOutput object.
        m_si3dRunStartYear = m_si3dOut.getStartYear();
        m_si3dRunStartTime = m_si3dOut.getJulianStartTime();
        GrHashList si3dOutProps = m_si3dOut.getProperties();
        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long)p.value;
        m_timeStepInc = ((double)tsInSec) / 86400;

        m_3DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("3-D Variables");
//        if (m_3DVarsSeries != null)
//            m_3DVarsArray = m_3DVarsSeries.getArray();
        m_2DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("2-D x-y Variables");
        if (m_2DVarsSeries != null) {
//            m_2DVarsArray = m_2DVarsSeries.getArray();

            DataDimension dim = m_2DVarsSeries.getDimension(0);
            if (dim != null) {
                // Store the scale factor for the dimension.
                GrHashList dimProps = dim.getProperties();
                p = (GrProperty)dimProps.get("unitScale");
                m_wseScale = (Double)p.value;
            }
        }

        // Get the 3D variables array.
        m_num3DVarsDims = m_3DVarsSeries.getNumDimensions();
//        m_3DVarsArray = m_3DVarsSeries.getArray();

        // Find the unit scales for the X, Y, and Z dimensions.
        GrProperty unitProp = (GrProperty)
            (m_3DVarsSeries.getDimension(0).getProperties().get("unitScale"));
        m_unitScales[0] = (Double)unitProp.value;
        unitProp = (GrProperty)
            (m_3DVarsSeries.getDimension(1).getProperties().get("unitScale"));
        m_unitScales[1] = (Double)unitProp.value;
        unitProp = (GrProperty)
            (m_3DVarsSeries.getDimension(2).getProperties().get("unitScale"));
        m_unitScales[2] = (Double)unitProp.value;

        // Create an array to hold Time-X-Y-Z data of the particle at every
        // time step for the duration of the run.
        m_num3DWetPoints = m_3DIndexArray.length / 3;
        m_numTimeSteps = m_3DVarsSeries.getCount() / m_num3DWetPoints;

        return true;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void setSpacing(double spacing) {
        
        m_xSpacing = spacing;
        m_ySpacing = spacing;
    }
    
//------------------------------------------------------------------------------
/**
 *
 */
    public void recalculate() {

        if (!m_bInitialized) {
            if (!initializeVariables())
                return;

            m_bInitialized = true;
        }

        createDiscreteCrossSection();
        calculateFlux();
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        if (m_bAutoRecalc) {
            if (gro == m_crossSectionInputSeries || gro == m_bath ||
                gro == m_si3dOut || gro == this) {

                if (gro == m_bath || gro == m_si3dOut)
                    m_bInitialized = false;

                recalculate();
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

        String s = "Flux Calculator";

        return s;
 	}
}






