/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-05
 * Date started:    2003-05-22
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.sfhydro.si3d.*;
import java.util.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class LagrangianFieldCalculator extends GrObject
    implements java.awt.event.ActionListener, gov.usgs.gr.tools.ThreadWorker {

    protected int m_numWetPoints = 0;
    protected double m_timeStepInc = 0;
    protected double m_si3dRunStartTime = 0;
    protected int[] m_index3d = null;
    protected int m_numTimeSteps = 0;
    protected int m_numTS = 0;
    protected double m_xSpacing = 0;
    protected double m_ySpacing = 0;
    protected double m_zSpacing = 0;
    protected int m_numDims = 0;
    protected ParticleTracker m_tracker = null;
    protected long m_fileOffset = 0;
    protected float[] m_writeCache = null;
    protected double m_uScale = 0;
    protected double m_vScale = 0;
    protected double m_wScale = 0;
    gov.usgs.gr.data.RafDataSeries m_outRafSeries = null;
    protected int m_numTimeSpaceNodes = 0;
    protected int m_timeSpaceNodeOffset = 0;
    protected int m_numWriteCacheNodes = 0;

    protected int m_blockSize = 0;
    protected int m_blockOffset = 0;
    protected int m_numThreads = 1;
    protected int m_numReadCacheSegments = 4;

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public LagrangianFieldCalculator() {

        setDefaults("gov.usgs.sfhydro.si3d.LagrangianFieldCalculator");
    }

//------------------------------------------------------------------------------
/**
 * Add a child object.  Accepts DataSeries, TimeMarker.  Only holds one DS at
 * a time.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

    	if (gro instanceof ParticleTracker) {
            super.add(gro);

            return true;
        }
        if (gro instanceof Bathymetry || gro instanceof Si3dOutput) {
            super.add(gro);

            return true;
        }
        if (gro instanceof GrObjectContainer) {
            super.add(gro);

            return true;
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * numThreads       Integer "Number of Threads"                             <p>
 * numReadCacheSegments       Integer "Number of Read Cache Segments"       <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

        // First add properties from the super class.
        GrHashList props = super.getProperties();

        props.put("numThreads", new GrProperty(
            m_numThreads, "Number of Threads"));
        props.put("numReadCacheSegments", new GrProperty(
            m_numReadCacheSegments,
            "Number of Read Cache Segments"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.sfhydro.si3d.LagrangianFieldCalculator#getProperties
 * getProperties() for a list of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

        GrProperty p;

        // For each property with a string value, pass it to the setProperty()
        // method to be handled there.
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                // Handle string properties by passing them to the
                // setProperty() method.
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                if (key.equals("numThreads")) {
                    int numThreads = (Integer)p.value;
                    if (numThreads > 0 && numThreads < 1000)
                        m_numThreads = numThreads;
                }
                if (key.equals("numReadCacheSegments")) {
                    int numSegs = (Integer)p.value;
                    if (numSegs > 0 && numSegs < 1000)
                        m_numReadCacheSegments = numSegs;
                }
            }
            // Handle string properties by passing them to the
            // setProperty() method.
            if (obj instanceof String) {
                setProperty(key, (String)obj);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.sfhydro.si3d.LagrangianFieldCalculator#getProperties
 * getProperties() for a list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

        if (key.equals("numThreads")) {
            int numThreads = Integer.parseInt(value);
            if (numThreads > 0 && numThreads < 1000)
                m_numThreads = numThreads;
        }
        if (key.equals("numReadCacheSegments")) {
            int numSegs = Integer.parseInt(value);
            if (numSegs > 0 && numSegs < 1000)
                m_numReadCacheSegments = numSegs;
        }
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 */
    public void createField() {

        java.text.SimpleDateFormat dateFormatter = new
            java.text.SimpleDateFormat ("MMM'.' dd',' yyyy 'at' hh:mm:ss");
        java.text.SimpleDateFormat hourFormatter = new
            java.text.SimpleDateFormat ("hh:mm:ss");
        GregorianCalendar gc1 = new GregorianCalendar();
        System.out.println("Creating Lagrangian Vector Field");
        System.out.println("Start time = "
            + dateFormatter.format(gc1.getTime()));

        long startTimeMillis = System.currentTimeMillis();
        java.text.NumberFormat percentForm =
            new java.text.DecimalFormat("0.00");

        Bathymetry bath = null;
        Si3dOutput si3dOut = null;
        m_tracker = null;

        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof ParticleTracker) {
                m_tracker = (ParticleTracker)(gro);
            }
        }

        // Check that the member variables have been initialized.
        if (m_tracker == null) {
            return;
        }

        Enumeration enumeration = m_tracker.children();
        while (enumeration.hasMoreElements()) {
            GrObject gro = (GrObject)enumeration.nextElement();
            if (gro instanceof Bathymetry) {
                bath = (Bathymetry)(gro);
            }
            if (gro instanceof Si3dOutput) {
                si3dOut = (Si3dOutput)gro;
            }
        }

        // Check that the member variables have been initialized.
        if (bath == null || si3dOut == null) {
            return;
        }

        m_index3d = bath.get3DIndexArray();

        // Get grid spacing information from the Bathymetry object.
        GrHashList bathProps = bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        m_xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        m_ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        m_zSpacing = (Double)p.value;

        DataSeries series2d = si3dOut.getContainerSeries("2-D x-y Variables");
        if (series2d == null)
            return;

        if (!(series2d instanceof RafDataSeries))
            return;

        RafDataSeries inRafSeries2D = (RafDataSeries)series2d;
        int numDims2D = inRafSeries2D.getNumDimensions();
        int numTimeSpaceNodes2D = inRafSeries2D.getCount();
        int readCacheSize2D = numTimeSpaceNodes2D * numDims2D;
        inRafSeries2D.createCacheSegments(1, readCacheSize2D);
        inRafSeries2D.fillCacheSegment(0, 0);
        System.out.println("Created 2D read cache of size " + readCacheSize2D
            + " of "
            + (inRafSeries2D.getCount() * inRafSeries2D.getNumDimensions()));

        DataSeries series3d = si3dOut.getContainerSeries("3-D Variables");
        if (series3d == null)
            return;

        if (!(series3d instanceof RafDataSeries))
            return;

        RafDataSeries inRafSeries = (RafDataSeries)series3d;
        m_numDims = inRafSeries.getNumDimensions();
        m_numTimeSpaceNodes = inRafSeries.getCount();

        m_numTimeSteps = series3d.getCount() / (m_index3d.length / 3);

        m_numWetPoints = m_index3d.length / 3;

        DataDimension dim = inRafSeries.getDimension(0);
        GrHashList dimProps = dim.getProperties();
        p = (GrProperty)dimProps.get("unitScale");
        m_uScale = (Float)p.value;

        dim = inRafSeries.getDimension(1);
        dimProps = dim.getProperties();
        p = (GrProperty)dimProps.get("unitScale");
        m_vScale = (Float)p.value;

        dim = inRafSeries.getDimension(2);
        dimProps = dim.getProperties();
        p = (GrProperty)dimProps.get("unitScale");
        m_wScale = (Float)p.value;

        m_si3dRunStartTime = si3dOut.getJulianStartTime();
        GrHashList si3dOutProps = si3dOut.getProperties();
        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long)p.value;

        m_timeStepInc = ((double)tsInSec) / 86400;

        System.out.println("Grid spacing = " + m_xSpacing
            + ", Time Step = " + tsInSec + " seconds");
        System.out.println("Number of Wet Points = " + m_numWetPoints
            + ", Number of Time Steps = " + m_numTimeSteps);
        System.out.println("Number of Time-Space Nodes = " + m_numTimeSpaceNodes
            + ", Number of Threads = " + m_numThreads);

        // Create read cache within the random access file.
        System.out.println("Reading data from disk into memory cache...");
        // Read cache only stores U, V, W
        int readCacheSize = m_numTimeSpaceNodes * m_numDims * 2;

        float[] allocationTestArray = null;
        while (allocationTestArray == null) {
            try {
                allocationTestArray =
                    new float[readCacheSize];
            }
            catch (OutOfMemoryError e) {
                readCacheSize /= 2;
                if (readCacheSize < 100)
                    return;
            }
        }

        allocationTestArray = new float[0];

        readCacheSize -= (readCacheSize % (24 * m_numReadCacheSegments));
        readCacheSize /= 2;
        int segmentSize = readCacheSize / m_numReadCacheSegments;
        int readCacheOffset = 0;
        inRafSeries.createCacheSegments(m_numReadCacheSegments, segmentSize);
        for (int i = 0; i < m_numReadCacheSegments; i++) {
            inRafSeries.fillCacheSegment(i, readCacheOffset);
            readCacheOffset += segmentSize;
        }
        int leadingCacheSegment = 0;
        System.out.println("Created read cache of size " + readCacheSize
            + " of "
            + (inRafSeries.getCount() * inRafSeries.getNumDimensions()));
        float hourSpan = 3600 * (float)readCacheSize / (float)m_numDims
            / (float)m_numWetPoints / (float)tsInSec;
        java.text.NumberFormat spanForm = new java.text.DecimalFormat("0.00");
        System.out.println("Span of " + spanForm.format(hourSpan) + " hours in "
            + m_numReadCacheSegments + " segments of " + segmentSize);

        // Create a write cache for holding results so they can be written
        // back to disk in large blocks.
        int writeCacheSize = segmentSize;
        while (m_writeCache == null) {
            try {
                m_writeCache = new float[writeCacheSize];
            }
            catch (OutOfMemoryError e) {
                writeCacheSize /= 2;
                if (writeCacheSize < 1)
                    return;
            }
        }
        m_writeCache = new float[writeCacheSize];

        m_numWriteCacheNodes = writeCacheSize / m_numDims;
        m_blockSize = m_numWriteCacheNodes / m_numThreads;
        m_blockSize += m_numThreads;

        System.out.println("Created write cache of size " + writeCacheSize
            + " of "
            + (inRafSeries.getCount() * inRafSeries.getNumDimensions()));

        GrHashList props = inRafSeries.getProperties();
        p = (GrProperty)props.get("file");
        String oldFilename = (String)p.value;
        String newFilename = oldFilename.replaceFirst(".bin", "_lg.bin");
        if (oldFilename.equals(newFilename)) {
            System.out.println(
                "3D velocity filename did not contain .bin.  " +
                "Exiting without filtering.");

            return;
        }

        m_outRafSeries = new gov.usgs.gr.data.RafDataSeries(m_numDims);
        m_outRafSeries.setProperty("file", newFilename);
        m_outRafSeries.openRAF("rwd");

        System.out.println("Writing Lagrangian velocities to " + newFilename);

        if (!m_tracker.initializeVariables())
            return;

        m_numTS = (int)Math.round(1.03333333333333 / m_timeStepInc);

        // Roughly m/s to m/day
        double outputScale = 1.0E-5f;
        m_uScale = outputScale / m_uScale;
        m_vScale = outputScale / m_vScale;
        m_wScale = outputScale / m_wScale;

        int previousNodeOffset = 0;
        int fileOffset = 0;

        System.out.println("Block Size = " + m_blockSize);

        double duration = 1.033333333333333 + m_timeStepInc;
        while (m_timeSpaceNodeOffset < m_numTimeSpaceNodes) {
            m_blockOffset = 0;
            // Filter the data using parallel processing.
            MultiThreader multiThreader =
                new MultiThreader(this, m_numThreads, "Lagrangian");
            multiThreader.runThreads();

            // Determine the number of nodes that were processed and ready
            // to be written to the file.
            int numNodesProcessed = m_timeSpaceNodeOffset - previousNodeOffset;
            int writeSize = numNodesProcessed * m_numDims;

            // Transfer the contents of the writeCache array to disk.
            System.out.println("Writing data to disk...");

            m_outRafSeries.writeValuesFrom(m_writeCache, 0, fileOffset,
                writeSize);

            double percentage = 100 * (float)m_timeSpaceNodeOffset
                / (double)m_numTimeSpaceNodes;
            gc1 = new GregorianCalendar();
            String timeString = hourFormatter.format(gc1.getTime());
            System.out.println("Processed through Time-Space Node "
                + m_timeSpaceNodeOffset + " of " + m_numTimeSpaceNodes
                + " (" + percentForm.format(percentage) + "%) "
                + timeString);

            // Advance the file offset for the next iteration.
            fileOffset += writeSize;
            previousNodeOffset = m_timeSpaceNodeOffset;

            while (fileOffset > (readCacheOffset -
                ((m_numReadCacheSegments - 1) * segmentSize))) {

                // Leapfrog the read cache as soon as the first one is finished.
                inRafSeries.fillCacheSegment(
                    leadingCacheSegment, readCacheOffset);
                System.out.println("Switched cache segment "
                    + leadingCacheSegment + " to offset " + readCacheOffset);
                readCacheOffset += segmentSize;
                leadingCacheSegment++;
                if (leadingCacheSegment >= m_numReadCacheSegments)
                    leadingCacheSegment = 0;
            }
        }

        // Delete the read cache and close the random access file.
        inRafSeries2D.createCacheSegments(0, 0);
        inRafSeries.createCacheSegments(0, 0);
        m_outRafSeries.closeRAF();


        // Calculate the elapsed runtime.
        long endTimeMillis = System.currentTimeMillis();

        long totalTimeSeconds = (endTimeMillis - startTimeMillis) / 1000;
        long hours = totalTimeSeconds / 3600;
        totalTimeSeconds -= hours * 3600;
        long minutes = totalTimeSeconds / 60;
        long seconds = totalTimeSeconds - (minutes * 60);

        gc1 = new GregorianCalendar();
        System.out.println("" + dateFormatter.format(gc1.getTime()));

        java.text.NumberFormat form = new java.text.DecimalFormat("00");
        System.out.println("Time elapsed while filtering file: "
            + form.format(hours) + ":" + form.format(minutes) + ":"
            + form.format(seconds));
    }

//------------------------------------------------------------------------------
/**
 * Filters one block of wet points.
 */
    public boolean performThreadWork() {

        // Create an array for storing the particle path.
        double[] path = new double[4 * m_numTS];

        int blockOffset;
        int timeSpaceNodeOffset;
        int blockSize;

        // Grab the next chunk of work that is left to be done.
        // Return false if no nodes are left to process, or if the
        // write cache has been filled.
        synchronized(this) {
            if (m_blockOffset >= m_numWriteCacheNodes
                || m_timeSpaceNodeOffset >= m_numTimeSpaceNodes) {

                return false;
            }

            // Determine where to start and how many nodes to process.
            // It will be the full amount unless it is the last block
            // at the end of the write cache or the end of all the nodes.
            blockOffset = m_blockOffset;
            timeSpaceNodeOffset = m_timeSpaceNodeOffset;
            blockSize = m_blockSize;
            if ((blockOffset + m_blockSize) > m_numWriteCacheNodes)
                blockSize = m_numWriteCacheNodes - blockOffset;
            if ((timeSpaceNodeOffset + blockSize) > m_numTimeSpaceNodes)
                blockSize = m_numTimeSpaceNodes - timeSpaceNodeOffset;

            // Update the shared offset variables so other methods and
            // threads know how far the node processing has proceeded.
            m_blockOffset += blockSize;
            m_timeSpaceNodeOffset += blockSize;
        }

        int lastPrintedIndex = timeSpaceNodeOffset - 1;

        for (int i = 0; i < blockSize; i++) {
            int writeCacheOffset = (i + blockOffset) * m_numDims;
            int ts = (i + timeSpaceNodeOffset) / m_numWetPoints;
            // Don't bother tracking particles if the start time is within
            // a tidal day of the end of the run.
            if ((ts + m_numTS) > m_numTimeSteps) {
                for (int n = 0; n < m_numDims; n++) {
                    m_writeCache[writeCacheOffset + n] = 0;
                }
                continue;
            }

            // Set up the starting position and time for tracking a particle
            // from this time-space node.
            double startTime = ts * m_timeStepInc + m_si3dRunStartTime;
            double u = 0;
            double v = 0;
            double w = 0;

            int wetPointOffset = (timeSpaceNodeOffset + i) % m_numWetPoints;
            int currentI = m_index3d[wetPointOffset * 3];
            int currentJ = m_index3d[wetPointOffset * 3 + 1];
            int currentK = m_index3d[wetPointOffset * 3 + 2];

            double xStart = ((double)currentI) * m_xSpacing;
            double yStart = ((double)currentJ) * m_ySpacing;
            double zStart = -((double)currentK - 0.5) * m_zSpacing;

            path[(m_numTS - 1) * 4] = -1;
            path[(m_numTS - 1) * 4 + 1] = -1;
            path[(m_numTS - 1) * 4 + 2] = -1;
            path[(m_numTS - 1) * 4 + 3] = -1;

            // Track the particle.
            path = m_tracker.track (0, m_numTS - 1, ts, path,
                xStart, yStart, zStart);

            // Compute the difference between the starting position and the
            // position one tidal day later.
            if (path[(m_numTS - 1) * 4] > 0) {
                double endX = path[(m_numTS - 1) * 4 + 1];
                double endY = path[(m_numTS - 1) * 4 + 2];
                double endZ = path[(m_numTS - 1) * 4 + 3];
                if (endX >= 0 && endY >= 0) {
                    u = endX - xStart;
                    v = endY - yStart;
                    w = endZ - zStart;
                }
            }

            // Put the results into the write cache array.
            try {
                m_writeCache[writeCacheOffset] = (float)(u * m_uScale);
                m_writeCache[writeCacheOffset + 1] = (float)(v * m_vScale);
                m_writeCache[writeCacheOffset + 2] = (float)(w * m_wScale);

                for (int n = 3; n < m_numDims; n++) {
                    m_writeCache[writeCacheOffset + n] = 0;
                }
            }
            catch (Exception e) {
                e.printStackTrace();
            }

            if ((i % (m_numWetPoints * 50)) == 0 && i > 0) {
                System.out.println("Processed " + (lastPrintedIndex + 1)
                    + " to " + (timeSpaceNodeOffset + i) + "  -- Thread "
                    + Thread.currentThread().getName());
                lastPrintedIndex = timeSpaceNodeOffset + i;
            }
        }

        System.out.println("Processed " + (lastPrintedIndex + 1) + " to "
            + (timeSpaceNodeOffset + blockSize - 1) + "  -- Thread "
            + Thread.currentThread().getName());

        return true;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void createLagrangianProfile() {


        Bathymetry bath = null;
        Si3dOutput si3dOut = null;
        m_tracker = null;

        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof ParticleTracker) {
                m_tracker = (ParticleTracker)(gro);
            }
        }

        // Check that the member variables have been initialized.
        if (m_tracker == null)
            return;

        if (!m_tracker.initializeVariables())
            return;

        DoubleDataSeries xyInputSeries =
            m_tracker.getContainerSeries("X-Y Input Series");

        if (xyInputSeries == null || xyInputSeries.getCount() < 1)
            return;

        double[] xyInputArray = xyInputSeries.getArray();
        double xVal = xyInputArray[0];
        double yVal = xyInputArray[1];

        Enumeration enumeration = m_tracker.children();
        while (enumeration.hasMoreElements()) {
            GrObject gro = (GrObject)enumeration.nextElement();
            if (gro instanceof Bathymetry) {
                bath = (Bathymetry)(gro);
            }
            if (gro instanceof Si3dOutput) {
                si3dOut = (Si3dOutput)gro;
            }
        }

        // Check that the member variables have been initialized.
        if (bath == null || si3dOut == null) {
            return;
        }

        m_index3d = bath.get3DIndexArray();

        // Get grid spacing information from the Bathymetry object.
        GrHashList bathProps = bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        double xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        double ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        double zSpacing = (Double)p.value;

        DataSeries series3d = si3dOut.getContainerSeries("3-D Variables");
        if (series3d == null)
            return;

        // Find the lookup index for surface and bottom heights.
        int lookup2D = -1;
        int currentI = (int)Math.round(xVal / xSpacing);
        int currentJ = (int)Math.round(yVal / ySpacing);
        DataSeries bathSeries = bath.getBathymetrySeries();
        int numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        int[] index2D = bath.get2DIndexArray();
        int[] lookupArray2D = bath.get2DLookupArray(index2D);
        int ijIndex = currentI + (currentJ * numI);
        if (ijIndex < 0 || ijIndex > lookupArray2D.length) {
            // Write (0, 0, 0) and move to the next time step.
            return;
        }

        // Check that the data is valid and the index is a wet point.
        if (lookupArray2D != null)
            lookup2D = lookupArray2D[ijIndex];

        if (lookup2D < 0) {
            // Write (0, 0, 0) and move to the next time step.
            return;
        }

        // Retrieve the depth at the current location from the array.
        double[] depthArray = bath.getAverageDepthArray();
        double depth = depthArray[lookup2D];

        double surfaceHeightAdjustment = 0;

        // Calculate the bottom adjustment in world units.
        // It is zero if K is not a bottom layer.
        double bottomLayerThickness = depth % zSpacing;
        if (bottomLayerThickness <= 0)
            bottomLayerThickness = zSpacing;
        double bottomAdjustment = zSpacing - bottomLayerThickness;

        FloatDataSeries varsSeries2D = (FloatDataSeries)
            si3dOut.getContainerSeries("2-D x-y Variables");
        DataDimension dim = varsSeries2D.getDimension(0);
        double wseScale = 1;
        if (dim != null) {
            // Store the scale factor for the dimension.
            GrHashList dimProps = dim.getProperties();
            p = (GrProperty)dimProps.get("unitScale");
            wseScale = (Double)p.value;
        }

        m_si3dRunStartTime = si3dOut.getJulianStartTime();
        GrHashList si3dOutProps = si3dOut.getProperties();
        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long)p.value;

        GrHashList trackerProps = m_tracker.getProperties();
        p = (GrProperty)trackerProps.get("tsDivs");
        int tsDivs = (Integer)p.value;

        m_timeStepInc = ((double)tsInSec) / 86400;
        m_timeStepInc /= (double)tsDivs;

        m_numTimeSteps = series3d.getCount() / (m_index3d.length / 3);
        m_numTimeSteps *= (double)tsDivs;
        m_numTS = (int)Math.round(1.03333333333333 / m_timeStepInc) + 1;
        m_numTimeSteps -= m_numTS;
        int numLayers = ((int)((depth - 0.00001) / zSpacing)) + 1;
        double[] path = new double[4 * m_numTS];
        int numHourlyDivs = (int)(1.0 / 24.0 / m_timeStepInc);
        double[] tzuvArray =
            new double[(m_numTimeSteps / numHourlyDivs) * numLayers * 4];

        double cmsScale = 100 / 1.03333333333333 / 86400;

        for (int ts = 0; ts < m_numTimeSteps; ts += numHourlyDivs) {
            double startTime = ts * m_timeStepInc + m_si3dRunStartTime;

            // Calculate the surface adjustment in world units.
            // It is zero if K is not a surface layer.
//            if (lookup2D >= 0 && varsSeries2D != null) {
//
//                try {
//                    surfaceHeightAdjustment = getVar2D(lookup2D
//                        + (ts * bathSeries.getNumPoints()), varsSeries2D);
//                    surfaceHeightAdjustment *= wseScale;
//                }
//                catch (Exception e) {
//                    System.err.println("Exception at ts = " + ts);
//                    e.printStackTrace();
//                }
//            }

            for (int currentK = 0; currentK < numLayers; currentK++) {
                double z = (double)(numLayers - currentK) * zSpacing;
                if (currentK == (numLayers - 1))
                    z -= (bottomAdjustment * 0.5);
                else
                    z -= (zSpacing * 0.5);
// Gets tidally filtered later, so don't bother with height.
//                if (currentK == 0)
//                    z += (surfaceHeightAdjustment * 0.5);

                // Set up the starting position and time for tracking a particle
                // from this time-space node.
                double u = 0;
                double v = 0;
                double w = 0;

                double xStart = ( (double) currentI) * xSpacing;
                double yStart = ( (double) currentJ) * ySpacing;
                double zStart = - ( (double) currentK - 0.5) * zSpacing;

                path[ (m_numTS - 1) * 4] = -1;
                path[ (m_numTS - 1) * 4 + 1] = -1;
                path[ (m_numTS - 1) * 4 + 2] = -1;
                path[ (m_numTS - 1) * 4 + 3] = -1;

                // Track the particle.
                path = m_tracker.track(0, m_numTS - 1, ts, path,
                                       xStart, yStart, zStart);

                // Compute the difference between the starting position and the
                // position one tidal day later.
                if (path[ (m_numTS - 1) * 4] > 0) {
                    double endX = path[ (m_numTS - 1) * 4 + 1];
                    double endY = path[ (m_numTS - 1) * 4 + 2];
                    double endZ = path[ (m_numTS - 1) * 4 + 3];
                    if (endX >= 0 && endY >= 0) {
                        u = endX - xStart;
                        v = endY - yStart;
                        w = endZ - zStart;

                        // Convert from meters per tidal day to cm/s
                        u *= cmsScale;
                        v *= cmsScale;
                        w *= cmsScale;
                    }
                }
                int tzuvOffset =
                    (ts / numHourlyDivs) * numLayers * 4 + currentK * 4;
                tzuvArray[tzuvOffset] = startTime;
                tzuvArray[tzuvOffset + 1] = z;
                tzuvArray[tzuvOffset + 2] = u;
                tzuvArray[tzuvOffset + 3] = v;
            }
        }

        DoubleDataSeries tzuvSeries = getContainerSeries("Profile Output");
        if (tzuvSeries == null) {
            tzuvSeries = new DoubleDataSeries(tzuvArray, 4);
            GrObjectContainer groc = new GrObjectContainer();
            groc.setName("Profile Output");
            groc.setProperty("title", "Profile Output");
            groc.add(tzuvSeries);
            add(groc);
        }
        else {
            tzuvSeries.setArray(tzuvArray);
        }

        gov.usgs.sfhydro.gr.tools.GodinFilter gf =
            new gov.usgs.sfhydro.gr.tools.GodinFilter();
        gf.setProperty("parabolicInterpolation", "true");
        gf.setProperty("fDims", "2,3");
        ArrayList seriesList = new ArrayList();
        seriesList.add(tzuvSeries);
//        gf.performAction (seriesList);
        tzuvSeries.deselectAll();
    }

//------------------------------------------------------------------------------
/**
 * Return the DoubleDataSeries within the specified ObjectContainer within this
 * object.
 * @param containerTitle    The title of the container.
 * @return  The series inside the specified container, null if not found.
 */
    public DoubleDataSeries getContainerSeries(String containerTitle) {

        // Look through the children of this object for an ObjectContainer
        // with a matching title.
        GrObjectContainer groc =
            GrObjectContainer.getChildContainer(this, containerTitle);

        if (groc != null) {
            // Look through the contents of the ObjectContainer for
            // a DoubleDataSeries.
            Enumeration enumeration = groc.children();
            while (enumeration.hasMoreElements()) {
                GrObject oc_child = (GrObject)enumeration.nextElement();
                if (oc_child instanceof DoubleDataSeries)
                    return (DoubleDataSeries)oc_child;
            }
        }

        // Could not find a variables series under a container with the
        // given title.
        return null;
    }

//------------------------------------------------------------------------------
/**
 * Returns the 2D variable at the specified offset.
 */
    protected float getVar2D(int offset, RafDataSeries varsSeries2D) {

        float var = 0;

//        if (varsSeries2D instanceof RafDataSeries) {
            var = ((RafDataSeries)varsSeries2D).getValueAt(offset);
//        }
//        else {
//            var = varsSeries2D.getArray()[offset];
//        }

        return var;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

    	m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * @return  A menu item that can be placed on a menu and send an ActionEvent
 *          back to this object when it is chosen.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"Lagrangian Calculator");
        javax.swing.JMenuItem item = new javax.swing.JMenuItem(
            "Create Lagrangian Field");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Create Lagrangian Profile");
        item.addActionListener(this);
        subMenu.add(item);

    	return subMenu;
    }

//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Create Lagrangian Field")) {
        	createField();
        }
        if (srcText.equals("Create Lagrangian Profile")) {
            createLagrangianProfile();
        }
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s =  "Lagrangian Calculator";

        return s;
    }
}






