/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-05
 * Date started:    2002-05-15
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.util.*;

//==============================================================================
/**
 * Calculates the path that a particle would take through a time-variable
 * velocity field.
 */

public class ParticlePathCalculator extends GrObject {
    
    /**
     * The threshold below which a number is assumed to be zero for calculation
     * purposes.
     */
    public double m_nearZero = 1E-11;
    /**
     * The scales used for multiplying the U, V, and W values when they are
     * read from the data file.
     */
    public double[] m_unitScales = {1, 1, 1};
    /** The Water Surface Elevation scale */
    public double m_wseScale = 1;
    /** An array containing the average depth for each water column. */
    public double[] m_avgDepthArray = null;
    /** The number of wet columns on the grid. */
    public int m_numWet2Dpoints = 0;
    /** The I-J indexes for each water column. */
    public int[] m_2DIndexArray = null;
    /** The indexes for retrieving 2D variables from the wet-only 2D array. */
    public int[] m_2DLookupArray = null;
    /** The data series containing data types that vary by time, X, and Y. */
    public RafDataSeries m_2DVarsSeries = null;
    /** The array stored in the 2D Vars Series. */
    public float[] m_2DVarsArray = null;
    /** The number of wet cells in the grid. */
    public int m_numWet3Dpoints = 0;
    /** The I-J-K indexes for each wet cell. */
    public int[] m_3DIndexArray = null;
    /** The index into the wet array for each cell on the grid.  -1 if dry. */
    public int[] m_3DLookupArray = null;
    /** The data series containing data types that vary by time, X, Y and Z. */
    public RafDataSeries m_3DVarsSeries = null;
    /** The array stored in the 3D Vars Series. */
    public float[] m_3DVarsArray = null;
    /** The number of dimensions for the 3D output. */
    public int m_num3DVarsDims = 0;
    
    /** X-spacing of the grid. */
    public double m_xSpacing = 1;
    /** Y-spacing of the grid. */
    public double m_ySpacing = 1;
    /** Z-spacing of the grid. */
    public double m_zSpacing = 1;
    /** Time of the start of the Si3D run. */
    public double m_si3dRunStartTime = 0;
    /** Si3D run time step increment. */
    public double m_timeStepInc = 1;
    /** The number of time steps in the Si3D output. */
    public int m_numTimeSteps = 0;
    /** The number of grid cells in the X direction. */
    public int m_numI = -1;
    /** The number of grid cells in the Y direction. */
    public int m_numJ = -1;
    /** The number of grid cells in the vertical direction. */
    public int m_numLayers = -1;
    
    /** Whether to save certain variables on the initial time step. */
    public boolean m_bSaveInitialRange = false;
    /** The initial Water Surface Elevation. */
    public double m_initialWSE = 0;
    /** The initial grid depth. */
    public double m_initialGridDepth = 0;
    
    /** Whether to use the W component of velocity in calculations. */
    public boolean m_bUseW = true;
    /** Whether to calculate drift. */
    public boolean m_bCalcDrift = true;
    /** Whether to calculate turbulence. */
    public boolean m_bCalcTurbulence = false;
    /** The boundaries of the grid. */
    public double[][] m_boundaries = null;
    /** The barriers of the grid. */
    public double[][] m_barriers = null;
    
    /**
     * The number of times to divide Si3D vector field model output to arrive
     * at the particle tracking time step.
     */
    public int m_tsDivs = 1;
    /**
     * Whether to only report the time the particle was defined instead of a
     * complete track.
     */
    protected boolean m_bTrackTimeOnly = false;
    /** And array of descriptions for the available types of interpolation. */
    public String[] m_interpolationTypes =
        new String[] {"None", "Linear", "Parabolic"};
    /**
     * The numeric code of the interpolation to be used.
     * 0 = none
     * 1 = linear
     * 2 = parabolic
     */
    public int m_interpolationIndex = 0;
    
    /** A cell velocity object used for intermediate calculations. */
    public CellDescriptor m_tmpCell1 = new CellDescriptor();
    /** A cell velocity object used for intermediate calculations. */
    public CellDescriptor m_tmpCell2 = new CellDescriptor();
    /** A cell velocity object used for intermediate calculations. */
    public CellDescriptor m_tmpCell3 = new CellDescriptor();
    /** A cell velocity object used for intermediate calculations. */
    public CellDescriptor m_tmpCell4 = new CellDescriptor();
    /** A cell velocity object used for intermediate calculations. */
    public CellDescriptor m_tmpCellLeft = new CellDescriptor();
    /** A cell velocity object used for intermediate calculations. */
    public CellDescriptor m_tmpCellRight = new CellDescriptor();
    
    protected String m_errorString = null;
    
    //==========================================================================
    /**
     * A data structure class for holding the I-J-K location and height of
     * a cell.
     */
    public class CellDescriptor {
        int i = 0;
        int j = 0;
        int k = 0;
        int ptmTS = -1;
        int lookup2D = -1;
        CellVariables vars = new CellVariables();
        ColumnVariables vars2D = new ColumnVariables();
        
        //----------------------------------------------------------------------
        /**
         * Determines if another object is equal to this one.
         * @param cell The object to compare to this one.
         * @return True if i, j, k, and ptmTS of the other cell matches this one
         */
        public boolean equals(CellDescriptor cell) {
            if (cell.ptmTS == ptmTS &&
                cell.i == i && cell.j == j && cell.k == k) {
                
                return true;
            } else {
                return false;
            }
        }
        
        //----------------------------------------------------------------------
        /**
         * Copys the local member values from this object into the argument
         * object.
         * @param cell The object to copy values into.
         */
        public void copyDescriptionInto(CellDescriptor cell) {
            cell.i = i;
            cell.j = j;
            cell.k = k;
            cell.ptmTS = ptmTS;
            cell.lookup2D = lookup2D;
        }
        
        //----------------------------------------------------------------------
        /**
         * Copys all the values from this object into the argument object.
         * @param cell The object to copy values into.
         */
        public void copyAllInto(CellDescriptor cell) {
            copyDescriptionInto(cell);
            vars.copyInto(cell.vars);
            vars2D.copyInto(cell.vars2D);
        }
    }
    
    //==========================================================================
    /**
     * A data structure class for holding the 2D variables for a water coloumn.
     */
    public class ColumnVariables  {
        double wseLocation = 0;
        double bottomLocation = 0;
        double wse = 0;
        double depth = 0;
        double surfaceHeightAdjustment = 0;
        double bottomAdjustment = 0;
        int bottomLayerK = 1;
        
        //----------------------------------------------------------------------
        /**
         * Copys all the values from this object into the argument object.
         * @param vars2D The object to copy values into.
         */
        public void copyInto(ColumnVariables vars2D) {
            vars2D.wseLocation = wseLocation;
            vars2D.bottomLocation = bottomLocation;
            vars2D.wse = wse;
            vars2D.depth = depth;
            vars2D.surfaceHeightAdjustment = surfaceHeightAdjustment;
            vars2D.bottomAdjustment = bottomAdjustment;
            vars2D.bottomLayerK = bottomLayerK;
        }
    }
    
    //==========================================================================
    /**
     * A data structure class for holding the velocities at the six sides of
     * a cell, as well as other variables.
     */
    public class CellVariables  {
        double uEast = 0;
        double uWest = 0;
        double vNorth = 0;
        double vSouth = 0;
        double wTop = 0;
        double wBottom = 0;
        double zEddyDiffusivityTop = 0;
        double zEddyDiffusivityBottom = 0;
        double layerHeight = 0;
        int[] lookups = new int[4];
        
        //----------------------------------------------------------------------
        /**
         * Copys all the values from this object into the argument object.
         * @param vars The variables object to copy values into.
         */
        public void copyInto(CellVariables vars) {
            vars.uEast = uEast;
            vars.uWest = uWest;
            vars.vNorth = vNorth;
            vars.vSouth = vSouth;
            vars.wTop = wTop;
            vars.wBottom = wBottom;
            vars.zEddyDiffusivityTop = zEddyDiffusivityTop;
            vars.zEddyDiffusivityBottom = zEddyDiffusivityBottom;
            vars.layerHeight = layerHeight;
            vars.lookups[0] = lookups[0];
            vars.lookups[1] = lookups[1];
            vars.lookups[2] = lookups[2];
            vars.lookups[3] = lookups[3];
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Creates an empty object.
     */
    public ParticlePathCalculator() {
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the current values for this object.
     * Properties include: <p>
     * useW                  Boolean     "Use W"                             <p>
     * tsDivs                Integer     "Time Step Divisions"               <p>
     * interpolationType     Integer     "Interpolation"                     <p>
     *      0 = none, 1 = linear, 2 = parabolic                              <p>
     * calcDrift             Boolean     "Calculate Vertical Drift"          <p>
     * calcTurbulence        Boolean     "Calculate Vertical Turbulence"     <p>
     * @return  A list of GrProperty objects.
     */
    public GrHashList getProperties() {
        
        GrHashList props = super.getProperties();
        
        props.put("useW", new GrProperty(m_bUseW, "Use W"));
        props.put("tsDivs", new GrProperty(m_tsDivs,
            "Time Step Divisions"));
        
        // Give the user some symbol choices...
        GrProperty p = new GrProperty("", "Interpolation");
        javax.swing.JComboBox jcb = new javax.swing.JComboBox();
        
        // Fill the symbol type combo box.
        for (int i = 0; i < m_interpolationTypes.length; i++) {
            String name = m_interpolationTypes[i];
            jcb.addItem(name);
            if (m_interpolationIndex == i) {
                jcb.setSelectedIndex(i);
                p.value = m_interpolationTypes[i];
            }
        }
        
        p.component = jcb;
        props.put("interpolationType", p);
        props.put("calcDrift", new GrProperty(m_bCalcDrift,
            "Calculate Vertical Drift"));
        props.put("calcTurbulence",
            new GrProperty(m_bCalcTurbulence, "Calculate Vertical Turbulence"));
        
        return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Set the values of one or more of this object's properties.
     * @see gov.usgs.sfhydro.si3d.ParticleTracker#getProperties getProperties() for
     * a list of accepted properties.
     * @param props A list of GrProperty objects specifying the new values for
     *              certain properties.
     */
    public void setProperties(HashMap props) {
        
        super.setProperties(props);
        
        GrProperty p;
        
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;
                
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                if (key.equals("useW"))
                    m_bUseW = (Boolean)p.value;
                if (key.equals("tsDivs")) {
                    int tsDivs = (Integer)p.value;
                    if (tsDivs > 0)
                        m_tsDivs = tsDivs;
                }
                if (key.equals("interpolationType")) {
                    // Check each name.
                    int numTypes = m_interpolationTypes.length;
                    for (int i = 0; i < numTypes; i++) {
                        String name = ((String)p.value).trim().toLowerCase();
                        if (name.equals(m_interpolationTypes[i].
                            toLowerCase())) {
                            
                            m_interpolationIndex = i;
                            break;
                        }
                    }
                }
                if (key.equals("calcDrift"))
                    m_bCalcDrift = (Boolean)p.value;
                if (key.equals("calcTurbulence"))
                    m_bCalcTurbulence = (Boolean)p.value;
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }
        
        m_bHasChanged = true;
        notifyObservers();
    }
    
//------------------------------------------------------------------------------
    /**
     * Set a single property.
     * @see gov.usgs.sfhydro.si3d.ParticleTracker#getProperties getProperties() for
     * a list of accepted properties.
     * @param key   The property name.
     * @param value The new value for the property.
     */
    public void setProperty(String key, String value) {
        
        super.setProperty(key, value);
        
        if (key.equals("useW"))
            m_bUseW = Boolean.parseBoolean(value);
        if (key.equals("tsDivs")) {
            int tsDivs = Integer.parseInt(value);
            if (tsDivs > 0)
                m_tsDivs = tsDivs;
        }
        if (key.equals("interpolationType")) {
            // Check each name.
            for (int i = 0; i < m_interpolationTypes.length; i++) {
                String name = value.trim().toLowerCase();
                if (name.equals(m_interpolationTypes[i].toLowerCase())) {
                    m_interpolationIndex = i;
                    break;
                }
            }
        }
        if (key.equals("calcDrift"))
            m_bCalcDrift = Boolean.parseBoolean(value);
        if (key.equals("calcTurbulence"))
            m_bCalcTurbulence = Boolean.parseBoolean(value);
    }
    
//------------------------------------------------------------------------------
    /**
     * Wrapper for the basic tracking method.
     * Converts start time and duration to start and end time steps and allocates
     * the array to hold the path.  Fills the portion of the array before and after
     * the start and end.
     * @param startTime The start time in Julian Days relative to the same year
     *                  as the Si3D output.
     * @param duration  The length of time to track the particle.
     * @param xStart    The X coordinate of the release point.
     * @param yStart    The Y coordinate of the release point.
     * @param zStart    The Z coordinate of the release point.
     * @return  A data series containing the Time-X-Y-Z particle path.  The number
     *          of time steps is (the number of Si3D output time steps minus one)
     *          multiplied by (the number of time step divisions) plus one.
     */
    protected DoubleDataSeries track(double startTime, double duration,
        double xStart, double yStart, double zStart) {
        
        // Don't proceed if there is not enough vector input data.
        if (m_numTimeSteps < 1)
            return null;
        
        double ptmTimeStepInc = m_timeStepInc / (double)m_tsDivs;
        int ptmNumTimeSteps = (m_numTimeSteps - 1) * m_tsDivs + 1;
        
        // Convert the start time to the nearest time step.
        int startTS = (int)Math.round((startTime - m_si3dRunStartTime)
        / ptmTimeStepInc + 1);
        
        // Perform bounds checking on the start time step.
        if (startTS < 1)
            startTS = 1;
        if (startTS > ptmNumTimeSteps)
            startTS = ptmNumTimeSteps;
        
        // Convert the end time to the nearest time step.
        int endTS = startTS +
            (int)Math.round(duration / ptmTimeStepInc + 1);
        
        // Perform bounds checking on the end time step.
        if (endTS < startTS || endTS >= ptmNumTimeSteps)
            endTS = ptmNumTimeSteps - 1;
        
        // Allocate an array to hold the path at the PTM time step.
        double[] path = new double[4 * ptmNumTimeSteps];
        
        // Fill in the points leading up to the start.
        for (int ts = 0; ts < (startTS - 1); ts++) {
            path[ts * 4] = ts * ptmTimeStepInc + m_si3dRunStartTime;
            path[ts * 4 + 1] = -1;
            path[ts * 4 + 2] = -1;
            path[ts * 4 + 3] = -1;
        }
        
        // Fill in the points after the end.
        for (int ts = (endTS + 1); ts < ptmNumTimeSteps; ts++) {
            path[ts * 4] = ts * ptmTimeStepInc + m_si3dRunStartTime;
            path[ts * 4 + 1] = -1;
            path[ts * 4 + 2] = -1;
            path[ts * 4 + 3] = 3;
        }
        
        // Initialize the first time step with the starting location of the
        // particle.
        path[(startTS - 1) * 4] = startTime;
        path[(startTS - 1) * 4 + 1] = xStart;
        path[(startTS - 1) * 4 + 2] = yStart;
        path[(startTS - 1) * 4 + 3] = zStart;
        
        // Track the particle.
        path = track(startTS, endTS, 0, path, xStart, yStart, zStart);
        
        return (new DoubleDataSeries(path, 4));
    }
    
//------------------------------------------------------------------------------
    /**
     * Track a particle through the Si3D vector field from the X-Y-Z release point
     * and starting time step until the ending time step.
     * @param startTS   The starting time step for calculations
     *                  (one past the time with the release point).
     * @param endTS     The ending time step.
     * @param offset    The time step at which the path array should start.
     * @param path      The array that will hold the Time-X-Y-Z path at each time.
     * @param xStart    The X coordinate of the release point.
     * @param yStart    The Y coordinate of the release point.
     * @param zStart    The Z coordinate of the release point.
     * @return The path array with the particle path coordinates in it.
     */
    protected double[] track(int startTS, int endTS, int offset, double[] path,
        double xStart, double yStart, double zStart) {
        
        // Create an array to hold the current particle coordinates measured
        // in cell units.
        double[] partCoords = {0, 0, 0};
        partCoords[0] = xStart / m_xSpacing;
        partCoords[1] = yStart / m_ySpacing;
        partCoords[2] = zStart / m_zSpacing;
        
        // Set the next position to an undefined value to signal that the
        // current I, J, and K need to be calculated on the first time step.
        int nextI = -1;
        int nextJ = -1;
        int nextK = -1;
        
        CellDescriptor cell = new CellDescriptor();
        CellVariables vars = cell.vars;
        
        // The closest a particle can get to the bottom or the surface.
        double boundaryPad = 1E-9;
        
        // The particle is defined until the first time it leaves the grid.
        boolean bParticleDefined = true;
        
        // Previous position at the start is the same as the start.
        double[] prevPartCoords = new double[3];
        prevPartCoords[0] = partCoords[0];
        prevPartCoords[1] = partCoords[1];
        prevPartCoords[2] = partCoords[2];
        
        // Calculate the time step of the particle tracking model as the time
        // step of the vector file divided by m_tsDivs.
        double ptmTimeStepInc = m_timeStepInc / (double)m_tsDivs;
        
        double tinyInc = 1E-4;
        double veryTinyInc = 1E-6;
        double superTinyInc = 1E-8;
        
        
        // Calculate and store the particle coordinates for each time step.
        for (int ts = startTS; ts <= endTS; ts++) {
            // Store the time for this step in the path.
            path[ts * 4] = (ts + offset) * ptmTimeStepInc + m_si3dRunStartTime;
            
            // Once the particle has left the grid, assign an undefined
            // location for all remaining times.
            if (!bParticleDefined) {
                path[ts * 4 + 1] = -1;
                path[ts * 4 + 2] = -1;
                path[ts * 4 + 3] = 3;
                continue;
            }
            
            // Initialize the time remaining to the full time step length.
            double timeRemaining = ptmTimeStepInc;
            
            // Set layer variables.
            vars.layerHeight = m_zSpacing;
            
            // Keep track of layer changes to prevent oscillations along a
            // layer boundary.
            int lastLayerChange = 0;
            long numCycles = 0;
            int tinyIncCounter = 0;
            int tinyIncCheckpointI = Integer.MIN_VALUE;
            int tinyIncCheckpointJ = Integer.MIN_VALUE;
            int tinyIncCheckpointK = Integer.MIN_VALUE;
            
            double zEddyDiffusivityTotal = 0;
            double zEddyDiffusivityAvg = 0;
            
            // Sub-cycle until the time remaining before the next time step
            // has been used up.  For each iteration, calculate the location
            // of the particle at the end of the time step.  If it is within
            // the current cell, store that location and stop sub-cycling.
            // If the location is outside the current cell, move the particle
            // just beyond the cell boundary and recalulate the time remaining.
            while (timeRemaining > 0) {
                cell.i = nextI;
                cell.j = nextJ;
                cell.k = nextK;
                cell.ptmTS = ts + offset;
                
                numCycles++;
                if ((numCycles % 1000000) == 0) {
                    System.out.println("Stuck in loop at ts = " + (ts + offset)
                    + ", i = " + cell.i + ", j = " + cell.j + ", k = "
                        + cell.k);
                    
                    path[ts * 4 + 1] = -1;
                    path[ts * 4 + 2] = -1;
                    path[ts * 4 + 3] = 3;
                    continue;
                }
                
                // If this is the first time step, calculate the I, J, and K
                // values from the coordinates of the release point.
                // Calculate the I, J, and K indexes for the cell that
                // currently holds the particle.
                cell.i = (int)Math.round(partCoords[0]);
                nextI = cell.i;
                cell.j = (int)Math.round(partCoords[1]);
                nextJ = cell.j;
                cell.k = (int)Math.round(-partCoords[2]);
                if (partCoords[2] > -1)
                    cell.k = 1;
                nextK = cell.k;
                
                // Check that the data is valid and the index is a wet point.
                int ijIndex = cell.i + (cell.j * m_numI);
                if (ijIndex < 0 || ijIndex > m_2DLookupArray.length) {
                    // If undefined, write (0, 0, 0) and move to the next
                    // time step.
                    path[ts * 4 + 1] = 0;
                    path[ts * 4 + 2] = 0;
                    path[ts * 4 + 3] = 0;
                    bParticleDefined = false;
                    break;
                }
                
                // Find the lookup index for surface and bottom heights.
                cell.lookup2D = -1;
                if (m_2DLookupArray != null)
                    cell.lookup2D = m_2DLookupArray[ijIndex];
                
                if (cell.lookup2D < 0) {
                    // If undefined, write (0, 0, 0) and move to the next
                    // time step.
                    path[ts * 4 + 1] = 0;
                    path[ts * 4 + 2] = 0;
                    path[ts * 4 + 3] = 0;
                    bParticleDefined = false;
                    break;
                }
                
                // Retrieve the velocities at the faces of the cell using
                // the specified type of interpolation between time steps
                // of the Si3D velocity field.
                getCellVariables(cell);
                scaleCellVariables(cell);
                
                if (m_bCalcDrift) {
                    double zDrift = cell.vars.zEddyDiffusivityTop -
                        cell.vars.zEddyDiffusivityBottom;
                    cell.vars.wTop += zDrift;
                    cell.vars.wBottom += zDrift;
                }
                
                if (m_bCalcTurbulence) {
                    zEddyDiffusivityAvg = (cell.vars.zEddyDiffusivityTop +
                        cell.vars.zEddyDiffusivityBottom) / 2;
                }
                
                // In case cell.k was adjusted, make sure nextK matches.
                nextK = cell.k;
                
                double layerHeightRatio = vars.layerHeight / m_zSpacing;
                
                // Calculate the locations of the faces of this cell in cell
                // units referenced to the lower-left corner of the grid.
                // Depth is negative measured from the top of layer 1 when
                // there is a surface adjustment of 0.
                double eastFace = (double)cell.i + 0.5;
                double westFace = (double)cell.i - 0.5;
                double northFace = (double)cell.j + 0.5;
                double southFace = (double)cell.j - 0.5;
                double topFace = -((double)cell.k) + 0.5
                    + (cell.vars2D.surfaceHeightAdjustment / m_zSpacing);
                double bottomFace = -((double)cell.k) - 0.5
                    + (cell.vars2D.bottomAdjustment / m_zSpacing);
                
                // Clamp the particle location to within the wet area of the
                // current layer.
                if (partCoords[2] > topFace) {
                    partCoords[2] = topFace;
                    if (cell.k == 1)
                        partCoords[2] -= boundaryPad;
                }
                if (partCoords[2] < bottomFace) {
                    partCoords[2] = bottomFace;
                    if (cell.k == cell.vars2D.bottomLayerK) {
                        partCoords[2] += boundaryPad;
                    }
                }
                
                // Calculate the position, relative to the cell's lower corner.
                double xRelativeToCell = partCoords[0] - westFace;
                double yRelativeToCell = partCoords[1] - southFace;
                double zRelativeToCell = (partCoords[2] - bottomFace)
                / layerHeightRatio;
                
                if (xRelativeToCell > -m_nearZero &&
                    xRelativeToCell < m_nearZero) {
                    
                    partCoords[0] = westFace + m_nearZero;
                }
                if (yRelativeToCell > -m_nearZero &&
                    yRelativeToCell < m_nearZero) {
                    
                    partCoords[1] = southFace + m_nearZero;
                }
                if (zRelativeToCell > -m_nearZero &&
                    zRelativeToCell < m_nearZero) {
                    
                    partCoords[2] = bottomFace + m_nearZero;
                }
                
                xRelativeToCell = partCoords[0] - westFace;
                yRelativeToCell = partCoords[1] - southFace;
                zRelativeToCell = (partCoords[2] - bottomFace)
                / layerHeightRatio;
                
                // Calculate intermediate variables for the simple advection
                // step of the Random Walk Model.
                double Ax = vars.uEast - vars.uWest;
                double Ay = vars.vNorth - vars.vSouth;
                double Az = vars.wTop - vars.wBottom;
                
                double Cx = vars.uWest / Ax;
                double Cy = vars.vSouth / Ay;
                double Cz = vars.wBottom / Az;
                
                // Calculate the particle's velocity across the cell.
                double uParticle = Ax * xRelativeToCell + vars.uWest;
                double vParticle = Ay * yRelativeToCell + vars.vSouth;
                double wParticle = Az * zRelativeToCell + vars.wBottom;
                
                // Calculate the x, y, and z coordinates at the end of the time
                // step.
                double xEndOfTimeStep = positionAtTime(Ax, vars.uWest, Cx,
                    xRelativeToCell, timeRemaining) + westFace;
                double yEndOfTimeStep = positionAtTime(Ay, vars.vSouth, Cy,
                    yRelativeToCell, timeRemaining) + southFace;
                double zEndOfTimeStep = positionAtTime(Az, vars.wBottom, Cz,
                    zRelativeToCell, timeRemaining);
                zEndOfTimeStep *= layerHeightRatio;
                zEndOfTimeStep += bottomFace;
                
                // Determine if the particle passed to a new layer as opposed
                // to breaking through the surface.
                boolean bChangedLayer = false;
                if (zEndOfTimeStep < bottomFace)
                    bChangedLayer = true;
                if (cell.k > 1) {
                    if (zEndOfTimeStep > topFace)
                        bChangedLayer = true;
                }
                
                // Check if the particle went outside of the cell with a
                // nonzero velocity.
                if ((xEndOfTimeStep > eastFace || xEndOfTimeStep < westFace
                    || yEndOfTimeStep > northFace || yEndOfTimeStep
                    < southFace || bChangedLayer) && !(
                    Math.abs(uParticle) < m_nearZero &&
                    Math.abs(vParticle) < m_nearZero &&
                    Math.abs(wParticle) < m_nearZero)) {
                    
                    // It went outside of the cell, so find out when and where
                    // and loop with a smaller remaining time...
                    
                    // Calculate intermediate values.
                    double xParticleC = xRelativeToCell + Cx;
                    double yParticleC = yRelativeToCell + Cy;
                    double zParticleC = zRelativeToCell + Cz;
                    double xLowerC = Cx;
                    double yLowerC = Cy;
                    double zLowerC = Cz;
                    double xUpperC = 1 + Cx;
                    double yUpperC = 1 + Cy;
                    double zUpperC = 1 + Cz;
                    
                    // For each dimension, determine the case that describes
                    // the given combination of values.
                    int caseX = findCase(Ax, vars.uWest, xParticleC, xLowerC,
                        xUpperC, uParticle);
                    int caseY = findCase(Ay, vars.vSouth, yParticleC, yLowerC,
                        yUpperC, vParticle);
                    int caseZ = findCase(Az, vars.wBottom, zParticleC, zLowerC,
                        zUpperC, wParticle);
                    
                    // Calculate distances from the walls.
                    double xLowerDist = westFace - partCoords[0];
                    double yLowerDist = southFace - partCoords[1];
                    double zLowerDist = bottomFace - partCoords[2];
                    zLowerDist /= layerHeightRatio;
                    double xUpperDist = eastFace - partCoords[0];
                    double yUpperDist = northFace - partCoords[1];
                    double zUpperDist = topFace - partCoords[2];
                    zUpperDist /= layerHeightRatio;
                    
                    // For each dimension, determine the time at which the
                    // particle would reach the face of the cell.
                    double tFaceX = timeToBoundary(Ax, vars.uWest, xParticleC,
                        xLowerC, xUpperC, xLowerDist, xUpperDist, caseX);
                    double tFaceY = timeToBoundary(Ay, vars.vSouth, yParticleC,
                        yLowerC, yUpperC, yLowerDist, yUpperDist, caseY);
                    double tFaceZ = timeToBoundary(Az, vars.wBottom, zParticleC,
                        zLowerC, zUpperC, zLowerDist, zUpperDist, caseZ);
                    
                    // Don't consider breaking through the surface as
                    // crossing a cell boundary for the purposes of finding
                    // the next boundary crossing.
                    if (cell.k == 1 && wParticle > 0) {
                        tFaceZ = Double.POSITIVE_INFINITY;
                    }
                    
                    // Prevent oscillations on a layer boundary.
                    boolean bDisableW = false;
                    if ((lastLayerChange > 0 && wParticle > 0) ||
                        (lastLayerChange < 0 && wParticle < 0)) {
                        
                        bDisableW = true;
                    }
                    
                    boolean bCellChangedI = false;
                    boolean bCellChangedJ = false;
                    boolean bCellChangedK = false;
                    
                    // If the particle was moved to an X boundary, then
                    // change the I index for the next time step.
                    if (tFaceX < tFaceY && tFaceX < tFaceZ) {
                        if (uParticle > 0) {
                            partCoords[0] = eastFace + m_nearZero;
                            nextI++;
                        } else {
                            partCoords[0] = westFace - m_nearZero;
                            nextI--;
                        }
                        
                        lastLayerChange = 0;
                        bCellChangedI = true;
                    }
                    
                    // If the particle was moved to a Y boundary, then
                    // change the J index for the next time step.
                    if (tFaceY < tFaceX && tFaceY < tFaceZ) {
                        if (vParticle > 0) {
                            partCoords[1] = northFace + m_nearZero;
                            nextJ++;
                        } else {
                            partCoords[1] = southFace - m_nearZero;
                            nextJ--;
                        }
                        lastLayerChange = 0;
                        bCellChangedJ = true;
                    }
                    
                    // Check if the particle exited the grid through an open
                    // boundary.  If so, set it as undefined so it will stop
                    // being tracked.
                    int num = wasLineCrossed(m_boundaries, prevPartCoords[0],
                        prevPartCoords[1], partCoords[0], partCoords[1]);
                    if (num >= 0 && prevPartCoords[0] > 0 &&
                        prevPartCoords[1] > 0) {
                        
                        bParticleDefined = false;
                        if (m_bTrackTimeOnly) {
                            path[0] =
                                path[ts * 4] + ptmTimeStepInc - timeRemaining;
                            path[1] = num;
                            
                            return path;
                        }
                    }
                    if (num >= 0 && prevPartCoords[0] <= 0 &&
                        prevPartCoords[1] <= 0) {
                        
                        System.out.println(
                            "Crossed into grid over boundary at ts " + ts);
                    }
                    
                    // Save the particle's current location so it can be
                    // determined if it crosses a boundary on the next time
                    // step.
                    prevPartCoords[0] = partCoords[0];
                    prevPartCoords[1] = partCoords[1];
                    prevPartCoords[2] = partCoords[2];
                    
                    // If the particle was moved to a Z boundary, then
                    // change the K index for the next time step.
                    if (tFaceZ < tFaceX && tFaceZ < tFaceY) {
                        if (wParticle > 0) {
                            partCoords[2] = topFace + m_nearZero;
                            nextK--;
                            lastLayerChange = -1;
                        } else {
                            partCoords[2] = bottomFace - m_nearZero;
                            nextK++;
                            lastLayerChange = 1;
                        }
                        bCellChangedK = true;
                    }
                    
                    // Determine which face was reached first and use that as
                    // the time increment.
                    double timeInc = tFaceX;
                    if (tFaceY < timeInc) timeInc = tFaceY;
                    if (tFaceZ < timeInc) timeInc = tFaceZ;
                    if (timeInc < m_nearZero)
                        timeInc = m_nearZero;
                    
                    if (timeInc < tinyInc) {
                        // Keep track of distance travelled.
                        if (tinyIncCounter == 0) {
                            tinyIncCheckpointI = nextI;
                            tinyIncCheckpointJ = nextJ;
                            tinyIncCheckpointK = nextK;
                            tinyIncCounter++;
                        } else {
                            // Remove the tiny inc test if the particle has
                            // travelled more than a couple of cells.
                            int iDiff = Math.abs(tinyIncCheckpointI - nextI);
                            int jDiff = Math.abs(tinyIncCheckpointJ - nextJ);
                            int kDiff = Math.abs(tinyIncCheckpointK - nextK);
                            if (iDiff > 2 || jDiff > 2 || kDiff > 2) {
                                tinyIncCounter = 0;
                            } else {
                                // Bump up the counter.
                                tinyIncCounter++;
                                if (timeInc < veryTinyInc) {
                                    tinyIncCounter++;
                                    if (timeInc < superTinyInc) {
                                        tinyIncCounter++;
                                    }
                                }
                                
                            }
                        }
                    } else {
                        tinyIncCounter = 0;
                    }
                    
                    // Don't move the particles any further on this time step
                    // if timeInc is infinity.
                    if (timeInc == Double.POSITIVE_INFINITY) {
                        timeRemaining = -1;
                    } else {
                        // Track the particle for the given time increment, at
                        // which time it should be at the face of the cell.
                        if (!bCellChangedI) {
                            partCoords[0] = positionAtTime(Ax, vars.uWest, Cx,
                                xRelativeToCell, timeInc) + westFace;
                        }
                        if (!bCellChangedJ) {
                            partCoords[1] = positionAtTime(Ay, vars.vSouth, Cy,
                                yRelativeToCell, timeInc) + southFace;
                        }
                        if (bDisableW) {
                            // In case this time step is aborted.
                            zEndOfTimeStep = partCoords[2];
                        } else {
                            if (!bCellChangedK) {
                                partCoords[2] = positionAtTime(Az,
                                    vars.wBottom, Cz, zRelativeToCell,
                                    timeInc);
                                partCoords[2] *= layerHeightRatio;
                                partCoords[2] += bottomFace;
                            }
                        }
                    }
                    
                    // Remove the time increment from the amount of time
                    // remaining in the current time step.
                    if (tinyIncCounter > 12) {
                        if (m_bCalcTurbulence) {
                            zEddyDiffusivityTotal +=
                                (zEddyDiffusivityAvg * timeRemaining);
                        }
                        
                        timeRemaining = -1;
                    } else {
                        if (m_bCalcTurbulence) {
                            zEddyDiffusivityTotal +=
                                (zEddyDiffusivityAvg * timeInc);
                        }
                        timeRemaining -= timeInc;
                    }
                } else {
                    if (m_bCalcTurbulence) {
                        zEddyDiffusivityTotal +=
                            (zEddyDiffusivityAvg * timeRemaining);
                    }
                    // The particle would still be inside the cell at the end
                    // of the time step, so finish this time step and set the
                    // timeRemaining flag to go to the next time step.
                    // Set particle coords and add to path.
                    // Go to next time step.
                    timeRemaining = -1;
                    
                    partCoords[0] = xEndOfTimeStep;
                    partCoords[1] = yEndOfTimeStep;
                    partCoords[2] = zEndOfTimeStep;
                }
                
                // Clamp the particle between the bottom and surface.
                if (partCoords[2] < cell.vars2D.bottomLocation)
                    partCoords[2] = cell.vars2D.bottomLocation + boundaryPad;
                if (partCoords[2] > cell.vars2D.wseLocation)
                    partCoords[2] = cell.vars2D.wseLocation - boundaryPad;
            }
            
            if (m_bCalcTurbulence) {
                double zTurb = Math.sqrt(6 * zEddyDiffusivityTotal) *
                    (Math.random() * 2 - 1);
                partCoords[2] += zTurb;
                
                // Clamp the particle between the bottom and surface.
                if (partCoords[2] < cell.vars2D.bottomLocation)
                    partCoords[2] = cell.vars2D.bottomLocation + boundaryPad;
                if (partCoords[2] > cell.vars2D.wseLocation)
                    partCoords[2] = cell.vars2D.wseLocation - boundaryPad;
            }
            
            // Convert the current particle location in cell units to world
            // units and store in the path array.
            path[ts * 4 + 1] = partCoords[0] * m_xSpacing;
            path[ts * 4 + 2] = partCoords[1] * m_ySpacing;
            path[ts * 4 + 3] = (partCoords[2] + 0.5) * m_zSpacing;
        }
        
        return path;
    }
    
//------------------------------------------------------------------------------
    /**
     * Calculate the one-dimensional position after a certain time increment of a
     * particle that starts at a 1D position with a 1D velocity.
     * @param A The A variable.
     * @param B The B variable.
     * @param C The C variable.
     * @param startPosition The start position of the particle.
     * @param timeInc       The time increment.
     * @return  The position of the particle after being moved.
     */
    protected double positionAtTime(double A, double B, double C,
        double startPosition, double timeInc) {
        
        double endPosition;
        
        // Calculate the position at the end of the time step based on the
        // velocities in the current cell.
        // Must check if A is zero before using it as a divisor.
        if (Math.abs(A) < m_nearZero) {
            endPosition = startPosition + (B * timeInc);
        } else {
            endPosition = (startPosition + C) *
                Math.exp(A * timeInc) - C;
        }
        
        return endPosition;
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds which case applies for the given set of variables.
     * Can be applied to any dimension.                                     <p>
     * Case 0: Velocity of Particle == 0                                    <p>
     * Case 1: A > 0 and Particle location + C > 0                          <p>
     * Case 2: A > 0 and Particle location + C < 0                          <p>
     * Case 3: A < 0 and Lower wall + C > 0                                 <p>
     * Case 4: A < 0 and Upper wall + C < 0                                 <p>
     * Case 5: A < 0 and Lower wall + C < 0 AND upper wall + C > 0          <p>
     * Case 6: A == 0 and B > 0                                             <p>
     * Case 7: A == 0 and B < 0                                             <p>
     * @param A             The A variable.
     * @param B             The B variable.
     * @param particleC     The C variable as calculated for the particle.
     * @param lowerC        The C variable as calculated for the lower wall.
     * @param upperC        The C variable as calculated for the upper wall.
     * @param velParticle   The velocity of the particle in the direction from the
     *                      lower wall toward the upper wall.
     * @return The integer code for the case.
     */
    protected int findCase(double A, double B, double particleC,
        double lowerC, double upperC, double velParticle) {
        
        // Velocity of Particle == 0
        if (Math.abs(velParticle) < m_nearZero) {
            return 0;
        } else {
            if (A > m_nearZero) {
                // A > 0
                if (particleC > m_nearZero)
                    // Particle location + C > 0
                    return 1;
                if (particleC < -m_nearZero)
                    // Particle location + C < 0
                    return 2;
            } else {
                // A <= 0
                if (A < -m_nearZero) {
                    // A < 0
                    if (lowerC > m_nearZero)
                        // Lower wall + C > 0
                        return 3;
                    if (upperC < -m_nearZero)
                        // Upper wall + C < 0
                        return 4;
                    if (lowerC < m_nearZero && upperC > -m_nearZero)
                        // Lower wall + C < 0 AND upper wall + C > 0
                        return 5;
                } else {
                    // A == 0
                    if (B > m_nearZero)
                        // B > 0
                        return 6;
                    if (B < -m_nearZero)
                        // B < 0
                        return 7;
                }
            }
        }
        
        return 0;
    }
    
//------------------------------------------------------------------------------
    /**
     * Calculate how long it will take the particle to reach the cell boundary
     * from its current position and given the provided 1D values calculated from
     * the velocities.
     * @param A             The A variable.
     * @param B             The B variable.
     * @param particleC     The C variable as calculated for the particle.
     * @param lowerC        The C variable as calculated for the lower wall.
     * @param upperC        The C variable as calculated for the upper wall.
     * @param lowerDist     The distance from the lower boundary.
     * @param upperDist     The distance from the upper boundary.
     * @param caseNum       The integer code describing the case.
     * @return  The amount of time it will take the particle to reach the cell
     *          boundary.
     */
    protected double timeToBoundary(double A, double B, double particleC,
        double lowerC, double upperC, double lowerDist, double upperDist,
        int caseNum) {
        
        switch (caseNum) {
            case 0:
                return Double.POSITIVE_INFINITY;
            case 1:
                double ratio = upperC / particleC;
                if (ratio < 0) {
                    int hey = 0;
                }
                double val = Math.log(ratio) / A;
                return val;
            case 2:
                ratio = lowerC / particleC;
                if (ratio < 0) {
                    int hey = 0;
                }
                val = Math.log(ratio) / A;
                return val;
            case 3:
                ratio = lowerC / particleC;
                if (ratio < 0) {
                    int hey = 0;
                }
                val = Math.log(ratio) / A;
                return val;
            case 4:
                ratio = upperC / particleC;
                if (ratio < 0) {
                    int hey = 0;
                }
                val = Math.log(ratio) / A;
                return val;
            case 5:
                return Double.POSITIVE_INFINITY;
            case 6:
                return (upperDist / B);
            case 7:
                return (lowerDist / B);
            default:
                return Double.POSITIVE_INFINITY;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Determines if a particle crossed a given line while moving from its
     * starting position to its ending position.
     * @param lines An array of line arrays, each containing for values for
     *              (x1, y1, x2, y2) defining the endpoints of the lines.
     * @param x1    The X-coordinate of the starting position.
     * @param y1    The Y-coordinate of the starting position.
     * @param x2    The X-coordinate of the ending position.
     * @param y2    The Y-coordinate of the ending position.
     * @return  The index in the lines array of the line segment that was crossed
     *          first, -1 if no line segment was crossed.
     */
    protected int wasLineCrossed(double[][] lines, double x1, double y1,
        double x2, double y2) {
        
        // Create an array for storing the distance to each line.
        double[] distances = new double[lines.length];
        
        // Check each line in the array.
        for (int i = 0; i < lines.length; i++) {
            double[] line = lines[i];
            // The line array must contain exactly four elements.
            if (line.length < 4)
                continue;
            
            // Initialize distance to largest positive value.
            distances[i] = Double.POSITIVE_INFINITY;
            
            // The line is vertical.
            if (line[0] == line[2]) {
                // If the particle started on one side and ended on the other...
                if ((x1 <= line[0] && x2 >= line[0]) ||
                    (x1 >= line[0] && x2 <= line[0])) {
                    
                    // Find out where the intersection occurred.
                    double intersectionY = Double.POSITIVE_INFINITY;
                    double xDiff = x2 - x1;
                    if (xDiff != 0) {
                        intersectionY =
                            ((line[0] - x1) * (y2 - y1) / xDiff) + y1;
                    } else {
                        if (x1 == line[0])
                            intersectionY = y1;
                    }
                    
                    if ((intersectionY <= line[3] && intersectionY >= line[1])||
                        (intersectionY >= line[3] && intersectionY <= line[1])){
                        // The lines intersected within the boundary segment.
                        distances[i] = Math.abs(intersectionY - y1);
                    }
                }
            }
            
            // The line is horizontal.
            if (line[1] == line[3]) {
                // If the particle started on one side and ended on the other...
                if ((y1 <= line[1] && y2 >= line[1]) ||
                    (y1 >= line[1] && y2 <= line[1])) {
                    
                    // Find out where the intersection occurred.
                    double intersectionX = Double.POSITIVE_INFINITY;
                    double yDiff = y2 - y1;
                    if (yDiff != 0) {
                        intersectionX =
                            ((line[1] - y1) * (x2 - x1) / yDiff) + x1;
                    } else {
                        if (y1 == line[1])
                            intersectionX = x1;
                    }
                    
                    if ((intersectionX <= line[2] && intersectionX >= line[0])||
                        (intersectionX >= line[2] && intersectionX <= line[0])){
                        // The lines intersected within the boundary segment.
                        distances[i] = Math.abs(intersectionX - x1);
                    }
                }
            }
        }
        
        // Check if any of the y distances are less than infinity and return the
        // smallest one.
        int smallestIndex = -1;
        double smallestDistance = Double.POSITIVE_INFINITY;
        for (int i = 0; i < distances.length; i++) {
            if (distances[i] < smallestDistance) {
                smallestIndex = i;
                smallestDistance = distances[i];
            }
        }
        
        return smallestIndex;
    }
    
//------------------------------------------------------------------------------
    /**
     * Gets the velocities at each face of the cell at the given time step.
     * Converts the velocities to units of grid cells per day.
     * If the time step is between time steps of the Si3D output vector field,
     * interpolates the values using the preceding and following values with
     * the method chosen by m_interpolationIndex.
     * @param cell  The cell location and variables.
     */
    public void getCellVariables(CellDescriptor cell) {
        
        // Get raw velocities using the chosen time-interpolation method.
        switch (m_interpolationIndex) {
            case 1:
                getVariablesLinear(cell);
                break;
            case 2:
                getVariablesParabolic(cell);
                break;
            default:
                getVariablesNearest(cell);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Gets the velocities at each face of the cell at the given time step.
     * Converts the velocities to units of grid cells per day.
     * If the time step is between time steps of the Si3D output vector field,
     * interpolates the values using the preceding and following values with
     * the method chosen by m_interpolationIndex.
     * @param cell  The cell location and variables.
     */
    public void scaleCellVariables(CellDescriptor cell) {
        
        CellVariables vars = cell.vars;
        
        // Multiply by the number of seconds in a day (86400) and
        // the unit scale and divide by the cell size.
        // Convert cm^2/s to grid cells per day.
        vars.uEast *= (86400 * m_unitScales[0] / m_xSpacing);
        vars.vNorth *= (86400 * m_unitScales[1] / m_ySpacing);
        vars.wTop *= (86400 * m_unitScales[2] / vars.layerHeight);
        vars.uWest *= (86400 * m_unitScales[0] / m_xSpacing);
        vars.vSouth *= (86400 * m_unitScales[1] / m_ySpacing);
        vars.wBottom *= (86400 * m_unitScales[2] / vars.layerHeight);
        if (m_bCalcDrift) {
            vars.zEddyDiffusivityTop *= (86400 * m_unitScales[3])
            / vars.layerHeight / vars.layerHeight;
            vars.zEddyDiffusivityBottom *= (86400 * m_unitScales[3])
            / vars.layerHeight / vars.layerHeight;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Gets the velocities at each face of the cell at the given time step.
     * No units conversion is performed.
     * If the time step is between time steps of the Si3D output vector field,
     * uses either the preceding or following values, whichever is nearer.
     * @param cell  The cell location and variables.
     */
    protected void getVariablesNearest(CellDescriptor cell) {
        
        // If the last temp cell already contains the values for this location
        // and time step, reuse them.
        if (m_tmpCell1.equals(cell)) {
            m_tmpCell1.copyAllInto(cell);
            return;
        }
        
        // Calculate the lookups for each face of the cell at the given
        // time step.
        int vecFieldTS = (int)Math.round((double)cell.ptmTS / (double)m_tsDivs);
        
        getLookups(cell, vecFieldTS);
        
        // Load the variables into the vars object.
        getVariables(cell);
        
        // Make corrections for walls.
        correctForWalls(cell);
        
        // Get layer height variables for the cell.
        getLayerHeight(cell);
        
        // Store the values for possible reuse next time step.
        cell.copyAllInto(m_tmpCell1);
    }
    
//------------------------------------------------------------------------------
    /**
     * Gets the velocities at each face of the cell at the given time step.
     * No units conversion is performed.
     * If the time step is between time steps of the Si3D output vector field, uses
     * linear interpolation between the preceding and following values.
     * @param cell  The cell location and variables.
     */
    protected void getVariablesLinear(CellDescriptor cell) {
        
        // If the last temp cell already contains the values for this location
        // and time step, reuse them.
        if (m_tmpCell2 != cell && m_tmpCell2.equals(cell)) {
            m_tmpCell2.copyAllInto(cell);
            correctForWalls(cell);
            return;
        }
        
        // Determine the vector field time steps before and after this PTM
        // time step.
        int vecFieldTS1 = cell.ptmTS / m_tsDivs;
        int remainder = cell.ptmTS % m_tsDivs;
        
        if (remainder == 0) {
            // Get layer height variables for the cell.
            getLayerHeight(cell);
            
            // Get the preceding time's lookups.
            getLookups(cell, vecFieldTS1);
            
            // If this is not between vector field time steps, use the time
            // step lookup that was just retrieved.
            // Load the velocities into the vars object.
            getVariables(cell);
            
        } else {
            int vecFieldTS2 = vecFieldTS1 + 1;
            
            // Perform a bounds check.
            if (vecFieldTS1 < 1)
                vecFieldTS1 = 1;
            if (vecFieldTS1 > m_numTimeSteps)
                vecFieldTS1 = m_numTimeSteps;
            if (vecFieldTS2 >= m_numTimeSteps)
                vecFieldTS2 = vecFieldTS1;
            
            cell.copyDescriptionInto(m_tmpCell1);
            m_tmpCell1.ptmTS = vecFieldTS1 * m_tsDivs;
            
            // If the last temp cell already contains the values for this
            // location and time step, reuse them.
            if (m_tmpCell2.equals(m_tmpCell1)) {
                m_tmpCell2.copyAllInto(m_tmpCell1);
            } else {
                // Store the preceding time step's velocities.
                getCellVariables(m_tmpCell1);
            }
            
            cell.copyDescriptionInto(m_tmpCell2);
            m_tmpCell2.ptmTS = vecFieldTS2 * m_tsDivs;
            
            // Store the following time step's velocities.
            getCellVariables(m_tmpCell2);
            
            // Interpolate the answer.
            // Loads the velocities into the vars object.
            double xVal = (double) remainder / (double) m_tsDivs;
            linearlyInterpolate(m_tmpCell1.vars, m_tmpCell2.vars, cell.vars,
                xVal);
            
            cell.vars.lookups[0] = m_tmpCell2.vars.lookups[0];
            cell.vars.lookups[1] = m_tmpCell2.vars.lookups[1];
            cell.vars.lookups[2] = m_tmpCell2.vars.lookups[2];
            cell.vars.lookups[3] = m_tmpCell2.vars.lookups[3];
            
            // Make corrections for walls.
            correctForWalls(cell);
            
            // Get layer height variables for the cell.
            getLayerHeight(cell);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Gets the variables at each face of the cell at the given time step.
     * No units conversion is performed.
     * If the time step is between time steps of the Si3D output vector field, uses
     * parabolic interpolation between the preceding two and following two values.
     * @param cell  The cell location and variables.
     */
    protected void getVariablesParabolic(CellDescriptor cell) {
        
        // If the last temp cell already contains the values for this location
        // and time step, reuse them.
        if (m_tmpCell3 != cell && m_tmpCell3.equals(cell)) {
            m_tmpCell3.copyAllInto(cell);
            correctForWalls(cell);
            return;
        }
        
        // Determine the vector field time steps before and after this PTM
        // time step.
        int remainder = cell.ptmTS % m_tsDivs;
        int vecFieldTS2 = cell.ptmTS / m_tsDivs;
        
        if (remainder == 0) {
            // Get layer height variables for the cell.
            getLayerHeight(cell);
            
            // Get the preceding time's lookups.
            getLookups(cell, vecFieldTS2);
            
            // If this is not between vector field time steps, use the time
            // step lookup that was just retrieved.
            // Load the velocities into the vars object.
            getVariables(cell);
        } else {
            int vecFieldTS1 = vecFieldTS2 - 1;
            int vecFieldTS3 = vecFieldTS2 + 1;
            int vecFieldTS4 = vecFieldTS2 + 2;
            
            // Perform a bounds check.
            if (vecFieldTS1 < 1)
                vecFieldTS1 = 1;
            if (vecFieldTS2 < 1)
                vecFieldTS2 = 1;
            if (vecFieldTS1 > m_numTimeSteps)
                vecFieldTS1 = m_numTimeSteps;
            if (vecFieldTS2 > m_numTimeSteps)
                vecFieldTS2 = m_numTimeSteps;
            if (vecFieldTS3 >= m_numTimeSteps)
                vecFieldTS3 = vecFieldTS2;
            if (vecFieldTS4 >= m_numTimeSteps)
                vecFieldTS4 = vecFieldTS3;
            
            // Store the other time step's velocities.
            cell.copyDescriptionInto(m_tmpCell1);
            m_tmpCell1.ptmTS = vecFieldTS1 * m_tsDivs;
            getCellVariables(m_tmpCell1);
            
            cell.copyDescriptionInto(m_tmpCell2);
            m_tmpCell2.ptmTS = vecFieldTS2 * m_tsDivs;
            getCellVariables(m_tmpCell2);
            
            cell.copyDescriptionInto(m_tmpCell3);
            m_tmpCell3.ptmTS = vecFieldTS3 * m_tsDivs;
            getCellVariables(m_tmpCell3);
            
            cell.copyDescriptionInto(m_tmpCell4);
            m_tmpCell4.ptmTS = vecFieldTS4 * m_tsDivs;
            getCellVariables(m_tmpCell4);
            
            // Interpolate the answer.
            // Loads the velocities into the vars object.
            double tVal = (double) remainder / (double) m_tsDivs;
            parabolicallyInterpolate(m_tmpCell1.vars, m_tmpCell2.vars,
                m_tmpCell3.vars, m_tmpCell4.vars, cell.vars, tVal);
            
            cell.vars.lookups[0] = m_tmpCell3.vars.lookups[0];
            cell.vars.lookups[1] = m_tmpCell3.vars.lookups[1];
            cell.vars.lookups[2] = m_tmpCell3.vars.lookups[2];
            cell.vars.lookups[3] = m_tmpCell3.vars.lookups[3];
            
            // Make corrections for walls.
            correctForWalls(cell);
            
            // Get layer height variables for the cell.
            getLayerHeight(cell);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Calculates the lookups for the given cell and vector field time step.
     * Returns four lookup indexes in the cell.vars.lookups array.  The first is
     * used for the east, north, and top faces of the cell.  The second is used for
     * the west side, the third for the south, and the fourth for the bottom.
     * @param cell  The cell location and variables.
     * @param ts    The vector field time step at which the lookups are required.
     */
    protected void getLookups(CellDescriptor cell, int ts) {
        
        int[] lookups = cell.vars.lookups;
        
        // Current lookup
        lookups[0] = getLookupIndex(cell.i, cell.j, cell.k,
            m_numI, m_numJ, m_numLayers, m_num3DVarsDims, ts,
            m_numWet3Dpoints, m_3DLookupArray);
        // West lookup
        lookups[1] = getLookupIndex((cell.i - 1), cell.j, cell.k,
            m_numI, m_numJ, m_numLayers, m_num3DVarsDims, ts,
            m_numWet3Dpoints, m_3DLookupArray);
        // South lookup
        lookups[2] = getLookupIndex(cell.i, (cell.j - 1), cell.k,
            m_numI, m_numJ, m_numLayers, m_num3DVarsDims, ts,
            m_numWet3Dpoints, m_3DLookupArray);
        // Bottom lookup
        lookups[3] = getLookupIndex(cell.i, cell.j, (cell.k + 1),
            m_numI, m_numJ, m_numLayers, m_num3DVarsDims, ts,
            m_numWet3Dpoints, m_3DLookupArray);
    }
    
//------------------------------------------------------------------------------
    /**
     * Gets the variables at each face of the cell at the given vector field
     * time step.  No units conversion is performed.
     * @param cell  The cell location and variables.
     */
    protected void getVariables(CellDescriptor cell) {
        
        // lookups   An array of lookup indexes for the faces of the cell.
        //          array element 0 contains east, north, and top lookups
        //          array element 1 contains the west lookup
        //          array element 2 contains the south lookup
        //          array element 3 contains the bottom lookup
        
        CellVariables vars = cell.vars;
        int[] lookups = vars.lookups;
        
        // Get the three variables from the current cell.
        vars.uEast = getVar3D(lookups[0], 0);
        vars.vNorth = getVar3D(lookups[0], 1);
        
        // W should be zero at the free surface.
        if (cell.k == 1 || !m_bUseW)
            vars.wTop = 0;
        else
            vars.wTop = getVar3D(lookups[0], 2);
        
        // Get the west value from the western neighbor cell.
        vars.uWest = getVar3D(lookups[1], 0);
        
        // Get the south value from the southern neighbor cell.
        vars.vSouth = getVar3D(lookups[2], 1);
        
        // Get the bottom value from the lower neighbor cell.
        vars.wBottom = getVar3D(lookups[3], 2);
        
        if (m_bCalcDrift) {
            // Get the top drift value from the current cell.
            vars.zEddyDiffusivityTop = getVar3D(lookups[0], 3);
            
            // Get the bottom drift value from the lower neighbor cell.
            vars.zEddyDiffusivityBottom = getVar3D(lookups[3], 3);
            
            // Clamp above zero.
            if (vars.zEddyDiffusivityTop < m_nearZero)
                vars.zEddyDiffusivityTop = m_nearZero;
            if (vars.zEddyDiffusivityBottom < m_nearZero)
                vars.zEddyDiffusivityBottom = m_nearZero;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Correct values that are at walls based on the values on the opposite side
     * of the cell.
     * @param cell  The cell location and variables.
     */
    protected void correctForWalls(CellDescriptor cell) {
        
        // lookups   An array of lookup indexes for the faces of the cell.
        //          array element 0 contains east, north, and top lookups
        //          array element 1 contains the west lookup
        //          array element 2 contains the south lookup
        //          array element 3 contains the bottom lookup
        
        CellVariables vars = cell.vars;
        int[] lookups = vars.lookups;
        
        // East wall, no west wall, moving west...
        if (lookups[0] < 0 && lookups[1] >= 0 && vars.uWest < 0)
            vars.uEast = vars.uWest * 0.1;
        
        // West wall, no east wall, moving east...
        if (lookups[1] < 0 && lookups[0] >= 0 && vars.uEast > 0)
            vars.uWest = vars.uEast * 0.1;
        
        // North wall, no south wall, moving south...
        if (lookups[0] < 0 && lookups[2] >= 0 && vars.vSouth < 0)
            vars.vNorth = vars.vSouth * 0.1;
        
        // South wall, no north wall, moving north...
        if (lookups[2] < 0 && lookups[0] >= 0 && vars.vNorth > 0)
            vars.vSouth = vars.vNorth * 0.1;
        
        // Top wall, no bottom wall, moving down...
        if (lookups[0] < 0 && lookups[3] >= 0 && vars.wBottom < 0)
            vars.wTop = vars.wBottom * 0.1;
        
        // Bottom wall, no top wall, moving up...
        if (lookups[3] < 0 && lookups[0] >= 0 && vars.wTop > 0)
            vars.wBottom = vars.wTop * 0.1;
    }
//------------------------------------------------------------------------------
    /**
     * For each face of a cell, find the velocity at a given time by linearly
     * interpolating between two sets of cell velocites defined at times 0 and 1.
     * @param vars1     The velocities at each face at time = 0.
     * @param vars2     The velocities at each face at time = 1.
     * @param varsNew   The velocities at each face are passed back in this object.
     * @param tVal      The time at which to interpolate (between 0 and 1).
     */
    protected void linearlyInterpolate(CellVariables vars1,
        CellVariables vars2, CellVariables varsNew, double tVal) {
        
        // Calculate the time between tVal and 1.
        double timeTill1 = 1 - tVal;
        
        // Return the exact original value if the points before and after
        // are equal.  Otherwise, linearly interpolate between the points.
        // Calculate all six faces of the cell.
        if (vars1.uEast == vars2.uEast)
            varsNew.uEast = vars1.uEast;
        else
            varsNew.uEast = (vars1.uEast * timeTill1) + (vars2.uEast * tVal);
        
        if (vars1.uWest == vars2.uWest)
            varsNew.uWest = vars1.uWest;
        else
            varsNew.uWest = (vars1.uWest * timeTill1) + (vars2.uWest * tVal);
        
        if (vars1.vNorth == vars2.vNorth)
            varsNew.vNorth = vars1.vNorth;
        else
            varsNew.vNorth = (vars1.vNorth * timeTill1) + (vars2.vNorth * tVal);
        
        if (vars1.vSouth == vars2.vSouth)
            varsNew.vSouth = vars1.vSouth;
        else
            varsNew.vSouth = (vars1.vSouth * timeTill1) + (vars2.vSouth * tVal);
        
        if (vars1.wTop == vars2.wTop)
            varsNew.wTop = vars1.wTop;
        else
            varsNew.wTop = (vars1.wTop * timeTill1) + (vars2.wTop * tVal);
        
        if (vars1.wBottom == vars2.wBottom)
            varsNew.wBottom = vars1.wBottom;
        else
            varsNew.wBottom = (vars1.wBottom * timeTill1)
            + (vars2.wBottom * tVal);
        
        if (m_bCalcDrift) {
            if (vars1.zEddyDiffusivityTop == vars2.zEddyDiffusivityTop)
                varsNew.zEddyDiffusivityTop = vars1.zEddyDiffusivityTop;
            else
                varsNew.zEddyDiffusivityTop = (vars1.zEddyDiffusivityTop
                    * timeTill1) + (vars2.zEddyDiffusivityTop * tVal);
            
            if (vars1.zEddyDiffusivityBottom == vars2.zEddyDiffusivityBottom)
                varsNew.zEddyDiffusivityBottom = vars1.zEddyDiffusivityBottom;
            else
                varsNew.zEddyDiffusivityBottom = (vars1.zEddyDiffusivityBottom
                    * timeTill1) + (vars2.zEddyDiffusivityBottom * tVal);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * For each face of a cell, find the velocity at a given time by parabolically
     * interpolating between two sets of cell velocites defined at times
     * -1, 0 1, and 2.
     * @param vars1     The velocities at each face at time = -1.
     * @param vars2     The velocities at each face at time = 0.
     * @param vars3     The velocities at each face at time = 1.
     * @param vars4     The velocities at each face at time = 2.
     * @param varsNew   The velocities at each face are passed back in this object.
     * @param tVal      The time at which to interpolate (between 0 and 1).
     */
    protected void parabolicallyInterpolate(CellVariables vars1,
        CellVariables vars2, CellVariables vars3, CellVariables vars4,
        CellVariables varsNew, double tVal) {
        
        // Calculate the time between tVal and 1.
        double timeTill1 = 1 - tVal;
        
        // Return the exact original value if the points before and after
        // are equal.  Otherwise, parabolically interpolate between the points.
        // Calculate all six faces of the cell.
        varsNew.uEast = getValueOnParabola(vars1.uEast, vars2.uEast,
            vars3.uEast, vars4.uEast, tVal, timeTill1);
        varsNew.uWest = getValueOnParabola(vars1.uWest, vars2.uWest,
            vars3.uWest, vars4.uWest, tVal, timeTill1);
        varsNew.vNorth = getValueOnParabola(vars1.vNorth, vars2.vNorth,
            vars3.vNorth, vars4.vNorth, tVal, timeTill1);
        varsNew.vSouth = getValueOnParabola(vars1.vSouth, vars2.vSouth,
            vars3.vSouth, vars4.vSouth, tVal, timeTill1);
        varsNew.wTop = getValueOnParabola(vars1.wTop, vars2.wTop,
            vars3.wTop, vars4.wTop, tVal, timeTill1);
        varsNew.wBottom = getValueOnParabola(vars1.wBottom, vars2.wBottom,
            vars3.wBottom, vars4.wBottom, tVal, timeTill1);
        if (m_bCalcDrift) {
            varsNew.zEddyDiffusivityTop = getValueOnParabola(
                vars1.zEddyDiffusivityTop,
                vars2.zEddyDiffusivityTop, vars3.zEddyDiffusivityTop,
                vars4.zEddyDiffusivityTop, tVal, timeTill1);
            varsNew.zEddyDiffusivityBottom = getValueOnParabola(
                vars1.zEddyDiffusivityBottom, vars2.zEddyDiffusivityBottom,
                vars3.zEddyDiffusivityBottom, vars4.zEddyDiffusivityBottom,
                tVal, timeTill1);
            
            // Clamp above zero.
            if (varsNew.zEddyDiffusivityTop < m_nearZero) {
                double minEDT = vars1.zEddyDiffusivityTop;
                if (vars2.zEddyDiffusivityTop < minEDT)
                    minEDT = vars2.zEddyDiffusivityTop;
                if (vars3.zEddyDiffusivityTop < minEDT)
                    minEDT = vars3.zEddyDiffusivityTop;
                if (vars4.zEddyDiffusivityTop < minEDT)
                    minEDT = vars4.zEddyDiffusivityTop;
                minEDT = (minEDT + m_nearZero) / 2;
                
                varsNew.zEddyDiffusivityTop = minEDT;
            }
            if (varsNew.zEddyDiffusivityBottom < m_nearZero) {
                double minEDB = vars1.zEddyDiffusivityBottom;
                if (vars2.zEddyDiffusivityBottom < minEDB)
                    minEDB = vars2.zEddyDiffusivityBottom;
                if (vars3.zEddyDiffusivityBottom < minEDB)
                    minEDB = vars3.zEddyDiffusivityBottom;
                if (vars4.zEddyDiffusivityBottom < minEDB)
                    minEDB = vars4.zEddyDiffusivityBottom;
                minEDB = (minEDB + m_nearZero) / 2;
                
                varsNew.zEddyDiffusivityBottom = minEDB;
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Uses parabolic interpolation to find the Y value on a curve through the
     * points (-1, y1), (0, y2), (1, y3), and (2, y4) at the X-location xVal.
     * Fits a parabola through points 1, 2, and 3 and a second parabola through
     * points 2, 3, and 4.  Returns the average of the parabola's Y values
     * at xVal.  The average is weighted fully toward parabola 1 at xVal=0 and
     * fully toward parabola 2 at xVal=1
     * @param y1    The Y value at point 1.
     * @param y2    The Y value at point 2.
     * @param y3    The Y value at point 3.
     * @param y4    The Y value at point 4.
     * @param xVal  The X value (between 0 and 1).
     * @param xVal2 (1 - xVal).
     * @return  The parabolically interpolated Y value.
     */
    protected double getValueOnParabola(double y1, double y2, double y3,
        double y4, double xVal, double xVal2) {
        
        // Fit a parabola through points 1, 2, and 3
        double[] coeffs1 = DoubleDataSeries.fitParabola(
            -1, y1, 0, y2, 1, y3);
        
        // Find the Y value on the parabola at xVal.
        double leftValue = (double)(coeffs1[0] * xVal * xVal
            + coeffs1[1] * xVal + coeffs1[2]);
        
        // Fit a parabola through points 2, 3, and 4
        double[] coeffs2 = DoubleDataSeries.fitParabola(
            0, y2, 1, y3, 2, y4);
        
        // Find the Y value on the parabola at xVal.
        double rightValue = (double)(coeffs2[0] * xVal * xVal
            + coeffs2[1] * xVal + coeffs2[2]);
        
        // Use a weighted average between the left and right values.
        leftValue *= xVal2;
        rightValue *= xVal;
        
        return (leftValue + rightValue);
    }
    
//------------------------------------------------------------------------------
    /**
     * Calculates the layer height of the cell at the given time step.
     * If the time step is between time steps of the Si3D output WSE field,
     * interpolates the values using the preceding and following values with
     * the method chosen by m_interpolationIndex.
     * @param cell  The cell location and variables.
     */
    protected void getLayerHeight(CellDescriptor cell) {
        
        CellVariables vars = cell.vars;
        ColumnVariables vars2D = cell.vars2D;
        
        // Retrieve the depth at the current location from the array.
        vars2D.depth = m_avgDepthArray[cell.lookup2D];
        vars2D.wse = 0;
        
        // Initialize local variables...
        vars2D.wseLocation = 0;
        vars2D.bottomLocation = 0;
        
        // Check that K is a defined layer.
        if (cell.k <= 0)
            cell.k = 1;
        vars2D.bottomLayerK = ((int)
        ((vars2D.depth - 0.00001) / m_zSpacing)) + 1;
        if (cell.k > vars2D.bottomLayerK) {
            cell.k = vars2D.bottomLayerK;
        }
        
        // Define the actual locations of the top and bottom of current
        // layer K...
        vars2D.bottomAdjustment = 0;
        vars2D.surfaceHeightAdjustment = 0;
        
        // Calculate the bottom adjustment in world units.
        // It is zero if K is not a bottom layer.
        if (cell.k == vars2D.bottomLayerK) {
            vars2D.bottomAdjustment = (cell.k * m_zSpacing) - vars2D.depth;
        }
        
        // Calculate the surface adjustment in world units.
        // It is zero if K is not a surface layer.
        if (cell.k == 1) {
            vars2D.wse = getWSE(cell.lookup2D, cell.ptmTS);
            if (cell.lookup2D >= 0 && m_2DVarsArray != null) {
                vars2D.surfaceHeightAdjustment = vars2D.wse;
            }
        }
        
        // Calculate the water surface elevation in grid coordinates.
        vars2D.wseLocation = (vars2D.wse / m_zSpacing) - 0.5;
        // Calculate the bottom location in grid coordinates.
        vars2D.bottomLocation =
            (-vars2D.depth / m_zSpacing) - 0.5;
        
        // Save certain member variables on the first time step.
        if (m_bSaveInitialRange) {
            int wseFieldTS =
                (int)Math.round((double)cell.ptmTS / (double)m_tsDivs);
            
            m_initialWSE = getVar2D(cell.lookup2D
                + (wseFieldTS * m_numWet2Dpoints)) * m_wseScale;
            m_initialGridDepth = -vars2D.depth;
            m_bSaveInitialRange = false;
        }
        
        // Calculate the layer height from the Z spacing and any
        // surface and bottom adjustments.
        vars.layerHeight = m_zSpacing;
        if (cell.k <= 1)
            vars.layerHeight += vars2D.surfaceHeightAdjustment;
        if (cell.k >= vars2D.bottomLayerK)
            vars.layerHeight -= vars2D.bottomAdjustment;
        
        return;
    }
//------------------------------------------------------------------------------
    /**
     * Gets the Water Surface Elevation the cell found at the provided lookup and
     * at the given PTM time step.
     * Multiplies the raw WSE by the WSE scale.
     * If the time step is between time steps of the Si3D output WSE field,
     * interpolates the value using the preceding and following values with
     * the method chosen by m_interpolationIndex.
     * @param lookup2D  The location lookup of the cell.
     * @param ts    The PTM time step at which the WSE is required.
     * @return  The WSE value for the cell at ts.
     */
    protected double getWSE(int lookup2D, int ts) {
        
        double wse = 0;
        
        // Retrieve the value using the selected interpolation type.
        switch (m_interpolationIndex) {
            // The default is no interpolation...
            case 0:
            default:
                // Retrieve the WSE value at the nearest available time step.
                int wseFieldTS = (int)Math.round((double)ts / (double)m_tsDivs);
                wse = getVar2D(lookup2D + (wseFieldTS * m_numWet2Dpoints));
                break;
                
                // For linear interpolation...
            case 1:
                // Calculate the WSE ts that precedes the given PTM ts
                int wseFieldTS1 = ts / m_tsDivs;
                int remainder = ts % m_tsDivs;
                
                // No need to interpolate if the PTM ts falls on the WSE ts.
                if (remainder == 0) {
                    wse = getVar2D(lookup2D + (wseFieldTS1 * m_numWet2Dpoints));
                    break;
                }
                
                // Calculate the WSE ts that follows and perform a bounds check.
                double wse1 =
                    getVar2D(lookup2D + (wseFieldTS1 * m_numWet2Dpoints));
                int wseFieldTS2 = wseFieldTS1 + 1;
                if (wseFieldTS2 >= m_numTimeSteps)
                    wseFieldTS2 = wseFieldTS1;
                
                // Retrieve the WSE value at the following time step.
                double wse2 =
                    getVar2D(lookup2D + (wseFieldTS2 * m_numWet2Dpoints));
                
                // Interpolate between the two values.
                double val1 = (double) remainder / (double) m_tsDivs;
                double val2 = 1 - val1;
                wse = (wse1 * val2) + (wse2 * val1);
                break;
                
                // For parabolic interpolation...
            case 2:
                // Determine the WSE ts before and after this PTM ts.
                wseFieldTS2 = ts / m_tsDivs;
                remainder = ts % m_tsDivs;
                
                // No need to interpolate if the PTM ts falls on the WSE ts.
                if (remainder == 0) {
                    wse = getVar2D(lookup2D + (wseFieldTS2 * m_numWet2Dpoints));
                    break;
                }
                
                wseFieldTS1 = wseFieldTS2 - 1;
                int wseFieldTS3 = wseFieldTS2 + 1;
                int wseFieldTS4 = wseFieldTS2 + 2;
                
                // Perform a bounds check.
                if (wseFieldTS1 < 1)
                    wseFieldTS1 = 1;
                if (wseFieldTS2 < 1)
                    wseFieldTS2 = 1;
                if (wseFieldTS1 > m_numTimeSteps)
                    wseFieldTS1 = m_numTimeSteps;
                if (wseFieldTS2 > m_numTimeSteps)
                    wseFieldTS2 = m_numTimeSteps;
                if (wseFieldTS3 >= m_numTimeSteps)
                    wseFieldTS3 = wseFieldTS2;
                if (wseFieldTS4 >= m_numTimeSteps)
                    wseFieldTS4 = wseFieldTS3;
                
                // Store the time step's WSE values.
                wse1 = getVar2D(lookup2D + (wseFieldTS1 * m_numWet2Dpoints));
                wse2 = getVar2D(lookup2D + (wseFieldTS2 * m_numWet2Dpoints));
                double wse3 =
                    getVar2D(lookup2D + (wseFieldTS3 * m_numWet2Dpoints));
                double wse4 =
                    getVar2D(lookup2D + (wseFieldTS4 * m_numWet2Dpoints));
                
                // Calculate the time after 0 and before 1.
                double tVal = (double) remainder / (double) m_tsDivs;
                double timeTill1 = 1 - tVal;
                
                // Get the parabolically interpolated answer.
                wse = getValueOnParabola(
                    wse1, wse2, wse3, wse4, tVal, timeTill1);
        }
        
        // Multiply by the WSE scale.
        wse *= m_wseScale;
        
        return wse;
    }
    
//------------------------------------------------------------------------------
    /**
     * Returns the lookup index of the specified cell for the vars3D array.
     * The lookup is a combination of the location within the grid and the time
     * step.  It will point to the index in an array of values stored for each
     * wet cell at each time step.
     * @param i             The I coordinate of the cell.
     * @param j             The J coordinate of the cell.
     * @param k             The K coordinate of the cell.
     * @param numI          The number of cells in I direction.
     * @param numJ          The number of cells in J direction.
     * @param numK          The number of cells in K direction.
     * @param numVarDims    The number of variable dimensions for the data.
     * @param ts            The time step of the data.
     * @param numWetPoints  The number of wet cells in the grid.
     * @param lookupArray   The array containing the location lookups of wet cells.
     * @return  The cell's lookup at the given time step.
     */
    protected int getLookupIndex(int i, int j, int k, int numI, int numJ,
        int numK, int numVarDims, int ts, int numWetPoints, int[] lookupArray) {
        
        // Set the lookup index to -1
        int lookupIndex = -1;
        
        // Perform a bounds check and then get the lookup from the array.
        // If the value is not in range, the already-assigned value of -1
        // will indicate it does not exist.
        if (i >= 0 && i < numI && j >= 0 && j < numJ && k >= 0 && k < numK) {
            
            lookupIndex = lookupArray[i * numK + j * numK * numI + k];
            if (lookupIndex >= 0)
                lookupIndex = (lookupIndex + ts * numWetPoints) * numVarDims;
        }
        
        return lookupIndex;
    }
    
//------------------------------------------------------------------------------
    /**
     * Returns the U, V, W, or other value of the specified cell from the 3D vars
     * series.  Handles the case of the series being stored in a random access file.
     * @param lookupIndex   The lookup index for finding the cell in the array of
     *                      wet cell values for all time steps.
     * @param dimNum        The dimension number for the type of variable needed.
     * @return  The value at the specified location, time step, and dimension.
     *          0 is returned if a value could not be found.
     */
    protected float getVar3D(int lookupIndex, int dimNum) {
        
        // Set the default return value to 0.
        float var = 0;
        
        // Random access file data series must be handled differently.
        if (m_3DVarsSeries instanceof RafDataSeries) {
            // Perform a bounds check first.
            if (lookupIndex >= 0) {
                var = ( (RafDataSeries) m_3DVarsSeries).getValueAt(
                    lookupIndex + dimNum);
            }
        }
        // Get the value straight from the array.
        else {
            // Perform a bounds check first.
            if (lookupIndex >= 0 &&
                (lookupIndex + dimNum) < m_3DVarsArray.length) {
                
                var = m_3DVarsArray[lookupIndex + dimNum];
            }
        }
        
        return var;
    }
    
//------------------------------------------------------------------------------
    /**
     * Returns the value at the specified index from the 2D vars series.
     * Handles the case of the series being stored in a random access file.
     * @param offset    The offset into the array for the value needed.
     *                  Assumes this is within the array bounds.
     * @return  The value at the specified offset.
     *          0 is returned if a value could not be found.
     */
    protected float getVar2D(int offset) {
        
        // Set the default return value to 0.
        float var = 0;
        
        // Retrieve the value, either from a random access file, or directly
        // from the array.
//        if (m_2DVarsSeries instanceof RafDataSeries) {
        var = ((RafDataSeries)m_2DVarsSeries).getValueAt(offset);
//        }
//        else {
//            var = m_2DVarsArray[offset];
//        }
        
        return var;
    }
    
//------------------------------------------------------------------------------
    /**
     * Creates a new array from a two-point line array (x1, y1, x2, y2).  The
     * new array's line is extended to cover the full cells at the end points.
     * Only horizontal or vertical lines are converted.  For example, if a
     * line starts at (1, 1) and ends at (4, 1) it is changed to (0.5, 1) to
     * (4.5, 1).
     * @param series    Series with interleaved X-Y endpoints of a line extending
     *                  to the center of a cell on each end.  Contains four
     *                  points (x1, y1, x2, y2).
     * @return  A new four-element array with (x1, y1, x2, y2) stretching to
     *          cover the entire cell on each end.
     */
    protected double[] cellBoundaryLine(DoubleDataSeries series) {
        
        // Get the original array and create a new array of the same size.
        double[] originalArray = series.getArray();
        double[] array = new double[originalArray.length];
        
        // Original array must have exactly four elements.
        if (originalArray.length != 4)
            return array;
        
        // Copy the values from the old array to the new array.
        array[0] = originalArray[0];
        array[1] = originalArray[1];
        array[2] = originalArray[2];
        array[3] = originalArray[3];
        
        //Determine if the line is vertical or horizontal.
        boolean bXSame = (Math.round(array[0]) == Math.round(array[2]));
        boolean bYSame = (Math.round(array[1]) == Math.round(array[3]));
        boolean bVertical;
        if (bXSame) {
            if (bYSame)
                bVertical = (series.getType() == 1);
            else
                bVertical = true;
        } else {
            if (bYSame)
                bVertical = false;
            else
                return array;
        }
        
        if (bVertical) {
            // If vertical, extend the top and bottom ends.
            if (array[1] < array[3]) {
                array[1] -= 0.501;
                array[3] += 0.501;
            } else {
                array[1] += 0.501;
                array[3] -= 0.501;
            }
        } else {
            // If horizontal, extend the left and right ends.
            if (array[0] < array[2]) {
                array[0] -= 0.501;
                array[2] += 0.501;
            } else {
                array[0] += 0.501;
                array[2] -= 0.501;
            }
        }
        
        return array;
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * If an error was detected during an operation, the class's error string
     * is appended.
     * @return  The string that describes this object.
     */
    public String toString() {
        
        String s = "Particle Path Calculator";
        
        if (m_errorString != null)
            s += m_errorString;
        
        return s;
    }
}

