/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-03
 * Date started:    2005-05-05
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.awt.Frame;
import java.io.*;
import java.util.*;
import javax.swing.*;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import gov.usgs.gr.tools.MultiThreader;

//==============================================================================
/**
 * Manages massive particle releases to create a spatially and temporally
 * varying map showing average particle tracking results.
 */

public class ParticleTrackCollator extends GrObject implements
    java.awt.event.ActionListener, gov.usgs.gr.tools.ThreadWorker {
    
    /**
     * Holds any error descriptions that should be shown to the user in
     * the properties dialog.
     */
    protected String m_errorString = null;
    /** The name of the data set's file. */
    protected String m_filename = "tracks.bin";
    
    protected int[][] m_rgbColumns = new int[4][];
    protected int[] m_columnAssignments = null;
    
    /** The working directory and ultimate output file. */
    protected File m_file =
        new File(System.getProperty("gov.usgs.gr.currentfile"));
    /** The default starting file name. */
    protected String m_startFileName = "";
    /** A handle to frame to use for showing dialogs. */
    protected Frame m_frame = null;
    
    protected Bathymetry m_bath = null;
    protected int m_numI;
    protected int m_numJ;
    protected int m_numK;
    protected double m_xSpacing;
    protected double m_ySpacing;
    protected double m_zSpacing;
    protected int[] m_2DLookupArray;
    protected int[] m_2DIndexArray;
    protected int[] m_3DLookupArray;
    protected int[] m_3DIndexArray;
    protected int m_numWetPoints2D = 0;
    protected int m_numTimes = 0;
    
    protected int m_cacheSize = 0;
    protected int m_cacheOffset = 0;
    protected boolean m_bUseByteCache = false;
    protected boolean m_bOomeOccurred = false;
    
    /** The data series containing data types that vary by time, X, and Y. */
    protected RafDataSeries m_3DVarsSeries = null;
    /** The data series containing data types that vary by time, X, and Y. */
    protected RafDataSeries m_2DVarsSeries = null;
    /** The general output from the Si3D model. */
    protected Si3dOutput m_si3dOut = null;
    /** The Water Surface Elevation scale */
    protected double m_wseScale = 1;
    
    /** A container that holds output data series containing particle paths. */
    protected GrObjectContainer m_outputContainer = null;
    /** A data series containing T-X-Y-R-G-B values. */
    protected DoubleDataSeries m_outputSeries6 = new DoubleDataSeries(6);
    /** A data series containing T-X-Y-V values. */
    protected DoubleDataSeries m_outputSeries4 = new DoubleDataSeries(4);
    /** A data series containing T-X-Y-P probability values. */
    protected DoubleDataSeries m_outputSeriesP = new DoubleDataSeries(
        3, "Collator Probability Map", 4);
    /** The percentage change below which optimization stops. */
    protected double m_optStop = 100;
    /**
     * The minimum percentage of particles that make it to the transit finish
     * line for a given cell that will cause the transit difference to be
     * calculated for that cell.
     *
     */
    protected double m_minTransitPortion = 0.5;
    
    protected int m_numThreads = 1;
    protected ParticleTracker[] m_ptArray = null;
    protected ArrayList<ParticleTracker> m_liveTrackers =
        new ArrayList<ParticleTracker>();
    protected ParticleTracker m_mainPT = null;
    
    //==========================================================================
    /**
     * An object that can be sorted and tell how an interleaved array can be
     * sorted.
     */
    public class SortableObject implements Comparable {
        int index;
        double[] array;
        static final int numDims = 10;
        static final double sec5 = 0.00005;
        
        //----------------------------------------------------------------------
        /**
         * Compares an object to this one for sorting.
         * @param o The object to compare to this one.
         */
        public int compareTo(Object o) {
            
            if (!(o instanceof SortableObject))
                return 0;
            
            SortableObject so = (SortableObject)o;
            
            int i1 = index * numDims;
            int i2 = so.index * numDims;
            
            // Compare times...
            if (array[i1] > (array[i2] + sec5))
                return 1;
            
            if (array[i1] < (array[i2] - sec5))
                return -1;
            
            // Times are equal, so compare release indexes...
            if (Math.round(array[i1 + 1]) > Math.round(array[i2 + 1]))
                return 1;
            
            if (Math.round(array[i1 + 1]) < Math.round(array[i2 + 1]))
                return -1;
            
            // Release indexes are equal, so compare finish lines...
            if (Math.round(array[i1 + 2]) > Math.round(array[i2 + 2]))
                return 1;
            
            if (Math.round(array[i1 + 2]) < Math.round(array[i2 + 2]))
                return -1;
            
            // Finish lines are equal, so compare finish times...
            if (array[i1 + 3] > array[i2 + 3])
                return 1;
            
            if (array[i1 + 3] < array[i2 + 3])
                return -1;
            
            return 0;
            
        }
    }
    
    //------------------------------------------------------------------------------
    /**
     * Creates an empty object.
     */
    public ParticleTrackCollator() {
        
        String parent = new File(System.getProperty("gov.usgs.gr.currentfile"))
            .getParent();
        m_startFileName = parent + "/" + "tracks.bin";
        
        m_file = new File(m_startFileName);
        
        m_outputContainer = new GrObjectContainer();
        m_outputContainer.setName("Output");
        m_outputContainer.setProperty("title", "Output");
        add(m_outputContainer);
        m_outputContainer.add(m_outputSeries6);
        m_outputContainer.add(m_outputSeries4);
        m_outputContainer.add(m_outputSeriesP);
        try {
            jbInit();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }
    
    //------------------------------------------------------------------------------
    /**
     * Adds an object as a child, but only if it is an instance of ParticleTracker.
     * @param gro   The object to add.
     * @return  Whether the object was successfully added.
     */
    public boolean add(GrObject gro) {
        
        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();
        
        // Add allowable objects...
        if (gro instanceof ParticleTracker) {
            super.add(gro);
            return true;
        }
        // Add allowable objects...
        if (gro instanceof IntDataSeries) {
            super.add(gro);
            return true;
        }
        
        
        // Add object container...
        if (gro instanceof GrObjectContainer) {
            GrObjectContainer groc = (GrObjectContainer)gro;
            GrHashList props = groc.getProperties();
            GrProperty p = (GrProperty)props.get("title");
            if (p == null)
                return false;
            
            // Compare the title of groc to each container and replace the
            // current one if there is a match.
            GrObjectContainer oldGroc = m_outputContainer;
            
            int index = m_children.size();
            if (oldGroc != null) {
                index = m_children.indexOf(oldGroc);
                oldGroc.deleteObserver(this);
                m_children.remove(oldGroc);
            }
            if (index < 0)
                index = 0;
            m_children.add(index, groc);
            groc.addObserver(this);
            groc.setName("");
            
            m_outputContainer = groc;
            
            // Look for DoubleDataSeries within the output
            // container and use them for output if they are
            // there.
            Enumeration enumeration =
                m_outputContainer.children();
            while (enumeration.hasMoreElements()) {
                GrObject child =
                    (GrObject) enumeration.nextElement();
                if (child instanceof DoubleDataSeries) {
                    DoubleDataSeries ds = (DoubleDataSeries)child;
                    if (ds.getNumDimensions() == 4) {
                        if (ds.getTitle().equals("Collator Probability Map"))
                            m_outputSeriesP = ds;
                        else
                            m_outputSeries4 = ds;
                    }
                    if (ds.getNumDimensions() == 6)
                        m_outputSeries6 = ds;
                }
            }
            
            // Make sure there are no other output series
            // in the output container besided those defined.
            m_outputContainer.clear();
            m_outputContainer.add(m_outputSeries6);
            m_outputContainer.add(m_outputSeries4);
            m_outputContainer.add(m_outputSeriesP);
            
            // Set default property values.
            m_outputSeries6.getDimension(0).
                setProperty("locked", "true");
            m_outputSeries6.getDimension(1).
                setProperty("locked", "true");
            
            m_outputSeries6.setTitle(
                "Collator RGB Output");
            
            m_outputSeries4.getDimension(0).
                setProperty("locked", "true");
            m_outputSeries4.getDimension(1).
                setProperty("locked", "true");
            
            m_outputSeries4.setTitle(
                "Collator Scalar Output");
            
            m_outputSeriesP.getDimension(0).
                setProperty("locked", "true");
            m_outputSeriesP.getDimension(1).
                setProperty("locked", "true");
            
            m_outputSeriesP.setTitle(
                "Collator Probability Map");
            
            
            m_bHasChanged = true;
            return true;
        }
        
        return false;
    }
    
    //------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public GrHashList getProperties() {
        
        GrHashList props = super.getProperties();
        
        props.put("filename", new GrProperty(
            m_filename, "File Name"));
        props.put("optStop", new GrProperty(
            m_optStop, "Optimization Stop Percentage"));
        
        String s = "";
        for (int i = 0; i < 4; i++) {
            if (m_rgbColumns[i] != null) {
                for (int j = 0; j < m_rgbColumns[i].length; j++) {
                    if (j > 0)
                        s += "+";
                    s += m_rgbColumns[i][j];
                }
            }
            
            if (i < 2)
                s += ", ";
        }
        props.put("rgbAssignments",
            new GrProperty(s, "RGB Boundary Assignments"));
        props.put("numThreads", new GrProperty(m_numThreads,
            "Number of Threads"));
        props.put("cacheSize", new GrProperty(m_cacheSize, "Cache Size"));
        props.put("cacheOffset", new GrProperty(m_cacheOffset, "Cache Offset"));
        props.put("useByteCache", new GrProperty(m_bUseByteCache,
            "Use Byte Cache"));
        props.put("minTransitPercentage", new GrProperty(
            m_minTransitPortion * 100, "Minimum Transit Percentage"));
        
        return props;
    }
    
    //------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public void setProperties(HashMap props) {
        
        super.setProperties(props);
        
        GrProperty p;
        
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;
                
                if (key.equals("optStop"))
                    m_optStop = (Double)p.value;
                
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                if (key.equals("numThreads"))
                    m_numThreads = (Integer)p.value;
                if (key.equals("cacheSize"))
                    m_cacheSize = (Integer) p.value;
                if (key.equals("cacheOffset"))
                    m_cacheOffset = (Integer)p.value;
                if (key.equals("useByteCache"))
                    m_bUseByteCache = (Boolean)p.value;
                if (key.equals("minTransitPercentage")) {
                    m_minTransitPortion = (Double)p.value;
                    m_minTransitPortion /= 100;
                    if (m_minTransitPortion > 1)
                        m_minTransitPortion = 1;
                }
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }
    }
    
    //------------------------------------------------------------------------------
    /**
     * Set a single property.
     */
    public void setProperty(String key, String value) {
        
        super.setProperty(key, value);
        
        if (key.equals("optStop"))
            m_optStop = Double.parseDouble(value);
        if (key.equals("filename")) {
            m_filename = value;
        }
        if (key.equals("rgbAssignments")) {
            int maxColumn = 0;
            String[] columnTokens = AsciiIO.parseLine(value, ",");
            for (int i = 0; i < 4; i++) {
                if (i >= columnTokens.length) {
                    m_rgbColumns[i] = new int[0];
                    continue;
                }
                
                m_rgbColumns[i] = AsciiIO.parseLineAsInts(columnTokens[i], "+");
                
                for (int j = 0; j < m_rgbColumns[i].length; j++) {
                    if (m_rgbColumns[i][j] > maxColumn)
                        maxColumn = m_rgbColumns[i][j];
                }
            }
            
            // Initialize all columns to no color (-1).
            m_columnAssignments = new int[maxColumn + 1];
            for (int i = 0; i < m_columnAssignments.length; i++) {
                m_columnAssignments[i] = -1;
            }
            
            // Set certain columns to tally on Red, Green, or Blue.
            for (int i = 0; i < 4; i++) {
                for (int j = 0; j < m_rgbColumns[i].length; j++) {
                    if (m_rgbColumns[i][j] >= 0)
                        m_columnAssignments[m_rgbColumns[i][j]] = i;
                }
            }
        }
        if (key.equals("numThreads"))
            m_numThreads = Integer.parseInt(value);
        if (key.equals("cacheSize"))
            m_cacheSize = Integer.parseInt(value);
        if (key.equals("cacheOffset"))
            m_cacheOffset = Integer.parseInt(value);
        if (key.equals("useByteCache"))
            m_bUseByteCache = Boolean.parseBoolean(value);
        if (key.equals("minTransitPercentage")) {
            m_minTransitPortion = Double.parseDouble(value);
            m_minTransitPortion /= 100;
            if (m_minTransitPortion > 1)
                m_minTransitPortion = 1;
        }
    }
    
    //------------------------------------------------------------------------------
    /**
     *
     */
    public void cacheRafData() {
        
        Si3dOutput si3dOutput = (Si3dOutput)firstChildInstanceOf(
            "gov.usgs.sfhydro.si3d.Si3dOutput", true);
        if (si3dOutput == null)
            return;
        
        m_si3dOut = si3dOutput;
        RafDataSeries series3D = (RafDataSeries)
            m_si3dOut.getContainerSeries("3-D Variables");
        
        // Find the 2D series and perform bounds checking on the cache size.
        RafDataSeries series2D = (RafDataSeries)
            m_si3dOut.getContainerSeries("2-D x-y Variables");
        
        if (m_cacheSize == 0) {
            series3D.createCacheSegments(1, 0);
            series2D.createCacheSegments(1, 0);
            series3D.createByteCacheSegments(1, 0);
            series2D.createByteCacheSegments(1, 0);
            System.gc();
            return;
        }
        
        // Get the size of the 3D series.
        int count = series3D.getCount();
        int numDims = series3D.getNumDimensions();
        int numPoints = count * numDims;
        
        // Get the size of the 2D series.
        int count2D = series2D.getCount();
        int numDims2D = series2D.getNumDimensions();
        int numPoints2D = count2D * numDims2D;
        
        // Use temp variables to adjust the cache size.
        int cacheSize = m_cacheSize;
        int cacheOffset = m_cacheOffset;
        int maxCacheSize = numPoints - cacheOffset;
        
        // Determine how much memory is available.
        System.gc();
        Runtime runtime = Runtime.getRuntime();
        long maxMemory = runtime.maxMemory();
        long totalMemory = runtime.totalMemory();
        long freeMemory = runtime.freeMemory();
        long availableMemory = maxMemory - totalMemory + freeMemory;
        
        // If cache size is less than zero, the number portion represents the
        // percentage of memory to use.
        double usePortion = 1;
        if (cacheSize < 0)
            usePortion = -(double)cacheSize / 100;
        
        // Convert from byte to float units, then use some portion of what's
        // available.
        int maxNumElements = (int)(availableMemory * 0.25 * usePortion);
        if ((maxCacheSize + numPoints2D) > maxNumElements) {
            maxCacheSize = maxNumElements * 2;
            m_bUseByteCache = true;
        }
        
        // Perform bounds checking on the cache size and offset.
        if (cacheOffset < 0 || cacheOffset >= numPoints) {
            cacheOffset = 0;
            cacheSize = 0;
        } else {
            if (cacheSize > maxCacheSize)
                cacheSize = maxCacheSize;
        }
        
        // If cache size is less than zero it is a signal to set it to the data
        // array size.
        if (cacheSize < 0)
            cacheSize = maxCacheSize;
        
        // Use up to half of the cache for the 2D file.
        int cacheSize2D = cacheSize / 2;
        if (cacheSize2D > numPoints2D)
            cacheSize2D = numPoints2D;
        cacheSize -= cacheSize2D;
        
        // Allocate the cache.
        System.out.println("Resizing cache to " + cacheSize
            + " elements, beginning at " + cacheOffset + "  (" + cacheSize2D
            + " 2D elements) (" + availableMemory + " available)");
        
        //        // Break the cache into 16 segments.
        //        int numSegments = 16;
        //        cacheSize /= numSegments;
        //        cacheSize += numSegments;
        //        cacheSize2D /= numSegments;
        //        cacheSize2D += numSegments;
        
        
        if (m_bUseByteCache) {
            series3D.createByteCacheSegments(1, cacheSize);
            series3D.fillByteCacheSegment(0, cacheOffset);
            series2D.createByteCacheSegments(1, cacheSize2D);
            series2D.fillByteCacheSegment(0, 0);
        } else {
            series3D.createCacheSegments(1, cacheSize);
            series3D.fillCacheSegment(0, cacheOffset);
            series2D.createCacheSegments(1, cacheSize2D);
            series2D.fillCacheSegment(0, 0);
        }
        
        //        series3D.createCacheSegments(numSegments, cacheSize);
        //        series2D.createCacheSegments(numSegments, cacheSize2D);
        //        for (int i = 0; i < numSegments; i++) {
        //            series3D.fillCacheSegment(i, cacheOffset + (i * cacheSize));
        //            series2D.fillCacheSegment(i, i * cacheSize2D);
        //        }
    }
    
    //------------------------------------------------------------------------------
    /**
     * For each point specified by the input arrays, release and track a particle.
     * Store the particle path in the output data series.
     */
    public String prepFiles() {
        
        // Set up the file chooser dialog.
        JFileChooser fc = new JFileChooser();
        fc.setDialogTitle("Choose File");
        fc.setCurrentDirectory(m_file);
        fc.setSelectedFile(m_file);
        fc.setDialogType(JFileChooser.SAVE_DIALOG);
        int returnVal = fc.showSaveDialog(m_frame);
        
        // If the user pressed OK...
        boolean bMerge = true;
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            m_file = fc.getSelectedFile();
            m_filename = m_file.getName();
            
            if (m_file.exists()) {
                returnVal = JOptionPane.showConfirmDialog(m_frame,
                    "Merge existing file rather than delete it?");
                if (returnVal == JOptionPane.NO_OPTION)
                    bMerge = false;
            }
            
            if (returnVal == JOptionPane.CANCEL_OPTION)
                return null;
        }
        
        int strIndex = m_filename.indexOf(".bin");
        String fileNameBase = null;
        if (strIndex < 0)
            fileNameBase = m_filename;
        else
            fileNameBase = m_filename.substring(0, strIndex);
        
        final String finalFileNameBase = fileNameBase;
        final boolean fbMerge = bMerge;
        
        // Check if there are existing files of type filename base.
        FilenameFilter fileNameFilter = new FilenameFilter() {
            public boolean accept(File dir, String name) {
                if (name.endsWith(".bin")) {
                    if (name.startsWith(finalFileNameBase)) {
                        if (!(fbMerge && name.equals(m_filename)))
                            return true;
                    }
                }
                
                return false;
            }
        };
        
        String[] namesArray = m_file.getParentFile().list(fileNameFilter);
        
        // If there are existing files, confirm overwrite and delete them all
        // before starting.
        if (namesArray.length > 0) {
            int choice = JOptionPane.YES_OPTION;
            choice = JOptionPane.showConfirmDialog(m_frame,
                "File(s) with same base name will be\nremoved before starting.",
                "Confirm",
                JOptionPane.OK_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE);
            
            if (choice == JOptionPane.YES_OPTION) {
                for (int i = 0; i < namesArray.length; i++) {
                    File matchingFile =
                        new File(m_file.getParent() + "/" + namesArray[i]);
                    boolean bSuccess = matchingFile.delete();
                }
            }
        }
        
        return fileNameBase;
    }
    
    //------------------------------------------------------------------------------
    /**
     * For each point specified by the input arrays, release and track a particle.
     * Store the particle path in the output data series.
     */
    public void recalculate(String fileNameBase) {
        
        // Read the input variables.
        int status = readVariables();
        if (status < 0) {
            System.err.println("Error reading input objects.");
            return;
        }
        
        
        ParticleTracker ptm = (ParticleTracker)firstChildInstanceOf(
            "gov.usgs.sfhydro.si3d.ParticleTracker", false);
        
        // Initialize to -1 for the first run.
        double changeFactor = -1;
        
        int numRuns = 1;
        
        // Clear old probability maps.
        while (true) {
            if (m_3DVarsSeries instanceof RafDataSeries) {
                ((RafDataSeries)m_3DVarsSeries).resetCacheStats();
            }
            GregorianCalendar gc1 = new GregorianCalendar();
            long time1 = gc1.getTimeInMillis();
            
            System.out.print("Releasing " + numRuns + ": ");
            List fileList = new ArrayList();
            List runFileList = null;
            int fileNum = 0;
            for (int i = 0; i < numRuns; i++) {
                System.out.print("" + (i+1));
                runFileList = generateTrackStats(fileNameBase, fileNum);
                if (runFileList == null) {
                    try {
                        for (int g = 0; g < 3; g++) {
                            System.gc();
                            Thread.currentThread().sleep(3000);
                        }
                        System.gc();
                    } catch (InterruptedException ex) {
                        ex.printStackTrace();
                    }

                    break;
                }
                fileNum += runFileList.size();
                fileList.addAll(runFileList);
                System.out.print("..");
            }
            
            System.out.println("done");
            
            GregorianCalendar gc2 = new GregorianCalendar();
            long time2 = gc2.getTimeInMillis();
            
            // Merge the files that were created back in with the main output
            // file and then delete the newly created ones.
            if (m_file.exists())
                fileList.add(m_file);
            File[] inFiles = (File[]) fileList.toArray(new File[0]);
            File[] delFiles = (File[]) fileList.toArray(new File[0]);
            String origOutName = m_file.getParent() + "/" + m_filename;
            File tempOutFile = new File(
                m_file.getParent() + "/" + m_filename + ".temp");
            mergeFiles(inFiles, tempOutFile);
            for (int i = 0; i < delFiles.length; i++) {
                boolean bSuccess = delFiles[i].delete();
                if (!bSuccess)
                    System.out.println("Could not delete " + delFiles[i]);
            }
                
            m_file = new File(origOutName);
            boolean bSuccess = tempOutFile.renameTo(m_file);

            // Return if OOME occurred.
            if (m_bOomeOccurred)
                return;
            
            // Write out the percentage of disk cache hits.
            if (m_3DVarsSeries instanceof RafDataSeries) {
                long readHits = ((RafDataSeries)m_3DVarsSeries).getNumReadHits();
                long readTotal = ((RafDataSeries)m_3DVarsSeries).getNumReads();
                long readMin = ((RafDataSeries)m_3DVarsSeries).getReadMin();
                long readMax = ((RafDataSeries)m_3DVarsSeries).getReadMax();
                System.out.println("Read hits: " + readHits + " / " + readTotal
                    + " = " + (100 * ( (float) readHits) / ( (float) readTotal))
                    + "%");
                System.out.println("Read Range: " + readMin + " to " + readMax);
                
                long readRange = readMax - readMin;
                long newCacheOffset = readMin - (long)(readRange * 0.005);
                System.out.println("Adjusting cache offset by " + 
                    (m_cacheOffset - newCacheOffset));
                m_cacheOffset = (int)newCacheOffset;
            }

            int numOutputDims = (m_minTransitPortion >= 0) ? 4 : 6;
            double[] oldOutputArray;
            if (numOutputDims == 4)
                oldOutputArray = m_outputSeries4.getArray();
            else
                oldOutputArray = m_outputSeries6.getArray();
            createAnimation();
            double[] newOutputArray;
            if (numOutputDims == 4)
                newOutputArray = m_outputSeries4.getArray();
            else
                newOutputArray = m_outputSeries6.getArray();
            double averageChange = calculateAverageChange(
                oldOutputArray, newOutputArray, numOutputDims);
            
            if (changeFactor < 0 && m_optStop < 99) {
                changeFactor = 1;
                System.out.println(
                    "Creating second comparison point before optimizing");
                continue;
            }
            
            GrProperty p = (GrProperty)m_mainPT.getProperties().get(
                "randomParticlesPerTS");
            int numReleases = (Integer)p.value * m_numThreads;
            changeFactor = (double)m_numTimes * (double)m_numWetPoints2D /
                ((double)numReleases * (double)numRuns);
            if ((averageChange * changeFactor) > 100) {
                System.out.println("m_numTimes * m_numWetPoints2D = " + m_numTimes + " * "
                    + m_numWetPoints2D + " = " + (m_numTimes * m_numWetPoints2D));
                System.out.println("numReleases * numRuns = " + numReleases + " * " + numRuns
                    + " = " + (numReleases * numRuns));
                System.out.println("changeFactor  = " + changeFactor);
                System.out.println("averageChange = " + averageChange);
            }
            averageChange *= changeFactor;
            double stopThreshold = (m_optStop < 1E-9 ? 1E7 : (100 / m_optStop));
            int coverageNumRuns = (int)Math.max(stopThreshold, changeFactor);
            
            
            if (averageChange < m_optStop || m_optStop > 99) {
                System.out.println("Stopped at change of " + averageChange
                    + " %  ( below " + m_optStop + " % )");
                return;
            }
            
            createProbabilityMap();
            
            GregorianCalendar gc3 = new GregorianCalendar();
            long time3 = gc3.getTimeInMillis();
            long runTime = time2 - time1;
            long optTime = time3 - time2;
            int optNumRuns =
                numRuns * (int)(100 * ((double)optTime / (double)runTime));
            int proposedNumRuns = Math.min(optNumRuns, coverageNumRuns);
            
            // Cap the run time at one hour.
            int oneHourNumRuns =
                (int)Math.round(3600000 / (double)runTime) * numRuns;
            if (proposedNumRuns > oneHourNumRuns)
                proposedNumRuns = oneHourNumRuns;
            
            // Cap the increase at 10 times the current number.
            if (proposedNumRuns > (numRuns * 5))
                numRuns *= 5;
            else
                numRuns = proposedNumRuns;
            
            // Cap the number of runs at 50.
            if (numRuns > 50)
                numRuns = 50;
            
            // Avoid underflow.
            if (numRuns < 1) {
                numRuns = 1;
            }
            
            // Write status information to the screen.
            String rtString = "Run time = " + runTime + " ms";
            if (runTime > 60000) {
                rtString += " (" + (int)(runTime / 60000) + ":"
                    + ((runTime % 60000) / 1000) + ")";
            }
            System.out.println(rtString);
            System.out.println("Opt time = " + optTime + " ms");
            System.out.println("Animation changed by " + averageChange + "%  " +
                "( X " + changeFactor + ")  ( above " + m_optStop + "% )");
        }
    }
    
    //------------------------------------------------------------------------------
    /**
     */
    public double calculateAverageChange(double[] oldArray, double[] newArray,
        int numDims) {
        
        if (oldArray.length != newArray.length)
            return 100;
        
        int numPoints = oldArray.length / numDims;
        double totalDiff = 0;
        for (int i = 0; i < numPoints; i++) {
            int offset = i * numDims;
            if (numDims == 4) {
                double oldValue = oldArray[offset + 3];
                double newValue = newArray[offset + 3];
                double diff = Math.abs(oldValue - newValue);
                double maxVal = Math.max(oldValue, newValue);
                if (maxVal != 0)
                    diff /= maxVal;
                if (diff > 1)
                    diff = 1;
                totalDiff += diff;
            } else {
                double redDiff =
                    Math.abs(oldArray[offset + 3] - newArray[offset + 3]);
                double greenDiff =
                    Math.abs(oldArray[offset + 4] - newArray[offset + 4]);
                double blueDiff =
                    Math.abs(oldArray[offset + 5] - newArray[offset + 5]);
                double colorDiff = Math.min(1, redDiff + greenDiff + blueDiff);
                
                totalDiff += colorDiff;
            }
        }
        
        double averageChange = totalDiff / (double)numPoints;
        
        return averageChange * 100;
    }
    
    //------------------------------------------------------------------------------
    /**
     */
    public List generateTrackStats(String fileNameBase, int fileNum) {
        
        ArrayList fileList = new ArrayList();
        
        int numInDims = 10;
        int numOutDims = 10;
        int ptmIndex = fileNum;
        // Set off each ParticleTracker child.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (!(gro instanceof ParticleTracker))
                continue;
            
            m_mainPT = (ParticleTracker)gro;
            double[] allStats = null;
            
            if (m_numThreads == 1) {
                m_mainPT.createTransitTimeDistribution();
                allStats = m_mainPT.getTrackStats();
            } else {
                m_ptArray = new ParticleTracker[m_numThreads];
                
                MultiThreader multiThreader =
                    new MultiThreader(this, m_numThreads, "PTM Stats");
                multiThreader.runThreads();
        
                // Clean up any old trackers that were left without a thread during
                // an OOME.
                while (m_liveTrackers.size() > 0) {
                    ParticleTracker oldTracker = m_liveTrackers.get(0);
                    oldTracker.clear();
                    m_liveTrackers.remove(0);
                }
                
                m_bOomeOccurred = multiThreader.getOomeOccurred();
                if (m_bOomeOccurred) {
                    return null;
                }
                
                for (int i = 0; i < m_ptArray.length; i++) {
                    double[] theseStats = m_ptArray[i].getTrackStats();
                    if (allStats == null) {
                        allStats = new double[theseStats.length * m_numThreads];
                    }
                    int offset = i * theseStats.length;
                    for (int j = 0; j < theseStats.length; j++) {
                        allStats[offset + j] = theseStats[j];
                    }
                }
            }
            
            int numPoints = allStats.length / numInDims;
            SortableObject[] sortableArray = new SortableObject[numPoints];
            for (int i = 0; i < numPoints; i++) {
                sortableArray[i] = new SortableObject();
                sortableArray[i].index = i;
                sortableArray[i].array = allStats;
            }
            Arrays.sort(sortableArray);
            double[] outStats = new double[numPoints * numOutDims];
            for (int i = 0; i < numPoints; i++) {
                int index = sortableArray[i].index;
                for (int j = 0; j < numOutDims; j++) {
                    outStats[i * numOutDims + j] =
                        ((double)allStats[index * numInDims + j]);
                }
            }
            
            // Determine path and filename.
            NumberFormat form = new DecimalFormat("000");
            String newFilename = m_file.getParent() + "/"
                + fileNameBase + "_" + form.format(ptmIndex + 1) + ".bin";
            
            // Sort stats by release time and then by release cell index
            RafDataSeries outRafSeries =
                new gov.usgs.gr.data.RafDataSeries(numOutDims);
            outRafSeries.setProperty("file", newFilename);
            outRafSeries.setProperty("diskType", "DOUBLE");
            outRafSeries.setProperty("memType", "DOUBLE");
            outRafSeries.setProperty("bigEndian", "true");
            outRafSeries.openRAF("rwd");
            
            outRafSeries.writeValuesFrom(outStats, 0, 0, outStats.length);
            outRafSeries.closeRAF();
            
            fileList.add(new File(newFilename));
            
            ptmIndex++;
        }
        
        return fileList;
    }
    
    //------------------------------------------------------------------------------
    /**
     *
     */
    public boolean performThreadWork() {
        
        ParticleTracker pt;
        synchronized (m_mainPT) {
            try {
                // Create a copy of the main PT that shares certain objects.
                pt = (ParticleTracker)(m_mainPT.getClass().newInstance());
                m_liveTrackers.add(pt);
                pt.setProperties(m_mainPT.getProperties());
                Enumeration enum1 = m_mainPT.children();
                while (enum1.hasMoreElements()) {
                    GrObject child = (GrObject)enum1.nextElement();
                    
                    // Determine if the child is the output container.
                    boolean bOutputContainer = false;
                    if (child instanceof GrObjectContainer) {
                        GrProperty p =
                            (GrProperty)child.getProperties().get("title");
                        if (((String)p.value).equals("Output"))
                            bOutputContainer = true;
                    }
                    
                    // Add the original child unless it is the particle path
                    // calculator or the output container.
                    if (child instanceof ParticlePathCalculator ||
                        child instanceof GrObjectContainer) {
                        
                        if (bOutputContainer) {
                        } else {
                            pt.add(child.createCopy(true));
                        }
                    } else {
                        pt.add(child);
                    }
                }
            } catch (IllegalAccessException iae) { return false;} catch (InstantiationException ie) {return false;}
        }
        
        pt.createTransitTimeDistribution();
        pt.clear();
        m_liveTrackers.remove(pt);
        
        synchronized (m_ptArray) {
            for (int i = 0; i < m_ptArray.length; i++) {
                if (m_ptArray[i] == null) {
                    m_ptArray[i] = pt;
                    break;
                }
            }
        }
        
        return false;
    }
    
    //------------------------------------------------------------------------------
    /**
     * Create a DoubleDataSeries with Time-X-Y-Red-Green-Blue values showing the
     * average values at each time and X-Y location.
     */
    public void createAnimation() {
        
        // Set the number of dimensions in the input file and open it.
        int numInDims = 10;
        RafDataSeries inRafSeries =
            new gov.usgs.gr.data.RafDataSeries(numInDims);
        inRafSeries.setProperty("file", m_file.getAbsolutePath());
        inRafSeries.setProperty("diskType", "DOUBLE");
        inRafSeries.setProperty("memType", "DOUBLE");
        inRafSeries.setProperty("bigEndian", "true");
        inRafSeries.createManagedCache(2, 1000);
        
        // Define local variables.
        int count = inRafSeries.getCount();
        double releaseTime = 0;
        double transitTime = 0;
        m_numTimes = 0;
        double lastTime = Double.NEGATIVE_INFINITY;
        // Define five seconds as a fraction of a day.
        double sec5 = 0.00005;
        
        // Don't count on evenly spaced times.  Read the entire file and count
        // up the time steps.
        for (int i = 0; i < count; i++) {
            int offset = i * numInDims;
            releaseTime = inRafSeries.getDoubleValueAt(offset);
            if (releaseTime > (lastTime + sec5)) {
                m_numTimes++;
                lastTime = releaseTime;
            }
        }
        
        // Create an out array to hold the time steps.
        int numOutDims = 6;
        if (m_minTransitPortion >= 0)
            numOutDims = 4;
        double[] outArray =
            new double[m_numTimes * m_numWetPoints2D * numOutDims];
        int[] countArray = new int[m_numWetPoints2D];
        int[] totalCounts = new int[m_numWetPoints2D];
        int[] numReleasesArray = null;
        if (m_minTransitPortion >= 0) {
            numReleasesArray = new int[m_numTimes * m_numWetPoints2D];
            Arrays.fill(numReleasesArray, 0);
        }
        // Initialize the X, Y, R, G, and B dimensions of the array.
        for (int j = 0; j < m_numTimes; j++) {
            for (int i = 0; i < m_numWetPoints2D; i++) {
                int index = (j * m_numWetPoints2D + i) * numOutDims;
                outArray[index + 1] = m_2DIndexArray[i * 2] * m_xSpacing;
                outArray[index + 2] = m_2DIndexArray[i * 2 + 1] * m_ySpacing;
                // Use -1 RGB values to indicate no releases and 0 to indicate
                // an unknown result.
                outArray[index + 3] = -1;
                if (m_minTransitPortion < 0) {
                    outArray[index + 4] = -1;
                    outArray[index + 5] = -1;
                }
                countArray[i] = 0;
                totalCounts[i] = 0;
            }
        }
        
        // Run through the list of particle releases.
        lastTime = Integer.MIN_VALUE;
        int counter = 0;
        for (int i = 0; i < count; i++) {
            // Read the time, release index, and which finish line was crossed.
            int offset = i * numInDims;
            releaseTime = inRafSeries.getDoubleValueAt(offset);
            int releaseIndex3D = (int)Math.round(
                inRafSeries.getDoubleValueAt(offset + 1));
            if (releaseIndex3D < 0) {
                System.out.println("Negative indexes encountered.  ("
                    + releaseIndex3D + " at " + i + ")");
                continue;
            }
            
            int currentI = m_3DIndexArray[releaseIndex3D * 3];
            int currentJ = m_3DIndexArray[releaseIndex3D * 3 + 1];
            int ijIndex = currentI + (currentJ * m_numI);
            int releaseIndex2D = m_2DLookupArray[ijIndex];
            
            int lineCrossed = (int)Math.round(
                inRafSeries.getDoubleValueAt(offset + 2));
            transitTime =
                inRafSeries.getDoubleValueAt(offset + 3) - releaseTime;
            
            // If this is the first index of a new time step.
            if (releaseTime != lastTime) {
                // If not the first time step...
                if (i != 0) {
                    for (int j = 0; j < m_numWetPoints2D; j++) {
                        // Write the time value to the out array for all
                        // the cells of the previous time step.
                        int index = (counter + j) * numOutDims;
                        outArray[index] = lastTime;
                        
                        // Normalize the RGB for each cell to the number
                        // released at that cell.
                        if (countArray[j] != 0) {
                            double numReleased = (double) countArray[j];
                            if (m_minTransitPortion >= 0) {
                                if (totalCounts[j] > 0) {
                                    double totalCount = totalCounts[j];
                                    double portion =
                                        numReleased / totalCount;
                                    if (portion > m_minTransitPortion)
                                        outArray[index + 3] /= numReleased;
                                    else
                                        outArray[index + 3] = -1;
                                } else {
                                    outArray[index + 3] = -1;
                                }
                            } else {
                                outArray[index + 3] /= numReleased;
                                outArray[index + 4] /= numReleased;
                                outArray[index + 5] /= numReleased;
                            }
                            countArray[j] = 0;
                            totalCounts[j] = 0;
                        }
                    }
                    
                    // Move the output counter to the next time step.
                    counter += m_numWetPoints2D;
                }
                
                // Keep track of this time so it can be used once the next
                // time appears.
                lastTime = releaseTime;
            }
            
            if (m_minTransitPortion >= 0) {
                // If it was previously -1 for no releases, change it to 0.
                int outIndex = (counter + releaseIndex2D) * numOutDims;
                // Increase the tally for the appropriate cell if the finish line
                // crossed matches one of the column assignments.
                if (lineCrossed >= 0 && lineCrossed < m_columnAssignments.length) {
                    if (m_columnAssignments[lineCrossed] == 2) {
                        
                        if (outArray[outIndex + 3] < 0) {
                            outArray[outIndex + 3] = 0;
                        }
                        outArray[(counter + releaseIndex2D) * numOutDims + 3]
                            += transitTime;
                        
                        // Increase the release counter for that cell.
                        countArray[releaseIndex2D]++;
                    }
                }
                totalCounts[releaseIndex2D]++;
            } else {
                // If it was previously -1 for no releases, change it to 0.
                int outIndex = (counter + releaseIndex2D) * numOutDims;
                if (outArray[outIndex + 3] < 0) {
                    outArray[outIndex + 3] = 0;
                    outArray[outIndex + 4] = 0;
                    outArray[outIndex + 5] = 0;
                }
                // Increase the tally for the appropriate cell if the finish line
                // crossed matches one of the column assignments.
                if (lineCrossed >= 0 && lineCrossed < m_columnAssignments.length) {
                    if (m_columnAssignments[lineCrossed] == 0)
                        outArray[(counter + releaseIndex2D) * numOutDims + 3]++;
                    if (m_columnAssignments[lineCrossed] == 1)
                        outArray[(counter + releaseIndex2D) * numOutDims + 4]++;
                    if (m_columnAssignments[lineCrossed] == 2)
                        outArray[(counter + releaseIndex2D) * numOutDims + 5]++;
                }
                // Increase the release counter for that cell.
                countArray[releaseIndex2D]++;
            }
        }
        
        // Close the in series.
        inRafSeries.closeRAF();
        
        // Finish processing the last time step.
        for (int j = 0; j < m_numWetPoints2D; j++) {
            int index = (counter + j) * numOutDims;
            outArray[index] = releaseTime;
            if (countArray[j] != 0) {
                double numReleased = (double) countArray[j];
                if (m_minTransitPortion >= 0) {
                    if (totalCounts[j] > 0) {
                        double totalCount = totalCounts[j];
                        double portion =
                            numReleased / totalCount;
                        if (portion > m_minTransitPortion)
                            outArray[index + 3] /= numReleased;
                        else
                            outArray[index + 3] = -1;
                    } else {
                        outArray[index + 3] = -1;
                    }
                } else {
                    outArray[index + 3] /= numReleased;
                    outArray[index + 4] /= numReleased;
                    outArray[index + 5] /= numReleased;
                }
                countArray[j] = 0;
                totalCounts[j] = 0;
            }
        }
        
        // Update the out series.
        if (m_minTransitPortion >= 0)
            m_outputSeries4.setArray(outArray);
        else
            m_outputSeries6.setArray(outArray);
    }
    
    //------------------------------------------------------------------------------
    /**
     *
     */
    public void findTransitDifference() {
        
        double[] outputArray1 = m_outputSeries4.getArray();
        createAnimation();
        double[] outputArray2 = m_outputSeries4.getArray();
        
        if (outputArray1.length != outputArray2.length) {
            m_errorString = "First transit output was not same size.";
            return;
        }
        
        for (int i = 3; i < outputArray2.length; i += 4) {
            if (outputArray2[i] <= 0 || outputArray1[i] <= 0)
                outputArray2[i] = 0;
            else
                outputArray2[i] = outputArray2[i] - outputArray1[i];
        }
        
        m_outputSeries4.arrayChanged();
        
        m_errorString = null;
    }
    
    //------------------------------------------------------------------------------
    /**
     * Creates a results summary.
     */
    public void createResultsSummary() {
        
        
        // Set the number of dimensions in the input file and open it.
        int numInDims = 10;
        RafDataSeries inRafSeries =
            new gov.usgs.gr.data.RafDataSeries(numInDims);
        inRafSeries.setProperty("file", m_file.getAbsolutePath());
        inRafSeries.setProperty("diskType", "DOUBLE");
        inRafSeries.setProperty("memType", "DOUBLE");
        inRafSeries.setProperty("bigEndian", "true");
        inRafSeries.createManagedCache(2, 1000);
        
        // Define local variables.
        int count = inRafSeries.getCount();
        // Create an array for storing the number of particles across each line.
        int[] lineTallies = new int[0];
        
        for (int i = 0; i < count; i++) {
            // Read the time, release index, and which finish line was crossed.
            int offset = i * numInDims;
            
            int lineCrossed =
                (int)Math.round(inRafSeries.getDoubleValueAt(offset + 2)) + 1;
            if (lineCrossed < 0)
                lineCrossed = 0;
            
            // Increase the size of the lineTallies array if it is too small.
            if (lineCrossed >= lineTallies.length) {
                int[] tempArray = new int[lineCrossed + 1];
                // Copy the existing values into the new array.
                for (int j = 0; j < lineTallies.length; j++) {
                    tempArray[j] = lineTallies[j];
                }
                for (int j = lineTallies.length; j < tempArray.length; j++) {
                    tempArray[j] = 0;
                }
                lineTallies = tempArray;
            }
            
            // Increase the tally for the finish line crossed.
            lineTallies[lineCrossed]++;
        }
        
        // Close the in series.
        inRafSeries.closeRAF();
        
        // Write out the tally summary
        System.out.println(" -- Results Summary --");
        
        for (int i = 1; i < lineTallies.length; i++) {
            double percentage = 100 *
                (double)lineTallies[i] / (double)count;
            if (lineTallies[i] > 0) {
                System.out.println("Finish line " + (i - 1) + ": "
                    + lineTallies[i] + "  ( " + percentage + "% )");
            }
        }
        if (lineTallies.length > 0) {
            if (lineTallies[0] > 0) {
                double percentage =
                    100 * (double)lineTallies[0] / (double)count;
                System.out.println("Did not finish: " + lineTallies[0]
                    + "  ( " + percentage + "% )");
            } else {
                System.out.println("All particles finished.");
            }
            System.out.println(" -- " + count + " particles released --");
        } else {
            System.out.println(" -- No particles found --");
        }
    }
    
    //------------------------------------------------------------------------------
    /**
     * Create a DoubleDataSeries with Time-X-Y-Red-Green-Blue values showing the
     * average values at each time and X-Z location across a cross section.
     */
    public void createCrossSectionAnimation(double x1, double y1,
        double x2, double y2, double xSpacing, double zSpacing) {
        
        // Set the number of dimensions in the input file and open it.
        int numInDims = 10;
        RafDataSeries inRafSeries =
            new gov.usgs.gr.data.RafDataSeries(numInDims);
        inRafSeries.setProperty("file", m_file.getAbsolutePath());
        inRafSeries.setProperty("diskType", "DOUBLE");
        inRafSeries.setProperty("memType", "DOUBLE");
        inRafSeries.setProperty("bigEndian", "true");
        inRafSeries.createManagedCache(2, 1000);
        
        // Define local variables.
        int count = inRafSeries.getCount();
        double releaseTime = 0;
        m_numTimes = 0;
        double lastTime = Double.NEGATIVE_INFINITY;
        // Define five seconds as a fraction of a day.
        double sec5 = 0.00005;
        
        // Temporary until code is added...
        boolean bTransit = false;
        
        // Don't count on evenly spaced times.  Read the entire file and count
        // up the time steps.
        for (int i = 0; i < count; i++) {
            int offset = i * numInDims;
            releaseTime = inRafSeries.getDoubleValueAt(offset);
            if (releaseTime > (lastTime + sec5)) {
                m_numTimes++;
                lastTime = releaseTime;
            }
        }
        
        double xDiff = x2 - x1;
        double yDiff = y2 - y1;
        double length = Math.sqrt((xDiff * xDiff) + (yDiff * yDiff));
        double cosTheta = xDiff / length;
        double sinTheta = yDiff / length;
        int numI = (int)length + 1;
        int numK = m_numK;
        int numXzPoints = numI * numK;
        
        // Create an out array to hold the time steps.
        int numOutDims = 6;
        double[] outArray =
            new double[m_numTimes * numXzPoints * numOutDims];
        int[] countArray = new int[numXzPoints];
        
        int minI = (int)(x1 / m_xSpacing);
        int minJ = (int)(y1 / m_ySpacing);
        int maxI = (int)(x2 / m_xSpacing) + 1;
        int maxJ = (int)(y2 / m_ySpacing) + 1;
        
        // Initialize the X, Y, R, G, and B dimensions of the array.
        for (int j = 0; j < m_numTimes; j++) {
            for (int i = 0; i < numI; i++) {
                for (int k = 0; k < numK; k++) {
                    int ijIndex = i + (k * numI);
                    int index = (j * numXzPoints + ijIndex) * numOutDims;
                    outArray[index + 1] = i * xSpacing;
                    outArray[index + 2] = k * -zSpacing;
                    // Use -1 RGB values to indicate no releases and 0 to indicate
                    // an unknown result.
                    outArray[index + 3] = -1;
                    outArray[index + 4] = -1;
                    outArray[index + 5] = -1;
                    countArray[i] = 0;
                }
            }
        }
        
        // Run through the list of particle releases.
        lastTime = Integer.MIN_VALUE;
        int counter = 0;
        for (int i = 0; i < count; i++) {
            // Read the time, release index, and which finish line was crossed.
            int offset = i * numInDims;
            releaseTime = inRafSeries.getDoubleValueAt(offset);
            int releaseIndex3D = (int)Math.round(
                inRafSeries.getDoubleValueAt(offset + 1));
            int currentI = m_3DIndexArray[releaseIndex3D * 3];
            int currentJ = m_3DIndexArray[releaseIndex3D * 3 + 1];
            
            if (currentI < minI || currentI > maxI || currentJ < minJ
                || currentJ > maxJ) {
                
                lastTime = releaseTime;
                continue;
            }
            
            double releaseX = inRafSeries.getDoubleValueAt(offset + 4);
            double releaseY = inRafSeries.getDoubleValueAt(offset + 5);
            
            double shiftedX = releaseX - x1;
            double shiftedY = releaseY - y1;
            double rotatedX = shiftedX * cosTheta - shiftedY * sinTheta;
            double rotatedY = shiftedY * cosTheta - shiftedX * sinTheta;
            int releaseI = (int)(rotatedX / xSpacing);
            int releaseJ = (int)((rotatedY / xSpacing) + 0.5);
            
            if (releaseI < 0 || releaseI >= numI || releaseJ < 0
                || releaseJ >= 1) {
                
                lastTime = releaseTime;
                continue;
            }
            
            double releaseZ = inRafSeries.getDoubleValueAt(offset + 6);
            int releaseK = (int)(releaseZ / zSpacing);
            if (releaseK < 0)
                releaseK = 0;
            int ijIndex = releaseI + (releaseK * numI);
            
            int lineCrossed =
                (int)Math.round(inRafSeries.getDoubleValueAt(offset + 2));
            
            // If this is the first index of a new time step.
            if (releaseTime != lastTime) {
                // If not the first time step...
                if (i != 0) {
                    for (int j = 0; j < numXzPoints; j++) {
                        // Write the time value to the out array for all
                        // the cells of the previous time step.
                        int index = (counter + j) * numOutDims;
                        outArray[index] = lastTime;
                        
                        // Normalize the RGB for each cell to the number
                        // released at that cell.
                        if (countArray[j] != 0) {
                            double numReleased = (double) countArray[j];
                            outArray[index + 3] /= numReleased;
                            outArray[index + 4] /= numReleased;
                            outArray[index + 5] /= numReleased;
                            countArray[j] = 0;
                        }
                    }
                    
                    // Move the output counter to the next time step.
                    counter += numXzPoints;
                }
                
                // Keep track of this time so it can be used once the next
                // time appears.
                lastTime = releaseTime;
            }
            
            // If it was previously -1 for no releases, change it to 0.
            if (outArray[(counter + ijIndex) * numOutDims + 3] < 0) {
                outArray[(counter + ijIndex) * numOutDims + 3] = 0;
                outArray[(counter + ijIndex) * numOutDims + 4] = 0;
                outArray[(counter + ijIndex) * numOutDims + 5] = 0;
            }
            
            // Increase the tally for the appropriate cell if the finish line
            // crossed matches one of the column assignments.
            if (lineCrossed >= 0 && lineCrossed < m_columnAssignments.length) {
                if (m_columnAssignments[lineCrossed] == 0)
                    outArray[ (counter + ijIndex) * numOutDims + 3]++;
                if (m_columnAssignments[lineCrossed] == 1)
                    outArray[ (counter + ijIndex) * numOutDims + 4]++;
                if (m_columnAssignments[lineCrossed] == 2)
                    outArray[ (counter + ijIndex) * numOutDims + 5]++;
            }
            
            // Increase the release counter for that cell.
            countArray[ijIndex]++;
        }
        
        // Close the in series.
        inRafSeries.closeRAF();
        
        // Finish processing the last time step.
        for (int j = 0; j < numXzPoints; j++) {
            int index = (counter + j) * numOutDims;
            outArray[index] = releaseTime;
            if (countArray[j] != 0) {
                double numReleased = (double) countArray[j];
                outArray[index + 3] /= numReleased;
                outArray[index + 4] /= numReleased;
                outArray[index + 5] /= numReleased;
                countArray[j] = 0;
            }
        }
        
        // Update the out series.
        if (bTransit)
            m_outputSeries4.setArray(outArray);
        else
            m_outputSeries6.setArray(outArray);
    }
    
    //------------------------------------------------------------------------------
    /**
     */
    public double[] createProbabilityMap() {
        
        GrObjectContainer groc =
            GrObjectContainer.getChildContainer(this, "Output");
        DoubleDataSeries series = (DoubleDataSeries)groc.findChildInstancesOf(
            "gov.usgs.gr.data.DoubleDataSeries", false);
        
        if (series == null)
            return new double[0];
        
        DoubleDataSeries animationSeries;
        if (m_minTransitPortion >= 0)
            animationSeries = m_outputSeries4;
        else
            animationSeries = m_outputSeries6;
        int numWetPoints2D = m_2DIndexArray.length / 2;
        int numTimes = animationSeries.getCount() / numWetPoints2D;
        double[] probArray = new double[animationSeries.getCount() * 4];
        double[] randomProbabilityInputArray =
            new double[animationSeries.getCount()];
        DoubleDataSeries randomProbabilityInputSeries =
            new DoubleDataSeries(randomProbabilityInputArray, 1);
        double[] tsWeightingInputArray = new double[numTimes];
        DoubleDataSeries tsWeightingInputSeries =
            new DoubleDataSeries(tsWeightingInputArray, 1);
        double[] animationArray = animationSeries.getArray();
        int numAnimDims = animationSeries.getNumDimensions();
        double dim3Max = animationSeries.getDimension(3).getMax();
        
        double[] depthArray = m_bath.getAverageDepthArray();
        
        // Check for an input thin-wall barrier series.
        IntDataSeries barrierSeries = (IntDataSeries)firstChildInstanceOf(
            "gov.usgs.gr.data.IntDataSeries", false);
        
        // Get lookup and reverse lookup arrays from first PTM Bath
        // Assume all points are there for each time step
        for (int timeStep = 0; timeStep < numTimes; timeStep++) {
            int animIndex = timeStep * numWetPoints2D * numAnimDims;
            double time = animationArray[animIndex];
            double tsProbTotal = 0;
            
            for (int index = 0; index < numWetPoints2D; index++) {
                // for each wet point, calculate prob and write to probArray
                int currentI = m_2DIndexArray[index * 2];
                int currentJ = m_2DIndexArray[index * 2 + 1];
                
                int t1Index = (timeStep * numWetPoints2D + index)
                    * numAnimDims;
                double t1Red = animationArray[t1Index + 3];
                double t1Green = 0;
                double t1Blue = 0;
                double t1Bright = 0;
                if (numAnimDims == 6) {
                    t1Green = animationArray[t1Index + 4];
                    t1Blue = animationArray[t1Index + 5];
                    t1Bright = t1Red + t1Green + t1Blue;
                }
                
                double defaultDiff = 0.01;
                if (t1Red < 0)
                    defaultDiff = 8;
                
                // Set the default diff value.
                double nDiff = defaultDiff;
                // Check that the north neighbor is in bounds.
                if ((currentJ + 1) < m_numJ && t1Red >= 0) {
                    // Check that the north neighbor is wet and reachable.
                    int ijIndex = currentI + ( (currentJ + 1) * m_numI);
                    int nIndex2D = m_2DLookupArray[ijIndex];
                    int combo1IJ = (currentI - 2) * 10000 + currentJ - 1;
                    int combo2IJ = (currentI - 1) * 10000 + currentJ - 1;
                    boolean bThinWall = existsThinWallBarrier(barrierSeries,
                        combo1IJ, combo2IJ);
                    if (nIndex2D >= 0 && !bThinWall) {
                        // Look up the north color values.
                        int nIndex = (timeStep * numWetPoints2D + nIndex2D)
                            * numAnimDims;
                        
                        // Only calculate if the neighbor has had releases.
                        double nRed = animationArray[nIndex + 3];
                        double nGreen = 0;
                        double nBlue = 0;
                        double nBright = 0;
                        if (nRed >= 0) {
                            if (numAnimDims == 6) {
                                nGreen = animationArray[nIndex + 4];
                                nBlue = animationArray[nIndex + 5];
                                nBright = nRed + nGreen + nBlue;
                            }
                            // Calculate the difference.
                            // Use the default minimum if near black.
                            double redDiff = Math.abs(t1Red - nRed);
                            double greenDiff = Math.abs(t1Green - nGreen);
                            double blueDiff = Math.abs(t1Blue - nBlue);
                            double brightDiff = Math.abs(t1Bright - nBright);
                            if (t1Bright > 0.01 ||
                                (redDiff + greenDiff + blueDiff) > 0.01) {
                                nDiff = redDiff + greenDiff + blueDiff +
                                    brightDiff;
                                // Make transit time adjustment, if needed.
                                if (numAnimDims == 4) {
                                    double maxRed = Math.max(t1Red, nRed);
                                    if (maxRed != 0)
                                        nDiff /= maxRed;
                                }
                            }
                            if (nDiff < 0 || nDiff > 12)
                                System.out.println("diff n out of range");
                        }
                    }
                }
                
                // Set the default diff value.
                double sDiff = defaultDiff;
                // Check that the south neighbor is in bounds.
                if ( (currentJ - 1) >= 0 && t1Red >= 0) {
                    // Check that the south neighbor is wet and reachable.
                    int ijIndex = currentI + ( (currentJ - 1) * m_numI);
                    int sIndex2D = m_2DLookupArray[ijIndex];
                    int combo1IJ = (currentI - 2) * 10000 + currentJ - 2;
                    int combo2IJ = (currentI - 1) * 10000 + currentJ - 2;
                    boolean bThinWall = existsThinWallBarrier(barrierSeries,
                        combo1IJ, combo2IJ);
                    if (sIndex2D >= 0 && !bThinWall) {
                        // Look up the south color values.
                        int sIndex = (timeStep * numWetPoints2D + sIndex2D)
                            * numAnimDims;
                        
                        // Only calculate if the neighbor has had releases.
                        double sRed = animationArray[sIndex + 3];
                        double sGreen = 0;
                        double sBlue = 0;
                        double sBright = 0;
                        if (sRed >= 0) {
                            if (numAnimDims == 6) {
                                sGreen = animationArray[sIndex + 4];
                                sBlue = animationArray[sIndex + 5];
                                sBright = sRed + sGreen + sBlue;
                            }
                            
                            // Calculate the difference.
                            // Use the default minimum if near black.
                            double redDiff = Math.abs(t1Red - sRed);
                            double greenDiff = Math.abs(t1Green - sGreen);
                            double blueDiff = Math.abs(t1Blue - sBlue);
                            double brightDiff = Math.abs(t1Bright - sBright);
                            if (t1Bright > 0.01 ||
                                (redDiff + greenDiff + blueDiff) > 0.01) {
                                sDiff = redDiff + greenDiff + blueDiff +
                                    brightDiff;
                                // Make transit time adjustment, if needed.
                                if (numAnimDims == 4) {
                                    double maxRed = Math.max(t1Red, sRed);
                                    if (maxRed != 0)
                                        sDiff /= maxRed;
                                }
                            }
                            if (sDiff < 0 || sDiff > 12)
                                System.out.println("diff s out of range");
                        }
                    }
                }
                
                // Set the default diff value.
                double eDiff = defaultDiff;
                // Check that the east neighbor is in bounds.
                if ( (currentI + 1) < m_numI && t1Red >= 0) {
                    // Check that the east neighbor is wet and reachable.
                    int ijIndex = (currentI + 1) + (currentJ * m_numI);
                    int eIndex2D = m_2DLookupArray[ijIndex];
                    int combo1IJ = (currentI - 1) * 10000 + currentJ - 2;
                    int combo2IJ = (currentI - 1) * 10000 + currentJ - 1;
                    boolean bThinWall = existsThinWallBarrier(barrierSeries,
                        combo1IJ, combo2IJ);
                    if (eIndex2D >= 0 && !bThinWall) {
                        // Look up the east color values.
                        int eIndex = (timeStep * numWetPoints2D + eIndex2D)
                            * numAnimDims;
                        
                        // Only calculate if the neighbor has had releases.
                        double eRed = animationArray[eIndex + 3];
                        double eGreen = 0;
                        double eBlue = 0;
                        double eBright = 0;
                        if (eRed >= 0) {
                            if (numAnimDims == 6) {
                                eGreen = animationArray[eIndex + 4];
                                eBlue = animationArray[eIndex + 5];
                                eBright = eRed + eGreen + eBlue;
                            }
                            
                            // Calculate the difference.
                            // Use the default minimum if near black.
                            double redDiff = Math.abs(t1Red - eRed);
                            double greenDiff = Math.abs(t1Green - eGreen);
                            double blueDiff = Math.abs(t1Blue - eBlue);
                            double brightDiff = Math.abs(t1Bright - eBright);
                            if (t1Bright > 0.01 ||
                                (redDiff + greenDiff + blueDiff) > 0.01) {
                                eDiff = redDiff + greenDiff + blueDiff +
                                    brightDiff;
                                // Make transit time adjustment, if needed.
                                if (numAnimDims == 4) {
                                    double maxRed = Math.max(t1Red, eRed);
                                    if (maxRed != 0)
                                        eDiff /= maxRed;
                                }
                            }
                            if (eDiff < 0 || eDiff > 12)
                                System.out.println("diff e out of range");
                        }
                    }
                }
                
                // Set the default diff value.
                double wDiff = defaultDiff;
                // Check that the west neighbor is in bounds.
                if ( (currentI - 1) >= 0 && t1Red >= 0) {
                    // Check that the west neighbor is wet and reachable.
                    int ijIndex = (currentI - 1) + (currentJ * m_numI);
                    int wIndex2D = m_2DLookupArray[ijIndex];
                    int combo1IJ = (currentI - 2) * 10000 + currentJ - 2;
                    int combo2IJ = (currentI - 2) * 10000 + currentJ - 1;
                    boolean bThinWall = existsThinWallBarrier(barrierSeries,
                        combo1IJ, combo2IJ);
                    if (wIndex2D >= 0 && !bThinWall) {
                        // Look up the west color values.
                        int wIndex = (timeStep * numWetPoints2D + wIndex2D)
                            * numAnimDims;
                        
                        // Only calculate if the neighbor has had releases.
                        double wRed = animationArray[wIndex + 3];
                        double wGreen = 0;
                        double wBlue = 0;
                        double wBright = 0;
                        if (wRed >= 0) {
                            if (numAnimDims == 6) {
                                wGreen = animationArray[wIndex + 4];
                                wBlue = animationArray[wIndex + 5];
                                wBright = wRed + wGreen + wBlue;
                            }
                            
                            // Calculate the difference.
                            // Use the default minimum if near black.
                            double redDiff = Math.abs(t1Red - wRed);
                            double greenDiff = Math.abs(t1Green - wGreen);
                            double blueDiff = Math.abs(t1Blue - wBlue);
                            double brightDiff = Math.abs(t1Bright - wBright);
                            if (t1Bright > 0.01 ||
                                (redDiff + greenDiff + blueDiff) > 0.01) {
                                wDiff = redDiff + greenDiff + blueDiff +
                                    brightDiff;
                                // Make transit time adjustment, if needed.
                                if (numAnimDims == 4) {
                                    double maxRed = Math.max(t1Red, wRed);
                                    if (maxRed != 0)
                                        wDiff /= maxRed;
                                }
                            }
                            if (wDiff < 0 || wDiff > 12)
                                System.out.println("diff w out of range");
                        }
                    }
                }
                
                // Set the default diff value.
                double t2Diff = defaultDiff;
                // Check that the next neighbor is defined.
                if ( (timeStep + 1) < numTimes && t1Red >= 0) {
                    // Look up the previous color values.
                    int t2Index = ( (timeStep + 1) * numWetPoints2D + index)
                        * numAnimDims;
                    
                    // Only calculate if the neighbor has had releases.
                    double t2Red = animationArray[t2Index + 3];
                    double t2Green = 0;
                    double t2Blue = 0;
                    double t2Bright = 0;
                    if (t2Red >= 0) {
                        if (numAnimDims == 6) {
                            t2Green = animationArray[t2Index + 4];
                            t2Blue = animationArray[t2Index + 5];
                            t2Bright = t2Red + t2Green + t2Blue;
                        }
                        
                        // Calculate the difference.
                        // Use the default minimum if near black.
                        double redDiff = Math.abs(t1Red - t2Red);
                        double greenDiff = Math.abs(t1Green - t2Green);
                        double blueDiff = Math.abs(t1Blue - t2Blue);
                        double brightDiff = Math.abs(t1Bright - t2Bright);
                        if (t1Bright > 0.01 ||
                            (redDiff + greenDiff + blueDiff) > 0.01) {
                            
                            t2Diff = 0.9 * (redDiff + greenDiff + blueDiff +
                                brightDiff);
                            // Make transit time adjustment, if needed.
                            if (numAnimDims == 4) {
                                double maxRed = Math.max(t1Red, t2Red);
                                if (maxRed != 0)
                                    t2Diff /= maxRed;
                            }
                        }
                        if (t2Diff < 0 || t2Diff > 12)
                            System.out.println("diff t2 out of range");
                    }
                }
                
                // Set the default diff value.
                double t0Diff = defaultDiff;
                // Check that the previous neighbor is defined.
                if ( (timeStep - 1) >= 0 && t1Red >= 0) {
                    // Look up the previous color values.
                    int t0Index = ( (timeStep - 1) * numWetPoints2D + index)
                        * numAnimDims;
                    
                    // Only calculate if the neighbor has had releases.
                    double t0Red = animationArray[t0Index + 3];
                    double t0Green = 0;
                    double t0Blue = 0;
                    double t0Bright = 0;
                    if (t0Red >= 0) {
                        if (numAnimDims == 6) {
                            t0Green = animationArray[t0Index + 4];
                            t0Blue = animationArray[t0Index + 5];
                            t0Bright = t0Red + t0Green + t0Blue;
                        }
                        
                        // Calculate the difference.
                        // Use the default minimum if near black.
                        double redDiff = Math.abs(t1Red - t0Red);
                        double greenDiff = Math.abs(t1Green - t0Green);
                        double blueDiff = Math.abs(t1Blue - t0Blue);
                        double brightDiff = Math.abs(t1Bright - t0Bright);
                        if (t1Bright > 0.01 ||
                            (redDiff + greenDiff + blueDiff) > 0.01) {
                            
                            t0Diff = 0.9 * (redDiff + greenDiff + blueDiff +
                                brightDiff);
                            // Make transit time adjustment, if needed.
                            if (numAnimDims == 4) {
                                double maxRed = Math.max(t1Red, t0Red);
                                if (maxRed != 0)
                                    t0Diff /= maxRed;
                            }
                        }
                        if (t0Diff < 0 || t0Diff > 12)
                            System.out.println("diff t0 out of range");
                    }
                }
                
                double depth = depthArray[index];
                double waterSurfaceElevation =
                    getVar2D((timeStep * numWetPoints2D + index)) * m_wseScale;
                double prob3D = nDiff + sDiff + eDiff + wDiff + t2Diff + t0Diff;
                double prob2D = prob3D / (depth + waterSurfaceElevation);
                int outIndex = (timeStep * numWetPoints2D) + index;
                randomProbabilityInputArray[outIndex] = prob2D;
                
                outIndex *= 4;
                probArray[outIndex] = time;
                probArray[outIndex + 1] = currentI * m_xSpacing;
                probArray[outIndex + 2] = currentJ * m_ySpacing;
                probArray[outIndex + 3] = prob3D;
                
                tsProbTotal += prob2D;
            }
            
            tsWeightingInputArray[timeStep] = tsProbTotal / numWetPoints2D;
        }
        
        tsWeightingInputSeries.arrayChanged();
        double meanWeighting = gov.usgs.gr.tools.MeanCalculator.
            calculateMean(tsWeightingInputSeries);
        tsWeightingInputSeries.selectAll();
        tsWeightingInputSeries.modifySelectedPoints(4, 0, meanWeighting);
        tsWeightingInputSeries.deselectAll();
        
        randomProbabilityInputSeries.arrayChanged();
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject) iter.next();
            if (! (gro instanceof ParticleTracker))
                continue;
            
            ParticleTracker pt = (ParticleTracker) gro;
            
            GrObjectContainer probInputGroc =
                GrObjectContainer.getChildContainer(pt,
                "Random Probability Input Series");
            probInputGroc.clear();
            probInputGroc.add(randomProbabilityInputSeries);
            
            GrObjectContainer tsWeightingInputGroc =
                GrObjectContainer.getChildContainer(pt,
                "Time Step Weighting Input Series");
            tsWeightingInputGroc.clear();
            tsWeightingInputGroc.add(tsWeightingInputSeries);
        }
        
        return probArray;
    }
    
    //------------------------------------------------------------------------------
    /**
     * Returns the value at the specified index from the 2D vars series.
     * Handles the case of the series being stored in a random access file.
     * @param offset    The offset into the array for the value needed.
     *                  Assumes this is within the array bounds.
     * @return  The value at the specified offset.
     *          0 is returned if a value could not be found.
     */
    protected float getVar2D(int offset) {
        
        // Set the default return value to 0.
        float var = 0;
        
        // Retrieve the value, either from a random access file, or directly
        // from the array.
        var = ((RafDataSeries)m_2DVarsSeries).getValueAt(offset);
        
        return var;
    }
    
    //------------------------------------------------------------------------------
    /**
     *
     */
    public boolean existsThinWallBarrier(IntDataSeries barrierSeries,
        int combo1IJ, int combo2IJ) {
        
        if (barrierSeries == null)
            return false;
        
        int[] oldArray = barrierSeries.getArray();
        
        // Check if the barrier is already in the series.
        int index = barrierSeries.search(combo1IJ);
        if (index >= 0 && oldArray[index * 2] == combo1IJ) {
            if (oldArray[index * 2 + 1] == combo2IJ) {
                return true;
            }
        } else {
            // Check the index below
            index--;
            if (index >= 0 && oldArray[index * 2] == combo1IJ) {
                if (oldArray[index * 2 + 1] == combo2IJ) {
                    return true;
                }
            } else {
                // Check the index above
                index += 2;
                if ((index * 2) < oldArray.length &&
                    oldArray[index * 2] == combo1IJ) {
                    
                    if (oldArray[index * 2 + 1] == combo2IJ) {
                        return true;
                    }
                }
            }
        }
        
        return false;
    }
    
    //------------------------------------------------------------------------------
    /**
     *
     */
    public void mergeFiles(File[] inFiles, File outFile) {
        
        double sec5 = 0.00005f;
        int numInDims = 10;
        int outputOffset = 0;
        int[] fileOffsets = new int[inFiles.length];
        double[][] inVals = new double[inFiles.length][numInDims];
        int[] counts = new int[inFiles.length];
        RafDataSeries[] inSeriesList = new RafDataSeries[inFiles.length];
        
        for (int i = 0; i < inFiles.length; i++) {
            inSeriesList[i] = new gov.usgs.gr.data.RafDataSeries(numInDims);
            inSeriesList[i].setProperty("file", inFiles[i].getAbsolutePath());
            inSeriesList[i].setProperty("diskType", "DOUBLE");
            inSeriesList[i].setProperty("memType", "DOUBLE");
            inSeriesList[i].setProperty("bigEndian", "true");
            inSeriesList[i].createManagedCache(2, 1000);
            counts[i] = inSeriesList[i].getCount() * numInDims;
            for (int j = 0; j < numInDims; j++) {
                inVals[i][j] = inSeriesList[i].getDoubleValueAt(j);
            }
            fileOffsets[i] = numInDims;
        }
        
        boolean bSuccess = outFile.delete();
        RafDataSeries outRafSeries =
            new gov.usgs.gr.data.RafDataSeries(numInDims);
        outRafSeries.setProperty("file", outFile.getAbsolutePath());
        outRafSeries.setProperty("diskType", "DOUBLE");
        outRafSeries.setProperty("memType", "DOUBLE");
        outRafSeries.setProperty("bigEndian", "true");
        outRafSeries.openRAF("rwd");
        
        long fileOffset = 0;
        double[] outArray = new double[2000 * numInDims];
        
        // Merge files...
        int numInputFiles = inFiles.length;
        while (numInputFiles > 0) {
            // Choose the next track to write into the output file.
            int lowestFileIndex = 0;
            double lowestReleaseIndex = 0;
            double lowestFinishIndex = 0;
            double lowestFinishTime = 0;
            double lowestTime = Double.MAX_VALUE;
            for (int i = 0; i < numInputFiles; i++) {
                // Check if new time is lower.
                if (lowestTime > (inVals[i][0] + sec5)) {
                    // New time is lower, so update lowest variables.
                    lowestTime = inVals[i][0];
                    lowestReleaseIndex = inVals[i][1];
                    lowestFinishIndex = inVals[i][2];
                    lowestFinishTime = inVals[i][3];
                    lowestFileIndex = i;
                } else {
                    // New time is not lower.  Check if it is equal.
                    if (lowestTime > (inVals[i][0] - sec5)) {
                        // New time is equal.  Check if release is lower.
                        if (Math.round(lowestReleaseIndex) >
                            Math.round(inVals[i][1])) {
                            
                            // Release is lower, so update lowest variables.
                            lowestTime = inVals[i][0];
                            lowestReleaseIndex = inVals[i][1];
                            lowestFinishIndex = inVals[i][2];
                            lowestFinishTime = inVals[i][3];
                            lowestFileIndex = i;
                        } else {
                            // Release is not lower.  Check if release is equal.
                            if (Math.round(lowestReleaseIndex) ==
                                Math.round(inVals[i][1])) {
                                
                                // Release is equal.
                                // Check if finish line is lower.
                                if (Math.round(lowestFinishIndex) >
                                    Math.round(inVals[i][2])) {
                                    
                                    // Finish line is lower,
                                    // so update lowest variables.
                                    lowestTime = inVals[i][0];
                                    lowestReleaseIndex = inVals[i][1];
                                    lowestFinishIndex = inVals[i][2];
                                    lowestFinishTime = inVals[i][3];
                                    lowestFileIndex = i;
                                } else {
                                    // Finish line is not lower.
                                    // Check if finish line is equal.
                                    if (Math.round(lowestFinishIndex) ==
                                        Math.round(inVals[i][2])) {
                                        
                                        // Finish line is equal.
                                        // Check if finish time is lower.
                                        if (lowestFinishTime > inVals[i][3]) {
                                            
                                            // Finish time is lower,
                                            // so update lowest variables.
                                            lowestTime = inVals[i][0];
                                            lowestReleaseIndex = inVals[i][1];
                                            lowestFinishIndex = inVals[i][2];
                                            lowestFinishTime = inVals[i][3];
                                            lowestFileIndex = i;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            
            // Write the next track to the output file.
            for (int j = 0; j < numInDims; j++) {
                ///// Hard wire fix for merging different times onto one frame...
                //                if (j == 0)
                //                    inVals[lowestFileIndex][j] = 0;
                outArray[outputOffset + j] = inVals[lowestFileIndex][j];
            }
            
            outputOffset += numInDims;
            
            if (outputOffset >= outArray.length) {
                outRafSeries.writeValuesFrom(
                    outArray, 0, fileOffset, outArray.length);
                fileOffset += outputOffset;
                outputOffset = 0;
            }
            
            if (counts[lowestFileIndex] > fileOffsets[lowestFileIndex]) {
                for (int j = 0; j < numInDims; j++) {
                    inVals[lowestFileIndex][j] = inSeriesList[lowestFileIndex].
                        getDoubleValueAt(fileOffsets[lowestFileIndex] + j);
                }
                fileOffsets[lowestFileIndex] += numInDims;
            } else {
                // The file has been read to the end so remove its entries
                // and move the remaining entries one index forward in their
                // respective arrays.
                inSeriesList[lowestFileIndex].closeRAF();
                for (int i = lowestFileIndex; i < (numInputFiles - 1); i++) {
                    inFiles[i] = inFiles[i + 1];
                    fileOffsets[i] = fileOffsets[i + 1];
                    for (int j = 0; j < numInDims; j++) {
                        inVals[i][j] = inVals[i + 1][j];
                    }
                    counts[i] = counts[i + 1];
                    inSeriesList[i] = inSeriesList[i + 1];
                }
                numInputFiles--;
            }
        }
        
        outRafSeries.writeValuesFrom(outArray, 0, fileOffset, outputOffset);
        
        outRafSeries.closeRAF();
    }
    
    //------------------------------------------------------------------------------
    /**
     * Convert a file in the old 7-dimension float format to the double format with
     * 10 dimensions and finish X, Y, and Z at the end.
     */
    public void convertFile7to10(File inFile, File outFile) {
        
        int numInDims = 7;
        int numOutDims = 10;
        double[] vals = new double[numInDims];
        vals[7] = -1;
        vals[8] = -1;
        vals[9] = 0;
        
        RafDataSeries inSeries = new gov.usgs.gr.data.RafDataSeries(numInDims);
        inSeries.setProperty("file", inFile.getAbsolutePath());
        inSeries.setProperty("diskType", "FLOAT");
        inSeries.setProperty("memType", "FLOAT");
        inSeries.setProperty("bigEndian", "true");
        inSeries.createManagedCache(2, 1000);
        
        boolean bSuccess = outFile.delete();
        RafDataSeries outSeries =
            new gov.usgs.gr.data.RafDataSeries(numInDims);
        outSeries.setProperty("file", outFile.getAbsolutePath());
        outSeries.setProperty("diskType", "DOUBLE");
        outSeries.setProperty("memType", "DOUBLE");
        outSeries.setProperty("bigEndian", "true");
        outSeries.openRAF("rwd");
        
        int numLines = inSeries.getCount();
        long offset = 0;
        for (int i = 0; i < numLines; i++) {
            for (int j = 0; j < numInDims; j++) {
                vals[j] = inSeries.getValueAt(j);
            }
            for (int j = 0; j < numOutDims; j++) {
                outSeries.writeValueAt(vals[j], offset);
                offset++;
            }
        }
        
        inSeries.closeRAF();
        outSeries.closeRAF();
    }
    
    //------------------------------------------------------------------------------
    /**
     * Initialize member variables by reading values from the Bathymetry and
     * Si3dOutput child objects.
     *
     * @return int, 0 if successful, -1 if child object was missing.
     */
    public synchronized int readVariables() {
        
        boolean bChanged = false;
        int status = 0;
        
        // Look for the child objects.
        Bathymetry bath = (Bathymetry)firstChildInstanceOf(
            "gov.usgs.sfhydro.si3d.Bathymetry", true);
        
        if (bath == null)
            return -1;
        
        m_bath = bath;
        
        Si3dOutput si3dOut = (Si3dOutput)firstChildInstanceOf(
            "gov.usgs.sfhydro.si3d.Si3dOutput", true);
        
        if (si3dOut == null)
            return -1;
        
        m_si3dOut = si3dOut;
        
        // Read variables...
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        m_xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        m_ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        m_zSpacing = (Double)p.value;
        
        //        double startTime = m_si3dOut.getJulianStartTime();
        //        GrHashList si3dOutProps = m_si3dOut.getProperties();
        //        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
        //        Long tsInSec = (Long)p.value;
        //        double timeStepInc = ((double)tsInSec) / 86400;
        
        m_2DIndexArray = m_bath.get2DIndexArray();
        m_2DLookupArray = m_bath.get2DLookupArray(m_2DIndexArray);
        m_3DIndexArray = m_bath.get3DIndexArray();
        m_3DLookupArray = m_bath.get3DLookupArray(m_3DIndexArray);
        DataSeries bathSeries = m_bath.getBathymetrySeries();
        m_3DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("3-D Variables");
        m_2DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("2-D x-y Variables");
        if (m_2DVarsSeries != null) {
            
            // Store the Water Surface Elevation scale.
            DataDimension dim = m_2DVarsSeries.getDimension(0);
            if (dim != null) {
                // Store the scale factor for the dimension.
                GrHashList dimProps = dim.getProperties();
                p = (GrProperty)dimProps.get("unitScale");
                m_wseScale = (Double)p.value;
            }
        }
        
        m_numWetPoints2D = m_2DIndexArray.length / 2;
        
        m_numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        m_numJ = (int)Math.round(bathSeries.getMax(1)) + 1;
        m_numK = (int)Math.round(bathSeries.getMax(2));
        
        return 0;
    }
    
    //------------------------------------------------------------------------------
    /**
     * Provide the list of menu items that execute commands on this object.
     * @return  A submenu of commands.
     */
    public javax.swing.JMenuItem getMenuCommands() {
        
        javax.swing.JMenu subMenu = new javax.swing.JMenu(
            "Particle Track Collator");
        javax.swing.JMenuItem item = new javax.swing.JMenuItem("Calculate");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Create Animation");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Find Transit Difference");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Create Probability Map");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Merge Binary Files");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Cache RAF Data");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Write Results Summary");
        item.addActionListener(this);
        subMenu.add(item);
        
        return subMenu;
    }
    
    //------------------------------------------------------------------------------
    /**
     * Handles action events, such as a recalculate command.
     * @param e The action event.
     */
    public void actionPerformed(java.awt.event.ActionEvent e) {
        
        // Check the source to see if it is a command.
        javax.swing.JMenuItem source =
            (javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        
        // Calculate command.
        if (srcText.equals("Calculate")) {
            // Note the start time.
            GregorianCalendar gc1 = new GregorianCalendar();
            long time1 = gc1.getTimeInMillis();
            
            m_frame = GrController.getParentFrame(source);
            String fileNameBase = prepFiles();
            
            // If the user said OK...
            if (fileNameBase != null) {
                // Recalculate.
                m_bOomeOccurred = false;
                int originalCacheSize = m_cacheSize;
                do {
                    if (m_bOomeOccurred) {
                        System.out.println("Out of memory error occurred "
                            + "while processing.");
                        GregorianCalendar gcOome = new GregorianCalendar();
                        java.text.SimpleDateFormat formatter = new
                            java.text.SimpleDateFormat(
                            "MMM'.' dd',' yyyy 'at' hh:mm:ss");
                        System.out.println(""
                            + formatter.format(gcOome.getTime()));
                        
                        if (m_cacheSize > 0)
                            m_cacheSize *= 0.9;
                        if (m_cacheSize < 0)
                            m_cacheSize++;
                        if (m_cacheSize < -100 || m_cacheSize > 1e12 ||
                            (m_cacheSize > -1 && m_cacheSize < 100)) {
                            
                            break;
                        }
                        System.out.println("Attempting to continue with "
                            + "smaller disk cache (" + m_cacheSize + ")");
                        
                    }
                    
                    m_bOomeOccurred = false;
                    
                    try {
                        // Allocate the disk cache.
                        if (m_cacheSize != 0)
                            cacheRafData();
                        
                        recalculate(fileNameBase);
                    } catch (OutOfMemoryError oome) {
                        m_bOomeOccurred = true;
                    }
                    
                    // Delete the cache to free memory.
                    if (m_cacheSize != 0) {
                        int temp = m_cacheSize;
                        System.out.println("Setting cache size to 0");
                        m_cacheSize = 0;
                        System.gc();
                        cacheRafData();
                        m_cacheSize = temp;
                        if (m_bOomeOccurred) {
                            try {
                                for (int i = 0; i < 3; i++) {
                                    System.out.println("\nGC");
                                    System.gc();
                                    Thread.currentThread().sleep(5000);
                                }
                                System.out.println("\nGC");
                                System.gc();
                            } catch (InterruptedException ex) {
                                ex.printStackTrace();
                            }
                        }
                    }
                } while (m_bOomeOccurred);
                
                m_cacheSize = originalCacheSize;
            }
            
            // Write out the elapsed time if it was greater than 10 seconds.
            GregorianCalendar gc2 = new GregorianCalendar();
            long time2 = gc2.getTimeInMillis();
            if ((time2 - time1) > 10000) {
                java.text.SimpleDateFormat formatter = new
                    java.text.SimpleDateFormat(
                    "MMM'.' dd',' yyyy 'at' hh:mm:ss");
                System.out.println("PTM Start: "
                    + formatter.format(gc1.getTime()));
                System.out.println("PTM End:   "
                    + formatter.format(gc2.getTime()));
            }
        }
        
        if (srcText.equals("Create Animation")) {
            m_frame = GrController.getParentFrame(source);
            File[] files = GrController.showFileDialog("Choose File", true,
                m_file, m_frame);
            if (files != null && files.length > 0) {
                m_file = files[0];
                m_filename = m_file.getName();
                
                int status = readVariables();
                if (status >= 0)
                    createAnimation();
                //createCrossSectionAnimation (5262, 2550, 5430, 2550, 1, 1);
            }
        }
        if (srcText.equals("Find Transit Difference")) {
            m_frame = GrController.getParentFrame(source);
            File[] files = GrController.showFileDialog("Choose File", true,
                m_file, m_frame);
            if (files != null && files.length > 0) {
                m_file = files[0];
                m_filename = m_file.getName();
                
                int status = readVariables();
                if (status >= 0)
                    findTransitDifference();
            }
        }
        if (srcText.equals("Create Probability Map")) {
            int status = readVariables();
            if (status >= 0) {
                double[] probArray = createProbabilityMap();
                m_outputSeriesP.setArray(probArray);
            }
        }
        if (srcText.equals("Merge Binary Files")) {
            m_frame = GrController.getParentFrame(source);
            File[] inFiles = GrController.showFileDialog("Choose Input Files",
                true, m_file, m_frame);
            if (inFiles != null && inFiles.length > 0) {
                File[] outFiles = GrController.showFileDialog(
                    "Choose Output File", true, m_file, m_frame);
                if (outFiles != null && outFiles.length > 0) {
                    mergeFiles(inFiles, outFiles[0]);
                }
            }
        }
        if (srcText.equals("Cache RAF Data")) {
            cacheRafData();
        }
        if (srcText.equals("Write Results Summary")) {
            m_frame = GrController.getParentFrame(source);
            File[] files = GrController.showFileDialog("Choose File", true,
                m_file, m_frame);
            if (files != null && files.length > 0) {
                m_file = files[0];
                m_filename = m_file.getName();
                
                int status = readVariables();
                if (status >= 0)
                    createResultsSummary();
            }
        }
    }
    
    //------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * If an error was detected during an operation, the class's error string
     * is appended.
     * @return  The string that describes this object.
     */
    public String toString() {
        
        String s = "Particle Track Collator";
        
        if (m_errorString != null)
            s += m_errorString;
        
        return s;
    }
    
    private void jbInit() throws Exception {
    }
}

