/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    2002-05-15
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.*;
import java.util.*;
import javax.swing.*;

//==============================================================================
/**
 * Calculates the path that a particle would take through a time-variable
 * velocity field.
 */

public class ParticleTracker extends GrObject implements
	java.awt.event.ActionListener {

    /** The path calculator that creates the paths. */
    protected ParticlePathCalculator m_calc = null;
    /** The bathymetry output from the Si3D model. */
    protected Bathymetry m_bath = null;
    /** The general output from the Si3D model. */
    protected Si3dOutput m_si3dOut = null;
    /** An array of statistics generated from tracking particles. */
    protected double[] m_trackStats = new double[0];

    /**
     * A container that holds a data series with the X-Y locations of the
     * release point of each particle.
     */
    protected GrObjectContainer m_xyInputContainer = null;
    /**
     * A container that holds a data series with the particle number and
     * depth of the release point of each particle.
     */
    protected GrObjectContainer m_nzInputContainer = null;
    /**
     * A container that holds a data series with the release time and
     * particle number of each particle.
     */
    protected GrObjectContainer m_tnInputContainer = null;
    /**
     * A container that holds a data series with the tracking duration of
     * each particle.
     */
    protected GrObjectContainer m_ndInputContainer = null;
    /**
     * A container that holds a data series with the X-Y-Z locations of the
     * release point of each particle.
     */
    protected GrObjectContainer m_xyzInputContainer = null;
    /**
     * A container that holds a data series with the Time-X-Y-Z release
     * point locations and tracking duration of each particle.
     */
    protected GrObjectContainer m_txyzdInputContainer = null;
    /**
     * A container that holds a data series with the X-Y locations of the
     * transit finish lines.
     */
    protected GrObjectContainer m_finishLineInputContainer = null;
    /**
     * A container that holds a data series with the probabilities for
     * releasing at each cell.
     */
    protected GrObjectContainer m_randomProbabilityInputContainer = null;
    /**
     * A container that holds a data series with the weighting factor for
     * releasing at each time step.
     */
    protected GrObjectContainer m_tsWeightingInputContainer = null;

    /** A container that holds output data series containing particle paths. */
    protected GrObjectContainer m_outputContainer = null;
    /** A data series containing X-Y particle path output. */
    protected DoubleDataSeries m_xyOutputSeries = new DoubleDataSeries(2);
    /** A data series containing Time-X particle path output. */
    protected DoubleDataSeries m_txOutputSeries = new TimeSeries(2);
    /** A data series containing Time-Y particle path output. */
    protected DoubleDataSeries m_tyOutputSeries = new TimeSeries(2);
    /** A data series containing Time-Z particle path output. */
    protected DoubleDataSeries m_tzOutputSeries = new TimeSeries(2);
    /** A data series containing Time-X-Y-Z particle path output. */
    protected DoubleDataSeries m_txyzOutputSeries = new DoubleDataSeries(4);
    /** A data series containing Lagrangian particle path output. */
    protected DoubleDataSeries m_lgOutputSeries = new DoubleDataSeries(4);
    /** A data series containing N-Water Surface Elevation. */
    protected DoubleDataSeries m_nWseSeries = new DoubleDataSeries(2);
    /** A data series containing grid depths. */
    protected DoubleDataSeries m_nGridDepthSeries = new DoubleDataSeries(2);
    /** A data series containing Probability Density Function output. */
    protected DoubleDataSeries m_pdfOutputSeries = new DoubleDataSeries(2);
    /** A data series containing Cumulative Density Function output. */
    protected DoubleDataSeries m_cdfOutputSeries = new DoubleDataSeries(2);
    /** A data series containing Cumulative Density Function output. */
    protected DoubleDataSeries m_transitTimesOutputSeries =
        new DoubleDataSeries(1);

    /**
     * The boundaries of the area for generating random particle release
     * points.  The values are xMin, yMin, zMin, xMax, yMax, and zMax.
     */
    protected double[] m_randomBounds = {0, 0, 0, 0, 0, 0};
    /** The start time for generating random particle releases. */
    protected double m_randomStartTime = 0;
    /** The duration for generating random particle releases. */
    protected double m_randomReleaseDuration = 1.04;
    /** The duration for generating random particle releases. */
    protected double m_randomTrackDuration = Double.MAX_VALUE;
    /** The number of random particles to generate per time step. */
    protected int m_randomParticlesPerTS = 1;
    /** The number of time steps to skip between generated random releases. */
    protected int m_randomTimeStep = 1;

    /**
     * A list of the current containers and their titles for
     * comparing to gro.
     */
    private String[] m_containerTitles = {"X-Y Input Series",
        "N-Z Input Series", "T-N Input Series", "N-D Input Series",
        "X-Y-Z Input Series", "T-X-Y-Z-D Input Series",
        "Finish Line Input Series", "Random Probability Input Series",
        "Time Step Weighting Input Series", "Output"};

    /**
     * Whether to automatically recalculate the paths each time an input
     * series changes.
     */
    protected boolean m_bAutoRecalc = false;
    /** The interval for reporting the distribution of transit times. */
    protected double m_distributionInterval = 1;
    /**
     * Holds any error descriptions that should be shown to the user in
     * the properties dialog.
     */
    protected String m_errorString = null;

//------------------------------------------------------------------------------
/**
 * Creates an empty object.
 */
    public ParticleTracker () {

        // Add a default calculator.
        add(new ParticlePathCalculator());

        // Add the containers.
        for (int i = 0; i < m_containerTitles.length; i++) {
           GrObjectContainer groc = new GrObjectContainer();
           groc.setName(m_containerTitles[i]);
           groc.setProperty("title", m_containerTitles[i]);
           add(groc);
        }

        // Add the output series to the output container.
        m_outputContainer.add(m_xyOutputSeries);
        m_outputContainer.add(m_txOutputSeries);
        m_outputContainer.add(m_tyOutputSeries);
        m_outputContainer.add(m_tzOutputSeries);
        m_outputContainer.add(m_txyzOutputSeries);
        m_outputContainer.add(m_lgOutputSeries);
        m_outputContainer.add(m_nWseSeries);
        m_outputContainer.add(m_nGridDepthSeries);
        m_outputContainer.add(m_pdfOutputSeries);
        m_outputContainer.add(m_cdfOutputSeries);
        m_outputContainer.add(m_transitTimesOutputSeries);
    }

//------------------------------------------------------------------------------
/**
 * Adds an object as a child, but only if it is an instance of Bathymetry,
 * Si3dOutput, DoubleDataSeries, GraphicalElement, or GrObjectContainer.
 * Removes the old child and replaces it with the new one.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();
        if (gro instanceof Bathymetry) {
            // Look for the old bathymetry and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof Bathymetry) iter.remove();
            }

            // Add the new formatter at the beginning of the list.
            super.add(gro);
            m_children.remove(gro);
            m_children.add(0, gro);
            m_bath = (Bathymetry)(gro);
            return true;
        }
        if (gro instanceof Si3dOutput) {
            // Look for the old bathymetry and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof Si3dOutput) iter.remove();
            }

            // Add the new formatter at the beginning of the list.
            super.add(gro);
            m_children.remove(gro);
            m_children.add(0, gro);
            m_si3dOut = (Si3dOutput)gro;
            return true;
        }
        if (gro instanceof ParticlePathCalculator) {
            // Look for the old calculator and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof ParticlePathCalculator) iter.remove();
            }

            // Add the new formatter at the beginning of the list.
            super.add(gro);
            m_children.remove(gro);
            m_children.add(0, gro);
            m_calc = (ParticlePathCalculator)(gro);
            return true;
        }

        // Add DS...
        if (gro instanceof DoubleDataSeries) {
        	DoubleDataSeries ds = (DoubleDataSeries)gro;

            m_children.add(ds);
            return true;
        }

        // Add graphical element...
        if (gro instanceof GraphicalElement) {
            m_children.add(gro);
            return true;
        }

        // Add object container...
        if (gro instanceof GrObjectContainer) {
            GrObjectContainer groc = (GrObjectContainer)gro;
            GrHashList props = groc.getProperties();
            GrProperty p = (GrProperty)props.get("title");
            if (p == null)
                return false;
            String title = (String)p.value;

            // Compare the title of groc to each container and replace the
            // current one if there is a match.
            for (int i = 0; i < m_containerTitles.length; i++) {
                if (title.equals(m_containerTitles[i])) {
                    GrObjectContainer oldGroc = null;

                    Iterator iter = m_children.iterator();
                    while (iter.hasNext()) {
                        GrObject child = (GrObject)iter.next();
                        if (child instanceof GrObjectContainer) {
                            GrObjectContainer oc = (GrObjectContainer)child;
                            GrHashList childProps = oc.getProperties();
                            GrProperty cp = (GrProperty)childProps.get("title");
                            if (cp.value.equals(title)) {
                                oldGroc = oc;
                                break;
                            }
                        }
                    }

                    int index = m_children.size();
                    if (oldGroc != null) {
                        index = m_children.indexOf(oldGroc);
                        oldGroc.deleteObserver(this);
                        m_children.remove(oldGroc);
                    }
                    m_children.add(index, groc);
                    groc.addObserver(this);
                    groc.setName("");

                    switch (i) {
                        case 0:
                            m_xyInputContainer = groc;
                            break;
                        case 1:
                            m_nzInputContainer = groc;
                            break;
                        case 2:
                            m_tnInputContainer = groc;
                            break;
                        case 3:
                            m_ndInputContainer = groc;
                            break;
                        case 4:
                            m_xyzInputContainer = groc;
                            break;
                        case 5:
                            m_txyzdInputContainer = groc;
                            break;
                        case 6:
                            m_finishLineInputContainer = groc;
                            break;
                        case 7:
                            m_randomProbabilityInputContainer = groc;
                            break;
                        case 8:
                            m_tsWeightingInputContainer = groc;
                            break;
                        case 9:
                            m_outputContainer = groc;

                            // Look for DoubleDataSeries within the output
                            // container and use them for output if they are
                            // there.
                            int numTS = 0;
                            int numXY = 0;
                            int numDS = 0;
                            Enumeration enumeration =
                                m_outputContainer.children();
                            while (enumeration.hasMoreElements()) {
                                GrObject child =
                                    (GrObject)enumeration.nextElement();
                                if (child instanceof DoubleDataSeries) {
                                    if (((DoubleDataSeries)child).
                                        getNumDimensions() == 1) {

                                        m_transitTimesOutputSeries =
                                            (DoubleDataSeries)child;
                                    }
                                    if (DoubleDataSeries.countDims(child, 2)  &&
                                        !(TimeSeries.countDims(child, 2))) {

                                         if (numXY == 0)
                                            m_xyOutputSeries =
                                                (DoubleDataSeries)child;
                                        if (numXY == 1)
                                            m_nWseSeries =
						(DoubleDataSeries)child;
                                        if (numXY == 2)
                                            m_nGridDepthSeries =
                                                (DoubleDataSeries)child;
                                         if (numXY == 3)
                                            m_pdfOutputSeries =
                                                (DoubleDataSeries)child;
                                         if (numXY == 4)
                                            m_cdfOutputSeries =
                                                (DoubleDataSeries)child;

                                        numXY++;
                                    }
                                    if (((DoubleDataSeries)child).
                                        getNumDimensions() == 4) {

                                        if (numDS == 0)
                                            m_txyzOutputSeries =
                                                (DoubleDataSeries)child;
                                        if (numDS == 1)
                                            m_lgOutputSeries =
                                                (DoubleDataSeries)child;
                                        numDS++;
                                    }
                                    if (child instanceof TimeSeries) {
                                        if (numTS == 0)
                                            m_txOutputSeries =(TimeSeries)child;
                                        if (numTS == 1)
                                            m_tyOutputSeries =(TimeSeries)child;
                                        if (numTS == 2)
                                            m_tzOutputSeries =(TimeSeries)child;
                                        numTS++;
                                    }
                                }
                            }

                            // Make sure there are no other output series
                            // in the output container besided those defined.
                            m_outputContainer.clear();
                            m_outputContainer.add(m_xyOutputSeries);
                            m_outputContainer.add(m_txOutputSeries);
                            m_outputContainer.add(m_tyOutputSeries);
                            m_outputContainer.add(m_tzOutputSeries);
                            m_outputContainer.add(m_txyzOutputSeries);
                            m_outputContainer.add(m_lgOutputSeries);
                            m_outputContainer.add(m_nWseSeries);
                            m_outputContainer.add(m_nGridDepthSeries);
                            m_outputContainer.add(m_pdfOutputSeries);
                            m_outputContainer.add(m_cdfOutputSeries);
                            m_outputContainer.add(m_transitTimesOutputSeries);

                            // Set default property values.
                            m_xyOutputSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_xyOutputSeries.getDimension(1).
                                    setProperty("locked", "true");
                            m_txOutputSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_txOutputSeries.getDimension(1).
                                    setProperty("locked", "true");
                            m_tyOutputSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_tyOutputSeries.getDimension(1).
                                    setProperty("locked", "true");
                            m_tzOutputSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_tzOutputSeries.getDimension(1).
                                    setProperty("locked", "true");
                            m_txyzOutputSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_txyzOutputSeries.getDimension(1).
                                    setProperty("locked", "true");
                            m_txyzOutputSeries.getDimension(2).
                                    setProperty("locked", "true");
                            m_txyzOutputSeries.getDimension(3).
                                    setProperty("locked", "true");
                            m_lgOutputSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_lgOutputSeries.getDimension(1).
                                    setProperty("locked", "true");
                            m_lgOutputSeries.getDimension(2).
                                    setProperty("locked", "true");
                            m_lgOutputSeries.getDimension(3).
                                    setProperty("locked", "true");
                            m_xyOutputSeries.setTitle(
                                    "X-Y Particle Path Components");
                            m_txOutputSeries.setTitle(
                                    "Time-X Particle Path Components");
                            m_tyOutputSeries.setTitle(
                                    "Time-Y Particle Path Components");
                            m_tzOutputSeries.setTitle(
                                    "Time-Z Particle Path Components");
                            m_txyzOutputSeries.setTitle(
                                    "Time-X-Y-Z Particle Paths");
                            m_lgOutputSeries.setTitle(
                                    "Time-X-Y-Z Lagrangian Particle Paths");
                            m_nWseSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_nWseSeries.getDimension(1).
                                    setProperty("locked", "true");
                            m_nWseSeries.setTitle(
                                    "N-Water Surface Elevation");
                            m_nGridDepthSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_nGridDepthSeries.getDimension(1).
                                    setProperty("locked", "true");
                            m_nGridDepthSeries.setTitle(
                                    "N-Grid Depth");
                            m_pdfOutputSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_pdfOutputSeries.getDimension(1).
                                    setProperty("locked", "true");
                            m_pdfOutputSeries.setTitle(
                                    "Probability Density");
                            m_cdfOutputSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_cdfOutputSeries.getDimension(1).
                                    setProperty("locked", "true");
                            m_cdfOutputSeries.setTitle(
                                    "Cumulative Density");
                            m_transitTimesOutputSeries.getDimension(0).
                                    setProperty("locked", "true");
                            m_transitTimesOutputSeries.setTitle(
                                    "Transit Times");
                            break;
                    }

                    m_bHasChanged = true;
                    return true;
                }
            }
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Overrides GrObject.remove() to keep containers from being removed.
 * Replacing existing GrObjectContainer children is supported through
 * the add(GrObject) method.
 * @param i     The index of the object to remove.
 * @return      The object that was removed.
 */
    public GrObject remove(int i) {

        if (m_children.get(i) instanceof GrObjectContainer)
            return null;

        if (m_children.get(i) instanceof ParticlePathCalculator)
            return null;

        // Let the super class handle removal.
        GrObject child = super.remove(i);

        if (child == m_bath) m_bath = null;
        if (child == m_si3dOut) m_si3dOut = null;

        return child;
    }

//------------------------------------------------------------------------------
/**
 * Overrides GrObject.remove() to keep containers from being removed.
 * Replacing existing GrObjectContainer children is supported through
 * the add(GrObject) method.
 * @param gro   The object to be removed.
 * @return      Whether the object was removed.
 */
    public boolean remove(GrObject gro) {

        if (gro instanceof GrObjectContainer ||
            gro instanceof ParticlePathCalculator) {

            return false;
        }
        else {
            return super.remove(gro);
        }
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * calcClass             String      "Calculator Class"                  <p>
 * autoRecalc            Boolean     "Automatic Recalculation"           <p>
 * useW                  Boolean     "Use W"                             <p>
 * tsDivs                Integer     "Time Step Divisions"               <p>
 * interpolationType     Integer     "Interpolation"                     <p>
 *      0 = none, 1 = linear, 2 = parabolic                              <p>
 * calcDrift             Boolean     "Calculate Vertical Drift"          <p>
 * calcTurbulence        Boolean     "Calculate Vertical Turbulence"     <p>
 * randomStartTime       Double      "Random Start Time"                 <p>
 * randomReleaseDuration Double      "Random Release Duration"           <p>
 * randomTrackDuration   Double      "Random Track Duration"             <p>
 * randomParticlesPerTS  Integer     "Random Particles Per TS"           <p>
 * randomTimeStep        Integer     "Random Time Step"                  <p>
 * randomBounds          String      "Random Bounds"                     <p>
 *      Delimited values for area of random release
 *      xMin, yMin, zMin, xMax, yMax, zMax                               <p>
 * distributionInterval  Double      "Distribution Interval"             <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("calcClass", new GrProperty(
            m_calc.getClass().getCanonicalName(),"Calculator Class"));
        props.put("autoRecalc", new GrProperty(
            m_bAutoRecalc, "Automatic Recalculation"));

        // Give the user some symbol choices...
        GrProperty p = new GrProperty("", "Interpolation");
        javax.swing.JComboBox jcb = new javax.swing.JComboBox();

        // Fill the symbol type combo box.
        for (int i = 0; i < m_calc.m_interpolationTypes.length; i++) {
            String name = m_calc.m_interpolationTypes[i];
            jcb.addItem(name);
            if (m_calc.m_interpolationIndex == i) {
                jcb.setSelectedIndex(i);
                p.value = m_calc.m_interpolationTypes[i];
            }
        }

        p.component = jcb;

        props.put("randomStartTime", new GrProperty(
            m_randomStartTime, "Random Start Time"));
        props.put("randomReleaseDuration", new GrProperty(
            m_randomReleaseDuration, "Random Release Duration"));
        props.put("randomTrackDuration", new GrProperty(
            m_randomTrackDuration, "Random Track Duration"));
        props.put("randomParticlesPerTS", new GrProperty(
            m_randomParticlesPerTS, "Random Particles Per TS"));
        props.put("randomTimeStep", new GrProperty(
            m_randomTimeStep, "Random Time Step"));

        String s = "";
        for (int i = 0; i < m_randomBounds.length; i++) {
            if (i > 0) s += ", ";
            s += m_randomBounds[i];
        }
        props.put("randomBounds", new GrProperty(s,
            "Random Bounds"));
        props.put("distributionInterval", new GrProperty(
            m_distributionInterval, "Distribution Interval"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.sfhydro.si3d.ParticleTracker#getProperties getProperties() for
 * a list of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                if (key.equals("autoRecalc"))
                    m_bAutoRecalc = (Boolean)p.value;
                if (key.equals("useW"))
                    m_calc.m_bUseW = (Boolean)p.value;
                if (key.equals("tsDivs")) {
                    int tsDivs = (Integer)p.value;
                    if (tsDivs > 0)
                        m_calc.m_tsDivs = tsDivs;
                }
                if (key.equals("interpolationType")) {
                    // Check each name.
                    int numTypes = m_calc.m_interpolationTypes.length;
                    for (int i = 0; i < numTypes; i++) {
                        String name = ((String)p.value).trim().toLowerCase();
                        if (name.equals(m_calc.m_interpolationTypes[i].
                            toLowerCase())) {

                            m_calc.m_interpolationIndex = i;
                            break;
                        }
                    }
                }
                if (key.equals("calcDrift"))
                    m_calc.m_bCalcDrift = (Boolean)p.value;
                if (key.equals("calcTurbulence"))
                    m_calc.m_bCalcTurbulence = (Boolean)p.value;
                if (key.equals("randomStartTime"))
                    m_randomStartTime = (Double)p.value;
                if (key.equals("randomReleaseDuration") ||
                    key.equals("randomDuration")) {

                    m_randomReleaseDuration = (Double)p.value;
                }
                if (key.equals("randomTrackDuration"))
                    m_randomTrackDuration = (Double)p.value;
                if (key.equals("randomParticlesPerTS"))
                    m_randomParticlesPerTS = (Integer)p.value;
                if (key.equals("randomTimeStep")) {
                    int iVal = (Integer)p.value;
//                    if (iVal > 0)
                        m_randomTimeStep = iVal;
                }
                if (key.equals("distributionInterval"))
                    m_distributionInterval = (Double)p.value;
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }

        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.sfhydro.si3d.ParticleTracker#getProperties getProperties() for
 * a list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("calcClass")) {
        	Object obj = GrController.createObject(value,
            	"gov.usgs.sfhydro.si3d.ParticlePathCalculator", null);
            if (obj != null) {
                ParticlePathCalculator newCalc = (ParticlePathCalculator)obj;
                newCalc.setProperties(m_calc.getProperties());
                m_calc = newCalc;
            }
        }
        if (key.equals("autoRecalc"))
            m_bAutoRecalc = Boolean.parseBoolean(value);
        if (key.equals("useW"))
            m_calc.m_bUseW = Boolean.parseBoolean(value);
        if (key.equals("tsDivs")) {
            int tsDivs = Integer.parseInt(value);
            if (tsDivs > 0)
                m_calc.m_tsDivs = tsDivs;
        }
        if (key.equals("randomStartTime"))
            m_randomStartTime = Double.parseDouble(value);
        if (key.equals("randomReleaseDuration") || key.equals("randomDuration"))
            m_randomReleaseDuration = Double.parseDouble(value);
        if (key.equals("randomTrackDuration"))
            m_randomTrackDuration = Double.parseDouble(value);
        if (key.equals("randomParticlesPerTS"))
            m_randomParticlesPerTS = Integer.parseInt(value);
        if (key.equals("randomTimeStep")) {
            int iVal = Integer.parseInt(value);
            m_randomTimeStep = iVal;
        }

        if (key.equals("randomBounds")) {
            double[] vals = AsciiIO.parseLineAsDoubles(value, ", \t");
            if (vals.length == 6)
                m_randomBounds = vals;
        }
        if (key.equals("distributionInterval"))
            m_distributionInterval = Double.parseDouble(value);
    }

//------------------------------------------------------------------------------
/**
 * Wrapper for the basic tracking method.
 * Converts start time and duration to start and end time steps and allocates
 * the array to hold the path.  Fills the portion of the array before and after
 * the start and end.
 * @param startTime The start time in Julian Days relative to the same year
 *                  as the Si3D output.
 * @param duration  The length of time to track the particle.
 * @param xStart    The X coordinate of the release point.
 * @param yStart    The Y coordinate of the release point.
 * @param zStart    The Z coordinate of the release point.
 * @return  A data series containing the Time-X-Y-Z particle path.  The number
 *          of time steps is (the number of Si3D output time steps minus one)
 *          multiplied by (the number of time step divisions) plus one.
 */
    protected DoubleDataSeries track (double startTime, double duration,
        double xStart, double yStart, double zStart) {

        return m_calc.track (startTime, duration, xStart, yStart, zStart);
    }

//------------------------------------------------------------------------------
/**
 * Track a particle through the Si3D vector field from the X-Y-Z release point
 * and starting time step until the ending time step.
 * @param startTS   The starting time step.
 * @param endTS     The ending time step.
 * @param offset    The time step at which the path array should start.
 * @param path      The array that will hold the Time-X-Y-Z path at each time.
 * @param xStart    The X coordinate of the release point.
 * @param yStart    The Y coordinate of the release point.
 * @param zStart    The Z coordinate of the release point.
 * @return The path array with the particle path coordinates in it.
 */
    protected double[] track (int startTS, int endTS, int offset, double[] path,
        double xStart, double yStart, double zStart) {

        return m_calc.track (startTS, endTS, offset, path,
            xStart, yStart, zStart);
    }

//------------------------------------------------------------------------------
/**
 * Return the DoubleDataSeries within the specified ObjectContainer within this
 * object.
 * @param containerTitle    The title of the container.
 * @return  The series inside the specified container, null if not found.
 */
    public DoubleDataSeries getContainerSeries(String containerTitle) {

        // Look through the children of this object for an ObjectContainer
        // with a matching title.
        GrObjectContainer groc =
            GrObjectContainer.getChildContainer(this, containerTitle);

        if (groc != null) {
            // Look through the contents of the ObjectContainer for
            // a DoubleDataSeries.
            Enumeration enumeration = groc.children();
            while (enumeration.hasMoreElements()) {
                GrObject oc_child = (GrObject)enumeration.nextElement();
                if (oc_child instanceof DoubleDataSeries)
                    return (DoubleDataSeries)oc_child;
            }
        }

        // Could not find a variables series under a container with the
        // given title.
        return null;
    }

//------------------------------------------------------------------------------
/**
 * For each point specified by the input arrays, release and track a particle.
 * Store the particle path in the output data series.
 */
    public void recalculate () {

        // Initialize variables from the Si3dOutput and Bathymetry.
        if (!initializeVariables())
            return;

        // Get the input series.
        DoubleDataSeries xyInputSeries = getContainerSeries("X-Y Input Series");
        DoubleDataSeries nzInputSeries = getContainerSeries("N-Z Input Series");
        DoubleDataSeries tnInputSeries = getContainerSeries("T-N Input Series");
        DoubleDataSeries ndInputSeries = getContainerSeries("N-D Input Series");
        DoubleDataSeries xyzInputSeries =
            getContainerSeries("X-Y-Z Input Series");
        DoubleDataSeries txyzdInputSeries =
            getContainerSeries("T-X-Y-Z-D Input Series");

        // If the Time-X-Y-Z input series was not provided directly, form it
        // from the series that were provided.
        if (txyzdInputSeries == null) {
            txyzdInputSeries = new DoubleDataSeries(5);
            if (xyzInputSeries == null) {
                if (xyInputSeries == null) {

                    return;
                }
                else {
                    // Get starting locations from the X-Y input file.
                    int numPoints = xyInputSeries.getNumPoints();
                    int numXyDims = xyInputSeries.getNumDimensions();
                    double[] txyzdArray = new double[numPoints * 5];
                    double[] xyArray = xyInputSeries.getArray();
                    for (int i = 0; i < numPoints; i++) {
                        txyzdArray[i * 5] = 0;
                        txyzdArray[i * 5 + 1] = xyArray[i * numXyDims];
                        txyzdArray[i * 5 + 2] = xyArray[i * numXyDims + 1];
                        txyzdArray[i * 5 + 3] = 0;
                        txyzdArray[i * 5 + 4] = -1;
                    }

                    // Add the z value if there is an N-Z series.
                    if (nzInputSeries != null) {
                        int numNzPoints = nzInputSeries.getNumPoints();
                        int numNzDims = xyInputSeries.getNumDimensions();
                        double[] nzArray = nzInputSeries.getArray();
                        for (int i = 0; i < numPoints; i++) {
                            if (i >= numNzPoints)
                                txyzdArray[i * 5 + 3] = 0;
                            else
                                txyzdArray[i * 5 + 3] =
                                    nzArray[i * numNzDims + 1];
                        }
                    }

                    // Add the time value if there is a Time-N series.
                    if (tnInputSeries != null) {
                        int numTnPoints = tnInputSeries.getNumPoints();
                        int numTnDims = tnInputSeries.getNumDimensions();
                        double[] tnArray = tnInputSeries.getArray();
                        for (int i = 0; i < numPoints; i++) {
                            if (i >= numTnPoints)
                                txyzdArray[i * 5] = 0;
                            else
                                txyzdArray[i * 5] = tnArray[i * numTnDims];
                        }
                    }

                    // Add the duration value if there is an N-Duration series.
                    if (ndInputSeries != null) {
                        int numNdPoints = ndInputSeries.getNumPoints();
                        int numNdDims = ndInputSeries.getNumDimensions();
                        double[] ndArray = ndInputSeries.getArray();
                        for (int i = 0; i < numPoints; i++) {
                            if (i >= numNdPoints)
                                txyzdArray[i * 5] = 0;
                            else
                                txyzdArray[i * 5 + 4] =
                                    ndArray[i * numNdDims + 1];
                        }
                    }

                    // Apply the newly created array.
                    txyzdInputSeries.setArray(txyzdArray);
                }
            }
            else {
                // The X-Y-Z were provided in one series, so just add a time
                // dimension and use them as input.
                int numPoints = xyzInputSeries.getNumPoints();
                int numXyzDims = xyzInputSeries.getNumDimensions();
                int numTxyzdDims = 5;
                double[] txyzdArray = new double[numPoints * numTxyzdDims];
                double[] xyzArray = xyzInputSeries.getArray();
                for (int i = 0; i < numPoints; i++) {
                    txyzdArray[i * numTxyzdDims] = 0;
                    txyzdArray[i * numTxyzdDims + 1] =
                        xyzArray[i * numXyzDims];
                    txyzdArray[i * numTxyzdDims + 2] =
                        xyzArray[i * numXyzDims + 1];
                    txyzdArray[i * numTxyzdDims + 3] =
                        xyzArray[i * numXyzDims + 2];
                }
                txyzdInputSeries.setArray(txyzdArray);
            }
        }

        // Find the bathymetry child.  Return if there is none.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof Bathymetry) m_bath = (Bathymetry)(gro);
        }

        if (m_bath == null) return;

        // Get the 2D and 3D variable series.
        m_calc.m_3DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("3-D Variables");
//        if (m_calc.m_3DVarsSeries != null)
//            m_calc.m_3DVarsArray = m_calc.m_3DVarsSeries.getArray();
        m_calc.m_2DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("2-D x-y Variables");

        // Get the scale factor for Water Surface Elevation.
        if (m_calc.m_2DVarsSeries != null) {
//            m_calc.m_2DVarsArray = m_calc.m_2DVarsSeries.getArray();

            DataDimension dim = m_calc.m_2DVarsSeries.getDimension(0);
            if (dim != null) {
                // Store the scale factor for the dimension.
                GrHashList dimProps = dim.getProperties();
                GrProperty p = (GrProperty)dimProps.get("unitScale");
                m_calc.m_wseScale = (Double)p.value;
            }
        }

        // Get other variables.
        int[] indexArray2D = m_bath.get2DIndexArray();
        m_calc.m_2DLookupArray = m_bath.get2DLookupArray(indexArray2D);
        int[] indexArray3D = m_bath.get3DIndexArray();
        DoubleDataSeries bathSeries = m_bath.getBathymetrySeries();
        m_calc.m_numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        m_calc.m_numJ = (int)Math.round(bathSeries.getMax(1)) + 1;

        // Create lists to hold the paths.
        ArrayList pathSeriesList = new ArrayList();
        ArrayList xyzPaths = new ArrayList();

        // Set the input descriptors.
        double[] txyzdArray = txyzdInputSeries.getArray();
        int numTxyzdPoints = txyzdInputSeries.getNumPoints();
        int numTxyzdDims = txyzdInputSeries.getNumDimensions();

        // Create an array to save the depth and WSE at the first time
        // step of each particle release.
        double[] nWseArray = new double[txyzdInputSeries.getCount() * 2];
        double[] nGridDepthArray = new double[txyzdInputSeries.getCount() * 2];

        // Declare the array to hold Lagrangian paths.
        double[] lgPaths = null;

        // Declare a variable for keeping track of elapsed calculation time.
        GregorianCalendar gc1 = new GregorianCalendar();
        long firstTimeInMillis = gc1.getTimeInMillis();
        long lastTimeInMillis = firstTimeInMillis;

        // Release a particle for each input release point.
        for (int i = 0; i < numTxyzdPoints; i++) {
            // Get input descriptor variables.
            double startTime = txyzdArray[i * numTxyzdDims];
            double x = txyzdArray[i * numTxyzdDims + 1];
            double y = txyzdArray[i * numTxyzdDims + 2];
            double z = txyzdArray[i * numTxyzdDims + 3];
            double duration = txyzdArray[i * numTxyzdDims + 4];

            // Reset the vertical range variables.
            m_calc.m_bSaveInitialRange = true;
            m_calc.m_initialWSE = 0;
            m_calc.m_initialGridDepth = 0;

            // Track the particle and store the path.
            DoubleDataSeries path = m_calc.track(startTime, duration, x, y, z);

            // If no path was generated, set the error string and return.
            if (path == null) {
                m_errorString = " -- Error -- Could not recalculate path";
                return;
            }

            // Create a Lagrangian path from the original path.
            DoubleDataSeries lgPath = (DoubleDataSeries)path.createCopy(true);
            lgPath.m_bParabolicInterpolation = true;
            lgPath.resample(1.0333333333333333, startTime, 0);

            int numLgPoints = lgPath.getCount();
            if (lgPaths == null) {
                // Allocate the Lagrangian path array.
                lgPaths = new double[numTxyzdPoints * numLgPoints * 4];
            }

            // Copy the Lagrangian path into the larger interleaved array
            // for all Lagrangian paths.
            if (numLgPoints <= (lgPaths.length / numTxyzdPoints / 4)) {
                double[] lgPathArray = lgPath.getArray();
                for (int j = 0; j < numLgPoints; j++) {
                    lgPaths[(i + j * numTxyzdPoints) * 4] =
                        lgPathArray[j * 4];
                    lgPaths[(i + j * numTxyzdPoints) * 4 + 1]
                        = lgPathArray[j * 4 + 1];
                    lgPaths[(i + j * numTxyzdPoints) * 4 + 2]
                        = lgPathArray[j * 4 + 2];
                    lgPaths[(i + j * numTxyzdPoints) * 4 + 3]
                        = lgPathArray[j * 4 + 3];
                }
            }

            // Allocate the path series, if it has not been already.
            if (path == null)
                path = new DoubleDataSeries(4);

            // Copy the Time and Z values into the Time-X output array.
            if (i == 0) {
                double[] tzArray = new double[path.getCount() * 2];
                double[] pathArray = path.getArray();
                int numPathDims = path.getNumDimensions();
                for (int j = 0; j < path.getCount(); j++) {
                    tzArray[j * 2] = pathArray[j * numPathDims];
                    tzArray[j * 2 + 1] = pathArray[j * numPathDims + 3];
                }
                m_tzOutputSeries.setArray(tzArray);
            }

            // Save the depth and WSE at the first time step.
            nWseArray[i * 2] = i;
            nWseArray[i * 2 + 1] = m_calc.m_initialWSE;
            nGridDepthArray[i * 2] = i;
            nGridDepthArray[i * 2 + 1] = m_calc.m_initialGridDepth;

            // If the particle travelled at least one cell, combine the
            // locations into an X-Y-Z series.
            if ((path.getMax(0) - path.getMin(0) > 1) ||
                (path.getMax(1) - path.getMin(1) > 1)) {

                GrHashList dims = new GrHashList();
                DataDimension dim = path.getDimension(1);
                dims.put(dim, dim.getAssociatedSeries());
                dim = path.getDimension(2);
                dims.put(dim, dim.getAssociatedSeries());
                dim = path.getDimension(3);
                dims.put(dim, dim.getAssociatedSeries());
                DoubleDataSeries xyzSeries =
                    (DoubleDataSeries)DimensionOrganizer.combine(dims, false);
                xyzPaths.add(xyzSeries.getArray());
                pathSeriesList.add(path);
            }

            // Write out a status message every 100 releases, as long as
            // at least a minute has passed since the last message.
            if (i > 1 && ((i % 100) == 0)) {
                GregorianCalendar gc2 = new GregorianCalendar();
                long nowInMillis = gc2.getTimeInMillis();
                if ((nowInMillis - lastTimeInMillis) > 60000) {
                    double totalElapsedTime = (double)
                        ((nowInMillis - firstTimeInMillis) / 6000);
                    totalElapsedTime /= 10;
                    System.out.println("Tracked " + i + " particles of "
                        + numTxyzdPoints + " in " + totalElapsedTime
                        + " minutes");
                    lastTimeInMillis = nowInMillis;
                }
            }
        }

        // Set the WSE and depth output series.
        m_nWseSeries.setArray(nWseArray);
        m_nGridDepthSeries.setArray(nGridDepthArray);

        // Count the number of points in all the paths and mark where each
        // one begins and ends.
        int numTimeSteps = 0;
        int totalPoints = 0;
        int[] breaks = new int[xyzPaths.size() * 3];
        int breakIndex = 0;
        iter = xyzPaths.iterator();
        while (iter.hasNext()) {
            double[] pathArray = (double[])iter.next();
            int numPathPoints = pathArray.length;
            breaks[breakIndex] = totalPoints / 3;
            totalPoints += numPathPoints;
            breaks[breakIndex + 1] = (totalPoints - 3) / 3;
            breakIndex += 3;

            if ((totalPoints / 3) > numTimeSteps)
                numTimeSteps = totalPoints / 3;
        }

        int txyzTotalPoints = (totalPoints / 3) * 4;

        // Create one array to hold all the paths and copy them into
        // this new array.
        double[] pathsArray = new double[totalPoints];
        int pathsIndex = 0;
        iter = xyzPaths.iterator();
        while (iter.hasNext()) {
            double[] pathArray = (double[])iter.next();
            for (int i = 0; i < pathArray.length; i++) {
                pathsArray[pathsIndex] = pathArray[i];
                pathsIndex++;
            }
        }

        // Create a Time-X-Y-Z output array for all released particles and
        // copy the paths into it.
        double[] txyzOutputArray = new double[txyzTotalPoints];
        int txyzIndex = 0;
        for (int i = 0; i < numTimeSteps; i++) {
            int pathNum = 0;
            double time = 0;
            iter = pathSeriesList.iterator();
            while (iter.hasNext()) {
                DoubleDataSeries pathSeries = (DoubleDataSeries)iter.next();
                if (pathSeries.getCount() <= i) break;

                double[] pathArray = pathSeries.getArray();
                int numDims = pathSeries.getNumDimensions();

                // Use the time from the first particle for all particles
                // on this time step.  This keeps rounding errors from
                // splitting the time step into multiple steps in the
                // TimeMarker.
                if (pathNum == 0)
                    time = pathArray[i * numDims];
                txyzOutputArray[txyzIndex] = time;
                txyzOutputArray[txyzIndex + 1] = pathArray[i * numDims + 1];
                txyzOutputArray[txyzIndex + 2] = pathArray[i * numDims + 2];
                txyzOutputArray[txyzIndex + 3] = pathArray[i * numDims + 3];
                txyzIndex += 4;
                pathNum++;
            }
        }

        m_txyzOutputSeries.setArray(txyzOutputArray);

        // Combine data into the X-Y output series.
        GrHashList dims = new GrHashList();
        DataDimension tDim = m_txyzOutputSeries.getDimension(0);
        DataDimension xDim = m_txyzOutputSeries.getDimension(1);
        DataDimension yDim = m_txyzOutputSeries.getDimension(2);
        DataDimension zDim = m_txyzOutputSeries.getDimension(3);
        dims.put(xDim, m_txyzOutputSeries);
        dims.put(yDim, m_txyzOutputSeries);
        DoubleDataSeries xySeries =
            (DoubleDataSeries)DimensionOrganizer.combine(dims, false);
        m_xyOutputSeries.setArray(xySeries.getArray());

        // Combine data into the Time-X output series.
        dims.clear();
        dims.put(tDim, m_txyzOutputSeries);
        dims.put(xDim, m_txyzOutputSeries);
        xySeries = (DoubleDataSeries)DimensionOrganizer.combine(dims, false);
        m_txOutputSeries.setArray(xySeries.getArray());

        // Combine data into the Time-Y output series.
        dims.clear();
        dims.put(tDim, m_txyzOutputSeries);
        dims.put(yDim, m_txyzOutputSeries);
        xySeries = (DoubleDataSeries)DimensionOrganizer.combine(dims, false);
        m_tyOutputSeries.setArray(xySeries.getArray());

        m_lgOutputSeries.setArray(lgPaths);
    }

//------------------------------------------------------------------------------
/**
 * Read the Bathymetry and Si3dOutput children and load their values into the
 * member variables of this class.
 * @return  Whether all the variables were read and initialized successfully.
 */
    protected boolean initializeVariables() {

        // Make sure the Bathymetry and Si3dOutput objects are defined.
        if (m_bath == null || m_si3dOut == null) {
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                GrObject gro = (GrObject)iter.next();
                if (gro instanceof Bathymetry) m_bath = (Bathymetry)(gro);
                if (gro instanceof Si3dOutput) m_si3dOut = (Si3dOutput)gro;
            }
        }

        if (m_bath == null || m_si3dOut == null) return false;

        // Get grid spacing information from the Bathymetry object.
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        m_calc.m_xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        m_calc.m_ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        m_calc.m_zSpacing = (Double)p.value;

        // Get time step information from the Si3dOutput object.
        m_calc.m_si3dRunStartTime = m_si3dOut.getJulianStartTime();
        GrHashList si3dOutProps = m_si3dOut.getProperties();
        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long)p.value;
        m_calc.m_timeStepInc = ((double)tsInSec) / 86400;

        // Find the dimensions of the grid and the number of wet 2D points.
        DoubleDataSeries bathSeries = m_bath.getBathymetrySeries();
        m_calc.m_numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        m_calc.m_numJ = (int)Math.round(bathSeries.getMax(1)) + 1;
        m_calc.m_numLayers = (int)Math.round(bathSeries.getMax(2));
        m_calc.m_numWet2Dpoints = bathSeries.getNumPoints();
        m_calc.m_avgDepthArray = m_bath.getAverageDepthArray();

        // Get array indexing information from the Bathymetry object.
        m_calc.m_3DIndexArray = m_bath.get3DIndexArray();

        // Get the rectangular array that holds -1 at dry points and the
        // index into the variables array at wet points.
        m_calc.m_3DLookupArray = m_bath.get3DLookupArray(m_calc.m_3DIndexArray);

        // Get the 2D and 3D variable series.
        m_calc.m_3DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("3-D Variables");
//        if (m_calc.m_3DVarsSeries != null)
//            m_calc.m_3DVarsArray = m_calc.m_3DVarsSeries.getArray();
        m_calc.m_2DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("2-D x-y Variables");
        if (m_calc.m_2DVarsSeries != null) {
//            m_calc.m_2DVarsArray = m_calc.m_2DVarsSeries.getArray();

            // Store the Water Surface Elevation scale.
            DataDimension dim = m_calc.m_2DVarsSeries.getDimension(0);
            if (dim != null) {
                // Store the scale factor for the dimension.
                GrHashList dimProps = dim.getProperties();
                p = (GrProperty)dimProps.get("unitScale");
                m_calc.m_wseScale = (Double)p.value;
            }
        }

        // Get the lookup and a index arrays.
        m_calc.m_2DIndexArray = m_bath.get2DIndexArray();
        m_calc.m_2DLookupArray = m_bath.get2DLookupArray(m_calc.m_2DIndexArray);
        m_calc.m_3DIndexArray = m_bath.get3DIndexArray();
        m_calc.m_3DLookupArray = m_bath.get3DLookupArray(m_calc.m_3DIndexArray);

        // Get the 3D variables array.
        m_calc.m_num3DVarsDims = m_calc.m_3DVarsSeries.getNumDimensions();
//        m_calc.m_3DVarsArray = m_calc.m_3DVarsSeries.getArray();

        // Find the unit scales for the X, Y, and Z dimensions.
        int numDims = m_calc.m_3DVarsSeries.getNumDimensions();
        m_calc.m_unitScales = new double[numDims];
        for (int i = 0; i < numDims; i++) {
            GrProperty unitProp = (GrProperty)(m_calc.m_3DVarsSeries.
                getDimension(i).getProperties().get("unitScale"));
            m_calc.m_unitScales[i] = (Double)unitProp.value;
        }

        // Create an array to hold Time-X-Y-Z data of the particle at every
        // time step for the duration of the run.
        m_calc.m_numWet3Dpoints = m_calc.m_3DIndexArray.length / 3;
        m_calc.m_numTimeSteps =
            m_calc.m_3DVarsSeries.getCount() / m_calc.m_numWet3Dpoints;

        // Set up an array of arrays holding the endpoints of each line
        // defining an open boundary.
        ArrayList dataList = new ArrayList();
        GrObjectContainer boundaryOC =
            GrObjectContainer.getChildContainer(m_bath, "Open Boundaries");
        if (boundaryOC != null) {
            Enumeration boundaryEnum = boundaryOC.children();
            while (boundaryEnum.hasMoreElements()) {
                GrObject gro = (GrObject)boundaryEnum.nextElement();
                if (gro instanceof DoubleDataSeries) {
                    dataList.add(gro);
                }
            }
            m_calc.m_boundaries = new double[dataList.size()][];
            Iterator iter = dataList.iterator();
            int lineIndex = 0;
            while (iter.hasNext()) {
                DoubleDataSeries series = (DoubleDataSeries)iter.next();
                m_calc.m_boundaries[lineIndex] =
                    m_calc.cellBoundaryLine(series);
                lineIndex++;
            }
        }

        // Set up an array of arrays holding the endpoints of each line
        // defining a thin-wall barrier.
        dataList.clear();
        GrObjectContainer barrierOC =
            GrObjectContainer.getChildContainer(m_bath, "Thin-wall Barriers");
        if (barrierOC != null) {
            Enumeration barrierEnum = barrierOC.children();
            while (barrierEnum.hasMoreElements()) {
                GrObject gro = (GrObject)barrierEnum.nextElement();
                if (gro instanceof DoubleDataSeries) {
                    dataList.add(gro);
                }
            }
            m_calc.m_barriers = new double[dataList.size()][];
            Iterator iter = dataList.iterator();
            int lineIndex = 0;
            while (iter.hasNext()) {
                DoubleDataSeries series = (DoubleDataSeries)iter.next();
                m_calc.m_barriers[lineIndex] = m_calc.cellBoundaryLine(series);
                lineIndex++;
            }
        }

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Releases particles at random locations and plots the distribution of transit
 * times to the finish line.
 */
    public void createTransitTimeDistribution() {

        // Initialize variables from the Si3dOutput and Bathymetry.
        if (!initializeVariables())
            return;

        DoubleDataSeries finishLineInputSeries = getContainerSeries(
            "Finish Line Input Series");

        double[][] oldBoundaries = m_calc.m_boundaries;
        if (finishLineInputSeries != null) {
            int numFinishLines = finishLineInputSeries.getCount() / 2;
            m_calc.m_boundaries =
                new double[oldBoundaries.length + numFinishLines][];

            for (int i = 0; i < oldBoundaries.length; i++) {
                m_calc.m_boundaries[i] = oldBoundaries[i];
            }

            double[] newLines = finishLineInputSeries.getArray();
            for (int i = 0; i < numFinishLines; i++) {
                double[] newArray = new double[4];
                newArray[0] = newLines[i * 4] / m_calc.m_xSpacing;
                newArray[1] = newLines[i * 4 + 1] / m_calc.m_ySpacing;
                newArray[2] = newLines[i * 4 + 2] / m_calc.m_xSpacing;
                newArray[3] = newLines[i * 4 + 3] / m_calc.m_ySpacing;

                double diffX = Math.abs(newArray[2] - newArray[0]);
                double diffY = Math.abs(newArray[3] - newArray[1]);
                if (diffX < diffY)
                    newArray[2] = newArray[0];
                else
                    newArray[3] = newArray[1];

                m_calc.m_boundaries[oldBoundaries.length + i] = newArray;
            }
        }

        // Clear prior input series.
        GrObjectContainer txyzdContainer =
            GrObjectContainer.getChildContainer(this, "T-X-Y-Z-D Input Series");
        txyzdContainer.clear();

        generateRandomReleaseInput();

        DoubleDataSeries txyzdInputSeries =
            getContainerSeries("T-X-Y-Z-D Input Series");

        // Set the input descriptors.
        double[] txyzdArray = txyzdInputSeries.getArray();
        int numTxyzdPoints = txyzdInputSeries.getNumPoints();
        int numTxyzdDims = txyzdInputSeries.getNumDimensions();
        // Allocate an array to hold the path at the PTM time step.

        double ptmTimeStepInc = m_calc.m_timeStepInc / (double)m_calc.m_tsDivs;
        int ptmNumTimeSteps = (m_calc.m_numTimeSteps - 1) * m_calc.m_tsDivs + 1;

        double[] times = new double[numTxyzdPoints];
        int numTrackStatDims = 10;
        m_trackStats = new double[numTxyzdPoints * numTrackStatDims];
        double[] path = new double[4 * ptmNumTimeSteps];
        m_calc.m_bTrackTimeOnly = true;

        // Declare a variable for keeping track of elapsed calculation time.
        GregorianCalendar gc1 = new GregorianCalendar();
        long firstTimeInMillis = gc1.getTimeInMillis();
        long lastTimeInMillis = firstTimeInMillis;

        DoubleDataSeries bathSeries = m_bath.getBathymetrySeries();
        int numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        double[] depthArray = m_bath.getAverageDepthArray();
        // Calculate the number of wet points in the I-J plane.
        int num2D = m_bath.get2DIndexArray().length / 2;

        // Release a particle for each input release point.
        for (int i = 0; i < numTxyzdPoints; i++) {
            // Get input descriptor variables.
            double startTime = txyzdArray[i * numTxyzdDims];
            double x = txyzdArray[i * numTxyzdDims + 1];
            double y = txyzdArray[i * numTxyzdDims + 2];
            double z = txyzdArray[i * numTxyzdDims + 3];
            double duration = txyzdArray[i * numTxyzdDims + 4];

            // Convert the start time to the nearest time step.
            int releaseTS = (int)Math.round((startTime -
                m_calc.m_si3dRunStartTime) / ptmTimeStepInc);

            // Perform bounds checking on the start time step.
            if (releaseTS < 0)
                releaseTS = 0;
            if (releaseTS >= ptmNumTimeSteps)
                releaseTS = ptmNumTimeSteps - 1;
            int startTS = releaseTS + 1;

            // Convert the end time to the nearest time step.
            int numTS = (int)Math.round(duration / ptmTimeStepInc);

            // Perform bounds checking on the end time step.
            if (numTS < 0 || numTS > (ptmNumTimeSteps - startTS - 1))
                numTS = ptmNumTimeSteps - startTS - 1;

//            m_trackStats[i * numTrackStatDims] = (double)startTS;
            m_trackStats[i * numTrackStatDims] = startTime;
            int currentI = (int)(x / m_calc.m_xSpacing + 0.5);
            int currentJ = (int)(y / m_calc.m_ySpacing + 0.5);

            // Determine cell height.
            int ijIndex = currentI + (currentJ * numI);
            int lookup2D = m_calc.m_2DLookupArray[ijIndex];
            if (lookup2D < 0) {
                System.err.println("Error:  release out of bounds.");
            }
            double waterSurfaceElevation = m_calc.getVar2D(lookup2D
                + (releaseTS * num2D)) * m_calc.m_wseScale;
            double depth = depthArray[lookup2D];
////// Check StartTS inconsistencies.
//            if (z < -depth) {
//                System.out.println("Release point Z changed from " + z + " to "
//                    + (-depth));
//                z = -depth;
//            }
//            if (z > waterSurfaceElevation) {
//                System.out.println("Release point Z changed from " + z + " to "
//                    + waterSurfaceElevation);
//                z = waterSurfaceElevation;
//            }

            int currentK = (int)(-z / m_calc.m_zSpacing) + 1;
            if (z > 0)
                currentK = 1;
            int offset = ((currentI * m_calc.m_numLayers)
                + (currentJ * m_calc.m_numLayers * m_calc.m_numI) + currentK);
            int lookupIndex = m_calc.m_3DLookupArray[offset];

// Check2
//System.out.println("" + i + " LookupIndex = " + lookupIndex + ", i = " + currentI
//    + ", j = " + currentJ + ", k = " + currentK + ", x = " + x
//    + ", y = " + y + ", z = " + z + ", offset = " + offset);
            m_trackStats[i * numTrackStatDims + 1] = lookupIndex;
            m_trackStats[i * numTrackStatDims + 4] = x;
            m_trackStats[i * numTrackStatDims + 5] = y;
            m_trackStats[i * numTrackStatDims + 6] = z;
            // Reset the vertical range variables.
            m_calc.m_bSaveInitialRange = true;
            m_calc.m_initialWSE = 0;
            m_calc.m_initialGridDepth = 0;

            path[0] = Double.POSITIVE_INFINITY;
            path[1] = -1;

            // Track the particle.
            path = m_calc.track(1, numTS, startTS, path, x, y, z);

//            // Only count the time if it crossed one of the new boundaries.
//            if (Math.round(path[1]) < oldBoundaries.length)
//                path[0] = Double.POSITIVE_INFINITY;

            times[i] = path[0] - startTime;

            double finishTime = path[0];
            double finishX = -1;
            double finishY = -1;
            double finishZ = -1;

            if (path[1] < 0) {
                finishTime = path[numTS * 4];
                finishX = path[numTS * 4 + 1];
                finishY = path[numTS * 4 + 2];
                finishZ = path[numTS * 4 + 3];
            }

            m_trackStats[i * numTrackStatDims + 2] = (double)path[1];
            m_trackStats[i * numTrackStatDims + 3] = finishTime;
            m_trackStats[i * numTrackStatDims + 7] = finishX;
            m_trackStats[i * numTrackStatDims + 8] = finishY;
            m_trackStats[i * numTrackStatDims + 9] = finishZ;

            // Write out a status message every 100 releases, as long as
            // at least a minute has passed since the last message.
            if (i > 1 && ((i % 100) == 0)) {
                GregorianCalendar gc2 = new GregorianCalendar();
                long nowInMillis = gc2.getTimeInMillis();
                if ((nowInMillis - lastTimeInMillis) > 60000) {
                    double totalElapsedTime = (double)
                        ((nowInMillis - firstTimeInMillis) / 6000);
                    totalElapsedTime /= 10;
                    System.out.println("Tracked " + i + " particles of "
                        + numTxyzdPoints + " in " + totalElapsedTime
                        + " minutes");
                    lastTimeInMillis = nowInMillis;
                }
            }
        }

        m_calc.m_bTrackTimeOnly = false;
        m_calc.m_boundaries = oldBoundaries;

        Arrays.sort(times);
        int numNonInfinite = Arrays.binarySearch(times, Double.MAX_VALUE);
        if (numNonInfinite < 0)
            numNonInfinite = -(numNonInfinite + 1);

        // Handle the case of all times being infinite
        if (numNonInfinite <= 0) {
            m_pdfOutputSeries.setArray(new double[0]);
            m_cdfOutputSeries.setArray(new double[0]);
            return;
        }

        double maxNonInfiniteTime = times[numNonInfinite - 1];
        int numDivs = (int)(maxNonInfiniteTime / m_distributionInterval) + 1;
        double halfInterval = m_distributionInterval / 2;
        double[] pdfArray = new double[numDivs * 2];
        int lastIndex = 0;
        for (int i = 0; i < numDivs; i++) {
            pdfArray[i * 2] = (double)i * m_distributionInterval + halfInterval;
            double divisionPoint = (double)(i + 1) * m_distributionInterval;

            int intervalSum = 0;
            for (int j = lastIndex; j < numTxyzdPoints; j++) {
                 if (times[j] > divisionPoint) {
                     lastIndex = j;
                     break;
                 }
                 intervalSum++;
            }

            pdfArray[i * 2 + 1] =
                ((double)intervalSum / (double)numTxyzdPoints) * 100;
        }

        m_pdfOutputSeries.setArray(pdfArray);

        double[] cdfArray = new double[pdfArray.length];
        double cumulativeSum = 0;
        for (int i = 0; i < numDivs; i++) {
            cdfArray[i * 2] = pdfArray[i * 2];
            cumulativeSum += pdfArray[i * 2 + 1];
            cdfArray[i * 2 + 1] = cumulativeSum;
        }

        m_cdfOutputSeries.setArray(cdfArray);

        m_transitTimesOutputSeries.setArray(times);

    }

//------------------------------------------------------------------------------
/**
 * Generates random release locations to use as input.  Locations are
 * distributed across wet cells within a bounded area.
 */
    protected void generateRandomReleaseInput() {

        // Make sure the Bathymetry and Si3dOutput objects are defined.
        if (m_bath == null || m_si3dOut == null) {
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                GrObject gro = (GrObject)iter.next();
                if (gro instanceof Bathymetry) m_bath = (Bathymetry)(gro);
                if (gro instanceof Si3dOutput) m_si3dOut = (Si3dOutput)gro;
            }
        }

        if (m_bath == null || m_si3dOut == null) return;

        // Get grid spacing information from the Bathymetry object.
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        double xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        double ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        double zSpacing = (Double)p.value;

        // Find the dimensions of the grid and the number of wet 2D points.
        DoubleDataSeries bathSeries = m_bath.getBathymetrySeries();
        int numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        int numJ = (int)Math.round(bathSeries.getMax(1)) + 1;
        int numGridLayers = (int)Math.round(bathSeries.getMax(2));
        double[] depthArray = m_bath.getAverageDepthArray();

        // Get array indexing information from the Bathymetry object.
        int[] indexArray3D = m_bath.get3DIndexArray();

        // Check for an input probability series.
        double[] probabilityArray = null;
        double[] tsWeightingArray = null;
        DoubleDataSeries probabilitySeries = getContainerSeries(
            "Random Probability Input Series");
        if (probabilitySeries != null)
            probabilityArray = probabilitySeries.getArray();
        DoubleDataSeries tsWeightingSeries = getContainerSeries(
            "Time Step Weighting Input Series");

        // Get the 2D versions.
        int[] indexArray2D = m_bath.get2DIndexArray();
        m_calc.m_2DLookupArray = m_bath.get2DLookupArray(indexArray2D);
        m_calc.m_3DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("3-D Variables");
        m_calc.m_2DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("2-D x-y Variables");
        if (m_calc.m_2DVarsSeries != null) {
//            m_calc.m_2DVarsArray = m_calc.m_2DVarsSeries.getArray();

            DataDimension dim = m_calc.m_2DVarsSeries.getDimension(0);
            if (dim != null) {
                // Store the scale factor for the dimension.
                GrHashList dimProps = dim.getProperties();
                p = (GrProperty)dimProps.get("unitScale");
                m_calc.m_wseScale = (Double)p.value;
            }
        }

        // Get array indexing information from the Bathymetry object.
        m_calc.m_3DIndexArray = m_bath.get3DIndexArray();
        m_calc.m_numWet3Dpoints = m_calc.m_3DIndexArray.length / 3;
        m_calc.m_numWet2Dpoints = bathSeries.getNumPoints();


        // 15 min = 0.010416667

        // Calculate the number of wet points in the I-J plane.
        int num2D = indexArray2D.length / 2;
        // Calculate the number of wet points in the I-J-K grid.
        int num3D = indexArray3D.length / 3;

        // Get time step information from the Si3dOutput object.
        double si3dRunStartTime = m_si3dOut.getJulianStartTime();
        if (m_randomStartTime < si3dRunStartTime)
            m_randomStartTime = si3dRunStartTime;
        GrHashList si3dOutProps = m_si3dOut.getProperties();
        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long)p.value;
        double timeStepInc = ((double)tsInSec) / 86400;

        int tsMultiplier = 1;
        if (m_randomTimeStep < 0) {
            tsMultiplier = -m_randomTimeStep;
        }
        timeStepInc /= tsMultiplier;
        int randomTimeStep = m_randomTimeStep;
        if (randomTimeStep < 0)
            randomTimeStep = 1;

        int startTS = (int)Math.round(
            (m_randomStartTime - si3dRunStartTime) / timeStepInc);
        int numTimeSteps = m_calc.m_3DVarsSeries.getCount() / num3D;
        if (numTimeSteps < 1)
            return;

        numTimeSteps *= tsMultiplier;
        if (startTS > numTimeSteps)
            startTS = numTimeSteps;

        int endTS = startTS +
            (int)Math.round(m_randomReleaseDuration / timeStepInc);

        if (endTS < startTS || endTS >= numTimeSteps)
            endTS = numTimeSteps - 1;

        // Determine how many releases there will be and allocate space
        // for them.
        int totalReleases = (((endTS - startTS) / randomTimeStep) + 1) *
            m_randomParticlesPerTS;
        double[] txyzdArray = new double[totalReleases * 5];

        // Get the boundaries of the release area.
        int minI = (int)Math.round(m_randomBounds[0] / xSpacing);
        int minJ = (int)Math.round(m_randomBounds[1] / ySpacing);
        int minK = (int)Math.round(m_randomBounds[2] / zSpacing);
        int maxI = (int)Math.round(m_randomBounds[3] / xSpacing);
        int maxJ = (int)Math.round(m_randomBounds[4] / ySpacing);
        int maxK = (int)Math.round(m_randomBounds[5] / zSpacing);

        double[] depthChecks = new double[m_calc.m_numWet3Dpoints + 1];
        double[] likelihoods = new double[m_calc.m_numWet3Dpoints + 1];
        int[] likelyLookups = new int[m_calc.m_numWet3Dpoints];
        int[] surfaceIndexes = new int[m_calc.m_numWet2Dpoints];
        double[] wse = null;

        likelihoods[0] = 0;
        int numLikely = 0;
        int surfaceCounter = 0;
        int outputOffset = 0;

        if (tsWeightingSeries != null) {
            tsWeightingArray = tsWeightingSeries.getArray();
            if (tsWeightingArray.length <= (endTS - startTS))
                tsWeightingArray = null;
        }

        // Calculate and store the particle coordinates for each time step
        // that has velocity data.
        for (int ts = startTS; ts <= endTS; ts += randomTimeStep) {
            // Store the time for this step in the path.
            double time = ts * timeStepInc + si3dRunStartTime;

            if (probabilityArray != null) {
                likelihoods[0] = 0;
                numLikely = 0;
                surfaceCounter = 0;
            }

            if (surfaceCounter == 0 || probabilityArray != null) {
                // Look at each wet non-surface cell and fill in the likelihood of it
                // being chosen.  Also note the indexes of the surface cells.
                for (int i = 0; i < m_calc.m_numWet3Dpoints; i++) {
                    int currentI = indexArray3D[i * 3];
                    int currentJ = indexArray3D[i * 3 + 1];

                    // Don't proceed if outside the release area.
                    if (currentI < minI || currentI > maxI ||
                        currentJ < minJ || currentJ > maxJ) {

                        continue;
                    }

                    int currentK = indexArray3D[i * 3 + 2];

                    // If it is a surface cell, mark it for later and skip ahead.
                    if (currentK == 1) {
                        surfaceIndexes[surfaceCounter] = i;
                        surfaceCounter++;
                        continue;
                    }

                    // Adjust the likelihood if it is a bottom cell.
                    int ijIndex = currentI + (currentJ * numI);
                    int lookup2D = m_calc.m_2DLookupArray[ijIndex];
                    double depth = depthArray[lookup2D];
                    int numLayers = (int)(depth / zSpacing) + 1;
                    double likelihood = zSpacing;
                    if (currentK == numLayers)
                        likelihood = depth - ( (currentK - 1) * zSpacing);

                    // Apply the spatially varying probability map, if available
                    if (probabilityArray != null) {
                        int index = lookup2D + ((ts - startTS) * num2D);
                        if (index < probabilityArray.length) {
                            double probability = probabilityArray[index];
                            likelihood *= probability;
                        }
                    }

                    likelyLookups[numLikely] = i;
                    likelihoods[numLikely + 1] = likelihoods[numLikely] + likelihood;
                    depthChecks[numLikely] = likelihood;
                    numLikely++;
                }

                for (int i = (numLikely + surfaceCounter); i < (likelihoods.length - 1);
                     i++) {

                    likelihoods[i + 1] = Double.MAX_VALUE;
                }

                wse = new double[surfaceCounter];
            }

            // Determine the likelihood of the surface cells being chosen.
            for (int i = 0; i < surfaceCounter; i++) {
                // Adjust the likelihood if it is a bottom cell.
                int currentI = indexArray3D[surfaceIndexes[i] * 3];
                int currentJ = indexArray3D[surfaceIndexes[i] * 3 + 1];
                int currentK = indexArray3D[surfaceIndexes[i] * 3 + 2];
                int ijIndex = currentI + (currentJ * numI);
                int lookup2D = m_calc.m_2DLookupArray[ijIndex];
                double waterSurfaceElevation = m_calc.getVar2D(lookup2D
                    + (ts / tsMultiplier * num2D)) * m_calc.m_wseScale;
                wse[i] = waterSurfaceElevation;
                double depth = depthArray[lookup2D];
                int numLayers = (int)(depth / zSpacing) + 1;
                double likelihood = zSpacing;
                if (currentK == numLayers)
                    likelihood = depth - ((currentK - 1) * zSpacing);
                likelihood += waterSurfaceElevation;

                // Apply the spatially varying probability map, if available
                if (probabilityArray != null) {
                    int index = lookup2D + ((ts - startTS) * num2D);
                    if (index < probabilityArray.length) {
                        double probability = probabilityArray[index];
                        likelihood *= probability;
                    }
                }

                likelyLookups[numLikely + i] = surfaceIndexes[i];
                likelihoods[numLikely + i + 1] =
                    likelihoods[numLikely + i] + likelihood;
                depthChecks[numLikely + i] = likelihood;
            }

            if ((numLikely + surfaceCounter) == 0)
                return;

            double rangeMax = likelihoods[numLikely + surfaceCounter];

            // Release each of the particles for this time step.
            int numReleases = m_randomParticlesPerTS;
            if (tsWeightingArray != null) {
                // Calculate the remaining space in the output array.
                int numRelLeft = ((txyzdArray.length - outputOffset) / 5);

                if (ts == endTS) {
                    // Use up any remaining releases on the final time step.
                    numReleases = numRelLeft;
                }
                else {
                    // Make sure there is at least one release for each
                    // remaining time step.
                    int numTsLeft = (endTS - ts) / randomTimeStep;
                    int allowableReleases = numRelLeft - numTsLeft;
                    numReleases = (int)Math.round(
                        (double)numReleases * tsWeightingArray[ts - startTS]);
                    numReleases = Math.min(numReleases, allowableReleases);
                }
            }

            for (int i = 0; i < numReleases; i++) {
                // Get a random number.
                double randomNum = Math.random() * rangeMax;
                int selectedIndex =
                    Arrays.binarySearch(likelihoods, randomNum);
                if (selectedIndex < 0)
                    selectedIndex = -(selectedIndex + 1);
                selectedIndex--;
                int lookup = likelyLookups[selectedIndex];

                int currentI = indexArray3D[lookup * 3];
                int currentJ = indexArray3D[lookup * 3 + 1];
                int currentK = indexArray3D[lookup * 3 + 2];

                // Determine cell height.
                int ijIndex = currentI + (currentJ * numI);
                int lookup2D = m_calc.m_2DLookupArray[ijIndex];
                double waterSurfaceElevation = m_calc.getVar2D(lookup2D
                    + (ts / tsMultiplier * num2D)) * m_calc.m_wseScale;
                double depth = depthArray[lookup2D];
                int numLayers = (int)(depth / zSpacing) + 1;
                double cellHeight = zSpacing;
                if (currentK == numLayers)
                    cellHeight = depth - ((currentK - 1) * zSpacing);
                if (currentK == 1)
                    cellHeight += waterSurfaceElevation;

                double cellTop = 0;
                if (currentK == 1)
                    cellTop = wse[selectedIndex - numLikely];
                else
                    cellTop = (double)(currentK - 1) * -zSpacing;

                // Get a random X-Y location within the cell.
                double x = Math.random() * xSpacing;
                double y = Math.random() * ySpacing;
                double z = Math.random() * cellHeight;

                // Check if beyond the end of the array.
                if ((outputOffset + 5) > txyzdArray.length) {
                    break;
                }

                // Store the release point and time.
                txyzdArray[outputOffset] = time;
                txyzdArray[outputOffset + 1] = ((double)currentI - 0.5)
                    * xSpacing + x;
                txyzdArray[outputOffset + 2] = ((double)currentJ - 0.5)
                    * ySpacing + y;
                txyzdArray[outputOffset + 3] = cellTop - z;
                txyzdArray[outputOffset + 4] = m_randomTrackDuration;

                outputOffset += 5;
            }
        }

        // Put the random release array into a series and add it to this object.
        DoubleDataSeries txyzdSeries = new DoubleDataSeries(5);
        txyzdSeries.setArray(txyzdArray);
        m_txyzdInputContainer.add(txyzdSeries);
    }

//------------------------------------------------------------------------------
/**
 * Generates random release locations along a cross section to use as input.
 * Locations are distributed across wet cells within the cross section.
 */
    protected void generateXSectRandomReleaseInput(CrossSectionGrid csg) {

        csg.createCrossSectionGrid();

        // Get the output of the recreated cross section grid.
        IntDataSeries ikSeries = csg.getIKSeries();
        int[] ikArray = ikSeries.getArray();
        double[] xyzLocations = csg.getXyz();
        int numXyzPoints = xyzLocations.length / 3;
        double csgSpacingX = csg.getSpacingX();
        double csgSpacingZ = csg.getSpacingZ();

        // Make sure the Bathymetry and Si3dOutput objects are defined.
        if (m_bath == null || m_si3dOut == null) {
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                GrObject gro = (GrObject)iter.next();
                if (gro instanceof Bathymetry) m_bath = (Bathymetry)(gro);
                if (gro instanceof Si3dOutput) m_si3dOut = (Si3dOutput)gro;
            }
        }

        if (m_bath == null || m_si3dOut == null) return;

        // Get grid spacing information from the Bathymetry object.
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        double xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        double ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        double zSpacing = (Double)p.value;

        // Find the dimensions of the grid and the number of wet 2D points.
        DoubleDataSeries bathSeries = m_bath.getBathymetrySeries();
        int numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        int numJ = (int)Math.round(bathSeries.getMax(1)) + 1;
        int numLayers = (int)Math.round(bathSeries.getMax(2));
        double[] depthArray = m_bath.getAverageDepthArray();

        // Get array indexing information from the Bathymetry object.
        int[] indexArray3D = m_bath.get3DIndexArray();

        // Check for an input probability series.
        double[] probabilityArray = null;
        double[] tsWeightingArray = null;
        DoubleDataSeries probabilitySeries = getContainerSeries(
            "Random Probability Input Series");
        if (probabilitySeries != null)
            probabilityArray = probabilitySeries.getArray();
        DoubleDataSeries tsWeightingSeries = getContainerSeries(
            "Time Step Weighting Input Series");

        // Get the 2D versions.
        int[] indexArray2D = m_bath.get2DIndexArray();
        m_calc.m_2DLookupArray = m_bath.get2DLookupArray(indexArray2D);
        m_calc.m_3DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("3-D Variables");
        m_calc.m_2DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("2-D x-y Variables");
        if (m_calc.m_2DVarsSeries != null) {
//            m_calc.m_2DVarsArray = m_calc.m_2DVarsSeries.getArray();

            DataDimension dim = m_calc.m_2DVarsSeries.getDimension(0);
            if (dim != null) {
                // Store the scale factor for the dimension.
                GrHashList dimProps = dim.getProperties();
                p = (GrProperty)dimProps.get("unitScale");
                m_calc.m_wseScale = (Double)p.value;
            }
        }

        // Get array indexing information from the Bathymetry object.
        m_calc.m_3DIndexArray = m_bath.get3DIndexArray();
        m_calc.m_numWet3Dpoints = m_calc.m_3DIndexArray.length / 3;
        m_calc.m_numWet2Dpoints = bathSeries.getNumPoints();


        // 15 min = 0.010416667

        // Calculate the number of wet points in the I-J plane.
        int num2D = indexArray2D.length / 2;
        // Calculate the number of wet points in the I-J-K grid.
        int num3D = indexArray3D.length / 3;

        // Get time step information from the Si3dOutput object.
        double si3dRunStartTime = m_si3dOut.getJulianStartTime();
        if (m_randomStartTime < si3dRunStartTime)
            m_randomStartTime = si3dRunStartTime;
        GrHashList si3dOutProps = m_si3dOut.getProperties();
        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long)p.value;
        double timeStepInc = ((double)tsInSec) / 86400;

        int tsMultiplier = 1;
        if (m_randomTimeStep < 0) {
            tsMultiplier = -m_randomTimeStep;
        }
        timeStepInc /= tsMultiplier;
        int randomTimeStep = m_randomTimeStep;
        if (randomTimeStep < 0)
            randomTimeStep = 1;

        int startTS = (int)Math.round(
            (m_randomStartTime - si3dRunStartTime) / timeStepInc);
        int numTimeSteps = m_calc.m_3DVarsSeries.getCount() / num3D;
        if (numTimeSteps < 1)
            return;

        numTimeSteps *= tsMultiplier;
        if (startTS > numTimeSteps)
            startTS = numTimeSteps;

        int endTS = startTS +
            (int)Math.round(m_randomReleaseDuration / timeStepInc);

        if (endTS < startTS || endTS >= numTimeSteps)
            endTS = numTimeSteps - 1;

        // Determine how many releases there will be and allocate space
        // for them.
        int totalReleases = (((endTS - startTS) / randomTimeStep) + 1) *
            m_randomParticlesPerTS;
        double[] txyzdArray = new double[totalReleases * 5];

        // Get the boundaries of the release area.
        int minI = (int)Math.round(m_randomBounds[0] / xSpacing);
        int minJ = (int)Math.round(m_randomBounds[1] / ySpacing);
        int maxI = (int)Math.round(m_randomBounds[3] / xSpacing);
        int maxJ = (int)Math.round(m_randomBounds[4] / ySpacing);

        double[] depthChecks = new double[m_calc.m_numWet3Dpoints + 1];
        double[] likelihoods = new double[m_calc.m_numWet3Dpoints + 1];
        int[] likelyLookups = new int[m_calc.m_numWet3Dpoints];
        int[] surfaceIndexes = new int[m_calc.m_numWet2Dpoints];
        double[] wse = null;

        likelihoods[0] = 0;
        int numLikely = 0;
        int surfaceCounter = 0;
        int outputOffset = 0;

        if (tsWeightingSeries != null) {
            tsWeightingArray = tsWeightingSeries.getArray();
            if (tsWeightingArray.length <= (endTS - startTS))
                tsWeightingArray = null;
        }

        // Calculate and store the particle coordinates for each time step
        // that has velocity data.
        for (int ts = startTS; ts <= endTS; ts += randomTimeStep) {
            // Store the time for this step in the path.
            double time = ts * timeStepInc + si3dRunStartTime;

            if (probabilityArray != null) {
                likelihoods[0] = 0;
                numLikely = 0;
                surfaceCounter = 0;
            }

            if (surfaceCounter == 0 || probabilityArray != null) {
                // Look at each wet non-surface cell and fill in the likelihood of it
                // being chosen.  Also note the indexes of the surface cells.
                for (int i = 0; i < m_calc.m_numWet3Dpoints; i++) {
                    int currentI = indexArray3D[i * 3];
                    int currentJ = indexArray3D[i * 3 + 1];

                    // Don't proceed if outside the release area.
                    if (currentI < minI || currentI > maxI ||
                        currentJ < minJ || currentJ > maxJ) {

                        continue;
                    }

                    int currentK = indexArray3D[i * 3 + 2];

                    // If it is a surface cell, mark it for later and skip ahead.
                    if (currentK == 1) {
                        surfaceIndexes[surfaceCounter] = i;
                        surfaceCounter++;
                        continue;
                    }

                    // Adjust the likelihood if it is a bottom cell.
                    int ijIndex = currentI + (currentJ * numI);
                    int lookup2D = m_calc.m_2DLookupArray[ijIndex];
                    double depth = depthArray[lookup2D];
                    double likelihood = zSpacing;
                    if (currentK == numLayers)
                        likelihood = depth - ( (currentK - 1) * zSpacing);

                    // Apply the spatially varying probability map, if available
                    if (probabilityArray != null) {
                        int index = lookup2D + ((ts - startTS) * num2D);
                        if (index < probabilityArray.length) {
                            double probability = probabilityArray[index];
                            likelihood *= probability;
                        }
                    }

                    likelyLookups[numLikely] = i;
                    likelihoods[numLikely + 1] = likelihoods[numLikely] + likelihood;
                    depthChecks[numLikely] = likelihood;
                    numLikely++;
                }

                for (int i = (numLikely + surfaceCounter); i < (likelihoods.length - 1);
                     i++) {

                    likelihoods[i + 1] = Double.MAX_VALUE;
                }

                wse = new double[surfaceCounter];
            }

            // Determine the likelihood of the surface cells being chosen.
            for (int i = 0; i < surfaceCounter; i++) {
                // Adjust the likelihood if it is a bottom cell.
                int currentI = indexArray3D[surfaceIndexes[i] * 3];
                int currentJ = indexArray3D[surfaceIndexes[i] * 3 + 1];
                int currentK = indexArray3D[surfaceIndexes[i] * 3 + 2];
                int ijIndex = currentI + (currentJ * numI);
                int lookup2D = m_calc.m_2DLookupArray[ijIndex];
                double waterSurfaceElevation = m_calc.getVar2D(lookup2D
                    + (ts / tsMultiplier * num2D)) * m_calc.m_wseScale;
                wse[i] = waterSurfaceElevation;
                double depth = depthArray[lookup2D];
                double likelihood = zSpacing;
                if (currentK == numLayers)
                    likelihood = depth - ((currentK - 1) * zSpacing);
                likelihood += waterSurfaceElevation;

                // Apply the spatially varying probability map, if available
                if (probabilityArray != null) {
                    int index = lookup2D + ((ts - startTS) * num2D);
                    if (index < probabilityArray.length) {
                        double probability = probabilityArray[index];
                        likelihood *= probability;
                    }
                }

                likelyLookups[numLikely + i] = surfaceIndexes[i];
                likelihoods[numLikely + i + 1] =
                    likelihoods[numLikely + i] + likelihood;
                depthChecks[numLikely + i] = likelihood;
            }

            if ((numLikely + surfaceCounter) == 0)
                return;

            double rangeMax = likelihoods[numLikely + surfaceCounter];

            // Release each of the particles for this time step.
            int numReleases = m_randomParticlesPerTS;
            if (tsWeightingArray != null) {
                // Calculate the remaining space in the output array.
                int numRelLeft = ((txyzdArray.length - outputOffset) / 5);

                if (ts == endTS) {
                    // Use up any remaining releases on the final time step.
                    numReleases = numRelLeft;
                }
                else {
                    // Make sure there is at least one release for each
                    // remaining time step.
                    int numTsLeft = (endTS - ts) / randomTimeStep;
                    int allowableReleases = numRelLeft - numTsLeft;
                    numReleases = (int)Math.round(
                        (double)numReleases * tsWeightingArray[ts - startTS]);
                    numReleases = Math.min(numReleases, allowableReleases);
                }
            }

            for (int i = 0; i < numReleases; i++) {
                // Get a random number.
                double randomNum = Math.random() * rangeMax;
                int selectedIndex =
                    Arrays.binarySearch(likelihoods, randomNum);
                if (selectedIndex < 0)
                    selectedIndex = -(selectedIndex + 1);
                selectedIndex--;
                int lookup = likelyLookups[selectedIndex];

                int currentI = indexArray3D[lookup * 3];
                int currentJ = indexArray3D[lookup * 3 + 1];
                int currentK = indexArray3D[lookup * 3 + 2];

                double cellHeight = likelihoods[selectedIndex + 1] -
                    likelihoods[selectedIndex];
                double cellTop = 0;
                if (currentK == 1)
                    cellTop = wse[selectedIndex - numLikely];
                else
                    cellTop = (double)currentK * zSpacing;

                // Get a random X-Y location within the cell.
                double x = Math.random() * xSpacing;
                double y = Math.random() * ySpacing;
                double z = Math.random() * cellHeight;

                // Check if beyond the end of the array.
                if ((outputOffset + 5) > txyzdArray.length) {
                    break;
                }

                // Store the release point and time.
                txyzdArray[outputOffset] = time;
                txyzdArray[outputOffset + 1] = ((double)currentI - 0.5)
                    * xSpacing + x;
                txyzdArray[outputOffset + 2] = ((double)currentJ - 0.5)
                    * ySpacing + y;
                txyzdArray[outputOffset + 3] = cellTop - z;
                txyzdArray[outputOffset + 4] = m_randomTrackDuration;

                outputOffset += 5;
            }
        }

        // Put the random release array into a series and add it to this object.
        DoubleDataSeries txyzdSeries = new DoubleDataSeries(5);
        txyzdSeries.setArray(txyzdArray);
        m_txyzdInputContainer.add(txyzdSeries);
    }
//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        if (m_bAutoRecalc) {
            if (gro == m_xyInputContainer || gro == m_nzInputContainer ||
                gro == m_tnInputContainer || gro == m_ndInputContainer) {

                recalculate();
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * @return  A submenu of commands.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"Particle Tracker");
        javax.swing.JMenuItem item = new javax.swing.JMenuItem("Calculate");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Generate Random Input");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Create Transit Time Distribution");
        item.addActionListener(this);
        subMenu.add(item);

    	return subMenu;
    }
//------------------------------------------------------------------------------
/**
 * Returns the stats array.
 * @return  The stats array for the last set of tracks.
 */
    public double[] getTrackStats() {

        return m_trackStats;
    }

//------------------------------------------------------------------------------
/**
 * Handles action events, such as a recalculate command.
 * @param e The action event.
 */
    public void actionPerformed(java.awt.event.ActionEvent e) {

        // Check the source to see if it is a command.
        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();

        // Calculate command.
        if (srcText.equals("Calculate")) {
            // Note the start time.
            GregorianCalendar gc1 = new GregorianCalendar();
            long time1 = gc1.getTimeInMillis();

            // Recalculate.
        	recalculate();

            // Write out the elapsed time if it was greater than 10 seconds.
            GregorianCalendar gc2 = new GregorianCalendar();
            long time2 = gc2.getTimeInMillis();
            if ((time2 - time1) > 10000) {
                java.text.SimpleDateFormat formatter = new
                    java.text.SimpleDateFormat(
                    "MMM'.' dd',' yyyy 'at' hh:mm:ss");
                System.out.println("PTM Start: "
                    + formatter.format(gc1.getTime()));
                System.out.println("PTM End:   "
                    + formatter.format(gc2.getTime()));
            }
        }

        // Generate Random Input command.
        if (srcText.equals("Generate Random Input")) {
            generateRandomReleaseInput();
        }

        // Create Transit Time Distribution command.
        if (srcText.equals("Create Transit Time Distribution")) {
            createTransitTimeDistribution();
        }
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * If an error was detected during an operation, the class's error string
 * is appended.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Particle Tracker";

        if (m_errorString != null)
            s += m_errorString;

        return s;
 	}
}

