/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-11-05
 * Date started:    2003-04-11
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.util.*;

//==============================================================================
/**
 * Bathymetry for the SI3D model, including barriers and boundaries.
 */

public class RawBathymetry extends GrObject
    implements java.awt.event.ActionListener {
    
    protected DoubleDataSeries m_bathSeries = null;
    protected IntDataSeries m_dryCellSeries = null;
    protected IntDataSeries m_twbSeries = null;
    protected double m_xSpacing = 1;
    protected double m_ySpacing = 1;
    protected double m_zSpacing = 1;
    protected int m_numI = 0;
    protected int m_numJ = 0;
    protected double m_totalVolume = Double.NEGATIVE_INFINITY;
    protected IntDataSeries m_wetIndexSeries = null;
    protected String m_bathFileName = null;
    protected String m_barrFileName = null;
    /** A title describing the data object. */
    protected String m_title = "";
    protected double m_totalRawVolume = 0;
    protected double m_reachableVolume = 0;
    protected double[] m_rawVolumes = {0};
    protected double m_totalRawArea = 0;
    protected double m_reachableArea = 0;
    protected double[] m_rawAreas = {0};
    protected double m_inputWaterLevel = 0;
    protected double m_inputCutoffDepth = Double.POSITIVE_INFINITY;
    protected double m_volumeLandValue = 0;
    
//------------------------------------------------------------------------------
    /**
     * Creates an empty filter object.
     */
    public RawBathymetry() {
        
        setDefaults("gov.usgs.sfhydro.si3d.RawBathymetry");
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides add in GrObject.
     */
    public boolean add(GrObject gro) {
        
        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();
        
        if (m_children.contains(gro)) return false;
        
        boolean bAdded = false;
        
        if (gro instanceof DoubleDataSeries) {
            DoubleDataSeries series = (DoubleDataSeries)gro;
            if (series.getNumDimensions() == 1) {
                m_bathSeries = series;
                m_bathSeries.setProperty("editorClassName",
                    "gov.usgs.sfhydro.si3d.BathymetryEditorPanel");
                m_totalVolume = Double.NEGATIVE_INFINITY;
                m_wetIndexSeries = null;
            }
            
            boolean bSuccess = super.add(gro);
            
            return bSuccess;
        }
        
        if (gro instanceof IntDataSeries) {
            IntDataSeries series = (IntDataSeries)gro;
            if (series.getNumDimensions() == 1) {
                if (m_dryCellSeries != null)
                    remove(m_dryCellSeries);
                m_dryCellSeries = series;
                m_totalVolume = Double.NEGATIVE_INFINITY;
                m_wetIndexSeries = null;
            }
            if (series.getNumDimensions() == 2) {
                if (m_twbSeries != null)
                    remove(m_twbSeries);
                m_twbSeries = series;
            }
            
            
            boolean bSuccess = super.add(gro);
            
            return bSuccess;
        }
        if (gro instanceof GrObjectContainer) {
            ((GrObjectContainer)gro).setName("");
            
            return super.add(gro);
        }
        if (gro instanceof Location) {
            
            return super.add(gro);
        }
        
        return bAdded;
    }
    
//------------------------------------------------------------------------------
    /**
     * Wrapper for clearing the list of children.
     */
    public synchronized void clear() {
        
        m_bathSeries = null;
        m_dryCellSeries = null;
        m_totalVolume = Double.NEGATIVE_INFINITY;
        m_wetIndexSeries = null;
        m_twbSeries = null;
        
        super.clear();
    }
    
//------------------------------------------------------------------------------
    /**
     * Wrapper for removing items from the list of children.
     */
    public synchronized GrObject remove(int i) {
        
        GrObject gro = super.remove(i);
        
        if (gro == m_bathSeries) {
            m_bathSeries = null;
            m_totalVolume = Double.NEGATIVE_INFINITY;
            m_wetIndexSeries = null;
        }
        if (gro == m_dryCellSeries) {
            m_dryCellSeries = null;
            m_totalVolume = Double.NEGATIVE_INFINITY;
            m_wetIndexSeries = null;
        }
        if (gro == m_twbSeries)
            m_twbSeries = null;
        
        return gro;
    }
    
//------------------------------------------------------------------------------
    /**
     * Wrapper for removing items from the list of children.
     */
    public synchronized boolean remove(GrObject gro) {
        
        if (gro == m_bathSeries) {
            m_bathSeries = null;
            m_totalVolume = Double.NEGATIVE_INFINITY;
            m_wetIndexSeries = null;
        }
        if (gro == m_dryCellSeries) {
            m_dryCellSeries = null;
            m_totalVolume = Double.NEGATIVE_INFINITY;
            m_wetIndexSeries = null;
        }
        if (gro == m_twbSeries)
            m_twbSeries = null;
        
        return super.remove(gro);
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public GrHashList getProperties() {
        
        GrHashList props = super.getProperties();
        
        props.put("title", new GrProperty(m_title, "Title"));
        props.put("xSpacing", new GrProperty(m_xSpacing, "X Spacing"));
        props.put("ySpacing", new GrProperty(m_ySpacing, "Y Spacing"));
        props.put("zSpacing", new GrProperty(m_zSpacing, "Z Spacing"));
        props.put("numI", new GrProperty(m_numI, "Number of I Points"));
        props.put("numJ", new GrProperty(m_numJ, "Number of J Points"));
        props.put("waterLevel",
            new GrProperty(m_inputWaterLevel, "Water Level"));
        if (m_totalVolume == Double.NEGATIVE_INFINITY)
            m_totalVolume = calculateTotalVolume();
        props.put("totalArea", new GrProperty(
            getWetIndexSeries().getCount() * m_xSpacing * m_xSpacing,
            "Total Area", false, false));
        props.put("totalVolume", new GrProperty(
            m_totalVolume, "Total Volume", false, false));
        if (m_totalRawArea != 0)
            props.put("totalRawArea", new GrProperty(
                m_totalRawArea, "Total Raw Area", false, false));
        if (m_totalRawVolume != 0)
            props.put("totalRawVolume", new GrProperty(
                m_totalRawVolume, "Total Raw Volume", false, false));
        if (m_reachableArea != 0)
            props.put("reachableArea", new GrProperty(
                m_reachableArea, "Reachable Area", false, false));
        if (m_reachableVolume != 0)
            props.put("reachableVolume", new GrProperty(
                m_reachableVolume, "Reachable Volume", false, false));
        
        return props;
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public void setProperties(HashMap props) {
        
        super.setProperties(props);
        
        GrProperty p;
        
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;
                
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                
                if (key.equals("xSpacing")) {
                    m_xSpacing = (Double)p.value;
                    m_totalVolume = Double.NEGATIVE_INFINITY;
                    m_wetIndexSeries = null;
                }
                if (key.equals("ySpacing"))
                    m_ySpacing = (Double)p.value;
                if (key.equals("zSpacing"))
                    m_zSpacing = (Double)p.value;
                if (key.equals("numI"))
                    m_numI = (Integer)p.value;
                if (key.equals("numJ"))
                    m_numJ = (Integer)p.value;
                if (key.equals("waterLevel"))
                    m_inputWaterLevel = (Double)p.value;
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Set a single property.
     */
    public void setProperty(String key, String value) {
        
        super.setProperty(key, value);
        
        if (key.equals("title")) m_title = value;
        if (key.equals("xSpacing")) {
            m_xSpacing = Double.parseDouble(value);
            m_totalVolume = Double.NEGATIVE_INFINITY;
            m_wetIndexSeries = null;
        }
        if (key.equals("ySpacing"))
            m_ySpacing = Double.parseDouble(value);
        if (key.equals("zSpacing"))
            m_zSpacing = Double.parseDouble(value);
        if (key.equals("numI"))
            m_numI = Integer.parseInt(value);
        if (key.equals("numJ"))
            m_numJ = Integer.parseInt(value);
        if (key.equals("waterLevel"))
            m_inputWaterLevel = Double.parseDouble(value);
    }
    
//------------------------------------------------------------------------------
    /**
     * .
     */
    public DoubleDataSeries getBathymetrySeries() {
        
        return m_bathSeries;
    }
    
//------------------------------------------------------------------------------
    /**
     * .
     */
    public IntDataSeries getDryCellSeries() {
        
        return m_dryCellSeries;
    }
    
//------------------------------------------------------------------------------
    /**
     * .
     */
    public IntDataSeries getBarrierSeries() {
        
        return m_twbSeries;
    }
    
//------------------------------------------------------------------------------
    /**
     * Returns an array of wet points, as interleaved I and J vertices.
     * If the array has not been allocated, it creates it and stores it for future
     * calls.  The array is invalidated whenever the bathymetry series is updated.
     */
    public IntDataSeries getWetIndexSeries() {
        
        if (m_bathSeries == null)
            return null;
        
        if (m_wetIndexSeries == null) {
            double[] depthArray = m_bathSeries.getArray();
            int numWet = 0;
            int dryStart = 0;
            
            // First pass counts the number of wet points so the array can
            // be allocated.
            for (int i = 0; i < (m_numI - 1); i++) {
                for (int j = 0; j < (m_numJ - 1); j++) {
                    double blZ = depthArray[j * m_numI + i];
                    double brZ = depthArray[j * m_numI + i + 1];
                    double trZ = depthArray[(j + 1) * m_numI + i + 1];
                    double tlZ = depthArray[(j + 1) * m_numI + i];
                    
                    if (blZ > 0 && brZ > 0 && trZ > 0 && tlZ > 0) {
                        numWet++;
                    }
                }
            }
            
            int numDry = 0;
            int[] dryCellArray = null;
            if (m_dryCellSeries != null) {
                dryCellArray = m_dryCellSeries.getArray();
                numDry = dryCellArray.length;
            }
            int currentIndex = 0;
            
            int[] wetIndexes = new int[numWet - numDry];
            
            for (int i = 0; i < (m_numI - 1); i++) {
                for (int j = 0; j < (m_numJ - 1); j++) {
                    double blZ = depthArray[j * m_numI + i];
                    double brZ = depthArray[j * m_numI + i + 1];
                    double trZ = depthArray[(j + 1) * m_numI + i + 1];
                    double tlZ = depthArray[(j + 1) * m_numI + i];
                    
                    if (blZ > 0 && brZ > 0 && trZ > 0 && tlZ > 0) {
                        boolean bDry = false;
                        if (dryCellArray != null) {
                            int comboIJ = i * 10000 + j;
                            int index = m_dryCellSeries.search(comboIJ, 0);
                            if (index >= 0 && dryCellArray[index] == comboIJ)
                                continue;
                        }
                        
                        if (!bDry) {
                            if (currentIndex < wetIndexes.length) {
                                int comboIJ = i * 10000 + j;
                                wetIndexes[currentIndex] = comboIJ;
                                currentIndex ++;
                            }
                        }
                    }
                }
            }
            
            m_wetIndexSeries = new IntDataSeries(wetIndexes, 1);
        }
        
        return m_wetIndexSeries;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public boolean isWet(int i, int j) {
        
        if (m_wetIndexSeries == null)
            getWetIndexSeries();
        int comboIJ = i * 10000 + j;
        int index = m_wetIndexSeries.search(comboIJ);
        int[] array = m_wetIndexSeries.getArray();
        if (array[index] == comboIJ)
            return true;
        else
            return false;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public double getDepthAt(int i, int j) {
        
        DoubleDataSeries series = getBathymetrySeries();
        double[] depthArray = series.getArray();
        int index = j * m_numI + i;
        double depth = Double.NaN;
        if (i >= 0 && i < m_numI && j >= 0 && j < m_numJ &&
            index >= 0 && index < depthArray.length) {
            
            depth = depthArray[index];
        }
        
        return depth;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public double calculateTotalVolume() {
        
        double totalVolume = 0;
        
        try {
            // Get data and properties from the bathymetry object.
            DoubleDataSeries series = getBathymetrySeries();
            double[] depthArray = series.getArray();
            
            double cellArea = m_xSpacing * m_xSpacing;
            
            int[] wetIndexes = getWetIndexSeries().getArray();
            for (int n = 0; n < wetIndexes.length; n++) {
                int i = wetIndexes[n] / 10000;
                int j = wetIndexes[n] - (i * 10000);
                double blZ = depthArray[j * m_numI + i];
                double brZ = depthArray[j * m_numI + i + 1];
                double trZ = depthArray[(j + 1) * m_numI + i + 1];
                double tlZ = depthArray[(j + 1) * m_numI + i];
                
                double avgDepth = (blZ + brZ + trZ + tlZ) / 4;
                totalVolume += (avgDepth * cellArea);
            }
        } catch (Exception e) {}
        
        return totalVolume;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public void calculateTotalRawVolume() {
        
        m_totalRawVolume = 0;
        m_totalRawArea = 0;
        m_rawVolumes = new double[0];
        m_rawAreas = new double[0];
        m_volumeLandValue =
            Math.max(0, m_inputWaterLevel) - (m_xSpacing / 10) - 2;
        
        try {
            // Get data and properties from the bathymetry object.
            DoubleDataSeries series = getBathymetrySeries();
            double[] depthArray = series.getArray();
            
            double cellArea = m_xSpacing * m_xSpacing;
            
            int[] dryCellArray = null;
            if (m_dryCellSeries != null) {
                dryCellArray = m_dryCellSeries.getArray();
            }
            int currentIndex = 0;
            
            m_rawVolumes = new double[(m_numI - 1) * (m_numJ - 1)];
            m_rawAreas = new double[(m_numI - 1) * (m_numJ - 1)];
            for (int i = 0; i < (m_numI - 1); i++) {
                for (int j = 0; j < (m_numJ - 1); j++) {
                    double blZ = depthArray[j * m_numI + i];
                    double brZ = depthArray[j * m_numI + i + 1];
                    double trZ = depthArray[(j + 1) * m_numI + i + 1];
                    double tlZ = depthArray[(j + 1) * m_numI + i];
                    
                    // Clamp the land value.
                    if (blZ < m_volumeLandValue)
                        blZ = m_volumeLandValue;
                    if (brZ < m_volumeLandValue)
                        brZ = m_volumeLandValue;
                    if (trZ < m_volumeLandValue)
                        trZ = m_volumeLandValue;
                    if (tlZ < m_volumeLandValue)
                        tlZ = m_volumeLandValue;
                    
                    // Clamp to the cutoff depth.
                    if (blZ > m_inputCutoffDepth)
                        blZ = m_inputCutoffDepth;
                    if (brZ > m_inputCutoffDepth)
                        brZ = m_inputCutoffDepth;
                    if (trZ > m_inputCutoffDepth)
                        trZ = m_inputCutoffDepth;
                    if (tlZ > m_inputCutoffDepth)
                        tlZ = m_inputCutoffDepth;
                    
                    // Shift the water level.
                    blZ += m_inputWaterLevel;
                    brZ += m_inputWaterLevel;
                    trZ += m_inputWaterLevel;
                    tlZ += m_inputWaterLevel;
                    
                    if (blZ > 0 || brZ > 0 || trZ > 0 || tlZ > 0) {
                        boolean bDry = false;
                        if (dryCellArray != null) {
                            int comboIJ = i * 10000 + j;
                            int index = m_dryCellSeries.search(comboIJ, 0);
                            if (index >= 0 && dryCellArray[index] == comboIJ)
                                continue;
                        }
                        
                        // Clamp the land value.
                        if (blZ < m_volumeLandValue)
                            blZ = m_volumeLandValue;
                        if (brZ < m_volumeLandValue)
                            brZ = m_volumeLandValue;
                        if (trZ < m_volumeLandValue)
                            trZ = m_volumeLandValue;
                        if (tlZ < m_volumeLandValue)
                            tlZ = m_volumeLandValue;
                                      
                        double avgDepth = 0;
                        double numWetCorners = 0;
                        if (blZ > 0) {
                            numWetCorners++;
                            avgDepth += blZ;
                        }
                        if (brZ > 0) {
                            numWetCorners++;
                            avgDepth += brZ;
                        }
                        if (trZ > 0) {
                            numWetCorners++;
                            avgDepth += trZ;
                        }
                        if (tlZ > 0) {
                            numWetCorners++;
                            avgDepth += tlZ;
                        }
                        
                        avgDepth /= 4;
                        
                        double wetArea = 0;
                        if (numWetCorners == 1)
                            wetArea = 0.125 * cellArea;
                        if (numWetCorners == 2)
                            wetArea = 0.5 * cellArea;
                        if (numWetCorners == 3)
                            wetArea = 0.875 * cellArea;
                        if (numWetCorners == 4)
                            wetArea = cellArea;
                            
                        double volume = avgDepth * cellArea;
                       
                        m_rawVolumes[j * (m_numI - 1) + i] = volume;
                        m_totalRawVolume += volume;
                        m_rawAreas[j * (m_numI - 1) + i] = wetArea;
                        m_totalRawArea += cellArea;
                    }
                }
            }
        } catch (Exception e) {}
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public void calculateReachableVolume(int startI, int startJ,
        double waterLevel, double cutoffDepth) {
        
        // Declare local variables.
        double totalVolume = 0;
        double reachableVolume = 0;
        DoubleDataSeries series = getBathymetrySeries();
       
        // If there is a cutoff depth, and it is above the deepest point,
        // calculate the volume below that depth.
        if (cutoffDepth < series.getMax(0) && cutoffDepth > -waterLevel)
            m_inputCutoffDepth = cutoffDepth;
        else
            m_inputCutoffDepth = Double.POSITIVE_INFINITY;
/*            // Store the input selected points, so they can be restored after
            // the lower volume is calculated.
            int[] selectedIndexes = series.getSelectedIndexes();
            int numSelected = series.getNumSelected();
            int[] boundaryIndexes = new int[numSelected];
            for (int i = 0; i < numSelected; i++) {
                boundaryIndexes[i] = selectedIndexes[i];
            }

            // Calculate the volume below the cutoff depth.
            m_inputWaterLevel = -cutoffDepth;
            calculateReachableVolume(startI, startJ);
            
            // Store the volume in temp variables.
            totalVolume = m_totalRawVolume;
            reachableVolume = m_reachableVolume;
        
            // Restore the original input selection.
            series.deselectAll();
            for (int i = 0; i < boundaryIndexes.length; i++) {
                series.selectPoint(boundaryIndexes[i]);
            }
        }
*/        
        // Calculate the volume below the water level.
        m_inputWaterLevel = waterLevel;
        calculateReachableVolume(startI, startJ);
        
        // If there was a cutoff volume, subtract it from the volume below
        // the water level.
        if (cutoffDepth < series.getMax(0)) {
            m_totalRawVolume -= totalVolume;
            m_reachableVolume -= reachableVolume;
        }
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public void calculateReachableVolume(int startI, int startJ) {
        
        // Initialize variables and arrays.
        m_reachableVolume = 0;
        m_reachableArea = 0;
        DoubleDataSeries series = getBathymetrySeries();
        double[] depthArray = series.getArray();
        calculateTotalRawVolume();
        int[] selectedIndexes = series.getSelectedIndexes();
        int numSelected = series.getNumSelected();
        boolean[] touchedIndexes = new boolean[selectedIndexes.length];
        Arrays.fill(touchedIndexes, false);
        for (int i = 0; i < numSelected; i++) {
            touchedIndexes[selectedIndexes[i]] = true;
        }
        int[] traceIndexes = new int[m_numI * m_numJ];
        int numSteps = 0;
        int startIndex = 0;
        boolean bMovingForward = true;
        
        // Use the first selected point as the start.
        if (numSelected > 0)
            startIndex = selectedIndexes[numSelected - 1];
        
        // Store the coordinates of the neighbor points that will be
        // tested in order from east counter-clockwise to southeast.
        // This order should minimize the frequency of needing to
        // shift values in the selectedIndexes array to maintain
        // ascending order.
        int[] testI = {1, 1, 0, -1, -1, -1, 0, 1};
        int[] testJ = {0, 1, 1, 1, 0, -1, -1, -1};
        
        // Convert the start index to I and J.
//            int startJ = startIndex / m_numI;
//            int startI = startIndex - (startJ * m_numI);
        
        // The start point is the only point in the trace.
        traceIndexes[0] = startJ * m_numI + startI;
        
        // Keeping looking until the trace has been rolled back past the
        // start point.
        while (numSteps >= 0) {
            // The current point is stored at the end of the trace.
            int currentIndex = traceIndexes[numSteps];
            
            // Convert the current index to I and J.
            int currentJ = currentIndex / m_numI;
            int currentI = currentIndex - (currentJ * m_numI);
            
            // The first time a point is encountered, add the four cells
            // around it to the volume total.
            if (bMovingForward) {
                if (currentI < (m_numI - 1) && currentJ < (m_numJ - 1)) {
                    // Add the volume to the total, then set vol to zero.
                    int lookupIndex = currentJ * (m_numI - 1) + currentI;
                    m_reachableVolume += m_rawVolumes[lookupIndex];
                    m_rawVolumes[lookupIndex] = 0;
                    m_reachableArea += m_rawAreas[lookupIndex];
                    m_rawAreas[lookupIndex] = 0;
                }
                // Check if SE cell is in bounds.
                if (currentI < (m_numI - 1) && currentJ > 0) {
                    // Add the volume to the total, then set vol to zero.
                    int lookupIndex =
                        (currentJ - 1) * (m_numI - 1) + currentI;
                    m_reachableVolume += m_rawVolumes[lookupIndex];
                    m_rawVolumes[lookupIndex] = 0;
                    m_reachableArea += m_rawAreas[lookupIndex];
                    m_rawAreas[lookupIndex] = 0;
                }
                // Check if SW cell is in bounds.
                if (currentI > 0 && currentJ > 0) {
                    // Add the volume to the total, then set vol to zero.
                    int lookupIndex =
                        (currentJ - 1) * (m_numI - 1) + (currentI - 1);
                    m_reachableVolume += m_rawVolumes[lookupIndex];
                    m_rawVolumes[lookupIndex] = 0;
                    m_reachableArea += m_rawAreas[lookupIndex];
                    m_rawAreas[lookupIndex] = 0;
                }
                // Check if NW cell is in bounds.
                if (currentI > 0 && currentJ < (m_numJ - 1)) {
                    // Add the volume to the total, then set vol to zero.
                    int lookupIndex =
                        currentJ * (m_numI - 1) + (currentI - 1);
                    m_reachableVolume += m_rawVolumes[lookupIndex];
                    m_rawVolumes[lookupIndex] = 0;
                    m_reachableArea += m_rawAreas[lookupIndex];
                    m_rawAreas[lookupIndex] = 0;
                }
            }
            
            // Initialize bMovingForward to false and set it to true at the
            // first neighbor that is wet and has not already been visited.
            bMovingForward = false;
            int nextIndex = -1;
            for (int i = 0; i < 8; i++) {
                int nextI = currentI + testI[i];
                int nextJ = currentJ + testJ[i];
                if (nextI < 0 || nextI >= m_numI ||
                    nextJ < 0 || nextJ >= m_numJ) {
                    
                    continue;
                }
                
                nextIndex = nextJ * m_numI + nextI;
                if (depthArray[nextIndex] > -m_inputWaterLevel &&
                    !touchedIndexes[nextIndex]) {
                    
                    bMovingForward = true;
                    break;
                }
                
            }
            
            // If a wet, unvisited neighbor was found, make it the point
            // to visit on the next iteration.
            if (bMovingForward) {
                // Mark the point as touched
                touchedIndexes[nextIndex] = true;
                
                // Move the trace one step forward.
                numSteps++;
                traceIndexes[numSteps] = nextIndex;
            } else {
                // Step backward to the previous point in the trace.
                numSteps--;
            }
        }
        
        // Select all the touched points, excluding the boundaries,
        // so they can be seen.
        for (int i = 0; i < numSelected; i++) {
            touchedIndexes[selectedIndexes[i]] = false;
        }
        series.deselectAll();
        for (int i = 0; i < touchedIndexes.length; i++) {
            if (touchedIndexes[i])
                series.selectPoint(i);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Overrides default update() in GrObject.  Called by an observed GrObject's
     * notifyObservers() method to handle changes to the observed object.
     * May use the notifyObservers() method to propagate the event handling up to
     * this object's observers.
     * @param gro   The observed object that changed.
     * @param arg   Additional information that the updated object may choose to
     *              send.
     */
    public void update(GrObject gro, Object arg) {
        
        if (gro == m_bathSeries || gro == m_dryCellSeries) {
            m_wetIndexSeries = null;
            m_totalVolume = Double.NEGATIVE_INFINITY;
            super.update(gro, arg);
        }
        if (gro == m_twbSeries) {
            super.update(gro, arg);
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Provide the list of menu items that execute commands on this object.
     * @return  A menu item that can be placed on a menu and send an ActionEvent
     *          back to this object when it is chosen.
     */
    public javax.swing.JMenuItem getMenuCommands() {
        
        javax.swing.JMenu subMenu = new javax.swing.JMenu(
            "Raw Bathymetry");
        javax.swing.JMenuItem item = new javax.swing.JMenuItem(
            "Save Bathymetry...");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem(
            "Save Barriers and Dry Cells...");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem(
            "Calculate Reachable Volume...");
        item.addActionListener(this);
        subMenu.add(item);
        
        return subMenu;
    }
    
//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {
        
        javax.swing.JMenuItem source =
            (javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Save Bathymetry...")) {
            gov.usgs.sfhydro.data.Si3dBathymetryFile bf =
                new gov.usgs.sfhydro.data.Si3dBathymetryFile();
            bf.add(this);
            if (m_bathFileName == null)
                m_bathFileName = System.getProperty("gov.usgs.gr.currentfile");
            bf.setProperty("file", m_bathFileName);
            bf.save(true);
            GrHashList props = bf.getProperties();
            GrProperty p = (GrProperty)props.get("file");
            m_bathFileName = (String)p.value;
        }
        if (srcText.equals("Save Barriers and Dry Cells...")) {
            gov.usgs.sfhydro.data.Si3dBarrierFile bf =
                new gov.usgs.sfhydro.data.Si3dBarrierFile();
            bf.add(this);
            if (m_barrFileName == null)
                m_barrFileName = System.getProperty("gov.usgs.gr.currentfile");
            bf.setProperty("file", m_barrFileName);
            bf.save(true);
            GrHashList props = bf.getProperties();
            GrProperty p = (GrProperty)props.get("file");
            m_barrFileName = (String)p.value;
        }
        if (srcText.equals("Save Barriers and Dry Cells...")) {
            gov.usgs.sfhydro.data.Si3dBarrierFile bf =
                new gov.usgs.sfhydro.data.Si3dBarrierFile();
            bf.add(this);
            if (m_barrFileName == null)
                m_barrFileName = System.getProperty("gov.usgs.gr.currentfile");
            bf.setProperty("file", m_barrFileName);
            bf.save(true);
            GrHashList props = bf.getProperties();
            GrProperty p = (GrProperty)props.get("file");
            m_barrFileName = (String)p.value;
        }
        if (srcText.equals("Calculate Reachable Volume...")) {
            calculateReachableVolume(0, 0);
        }
        
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {
        
        String s = "Si3D Raw Bathymetry";
        if (m_title.length() > 0)
            s += (": " + m_title);
        
        return s;
    }
}

