/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2002-04-25
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import java.text.*;
import java.util.*;

//==============================================================================
/**
 * Bathymetry for the SI3D model, including barriers and boundaries.
 */

public class Si3dOutput extends GrObject {

    protected String m_runTitle = "";
    protected String m_runNumber = "";
    protected GregorianCalendar m_startTime = null;
    protected long m_timeStepInSeconds = 0;

//------------------------------------------------------------------------------
/**
 * Creates an empty filter object.
 */
    public Si3dOutput () {

        setDefaults("gov.usgs.sfhydro.si3d.Si3dOutput");
    }

//------------------------------------------------------------------------------
/**
 * Overrides add in GrObject.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

		boolean bAdded = false;

        if (gro instanceof GrObjectContainer) {
            ((GrObjectContainer)gro).setName("");

            return super.add(gro);
        }

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("runTitle", new GrProperty(
        	m_runTitle, "Run Title"));
        props.put("runNumber", new GrProperty(
        	m_runNumber, "Run Number"));

        SimpleDateFormat formatter =
            new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
        formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
        StringBuffer buf = formatter.format(m_startTime.getTime(),
            new StringBuffer(), new FieldPosition(0));
        props.put("startTime", new GrProperty(
        	"" + buf.toString(), "Start Time"));
        props.put("timeStepInSeconds", new GrProperty(
            m_timeStepInSeconds, "Time Step in Seconds"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
				p = (GrProperty)obj;

                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }

            	if (key.equals("startTime")) {
                    SimpleDateFormat formatter =
                        new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
                    formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
                    m_startTime = new GregorianCalendar(
                        TimeZone.getTimeZone("UTC"));
                    m_startTime.clear();
                    ParsePosition pos = new ParsePosition(0);
                    m_startTime.setTime(formatter.parse((String)p.value, pos));
                }
            	if (key.equals("timeStepInSeconds"))
            		m_timeStepInSeconds = (Long)p.value;
            }
            if (obj instanceof String) {
				String value = (String)obj;
                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("runTitle"))
            m_runTitle = value;
        if (key.equals("runNumber"))
            m_runNumber = value;
        if (key.equals("startTime")) {
            SimpleDateFormat formatter =
                new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
            formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
            m_startTime = new GregorianCalendar(
                TimeZone.getTimeZone("UTC"));
            m_startTime.clear();
            ParsePosition pos = new ParsePosition(0);
            m_startTime.setTime(formatter.parse(value, pos));
        }
        if (key.equals("timeStepInSeconds"))
            m_timeStepInSeconds = Long.parseLong(value);
    }

//------------------------------------------------------------------------------
/**
 * Return the start day and time as a Julian decimal.
 */
    public double getJulianStartTime() {

        if (m_startTime == null) return 0;

        double jDayTime = TimeDataFile.julianDecimal(m_startTime,
            m_startTime.get(Calendar.YEAR));

        return jDayTime;
    }

//------------------------------------------------------------------------------
/**
 * Return the start year.
 */
    public int getStartYear() {

        if (m_startTime == null) return 0;

        return m_startTime.get(Calendar.YEAR);
    }

//------------------------------------------------------------------------------
/**
 * Return the DataSeries within the specified ObjectContainer within this
 * object.
 */
    public DataSeries getContainerSeries(String containerTitle) {

        // Look through the children of this object for an ObjectContainer
        // with a matching title.
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof GrObjectContainer) {
                GrObjectContainer oc = (GrObjectContainer)gro;
                GrHashList props = oc.getProperties();
                GrProperty p = (GrProperty)props.get("title");
                if (p.value.equals(containerTitle)) {
                    // Look through the contents of the ObjectContainer for
                    // a DataSeries.
                    Enumeration enumeration = oc.children();
                    while (enumeration.hasMoreElements()) {
                        GrObject oc_child = (GrObject)enumeration.nextElement();
                        if (oc_child instanceof DataSeries)
                            return (DataSeries)oc_child;
                    }
                }
            }
        }

        // Could not find a variables series under a container with the
        // given title.
        return null;
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Si3D Output";

        return s;
 	}
}

