/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-03
 * Date started:    2004-08-24
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.*;
import gov.usgs.gr.visual.gl.*;
import gov.usgs.sfhydro.si3d.gl.*;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.*;
import java.awt.geom.*;
import java.io.*;
import java.util.*;
import javax.swing.*;

//==============================================================================
/**
 * An instance of this class can be placed as a child of a PageOfGraphs object
 * to provide an interface to the user for viewing or processing Si3D model
 * output.  Causes a panel to be shown on the toolbar.
 */
public class Si3dPostProcessor extends gov.usgs.gr.GrObject
    implements gov.usgs.gr.visual.gui.GuiPanelProvider {

    /** An instance of the controller associated with this object. */
    GrController m_grCon = null;
    /** An instance of the page associated with this object. */
    PageOfGraphs m_page = null;
    /** The panel created by this class for displaying on a toolbar. */
    protected JPanel m_panel = new JPanel();
    /** The configuration dialog shown to the user. */
    protected JDialog m_dialog = null;
    /** The working directory. */
    protected File m_workingDir = new File("/");
    /** The label that shows the user the chosen working directory. */
    protected JLabel m_workingDirLabel = new JLabel("/");
    /**
     * The label that shows whether the spacefile was found in the chosen
     * working directory.
     */
    protected JLabel m_outputFoundLabel = new JLabel(
        "Si3D output not found in spacefile.xml");
    /**
     * The label that shows whether the bathymetry was found in the chosen
     * working directory.
     */
    protected JLabel m_bathFoundLabel = new JLabel(
        "Si3D bathymetry not found in si3d_bathy.xml");
    /** A list of objects that contain si3d output. */
    protected List m_parentObjectList = new ArrayList();
    protected String m_startFileName = "";
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/Si3dPostProcessor.gif";

//------------------------------------------------------------------------------
/**
 * Constructor for creating a default object.
 */
    public Si3dPostProcessor() {

        m_startFileName = System.getProperty("gov.usgs.gr.currentfile");
        // Create the editing panel once so it can be passed out later
        // as needed.
        createPanel();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * workingDir    String  "Working Directory"                           <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

        // First add properties from the super class.
        GrHashList props = super.getProperties();

        props.put("workingDir", new GrProperty(m_workingDir.getAbsolutePath(),
            "Working Directory"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.sfhydro.si3d.Si3dPostProcessor#getProperties getProperties()
 * for a list of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

        GrProperty p;

        // For each property with a string value, pass it to the setProperty()
        // method to be handled there.
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                // Handle string properties by passing them to the
                // setProperty() method.
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
            }
            // Handle string properties by passing them to the
            // setProperty() method.
            if (obj instanceof String) {
                setProperty(key, (String)obj);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.sfhydro.si3d.Si3dPostProcessor#getProperties getProperties()
 * for a list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

        if (key.equals("workingDir")) {
            m_workingDir = new File(value);
            m_workingDirLabel.setText(value);
        }
    }

//------------------------------------------------------------------------------
/**
 * Adds an object as a child, but only if it is an instance of Graph, an
 * IntDataSeries with two or three dimensions, or a DataSeries with three
 * dimensions.  Removes the old child, if a matching one can be found, and
 * replaces it with the new one.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        if (m_children.contains(gro)) return false;

        boolean bAdded = false;

        return bAdded;
    }

//------------------------------------------------------------------------------
/**
 * Create the editor panel that will be displayed on the tool bar.
 */
    public void createPanel() {

        // Set up the panel's border and layout.
        m_panel.setBorder(BorderFactory.createTitledBorder("Si3d"));
        m_panel.setLayout(new GridLayout(1, 1));

        // Create an "Add Points" button.
        // Add it to the main panel below the other components.
        JButton setupButton = new JButton("Set up...");
        // Define a method to handle the Add button being clicked.
        java.awt.event.ActionListener setupAL =
            new java.awt.event.ActionListener() {

            public void actionPerformed(ActionEvent e) {
                if (m_dialog == null) {
                    m_dialog = new JDialog(
                        GrController.getParentFrame(m_panel));
                    createDialog();
                }

                javax.swing.SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        m_dialog.setVisible(true);
                    }
                });
            }
        };
        setupButton.addActionListener(setupAL);
        setupButton.setActionCommand("Set up...");
        setupButton.setToolTipText("Set up...");
        byte[] byteArray = {0};
        setupButton.setIcon(new ImageIcon(byteArray));
        setupButton.setBorderPainted(true);
        setupButton.setMargin(new Insets(0, 0, 0, 0));
        m_panel.add(setupButton);

        m_panel.setMaximumSize(new java.awt.Dimension(400, 400));

    }

//------------------------------------------------------------------------------
/**
 * Create the editor panel that will be displayed on the tool bar.
 */
    public void createDialog() {

        // Define a method to handle the Add button being clicked.
        java.awt.event.ActionListener currentAL =
            new java.awt.event.ActionListener() {

            public void actionPerformed(ActionEvent e) {
                ((Component)GrController.getParentFrame(m_dialog))
                    .setCursor(new Cursor(Cursor.WAIT_CURSOR));

                m_workingDir = new File(m_startFileName);
                if (!m_workingDir.isDirectory()) {
                    m_workingDir = m_workingDir.getParentFile();
                }
                changeDataSource();

                ((Component)GrController.getParentFrame(m_dialog))
                    .setCursor(Cursor.getDefaultCursor());
            }
        };

        // Define a method to handle the Add button being clicked.
        java.awt.event.ActionListener browseAL =
            new java.awt.event.ActionListener() {

            public void actionPerformed(ActionEvent e) {
                ((Component)GrController.getParentFrame(m_dialog))
                    .setCursor(new Cursor(Cursor.WAIT_CURSOR));
                // Set up the file chooser dialog.
                JFileChooser fc = new JFileChooser();
                fc.setDialogTitle("Choose working directory");
                fc.setCurrentDirectory(m_workingDir);
                fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
                int returnVal = fc.showOpenDialog(m_dialog);
                if (returnVal == JFileChooser.APPROVE_OPTION) {
                    m_workingDir = fc.getSelectedFile();
                    changeDataSource();
                }
                ((Component)GrController.getParentFrame(m_dialog))
                    .setCursor(Cursor.getDefaultCursor());
            }
        };

        m_dialog.setTitle("Si3d Post Processor");
        m_dialog.setModal(true);
        m_dialog.setResizable(true);
        m_dialog.getRootPane().setMinimumSize(new Dimension(500, 250));

        // Create a panel to show the path title.
        JPanel workingDirPanel = new JPanel();
        workingDirPanel.setBorder(BorderFactory.createTitledBorder(
            "Working Directory"));
        workingDirPanel.setLayout(new BoxLayout(workingDirPanel,
            BoxLayout.Y_AXIS));
        workingDirPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
        m_workingDirLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
        workingDirPanel.add(m_workingDirLabel);
        JButton currentButton = new JButton("Use Current");
        currentButton.addActionListener(currentAL);
        currentButton.setToolTipText("Use Current Directory as Data Source");

        JButton browseButton = new JButton("Browse...");
        browseButton.addActionListener(browseAL);
        browseButton.setToolTipText("Browse...");

        JPanel dirButtonPanel = new JPanel();
        dirButtonPanel.setLayout(new BoxLayout(dirButtonPanel,
            BoxLayout.X_AXIS));
        dirButtonPanel.add(currentButton);
        dirButtonPanel.add(Box.createRigidArea(new Dimension(5,0)));
        dirButtonPanel.add(browseButton);

        dirButtonPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
        workingDirPanel.add(dirButtonPanel);
        workingDirPanel.add(Box.createRigidArea(new Dimension(0,5)));
        m_outputFoundLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
        workingDirPanel.add(m_outputFoundLabel);
        m_bathFoundLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
        workingDirPanel.add(m_bathFoundLabel);

        // Add the OK button to the bottom.
        JButton closeButton = new JButton("Close");
        closeButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(ActionEvent e) {
                m_dialog.dispose();
            }
        });

        // Add the OK button to the bottom.
        JPanel bottomPanel = new JPanel();
        bottomPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
        closeButton.setAlignmentX(Component.LEFT_ALIGNMENT);
        bottomPanel.add(closeButton);

        JPanel mainPanel = new JPanel();
        mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
        mainPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
        mainPanel.add(workingDirPanel);
        mainPanel.add(Box.createRigidArea(new Dimension(0,5)));
        mainPanel.add(bottomPanel);

        m_dialog.getContentPane().add(mainPanel);
        m_dialog.pack();
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void changeDataSource() {

        findParentPage();
        m_workingDirLabel.setText(m_workingDir.getAbsolutePath());

        m_parentObjectList.clear();

        File spacefile = new File(m_workingDir.getAbsolutePath()
            + "/spacefile.xml");
        if (spacefile.exists()) {
            m_outputFoundLabel.setText(
                "Si3D output found in spacefile.xml      ");
            XMLProjectFile xmlFile =
                readSi3dObjects(spacefile.getAbsolutePath());
            updateSi3dObjects(xmlFile, "gov.usgs.sfhydro.si3d.Si3dOutput");
            updatePageTitle(xmlFile);
        }
        else {
            m_outputFoundLabel.setText(
                "Si3D output not found in spacefile.xml");
        }

        File bathfile = new File(m_workingDir.getAbsolutePath()
            + "/si3d_bathy.xml");
        if (bathfile.exists()) {
            m_bathFoundLabel.setText(
                "Si3D output found in si3d_bathy.xml      ");
            try {
                XMLProjectFile xmlFile =
                    readSi3dObjects(bathfile.getAbsolutePath());
                updateTriFanObjects(xmlFile);
                Bathymetry bath = (Bathymetry)
                updateSi3dObjects(xmlFile, "gov.usgs.sfhydro.si3d.Bathymetry");
            }
            catch (Exception e2) {}
        }
        else {
            m_bathFoundLabel.setText(
            "Si3D output not found in si3d_bathy.xml");
        }
        reinitializeParentObjects();
    }

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return The editor panel created by this class at initialization time.
 */
    public JPanel getPanel() {

        return m_panel;
    }

//------------------------------------------------------------------------------
/**
 * Tells this interface what the current controller and page are.
 * @param grCon   The GrController to use.
 * @param page    The PageOfGraphs to use.
 */
    public void setControllerAndPage(GrController grCon, PageOfGraphs page) {

        m_grCon = grCon;
        m_page = page;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public XMLProjectFile readSi3dObjects(String xmlFileName) {

        XMLProjectFile xmlFile = new XMLProjectFile();
        xmlFile.setProperty("file", xmlFileName);
        // Read the GrFile.
        try {
            xmlFile.read();
        }
        catch (IOException ioe) {
            m_outputFoundLabel.setText(
                "Si3D output not readable in spacefile.xml");
        }

        return xmlFile;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public GrObject updateSi3dObjects(XMLProjectFile xmlFile,
        String className) {

        GrObject newGro = (GrObject)xmlFile.firstChildInstanceOf(
            className, true);
        if (newGro != null && m_page != null) {
            List list = m_page.findChildInstancesOf(className, true);
            Iterator iter = list.iterator();
            while (iter.hasNext()) {
                GrObject oldGro = (GrObject)iter.next();
                List parentList = oldGro.getObservers();
                Iterator parentIter = parentList.iterator();
                while (parentIter.hasNext()) {
                    GrObject parent = (GrObject)parentIter.next();
                    parent.remove(oldGro);
                    parent.add(newGro);
                    if (!m_parentObjectList.contains(parent))
                        m_parentObjectList.add(parent);
                }
            }
        }
        else {
            m_outputFoundLabel.setText(
                "Si3D output not readable in spacefile.xml");
        }

        return newGro;
    }
//------------------------------------------------------------------------------
/**
 *
 */
    protected void updatePageTitle(XMLProjectFile xmlFile) {

        String si3dName = "gov.usgs.sfhydro.si3d.Si3dOutput";

        Si3dOutput si3dOut = (Si3dOutput)xmlFile.firstChildInstanceOf(
            si3dName, true);
        if (si3dOut == null || m_page == null)
            return;

        GrHashList si3dOutProps = si3dOut.getProperties();
        GrProperty p = (GrProperty)si3dOutProps.get("runTitle");
        String title = (String)p.value;
        p = (GrProperty)si3dOutProps.get("runNumber");
        title += (" [" + (String)p.value + "]");
        m_page.setTitle(title);
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void updateTriFanObjects(XMLProjectFile xmlFile) {

        String bathName = "gov.usgs.sfhydro.si3d.Bathymetry";

        Bathymetry newBath = (Bathymetry)xmlFile.firstChildInstanceOf(
            bathName, true);
        if (newBath == null || m_page == null)
            return;

        DataSeries newBathSeries = newBath.getBathymetrySeries();
        GrHashList bathProps = newBath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        double xSpacing = (Double)p.value;

        List list = m_page.findChildInstancesOf(bathName, true);
        Iterator iter = list.iterator();
        while (iter.hasNext()) {
            Bathymetry oldBath = (Bathymetry)iter.next();
            DataSeries oldBathSeries = oldBath.getBathymetrySeries();

            List parentList = oldBathSeries.getObservers();
            Iterator parentIter = parentList.iterator();
            while (parentIter.hasNext()) {
                GrObject parent = (GrObject)parentIter.next();
                if (parent instanceof GlDiscontinuousTriFanSurface) {
                    GlDiscontinuousTriFanSurface triFan =
                        (GlDiscontinuousTriFanSurface)parent;
                    Iterator childIter =
                        triFan.getChildList().iterator();
                    while (childIter.hasNext()) {
                        GrObject child = (GrObject)childIter.next();
                        if (child instanceof DataSeries)
                            triFan.remove(child);
                    }
                    triFan.setProperty("gridWidth", "" + xSpacing);
                    triFan.add(newBathSeries);
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Reinitializes objects in the m_parentObjectsList.
 */
    public void reinitializeParentObjects() {

        Iterator parentIter = m_parentObjectList.iterator();
        while (parentIter.hasNext()) {
            GrObject parent = (GrObject)parentIter.next();
            if (parent instanceof TimeSeriesExtractor) {
                TimeSeriesExtractor tse = (TimeSeriesExtractor)parent;
                tse.recalculate();
            }
            if (parent instanceof FluxCalculator) {
                FluxCalculator fluxCalc = (FluxCalculator)parent;
                fluxCalc.recalculate();
            }
            if (parent instanceof Gl3dVectors) {
                Gl3dVectors vec = (Gl3dVectors)parent;
                DataSeries oldTimeSeries = null;

                // Find and keep a reference to the old 1D time series.
                Enumeration enumeration = vec.children();
                while (enumeration.hasMoreElements()) {
                    GrObject vecChild = (GrObject)enumeration.nextElement();
                    if (vecChild instanceof DataSeries) {
                        if (((DataSeries)vecChild).getNumDimensions() == 1) {
                            oldTimeSeries = (DataSeries)vecChild;
                            break;
                        }
                    }
                }

                // Reinitialize the vector.
                vec.reinitialize();

                // Find and keep a reference to the new 1D time series.
                DataSeries newTimeSeries = null;
                enumeration = vec.children();
                while (enumeration.hasMoreElements()) {
                    GrObject vecChild = (GrObject)enumeration.nextElement();
                    if (vecChild instanceof DataSeries) {
                        if (((DataSeries)vecChild).getNumDimensions() == 1)
                            newTimeSeries = (DataSeries)vecChild;
                    }
                }

                if (oldTimeSeries != null && newTimeSeries != null) {
                    // Replace the old time series with the new one in the time
                    // marker object.
                    List tmList = m_page.findChildInstancesOf(
                        "gov.usgs.gr.tools.TimeMarker", true);
                    Iterator tmIter = tmList.iterator();
                    while (tmIter.hasNext()) {
                        TimeMarker tm = (TimeMarker)tmIter.next();

                        // Look for the old series in the TM and make a list
                        // of all children of the TM that should come after
                        // it in child list.
                        boolean bOldSeriesFound = false;
                        List afterList = new ArrayList();
                        Enumeration tmChildEnum = tm.children();
                        while (tmChildEnum.hasMoreElements()) {
                            if (bOldSeriesFound) {
                                afterList.add(tmChildEnum.nextElement());
                            }
                            else {
                                if (tmChildEnum.nextElement() == oldTimeSeries)
                                    bOldSeriesFound = true;
                            }
                        }

                        // Remove the old series and add the new series to
                        // the end of the list.
                        tm.remove(oldTimeSeries);
                        tm.add(newTimeSeries);

                        // Remove the children that came after the old series
                        // and add them to the end of the list, so they will
                        // follow the new time series.
                        Iterator afterIter = afterList.iterator();
                        while (afterIter.hasNext()) {
                            GrObject afterObj = (GrObject)afterIter.next();
                            tm.remove(afterObj);
                            tm.add(afterObj);
                        }
                    }
                }
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Finds the PageofGraphs that contains m_zoom.
 */
    public void findParentPage() {

        boolean bPageFound = false;
        List observers = getObservers();
        Iterator obIter = observers.iterator();
        while (obIter.hasNext()) {
            // Look for a page parent of the graph.
            GrObject gro = (GrObject)obIter.next();
            if (gro instanceof PageOfGraphs) {
                m_page = (PageOfGraphs)gro;
                bPageFound = true;
                break;
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

        String s = "Si3d Post Processor";

        return s;
     }
}

