/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2008-10-10
 * Date started:    2002-06-12
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.sfhydro.si3d.*;
import java.util.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class TimeSeriesExtractor extends GrObject {

    protected double m_nearZero = 1E-14;
    protected double[] m_unitScales = {1, 1, 1};
    protected double m_wseScale = 1;

    protected RafDataSeries m_2DVarsSeries = null;
    protected int[] m_2DLookupArray = null;
    protected float[] m_2DVarsArray = null;

    protected RafDataSeries m_3DVarsSeries = null;
    protected float[] m_3DVarsArray = null;

    protected Bathymetry m_bath = null;
    protected Si3dOutput m_si3dOut = null;
    protected GrObjectContainer m_outputContainer = new GrObjectContainer();
    protected DoubleDataSeries m_locationSeries = new DoubleDataSeries(2);
    protected TimeSeries m_tsOutputSeries = new TimeSeries(2);
    protected DoubleDataSeries m_tzuvOutputSeries = new DoubleDataSeries(4);

    protected double m_xSpacing = 1;
    protected double m_ySpacing = 1;
    protected double m_zSpacing = 1;
    protected double m_si3dRunStartTime = 0;
    protected int m_si3dRunStartYear = 0;
    protected double m_timeStepInc = 1;
    protected int m_numTimeSteps = 0;
    protected int m_numI = -1;
    protected int m_numJ = -1;
    protected int m_numLayers = -1;
    protected int m_numWet2Dpoints = 0;
    protected double[] m_avgDepthArray = null;
    protected int[] m_3DIndexArray = null;
    protected int[] m_3DLookupArray = null;
    protected int m_num3DVarsDims = 0;
    protected int[] m_2DIndexArray = null;
    protected int m_num3DWetPoints = 0;

    protected boolean m_bInitialized = false;
    protected boolean m_bAutoRecalc = false;
    protected boolean m_bGodinFilter = false;
    protected double m_scale = 1;

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public TimeSeriesExtractor(){

        setDefaults("gov.usgs.gr.visual.TimeSeriesExtractor");

        m_outputContainer.setName("Output");
        m_outputContainer.setProperty("title", "Output");
        m_outputContainer.add(m_tsOutputSeries);
        m_outputContainer.add(m_tzuvOutputSeries);
        add(m_outputContainer);
    }


//------------------------------------------------------------------------------
/**
 * Overrides     in GrObject.  Only allows DataDimensions to be added.
 * @param gro   The object to add.
 * @return  Whether the object was successfully added.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        if (gro instanceof Bathymetry) {
            // Look for the old bathymetry and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof Bathymetry) iter.remove();
            }

            // Add it at the beginning of the list.
            super.add(gro);
            m_children.remove(gro);
            m_children.add(0, gro);
            m_bath = (Bathymetry)(gro);
            m_bInitialized = false;
            return true;
        }
        if (gro instanceof Si3dOutput) {
            // Look for the old one and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (o instanceof Si3dOutput) iter.remove();
            }

            // Add it at the beginning of the list.
            super.add(gro);
            m_children.remove(gro);
            m_children.add(0, gro);
            m_si3dOut = (Si3dOutput)gro;
            m_bInitialized = false;
            return true;
        }
        if (DoubleDataSeries.countDims(gro, 2)) {
            // Look for the old one and remove it.
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                Object o = iter.next();
                if (DoubleDataSeries.countDims(o, 2))
		    iter.remove();
            }

            // Add it at the beginning of the list.
            super.add(gro);
            m_children.remove(gro);
            m_children.add(0, gro);
            m_locationSeries = (DoubleDataSeries)gro;
            return true;
        }
        if (gro instanceof GrObjectContainer) {
            GrObjectContainer groc = (GrObjectContainer)gro;
            GrHashList props = groc.getProperties();
            GrProperty p = (GrProperty)props.get("title");
            if (p == null)
                return false;
            String title = (String)p.value;

            // Compare the title of groc to each container and replace the
            // current one if there is a match.
            if (title.equals("Output")) {
                GrObjectContainer oldGroc = m_outputContainer;

                int index = m_children.size();
                if (oldGroc != null) {
                    index = m_children.indexOf(oldGroc);
                    oldGroc.deleteObserver(this);
                    m_children.remove(oldGroc);
                    if (index < 0)
                        index = 0;
                }
                m_children.add(index, groc);
                groc.addObserver(this);
                groc.setName("");

                m_outputContainer = groc;

                // Look for DoubleDataSeries within the output container
                // and use them for output if they are there.
                Enumeration enumeration = m_outputContainer.children();
                while (enumeration.hasMoreElements()) {
                    GrObject child = (GrObject)enumeration.nextElement();
                    if (child instanceof DoubleDataSeries) {
                        if (child instanceof TimeSeries) {
                            m_tsOutputSeries = (TimeSeries)child;
                        }
                        else {
                            m_tzuvOutputSeries = (DoubleDataSeries)child;
                        }
                    }
                }

                // Make sure there are no other output series
                // in the output container besided those defined.
                m_outputContainer.clear();
                m_outputContainer.add(m_tsOutputSeries);
                m_outputContainer.add(m_tzuvOutputSeries);

                m_tsOutputSeries.getDimension(0).
                        setProperty("locked", "true");
                m_tsOutputSeries.getDimension(1).
                        setProperty("locked", "true");
                m_tsOutputSeries.setTitle(
                        "Si3D Time Series");

                m_tzuvOutputSeries.getDimension(0).
                        setProperty("locked", "true");
                m_tzuvOutputSeries.getDimension(1).
                        setProperty("locked", "true");
                m_tzuvOutputSeries.getDimension(2).
                        setProperty("locked", "true");
                m_tzuvOutputSeries.getDimension(3).
                        setProperty("locked", "true");
                m_tzuvOutputSeries.setTitle(
                        "Si3D Velocity Profile Series");

                m_bHasChanged = true;
                return true;
            }
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Wrapper for removing items from the list of children.
 */
    public GrObject remove(int i) {

        if (m_children.get(i) instanceof GrObjectContainer)
            return null;

        // Let the super class handle removal.
        GrObject child = super.remove(i);

        if (child == m_bath) m_bath = null;
        if (child == m_si3dOut) m_si3dOut = null;
        if (child == m_locationSeries)
            m_locationSeries = null;

        return child;
    }

//------------------------------------------------------------------------------
/**
 * Removes child gro from the graph.  Does not remove object containers.
 */
    public boolean remove(GrObject gro) {

        if (gro == m_bath) m_bath = null;
        if (gro == m_si3dOut) m_si3dOut = null;
        if (gro == m_locationSeries)
            m_locationSeries = null;

        if (gro instanceof GrObjectContainer) {
            return false;
        }
        else {
            return super.remove(gro);
        }
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("autoRecalc", new GrProperty(
            m_bAutoRecalc, "Automatic Recalculation"));
        props.put("scale", new GrProperty(m_scale, "Scale"));
        props.put("godinFilter", new GrProperty(
            m_bGodinFilter, "Apply Godin Filter"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

        super.setProperties(props);
        
        boolean bNeedsUpdate = false;

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                if (key.equals("autoRecalc")) {
                    m_bAutoRecalc = (Boolean)p.value;
                    bNeedsUpdate = true;
                }
                if (key.equals("scale")) {
                    m_scale = (Double)p.value;
                    bNeedsUpdate = true;
                }
                if (key.equals("godinFilter")) {
                    m_bGodinFilter = (Boolean)p.value;
                    bNeedsUpdate = true;
                }
            }
            if (obj instanceof String) {
                String value = (String)obj;
                setProperty(key, value);
            }
        }

        m_bHasChanged = true;
        if (bNeedsUpdate)
            update(this, null);
        
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("autoRecalc"))
            m_bAutoRecalc = Boolean.parseBoolean(value);
        if (key.equals("scale"))
            m_scale = Double.parseDouble(value);
        if (key.equals("godinFilter"))
            m_bGodinFilter = Boolean.parseBoolean(value);
    }

//------------------------------------------------------------------------------
/**
 * Return the DoubleDataSeries within the specified ObjectContainer within this
 * object.
 */
    public DoubleDataSeries getContainerSeries(String containerTitle) {

        // Look through the children of this object for an ObjectContainer
        // with a matching title.
        GrObjectContainer groc =
            GrObjectContainer.getChildContainer(this, containerTitle);

        if (groc != null) {
            // Look through the contents of the ObjectContainer for
            // a DoubleDataSeries.
            Enumeration enumeration = groc.children();
            while (enumeration.hasMoreElements()) {
                GrObject oc_child = (GrObject)enumeration.nextElement();
                if (oc_child instanceof DoubleDataSeries)
                    return (DoubleDataSeries)oc_child;
            }
        }

        // Could not find a variables series under a container with the
        // given title.
        return null;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void retrieveVariables () {

        if (m_locationSeries == null || m_locationSeries.getNumPoints() < 1)
            return;

        double[] tsArray = new double[m_numTimeSteps * 2];

        double[] locationArray = m_locationSeries.getArray();
        int locationI = (int)Math.round(locationArray[0] / m_xSpacing);
        int locationJ = (int)Math.round(locationArray[1] / m_ySpacing);

        int ijIndex = locationI + (locationJ * m_numI);
        int lookup2D = -1;

        // Check that the data is valid and the index is a wet point.
        if (m_2DLookupArray != null && ijIndex < m_2DLookupArray.length) {
            lookup2D = m_2DLookupArray[ijIndex];
        }

        // Return null if the coordinate is land.
        if (lookup2D < 0) {
            return;
        }

        DoubleDataSeries bathSeries = m_bath.getBathymetrySeries();
        double[] bathArray = bathSeries.getArray();
        int numBathDims = bathSeries.getNumDimensions();
        int numColumnLayers =
            (int)Math.round(bathArray[lookup2D * numBathDims + 2]);
        double[] depthArray = m_bath.getAverageDepthArray();
        double depth = depthArray[lookup2D];
        double bottomLayerHeight =
            depth - ((numColumnLayers - 1) * m_zSpacing);
        double bottomAdjustment = m_zSpacing - bottomLayerHeight;

//        double[] tzuvArray = new double[m_numTimeSteps * numColumnLayers * 4];

        for (int timeStep = 0; timeStep < m_numTimeSteps; timeStep++) {
            double time = (double)timeStep * m_timeStepInc + m_si3dRunStartTime;

            int layer = 1;

            double wse = getWaterSurfaceElevation(
                locationI, locationJ, timeStep);

            tsArray[timeStep * 2] = time;
            tsArray[timeStep * 2 + 1] = wse * m_scale;

/*
            for (layer = 1; layer <= numColumnLayers; layer++) {
                int currentLookup = getLookupIndex(locationI, locationJ,
                    layer, m_numI, m_numJ, m_numLayers, m_num3DVarsDims,
                    timeStep, m_num3DWetPoints, m_3DLookupArray);

                double layerHeight = m_zSpacing;
                if (layer == numColumnLayers)
                    layerHeight = bottomLayerHeight;
                if (layer == 1) {
                    layerHeight += wse;
                }

                double uVel = getVar3D(currentLookup, 0);
                uVel *= m_unitScales[0];
                double vVel = getVar3D(currentLookup, 1);
                vVel *= m_unitScales[1];

                double z = (double)(numColumnLayers - layer + 1) * m_zSpacing;
                if (layer == numColumnLayers)
                    z -= (bottomAdjustment * 0.5);
                else
                    z -= (m_zSpacing * 0.5);

                int tzuvOffset =
                    timeStep * numColumnLayers * 4 + ((layer - 1) * 4);
                tzuvArray[tzuvOffset] = time;
                tzuvArray[tzuvOffset + 1] = z;
                tzuvArray[tzuvOffset + 2] = uVel * 100;
                tzuvArray[tzuvOffset + 3] = vVel * 100;
            }
 */
        }

//        m_tzuvOutputSeries.setArray(tzuvArray);

        m_tsOutputSeries.setReferenceYear(m_si3dRunStartYear);
        m_tsOutputSeries.setArray(tsArray);
        if (tsArray[0] > 275 && tsArray[tsArray.length - 2] > 355) {
            m_tsOutputSeries.syncToReferenceYear(m_si3dRunStartYear + 1);
            m_tsOutputSeries.arrayChanged();
        }

        if (m_bGodinFilter) {
            ArrayList dataList = new ArrayList();
            dataList.add(m_tsOutputSeries);
            gov.usgs.sfhydro.gr.tools.GodinFilter gf =
                new gov.usgs.sfhydro.gr.tools.GodinFilter();
            gf.setProperty("parabolicInterpolation", "true");
            gf.performAction(dataList);
            m_tsOutputSeries.deselectAll();

            /*
            gf.setProperty("fDims", "2,3");
            dataList.clear();
            /// ??? Is this filtering the z value?
            dataList.add(m_tzuvOutputSeries);
            gf.performAction (dataList);
            m_tzuvOutputSeries.deselectAll();
             */
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    protected double getWaterSurfaceElevation(int i, int j, int timeStep) {

        int lookup2D = -1;
        int ijIndex = i + (j * m_numI);
        if (ijIndex < 0 || ijIndex > m_2DLookupArray.length) {
            return 0;
        }

        // Check that the data is valid and the index is a
        // wet point.
        if (m_2DLookupArray != null)
            lookup2D = m_2DLookupArray[ijIndex];

        if (lookup2D < 0)
            return 0;

        double waterSurfaceElevation = getVar2D(lookup2D
            + (timeStep * m_numWet2Dpoints)) * m_wseScale;

        return waterSurfaceElevation;
    }

//------------------------------------------------------------------------------
/**
 * Returns the lookup index of the specified cell for the vars3D array.
 */
    protected int getLookupIndex(int i, int j, int k, int numI, int numJ,
        int numK, int numVarDims, int ts, int numWetPoints, int[] lookupArray) {

        int lookupIndex = -1;

        // Get the bottom value from the lower neighbor cell.
        if (i >= 0 && i < numI && j >= 0 && j < numJ && k >= 0 && k < numK) {

            lookupIndex = lookupArray[i * numK + j * numK * numI + k];
            if (lookupIndex >= 0)
                lookupIndex = (lookupIndex + ts * numWetPoints) * numVarDims;
        }

        return lookupIndex;
    }

//------------------------------------------------------------------------------
/**
 * Returns the U, V, or W components of the specified cell.
 */
    protected double getVar3D(int lookupIndex, int dimNum) {

        double var = 0;

        if (m_3DVarsSeries instanceof RafDataSeries) {
            if (lookupIndex >= 0) {
                var = ((RafDataSeries)m_3DVarsSeries).getValueAt(
                    lookupIndex + dimNum);
            }
        }
        else {
            if (lookupIndex >= 0 &&
                (lookupIndex + dimNum) < m_3DVarsArray.length) {

                var = m_3DVarsArray[lookupIndex + dimNum];
            }
        }

        return var;
    }

//------------------------------------------------------------------------------
/**
 * Returns the 2D variable at the specified offset.
 */
    protected double getVar2D(int offset) {

        double var = 0;

        if (m_2DVarsSeries instanceof RafDataSeries) {
            var = ((RafDataSeries)m_2DVarsSeries).getValueAt(offset);
        }
        else {
            var = m_2DVarsArray[offset];
        }

        return var;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    protected boolean initializeVariables() {

        // Make sure the Bathymetry and Si3dOutput objects are defined.
        if (m_bath == null || m_si3dOut == null) {
            Iterator iter = m_children.iterator();
            while (iter.hasNext()) {
                GrObject gro = (GrObject)iter.next();
                if (gro instanceof Bathymetry) m_bath = (Bathymetry)(gro);
                if (gro instanceof Si3dOutput) m_si3dOut = (Si3dOutput)gro;
            }
        }

        if (m_bath == null || m_si3dOut == null) return false;

        // Get grid spacing information from the Bathymetry object.
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        m_xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        m_ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        m_zSpacing = (Double)p.value;

        // Get time step information from the Si3dOutput object.
        m_si3dRunStartYear = m_si3dOut.getStartYear();
        m_si3dRunStartTime = m_si3dOut.getJulianStartTime();
        GrHashList si3dOutProps = m_si3dOut.getProperties();
        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long)p.value;
        m_timeStepInc = ((double)tsInSec) / 86400;

        // Find the dimensions of the grid and the number of wet 2D points.
        DoubleDataSeries bathSeries = m_bath.getBathymetrySeries();
        m_numI = (int)Math.round(bathSeries.getMax(0)) + 1;
        m_numJ = (int)Math.round(bathSeries.getMax(1)) + 1;
        m_numLayers = (int)Math.round(bathSeries.getMax(2));
        m_numWet2Dpoints = bathSeries.getNumPoints();
        m_avgDepthArray = m_bath.getAverageDepthArray();

        // Get array indexing information from the Bathymetry object.
        m_3DIndexArray = m_bath.get3DIndexArray();
        // Get the rectangular array that holds -1 at dry points and the
        // index into the variables array at wet points.
        m_3DLookupArray = m_bath.get3DLookupArray(m_3DIndexArray);

        m_3DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("3-D Variables");
//        if (m_3DVarsSeries != null)
//            m_3DVarsArray = m_3DVarsSeries.getArray();
        m_2DVarsSeries = (RafDataSeries)
            m_si3dOut.getContainerSeries("2-D x-y Variables");
        if (m_2DVarsSeries != null) {
//            m_2DVarsArray = m_2DVarsSeries.getArray();

            DataDimension dim = m_2DVarsSeries.getDimension(0);
            if (dim != null) {
                // Store the scale factor for the dimension.
                GrHashList dimProps = dim.getProperties();
                p = (GrProperty)dimProps.get("unitScale");
                m_wseScale = (Double)p.value;
            }
        }

        m_2DIndexArray = m_bath.get2DIndexArray();
        m_2DLookupArray = m_bath.get2DLookupArray(m_2DIndexArray);
        m_3DIndexArray = m_bath.get3DIndexArray();
        m_3DLookupArray = m_bath.get3DLookupArray(m_3DIndexArray);

        // Get the 3D variables array.
        m_num3DVarsDims = m_3DVarsSeries.getNumDimensions();
//        m_3DVarsArray = m_3DVarsSeries.getArray();

        // Find the unit scales for the X, Y, and Z dimensions.
        GrProperty unitProp = (GrProperty)
            (m_3DVarsSeries.getDimension(0).getProperties().get("unitScale"));
        m_unitScales[0] = (Double)unitProp.value;
        unitProp = (GrProperty)
            (m_3DVarsSeries.getDimension(1).getProperties().get("unitScale"));
        m_unitScales[1] = (Double)unitProp.value;
        unitProp = (GrProperty)
            (m_3DVarsSeries.getDimension(2).getProperties().get("unitScale"));
        m_unitScales[2] = (Double)unitProp.value;

        // Create an array to hold Time-X-Y-Z data of the particle at every
        // time step for the duration of the run.
        m_num3DWetPoints = m_3DIndexArray.length / 3;
        m_numTimeSteps = m_3DVarsSeries.getCount() / m_num3DWetPoints;

        return true;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void recalculate() {

        if (!m_bInitialized) {
            if (!initializeVariables())
                return;

            m_bInitialized = true;
        }

        retrieveVariables();
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        if (m_bAutoRecalc) {
            if (gro == m_locationSeries || gro == m_bath ||
                gro == m_si3dOut || gro == this) {

                if (gro == m_bath || gro == m_si3dOut)
                    m_bInitialized = false;

                recalculate();
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s = "Time Series Extractor";

        return s;
    }
}






