/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-05
 * Date started:    2003-02-23
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.sfhydro.si3d.*;
import java.util.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class VelocityFilter extends GrObject
    implements java.awt.event.ActionListener, gov.usgs.gr.tools.ThreadWorker {

    protected int m_blockSize = 0;
    protected int m_blockOffset = 0;
    protected int m_numWetPoints = 0;
    protected int m_numTimeSteps = 0;
    protected int m_numDims = 0;
    protected float m_uScale = 0;
    protected float m_vScale = 0;
    protected float m_wScale = 0;
    protected float m_outputScale = 1;
    protected RafDataSeries m_rafSeries = null;
    float[] m_times = new float[0];
    FloatDataSeries m_intervalSeries = null;
    protected gov.usgs.gr.data.RafDataSeries m_outRafSeries = null;

    protected boolean[] m_bThreadStates = new boolean[0];
    protected int m_numThreads = 1;

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public VelocityFilter(){

        setDefaults("gov.usgs.sfhydro.si3d.VelocityFilter");
    }

//------------------------------------------------------------------------------
/**
 * Add a child object.  Accepts DataSeries, TimeMarker.  Only holds one DS at
 * a time.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
        	gro = ((ObjectEffigy)gro).getSubject();

    	if (m_children.contains(gro)) return false;

    	if (gro instanceof DataSeries) {
            super.add(gro);

            return true;
        }
        if (gro instanceof Bathymetry || gro instanceof Si3dOutput) {
            super.add(gro);

            return true;
        }
        if (gro instanceof GrObjectContainer) {
            super.add(gro);

            return true;
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the current values for this object.
 * Properties include: <p>
 * numThreads       Integer "Number of Threads"                             <p>
 * outputScale      Float   "Output Scale"                                  <p>
 * @return  A list of GrProperty objects.
 */
    public GrHashList getProperties() {

        // First add properties from the super class.
        GrHashList props = super.getProperties();

        props.put("numThreads", new GrProperty(m_numThreads,
            "Number of Threads"));
        props.put("outputScale", new GrProperty(m_outputScale, "Output Scale"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Set the values of one or more of this object's properties.
 * @see gov.usgs.sfhydro.si3d.VelocityFilter#getProperties getProperties() for
 * a list of accepted properties.
 * @param props A list of GrProperty objects specifying the new values for
 *              certain properties.
 */
    public void setProperties(HashMap props) {

        GrProperty p;

        // For each property with a string value, pass it to the setProperty()
        // method to be handled there.
        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                // Handle string properties by passing them to the
                // setProperty() method.
                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                if (key.equals("numThreads")) {
                    int numThreads = (Integer)p.value;
                    if (numThreads > 0 && numThreads < 1000)
                        m_numThreads = numThreads;
                }
                if (key.equals("outputScale")) {
                    m_outputScale = (Float)p.value;
                }
            }
            // Handle string properties by passing them to the
            // setProperty() method.
            if (obj instanceof String) {
                setProperty(key, (String)obj);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 * @see gov.usgs.sfhydro.si3d.VelocityFilter#getProperties getProperties() for
 * a list of accepted properties.
 * @param key   The property name.
 * @param value The new value for the property.
 */
    public void setProperty(String key, String value) {

        if (key.equals("numThreads")) {
            int numThreads = Integer.parseInt(value);
            if (numThreads > 0 && numThreads < 1000)
                m_numThreads = numThreads;
        }
        if (key.equals("outputScale")) {
            m_outputScale = Float.parseFloat(value);
        }
    }

//------------------------------------------------------------------------------
/**
 * Issue the OpenGL commands to draw this series.
 */
    public void filter() {

        if (m_outRafSeries != null)
            throw new ConcurrentModificationException();

        long startTimeMillis = System.currentTimeMillis();

        Bathymetry bath = null;
        Si3dOutput si3dOut = null;

        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject)iter.next();
            if (gro instanceof Bathymetry) {
                bath = (Bathymetry)(gro);
            }
            if (gro instanceof Si3dOutput) {
                si3dOut = (Si3dOutput)gro;
            }
        }

        // Check that the member variables have been initialized.
        if (bath == null || si3dOut == null) {
            return;
        }

        int[] index3d = bath.get3DIndexArray();
        DataSeries series3d = si3dOut.getContainerSeries("3-D Variables");
        if (series3d == null)
            return;

        if (!(series3d instanceof RafDataSeries))
            return;

        m_rafSeries = (RafDataSeries)series3d;
        m_numDims = m_rafSeries.getNumDimensions();

        DataDimension dim = m_rafSeries.getDimension(0);
        GrHashList dimProps = dim.getProperties();
        GrProperty p = (GrProperty)dimProps.get("unitScale");
        m_uScale = (Float)p.value;

        dim = m_rafSeries.getDimension(1);
        dimProps = dim.getProperties();
        p = (GrProperty)dimProps.get("unitScale");
        m_vScale = (Float)p.value;

        dim = m_rafSeries.getDimension(2);
        dimProps = dim.getProperties();
        p = (GrProperty)dimProps.get("unitScale");
        m_wScale = (Float)p.value;

        m_numTimeSteps = series3d.getCount() / (index3d.length / 3);

        float startTime = (float)si3dOut.getJulianStartTime();
        GrHashList si3dOutProps = si3dOut.getProperties();
        p = (GrProperty)si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long)p.value;
        float timeStepInc = ((float)tsInSec) / 86400;

        // Create the Vector Animation and populate it with Time, (X, Y, Z)
        // and (U, V, W) arrays.
        // Round to the nearest minute.
        m_times = new float[m_numTimeSteps];
        for (int i = 0; i < m_numTimeSteps; i++) {
            m_times[i] = startTime + (i * timeStepInc) + 0.0003472222222222222f;
        }

        m_numWetPoints = index3d.length / 3;

        GrHashList props = m_rafSeries.getProperties();
        p = (GrProperty)props.get("file");
        String oldFilename = (String)p.value;
        String newFilename = oldFilename.replaceFirst(".bin", "_filt.bin");
        if (oldFilename.equals(newFilename)) {
            System.out.println(
                "3D velocity filename did not contain .bin.  " +
                "Exiting without filtering.");

            return;
        }

        m_outRafSeries = new gov.usgs.gr.data.RafDataSeries(m_numDims);
        m_outRafSeries.setProperty("file", newFilename);
        m_outRafSeries.openRAF("rwd");

        System.out.println("Writing filtered velocities to " + newFilename);

        m_blockSize = m_numWetPoints;
        float[] allocationTestArray = null;
        while (allocationTestArray == null) {
            try {
                allocationTestArray =
                    new float[m_blockSize * m_numTimeSteps * m_numDims];
            }
            catch (OutOfMemoryError e) {
                m_blockSize /= 2;
                if (m_blockSize < 2)
                    return;
            }
        }

        m_blockSize /= 2;
        m_blockSize /= m_numThreads;
        m_blockSize += m_numThreads;
        allocationTestArray = new float[0];

        System.out.println("Number of Wet Points = " + m_numWetPoints
            + ", Block size = " + m_blockSize
            + ", Number of Threads = " + m_numThreads);

        // Filter the data using parallel processing.
        MultiThreader multiThreader =
            new MultiThreader(this, m_numThreads, "Filter");
        multiThreader.runThreads();

        m_outRafSeries.closeRAF();

        m_times = new float[0];
        m_intervalSeries = null;
        m_outRafSeries = null;

        long endTimeMillis = System.currentTimeMillis();

        long totalTimeSeconds = (endTimeMillis - startTimeMillis) / 1000;
        long hours = totalTimeSeconds / 3600;
        totalTimeSeconds -= hours * 3600;
        long minutes = totalTimeSeconds / 60;
        long seconds = totalTimeSeconds - (minutes * 60);
        java.text.NumberFormat form = new java.text.DecimalFormat("00");
        System.out.println("Time elapsed while filtering file: "
            + form.format(hours) + ":" + form.format(minutes) + ":"
            + form.format(seconds));
    }

//------------------------------------------------------------------------------
/**
 * Filters one block of wet points.
 */
    public boolean performThreadWork() {

        int blockOffset;

        synchronized(this) {
            if (m_blockOffset >= m_numWetPoints)
                return false;

            blockOffset = m_blockOffset;
            m_blockOffset += m_blockSize;
        }

        FloatDataSeries uSeries = new FloatDataSeries(2);
        FloatDataSeries vSeries = new FloatDataSeries(2);
        FloatDataSeries wSeries = new FloatDataSeries(2);
        float[] uArray = new float[m_numTimeSteps * 2];
        float[] vArray = new float[m_numTimeSteps * 2];
        float[] wArray = new float[m_numTimeSteps * 2];

        gov.usgs.sfhydro.gr.tools.GodinFilter gf =
            new gov.usgs.sfhydro.gr.tools.GodinFilter();
        gf.setProperty("parabolicInterpolation", "true");
        ArrayList seriesList = new ArrayList();
        seriesList.add(uSeries);
        seriesList.add(vSeries);
        seriesList.add(wSeries);

        int stride = m_blockSize;
        if ((blockOffset + m_blockSize) > m_numWetPoints)
            stride = m_numWetPoints - blockOffset;

        float[] strideArray = new float[stride * m_numTimeSteps * m_numDims];

        synchronized (this) {
            for (int ts = 0; ts < m_numTimeSteps; ts++) {
                long fileOffset = (blockOffset + (ts * m_numWetPoints))
                    * m_numDims;
                m_rafSeries.getValuesInto(strideArray, ts * stride * m_numDims,
                    fileOffset, stride * m_numDims);
            }
        }

        System.out.println("Read nodes " + blockOffset + " to "
            + (blockOffset + stride - 1) + " from disk" + "  -- Thread "
            + Thread.currentThread().getName());

        int lastPrintedIndex = blockOffset - 1;

        for (int j = 0; j < stride; j++) {
            for (int ts = 0; ts < m_numTimeSteps; ts++) {
                int offset = (j + (ts * stride)) * m_numDims;
                uArray[ts * 2] = m_times[ts];
                vArray[ts * 2] = m_times[ts];
                wArray[ts * 2] = m_times[ts];
                uArray[ts * 2 + 1] = strideArray[offset] * m_uScale;
                vArray[ts * 2 + 1] = strideArray[offset + 1] * m_vScale;
                wArray[ts * 2 + 1] = strideArray[offset + 2] * m_wScale;
            }

            float firstTime = uArray[0];
            float lastTime = uArray[uArray.length - 2];

            if (m_intervalSeries == null) {
                m_intervalSeries = new FloatDataSeries(uArray, 2);
            }

            uSeries.setArray(uArray);
            vSeries.setArray(vArray);
            wSeries.setArray(wArray);
            gf.performAction (seriesList);

            float[] uFiltArray = uSeries.getArray();
            uFiltArray[0] = firstTime;
            uFiltArray[1] = 0;
            uFiltArray[uFiltArray.length - 2] = lastTime;
            uFiltArray[uFiltArray.length - 1] = 0;
            uSeries.arrayChanged();
            uSeries.resample(m_intervalSeries);

            float[] vFiltArray = vSeries.getArray();
            vFiltArray[0] = firstTime;
            vFiltArray[1] = 0;
            vFiltArray[vFiltArray.length - 2] = lastTime;
            vFiltArray[vFiltArray.length - 1] = 0;
            vSeries.arrayChanged();
            vSeries.resample(m_intervalSeries);

            float[] wFiltArray = wSeries.getArray();
            wFiltArray[0] = firstTime;
            wFiltArray[1] = 0;
            wFiltArray[wFiltArray.length - 2] = lastTime;
            wFiltArray[wFiltArray.length - 1] = 0;
            wSeries.arrayChanged();
            wSeries.resample(m_intervalSeries);

            float[] uOutArray = uSeries.getArray();
            float[] vOutArray = vSeries.getArray();
            float[] wOutArray = wSeries.getArray();

            // Put the results in the shared array to be written to the
            // file later.
            for (int ts = 0; ts < m_numTimeSteps; ts++) {
                int offset = (j + (ts * stride)) * m_numDims;
                strideArray[offset] =
                    uOutArray[ts * 2 + 1] * m_outputScale / m_uScale;
                strideArray[offset + 1] =
                    vOutArray[ts * 2 + 1] * m_outputScale / m_vScale;
                strideArray[offset + 2] =
                    wOutArray[ts * 2 + 1] * m_outputScale / m_wScale;
            }
            if (((j % 1000) == 0 && j > 0) || j == (stride - 1)) {
                System.out.println("Filtered nodes " + (lastPrintedIndex + 1)
                    + " to " + (blockOffset + j) + "  -- Thread "
                    + Thread.currentThread().getName());
                lastPrintedIndex = blockOffset + j;
            }
        }

        synchronized (this) {
            System.out.println("Writing " + blockOffset + " to "
                + (blockOffset + stride - 1) + "  -- Thread "
                + Thread.currentThread().getName());

            for (int ts = 0; ts < m_numTimeSteps; ts++) {
                int fileOffset = (blockOffset + (ts * m_numWetPoints))
                    * m_numDims;
                m_outRafSeries.writeValuesFrom(strideArray,
                    ts * stride * m_numDims,
                    fileOffset, stride * m_numDims);
            }
        }

        return true;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

    	m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of menu items that execute commands on this object.
 * @return  A menu item that can be placed on a menu and send an ActionEvent
 *          back to this object when it is chosen.
 */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
        	"Si3D Velocity Filter");
        javax.swing.JMenuItem item = new javax.swing.JMenuItem(
            "Create Filtered File");
        item.addActionListener(this);
        subMenu.add(item);

    	return subMenu;
    }

//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
        	(javax.swing.JMenuItem)(e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Create Filtered File")) {
        	filter();
        }
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString(){

        String s =  "Si3d Velocity Filter";

        return s;
    }
}






