/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2007-05-02
 * Date started:    2002-06-12
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */
package gov.usgs.sfhydro.si3d.gl;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.tools.*;
import gov.usgs.gr.visual.*;
import gov.usgs.gr.visual.gl.*;
import gov.usgs.sfhydro.si3d.*;
import java.awt.geom.*;
import java.io.PrintStream;
import java.util.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class Gl3dVectors extends gov.usgs.gr.visual.GraphicalElement
        implements GlElement, java.awt.event.ActionListener {

    /** The OpenGL GLAutoDrawable component. */
    protected GLAutoDrawable gld = null;
    /**
     * The main OpenGL object with all the gl drawing methods.  For
     * readability, the m_ prefix wasn't used.
     */
    protected GL gl = null;
    /** The OpenGL utility class with all the glu drawing methods. */
    protected GLU glu = null;
    /** Whether to interpret the third and fourth dims as U and V as
     *  opposed to direction and speed.
     *  */
    protected boolean m_bUsingUV = false;
    protected double[][] m_dataArrays = new double[7][];
    protected int[] m_seriesNums = {0, 0, 0, 0, 0, 0, 0};
    protected int[] m_dimNums = {0, 1, 2, -1, 3, 4, -1};
    protected int[] m_numDims = {0, 0, 0, 0, 0, 0, 0};
    protected double[] m_dimScales = {1, 1, 1, 1, 1, 1, 1};
    protected DataSeries[] m_seriesArray = new DataSeries[7];
    protected String[] m_symbolTypes;
    protected boolean m_bAnimationInitialized = false;
    protected RafDataSeries m_2DVarsSeries = null;
    protected RafDataSeries m_3DVarsSeries = null;
    protected int m_rafCacheSize2D = -1;
    protected long m_rafCacheOffset2D = -1;
    protected int[] m_3DIndexArray = null;
    protected int[] m_3DLookupArray = null;
    protected int[] m_2DLookupArray = null;
//    protected float[] m_2DVarsArray = null;
    protected float[] m_3DVarsArray = null;
    protected double m_wseScale = 1;
    protected double m_xSpacing = 1;
    protected double m_ySpacing = 1;
    protected double m_zSpacing = 1;
    protected int m_numI = -1;
    protected int m_numJ = -1;
    protected int m_numK = -1;
    protected int[] m_westArray = null;
    protected int[] m_southArray = null;
    protected int[] m_eastArray = null;
    protected int[] m_northArray = null;
    protected double m_xOutputSpacing = 1000;
    protected double m_yOutputSpacing = 1000;
    protected double m_zOutputSpacing = 2;
    protected double m_xOutputMin = 0;
    protected double m_yOutputMin = 0;
    protected double m_zOutputMin = 0;
    protected double m_xOutputMax = Double.MAX_VALUE;
    protected double m_yOutputMax = Double.MAX_VALUE;
    protected double m_zOutputMax = Double.MAX_VALUE;
    protected double[] m_xyLocations = null;
    protected int[] m_outputLocations = null;
    protected double[] m_profileSticks = null;
    protected boolean m_bDrawProfileSticks = true;
    protected boolean m_bDrawComponents = false;
    protected boolean m_bDrawCenterValues = true;
    protected boolean m_bComponentsAreStaggered = true;
    protected boolean m_bWriteValuesToFile = false;
    protected PrintStream m_outputPrintStream = null;
    protected double fileTime = Double.MIN_VALUE;
    protected Bathymetry m_bath = null;
    protected Si3dOutput m_si3dOut = null;
    protected String m_errorString = null;
    protected Rectangle2D m_lastContentTrans =
            new Rectangle2D.Double(0, 0, 0, 0);
    protected double m_scale = 1;
    /** The color to use when drawing. */
    protected float[] m_color = {1, 0.2f, 0.2f};
    protected TimeMarker m_timeMarker = new TimeMarker();
    protected ColorScheme m_colorScheme = null;
    protected double m_zScale = 0;
    /**
     * The numeric code of the symbol to be used.
     *  0 = simple arrow.
     *  1 = arrow with black inner area.
     *  2 = bar
     *  3 = line
     *  4 = 3D arrow with head with diamond cross-section.
     */
    protected int m_symbolIndex = 1;
    /** The tree icon to use for this class. */
    public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GlVectorAnimation.gif";

//------------------------------------------------------------------------------
    /**
     * Constructor for creating a blank data series.
     */
    public Gl3dVectors() {

        m_bUsingUV = true;
        m_zScale = -5;
        m_scale = 300;

        m_symbolTypes = new String[]{"Simple Arrow", "Two Tone Arrow",
            "Narrow Arrow", "Line Arrow", "Simple Bar", "Line"
        };

        setDefaults("gov.usgs.gr.visual.gl.GlVectorAnimation");
    }

//------------------------------------------------------------------------------
    /**
     * Add a child object.  Accepts DataSeries, TimeMarker.  Only holds one DS at
     * a time.
     */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy) {
            gro = ((ObjectEffigy) gro).getSubject();
        }

        if (m_children.contains(gro)) {
            return false;
        }

        if (gro instanceof DataSeries) {
            super.add(gro);

            resetDataDims();

            return true;
        }
        if (gro instanceof Bathymetry || gro instanceof Si3dOutput) {
            super.add(gro);

//            createVectorAnimation();

            return true;
        }
        if (gro instanceof GraphicalElement) {

            return super.add(gro);
        }
        if (gro instanceof GrObjectContainer) {
            super.add(gro);

            return true;
        }
        if (gro instanceof ColorScheme) {
            if (m_colorScheme != null) {
                remove((GrObject) m_colorScheme);
            }
            super.add(gro);
            m_colorScheme = (ColorScheme) gro;

            return true;
        }

        return false;
    }

//------------------------------------------------------------------------------
    /**
     * Wrapper for removing items from the list of children and resetting data
     * dimensions afterward.
     */
    public boolean remove(GrObject gro) {

        // Let the super class handle removal.
        boolean bFound = super.remove(gro);
        resetDataDims();

        if (gro == m_bath) {
            m_bath = null;
        }
        if (gro == m_si3dOut) {
            m_si3dOut = null;
        }

        return bFound;
    }

//------------------------------------------------------------------------------
    /**
     * Wrapper for removing items from the list of children and resetting data
     * dimensions afterward.
     */
    public GrObject remove(int i) {

        // Let the super class handle removal.
        GrObject child = super.remove(i);
        resetDataDims();

        if (child == m_bath) {
            m_bath = null;
        }
        if (child == m_si3dOut) {
            m_si3dOut = null;
        }

        return child;
    }

//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0) {
            setTreeIcon();
        }

        return (javax.swing.ImageIcon) TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null) {
            icon = super.setTreeIcon();
        }

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }

//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("scale", new GrProperty(m_scale, "Scale"));
        props.put("zScale", new GrProperty(m_zScale, "Z Scale"));
        props.put("usingUV", new GrProperty(
                m_bUsingUV, "U-V instead of Direction-Magnitude"));
        props.put("drawProfileSticks", new GrProperty(
                m_bDrawProfileSticks, "Draw Profile Sticks"));
        props.put("drawComponents", new GrProperty(
                m_bDrawComponents, "Draw Components"));
        props.put("drawCenterValues", new GrProperty(
                m_bDrawCenterValues, "Draw Center Values"));
        props.put("componentsAreStaggered", new GrProperty(
                m_bComponentsAreStaggered, "Components Are Staggered"));
        props.put("writeValuesToFile", new GrProperty(
                m_bWriteValuesToFile, "Write Values to File"));

        String s = "";
        for (int i = 0; i < m_seriesNums.length; i++) {
            if (i > 0) {
                s += ", ";
            }
            s += m_seriesNums[i];
        }
        props.put("seriesNums", new GrProperty(s,
                "Series Numbers for T,X,Y,Z,U,V,W"));

        s = "";
        for (int i = 0; i < m_dimNums.length; i++) {
            if (i > 0) {
                s += ", ";
            }
            s += m_dimNums[i];
        }
        props.put("dimNums", new GrProperty(s,
                "Dimension Numbers for T,X,Y,Z,U,V,W"));


        // Give the user some symbol choices...
        GrProperty p = new GrProperty("", "Symbol Type");
        javax.swing.JComboBox jcb = new javax.swing.JComboBox();

        // Fill the symbol type combo box.
        for (int i = 0; i < m_symbolTypes.length; i++) {
            String name = m_symbolTypes[i];
            jcb.addItem(name);
            if (m_symbolIndex == i) {
                jcb.setSelectedIndex(i);
                p.value = m_symbolTypes[i];
            }
        }

        p.component = jcb;
        props.put("symbolType", p);

        ColorProperty cp = new ColorProperty();
        cp.description = "Color";
        cp.setColor(m_color);
        props.put("color", cp);

        props.put("xOutputMin", new GrProperty(m_xOutputMin, "X Min"));
        props.put("xOutputMax", new GrProperty(m_xOutputMax, "X Max"));
        props.put("xOutputSpacing", new GrProperty(m_xOutputSpacing,
                "X Spacing"));
        props.put("yOutputMin", new GrProperty(m_yOutputMin, "Y Min"));
        props.put("yOutputMax", new GrProperty(m_yOutputMax, "Y Max"));
        props.put("yOutputSpacing", new GrProperty(m_yOutputSpacing,
                "Y Spacing"));
        props.put("zOutputMin", new GrProperty(m_zOutputMin, "Z Min"));
        props.put("zOutputMax", new GrProperty(m_zOutputMax, "Z Max"));
        props.put("zOutputSpacing", new GrProperty(m_zOutputSpacing,
                "Z Spacing"));

        return props;
    }

//------------------------------------------------------------------------------
    /**
     * Provide the list of properties with the latest values.
     */
    public void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String) iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty) obj;

                if (p.value instanceof String) {
                    setProperty(key, (String) p.value);
                    continue;
                }

                if (key.equals("scale")) {
                    m_scale = (Double) p.value;
                }
                if (key.equals("zScale")) {
                    m_zScale = (Double) p.value;
                }
                if (key.equals("usingUV")) {
                    m_bUsingUV = (Boolean) p.value;
                }
                if (key.equals("drawProfileSticks")) {
                    m_bDrawProfileSticks = (Boolean) p.value;
                }
                if (key.equals("drawComponents")) {
                    m_bDrawComponents = (Boolean) p.value;
                }
                if (key.equals("drawCenterValues")) {
                    m_bDrawCenterValues = (Boolean) p.value;
                }
                if (key.equals("componentsAreStaggered")) {
                    m_bComponentsAreStaggered = (Boolean) p.value;
                }
                if (key.equals("writeValuesToFile")) {
                    boolean newValue = (Boolean) p.value;
                    if (newValue != m_bWriteValuesToFile) {
                        m_bWriteValuesToFile = newValue;
                        if (m_bWriteValuesToFile) {

                            if (m_outputPrintStream == null) {
                                try {
                                    m_outputPrintStream = new PrintStream("vectors.txt");
                                } catch (Exception e) {
                                    e.printStackTrace();
                                    m_bWriteValuesToFile = false;
                                }
                            }

                        } else {
                            if (m_outputPrintStream != null) {
                                m_outputPrintStream.close();
                                m_outputPrintStream = null;
                            }

                        }

                    }
                }
                if (key.equals("symbolType")) {
                    // Check each name.
                    for (int i = 0; i < m_symbolTypes.length; i++) {
                        String name = ((String) p.value).trim().toLowerCase();
                        if (name.equals(m_symbolTypes[i].toLowerCase())) {
                            m_symbolIndex = i;
                            break;
                        }
                    }
                }
                if (key.equals("color")) {
                    ColorProperty cp = (ColorProperty) p;
                    m_color = cp.getColorArray();
                }
                if (key.equals("xOutputMin")) {
                    m_xOutputMin = (Double) p.value;
                }
                if (key.equals("xOutputMax")) {
                    m_xOutputMax = (Double) p.value;
                }
                if (key.equals("xOutputSpacing")) {
                    m_xOutputSpacing = (Double) p.value;
                }
                if (key.equals("yOutputMin")) {
                    m_yOutputMin = (Double) p.value;
                }
                if (key.equals("yOutputMax")) {
                    m_yOutputMax = (Double) p.value;
                }
                if (key.equals("yOutputSpacing")) {
                    m_yOutputSpacing = (Double) p.value;
                }
                if (key.equals("zOutputMin")) {
                    m_zOutputMin = (Double) p.value;
                }
                if (key.equals("zOutputMax")) {
                    m_zOutputMax = (Double) p.value;
                }
                if (key.equals("zOutputSpacing")) {
                    m_zOutputSpacing = (Double) p.value;
                }
            }
            if (obj instanceof String) {
                String value = (String) obj;
                setProperty(key, value);
            }
        }

        updateDrawOffsets();
        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
    /**
     * Set a single property.
     */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("scale")) {
            m_scale = Double.parseDouble(value);
        }
        if (key.equals("zScale")) {
            m_zScale = Double.parseDouble(value);
        }
        if (key.equals("usingUV")) {
            m_bUsingUV = Boolean.parseBoolean(value);
        }
        if (key.equals("drawProfileSticks")) {
            m_bDrawProfileSticks = Boolean.parseBoolean(value);
        }
        if (key.equals("drawComponents")) {
            m_bDrawComponents = Boolean.parseBoolean(value);
        }
        if (key.equals("drawCenterValues")) {
            m_bDrawCenterValues = Boolean.parseBoolean(value);
        }
        if (key.equals("componentsAreStaggered")) {
            m_bComponentsAreStaggered = Boolean.parseBoolean(value);
        }
        if (key.equals("writeValuesToFile")) {
            boolean newValue = Boolean.parseBoolean(value);
            if (newValue != m_bWriteValuesToFile) {
                m_bWriteValuesToFile = newValue;
                if (m_bWriteValuesToFile) {

                    if (m_outputPrintStream == null) {
                        try {
                            m_outputPrintStream = new PrintStream("vectors.txt");
                        } catch (Exception e) {
                            e.printStackTrace();
                            m_bWriteValuesToFile = false;
                        }
                    }

                } else {
                    if (m_outputPrintStream != null) {
                        m_outputPrintStream.close();
                        m_outputPrintStream = null;
                    }

                }

            }
        }
        if (key.equals("seriesNums")) {
            m_seriesNums = AsciiIO.parseLineAsInts(value, ", \t");
        }
        if (key.equals("dimNums")) {
            m_dimNums = AsciiIO.parseLineAsInts(value, ", \t");
        }
        if (key.equals("symbolType")) {
            // Check each name.
            for (int i = 0; i < m_symbolTypes.length; i++) {
                String name = value.trim().toLowerCase();
                if (name.equals(m_symbolTypes[i].toLowerCase())) {
                    m_symbolIndex = i;
                    break;
                }
            }
        }
        if (key.equals("color")) {
            m_color = ColorProperty.getColorArray(value);
        }
        if (key.equals("xOutputMin")) {
            m_xOutputMin = Double.parseDouble(value);
        }
        if (key.equals("xOutputMax")) {
            m_xOutputMax = Double.parseDouble(value);
        }
        if (key.equals("xOutputSpacing")) {
            m_xOutputSpacing = Double.parseDouble(value);
        }
        if (key.equals("yOutputMin")) {
            m_yOutputMin = Double.parseDouble(value);
        }
        if (key.equals("yOutputMax")) {
            m_yOutputMax = Double.parseDouble(value);
        }
        if (key.equals("yOutputSpacing")) {
            m_yOutputSpacing = Double.parseDouble(value);
        }
        if (key.equals("zOutputMin")) {
            m_zOutputMin = Double.parseDouble(value);
        }
        if (key.equals("zOutputMax")) {
            m_zOutputMax = Double.parseDouble(value);
        }
        if (key.equals("zOutputSpacing")) {
            m_zOutputSpacing = Double.parseDouble(value);
        }

        updateDrawOffsets();
    }

//------------------------------------------------------------------------------
    /**
     * Reread the DataSeries children and set the DataSeries and dimension for
     * Time, X, Y, Z, U, V, and W.
     */
    public void resetDataDims() {

        // Check that the member variables have been initialized.
        if (!m_bAnimationInitialized) {
            int status = readFromChildren();
            if (status < 0) {
                return;
            }
        }

        int dsIndex = 0;
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            if (dsIndex >= 7) {
                break;
            }
            GrObject gro = (GrObject) iter.next();
            if (gro instanceof DataSeries) {
                DataSeries series = (DataSeries) gro;
                if (gro instanceof DoubleDataSeries) {
                    m_dataArrays[dsIndex] =
                            ((DoubleDataSeries) series).getArray();
                } else {
                    m_dataArrays[dsIndex] = new double[0];
                }
                m_numDims[dsIndex] = series.getNumDimensions();
                m_seriesArray[dsIndex] = series;

                dsIndex++;
            }
        }

        for (int i = dsIndex; i < 7; i++) {
            m_dataArrays[i] = null;
            m_numDims[i] = 0;
            m_seriesArray[i] = null;
            m_dimScales[i] = 1;
        }

        for (int i = 0; i < 7; i++) {
            // Skip if the series is null.
            DataSeries series = m_seriesArray[m_seriesNums[i]];
            if (series == null) {
                m_dimScales[i] = 1;
                continue;
            }

            // Skip if the dimension is null.
            DataDimension dim = series.getDimension(m_dimNums[i]);
            if (dim == null) {
                m_dimScales[i] = 1;
                continue;
            }

            // Store the scale factor for the dimension.
            GrHashList dimProps = dim.getProperties();
            GrProperty p = (GrProperty) dimProps.get("unitScale");
            double unitScale = (Double) p.value;
//            p = (GrProperty)dimProps.get("scale");
//            float scale = (Float)p.value;
//            scale *= unitScale;
            m_dimScales[i] = unitScale;
        }

    }

//------------------------------------------------------------------------------
    /**
     * Recompute the offsets where output will be drawn.
     */
    public synchronized void updateDrawOffsets() {

        if (m_bath == null || m_si3dOut == null) {
            createVectorAnimation();
            if (m_bath == null || m_si3dOut == null) {
                return;
            }
        }

        // Get grid spacing information from the Bathymetry object.
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty) bathProps.get("xSpacing");
        double xSpacing = (Double) p.value;
        p = (GrProperty) bathProps.get("ySpacing");
        double ySpacing = (Double) p.value;
        p = (GrProperty) bathProps.get("zSpacing");
        double zSpacing = (Double) p.value;

        // Find the dimensions of the grid and the number of wet 2D points.
        DataSeries bathSeries = m_bath.getBathymetrySeries();
        int numI = (int) Math.round(bathSeries.getMax(0)) + 1;
        int numJ = (int) Math.round(bathSeries.getMax(1)) + 1;
        int numLayers = (int) Math.round(bathSeries.getMax(2));
//        int numWet2Dpoints = bathSeries.getNumPoints();
//        float[] depthArray = m_bath.getAverageDepthArray();

        // Get array indexing information from the Bathymetry object.
        int[] indexArray = m_bath.get3DIndexArray();
        // Get the rectangular array that holds -1 at dry points and the
        // index into the variables array at wet points.
        int[] lookupArray = m_bath.get3DLookupArray(indexArray);

        double xMax = bathSeries.getMax(0) * xSpacing;
        double yMax = bathSeries.getMax(1) * ySpacing;
        double zMax = bathSeries.getMax(2) * zSpacing;

        if (m_xOutputMin < 0) {
            m_xOutputMin = 0;
        }
        if (m_yOutputMin < 0) {
            m_yOutputMin = 0;
        }
        if (m_zOutputMin < 0) {
            m_zOutputMin = 0;
        }
        if (m_xOutputMax > xMax) {
            m_xOutputMax = xMax;
        }
        if (m_yOutputMax > yMax) {
            m_yOutputMax = yMax;
        }
        if (m_zOutputMax > zMax) {
            m_zOutputMax = zMax;
        }
        if (m_xOutputSpacing < xSpacing) {
            m_xOutputSpacing = xSpacing;
        }
        if (m_yOutputSpacing < ySpacing) {
            m_yOutputSpacing = ySpacing;
        }
        if (m_zOutputSpacing < zSpacing) {
            m_zOutputSpacing = zSpacing;
        }

        double extraX = 0.15 * m_contentTrans.getWidth();
        double extraY = 0.15 * m_contentTrans.getHeight();
        double xDrawMin = m_contentTrans.getMinX() - extraX;
        double xDrawMax = m_contentTrans.getMaxX() + extraX;
        double yDrawMin = m_contentTrans.getMinY() - extraY;
        double yDrawMax = m_contentTrans.getMaxY() + extraY;
        if (m_xOutputMin > xDrawMin) {
            xDrawMin = m_xOutputMin;
        }
        if (m_yOutputMin > yDrawMin) {
            yDrawMin = m_yOutputMin;
        }
        if (m_xOutputMax < xDrawMax) {
            xDrawMax = m_xOutputMax;
        }
        if (m_yOutputMax < yDrawMax) {
            yDrawMax = m_yOutputMax;
        }

        // Look for a child series in an object container that specifies
        // the X-Y locations
        m_xyLocations = null;
        Iterator iter = m_children.iterator();
        outer:
        while (iter.hasNext()) {
            GrObject gro = (GrObject) iter.next();
            if (gro instanceof GrObjectContainer) {
                Enumeration enumeration = gro.children();
                while (enumeration.hasMoreElements()) {
                    GrObject child = (GrObject) enumeration.nextElement();
                    if (DoubleDataSeries.countDims(child, 2)) {
                        m_xyLocations = ((DoubleDataSeries) child).getArray();
                        break outer;
                    }
                }
            }
        }

        if (m_xyLocations == null) {
            ArrayList<Double> xyLocationList = new ArrayList<Double>();

            for (double xVal = (double) xDrawMin; xVal <= xDrawMax;
                    xVal += m_xOutputSpacing) {

                for (double yVal = (double) yDrawMin; yVal <= yDrawMax;
                        yVal += m_yOutputSpacing) {

                    int currentI = (int) Math.round(xVal / xSpacing);
                    int currentJ = (int) Math.round(yVal / ySpacing);
                    int currentK = 1;

                    int lookupIndex = -1;

                    // Find out if the cell is wet.
                    if (currentI >= 0 && currentI < numI && currentJ >= 0 &&
                            currentJ < numJ && currentK >= 0 &&
                            currentK < numLayers) {

                        lookupIndex = lookupArray[currentI * numLayers +
                                currentJ * numLayers * numI + currentK];
                        if (lookupIndex >= 0) {
                            xyLocationList.add(xVal);
                            xyLocationList.add(yVal);
                        }
                    }
                }
            }

            m_xyLocations = new double[xyLocationList.size()];
            int index = 0;
            for (double dVal : xyLocationList) {
                m_xyLocations[index++] = dVal;
            }
        }

        createProfileSticks();

        ArrayList<Integer> outputLocationList = new ArrayList();

        for (int i = 0; i < m_xyLocations.length; i += 2) {
            double xVal = m_xyLocations[i];
            double yVal = m_xyLocations[i + 1];

            int lastK = Integer.MIN_VALUE;
            for (double zVal = m_zOutputMin; zVal <= m_zOutputMax;
                    zVal += m_zOutputSpacing) {

                int currentI = (int) Math.round(xVal / xSpacing);
                int currentJ = (int) Math.round(yVal / ySpacing);
                int currentK = (int) Math.round(zVal / zSpacing);
                if (currentK < 1) {
                    currentK = 1;
                }
                if (currentK == lastK) {
                    continue;
                }
                lastK = currentK;

                int lookupIndex = -1;

                // Find out if the cell is wet.
                if (currentI >= 0 && currentI < numI && currentJ >= 0 &&
                        currentJ < numJ && currentK >= 0 &&
                        currentK < numLayers) {

                    lookupIndex = lookupArray[currentI * numLayers +
                            currentJ * numLayers * numI + currentK];
                    if (lookupIndex >= 0) {
                        outputLocationList.add(lookupIndex);
                    }
                }
            }
        }

        m_outputLocations = new int[outputLocationList.size()];
        int index = 0;
        for (Integer val : outputLocationList) {
            m_outputLocations[index] = val;
            index++;
        }
    }

//------------------------------------------------------------------------------
    /**
     * Recompute the offsets where output will be drawn.
     */
    public synchronized void createProfileSticks() {

        DataSeries bathSeries = m_bath.getBathymetrySeries();
        double[] depthArray = m_bath.getAverageDepthArray();

        // Get grid spacing information from the Bathymetry object.
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty) bathProps.get("xSpacing");
        double xSpacing = (Double) p.value;
        p = (GrProperty) bathProps.get("ySpacing");
        double ySpacing = (Double) p.value;

        m_profileSticks = new double[m_xyLocations.length * 3];

        for (int i = 0; i < m_xyLocations.length; i += 2) {
            double xVal = m_xyLocations[i];
            double yVal = m_xyLocations[i + 1];

            m_profileSticks[i * 3] = xVal - xSpacing;
            m_profileSticks[i * 3 + 1] = yVal - ySpacing;
            m_profileSticks[i * 3 + 2] = 0;
            m_profileSticks[i * 3 + 3] = xVal - xSpacing;
            m_profileSticks[i * 3 + 4] = yVal - ySpacing;
            m_profileSticks[i * 3 + 5] = 0;

            // Find the lookup index for surface and bottom heights.
            int lookup2D = -1;
            int currentI = (int) Math.round(xVal / xSpacing);
            int currentJ = (int) Math.round(yVal / ySpacing);
            int ijIndex = currentI + (currentJ * m_numI);
            if (ijIndex < 0 || ijIndex > m_2DLookupArray.length) {
                // Write (0, 0, 0) and move to the next time step.
                break;
            }

            // Check that the data is valid and the index is a wet point.
            if (m_2DLookupArray != null) {
                lookup2D = m_2DLookupArray[ijIndex];
            }

            if (lookup2D < 0) {
                // Write (0, 0, 0) and move to the next time step.
                break;
            }

            // Retrieve the depth at the current location from the array.
            double depth = depthArray[lookup2D];

            m_profileSticks[i * 3 + 5] = depth * m_zScale;
        }

    }

//------------------------------------------------------------------------------
    /**
     *
     */
    public synchronized int readFromChildren() {

        boolean bChanged = false;
        int status = 0;

        Bathymetry oldBath = m_bath;
        Si3dOutput oldSi3dOut = m_si3dOut;
        m_bath = null;
        m_si3dOut = null;

        int numSeries = 0;
        Iterator iter = m_children.iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject) iter.next();
            if (gro instanceof Bathymetry) {
                m_bath = (Bathymetry) (gro);
                if (gro != oldBath) {
                    bChanged = true;
                }
            }
            if (gro instanceof Si3dOutput) {
                m_si3dOut = (Si3dOutput) gro;
                if (gro != oldSi3dOut) {
                    bChanged = true;
                }
            }
            if (gro instanceof DataSeries) {
                if (gro != m_seriesArray[numSeries]) {
                    bChanged = true;
                }
                numSeries++;
            }
        }

        if (numSeries < 3) {
            status = -1;
        } else {
            if (bChanged) {
                status = 1;
            }
        }

        if (m_bath == null || m_si3dOut == null) {
            return -2;
        }

        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty) bathProps.get("xSpacing");
        m_xSpacing = (Double) p.value;
        p = (GrProperty) bathProps.get("ySpacing");
        m_ySpacing = (Double) p.value;
        p = (GrProperty) bathProps.get("zSpacing");
        m_zSpacing = (Double) p.value;

//        double startTime = m_si3dOut.getJulianStartTime();
        GrHashList si3dOutProps = m_si3dOut.getProperties();
        p = (GrProperty) si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long) p.value;
//        double timeStepInc = ((double)tsInSec) / 86400;

        m_3DIndexArray = m_bath.get3DIndexArray();
        m_3DLookupArray = m_bath.get3DLookupArray(m_3DIndexArray);
        DataSeries bathSeries = m_bath.getBathymetrySeries();

        // For every wet index, create an array of south indexes and an
        // array of west indexes.
        m_numI = (int) Math.round(bathSeries.getMax(0)) + 1;
        m_numJ = (int) Math.round(bathSeries.getMax(1)) + 1;
        m_numK = (int) Math.round(bathSeries.getMax(2));

        m_3DVarsSeries = (RafDataSeries) m_si3dOut.getContainerSeries("3-D Variables");

//        int numTimeSteps = m_3DVarsSeries.getCount() /
//            (m_3DIndexArray.length / 3);

        RafDataSeries vars2dSeries = (RafDataSeries) m_si3dOut.getContainerSeries(
                "2-D x-y Variables");
        if (vars2dSeries != null) {
//            m_2DVarsArray = vars2dSeries.getArray();

            DataDimension dim = vars2dSeries.getDimension(0);
            if (dim != null) {
                // Store the scale factor for the dimension.
                GrHashList dimProps = dim.getProperties();
                p = (GrProperty) dimProps.get("unitScale");
                m_wseScale = (Double) p.value;
            }
        }
        int[] index2D = m_bath.get2DIndexArray();
        m_2DLookupArray = m_bath.get2DLookupArray(index2D);

        return status;
    }

//------------------------------------------------------------------------------
    /**
     *
     */
    public synchronized void createVectorAnimation() {

        int status = readFromChildren();
        boolean bChanged = (status > 0) ? true : false;

        // Return if there weren't three dataseries children.
        if (status == -1) {
            return;
        }

        if (!bChanged && m_bAnimationInitialized) {
            return;
        }

        m_bAnimationInitialized = false;

        if (m_bath == null || m_si3dOut == null) {
            return;
        }

        int numWetPoints = m_3DIndexArray.length / 3;
        m_westArray = new int[numWetPoints];
        m_southArray = new int[numWetPoints];
        m_eastArray = new int[numWetPoints];
        m_northArray = new int[numWetPoints];
        for (int i = 0; i < numWetPoints; i++) {
            int iIndex = m_3DIndexArray[i * 3];
            int jIndex = m_3DIndexArray[i * 3 + 1];
            int kIndex = m_3DIndexArray[i * 3 + 2];
            if (iIndex > 0) {
                m_westArray[i] = m_3DLookupArray[(iIndex - 1) * m_numK +
                        jIndex * m_numK * m_numI + kIndex];
            } else {
                m_westArray[i] = -1;
            }
            if (jIndex > 0) {
                m_southArray[i] = m_3DLookupArray[iIndex * m_numK +
                        (jIndex - 1) * m_numK * m_numI + kIndex];
            } else {
                m_southArray[i] = -1;
            }
            if (jIndex < (m_numJ - 1)) {
                m_northArray[i] = m_3DLookupArray[iIndex * m_numK +
                        (jIndex + 1) * m_numK * m_numI + kIndex];
            } else {
                m_northArray[i] = -1;
            }
            if (iIndex < (m_numI - 1)) {
                m_eastArray[i] = m_3DLookupArray[(iIndex + 1) * m_numK +
                        jIndex * m_numK * m_numI + kIndex];
            } else {
                m_eastArray[i] = -1;
            }
        }

        if (m_3DVarsSeries == null) {
            return;
        }

        int[] array1 = {0, 1, 1, 1, 2, 2, 2};
        m_seriesNums = array1;
        int[] array2 = {0, 0, 1, 2, 0, 1, 2};
        m_dimNums = array2;

        m_bAnimationInitialized = true;

        resetDataDims();
    }

//------------------------------------------------------------------------------
    /**
     *
     */
    public synchronized void reinitialize() {

        Iterator iter = new ArrayList(m_children).iterator();
        while (iter.hasNext()) {
            GrObject gro = (GrObject) iter.next();
            if (gro instanceof DataSeries) {
                remove(gro);
            }
        }

        m_bAnimationInitialized = false;
        m_bath = null;
        m_si3dOut = null;

        createChildSeries();
        updateDrawOffsets();
    }

//------------------------------------------------------------------------------
    /**
     *
     */
    public synchronized void createChildSeries() {

        if (m_bath == null || m_si3dOut == null) {
            readFromChildren();
            if (m_bath == null || m_si3dOut == null) {
                return;
            }
        }

        m_3DIndexArray = m_bath.get3DIndexArray();
        m_3DVarsSeries = (RafDataSeries) m_si3dOut.getContainerSeries("3-D Variables");
        if (m_3DVarsSeries == null) {
            return;
        }

        int numTimeSteps = m_3DVarsSeries.getCount() /
                (m_3DIndexArray.length / 3);

        double startTime = m_si3dOut.getJulianStartTime();
        GrHashList si3dOutProps = m_si3dOut.getProperties();
        GrProperty p = (GrProperty) si3dOutProps.get("timeStepInSeconds");
        Long tsInSec = (Long) p.value;
        double timeStepInc = ((double) tsInSec) / 86400;

        // Create the Vector Animation and populate it with Time, (X, Y, Z)
        // and (U, V, W) arrays.
        // Round to the nearest minute.
        double[] timesArray = new double[numTimeSteps];
        for (int i = 0; i < numTimeSteps; i++) {
            timesArray[i] = startTime + (i * timeStepInc) + 0.00034722222222222;
        }
        DoubleDataSeries timesSeries = new DoubleDataSeries(timesArray, 1);
        timesSeries.setTitle("Times");
        timesSeries.addObserver(this);
        m_children.add(timesSeries);
        m_children.add(m_bath.get3DPositionSeries(m_3DIndexArray));
        m_children.add(m_3DVarsSeries);

    }

//------------------------------------------------------------------------------
    /**
     * Hide or show this series.
     * @param  bShow    True if this curve should be drawn.
     */
    public void show(boolean bShow) {
        m_bShow = bShow;
    }

//------------------------------------------------------------------------------
    /**
     * Sets the OpenGL objects to use for accessing drawing commands.
     * @param newgld   The OpenGL GLAutoDrawable component.
     * @param newgl    The OpenGL object containing the drawing commands.
     * @param newglu   The OpenGL utilites object.
     */
    public void setOpenGL(GLAutoDrawable newgld, GL newgl, GLU newglu) {
        gld = newgld;
        gl = newgl;
        glu = newglu;
    }

//------------------------------------------------------------------------------
    /**
     * Issue the OpenGL commands to draw this series.
     */
    public void createWaterSurface() {

        m_2DVarsSeries = (RafDataSeries) m_si3dOut.getContainerSeries(
                "2-D x-y Variables");
        int[] indexArray = m_bath.get2DIndexArray();
        DoubleDataSeries pos2dSeries = m_bath.get2DPositionSeries(indexArray);

        if (m_2DVarsSeries == null || pos2dSeries == null) {
            return;
        }

        int numVarDims = m_2DVarsSeries.getNumDimensions();
        int numPosDims = pos2dSeries.getNumDimensions();

        int numVarPoints = m_2DVarsSeries.getCount();
        int numPosPoints = pos2dSeries.getCount();

        double[] pos2dArray = pos2dSeries.getArray();
//        float[] vars2dArray = m_2DVarsSeries.getArray();
        double[] newArray = new double[numVarPoints * 4];

        for (int i = 0; i < numVarPoints; i++) {
            try {
                newArray[i * 4] = m_dataArrays[m_seriesNums[0]][i / numPosPoints];
                newArray[i * 4 + 1] =
                        pos2dArray[(i % numPosPoints) * numPosDims] - 1;
                newArray[i * 4 + 2] =
                        pos2dArray[(i % numPosPoints) * numPosDims + 1] - 1;
//            newArray[i * 4 + 3] = vars2dArray[i * numVarDims];
                newArray[i * 4 + 3] = getVar2D(i * numVarDims);
            } catch (Exception e) {
                System.err.println("Exception at i = " + i);
                e.printStackTrace();
                return;
            }
        }

        DoubleDataSeries newSeries = new DoubleDataSeries(newArray, 4);
        GlRectAnimation rectAnim = new GlRectAnimation();
        rectAnim.add(newSeries);
        add(rectAnim);
    }

//------------------------------------------------------------------------------
    /**
     * Issue the OpenGL commands to draw this series.
     */
    public synchronized void drawGL() {

        // Before drawing, check if this object is supposed to be shown.
        if (!m_bShow) {
            return;
        }

        m_errorString = null;

        // For calculating WSE from 2D series.
        if (m_bath == null) {
            m_errorString = "No Bathymetry";

            return;
        }

        // Check that the member variables have been initialized.
        if (!m_bAnimationInitialized) {
            createVectorAnimation();
            if (!m_bAnimationInitialized) {
                m_errorString = "Not Initialized";

                return;
            }
        }

        if (!m_contentTrans.equals(m_lastContentTrans) &&
                m_contentTrans.getWidth() != 0 && m_contentTrans.getHeight() != 0) {

            m_outputLocations = null;
            m_lastContentTrans = (Rectangle2D) m_contentTrans.clone();
        }

        if (m_outputLocations == null) {
            updateDrawOffsets();
        }

        try {
            if (m_seriesArray[m_seriesNums[1]].getCount() !=
                    m_seriesArray[m_seriesNums[2]].getCount()) {

                m_errorString = "X and Y counts not equal";

                return;
            }

            if (m_seriesArray[m_seriesNums[4]].getCount() !=
                    m_seriesArray[m_seriesNums[5]].getCount()) {

                m_errorString = "U and V counts not equal";

                return;
            }

            if (m_seriesArray[m_seriesNums[1]].getCount() !=
                    m_seriesArray[m_seriesNums[3]].getCount()) {
                m_dimNums[3] = -1;
            }

            if (m_seriesArray[m_seriesNums[4]].getCount() !=
                    m_seriesArray[m_seriesNums[6]].getCount()) {
                m_dimNums[6] = -1;
            }
        } catch (NullPointerException e) {
            m_errorString = "Series array not found";
            return;
        } catch (ArrayIndexOutOfBoundsException e) {
            m_errorString = "Series array not found";
            return;
        }

        if (m_bDrawProfileSticks) {
            drawProfileSticks();
        }

        // Save the previous line color and pattern.
        gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

        if (m_symbolIndex >= 4) {
            gl.glPushAttrib(gl.GL_POLYGON_BIT);
            gl.glPolygonMode(gl.GL_FRONT, gl.GL_FILL);
            gl.glCullFace(gl.GL_BACK);
            gl.glEnable(gl.GL_CULL_FACE);
        }

        double wScale = m_scale / Math.abs(m_zScale);

//////////   Work here on calculating the number of iterations for the
//// draw loop and the offsets for the T, Loc, and Vec arrays.
/////////////////
        long locOffset = 0;
        long vecOffset = 0;
        long endOffset = 0;
        if (m_seriesArray[m_seriesNums[0]].getCount() !=
                m_seriesArray[m_seriesNums[1]].getCount() ||
                m_seriesArray[m_seriesNums[4]].getCount() !=
                m_seriesArray[m_seriesNums[1]].getCount()) {

            // Not a one-to-one relationship between times, locations,
            // and vectors.
//            if ((m_seriesArray[m_seriesNums[0]].getCount() *
//                m_seriesArray[m_seriesNums[1]].getCount())
//                != m_seriesArray[m_seriesNums[4]].getCount()) {

//                m_errorString = "UVW count not T*XYZ";

            // Number of times multiplied by number of locations is
            // not equal to number of vectors, so don't draw anything.
//                return;
//            }
//            else {
            // Number of times multiplied by number of locations is
            // equal to number of vectors, so proceed with drawing.
            vecOffset = m_seriesArray[m_seriesNums[0]].getBegin() *
                    m_seriesArray[m_seriesNums[1]].getCount();
            endOffset = m_seriesArray[m_seriesNums[1]].getCount() - 1;
//            }
        } else {
            // One-to-one relationship between times, locations, and vectors.
            locOffset = m_seriesArray[m_seriesNums[0]].getBegin();
            vecOffset = 0;
            endOffset = m_seriesArray[m_seriesNums[0]].getEnd();
        }

        if (m_2DVarsSeries == null) {
            m_2DVarsSeries = (RafDataSeries) m_si3dOut.getContainerSeries("2-D x-y Variables");
        }

//        m_2DVarsArray = m_2DVarsSeries.getArray();

        // Set up disk caching if the 2D series is in a random access file.
        if (m_2DVarsSeries instanceof RafDataSeries) {
            int numWetPoints = m_bath.getBathymetrySeries().getCount();
            int numDims = m_2DVarsSeries.getNumDimensions();
            int rafCacheSize = numWetPoints * numDims;
            if (rafCacheSize != m_rafCacheSize2D) {
                m_rafCacheSize2D = rafCacheSize;
                ((RafDataSeries) m_2DVarsSeries).createCacheSegments(1,
                        m_rafCacheSize2D);
            }
            long rafCacheOffset = m_seriesArray[m_seriesNums[0]].getBegin() * numWetPoints * numDims;
            if (rafCacheOffset != m_rafCacheOffset2D) {
                m_rafCacheOffset2D = rafCacheOffset;
                ((RafDataSeries) m_2DVarsSeries).fillCacheSegment(0,
                        m_rafCacheOffset2D);
            }
        }

//        synchronized (m_3DVarsSeries) {
        // Set up disk caching if the 3D series is in a random access file.
        if (m_3DVarsSeries instanceof RafDataSeries) {
            int numWetPoints = m_3DIndexArray.length / 3;
            int numDims = m_3DVarsSeries.getNumDimensions();
            int[] existingCacheSizes = ((RafDataSeries) m_3DVarsSeries).getCacheSizes();
            int rafCacheSize = numWetPoints * numDims;
            if (existingCacheSizes.length == 0 ||
                    rafCacheSize != existingCacheSizes[0]) {

                ((RafDataSeries) m_3DVarsSeries).createCacheSegments(1,
                        rafCacheSize);
            }
            long rafCacheOffset = vecOffset * numDims;
            long[] existingCacheOffsets = ((RafDataSeries) m_3DVarsSeries).getCacheOffsets();
            if (existingCacheOffsets.length == 0 ||
                    rafCacheOffset != existingCacheOffsets[0]) {

                ((RafDataSeries) m_3DVarsSeries).fillCacheSegment(0,
                        rafCacheOffset);
            }
        }

        DataSeries bathSeries = m_bath.getBathymetrySeries();
        double[] depthArray = m_bath.getAverageDepthArray();

        int timeIndex = m_seriesArray[m_seriesNums[0]].getBegin();
        double newFileTime = (((DoubleDataSeries) m_seriesArray[m_seriesNums[0]]).getArray())[timeIndex];

        for (int i = 0; i < m_outputLocations.length; i++) {
            int offset = m_outputLocations[i];
            int index = offset * m_numDims[m_seriesNums[1]] + m_dimNums[1];
            if (index >= m_dataArrays[m_seriesNums[1]].length) {
                System.err.println("index = " + index + ", length = " +
                        m_dataArrays[m_seriesNums[1]].length);
                break;
            }
            double x = m_dataArrays[m_seriesNums[1]][offset * m_numDims[m_seriesNums[1]] + m_dimNums[1]];
            x *= m_dimScales[1];
            x -= m_xSpacing;
            double y = m_dataArrays[m_seriesNums[2]][offset * m_numDims[m_seriesNums[2]] + m_dimNums[2]];
            y *= m_dimScales[2];
            y -= m_ySpacing;
            double z = 0;
            if (m_dimNums[3] >= 0) {
                z = m_dataArrays[m_seriesNums[3]][offset * m_numDims[m_seriesNums[3]] + m_dimNums[3]];
                z *= m_dimScales[3];
            }

            double bottomAdjustment = 0;
            double surfaceHeightAdjustment = 0;
            // Find the lookup index for surface and bottom heights.
            int lookup2D = -1;
            int currentI = m_3DIndexArray[offset * 3];
            int currentJ = m_3DIndexArray[offset * 3 + 1];
            int currentK = m_3DIndexArray[offset * 3 + 2];
            int ijIndex = currentI + (currentJ * m_numI);
            if (ijIndex < 0 || ijIndex > m_2DLookupArray.length) {
                // Write (0, 0, 0) and move to the next time step.
                break;
            }

            // Check that the data is valid and the index is a wet point.
            if (m_2DLookupArray != null) {
                lookup2D = m_2DLookupArray[ijIndex];
            }

            if (lookup2D < 0) {
                // Write (0, 0, 0) and move to the next time step.
                break;
            }

            // Retrieve the depth at the current location from the array.
            double depth = depthArray[lookup2D];

            // Calculate the bottom adjustment in world units.
            // It is zero if K is not a bottom layer.
            double bottomLayerThickness = depth % m_zSpacing;
            if (bottomLayerThickness <= 0) {
                bottomLayerThickness = m_zSpacing;
            }
            if (z > (depth - bottomLayerThickness)) {
                bottomAdjustment = m_zSpacing - bottomLayerThickness;
//System.out.println("Bottom " + currentI + ", " + currentJ + ", " + currentK
//    + " -- " + z + " -- " + depth);
            }

            DataDimension dim = m_2DVarsSeries.getDimension(0);
            if (dim != null) {
                // Store the scale factor for the dimension.
                GrHashList dimProps = dim.getProperties();
                GrProperty p = (GrProperty) dimProps.get("unitScale");
                m_wseScale = (Double) p.value;
            }

            // Calculate the surface adjustment in world units.
            // It is zero if K is not a surface layer.
            if (z <= m_zSpacing) {
                if (lookup2D >= 0 && m_2DVarsSeries != null) {

                    int ts = m_seriesArray[m_seriesNums[0]].getBegin();
                    try {
//////// Remove this to avoid tidal bouncing on filtered profiles
                        surfaceHeightAdjustment = getVar2D(lookup2D + (ts * bathSeries.getNumPoints())) * m_wseScale;
                    } catch (Exception e) {
                        System.err.println("Exception at ts = " + ts);
                        e.printStackTrace();
                    }
                }
            }

            z -= (m_zSpacing * 0.5);
            z -= (bottomAdjustment * 0.5);
            z -= (surfaceHeightAdjustment * 0.5);
            if (m_zScale != 0) {
                z *= m_zScale;
            }

            // Set the color.
            if (m_colorScheme == null) {
                gl.glColor3fv(m_color, 0);
            } else {
                gl.glColor3fv(m_colorScheme.getColor(z), 0);
            }

            if (m_zScale == 0) {
                z = 0;
            }

            double uvTheta = 0;
            double uwTheta = 0;
            double len;
            if (m_bUsingUV) {
                double u = 0;
                double v = 0;
                double uEast = getVar3D(offset + vecOffset, 4);
                double vNorth = getVar3D(offset + vecOffset, 5);

                if (m_bComponentsAreStaggered) {
                    // Assumes the U and V components at this point are
                    // staggered.  U is centered along the east cell face.
                    // V is centered along the west cell face.  The U
                    // value at the center of the cell is the average of
                    // the east (this point) and west (neighboring point)
                    // sides of the cell.  The V value at the center is the
                    // average of the north and south values.

                    double uWest = 0;
                    double vSouth = 0;
                    boolean bWestWet = false;
                    boolean bSouthWet = false;
                    boolean bNorthWet = true;
                    boolean bEastWet = true;

                    if (offset >= m_westArray.length) {
                        break;
                    }

                    if (m_westArray[offset] >= 0) {
                        uWest = getVar3D(m_westArray[offset] + vecOffset, 4);
                        bWestWet = true;
                    }
                    if (m_southArray[offset] >= 0) {
                        vSouth = getVar3D(m_southArray[offset] + vecOffset, 5);
                        bSouthWet = true;
                    }
                    if (m_northArray[offset] < 0) {
                        bNorthWet = false;
                    }
                    if (m_eastArray[offset] < 0) {
                        bEastWet = false;
                    }

                    if (bWestWet && bEastWet) {
                        u = (uEast + uWest) / 2;
                    }
                    if (bWestWet && !bEastWet) {
                        u = uWest * 0.75;
                    }
                    if (!bWestWet && bEastWet) {
                        u = uEast * 0.75;
                    }
                    if (bSouthWet && bNorthWet) {
                        v = (vNorth + vSouth) / 2;
                    }
                    if (bSouthWet && !bNorthWet) {
                        v = vSouth * 0.75;
                    }
                    if (!bSouthWet && bNorthWet) {
                        v = vNorth * 0.75;
                    }
                } else {
                    // Assumes the U and V components at this point are both
                    // located at the center.
                    u = uEast;
                    v = vNorth;
                }

                double w = 0;
                if (m_dimNums[6] >= 0) {
                    w = getVar3D(offset + vecOffset, 6);
                    w *= m_zScale;
                }

                uvTheta = Math.atan2(v, u);
                uvTheta *= (180 / Math.PI);
                double uvLen = Math.sqrt((u * u) + (v * v));
                if (m_dimNums[6] >= 0) {
                    uwTheta = Math.atan2(w, uvLen);
                    uwTheta *= (180 / Math.PI);
                    len = Math.sqrt((u * u) + (v * v) + (w * w));
                } else {
                    len = uvLen;
                }

                if (m_bWriteValuesToFile) {
                    if (fileTime != newFileTime) {
                        double fileW = getVar3D(offset + vecOffset, 6);
                        String sVal = "" + newFileTime;
                        sVal += (" " + (float) x + " " + (float) y + " " + (float) z);
                        sVal += (" " + (float) u + " " + (float) v + " " + (float) fileW);
                        m_outputPrintStream.println(sVal);
                    }
                }

///////////////////
                if (m_bDrawComponents) {
                    gl.glPushMatrix();
                    gl.glTranslated(x, y, z);

                    gl.glBegin(gl.GL_LINES);
                    gl.glVertex2d(m_xSpacing / 2, 0);
                    gl.glVertex2d((m_xSpacing / 2) + (uEast * m_scale), 0);
                    gl.glVertex2d(0, m_ySpacing / 2);
                    gl.glVertex2d(0, (m_ySpacing / 2) + (vNorth * m_scale));
                    gl.glEnd();

                    gl.glPopMatrix();
                }
//////////////////
            } else {
                uvTheta = getVar3D(offset + vecOffset, 4);
                len = getVar3D(offset + vecOffset, 5);
                if (m_dimNums[6] >= 0) {
                    uwTheta = getVar3D(offset + vecOffset, 6);
                }
            }

            if (m_bDrawCenterValues) {
                gl.glPushMatrix();
                //            gl.glScalef(1, 1, m_zScale);
                gl.glTranslated(x, y, z);
                if (m_dimNums[6] >= 0) {
                    gl.glRotated(uwTheta, 0, 1, 0);
                }
                gl.glRotated(uvTheta, 0, 0, 1);
                gl.glScaled(len, len, len);
                gl.glScaled(m_scale, m_scale, m_scale / 10);

                switch (m_symbolIndex) {
                    case 0:
                        drawArrow();
                        break;
                    case 1:
                        drawTwoToneArrow();
                        break;
                    case 2:
                        drawNarrowArrow();
                        break;
                    case 3:
                        // Draw line arrow.
                        gl.glBegin(gl.GL_LINES);
                        gl.glVertex2f(0, 0);
                        gl.glVertex2f(1, 0);
                        gl.glVertex2f(1, 0);
                        gl.glVertex2f(0.8f, 0.07f);
                        gl.glVertex2f(1, 0);
                        gl.glVertex2f(0.8f, -0.06f);
                        gl.glEnd();
                        break;
                    case 4:
                        drawBar();
                        break;
                    case 5:
                        gl.glBegin(gl.GL_LINES);
                        gl.glVertex2f(0, 0);
                        gl.glVertex2f(1, 0);
                        gl.glEnd();
                        break;
                    case 6:
                        drawTwoToneDiamondArrow();
                        break;
                }
                gl.glPopMatrix();
            }
        }
//        } // Synchronized block

        fileTime = newFileTime;

        // Restore the previous line attributes.
        gl.glPopAttrib();
        if (m_symbolIndex >= 4) {
            gl.glPopAttrib();
            gl.glDisable(gl.GL_CULL_FACE);
        }

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
    /**
     *
     */
    public void drawProfileSticks() {

        // Save the previous line color and pattern.
        gl.glPushAttrib(gl.GL_LINE_BIT | gl.GL_CURRENT_BIT);

        // Point OpenGL to the interleaved x-y array.
        gl.glColor3fv(m_color, 0);
//        gl.glVertexPointer(3, gl.GL_FLOAT, 0, m_profileSticks);
//        gl.glDrawArrays(gl.GL_LINES, 0, m_profileSticks.length / 3);

//        System.out.println("GL Error = " + gl.glGetError());


        gl.glBegin(gl.GL_LINES);
        for (int i = 0; i < m_profileSticks.length / 3; i++) {
            double x = m_profileSticks[i * 3];
            double y = m_profileSticks[i * 3 + 1];
            double z = m_profileSticks[i * 3 + 2];
            gl.glVertex3d(x, y, z);
        }
        gl.glEnd();



        // Restore the previous line attributes.
        gl.glPopAttrib();

        m_bNeedsRedraw = false;
    }

//------------------------------------------------------------------------------
    /**
     * Issue the OpenGL commands to draw one vector arrow.
     */
    public void drawTwoToneArrow() {

        drawArrow();

        gl.glTranslatef(0.15f, 0, 0);
        gl.glScalef(0.8f, 0.6f, 1);

        // Set the color.
        gl.glPushAttrib(gl.GL_CURRENT_BIT);
        gl.glColor3f(0, 0, 0);

        drawArrow();

        gl.glPopAttrib();
    }

//------------------------------------------------------------------------------
    /**
     * Issue the OpenGL commands to draw one 3D arrow with a diamond cross section
     * head.
     */
    public void drawTwoToneDiamondArrow() {

        gl.glPushAttrib(gl.GL_CURRENT_BIT);

        gl.glBegin(gl.GL_TRIANGLES);
        gl.glColor3f(0.5f, 0.5f, 0.5f);
        gl.glVertex3f(0.5f, 0, 0.05f);
        gl.glVertex3f(0.5f, 0, -0.05f);
        gl.glVertex3f(0.5f, -0.05f, 0);
        gl.glVertex3f(0.5f, 0.05f, 0);
        gl.glVertex3f(0.5f, 0, -0.05f);
        gl.glVertex3f(0.5f, 0, 0.05f);
        gl.glEnd();

        gl.glColor3fv(m_color, 0);
        gl.glBegin(gl.GL_LINES);
        gl.glVertex2f(0, 0);
        gl.glVertex2f(1, 0);
        gl.glEnd();

        gl.glBegin(gl.GL_TRIANGLES);
        gl.glVertex3f(0.5f, 0.05f, 0);
        gl.glVertex3f(0.5f, 0, 0.05f);
        gl.glVertex3f(1, 0, 0);

        gl.glColor3f(m_color[0] * 0.5f, m_color[1] * 0.5f,
                m_color[1] * 0.5f);

        gl.glVertex3f(0.5f, 0, 0.05f);
        gl.glVertex3f(0.5f, -0.05f, 0);
        gl.glVertex3f(1, 0, 0);
        gl.glVertex3f(0.5f, 0, -0.05f);
        gl.glVertex3f(0.5f, 0.05f, 0);
        gl.glVertex3f(1, 0, 0);

        gl.glColor3f(0, 0, 0);
        gl.glVertex3f(0.5f, -0.05f, 0);
        gl.glVertex3f(0.5f, 0, -0.05f);
        gl.glVertex3f(1, 0, 0);

        gl.glPopAttrib();

        gl.glEnd();
    }

//------------------------------------------------------------------------------
    /**
     * Issue the OpenGL commands to draw one vector arrow.
     */
    public void drawArrow() {

        gl.glBegin(gl.GL_QUADS);
        gl.glVertex2f(0, -0.05f);
        gl.glVertex2f(0, 0.05f);
        gl.glVertex2f(0.5f, 0.05f);
        gl.glVertex2f(0.5f, -0.05f);
        gl.glEnd();

        gl.glBegin(gl.GL_TRIANGLES);
        gl.glVertex2f(0.5f, 0.2f);
        gl.glVertex2f(1, 0);
        gl.glVertex2f(0.5f, -0.2f);
        gl.glEnd();
    }

//------------------------------------------------------------------------------
    /**
     * Issue the OpenGL commands to draw one vector arrow.
     */
    public void drawNarrowArrow() {

        gl.glBegin(gl.GL_QUADS);
        gl.glVertex2f(0, -0.025f);
        gl.glVertex2f(0, 0.025f);
        gl.glVertex2f(0.5f, 0.025f);
        gl.glVertex2f(0.5f, -0.025f);
        gl.glEnd();

        gl.glBegin(gl.GL_TRIANGLES);
        gl.glVertex2f(0.5f, 0.1f);
        gl.glVertex2f(1, 0);
        gl.glVertex2f(0.5f, -0.1f);
        gl.glEnd();
    }

//------------------------------------------------------------------------------
    /**
     * Issue the OpenGL commands to draw a bar.
     */
    public void drawBar() {

        gl.glBegin(gl.GL_QUADS);
        gl.glVertex2f(0, -0.1f);
        gl.glVertex2f(0, 0.1f);
        gl.glVertex2f(1, 0.1f);
        gl.glVertex2f(1, -0.1f);
        gl.glEnd();
    }

//------------------------------------------------------------------------------
    /**
     * Returns the U, V, or W components of the specified cell.
     */
    protected double getVar3D(long lookupIndex, int dimNum) {

        double var = 0;

        DataSeries series = m_seriesArray[m_seriesNums[dimNum]];
        long offset = lookupIndex * m_numDims[m_seriesNums[dimNum]] + (long) m_dimNums[dimNum];

        if (series instanceof RafDataSeries) {
            if (lookupIndex >= 0) {
                var = ((RafDataSeries) series).getValueAt(offset);
                var *= m_dimScales[dimNum];
            }
        } else {
            double[] array = m_dataArrays[m_seriesNums[dimNum]];
            if (lookupIndex >= 0 && offset < array.length) {
                var = array[(int) offset];
                var *= m_dimScales[dimNum];
            }
        }

        return var;
    }

//------------------------------------------------------------------------------
    /**
     * Returns the 2D variable at the specified offset.
     */
    protected float getVar2D(int offset) {

        float var = 0;

//        if (m_2DVarsSeries instanceof RafDataSeries) {
        var = ((RafDataSeries) m_2DVarsSeries).getValueAt(offset);
//        }
//        else {
//            var = m_2DVarsArray[offset];
//        }

        return var;
    }

//------------------------------------------------------------------------------
    /**
     * Overrides default update() in GrObject.  Called by an observed GrObject's
     * notifyObservers() method to handle changes to the observed object.
     * May use the notifyObservers() method to propagate the event handling up to
     * this object's observers.
     * @param gro   The observed object that changed.
     * @param arg   Additional information that the updated object may choose to
     *              send.
     */
    public void update(GrObject gro, Object arg) {

        // Check if the object is a graphical element that needs to be redrawn.
        if (gro instanceof DataSeries) {
            m_bNeedsRedraw = true;
        }

        // Check if the object is a container.  That would indicate the
        // output locations have changed.
        if (gro instanceof GrObjectContainer) {
            updateDrawOffsets();
            m_bNeedsRedraw = true;
        }

        if (gro == m_bath) {
            m_bath = null;
            createVectorAnimation();
        }

        if (gro == m_si3dOut) {
            m_si3dOut = null;
            createVectorAnimation();
        }

        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
    /**
     * Change the color of the curve.
     * @param red   The red components value between 0.0 and 1.0
     * @param green The green components value between 0.0 and 1.0
     * @param blue  The blue components value between 0.0 and 1.0
     */
    public void setColor(float red, float green, float blue) {
        m_color[0] = red;
        m_color[1] = green;
        m_color[2] = blue;
    }

//------------------------------------------------------------------------------
    /**
     * Change the color of the curve (see other setColor routine).
     * @param color An array of floats decribing the red, green, and blue.
     */
    public void setColor(float[] color) {
        setColor(color[0], color[1], color[2]);
    }

//------------------------------------------------------------------------------
    /**
     * Provide the list of menu items that execute commands on this object.
     * @return  A menu item that can be placed on a menu and send an ActionEvent
     *          back to this object when it is chosen.
     */
    public javax.swing.JMenuItem getMenuCommands() {

        javax.swing.JMenu subMenu = new javax.swing.JMenu(
                "GL 3D Vectors");
        javax.swing.JMenuItem item = new javax.swing.JMenuItem(
                "Create Water Surface");
        item.addActionListener(this);
        subMenu.add(item);
        item = new javax.swing.JMenuItem("Reinitialize");
        item.addActionListener(this);
        subMenu.add(item);

        return subMenu;
    }

//------------------------------------------------------------------------------
    public void actionPerformed(java.awt.event.ActionEvent e) {

        javax.swing.JMenuItem source =
                (javax.swing.JMenuItem) (e.getSource());
        String srcText = source.getText();
        if (srcText.equals("Create Water Surface")) {
            createWaterSurface();
        }
        if (srcText.equals("Reinitialize")) {
            reinitialize();
        }
    }

//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {

        String s = null;
        if (m_bShow) {
            s = "GL Si3d Vectors: ";
        } else {
            s = "(GL Si3d Vectors): ";
        }

        if (m_errorString != null) {
            s += m_errorString;
        } else {
            s += m_title;
        }

        return s;
    }
}






