/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 *                  California District, Sacramento
 *                  San Francisco Bay Hydrodynamics Project
 * @version         2008-10-14
 * Date started:    2006-10-19
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d.gui;

import gov.usgs.gr.visual.gui.FileDropHandler;
import java.io.*;

//==============================================================================
/**
 * A graphing application in a frame.  Opens Si3D bathymetry data files and
 * displays them.  Volume and area calculations can be performed on the grids.
 */
public class GrBathCalc extends gov.usgs.gr.Gr {
    
    protected double m_gridSpacing = 10;
    protected GrBathCalcFrame m_calcFrame = null;
    
//------------------------------------------------------------------------------
    /**
     * Main constructor for the GrBathCalc application.
     * @param title The string that will appear in the window's title bar.
     */
    public GrBathCalc(String title) {
	
	// Set up this window frame.
	super(title);
    }    
    
//------------------------------------------------------------------------------
    /**
     * Setup the contents of this window and read in data files that were given
     * on the command line.
     */
    public void setupAndShow() {
	
	// Get the Gr Home property and set the current file property.
	String grHome = System.getProperty("gov.usgs.gr.grhome");
	
	// If the property was not found, set it to a default value.
	if (grHome == null) {
	    System.setProperty("gov.usgs.gr.grhome",
		System.getProperty("user.dir"));
	    grHome = System.getProperty("gov.usgs.gr.grhome");
	}
	
	grHome.replace('\\', '/');
	System.setProperty("gov.usgs.gr.currentfile", grHome);
	
	// Look for Gr configuration files in XML format for default
	// object properties.  Create a list of files in the config directory
	// within the Gr home directory that begin with "gr_config" and end
	// with ".xml"
	String[] fileNames = new File(grHome + "/config").list(
	    new FilenameFilter() {
	    
	    public boolean accept(File dir, String name) {
		name = name.toLowerCase();
		return (name.startsWith("gr_config") && name.endsWith(".xml"));
	    }
	});
	
	// Read the defaults files.
	if (fileNames == null) fileNames = new String[0];
	for (int i = 0; i < fileNames.length; i++) {
	    configureDefaults(grHome + "/config/" + fileNames[i]);
	}
	
	File configFile = new File(grHome + "/gr_config.grs");
	
	// Create a controller for the program and set it to use OpenGL
	// for the view.
	m_grCon = new GrBathCalcController();
	m_grCon.setUsingOpenGL(m_bUsingOpenGL);
	GrBathCalcController calcCon = (GrBathCalcController)m_grCon;
	m_calcFrame = new GrBathCalcFrame(calcCon);
	m_calcFrame.setTitle(this.getTitle());
	calcCon.setupOnly(m_calcFrame);
	
	// Read the Gr configuration file to set the data format.
	// Check the Gr directory and the user's home directory for
	// a file called gr_config.grs.
	if (configFile.canRead())
	    m_grCon.openScript(configFile);
	
	// Open any files that were specified on the command line.
	for (String filename : m_filesToOpen)
	    m_grCon.openFile(new File(filename));
	
        // Enable Drag and Drop
        FileDropHandler fdh = new FileDropHandler(m_calcFrame, m_grCon);
        
	m_calcFrame.setVisible(true);
	
	// Let the user start off by opening a data file if none were
	// given on the command line.
	if (m_filesToOpen.size() == 0)
	    m_grCon.openFile();
    }
    
//------------------------------------------------------------------------------
/*
 * Returns a string describing this version of GrBathCalc.
 * @return	A string describing this version of GrBathCalc.
 */
    public static String getVersion() {
	return ("2008-10-14");
    }
    
//------------------------------------------------------------------------------
/*
 * Create an instance of the GrBathCalc program.
 */
    static public void main(String args[]) {
	
        System.out.println("USGS GrBathCalc version " + getVersion());
	GrBathCalc grBathCalc = new GrBathCalc("GrBathCalc");
	grBathCalc.parseCommandLine(args);
	grBathCalc.setupAndShow();
    }
}











