/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2008-03-10
 * Date started:    2006-10-19
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d.gui;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import gov.usgs.gr.visual.gl.*;
import gov.usgs.gr.visual.j2d.*;
import gov.usgs.gr.visual.gui.*;
import gov.usgs.sfhydro.si3d.*;
import gov.usgs.sfhydro.si3d.gl.GlBathymetryGrid;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.HeadlessException;
import java.awt.Rectangle;
import java.awt.event.*;
import java.awt.geom.*;
import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.util.*;
import javax.swing.*;

//==============================================================================
/**
 * A factory that creates either a Java2D or OpenGL drawing panel and handles
 * the GUI responsibilities for the application.
 */
public class GrBathCalcController extends gov.usgs.gr.GrController {
    
    GrBathCalcFrame m_frame = null;
    
    public GrBathCalcGuiOptions m_options = new GrBathCalcGuiOptions();
    
    protected boolean m_bChooseMode = false;
    protected boolean m_bDefineMode = false;
    
    protected RawBathymetry m_bath = null;
    protected int m_numI = 0;
    protected int m_numJ = 0;
    protected FluxCalculator m_fluxCalc = null;
    protected DoubleDataSeries m_discreteInput = null;
    protected DoubleDataSeries m_discreteOutput = null;
    protected GlParticleAnimation m_partAnim = null;
    
    protected Curve m_deleteCurve = null;
    protected Curve m_editCurve = null;
    protected int m_drawCurveOffset = 0;
    
    protected String m_boundarySetName = "";
    protected int[][] m_boundarySet = new int[0][];
    
    protected int m_currentSpan = 1;
    protected int m_dotIndex = 0;
    
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/GrController.gif";
    
//------------------------------------------------------------------------------
    /**
     * Create a GrController through the super class.
     */
    public GrBathCalcController() {
	
	super();
	
	m_bZoomMode = false;
	m_bPanMode = true;
        m_editablePointSize = 9;
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a GrController and wrap it around an existing page of graphs.
     * @param page	The page to place inside the new conroller.
     */
    public GrBathCalcController(PageOfGraphs page) {
	
	super(page);
	
	m_bZoomMode = false;
	m_bPanMode = true;
        m_editablePointSize = 9;
	
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a GrController as a copy of another GrController.  Share the data.
     * @param	grCon	The template controller to use in creating this one.
     */
    public GrBathCalcController(GrController grCon) {
	
	super(grCon);
	
	m_bPanMode = true;
	m_bZoomMode = false;
        m_editablePointSize = 9;
	
    }
    
//------------------------------------------------------------------------------
    /**
     * Create a string representation of this object.
     * @return  The string that describes this object.
     */
    public String toString() {
	
	String s = "GrBathCalc";
	return s;
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An XML name of this element.
     */
    public String getXmlElementName() {
	return "gov.usgs.gr";
    }
    
//------------------------------------------------------------------------------
    /**
     * Accessor method.
     * @return An ImageIcon to use when representing this object in a tree.
     */
    public javax.swing.ImageIcon getTreeIcon() {
	
	if (TREE_ICON_HOLDER.size() <= 0)
	    setTreeIcon();
	
	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }
    
//------------------------------------------------------------------------------
    /**
     * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
     */
    protected javax.swing.ImageIcon setTreeIcon() {
	
	javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);
	
	if (icon == null)
	    icon = super.setTreeIcon();
	
	TREE_ICON_HOLDER.clear();
	TREE_ICON_HOLDER.add(icon);
	
	return icon;
    }
    
//------------------------------------------------------------------------------
    /**
     * Setup the contents of this window.
     * @param frame	The frame to put the GUI and drawing panel in.
     */
    public void setupOnly(GrBathCalcFrame frame) {
	
	m_frame = frame;
	
	// Find out the width and height of the screen in pixels.
	Dimension d = frame.getToolkit().getScreenSize();
	
	// Set the icon in the window to be the Gr icon.
	ImageIcon image = new javax.swing.ImageIcon(Thread.currentThread().
	    getContextClassLoader().getResource("gr.gif"));
	frame.setIconImage(image.getImage());
	
	m_frame.setUnitIndexes(m_options.m_areaUnitsIndex,
	    m_options.m_volUnitsIndex);
	
	// Create a GUI-options object and make it a child of this controller.
	// If the GUI-options doesn't have a preferred look and feel, set it
	// to be the current look and feel.
	GrProperty p = new GrProperty(
	    UIManager.getLookAndFeel().getClass().getName());
	GrHashList fallBackProps = new GrHashList();
	fallBackProps.put("lookAndFeel", p);
	GuiOptions guiOptions = new GuiOptions();
	GrHashList guiProps = guiOptions.getProperties();
	p = (GrProperty)guiProps.get("lookAndFeel");
	if (p.value == null)
	    guiOptions.setProperties(fallBackProps);
	add(guiOptions);
	
	// Create the main graphics panel as either OpenGL or Java2D.
	if (GlGrPanel.openGlIsAvailable() && m_bUsingOpenGL) {
//        	m_pageProxy = (JPanel)new GlGrPanel();
	    m_pageProxy = new GlGrPanel();
	} else {
	    // Didn't or couldn't create an OpenGL panel, so use Java2D...
	    m_bUsingOpenGL = false;
	    m_pageProxy = new J2dGrPanel();
	}
	
	// Hook this controller to the panel.
	m_pageProxy.setController(this);
	
	// Set the size of the panel to be 70% of the width and height of
	// the screen.
	Component component = m_pageProxy.getComponent();
	component.setSize((int)(d.width * 0.7), (int)(d.height * 0.7));
	frame.setComponent(component);
	frame.setSize((int)(d.width * 0.7), (int)(d.height * 0.7));
	frame.pack();
    }
    
//------------------------------------------------------------------------------
    /**
     * Opens the file that is passed in.  If the file type can't be determined by
     * the extension (GS, GRS, GRT, or XML) the user is asked to confirm the format.
     * Once the format is determined, the file is read and the contents are added
     * as data or graphics.
     * @param file  The file to open.
     */
    public void openFile(File file) {
	
	// Return if the file isn't valid.
	if (file == null)
	    return;
	
	String fileName = file.getAbsolutePath();
	fileName.replace('\\', '/');
	if (fileName.endsWith(".xml")) {
	    super.openFile(file);
	    return;
	}
        
        // Check if the user wants to close the current page.
        if (m_page != null) {
            int option = JOptionPane.showConfirmDialog(null,
                "Close current session and open new grid?", "Open New File",
                JOptionPane.YES_NO_OPTION);
            if (option == JOptionPane.NO_OPTION)
                return;
            
            clear();
        }

	// Try to determine the format.
        m_dataFormatClass = identifyFormat(file);
	
        if (m_dataFormatClass == null)
            m_dataFormatClass = "gov.usgs.sfhydro.data.Si3dBathymetryFile";
        
	// Check if the chosen format is derived from GrFile.
	Class grFileClass = null;
	Class formatClass = null;
	try {
	    grFileClass = Class.forName("gov.usgs.gr.GrFile");
	    formatClass = Class.forName(m_dataFormatClass);
	} catch (ClassNotFoundException e) {return;}
	
	// Show an hourglass until the job is complete.
	((Component)getParentFrame(m_pageProxy.getComponent()))
	.setCursor(new Cursor(Cursor.WAIT_CURSOR));
	
	GrFile grFile = getGrFile(file.getAbsolutePath());
        
        Runnable runnable = null;
        final InitialInputFrame inputFrame = new InitialInputFrame(m_options);
        final List userInputReceived = new ArrayList();
        final GrFile theFile = grFile;

        if (m_dataFormatClass.equals(
            "gov.usgs.sfhydro.data.Si3dBathymetryFile")) {

    //	inputFrame.setTitle("GrBathCalc");
            runnable = new Runnable() {
                public void run() {
    //		Dimension d = inputFrame.getToolkit().getScreenSize();
    //		inputFrame.setLocation((int)(d.width * 0.2),
    //		    (int)(d.height * 0.2));
    //		inputFrame.setAlwaysOnTop(true);
    //		inputFrame.setVisible(true);

                    boolean bInputWasValid = false;
                    while (!bInputWasValid) {
                        String userInput = "" + m_options.m_gridSpacing;
                        userInput = (String)JOptionPane.showInputDialog(m_frame,
                            "Grid Spacing in Meters", "Verify",
                            JOptionPane.QUESTION_MESSAGE, null, null, userInput);

                        // Quit the program if cancelled
                        if (userInput == null)
                            System.exit(0);

                        try {
                            m_options.m_gridSpacing =
                               (int)Math.round(Double.parseDouble(userInput));
                            bInputWasValid = true;
                        } catch (NumberFormatException ex) {
                            JOptionPane.showMessageDialog(m_frame,
                                "The value is not valid.  Please enter an integer value.",
                                "Parsing Error",
                                JOptionPane.ERROR_MESSAGE);
                        }
                    }
                    userInputReceived.add(new Integer(0));
                }
            };
        }
        else {
            runnable = new Runnable() {
                public void run() {
                    if (m_dataFormatClass.equals(
                        "gov.usgs.sfhydro.data.ArcGridAsciiFile")) {
                        
                        theFile.showOptionsDialog(m_frame);
                        userInputReceived.add(new Integer(0));
                    }
                }
            };
        }
        
        if (SwingUtilities.isEventDispatchThread())
            runnable.run();
        else 
            try {
                SwingUtilities.invokeAndWait(runnable);
            } catch (InterruptedException ex) {
                ex.printStackTrace();
            } catch (InvocationTargetException ex) {
                ex.printStackTrace();
            }
	
	// Read the GrFile.
	try {
	    grFile.read();
	    
	    m_bath = (RawBathymetry)grFile.firstChildInstanceOf(
		"gov.usgs.sfhydro.si3d.RawBathymetry", false);
	    DoubleDataSeries series = m_bath.getBathymetrySeries();
	    series.setDimensionLocked(0, true);
	    PageOfGraphs page = (PageOfGraphs)grFile.firstChildInstanceOf(
		"gov.usgs.gr.visual.PageOfGraphs", false);
	    Graph graph = (Graph)page.firstChildInstanceOf(
		"gov.usgs.gr.visual.Graph", false);
	    graph.setProperty("aspect1to1", "true");
	    GlBathymetryGrid glBath = (GlBathymetryGrid)graph.
                firstChildInstanceOf(
		"gov.usgs.sfhydro.si3d.gl.GlBathymetryGrid", false);
	    glBath.setProperty("drawPartiallyWetThreshold", "10000");
	    glBath.setProperty("drawGridThreshold", "3");
	    glBath.setProperty("drawDepthsThreshold", "2");
	    glBath.setProperty("drawUvwThreshold", "1");
	    glBath.setProperty("drawSolidEdges", "false");
	    glBath.setProperty("csCoefficient", "0.4");
            glBath.regenerateColorScheme();
            
            double[] partArray = {0, 0};
            DoubleDataSeries partSeries = new DoubleDataSeries(partArray, 2);
            m_partAnim = new GlParticleAnimation(partSeries);
            m_partAnim.setProperty("seriesNums", "0,1,1,0,0");
            m_partAnim.setProperty("dimNums", "-1,0,1,-1,-1");
            m_partAnim.setProperty("scale", "2000");
            m_partAnim.setProperty("symbolType", "Circle");
            m_partAnim.setProperty("show", "false");
            graph.add(m_partAnim);
	    
	    GrHashList bathProps = m_bath.getProperties();
	    GrProperty p = (GrProperty)bathProps.get("numI");
	    m_numI = (Integer)p.value;
	    p = (GrProperty)bathProps.get("numJ");
	    m_numJ = (Integer)p.value;
	    m_frame.setGridSize(m_numI, m_numJ);
	    
	    
	    try {
		
		while (userInputReceived.size() < 1) {
		    Thread.sleep(200);
		}
	    } catch (InterruptedException ex) {
		ex.printStackTrace();
	    }
            
	    
	    m_bath.setProperty("xSpacing", "" + m_options.m_gridSpacing);
	    m_bath.setProperty("ySpacing", "" + m_options.m_gridSpacing);
	    m_bath.notifyObservers();
	    m_frame.setGridSpacing(m_options.m_gridSpacing);
	    
	    m_fluxCalc = new FluxCalculator();
	    m_fluxCalc.setSpacing(m_options.m_gridSpacing);
	    m_discreteInput = new DoubleDataSeries(2);
	    m_fluxCalc.add(m_discreteInput);
	    m_discreteOutput =
		(DoubleDataSeries)m_fluxCalc.getContainerSeries("Output");
	    
	} catch (IOException e) {
	    e.printStackTrace();
	    
	    JOptionPane.showMessageDialog(getParentFrame(
		m_pageProxy.getComponent()),
		"File " + file.getAbsolutePath() + " could not be loaded.",
		"IO Exception", JOptionPane.ERROR_MESSAGE);
	    return;
	}
	
	// Add the file as a child of the controller.
	add(grFile);
	
	// Declare new local variables.
	boolean bBareSeries = false;
	Enumeration enumeration;
	
	// Children of the file are usually added unless the file is a
	// template.
	if (m_bAddChildrenWhenOpening) {
	    // Add certain children of the file to this controller.
	    enumeration = grFile.children();
	    while (enumeration.hasMoreElements()) {
		GrObject gro = (GrObject)enumeration.nextElement();
		while (gro instanceof ObjectEffigy)
		    gro = ((ObjectEffigy)gro).getSubject();
		
		// Child is a page...
		if (gro instanceof PageOfGraphs) {
		    // Add it as a child to this controller.
		    add(gro);
		    
		    // If the page title is blank, set it equal to the
		    // file name.
		    PageOfGraphs page = (PageOfGraphs)gro;
		    String pageTitle = page.getTitle();
		    if (pageTitle == null || pageTitle.equals("")) {
			pageTitle = file.getName();
			page.setTitle(pageTitle);
		    }
		    
		    // Set the page as the current page and log the open
		    // command
		    setPage(page);
		    logCommand("Open " + file.getAbsolutePath());
		    
		    
		    // Look for a TimeMarker object as a child of the page.
		    boolean bTime = false;
		    Enumeration pageChildren = page.children();
		    while (pageChildren.hasMoreElements()) {
			GrObject pcGro = (GrObject)pageChildren.nextElement();
			if (pcGro instanceof gov.usgs.gr.tools.TimeMarker) {
			    bTime = true;
			    break;
			}
		    }
		    
		    if (bTime)
			usePanMode(true);
		}
		
		// Child is a GUI Options object...
		if (gro instanceof GuiOptions) {
		    add(gro);
		}
		
		// Child is a data series...
		if (gro instanceof DataSeries) bBareSeries = true;
	    }
	}
	
	// Change the title of the window to the name of the file.
	updateWindowTitle(file.getName());
	
	// Trigger a repaint of the updated page contents.
	zoomAll();
//	m_pageProxy.repaint();
    }
    
//------------------------------------------------------------------------------
    /**
     * Delete all selected points on the current page or, if there are no selected
     * points, prompt the user to delete the selected graphs.
     */
    public void deleteSelected() {
	
    }
    
//------------------------------------------------------------------------------
    /**
     */
    public void setMode(int index) {
	
	// If leaving add mode
	if (m_bDefineMode && index != 2) {
	    // Remove the drawing curve if it has fewer than two points.
	    m_drawModeCurve = null;
	}
	
	if (m_page != null && (m_bPanMode || m_bChooseMode) && (index == 2 || index == 3)) {
            m_page.deferUpdates();
	    m_bQuickDraw = true;
	    modifyDrawingStyle();
	    m_bath.getBathymetrySeries().deselectAll();
	    showCurves(true);
//            showEditingSymbols(true);
            m_page.processDeferredUpdates();
///	    triggerRepaint();
	}
	if (m_page != null && m_bDefineMode && (index == 0 || index == 1)) {
	    discretizeBoundaries();
	    selectBoundaries();
	    showCurves(false);
///	    m_bQuickDraw = false;
	    modifyDrawingStyle();
//            showEditingSymbols(false);
	    triggerRepaint();
	}
	
	// Set select mode on or off and set pan and zoom to the opposite value.
	m_bPanMode = (index == 0 ? true : false);
	m_bChooseMode = (index == 1 ? true : false);
	m_bDefineMode = (index == 2 ? true : false);
    }
    
//------------------------------------------------------------------------------
    /**
     */
    public void manageBoundarySet() {
	
	
	final BoundarySetManager bsmFrame = new BoundarySetManager();
	bsmFrame.setController(this);
	bsmFrame.setSuggestedSuffix("_" + m_options.m_gridSpacing + "m");
	bsmFrame.setTitle("Boundaries");
	try {
	    Runnable runnable = new Runnable() {
		public void run() {
		    Dimension d = bsmFrame.getToolkit().getScreenSize();
		    bsmFrame.setLocation((int)(d.width * 0.2),
			(int)(d.height * 0.2));
		    bsmFrame.setAlwaysOnTop(true);
		    
		    boolean bPresent = false;
                    if (m_page != null) {
                        Graph graph = m_page.getGraphList().get(0);
                        List<Curve> curves = graph.getCurves();
                        if (curves.size() > 0)
                            bPresent = true;
                    }
		    bsmFrame.setBoundariesPresent(bPresent);
		    
		    bsmFrame.setVisible(true);
		}
	    };
	    if (SwingUtilities.isEventDispatchThread())
		runnable.run();
	    else
		SwingUtilities.invokeAndWait(runnable);
	} catch (HeadlessException ex) {
	    ex.printStackTrace();
	} catch (InterruptedException ex) {
	    ex.printStackTrace();
	} catch (InvocationTargetException ex) {
	    ex.printStackTrace();
	}
//        if (m_boundarySetName.length() < 1) {
//            saveBoundaries();
//        }
//        else {
//            openBoundaries();
//        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Zoom in on a smaller area of the zooms on all selected graphs.
     */
    public void zoomIn() {
	
	// Don't proceed if there is no current page.
	if (m_page == null) return;
	
	// Show an hourglass until the job is complete.
	((Component)getParentFrame(m_pageProxy.getComponent()))
	.setCursor(new Cursor(Cursor.WAIT_CURSOR));
	
	int xSpan = 0;
	int ySpan = 0;
	
	// Change the x-axis for all currently selected graphs.
	Iterator iter = m_page.getSelectedXZooms().iterator();
	while (iter.hasNext()) {
	    Zoom zoom = (Zoom)iter.next();
	    zoom.zoomIn();
	    int numColumns = 1;
	    zoom.autoScale(m_page.getWidth() / numColumns / 80);
	    xSpan = (int)
                (zoom.getViewLength() / (double)m_options.m_gridSpacing);
	}
	
	// Change the y-axis for all currently selected graphs.
	iter = m_page.getSelectedYZooms().iterator();
	while (iter.hasNext()) {
	    Zoom zoom = (Zoom)iter.next();
	    zoom.zoomIn();
	    int numRows = m_page.getShowList().size();
	    zoom.autoScale(m_page.getHeight() / numRows / 60);
	    ySpan = (int)
                (zoom.getViewLength() / (double)m_options.m_gridSpacing);
	}
	
	m_currentSpan = (int)(xSpan + ySpan / 2);
	modifyDrawingStyle();
    }
    
//------------------------------------------------------------------------------
    /**
     * Zoom out to show a larger area of the graph.
     */
    public void zoomOut() {
	
	// Don't proceed if there is no current page.
	if (m_page == null) return;
	
	// Show an hourglass until the job is complete.
	((Component)getParentFrame(m_pageProxy.getComponent()))
	.setCursor(new Cursor(Cursor.WAIT_CURSOR));
	
	
	int xSpan = 0;
	int ySpan = 0;
	
	// Change the x-axis for all currently selected graphs.
	Iterator iter = m_page.getSelectedXZooms().iterator();
	while (iter.hasNext()) {
	    Zoom zoom = (Zoom)iter.next();
	    zoom.zoomOut();            // Autoscale...
	    int numColumns = 1;
	    zoom.autoScale(m_page.getWidth() / numColumns / 80);
	    
	    xSpan = (int)
	    (zoom.getViewLength() / (double)m_options.m_gridSpacing);
	}
	
	// Change the y-axis for all currently selected graphs.
	iter = m_page.getSelectedYZooms().iterator();
	while (iter.hasNext()) {
	    Zoom zoom = (Zoom)iter.next();
	    zoom.zoomOut();
	    int numRows = m_page.getShowList().size();
	    zoom.autoScale(m_page.getHeight() / numRows / 60);
	    ySpan = (int)
	    (zoom.getViewLength() / (double)m_options.m_gridSpacing);
	}
	
	m_currentSpan = (int)(xSpan + ySpan / 2);
	modifyDrawingStyle();
    }
    
//------------------------------------------------------------------------------
    /**
     * Zoom out to show all the data in the graph.
     */
    public void zoomAll() {
	
	// Don't proceed if there is no current page.
	if (m_page == null) return;
	
	// Show an hourglass until the job is complete.
	((Component)getParentFrame(m_pageProxy.getComponent()))
	.setCursor(new Cursor(Cursor.WAIT_CURSOR));
	
	int xSpan = 0;
	int ySpan = 0;
	
	m_page.updateZoomContents();
	// Change the x-axis for all currently selected graphs.
	Iterator iter = m_page.getSelectedXZooms().iterator();
	while (iter.hasNext()) {
	    Zoom zoom = (Zoom)iter.next();
	    int numColumns = 1;
	    zoom.zoomAll(m_page.getWidth() / numColumns / 80);
	    
	    xSpan = (int)
	    (zoom.getViewLength() / (double)m_options.m_gridSpacing);
	}
	// Change the y-axis for all currently selected graphs.
	iter = m_page.getSelectedYZooms().iterator();
	while (iter.hasNext()) {
	    Zoom zoom = (Zoom)iter.next();
	    int numRows = m_page.getShowList().size();
	    zoom.zoomAll(m_page.getHeight() / numRows / 80);
	    
	    ySpan += (int)
	    (zoom.getViewLength() / (double)m_options.m_gridSpacing);
	}
	
	m_currentSpan = (int)(xSpan + ySpan / 2);
	modifyDrawingStyle();
    }
    
//------------------------------------------------------------------------------
    /**
     * Zoom out to show a larger area of the graph.
     */
    public void modifyDrawingStyle() {
	
	if (m_bQuickDraw) {
	    useFastDrawingStyle();
	    return;
	}

        GlBathymetryGrid glBath = (GlBathymetryGrid)firstChildInstanceOf(
	    "gov.usgs.sfhydro.si3d.gl.GlBathymetryGrid", true);
	if (m_currentSpan > 1500) {
	    glBath.setProperty("drawWetSquaresThreshold", "5");
	} else {
	    glBath.setProperty("drawWetSquaresThreshold", "50000");
	}
	if (m_currentSpan < 26) {
	    glBath.setProperty("drawGridThreshold", "200");
	} else {
	    glBath.setProperty("drawGridThreshold", "3");
	}
	if (m_currentSpan < 13) {
	    glBath.setProperty("drawDepthsThreshold", "199");
	} else {
	    glBath.setProperty("drawDepthsThreshold", "2");
	}
	
        int lod = Integer.MAX_VALUE;
        if (m_currentSpan > 0)
            lod = Math.round(300f / (float)m_currentSpan);
	glBath.setProperty("levelOfDetail", "" + lod);        
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public void useFastDrawingStyle() {
	
	GlBathymetryGrid glBath = (GlBathymetryGrid)firstChildInstanceOf(
	    "gov.usgs.sfhydro.si3d.gl.GlBathymetryGrid", true);
	glBath.setProperty("drawWetSquaresThreshold", "5");
	glBath.setProperty("drawGridThreshold", "3");
	glBath.setProperty("drawDepthsThreshold", "2");
	glBath.setProperty("levelOfDetail", "1");
    }
    
//------------------------------------------------------------------------------
    public void writeAreaAndVolume() {
	
	// Area
	GrHashList bathProps = m_bath.getProperties();
	GrProperty p = (GrProperty)bathProps.get("totalRawArea");
	double totalArea = (Double)p.value;
	p = (GrProperty)bathProps.get("reachableArea");
	double reachableArea = (Double)p.value;
	m_frame.setArea(totalArea, reachableArea);
	
	// Volume
	p = (GrProperty)bathProps.get("totalRawVolume");
	double totalVolume = (Double)p.value;
	p = (GrProperty)bathProps.get("reachableVolume");
	double reachableVolume = (Double)p.value;
	m_frame.setVolume(totalVolume, reachableVolume);
    }
    
//------------------------------------------------------------------------------
    public void discretizeBoundaries() {
	
	Graph graph = m_page.getGraphList().get(0);
	List<Curve> curves = graph.getCurves();
	m_boundarySet = new int[curves.size()][];
	int boundaryIndex = 0;
	for (Curve curve : curves) {
	    m_discreteInput.setArray(curve.getSeries().getArray());
	    m_fluxCalc.createDiscreteCrossSection();
	    double[] discreteArray = m_discreteOutput.getArray();
	    int numPoints = m_discreteOutput.getNumPoints();
	    m_boundarySet[boundaryIndex] = new int[numPoints];
	    for (int n = 0; n < numPoints; n++) {
		int iCoord = (int)Math.round(discreteArray[n * 2]
		    / m_options.m_gridSpacing);
		int jCoord = (int)Math.round(discreteArray[n * 2 + 1]
		    / m_options.m_gridSpacing);
		int index = jCoord * m_numI + iCoord;
		m_boundarySet[boundaryIndex][n] = index;
	    }
	    boundaryIndex++;
	}
    }
    
//------------------------------------------------------------------------------
    public void selectBoundaries() {
	
	DoubleDataSeries bathSeries = m_bath.getBathymetrySeries();
	bathSeries.deselectAll();
	
	int numBoundaries = m_boundarySet.length;
	for (int i = 0; i < numBoundaries; i++) {
	    int[] pointArray = m_boundarySet[i];
	    int numPoints = pointArray.length;
	    for (int n = 0; n < numPoints; n++) {
		bathSeries.selectPointSafely(pointArray[n]);
	    }
	}
    }
    
//------------------------------------------------------------------------------
    public void showCurves(boolean bShow) {
	
	Graph graph = m_page.getGraphList().get(0);
	List<Curve> curves = graph.getCurves();
	for (Curve curve : curves) {
	    curve.setShow(bShow);
	}
	
    }
    
//------------------------------------------------------------------------------
    public void saveBoundaries(String setName) {
	
	Graph graph = m_page.getGraphList().get(0);
	List<Curve> curves = graph.getCurves();
	XMLProjectFile projFile = new XMLProjectFile();
	GrObjectContainer groc = new GrObjectContainer();
	for (Curve curve : curves) {
	    groc.add(curve);
	}
	projFile.add(groc);
	
	String userHome = System.getProperty("user.home");
	userHome.replace('\\', '/');
	
	String fileDir = userHome + "/GrBathCalc/Boundaries";
	projFile.setProperty("file", fileDir + "/" + setName + "_bnd.xml");
	
	// Write the file to disk.
	try {
	    (new File(fileDir)).mkdirs();
	    projFile.write();
	} catch (java.io.IOException ioe) {
	    ioe.printStackTrace();
	}
    }
    
//------------------------------------------------------------------------------
    public void openBoundaries(String setName, boolean bReplace) {
	
	String userHome = System.getProperty("user.home");
	userHome.replace('\\', '/');
	
	String fileName = userHome + "/GrBathCalc/Boundaries/" +
	    setName + "_bnd.xml";
	
	try {
	    Graph graph = (Graph)m_page.firstChildInstanceOf(
		"gov.usgs.gr.visual.gl.GlGraph", false);
	    List<GlCurve> newCurves = new ArrayList();
	    if (setName != null) {
		openFile(new File(fileName));

		List fileList = this.findChildInstancesOf(
		    "gov.usgs.gr.GrFile", false);
		GrFile file = (GrFile)fileList.get(fileList.size() - 1);
		newCurves = (List<GlCurve>)file.findChildInstancesOf(
		    "gov.usgs.gr.visual.gl.GlCurve", true);
	    }
	    if (bReplace) {
		List<GlCurve> oldCurves = (List<GlCurve>)graph.findChildInstancesOf(
		    "gov.usgs.gr.visual.gl.GlCurve", false);
		for (GrObject gro : oldCurves) {
		    graph.remove(gro);
		}
	    }
	    for (GrObject gro : newCurves) {
		graph.add(gro);
	    }
	} finally {
	}

	if (m_bPanMode || m_bChooseMode) {
	    showCurves(false);
	    discretizeBoundaries();
	    selectBoundaries();
	    triggerRepaint();
	}	
        else {
	    showCurves(true);
	    triggerRepaint();
            
        }
    }
    
//------------------------------------------------------------------------------
    public boolean showSelectionMarker(Graph graph, int mouseX, int mouseY,
	Rectangle2D.Double userRect) {
	
        // Initialize local variables.
        boolean bNeedRedraw = false;
	m_deleteCurve = null;
	m_editCurve = null;
	boolean bShowParticle = false;
	boolean bShowMid = false;
	double partX = 0;
	double partY = 0;
        int newDotIndex = 0;
        
        // Set the width of the selection area in pixels and calculate the
        // equivalent width and half-width in user coordinates.
	int pixelWidth = 12;
	Rectangle s2Rect = new Rectangle(
	    mouseX + pixelWidth, mouseY, 0, 0);
	Rectangle2D.Double u2Rect = graph.screenToUser(s2Rect);
	double width = u2Rect.x - userRect.x;
	double halfWidth = width / 2;

        // Get a list of the curves in the graph.
        List<Curve> curves = new ArrayList<Curve>(graph.getCurves());
        
//// Redundant?	
//	if (m_drawModeCurve != null)
//	    curves.add(m_drawModeCurve);
        
        // Look at each curve on the graph...
	for (int i = curves.size() - 1; i >= 0; i--) {
	    Curve curve = curves.get(i);
	    double[] curveArray = curve.getSeries().getArray();
            
            // Calculate the distance between the mouse and the first endpoint
            // of the curve.  Turn the dot on if is close enough.
	    double dist1 = calculateDistanceBetween(userRect.x,
		userRect.y, curveArray[0], curveArray[1]);
	    if (dist1 < halfWidth) {
		bShowParticle = true;
	    }
            
            // Calculate the distance between the mouse and the last endpoint
            // of the curve.  Turn the dot on if is close enough.
	    double dist2 = calculateDistanceBetween(userRect.x,
		userRect.y, curveArray[2], curveArray[3]);
	    if (dist2 < halfWidth) {
		 bShowParticle = true;                           
	    }
            
            // Calculate the distance between the mouse and the midpoint
            // of the curve.  Turn the dot on if is close enough.
	    double midX = (curveArray[0] + curveArray[2]) / 2;
	    double midY = (curveArray[1] + curveArray[3]) / 2;
	    double distMid = calculateDistanceBetween(userRect.x,
		userRect.y, midX, midY);
	    if (distMid < halfWidth) {
		 bShowParticle = true;                           
	    }
            
            // If the mouse was close enough to one or more of the points
            // (first, last, or middle) then determine which one was closest.
	    if (bShowParticle) {
		if (dist1 <= dist2) {
		    if (dist1 <= distMid) {
                        // Closest to first endpoint.
			partX = curveArray[0];
			partY = curveArray[1];
			m_drawCurveOffset = 0;
                        newDotIndex = i * 3 + 1;
		    }
		    else {
                        // Closest to midpoint.
			partX = midX;
			partY = midY; 
			bShowMid = true;
		    }
		}
		else {
		    if (dist2 <= distMid) {
                        // Closest to last endpoint.
			partX = curveArray[2];
			partY = curveArray[3];                                    
			m_drawCurveOffset = 2;
                        newDotIndex = i * 3 + 1;
		    }
		    else {
                        // Closest to midpoint.
			partX = midX;
			partY = midY; 
			bShowMid = true;
		    }
		}
		
		if (bShowMid) {
		    m_deleteCurve = curve;
                    newDotIndex = i * 3 + 3;
                }
		else {
		    m_editCurve = curve;
                }
		
		break;                            
	    }
	}
        
        // Don't redraw if nothing has changed.
        if (newDotIndex != m_dotIndex) {
            m_dotIndex = newDotIndex;
            bNeedRedraw = true;
        }
	
        // If editing the curve, set the endpoint value to the mouse location. 
	if (m_drawModeCurve != null) {
	    int yPos = m_page.getHeight() - m_mouseY;
	    DoubleDataSeries series = m_drawModeCurve.getSeries();
	    double[] drawArray = series.getArray();
	    drawArray[m_drawCurveOffset] = userRect.x;
	    drawArray[m_drawCurveOffset + 1] = userRect.y;
	    series.arrayChanged();
            bNeedRedraw = true;
	}

        // Turn the dot on or off, as determined above.
	if (bShowParticle) {
	    m_partAnim.setProperty("show", "true");
	    if (bShowMid) {
		m_partAnim.setProperty("color", "red");

		width *= 2;
	    }
	    else {
		m_partAnim.setProperty("color", "blue");                        
	    }
	    m_partAnim.setProperty("scale", "" + width);
            if (m_drawModeCurve != null)
                m_partAnim.setProperty("symbolType", "Bull's Eye");
            else
                m_partAnim.setProperty("symbolType", "Circle");
	    DoubleDataSeries series = (DoubleDataSeries)m_partAnim.
		firstChildInstanceOf("gov.usgs.gr.data.DoubleDataSeries",
		false);
	    double[] partArray = series.getArray();
	    partArray[0] = partX;
	    partArray[1] = partY;
	    series.arrayChanged();
	}
	else {
	    m_partAnim.setProperty("show", "false");
	}
        
        return bNeedRedraw;
    }
    
//------------------------------------------------------------------------------
    /**
     * Handle mouse movement within the page area.  Updates nearest point
     * coordinates on the status bar and shows and hides the side bar of the graphs.
     * @param e     The mouse event object containing the mouse position.
     */
    public void mouseMoved(MouseEvent e) {
	
//        System.out.println("Mouse moved to " + e.getX() + ", " + e.getY());
	// Don't proceed if there is no current page.
	if (m_page == null) return;
	synchronized (m_page) {
	    if (m_mouseHandler != null) {
		m_mouseHandler.mouseMoved(e);
	    }
	    
	    // Initialize the X-coordinate text string and the formatter.
	    String xCoord = "";
	    Point2D.Double p = null;
	    
	    // Find the graph nearest to the mouse.
	    Graph graph = m_page.getNearestChild(e.getX(), e.getY());
	    
	    // Convert the screen pixel coordinates to user space coordinates.
	    Rectangle screenRect = new Rectangle(e.getX(),e.getY(), 0, 0);
	    Rectangle2D.Double userRect =
		m_page.screenToUser(screenRect, graph);
	    
	    if (userRect != null) {
		p = new Point2D.Double(userRect.x, userRect.y);
		int iCoord = (int)Math.round(p.x / m_options.m_gridSpacing);
		int jCoord = (int)Math.round(p.y / m_options.m_gridSpacing);
		// Set coordinate text.
		String ijString = "" + iCoord + ", " + jCoord;
		String xyString = "" + String.format("%g", p.x) + ", "
		    + String.format("%g", p.y);
		String depthString = "LAND";
		double depth = m_bath.getDepthAt(iCoord, jCoord);
		if (Double.isNaN(depth))
		    depthString = "";
		else if (depth > -5)
		    depthString = String.format("%6.1f", depth);
		m_frame.setCoords(ijString, xyString, depthString);
                
                if (m_bDefineMode) {
                    boolean bNeedRedraw = showSelectionMarker(
                        graph, e.getX(), e.getY(), userRect);
                    if (bNeedRedraw)
                        m_pageProxy.repaint();        
                }
	    }
	    
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Handle mouse dragging within the page area.  Dragging can be interpreted
     * as panning, defining a zoom rectangle, or moving a data point.
     * @param e     The mouse event object containing the mouse position.
     */
    public void mouseDragged(MouseEvent e) {
//System.out.println("Mouse dragged to " + e.getX() + ", " + e.getY());
	// Don't proceed if there is no current page.
	if (m_page == null) return;
	synchronized (m_page) {
	    
	    Graph graph0 = (Graph)m_page.getSelectedGraphs().get(0);
	    
	    // Declare local variables.
	    int xDiff, yDiff;
	    
	    // If in pan mode with the right mouse button down...
	    if (m_bPanMode) {
		// Get the difference in mouse position since the last call.
		// Make the difference zero if the axis is locked.
		xDiff = m_bLockX ? 0 : m_mouseX - e.getX();
		yDiff = m_bLockY ? 0 : e.getY() - m_mouseY;
		
		// Send the pan message to the selected graphs.
		m_page.pan(xDiff, yDiff);
		
		// Save the current mouse position for future use.
		m_mouseX = e.getX();
		m_mouseY = e.getY();

		// Trigger a repaint of the updated contents.
		m_pageProxy.repaint();
	    }
	    if (m_bDefineMode) {
		mouseMoved(e);
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Handle mouse presses within the page area.  Set up variables for comparison
     * during later mouse drag and release events.
     * @param e     The mouse event object containing the mouse position.
     */
    public void mousePressed(MouseEvent e) {
	
	// Don't proceed if there is no current page.
	if (m_page == null) return;
	synchronized (m_page) {
	    // Save the mouse position for future comparisons.
	    m_mouseX = e.getX();
	    m_mouseY = e.getY();

	    // Find the graph nearest to the mouse.
	    Graph graph = m_page.getNearestChild(e.getX(), e.getY());
	    
	    Rectangle screenRect = new Rectangle(e.getX(),e.getY(), 0, 0);
	    Rectangle2D.Double userRect =
		m_page.screenToUser(screenRect, graph);
	    
	    if (m_bChooseMode) {
		if (m_bath == null)
		    return;
				
		int i = (int)Math.round(userRect.x / m_options.m_gridSpacing);
		int j = (int)Math.round(userRect.y / m_options.m_gridSpacing);
		
		double waterLevel = m_frame.getWaterLevel();
		double cutoffDepth = m_frame.getCutoffDepth(waterLevel);
		
		double depth = m_bath.getDepthAt(i, j);
		if (!Double.isNaN(depth) && depth > 0) {
		    selectBoundaries();
		    m_bath.calculateReachableVolume(i, j, waterLevel,
			cutoffDepth);
		    writeAreaAndVolume();
		} else {
		    m_bath.getBathymetrySeries().deselectAll();
		    selectBoundaries();
		}
	    }
	    
	    // Add a point to a new curve that is being drawn by the user.
	    if (m_bDefineMode) {
		// If this is the first click inside this graph, set up
		// new curve for the graph.
		if (m_drawModeCurve == null && m_editCurve == null) {
		    if (m_deleteCurve != null) {
			graph.remove(m_deleteCurve);
			m_deleteCurve = null;
			m_partAnim.setProperty("symbolType", "Bull's Eye");
			
			m_pageProxy.repaint();
			
			return;
		    }
		    // Allow the new series to be edited in both directions.
		    DoubleDataSeries newSeries = new DoubleDataSeries(2);
		    newSeries.getDimension(0).setDimensionLocked(false);
		    newSeries.getDimension(1).setDimensionLocked(false);
		    m_drawModeCurve = new GlCurve(newSeries);
		    m_drawModeCurve.setLineWidth(5);
//                    m_drawModeCurve.setProperty("editablePointSize", "9");
		    m_drawModeCurve.setColor(0, 0, 1);
		    m_drawModeGraph = graph;
		    
		    // Make sure the curve that is being drawn is added
		    // to the graph.
		    if (graph != null) {
			graph.add(m_drawModeCurve);
		    }
		}
		if (m_drawModeCurve == null && m_editCurve != null) {
		    m_drawModeCurve = m_editCurve;
		    m_editCurve = null;
		    
		    return;
		}
		
		// Get the data series for the draw curve.
		DoubleDataSeries series = m_drawModeCurve.getSeries();
		double[] oldArray = series.getArray();
		int oldSize = oldArray.length;
		double[] newArray = null;
		if (oldSize != 4) {
		    newArray = new double[4];
		    m_drawCurveOffset = 0;
		}
		else {
		    newArray = oldArray;
		    m_drawCurveOffset = 2;
		}
				
		// Calculate the position of the graph user area.
		Rectangle rect = graph.getDisplayRect();
		int yPos = m_page.getHeight() - m_mouseY;
		
		if (oldSize > 0) {
		    // Save the current line for later undo.
		    DataSnapShot snapShot = new DataSnapShot();
		    snapShot.store(series);
		    saveDataSnapShot(snapShot);
		}
		
		// Append the most recent point to the end of the
		// curve's series' data array.
		Point2D.Double p = graph.getCoords(
		    m_mouseX - rect.x, yPos);
		if (newArray != oldArray) {
		    newArray[2] = userRect.x;
		    newArray[3] = userRect.y;
		    newArray[0] = newArray[2];
		    newArray[1] = newArray[3];
		}
		
		if (newArray == oldArray)
		    series.arrayChanged();
		else
		    series.setArray(newArray);

		showSelectionMarker(graph, e.getX(), e.getY(), userRect);
				
		if ((newArray[0] != newArray[2]) ||
		    (newArray[1] != newArray[3])) {
		    
			m_drawModeCurve = null;
		}
		
	    }
	    	    
	    // Trigger a repaint of the updated contents.
	    m_pageProxy.repaint();
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Handle mouse being released within the page area.  Look at variables that
     * were set during earlier mouse press and drag events.
     * @param e     The mouse event.
     */
    public void mouseReleased(MouseEvent e) {
	
	// Don't proceed if there is no current page.
	if (m_page == null) return;
	synchronized (m_page) {
	    
	    if (m_bDefineMode) {
		// If this is the first click inside this graph, set up
		// new curve for the graph.
		if (m_drawModeCurve != null) {
		    mousePressed(e);
		}
	    }
	}
    }
    
//------------------------------------------------------------------------------
    /**
     * Handle the mouse leaving the drawing area.
     * @param e     The mouse event.
     */
    public void mouseExited(MouseEvent e) {
	
//        System.out.println("Mouse exited at " + e.getX() + ", " + e.getY());
	// Don't proceed if there is no current page.
	if (m_page == null) return;
	synchronized (m_page) {	    
	    m_drawModeCurve = null;
	}    
    }
    
//------------------------------------------------------------------------------
    /**
     * Selects points on the page based on the coordinates and key modifiers.
     * Updates the selection rectangle for the next page redraw.
     * @param xMax	The new x coordinate of the corner of the selection window.
     * @param yMax	The new y coordinate of the corner of the selection window.
     * @param bCtrl	True if the Control key is held down on the key board.
     * @param bShift	True if the Shift key is held down on the key board.
     * @param bAlt	True if the Alt key is held down on the key board.
     */
/*    public void selectWindow(int xMax, int yMax, boolean bCtrl, boolean bShift,
	boolean bAlt) {
	
	// Don't proceed if there is no current page.
	if (m_page == null) return;
	
	// If the mouse hasn't moved since it was pressed, select a 6X6 pixel
	// area centered on the mouse location.
	int mouseX = m_mouseX;
	int mouseY = m_mouseY;
	
	if (xMax == mouseX) {
	    mouseX -= 3;
	    mouseY -= 3;
	    xMax += 3;
	    yMax += 3;
	}
	
	m_page.deferUpdates();
	
	// Update the selection rectangle.
	m_page.setZoomWindow(mouseX, mouseY, xMax, yMax);
	
	// Select the points within the selection region.
	m_page.selectWindow(!bAlt, !bShift, !(bAlt || bCtrl));
	
	m_page.processDeferredUpdates();
    }
*/
        
//------------------------------------------------------------------------------
    /**
     *
     */
    public double calculateDistanceBetween(double x1, double y1, double x2,
        double y2) {
        
        double xDiff = x2 - x1;
        double yDiff = y2 - y1;
        double diff = Math.sqrt((xDiff * xDiff) + (yDiff * yDiff));
        
        return diff;
    }
    
//------------------------------------------------------------------------------
    /**
     * Update the window title to match the current page title.  If the page
     * doesn't have a title, use the suggested title.  The window title is trimmed
     * to 80 charactersand then has " - Gr" appended to it.
     * If there is no current page, use "USGS Gr" as the title.
     * @param suggestedTitle    The title to use if the page doesn't have one.
     */
    public void updateWindowTitle(String suggestedTitle) {
        
        // If there is no current page, use "USGS Gr" as the title.
        if (m_page == null) {
            getParentFrame(m_pageProxy.getComponent()).setTitle(
                "USGS GrBathCalc");
            return;
        }
        
        // If the page doesn't have a title, use the suggested title.
        String pageTitle = m_page.getTitle();
        if (pageTitle == null || pageTitle.equals("")) {
            pageTitle = suggestedTitle;
            m_page.setTitle(pageTitle);
        }
        
        // Trim the page title to 80 characters.
        if (pageTitle.length() > 80)
            pageTitle = pageTitle.substring(0, 80);
        
        // If the page title doesn't match the window title, update the
        // window.
        if (!m_pageTitle.equals(pageTitle)) {
            m_pageTitle = pageTitle;
            getParentFrame(m_pageProxy.getComponent()).
                    setTitle(pageTitle + " - GrBathCalc");
        }
    }
    
//------------------------------------------------------------------------------
    /**
     * Finds the frame that is ultimately the parent of the given component.
     * @param component The component to begin searching upward from.
     * @return  The frame that was eventually found, or a new frame if none.
     */
    public static Frame getParentFrame(Component component) {
	
	// Initialize the frame to null.
	Frame frame = null;
	
	// Return null if the component is null;
	if (component == null) return new Frame();
	
	// Ascend the hierarchy of containers until the parent frame is found.
	Container container = component.getParent();
	while (container != null && !(container instanceof Frame)) {
	    container = container.getParent();
	    if (container instanceof Frame) frame =
		(Frame)container;
	}
	
	// If no frame was found, return a new frame.
	if (frame == null) return new Frame();
	else return frame;
    }
    
//------------------------------------------------------------------------------
    /**
     *
     */
    public void writeGuiOptions() {
	
	String grHome = System.getProperty("gov.usgs.gr.grhome");
	grHome.replace('\\', '/');
	
	XMLProjectFile projFile = new XMLProjectFile();
	projFile.setProperty("file",
	    grHome + "/config/gr_config_GrBathCalc.xml");
	projFile.add(m_options);
	
	// Write the file to disk.
	try {
	    projFile.write();
	} catch (java.io.IOException ioe) {
	    ioe.printStackTrace();
	}
    }
}


