/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2003-08-18
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d.ps;

import gov.usgs.gr.*;
import gov.usgs.gr.visual.*;
import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;

//==============================================================================
/**
 * Contains options for a PsBathymetryGrid.  If this object is a child of
 * a GlBathymetryGrid, it will be transferred to the PostScript object.
 */
public class Grid2PsOptions extends gov.usgs.gr.GrObject {

    /** The output file name. */
    protected int m_levelsOfDetail = 1;
    /** The color to use when drawing partially wet cells. */
    protected float[] m_textColor = {0, 0, 0};
    /** The bleach level to use when text is on. */
    protected float m_textOnBleach = 0.75f;
    /** The bleach level to use when text is off. */
    protected float m_textOffBleach = 0;
    /** The thickness of the line in pixels. */
    protected float m_gridLineWidth = 0.2f;
    /** The color of this curve line in an RGB array. */
    protected float[] m_gridColor = {0, 0, 0};
    /**
     * Predefined colors stored as RGB triples in float arrays.
     * The order is green, magenta, cyan, yellow, white, red.
     */
    protected float[][] m_colors = {{0, 1, 0}, {1, 0, 1}, {0, 1, 1}, {1, 1, 0},
        {1, 1, 1}, {1, 0, 0}};
    /** Predefined line patterns to use. */
    protected short[] m_linePatterns = {Gr.SOLID_LINE_PATTERN,
        Gr.DASHED_LINE_PATTERN, Gr.DOTTED_LINE_PATTERN,
        Gr.DASH_DOT_LINE_PATTERN, Gr.LOOSE_DASHED_LINE_PATTERN,
        Gr.LOOSE_DOTTED_LINE_PATTERN, Gr.DASH_DOT_DASH_LINE_PATTERN,
        Gr.TIGHT_DASHED_LINE_PATTERN, Gr.TIGHT_DOTTED_LINE_PATTERN};
    /** The dash pattern of the line. */
    protected short m_gridLinePattern = Gr.SOLID_LINE_PATTERN;
    /** Whether to show the grid. */
    protected boolean m_bShowGrid = false;
    /** Whether to draw the grid over the wet points. */
    protected boolean m_bDrawGridOverWetPoints = false;
    /** The tree icon to use for this class. */
	public static ArrayList TREE_ICON_HOLDER = new ArrayList(1);
    /** The location of the tree icon. */
    protected String m_iconFile = "GrObject/PsOptions.gif";

//------------------------------------------------------------------------------
/**
 * Construct an empty object.
 */
    public Grid2PsOptions () {

        setDefaults("gov.usgs.gr.visual.ps.PsOptions");
    }

//------------------------------------------------------------------------------
/**
 * Create a string representation of this object.
 * @return  The string that describes this object.
 */
    public String toString() {

    	String s = "Grid2Ps Options";

        return s;
 	}

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

    	GrHashList props = super.getProperties();

        props.put("levelsOfDetail", new GrProperty(m_levelsOfDetail,
            "Level of Detail"));
        props.put("textOnBleach", new GrProperty(
            m_textOnBleach, "Bleach with Text On"));
        props.put("textOffBleach", new GrProperty(
            m_textOffBleach, "Bleach with Text Off"));

        ColorProperty cp = new ColorProperty();
        cp.description = "Text Color";
        cp.setColor(m_textColor);
        props.put("textColor", cp);

        cp = new ColorProperty();
        cp.description = "Grid Color";
        cp.setColor(m_gridColor);
        props.put("gridColor", cp);

        props.put("gridLineWidth", new GrProperty(
            m_gridLineWidth, "Grid Line Width"));

        // Give the user some line type choices...
        GrProperty p = new GrProperty("", "Line Pattern");
        javax.swing.JComboBox jcb = new javax.swing.JComboBox();
        String[] patterns = {
            "Solid", "Dashed", "Dotted", "Dash Dot", "Loose Dashed",
            "Loose Dotted", "Dash Dot Dash", "Tight Dashed", "Tight Dotted"
            };

        // Fill the line pattern combo box.
        for (int i = 0; i < patterns.length; i++) {
            String name = patterns[i];
            jcb.addItem(name);
            if (m_linePatterns[i] == m_gridLinePattern) {
                jcb.setSelectedIndex(i);
                p.value = patterns[i];
            }
        }

        p.component = jcb;
		props.put("gridLinePattern", p);

        props.put("showGrid", new GrProperty(m_bShowGrid, "Show Grid"));
        props.put("drawGridOverWetCells", new GrProperty(
            m_bDrawGridOverWetPoints, "Draw Grid Over Wet Cells"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

		super.setProperties(props);

    	GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
        	String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }
                if (key.equals("levelsOfDetail"))
                    m_levelsOfDetail = (Integer)p.value;
                if (key.equals("textOnBleach"))
                    m_textOnBleach = (Float)p.value;
                if (key.equals("textOffBleach"))
                    m_textOffBleach = (Float)p.value;
                if (key.equals("textColor")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_textColor = cp.getColorArray();
                }
                if (key.equals("gridLineWidth"))
                    m_gridLineWidth = (Float)p.value;
                if (key.equals("gridColor")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_gridColor = cp.getColorArray();
                }
                if (key.equals("showGrid"))
                    m_bShowGrid = (Boolean)p.value;
                if (key.equals("drawGridOverWetCells"))
                    m_bDrawGridOverWetPoints = (Boolean)p.value;
            }
            if (obj instanceof String) {
            	String value = (String)obj;

                setProperty(key, value);
            }
        }
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperty(String key, String value) {

    	super.setProperty(key, value);

        if (key.equals("levelsOfDetail"))
            m_levelsOfDetail = Integer.parseInt(value);
        if (key.equals("textOnBleach"))
            m_textOnBleach = Float.parseFloat(value);
        if (key.equals("textOffBleach"))
            m_textOffBleach = Float.parseFloat(value);
        if (key.equals("textColor")) {
            m_textColor = ColorProperty.getColorArray(value);
        }
        if (key.equals("gridColor")) {
            m_gridColor = ColorProperty.getColorArray(value);
        }
        if (key.equals("gridLineWidth"))
            m_gridLineWidth = Float.parseFloat(value);
        if (key.equals("gridLinePattern")) {
            String[] patterns = {
                "solid", "dashed", "dotted", "dash dot",
                "loose dashed", "loose dotted", "dash dot dash",
                "tight dashed", "tight dotted"
            };

            // Check each name.
            for (int i = 0; i < patterns.length; i++) {
                String name = value.trim().toLowerCase();
                if (name.equals(patterns[i])) {
                    m_gridLinePattern = m_linePatterns[i];
                    break;
                }
            }
        }
        if (key.equals("showGrid"))
            m_bShowGrid = Boolean.parseBoolean(value);
        if (key.equals("drawGridOverWetCells"))
            m_bDrawGridOverWetPoints = Boolean.parseBoolean(value);
	}

//------------------------------------------------------------------------------
/**
 * Accessor method.
 * @return An ImageIcon to use when representing this object in a tree.
 */
    public javax.swing.ImageIcon getTreeIcon() {

        if (TREE_ICON_HOLDER.size() <= 0)
            setTreeIcon();

    	return (javax.swing.ImageIcon)TREE_ICON_HOLDER.get(0);
    }

//------------------------------------------------------------------------------
/**
 * Loads the tree icon from a file and stores it in the TREE_ICON_HOLDER.
 */
    protected javax.swing.ImageIcon setTreeIcon() {

        javax.swing.ImageIcon icon = loadTreeIcon(m_iconFile);

        if (icon == null)
            icon = super.setTreeIcon();

        TREE_ICON_HOLDER.clear();
        TREE_ICON_HOLDER.add(icon);

        return icon;
    }
}




