/**
 * @(#)             Gr graphing application
 * @author          John Donovan
 *                  U.S. Geological Survey, Water Resources Division
 * @version         2006-06-06
 * Date started:    2003-08-12
 *
 * Copyright (C) 1998-2008 John M. Donovan
 * See the file Gr.java in the package gov.usgs.gr for licensing and
 * warranty information.  Also see license.txt in the package gov.usgs.
 */

package gov.usgs.sfhydro.si3d.ps;

import gov.usgs.gr.*;
import gov.usgs.gr.data.*;
import gov.usgs.gr.visual.*;
import gov.usgs.gr.visual.ps.*;
import gov.usgs.sfhydro.si3d.*;
import gov.usgs.sfhydro.si3d.gl.*;
import java.awt.Rectangle;
import java.awt.geom.*;
import java.io.*;
import java.text.*;
import java.util.*;

//==============================================================================
/**
 * A series of discontinuous triangle fan surfaces.
 */
public class PsBathymetryGrid extends gov.usgs.gr.visual.GraphicalElement
    implements PsElement, PsDrawingObject {

    /** The writer for the output file. */
    protected PrintWriter ps;
    /** The lowest value on the x axis. */
    protected double m_minx = 0;
    /** The highest value on the x axis. */
    protected double m_maxx = 1;
    /** The lowest value on the y axis. */
    protected double m_miny = 0;
    /** The highest value on the y axis. */
    protected double m_maxy = 1;

    /**
     * Describes the size of the graph area after it has been transformed
     * to a regular shape.
     */
    protected Rectangle2D.Double m_local = new Rectangle2D.Double(0, 0, 1, 1);

    protected Bathymetry m_bath = null;
    protected RawBathymetry m_rawBath = null;
    protected Grid2PsOptions m_grid2psOptions = null;
    protected double m_xSpacing = 0;
    protected double m_ySpacing = 0;
    protected double m_zSpacing = 0;
    protected DoubleDataSeries m_bathSeries = null;
    protected DoubleDataSeries m_rawBathSeries = null;
    protected int[] m_drawIndexes = null;
    protected boolean m_bDrawIndexesValid = false;
    protected int m_numWetPoints = 0;
    protected int m_numI = 0;
    protected int m_numJ = 0;
    protected int m_numK = 0;
    protected double m_zScale = 1;
    protected double m_xOffset = 0;
    protected double m_yOffset = 0;
    protected boolean m_bDrawVerticalCenterLines = true;
    protected boolean m_bDrawCells = false;
    protected boolean m_bDrawLayers = false;
    protected int m_iViewMin = 0;
    protected int m_iViewMax = 0;
    protected int m_jViewMin = 0;
    protected int m_jViewMax = 0;
    protected int m_drawWetSquaresThreshold = 10000;
    protected int m_drawGridThreshold = 150;
    protected int m_drawDepthsThreshold = 90;
    protected int m_drawUvwThreshold = 50;
    protected double m_highestWaterLevel = 0;
    protected double m_lowestLandLevel = 0;
    protected double m_deepestColorLevel = 0;
    protected double m_csCoefficient = 0;
    protected boolean m_bDrawVelLocations = false;
    protected ColorScheme m_colorScheme = new InterpolatingColorScheme();
    protected double[] m_orphanCorners = null;
    protected double[] m_wetDiagonals = null;
    /** The color to use when drawing partially wet cells. */
    protected float[] m_partiallyWetColor = {0, 0.3f, 0};
    /** Whether to draw partially wet cells. */
    protected boolean m_bDrawPartiallyWet = true;
    /** Whether to draw markers for depths that aren't on a fully wet cell. */
    protected boolean m_bDrawOrphanCorners = false;
    /** Whether to draw markers for two wet diagonal neighbors. */
    protected boolean m_bDrawWetDiagonals = false;
    /** The color to use when drawing land cells. */
    protected float[] m_landColor = {0.9f, 0.85f, 0.7f};
    /** The color to use when drawing dry cells. */
    protected float[] m_dryCellColor = {0.4f, 0.37f, 0.1f};
    /** The color to use when drawing thin wall barriers. */
    protected float[] m_barrierColor = {1, 1, 0};
    /** Whether to draw partially wet cells. */
    protected boolean m_bDrawLand = true;
    /** Whether to draw partially the surface of each cell. */
    protected boolean m_bDrawSurface = true;
    /** The color of the lines. */
    protected float[] m_color = {0, 1, 0};
    /** Whether to draw depth text. */
    protected boolean m_bDrawDepthText = false;

//------------------------------------------------------------------------------
/**
 * Constructor for creating a blank data series.
 */
    public PsBathymetryGrid(){

        setDefaults("gov.usgs.gr.visual.ps.PsCurve");
    }

//------------------------------------------------------------------------------
/**
 * Constructor for cloning this curve from an existing one.
 */
    public PsBathymetryGrid(GlBathymetryGrid glBathGrid) {

        // Copy the other curve's attributes.
        Enumeration enumeration = glBathGrid.children();
        while (enumeration.hasMoreElements()) {
            GrObject child = (GrObject)enumeration.nextElement();
            add(child);
        }

        setProperties(glBathGrid.getProperties());
    }
//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public GrHashList getProperties() {

        GrHashList props = super.getProperties();

        props.put("zScale", new GrProperty(m_zScale, "Z Scale"));
        props.put("drawVerticalCenterLines", new GrProperty(
            m_bDrawVerticalCenterLines, "Draw Vertical Center Lines"));
        props.put("drawCells", new GrProperty(
            m_bDrawCells, "Draw Cell Outlines"));
        props.put("drawLayers", new GrProperty(
            m_bDrawLayers, "Draw Layers"));
        props.put("drawSurface", new GrProperty(
            m_bDrawSurface, "Draw Surface"));

        ColorProperty cp = new ColorProperty();
        cp.description = "Color";
        cp.setColor(m_color);
        props.put("color", cp);

        props.put("drawOrphanCorners", new GrProperty(
            m_bDrawOrphanCorners, "Mark Orphan Corners"));
        props.put("drawWetDiagonals", new GrProperty(
            m_bDrawWetDiagonals, "Mark 2 Wet Diagonals"));
        props.put("drawPartiallyWet", new GrProperty(
            m_bDrawPartiallyWet, "Draw Partially Wet Cells"));

        cp = new ColorProperty();
        cp.description = "Partially Wet Color";
        cp.setColor(m_partiallyWetColor);
        props.put("partiallyWetColor", cp);

        props.put("drawLand", new GrProperty(m_bDrawLand, "Draw Land"));

        cp = new ColorProperty();
        cp.description = "Land Color";
        cp.setColor(m_landColor);
        props.put("landColor", cp);

        cp = new ColorProperty();
        cp.description = "Dry Cell Color";
        cp.setColor(m_dryCellColor);
        props.put("dryCellColor", cp);

        cp = new ColorProperty();
        cp.description = "Barrier Color";
        cp.setColor(m_barrierColor);
        props.put("barrierColor", cp);

//        props.put("drawSoftEdge", new GrProperty(
//        	m_bDrawSoftEdge, "Draw Soft Edge"));
        props.put("drawWetSquaresThreshold", new GrProperty(
            m_drawWetSquaresThreshold,
            "Draw Wet Squares Threshold"));
        props.put("drawGridThreshold", new GrProperty(
            m_drawGridThreshold, "Draw Grid Threshold"));
        props.put("drawDepthsThreshold", new GrProperty(
            m_drawDepthsThreshold, "Draw Depths Threshold"));
        props.put("drawUvwThreshold", new GrProperty(
            m_drawUvwThreshold, "Draw Uvw Threshold"));
        props.put("highestWater", new GrProperty(
            m_highestWaterLevel, "Highest Water Level"));
        props.put("lowestLand", new GrProperty(
            m_lowestLandLevel, "Lowest Land Level"));
        props.put("deepestLevel", new GrProperty(
            m_deepestColorLevel, "Deepest Color Level"));
        props.put("csCoefficient", new GrProperty(
            m_csCoefficient, "Color Scheme Coefficient (-1 to 1)"));

        return props;
    }

//------------------------------------------------------------------------------
/**
 * Provide the list of properties with the latest values.
 */
    public void setProperties(HashMap props) {

        super.setProperties(props);

        GrProperty p;

        Iterator iter = props.keySet().iterator();
        while (iter.hasNext()) {
            String key = (String)iter.next();
            Object obj = props.get(key);
            if (obj instanceof GrProperty) {
                p = (GrProperty)obj;

                if (p.value instanceof String) {
                    setProperty(key, (String)p.value);
                    continue;
                }

                if (key.equals("zScale"))
                    m_zScale = (Double)p.value;
                if (key.equals("drawVerticalCenterLines"))
                    m_bDrawVerticalCenterLines = (Boolean)p.value;
                if (key.equals("drawCells"))
                    m_bDrawCells = (Boolean)p.value;
                if (key.equals("drawLayers"))
                    m_bDrawLayers = (Boolean)p.value;
                if (key.equals("drawSurface"))
                    m_bDrawSurface = (Boolean)p.value;
                if (key.equals("color")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_color = cp.getColorArray();
                }
                if (key.equals("drawOrphanCorners"))
                    m_bDrawOrphanCorners = (Boolean)p.value;
                if (key.equals("drawWetDiagonals"))
                    m_bDrawWetDiagonals = (Boolean)p.value;
                if (key.equals("drawPartiallyWet"))
                    m_bDrawPartiallyWet = (Boolean)p.value;
                if (key.equals("partiallyWetColor")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_partiallyWetColor = cp.getColorArray();
                }
                if (key.equals("drawLand"))
                    m_bDrawLand = (Boolean)p.value;
                if (key.equals("landColor")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_landColor = cp.getColorArray();
                }
                if (key.equals("dryCellColor")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_dryCellColor = cp.getColorArray();
                }
                if (key.equals("barrierColor")) {
                    ColorProperty cp = (ColorProperty)p;
                    m_barrierColor = cp.getColorArray();
                }
                if (key.equals("drawWetSquaresThreshold"))
                    m_drawWetSquaresThreshold = (Integer)p.value;
                if (key.equals("drawGridThreshold"))
                    m_drawGridThreshold = (Integer)p.value;
                if (key.equals("drawDepthsThreshold"))
                    m_drawDepthsThreshold = (Integer)p.value;
                if (key.equals("drawUvwThreshold"))
                    m_drawUvwThreshold = (Integer)p.value;
                if (key.equals("highestWater"))
                    m_highestWaterLevel = (Double)p.value;
                if (key.equals("lowestLand"))
                    m_lowestLandLevel = (Double)p.value;
                if (key.equals("deepestLevel"))
                    m_deepestColorLevel = (Double)p.value;
                if (key.equals("csCoefficient")) {
                    m_csCoefficient = (Double)p.value;
                    if (m_csCoefficient < -1)
                        m_csCoefficient = -1;
                    if (m_csCoefficient > 1)
                        m_csCoefficient = 1;
                }
            }
        }

        m_bNeedsRedraw = true;
        m_bHasChanged = true;
        notifyObservers();
    }

//------------------------------------------------------------------------------
/**
 * Set a single property.
 */
    public void setProperty(String key, String value) {

        super.setProperty(key, value);

        if (key.equals("zScale"))
            m_zScale = Double.parseDouble(value);
        if (key.equals("drawVerticalCenterLines"))
            m_bDrawVerticalCenterLines = Boolean.parseBoolean(value);
        if (key.equals("drawCells"))
            m_bDrawCells = Boolean.parseBoolean(value);
        if (key.equals("drawLayers"))
            m_bDrawLayers = Boolean.parseBoolean(value);
        if (key.equals("drawSurface"))
            m_bDrawSurface = Boolean.parseBoolean(value);
        if (key.equals("color")) {
            m_color = ColorProperty.getColorArray(value);
        }
        if (key.equals("drawOrphanCorners"))
            m_bDrawOrphanCorners = Boolean.parseBoolean(value);
        if (key.equals("drawWetDiagonals"))
            m_bDrawWetDiagonals = Boolean.parseBoolean(value);
        if (key.equals("drawPartiallyWet"))
            m_bDrawPartiallyWet = Boolean.parseBoolean(value);
        if (key.equals("partiallyWetColor")) {
            m_partiallyWetColor = ColorProperty.getColorArray(value);
        }
        if (key.equals("drawLand"))
            m_bDrawLand = Boolean.parseBoolean(value);
        if (key.equals("landColor"))
            m_landColor = ColorProperty.getColorArray(value);
        if (key.equals("dryCellColor"))
            m_dryCellColor = ColorProperty.getColorArray(value);
        if (key.equals("barrierColor"))
            m_barrierColor = ColorProperty.getColorArray(value);
        if (key.equals("drawWetSquaresThreshold"))
            m_drawWetSquaresThreshold = Integer.parseInt(value);
        if (key.equals("drawGridThreshold"))
            m_drawGridThreshold = Integer.parseInt(value);
        if (key.equals("drawDepthsThreshold"))
            m_drawDepthsThreshold = Integer.parseInt(value);
        if (key.equals("drawUvwThreshold"))
            m_drawUvwThreshold = Integer.parseInt(value);
        if (key.equals("highestWater"))
            m_highestWaterLevel = Double.parseDouble(value);
        if (key.equals("lowestLand"))
            m_lowestLandLevel = Double.parseDouble(value);
        if (key.equals("deepestLevel"))
            m_deepestColorLevel = Double.parseDouble(value);
        if (key.equals("csCoefficient")) {
            m_csCoefficient = Double.parseDouble(value);
            if (m_csCoefficient < -1)
                m_csCoefficient = -1;
            if (m_csCoefficient > 1)
                m_csCoefficient = 1;
        }

        m_bNeedsRedraw = true;
    }

//------------------------------------------------------------------------------
/**
 * Add an object to the curve.  Only accepts DoubleDataSeries.  Only holds one at
 * a time.
 */
    public boolean add(GrObject gro) {

        // If gro is an object effigy, get the eventual subject and try
        // adding that.
        while (gro instanceof gov.usgs.gr.ObjectEffigy)
            gro = ((ObjectEffigy)gro).getSubject();

        if (m_children.contains(gro)) return false;

        if (gro instanceof Bathymetry) {
            if (m_bath != null)
                remove(m_bath);
            setBathymetry((Bathymetry)gro);

            return super.add(gro);
        }

        if (gro instanceof RawBathymetry) {
            if (m_rawBath != null)
                remove(m_rawBath);
            setRawBathymetry((RawBathymetry)gro);

            return super.add(gro);
        }
        if (gro instanceof ColorScheme) {
            if (m_colorScheme != null)
                remove((GrObject)m_colorScheme);
            super.add(gro);
            m_colorScheme = (ColorScheme)gro;

            return true;
        }
        if (gro instanceof Grid2PsOptions) {
            if (m_grid2psOptions != null)
                remove((GrObject)m_grid2psOptions);
            super.add(gro);
            m_grid2psOptions = (Grid2PsOptions)gro;

            return true;
        }

        return false;
    }

//------------------------------------------------------------------------------
/**
 * Sets the Bathymetry object.
 */
    public void setBathymetry(Bathymetry bath) {

        m_bath = bath;
        GrHashList bathProps = m_bath.getProperties();
        GrProperty p = (GrProperty)bathProps.get("xSpacing");
        m_xSpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("ySpacing");
        m_ySpacing = (Double)p.value;
        p = (GrProperty)bathProps.get("zSpacing");
        m_zSpacing = (Double)p.value;
        m_bathSeries = m_bath.getBathymetrySeries();
        m_numI = (int)Math.round(m_bathSeries.getMax(0)) + 1;
        m_numJ = (int)Math.round(m_bathSeries.getMax(1)) + 1;
        m_numK = (int)Math.round(m_bathSeries.getMax(2));
    }

//------------------------------------------------------------------------------
/**
 * Sets the Bathymetry object.
 */
    public void setRawBathymetry(RawBathymetry rawBath) {

        m_rawBath = rawBath;
        GrHashList props = m_rawBath.getProperties();
        GrProperty p = (GrProperty)props.get("xSpacing");
        m_xSpacing = (Double)p.value;
        p = (GrProperty)props.get("ySpacing");
        m_ySpacing = (Double)p.value;
        p = (GrProperty)props.get("zSpacing");
        m_zSpacing = (Double)p.value;
        p = (GrProperty)props.get("numI");
        m_numI = (Integer)p.value;
        p = (GrProperty)props.get("numJ");
        m_numJ = (Integer)p.value;
        m_rawBathSeries = m_rawBath.getBathymetrySeries();

        m_highestWaterLevel = -m_zSpacing;
        m_deepestColorLevel = m_rawBathSeries.getMax(0);

        m_bDrawIndexesValid = false;

        checkGrid();
    }

//------------------------------------------------------------------------------
/**
 * Sets the Bathymetry object.
 */
    public void initializeColorScheme() {

        InterpolatingColorScheme cs = new InterpolatingColorScheme();
        cs.setProperty("colorValue0", "(139,106,52) 0.0");
        cs.setProperty("colorValue1", "(0,204,153) 1.5");
        cs.setProperty("colorValue2", "(0,210,210) 4.0");
        cs.setProperty("colorValue3", "(0,153,153) 10.0");
        cs.setProperty("colorValue4", "(100,150,255) 20.0");
        cs.setProperty("colorValue5", "(70,70,255) 30.0");
        cs.setProperty("colorValue6", "(135,105,143) 50.0");
        cs.setProperty("colorValue7", "(146,60,111) 80.0");
        add(cs);
    }

//------------------------------------------------------------------------------
/**
 * Sets the PrintWriter to use for output.
 * @param pw    The PrintWriter object.
 */
    public void setPsFile(PrintWriter pw) {
        ps = pw;
    }

//------------------------------------------------------------------------------
/**
 * Issue the PostScript commands to draw this series.
 * <p><b> Pre Conditions </b><br>
 *      The origin has been moved to (0, 0), scaling has occurred, and
 *      line type is solid.
 */
    public void drawPS(){

        if (!m_bShow) return;

        // Before drawing anything, check if the curve is supposed to be shown.
        if (!m_bShow || (m_bath == null && m_rawBath == null
            && m_colorScheme == null)) {

            m_bNeedsRedraw = false;

            return;
        }

        boolean bShowGrid = false;
        if (m_grid2psOptions != null) {
            GrHashList grid2psProps = m_grid2psOptions.getProperties();
            GrProperty p = (GrProperty)grid2psProps.get("showGrid");
            bShowGrid = (Boolean)p.value;
        }

        Graph graph = null;
        List observers = getObservers();
        Iterator obIter = observers.iterator();
        while (obIter.hasNext()) {
            // Look for a page parent of the graph.
            GrObject gro = (GrObject)obIter.next();
            if (gro instanceof Graph) {
                graph = (Graph)gro;
            }
        }

        if (graph != null) {
            double xMin = graph.getXAxis().getZoom().getViewMin();
            double xMax = graph.getXAxis().getZoom().getViewMax();
            double yMin = graph.getYAxis().getZoom().getViewMin();
            double yMax = graph.getYAxis().getZoom().getViewMax();

            m_iViewMin = (int)Math.round(xMin / m_xSpacing) - 1;
            m_iViewMax = (int)Math.round(xMax / m_xSpacing) + 2;
            m_jViewMin = (int)Math.round(yMin / m_ySpacing) - 1;
            m_jViewMax = (int)Math.round(yMax / m_ySpacing) + 2;

            if (bShowGrid) {
                m_iViewMin += 5;
                m_iViewMax -= 5;
                m_jViewMin += 5;
                m_jViewMax -= 5;
            }

            if (m_iViewMin < 0) m_iViewMin = 0;
            if (m_iViewMax > m_numI) m_iViewMax = m_numI;
            if (m_iViewMax < m_iViewMin) m_iViewMax = m_iViewMin;
            if (m_jViewMin < 0) m_jViewMin = 0;
            if (m_jViewMax > m_numJ) m_jViewMax = m_numJ;
            if (m_jViewMax < m_jViewMin) m_jViewMax = m_jViewMin;
        }

        drawRaw();

    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void drawRaw(){

        if (m_rawBathSeries == null)
            return;

        // Draw grid lines.
        int iSpan = m_iViewMax - m_iViewMin;
        int jSpan = m_jViewMax - m_jViewMin;
        if (iSpan < m_drawGridThreshold && jSpan < m_drawGridThreshold &&
            iSpan > 0 && jSpan > 0) {

            if (iSpan < m_drawUvwThreshold && jSpan < m_drawUvwThreshold) {
                m_bDrawVelLocations = true;
                m_bDrawDepthText = true;
                drawDepthSquares();
                drawDepthNumbers();
            }
            else {
                m_bDrawVelLocations = false;
                if (iSpan < m_drawDepthsThreshold &&
                    jSpan < m_drawDepthsThreshold) {

                    m_bDrawDepthText = true;
                    drawDepthSquares();
                    drawDepthNumbers();
                }
                else {
                    m_bDrawDepthText = false;
                    drawDepthSquares();
                }
            }
        }
        else {
            // Draw color squares representing depth.
            m_bDrawDepthText = false;
               drawDepthSquares();
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void drawGrid(){

        float[] gridColor = {0, 0, 0};
        float gridLineWidth = 0.2f;
        String gridLinePattern = "dashed";
        if (m_grid2psOptions != null) {
            GrHashList grid2PsProps = m_grid2psOptions.getProperties();
            GrProperty p = (GrProperty)grid2PsProps.get("gridColor");
            ColorProperty cp = (ColorProperty)p;
            gridColor = cp.getColorArray();
            p = (GrProperty)grid2PsProps.get("gridLineWidth");
            gridLineWidth = (Float)p.value;

            String[] patterns = {
                "solid", "dashed", "dotted", "dash dot",
                "loose dashed", "loose dotted", "dash dot dash",
                "tight dashed", "tight dotted"
            };

            p = (GrProperty)grid2PsProps.get("gridLinePattern");
            gridLinePattern = ((String)p.value).toLowerCase();
        }

        double xScale = (double)m_local.width / (m_maxx - m_minx);
        double yScale = (double)m_local.height / (m_maxy - m_miny);

        NumberFormat form = new DecimalFormat("0.##");

        ps.println("\n% Draw grid lines");

        ps.println("gsave");

        String font = "Times-Bold-Italic";
        double size = 12;
        if (psOptions != null) {
            font = psOptions.getPsFont();
            size = psOptions.getPsFontSize();
        }
        size = size * (xScale * m_xSpacing * 0.05);
        ps.println("/" + font + " findfont " + form.format(size)
            + " scalefont setfont");

        ps.println(gridColor[0] + " " + gridColor[1] + " "
            + gridColor[2] + " setrgbcolor");
        ps.println(form.format(gridLineWidth) + " thick setlinewidth");

        // Turn on a regular, dotted line pattern.
        ps.println("\n" + "% Draw grid lines\n" + "loosedot");
        if (gridLinePattern.equals("solid")) {
                ps.println("solid");
        }
        if (gridLinePattern.equals("dashed")) {
            ps.println("dash");
        }
        if (gridLinePattern.equals("dotted")) {
            ps.println("dot");
        }
        if (gridLinePattern.equals("loose dotted")) {
            ps.println("loosedot");
        }
        if (gridLinePattern.equals("dash dot dash")) {
            ps.println("chaindash");
        }
        if (gridLinePattern.equals("dash dot")) {
            ps.println("chaindot");
        }
        if (gridLinePattern.equals("loose dashed")) {
            ps.println("vardash");
        }

        ps.println(" " + (-m_minx * xScale) + " " + (-m_miny * yScale)
            + " translate");
        // Vertical grid lines...
        ps.println("" + (((m_iViewMin + m_iViewMax - 1) * m_xSpacing) * xScale)
            + " "
            + (((m_jViewMin) * m_ySpacing) * yScale) + " translate 90 rotate");
        for (int i = m_iViewMin; i < m_iViewMax; i++) {
            String gridString = "" + (m_iViewMax - i + m_iViewMin);
            double pageX = yScale * (-1.3 * m_ySpacing);
            double pageY = xScale * (((double)i - 0.3) * m_xSpacing);
            ps.println("(" + gridString + ") " + form.format(pageX) + " "
                + form.format(pageY) + " rightshow");
            pageX = yScale *
                (((double)(m_jViewMax - m_jViewMin) + 0.3) * m_ySpacing);
            ps.println("(" + gridString + ") " + form.format(pageX) + " "
                + form.format(pageY) + " leftshow");

            ps.print("np ");
            pageX = xScale * -m_ySpacing;
            pageY = xScale * ((double)i * m_xSpacing);
            ps.print(form.format(pageX) + " " + form.format(pageY)
                + " mv ");
            pageX = yScale *
                (((double)(m_jViewMax - m_jViewMin)) * m_ySpacing);
            ps.print(form.format(pageX) + " " + form.format(pageY)
                + " l ");
            ps.println("st");
        }
        ps.println("-90 rotate " + (((m_iViewMin + m_iViewMax - 1) * m_xSpacing)
            * -xScale)
            + " " + (((m_jViewMin) * m_ySpacing) * -yScale) + " translate");

        // Horizontal grid lines...
        for (int j = m_jViewMin; j < m_jViewMax; j++) {
            String gridString = "" + (j + 1);
            double pageX = xScale *
                (((double)(m_iViewMin) - 1.3) * m_xSpacing);
            double pageY = yScale * (((double)j - 0.3) * m_ySpacing);
            ps.println("(" + gridString + ") " + form.format(pageX) + " "
                + form.format(pageY) + " rightshow");
            pageX = xScale *
                (((double)(m_iViewMax) + 0.3) * m_xSpacing);
            ps.println("(" + gridString + ") " + form.format(pageX) + " "
                + form.format(pageY) + " leftshow");

            ps.print("np ");
            pageX = xScale *
                (((double)(m_iViewMin) - 1) * m_xSpacing);
            pageY = yScale * ((double)j * m_ySpacing);
            ps.print(form.format(pageX) + " " + form.format(pageY)
                + " mv ");
            pageX = xScale *
                (((double)(m_iViewMax)) * m_xSpacing);
            ps.print(form.format(pageX) + " " + form.format(pageY)
                + " l ");
            ps.println("st");
        }
        ps.println(" " + (m_minx * xScale) + " " + (m_miny * yScale)
            + " translate");
        ps.println("solid");

        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void drawDepthSquares() {

        boolean bShowGrid = false;
        boolean bDrawGridOverWetCells = false;
        float[] gridColor = {0, 0, 0};
        float[] textColor = {0, 0, 0};
        float textOnBleach = 0.75f;
        float textOffBleach = 0;
        if (m_grid2psOptions != null) {
            GrHashList grid2psProps = m_grid2psOptions.getProperties();
            GrProperty p = (GrProperty)grid2psProps.get("showGrid");
            bShowGrid = (Boolean)p.value;
            p = (GrProperty)grid2psProps.get("drawGridOverWetCells");
            bDrawGridOverWetCells = (Boolean)p.value;
            p = (GrProperty)grid2psProps.get("gridColor");
            ColorProperty cp = (ColorProperty)p;
            gridColor = cp.getColorArray();
            p = (GrProperty)grid2psProps.get("textColor");
            cp = (ColorProperty)p;
            textColor = cp.getColorArray();
            p = (GrProperty)grid2psProps.get("textOnBleach");
            textOnBleach = (Float)p.value;
            p = (GrProperty)grid2psProps.get("textOffBleach");
            textOffBleach = (Float)p.value;
        }

        double[] array = m_rawBathSeries.getArray();
        if (m_drawIndexes == null ||
            (m_drawIndexes.length != (array.length * 2))) {

            m_drawIndexes = new int[array.length * 2];
            m_bDrawIndexesValid = false;
        }

        double minZ = m_rawBathSeries.getMin(0);
        double maxZ = m_rawBathSeries.getMax(0);
        double spanZ = maxZ - minZ;

        NumberFormat form = new DecimalFormat("0.##");

        ps.println("\n% Draw depth squares");

        ps.println("gsave");

        double xScale = (double)m_local.width / (m_maxx - m_minx);
        double yScale = (double)m_local.height / (m_maxy - m_miny);

        // Draw a background rectangle so the grid and depth text will
        // show up regardless of color.
        if (bShowGrid) {
            float[] bgColor = {0, 0, 0};
            bgColor[0] = gridColor[0];
            bgColor[1] = gridColor[1];
            bgColor[2] = gridColor[2];
            bgColor[0] = 1 - bgColor[0];
            bgColor[1] = 1 - bgColor[1];
            bgColor[2] = 1 - bgColor[2];

            ps.println(bgColor[0] + " " + bgColor[1] + " "
                + bgColor[2] + " setrgbcolor");

            ps.print("np ");
            double pageX = xScale *
                (((double)(m_iViewMin) - 3) * m_xSpacing - m_minx);
            double pageY = yScale *
                (((double)(m_jViewMin) - 3) * m_ySpacing - m_miny);
            ps.print(form.format(pageX) + " " + form.format(pageY) + " mv ");
            pageX = xScale *
                (((double)(m_iViewMax) + 2) * m_xSpacing - m_minx);
            ps.print(form.format(pageX) + " " + form.format(pageY) + " l ");
            pageY = yScale *
                (((double)(m_jViewMax) + 2) * m_ySpacing - m_miny);
            ps.print(form.format(pageX) + " " + form.format(pageY) + " l ");
            pageX = xScale *
                (((double)(m_iViewMin) - 3) * m_xSpacing - m_minx);
            ps.print(form.format(pageX) + " " + form.format(pageY) + " l ");
            ps.println("fill");

            if (m_bDrawDepthText) {
                bgColor[0] = gridColor[0];
                bgColor[1] = gridColor[1];
                bgColor[2] = gridColor[2];
                bgColor[0] = (bgColor[0] + textColor[0]) / 2;
                bgColor[1] = (bgColor[1] + textColor[1]) / 2;
                bgColor[2] = (bgColor[2] + textColor[2]) / 2;
                bgColor[0] = 1 - bgColor[0];
                bgColor[1] = 1 - bgColor[1];
                bgColor[2] = 1 - bgColor[2];

                ps.println(bgColor[0] + " " + bgColor[1] + " "
                    + bgColor[2] + " setrgbcolor");

                ps.print("np ");
                pageX = xScale *
                    (((double)(m_iViewMin) - 1) * m_xSpacing - m_minx);
                pageY = yScale *
                    (((double)(m_jViewMin) - 1) * m_ySpacing - m_miny);
                ps.print(form.format(pageX) + " " + form.format(pageY) + " mv ");
                pageX = xScale *
                    (((double)(m_iViewMax)) * m_xSpacing - m_minx);
                ps.print(form.format(pageX) + " " + form.format(pageY) + " l ");
                pageY = yScale *
                    (((double)(m_jViewMax)) * m_ySpacing - m_miny);
                ps.print(form.format(pageX) + " " + form.format(pageY) + " l ");
                pageX = xScale *
                    (((double)(m_iViewMin) - 1) * m_xSpacing - m_minx);
                ps.print(form.format(pageX) + " " + form.format(pageY) + " l ");
                ps.println("fill");
            }
        }

        // Draw the land as a big rectangle under the wet cells.
        if (m_bDrawLand && m_zScale == 0) {
            ps.println(m_landColor[0] + " " + m_landColor[1] + " "
                + m_landColor[2] + " setrgbcolor");
            ps.print("np ");
            double pageX = xScale *
                (((double)(m_iViewMin)) * m_xSpacing - m_minx);
            double pageY = yScale *
                (((double)(m_jViewMin)) * m_ySpacing - m_miny);
            ps.print(form.format(pageX) + " " + form.format(pageY) + " mv ");
            pageX = xScale *
                (((double)(m_iViewMax) - 1) * m_xSpacing - m_minx);
            ps.print(form.format(pageX) + " " + form.format(pageY) + " l ");
            pageY = yScale *
                (((double)(m_jViewMax) - 1) * m_ySpacing - m_miny);
            ps.print(form.format(pageX) + " " + form.format(pageY) + " l ");
            pageX = xScale *
                (((double)(m_iViewMin)) * m_xSpacing - m_minx);
            ps.print(form.format(pageX) + " " + form.format(pageY) + " l ");
            ps.println("fill");
        }

        // translate (m_xOffset, m_yOffset)
        int iSpan = m_iViewMax - m_iViewMin;
        int jSpan = m_jViewMax - m_jViewMin;

        if (bShowGrid && !bDrawGridOverWetCells)
            drawGrid();

        // Check if the user has specified a value for the number of divisions
        // within each square.
        int numDivs = 1;
        if (m_grid2psOptions != null) {
            GrHashList grid2psProps = m_grid2psOptions.getProperties();
            GrProperty p = (GrProperty)grid2psProps.get("levelsOfDetail");
            numDivs = (Integer)p.value;
        }

        // Set the bleach for the depth color scheme.
        if (m_bDrawDepthText)
            ((GrObject)m_colorScheme).setProperty("bleach", "" + textOnBleach);
        else
            ((GrObject)m_colorScheme).setProperty("bleach", "" + textOnBleach);

        for (int i = m_iViewMin; i < (m_iViewMax - 1); i++) {
            for (int j = m_jViewMin; j < (m_jViewMax - 1); j++) {
                double x = xScale * (i * m_xSpacing - m_minx);
                double y = yScale * (j * m_ySpacing - m_miny);
                double width = xScale * m_xSpacing;
                double height = yScale * m_ySpacing;

                drawDepthSquare(i, j, array, x, y, width, height, numDivs);

                int numWritten = (i * (m_iViewMax - m_iViewMin) + j)
                    * numDivs * numDivs;
                if (numWritten % 1000 == 0) {
                    ps.flush();
                }
            }
        }

        ((GrObject)m_colorScheme).setProperty("bleach", "0");

        // Draw dry cells.
        IntDataSeries dryCellSeries = m_rawBath.getDryCellSeries();
        if (dryCellSeries != null) {
            int[] dryCellArray = dryCellSeries.getArray();
            int numDryCells = dryCellArray.length;
            ps.println(m_dryCellColor[0] + " " + m_dryCellColor[1] + " "
                + m_dryCellColor[2] + " setrgbcolor");
            for (int n = 0; n < numDryCells; n++) {
                double i = dryCellArray[n] / 10000;
                double j = dryCellArray[n] - (i * 10000);

                if (i < m_iViewMin || i >= m_iViewMax ||
                    j < m_jViewMin || j >= m_jViewMax)

                    continue;

                double x = xScale * (i * m_xSpacing - m_minx);
                double y = yScale * (j * m_ySpacing - m_miny);
                double width = xScale * m_xSpacing;
                double height = yScale * m_ySpacing;

                ps.print(form.format(x) + " " +
                    form.format(y) + " mv ");
                ps.print(form.format(width) + " 0 rl ");
                ps.print("0 " + form.format(height) + " rl ");
                ps.print(form.format(-width) + " 0 rl ");
                ps.println("fill");
            }
        }

        drawWetDiagonals();

        // Draw thin-wall barriers.
        drawBarriers();

        if (bShowGrid && bDrawGridOverWetCells)
            drawGrid();

        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void drawDepthSquare(int i, int j, double[] array, double x, double y,
        double width, double height, int numDivs) {

        double blZ = array[j * m_numI + i];
        double brZ = array[j * m_numI + i + 1];
        double trZ = array[(j + 1) * m_numI + i + 1];
        double tlZ = array[(j + 1) * m_numI + i];

        int numWetCorners = 0;
        if (blZ >= m_highestWaterLevel) numWetCorners++;
        if (brZ >= m_highestWaterLevel) numWetCorners++;
        if (trZ >= m_highestWaterLevel) numWetCorners++;
        if (tlZ >= m_highestWaterLevel) numWetCorners++;

        if (numWetCorners == 0)
            return;

        // Draw cell as partially wet.
        if (numWetCorners > 0 && numWetCorners < 4) {
            NumberFormat form = new DecimalFormat("0.###");
            ps.print(form.format(x) + " " + form.format(y) + " ");
            ps.print(form.format(width) + " " + form.format(height) + " ");
            ps.print(form.format(m_partiallyWetColor[0]) + " " +
                form.format(m_partiallyWetColor[1]) + " "
                + form.format(m_partiallyWetColor[2]));
            ps.println(" rcfl");
            return;
        }

        if (numDivs < 1 || numDivs > 32000)
        numDivs = 1;
        double xInc = width / (double)numDivs;
        double yInc = height / (double)numDivs;
        double[] cornerDepths = new double[4];

        // Draw the cell as an array of colored squares of size
        // numDivs X numDivs.
        for (int innerI = 0; innerI < numDivs; innerI++) {
            for (int innerJ = 0; innerJ < numDivs; innerJ++) {
                // Determine the depths of the four corners of this
                // inner cell by interpolating.
                for (int index = 0; index < 4; index++) {
                    double xPos, yPos;
                    if (index == 0 || index == 3)
                        xPos = (double)innerI / (double)numDivs;
                    else
                        xPos = (double)(innerI + 1) / (double)numDivs;
                    if (index == 0 || index == 1)
                        yPos = (double)innerJ / (double)numDivs;
                    else
                        yPos = (double)(innerJ + 1) / (double)numDivs;
                    double bottomZ = blZ * (1 - xPos) + brZ * xPos;
                    double topZ = tlZ * (1 - xPos) + trZ * xPos;
                    cornerDepths[index] =
                        bottomZ * (1 - yPos) + topZ * yPos;
                }

                // Use the interpolated depths to arrive at an average
                // depth for this inner square.  Look up the color for
                // the average depth;
                double avgZ = (cornerDepths[0] + cornerDepths[1]
                    + cornerDepths[2] + cornerDepths[3]) / 4;
                float[] colorArray = m_colorScheme.getColor(avgZ);
                NumberFormat form = new DecimalFormat("0.###");
                ps.print(form.format(x + ((double)innerI * xInc)) + " " +
                    form.format(y + ((double)innerJ * yInc)) + " ");
                ps.print(form.format(xInc) + " " + form.format(yInc) + " ");
                ps.print(form.format(colorArray[0]) + " " +
                    form.format(colorArray[1]) + " "
                    + form.format(colorArray[2]));
                ps.println(" rcfl");
            }
        }

        return;
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void drawBarriers() {

        IntDataSeries barrierSeries = m_rawBath.getBarrierSeries();
        if (barrierSeries != null) {
            int[] barrierArray = barrierSeries.getArray();
            int numBarriers = barrierArray.length / 2;

            double xScale = (double)m_local.width / (m_maxx - m_minx);
            double yScale = (double)m_local.height / (m_maxy - m_miny);

            // Save the calling function's graphics state.
            ps.println("gsave");
            ps.println(m_barrierColor[0] + " " + m_barrierColor[1] + " "
                + m_barrierColor[2] + " setrgbcolor");
            NumberFormat form = new DecimalFormat("0.##");
            ps.println(form.format(1) + " thick setlinewidth");
            for (int n = 0; n < numBarriers; n++) {
                int i1 = barrierArray[n * 2] / 10000;
                int j1 = barrierArray[n * 2 ] - (i1 * 10000);
                int i2 = barrierArray[n * 2 + 1] / 10000;
                int j2 = barrierArray[n * 2 + 1] - (i2 * 10000);

                if (i1 < m_iViewMin || i1 >= m_iViewMax ||
                    i2 < m_iViewMin || i2 >= m_iViewMax ||
                    j1 < m_jViewMin || j1 >= m_jViewMax ||
                    j2 < m_jViewMin || j2 >= m_jViewMax)

                    continue;

                double x1 = (i1 == i2) ? ((double)i1) - 0.15 : i1;
                double x2 = (i1 == i2) ? ((double)i2) + 0.15 : i2;
                double y1 = (j1 == j2) ? ((double)j1) - 0.15 : j1;
                double y2 = (j1 == j2) ? ((double)j2) + 0.15 : j2;

                // Slightly lengthen the line so it overlaps a little.
                if (i1 > i2) {
                    x1 += 0.01;
                    x2 -= 0.01;
                }
                if (i2 > i1) {
                    x1 -= 0.01;
                    x2 += 0.01;
                }
                if (j1 > j2) {
                    y1 += 0.01;
                    y2 -= 0.01;
                }
                if (j2 > j1) {
                    y1 -= 0.01;
                    y2 += 0.01;
                }

                ps.print("np ");
                double pageX1 = xScale * ((double)x1 * m_xSpacing - m_minx);
                double pageY1 = yScale * ((double)y1 * m_ySpacing - m_miny);
                double pageX2 = xScale * ((double)x2 * m_xSpacing - m_minx);
                double pageY2 = yScale * ((double)y2 * m_ySpacing - m_miny);
                ps.print(form.format(pageX1) + " " + form.format(pageY1)
                    + " mv ");
                ps.print(form.format(pageX2) + " " + form.format(pageY1)
                    + " l ");
                ps.print(form.format(pageX2) + " " + form.format(pageY2)
                    + " l ");
                ps.print(form.format(pageX1) + " " + form.format(pageY2)
                    + " l ");
                ps.println("fill");
            }
            ps.println("grestore");
        }
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void drawDepthNumbers() {

        double[] array = m_rawBathSeries.getArray();

        double xScale = (double)m_local.width / (m_maxx - m_minx);
        double yScale = (double)m_local.height / (m_maxy - m_miny);

        NumberFormat form = new DecimalFormat("0.0");

        // Save the calling function's graphics state.
        ps.println("gsave");

        float[] textColor = {0, 0, 0};
        if (m_grid2psOptions != null) {
            GrHashList grid2PsProps = m_grid2psOptions.getProperties();
            GrProperty p = (GrProperty)grid2PsProps.get("textColor");
            ColorProperty cp = (ColorProperty)p;
            textColor = cp.getColorArray();
        }
        ps.println(form.format(textColor[0]) + " " + form.format(textColor[1])
            + " " + form.format(textColor[2]) + " setrgbcolor");
        String font = "Times-Bold-Italic";
        double size = 12;
        if (psOptions != null) {
            font = psOptions.getPsFont();
            size = psOptions.getPsFontSize();
        }
        size = size * (xScale * m_xSpacing * 0.05);
        ps.println("/" + font + " findfont " + form.format(size)
            + " scalefont setfont");

        for (int i = m_iViewMin; i < m_iViewMax; i++) {
            for (int j = m_jViewMin; j < m_jViewMax; j++) {
                double z = array[j * m_numI + i];
                int depth10 = (int)Math.round(z * 10);
                if (z < m_highestWaterLevel) continue;

                String depthString = null;
                depthString = "" + depth10;
                double pageX = xScale *
                    (((double)i + 0.2) * m_xSpacing - m_minx);
                double pageY = yScale * ((double)j
                    * m_ySpacing - m_miny) - (size * 0.3);
                ps.println("(" + depthString + ") 40 " + form.format(pageX) + " "
                    + form.format(pageY) + " rotshow");


            }
        }

        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 *
 */
    public void drawWetDiagonals() {

        ps.println("\n" + "% Draw wet diagonals.");
        ps.println("gsave");
        ps.println(m_dryCellColor[0] + " " + m_dryCellColor[1] + " "
            + m_dryCellColor[2] + " setrgbcolor");

        NumberFormat form = new DecimalFormat("0.###");

        double xScale = (double)m_local.width / (m_maxx - m_minx);
        double yScale = (double)m_local.height / (m_maxy - m_miny);

        ps.println(form.format(m_partiallyWetColor[0]) + " " +
            form.format(m_partiallyWetColor[1]) + " "
            + form.format(m_partiallyWetColor[2]));

        for (int n = 0; n < m_wetDiagonals.length; n += 2) {
            double i = m_wetDiagonals[n];
            if ((n + 1) >= m_wetDiagonals.length)
                break;
            double j = m_wetDiagonals[n + 1];

            if (i < m_iViewMin || i > m_iViewMax
                || j < m_jViewMin || j > m_jViewMax) {

                continue;
            }

            double x = xScale * ((i - 0.55) * m_xSpacing - m_minx);
            double y = yScale * (j * m_ySpacing - m_miny);
            ps.print("np " + form.format(x) + " " + form.format(y) + " mv ");
            x = xScale * (i * m_xSpacing - m_minx);
            y = yScale * ((j - 0.55) * m_ySpacing - m_miny);
            ps.print(form.format(x) + " " + form.format(y) + " l ");
            x = xScale * ((i + 0.55) * m_xSpacing - m_minx);
            y = yScale * (j * m_ySpacing - m_miny);
            ps.print(form.format(x) + " " + form.format(y) + " l ");
            x = xScale * (i * m_xSpacing - m_minx);
            y = yScale * ((j + 0.55) * m_ySpacing - m_miny);
            ps.print(form.format(x) + " " + form.format(y) + " l ");
            ps.println(" fill");
        }

        ps.println("grestore");
    }

//------------------------------------------------------------------------------
/**
 * Attribute modification method.
 * @param local Rectangle discribing the size of the graph area in local
 * 				coordinates.
 */
    public void setLocalRect(Rectangle2D.Double local) {
        m_local = local;
    }

//------------------------------------------------------------------------------
/**
 * Adjust the range of the axis, so the curve can be scaled to fit.
 * @param minx  The lower end of the X range.
 * @param maxx  The upper end of the X range.
 * @param miny  The lower end of the Y range.
 * @param maxy  The upper end of the Y range.
 */
    public void setAxisRange(double minx, double maxx, double miny, double maxy) {
        m_minx = minx;
        m_maxx = maxx;
        m_miny = miny;
        m_maxy = maxy;
    }

//------------------------------------------------------------------------------
/**
 * Find the minimum and maximum boundary values for this object as it is shown
 * for the specified dimension.
 * @param dimension The dimension to report on.  X=0, Y=1.
 * @return The minimum (stored as point.x) and maximum (stored as point.y)
 *         values for the graph contents.   null if hidden.
 */
    public Point2D.Double getBounds(int dimension){

        if (m_bath == null && m_rawBath == null) return null;

        if (dimension == 0) {
            Point2D.Double p = new Point2D.Double(m_xOffset,
                m_xOffset + ((double)(m_numI - 1) * m_xSpacing));
            return p;
        }
        if (dimension == 1) {
            Point2D.Double p = new Point2D.Double(m_yOffset,
                m_yOffset + ((double)(m_numJ - 1) * m_ySpacing));
            return p;
        }

        return null;
    }

//------------------------------------------------------------------------------
/**
 * Overrides default update() in GrObject.  Called by an observed GrObject's
 * notifyObservers() method to handle changes to the observed object.
 * May use the notifyObservers() method to propagate the event handling up to
 * this object's observers.
 * @param gro   The observed object that changed.
 * @param arg   Additional information that the updated object may choose to
 *              send.
 */
    public void update(GrObject gro, Object arg) {

        if (gro == m_rawBath) {
            setRawBathymetry(m_rawBath);
        }
    }

//------------------------------------------------------------------------------
/**
 * Looks for problems with the grid.
 */
    public void checkGrid() {

        ArrayList orphanCornerList = new ArrayList();
        ArrayList wetDiagonalList = new ArrayList();
        double[] array = m_rawBathSeries.getArray();
        IntDataSeries dryCellSeries = m_rawBath.getDryCellSeries();
        boolean[] bWet = new boolean[(m_numI - 1) * (m_numJ - 1)];
        RawBathymetryCalculator calc = new RawBathymetryCalculator();

        // Look at each cell in the grid.
        for (int i = 0; i < (m_numI - 1); i++) {
            for (int j = 0; j < (m_numJ - 1); j++) {
                bWet[j * (m_numI - 1) + i] =
                    calc.isWet(i, j, array, m_numI, m_numJ, dryCellSeries);
            }
        }

        for (int i = 0; i < m_numI; i++) {
            for (int j = 0; j < m_numJ; j++) {
                double depth = array[j * m_numI + i];
                if (depth < m_highestWaterLevel)
                    continue;

                boolean bBLwet = false;
                boolean bBRwet = false;
                boolean bTRwet = false;
                boolean bTLwet = false;
                boolean bEdge = false;

                if (i > 0 && j > 0) {
                    if (bWet[(j - 1) * (m_numI - 1) + (i - 1)]) {
                        bBLwet = true;
                    }
                }
                else {
                    bEdge = true;
                }
                if (i < (m_numI - 1) && j > 0) {
                    if (bWet[(j - 1) * (m_numI - 1) + i]) {
                        bBRwet = true;
                    }
                }
                else {
                    bEdge = true;
                }
                if (i < (m_numI - 1) && j < (m_numJ - 1)) {
                    if (bWet[j * (m_numI - 1) + i]) {
                        bTRwet = true;
                    }
                }
                else {
                    bEdge = true;
                }
                if (i > 0 && j < (m_numJ - 1)) {
                    if (bWet[j * (m_numI - 1) + (i - 1)]) {
                        bTLwet = true;
                    }
                }
                else {
                    bEdge = true;
                }

                if (!bBLwet && !bBRwet && !bTRwet && !bTLwet) {
                    orphanCornerList.add(i);
                    orphanCornerList.add(j);
                }
                if (!bEdge) {
                   if ((bBLwet && !bBRwet && bTRwet && !bTLwet)
                        || (!bBLwet && bBRwet && !bTRwet && bTLwet)) {

                        wetDiagonalList.add(i);
                        wetDiagonalList.add(j);
                   }
                }
            }
        }

        m_orphanCorners = new double[orphanCornerList.size()];
        int index = 0;
        Iterator iter = orphanCornerList.iterator();
        while (iter.hasNext()) {
            m_orphanCorners[index] = (double)((Integer)iter.next());
            index++;
        }

        m_wetDiagonals = new double[wetDiagonalList.size()];
        index = 0;
        iter = wetDiagonalList.iterator();
        while (iter.hasNext()) {
            m_wetDiagonals[index] = (double)((Integer)iter.next());
            index++;
        }
    }

//------------------------------------------------------------------------------
    public void regenerateColorScheme() {

        if (m_colorScheme == null)
            add(new InterpolatingColorScheme());
        if (!(m_colorScheme instanceof InterpolatingColorScheme))
            add(new InterpolatingColorScheme());

        // Use the parabola to interpolate the y value.
        double[] coeffs = DoubleDataSeries.fitParabola(
            0, 0, 0.5, m_csCoefficient, 1, 1);
        double exponent = 1;
        if (m_csCoefficient > 0 && m_csCoefficient < 0.99999)
            exponent = 1 / (1 - m_csCoefficient);
        if (m_csCoefficient < 0 && m_csCoefficient > -0.99999)
            exponent = 1 + m_csCoefficient;


        double range = m_deepestColorLevel - m_lowestLandLevel;

        float colors[][] = {{139,106,52}, {0,204,153}, {0,100,100},
            {0,186,210}, {95,37,181}, {154,139,232}, {135,88,113},
            {205,128,163}, {197,92,61}, {241,205,120}};

        InterpolatingColorScheme cs =
            (InterpolatingColorScheme)m_colorScheme;
        cs.setColorAtValue("CLEAR");
        cs.setColorAtValue("(255,255,0)" + m_highestWaterLevel);
//        cs.setColorAtValue("(139,106,52)" + 0);


        // Format the number for output.
        java.text.NumberFormat form = null;
        java.text.NumberFormat form0 = new java.text.DecimalFormat("0");
        java.text.NumberFormat form1 = new java.text.DecimalFormat("0.#");
        java.text.NumberFormat form2 = new java.text.DecimalFormat("0.##");
        java.text.NumberFormat form3 = new java.text.DecimalFormat("0.###");
        java.text.NumberFormat form4 = new java.text.DecimalFormat("0.####");
        java.text.NumberFormat form5 = new java.text.DecimalFormat("0.#####");
        java.text.NumberFormat form6 = new java.text.DecimalFormat("0.######");

        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < 10; i++) {
            double val = Math.pow(0.1 * (double)i, exponent);
            val *= range;
            val += m_lowestLandLevel;

            // Choose the approriate format depending on the size of the
            // number.
            if (val > 1000000)  form = form0;
            else {
                if (val > 100000) form = form0;
                else {
                    if (val > 10000)  form = form1;
                    else {
                        if (val > 1000) form = form2;
                        else {
                            if (val > 100) form = form3;
                            else {
                                if (val > 10) form = form4;
                                else form = form5;
                            }
                        }
                    }
                }
            }

            cs.setColorAtValue("(" + colors[i][0] + "," + colors[i][1] + ","
                + colors[i][2] + ")" + val);
        }
    }
}






